# -*- coding: utf-8 -*-

# Copyright 2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from __future__ import annotations

import math

from interiorcad.backend.workpiece import Shape
from interiorcad.interface.geometry import isNearlyLesser, isNearlyGreater, isNearlyEqual, isNearlyLesserOrEqual
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polylinesegment import PolylineSegment
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.surface import NCMacro2D
from interiorcad.interface.surface.acangles import normalizeAngle


class VerticalClamexEntryVectorFinder:
    """Find data, like the entry vector for the saw axis or the corresponding angle, for a vertical Clamex NC Macro

    For this calculation, the nearest edge must be found, as the saw must come from outside the Workpiece"""

    def __init__(self: VerticalClamexEntryVectorFinder, ncmacro: NCMacro2D, workpieceShape: Shape, maximumDistance: float = math.inf) -> None:
        assert ncmacro.isVertical(), "This class only works with vertical NC Macros used for Clamex"
        assert workpieceShape.isValid(), "Workpiece Shape should not be empty or otherwise invalid"
        assert isNearlyGreater(maximumDistance, 0.0), "Maximum distance must be positive and not zero"

        self.ncmacro: NCMacro2D = ncmacro
        self.workpieceShape: Shape = workpieceShape
        self.maximumDistance: float = maximumDistance

        self.minimalDistantSegment: PolylineSegment = None
        self.isMinimalDistantSegmentParallel: bool | None = None

        self.entryVector: Vector | None = None

    def calculate(self: VerticalClamexEntryVectorFinder) -> None:
        self.findMinimalDistantSegment()
        self.entryVector = self.getCalculateEntryVector()

    def findMinimalDistantSegment(self: VerticalClamexEntryVectorFinder) -> None:
        segments = self.workpieceShape.polyline.getSegments()
        minimalDistance = self.maximumDistance
        self.minimalDistantSegment = None
        self.isMinimalDistantSegmentParallel = None
        for segment in segments:
            if self.isLinearSegment(segment):
                distance = self.calculateDistance(segment)
                maximumDistanceAllowedForSegment = self.getMaximumDistanceAllowedForSegment(segment)
                if maximumDistanceAllowedForSegment is None or isNearlyGreater(distance, maximumDistanceAllowedForSegment):
                    continue

                isSegmentParallel = self.isParallel(segment)

                betterEdge = False

                if self.minimalDistantSegment is None and isNearlyLesserOrEqual(distance, self.maximumDistance):
                    # For the first match, exactly the maximum distance counts too
                    betterEdge = True

                elif not self.isMinimalDistantSegmentParallel and isSegmentParallel and isNearlyLesserOrEqual(distance, self.maximumDistance):
                    # A parallel segment has priority over any other non-parallel segment found, as long the distance is correct (exactly the maximum distance counts too)
                    betterEdge = True

                elif isNearlyLesser(distance, minimalDistance):
                    betterEdge = not self.isMinimalDistantSegmentParallel or (self.isMinimalDistantSegmentParallel and isSegmentParallel)

                if betterEdge:
                    minimalDistance = distance
                    self.minimalDistantSegment = segment
                    self.isMinimalDistantSegmentParallel = isSegmentParallel

    def getMaximumDistanceAllowedForSegment(self: VerticalClamexEntryVectorFinder, segment: PolylineSegment) -> float | None:
        """Calculate the maximal distance allowed counting the angle of the NC Macro compared to the segment

        Keyword: cosinus of angle
        """

        # We avoid to use perp on both to speed up calculation. (It would be more mathematically correct)
        ncMacroDirection = self.getNCMacroDirection()
        segmentDirection = self.getNormalizedSegment(segment)

        factor = abs(ncMacroDirection.dot(segmentDirection))
        if isNearlyGreater(factor, 0.0):
            return self.maximumDistance * factor

    def getEntryVector(self: VerticalClamexEntryVectorFinder) -> Vector | None:
        return self.entryVector

    def getCalculateEntryVector(self: VerticalClamexEntryVectorFinder) -> Vector | None:
        """Vector in which the tool should enter the workpiece

        Note: that is not the plane of the saw, but the vector of the saw's axis."""

        # We give directly a 3D Vector not to have to convert it later. (Our angle functions expect Vector, not Point)

        if self.minimalDistantSegment:
            segmentNormal = self.getNormalizedSegment(self.minimalDistantSegment).perp()

            # We assume that for a Clamex, the NC macro is always inside compared to the edges.
            roughVector = self.ncmacro.position - self.minimalDistantSegment.start.position
            dot = segmentNormal.dot(roughVector)
            if isNearlyEqual(dot, 0.0):
                if not self.workpieceShape.polyline.isClockwise():
                    segmentNormal = - segmentNormal
            elif isNearlyLesser(dot, 0.0):
                segmentNormal = - segmentNormal

            if self.isMinimalDistantSegmentParallel:
                # We know that the NC Macro is parallel to the segment, so use the segment as reference.
                return Vector(segmentNormal.x, segmentNormal.y, 0.0)
            else:
                # We know that the NC Macro is not parallel to the segment, so we have no other choice that to use the orientation of the NC Macro
                ncMacroNormal = self.getNCMacroDirection().perp()
                dot2 = ncMacroNormal.dot(segmentNormal)
                if isNearlyEqual(dot2, 0.0):
                    return None  # A perpendicular NC Macro cannot be handled from the corresponding edge
                elif isNearlyLesser(dot2, 0.0):
                    ncMacroNormal = - ncMacroNormal

                return Vector(ncMacroNormal.x, ncMacroNormal.y)

    def getAngle(self: VerticalClamexEntryVectorFinder) -> float | None:
        if self.entryVector:
            angleOfEntryVector = VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(self.entryVector)
            # isFlipped is irrelevant in this case, as we have searched from where the machining should be made
            return normalizeAngle(90.0 + angleOfEntryVector)

    def getCorrespondingSegment(self: VerticalClamexEntryVectorFinder) -> PolylineSegment | None:
        return self.minimalDistantSegment

    def calculateDistance(self: VerticalClamexEntryVectorFinder, segment: PolylineSegment) -> float:
        ncMacroDirection = self.getNCMacroDirection()
        return max(
            segment.distance(self.ncmacro.position + ncMacroDirection * (- self.ncmacro.deltax / 2.0)),
            segment.distance(self.ncmacro.position + ncMacroDirection * (self.ncmacro.deltax / 2.0))
        )

    def isLinearSegment(self: VerticalClamexEntryVectorFinder, segment: PolylineSegment) -> bool:
        return segment.start.isCornerVertex() and segment.end.isCornerVertex()

    def isParallel(self: VerticalClamexEntryVectorFinder, segment: PolylineSegment) -> bool:
        ncMacroDirection = self.getNCMacroDirection()
        segmentDirection = self.getNormalizedSegment(segment)
        return ncMacroDirection == segmentDirection or ncMacroDirection == - segmentDirection

    def getNormalizedSegment(self: VerticalClamexEntryVectorFinder, segment: PolylineSegment) -> Point:
        return (segment.end.position - segment.start.position).normalized()

    def getNCMacroDirection(self: VerticalClamexEntryVectorFinder) -> Point:
        ncMacroVectorU = self.ncmacro.matrix.u
        return Point(ncMacroVectorU.x, ncMacroVectorU.y)
