# -*- coding: utf-8 -*-

# Copyright 2019-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import operator

from interiorcad.backend.cmoduleadapters.utilities import convertVerticesFromNamedTupleToDict, convertVisibleCornerVerticesFromPositionsToDict, createIdentityMatrixAsNamedTuple, convertVerticesFromPositionsToDict, convertVerticesFromPositionTuplesToDict

from interiorcad.interface.geometry.clockwise import arePositionsClockwise

# TODO IT-250: it is impractical that these radiusCompensation constants have to be re-declared -- NG
RADIUS_COMPENSATION_NONE = 0
RADIUS_COMPENSATION_LEFT = 1
RADIUS_COMPENSATION_RIGHT = 2

try:
    # The module _contourpath exists from interiorcad 2019 F6 on
    import _contourpath

    def __getAdaptedRadiusCompensationOldSchool(vertices, radiusCompensation):
        isClockwise = arePositionsClockwise([vertex.position for vertex in vertices])
        return RADIUS_COMPENSATION_RIGHT if operator.xor(radiusCompensation == RADIUS_COMPENSATION_RIGHT, isClockwise) else RADIUS_COMPENSATION_LEFT

    def __getLineWithDistanceOldSchool(vertices, radiusCompensation, distance):
        line = SimpleLineSegment(vertices[0].position, vertices[1].position)

        if radiusCompensation == RADIUS_COMPENSATION_RIGHT:
            line.moveRight(distance)
        else:
            line.moveLeft(distance)

        return convertVerticesFromPositionTuplesToDict(line.getStartAndEndPoints())

    def getContourPathWithDistanceOldSchool(vertices, radiusCompensation, distance):
        # getContourPathWithDistance used to need a matrix, but the C++ code did not care about it.
        if len(vertices) > 2:
            newVertices = _contourpath.getContourPathWithDistance(
                tuple(vertex.asTuple() for vertex in vertices),  # The result cannot be a generator, as _contourpath.getContourPathWithDistance excepts a PySequence_, which a generator is not.
                createIdentityMatrixAsNamedTuple(),
                __getAdaptedRadiusCompensationOldSchool(vertices, radiusCompensation),
                distance
            )
            return convertVerticesFromNamedTupleToDict(newVertices)
        elif len(vertices) == 2:
            return __getLineWithDistanceOldSchool(vertices, radiusCompensation, distance)
        else:
            assert False, "At least two vertices expected!"

        # In interiorcad 2019 F6 there used to be a PolygonConverter().convertVertices here. Removed in interiorcad 2019 F7, as it is too high-level code for a cmoduleadapter

    # New in interiorcad 2020 F3
    def getOffsetContourPathNewSchool(vertices, radiusCompensation, distance):
        return _contourpath.getOffsetContourPath(
            tuple(vertex.asTuple() for vertex in vertices),  # The result cannot be a generator, as _contourpath.getOffsetContourPath excepts a PySequence_, which a generator is not.
            radiusCompensation,
            distance
        )

    # New in interiorcad 2020 F3
    def getOffsetContourPath(vertices, radiusCompensation, distance):
        if distance == 0 or radiusCompensation == RADIUS_COMPENSATION_NONE:
            return convertVerticesFromPositionsToDict(vertices)
        elif hasattr(_contourpath, 'getOffsetContourPath'):  # New in interiorcad 2020 F3
            return getOffsetContourPathNewSchool(vertices, radiusCompensation, distance)
        elif hasattr(_contourpath, 'getContourPathWithDistance'):
            return getContourPathWithDistanceOldSchool(vertices, radiusCompensation, distance)
        else:
            raise RuntimeError("Cannot find a function for getOffsetContourPath! (Too old version of interiorcad (before 2019 F6)?")  # Do not translate, as an old interirocad has surely no translation for it.

    # compatibility function for old code, up to interiorcad 2020 F2; do NOT use for new code, use getOffsetContourPath instead
    def getContourPathWithDistance(vertices, matrix, radiusCompensation, distance):
        adaptedRadiusCompensation = RADIUS_COMPENSATION_RIGHT if operator.xor(radiusCompensation == RADIUS_COMPENSATION_RIGHT, matrix.isFlipped()) else RADIUS_COMPENSATION_LEFT
        return getOffsetContourPath(vertices, adaptedRadiusCompensation, distance)


except ImportError:

    from interiorcad.interface.geometry.center3points import getCenterOfThreePointsOnArc
    from interiorcad.interface.geometry.cornervertex import CornerVertex
    from interiorcad.interface.geometry.simplelinesegment import SimpleLineSegment

    def __getShiftedPosition(center, position, newRadius):
        v = (position - center).normalized()
        return center + v * newRadius

    def __createCornerVertex(position):
        return {
            'x': position.x,
            'y': position.y,
            'type': 0,
            'radius': 0,
            'visible': True,
        }

    def __createRadiusVertex(position, radius):
        return {
            'x': position.x,
            'y': position.y,
            'type': 4,
            'radius': radius,
            'visible': True,
        }

    def __isSamePosition(vertex, position):
        return vertex['x'] == position.x and vertex['y'] == position.y

    def __getStupidOffsetPolygonalContourPath(vertices, radiusCompensation, distance):
        """This function is meant just for acceptance tests to work and therefore the code shifts the path in an over-simplistic way. -- NG"""

        newPositions = []

        i = 0
        while i < len(vertices):
            assert vertices[i].isCornerVertex() or vertices[i].isRadiusVertex(), "This over-simplistic code supports only corner and radius vertices!"
            j = (i + 1) % len(vertices)
            if vertices[i].visible:
                if vertices[i].isCornerVertex() and vertices[j].isCornerVertex():
                    line = SimpleLineSegment(vertices[i].position, vertices[j].position)
                    if radiusCompensation == RADIUS_COMPENSATION_RIGHT:
                        line.moveRight(distance)
                    else:
                        line.moveLeft(distance)

                    if not newPositions or not __isSamePosition(newPositions[-1], line.getStartPoint()):
                        newPositions.append(__createCornerVertex(line.getStartPoint()))
                    newPositions.append(__createCornerVertex(line.getEndPoint()))
                    i += 1

                elif vertices[i].isCornerVertex() and vertices[j].isRadiusVertex():
                    k = (j + 1) % len(vertices)
                    if vertices[k].isCornerVertex():
                        radius = vertices[j].radius
                        center = getCenterOfThreePointsOnArc(vertices[i].position, vertices[j].position, vertices[k].position, radius)
                        newRadius = (radius - distance) if radiusCompensation == RADIUS_COMPENSATION_RIGHT else (radius + distance)
                        newPosition1 = __getShiftedPosition(center, vertices[i].position, newRadius)
                        if not newPositions or not __isSamePosition(newPositions[-1], newPosition1):
                            newPositions.append(__createCornerVertex(newPosition1))
                        newPositions.append(__createRadiusVertex(__getShiftedPosition(center, vertices[j].position, newRadius), newRadius))
                        newPositions.append(__createCornerVertex(__getShiftedPosition(center, vertices[k].position, newRadius)))
                        i += 2
                    else:
                        assert False, "This over-simplistic code does not support the situation where the radisu vertex is not followed by a corner vertex"

                else:
                    assert False, "This over-simplistic code does not support this situation"
            else:
                i += 1

        # TODO IT-243/IT-276: remove the last vertex if the path is really closed (and make the one before visible) -- NG
        if newPositions:
            newPositions[-1]['visible'] = False

        return newPositions

    def getOffsetContourPath(vertices, radiusCompensation, distance):
        if distance == 0 or radiusCompensation == RADIUS_COMPENSATION_NONE:
            return convertVerticesFromPositionsToDict(vertices)
        else:
            return __getStupidOffsetPolygonalContourPath(vertices, radiusCompensation, distance)
