# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import atexit
import sqlite3
from collections import namedtuple
from os import path

import interiorcad.backend.paths

CoveringsData = namedtuple("CoveringsData", "description supplierid supplier waste unit price markup amountperunit thickness group")


class MockCoverings:
    def __init__(self, _pathname=None):
        self.alreadyOpened = False
        self.databasename = None

    def __del__(self):
        self.closeDB()

    def __enter__(self):
        self.openDB()
        return self

    def __exit__(self, _exitType, _value, _traceback):
        self.closeDB()
        return True

    def openDB(self):
        if not self.alreadyOpened:
            assert self.databasename, "No database defined."
            assert path.isfile(self.databasename), "Database not found."
            self.db = sqlite3.connect(self.databasename)
            self.db.row_factory = lambda cursor, row: self.generateNamedTuple(cursor, row)
            self.alreadyOpened = True
            self.assertDatabaseIsValid()

    def closeDB(self):
        if self.alreadyOpened:
            self.db.close()
            self.alreadyOpened = False

    def __generateSQLResult(self, query):
        self.openDB()
        sqlCursor = self.db.cursor()
        sqlCursor.execute(query)
        for rowTuple in sqlCursor:
            yield rowTuple

    def generateNamedTuple(self, cursor, row):
        return CoveringsData(*row)

    def getCoveringsData(self, materialID):
        # The WHERE clause is to make mockCoverings.getCoveringsData as similar as possible to XGEmbeddedPython.getCoveringsData.
        sqlQuery = "SELECT DISTINCT description, supplier_id, supplier, waste, unit, cost, markup, amount_per_unit, thickness, id FROM coverings WHERE id=" + materialID
        return list(self.__generateSQLResult(sqlQuery))

    def assertDatabaseIsValid(self):
        sqlCursor = self.db.cursor()
        try:
            sqlCursor.execute("SELECT * FROM datatrackertable")
        except sqlite3.OperationalError as e:
            raise Exception("Database is not valid") from e

    def __setattr__(self, name, value):
        if name == "databasename":
            self.closeDB()
        self.__dict__[name] = value


mockCoveringsInstance = MockCoverings()


def setTestDatabase(dbFilename):
    dbFilepath = dbFilename
    if not path.isabs(dbFilename):
        dbFilepath = path.join(interiorcad.backend.paths.mockDocuments, dbFilename)

    assert path.isfile(dbFilepath), "Can't find database at " + dbFilepath

    mockCoveringsInstance.databasename = dbFilepath


def closeDatabaseConnection():
    mockCoveringsInstance.closeDB()


atexit.register(closeDatabaseConnection)  # closeDatabaseConnection gets called, when python shell exits.
