# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import os

from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.surfacedimensioncreator import SurfaceDimensionCreator, SurfaceDimension
from interiorcad.backend.workpiece.reader import WorkpieceReader
from interiorcad.backend.workpiececreationstrategy.creator import WorkpieceCreationStrategy
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.edges import EdgeData, Edges
from interiorcad.interface.geometry.point import Point


def getSurfaceDimensionCreator(workpiece, configuration, strategy):
    reader = WorkpieceReader(workpiece)
    sizingAddition = configuration.cut_size_add
    return SurfaceDimensionCreator(workpieceReader=reader, strategy=strategy, needsMain=True, needsOpposite=False, sizingAddition=sizingAddition)


def getPartDimensions(workpiece, configuration, strategy):
    creator = getSurfaceDimensionCreator(workpiece, configuration, strategy)
    return creator.createCuttingList()


def getFinishedSizePartDimensions(workpiece, configuration, strategy):
    creator = getSurfaceDimensionCreator(workpiece, configuration, strategy)
    return creator.createFinishedSize()


def getFinishedSizePartDimensionsWithAddition(workpiece, configuration, strategy):
    creator = getSurfaceDimensionCreator(workpiece, configuration, strategy)
    return creator.createFinishedSizeWithAddition()


def getSawnSizePartDimensions(workpiece, configuration, strategy):
    creator = getSurfaceDimensionCreator(workpiece, configuration, strategy)
    return creator.createSawnSize()


def getSawnSizePartDimensionsWithAddition(workpiece, configuration, strategy):
    creator = getSurfaceDimensionCreator(workpiece, configuration, strategy)
    return creator.createSawnSizeWithAddition()


def setFakePartDimensions(entry):
    fakePartDimensions = SurfaceDimension()
    fakePartDimensions.size = Point(entry.boardLength, entry.boardWidth)

    entry.finishedSizePartDimensions = fakePartDimensions
    entry.finishedSizePartDimensionsGrainDirection = fakePartDimensions
    entry.finishedSizePartDimensionsWithAddition = fakePartDimensions
    entry.finishedSizePartDimensionsWithAdditionGrainDirection = fakePartDimensions
    entry.sawnSizePartDimensions = fakePartDimensions
    entry.sawnSizePartDimensionsGrainDirection = fakePartDimensions
    entry.sawnSizePartDimensionsWithAddition = fakePartDimensions
    entry.sawnSizePartDimensionsWithAdditionGrainDirection = fakePartDimensions


def setFakeEdges(entry):
    fakeEdges = Edges(rectangular=4 * [EdgeData()])
    fakeEdges.ensureNotNoneRectangularEdges()
    fakeEdges.polygonal = ()
    entry.edges = fakeEdges
    entry.edgesGrainDirection = fakeEdges


# FIXME: Duplicate von interiorcad.cuttinglistexport.base.CuttinglistExportBase -- NG (IT-210)
GRAIN_LENGTHWISE, GRAIN_CROSSWISE, GRAIN_NONE = 0, 1, 2

# Edges Front Back Left Right = 1234. After rotating 90° clockwise it must be Left Right Back Front = 3421
EDGE_CODE_CROSS_GRAIN_TRANSLATION = str.maketrans({'1': '3', '2': '4', '3': '2', '4': '1'})


class WorkpieceEntry:
    MERGE_FOR_MATERIAL_LIST = 0
    MERGE_FOR_PARTS_LIST = 1

    CUMULATE_NONE = 0
    CUMULATE_SAME_MATERIAL_AND_DIMENSION = 1
    CUMULATE_SAME_MATERIAL_AND_DIMENSION_AND_DESCRIPTION = 2

    def __init__(self, workpiece, configuration, useCoveringGrainDirection):
        self.workpiece = copy.deepcopy(workpiece)

        self.workpieceUUID = workpiece.uuid
        self.configuration = configuration
        self.strategy = WorkpieceCreationStrategy.createSizeStrategy(workpiece, configuration)
        self.constructionGroupComponentID = ExportDataCollector.getConstructionGroupComponentID(self.workpieceUUID)

        partDimensions = getPartDimensions(self.workpiece, self.configuration, self.strategy)
        self.boardLength = partDimensions.getLength()
        self.boardWidth = partDimensions.getWidth()

        self.finishedSizePartDimensions = getFinishedSizePartDimensions(self.workpiece, self.configuration, self.strategy)
        self.finishedSizePartDimensionsGrainDirection = getFinishedSizePartDimensions(self.workpiece, self.configuration, self.strategy)

        self.finishedSizePartDimensionsWithAddition = getFinishedSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy)
        self.finishedSizePartDimensionsWithAdditionGrainDirection = getFinishedSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy)

        self.sawnSizePartDimensions = getSawnSizePartDimensions(self.workpiece, self.configuration, self.strategy)
        self.sawnSizePartDimensionsGrainDirection = getSawnSizePartDimensions(self.workpiece, self.configuration, self.strategy)

        self.sawnSizePartDimensionsWithAddition = getSawnSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy)
        self.sawnSizePartDimensionsWithAdditionGrainDirection = getSawnSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy)

        self.thickness = self.workpiece.thickness
        self.boardThickness = self.workpiece.boardThickness

        if useCoveringGrainDirection:
            self.grain = ExportDataCollector.getCoveringGrainDirection(self.uuid)
        else:
            self.grain = ExportDataCollector.getBoardGrainDirection(self.uuid)

        if self.grain == GRAIN_CROSSWISE:
            self.boardLengthGrainDirection = self.boardWidth
            self.boardWidthGrainDirection = self.boardLength
            self.finishedSizePartDimensionsGrainDirection.rotateRight()
            self.finishedSizePartDimensionsWithAdditionGrainDirection.rotateRight()
            self.sawnSizePartDimensionsGrainDirection.rotateRight()
            self.sawnSizePartDimensionsWithAdditionGrainDirection.rotateRight()

        else:  # GRAIN_LENGTHWISE, GRAIN_NONE
            self.boardLengthGrainDirection = self.boardLength
            self.boardWidthGrainDirection = self.boardWidth

        self.itemID = ExportDataCollector.getItemID(self.uuid)

        self.description = ExportDataCollector.getDescription(self.uuid)

        self.ncProgramPathSizing = ExportDataCollector.getNCProgramNameSizing(self.workpieceUUID)
        self.ncProgramPathMain = ExportDataCollector.getNCProgramNameMain(self.workpieceUUID)
        self.ncProgramPathOpposite = ExportDataCollector.getNCProgramNameOpposite(self.workpieceUUID)
        self.ncProgramNameSizing = os.path.basename(self.ncProgramPathSizing)
        self.ncProgramNameMain = os.path.basename(self.ncProgramPathMain)
        self.ncProgramNameOpposite = os.path.basename(self.ncProgramPathOpposite)

        self.edges = ExportDataCollector.getEdges(self.workpieceUUID)
        self.edges.ensureNotNoneRectangularEdges()

        self.edgesGrainDirection = copy.deepcopy(self.edges)
        if self.grain == GRAIN_CROSSWISE:
            self.edgesGrainDirection.polygonal = None  # TODO: correct polygonal edges information -- NG (IT-212)

            self.edgesGrainDirection.front = self.edges.right
            self.edgesGrainDirection.back = self.edges.left
            self.edgesGrainDirection.left = self.edges.front
            self.edgesGrainDirection.right = self.edges.back

            # Edges Front Back Left Right = 1234. After rotating 90° clockwise it must be Left Right Back Front = 3421
            self.edgesGrainDirection.edgecode = self.edgesGrainDirection.edgecode.translate(EDGE_CODE_CROSS_GRAIN_TRANSLATION)

        # Compatibility member variables for old code; do NOT use for new code! Use the function of self.edges directly (interiorcad 2018 F6)
        self.polygonalEdgesID = ', '.join(self.edges.getPolygonalExistingEdgeSupplierIDs())
        self.polygonalEdges = ', '.join(self.edges.getPolygonalExistingEdgeIDs())

        self.count = workpiece.count

    # Compatibility variable for old code; do NOT use for new code!
    @property
    def ncPropramPathMain(self):
        return self.ncProgramPathMain

    # Compatibility variable for old code; do NOT use for new code!
    @property
    def ncPropramNameMain(self):
        return self.ncProgramNameMain

    # Compatibility variable for old code; do NOT use for new code! Use self.boardLength directly
    def getPartLength(self):
        return self.boardLength

    # Compatibility variable for old code; do NOT use for new code! Use self.boardWidth directly
    def getPartWidth(self):
        return self.boardWidth

    def isWorkpiece(self):
        return True

    def isFitting(self):
        return False

    def isCovering(self):
        return False

    def isBoughtInEntry(self):
        return False

    def canBeCumulated(self, entry, **accumulationOptions):
        if accumulationOptions['workpiecesAccumulationType'] == self.CUMULATE_SAME_MATERIAL_AND_DIMENSION:
            cumulate = self.__canBeCumulatedByMaterialAndDimension(entry)
        elif accumulationOptions['workpiecesAccumulationType'] == self.CUMULATE_SAME_MATERIAL_AND_DIMENSION_AND_DESCRIPTION:
            cumulate = self.__canBeCumulatedByMaterialAndDimensionAndDescription(entry)
        else:  # CUMULATE_NONE
            cumulate = False

        if accumulationOptions['workpiecesMergeType'] == self.MERGE_FOR_MATERIAL_LIST:
            pass
        elif accumulationOptions['workpiecesMergeType'] == self.MERGE_FOR_PARTS_LIST:
            cumulate = cumulate and self.__canBeCumulatedDueToExtraCriteriaForPartsLists(entry, **accumulationOptions)

        return cumulate

    def __canBeCumulatedByMaterialAndDimension(self, entry):
        return all((
            self.itemID == entry.itemID,
            isNearlyEqual(self.boardLength, entry.boardLength),
            isNearlyEqual(self.boardWidth, entry.boardWidth),
            self.grain == entry.grain,
        ))

    def __canBeCumulatedByMaterialAndDimensionAndDescription(self, entry):
        return all((
            self.__canBeCumulatedByMaterialAndDimension(entry),
            self.description.strip() == entry.description.strip(),
        ))

    def __canBeCumulatedDueToExtraCriteriaForPartsLists(self, entry, **accumulationOptions):
        def checkDataTrackerString(key):
            export = accumulationOptions['export']
            return export.getDataAsString(key=key, workpiece=self.workpiece) == export.getDataAsString(key=key, workpiece=entry.workpiece)

        return all((
            checkDataTrackerString(key='part_covering1'),
            checkDataTrackerString(key='part_covering2'),

            checkDataTrackerString(key='part_laminate1'),
            checkDataTrackerString(key='part_laminate2'),
            checkDataTrackerString(key='part_laminate3'),

            self.edges == entry.edges,
        ))

    def modifyAfterMerge(self, **accumulationOptions):
        if 'workpiecesMergeType' in accumulationOptions and accumulationOptions['workpiecesMergeType'] == self.MERGE_FOR_MATERIAL_LIST:
            self.workpiece.uuid = ''  # Reset the UUID, as no data can be further loaded from the datatracker, as it will not match all former entries
            self.__resetEdges()
        elif 'workpiecesMergeType' in accumulationOptions and accumulationOptions['workpiecesMergeType'] == self.MERGE_FOR_PARTS_LIST:
            pass  # We do not reset the UUID, as we assume that we still can load further data from the datatracker and it will be valid for all former entries (unlike in the mode MERGE_FOR_MATERIAL_LIST

    def __resetEdges(self):
        setFakeEdges(self)
        self.polygonalEdgesID = ''
        self.polygonalEdges = ''
        self.workpiece.edges = self.edges  # The edges in self.workpiece must also be reset, despite IC-16072

    def __getattr__(self, key):
        return getattr(self.workpiece, key)


class CoveringEntry:
    def __init__(self, workpiece, configuration, covering, coveringThickness):
        self.workpiece = copy.deepcopy(workpiece)

        # Clear the UUID as the covering has no UUID, especially not the one of the workpiece (somewhere there is unfortunate code that uses directly workpiece.uuid)
        self.workpieceUUID = workpiece.uuid
        self.uuid = ''
        self.workpiece.uuid = ''

        self.configuration = configuration
        self.strategy = WorkpieceCreationStrategy.createSizeStrategy(workpiece, configuration)

        partDimensions = getPartDimensions(self.workpiece, self.configuration, self.strategy)
        # TODO: depending on how the coverings are made, either finished size or sawn size ( + 2* addition ) is wanted. We cannot choose directly for now (IT-205)
        self.boardLength = partDimensions.getLength() + 2 * configuration.additionalCoveringAddition
        self.boardWidth = partDimensions.getWidth() + 2 * configuration.additionalCoveringAddition

        self.finishedSizePartDimensions = self.addDoubleCoveringAdditionToDimension(getFinishedSizePartDimensions(self.workpiece, self.configuration, self.strategy))
        self.finishedSizePartDimensionsGrainDirection = self.addDoubleCoveringAdditionToDimension(getFinishedSizePartDimensions(self.workpiece, self.configuration, self.strategy))

        self.finishedSizePartDimensionsWithAddition = self.addDoubleCoveringAdditionToDimension(getFinishedSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy))
        self.finishedSizePartDimensionsWithAdditionGrainDirection = self.addDoubleCoveringAdditionToDimension(getFinishedSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy))

        self.sawnSizePartDimensions = self.addDoubleCoveringAdditionToDimension(getSawnSizePartDimensions(self.workpiece, self.configuration, self.strategy))
        self.sawnSizePartDimensionsGrainDirection = self.addDoubleCoveringAdditionToDimension(getSawnSizePartDimensions(self.workpiece, self.configuration, self.strategy))

        self.sawnSizePartDimensionsWithAddition = self.addDoubleCoveringAdditionToDimension(getSawnSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy))
        self.sawnSizePartDimensionsWithAdditionGrainDirection = self.addDoubleCoveringAdditionToDimension(getSawnSizePartDimensionsWithAddition(self.workpiece, self.configuration, self.strategy))

        self.grain = ExportDataCollector.getCoveringGrainDirection(self.workpieceUUID)

        if self.grain == GRAIN_CROSSWISE:
            self.boardLengthGrainDirection = self.boardWidth
            self.boardWidthGrainDirection = self.boardLength
            self.finishedSizePartDimensionsGrainDirection.rotateRight()
            self.finishedSizePartDimensionsWithAdditionGrainDirection.rotateRight()
            self.sawnSizePartDimensionsGrainDirection.rotateRight()
            self.sawnSizePartDimensionsWithAdditionGrainDirection.rotateRight()

        else:  # GRAIN_LENGTHWISE, GRAIN_NONE
            self.boardLengthGrainDirection = self.boardLength
            self.boardWidthGrainDirection = self.boardWidth

        self.thickness = coveringThickness
        self.boardThickness = coveringThickness

        self.itemID = covering

        self.ncProgramPathSizing = ExportDataCollector.getNCProgramNameSizing(self.workpieceUUID)
        self.ncProgramPathMain = ExportDataCollector.getNCProgramNameMain(self.workpieceUUID)
        self.ncProgramPathOpposite = ExportDataCollector.getNCProgramNameOpposite(self.workpieceUUID)
        self.ncProgramNameSizing = os.path.basename(self.ncProgramPathSizing)
        self.ncProgramNameMain = os.path.basename(self.ncProgramPathMain)
        self.ncProgramNameOpposite = os.path.basename(self.ncProgramPathOpposite)

        self.count = workpiece.count

        setFakeEdges(self)

    def addDoubleCoveringAdditionToDimension(self, dimensions):
        dimensions.size.x += 2 * self.configuration.additionalCoveringAddition
        dimensions.size.y += 2 * self.configuration.additionalCoveringAddition
        return dimensions

    def isWorkpiece(self):
        return False

    def isFitting(self):
        return False

    def isCovering(self):
        return True

    def isBoughtInEntry(self):
        return False

    def canBeCumulated(self, entry, **accumulationOptions):
        return all((
            entry.isCovering(),
            'shouldAccumulateCoverings' in accumulationOptions and accumulationOptions['shouldAccumulateCoverings'],
            self.itemID == entry.itemID,
            isNearlyEqual(self.boardLength, entry.boardLength),
            isNearlyEqual(self.boardWidth, entry.boardWidth),
            self.grain == entry.grain,
        ))

    def modifyAfterMerge(self, **accumulationOptions):
        pass

    def __getattr__(self, key):
        return None


class FittingEntry:
    def __init__(self, uuidOfFitting, fakeWorkpiece, count):
        self.uuid = uuidOfFitting
        self.parentUUID = ''
        self.workpiece = fakeWorkpiece

        self.boardLength = 0
        if ExportDataCollector.isBoxObjectType(self.uuid):
            self.boardLength = ExportDataCollector.getBoxObjectResizeInfo(self.uuid)

        self.boardWidth = 0

        setFakePartDimensions(self)

        self.grain = GRAIN_NONE

        self.boardLengthGrainDirection = self.boardLength
        self.boardWidthGrainDirection = self.boardWidth

        self.thickness = 0
        self.boardThickness = 0

        self.itemID = ExportDataCollector.getItemID(self.uuid)

        self.ncProgramPathSizing = ''
        self.ncProgramPathMain = ''
        self.ncProgramPathOpposite = ''
        self.ncProgramNameSizing = ''
        self.ncProgramNameMain = ''
        self.ncProgramNameOpposite = ''

        self.name = ExportDataCollector.getDescription(self.uuid)
        self.description = self.name

        self.count = count * ExportDataCollector.getBoardCount(self.uuid)

        setFakeEdges(self)

    # Compatibility variable for old code; do NOT use for new code!
    @property
    def ncPropramPathMain(self):
        return self.ncProgramPathMain

    # Compatibility variable for old code; do NOT use for new code!
    @property
    def ncPropramNameMain(self):
        return self.ncProgramNameMain

    def isWorkpiece(self):
        return False

    def isFitting(self):
        return True

    def isCovering(self):
        return False

    def isBoughtInEntry(self):
        return False

    def canBeCumulated(self, entry, **accumulationOptions):
        return all((
            'shouldAccumulateFittings' in accumulationOptions and accumulationOptions['shouldAccumulateFittings'],
            self.itemID == entry.itemID,
        ))

    def modifyAfterMerge(self, **accumulationOptions):
        pass

    def __getattr__(self, key):
        return None


class BoughtInEntry:
    # See enum class WoodenDrawerComponentToBaseImplementation in C++
    IMPLEMENTATION_NONE, IMPLEMENTATION_DADO, IMPLEMENTATION_RABBET, IMPLEMENTATION_STANDALONE, IMPLEMENTATION_FLUSH = range(0, 5)

    def __init__(self, uuidOfBoughtIn, fakeWorkpiece, count):
        self.uuid = uuidOfBoughtIn
        self.parentUUID = fakeWorkpiece.parentUUID
        self.workpiece = fakeWorkpiece

        # TODO IT-236: Length/Width/Height of a drawer are *not* in X/Y/Z like for a Custom Part or Cabinet. What should we do here? Cobert the axis? Leave 0? -- NG
        self.boardLength = 0
        self.boardWidth = 0

        setFakePartDimensions(self)

        self.grain = GRAIN_NONE

        self.boardLengthGrainDirection = self.boardLength
        self.boardWidthGrainDirection = self.boardWidth

        self.thickness = 0
        self.boardThickness = 0

        self.itemID = ExportDataCollector.getItemID(self.uuid)

        self.ncProgramPathSizing = ''
        self.ncProgramPathMain = ''
        self.ncProgramPathOpposite = ''
        self.ncProgramNameSizing = ''
        self.ncProgramNameMain = ''
        self.ncProgramNameOpposite = ''

        self.name = ExportDataCollector.getUniqueName(self.uuid)
        self.description = ExportDataCollector.getDescription(self.uuid)

        self.count = count * ExportDataCollector.getBoardCount(self.uuid)

        setFakeEdges(self)

    # Compatibility variable for old code; do NOT use for new code!
    @property
    def ncPropramPathMain(self):
        return self.ncProgramPathMain

    # Compatibility variable for old code; do NOT use for new code!
    @property
    def ncPropramNameMain(self):
        return self.ncProgramNameMain

    def isWorkpiece(self):
        return False

    def isFitting(self):
        return False

    def isCovering(self):
        return False

    def canBeCumulated(self, _entry, **_accumulationOptions):
        return False

    def isBoughtInEntry(self):
        return True

    def modifyAfterMerge(self, **accumulationOptions):
        pass

    def __getattr__(self, key):
        return None


class PIOContainerEntry:
    def __init__(self, uuid, fakeWorkpiece, count):
        self.uuid = uuid
        self.workpiece = fakeWorkpiece
        self.count = count

    def isWorkpiece(self):
        return False

    def isFitting(self):
        return False

    def isCovering(self):
        return False

    def canBeCumulated(self, _entry, **_accumulationOptions):
        return False

    def isBoughtInEntry(self):
        return False

    def __getattr__(self, key):
        return None
