# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import functools
import json

try:
    import _interiorcaddatatracker
    datatracker = _interiorcaddatatracker

except ImportError:
    import interiorcad.backend.datatracker.mockdatatracker as mockdatatracker
    datatracker = mockdatatracker.mockDatatrackerInstance

    def setTestDatabase(dbFilename):
        global clearCaches
        mockdatatracker.setTestDatabase(dbFilename)
        clearCaches()

    def closeDatabaseConnection():
        mockdatatracker.closeDatabaseConnection()


CACHE_LENGTH = None


def clearCaches():
    for item in globals().values():
        if hasattr(item, 'cache_clear') and callable(item.cache_clear):
            item.cache_clear()


def beginConsuming():
    pass


def endConsuming():
    pass


if hasattr(datatracker, 'getExportUUIDs'):  # From interiorcad 2018 F5 on

    @functools.lru_cache(maxsize=CACHE_LENGTH)
    def getExportUUIDs():
        return datatracker.getExportUUIDs()

else:

    @functools.lru_cache(maxsize=CACHE_LENGTH)
    def getUUIDListForExport():  # Up to interiorcad 2018 F4
        return tuple(uuid for uuid in datatracker.getUUIDListForExport())


if hasattr(datatracker, 'getUnfilteredExportUUIDs'):  # New in interiorcad 2021 F4.4

    @functools.lru_cache(maxsize=CACHE_LENGTH)
    def getUnfilteredExportUUIDs():
        return datatracker.getUnfilteredExportUUIDs()

else:

    @functools.lru_cache(maxsize=CACHE_LENGTH)
    def getUUIDListForExport():  # Up to interiorcad 2018 F4
        return tuple(uuid for uuid in datatracker.getUUIDListForExport())


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getItemAsString(uuid, subid, key):
    assert subid == 0, "SubID is not 0"
    return datatracker.getItemAsString(uuid, subid, key)


# Do NOT cache
def getItemAsStringWithoutCaching(uuid, subid, key):
    return getItemAsString.__wrapped__(uuid, subid, key)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getItemAsLong(uuid, subid, key):
    assert subid == 0, "SubID is not 0"
    return datatracker.getItemAsLong(uuid, subid, key)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getItemAsDouble(uuid, subid, key):
    assert subid == 0, "SubID is not 0"
    return datatracker.getItemAsDouble(uuid, subid, key)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getItemAsBool(uuid, subid, key):
    assert subid == 0, "SubID is not 0"
    return datatracker.getItemAsBool(uuid, subid, key)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getPolygon2D(uuid, subid, key):
    assert subid == 0, "SubID is not 0"

    # From interiorcad 2018 F3 on, where the path is in a JSON text
    jsonString = datatracker.getItemAsString(uuid, 0, key)
    if jsonString:
        raw = json.loads(jsonString)
        return raw.get('vertices', dict())
    elif hasattr(datatracker, 'getItemAsPolygon2D'):  # Up to interiorcad 2018 F2
        legacyKeyPrefix = key.replace('_json', '_')
        return [
            {
                'x': legacyVertex.x,
                'y': legacyVertex.y,
                'type': legacyVertex.type,
                'radius': legacyVertex.radius,
                'visible': legacyVertex.visible,
            }
            for legacyVertex in datatracker.getItemAsPolygon2D(uuid, subid, legacyKeyPrefix)
        ]
    elif hasattr(datatracker, 'getItemAsLegacyPolygon2D'):  # Mock for Tests
        legacyKeyPrefix = key.replace('_json', '_')
        return datatracker.getItemAsLegacyPolygon2D(uuid, subid, legacyKeyPrefix)
    else:
        # interiorcad may have not written any path data if the path is corrupt. (IC-14240)
        return []


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getGridPoints(uuid, subid, key):
    assert subid == 0, "SubID is not 0"
    jsonString = datatracker.getItemAsString(uuid, 0, key)
    if jsonString:
        raw = json.loads(jsonString)
        return raw.get('gridpoints')
    else:
        return []


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getGridPointsCoordinateSystem(uuid, subid, key):
    assert subid == 0, "SubID is not 0"
    jsonString = datatracker.getItemAsString(uuid, 0, key)
    if jsonString:
        return json.loads(jsonString)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getChildUUIDsByPartType(parentUUID, partType):
    return datatracker.getChildUUIDsByPartType(parentUUID, partType)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def getChildFittingUUIDs(parentUUID):
    return datatracker.getChildFittingUUIDs(parentUUID)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def createKeyWithPostFixKey(key, postfixKey):
    return datatracker.createKeyWithPostFixKey(key, postfixKey)


@functools.lru_cache(maxsize=CACHE_LENGTH)
def createKeyWithPostFixKeyAndID(key, postfixKey, xg_id):
    return datatracker.createKeyWithPostFixKeyAndID(key, postfixKey, xg_id)


# This is the (modified) old buggy implementation of ExportDataCollector.getRoutingUUIDs.
def getLegacyRoutingUUIDs(customPartUUID):
    beginConsuming()
    for uuid, count in getExportUUIDs():
        parentUUID = getItemAsString(uuid, 0, 'rout_partuuid')
        if customPartUUID == parentUUID:
            yield uuid
    endConsuming()


def getRoutingUUIDs(customPartUUID):
    if hasattr(datatracker, 'getRoutingUUIDs'):  # From interiorcad 2017 F7 on
        return datatracker.getRoutingUUIDs(customPartUUID)
    else:
        # Until interiorcad 2017 F6, there was no way to ask C++, therefore improvise instead, using the old buggy code.
        return getLegacyRoutingUUIDs(customPartUUID)


# From interiorcad 2018 F3 on
def getMatrix(uuid, subid, key):
    return datatracker.getMatrix(uuid, subid, key)


# Do NOT cache!
def setNCProgramNames(uuid, exportName, sizingName, mainName, oppositeName):
    # Careful: for performance reasons, we do not reset the cache. However it means that the keys set here must be retrieved by getItemAsStringWithoutCaching
    try:
        return datatracker.setNCProgramNames(uuid, exportName, sizingName, mainName, oppositeName)
    except TypeError:  # up to interiorcad 2017, there were only 4 parameters (without export name)
        return datatracker.setNCProgramNames(uuid, sizingName, mainName, oppositeName)
