# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import json

from interiorcad.backend.datatracker import datatracker, datatrackeritemtypes
from interiorcad.backend.datatracker.polygonconverter import PolygonConverter
from interiorcad.interface.geometry.edges import EdgeData, Edges
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.geometry.vector import Vector


class ExportDataCollector:
    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code, use getAllCustomPartUUIDsAndCounts instead
    @staticmethod
    def getCustomPartUUIDs():
        return [
            uuid
            for uuid in ExportDataCollector.getObjectUUIDs(datatrackeritemtypes.ITEMTYPE_CUSTOMPART)
            if not ExportDataCollector.getUserPerceivedParentUUID(uuid)
        ]

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code
    @staticmethod
    def getPIOContainerUUIDs():
        return list(itertools.chain.from_iterable(ExportDataCollector.getObjectUUIDs(containerType) for containerType in datatrackeritemtypes.PIO_CONTAINER_TYPES))

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code
    @staticmethod
    def getCustomPartUUIDsInsidePIOContainerObjects():
        uuids = set()
        for parentUUID in ExportDataCollector.getPIOContainerUUIDs():
            uuids.update(ExportDataCollector.getSubCustomPartsFromPIOContainer(parentUUID))
        return uuids

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code, use getAllCustomPartUUIDsAndCounts instead
    @staticmethod
    def getAllCustomPartUUIDs():
        # sorted was needed to have stable tests -- NG
        return sorted(uuid for (uuid, _count) in ExportDataCollector.getAllCustomPartUUIDsAndCounts())

    # from interiorcad 2018 F5 on
    @staticmethod
    def getAllCustomPartUUIDsAndCounts():
        # TODO: how to have stable tests without needing a "sorted"? -- NG (IT-194/IT-199/IT-205)
        return sorted(ExportDataCollector.generateObjectUUIDsAndCounts(datatrackeritemtypes.ITEMTYPE_CUSTOMPART))

    # New in interiorcad 2021 F4.4
    @staticmethod
    def getUnfiltertedAllCustomPartUUIDsAndCounts():
        return sorted(ExportDataCollector.generateUnfilteredObjectUUIDsAndCounts(datatrackeritemtypes.ITEMTYPE_CUSTOMPART))

    # from interiorcad 2019 F6 on (officially only from interiorcad 2020 F1 on)
    @staticmethod
    def getAllBoughtInsUUIDsAndCounts():
        # TODO: how to have stable tests without needing a "sorted"? -- NG (IT-194/IT-199/IT-205/IT-236)
        return sorted(ExportDataCollector.generateObjectUUIDsAndCounts(datatrackeritemtypes.ITEMTYPE_BOUGHTIN))

    # compatibility function for old code, up to interiorcad 2020 F1 (excluded); do NOT use for new code, use getSplitCustomPartAndPIOContainerUUIDsAndCounts instead
    # from interiorcad 2018 F5 on
    @staticmethod
    def getSplitFreeCustomPartAndPIOContainerUUIDsAndCounts():
        freeCustomParts = []
        pioContainers = []

        for uuid, count in ExportDataCollector.getExportUUIDsAndCounts():
            partType = ExportDataCollector.getPartType(uuid)
            if partType == datatrackeritemtypes.ITEMTYPE_CUSTOMPART:
                if not ExportDataCollector.getUserPerceivedParentUUID(uuid):
                    freeCustomParts.append((uuid, count))
            elif partType in datatrackeritemtypes.PIO_CONTAINER_TYPES:
                pioContainers.append((uuid, count))

        return freeCustomParts, pioContainers

    # from interiorcad 2020 F1 on
    @staticmethod
    def getSplitCustomPartAndPIOContainerUUIDsAndCounts():
        freeCustomParts = []
        childCustomParts = []
        pioContainers = []

        for uuid, count in ExportDataCollector.getExportUUIDsAndCounts():
            partType = ExportDataCollector.getPartType(uuid)
            if partType == datatrackeritemtypes.ITEMTYPE_CUSTOMPART:
                if ExportDataCollector.getUserPerceivedParentUUID(uuid):
                    childCustomParts.append((uuid, count))
                else:
                    freeCustomParts.append((uuid, count))
            elif partType in datatrackeritemtypes.PIO_CONTAINER_TYPES:
                pioContainers.append((uuid, count))

        return freeCustomParts, pioContainers, childCustomParts

    @staticmethod
    def getSubCustomPartsFromPIOContainer(parentUUID):
        datatracker.beginConsuming()
        uuids = datatracker.getChildUUIDsByPartType(parentUUID, datatrackeritemtypes.ITEMTYPE_CUSTOMPART)
        datatracker.endConsuming()
        return uuids

    @staticmethod
    def getRawDataAsString(uuid, key):
        return datatracker.getItemAsString(uuid, 0, key)

    @staticmethod
    def getRawDataAsFloat(uuid, key):
        return datatracker.getItemAsDouble(uuid, 0, key)

    @staticmethod
    def getRawDataAsInt(uuid, key):
        return datatracker.getItemAsLong(uuid, 0, key)

    @staticmethod
    def getUniqueName(uuid):
        if uuid:
            datatracker.beginConsuming()
            partName = datatracker.getItemAsString(uuid, 0, 'part_uniquename')
            if partName == "":
                partName = datatracker.getItemAsString(uuid, 0, 'unit_unitname')
            datatracker.endConsuming()
            return partName
        else:
            return ''

    # New in interiorcad 2020 F8
    @staticmethod
    def getCabinetUniqueName(uuid):
        if uuid:
            return ExportDataCollector.getUniqueName(ExportDataCollector.getUserPerceivedParentUUID(uuid))
        else:
            return ''

    # For working with C++ code up to interiorcad 2017 F4, do NOT use directly
    @staticmethod
    def __generateSimulatedExportUUIDsAndCounts():
        uuids = set()

        datatracker.beginConsuming()

        for uuid in datatracker.getUUIDListForExport():
            partType = ExportDataCollector.getPartType(uuid)
            if partType == datatrackeritemtypes.ITEMTYPE_CUSTOMPART:
                uuids.add(uuid)
            elif partType in datatrackeritemtypes.PIO_CONTAINER_TYPES:
                uuids.add(uuid)
                uuids.update(ExportDataCollector.getSubCustomPartsFromPIOContainer(uuid))

        datatracker.endConsuming()

        return uuids

    # from interiorcad 2018 F5 on
    @staticmethod
    def getExportUUIDsAndCounts():
        if hasattr(datatracker, 'getExportUUIDs'):  # From interiorcad 2018 F5 on
            return datatracker.getExportUUIDs()
        else:
            return [(uuid, 1) for uuid in ExportDataCollector.__generateSimulatedExportUUIDsAndCounts()]

    # New in interiorcad 2021 F4.4
    @staticmethod
    def getUnfilteredExportUUIDsAndCounts():
        if hasattr(datatracker, 'getUnfilteredExportUUIDs'):  # New in interiorcad 2021 F4.4
            return datatracker.getUnfilteredExportUUIDs()
        else:
            return [(uuid, 1) for uuid in ExportDataCollector.__generateSimulatedExportUUIDsAndCounts()]

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code, use generateObjectUUIDsAndCounts instead
    @staticmethod
    def getObjectUUIDs(pioUniversalName):
        return [uuid for (uuid, _count) in ExportDataCollector.generateObjectUUIDsAndCounts(pioUniversalName)]

    # from interiorcad 2018 F5 on
    @staticmethod
    def generateObjectUUIDsAndCounts(pioUniversalName):
        datatracker.beginConsuming()
        for uuid, count in ExportDataCollector.getExportUUIDsAndCounts():
            if pioUniversalName in ExportDataCollector.getPartType(uuid):
                yield uuid, count
        datatracker.endConsuming()

    # New in interiorcad 2021 F4.4
    @staticmethod
    def generateUnfilteredObjectUUIDsAndCounts(pioUniversalName):
        datatracker.beginConsuming()
        for uuid, count in ExportDataCollector.getUnfilteredExportUUIDsAndCounts():
            if pioUniversalName in ExportDataCollector.getPartType(uuid):
                yield uuid, count
        datatracker.endConsuming()

    @staticmethod
    def getItemID(uuid):
        return datatracker.getItemAsString(uuid, 0, 'part_itemid')

    @staticmethod
    def getDescription(uuid):
        return datatracker.getItemAsString(uuid, 0, 'part_description')

    @staticmethod
    def getBoardCount(uuid):
        return datatracker.getItemAsLong(uuid, 0, 'part_boardcount')

    @staticmethod
    def getUnitGroup(uuid):
        return datatracker.getItemAsLong(uuid, 0, 'part_unitgroup')

    @staticmethod
    def getBoardGrainDirection(uuid):
        return datatracker.getItemAsLong(uuid, 0, 'part_boardgraindirection')

    @staticmethod
    def getCoveringGrainDirection(uuid):
        return datatracker.getItemAsLong(uuid, 0, 'part_coveringgraindirection')

    @staticmethod
    def getNCProgramNameSizing(uuid):
        return datatracker.getItemAsStringWithoutCaching(uuid, 0, 'part_ncprogrampathsizing')

    @staticmethod
    def getNCProgramNameMain(uuid):
        return datatracker.getItemAsStringWithoutCaching(uuid, 0, 'part_ncprogrampathfront')

    @staticmethod
    def getNCProgramNameOpposite(uuid):
        return datatracker.getItemAsStringWithoutCaching(uuid, 0, 'part_ncprogrampathback')

    @staticmethod
    def getBoxObjectResizeInfo(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'part_boxobject_resizevalue')

    @staticmethod
    def getConstructionGroupComponentID(uuid):
        return datatracker.getItemAsLong(uuid, 0, 'part_constructiongroupcomponentid')

    # The information is available in the Datatracker only from interiorcad F9 on
    @staticmethod
    def getBoughtInUUIDOfPart(uuid):
        return datatracker.getItemAsString(uuid, 0, 'part_boughtin_uuid')

    # compatibility function for old code; do NOT use for new code
    @staticmethod
    def getSizingProgramName(uuid):
        return ExportDataCollector.getNCProgramNameSizing(uuid=uuid)

    # compatibility function for old code; do NOT use for new code
    @staticmethod
    def getMainProgramName(uuid):
        return ExportDataCollector.getNCProgramNameMain(uuid=uuid)

    # compatibility function for old code; do NOT use for new code
    @staticmethod
    def getOppositeProgramName(uuid):
        return ExportDataCollector.getNCProgramNameOpposite(uuid=uuid)

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code, use generateExportedFittingsUUIDsAndCounts instead
    @staticmethod
    def getFittingUUIDsForExport():
        uniqueFittingUUIDs = set()
        for customPartUUID in ExportDataCollector.getAllCustomPartUUIDs():
            uniqueFittingUUIDs.update(ExportDataCollector.getFittingUUIDsFromCustomPart(customPartUUID))
        return uniqueFittingUUIDs

    @staticmethod
    def getFittingUUIDsFromCustomPart(customPartUUID):
        datatracker.beginConsuming()
        fittingUUIDs = datatracker.getChildFittingUUIDs(customPartUUID)
        datatracker.endConsuming()
        return fittingUUIDs

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code, use generateExportedFittingsUUIDsAndCounts instead
    @staticmethod
    def getAllFittingsFromExportedUUIDs():
        return ExportDataCollector.getFittingUUIDsForExport()

    # from interiorcad 2018 F5 on
    @staticmethod
    def generateExportedFittingsUUIDsAndCounts():
        if hasattr(datatracker, 'getExportUUIDs'):  # From interiorcad 2018 F5 on
            for uuid, count in ExportDataCollector.getExportUUIDsAndCounts():
                if ExportDataCollector.isFitting(uuid):
                    yield uuid, count
        else:
            for uuid in ExportDataCollector.getFittingUUIDsForExport():
                yield uuid, 1

    @staticmethod
    def getUserPerceivedParentUUID(uuid):
        parent = datatracker.getItemAsString(uuid, 0, 'part_userparentuuid')  # From interiorcad 2017 F7 on
        if not parent:
            parent = datatracker.getItemAsString(uuid, 0, 'part_parentuuid')  # Until interiorcad 2017 F6
        return parent

    @staticmethod
    def getPIOContainerUUID(uuid):
        return datatracker.getItemAsString(uuid, 0, 'part_piocontaineruuid')  # Entry only available from interiorcad 2017 F7 on

    @staticmethod
    def getUserParentPIOContainerUUID(uuid):
        return datatracker.getItemAsString(uuid, 0, 'part_userparentpiocontaineruuid')

    # compatibility function for old code; do NOT use for new code, replaced by ExportDataCollector.getUserPerceivedParentUUID
    @staticmethod
    def getParentUUID(uuid):
        return ExportDataCollector.getUserPerceivedParentUUID(uuid)

    @staticmethod
    def getFittingParentUUID(uuid):
        return datatracker.getItemAsString(uuid, 0, 'fitting_parentuuid1')

    @staticmethod
    def isFitting(uuid):
        # We assume that a UUID defines a fitting, if it has at least one fitting parent
        return bool(ExportDataCollector.getFittingParentUUID(uuid))

    @staticmethod
    def getCustomPartThickness(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'unit_unitheight')

    @staticmethod
    def getCustomPartBoardThickness(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'part_depth')

    @staticmethod
    def getCustomPartIsMirrored(uuid):
        return datatracker.getItemAsBool(uuid, 0, 'part_objectismirrored')

    @staticmethod
    def __getSizeBounds(uuid, keys):
        bounds = Rectangle()
        bounds.bottom = datatracker.getItemAsDouble(uuid, 0, keys[0])
        bounds.left = datatracker.getItemAsDouble(uuid, 0, keys[1])
        bounds.right = datatracker.getItemAsDouble(uuid, 0, keys[2])
        bounds.top = datatracker.getItemAsDouble(uuid, 0, keys[3])
        return bounds

    @staticmethod
    def getFinishedSizeBounds(uuid):
        return ExportDataCollector.__getSizeBounds(uuid, ('part_workpieceboundbottom', 'part_workpieceboundleft', 'part_workpieceboundright', 'part_workpieceboundtop'))

    @staticmethod
    def getSawnSizeBounds(uuid):
        return ExportDataCollector.__getSizeBounds(uuid, ('part_workpiecesawnboundbottom', 'part_workpiecesawnboundleft', 'part_workpiecesawnboundright', 'part_workpiecesawnboundtop'))

    @staticmethod
    def getPartType(uuid):
        return datatracker.getItemAsString(uuid, 0, 'part_type')

    @staticmethod
    def getTypedRoutings(customPartUUID):
        routings = ExportDataCollector.getRoutingUUIDs(customPartUUID)
        return tuple((ExportDataCollector.getPartType(uuid), uuid) for uuid in routings)

    @staticmethod
    def isDrillingType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_DRILLING

    @staticmethod
    def isGridDrillingType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_GRIDDRILLING

    @staticmethod
    def isDadoRabbetType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_DADORABBET

    @staticmethod
    def isMiterType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_MITER

    @staticmethod
    def isContourType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_CONTOUR

    @staticmethod
    def isCenterPunchType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_CENTERPUNCH

    @staticmethod
    def isNCMacroType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_NCMACRO

    @staticmethod
    def isChamferType(itemType):
        return itemType == datatrackeritemtypes.ITEMTYPE_CHAMFER

    @staticmethod
    def isBoxObjectType(uuid):
        return ExportDataCollector.getPartType(uuid) == datatrackeritemtypes.ITEMTYPE_BOXOBJECT

    @staticmethod
    def getRoutingUUIDs(customPartUUID):
        return datatracker.getRoutingUUIDs(customPartUUID)

    # Up to interiorcad 2018 F2
    @staticmethod
    def getMachiningMatrixLegacy(uuid):
        datatracker.beginConsuming()
        u = Vector(
            x=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixux'),
            y=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixuy'),
            z=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixuz')
        )
        v = Vector(
            x=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixvx'),
            y=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixvy'),
            z=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixvz')
        )
        w = Vector(
            x=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixwx'),
            y=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixwy'),
            z=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixwz')
        )
        offset = Vector(
            x=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixoffsetx'),
            y=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixoffsety'),
            z=datatracker.getItemAsDouble(uuid, 0, 'rout_matrixoffsetz')
        )
        datatracker.endConsuming()
        return Matrix(u=u, v=v, w=w, offset=offset)

    @staticmethod
    def getMachiningMatrix(uuid):
        try:
            # From interiorcad 2018 F3 on, where the module _interiorcaddatatracker has the method "getMatrix"
            datatracker.beginConsuming()
            matrixTuple = datatracker.getMatrix(uuid, 0, 'rout_matrixblob')
            datatracker.endConsuming()
            return Matrix(
                Vector(matrixTuple.a00, matrixTuple.a01, matrixTuple.a02),
                Vector(matrixTuple.a10, matrixTuple.a11, matrixTuple.a12),
                Vector(matrixTuple.a20, matrixTuple.a21, matrixTuple.a22),
                Vector(matrixTuple.xOffset, matrixTuple.yOffset, matrixTuple.zOffset)
            )

        except (AttributeError, LookupError):
            return ExportDataCollector.getMachiningMatrixLegacy(uuid)

    @staticmethod
    def getDrillingDiameter(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_diameter')

    @staticmethod
    def getMachiningDepth(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_depth')

    @staticmethod
    def getMachiningWidth(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_width')

    @staticmethod
    def getLineMachiningLength(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_length')

    @staticmethod
    def getLineMachiningStartAddition(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_startaddition')

    @staticmethod
    def getLineMachiningEndAddition(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_endaddition')

    @staticmethod
    def getDadoRabbetStartInset(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_insetstart')

    @staticmethod
    def getDadoRabbetEndInset(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_insetend')

    # compatibility function for old code, up to interiorcad 2022 F1.1. Please use getAngle instead
    @staticmethod
    def getMiterAngle(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_angle')

    # New in interiorcad 2022 F2
    @staticmethod
    def getAngle(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_angle')

    @staticmethod
    def getRadiusCompensation(uuid):
        return datatracker.getItemAsLong(uuid, 0, 'rout_radiuscompensation')

    @staticmethod
    def getMacroName(uuid):
        return datatracker.getItemAsString(uuid, 0, 'rout_macroname')

    @staticmethod
    def getDeltaX(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_deltax')

    @staticmethod
    def getDeltaY(uuid):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_deltay')

    @staticmethod
    def getApproach(uuid):
        return datatracker.getItemAsString(uuid, 0, 'rout_approach')

    @staticmethod
    def getDeparture(uuid):
        return datatracker.getItemAsString(uuid, 0, 'rout_departure')

    @staticmethod
    def hasContourMachineOptions(uuid):
        return datatracker.getItemAsBool(uuid, 0, 'rout_hasmachineoptions')

    @staticmethod
    def getContourSelectedMachine(uuid):
        return datatracker.getItemAsString(uuid, 0, 'rout_machine')

    @staticmethod
    def getContourMultiUsageEntries(uuid):
        return datatracker.getItemAsLong(uuid, 0, 'rout_multiusageentries')

    @staticmethod
    def getContourClearOut(uuid, index):
        return datatracker.getItemAsBool(uuid, 0, 'rout_clearout' + str(index))

    @staticmethod
    def getContourCutThrough(uuid, index):
        return datatracker.getItemAsBool(uuid, 0, 'rout_cutthrough' + str(index))

    @staticmethod
    def getContourRouter(uuid, index):
        return datatracker.getItemAsString(uuid, 0, 'rout_router' + str(index))

    @staticmethod
    def getContourDepthPerStep(uuid, index):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_depthperstep' + str(index))

    @staticmethod
    def getContourDistance(uuid, index):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_distance' + str(index))

    @staticmethod
    def getContourFeedRate(uuid, index):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_feedrate' + str(index))

    @staticmethod
    def getContourDepth(uuid, index):
        return datatracker.getItemAsDouble(uuid, 0, 'rout_depth' + str(index))

    @staticmethod
    def getNCMacroCode(uuid):
        code = datatracker.getItemAsString(uuid, 0, 'rout_macrocode')
        # VW's multi-line text field uses CR. Corrected in interiorcad 2020 F6.2 where the CR is replaced by LF. (IC-22118)
        # For older documents, we make the replace here too.
        return code.replace('\r', '\n')

    @staticmethod
    def getNCMacroParametersAsStrings(uuid):
        count = datatracker.getItemAsLong(uuid, 0, 'record_count')
        for i in range(count):
            ### IT-144 Unschoen das wir nur Strings bekommen, nicht die echte Datentypen aus sqlite3. -- NG
            yield datatracker.getItemAsString(uuid, 0, 'record_name' + str(i)), datatracker.getItemAsString(uuid, 0, 'record_value' + str(i))

    # From interiorcad 2018 F3 on
    @staticmethod
    def getPolygon2D(uuid, key):
        vertices = datatracker.getPolygon2D(uuid, 0, key)
        return PolygonConverter().convertPolyline(vertices)

    # From interiorcad 2022 F2.1 on (renamed in that version)
    @staticmethod
    def getGridPoints(uuid):
        return datatracker.getGridPoints(uuid, 0, 'rout_gridpoints')

    # From interiorcad 2022 F2.1 on
    @staticmethod
    def getGridPointsCoordinateSystem(uuid):
        return datatracker.getGridPointsCoordinateSystem(uuid, 0, 'rout_gridpointscoordinatesystem')

    # From interiorcad 2022 F2.1 on
    @staticmethod
    def getDrillingThrough(uuid):
        return datatracker.getItemAsBool(uuid, 0, 'rout_drillthrough')

    @staticmethod
    def getContourPath(uuid):
        return ExportDataCollector.getPolygon2D(uuid, 'polygon2d_json')

    @staticmethod
    def getFinishedSizeSizingContour(customPartUUID):
        return ExportDataCollector.getPolygon2D(customPartUUID, 'polygon2d_json')

    @staticmethod
    def getSawnSizeSizingContour(customPartUUID):
        return ExportDataCollector.getPolygon2D(customPartUUID, 'sawnpolygon2d_json')

    @staticmethod
    def getDocumentName(documentUUID):
        datatracker.beginConsuming()
        documentName = datatracker.getItemAsString(documentUUID, 0, 'part_documentname')
        datatracker.endConsuming()
        return documentName

    # New in Epic ID-427 (interiorcad 2023 F5.0)
    @staticmethod
    def getProjectPosition(uuid):
        datatracker.beginConsuming()
        position = datatracker.getItemAsString(uuid, 0, 'jobd_positionnumber')
        datatracker.endConsuming()
        return position

    @staticmethod
    def __getPolygonalEdges(customPartUUID, shouldSwapEdgeData):
        datatracker.beginConsuming()
        edgesAsJSONString = datatracker.getItemAsString(customPartUUID, 0, 'part_edges_json')
        datatracker.endConsuming()

        if edgesAsJSONString:
            return ExportDataCollector.__getPolygonalEdgesJSON(edgesAsJSONString, shouldSwapEdgeData)
        else:
            return ExportDataCollector.__getPolygonalEdgesLegacy(customPartUUID)

    @staticmethod
    def __getRectangularEdges(customPartUUID, shouldSwapEdgeData):
        datatracker.beginConsuming()
        edgesAsJSONString = datatracker.getItemAsString(customPartUUID, 0, 'part_edges_json')
        datatracker.endConsuming()

        if edgesAsJSONString:
            return ExportDataCollector.__getRectangularEdgesJSON(edgesAsJSONString, shouldSwapEdgeData)
        else:
            return ExportDataCollector.__getRectangularEdgesLegacy(customPartUUID)

    @staticmethod
    def __getRectangularEdgesJSON(edgesAsJSONString, shouldSwapEdgeData):
        edgesAsJSON = json.loads(edgesAsJSONString)

        hasRectangularEdges = 'rectangular' in edgesAsJSON
        if hasRectangularEdges:
            result = list()
            for index in range(1, 5):
                rectangularEdges = edgesAsJSON['rectangular']
                rectangularEdgeIndex = '_'.join(('edge', str(index)))

                currentEdge = EdgeData(
                    supplierID=rectangularEdges[rectangularEdgeIndex]['id'],
                    thickness=rectangularEdges[rectangularEdgeIndex]['thickness'],
                    length=rectangularEdges[rectangularEdgeIndex]['length'],
                    jointStart=rectangularEdges[rectangularEdgeIndex]['startjoint'],
                    jointEnd=rectangularEdges[rectangularEdgeIndex]['endjoint'],
                    isPolygonal=rectangularEdges[rectangularEdgeIndex].get('ispolygonal', True),  # As it is rectangular, we know it is polygonal, even if the information is not in the datatracker
                )

                if shouldSwapEdgeData:
                    currentEdge.jointStart, currentEdge.jointEnd = currentEdge.jointEnd, currentEdge.jointStart

                result.append(currentEdge)
        else:
            result = 4 * [EdgeData()]

        return result

    @staticmethod
    def __getRectangularEdgesLegacy(customPartUUID):
        result = list()
        for index in range(1, 5):
            result.append(EdgeData(
                supplierID=datatracker.getItemAsString(customPartUUID, 0, 'part_edge{index}'.format(index=index)),
                thickness=datatracker.getItemAsDouble(customPartUUID, 0, 'part_edge{index}thickness'.format(index=index)),
                length=datatracker.getItemAsDouble(customPartUUID, 0, 'part_edge{index}length'.format(index=index)),
                jointStart=datatracker.getItemAsString(customPartUUID, 0, 'part_edge{index}startjoint'.format(index=index)),
                jointEnd=datatracker.getItemAsString(customPartUUID, 0, 'part_edge{index}endjoint'.format(index=index)),
                isPolygonal=True,  # We know this is a rectangle, so we can affirm that it is polygonal.
            ))
        return result

    @staticmethod
    def getEdges(customPartUUID):
        # From interiorcad 2020 F9. Before this version, polygon2d_original_direction would return an empty path, so we cannot fix the edge data
        originalPathDirection = datatracker.getItemAsString(customPartUUID, 0, 'polygon2d_original_direction')
        shouldSwapEdgeData = originalPathDirection == 'ccw'

        return Edges(
            polygonal=ExportDataCollector.__getPolygonalEdges(customPartUUID, shouldSwapEdgeData),
            rectangular=ExportDataCollector.__getRectangularEdges(customPartUUID, shouldSwapEdgeData),
            edgecode=ExportDataCollector.__getEdgeCode(customPartUUID)
        )

    @staticmethod
    def __getPolygonalEdgesJSON(edgesAsJSONString, shouldSwapEdgeData):
        result = list()
        edgesAsJSON = json.loads(edgesAsJSONString)
        for edge in edgesAsJSON.get('polygonal'):
            result.append(EdgeData(
                supplierID=edge['id'],
                thickness=edge['thickness'],
                length=edge['length'],
                jointStart=edge['startjoint'],
                jointEnd=edge['endjoint'],
                isPolygonal=edge.get('ispolygonal', False),  # If the data is not in the datatracker we cannot assume that it is polygonal, as we have no clue.
            ))

        if shouldSwapEdgeData:
            result.reverse()

        return result

    @staticmethod
    def __getPolygonalEdgesLegacy(customPartUUID):
        polygonalEdgeCount = datatracker.getItemAsLong(customPartUUID, 0, 'part_edgescount')

        result = list()
        for index in range(polygonalEdgeCount):
            result.append(EdgeData(
                supplierID=datatracker.getItemAsString(customPartUUID, 0, 'part_edgeid' + str(index)),
                thickness=datatracker.getItemAsDouble(customPartUUID, 0, 'part_edgethickness' + str(index)),
                length=datatracker.getItemAsDouble(customPartUUID, 0, 'part_edgelength' + str(index)),
                jointStart=datatracker.getItemAsString(customPartUUID, 0, 'part_edgestartjoint' + str(index)),
                jointEnd=datatracker.getItemAsString(customPartUUID, 0, 'part_edgeendjoint' + str(index)),
                isPolygonal=False,  # We have no data in datatracker to know it. Assume false
            ))
        return result

    @staticmethod
    def __getEdgeCode(customPartUUID):
        datatracker.beginConsuming()
        edgesAsJSONString = datatracker.getItemAsString(customPartUUID, 0, 'part_edges_json')
        datatracker.endConsuming()

        if edgesAsJSONString:
            return ExportDataCollector.__getEdgeCodeJSON(edgesAsJSONString)
        else:
            return ''

    @staticmethod
    def __getEdgeCodeJSON(edgesAsJSONString):
        edgesAsJSON = json.loads(edgesAsJSONString)
        return edgesAsJSON['edgecode']
