# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import getpass
import os
import time

import interiorcad.backend.datatracker.datatracker  # TODO IT-252: why using the datatracker directly and not ExportDataCollector? -- NG
import interiorcad.backend.datatracker.projectpositions
import interiorcad.backend.openservices
import interiorcad.backend.systemdirectories
from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.filenamecleaner import FilenameCleaner
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP, getConstructionGroupName, getConstructionGroupComponentName
from interiorcad.interface.configuration import DATETIME_NOW


class ExportHelper:
    @staticmethod
    def CreatePathIfNotExists(newPath):
        try:
            if os.path.isdir(newPath):
                pass  # os.makedirs would complain if the permissions are not the ones os.makedirs was ordered to create
            elif os.path.exists(newPath):
                errorMessage = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 4, "Cannot create directory {path}, as it exists but is not a directory!")
                raise RuntimeError(errorMessage.format(path=newPath))
            else:
                os.makedirs(newPath, exist_ok=True)
        except OSError:  # throws a WindowsError when a directory was deleted just before os.makedirs is called.
            time.sleep(1)
            os.makedirs(newPath, exist_ok=True)

    @staticmethod
    def createDictionaryForExportPath():
        infos = interiorcad.backend.systemdirectories.getSystemDirectories()
        ExportHelper.updateDictionaryWithDateTime(infos)
        return infos

    @staticmethod
    def openExportPath(exportConfiguration):
        path = ExportHelper.createExportPathWithoutWorkpiece(exportConfiguration)
        interiorcad.backend.openservices.openPath(path)

    @staticmethod
    def sanitizePath(path):
        if os.sep != '/':
            path = path.replace('/', os.sep)
        # Do not use os.path.join, as otherwise an absolute path would not remain absolute.
        return os.path.normpath(os.sep.join(map(str.strip, path.split(os.sep))))

    @staticmethod
    def createExportPathFromConfigurationWithDictionary(exportConfiguration, pathDictionary):
        path = exportConfiguration.getPathForExport()
        try:
            path = path.format_map(pathDictionary)
            return ExportHelper.sanitizePath(path)
        except KeyError as e:
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 49, "Placeholder {placeholder} of subfolder of export path is wrong! Accepted placeholders: {keys}")
            raise KeyError(message.format(placeholder=str(e), keys=tuple(pathDictionary.keys()))) from e

    PROJECT_DETAILS = {
        "DocumentName": "part_documentname",  # Old name of VWDocumentName. Not strictly a project detail, but close enough.
        "VWDocumentName": "part_documentname",  # Not strictly a project detail, but close enough.
        "ProjectID": "jobd_projectid",
        "ProjectCommission": "jobd_projectcomission",
        "ProjectConstruction": "jobd_projectconstruction",
        "PositionNumber": "jobd_positionnumber",
        "PositionQuantity": "jobd_positionquantity",
        "Administrator": "jobd_administrator",
        "PostalCode": "jobd_postalcode",
        "City": "jobd_city",
        "StreetName": "jobd_streetname",
        "FirstName": "jobd_firstname",
        "Title": "jobd_title",
        "Company": "jobd_company",
        "LastName": "jobd_lastname",
        "DeliveryDate": "jobd_deliverydate",
        "DeliveryWeek": "jobd_deliveryweek",
        "DeliveryYear": "jobd_deliveryyear",
        "CustomerNumber": "jobd_customernumber",
    }

    @staticmethod
    def updateDictionaryWithProjectDetails(pathDictionary, documentUUID):
        interiorcad.backend.datatracker.datatracker.beginConsuming()
        for (exportKey, dataTrackerKey) in ExportHelper.PROJECT_DETAILS.items():
            pathDictionary[exportKey] = interiorcad.backend.datatracker.datatracker.getItemAsString(documentUUID, 0, dataTrackerKey)

        try:
            pathDictionary['UserName'] = getpass.getuser()
        except ImportError:
            pathDictionary['UserName'] = os.path.split(os.path.expanduser("~"))[1]  # Home Path e.g. C:\Users\<UserName>
        else:
            pathDictionary['UserName'] = ''

        pathDictionary['VWDocumentFolder'] = os.path.dirname(interiorcad.backend.datatracker.datatracker.getItemAsString(documentUUID, 0, "part_documentpath"))

        interiorcad.backend.datatracker.datatracker.endConsuming()

    DATETIME_FORMATS = {
        'ISODateTime': '{:%Y%m%d %H%M%S}',
        'ISODate': '{:%Y%m%d}',
        'ISOTime': '{:%H%M%S}',
    }

    @staticmethod
    def updateDictionaryWithDateTime(dictionary):
        dictionary['FreeDateTime'] = DATETIME_NOW
        for (exportKey, formatTemplate) in ExportHelper.DATETIME_FORMATS.items():
            dictionary[exportKey] = formatTemplate.format(DATETIME_NOW)

    @staticmethod
    def getPositionData(positionNumber):
        return interiorcad.backend.datatracker.projectpositions.getDescriptionAndQuantityOfProjectPositionNamed(ExportHelper.getDocumentUUID(), positionNumber)

    @staticmethod
    def updateDictionaryWithDocumentData(pathDictionary, positionNumber):
        documentUUID = ExportHelper.getDocumentUUID()
        if documentUUID:
            ExportHelper.updateDictionaryWithProjectDetails(pathDictionary, documentUUID)
            pathDictionary['PositionDescription'], pathDictionary['PositionQuantity'] = ExportHelper.getPositionData(positionNumber)

    @staticmethod
    def __censurePotentialSeparators(name):
        return name.replace(os.sep, '').replace('/', '')

    @staticmethod
    def updateDictionaryWithPartData(pathDictionary, uuid):
        interiorcad.backend.datatracker.datatracker.beginConsuming()
        pathDictionary['PartName'] = interiorcad.backend.datatracker.datatracker.getItemAsString(uuid, 0, 'part_uniquename')
        pathDictionary['Material'] = interiorcad.backend.datatracker.datatracker.getItemAsString(uuid, 0, 'part_itemid')
        pathDictionary['CabinetName'] = ExportDataCollector.getCabinetUniqueName(uuid)
        pathDictionary['SideName'] = ''  # We need a dummy entry for the log file
        pathDictionary['SideNameNesting'] = ''  # We need a dummy entry for the log file
        pathDictionary['ExportFormat'] = ''  # We need a dummy entry for the log file  TODO IT-244 The log file will be written to another path than the nc files --mjk

        positionNumber = interiorcad.backend.datatracker.datatracker.getItemAsString(uuid, 0, 'jobd_positionnumber')
        ExportHelper.updateDictionaryWithDocumentData(pathDictionary, positionNumber)
        pathDictionary['PositionNumber'] = positionNumber

        constructionGroupID = interiorcad.backend.datatracker.datatracker.getItemAsLong(uuid, 0, 'part_constructiongroupid')
        constructionGroupComponentID = interiorcad.backend.datatracker.datatracker.getItemAsLong(uuid, 0, 'part_constructiongroupcomponentid')
        pathDictionary['ConstructionGroupName'] = pathDictionary['CGName'] = ExportHelper.__censurePotentialSeparators(getConstructionGroupName(constructionGroupID, constructionGroupComponentID))
        pathDictionary['ConstructionGroupComponentName'] = pathDictionary['CGCName'] = ExportHelper.__censurePotentialSeparators(getConstructionGroupComponentName(constructionGroupComponentID))

        interiorcad.backend.datatracker.datatracker.endConsuming()

    @staticmethod
    def createExportPathForCustomPart(exportConfiguration, uuid):
        exportPathDict = ExportHelper.createDictionaryForExportPath()
        ExportHelper.updateDictionaryWithPartData(exportPathDict, uuid)
        return ExportHelper.createExportPathFromConfigurationWithDictionary(exportConfiguration, exportPathDict)

    @staticmethod
    def mergePositionInfos(lhs, rhs, placeholder=None):
        newPositionInfo = {}
        for key in lhs.keys():
            if key in rhs and lhs[key] and rhs[key] and rhs[key] not in lhs[key].split(', '):
                newPositionInfo[key] = ', '.join((lhs[key], rhs[key]))
            elif key in rhs and ((not lhs[key] and not rhs[key]) or rhs[key] in lhs[key].split(', ')):
                newPositionInfo[key] = lhs[key]
            elif placeholder is not None:
                if key not in rhs or not rhs[key]:
                    newPositionInfo[key] = ', '.join((lhs[key], placeholder))
                elif not lhs[key]:
                    newPositionInfo[key] = ', '.join((placeholder, rhs[key]))
            elif key in rhs and not lhs[key] and rhs[key]:
                newPositionInfo[key] = rhs[key]
            else:
                newPositionInfo[key] = lhs[key]

        for key in rhs.keys():
            if key not in lhs:
                if placeholder:
                    newPositionInfo[key] = ', '.join((placeholder, rhs[key]))
                else:
                    newPositionInfo[key] = rhs[key]

        return newPositionInfo

    @staticmethod
    def mergeSimpleStringsAsCommaSeparatedString(lhs, rhs):
        if not lhs:
            return rhs
        elif lhs != rhs:
            return ','.join((lhs, rhs))
        else:
            return lhs

    @staticmethod
    def mergeSimpleStringsAsList(lhs, rhs) -> list:
        left = list(lhs) if isinstance(lhs, (list, tuple, set)) else [lhs]
        right = list(rhs) if isinstance(rhs, (list, tuple, set)) else [rhs]

        if not lhs:
            return right
        elif not rhs:
            return left
        else:
            return left + right

    @staticmethod
    def createExportPathWithoutWorkpiece(exportConfiguration):
        return ExportHelper.createExportPathForCustomPart(exportConfiguration=exportConfiguration, uuid='')

    @staticmethod
    def getUUIDOfFirstSelectedPart():
        uuids = ExportDataCollector.getExportUUIDsAndCounts()
        return uuids[0][0] if uuids else ''

    @staticmethod
    def createExportPathForFirstSelectedCustomPart(exportConfiguration):
        uuid = ExportHelper.getUUIDOfFirstSelectedPart()

        return ExportHelper.createExportPathForCustomPart(exportConfiguration, uuid)

    @staticmethod
    def getDocumentUUID():
        uuid = ExportHelper.getUUIDOfFirstSelectedPart()
        return interiorcad.backend.datatracker.datatracker.getItemAsString(uuid, 0, 'part_documentid')

    @staticmethod
    def createProjectInfoOfFirstSelectedCustomPart():
        documentUUID = ExportHelper.getDocumentUUID()

        projectInfo = {}
        ExportHelper.updateDictionaryWithProjectDetails(projectInfo, documentUUID)
        ExportHelper.updateDictionaryWithDateTime(projectInfo)
        return projectInfo

    @staticmethod
    def isProjectIDValid(_export, **keywords):
        # HACK: we do not want to consider the default project of interiorcad.sta as being valid (see also IC-6968)
        return keywords['ProjectID'] and keywords['ProjectID'] != 'A00023'

    @staticmethod
    def getProjectOutputFileNameVariables(export, **_keywords):
        # projectInfo = ExportHelper.createProjectInfoOfFirstSelectedCustomPart()
        projectInfo = export.getProjectInfo()
        return {
            'DocumentName': projectInfo['DocumentName'],
            'ProjectID': projectInfo['ProjectID'],
            'ProjectCommission': projectInfo['ProjectCommission'],
            'ProjectConstruction': projectInfo['ProjectConstruction'],
        }

    @staticmethod
    def createReducedOutputFileNameVariables(export, **keywords):
        outputFileName = keywords['outputFileName']
        outputVariables = export.getOutputFileNameVariables(**keywords)
        for key in outputVariables:
            if not outputVariables[key]:
                outputFileName = outputFileName.replace('{' + key + '}', '')
        outputFileName = FilenameCleaner().clean(filename=outputFileName)
        return outputFileName

    @staticmethod
    def getEdgesFromEntryOrWorkpiece(**keywords):
        if 'entry' not in keywords and 'workpiece' not in keywords:
            raise KeyError("No 'entry' or 'workpiece' found in keywords.")
        if 'entry' in keywords and keywords['entry'].edges:
            return keywords['entry'].edges
        elif 'workpiece' in keywords and keywords['workpiece'].edges:
            return keywords['workpiece'].edges

    @staticmethod
    def readEdgeData(edge, what):
        if edge:
            mappedEdgeData = {'supplierID': lambda: edge.supplierID, 'thickness': lambda: edge.thickness, 'length': lambda: edge.length, 'startjoint': lambda: edge.jointStart, 'endjoint': lambda: edge.jointEnd}
            if what not in mappedEdgeData:
                raise KeyError("Invalid variable name for edge: '{what}'".format(what=what))
            return mappedEdgeData[what]()

    @staticmethod
    def getRectangularEdgeData(**keywords):
        assert 'key' in keywords and keywords['key'].startswith('part_edge'), "Unexpected **keywords - this method needs 'key' to start with 'part_edge'"
        edges = ExportHelper.getEdgesFromEntryOrWorkpiece(**keywords)
        if edges:
            key = keywords['key']
            what = key[10:] or 'supplierID'
            return ExportHelper.readEdgeData(edges.getRectangularEdge(int(key[9])), what or 'supplierID')

    @staticmethod
    def getKeyOrData(**keywords):
        if 'key' in keywords:
            if keywords['key'].startswith('part_edge'):
                return ExportHelper.getRectangularEdgeData(**keywords)
            else:
                uuid = keywords['uuid'] if 'uuid' in keywords else keywords['workpiece'].uuid
                return keywords['getter'](uuid, keywords['key'])
        else:
            return keywords['data']

    @staticmethod
    def getDataAsString(**keywords):
        return ExportHelper.getKeyOrData(getter=ExportDataCollector.getRawDataAsString, **keywords) or ''

    @staticmethod
    def getDataAsInt(**keywords):
        return ExportHelper.getKeyOrData(getter=ExportDataCollector.getRawDataAsInt, **keywords) or 0

    @staticmethod
    def getDataAsFloat(**keywords):
        return ExportHelper.getKeyOrData(getter=ExportDataCollector.getRawDataAsFloat, **keywords) or 0.0
