# -*- coding: utf-8 -*-

# Copyright 2011-2018, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import sqlite3
from os import path
import unittest
import atexit
from collections import namedtuple
import struct

import interiorcad.backend.paths


class MockDatatracker:
    def __init__(self, _pathname=None):
        self.alreadyOpened = False
        self.databasename = None

    def __del__(self):
        self.closeDB()

    def __enter__(self):
        self.openDB()
        return self

    def __exit__(self, _exitType, _value, _traceback):
        self.closeDB()
        return True

    def openDB(self):
        if not self.alreadyOpened:
            assert self.databasename, "No database defined."
            assert path.isfile(self.databasename), "Database not found."
            self.db = sqlite3.connect(self.databasename)
            self.alreadyOpened = True
            self.assertDatabaseIsValid()

    def closeDB(self):
        if self.alreadyOpened:
            self.db.close()
            self.alreadyOpened = False

    def beginConsuming(self):
        self.openDB()

    def endConsuming(self):
        self.openDB()

    def __generateSQLResult(self, query):
        self.openDB()
        sqlCursor = self.db.cursor()
        sqlCursor.execute(query)
        for rowTuple in sqlCursor:
            yield rowTuple

    def getExportUUIDs(self):
        # The WHERE clause is to make mockdatatracker.getExportUUIDs as similar as possible to XGEmbeddedPython.getExportUUIDs.
        sqlQuery = "SELECT DISTINCT uuid, 1 FROM datatrackertable WHERE uuid NOT IN (SELECT DISTINCT uuid FROM datatrackertable WHERE key = 'part_type' and value = 'XG Document')"
        return list(self.__generateSQLResult(sqlQuery))

    def getUnfilteredExportUUIDs(self):
        return self.getExportUUIDs()

    def getValueForKey(self, uuid, subid, key):
        self.openDB()
        sqlCursor = self.db.cursor()
        sqlCursor.execute("SELECT value FROM datatrackertable WHERE uuid='" + uuid + "' AND subindex=" + str(subid) + " AND key='" + key + "'")
        sqlResult = sqlCursor.fetchone()
        value = None
        if sqlResult:
            value = sqlResult[0]
        return value

    def getItemAsString(self, uuid, subid, key):
        value = self.getValueForKey(uuid, subid, key)
        valueAsString = ""
        if value:
            valueAsString = self._getValueAsString(value)
        return valueAsString

    def getItemAsLong(self, uuid, subid, key):
        value = self.getValueForKey(uuid, subid, key)
        return self._getValueAsLong(value)

    def getItemAsDouble(self, uuid, subid, key):
        value = self.getValueForKey(uuid, subid, key)
        return self._getValueAsDouble(value)

    def getItemAsBool(self, uuid, subid, key):
        value = self.getValueForKey(uuid, subid, key)
        return self._getValueAsBool(value)

    def getItemAsLegacyPolygon2D(self, uuid, subid, key):
        numItems = self.getItemAsLong(uuid, subid, self.createKeyWithPostFixKey(key, "NumItems"))
        return [
            {
                'x': self.getValueForKey(uuid, subid, self.createKeyWithPostFixKeyAndID(key, "_point_x", index)),
                'y': self.getValueForKey(uuid, subid, self.createKeyWithPostFixKeyAndID(key, "_point_y", index)),
                'type': self.getValueForKey(uuid, subid, self.createKeyWithPostFixKeyAndID(key, "_point_type", index)),
                'radius': self.getValueForKey(uuid, subid, self.createKeyWithPostFixKeyAndID(key, "_point_radius", index)),
                'visible': self.getValueForKey(uuid, subid, self.createKeyWithPostFixKeyAndID(key, "_point_visible", index)),
            }
            for index in range(numItems)
        ]

    Matrix = namedtuple("Matrix", "a00 a01 a02 a10 a11 a12 a20 a21 a22 xOffset yOffset zOffset")

    # 12 doubles (8 bytes each) expected, in native format
    MatrixBlobFormat = '@dddddddddddd'
    assert struct.calcsize(MatrixBlobFormat) == 12 * 8, 'Binary compatibility broken.'

    def getMatrix(self, uuid, subid, key):
        blob = self.getValueForKey(uuid, subid, key)
        try:
            return self.Matrix(*struct.unpack(self.MatrixBlobFormat, blob))
        except (TypeError, struct.error) as e:
            raise LookupError from e

    def getChildUUIDsByPartType(self, parentUUID, partType):
        sqlQuery = "SELECT DISTINCT uuid FROM datatrackertable WHERE key='part_type' AND value='" + partType + "' AND uuid in (SELECT DISTINCT uuid FROM datatrackertable WHERE key IN ('part_parentuuid', 'part_userparentuuid', 'part_piocontaineruuid') AND value='" + str(parentUUID) + "')"
        return [uuid for (uuid,) in self.__generateSQLResult(sqlQuery)]

    def getChildFittingUUIDs(self, parentUUID):
        sqlQuery = "SELECT DISTINCT uuid FROM datatrackertable WHERE value='" + parentUUID + "' AND key LIKE 'fitting_parentuuid%'"
        return [uuid for (uuid,) in self.__generateSQLResult(sqlQuery)]

    # See also DataTrackerSQL::CreateRoutingUUIDs in C++
    def getRoutingUUIDs(self, customPartUUID):
        sqlQuery = "SELECT DISTINCT uuid FROM datatrackertable WHERE key='rout_partuuid' AND value='" + customPartUUID + "'"
        return [uuid for (uuid,) in self.__generateSQLResult(sqlQuery)]

    def createKeyWithPostFixKey(self, key, postfixKey):
        return key + postfixKey

    def createKeyWithPostFixKeyAndID(self, key, postfixKey, xg_id):
        return key + str(xg_id) + postfixKey

    def assertDatabaseIsValid(self):
        sqlCursor = self.db.cursor()
        try:
            sqlCursor.execute("SELECT * FROM datatrackertable")
        except sqlite3.OperationalError as e:
            raise Exception("Database is not valid") from e

    def _getValueAsString(self, value):
        return str(value)

    def _getValueAsLong(self, value):
        valueAsLong = 0
        try:
            if value:
                valueAsLong = int(value)
        except ValueError:
            pass
        return valueAsLong

    def _getValueAsDouble(self, value):
        valueAsDouble = 0.0
        try:
            if value:
                valueAsDouble = float(value)
        except ValueError:
            pass

        return valueAsDouble

    def _getValueAsBool(self, value):
        valueAsLong = self._getValueAsLong(value)
        return valueAsLong == 1

    def __setattr__(self, name, value):
        if name == "databasename":
            self.closeDB()
        self.__dict__[name] = value

    def setNCProgramNames(self, dummyUUID, dummyExportName, summySizingName, dummyMainName, dummyOppositeName):
        pass


mockDatatrackerInstance = MockDatatracker()


def setTestDatabase(dbFilename):
    dbFilepath = dbFilename
    if not path.isabs(dbFilename):
        dbFilepath = path.join(interiorcad.backend.paths.mockDocuments, dbFilename)

    assert path.isfile(dbFilepath), "Can't find database at " + dbFilepath

    mockDatatrackerInstance.databasename = dbFilepath


def closeDatabaseConnection():
    mockDatatrackerInstance.closeDB()


atexit.register(closeDatabaseConnection)  # closeDatabaseConnection gets called, when python shell exits.


class MockDatatrackerSimpleTests(unittest.TestCase):
    def testGetValueAsString(self):
        valueAsString = mockDatatrackerInstance._getValueAsString(float(5.01))
        self.assertIsInstance(valueAsString, str)
        self.assertEqual(valueAsString, "5.01")

        valueAsString = mockDatatrackerInstance._getValueAsString(int(5))
        self.assertIsInstance(valueAsString, str)
        self.assertEqual(valueAsString, "5")

        valueAsString = mockDatatrackerInstance._getValueAsString(None)
        self.assertIsInstance(valueAsString, str)
        self.assertEqual(valueAsString, "None")

        valueAsString = mockDatatrackerInstance._getValueAsString("The String")
        self.assertIsInstance(valueAsString, str)
        self.assertEqual(valueAsString, "The String")

    def testGetValueAsLong(self):
        valueAsLong = mockDatatrackerInstance._getValueAsLong(5.01)
        self.assertIsInstance(valueAsLong, int)
        self.assertEqual(valueAsLong, 5)

        valueAsLong = mockDatatrackerInstance._getValueAsLong(17)
        self.assertIsInstance(valueAsLong, int)
        self.assertEqual(valueAsLong, 17)

        valueAsLong = mockDatatrackerInstance._getValueAsLong("hurz")
        self.assertIsInstance(valueAsLong, int)
        self.assertEqual(valueAsLong, 0)

        valueAsLong = mockDatatrackerInstance._getValueAsLong("01234")
        self.assertIsInstance(valueAsLong, int)
        self.assertEqual(valueAsLong, 1234)

        valueAsLong = mockDatatrackerInstance._getValueAsLong(None)
        self.assertIsInstance(valueAsLong, int)
        self.assertEqual(valueAsLong, 0)

    def testGetValueAsDouble(self):
        valueAsDouble = mockDatatrackerInstance._getValueAsDouble(5.01)
        self.assertIsInstance(valueAsDouble, float)
        self.assertEqual(valueAsDouble, 5.01)

        valueAsDouble = mockDatatrackerInstance._getValueAsDouble(17)
        self.assertIsInstance(valueAsDouble, float)
        self.assertEqual(valueAsDouble, 17.0)

        valueAsDouble = mockDatatrackerInstance._getValueAsDouble("hurz")
        self.assertIsInstance(valueAsDouble, float)
        self.assertEqual(0.0, valueAsDouble)

        valueAsDouble = mockDatatrackerInstance._getValueAsDouble("01234")
        self.assertIsInstance(valueAsDouble, float)
        self.assertEqual(1234.0, valueAsDouble)

        valueAsDouble = mockDatatrackerInstance._getValueAsDouble(None)
        self.assertIsInstance(valueAsDouble, float)
        self.assertEqual(0.0, valueAsDouble)

    def testGetValueAsBool(self):
        valueAsBool = mockDatatrackerInstance._getValueAsBool("1")
        self.assertTrue(valueAsBool)

        valueAsBool = mockDatatrackerInstance._getValueAsBool("0")
        self.assertFalse(valueAsBool)

        valueAsBool = mockDatatrackerInstance._getValueAsBool("5")
        self.assertFalse(valueAsBool)
