# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.vertex import Vertex


class PolygonConverter:
    def convertPolyline(self, vertices):
        return Polyline.createFromIterable(self.convertVertices(vertices))

    def convertVertices(self, vertices):
        return self.__getPathWithValidRadius([self.convertVertex(vertex) for vertex in vertices])

    def convertVertex(self, vertex):
        position = Point(vertex['x'], vertex['y'])

        vertexType = vertex['type']
        # TODO IT-233: can we change the code to use the type in the derived classes? -- NG
        if vertexType == Vertex.CORNER_VERTEX:
            convertedVertex = CornerVertex(position)
        elif vertexType == Vertex.RADIUS_VERTEX:
            convertedVertex = RadiusVertex(position, radius=vertex['radius'])
        else:
            assert True, 'Unknown vertex type {type}'.format(type=vertexType)
            convertedVertex = CornerVertex(position)

        convertedVertex.visible = vertex['visible']

        return convertedVertex

    def __getPathWithValidRadius(self, vertices):
        # In general: VW has normally no radius for radius vertices. It is just interiorcad that makes sure that the datatracker is correctly filled. However, be prepared for the day, where we do not do it anymore.
        #
        # In particular: in interiorcad 2019 F6, F7, F8 and 2020 F1, the radius of a radius vertex could be 0, as the C++ code called by _contourpath.getContourPathWithDistance was buggy. For such versions, we need to correct broken radius IC-20343.

        for i in range(len(vertices)):
            vertex = vertices[i]
            if vertex.type == Vertex.RADIUS_VERTEX and isNearlyEqual(vertex.radius, 0.0):
                j = (i+1) % len(vertices)
                vertex.radius = Polyline.calculateRadiusFrom3Points(vertices[i-1].position, vertices[i].position, vertices[j].position)

            yield vertex


class PolygonConverterTests(unittest.TestCase):
    def setUp(self):
        self.converter = PolygonConverter()

    def testConvertCornerVertexTuple(self):
        converted = self.converter.convertVertex(self.cornerVertex())
        expected = self.expectedCornerVertex()
        self.assertEqual(expected, converted)

    def testConvertRadiusVertexTuple(self):
        converted = self.converter.convertVertex(self.radiusVertex())
        expected = self.expectedRadiusVertex()
        self.assertEqual(expected, converted)

    def testConvertPolyline(self):
        polyline = [self.cornerVertex(), self.radiusVertex()]
        converted = self.converter.convertPolyline(polyline)
        expected = Polyline()
        expected.addVertex(self.expectedCornerVertex())
        expected.addVertex(self.expectedRadiusVertex())
        self.assertEqual(expected, converted)

    def cornerVertex(self):
        return {'x': 1, 'y': 2, 'type': Vertex.CORNER_VERTEX, 'radius': 0, 'visible': True}

    def expectedCornerVertex(self):
        v = CornerVertex(Point(1, 2))
        v.visible = True
        return v

    def radiusVertex(self):
        return {'x': 11, 'y': 22, 'type': Vertex.RADIUS_VERTEX, 'radius': 33, 'visible': True}

    def expectedRadiusVertex(self):
        expected = RadiusVertex(Point(11, 22), radius=33)
        expected.visible = True
        return expected
