# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import json
import logging
import os.path
import unittest
from collections import OrderedDict
from functools import lru_cache

import interiorcad.backend.exportdatasafe
from interiorcad.backend.exportmainpath import getAuxiliaryFilesPath
from interiorcad.backend.filenamecleaner import FilenameCleaner
from interiorcad.backend.jsonfilewriter import JSONFileWriter
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP


@lru_cache()
def getJSONConfigurationPaths():
    writeDictDosFilenamesFile = os.path.join(getAuxiliaryFilesPath(), 'DosFilenames Dictionary.json')
    readDictDosFilenamesFile = writeDictDosFilenamesFile
    if interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS and not os.path.exists(readDictDosFilenamesFile):
        machineConfigurationFile = interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS[0][1]
        # Only the NC export can find the old file, as it is in exportstarter/nc-export.
        if machineConfigurationFile:
            readDictDosFilenamesFile = os.path.join(os.path.dirname(machineConfigurationFile), 'DosFilenames Dictionary.json')

    return readDictDosFilenamesFile, writeDictDosFilenamesFile


class DosFilenameCreator:
    def deleteDosExcludes(self, name):
        excludes = ['"', '*', '+', ',', '/', ':', ';', '<', '=', '>', '?', '\\', '[', ']', '|', '.', '_', ' ']   # '_' isn't an exclude, but better for shorter names
        for value in excludes:
            name = name.replace(value, '')
        return name

    def filenameCleanerInclDeleteDosExcludes(self, filename):
        filenameCleaner = FilenameCleaner()
        filename = filenameCleaner.clean(filename)   # convert äöüÄÖÜ before deleteDosExcludes and shortenFilenames
        filename = self.deleteDosExcludes(filename)
        return filename

    def defaultDictDosFilenames(self):
        return json.loads("""{
            "Seite": "Se",
            "Boden": "Bo",
            "Konstruktionsboden": "Kb",
            "Fachboden": "Fb",
            "Mittelseitenstollen": "Mst",
            "Mittelseiten": "Ms",
            "Mittelseite": "Ms",
            "Travers": "Tr",
            "Ruckwand": "Rw",
            "Tur": "Fr",
            "Blindfront": "BFr",
            "Doppel": "D",
            "Holzschubkasten": "HSk",
            "Systemschubkasten": "SSk",
            "Schubkasten": "Sk",
            "Vorderstuck": "Vs",
            "Hinterstuck": "Hs",
            "Blende": "Bl",
            "Sockelverstarker": "Sov",
            "Sockel": "So",
            "Stollen": "St",
            "Klappe": "Kl",
            "Hangeleiste": "Hl",
            "Rahmenfries": "Rf",
            "Mittelfries": "Mf",
            "FrontrahmenFries": "Frf",
            "FrontrahmenMittelfries": "Frmf",
            "Frontrahmen": "Frr",
            "Front": "Fr",
            "Rahmenfullung": "Rfu",
  
            "Links": "Li",
            "Rechts": "Re",
            "Oben": "Ob",
            "Unten": "Un",
            "Vorne": "Vo",
            "Hinten": "Hi",
            "Schrag": "S",
            "Aufrecht": "A",
            "Waagrecht": "W",
            "Mitte": "Mi",
  
            "Bauteil": "Bt",
  
            "Haupt": "H",
            "Neben": "N"
        }""", object_pairs_hook=OrderedDict)

    def getDOSFileNames(self):
        readDictDosFilenamesFile, writeDictDosFilenamesFile = getJSONConfigurationPaths()
        rewriteUserFile = readDictDosFilenamesFile != writeDictDosFilenamesFile or not os.path.exists(writeDictDosFilenamesFile)

        dictDosFilenames = self.defaultDictDosFilenames()
        try:
            with open(readDictDosFilenamesFile, 'r') as handle:
                dictDosFilenames = json.load(handle, object_pairs_hook=OrderedDict)
        except FileNotFoundError:
            rewriteUserFile = True
        except Exception as e:
            logging.warning(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 81, "Error while reading the JSON file '{filepath}': {exception}").format(filepath=writeDictDosFilenamesFile, exception=str(e)))
            logging.info(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 83, "Note the original JSON syntax when making changes. Only the ASCII characters 32 to 127 will work correctly, special characters require Unicode characters. The JSON file will be re-created if you delete it."))
        if self.defaultDictDosFilenames() != dictDosFilenames:
            for defaultKey in self.defaultDictDosFilenames():
                if defaultKey not in dictDosFilenames:
                    dictDosFilenames[defaultKey] = self.defaultDictDosFilenames()[defaultKey]
                    rewriteUserFile = True

        if rewriteUserFile:
            JSONFileWriter().writeFile(writeDictDosFilenamesFile, dictDosFilenames)

        return dictDosFilenames

    def shortenFilenames(self, filename):
        dictDosFilenames = self.getDOSFileNames()

        for key in dictDosFilenames:
            filename = filename.replace(key, dictDosFilenames[key])
        return filename

    def createDosFilename(self, filename):
        filename = self.filenameCleanerInclDeleteDosExcludes(filename)
        filename = self.shortenFilenames(filename)
        return filename


class DosFilenameCreatorTests(unittest.TestCase):
    class TestableDosFilenameCreator(DosFilenameCreator):
        def getDOSFileNames(self):
            return self.defaultDictDosFilenames()

    def setUp(self):
        self.creator = self.TestableDosFilenameCreator()

    def testSubstituteCustomPartnames(self):
        self.assertEqual('FrFarbigBeschlageSeLi1H', self.creator.createDosFilename('Front_Farbig_+_Beschlage_Seite_Links1_H'))

    def testRemoveDos83Excludes(self):
        self.assertEqual('', self.creator.createDosFilename(r'"*+,/:;<=>?\[]|._ '))

    def testSubstituteGermanCabinetNames(self):
        """Tests names from dev\XG Cabinet\XGCabinet-D.vwr\Strings\11010 Cabinet Part Base Name.vwstrings"""

        self.assertEqual('Bt1', self.creator.createDosFilename("Bauteil-1"))
        self.assertEqual('SeLi1', self.creator.createDosFilename("Seite Links-1"))
        self.assertEqual('SeRe1', self.creator.createDosFilename("Seite Rechts-1"))
        self.assertEqual('StLiVo1', self.creator.createDosFilename("Stollen Links Vorne-1"))
        self.assertEqual('StLiHi1', self.creator.createDosFilename("Stollen Links Hinten-1"))
        self.assertEqual('StReVo1', self.creator.createDosFilename("Stollen Rechts Vorne-1"))
        self.assertEqual('StReHi1', self.creator.createDosFilename("Stollen Rechts Hinten-1"))
        self.assertEqual('BoOb1', self.creator.createDosFilename("Boden Oben-1"))
        self.assertEqual('BoUn1', self.creator.createDosFilename("Boden Unten-1"))
        self.assertEqual('Rw1', self.creator.createDosFilename("Rückwand-1"))
        self.assertEqual('SoVo1', self.creator.createDosFilename("Sockel Vorne-1"))
        self.assertEqual('BlOb1', self.creator.createDosFilename("Blende Oben-1"))
        self.assertEqual('Fr1', self.creator.createDosFilename("Tür-1"))
        self.assertEqual('DturLi1', self.creator.createDosFilename("Doppeltür Links-1"))
        self.assertEqual('DturRe1', self.creator.createDosFilename("Doppeltür Rechts-1"))
        self.assertEqual('SkFr1', self.creator.createDosFilename("Schubkasten Front-1"))
        self.assertEqual('Ms1', self.creator.createDosFilename("Mittelseite-1"))
        self.assertEqual('MstVo1', self.creator.createDosFilename("Mittelseitenstollen Vorne-1"))
        self.assertEqual('MstHi1', self.creator.createDosFilename("Mittelseitenstollen Hinten-1"))
        self.assertEqual('Kb1', self.creator.createDosFilename("Konstruktionsboden-1"))
        self.assertEqual('TrObVo1', self.creator.createDosFilename("Travers Oben Vorne-1"))
        self.assertEqual('TrObHi1', self.creator.createDosFilename("Travers Oben Hinten-1"))
        self.assertEqual('TrUnVo1', self.creator.createDosFilename("Travers Unten Vorne-1"))
        self.assertEqual('TrUnHi1', self.creator.createDosFilename("Travers Unten Hinten-1"))
        self.assertEqual('TrVo1', self.creator.createDosFilename("Travers Vorne-1"))
        self.assertEqual('TrHi1', self.creator.createDosFilename("Travers Hinten-1"))
        self.assertEqual('Fb1', self.creator.createDosFilename("Fachboden-1"))
        self.assertEqual('BlLi1', self.creator.createDosFilename("Blende Links-1"))
        self.assertEqual('BlRe1', self.creator.createDosFilename("Blende Rechts-1"))
        self.assertEqual('SoHi1', self.creator.createDosFilename("Sockel Hinten-1"))
        self.assertEqual('SoLi1', self.creator.createDosFilename("Sockel Links-1"))
        self.assertEqual('SoRe1', self.creator.createDosFilename("Sockel Rechts-1"))
        self.assertEqual('Sov1', self.creator.createDosFilename("Sockelverstärker-1"))
        self.assertEqual('BFr1', self.creator.createDosFilename("Blindfront-1"))
        self.assertEqual('RfLiA1', self.creator.createDosFilename("Rahmenfries Links Aufrecht-1"))
        self.assertEqual('RfReA1', self.creator.createDosFilename("Rahmenfries Rechts Aufrecht-1"))
        self.assertEqual('RfUnW1', self.creator.createDosFilename("Rahmenfries Unten Waagrecht-1"))
        self.assertEqual('RfObW1', self.creator.createDosFilename("Rahmenfries Oben Waagrecht-1"))
        self.assertEqual('MfA1', self.creator.createDosFilename("Mittelfries Aufrecht-1"))
        self.assertEqual('MfW1', self.creator.createDosFilename("Mittelfries Waagrecht-1"))
        self.assertEqual('HlOb1', self.creator.createDosFilename("Hängeleiste Oben-1"))
        self.assertEqual('HlUn1', self.creator.createDosFilename("Hängeleiste Unten-1"))
        self.assertEqual('HlMi1', self.creator.createDosFilename("Hängeleiste Mitte-1"))
        self.assertEqual('Rfu1', self.creator.createDosFilename("Rahmenfüllung-1"))
        self.assertEqual('FrfLiA1', self.creator.createDosFilename("Frontrahmen Fries Links Aufrecht-1"))
        self.assertEqual('FrfReA1', self.creator.createDosFilename("Frontrahmen Fries Rechts Aufrecht-1"))
        self.assertEqual('FrfUnW1', self.creator.createDosFilename("Frontrahmen Fries Unten Waagrecht-1"))
        self.assertEqual('FrfObW1', self.creator.createDosFilename("Frontrahmen Fries Oben Waagrecht-1"))
        self.assertEqual('FrrMfA1', self.creator.createDosFilename("Frontrahmen Mittelfries Aufrecht-1"))
        self.assertEqual('FrrMfW1', self.creator.createDosFilename("Frontrahmen Mittelfries Waagrecht-1"))
        self.assertEqual('Kl1', self.creator.createDosFilename("Klappe-1"))
        self.assertEqual('Frr1', self.creator.createDosFilename("Frontrahmen-1"))

    def testSubstituteGermanDrawerNames(self):
        """Tests names from dev\XG Custom Part\XGCustomPart-D.vwr\Strings\13001 Other strings.vwstrings"""

        self.assertEqual('SSkBo1', self.creator.createDosFilename("Systemschubkasten Boden-1"))  # Index 25
        self.assertEqual('SSkRw1', self.creator.createDosFilename("Systemschubkasten Rückwand-1"))  # Index 26
        self.assertEqual('HSkBo1', self.creator.createDosFilename("Holzschubkasten Boden-1"))  # Index 37
        self.assertEqual('HSkHs1', self.creator.createDosFilename("Holzschubkasten Hinterstück-1"))  # Index 28
        self.assertEqual('HSkSeLi1', self.creator.createDosFilename("Holzschubkasten Seite Links-1"))  # Index 39
        self.assertEqual('HSkSeRe1', self.creator.createDosFilename("Holzschubkasten Seite Rechts-1"))  # Index 40
        self.assertEqual('HSkVs1', self.creator.createDosFilename("Holzschubkasten Vorderstück-1"))  # Index 41

    def testSubstituteGermanSlantedNames(self):
        """Tests names from dev\XG Custom Part\XGCustomPart-D.vwr\Strings\13001 Other strings.vwstrings"""
        self.assertEqual('SeLiS1', self.creator.createDosFilename("Seite Links Schräg-1")) # Index 27
        self.assertEqual('SeReS1', self.creator.createDosFilename("Seite Rechts Schräg-1"))  # Index 28
