# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import json
import logging
import os.path
import unittest
from collections import OrderedDict
from functools import lru_cache

from interiorcad.backend.exportmainpath import getAuxiliaryFilesPath
from interiorcad.backend.jsonfilewriter import JSONFileWriter
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP


@lru_cache()
def getJSONConfigurationPaths():
    writeDictFilenameAndPathFilterFile = os.path.join(getAuxiliaryFilesPath(), 'FilenameAndPathFilter Dictionary.json')
    readDictFilenameAndPathFilterFile = writeDictFilenameAndPathFilterFile

    # FilenameAndPathFilter Dictionary.json was never in the folder 'nc-export'. Therefore, there is no need for such code, unlike in interiorcad.backend.dosfilenamecreator.getJSONConfigurationPaths

    return readDictFilenameAndPathFilterFile, writeDictFilenameAndPathFilterFile


class FilenameAndPathFilter:
    def defaultDictionary(self):
        return json.loads(r"""{
            "_ReadMe": "Note the original JSON syntax when making changes. Only the ASCII characters 32 to 127 will work correctly, special characters require Unicode characters in C/C++/Java source code format.",
            "diacritics": {
                "\u00e4": "a",
                "\u00f6": "o",
                "\u00fc": "u",
                "\u00c4": "A",
                "\u00d6": "O",
                "\u00dc": "U",
                "\u00df": "s"
            },
            "replacements": {
                " ": "_",
                "-": "",

                "\\": "\\",
                "/": "/",
                ":": ":",
                "*": "*",
                "?": "?",
                "\"": "\"",
                "<": "<",
                ">": ">",
                "|": "|",

                "\u00a7": "\u00a7",
                "$": "$",
                "%": "%",
                "&": "&",
                "(": "(",
                ")": ")",
                "{": "{",
                "}": "}",
                "[": "[",
                "]": "]",
                "=": "=",
                "\u00b4": "\u00b4",
                "`": "`",
                "'": "'",
                "#": "#",
                "+": "+",
                ",": ",",
                ".": ".",
                ";": ";",
                "~": "~",
                "@": "at",
                "\u20ac": "E",
                "\u00b2": "2",
                "\u00b3": "3"
            },
            "shortnames": {
                "Seite": "Se",
                "Boden": "Bo",
                "Konstruktionsboden": "Kb",
                "Fachboden": "Fb",
                "Mittelseitenstollen": "Mst",
                "Mittelseiten": "Ms",
                "Mittelseite": "Ms",
                "Travers": "Tr",
                "Ruckwand": "Rw",
                "Tur": "Fr",
                "Blindfront": "BFr",
                "Doppel": "D",
                "Holzschubkasten": "HSk",
                "Systemschubkasten": "SSk",
                "Schubkasten": "Sk",
                "Vorderstuck": "Vs",
                "Hinterstuck": "Hs",
                "Blende": "Bl",
                "Sockelverstarker": "Sov",
                "Sockel": "So",
                "Stollen": "St",
                "Klappe": "Kl",
                "Hangeleiste": "Hl",
                "Rahmenfries": "Rf",
                "Mittelfries": "Mf",
                "FrontrahmenFries": "Frf",
                "FrontrahmenMittelfries": "Frmf",
                "Frontrahmen": "Frr",
                "Front": "Fr",
                "Rahmenfullung": "Rfu",

                "Links": "Li",
                "Rechts": "Re",
                "Oben": "Ob",
                "Unten": "Un",
                "Vorne": "Vo",
                "Hinten": "Hi",
                "Schrag": "S",
                "Aufrecht": "A",
                "Waagrecht": "W",
                "Mitte": "Mi",

                "Bauteil": "Bt",

                "Haupt": "H",
                "Neben": "N"
            }
        }""", object_pairs_hook=OrderedDict)

    def getFilenameAndPathFilterDictionary(self, useIndividualFilters):
        readDictFilenameAndPathFilterFile, writeDictFilenameAndPathFilterFile = getJSONConfigurationPaths()
        rewriteUserFile = readDictFilenameAndPathFilterFile != writeDictFilenameAndPathFilterFile or not os.path.exists(writeDictFilenameAndPathFilterFile)

        FilenameAndPathFilterDict = self.defaultDictionary()
        try:
            with open(readDictFilenameAndPathFilterFile, 'r') as handle:
                FilenameAndPathFilterDict = json.load(handle, object_pairs_hook=OrderedDict)
        except FileNotFoundError:
            rewriteUserFile = True
        except Exception as e:
            logging.warning(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 81, "Error while reading the JSON file '{filepath}': {exception}").format(filepath=writeDictFilenameAndPathFilterFile, exception=str(e)))
            logging.info(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 83, "Note the original JSON syntax when making changes. Only the ASCII characters 32 to 127 will work correctly, special characters require Unicode characters. The JSON file will be re-created if you delete it."))
        if self.defaultDictionary() != FilenameAndPathFilterDict:
            for defaultKey in self.defaultDictionary():
                if defaultKey not in FilenameAndPathFilterDict:
                    FilenameAndPathFilterDict[defaultKey] = self.defaultDictionary()[defaultKey]
                    rewriteUserFile = True

        if useIndividualFilters and rewriteUserFile:
            JSONFileWriter().writeFile(writeDictFilenameAndPathFilterFile, FilenameAndPathFilterDict)

        return FilenameAndPathFilterDict

    def filterKeysFromDictInName(self, name, selectedDict):
        for key in selectedDict:
            if key in name:
                name = name.replace(key, selectedDict[key])
        return name

    def replaceDiacritics(self, name, entireDict):
        return self.filterKeysFromDictInName(name=name, selectedDict=entireDict['diacritics'])

    def replaceReplacements(self, name, entireDict):
        return self.filterKeysFromDictInName(name=name, selectedDict=entireDict['replacements'])

    def shortenNames(self, name, entireDict):
        return self.filterKeysFromDictInName(name=name, selectedDict=entireDict['shortnames'])

    def replaceUnallowedCharacters(self, name):
        unallowedCharacters = list(r'\/:*?"<>|')
        for key in unallowedCharacters:
            if key in name:
                name = name.replace(key, '')  # TODO: optional individual character  --mjk
        return name

    # replace('__', '_' does NOT always work
    @staticmethod
    def getSimplifiedMultipleUnderscores(value):
        previousCharacter = None
        result = ''
        for character in value:
            if character != '_' or previousCharacter != '_':
                result += character
            previousCharacter = character
        return result

    @staticmethod
    def deleteLeadingAndTrailingUnderscores(value):
        return value.strip('_')

    def filenameAndPathFilter(self, filterTerm, filterValue, useIndividualFilters):
        entireDict = self.getFilenameAndPathFilterDictionary(useIndividualFilters=useIndividualFilters)

        result = filterTerm
        result = self.replaceDiacritics(result, entireDict=entireDict) if 'diacritics' in filterValue else result
        result = self.replaceReplacements(result, entireDict=entireDict) if 'replacements' in filterValue else result
        result = self.shortenNames(result, entireDict=entireDict) if 'shortnames' in filterValue else result
        result = self.replaceUnallowedCharacters(result)
        result = self.getSimplifiedMultipleUnderscores(result)
        result = self.deleteLeadingAndTrailingUnderscores(result)
        return result


class TestFilenameAndPathFilterCreator(unittest.TestCase):
    class FilenameAndPathFilterCreator(FilenameAndPathFilter):
        def getFilenameAndPathFilterDictionary(self, useIndividualFilters):
            return self.defaultDictionary()

    def setUp(self):
        self.creator = self.FilenameAndPathFilterCreator()


class ReplaceExcludesTests(TestFilenameAndPathFilterCreator):
    def testReplaceExcludes(self):
        self.assertEqual('+,;=[]._ ', self.creator.replaceUnallowedCharacters(r'"*+,/:;<=>?\[]|._ '))

    def testRemoveSlash(self):
        self.assertEqual('MainSide', self.creator.replaceUnallowedCharacters('/Main/Side/'))


class ReplaceDiacriticsTests(TestFilenameAndPathFilterCreator):
    def testReplaceUmlautsWithConsonants(self):
        self.assertEqual('aouAOU', self.creator.replaceDiacritics('\xe4\xf6\xfc\xc4\xd6\xdc', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))

    def testReplaceSpecialSWithNormalS(self):
        self.assertEqual('s', self.creator.replaceDiacritics('\xdf', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))


class ReplaceReplacementsTests(TestFilenameAndPathFilterCreator):
    def testReplaceSpacesWithUnderscore(self):
        self.assertEqual('Bauteil_2_Main', self.creator.replaceReplacements('Bauteil 2 Main', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))

    def testReplaceDashesWithUnderscrores(self):
        self.assertEqual('Bauteil2_Main', self.creator.replaceReplacements('Bauteil-2_Main', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))

    def testReplaceNearlyUnallowedCharacters(self):
        self.assertEqual('"*+,/:;<=>?\[]|.__', self.creator.replaceReplacements(r'"*+,/:;<=>?\[]|._ ', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))


class ShortenNamesTests(TestFilenameAndPathFilterCreator):
    def testReplaceSeiteLinks2Haupt(self):
        self.assertEqual('Se Li-2_H', self.creator.shortenNames('Seite Links-2_Haupt', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))

    def testReplaceBodenOben2Neben(self):
        self.assertEqual('Bo Ob-2_N', self.creator.shortenNames('Boden Oben-2_Neben', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))

    def testReplaceBauteil2Haupt(self):
        self.assertEqual('Bt-2_H', self.creator.shortenNames('Bauteil-2_Haupt', self.creator.getFilenameAndPathFilterDictionary(useIndividualFilters=False)))


class FilenameAndPathFilterTests(TestFilenameAndPathFilterCreator):
    def testRemoveDos83Excludes(self):
        self.assertEqual('+,;=[].', self.creator.filenameAndPathFilter(filterTerm=r'"*+,/:;<=>?\[]|._ ', filterValue='diacritics replacements shortnames', useIndividualFilters=False))

    def testGreekCapitalLetterAlphaShouldBeKept(self):
        self.assertEqual('\u0391', self.creator.filenameAndPathFilter(filterTerm='\u0391', filterValue='diacritics replacements shortnames', useIndividualFilters=False))

    def testSubstituteCustomPartnames(self):
        self.assertEqual('Fr_Farbig_+_Beschlage_Se_Li1_H', self.creator.filenameAndPathFilter(filterTerm='Front_Farbig_+_Beschlage_Seite_Links1_H', filterValue='diacritics replacements shortnames', useIndividualFilters=False))


class GetSimplifiedMultipleUnderscoresTests(unittest.TestCase):
    def testAnEmptyTextShouldBeKept(self):
        self.assertEqual('', FilenameAndPathFilter.getSimplifiedMultipleUnderscores(''))

    def testLowerCaseLettersAndSpacesShouldBeKept(self):
        self.assertEqual('the quick brown fox jumps over the lazy dog', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('the quick brown fox jumps over the lazy dog'))

    def testUpperCaseLettersAndSpacesShouldBeKept(self):
        self.assertEqual('THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG'))

    def testOneUnderScoreShouldBeKept(self):
        self.assertEqual('_', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('_'))

    def testOnlyOneUnderScoreOfTwoConsicutiveOnesShouldBeKept(self):
        self.assertEqual('_', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('__'))

    def testSingleUnderScoreBetweenLettersShouldBeKept(self):
        self.assertEqual('A_b_C', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('A_b_C'))

    def testAvoidDoubleUnderscores(self):
        self.assertEqual('Main_Side', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('Main___Side'))

    def testAvoidDoubleUnderscores2(self):
        self.assertEqual('Main_Side', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('Main_____Side'))

    def testSimulateOriginalBugUnderPython3_3(self):
        self.assertEqual('Main_Side', FilenameAndPathFilter.getSimplifiedMultipleUnderscores('Main_________Side'))


class DeleteLeadingAndTrailingUnderscores(unittest.TestCase):
    def testDeleteLeadingAndTrailingUnderscoresFront(self):
        self.assertEqual('Main Side', FilenameAndPathFilter.deleteLeadingAndTrailingUnderscores('_Main Side'))

    def testDeleteLeadingAndTrailingUnderscoresBack(self):
        self.assertEqual('Main Side', FilenameAndPathFilter.deleteLeadingAndTrailingUnderscores('Main Side_'))

    def testDeleteLeadingAndTrailingUnderscoresFrontAndBack(self):
        self.assertEqual('Main Side', FilenameAndPathFilter.deleteLeadingAndTrailingUnderscores('_Main Side_'))
