# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest


class FilenameCleaner:
    def __init__(self):
        self.map = {
            '\xe4': 'a',
            '\xf6': 'o',
            '\xfc': 'u',
            '\xc4': 'A',
            '\xd6': 'O',
            '\xdc': 'U',
            '\xdf': 's'
        }  # Umlauts should be replaced every time

        self.individualCharacterReplacements = {
            ' ': '_',
            '-': '',
            '/': '',
            '.': '_'
        }

        self.map.update(self.individualCharacterReplacements)

    # replace('__', '_' does NOT always work
    @staticmethod
    def getSimplifiedMultipleUnderscores(value):
        previousCharacter = None
        result = ''
        for character in value:
            if character != '_' or previousCharacter != '_':
                result += character
            previousCharacter = character
        return result

    @staticmethod
    def deleteLeadingAndTrailingUnderscores(value):
        return value.strip('_')

    def clean(self, filename):
        result = FilenameCleaner.getSimplifiedMultipleUnderscores(filename.translate(str.maketrans(self.map)))
        result = FilenameCleaner.deleteLeadingAndTrailingUnderscores(result)
        return result


class FilenameCleanerTests(unittest.TestCase):
    def testReplaceSpacesWithUnderscore(self):
        cleaner = FilenameCleaner()
        self.assertEqual('Bauteil_2_Main', cleaner.clean('Bauteil 2 Main'))

    def testRemoveDashes(self):
        cleaner = FilenameCleaner()
        self.assertEqual('Bauteil2Main', cleaner.clean('Bauteil-2-Main'))

    def testReplaceUmlautsWithConsonants(self):
        cleaner = FilenameCleaner()
        self.assertEqual('aouAOU', cleaner.clean('\xe4\xf6\xfc\xc4\xd6\xdc'))

    def testReplaceSpecialSWithNormalS(self):
        cleaner = FilenameCleaner()
        self.assertEqual('s', cleaner.clean('\xdf'))

    def testRemoveSlash(self):
        cleaner = FilenameCleaner()
        self.assertEqual('MainSide', cleaner.clean('/Main/Side/'))

    def testAvoidDoubleUnderscores(self):
        cleaner = FilenameCleaner()
        self.assertEqual('Main_Side', cleaner.clean('Main - Side'))

    def testGreekCapitalLetterAlphaShouldBeKept(self):
        cleaner = FilenameCleaner()
        self.assertEqual('\u0391', cleaner.clean('\u0391'))

    def testAvoidDoubleUnderscores2(self):
        cleaner = FilenameCleaner()
        self.assertEqual('Main_Side', cleaner.clean('Main - - - - Side'))


class GetSimplifiedMultipleUnderscoresTests(unittest.TestCase):
    def testAnEmptyTextShouldBeKept(self):
        self.assertEqual('', FilenameCleaner.getSimplifiedMultipleUnderscores(''))

    def testLowerCaseLettersAndSpacesShouldBeKept(self):
        self.assertEqual('the quick brown fox jumps over the lazy dog', FilenameCleaner.getSimplifiedMultipleUnderscores('the quick brown fox jumps over the lazy dog'))

    def testUpperCaseLettersAndSpacesShouldBeKept(self):
        self.assertEqual('THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG', FilenameCleaner.getSimplifiedMultipleUnderscores('THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG'))

    def testOneUnderScoreShouldBeKept(self):
        self.assertEqual('_', FilenameCleaner.getSimplifiedMultipleUnderscores('_'))

    def testOnlyOneUnderScoreOfTwoConsicutiveOnesShouldBeKept(self):
        self.assertEqual('_', FilenameCleaner.getSimplifiedMultipleUnderscores('__'))

    def testSingleUnderScoreBetweenLettersShouldBeKept(self):
        self.assertEqual('A_b_C', FilenameCleaner.getSimplifiedMultipleUnderscores('A_b_C'))

    def testSimulateOriginalBugUnderPython3_3(self):
        self.assertEqual('Main_Side', FilenameCleaner.getSimplifiedMultipleUnderscores('Main_________Side'))
