# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest

from interiorcad.backend.cmoduleadapters.jointsizingpath import getJointSizingPath, getJointPathClockWiseAndBounds
from interiorcad.backend.datatracker.polygonconverter import PolygonConverter
from interiorcad.backend.joint.calculator import JointCalculator
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.backend.workpiece.shape import Shape
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.geometry.size import Size


class JointSizing:
    def __init__(self, workpieceReader, minJoint, maxJoint):
        self.workpieceReader = workpieceReader
        self.minJoint = minJoint
        self.maxJoint = maxJoint

    # New function in interiorcad 2019 F7
    def getJointSizingShapeAndBounds(self):
        # TODO IT-240: do we want an option to always use the C++ code, even for rectangular? -- NG
        if self.workpieceReader.isRectangular():
            return Shape.createRectangularSizingFromBounds(self.getRectangularJointSizingBounds()), self.getRectangularJointSizingBounds()
        else:
            return self.getSlowJointSizingShapeAndBounds()

    # New function in interiorcad 2019 F7
    def getSlowJointSizingShapeAndBounds(self):
        uuid = self.workpieceReader.workpiece.uuid
        result = getJointPathClockWiseAndBounds(uuid, self.minJoint, self.maxJoint)
        if result and result[0] and result[1]:
            polyline = PolygonConverter().convertPolyline(result[0])
            bounds = Rectangle.createFromValues(*result[1])
            return Shape(polyline), bounds
        else:
            errorMessage = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 37, "Joint cut size could not be calculated! Workpiece: {workpiece}")
            raise RuntimeError(errorMessage.format(workpiece=self.workpieceReader.workpiece.name))

    # compatibility function for old code (up to interiorcad 2019 F6); do NOT use for new code, use getJointSizingShapeAndBounds instead
    def getCalculatedPolygonalJointSizingPolyline(self):
        uuid = self.workpieceReader.workpiece.uuid
        positions = getJointSizingPath(uuid, self.minJoint, self.maxJoint)
        if positions:
            return Polyline.createFromCornerVertexPositionTuples(positions)
        else:
            errorMessage = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 37, "Joint cut size could not be calculated! Workpiece: {workpiece}")
            raise RuntimeError(errorMessage.format(workpiece=self.workpieceReader.workpiece.name))

    # compatibility function for old code (up to interiorcad 2019 F6); do NOT use for new code, use getJointSizingShapeAndBounds instead
    def getJointSizingShape(self):
        if self.workpieceReader.isRectangular():
            return Shape.createRectangularSizingFromBounds(self.getRectangularJointSizingBounds())
        else:
            return Shape(self.getCalculatedPolygonalJointSizingPolyline())

    # compatibility function for old code (up to interiorcad 2019 F6); do NOT use for new code, use getJointSizingShapeAndBounds instead
    def getPolygonalJointSizingBounds(self):
        polyline = self.getCalculatedPolygonalJointSizingPolyline()

        maxXVertex = max(polyline.vertices, key=lambda vertex: vertex.position.x)
        maxYVertex = max(polyline.vertices, key=lambda vertex: vertex.position.y)
        minXVertex = min(polyline.vertices, key=lambda vertex: vertex.position.x)
        minYVertex = min(polyline.vertices, key=lambda vertex: vertex.position.y)

        return Rectangle(
            minXVertex.position.x,
            minYVertex.position.y,
            maxXVertex.position.x,
            maxYVertex.position.y
        )

    # compatibility function for old code (up to interiorcad 2019 F6); do NOT use for new code, use getJointSizingShapeAndBounds instead
    def getJointSizingBounds(self):
        if self.workpieceReader.isRectangular():
            return self.getRectangularJointSizingBounds()
        else:
            return self.getPolygonalJointSizingBounds()

    # compatibility function for old code (up to interiorcad 2019 F6); do NOT use for new code, use getJointSizingShapeAndBounds instead
    def getRectangularJointSizingBounds(self):
        additionForEdges = self.workpieceReader.getFinishedSizeAddition()
        jointCalculator = JointCalculator(self.minJoint, self.maxJoint)
        jointSizeAddition = jointCalculator.getJointAddition(additionForEdges)

        jointSize = self.workpieceReader.getSawnSizingBoundsFromWorkpiece()
        left = jointSize.left - jointSizeAddition.left
        right = jointSize.right + jointSizeAddition.right
        top = jointSize.top + jointSizeAddition.top
        bottom = jointSize.bottom - jointSizeAddition.bottom
        return Rectangle(left=left, bottom=bottom, right=right, top=top)

    # compatibility function for old code (up to interiorcad 2019 F6); do NOT use for new code, use getJointSizingShapeAndBounds instead
    def getJointSizingSize(self):
        bounds = self.getJointSizingBounds()
        return Size(bounds.getLength(), bounds.getWidth())


class RectangularJointSizingTests(unittest.TestCase):
    class MockWorkpieceReader:
        def __init__(self):
            self.sawnSize = Rectangle(0, 0, 200, 100)
            self.finishedSizeAddition = Rectangle(0, 0, 0, 0)
            self.sawnSizeSizingPoly = None

        def getFinishedSizeAddition(self):
            return self.finishedSizeAddition

        def getSawnSizingBoundsFromWorkpiece(self):
            return self.sawnSize

        def getSawnSizeSizing(self):
            return Shape(self.sawnSizeSizingPoly)

        def isRectangular(self):
            return self.sawnSizeSizingPoly.isRectangular()

        def isPolygonal(self):
            return self.sawnSizeSizingPoly.isPolygonal()

    def setUp(self):
        self.mockReader = self.MockWorkpieceReader()
        self.mockReader.sawnSizeSizingPoly = Polyline(
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(200, 100)),
            CornerVertex(Point(200, 0))
        )

    def testWorkpieceWithoutEdgesShouldNotUseJointingForBounds(self):
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedBounds = Rectangle(0, 0, 200, 100)
        self.assertEqual(expectedBounds, creator.getJointSizingBounds())

    def testWorkpieceWithEdgesShouldUseJointingForBounds(self):
        self.mockReader.finishedSizeAddition = Rectangle(1, 2, 3, 4)
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedBounds = Rectangle(-1, -2, 203, 104)
        self.assertEqual(expectedBounds, creator.getJointSizingBounds())

    def testWorkpieceWithThickEdgesShouldUseMaxJointForBounds(self):
        self.mockReader.finishedSizeAddition = Rectangle(6, 7, 8, 9)
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedBounds = Rectangle(-5, -5, 205, 105)
        self.assertEqual(expectedBounds, creator.getJointSizingBounds())

    def testWorkpieceWithSomeThickEdgesShouldUseJointCorrectlyForBounds(self):
        self.mockReader.finishedSizeAddition = Rectangle(1, 2, 8, 9)
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedBounds = Rectangle(-1, -2, 205, 105)
        self.assertEqual(expectedBounds, creator.getJointSizingBounds())

    def testWorkpieceWithEdgesSmallerThanMinJointShouldUseMinJointForBounds(self):
        self.mockReader.finishedSizeAddition = Rectangle(0.5, 0.5, 0.5, 0.5)
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedBounds = Rectangle(-1, -1, 201, 101)
        self.assertEqual(expectedBounds, creator.getJointSizingBounds())

    def testWorkpieceWithEdgesButWithoutMinMaxJointShouldNotAddSizingAdditionsForBounds(self):
        self.mockReader.finishedSizeAddition = Rectangle(1, 2, 3, 4)
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=0, maxJoint=0)
        expectedBounds = Rectangle(0, 0, 200, 100)
        self.assertEqual(expectedBounds, creator.getJointSizingBounds())

    def testWorkpieceWithoutEdgesShouldNotUseJointingForSizing(self):
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedSizing = Polyline(
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(200, 100)),
            CornerVertex(Point(200, 0)),
        )
        self.assertEqual(Shape(expectedSizing), creator.getJointSizingShape())

    def testWorkpieceWithEdgesShouldUseJointingForSizing(self):
        self.mockReader.finishedSizeAddition = Rectangle(1, 2, 3, 4)
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedSizing = Polyline(
            CornerVertex(Point(-1, -2)),
            CornerVertex(Point(-1, 104)),
            CornerVertex(Point(203, 104)),
            CornerVertex(Point(203, -2)),
        )
        self.assertEqual(Shape(expectedSizing), creator.getJointSizingShape())

    def testWorkpieceWithEdgesShouldUseMaxJointForSizing(self):
        self.mockReader.finishedSizeAddition = Rectangle(6, 7, 8, 9)
        creator = JointSizing(workpieceReader=self.mockReader, minJoint=1, maxJoint=5)
        expectedSizing = Polyline(
            CornerVertex(Point(-5, -5)),
            CornerVertex(Point(-5, 105)),
            CornerVertex(Point(205, 105)),
            CornerVertex(Point(205, -5)),
        )
        self.assertEqual(Shape(expectedSizing), creator.getJointSizingShape())
