# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest

from interiorcad.backend.ncexportmigrated.nclinesegment import NCLineSegment
from interiorcad.interface.geometry import isNearlyLesserOrEqual, isNearlyGreater
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point


def splitSectionsInXDistance(segments, maximumXDistance):
    def getRemainingXDistance(x):
        if isNearlyGreater(math.fabs(x - minX), maximumXDistance):
            return maximumXDistance - intermediaryStartPoint.x + minX
        elif isNearlyGreater(math.fabs(x - maxX), maximumXDistance):
            return maximumXDistance - intermediaryStartPoint.x + maxX

    if isNearlyLesserOrEqual(maximumXDistance, 0.0):
        maximumXDistance = float('+inf')

    section = []
    minX = maxX = None

    for segment in segments:
        assert segment.isLineSegment(), "Only Line Segments are supported yet"  # TODO IT-254 implement Radius Segment
        startPoint = segment.getStartVertex().position
        endPoint = segment.getEndVertex().position
        directionVector = (endPoint - startPoint).normalized()

        intermediaryStartPoint = startPoint
        minX = intermediaryStartPoint.x if minX is None else min(minX, intermediaryStartPoint.x)
        maxX = intermediaryStartPoint.x if maxX is None else max(maxX, intermediaryStartPoint.x)

        while getRemainingXDistance(endPoint.x):
            remainingXDistance = getRemainingXDistance(endPoint.x)
            magnitude = math.fabs(remainingXDistance / directionVector.dot(Point(1.0, 0.0)))
            assert magnitude, 'Magnitude should not be zero!'

            intermediaryEndPoint = intermediaryStartPoint + directionVector * magnitude
            section.append(NCLineSegment(CornerVertex(intermediaryStartPoint), CornerVertex(intermediaryEndPoint)))

            yield section
            section = []

            intermediaryStartPoint = intermediaryEndPoint
            minX = maxX = intermediaryStartPoint.x

        if intermediaryStartPoint != endPoint:
            section.append(NCLineSegment(CornerVertex(intermediaryStartPoint), CornerVertex(endPoint)))
            minX = min(minX, endPoint.x)
            maxX = max(maxX, endPoint.x)

    if section:
        yield section


class SimpleTests(unittest.TestCase):
    MAXIMUM_SEGMENT_X_DISTANCE = 105

    def testXShort(self):
        segments = NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(100, 10))),
        sections = tuple(splitSectionsInXDistance(segments, self.MAXIMUM_SEGMENT_X_DISTANCE))

        expectedSections = tuple((
            [NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(100, 10))),],
        ))

        self.assertEqual(expectedSections, sections)

    def testYShort(self):
        segments = NCLineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, 100))),
        sections = tuple(splitSectionsInXDistance(segments, self.MAXIMUM_SEGMENT_X_DISTANCE))

        expectedSections = tuple((
            [NCLineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, 100))),],
        ))

        self.assertEqual(expectedSections, sections)

    def testXLongShouldBeSplit(self):
        segments = NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(500, 10))),
        sections = tuple(splitSectionsInXDistance(segments, self.MAXIMUM_SEGMENT_X_DISTANCE))

        expectedSections = tuple((
            [NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(105, 10))),],
            [NCLineSegment(CornerVertex(Point(105, 10)), CornerVertex(Point(210, 10))),],
            [NCLineSegment(CornerVertex(Point(210, 10)), CornerVertex(Point(315, 10))),],
            [NCLineSegment(CornerVertex(Point(315, 10)), CornerVertex(Point(420, 10))),],
            [NCLineSegment(CornerVertex(Point(420, 10)), CornerVertex(Point(500, 10))),],
        ))

        self.assertEqual(expectedSections, sections)

    def testYLongShouldNotBeSplit(self):
        segments = NCLineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, 500))),
        sections = tuple(splitSectionsInXDistance(segments, self.MAXIMUM_SEGMENT_X_DISTANCE))

        expectedSections = tuple((
            [NCLineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, 500))),],
        ))

        self.assertEqual(expectedSections, sections)

    def testXTwoShortSegmentsShouldBeSplit(self):
        segments = (
            NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(100, 10))),
            NCLineSegment(CornerVertex(Point(100, 10)), CornerVertex(Point(200, 10))),
        )

        sections = tuple(splitSectionsInXDistance(segments, self.MAXIMUM_SEGMENT_X_DISTANCE))

        expectedSections = tuple((
            [
                NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(100, 10))),
                NCLineSegment(CornerVertex(Point(100, 10)), CornerVertex(Point(105, 10))),
            ],
            [
                NCLineSegment(CornerVertex(Point(105, 10)), CornerVertex(Point(200, 10))),
            ],
        ))

        self.assertEqual(expectedSections, sections)

    def testXOneLargeSegmentAndBackShouldBeSplit(self):
        segments = (
            NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(200, 10))),
            NCLineSegment(CornerVertex(Point(200, 10)), CornerVertex(Point(0, 10))),
        )
        sections = tuple(splitSectionsInXDistance(segments, self.MAXIMUM_SEGMENT_X_DISTANCE))

        expectedSections = tuple((
            [NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(105, 10))),],
            [
                NCLineSegment(CornerVertex(Point(105, 10)), CornerVertex(Point(200, 10))),
                NCLineSegment(CornerVertex(Point(200, 10)), CornerVertex(Point(95, 10))),
            ],
            [NCLineSegment(CornerVertex(Point(95, 10)), CornerVertex(Point(0, 10))),],
        ))

        self.assertEqual(expectedSections, sections)
