# -*- coding: utf-8 -*-

# Copyright 2015-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest.mock
import copy

from interiorcad.backend.ncexportmigrated.nclinesegment import NCLineSegment
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.simplelinesegment import SimpleLineSegment
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.backend.ncexportmigrated.ncradiussegment import NCRadiusSegment


def getStartEndLine(machining):
    return (
        NCLineSegment(CornerVertex(machining.getStartPoint()), CornerVertex(machining.getEndPoint())),
    )


# Old name, for compatibility (for dados)
def getMiddleLine(dado):
    return getStartEndLine(dado)


def __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart, securityDistanceAtEnd):
    middleLine = dado.getLineSegment()

    if not isNearlyEqual(securityDistanceAtStart, 0):
        middleLine.increaseAtStart(securityDistanceAtStart)
    if not isNearlyEqual(securityDistanceAtEnd, 0):
        middleLine.increaseAtEnd(securityDistanceAtEnd)

    rightLine = copy.copy(middleLine)
    rightLine.moveRight(dado.width / 2)

    leftLine = copy.copy(middleLine)
    leftLine.moveLeft(dado.width / 2)

    return leftLine, rightLine


def getPerimeterCW(dado):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(leftLine.getStartPoint())),
    )


def getPerimeterCCW(dado):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(rightLine.getStartPoint())),
    )


def getPerimeterInsetAtEndCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(leftLine.getStartPoint())),
    )


def getPerimeterInsetAtEndCCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(rightLine.getStartPoint())),
    )


def getPerimeterInsetAtStartCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=securityDistance)
    return (
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(leftLine.getStartPoint())),
    )


def getPerimeterInsetAtStartCCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=securityDistance)
    return (
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(rightLine.getStartPoint())),
    )


def getPerimeterThroughCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=securityDistance)
    return (
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(leftLine.getStartPoint())),
    )


def getPerimeterThroughCCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=securityDistance)
    return (
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(rightLine.getStartPoint())),
    )


def getDadoContourBothInsetCW(dado):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(leftLine.getMiddlePoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getMiddlePoint())),
    )


def getDadoContourBothInsetCCW(dado):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(rightLine.getMiddlePoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getMiddlePoint())),
    )


def getDadoContourThroughCW(dado, securityDistance, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=securityDistance)

    leftLineVector = (leftLine.getEndPoint() - leftLine.getStartPoint())
    leftLineVector.normalize()

    rightLineVector = (rightLine.getEndPoint() - rightLine.getStartPoint())
    rightLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint() + leftLine.getDirectionVector() * routerRadius)),
        NCLineSegment(CornerVertex(leftLine.getEndPoint() + leftLine.getDirectionVector() * routerRadius), CornerVertex(rightLine.getEndPoint() + rightLine.getDirectionVector() * routerRadius)),
        NCLineSegment(CornerVertex(rightLine.getEndPoint() + rightLine.getDirectionVector() * routerRadius), CornerVertex(rightLine.getStartPoint())),
    )


def getDadoContourThroughCCW(dado, securityDistance, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=securityDistance)

    leftLineVector = (leftLine.getEndPoint() - leftLine.getStartPoint())
    leftLineVector.normalize()

    rightLineVector = (rightLine.getEndPoint() - rightLine.getStartPoint())
    rightLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint() + rightLine.getDirectionVector() * routerRadius)),
        NCLineSegment(CornerVertex(rightLine.getEndPoint() + rightLine.getDirectionVector() * routerRadius), CornerVertex(leftLine.getEndPoint() + leftLine.getDirectionVector() * routerRadius)),
        NCLineSegment(CornerVertex(leftLine.getEndPoint() + leftLine.getDirectionVector() * routerRadius), CornerVertex(leftLine.getStartPoint())),
    )


def getDadoContourInsetAtEndCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint())),
    )


def getDadoContourInsetAtEndCCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=0)
    return (
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint())),
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(leftLine.getEndPoint())),
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint())),
    )


def getDadoContourInsetAtStartCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=securityDistance)
    return (
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint())),
    )


def getDadoContourInsetAtStartCCW(dado, securityDistance):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=securityDistance)
    return (
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint())),
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(rightLine.getStartPoint())),
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint())),
    )


def getDadoRoundedContourBothInsetCW(dado, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=0)

    endLineVector = (rightLine.getEndPoint() - leftLine.getEndPoint())
    endLineVector.normalize()

    startLineVector = (leftLine.getStartPoint() - rightLine.getStartPoint())
    startLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(leftLine.getMiddlePoint()), CornerVertex(leftLine.getEndPoint() - leftLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getEndPoint(), - leftLine.getDirectionVector(), endLineVector, routerRadius),
        NCLineSegment(CornerVertex(leftLine.getEndPoint() + endLineVector * routerRadius), CornerVertex(rightLine.getEndPoint() - endLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getEndPoint(), - endLineVector, - rightLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(rightLine.getEndPoint() - leftLine.getDirectionVector() * routerRadius), CornerVertex(rightLine.getStartPoint() + leftLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getStartPoint(), leftLine.getDirectionVector(), startLineVector, routerRadius),
        NCLineSegment(CornerVertex(rightLine.getStartPoint() + startLineVector * routerRadius), CornerVertex(leftLine.getStartPoint() - startLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getStartPoint(), - startLineVector, leftLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(leftLine.getStartPoint() + leftLine.getDirectionVector() * routerRadius), CornerVertex(leftLine.getMiddlePoint())),
    )


def getDadoRoundedContourBothInsetCCW(dado, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=0)

    endLineVector = (leftLine.getEndPoint() - rightLine.getEndPoint())
    endLineVector.normalize()

    startLineVector = (rightLine.getStartPoint() - leftLine.getStartPoint())
    startLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(rightLine.getMiddlePoint()), CornerVertex(rightLine.getEndPoint() - rightLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getEndPoint(), - rightLine.getDirectionVector(), endLineVector, routerRadius),
        NCLineSegment(CornerVertex(rightLine.getEndPoint() + endLineVector * routerRadius), CornerVertex(leftLine.getEndPoint() - endLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getEndPoint(), - endLineVector, - leftLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(leftLine.getEndPoint() - leftLine.getDirectionVector() * routerRadius), CornerVertex(leftLine.getStartPoint() + leftLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getStartPoint(), leftLine.getDirectionVector(), startLineVector, routerRadius),
        NCLineSegment(CornerVertex(leftLine.getStartPoint() + startLineVector * routerRadius), CornerVertex(rightLine.getStartPoint() - startLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getStartPoint(), - startLineVector, rightLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(rightLine.getStartPoint() + rightLine.getDirectionVector() * routerRadius), CornerVertex(rightLine.getMiddlePoint())),
    )


def getDadoRoundedContourInsetAtEndCW(dado, securityDistance, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=0)

    endLineVector = (rightLine.getEndPoint() - leftLine.getEndPoint())
    endLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(leftLine.getStartPoint()), CornerVertex(leftLine.getEndPoint() - leftLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getEndPoint(), - leftLine.getDirectionVector(), endLineVector, routerRadius),
        NCLineSegment(CornerVertex(leftLine.getEndPoint() + endLineVector * routerRadius), CornerVertex(rightLine.getEndPoint() - endLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getEndPoint(), - endLineVector, - rightLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(rightLine.getEndPoint() - rightLine.getDirectionVector() * routerRadius), CornerVertex(rightLine.getStartPoint())),
    )


def getDadoRoundedContourInsetAtEndCCW(dado, securityDistance, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=securityDistance, securityDistanceAtEnd=0)

    endLineVector = (leftLine.getEndPoint() - rightLine.getEndPoint())
    endLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(rightLine.getStartPoint()), CornerVertex(rightLine.getEndPoint() - rightLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getEndPoint(), - rightLine.getDirectionVector(), endLineVector, routerRadius),
        NCLineSegment(CornerVertex(rightLine.getEndPoint() + endLineVector * routerRadius), CornerVertex(leftLine.getEndPoint() - endLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getEndPoint(), - endLineVector, - leftLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(leftLine.getEndPoint() - leftLine.getDirectionVector() * routerRadius), CornerVertex(leftLine.getStartPoint())),
    )


def getDadoRoundedContourInsetAtStartCW(dado, securityDistance, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=securityDistance)

    startLineVector = (leftLine.getStartPoint() - rightLine.getStartPoint())
    startLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(rightLine.getEndPoint()), CornerVertex(rightLine.getStartPoint() + rightLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getStartPoint(), leftLine.getDirectionVector(), startLineVector, routerRadius),
        NCLineSegment(CornerVertex(rightLine.getStartPoint() + startLineVector * routerRadius), CornerVertex(leftLine.getStartPoint() - startLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getStartPoint(), - startLineVector, leftLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(leftLine.getStartPoint() + leftLine.getDirectionVector() * routerRadius), CornerVertex(leftLine.getEndPoint())),
    )


def getDadoRoundedContourInsetAtStartCCW(dado, securityDistance, routerRadius):
    leftLine, rightLine = __getRightAndLeftLinesWithSecurityDistance(dado, securityDistanceAtStart=0, securityDistanceAtEnd=securityDistance)

    startLineVector = (rightLine.getStartPoint() - leftLine.getStartPoint())
    startLineVector.normalize()

    return (
        NCLineSegment(CornerVertex(leftLine.getEndPoint()), CornerVertex(leftLine.getStartPoint() + rightLine.getDirectionVector() * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(leftLine.getStartPoint(), leftLine.getDirectionVector(), startLineVector, routerRadius),
        NCLineSegment(CornerVertex(leftLine.getStartPoint() + startLineVector * routerRadius), CornerVertex(rightLine.getStartPoint() - startLineVector * routerRadius)),
        NCRadiusSegment.createRoundedSegmentAtCorner(rightLine.getStartPoint(), - startLineVector, rightLine.getDirectionVector(), routerRadius),
        NCLineSegment(CornerVertex(rightLine.getStartPoint() + leftLine.getDirectionVector() * routerRadius), CornerVertex(rightLine.getEndPoint())),
    )


def getDadoPerimeterCW(dado, securityDistance):
    if dado.isInsetBoth():
        return getPerimeterCW(dado)
    elif dado.isInsetStart():
        return getPerimeterInsetAtStartCW(dado, securityDistance)
    elif dado.isInsetEnd():
        return getPerimeterInsetAtEndCW(dado, securityDistance)
    else:
        return getPerimeterThroughCW(dado, securityDistance)


def getDadoPerimeterCCW(dado, securityDistance):
    if dado.isInsetBoth():
        return getPerimeterCCW(dado)
    elif dado.isInsetStart():
        return getPerimeterInsetAtStartCCW(dado, securityDistance)
    elif dado.isInsetEnd():
        return getPerimeterInsetAtEndCCW(dado, securityDistance)
    else:
        return getPerimeterThroughCCW(dado, securityDistance)


def getDadoContourCW(dado, securityDistance, routerRadius):
    if dado.isInsetBoth():
        return getDadoContourBothInsetCW(dado)
    elif dado.isInsetStart():
        return getDadoContourInsetAtStartCW(dado, securityDistance)
    elif dado.isInsetEnd():
        return getDadoContourInsetAtEndCW(dado, securityDistance)
    else:
        return getDadoContourThroughCW(dado, securityDistance, routerRadius)


def getDadoContourCCW(dado, securityDistance, routerRadius):
    if dado.isInsetBoth():
        return getDadoContourBothInsetCCW(dado)
    elif dado.isInsetStart():
        return getDadoContourInsetAtStartCCW(dado, securityDistance)
    elif dado.isInsetEnd():
        return getDadoContourInsetAtEndCCW(dado, securityDistance)
    else:
        return getDadoContourThroughCCW(dado, securityDistance, routerRadius)


def getDadoRoundedContourCW(dado, securityDistance, routerRadius):
    if isNearlyEqual(routerRadius, 0.0):
        return getDadoContourCW(dado, securityDistance, routerRadius)
    elif dado.isInsetBoth():
        return getDadoRoundedContourBothInsetCW(dado, routerRadius)
    elif dado.isInsetStart():
        return getDadoRoundedContourInsetAtStartCW(dado, securityDistance, routerRadius)
    elif dado.isInsetEnd():
        return getDadoRoundedContourInsetAtEndCW(dado, securityDistance, routerRadius)
    else:
        return getDadoContourThroughCW(dado, securityDistance, routerRadius)


def getDadoRoundedContourCCW(dado, securityDistance, routerRadius):
    if isNearlyEqual(routerRadius, 0.0):
        return getDadoContourCCW(dado, securityDistance, routerRadius)
    elif dado.isInsetBoth():
        return getDadoRoundedContourBothInsetCCW(dado, routerRadius)
    elif dado.isInsetStart():
        return getDadoRoundedContourInsetAtStartCCW(dado, securityDistance, routerRadius)
    elif dado.isInsetEnd():
        return getDadoRoundedContourInsetAtEndCCW(dado, securityDistance, routerRadius)
    else:
        return getDadoContourThroughCCW(dado, securityDistance, routerRadius)


class MockDado:
    def __init__(self, start, end, width):
        self.start = start
        self.end = end
        self.width = width

    def getStartPoint(self):
        return self.start

    def getEndPoint(self):
        return self.end

    def getLineSegment(self):
        return SimpleLineSegment(self.getStartPoint(), self.getEndPoint())


class SimpleTests(unittest.TestCase):
    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testMiddleLineX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(0, 0)), endPt=CornerVertex(Point(100, 0))),
        )
        self.assertEqual(expected, getMiddleLine(self.dadoX))

    def testMiddleLineY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(0, 0)), endPt=CornerVertex(Point(0, 100))),
        )
        self.assertEqual(expected, getMiddleLine(self.dadoY))

    def testPerimeterCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(0, 5)), endPt=CornerVertex(Point(100, 5))),
            NCLineSegment(startPt=CornerVertex(Point(100, 5)), endPt=CornerVertex(Point(100, -5))),
            NCLineSegment(startPt=CornerVertex(Point(100, -5)), endPt=CornerVertex(Point(0, -5))),
            NCLineSegment(startPt=CornerVertex(Point(0, -5)), endPt=CornerVertex(Point(0, 5))),
        )
        self.assertEqual(expected, getPerimeterCW(self.dadoX))

    def testPerimeterCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(0, -5)), endPt=CornerVertex(Point(100, -5))),
            NCLineSegment(startPt=CornerVertex(Point(100, -5)), endPt=CornerVertex(Point(100, 5))),
            NCLineSegment(startPt=CornerVertex(Point(100, 5)), endPt=CornerVertex(Point(0, 5))),
            NCLineSegment(startPt=CornerVertex(Point(0, 5)), endPt=CornerVertex(Point(0, -5))),
        )
        self.assertEqual(expected, getPerimeterCCW(self.dadoX))

    def testPerimeterCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, 0)), endPt=CornerVertex(Point(-5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 100)), endPt=CornerVertex(Point(5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(5, 100)), endPt=CornerVertex(Point(5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(5, 0)), endPt=CornerVertex(Point(-5, 0))),
        )
        self.assertEqual(expected, getPerimeterCW(self.dadoY))

    def testPerimeterCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, 0)), endPt=CornerVertex(Point(5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(5, 100)), endPt=CornerVertex(Point(-5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 100)), endPt=CornerVertex(Point(-5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 0)), endPt=CornerVertex(Point(5, 0))),
        )
        self.assertEqual(expected, getPerimeterCCW(self.dadoY))


class DadoContourBothInsetTests(unittest.TestCase):
    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(50, 5)), endPt=CornerVertex(Point(100, 5))),
            NCLineSegment(startPt=CornerVertex(Point(100, 5)), endPt=CornerVertex(Point(100, -5))),
            NCLineSegment(startPt=CornerVertex(Point(100, -5)), endPt=CornerVertex(Point(0, -5))),
            NCLineSegment(startPt=CornerVertex(Point(0, -5)), endPt=CornerVertex(Point(0, 5))),
            NCLineSegment(startPt=CornerVertex(Point(0, 5)), endPt=CornerVertex(Point(50, 5))),
        )
        self.assertEqual(expected, getDadoContourBothInsetCW(self.dadoX))

    def testCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(50, -5)), endPt=CornerVertex(Point(100, -5))),
            NCLineSegment(startPt=CornerVertex(Point(100, -5)), endPt=CornerVertex(Point(100, 5))),
            NCLineSegment(startPt=CornerVertex(Point(100, 5)), endPt=CornerVertex(Point(0, 5))),
            NCLineSegment(startPt=CornerVertex(Point(0, 5)), endPt=CornerVertex(Point(0, -5))),
            NCLineSegment(startPt=CornerVertex(Point(0, -5)), endPt=CornerVertex(Point(50, -5))),
        )
        self.assertEqual(expected, getDadoContourBothInsetCCW(self.dadoX))

    def testCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, 50)), endPt=CornerVertex(Point(-5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 100)), endPt=CornerVertex(Point(5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(5, 100)), endPt=CornerVertex(Point(5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(5, 0)), endPt=CornerVertex(Point(-5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 0)), endPt=CornerVertex(Point(-5, 50))),
        )
        self.assertEqual(expected, getDadoContourBothInsetCW(self.dadoY))

    def testCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, 50)), endPt=CornerVertex(Point(5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(5, 100)), endPt=CornerVertex(Point(-5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 100)), endPt=CornerVertex(Point(-5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 0)), endPt=CornerVertex(Point(5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(5, 0)), endPt=CornerVertex(Point(5, 50))),
        )
        self.assertEqual(expected, getDadoContourBothInsetCCW(self.dadoY))


class DadoRoundedContourBothInsetTests(unittest.TestCase):
    ROUTER_RADIUS = 4

    BIG = 98.82842712474618
    SMALL = 3.82842712474619

    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(50, 5)), endPt=CornerVertex(Point(96, 5))),
            NCRadiusSegment(startPt=CornerVertex(Point(96, 5)), radiusPt=RadiusVertex(position=Point(self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(100, 1))),
            NCLineSegment(startPt=CornerVertex(Point(100, 1)), endPt=CornerVertex(Point(100, -1))),
            NCRadiusSegment(startPt=CornerVertex(Point(100, -1)), radiusPt=RadiusVertex(position=Point(self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(96, -5))),
            NCLineSegment(startPt=CornerVertex(Point(96, -5)), endPt=CornerVertex(Point(4, -5))),
            NCRadiusSegment(startPt=CornerVertex(Point(4, -5)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(0, -1))),
            NCLineSegment(startPt=CornerVertex(Point(0, -1)), endPt=CornerVertex(Point(0, 1))),
            NCRadiusSegment(startPt=CornerVertex(Point(0, 1)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(4, 5))),
            NCLineSegment(startPt=CornerVertex(Point(4, 5)), endPt=CornerVertex(Point(50, 5))),
        )
        self.assertEqual(expected, getDadoRoundedContourBothInsetCW(self.dadoX, self.ROUTER_RADIUS))

    def testCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(50, -5)), endPt=CornerVertex(Point(96, -5))),
            NCRadiusSegment(startPt=CornerVertex(Point(96, -5)), radiusPt=RadiusVertex(position=Point(self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(100, -1))),
            NCLineSegment(startPt=CornerVertex(Point(100, -1)), endPt=CornerVertex(Point(100, 1))),
            NCRadiusSegment(startPt=CornerVertex(Point(100, 1)), radiusPt=RadiusVertex(position=Point(self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(96, 5))),
            NCLineSegment(startPt=CornerVertex(Point(96, 5)), endPt=CornerVertex(Point(4, 5))),
            NCRadiusSegment(startPt=CornerVertex(Point(4, 5)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(0, 1))),
            NCLineSegment(startPt=CornerVertex(Point(0, 1)), endPt=CornerVertex(Point(0, -1))),
            NCRadiusSegment(startPt=CornerVertex(Point(0, -1)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(4, -5))),
            NCLineSegment(startPt=CornerVertex(Point(4, -5)), endPt=CornerVertex(Point(50, -5))),
        )
        self.assertEqual(expected, getDadoRoundedContourBothInsetCCW(self.dadoX, self.ROUTER_RADIUS))

    def testCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, 50)), endPt=CornerVertex(Point(-5, 96))),
            NCRadiusSegment(startPt=CornerVertex(Point(-5, 96)), radiusPt=RadiusVertex(position=Point(-self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(-1, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-1, 100)), endPt=CornerVertex(Point(1, 100))),
            NCRadiusSegment(startPt=CornerVertex(Point(1, 100)), radiusPt=RadiusVertex(position=Point(self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(5, 96))),
            NCLineSegment(startPt=CornerVertex(Point(5, 96)), endPt=CornerVertex(Point(5, 4))),
            NCRadiusSegment(startPt=CornerVertex(Point(5, 4)), radiusPt=RadiusVertex(position=Point(self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(1, 0))),
            NCLineSegment(startPt=CornerVertex(Point(1, 0)), endPt=CornerVertex(Point(-1, 0))),
            NCRadiusSegment(startPt=CornerVertex(Point(-1, 0)), radiusPt=RadiusVertex(position=Point(-self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(-5, 4))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 4)), endPt=CornerVertex(Point(-5, 50))),
        )
        self.assertEqual(expected, getDadoRoundedContourBothInsetCW(self.dadoY, self.ROUTER_RADIUS))

    def testCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, 50)), endPt=CornerVertex(Point(5, 96))),
            NCRadiusSegment(startPt=CornerVertex(Point(5, 96)), radiusPt=RadiusVertex(position=Point(self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(1, 100))),
            NCLineSegment(startPt=CornerVertex(Point(1, 100)), endPt=CornerVertex(Point(-1, 100))),
            NCRadiusSegment(startPt=CornerVertex(Point(-1, 100)), radiusPt=RadiusVertex(position=Point(-self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(-5, 96))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 96)), endPt=CornerVertex(Point(-5, 4))),
            NCRadiusSegment(startPt=CornerVertex(Point(-5, 4)), radiusPt=RadiusVertex(position=Point(-self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(-1, 0))),
            NCLineSegment(startPt=CornerVertex(Point(-1, 0)), endPt=CornerVertex(Point(1, 0))),
            NCRadiusSegment(startPt=CornerVertex(Point(1, 0)), radiusPt=RadiusVertex(position=Point(self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(5, 4))),
            NCLineSegment(startPt=CornerVertex(Point(5, 4)), endPt=CornerVertex(Point(5, 50))),
        )
        self.assertEqual(expected, getDadoRoundedContourBothInsetCCW(self.dadoY, self.ROUTER_RADIUS))


class DadoContourThroughTests(unittest.TestCase):
    SECURITY_DISTANCE = 20
    ROUTER_RADIUS = 4

    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-20, 5)), endPt=CornerVertex(Point(124, 5))),
            NCLineSegment(startPt=CornerVertex(Point(124, 5)), endPt=CornerVertex(Point(124, -5))),
            NCLineSegment(startPt=CornerVertex(Point(124, -5)), endPt=CornerVertex(Point(-20, -5))),
        )
        self.assertEqual(expected, getDadoContourThroughCW(self.dadoX, DadoContourThroughTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-20, -5)), endPt=CornerVertex(Point(124, -5))),
            NCLineSegment(startPt=CornerVertex(Point(124, -5)), endPt=CornerVertex(Point(124, 5))),
            NCLineSegment(startPt=CornerVertex(Point(124, 5)), endPt=CornerVertex(Point(-20, 5))),
        )
        self.assertEqual(expected, getDadoContourThroughCCW(self.dadoX, DadoContourThroughTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, -20)), endPt=CornerVertex(Point(-5, 124))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 124)), endPt=CornerVertex(Point(5, 124))),
            NCLineSegment(startPt=CornerVertex(Point(5, 124)), endPt=CornerVertex(Point(5, -20))),
        )
        self.assertEqual(expected, getDadoContourThroughCW(self.dadoY, DadoContourThroughTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, -20)), endPt=CornerVertex(Point(5, 124))),
            NCLineSegment(startPt=CornerVertex(Point(5, 124)), endPt=CornerVertex(Point(-5, 124))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 124)), endPt=CornerVertex(Point(-5, -20))),
        )
        self.assertEqual(expected, getDadoContourThroughCCW(self.dadoY, DadoContourThroughTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))


class DadoContourInsetAtEndTests(unittest.TestCase):
    SECURITY_DISTANCE = 20

    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-20, 5)), endPt=CornerVertex(Point(100, 5))),
            NCLineSegment(startPt=CornerVertex(Point(100, 5)), endPt=CornerVertex(Point(100, -5))),
            NCLineSegment(startPt=CornerVertex(Point(100, -5)), endPt=CornerVertex(Point(-20, -5))),
        )
        self.assertEqual(expected, getDadoContourInsetAtEndCW(self.dadoX, DadoContourInsetAtEndTests.SECURITY_DISTANCE))

    def testCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-20, -5)), endPt=CornerVertex(Point(100, -5))),
            NCLineSegment(startPt=CornerVertex(Point(100, -5)), endPt=CornerVertex(Point(100, 5))),
            NCLineSegment(startPt=CornerVertex(Point(100, 5)), endPt=CornerVertex(Point(-20, 5))),
        )
        self.assertEqual(expected, getDadoContourInsetAtEndCCW(self.dadoX, DadoContourInsetAtEndTests.SECURITY_DISTANCE))

    def testCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, -20)), endPt=CornerVertex(Point(-5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 100)), endPt=CornerVertex(Point(5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(5, 100)), endPt=CornerVertex(Point(5, -20))),
        )
        self.assertEqual(expected, getDadoContourInsetAtEndCW(self.dadoY, DadoContourInsetAtEndTests.SECURITY_DISTANCE))

    def testCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, -20)), endPt=CornerVertex(Point(5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(5, 100)), endPt=CornerVertex(Point(-5, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 100)), endPt=CornerVertex(Point(-5, -20))),
        )
        self.assertEqual(expected, getDadoContourInsetAtEndCCW(self.dadoY, DadoContourInsetAtEndTests.SECURITY_DISTANCE))


class DadoRoundedContourInsetAtEndTests(unittest.TestCase):
    SECURITY_DISTANCE = 20
    ROUTER_RADIUS = 4

    BIG = 98.82842712474618
    SMALL = 3.82842712474619

    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-20, 5)), endPt=CornerVertex(Point(96, 5))),
            NCRadiusSegment(startPt=CornerVertex(Point(96, 5)), radiusPt=RadiusVertex(position=Point(self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(100, 1))),
            NCLineSegment(startPt=CornerVertex(Point(100, 1)), endPt=CornerVertex(Point(100, -1))),
            NCRadiusSegment(startPt=CornerVertex(Point(100, -1)), radiusPt=RadiusVertex(position=Point(self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(96, -5))),
            NCLineSegment(startPt=CornerVertex(Point(96, -5)), endPt=CornerVertex(Point(-20, -5))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtEndCW(self.dadoX, DadoContourInsetAtEndTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-20, -5)), endPt=CornerVertex(Point(96, -5))),
            NCRadiusSegment(startPt=CornerVertex(Point(96, -5)), radiusPt=RadiusVertex(position=Point(self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(100, -1))),
            NCLineSegment(startPt=CornerVertex(Point(100, -1)), endPt=CornerVertex(Point(100, 1))),
            NCRadiusSegment(startPt=CornerVertex(Point(100, 1)), radiusPt=RadiusVertex(position=Point(self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(96, 5))),
            NCLineSegment(startPt=CornerVertex(Point(96, 5)), endPt=CornerVertex(Point(-20, 5))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtEndCCW(self.dadoX, DadoContourInsetAtEndTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, -20)), endPt=CornerVertex(Point(-5, 96))),
            NCRadiusSegment(startPt=CornerVertex(Point(-5, 96)), radiusPt=RadiusVertex(position=Point(-self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(-1, 100))),
            NCLineSegment(startPt=CornerVertex(Point(-1, 100)), endPt=CornerVertex(Point(1, 100))),
            NCRadiusSegment(startPt=CornerVertex(Point(1, 100)), radiusPt=RadiusVertex(position=Point(self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(5, 96))),
            NCLineSegment(startPt=CornerVertex(Point(5, 96)), endPt=CornerVertex(Point(5, -20))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtEndCW(self.dadoY, DadoContourInsetAtEndTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, -20)), endPt=CornerVertex(Point(5, 96))),
            NCRadiusSegment(startPt=CornerVertex(Point(5, 96)), radiusPt=RadiusVertex(position=Point(self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(1, 100))),
            NCLineSegment(startPt=CornerVertex(Point(1, 100)), endPt=CornerVertex(Point(-1, 100))),
            NCRadiusSegment(startPt=CornerVertex(Point(-1, 100)), radiusPt=RadiusVertex(position=Point(-self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(-5, 96))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 96)), endPt=CornerVertex(Point(-5, -20))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtEndCCW(self.dadoY, DadoContourInsetAtEndTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))


class DadoContourInsetAtStartTests(unittest.TestCase):
    SECURITY_DISTANCE = 20

    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(120, -5)), endPt=CornerVertex(Point(0, -5))),
            NCLineSegment(startPt=CornerVertex(Point(0, -5)), endPt=CornerVertex(Point(0, 5))),
            NCLineSegment(startPt=CornerVertex(Point(0, 5)), endPt=CornerVertex(Point(120, 5))),
        )
        self.assertEqual(expected, getDadoContourInsetAtStartCW(self.dadoX, DadoContourInsetAtStartTests.SECURITY_DISTANCE))

    def testCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(120, 5)), endPt=CornerVertex(Point(0, 5))),
            NCLineSegment(startPt=CornerVertex(Point(0, 5)), endPt=CornerVertex(Point(0, -5))),
            NCLineSegment(startPt=CornerVertex(Point(0, -5)), endPt=CornerVertex(Point(120, -5))),
        )
        self.assertEqual(expected, getDadoContourInsetAtStartCCW(self.dadoX, DadoContourInsetAtStartTests.SECURITY_DISTANCE))

    def testCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, 120)), endPt=CornerVertex(Point(5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(5, 0)), endPt=CornerVertex(Point(-5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 0)), endPt=CornerVertex(Point(-5, 120))),
        )
        self.assertEqual(expected, getDadoContourInsetAtStartCW(self.dadoY, DadoContourInsetAtStartTests.SECURITY_DISTANCE))

    def testCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, 120)), endPt=CornerVertex(Point(-5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 0)), endPt=CornerVertex(Point(5, 0))),
            NCLineSegment(startPt=CornerVertex(Point(5, 0)), endPt=CornerVertex(Point(5, 120))),
        )
        self.assertEqual(expected, getDadoContourInsetAtStartCCW(self.dadoY, DadoContourInsetAtStartTests.SECURITY_DISTANCE))


class DadoRoundedContourInsetAtStartTests(unittest.TestCase):
    SECURITY_DISTANCE = 20
    ROUTER_RADIUS = 4

    BIG = 98.82842712474618
    SMALL = 3.82842712474619

    def setUp(self):
        self.dadoX = MockDado(start=Point(0, 0), end=Point(100, 0), width=10)
        self.dadoY = MockDado(start=Point(0, 0), end=Point(0, 100), width=10)

    def testCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(120, -5)), endPt=CornerVertex(Point(4, -5))),
            NCRadiusSegment(startPt=CornerVertex(Point(4, -5)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(0, -1))),
            NCLineSegment(startPt=CornerVertex(Point(0, -1)), endPt=CornerVertex(Point(0, 1))),
            NCRadiusSegment(startPt=CornerVertex(Point(0, 1)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(4, 5))),
            NCLineSegment(startPt=CornerVertex(Point(4, 5)), endPt=CornerVertex(Point(120, 5))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtStartCW(self.dadoX, DadoContourInsetAtStartTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCCWX(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(120, 5)), endPt=CornerVertex(Point(4, 5))),
            NCRadiusSegment(startPt=CornerVertex(Point(4, 5)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(0, 1))),
            NCLineSegment(startPt=CornerVertex(Point(0, 1)), endPt=CornerVertex(Point(0, -1))),
            NCRadiusSegment(startPt=CornerVertex(Point(0, -1)), radiusPt=RadiusVertex(position=Point(100 - self.BIG, -self.SMALL), radius=4), endPt=CornerVertex(Point(4, -5))),
            NCLineSegment(startPt=CornerVertex(Point(4, -5)), endPt=CornerVertex(Point(120, -5))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtStartCCW(self.dadoX, DadoContourInsetAtStartTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(5, 120)), endPt=CornerVertex(Point(5, 4))),
            NCRadiusSegment(startPt=CornerVertex(Point(5, 4)), radiusPt=RadiusVertex(position=Point(self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(1, 0))),
            NCLineSegment(startPt=CornerVertex(Point(1, 0)), endPt=CornerVertex(Point(-1, 0))),
            NCRadiusSegment(startPt=CornerVertex(Point(-1, 0)), radiusPt=RadiusVertex(position=Point(-self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(-5, 4))),
            NCLineSegment(startPt=CornerVertex(Point(-5, 4)), endPt=CornerVertex(Point(-5, 120))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtStartCW(self.dadoY, DadoContourInsetAtStartTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))

    def testCCWY(self):
        expected = (
            NCLineSegment(startPt=CornerVertex(Point(-5, 120)), endPt=CornerVertex(Point(-5, 4))),
            NCRadiusSegment(startPt=CornerVertex(Point(-5, 4)), radiusPt=RadiusVertex(position=Point(-self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(-1, 0))),
            NCLineSegment(startPt=CornerVertex(Point(-1, 0)), endPt=CornerVertex(Point(1, 0))),
            NCRadiusSegment(startPt=CornerVertex(Point(1, 0)), radiusPt=RadiusVertex(position=Point(self.SMALL, 100 - self.BIG), radius=4), endPt=CornerVertex(Point(5, 4))),
            NCLineSegment(startPt=CornerVertex(Point(5, 4)), endPt=CornerVertex(Point(5, 120))),
        )
        self.assertEqual(expected, getDadoRoundedContourInsetAtStartCCW(self.dadoY, DadoContourInsetAtStartTests.SECURITY_DISTANCE, self.ROUTER_RADIUS))


class CallOfTheRightFunctionsDueToInsetsTests(unittest.TestCase):
    SECURITY_DISTANCE = 100
    ROUTER_RADIUS = 4

    def setUp(self):
        self.dado = unittest.mock.Mock()

    @unittest.mock.patch(__name__ + '.getDadoContourBothInsetCW')
    def testDadoContourBothInsetCW(self, patcher):
        self.dado.isInsetBoth.return_value = True
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = True
        getDadoContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourInsetAtStartCW')
    def testDadoContourInsetAtStartCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = False
        getDadoContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourInsetAtEndCW')
    def testDadoContourInsetAtEndCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = True
        getDadoContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourThroughCW')
    def testDadoContourThroughCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = False
        getDadoContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourBothInsetCCW')
    def testDadoContourBothInsetCCW(self, patcher):
        self.dado.isInsetBoth.return_value = True
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = True
        getDadoContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourInsetAtStartCCW')
    def testDadoContourInsetAtStartCCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = False
        getDadoContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourInsetAtEndCCW')
    def testDadoContourInsetAtEndCCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = True
        getDadoContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourThroughCCW')
    def testDadoContourThroughCCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = False
        getDadoContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)


class CallOfTheRightFunctionsDueToRoundedInsetsTests(unittest.TestCase):
    SECURITY_DISTANCE = 100
    ROUTER_RADIUS = 4

    def setUp(self):
        self.dado = unittest.mock.Mock()

    @unittest.mock.patch(__name__ + '.getDadoRoundedContourBothInsetCW')
    def testDadoRoundedContourBothInsetCW(self, patcher):
        self.dado.isInsetBoth.return_value = True
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = True
        getDadoRoundedContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoRoundedContourInsetAtStartCW')
    def testDadoRoundedContourInsetAtStartCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = False
        getDadoRoundedContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoRoundedContourInsetAtEndCW')
    def testDadoRoundedContourInsetAtEndCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = True
        getDadoRoundedContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourThroughCW')
    def testDadoRoundedContourThroughCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = False
        getDadoRoundedContourCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoRoundedContourBothInsetCCW')
    def testDadoRoundedContourBothInsetCCW(self, patcher):
        self.dado.isInsetBoth.return_value = True
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = True
        getDadoRoundedContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoRoundedContourInsetAtStartCCW')
    def testDadoRoundedContourInsetAtStartCCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = True
        self.dado.isInsetEnd.return_value = False
        getDadoRoundedContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoRoundedContourInsetAtEndCCW')
    def testDadoRoundedContourInsetAtEndCCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = True
        getDadoRoundedContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)

    @unittest.mock.patch(__name__ + '.getDadoContourThroughCCW')
    def testDadoRoundedContourThroughCCW(self, patcher):
        self.dado.isInsetBoth.return_value = False
        self.dado.isInsetStart.return_value = False
        self.dado.isInsetEnd.return_value = False
        getDadoRoundedContourCCW(self.dado, self.SECURITY_DISTANCE, self.ROUTER_RADIUS)
        self.assertTrue(patcher.called)
