# -*- coding: utf-8 -*-

# Copyright 2011-2014, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest

from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline

from interiorcad.backend.ncexportmigrated.nclinesegment import NCLineSegment
from interiorcad.backend.ncexportmigrated.ncradiussegment import NCRadiusSegment


class NCLineAndRadiusSegment:
    @staticmethod
    def __generateVerticesAndTheirFollowersIfNotIdentical(vertices):
        for index, vertex in enumerate(vertices, start=0):
            nextPointIndex = (index + 1) % len(vertices)
            if vertices[nextPointIndex] == vertex:
                continue

            yield vertices[index], vertices[nextPointIndex]

    @staticmethod
    def __generateLineAndRadiusSegments(polylinePath):
        segmentsWithVisibility = []
        if polylinePath.vertices:
            lastEndVertex = polylinePath.vertices[0]
            for vertex, nextVertex in NCLineAndRadiusSegment.__generateVerticesAndTheirFollowersIfNotIdentical(polylinePath.vertices):
                if vertex.isCornerVertex():
                    lastEndVertex = NCLineAndRadiusSegment.__generateLineSegment(vertex, nextVertex, lastEndVertex, segmentsWithVisibility)
                elif vertex.isRadiusVertex():
                    lastEndVertex = NCLineAndRadiusSegment.__generateRadiusSegment(vertex, nextVertex, lastEndVertex, segmentsWithVisibility)
                else:
                    assert False, 'Invalid vertex type'
        return segmentsWithVisibility

    @staticmethod
    def __generateLineSegment(vertex, nextVertex, lastEndVertex, segmentsWithVisibility):
        if nextVertex.isCornerVertex():
            segmentsWithVisibility.append((vertex.visible, NCLineSegment(lastEndVertex, nextVertex)))
            lastEndVertex = nextVertex
        return lastEndVertex

    @staticmethod
    def __generateRadiusSegment(vertex, nextVertex, lastEndVertex, segmentsWithVisibility):
        segmentsWithVisibility.append((vertex.visible, NCRadiusSegment(lastEndVertex, vertex, nextVertex)))
        lastEndVertex = nextVertex
        return lastEndVertex

    @staticmethod
    def generateLineAndRadiusSegments(polylinePath):
        return tuple(segment for dummyVisible, segment in NCLineAndRadiusSegment.__generateLineAndRadiusSegments(polylinePath))

    @staticmethod
    def generateVisibleLineAndRadiusSegments(polylinePath):
        return tuple(segment for visible, segment in NCLineAndRadiusSegment.__generateLineAndRadiusSegments(polylinePath) if visible)


class Tests(unittest.TestCase):
    def setUp(self):
        self.polyline = Polyline(
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(100, 200)),
        )

    def getExpectedSegments(self, startVisible, endVisible):
        return (
            NCLineSegment(CornerVertex(position=Point(0, 0), visible=startVisible), CornerVertex(position=Point(100, 0), visible=endVisible)),
            NCLineSegment(CornerVertex(position=Point(100, 0), visible=startVisible), CornerVertex(position=Point(100, 200), visible=endVisible)),
            NCLineSegment(CornerVertex(position=Point(100, 200), visible=startVisible), CornerVertex(position=Point(0, 0), visible=endVisible))
        )

    def testGenerateLineAndRadiusSegments(self):
        expectedSegments = self.getExpectedSegments(startVisible=True, endVisible=True)
        self.assertEqual(expectedSegments, NCLineAndRadiusSegment.generateLineAndRadiusSegments(self.polyline))

    def testGenerateLineAndRadiusSegmentsShouldNotCareAboutVisibility(self):
        for vertex in self.polyline.vertices:
            vertex.visible = False

        expectedSegments = self.getExpectedSegments(startVisible=False, endVisible=False)
        self.assertEqual(expectedSegments, NCLineAndRadiusSegment.generateLineAndRadiusSegments(self.polyline))

    def testGenerateVisibleLineAndRadiusSegments(self):
        expectedSegments = self.getExpectedSegments(startVisible=True, endVisible=True)
        self.assertEqual(expectedSegments, NCLineAndRadiusSegment.generateVisibleLineAndRadiusSegments(self.polyline))

    def testGenerateVisibleLineAndRadiusSegmentsShouldReturnAnEmptyListIfAllVerticesAreNotVisible(self):
        for vertex in self.polyline.vertices:
            vertex.visible = False

        expectedSegments = ()
        self.assertEqual(expectedSegments, NCLineAndRadiusSegment.generateVisibleLineAndRadiusSegments(self.polyline))

    def testGenerateVisibleLineAndRadiusSegmentsShouldNotReturnTheLastSegmentIfItIsNotVisible(self):
        self.polyline.vertices[2].visible = False

        expectedSegments = (
            NCLineSegment(CornerVertex(position=Point(0, 0), visible=True), CornerVertex(position=Point(100, 0), visible=True)),
            NCLineSegment(CornerVertex(position=Point(100, 0), visible=True), CornerVertex(position=Point(100, 200), visible=False)),
        )
        self.assertEqual(expectedSegments, NCLineAndRadiusSegment.generateVisibleLineAndRadiusSegments(self.polyline))


class BugIC3424Tests(unittest.TestCase):
    def setUp(self):
        self.polyline = Polyline(
            CornerVertex(Point(0, 0)),
            RadiusVertex(Point(200, -200), radius=200),
            CornerVertex(Point(0, -400)),
            CornerVertex(Point(-400, -400)),
            CornerVertex(Point(-400, 0)),
        )

    def testSplitOfRoundSegment(self):
        expectedSegments = (
            NCRadiusSegment(startPt=CornerVertex(Point(0, 0), visible=True), radiusPt=RadiusVertex(position=Point(200, -200), visible=True, radius=200.00), endPt=CornerVertex(position=Point(0, -400), visible=True)),
            NCLineSegment(startPt=CornerVertex(position=Point(0, -400), visible=True), endPt=CornerVertex(position=Point(-400, -400), visible=True)),
            NCLineSegment(startPt=CornerVertex(position=Point(-400, -400), visible=True), endPt=CornerVertex(position=Point(-400, 0), visible=True)),
            NCLineSegment(startPt=CornerVertex(position=Point(-400, 0), visible=True), endPt=CornerVertex(position=Point(0, 0), visible=True))
        )
        self.assertEqual(expectedSegments, NCLineAndRadiusSegment.generateLineAndRadiusSegments(self.polyline))
