# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest
import math

from interiorcad.interface.geometry.clockwise import isClockwise
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.geometry.center3points import getCenterOfThreePointsOnArc


class NCRadiusSegment:
    def __init__(self, startPt=0, radiusPt=0, endPt=0):
        self.startVertex = startPt
        self.radiusVertex = radiusPt
        self.endVertex = endPt

    def isLineSegment(self):
        return False

    def __eq__(self, rhs):
        return all((rhs is not None, self.startVertex == rhs.startVertex, self.radiusVertex == rhs.radiusVertex, self.endVertex == rhs.endVertex))

    def __repr__(self):
        return "NCRadiusSegment(startPt={}, radiusPt={}, endPt={})".format(self.startVertex, self.radiusVertex, self.endVertex)

    def isClockwise(self):
        return isClockwise(self.startVertex.position, self.radiusVertex.position, self.endVertex.position)

    def isArcGreater180(self):
        firstVector = Vector()
        firstVector.x = self.startVertex.position.x - self.radiusVertex.position.x
        firstVector.y = self.startVertex.position.y - self.radiusVertex.position.y
        firstVector.z = 0.0

        secondVector = Vector()
        secondVector.x = self.endVertex.position.x - self.radiusVertex.position.x
        secondVector.y = self.endVertex.position.y - self.radiusVertex.position.y
        secondVector.z = 0.0

        return VectorMath.angleBetween(secondVector, firstVector) < 90.0

    def getDirection(self, segmentDirectionCW, segmentDirectionCCW, segmentDirectionCWGT180, segmentDirectionCCWGT180):
        if self.isClockwise():
            if self.isArcGreater180():
                direction = segmentDirectionCWGT180
            else:
                direction = segmentDirectionCW
        else:
            if self.isArcGreater180():
                direction = segmentDirectionCCWGT180
            else:
                direction = segmentDirectionCCW

        return direction

    def getStartVertex(self):
        return self.startVertex

    def getEndVertex(self):
        return self.endVertex

    def getRadius(self):
        return self.radiusVertex.radius

    def getCenter(self):
        return getCenterOfThreePointsOnArc(self.startVertex.position, self.radiusVertex.position, self.endVertex.position, self.getRadius())

    @classmethod
    def createRoundedSegmentAtCorner(cls, cornerPoint, deltaAtStartDirection, deltaAtEndDirection, radius):
        startVertex = CornerVertex(cornerPoint + deltaAtStartDirection * radius)
        endVertex = CornerVertex(cornerPoint + deltaAtEndDirection * radius)

        middlePoint = (cornerPoint + (deltaAtStartDirection + deltaAtEndDirection) * radius)

        radiusPoint = - deltaAtStartDirection - deltaAtEndDirection
        radiusPoint.normalize()
        radiusPoint *= radius
        radiusPoint += middlePoint
        radiusVertex = RadiusVertex(position=radiusPoint, radius=radius)

        return cls(startPt=startVertex, radiusPt=radiusVertex, endPt=endVertex)


class NCRadiusSegmentTests(unittest.TestCase):
    def setUp(self):
        self.SEGMENT_DIRECTION_CW = 0
        self.SEGMENT_DIRECTION_CCW = 1
        self.SEGMENT_DIRECTION_CW_GT_180 = 2
        self.SEGMENT_DIRECTION_CCW_GT_180 = 3

    def expectedDirection(self, segment):
        return segment.getDirection(self.SEGMENT_DIRECTION_CW, self.SEGMENT_DIRECTION_CCW, self.SEGMENT_DIRECTION_CW_GT_180, self.SEGMENT_DIRECTION_CCW_GT_180)

    def testSegmentShouldBeClockwise(self):
        startPt = CornerVertex(Point(150.0, 100.0))
        radiusPt = RadiusVertex(Point(185.355339, 85.355339), radius=50.0)
        endPt = CornerVertex(Point(200.0, 50.0))
        segment = NCRadiusSegment(startPt, radiusPt, endPt)

        self.assertTrue(segment.isClockwise())

    def testSegmentShouldNotBeClockwise(self):
        startPt = CornerVertex(Point(200.0, 50.0))
        radiusPt = RadiusVertex(Point(185.355339, 85.355339), radius=50.0)
        endPt = CornerVertex(Point(150.0, 100.0))
        segment = NCRadiusSegment(startPt, radiusPt, endPt)

        self.assertFalse(segment.isClockwise())

    def testSegmentShouldBeCW(self):
        startPt = CornerVertex(Point(150.0, 100.0))
        radiusPt = RadiusVertex(Point(185.355339, 85.355339), radius=50.0)
        endPt = CornerVertex(Point(200.0, 50.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CW)

    def testSegmentShouldBeCCW(self):
        startPt = CornerVertex(Point(200.0, 50.0))
        radiusPt = RadiusVertex(Point(185.355339, 85.355339), radius=50.0)
        endPt = CornerVertex(Point(150.0, 100.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CCW)

    def testSegmentShouldBeCWGT180(self):
        startPt = CornerVertex(Point(150.0, 100.0))
        radiusPt = RadiusVertex(Point(216.513, 116.513), radius=40.0)
        endPt = CornerVertex(Point(200.0, 50.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CW_GT_180)

    def testSegmentShouldBeCCWGT180(self):
        startPt = CornerVertex(Point(200.0, 50.0))
        radiusPt = RadiusVertex(Point(216.513, 116.513), radius=40.0)
        endPt = CornerVertex(Point(150.0, 100.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CCW_GT_180)

    def testSegmentShouldBeCWXilog(self):
        startPt = CornerVertex(Point(200.0, 50.0))
        radiusPt = RadiusVertex(Point(185.355339, 14.644), radius=50.0)
        endPt = CornerVertex(Point(150.0, 0.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CW)

    def testSegmentShouldBeCWWXilog(self):
        startPt = CornerVertex(Point(200.0, 50.0))
        radiusPt = RadiusVertex(Point(164.644, 35.355), radius=50.0)
        endPt = CornerVertex(Point(150.0, 0.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CCW)

    def testSegmentShouldBeCCWGT180Xilog(self):
        startPt = CornerVertex(Point(200.0, 50.0))
        radiusPt = RadiusVertex(Point(133.487, 66.513), radius=40.0)
        endPt = CornerVertex(Point(150.0, 0.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CCW_GT_180)

    def testSegmentShouldBeCWGT180Xilog(self):
        startPt = CornerVertex(Point(200.0, 66.513))
        radiusPt = RadiusVertex(Point(216.513, 0.0), radius=40.0)
        endPt = CornerVertex(Point(150.0, 16.513))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)
        direction = self.expectedDirection(segment)
        self.assertEqual(direction, self.SEGMENT_DIRECTION_CW_GT_180)

    def testIsGreaterThan180(self):
        startPt = CornerVertex(Point(0.0, 0.0))
        radiusPt = RadiusVertex(Point(-356.0, 158.0), radius=213.0)
        endPt = CornerVertex(Point(0.0, 315.0))

        segment = NCRadiusSegment(startPt, radiusPt, endPt)

        self.assertTrue(segment.isArcGreater180())


class CreateRoundedSegmentTests(unittest.TestCase):
    def testBottomToRight(self):
        cornerPoint = Point(0, 0)
        startVector = Point(0, -1)
        endVector = Point(1, 0)
        radius = 10

        startPt = CornerVertex(Point(0, -10))
        recess = (math.sqrt(2) - 1) * radius / math.sqrt(2)
        radiusPt = RadiusVertex(Point(recess, -recess), radius=10)
        endPt = CornerVertex(Point(10, 0))
        expectedSegment = NCRadiusSegment(startPt, radiusPt, endPt)

        self.assertEqual(expectedSegment, NCRadiusSegment.createRoundedSegmentAtCorner(cornerPoint, startVector, endVector, radius))
