# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest

from interiorcad.backend.ncexportmigrated.nclinesegment import NCLineSegment
from interiorcad.backend.ncexportmigrated.ncradiussegment import NCRadiusSegment
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polylinesegment import PolylineSegment
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.simplelinesegment import SimpleLineSegment


class RabbetToNCSegments:
    @staticmethod
    def __getParallelLineSegment(rabbet, insetSecurityDistance):
        pathLineSegment = rabbet.getPathLineSegment()

        startDot = (rabbet.getStartPoint() - pathLineSegment.getStartPoint()).dot(pathLineSegment.getDirectionVector())
        endDot = (rabbet.getEndPoint() - pathLineSegment.getEndPoint()).dot(pathLineSegment.getDirectionVector())
        newStart = pathLineSegment.getStartPoint() + pathLineSegment.getDirectionVector() * startDot
        newEnd = pathLineSegment.getEndPoint() + pathLineSegment.getDirectionVector() * endDot

        parallelLineSegment = SimpleLineSegment(newStart, newEnd)
        parallelVector = parallelLineSegment.getPerpendicularVector()
        if (newStart - rabbet.getStartPoint()).dot(parallelVector) < 0:
            parallelVector = - parallelVector
        parallelLineSegment.moveByVector(parallelVector * insetSecurityDistance)

        return parallelLineSegment

    @staticmethod
    def generateRabbetSegment(rabbet, insetSecurityDistance, straightSecurityDistance):
        parallelLineSegment = RabbetToNCSegments.__getParallelLineSegment(rabbet, insetSecurityDistance)

        if rabbet.isInsetStart():
            yield NCLineSegment(CornerVertex(parallelLineSegment.getStartPoint()), CornerVertex(rabbet.getStartPoint()))
        elif straightSecurityDistance:
            startPoint = rabbet.getStartPoint()
            startPoint -= rabbet.directionVector * straightSecurityDistance
            yield NCLineSegment(CornerVertex(startPoint), CornerVertex(rabbet.getStartPoint()))

        yield NCLineSegment(CornerVertex(rabbet.getStartPoint()), CornerVertex(rabbet.getEndPoint()))

        if rabbet.isInsetEnd():
            yield NCLineSegment(CornerVertex(rabbet.getEndPoint()), CornerVertex(parallelLineSegment.getEndPoint()))
        elif straightSecurityDistance:
            endPoint = rabbet.getEndPoint()
            endPoint += rabbet.directionVector * straightSecurityDistance
            yield NCLineSegment(CornerVertex(rabbet.getEndPoint()), CornerVertex(endPoint))

    @staticmethod
    def generateRabbetSegmentsWithSecurityDistance(rabbet, insetSecurityDistance, straightSecurityDistance):
        return RabbetToNCSegments.generateRabbetSegment(rabbet, insetSecurityDistance, straightSecurityDistance)

    @staticmethod
    def generateRabbetSegmentsWithInsets(rabbet):
        yield from RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=0, straightSecurityDistance=None)

    @staticmethod
    def generateRabbetRoundedSegmentsWithSecurityDistance(rabbet, insetSecurityDistance, straightSecurityDistance, routerRadius):
        parallelLineSegment = RabbetToNCSegments.__getParallelLineSegment(rabbet, insetSecurityDistance)

        if rabbet.isInsetStart():
            startLineVector = (rabbet.getStartPoint() - parallelLineSegment.getStartPoint())
            startLineVector.normalize()

            yield NCLineSegment(CornerVertex(parallelLineSegment.getStartPoint()), CornerVertex(rabbet.getStartPoint() - startLineVector * routerRadius))
            yield NCRadiusSegment.createRoundedSegmentAtCorner(rabbet.getStartPoint(), - startLineVector, parallelLineSegment.getDirectionVector(), routerRadius)
        elif straightSecurityDistance:
            startPoint = rabbet.getStartPoint()
            startPoint -= rabbet.directionVector * straightSecurityDistance
            yield NCLineSegment(CornerVertex(startPoint), CornerVertex(rabbet.getStartPoint()))

        startPoint = rabbet.getStartPoint()
        endPoint = rabbet.getEndPoint()
        if rabbet.isInsetStart():
            startPoint += parallelLineSegment.getDirectionVector() * routerRadius
        if rabbet.isInsetEnd():
            endPoint -= parallelLineSegment.getDirectionVector() * routerRadius

        yield NCLineSegment(CornerVertex(startPoint), CornerVertex(endPoint))

        if rabbet.isInsetEnd():
            endLineVector = (parallelLineSegment.getEndPoint() - rabbet.getEndPoint())
            endLineVector.normalize()

            yield NCRadiusSegment.createRoundedSegmentAtCorner(rabbet.getEndPoint(), - parallelLineSegment.getDirectionVector(), endLineVector, routerRadius)
            yield NCLineSegment(CornerVertex(rabbet.getEndPoint() + endLineVector * routerRadius), CornerVertex(parallelLineSegment.getEndPoint()))
        elif straightSecurityDistance:
            endPoint = rabbet.getEndPoint()
            endPoint += rabbet.directionVector * straightSecurityDistance
            yield NCLineSegment(CornerVertex(rabbet.getEndPoint()), CornerVertex(endPoint))

    @staticmethod
    def generateRabbetSegmentsClosedPerimeter(rabbet):
        insetSecurityDistance = 0

        parallelLineSegment = RabbetToNCSegments.__getParallelLineSegment(rabbet, insetSecurityDistance)

        yield NCLineSegment(CornerVertex(parallelLineSegment.getEndPoint()), CornerVertex(parallelLineSegment.getStartPoint()))
        yield NCLineSegment(CornerVertex(parallelLineSegment.getStartPoint()), CornerVertex(rabbet.getStartPoint()))
        yield NCLineSegment(CornerVertex(rabbet.getStartPoint()), CornerVertex(rabbet.getEndPoint()))
        yield NCLineSegment(CornerVertex(rabbet.getEndPoint()), CornerVertex(parallelLineSegment.getEndPoint()))


class MockRabbet:
    def __init__(self, startPoint, endPoint, startInset=0, endInset=0, segment=None):
        self.startPoint = startPoint
        self.endPoint = endPoint
        self.startInset = startInset
        self.endInset = endInset
        self.segment = segment

    def getStartPoint(self):
        return self.startPoint

    def getEndPoint(self):
        return self.endPoint

    def isInset(self):
        return self.isInsetStart() or self.isInsetEnd()

    def isInsetBoth(self):
        return self.isInsetStart() and self.isInsetEnd()

    def isInsetStart(self):
        return not isNearlyEqual(self.startInset, 0)

    def isInsetEnd(self):
        return not isNearlyEqual(self.endInset, 0)

    def getPathLineSegment(self):
        return SimpleLineSegment(self.segment.start.position, self.segment.end.position)

    def getDeltaVector(self):
        return self.endPoint - self.startPoint

    @property
    def directionVector(self):
        vector = self.getDeltaVector()
        vector.normalize()
        return vector


class GetParallelLineSegmentTests(unittest.TestCase):
    @staticmethod
    def getParallelLineSegment(rabbet, insetSecurityDistance):
        return RabbetToNCSegments._RabbetToNCSegments__getParallelLineSegment(rabbet, insetSecurityDistance)

    def setUp(self):
        self.INSET_SECURITY_DISTANCE = 100

    def testBottomCW(self):
        rabbet = MockRabbet(Point(0, 10), Point(500, 10), segment=PolylineSegment(CornerVertex(Point(500, 0)), CornerVertex(Point(0, 0))))
        lineSegment = GetParallelLineSegmentTests.getParallelLineSegment(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE)
        expectedLineSegment = SimpleLineSegment(Point(0, -100), Point(500, -100))
        self.assertEqual(expectedLineSegment, lineSegment)

    def testLeftCW(self):
        rabbet = MockRabbet(Point(10, 500), Point(10, 0), segment=PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 500))))
        lineSegment = GetParallelLineSegmentTests.getParallelLineSegment(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE)
        expectedLineSegment = SimpleLineSegment(Point(-100, 500), Point(-100, 0))
        self.assertEqual(expectedLineSegment, lineSegment)

    def testBottomCCW(self):
        rabbet = MockRabbet(Point(500, 10), Point(0, 10), segment=PolylineSegment(CornerVertex(Point(500, 0)), CornerVertex(Point(0, 0))))
        lineSegment = GetParallelLineSegmentTests.getParallelLineSegment(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE)
        expectedLineSegment = SimpleLineSegment(Point(500, -100), Point(0, -100))
        self.assertEqual(expectedLineSegment, lineSegment)

    def testLeftCCW(self):
        rabbet = MockRabbet(Point(10, 0), Point(10, 500), segment=PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 500))))
        lineSegment = GetParallelLineSegmentTests.getParallelLineSegment(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE)
        expectedLineSegment = SimpleLineSegment(Point(-100, 0), Point(-100, 500))
        self.assertEqual(expectedLineSegment, lineSegment)


class GenerateRabbetSegmentsWithSecurityDistanceTests(unittest.TestCase):
    def setUp(self):
        self.INSET_SECURITY_DISTANCE = 100
        self.STRAIGHT_SECURITY_DISTANCE = 200
        self.X_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(500, 0)), CornerVertex(Point(0, 0)))
        self.Y_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 500)))

    def testBottom(self):
        rabbet = MockRabbet(Point(0, 10), Point(500, 10), segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(-200, 10)), CornerVertex(Point(0, 10))),
            NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(500, 10))),
            NCLineSegment(CornerVertex(Point(500, 10)), CornerVertex(Point(700, 10))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testLeft(self):
        rabbet = MockRabbet(Point(10, 500), Point(10, 0), segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(10, 700)), CornerVertex(Point(10, 500))),
            NCLineSegment(CornerVertex(Point(10, 500)), CornerVertex(Point(10, 0))),
            NCLineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, -200))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetBottom(self):
        rabbet = MockRabbet(Point(20, 10), Point(490, 10), 20, 10, segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE))

        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(20, -100)), CornerVertex(Point(20, 10))),
            NCLineSegment(CornerVertex(Point(20, 10)), CornerVertex(Point(490, 10))),
            NCLineSegment(CornerVertex(Point(490, 10)), CornerVertex(Point(490, -100))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetLeft(self):
        rabbet = MockRabbet(Point(10, 490), Point(10, 20), 10, 20, segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(-100, 490)), CornerVertex(Point(10, 490))),
            NCLineSegment(CornerVertex(Point(10, 490)), CornerVertex(Point(10, 20))),
            NCLineSegment(CornerVertex(Point(10, 20)), CornerVertex(Point(-100, 20))),
        )
        self.assertEqual(expectedSegmentList, segmentList)


class GenerateRabbetSegmentsWithSecurityDistanceNoStraighTests(unittest.TestCase):
    def setUp(self):
        self.X_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(500, 0)), CornerVertex(Point(0, 0)))
        self.Y_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 500)))

    def testBottom(self):
        rabbet = MockRabbet(Point(0, 10), Point(500, 10), segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithInsets(rabbet=rabbet))
        expectedSegmentList = (NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(500, 10))),)
        self.assertEqual(expectedSegmentList, segmentList)

    def testLeft(self):
        rabbet = MockRabbet(Point(10, 500), Point(10, 0), segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithInsets(rabbet=rabbet))
        expectedSegmentList = (NCLineSegment(CornerVertex(Point(10, 500)), CornerVertex(Point(10, 0))),)
        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetBottom(self):
        rabbet = MockRabbet(Point(20, 10), Point(490, 10), 20, 10, segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithInsets(rabbet=rabbet))

        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(20, 0)), CornerVertex(Point(20, 10))),
            NCLineSegment(CornerVertex(Point(20, 10)), CornerVertex(Point(490, 10))),
            NCLineSegment(CornerVertex(Point(490, 10)), CornerVertex(Point(490, 0))),
        )

        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetLeft(self):
        rabbet = MockRabbet(Point(10, 490), Point(10, 20), 10, 20, segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithInsets(rabbet=rabbet))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(0, 490)), CornerVertex(Point(10, 490))),
            NCLineSegment(CornerVertex(Point(10, 490)), CornerVertex(Point(10, 20))),
            NCLineSegment(CornerVertex(Point(10, 20)), CornerVertex(Point(0, 20))),
        )
        self.assertEqual(expectedSegmentList, segmentList)


class GenerateRabbetSegmentsWithInsetTests(unittest.TestCase):
    def setUp(self):
        self.INSET_SECURITY_DISTANCE = 100
        self.X_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(500, 0)), CornerVertex(Point(0, 0)))
        self.Y_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 500)))

    def testBottom(self):
        rabbet = MockRabbet(Point(0, 10), Point(500, 10), segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=None))
        expectedSegmentList = (NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(500, 10))),)
        self.assertEqual(expectedSegmentList, segmentList)

    def testLeft(self):
        rabbet = MockRabbet(Point(10, 500), Point(10, 0), segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=None))
        expectedSegmentList = (NCLineSegment(CornerVertex(Point(10, 500)), CornerVertex(Point(10, 0))),)
        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetBottom(self):
        rabbet = MockRabbet(Point(20, 10), Point(490, 10), 20, 10, segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=None))

        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(20, -100)), CornerVertex(Point(20, 10))),
            NCLineSegment(CornerVertex(Point(20, 10)), CornerVertex(Point(490, 10))),
            NCLineSegment(CornerVertex(Point(490, 10)), CornerVertex(Point(490, -100))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetLeft(self):
        rabbet = MockRabbet(Point(10, 490), Point(10, 20), 10, 20, segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=None))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(-100, 490)), CornerVertex(Point(10, 490))),
            NCLineSegment(CornerVertex(Point(10, 490)), CornerVertex(Point(10, 20))),
            NCLineSegment(CornerVertex(Point(10, 20)), CornerVertex(Point(-100, 20))),
        )
        self.assertEqual(expectedSegmentList, segmentList)


class GenerateRabbetRoundedSegmentsTests(unittest.TestCase):
    ROUTER_RADIUS = 4
    INSET_SECURITY_DISTANCE = 100
    STRAIGHT_SECURITY_DISTANCE = 200

    BIG = 21.17157287525381
    SMALL = 8.82842712474619

    def setUp(self):
        self.X_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(500, 0)), CornerVertex(Point(0, 0)))
        self.Y_PATH_SEGMENT = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 500)))

    def testBottom(self):
        rabbet = MockRabbet(Point(0, 10), Point(500, 10), segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetRoundedSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE, routerRadius=self.ROUTER_RADIUS))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(-200, 10)), CornerVertex(Point(0, 10))),
            NCLineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(500, 10))),
            NCLineSegment(CornerVertex(Point(500, 10)), CornerVertex(Point(700, 10))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testLeft(self):
        rabbet = MockRabbet(Point(10, 500), Point(10, 0), segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetRoundedSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE, routerRadius=self.ROUTER_RADIUS))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(10, 700)), CornerVertex(Point(10, 500))),
            NCLineSegment(CornerVertex(Point(10, 500)), CornerVertex(Point(10, 0))),
            NCLineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, -200))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetBottom(self):
        rabbet = MockRabbet(Point(20, 10), Point(490, 10), 20, 10, segment=self.X_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetRoundedSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE, routerRadius=self.ROUTER_RADIUS))

        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(20, -100)), CornerVertex(Point(20, 6))),
            NCRadiusSegment(startPt=CornerVertex(Point(20, 6)), radiusPt=RadiusVertex(position=Point(self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(24, 10))),
            NCLineSegment(CornerVertex(Point(24, 10)), CornerVertex(Point(486, 10))),
            NCRadiusSegment(startPt=CornerVertex(Point(486, 10)), radiusPt=RadiusVertex(position=Point(510 - self.BIG, self.SMALL), radius=4), endPt=CornerVertex(Point(490, 6))),
            NCLineSegment(CornerVertex(Point(490, 6)), CornerVertex(Point(490, -100))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testInsetLeft(self):
        rabbet = MockRabbet(Point(10, 490), Point(10, 20), 10, 20, segment=self.Y_PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetRoundedSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=self.STRAIGHT_SECURITY_DISTANCE, routerRadius=self.ROUTER_RADIUS))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(-100, 490)), CornerVertex(Point(6, 490))),
            NCRadiusSegment(startPt=CornerVertex(Point(6, 490)), radiusPt=RadiusVertex(position=Point(self.SMALL, 510 - self.BIG), radius=4), endPt=CornerVertex(Point(10, 486))),
            NCLineSegment(CornerVertex(Point(10, 486)), CornerVertex(Point(10, 24))),
            NCRadiusSegment(startPt=CornerVertex(Point(10, 24)), radiusPt=RadiusVertex(position=Point(self.SMALL, self.BIG), radius=4), endPt=CornerVertex(Point(6, 20))),
            NCLineSegment(CornerVertex(Point(6, 20)), CornerVertex(Point(-100, 20))),
        )
        self.assertEqual(expectedSegmentList, segmentList)


class AngleInsetTests(unittest.TestCase):
    def setUp(self):
        self.INSET_SECURITY_DISTANCE = 100
        self.LENGTH = 500

    def testInset45Left(self):
        position = self.LENGTH * math.sqrt(2) / 2
        self.PATH_SEGMENT = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(position, position)))
        insetLeft = 10
        insetRight = 20
        rabbet = MockRabbet(Point(-insetLeft * math.sqrt(2) / 2, -insetLeft * math.sqrt(2) / 2), Point((500 + insetRight) * math.sqrt(2) / 2, (500 + insetRight) * math.sqrt(2) / 2), insetLeft, insetRight, segment=self.PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=None))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(63.63961030678928, -77.78174593052023)), CornerVertex(Point(-7.0710678118654755, -7.0710678118654755))),
            NCLineSegment(CornerVertex(Point(-7.0710678118654755, -7.0710678118654755)), CornerVertex(Point(367.6955262170047, 367.6955262170047))),
            NCLineSegment(CornerVertex(Point(367.6955262170047, 367.6955262170047)), CornerVertex(Point(438.4062043356595, 296.98484809834997))),
        )
        self.assertEqual(expectedSegmentList, segmentList)

    def testInset30Left(self):
        ANGLE = math.radians(30)
        cos = math.cos(ANGLE)
        sin = math.sin(ANGLE)

        self.PATH_SEGMENT = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(self.LENGTH * cos, self.LENGTH * sin)))
        insetLeft = 10
        insetRight = 20
        rabbet = MockRabbet(Point(-insetLeft * cos, -insetLeft * sin), Point((500 + insetRight) * cos, (500 + insetRight) * sin), insetLeft, insetRight, segment=self.PATH_SEGMENT)
        segmentList = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=self.INSET_SECURITY_DISTANCE, straightSecurityDistance=None))
        expectedSegmentList = (
            NCLineSegment(CornerVertex(Point(41.33974596215561, -91.60254037844388)), CornerVertex(Point(-8.660254037844387, -5))),
            NCLineSegment(CornerVertex(Point(-8.660254037844387, -5)), CornerVertex(Point(450.33320996790815, 260))),
            NCLineSegment(CornerVertex(Point(450.33320996790815, 260)), CornerVertex(Point(500.33320996790815, 173.39745962155612))),
        )
        self.assertEqual(expectedSegmentList, segmentList)
