# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import math
import unittest.mock

from interiorcad.backend.workpiece.contour import Contour
from interiorcad.backend.workpiece.drilling import Drilling
from interiorcad.backend.workpiece.machininganalyzer import MachiningsOnWorkpieceGetter
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.vector import Vector


class PiercingMachiningsTurnAroundFinder:
    def search(self, turnableMain, turnableOpposite, otherMain, otherOpposite):
        result = []
        if turnableMain and turnableOpposite and not otherMain:
            result = turnableOpposite
        if turnableMain and otherOpposite and not otherMain:
            result = turnableMain
        if turnableOpposite and otherMain and not otherOpposite:
            result = turnableOpposite
        return result


class PiercingVerticalMachiningsSideChanger:
    def process(self, workpiece):
        self.workpiece = workpiece

        for machining in self.findMachiningsToTurnAround():
            self.changeSide(machining)

    def findMachiningsToTurnAround(self):
        machinings = MachiningsOnWorkpieceGetter(self.workpiece)

        piercingFromTop = tuple(filter(self.canMachiningChangeSide, machinings.getAllMachiningsFromTop()))
        piercingFromBottom = tuple(filter(self.canMachiningChangeSide, machinings.getAllMachiningsFromBottom()))
        otherFromTop = tuple(itertools.filterfalse(self.canMachiningChangeSide, machinings.getAllMachiningsFromTop()))
        otherFromBottom = tuple(itertools.filterfalse(self.canMachiningChangeSide, machinings.getAllMachiningsFromBottom()))
        return PiercingMachiningsTurnAroundFinder().search(turnableMain=piercingFromTop, turnableOpposite=piercingFromBottom, otherMain=otherFromTop, otherOpposite=otherFromBottom)

    def canMachiningChangeSide(self, machining):
        if self.isSupportedDrilling(machining):
            return machining.depth >= self.workpiece.thickness
        elif self.isSupportedContour(machining):
            return self.areAllContourPassesThroughtOut(machining)

    def changeSide(self, machining):
        # This math for mirroring matrix of a machining is not correct. What we do here is changing the coordinate system from LHS to RHS
        # It seems this 'incorrect' math is enough for our system. So we leave it for now as is.
        # (correct would be to implement method: machining.matrix.scaleWithCenter(self.workpiece.thickness/2, -1, 1, -1) (see VWFC VWTransformMatrix::ScaleWithCenter(...))
        if self.isSupportedDrilling(machining):
            machining.matrix.w.z *= -1
            machining.matrix.origin.z = self.workpiece.thickness - machining.matrix.origin.z
        elif self.isSupportedContour(machining):
            machining.matrix.w.z *= -1
            machining.matrix.origin.z = self.workpiece.thickness - machining.matrix.origin.z
            machining.path.changeDirectionWithVisibility()

    def isSupportedDrilling(self, machining):
        return isinstance(machining, Drilling) and self.isVerticalUpOrDown(machining)

    def isSupportedContour(self, machining):
        return isinstance(machining, Contour) and self.isVerticalUpOrDown(machining)

    def areAllContourPassesThroughtOut(self, machining):
        if machining.multiUsageRoutings:
            return all(routing.depth >= self.workpiece.thickness for routing in machining.multiUsageRoutings)
        else:  # TODO IT-271: BothSided.db has apparently no multiUsageRoutings, so it has to be checked the old way -- NG
            return machining.depth >= self.workpiece.thickness

    def isVerticalUpOrDown(self, machining):
        return isNearlyEqual(math.fabs(machining.matrix.w.z), 1.0)


class PiercingMachiningsTurnAroundFinderTests(unittest.TestCase):
    def setUp(self):
        self.finder = PiercingMachiningsTurnAroundFinder()
        self.turnableMain = []
        self.turnableOpposit = []
        self.otherMain = []
        self.otherOpposit = []

    def __search(self):
        return self.finder.search(turnableMain=self.turnableMain, turnableOpposite=self.turnableOpposit, otherMain=self.otherMain, otherOpposite=self.otherOpposit)

    def test_NoMachiningsFindsNothing(self):
        self.assertEqual([], self.__search())

    def test_OneTurnableOnMainSideFindsNothing(self):
        self.turnableMain.append("1")
        self.assertEqual([], self.__search())

    def test_OneTurnableOnOppositSideFindsNothing(self):
        self.turnableOpposit.append("1")
        self.assertEqual([], self.__search())

    def test_OneTurnableOnMainSideOneOtherOnOppositFindsTurnable(self):
        self.turnableMain.append("1")
        self.otherOpposit.append("2")
        self.assertEqual(["1"], self.__search())

    def test_OneTurnableOnOppositSideOneOtherOnMainFindsTurnable(self):
        self.turnableOpposit.append("1")
        self.otherMain.append("2")
        self.assertEqual(["1"], self.__search())

    def test_OneTurnableOnOppositAndMainSideOneOtherOnOppositAndMainFindsNothing(self):
        self.turnableMain.append("1")
        self.turnableOpposit.append("2")
        self.otherMain.append("3")
        self.otherOpposit.append("4")
        self.assertEqual([], self.__search())

    def test_OneTurnableOnMainOneTurnableOnOppositFindsTurnableOnOppositSide(self):
        self.turnableMain.append("1")
        self.turnableOpposit.append("2")
        self.assertEqual(["2"], self.__search())

    def test_OneTurnableOnMainOneTurnableAndOneOtherOnOppositFindsTurnableOnMainSide(self):
        self.turnableMain.append("1")
        self.turnableOpposit.append("2")
        self.otherOpposit.append("3")
        self.assertEqual(["1"], self.__search())

    def test_OneTurnableAndOneOtherOnMainOneTurnableOnOppositFindsTurnableOnOppositSide(self):
        self.turnableMain.append("1")
        self.otherMain.append("2")
        self.turnableOpposit.append("3")
        self.assertEqual(["3"], self.__search())

    def test_OneOtherOnMainOneTurnableOnOppositFindsTurnableOnOppositSide(self):
        self.otherMain.append("1")
        self.turnableOpposit.append("2")
        self.assertEqual(["2"], self.__search())


class ChangeSideTests(unittest.TestCase):
    THICKNESS = 20

    def setUp(self):
        self.changer = PiercingVerticalMachiningsSideChanger()
        self.changer.workpiece = unittest.mock.Mock()
        self.changer.workpiece.thickness = self.THICKNESS

    def testDrillingShouldChangeNeitherDepthNorDiameterNorUUID(self):
        drilling = Drilling()
        drilling.uuid = 'MyUUID'
        drilling.depth = self.THICKNESS
        drilling.diameter = 8

        self.changer.changeSide(drilling)

        self.assertEqual('MyUUID', drilling.uuid)
        self.assertEqual(8, drilling.diameter)
        self.assertEqual(self.THICKNESS, drilling.depth)

    def testDrillingShouldChangeOrigin(self):
        drilling = Drilling()
        self.changer.changeSide(drilling)

        self.assertEqual(Vector(0, 0, self.THICKNESS), drilling.matrix.origin)

    def testDrillingShouldChangeMatrix(self):
        drilling = Drilling()
        self.changer.changeSide(drilling)

        self.assertEqual(Vector(0, 0, -1), drilling.matrix.w)

    def testContourShouldChangeNeitherDepthNorDiameterNorUUID(self):
        contour = Contour()
        contour.uuid = 'MyUUID'
        contour.depth = self.THICKNESS
        contour.width = 8
        contour.path = unittest.mock.Mock()

        self.changer.changeSide(contour)

        self.assertEqual('MyUUID', contour.uuid)
        self.assertEqual(8, contour.width)
        self.assertEqual(self.THICKNESS, contour.depth)

    def testContourShouldChangeOrigin(self):
        contour = Contour()
        contour.path = unittest.mock.Mock()

        self.changer.changeSide(contour)

        self.assertEqual(Vector(0, 0, self.THICKNESS), contour.matrix.origin)

    def testContourShouldChangeMatrix(self):
        contour = Contour()
        contour.path = unittest.mock.Mock()

        self.changer.changeSide(contour)

        self.assertEqual(Vector(0, 0, -1), contour.matrix.w)

    def testContourShouldChangeOrderOfVerticesInsteadOfTheRadiusCompensation(self):
        contour = Contour()
        contour.path = unittest.mock.Mock()
        contour.radiusCompensation = Contour.RADIUS_COMPENSATION_RIGHT

        self.changer.changeSide(contour)

        self.assertEqual(Contour.RADIUS_COMPENSATION_RIGHT, contour.radiusCompensation)
        contour.path.changeDirectionWithVisibility.assert_called_once_with()


class NonVerticalTests(unittest.TestCase):
    THICKNESS = 20

    def setUp(self):
        self.changer = PiercingVerticalMachiningsSideChanger()
        self.changer.workpiece = unittest.mock.Mock()
        self.changer.workpiece.thickness = self.THICKNESS

    def testFiveAxisDrillingShouldNotBeAllowedToChangeSide(self):
        drilling = Drilling()
        drilling.matrix = Matrix(w=Vector(0.707, 0, 0.707))
        drilling.depth = NonVerticalTests.THICKNESS + 2
        self.assertFalse(self.changer.canMachiningChangeSide(drilling))

    def testFiveAxisDrillingShouldNotBeModifiedAfterChangeSide(self):
        drilling = Drilling()
        drilling.matrix = Matrix(w=Vector(0.707, 0, 0.707))
        drilling.depth = NonVerticalTests.THICKNESS + 2

        self.changer.changeSide(drilling)

        self.assertEqual(Vector(0.707, 0, 0.707), drilling.matrix.w)

    def testHorizontalContourShouldNotBeAllowedToChangeSide(self):
        contour = Contour()
        contour.path = unittest.mock.Mock()
        contour.matrix = Matrix(w=Vector(0, 1, 0))
        contour.depth = NonVerticalTests.THICKNESS + 2

        self.assertFalse(self.changer.canMachiningChangeSide(contour))

    def testHorizontalContourShouldNotBeModifiedAfterChangeSide(self):
        contour = Contour()
        contour.path = unittest.mock.Mock()
        contour.matrix = Matrix(w=Vector(0, 1, 0))
        contour.depth = NonVerticalTests.THICKNESS + 2

        self.changer.changeSide(contour)

        self.assertEqual(Vector(0, 1, 0), contour.matrix.w)