# -*- coding: utf-8 -*-

# Copyright 2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from __future__ import annotations

import logging
from typing import Generator

from interiorcad.backend.clamex.verticalclamexentryvectorfinder import VerticalClamexEntryVectorFinder
from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.surface import Surface, NCMacro2D
from interiorcad.interface.surface.acangles import normalizeAngle

CLAMEX_C_ANGLE = 'clamexcangle'
CLAMEX_C_ANGLE_180 = 'clamexcangle180'  # Variant, where the +180° is added for non-vertical cases


class ClamexCAngleProcessor:
    def __init__(self: ClamexCAngleProcessor, maxDistanceFromEdge: float, innerPlacementExport: str):
        self.maxDistanceFromEdge = maxDistanceFromEdge
        self.innerPlacementExport = innerPlacementExport

    def process(self: ClamexCAngleProcessor, surface: Surface) -> None:
        keptNCMacros = self.processNCMacros(surface)
        surface.ncmacros = keptNCMacros

    def processNCMacros(self: ClamexCAngleProcessor, surface: Surface) -> Generator[NCMacro2D]:
        for ncmacro in surface.ncmacros:
            if self.isUsingClamexCAngle(ncmacro):
                if ncmacro.isVertical():
                    yield from self.processVerticalClamex(surface, ncmacro)
                else:
                    yield from self.processNonVerticalClamex(ncmacro)
            else:
                yield ncmacro

    @staticmethod
    def isUsingClamexCAngle(ncmacro: NCMacro2D) -> bool:
        # Do not search the closing curly bracket, as there might be a formatting, e.g. {clamexcangle:6.2f}
        return ncmacro.isCodeContaining('{' + CLAMEX_C_ANGLE) or ncmacro.isCodeContaining('{' + CLAMEX_C_ANGLE_180)

    def processVerticalClamex(self: ClamexCAngleProcessor, surface: Surface, ncmacro: NCMacro2D) -> Generator[NCMacro2D]:
        finder = VerticalClamexEntryVectorFinder(ncmacro, surface.finalFinishedSizeShape, self.maxDistanceFromEdge)
        finder.calculate()
        ncmacro.clamexcangle = ncmacro.clamexcangle180 = finder.getAngle()  # As we are in the vertical case, clamexcangle180 does NOT get an extra +180°

        if ncmacro.clamexcangle is not None:
            self.warnForVerticalClamexCloseToEdge(finder, surface, ncmacro)
            yield ncmacro
        else:
            yield from self.processClamexFurtherAwayFromEdge(surface, ncmacro)

    def warnForVerticalClamexCloseToEdge(self: ClamexCAngleProcessor, finder: VerticalClamexEntryVectorFinder, surface: Surface, ncmacro: NCMacro2D) -> None:
        message = None

        if not surface.finalFinishedSizeShape.isPolygonal():
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 87,
                "Vertical Clamex is on a Workpiece that is neither rectangular nor polygonal. Be careful when producing. Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")
        elif not finder.isMinimalDistantSegmentParallel:
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 85,
                "Vertical Clamex is not parallel to an edge of the Workpiece. Be careful when producing. Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")
        elif self.isCorrespondingSegmentInsideBounds(finder, surface):
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 86,
                "Vertical Clamex is not on the bounding bound of the Workpiece. Be careful when producing. Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")

        self.logwarn(message, surface, ncmacro, finder.getAngle())

    def processClamexFurtherAwayFromEdge(self: ClamexCAngleProcessor, surface: Surface, ncmacro: NCMacro2D) -> None:
        export = False

        if self.innerPlacementExport == 'all':
            export = True
            if not surface.finalFinishedSizeShape.isPolygonal():
                message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 88,
                    "Vertical Clamex is further away from any edge than the maximal distance. Be careful when producing. Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")
            else:
                message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 89,
                    "Vertical Clamex is on a Workpiece that is neither rectangular nor polygonal. Be careful when producing. Additionally the Clamex is further away from any polygonal edge than the maximal distance. Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")
        elif self.innerPlacementExport == 'polygonal':
            export = not surface.finalFinishedSizeShape.isPolygonal()
            if export:
                message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 88,
                    "Vertical Clamex is further away from any edge than the maximal distance. Be careful when producing. Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")
            else:
                message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 90,
                    "Skipping vertical Clamex on a Workpiece that is neither rectangular nor polygonal. Maximal distance from edge exceeded? Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")
        else:  # assumed 'none'
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 84,
                "Skipping vertical Clamex where the Clamex is further away from any polygonal edge than the maximal distance. Cabinet {cabinet}, Workpiece {workpiece}, Position {position}, C Angle {rotation}")

        self.logwarn(message, surface, ncmacro, ncmacro.cAngle)

        if export:
            ncmacro.clamexcangle = ncmacro.clamexcangle180 = ncmacro.cAngle  # As we are in the vertical case, clamexcangle180 does NOT get an extra +180°
            yield ncmacro

    def logwarn(self: ClamexCAngleProcessor, message: str, surface: Surface, ncmacro: NCMacro2D, rotation: float) -> None:
        if message:
            pioContainerName = ExportDataCollector.getCabinetUniqueName(surface.workpieceUUID)
            logging.warning(message.format(cabinet=pioContainerName, workpiece=surface.workpieceName, position=ncmacro.position, rotation=rotation))

    def processNonVerticalClamex(self: ClamexCAngleProcessor, ncmacro: NCMacro2D) -> Generator[NCMacro2D]:
        ncmacro.clamexcangle = ncmacro.cAngle
        ncmacro.clamexcangle180 = normalizeAngle(ncmacro.cAngle + 180.0)
        yield ncmacro

    def isCorrespondingSegmentInsideBounds(self: ClamexCAngleProcessor, finder: VerticalClamexEntryVectorFinder, surface: Surface) -> bool:
        bounds: Rectangle = surface.getFinishedSizeBounds()
        return any(bounds.isPointStrictlyInside(vertex.position) for vertex in finder.minimalDistantSegment.asTuple())
