# -*- coding: utf-8 -*-

# Copyright 2018, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import logging
import unittest.mock

from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.interface.surface.contour2d import Contour2D


class ContourApproachFilter:
    def process(self, surface):
        for contour in surface.contoursVertical:
            if not contour.areApproachDepartureValid():
                contour.forceApproachDepartureNotToBeAnArc()
                message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 58, "Contour without radius compensation, but with quarter arc approach or departure corrected to be linear! Workpiece: {workpiece}")
                logging.warning(message.format(workpiece=surface.workpieceName))


class Tests(unittest.TestCase):
    def setUp(self):
        self.filter = ContourApproachFilter()
        self.surface = unittest.mock.Mock()
        self.surface.contoursVertical = []

    def testNoContourShouldNotLogAnything(self):
        with unittest.mock.patch('logging.warning') as patcher:
            self.filter.process(self.surface)
            self.assertFalse(patcher.called)

    def testContourWithoutShouldNotTriggerWarning(self):
        contour = Contour2D()
        contour.approach = 'without'
        contour.departure = 'without'
        contour.radiusCompensation = Contour2D.RADIUS_COMPENSATION_NONE
        self.surface.contoursVertical.append(contour)

        with unittest.mock.patch('logging.warning') as patcher:
            self.filter.process(self.surface)
            self.assertFalse(patcher.called)

        expectedContour = Contour2D()
        expectedContour.approach = 'without'
        expectedContour.departure = 'without'
        expectedContour.radiusCompensation = Contour2D.RADIUS_COMPENSATION_NONE

        self.assertEqual([expectedContour], self.surface.contoursVertical)

    def testContourWithLinearShouldNotTriggerWarning(self):
        contour = Contour2D()
        contour.approach = 'linear blabla'
        contour.departure = 'linear blabla'
        contour.radiusCompensation = Contour2D.RADIUS_COMPENSATION_NONE
        self.surface.contoursVertical.append(contour)

        with unittest.mock.patch('logging.warning') as patcher:
            self.filter.process(self.surface)
            self.assertFalse(patcher.called)

        expectedContour = Contour2D()
        expectedContour.approach = 'linear blabla'
        expectedContour.departure = 'linear blabla'
        expectedContour.radiusCompensation = Contour2D.RADIUS_COMPENSATION_NONE

        self.assertEqual([expectedContour], self.surface.contoursVertical)

    def testContourWithArcShouldTriggerWarningAndBeCorrected(self):
        contour = Contour2D()
        contour.approach = 'arc blabla'
        contour.departure = 'arc blabla'
        contour.radiusCompensation = Contour2D.RADIUS_COMPENSATION_NONE
        self.surface.contoursVertical.append(contour)

        with unittest.mock.patch('logging.warning') as patcher:
            self.filter.process(self.surface)
            self.assertTrue(patcher.called)

        expectedContour = Contour2D()
        expectedContour.approach = 'linear blabla'
        expectedContour.departure = 'linear blabla'
        expectedContour.radiusCompensation = Contour2D.RADIUS_COMPENSATION_NONE

        self.assertEqual([expectedContour], self.surface.contoursVertical)
