# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import logging
import unittest.mock

from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP


class ContourToolFilter:
    def __init__(self, machineConfiguration):
        self.machineConfiguration = machineConfiguration

    def process(self, surface):
        validContours = list(self.checksAllContours(surface))
        surface.setAllContours(validContours)

    def checksAllContours(self, surface):
        for contour in surface.getAllContours():
            if contour.hasMachineOptions:
                yield from self.checkMultiContour(surface, contour)
            else:
                yield from self.checkSimpleContour(surface, contour)

    def checkSimpleContour(self, surface, contour):
        if contour.tool:
            yield contour
        elif contour.isVertical():
            self.informAboutUnsupportedVerticalContour(surface, contour)
        else:
            self.informAboutUnsupportedHorizontalContour(surface, contour)

    def checkMultiContour(self, surface, contour):
        if contour.tool:
            firstInvalidToolName = contour.getFirstInvalidToolName(self.machineConfiguration)
            if firstInvalidToolName is None:
                yield contour
            else:
                self.informAboutMissingToolName(surface, contour, firstInvalidToolName)
        else:
            self.informAboutMissingToolName(surface, contour, contour.multiUsageRoutings[0].toolName)  # We are always checking the main tool of the conotur. Therefore it is always 0

    def informAboutUnsupportedVerticalContour(self, surface, contour):
        message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 45, "Suppressing vertical contour, as a tool with the expected width and depth was not found! Workpiece: {workpiece}, Contour width: {width}, depth: {depth}")
        logging.warning(message.format(workpiece=surface.workpieceName, width=contour.width, depth=contour.depth))

    def informAboutUnsupportedHorizontalContour(self, surface, contour):
        message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 66, "Suppressing horizontal contour, as a tool with the expected width and depth was not found! Workpiece: {workpiece}, Contour width: {width}, depth: {depth}")
        logging.warning(message.format(workpiece=surface.workpieceName, width=contour.width, depth=contour.depth))

    def informAboutMissingToolName(self, surface, contour, toolName):
        message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 78, "Suppressing contour, as the tool name was not found! Workpiece: {workpiece}, Contour width: {width}, depth: {depth}, tool name: '{toolName}'")
        logging.warning(message.format(workpiece=surface.workpieceName, width=contour.width, depth=contour.depth, toolName=toolName))


class Tests(unittest.TestCase):
    class MockSurface(unittest.mock.Mock):
        def __init__(self, *args, **keywords):
            super().__init__(*args, **keywords)
            self.contours = []

        def getAllContours(self):
            return self.contours

        def setAllContours(self, machinings):
            self.contours = machinings

    def setUp(self):
        self.machineConfiguration = unittest.mock.Mock()
        self.filter = ContourToolFilter(self.machineConfiguration)
        self.surface = self.MockSurface()

    def testNoContourShouldNotLogAnything(self):
        with unittest.mock.patch('logging.warning') as patcher:
            self.filter.process(self.surface)
            self.assertFalse(patcher.called)

    def testContourWithoutToolShouldWarnAndDiscardContour(self):
        contour = unittest.mock.Mock()
        contour.tool = None
        contour.hasMachineOptions = False
        self.surface.contours.append(contour)

        with unittest.mock.patch('logging.warning') as patcher:
            self.filter.process(self.surface)
            self.assertTrue(patcher.called)

        self.assertEqual([], self.surface.getAllContours())

    def testContourWithToolShouldNotWarnAndBeKept(self):
        contour = unittest.mock.Mock()
        contour.tool = unittest.mock.Mock()
        contour.hasMachineOptions = False
        self.surface.contours.append(contour)

        with unittest.mock.patch('logging.warning') as patcher:
            self.filter.process(self.surface)
            self.assertFalse(patcher.called)

        self.assertEqual([contour], self.surface.getAllContours())
