# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import unittest

from interiorcad.interface.geometry.point import Point


class DadosSorter:
    def process(self, surface):
        surface.dadosVertical = list(itertools.chain(
            self.__sortByPositionAlongXAxis(dado for dado in surface.dadosVertical if dado.isAlongXAxis()),
            self.__sortByPositionAlongYAxis(dado for dado in surface.dadosVertical if dado.isAlongYAxis()),
            (dado for dado in surface.dadosVertical if self.__isFreeDado(dado)),
        ))

    def __isFreeDado(self, dado):
        return not dado.isAlongXAxis() and not dado.isAlongYAxis()

    def __sortByPositionAlongXAxis(self, dados):
        return sorted(dados, key=lambda dados: dados.getStartPoint().y)

    def __sortByPositionAlongYAxis(self, dados):
        return sorted(dados, key=lambda dados: dados.getStartPoint().x)


class SortDadosProcessorTests(unittest.TestCase):
    class MockFreeDado:
        def __init__(self):
            self.startPoint = Point(0, 0)

        def getStartPoint(self):
            return self.startPoint

        def isAlongXAxis(self):
            return False

        def isAlongYAxis(self):
            return False

    class MockDadoX(MockFreeDado):
        def isAlongXAxis(self):
            return True

    class MockDadoY(MockFreeDado):
        def isAlongYAxis(self):
            return True

    def setUp(self):
        self.processor = DadosSorter()
        self.surface = unittest.mock.Mock()
        self.surface.dadosVertical = []
        self.freeDado = self.MockFreeDado()
        self.dadoX = self.MockDadoX()
        self.dadoX2 = self.MockDadoX()
        self.dadoY = self.MockDadoY()
        self.dadoY2 = self.MockDadoY()

    def addDadosToSurface(self, *dados):
        self.surface.dadosVertical.extend(dados)

    def testSurfaceWithoutDadosShouldHaveNoDados(self):
        self.processor.process(self.surface)
        self.assertEqual([], self.surface.dadosVertical)

    def testSurfaceWithOneDadoShouldHaveOneDado(self):
        self.addDadosToSurface(self.freeDado)
        self.processor.process(self.surface)
        expectedDados = [self.freeDado]
        self.assertEqual(expectedDados, self.surface.dadosVertical)

    def testSurfaceWithTwoDadosYXShouldHaveTwoDadosXY(self):
        self.addDadosToSurface(self.dadoY, self.dadoX)
        self.processor.process(self.surface)
        expectedDados = [self.dadoX, self.dadoY]
        self.assertEqual(expectedDados, self.surface.dadosVertical)

    def testSurfaceWithTwoDadosXYShouldHaveTwoDadosXY(self):
        self.addDadosToSurface(self.dadoX, self.dadoY)
        self.processor.process(self.surface)
        expectedDados = [self.dadoX, self.dadoY]
        self.assertEqual(expectedDados, self.surface.dadosVertical)

    def testSurfaceWithThreeDadosFreeYX_ShouldHaveThreeDadosXYFree(self):
        self.addDadosToSurface(self.freeDado, self.dadoX, self.dadoY)
        self.processor.process(self.surface)
        expectedDados = [self.dadoX, self.dadoY, self.freeDado]
        self.assertEqual(expectedDados, self.surface.dadosVertical)

    def testSurfaceWithDadosXXShouldBeOrderedLeftToRight(self):
        self.dadoX.startPoint = Point(0, 0)
        self.dadoX2.startPoint = Point(0, 1)
        self.addDadosToSurface(self.dadoX2, self.dadoX)
        self.processor.process(self.surface)
        expectedDados = [self.dadoX, self.dadoX2]
        self.assertEqual(expectedDados, self.surface.dadosVertical)

    def testSurfaceWithDadosYYShouldBeOrderedBottomToTop(self):
        self.dadoY.startPoint = Point(0, 0)
        self.dadoY2.startPoint = Point(1, 0)
        self.addDadosToSurface(self.dadoY2, self.dadoY)
        self.processor.process(self.surface)
        expectedDados = [self.dadoY, self.dadoY2]
        self.assertEqual(expectedDados, self.surface.dadosVertical)
