# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import unittest.mock

from interiorcad.interface.geometry.point import Point


class RabbetsSorter:
    POSITIVE_X_AXIS = Point(1, 0)
    POSITIVE_Y_AXIS = Point(0, 1)
    NEGATIVE_X_AXIS = Point(-1, 0)
    NEGATIVE_Y_AXIS = Point(0, -1)

    def process(self, surface):
        surface.rabbetsVertical = list(itertools.chain(
            self.__sortByPosition([rabbet for rabbet in surface.rabbetsVertical if rabbet.directionVector == self.POSITIVE_X_AXIS]),
            self.__sortByPosition([rabbet for rabbet in surface.rabbetsVertical if rabbet.directionVector == self.POSITIVE_Y_AXIS]),
            self.__sortByPosition([rabbet for rabbet in surface.rabbetsVertical if rabbet.directionVector == self.NEGATIVE_X_AXIS]),
            self.__sortByPosition([rabbet for rabbet in surface.rabbetsVertical if rabbet.directionVector == self.NEGATIVE_Y_AXIS]),
            (rabbet for rabbet in surface.rabbetsVertical if self.__isFreeRabbet(rabbet)),
        ))

    def __isFreeRabbet(self, rabbet):
        return rabbet.directionVector not in (self.POSITIVE_X_AXIS, self.POSITIVE_Y_AXIS, self.NEGATIVE_X_AXIS, self.NEGATIVE_Y_AXIS)

    def __sortByPosition(self, rabbets):
        if rabbets:
            reverse = rabbets[0].directionVector in (self.NEGATIVE_X_AXIS, self.NEGATIVE_Y_AXIS)
            return sorted(rabbets, key=lambda rabbet: rabbet.getStartPoint().magnitude(), reverse=reverse)
        else:
            return ()


class SortRabbetsTests(unittest.TestCase):
    class MockRabbet:
        def __init__(self):
            self.startPoint = Point(0, 0)

        def getStartPoint(self):
            return self.startPoint

    class MockRabbetBack(MockRabbet):
        @property
        def directionVector(self):
            return Point(-1, 0)

    class MockRabbetFront(MockRabbet):
        @property
        def directionVector(self):
            return Point(1, 0)

    class MockRabbetLeft(MockRabbet):
        @property
        def directionVector(self):
            return Point(0, -1)

    class MockRabbetRight(MockRabbet):
        @property
        def directionVector(self):
            return Point(0, 1)

    class MockFreeRabbet(MockRabbet):
        @property
        def directionVector(self):
            return Point(0.5, 0.5)

    def setUp(self):
        self.surface = unittest.mock.Mock()
        self.surface.rabbetsVertical = []
        self.processor = RabbetsSorter()
        self.rabbetBack = self.MockRabbetBack()
        self.rabbetBack2 = self.MockRabbetBack()
        self.rabbetFront = self.MockRabbetFront()
        self.rabbetFront2 = self.MockRabbetFront()
        self.rabbetLeft = self.MockRabbetLeft()
        self.rabbetLeft2 = self.MockRabbetLeft()
        self.rabbetRight = self.MockRabbetRight()
        self.rabbetRight2 = self.MockRabbetRight()
        self.freeRabbet = self.MockFreeRabbet()

    def addRabbetsToSurface(self, *rabbets):
        self.surface.rabbetsVertical.extend(rabbets)

    def testSurfaceWithoutRabbetsShouldHaveNoRabbets(self):
        self.processor.process(self.surface)
        self.assertEqual([], self.surface.rabbetsVertical)

    def testSurfaceWithOneRabbetShouldHaveOneRabbet(self):
        self.addRabbetsToSurface(self.freeRabbet)
        self.processor.process(self.surface)
        expectedRabbets = [self.freeRabbet]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithBackAndFrontRabbetShouldSortToFrontAndBackRabbet(self):
        self.addRabbetsToSurface(self.rabbetBack, self.rabbetFront)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetFront, self.rabbetBack]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithFrontAndBackRabbetShouldSortToFrontAndBackRabbet(self):
        self.addRabbetsToSurface(self.rabbetFront, self.rabbetBack)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetFront, self.rabbetBack]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithLeftAndRightRabbetShouldSortToRightAndLeftRabbet(self):
        self.addRabbetsToSurface(self.rabbetLeft, self.rabbetRight)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetRight, self.rabbetLeft]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithFrontRightBackLeftRabbetShouldSortToFrontRightBackLeftRabbet(self):
        self.addRabbetsToSurface(self.rabbetFront, self.rabbetRight, self.rabbetBack, self.rabbetLeft)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetFront, self.rabbetRight, self.rabbetBack, self.rabbetLeft]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithFrontRightBackLeftFreeRabbetShouldSortFreeRabbetAtTheEnd(self):
        self.addRabbetsToSurface(self.rabbetFront, self.rabbetRight, self.rabbetBack, self.rabbetLeft, self.freeRabbet)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetFront, self.rabbetRight, self.rabbetBack, self.rabbetLeft, self.freeRabbet]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithTwoFrontRabbetsShouldSortRabbetsFromLeftToRight(self):
        self.rabbetFront.startPoint = Point(0, 0)
        self.rabbetFront2.startPoint = Point(1, 0)
        self.addRabbetsToSurface(self.rabbetFront2, self.rabbetFront)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetFront, self.rabbetFront2]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithTwoBackRabbetsShouldSortRabbetsFromRightToLeft(self):
        self.rabbetBack.startPoint = Point(0, 0)
        self.rabbetBack2.startPoint = Point(1, 0)
        self.addRabbetsToSurface(self.rabbetBack, self.rabbetBack2)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetBack2, self.rabbetBack]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithTwoLeftRabbetsShouldSortRabbetsFromTopToBottom(self):
        self.rabbetLeft.startPoint = Point(0, 0)
        self.rabbetLeft2.startPoint = Point(0, 1)
        self.addRabbetsToSurface(self.rabbetLeft, self.rabbetLeft2)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetLeft2, self.rabbetLeft]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)

    def testSurfaceWithTwoRightRabbetsShouldSortRabbetsFromBottomToTop(self):
        self.rabbetRight.startPoint = Point(0, 1)
        self.rabbetRight2.startPoint = Point(0, 0)
        self.addRabbetsToSurface(self.rabbetRight, self.rabbetRight2)
        self.processor.process(self.surface)
        expectedRabbets = [self.rabbetRight2, self.rabbetRight]
        self.assertEqual(expectedRabbets, self.surface.rabbetsVertical)
