# -*- coding: utf-8 -*-

# Copyright 2011-2014, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.backend.ncexportmigrated.ncradiussegment import NCRadiusSegment
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.polylinesegment import PolylineSegment
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.surface import Surface


class SizingContourProcessor:
    def process(self, surface):
        self.surface = surface
        if self.surface.hasSizing():
            self.splitRadiusSegmentWhenStartPointIsOfTypeRadius()

    def getFirstRadiusSegment(self):
        polyline = self.surface.sizing.polyline
        assert polyline.numVertices() > 2

        firstVertex = polyline.getVertex(polyline.numVertices() - 1)
        lastVertex = polyline.getVertex(1)
        radiusVertex = polyline.getVertex(0)

        return NCRadiusSegment(firstVertex, radiusVertex, lastVertex)

    def defineMidPointOfStartEndVertex(self, startVertex, endVertex):
        startEndVertexSegment = PolylineSegment(startVertex, endVertex)
        midPointOfStartEndVertexSegment = startEndVertexSegment.calculateMidPoint().position
        return midPointOfStartEndVertexSegment

    def normalizeVector(self, radiusVertex, midPointOfStartEndVertexSegment):
        vectorRadiusVertexToCenter = midPointOfStartEndVertexSegment - radiusVertex.position
        vectorRadiusVertexToCenter.normalize()
        return vectorRadiusVertexToCenter

    def defineRadiusVertexOfNewSegment(self, radiusVertex, circleCenterPoint, midPointOfPolylineSegment):
        vectorCircleCenterToMidPointStartRadiusVertex = midPointOfPolylineSegment - circleCenterPoint
        vectorCircleCenterToMidPointStartRadiusVertex.normalize()
        return circleCenterPoint + vectorCircleCenterToMidPointStartRadiusVertex * radiusVertex.radius

    def defineCircleCenterPointFromRadiusAndNormalizedVector(self, radiusVertex, vectorRadiusVertexToCenterNormalized):
        circleCenterPoint = radiusVertex.position + vectorRadiusVertexToCenterNormalized * radiusVertex.radius
        return circleCenterPoint

    def defineMidPointOfPolylineSegment(self, startVertex, radiusVertex):
        startRadiusVertexSegment = PolylineSegment(startVertex, radiusVertex)
        midPointOfStartRadiusVertexSegment = startRadiusVertexSegment.calculateMidPoint().position
        return midPointOfStartRadiusVertexSegment

    def createFirstRadiusSegment(self, startVertex, radiusVertex, circleCenterPoint):
        midPointOfStartRadiusVertexSegment = self.defineMidPointOfPolylineSegment(startVertex, radiusVertex)
        newFirstRadiusVertex = self.defineRadiusVertexOfNewSegment(radiusVertex, circleCenterPoint, midPointOfStartRadiusVertexSegment)
        firstRadiusSegment = NCRadiusSegment(CornerVertex(startVertex.position, True), RadiusVertex(newFirstRadiusVertex, True, radiusVertex.radius), CornerVertex(radiusVertex.position, True))
        return firstRadiusSegment

    def createSecondRadiusSegment(self, radiusVertex, endVertex, circleCenterPoint):
        midPointOfEndRadiusVertexSegment = self.defineMidPointOfPolylineSegment(endVertex, radiusVertex)
        newSecondRadiusVertex = self.defineRadiusVertexOfNewSegment(radiusVertex, circleCenterPoint, midPointOfEndRadiusVertexSegment)
        secondRadiusSegment = NCRadiusSegment(CornerVertex(radiusVertex.position, True), RadiusVertex(newSecondRadiusVertex, True, radiusVertex.radius), CornerVertex(endVertex.position, True))
        return secondRadiusSegment

    def defineCircleCenterPointFromStartEndVertex(self, startVertex, radiusVertex, endVertex):
        midPointOfStartEndVertexSegment = self.defineMidPointOfStartEndVertex(startVertex, endVertex)
        vectorRadiusVertexToCenterNormalized = self.normalizeVector(radiusVertex, midPointOfStartEndVertexSegment)
        circleCenterPoint = self.defineCircleCenterPointFromRadiusAndNormalizedVector(radiusVertex, vectorRadiusVertexToCenterNormalized)
        return circleCenterPoint

    def splitRadiusSegment(self, segmentToSplit):
        circleCenterPoint = self.defineCircleCenterPointFromStartEndVertex(segmentToSplit.startVertex, segmentToSplit.radiusVertex, segmentToSplit.endVertex)

        firstRadiusSegment = self.createFirstRadiusSegment(segmentToSplit.startVertex, segmentToSplit.radiusVertex, circleCenterPoint)
        secondRadiusSegment = self.createSecondRadiusSegment(segmentToSplit.radiusVertex, segmentToSplit.endVertex, circleCenterPoint)

        return firstRadiusSegment, secondRadiusSegment

    def splitRadiusSegmentWhenStartPointIsOfTypeRadius(self):
        polyline = self.surface.sizing.polyline
        if polyline:
            if not polyline.polylineStartsWithCornerVertex():
                self.removeLastRadiusPointIfEqualizesWithFirst()

                firstRadiusSegment = self.getFirstRadiusSegment()
                firstSegment, secondSegment = self.splitRadiusSegment(firstRadiusSegment)

                polyline.insertVertexAfter(0, secondSegment.radiusVertex)
                polyline.insertVertexAfter(polyline.numVertices() - 1, firstSegment.radiusVertex)  # ### IT-114: kann man die [polyline.numVertices() - 1] in [-1] umstellen? -- NG

                polyline.vertices[0] = secondSegment.startVertex

                if polyline.getVertex(0) != polyline.getVertex(-1):
                    polyline.addVertex(polyline.getVertex(0))

    def removeLastRadiusPointIfEqualizesWithFirst(self):
        polyline = self.surface.sizing.polyline
        firstVertex = polyline.vertices[0]
        lastVertex = polyline.vertices[-1]

        if firstVertex == lastVertex:
            del polyline.vertices[-1]


class SizingContourPreprarerTest(unittest.TestCase):
    def setUp(self):
        self.polylinePathStartsWithCorner = Polyline()
        self.polylinePathStartsWithCorner.addVertex(CornerVertex(Point(0, 0)))
        self.polylinePathStartsWithCorner.addVertex(CornerVertex(Point(100, 0)))
        self.polylinePathStartsWithCorner.addVertex(RadiusVertex(Point(200, 10)))
        self.polylinePathStartsWithCorner.addVertex(CornerVertex(Point(100, 100)))

        self.polylinePathStartsWithRadius = Polyline()
        self.polylinePathStartsWithRadius.addVertex(RadiusVertex(Point(200, 10), True, 150.0))
        self.polylinePathStartsWithRadius.addVertex(CornerVertex(Point(100, 100)))
        self.polylinePathStartsWithRadius.addVertex(CornerVertex(Point(100, 0)))
        self.polylinePathStartsWithRadius.addVertex(CornerVertex(Point(0, 0)))

        self.surface = Surface()
        self.sizingContourProcessor = SizingContourProcessor()
        self.sizingContourProcessor.surface = self.surface

    def testGetFirstRadiusSegmentFromPolyline(self):
        self.surface.sizing.polyline = self.polylinePathStartsWithRadius
        expectedSegment = NCRadiusSegment(CornerVertex(Point(0, 0), True), RadiusVertex(Point(200, 10), True, 150.0), CornerVertex(Point(100, 100), True))
        self.assertEqual(expectedSegment, self.sizingContourProcessor.getFirstRadiusSegment())

    def testSplitRadiusSegmentShouldCreateTwoNewRadiusSegments(self):
        self.sizingContourProcessor.surface = self.polylinePathStartsWithRadius

        segmentToSplit = NCRadiusSegment(CornerVertex(Point(0.0, 0.0), True), RadiusVertex(Point(75.0, 33.856217), True, 100.0), CornerVertex(Point(150.0, 0.0)))
        firstSegment, secondSegment = self.sizingContourProcessor.splitRadiusSegment(segmentToSplit)

        firstExpectedSegment = NCRadiusSegment(CornerVertex(Point(0.0, 0.0), True), RadiusVertex(Point(33.856217, 25), True, 100.0), CornerVertex(Point(75.0, 33.856217), True))
        secondExpectedSegment = NCRadiusSegment(CornerVertex(Point(75.0, 33.856217), True), RadiusVertex(Point(116.143783, 25), True, 100.0), CornerVertex(Point(150.0, 0.0), True))

        self.assertEqual(firstSegment, firstExpectedSegment)
        self.assertEqual(secondSegment, secondExpectedSegment)

    def testPolylineWithRadiusStartPointShouldSplitFirstRadiusPoint(self):
        polyline = Polyline()
        polyline.addVertex(RadiusVertex(Point(75.0, 33.856217), True, 100.0))
        polyline.addVertex(CornerVertex(Point(150.0, 0.0)))
        polyline.addVertex(CornerVertex(Point(0.0, 0.0), True))
        self.surface.sizing.polyline = polyline

        self.sizingContourProcessor.splitRadiusSegmentWhenStartPointIsOfTypeRadius()

        expectedPolyline = Polyline()
        expectedPolyline.addVertex(CornerVertex(Point(75.0, 33.856217), True))
        expectedPolyline.addVertex(RadiusVertex(Point(116.143783, 25), True, 100.0))
        expectedPolyline.addVertex(CornerVertex(Point(150.0, 0.0), True))
        expectedPolyline.addVertex(CornerVertex(Point(0.0, 0.0), True))
        expectedPolyline.addVertex(RadiusVertex(Point(33.856217, 25), True, 100.0))
        expectedPolyline.addVertex(CornerVertex(Point(75.0, 33.856217), True))

        self.assertEqual(expectedPolyline, self.sizingContourProcessor.surface.sizing.polyline)

    def testLastPointShouldBeRemovedWhenFirstAndLastPointIsEqualAndOfTypeRadius(self):
        polyline = Polyline()
        polyline.addVertex(RadiusVertex(Point(75.0, 33.856217), True, 100.0))
        polyline.addVertex(CornerVertex(Point(150.0, 0.0)))
        polyline.addVertex(CornerVertex(Point(0.0, 0.0), True))
        polyline.addVertex(RadiusVertex(Point(75.0, 33.856217), True, 100.0))
        self.surface.sizing.polyline = polyline

        self.sizingContourProcessor.removeLastRadiusPointIfEqualizesWithFirst()

        expectedPolyline = Polyline()
        expectedPolyline.addVertex(RadiusVertex(Point(75.0, 33.856217), True, 100.0))
        expectedPolyline.addVertex(CornerVertex(Point(150.0, 0.0)))
        expectedPolyline.addVertex(CornerVertex(Point(0.0, 0.0), True))

        self.assertEqual(expectedPolyline, self.sizingContourProcessor.surface.sizing.polyline)

    def testPolylineWithRadiusStartPointShouldSplitFirstRadiusPointWhenFirstAndSecondPointIsEqual(self):
        polyline = Polyline()
        polyline.addVertex(RadiusVertex(Point(75.0, 33.856217), True, 100.0))
        polyline.addVertex(CornerVertex(Point(150.0, 0.0)))
        polyline.addVertex(CornerVertex(Point(0.0, 0.0), True))
        polyline.addVertex(RadiusVertex(Point(75.0, 33.856217), True, 100.0))
        self.surface.sizing.polyline = polyline

        self.sizingContourProcessor.splitRadiusSegmentWhenStartPointIsOfTypeRadius()

        expectedPolyline = Polyline()
        expectedPolyline.addVertex(CornerVertex(Point(75.0, 33.856217), True))
        expectedPolyline.addVertex(RadiusVertex(Point(116.143783, 25), True, 100.0))
        expectedPolyline.addVertex(CornerVertex(Point(150.0, 0.0), True))
        expectedPolyline.addVertex(CornerVertex(Point(0.0, 0.0), True))
        expectedPolyline.addVertex(RadiusVertex(Point(33.856217, 25), True, 100.0))
        expectedPolyline.addVertex(CornerVertex(Point(75.0, 33.856217), True))

        self.assertEqual(expectedPolyline, self.sizingContourProcessor.surface.sizing.polyline)
