# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import logging
import unittest.mock

from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP


class SizingToolFilter:
    def __init__(self, machineConfiguration):
        self.machineConfiguration = machineConfiguration

    def process(self, surface):
        sizing = surface.sizing
        if sizing:
            if not sizing.tool:
                sizing.assignFallbackTool(surface, self.machineConfiguration)
                if sizing.tool:
                    self.informFallbackToolUsed(surface, sizing.tool)
                else:
                    self.informAboutUnsupportedSizing(surface)
                    surface.resetSizing()
            elif not surface.sizing.tool.exportMachinings:
                surface.resetSizing()
        elif sizing is not None:
            # If there is no sizing, you do not want a tool for it 8due to tool sorting) IC-22425
            sizing.tool = None
            sizing.toolsID = None

    def informFallbackToolUsed(self, surface, tool):
        message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 75, "The usable length {length} of the sizing router is shorter than the thickness {thickness} of the Workpiece {workpiece}. Using longest sizing router instead.")
        logging.warning(message.format(workpiece=surface.workpieceName, length=tool.getSortableUsableLength(), thickness=surface.thickness))

    def informAboutUnsupportedSizing(self, surface):
        message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 43, "Suppressing sizing, as no tool were found for sizings! Workpiece: {workpiece}")
        logging.warning(message.format(workpiece=surface.workpieceName))


class Tests(unittest.TestCase):
    def setUp(self):
        self.filter = SizingToolFilter(machineConfiguration=unittest.mock.Mock())
        self.surface = unittest.mock.Mock()

    def testNoSizingShouldNotLogAnything(self):
        with unittest.mock.patch('logging.warning') as patcher:
            self.surface.sizing = None
            self.filter.process(self.surface)
            self.assertFalse(patcher.called)

    def testSizingWithoutToolShouldWarn(self):
        with unittest.mock.patch('logging.warning') as patcher:
            self.surface.sizing = unittest.mock.Mock()
            self.surface.sizing.tool = None
            self.filter.process(self.surface)
            self.assertTrue(patcher.called)
            self.assertTrue(self.surface.resetSizing.called)
