# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import unittest

from interiorcad.backend.utilities.iterationtools import partition
from interiorcad.interface.geometry import isNearlyEqual, EPSILON_DIGITS
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.surface.drilling2d import Drilling2D
from interiorcad.interface.surface.griddrilling2D import GridDrilling2D


class DrillingGroupHelper:
    def couldBothDrillingsBePartOfAGridDrilling(self, left, right, expectedDeltaVector):
        return all((
            (right.position - left.position) == expectedDeltaVector,
            left.depth == right.depth,
        ))

    def processPotentialGridDrilling(self, candidateDrillings, verticalGridDrillings, verticalDrillings):
        if len(candidateDrillings) >= self.gridDrillingMinimumNumberOfDrillings:
            gridDrilling = GridDrilling2D.createFromSingleDrilling(candidateDrillings[0], gridNormalizedDirection=self.expectedNormalizedDeltaVector, gridDistance=self.gridDrillingDistance, countDrillings=len(candidateDrillings))
            verticalGridDrillings.append(gridDrilling)
        else:
            verticalDrillings.extend(candidateDrillings)
        candidateDrillings.clear()

    def groupIntoGridDrillingsAlongAxis(self, verticalGridDrillings, verticalDrillings):
        self.expectedDeltaVector = self.expectedNormalizedDeltaVector * self.gridDrillingDistance
        candidateDrillings = []
        for currentDrilling in self.sortedDrillings:
            if candidateDrillings and not self.couldBothDrillingsBePartOfAGridDrilling(candidateDrillings[-1], currentDrilling, self.expectedDeltaVector):
                self.processPotentialGridDrilling(candidateDrillings, verticalGridDrillings, verticalDrillings)
            candidateDrillings.append(currentDrilling)

        self.processPotentialGridDrilling(candidateDrillings, verticalGridDrillings, verticalDrillings)


class VerticalGridDrillingsFilter:
    def __init__(self, **keywords):
        self.gridDrillingXDistance = keywords['gridDrillingXDistance']
        self.gridDrillingXMinimumNumberOfDrillings = keywords['gridDrillingXMinimumNumberOfDrillings']
        self.gridDrillingYDistance = keywords['gridDrillingYDistance']
        self.gridDrillingYMinimumNumberOfDrillings = keywords['gridDrillingYMinimumNumberOfDrillings']
        self.xaxis = keywords['xaxis']
        self.yaxis = keywords['yaxis']
        self.xAxisGridDrillDiameters = keywords['xAxisGridDrillDiameters']
        self.yAxisGridDrillDiameters = keywords['yAxisGridDrillDiameters']

    def groupIntoGridDrillingsAlongXAxis(self, drillingsForY, verticalGridDrillings, verticalDrillings):
        helper = DrillingGroupHelper()
        helper.sortedDrillings = sorted(drillingsForY, key=lambda drilling: drilling.position.x)
        helper.expectedNormalizedDeltaVector = Vector(1.0, 0.0, 0.0)
        helper.gridDrillingDistance = self.gridDrillingXDistance
        helper.gridDrillingMinimumNumberOfDrillings = self.gridDrillingXMinimumNumberOfDrillings
        helper.groupIntoGridDrillingsAlongAxis(verticalGridDrillings, verticalDrillings)

    def groupIntoGridDrillingsAlongYAxis(self, drillingsForX, verticalGridDrillings, verticalDrillings):
        helper = DrillingGroupHelper()
        helper.sortedDrillings = sorted(drillingsForX, key=lambda drilling: drilling.position.y)
        helper.expectedNormalizedDeltaVector = Vector(0, 1)
        helper.gridDrillingDistance = self.gridDrillingYDistance
        helper.gridDrillingMinimumNumberOfDrillings = self.gridDrillingYMinimumNumberOfDrillings
        helper.groupIntoGridDrillingsAlongAxis(verticalGridDrillings, verticalDrillings)

    def partitionAccordingToDiameter(self, drillings, expectedDiameter):
        return partition(lambda drilling: isNearlyEqual(drilling.diameter, expectedDiameter), drillings)

    def processVerticalDrillingsAlongAxis(self, _surface, verticalDrillings, verticalGridDrillings, supportedDiameter, grouper, key):
        verticalDrillingsOfOtherDiameters, verticalDrillingsOfRightDiameters = self.partitionAccordingToDiameter(verticalDrillings, supportedDiameter)
        remainingVerticalDrillingsOfRightDiameter = []
        for _dummy, drillingsToProcess in itertools.groupby(sorted(verticalDrillingsOfRightDiameters, key=key), key=key):
            grouper(drillingsToProcess, verticalGridDrillings, remainingVerticalDrillingsOfRightDiameter)

        verticalDrillings[:] = list(itertools.chain(remainingVerticalDrillingsOfRightDiameter, verticalDrillingsOfOtherDiameters))

    def processVerticalDrillingsAlongXAxis(self, surface, verticalDrillings, verticalGridDrillings):
        def key(drilling):
            return drilling.zPosition, round(drilling.position.y, EPSILON_DIGITS)

        for diameter in self.xAxisGridDrillDiameters:
            self.processVerticalDrillingsAlongAxis(surface, verticalDrillings, verticalGridDrillings, diameter, self.groupIntoGridDrillingsAlongXAxis, key)

    def processVerticalDrillingsAlongYAxis(self, surface, verticalDrillings, verticalGridDrillings):
        def key(drilling):
            return drilling.zPosition, round(drilling.position.x, EPSILON_DIGITS)

        for diameter in self.yAxisGridDrillDiameters:
            self.processVerticalDrillingsAlongAxis(surface, verticalDrillings, verticalGridDrillings, diameter, self.groupIntoGridDrillingsAlongYAxis, key)

    def process(self, surface):
        nonDrillthrough, drillThrough = partition(lambda drilling: drilling.depth >= surface.thickness, surface.getVerticalDrillings())
        verticalGridDrillings = surface.verticalGridDrillings
        nonDrillthrough = list(nonDrillthrough)

        if self.xaxis:
            self.processVerticalDrillingsAlongXAxis(surface, nonDrillthrough, verticalGridDrillings)

        if self.yaxis:
            self.processVerticalDrillingsAlongYAxis(surface, nonDrillthrough, verticalGridDrillings)

        surface.setVerticalGridDrillings(verticalGridDrillings)
        surface.setVerticalDrillings(itertools.chain(drillThrough, nonDrillthrough))


class MockSurface:
    thickness = 20

    def __init__(self):
        self.verticalDrillings = []
        self.verticalGridDrillings = []

    def getVerticalDrillings(self):
        return self.verticalDrillings

    def setVerticalDrillings(self, drillings):
        self.verticalDrillings = drillings

    def setVerticalGridDrillings(self, gridDrillings):
        self.verticalGridDrillings = gridDrillings


class DrillingGroupHelperTests(unittest.TestCase):
    def setUp(self):
        self.helper = DrillingGroupHelper()
        self.helper.expectedNormalizedDeltaVector = Vector(1.0, 0.0, 0.0)
        self.helper.gridDrillingDistance = 32
        self.helper.gridDrillingMinimumNumberOfDrillings = 2
        self.verticalGridDrillings = []
        self.verticalDrillings = []

    def groupIntoGridDrillingsAlongAxis(self):
        assert self.inputDrillings == tuple(sorted(self.inputDrillings, key=lambda drilling: drilling.position.x)), "inputDrillings is not a correctly sorted tuple!"
        self.helper.sortedDrillings = self.inputDrillings
        self.helper.groupIntoGridDrillingsAlongAxis(self.verticalGridDrillings, self.verticalDrillings)

    def testOneDrillingShouldRemainOneDrilling(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
        )
        expectedVerticalDrillings = list(self.inputDrillings)

        self.groupIntoGridDrillingsAlongAxis()

        self.assertEqual(expectedVerticalDrillings, self.verticalDrillings)
        self.assertEqual([], self.verticalGridDrillings)

    def testTwoDrillings32mmApart(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=5, depth=10),
        )
        self.groupIntoGridDrillingsAlongAxis()

        expectedGridDrillings = [GridDrilling2D(position=Point(64, 100), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10)]
        self.assertEqual([], self.verticalDrillings)
        self.assertEqual(expectedGridDrillings, self.verticalGridDrillings)

    def testTwoDrillings32mmApartShouldNotBeAGridDrillingWith3DrillingsMinimum(self):
        self.helper.gridDrillingMinimumNumberOfDrillings = 3
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=5, depth=10),
        )
        expectedVerticalDrillings = list(self.inputDrillings)

        self.groupIntoGridDrillingsAlongAxis()

        self.assertEqual(expectedVerticalDrillings, self.verticalDrillings)
        self.assertEqual([], self.verticalGridDrillings)

    def testThreeDrillings32mmApart(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(128, 100), diameter=5, depth=10),
        )
        self.groupIntoGridDrillingsAlongAxis()

        expectedGridDrillings = [GridDrilling2D(position=Point(64, 100), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=3, diameter=5, depth=10)]
        self.assertEqual([], self.verticalDrillings)
        self.assertEqual(expectedGridDrillings, self.verticalGridDrillings)

    def testAnotherDepthShouldStopTheGridDrilling(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(128, 100), diameter=5, depth=7),
        )

        self.groupIntoGridDrillingsAlongAxis()

        expectedGridDrillings = [GridDrilling2D(position=Point(64, 100), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10)]

        self.assertEqual([Drilling2D.createFromTop(position=Point(128, 100), diameter=5, depth=7)], self.verticalDrillings)
        self.assertEqual(expectedGridDrillings, self.verticalGridDrillings)

    def testTwoGroupsOfTwoDrillingsShouldGiveTwoGridDrillings(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(32, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(64, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(100, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(132, 50), diameter=5, depth=10),
        )

        self.groupIntoGridDrillingsAlongAxis()

        expectedGridDrillings = [
            GridDrilling2D(position=Point(32, 50), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10),
            GridDrilling2D(position=Point(100, 50), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10),
        ]

        self.assertEqual([], self.verticalDrillings)
        self.assertEqual(expectedGridDrillings, self.verticalGridDrillings)

    def testASingleDrillingAtStartShouldRemainADrilling(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(10, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(32, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(64, 50), diameter=5, depth=10),
        )

        self.groupIntoGridDrillingsAlongAxis()

        expectedGridDrillings = [
            GridDrilling2D(position=Point(32, 50), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10),
        ]

        self.assertEqual([Drilling2D.createFromTop(position=Point(10, 50), diameter=5, depth=10)], self.verticalDrillings)
        self.assertEqual(expectedGridDrillings, self.verticalGridDrillings)

    def testASingleDrillingAtEndShouldRemainADrilling(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(32, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(64, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(100, 50), diameter=5, depth=10),
        )

        self.groupIntoGridDrillingsAlongAxis()

        expectedGridDrillings = [
            GridDrilling2D(position=Point(32, 50), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10),
        ]

        self.assertEqual([Drilling2D.createFromTop(position=Point(100, 50), diameter=5, depth=10)], self.verticalDrillings)
        self.assertEqual(expectedGridDrillings, self.verticalGridDrillings)

    def testASingleDrillingBetweenTwoGroupsOfDrillingsShouldRemainADrilling(self):
        self.inputDrillings = (
            Drilling2D.createFromTop(position=Point(32, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(64, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(80, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(100, 50), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(132, 50), diameter=5, depth=10),
        )

        self.groupIntoGridDrillingsAlongAxis()

        expectedGridDrillings = [
            GridDrilling2D(position=Point(32, 50), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10),
            GridDrilling2D(position=Point(100, 50), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=10),
        ]

        self.assertEqual([Drilling2D.createFromTop(position=Point(80, 50), diameter=5, depth=10)], self.verticalDrillings)
        self.assertEqual(expectedGridDrillings, self.verticalGridDrillings)


class AlongXAxis32mmTests(unittest.TestCase):
    def setUp(self):
        self.surface = MockSurface()
        self.filter = VerticalGridDrillingsFilter(
            gridDrillingXDistance=32,
            xAxisGridDrillDiameters=(5,),
            gridDrillingXMinimumNumberOfDrillings=2,
            gridDrillingYDistance=32,
            yAxisGridDrillDiameters=(5,),
            gridDrillingYMinimumNumberOfDrillings=2,
            xaxis=True,
            yaxis=False,
        )

    def testOneDrillingShouldRemainOneDrilling(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual([], self.surface.verticalGridDrillings)

    def testThreeDrillings32mmApart(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(128, 100), diameter=5, depth=10),
        )
        self.filter.process(self.surface)

        expectedGridDrillings = [GridDrilling2D(position=Point(64, 100), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=3, diameter=5, depth=10)]
        self.assertEqual((), tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testTwoDrillings32mmApartUnsupportedDiameter(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=8, depth=10),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=8, depth=10),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual([], self.surface.verticalGridDrillings)

    def testThreeDrillings32mmApartButDrillThroughShouldNotMakeAGridDrilling(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=20),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=5, depth=20),
            Drilling2D.createFromTop(position=Point(128, 100), diameter=5, depth=20),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        expectedGridDrillings = []
        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testThreeDrillings32mmApartButAlongYAxisShouldNotMakeAGridDrilling(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(100, 64), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(100, 96), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(100, 128), diameter=5, depth=10),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        expectedGridDrillings = []
        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testRealLifeSituationIC22799(self):  # IC-22799
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(239.33333334749977, 60.000842166988605), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(271.3333333474998, 60.000842166988605), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(303.33333334749955, 60.00084216698855), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(335.33333334749955, 60.00084216698855), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(367.33333334749955, 60.00084216698855), diameter=5, depth=10),

            Drilling2D.createFromTop(position=Point(239.33333334749977, 274.0008421669886), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(271.3333333474998, 274.0008421669886), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(303.33333334749955, 274.0008421669886), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(335.33333334749955, 274.0008421669886), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(367.33333334749955, 274.00084216698855), diameter=5, depth=10),
        )

        self.filter.process(self.surface)

        expectedVerticalDrillings = ()
        expectedGridDrillings = [
            GridDrilling2D(position=Point(239.33333334749977, 60.00084216698855), drillingDirection=Vector(0.0, 0.0, 1.0), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=5, diameter=5.0, depth=10.0),
            GridDrilling2D(position=Point(239.33333334749977, 274.00084216698855), drillingDirection=Vector(0.0, 0.0, 1.0), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=5, diameter=5.0, depth=10.0),
        ]

        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)


class AlongYAxis32mmTests(unittest.TestCase):
    def setUp(self):
        self.surface = MockSurface()
        self.filter = VerticalGridDrillingsFilter(
            gridDrillingXDistance=32,
            xAxisGridDrillDiameters=(5,),
            gridDrillingXMinimumNumberOfDrillings=2,
            gridDrillingYDistance=32,
            yAxisGridDrillDiameters=(5,),
            gridDrillingYMinimumNumberOfDrillings=2,
            xaxis=False,
            yaxis=True,
        )

    def testOneDrillingShouldRemainOneDrilling(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(100, 64), diameter=5, depth=10),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual([], self.surface.verticalGridDrillings)

    def testThreeDrillings32mmApart(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(100, 64), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(100, 96), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(100, 128), diameter=5, depth=10),
        )
        self.filter.process(self.surface)

        expectedGridDrillings = [GridDrilling2D(position=Point(100, 64), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=3, diameter=5, depth=10)]
        self.assertEqual((), tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testTwoDrillings32mmApartUnsupportedDiameter(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(100, 64), diameter=8, depth=10),
            Drilling2D.createFromTop(position=Point(100, 96), diameter=8, depth=10),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual([], self.surface.verticalGridDrillings)

    def testThreeDrillings32mmApartButDrillThroughShouldNotMakeAGridDrilling(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(100, 64), diameter=5, depth=20),
            Drilling2D.createFromTop(position=Point(100, 96), diameter=5, depth=20),
            Drilling2D.createFromTop(position=Point(100, 128), diameter=5, depth=20),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        expectedGridDrillings = []
        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testThreeDrillings32mmApartButAlongXAxisShouldNotMakeAGridDrilling(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(64, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(96, 100), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(128, 100), diameter=5, depth=10),
        )
        expectedVerticalDrillings = self.surface.verticalDrillings

        self.filter.process(self.surface)

        expectedGridDrillings = []
        self.assertEqual(expectedVerticalDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)


class OtherParametersForXThanForYTests(unittest.TestCase):
    def setUp(self):
        self.surface = MockSurface()
        self.filter = VerticalGridDrillingsFilter(
            gridDrillingXDistance=25,
            xAxisGridDrillDiameters=(8,),
            gridDrillingXMinimumNumberOfDrillings=2,
            gridDrillingYDistance=32,
            yAxisGridDrillDiameters=(5,),
            gridDrillingYMinimumNumberOfDrillings=2,
            xaxis=True,
            yaxis=True,
        )

    def testOneGridDrillinAlongXOneAlongY(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(25, 44), diameter=8, depth=10),
            Drilling2D.createFromTop(position=Point(100, 64), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(50, 44), diameter=8, depth=10),
            Drilling2D.createFromTop(position=Point(100, 96), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(75, 44), diameter=8, depth=10),
            Drilling2D.createFromTop(position=Point(100, 128), diameter=5, depth=10),
        )
        self.filter.process(self.surface)

        expectedGridDrillings = [
            GridDrilling2D(position=Point(25, 44), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=25, countDrillings=3, diameter=8, depth=10),
            GridDrilling2D(position=Point(100, 64), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=3, diameter=5, depth=10),
        ]

        self.assertEqual((), tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testOneGridDrillinAlongXWithAdditionalSingle(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(25, 44), diameter=8, depth=10),  # X
            Drilling2D.createFromTop(position=Point(50, 44), diameter=8, depth=10),  # X
            Drilling2D.createFromTop(position=Point(11, 500), diameter=8, depth=10),  # Single
            Drilling2D.createFromTop(position=Point(75, 44), diameter=8, depth=10),  # X
        )
        self.filter.process(self.surface)

        expectedGridDrillings = [
            GridDrilling2D(position=Point(25, 44), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=25, countDrillings=3, diameter=8, depth=10),
        ]

        expectedSingleDrillings = (
            Drilling2D.createFromTop(position=Point(11, 500), diameter=8, depth=10),
        )

        self.assertEqual(expectedSingleDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testOneGridDrillinAlongYWithAdditionalSingleDrillings(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(100, 64), diameter=5, depth=10),  # Y
            Drilling2D.createFromTop(position=Point(55, 77), diameter=5, depth=10),  # Single
            Drilling2D.createFromTop(position=Point(100, 96), diameter=5, depth=10),  # Y
            Drilling2D.createFromTop(position=Point(100, 128), diameter=5, depth=10),  # Y
        )
        self.filter.process(self.surface)

        expectedGridDrillings = [
            GridDrilling2D(position=Point(100, 64), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=3, diameter=5, depth=10),
        ]

        expectedSingleDrillings = (
            Drilling2D.createFromTop(position=Point(55, 77), diameter=5, depth=10),
        )

        self.assertEqual(expectedSingleDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)

    def testOneGridDrillinAlongXOneAlongYWithAdditionalSingleDrillings(self):
        self.surface.verticalDrillings = (
            Drilling2D.createFromTop(position=Point(25, 44), diameter=8, depth=10),  # X
            Drilling2D.createFromTop(position=Point(100, 64), diameter=5, depth=10),  # Y
            Drilling2D.createFromTop(position=Point(55, 77), diameter=5, depth=10),  # Single
            Drilling2D.createFromTop(position=Point(50, 44), diameter=8, depth=10),  # X
            Drilling2D.createFromTop(position=Point(100, 96), diameter=5, depth=10),  # Y
            Drilling2D.createFromTop(position=Point(11, 500), diameter=8, depth=10),  # Single
            Drilling2D.createFromTop(position=Point(75, 44), diameter=8, depth=10),  # X
            Drilling2D.createFromTop(position=Point(100, 128), diameter=5, depth=10),  # Y
        )
        self.filter.process(self.surface)

        expectedGridDrillings = [
            GridDrilling2D(position=Point(25, 44), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=25, countDrillings=3, diameter=8, depth=10),
            GridDrilling2D(position=Point(100, 64), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=3, diameter=5, depth=10),
        ]

        expectedSingleDrillings = (
            Drilling2D.createFromTop(position=Point(55, 77), diameter=5, depth=10),
            Drilling2D.createFromTop(position=Point(11, 500), diameter=8, depth=10),
        )

        self.assertEqual(expectedSingleDrillings, tuple(self.surface.verticalDrillings))
        self.assertEqual(expectedGridDrillings, self.surface.verticalGridDrillings)
