# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import math
import unittest

from interiorcad.backend.workpiece.complexgriddrilling import GridPointsCoordinateSystem
from interiorcad.interface.geometry import isNearlyEqual, isNearlyGreaterOrEqual
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.surface import VerticalDrilling, GridDrilling2D
from interiorcad.interface.surface.complexgriddrilling2D import ComplexGridDrilling2D


class VerticalGridDrillingsReplacer:
    def __init__(self, xaxis, xAxisGridDrillDiameters, yaxis, yAxisGridDrillDiameters, crotatedaxis, crotatedAxisGridDrillingDiameters):
        self.xaxis = xaxis
        self.xAxisGridDrillDiameters = xAxisGridDrillDiameters
        self.yaxis = yaxis
        self.yAxisGridDrillDiameters = yAxisGridDrillDiameters
        self.crotatedaxis = crotatedaxis
        self.crotatedAxisGridDrillingDiameters = crotatedAxisGridDrillingDiameters
        self.surfaceThickness = None

    def process(self, surface):
        self.surfaceThickness = surface.thickness
        surface.machinings = list(self.processAllMachinings(surface.machinings))
        self.surfaceThickness = None

    def processAllMachinings(self, machinings):
        for machining in machinings:
            yield from self.processMachining(machining)

    def processMachining(self, machining):
        if isinstance(machining, ComplexGridDrilling2D):
            yield from self.processGridDrilling(machining)
        else:
            yield machining

    def processGridDrilling(self, machining):
        def key(point):
            return math.fsum((machining.gridPointsCoordinateSystem.xaxis.y * point.x, machining.gridPointsCoordinateSystem.yaxis.y * point.y))

        gridNormalizedDirection = Vector(*machining.matrix.transform3D(machining.gridPointsCoordinateSystem.xaxis.x, machining.gridPointsCoordinateSystem.xaxis.y, 0.0)) - machining.matrix.origin
        if self.canExportGridDrilling(machining, gridNormalizedDirection):
            for _y, gridPoints in itertools.groupby(sorted(machining.gridPoints, key=key), key=key):
                yield from self.processLine(machining, tuple(gridPoints), gridNormalizedDirection)
        else:
            yield from self.createReplecementDrillings(machining, machining.gridPoints)

    def canExportGridDrilling(self, machining, gridNormalizedDirection):
        # TODO InnoWeek 2022-01: logging why the grid drilling must be ungrouped-- NG
        if machining.isVertical():
            if machining.drillThrough or isNearlyGreaterOrEqual(machining.depth, self.surfaceThickness):
                return False  # TODO InnoWeek 2022-01: support drill-through grid drillings  -- NG
            if gridNormalizedDirection == Vector(1.0, 0.0, 0.0) or gridNormalizedDirection == Vector(-1.0, 0.0, 0.0):
                return self.xaxis and machining.diameter in self.xAxisGridDrillDiameters
            elif gridNormalizedDirection == Vector(0.0, 1.0,  0.0) or gridNormalizedDirection == Vector(0.0, -1.0, 0.0):
                return self.yaxis and machining.diameter in self.yAxisGridDrillDiameters
            else:
                return self.crotatedaxis and machining.diameter in self.crotatedAxisGridDrillingDiameters

    def processLine(self, machining, gridPoints, gridNormalizedDirection):
        countGridPoints = len(gridPoints)
        if countGridPoints >= 2:
            # TODO InnoWeek 2022-01: cleaner code for drillings before and after the grid drilling -- NG
            middleIndex = countGridPoints // 2
            gridDistance = (gridPoints[middleIndex] - gridPoints[middleIndex - 1]).magnitude()

            for i in range(0, countGridPoints-1):
                if isNearlyEqual((gridPoints[i+1] - gridPoints[i]).magnitude(), gridDistance):
                    break

                yield self.createDrilling(machining, gridPoints[i])

            count = 1
            drillingsAfter = False
            for j in range(i, countGridPoints-1):
                if isNearlyEqual((gridPoints[j + 1] - gridPoints[j]).magnitude(), gridDistance):
                    count += 1
                else:
                    drillingsAfter = True
                    break

            gridDrilling = GridDrilling2D(position=gridPoints[i], drillingDirection=machining.matrix.w, gridNormalizedDirection=gridNormalizedDirection, gridDistance=gridDistance, countDrillings=count, diameter=machining.diameter, depth=machining.depth)
            gridDrilling.matrix = machining.matrix.getTranslatedBefore(gridPoints[i].x, gridPoints[i].y, 0)
            yield gridDrilling

            if drillingsAfter:
                for k in range(j+1, countGridPoints):
                    yield self.createDrilling(machining, gridPoints[k])

        else:
            yield from self.createReplecementDrillings(machining, gridPoints)

    def createReplecementDrillings(self, machining, gridPoints):
        for gridPoint in gridPoints:
            yield self.createDrilling(machining, gridPoint)

    def createDrilling(self, machining, gridPoint):
        drilling = VerticalDrilling.createFromTop(gridPoint, machining.diameter, machining.depth)
        drilling.matrix = machining.matrix.getTranslatedBefore(gridPoint.x, gridPoint.y, 0)
        drilling.drillThrough = machining.drillThrough
        if drilling.drillThrough:
            # The depth must be set for drill-through, as the code is used that single drillings hve a depth beimng the board thickness, already in teh Datatracker
            drilling.depth = self.surfaceThickness
        return drilling


class AlongXTests(unittest.TestCase):
    DIAMETER = 5.0
    DEPTH = 3.0

    def setUp(self):
        self.replacer = VerticalGridDrillingsReplacer(
            xaxis=True, xAxisGridDrillDiameters=(self.DIAMETER,),
            yaxis=True, yAxisGridDrillDiameters=(self.DIAMETER,),
            crotatedaxis=True, crotatedAxisGridDrillingDiameters=(self.DIAMETER,),
        )
        self.replacer.surfaceThickness = 19.0

        self.complexGridDrilling = ComplexGridDrilling2D()
        self.complexGridDrilling.diameter = self.DIAMETER
        self.complexGridDrilling.depth = self.DEPTH
        self.complexGridDrilling.gridPoints = []
        self.complexGridDrilling.gridPointsCoordinateSystem = GridPointsCoordinateSystem(
            xaxis=Point(1.0, 0.0),
            yaxis=Point(0.0, 1.0),
            origin=Point(0.0, 0.0),
        )
        self.complexGridDrilling.drillThrough = False

    def testOnePointShouldBecomeASingleDrilling(self):
        self.complexGridDrilling.gridPoints = [Point(1.0, 2.0)]

        expected = (VerticalDrilling.createFromTop(position=Point(1.0, 2.0), diameter=self.DIAMETER, depth=self.DEPTH),)

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testTwoPointsDifferingInYShouldBecomeTwoDrillings(self):
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(1.0, 22.0),
        ]

        expected = (
            VerticalDrilling.createFromTop(position=Point(1.0, 2.0), diameter=self.DIAMETER, depth=self.DEPTH),
            VerticalDrilling.createFromTop(position=Point(1.0, 22.0), diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testTwoPointsDifferingOnlyinXShouldBecomeOneGridDrilling(self):
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(11.0, 2.0),
        ]

        expected = (
            GridDrilling2D(position=Point(1.0, 2.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=2, diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testTThreePointsOfEqualDistanceDifferingOnlyinXShouldBecomeOneGridDrilling(self):
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(11.0, 2.0),
            Point(21.0, 2.0),
        ]

        expected = (
            GridDrilling2D(position=Point(1.0, 2.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=3, diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testThreePointsWhereOneDiffersInYShouldBecomeOneGridDrillingAndOneDrilling(self):
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(11.0, 2.0),
            Point(1.0, 22.0),
        ]

        expected = (
            GridDrilling2D(position=Point(1.0, 2.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=2, diameter=self.DIAMETER, depth=self.DEPTH),
            VerticalDrilling.createFromTop(position=Point(1.0, 22.0), diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testFourPointsGroupedByTwoInYShouldBecomeTwoGridDrillings(self):
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(11.0, 2.0),
            Point(1.0, 22.0),
            Point(11.0, 22.0),
        ]

        expected = (
            GridDrilling2D(position=Point(1.0, 2.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=2, diameter=self.DIAMETER, depth=self.DEPTH),
            GridDrilling2D(position=Point(1.0, 22.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=2, diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testAPointAfterThreePointsShouldBeExportedAsSeparatedDrilling(self):
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(11.0, 2.0),
            Point(21.0, 2.0),
            Point(111.0, 2.0),
        ]

        expected = (
            GridDrilling2D(position=Point(1.0, 2.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=3, diameter=self.DIAMETER, depth=self.DEPTH),
            VerticalDrilling.createFromTop(position=Point(111.0, 2.0), diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testAPointBeforeThreePointsShouldBeExportedAsSeparatedDrilling(self):
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(111.0, 2.0),
            Point(121.0, 2.0),
            Point(131.0, 2.0),
        ]

        expected = (
            VerticalDrilling.createFromTop(position=Point(1.0, 2.0), diameter=self.DIAMETER, depth=self.DEPTH),
            GridDrilling2D(position=Point(111.0, 2.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=3, diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testThreeRegularlySpacedPointsShouldBeExportedAsOneGridDrilling(self):
        # This test is to test that the algorith does not fail with only 3 points

        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(11.0, 2.0),
            Point(21.0, 2.0),
        ]

        expected = (
            GridDrilling2D(position=Point(1.0, 2.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=10, countDrillings=3, diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testThreeRegularlySpacedPointsInDrillThroughModeShouldBeExportedCorrectlyAsDrillThorughDrillings(self):
        def createDrillThroughDrilling(position):
            drilling = VerticalDrilling(position=position, diameter=self.DIAMETER, depth=self.replacer.surfaceThickness)
            drilling.drillThrough = True
            return drilling

        self.complexGridDrilling.drillThrough = True
        self.complexGridDrilling.gridPoints = [
            Point(1.0, 2.0),
            Point(11.0, 2.0),
            Point(21.0, 2.0),
        ]

        expected = (
            createDrillThroughDrilling(position=Point(1.0, 2.0)),
            createDrillThroughDrilling(position=Point(11.0, 2.0)),
            createDrillThroughDrilling(position=Point(21.0, 2.0)),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))


class AlongYTests(unittest.TestCase):
    DIAMETER = 5.0
    DEPTH = 3.0

    def setUp(self):
        self.replacer = VerticalGridDrillingsReplacer(
            xaxis=True, xAxisGridDrillDiameters=(self.DIAMETER,),
            yaxis=True, yAxisGridDrillDiameters=(self.DIAMETER,),
            crotatedaxis=True, crotatedAxisGridDrillingDiameters=(self.DIAMETER,),
        )
        self.replacer.surfaceThickness = 19.0

        self.complexGridDrilling = ComplexGridDrilling2D()
        self.complexGridDrilling.diameter = self.DIAMETER
        self.complexGridDrilling.depth = self.DEPTH
        self.complexGridDrilling.gridPoints = []
        self.complexGridDrilling.gridPointsCoordinateSystem = GridPointsCoordinateSystem(
            xaxis=Point(0.0, 1.0),
            yaxis=Point(1.0, 0.0),
            origin=Point(0.0, 0.0),
        )
        self.complexGridDrilling.drillThrough = False

    def testOnePointShouldBecomeASingleDrilling(self):
        self.complexGridDrilling.gridPoints = [Point(2.0, 1.0)]

        expected = (VerticalDrilling.createFromTop(position=Point(2.0, 1.0), diameter=self.DIAMETER, depth=self.DEPTH),)

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testTwoPointsDifferingInXShouldBecomeTwoDrillings(self):
        self.complexGridDrilling.gridPoints = [
            Point(2.0, 1.0),
            Point(22.0, 1.0),
        ]

        expected = (
            VerticalDrilling.createFromTop(position=Point(2.0, 1.0), diameter=self.DIAMETER, depth=self.DEPTH),
            VerticalDrilling.createFromTop(position=Point(22.0, 1.0), diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))

    def testTwoPointsDifferingOnlyinYShouldBecomeOneGridDrilling(self):
        self.complexGridDrilling.gridPoints = [
            Point(2.0, 1.0),
            Point(2.0, 11.0),
        ]

        expected = (
            GridDrilling2D(position=Point(2.0, 1.0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0.0, 1.0, 0.0), gridDistance=10, countDrillings=2, diameter=self.DIAMETER, depth=self.DEPTH),
        )

        self.assertEqual(expected, tuple(self.replacer.processMachining(self.complexGridDrilling)))
