# -*- coding: utf-8 -*-

# Copyright 2015, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest
import copy

from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector


class EntryCriteriaSorter:
    ASCENDING = 0
    DESCENDING = 1

    def __init__(self, criteria=None):
        self.criteria = criteria if criteria is not None else []

    def sort(self, entries):
        if self.criteria:
            for criterium in reversed(self.criteria):
                sortReversed = criterium['order'] if 'order' in criterium else self.ASCENDING
                if 'entry' in criterium:
                    entries.sort(key=criterium['entry'], reverse=sortReversed)
                elif 'key' in criterium:
                    if 'type' not in criterium or criterium['type'] == 'string':
                        getter = ExportDataCollector.getRawDataAsString
                    elif criterium['type'] == 'float':
                        getter = ExportDataCollector.getRawDataAsFloat
                    elif criterium['type'] == 'int':
                        getter = ExportDataCollector.getRawDataAsInt
                    else:
                        getter = ExportDataCollector.getRawDataAsString

                    entries.sort(key=lambda entry: getter(entry.uuid, criterium['key']))


class SortingTests(unittest.TestCase):
    def setUp(self):
        self.data = [
            {'pet': 'Cat', 'flower': 'Rose', 'wood': 'Beech', 'integer': 4},
            {'pet': 'Parrot', 'flower': 'Tulip', 'wood': 'Walnut', 'integer': 9},
            {'pet': 'Dog', 'flower': 'Lily', 'wood': 'Pine', 'integer': 3},
            {'pet': 'Mouse', 'flower': 'Lily', 'wood': 'Oak', 'integer': 6},
        ]

    def testWithoutSortingCriteriaNoSortingShouldBeDone(self):
        expectedData = copy.copy(self.data)
        sorter = EntryCriteriaSorter()
        sorter.sort(self.data)
        self.assertEqual(expectedData, self.data)

    def testSortInAscendingOrderFirstColumn(self):
        expectedData = [
            {'pet': 'Cat', 'flower': 'Rose', 'wood': 'Beech', 'integer': 4},
            {'pet': 'Dog', 'flower': 'Lily', 'wood': 'Pine', 'integer': 3},
            {'pet': 'Mouse', 'flower': 'Lily', 'wood': 'Oak', 'integer': 6},
            {'pet': 'Parrot', 'flower': 'Tulip', 'wood': 'Walnut', 'integer': 9},
        ]
        criteria = (
            {'entry': lambda entry: entry['pet'], 'order': EntryCriteriaSorter.ASCENDING},
        )

        sorter = EntryCriteriaSorter(criteria)
        sorter.sort(self.data)
        self.assertEqual(expectedData, self.data)

    def testLesserSignShouldSortInAscendingOrder(self):
        expectedData = [
            {'pet': 'Parrot', 'flower': 'Tulip', 'wood': 'Walnut', 'integer': 9},
            {'pet': 'Mouse', 'flower': 'Lily', 'wood': 'Oak', 'integer': 6},
            {'pet': 'Dog', 'flower': 'Lily', 'wood': 'Pine', 'integer': 3},
            {'pet': 'Cat', 'flower': 'Rose', 'wood': 'Beech', 'integer': 4},
        ]

        criteria = (
            {'entry': lambda entry: entry['pet'], 'order': EntryCriteriaSorter.DESCENDING},
        )

        sorter = EntryCriteriaSorter(criteria)
        sorter.sort(self.data)
        self.assertEqual(expectedData, self.data)

    def testSortingTwoColumnsShouldSortInAscendingOrder(self):
        expectedData = [
            {'pet': 'Dog', 'flower': 'Lily', 'wood': 'Pine', 'integer': 3},
            {'pet': 'Mouse', 'flower': 'Lily', 'wood': 'Oak', 'integer': 6},
            {'pet': 'Cat', 'flower': 'Rose', 'wood': 'Beech', 'integer': 4},
            {'pet': 'Parrot', 'flower': 'Tulip', 'wood': 'Walnut', 'integer': 9},
        ]
        criteria = (
            {'entry': lambda entry: entry['flower'], 'order': EntryCriteriaSorter.ASCENDING},
            {'entry': lambda entry: entry['pet'], 'order': EntryCriteriaSorter.ASCENDING},
        )

        sorter = EntryCriteriaSorter(criteria)
        sorter.sort(self.data)
        self.assertEqual(expectedData, self.data)

    def testSortingTwoColumnsOneAscendignOneDescendingShouldSortAsExpected(self):
        expectedData = [
            {'pet': 'Mouse', 'flower': 'Lily', 'wood': 'Oak', 'integer': 6},
            {'pet': 'Dog', 'flower': 'Lily', 'wood': 'Pine', 'integer': 3},
            {'pet': 'Cat', 'flower': 'Rose', 'wood': 'Beech', 'integer': 4},
            {'pet': 'Parrot', 'flower': 'Tulip', 'wood': 'Walnut', 'integer': 9},
        ]
        criteria = (
            {'entry': lambda entry: entry['flower'], 'order': EntryCriteriaSorter.ASCENDING},
            {'entry': lambda entry: entry['pet'], 'order': EntryCriteriaSorter.DESCENDING},
        )

        sorter = EntryCriteriaSorter(criteria)
        sorter.sort(self.data)
        self.assertEqual(expectedData, self.data)
