# -*- coding: utf-8 -*-

# Copyright 2011-2018, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest

from interiorcad.backend.surfacecreator.surfacesidedecider import SurfaceSideDecider
from interiorcad.backend.workpiececreationstrategy.creator import WorkpieceCreationStrategy


class Surfaces:
    def __init__(self):
        self.main = None
        self.opposite = None
        self.sizing = None

    def hasMain(self):
        return self.main is not None

    def hasOpposite(self):
        return self.opposite is not None

    def hasSizing(self):
        return self.sizing is not None

    def setMain(self, main):
        self.main = main

    def setOpposite(self, opposite):
        self.opposite = opposite

    def setSizing(self, sizing):
        self.sizing = sizing


class SurfaceCreationStrategy:
    def __init__(self, surfaceSideDecider, surfaceCreator):
        self.surfaceSideDecider = surfaceSideDecider
        self.surfaceCreator = surfaceCreator
        self.surfaceOrigin = surfaceCreator.options.surfaceOrigin

    def create(self):
        surfaces = Surfaces()
        if self.surfaceSideDecider.needsMain():
            surfaces.setMain(self.surfaceCreator.createMain())
        if self.surfaceSideDecider.needsOpposite():
            surfaces.setOpposite(self.surfaceCreator.createOpposite())
        if self.surfaceSideDecider.needsSizing():
            surfaces.setSizing(self.surfaceCreator.createSizing())
        return surfaces


class MockSurfaceCreator(unittest.mock.Mock):
    def createMain(self):
        return True

    def createOpposite(self):
        return True

    def createSizing(self):
        return True


class MockWorkpieceAnalyzer:
    def __init__(self):
        self.setHasNone()

    def hasMachiningsFromTop(self):
        return self.hasFromTop

    def hasMachiningsFromBottom(self):
        return self.hasFromBottom

    def hasMachiningsHorizontal(self):
        return self.hasHorizontal

    def setHasNone(self):
        self.hasFromTop = False
        self.hasFromBottom = False
        self.hasHorizontal = False

    def setHasFromTop(self):
        self.hasFromTop = True

    def setHasFromBottom(self):
        self.hasFromBottom = True

    def setHasHorizontal(self):
        self.hasHorizontal = True

    def setHasAll(self):
        self.hasFromTop = True
        self.hasFromBottom = True
        self.hasHorizontal = True


class SurfaceCreationStrategyTests(unittest.TestCase):
    class MockRectangularWorkpiece:
        def isRectangular(self):
            return True

    class MockConfiguration:
        pass

    def createPreSurfaceCreator(self, workpieceAnalyzer, separateSizing):
        configuration = self.MockConfiguration()
        configuration.accurateSaw = False
        configuration.separateSizing = separateSizing
        workpiece = self.MockRectangularWorkpiece()
        strategy = WorkpieceCreationStrategy.createSizingStrategy(workpiece, configuration)
        sideDecider = SurfaceSideDecider(workpiece, strategy, shouldAllwaysCreateMainSide=False, allowHorizontalMachiningsOnMainSide=True)
        sideDecider.setWorkpieceAnalyzer(workpieceAnalyzer)
        surfaceCreator = MockSurfaceCreator()
        creator = SurfaceCreationStrategy(sideDecider, surfaceCreator)
        return creator


class SurfaceCreationStrategyTests_NoMachinings(SurfaceCreationStrategyTests):
    class MockWorkpieceAnalyzer_NoMachinings(MockWorkpieceAnalyzer):
        def __init__(self):
            super().__init__()

    def createPreSurfaceCreator_NoMachinings_SeparateSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_NoMachinings(), separateSizing=True)

    def createPreSurfaceCreator_NoMachinings_InlineSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_NoMachinings(), separateSizing=False)

    def test_AccurateSaw_SeparateSizing_CreateSizingAndCuttinglist(self):
        creator = self.createPreSurfaceCreator_NoMachinings_SeparateSizing()
        preSurfaces = creator.create()
        self.assertFalse(preSurfaces.hasMain())
        self.assertFalse(preSurfaces.hasOpposite())
        self.assertTrue(preSurfaces.hasSizing())

    def test_AccurateSaw_InlineSizing_CreateMainAndCuttinglist(self):
        creator = self.createPreSurfaceCreator_NoMachinings_InlineSizing()
        preSurfaces = creator.create()
        self.assertTrue(preSurfaces.hasMain())
        self.assertFalse(preSurfaces.hasOpposite())
        self.assertFalse(preSurfaces.hasSizing())


class SurfaceCreationStrategyTests_MachiningsFromBottom(SurfaceCreationStrategyTests):
    class MockWorkpieceAnalyzer_MachiningsFromBottom(MockWorkpieceAnalyzer):
        def __init__(self):
            super().__init__()
            self.setHasFromBottom()

    def createPreSurfaceCreator_MachiningFromBottom_InlineSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsFromBottom(), separateSizing=False)

    def createPreSurfaceCreator_MachiningFromBottom_SeparateSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsFromBottom(), separateSizing=True)

    def test_AccurateSaw_InlineSizing_CreateOppositeAndCuttinglist(self):
        creator = self.createPreSurfaceCreator_MachiningFromBottom_InlineSizing()
        preSurfaces = creator.create()
        self.assertFalse(preSurfaces.hasMain())
        self.assertTrue(preSurfaces.hasOpposite())
        self.assertFalse(preSurfaces.hasSizing())

    def test_AccurateSaw_SeparateSizing_CreateOppositeAndSizingAndCuttinglist(self):
        creator = self.createPreSurfaceCreator_MachiningFromBottom_SeparateSizing()
        preSurfaces = creator.create()
        self.assertFalse(preSurfaces.hasMain())
        self.assertTrue(preSurfaces.hasOpposite())
        self.assertTrue(preSurfaces.hasSizing())


class SurfaceCreationStrategyTests_MachiningsFromTop(SurfaceCreationStrategyTests):
    class MockWorkpieceAnalyzer_MachiningsFromTop(MockWorkpieceAnalyzer):
        def __init__(self):
            super().__init__()
            self.setHasFromTop()

    def createSurfaceCreationStrategy_MachiningFromTop_InlineSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsFromTop(), separateSizing=False)

    def createSurfaceCreationStrategy_MachiningFromTop_SeparateSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsFromTop(), separateSizing=True)

    def test_AccurateSaw_InlineSizing_CreateOppositeAndCuttinglist(self):
        creator = self.createSurfaceCreationStrategy_MachiningFromTop_InlineSizing()
        preSurfaces = creator.create()
        self.assertTrue(preSurfaces.hasMain())
        self.assertFalse(preSurfaces.hasOpposite())
        self.assertFalse(preSurfaces.hasSizing())

    def test_AccurateSaw_SeparateSizing_CreateOppositeAndSizingAndCuttinglist(self):
        creator = self.createSurfaceCreationStrategy_MachiningFromTop_SeparateSizing()
        preSurfaces = creator.create()
        self.assertTrue(preSurfaces.hasMain())
        self.assertFalse(preSurfaces.hasOpposite())
        self.assertTrue(preSurfaces.hasSizing())


class SurfaceCreationStrategyTests_MachiningsFromTopAndBottom(SurfaceCreationStrategyTests):
    class MockWorkpieceAnalyzer_MachiningsFromTopAndBottom(MockWorkpieceAnalyzer):
        def __init__(self):
            super().__init__()
            self.setHasFromTop()
            self.setHasFromBottom()

    def createSurfaceCreationStrategy_MachiningFromTopAndBottom_InlineSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsFromTopAndBottom(), separateSizing=False)

    def createSurfaceCreationStrategy_MachiningFromTopAndBottom_SeparateSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsFromTopAndBottom(), separateSizing=True)

    def test_AccurateSaw_InlineSizing_CreateOppositeAndCuttinglist(self):
        creator = self.createSurfaceCreationStrategy_MachiningFromTopAndBottom_InlineSizing()
        preSurfaces = creator.create()
        self.assertTrue(preSurfaces.hasMain())
        self.assertTrue(preSurfaces.hasOpposite())
        self.assertFalse(preSurfaces.hasSizing())

    def test_AccurateSaw_SeparateSizing_CreateOppositeAndSizingAndCuttinglist(self):
        creator = self.createSurfaceCreationStrategy_MachiningFromTopAndBottom_SeparateSizing()
        preSurfaces = creator.create()
        self.assertTrue(preSurfaces.hasMain())
        self.assertTrue(preSurfaces.hasOpposite())
        self.assertTrue(preSurfaces.hasSizing())


class SurfaceCreationStrategyTests_MachiningsHorizontal(SurfaceCreationStrategyTests):
    class MockWorkpieceAnalyzer_MachiningsHorizontal(MockWorkpieceAnalyzer):
        def __init__(self):
            super().__init__()
            self.setHasHorizontal()

    def createSurfaceCreationStrategy_MachiningHorizontal_InlineSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsHorizontal(), separateSizing=False)

    def createSurfaceCreationStrategy_MachiningHorizontal_SeparateSizing(self):
        return self.createPreSurfaceCreator(self.MockWorkpieceAnalyzer_MachiningsHorizontal(), separateSizing=True)

    def test_AccurateSaw_InlineSizing_CreateOppositeAndCuttinglist(self):
        creator = self.createSurfaceCreationStrategy_MachiningHorizontal_InlineSizing()
        preSurfaces = creator.create()
        self.assertTrue(preSurfaces.hasMain())
        self.assertFalse(preSurfaces.hasOpposite())
        self.assertFalse(preSurfaces.hasSizing())

    def test_AccurateSaw_SeparateSizing_CreateOppositeAndSizingAndCuttinglist(self):
        creator = self.createSurfaceCreationStrategy_MachiningHorizontal_SeparateSizing()
        preSurfaces = creator.create()
        self.assertTrue(preSurfaces.hasMain())
        self.assertFalse(preSurfaces.hasOpposite())
        self.assertTrue(preSurfaces.hasSizing())
