# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import operator

from interiorcad.interface.surface import Surface
from interiorcad.interface.surface.origins import isOriginBack, isOriginRight
from interiorcad.interface.surface.shape2d import Shape2D


class BaseSurfaceCreator:
    def __init__(self, reader, dimensionCreator, sizingCreator, postProcessor, options):
        self.reader = reader
        self.dimensionCreator = dimensionCreator
        self.sizingCreator = sizingCreator
        self.postProcessor = postProcessor
        self.options = options

    def create(self):
        surface = Surface()
        self.writeBaseData(surface)
        self.writeDimensionData(surface)
        self.writeSizingData(surface)
        self.writeMachinings(surface)
        self.getEdges(surface)
        self.mirrorSurface(surface)
        self.grainConsiderationMode(surface)
        self.postProcessor.process(surface)
        return surface

    def writeBaseData(self, surface):
        surface.thickness = self.reader.getThickness()
        surface.workpieceName = self.reader.getName()
        surface.workpieceUUID = self.reader.getUUID()
        surface.count = self.reader.getCount()
        surface.grain = self.reader.getGrain()
        surface.sideType = self.getSideType()

    def writeDimensionData(self, surface):
        dimensionData = self.getDimensionData()
        surface.sawnSize = dimensionData.size
        surface.finishedSizeAddition = dimensionData.additionForEdges
        surface.sizingAddition = dimensionData.sizingAddition

    def writeSizingData(self, surface):
        sizingData = self.getSizingData()
        surface.setSizing(sizingData.sizing, sizingData.sizingBounds)

        # we can us copy.deepcopy directly, because we have implemented Polyline.__deepcopy__
        surface.finalFinishedSizeShape = Shape2D(polyline=copy.deepcopy(self.reader.getFinishedSizeShape().polyline))
        surface.finishedSizeSizing = self.reader.getFinishedSizeSizing()

    def writeMachinings(self, surface):
        pass

    def getEdges(self, surface):
        surface.edges = copy.deepcopy(self.reader.getEdges())

    def mirrorSurface(self, surface):
        if isOriginBack(self.options.surfaceOrigin):
            surface.mirrorOnXAxis()

        if operator.xor(operator.xor(self.reader.isMirrored(), isOriginRight(self.options.surfaceOrigin)), self.isOppositeSide()):
            surface.mirrorOnYAxis()

        if self.isOppositeSide():
            surface.mirrorOnXYPlane()

    def grainConsiderationMode(self, surface):
        if not self.isOppositeSide() and self.reader.isGrainCrosswise():
            if self.options.grainConsiderationMode == 'rotate-left':
                surface.rotateLeft()
            elif self.options.grainConsiderationMode == 'rotate-right':
                surface.rotateRight()
