# -*- coding: utf-8 -*-

# Copyright 2011-2018, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest

from interiorcad.backend.workpiece.machininganalyzer import MachiningsOnWorkpieceGetter


class SurfaceSideDecider:
    def __init__(self, workpiece, strategy, shouldAllwaysCreateMainSide, allowHorizontalMachiningsOnMainSide):
        self.workpiece = workpiece
        self.workpieceAnalyzer = MachiningsOnWorkpieceGetter(workpiece)
        self.strategy = strategy
        self.shouldAllwaysCreateMainSide = shouldAllwaysCreateMainSide
        self.allowHorizontalMachiningsOnMainSide = allowHorizontalMachiningsOnMainSide

    def __hasMachiningsFromBottom(self):
        return self.workpieceAnalyzer.hasMachiningsFromBottom()

    def __shouldHorizontalMachiningsBeOnMain(self):
        return self.allowHorizontalMachiningsOnMainSide and any((
            not self.__hasMachiningsFromBottom(),
            self.workpiece.isRectangular() and self.workpieceAnalyzer.hasMachiningsFromTop(),
        ))

    def needsMain(self):
        return any((
            self.shouldAllwaysCreateMainSide,
            self.workpieceAnalyzer.hasMachiningsFromTop(),
            self.workpieceAnalyzer.hasMachiningsHorizontal() and self.__shouldHorizontalMachiningsBeOnMain(),
            not self.needsOpposite() and not self.needsSizing(),
        ))

    def needsOpposite(self):
        return any((
            self.__hasMachiningsFromBottom(),
            self.needsHorizontalMachiningsOnOpposite(),
        ))

    def needsSizing(self):
        return self.strategy.needsSizing()

    def needsHorizontalMachiningsOnMain(self):
        return self.workpieceAnalyzer.hasMachiningsHorizontal() and self.__shouldHorizontalMachiningsBeOnMain()

    def needsHorizontalMachiningsOnOpposite(self):
        return self.workpieceAnalyzer.hasMachiningsHorizontal() and not self.__shouldHorizontalMachiningsBeOnMain()

    def setWorkpieceAnalyzer(self, workpieceAnalyzer):
        self.workpieceAnalyzer = workpieceAnalyzer


class SurfaceSideDeciderTests(unittest.TestCase):
    class MockAnalyzer_NoMachinings:
        def hasMachiningsFromTop(self):
            return False

        def hasMachiningsFromBottom(self):
            return False

        def hasMachiningsHorizontal(self):
            return False

    class MockRectangularWorkpiece:
        def isRectangular(self):
            return True

    class InlineSizingStrategy:
        def needsSizing(self):
            return False

    class SeparateSizingStrategy:
        def needsSizing(self):
            return True

    def createDefaultDecider(self, strategy, shouldAllwaysCreateMainSide=False, allowHorizontalMachiningsOnMainSide=True):
        return SurfaceSideDecider(workpiece=self.MockRectangularWorkpiece(), strategy=strategy, shouldAllwaysCreateMainSide=shouldAllwaysCreateMainSide, allowHorizontalMachiningsOnMainSide=allowHorizontalMachiningsOnMainSide)

    def createDecider_NoMachinings(self, strategy):
        decider = self.createDefaultDecider(strategy)
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_NoMachinings())
        return decider

    def testNoMachiningsWithoutSeperateSizing_ShouldHaveOnlyMainSurface(self):
        decider = self.createDecider_NoMachinings(self.InlineSizingStrategy())
        self.assertTrue(decider.needsMain())
        self.assertFalse(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testNoMachiningsWithSeperateSizing_ShouldHaveOnlySizingSurface(self):
        decider = self.createDecider_NoMachinings(self.SeparateSizingStrategy())
        self.assertFalse(decider.needsMain())
        self.assertFalse(decider.needsOpposite())
        self.assertTrue(decider.needsSizing())

    class MockAnalyzer_OnlyMachiningsFromTop(MockAnalyzer_NoMachinings):
        def hasMachiningsFromTop(self):
            return True

    class MockAnalyzer_OnlyMachiningsFromBottom(MockAnalyzer_NoMachinings):
        def hasMachiningsFromBottom(self):
            return True

    class MockAnalyzer_OnlyHorizontalMachinings(MockAnalyzer_NoMachinings):
        def hasMachiningsHorizontal(self):
            return True

    class MockAnalyzer_OnlyMachiningsFromTopAndHorizontal(MockAnalyzer_NoMachinings):
        def hasMachiningsFromTop(self):
            return True

        def hasMachiningsHorizontal(self):
            return True

    def testMachiningsFromTopWithoutSeperateSizing_ShouldHaveOnlyMainSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy())
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyMachiningsFromTop())
        self.assertTrue(decider.needsMain())
        self.assertFalse(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testMachiningsFromTopWithSeperateSizing_ShouldHaveMainAndSizingSurface(self):
        decider = self.createDefaultDecider(self.SeparateSizingStrategy())
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyMachiningsFromTop())
        self.assertTrue(decider.needsMain())
        self.assertFalse(decider.needsOpposite())
        self.assertTrue(decider.needsSizing())

    def testMachiningsFromBottomShouldHaveOppositeSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy())
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyMachiningsFromBottom())
        self.assertFalse(decider.needsMain())
        self.assertTrue(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testMachiningsFromBottomWithForcedMainShouldHaveMainAndOppositeSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy(), shouldAllwaysCreateMainSide=True)
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyMachiningsFromBottom())
        self.assertTrue(decider.needsMain())
        self.assertTrue(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testMachiningsHorizontalShouldHaveMainSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy())
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyHorizontalMachinings())
        self.assertTrue(decider.needsMain())
        self.assertFalse(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testMachiningsHorizontalWithDeniedMainShouldHaveOppositeSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy(), allowHorizontalMachiningsOnMainSide=False)
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyHorizontalMachinings())
        self.assertFalse(decider.needsMain())
        self.assertTrue(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testMachiningsHorizontalWithDeniedButForcedMainShouldHaveMainAndOppositeSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy(), shouldAllwaysCreateMainSide=True, allowHorizontalMachiningsOnMainSide=False)
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyHorizontalMachinings())
        self.assertTrue(decider.needsMain())
        self.assertTrue(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testMachiningsTopAndHorizontalShouldHaveMainSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy())
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyMachiningsFromTopAndHorizontal())
        self.assertTrue(decider.needsMain())
        self.assertFalse(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())

    def testMachiningsTopAndHorizontalWithDeniedMainShouldHaveMainAndOppositeSurface(self):
        decider = self.createDefaultDecider(self.InlineSizingStrategy(), allowHorizontalMachiningsOnMainSide=False)
        decider.setWorkpieceAnalyzer(self.MockAnalyzer_OnlyMachiningsFromTopAndHorizontal())
        self.assertTrue(decider.needsMain())
        self.assertTrue(decider.needsOpposite())
        self.assertFalse(decider.needsSizing())
