# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import unittest

from interiorcad.backend.workpiece.centerpunch import CenterPunch
from interiorcad.backend.workpiece.chamfer import Chamfer
from interiorcad.backend.workpiece.contour import Contour
from interiorcad.backend.workpiece.dadorabbet import DadoRabbet
from interiorcad.backend.workpiece.drilling import Drilling
from interiorcad.backend.workpiece.miter import Miter
from interiorcad.backend.workpiece.ncmacro import NCMacro
from interiorcad.backend.workpiece.shape import Shape
from interiorcad.interface.geometry.edges import Edges
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.rectangle import Rectangle

# FIXME (NESTING): Duplicate von interiorcad.cuttinglistexport.base.CuttinglistExportBase and WorkpieceEntry -- NG (InnoWeek 2018-12)
GRAIN_LENGTHWISE, GRAIN_CROSSWISE, GRAIN_NONE = 0, 1, 2


class WorkpieceStringRepresentation:
    def getDefault(self, wp):
        nameAddition = ''
        if len(wp.name) > 0:
            nameAddition = ' - "{name}"'.format(name=wp.name)
        return "Workpiece - {id}{name}".format(id=wp.uuid, name=nameAddition)

    def getDetailed(self, wp):
        drillingsBlock = self.createDrillingsBlock(wp.drillings)
        finishedSizeShapeBlock = self.createShapeBlock("Finished Size", wp.finishedSizeShape)
        sawnSizeShapeBlock = self.createShapeBlock("Sawn Size", wp.sawnSizeShape)
        header = self.getDefault(wp)
        if wp.parentUUID:
            header += '\n\tParent Cabinet: {}'.format(str(wp.parentUUID))
        return '\n'.join((
            header,
            drillingsBlock,
            finishedSizeShapeBlock,
            sawnSizeShapeBlock,
        ))

    def createDrillingsBlock(self, drillings):
        name = self.createNameByCount('Drilling', len(drillings))
        head = ' '.join(('\t{}', '{}')).format(len(drillings), name)
        body = '\n'.join(('\t\t{}'.format(str(d)) for d in drillings))
        return '\n'.join((head, body)).rstrip('\n')

    def createNameByCount(self, singular, count):
        if count > 1:
            name = singular + 's:'
        elif count == 1:
            name = singular + ':'
        else:
            name = singular + 's'
        return name

    def createShapeBlock(self, shapeType, shape):
        shapeBlock = ''
        if shape.isValid():
            shapeBlock = '\t{} Shape:\n\t\t{}'.format(shapeType, str(shape))
        else:
            shapeBlock = '\tNo {} Shape'.format(shapeType)
        return shapeBlock


class Workpiece:
    def __init__(self, uuid=''):
        self.uuid = uuid
        self.name = ''
        self.parentUUID = ''
        self.thickness = 0
        self.boardThickness = 0
        self.isMirrored = False
        self.machinings = []
        self.finishedSizeShape = Shape()
        self.finishedSizeBounds = Rectangle()
        self.sawnSizeShape = Shape()
        self.sawnSizeBounds = Rectangle()
        self.jointingSizeShape = Shape()
        self.jointingSizeBounds = Rectangle()
        self.edges = Edges()
        self.count = 1
        self.grain = GRAIN_NONE

    def bottomLeft(self):
        return Point(self.finishedSizeBounds.left, self.finishedSizeBounds.bottom)

    def isRectangular(self):
        return self.finishedSizeShape.isRectangular()

    def isPolygonal(self):
        return self.finishedSizeShape.isPolygonal()

    def hasFinishedSizeShape(self):
        return self.finishedSizeShape.isValid()

    def hasSawnSizeShape(self):
        return self.sawnSizeShape.isValid()

    def __getClassMachinings(self, wantedMachiningClasses):
        return list(filter(lambda machining: isinstance(machining, wantedMachiningClasses), self.machinings))

    def __setClassMachinings(self, wantedMachiningClasses, machinings):
        nonClassDrillings = itertools.filterfalse(lambda machining: isinstance(machining, wantedMachiningClasses), self.machinings)
        self.machinings = list(itertools.chain(machinings, nonClassDrillings))

    # compatibility function for old code (up to interiorcad 2022 F1.1); for new code use directly self.machinings if possible
    @property
    def drillings(self):
        return self.__getClassMachinings(Drilling)

    # compatibility function for old code (up to interiorcad 2022 F1.1); for new code use directly self.machinings if possible
    @property
    def centerPunches(self):
        return self.__getClassMachinings(CenterPunch)

    # compatibility function for old code (up to interiorcad 2022 F1.1); for new code use directly self.machinings if possible
    @property
    def dadosrabbets(self):
        return self.__getClassMachinings(DadoRabbet)

    # compatibility function for old code (up to interiorcad 2022 F1.1); for new code use directly self.machinings if possible
    @property
    def miters(self):
        return self.__getClassMachinings(Miter)

    # compatibility function for old code (up to interiorcad 2022 F1.1); for new code use directly self.machinings if possible
    @property
    def contours(self):
        return self.__getClassMachinings(Contour)

    # compatibility function for old code (up to interiorcad 2022 F1.1); for new code use directly self.machinings if possible
    @property
    def ncmacros(self):
        return self.__getClassMachinings(NCMacro)

    # compatibility function for old code (up to interiorcad 2022 F1.1); for new code use directly self.machinings if possible
    @property
    def chamfers(self):
        return self.__getClassMachinings(Chamfer)

    def getAllBounds(self):
        return [self.finishedSizeBounds, self.sawnSizeBounds, self.jointingSizeBounds]

    def getAllMachinings(self):
        return self.machinings

    def getAllSizings(self):
        return [self.finishedSizeShape, self.sawnSizeShape, self.jointingSizeShape]

    def __str__(self):
        return WorkpieceStringRepresentation().getDetailed(self)

    def __repr__(self):
        return WorkpieceStringRepresentation().getDefault(self)


class WorkpieceTest(unittest.TestCase):
    def setUp(self):
        self.wp = Workpiece()

    def testDefaultWorkpieceHasZeroThickness(self):
        self.assertEqual(0, self.wp.thickness)

    def testCanSetThickness(self):
        self.wp.thickness = 19.0
        self.assertEqual(19.0, self.wp.thickness)

    def testDefaultWorkpieceHasNoShape(self):
        self.assertFalse(self.wp.hasFinishedSizeShape())

    def testDefaultWorkpieceHasNoUUID(self):
        self.assertEqual('', self.wp.uuid)

    def testCanCreateWorkpieceWithUUID(self):
        wp = Workpiece('NEW_UUID')
        self.assertEqual('NEW_UUID', wp.uuid)

    def testDefaultWorkpieceHasStringRepresentation(self):
        self.assertTrue(str(self.wp))

    def testDefaultWorkpieceHasEmptyFinishedSizeBounds(self):
        self.assertEqual(Rectangle(), self.wp.finishedSizeBounds)

    def testDefaultWorkpieceHasEmptySawnSizeBounds(self):
        self.assertEqual(Rectangle(), self.wp.sawnSizeBounds)

    def testCanGetFinishedSizeBounds(self):
        expectedBounds = Rectangle(0, 0, 100, 100)
        self.wp.finishedSizeBounds = Rectangle(0, 0, 100, 100)
        self.assertEqual(expectedBounds, self.wp.finishedSizeBounds)

    def testCanGetSawnSizeBounds(self):
        expectedBounds = Rectangle(0, 0, 100, 100)
        self.wp.sawnSizeBounds = Rectangle(0, 0, 100, 100)
        self.assertEqual(expectedBounds, self.wp.sawnSizeBounds)

    def testDefaultWorkpieceHasEmptyName(self):
        self.assertEqual('', self.wp.name)

    def testSetName(self):
        self.wp.name = 'NEW_NAME'
        self.assertEqual('NEW_NAME', self.wp.name)

    def testGetAllBounds(self):
        bounds = self.wp.getAllBounds()

        self.assertIsInstance(bounds, list)
        self.assertEqual(3, len(bounds))
        self.assertIs(self.wp.finishedSizeBounds, bounds[0])
        self.assertIs(self.wp.sawnSizeBounds, bounds[1])
        self.assertIs(self.wp.jointingSizeBounds, bounds[2])

    def testGetAllSizings(self):
        allSizing = self.wp.getAllSizings()

        self.assertIsInstance(allSizing, list)
        self.assertEqual(3, len(allSizing))
        self.assertIs(self.wp.finishedSizeShape, allSizing[0])
        self.assertIs(self.wp.sawnSizeShape, allSizing[1])
        self.assertIs(self.wp.jointingSizeShape, allSizing[2])

    def testGetPosition(self):
        self.wp.finishedSizeBounds = Rectangle(0, 0, 100, 200)
        self.wp.sawnSizeBounds = Rectangle(5, 5, 95, 195)
        self.assertEqual(Point(0, 0), self.wp.bottomLeft())

    def testGetPositionOfMovedWorkpiece(self):
        self.wp.finishedSizeBounds = Rectangle(50, 100, 100, 200)
        self.wp.sawnSizeBounds = Rectangle(55, 55, 95, 195)
        self.assertEqual(Point(50, 100), self.wp.bottomLeft())

    def testIsWorkpieceRectangularShouldReturnFalse(self):
        self.assertFalse(self.wp.isRectangular())

    def testDefaultWorkpieceIsNotMirrored(self):
        self.assertFalse(self.wp.isMirrored)
