# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest
from collections import namedtuple

from interiorcad.backend.workpiece.machining import Machining
from interiorcad.interface.geometry.matrix import Matrix

MultiUsageRouting = namedtuple('MultiUsageRouting', 'toolName depthPerStep distance feedRate depth clearOut cutThrough')


class Contour(Machining):
    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    def __init__(self, uuid=''):
        super().__init__(uuid)
        self.depth = 0
        self.width = 0
        self.radiusCompensation = Contour.RADIUS_COMPENSATION_NONE
        self.approach = ''
        self.departure = ''
        self.clearOut = False
        self.cutThrough = False
        self.hasMachineOptions = False
        self.selectedMachine = None
        self.multiUsageRoutings = None

    def __str__(self):
        result = ' '.join((
            'Contour:',
            super().__str__() + ';',
            'width({:.{}f}),'.format(self.width, 2),
            'depth({:.{}f})'.format(self.depth, 2),
        ))
        return result

    def getDepth(self):
        return self.depth

    def getWidth(self):
        return self.width

    def getRadiusCompensation(self):
        return self.radiusCompensation


class Tests(unittest.TestCase):
    def setUp(self):
        self.contour = Contour()

    def testDefaultHasZeroWidth(self):
        self.assertEqual(0, self.contour.getWidth())

    def testDefaultHasZeroDepth(self):
        self.assertEqual(0, self.contour.getDepth())

    def testDefaultHasDefaultMatrix(self):
        self.assertEqual(Matrix(), self.contour.getMatrix())

    def testDefaultHasEmptyUUID(self):
        self.assertEqual('', self.contour.uuid)

    def testCanSetWidth(self):
        self.contour.width = 15
        self.assertEqual(15, self.contour.getWidth())

    def testCanSetDepth(self):
        self.contour.depth = 15
        self.assertEqual(15, self.contour.getDepth())

    def testCanSetRadiusCompensation(self):
        self.contour.radiusCompensation = self.contour.RADIUS_COMPENSATION_RIGHT
        self.assertEqual(self.contour.RADIUS_COMPENSATION_RIGHT, self.contour.getRadiusCompensation())

    def testCanSetMatrix(self):
        matrix = Matrix('u', 'v', 'w', 'origin')
        self.contour.matrix = matrix
        self.assertEqual('u', self.contour.matrix.u)
        self.assertEqual('v', self.contour.matrix.v)
        self.assertEqual('w', self.contour.matrix.w)
        self.assertEqual('origin', self.contour.matrix.origin)

    def testCanSetUUID(self):
        self.contour.uuid = 'NEW_UUID'
        self.assertEqual('NEW_UUID', self.contour.getUUID())

    def testCanSetUUIDInConstructor(self):
        contour = Contour('NEW_UUID')
        self.assertEqual('NEW_UUID', contour.getUUID())
