# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.backend.workpiece import Workpiece
from interiorcad.backend.workpiece.mover import WorkpieceMover
from interiorcad.interface.geometry.point import Point
from interiorcad.backend.workpiece.shape import Shape
from interiorcad.backend.workpiece.machiningcreator import MachiningCreator
from interiorcad.backend.joint.sizing import JointSizing
from interiorcad.backend.workpiece.reader import WorkpieceReader


class WorkpieceCreator:
    def __init__(self, dataCollector, jointingInfo):
        self.dataCollector = dataCollector
        self.jointingInfo = jointingInfo

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code
    def createAllWorkpiecesForExport(self):
        uuids = self.dataCollector.getCustomPartUUIDs()
        return self.createWorkpieces(uuids)

    # compatibility function for old code, up to interiorcad 2018 F4; do NOT use for new code, use createCountedWorkpieces instead
    def createWorkpieces(self, uuids):
        return (self.createNCWorkpiece(uuid) for uuid in uuids)

    # from interiorcad 2018 F5 on
    def createCountedWorkpieces(self, uuidsAndCounts):
        return (self.createNCWorkpiece(uuid, count) for uuid, count in uuidsAndCounts)

    def createNCWorkpiece(self, uuid, count=1):
        workpiece = Workpiece(uuid)
        return self.__createWorkpiece(workpiece, uuid, count)

    def __createWorkpiece(self, workpiece, uuid, count):
        self.__setMajorPropertiesOfWorkpiece(workpiece, uuid, count)
        self.__addMachinings(workpiece, uuid)
        self.__setBoundsAndShapes(workpiece, uuid)
        self.__getEdges(workpiece, uuid)

        WorkpieceMover(workpiece).moveBottomLeftTo(Point(0, 0))

        return workpiece

    def hasCovering(self, uuid):
        return any((
            self.dataCollector.getRawDataAsString(uuid=uuid, key='part_covering1'),
            self.dataCollector.getRawDataAsFloat(uuid=uuid, key='part_covering1thickness'),
            self.dataCollector.getRawDataAsString(uuid=uuid, key='part_covering2'),
            self.dataCollector.getRawDataAsFloat(uuid=uuid, key='part_covering2thickness'),
        ))

    def __setMajorPropertiesOfWorkpiece(self, workpiece, uuid, count):
        workpiece.name = self.dataCollector.getUniqueName(uuid)
        workpiece.parentUUID = self.dataCollector.getUserPerceivedParentUUID(uuid)
        workpiece.thickness = self.dataCollector.getCustomPartThickness(uuid)
        workpiece.boardThickness = self.dataCollector.getCustomPartBoardThickness(uuid)
        workpiece.isMirrored = self.dataCollector.getCustomPartIsMirrored(uuid)
        workpiece.count = count * self.dataCollector.getBoardCount(uuid)

        if self.hasCovering(uuid):
            workpiece.grain = self.dataCollector.getCoveringGrainDirection(uuid)
        else:
            workpiece.grain = self.dataCollector.getBoardGrainDirection(uuid)

    def __setBoundsAndShapes(self, workpiece, uuid):
        workpiece.finishedSizeBounds = self.dataCollector.getFinishedSizeBounds(uuid)
        workpiece.finishedSizeShape = self.__createFinishedSizeSizing(uuid)

        workpiece.sawnSizeBounds = self.dataCollector.getSawnSizeBounds(uuid)
        workpiece.sawnSizeShape = self.__createSawnSizeSizing(uuid)

        if self.jointingInfo.jointing:
            workpiece.jointingSizeShape, workpiece.jointingSizeBounds = self.__getJointSizingShapeAndBounds(workpiece)

    def __getEdges(self, workpiece, uuid):
        workpiece.edges = self.dataCollector.getEdges(uuid)

    def __addMachinings(self, workpiece, uuid):
        machiningCreator = MachiningCreator(self.dataCollector)
        workpiece.machinings = machiningCreator.createMachinings(uuid)

    def __createFinishedSizeSizing(self, uuid):
        polyline = self.dataCollector.getFinishedSizeSizingContour(uuid)
        return Shape(polyline)

    def __createSawnSizeSizing(self, uuid):
        polyline = self.dataCollector.getSawnSizeSizingContour(uuid)
        return Shape(polyline)

    def __getJointSizingShapeAndBounds(self, workpiece):
        reader = WorkpieceReader(workpiece)
        jointSizing = JointSizing(reader, self.jointingInfo.minJoint, self.jointingInfo.maxJoint)
        return jointSizing.getJointSizingShapeAndBounds()
