# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.point import Point
from interiorcad.backend.workpiece.linemachining import LineMachining
from interiorcad.interface.geometry.vector import Vector


class DadoRabbet(LineMachining):
    def __init__(self, uuid=''):
        super().__init__(uuid)
        self.depth = 0

    def __str__(self):
        result = ' '.join((
            'Dado/Rabbet:',
            super().__str__() + ';',
            'depth({:.{}f})'.format(self.depth, 2),
        ))
        return result

    def getDepth(self):
        return self.depth


class Tests(unittest.TestCase):
    def setUp(self):
        self.dadoRabbet = DadoRabbet()

    def testDefaultHasZeroWidth(self):
        self.assertEqual(0, self.dadoRabbet.getWidth())

    def testDefaultHasZeroDepth(self):
        self.assertEqual(0, self.dadoRabbet.getDepth())

    def testDefaultHasZeroLength(self):
        self.assertEqual(0, self.dadoRabbet.getLength())

    def testDefaultHasDefaultMatrix(self):
        self.assertEqual(Matrix(), self.dadoRabbet.getMatrix())

    def testDefaultHasEmptyUUID(self):
        self.assertEqual('', self.dadoRabbet.uuid)

    def testCanSetWidth(self):
        self.dadoRabbet.width = 15
        self.assertEqual(15, self.dadoRabbet.getWidth())

    def testCanSetDepth(self):
        self.dadoRabbet.depth = 15
        self.assertEqual(15, self.dadoRabbet.getDepth())

    def testCanSetLength(self):
        self.dadoRabbet.length = 5
        self.assertEqual(5, self.dadoRabbet.getLength())

    def testCanSetMatrix(self):
        matrix = Matrix('u', 'v', 'w', 'origin')
        self.dadoRabbet.matrix = matrix
        self.assertEqual('u', self.dadoRabbet.matrix.u)
        self.assertEqual('v', self.dadoRabbet.matrix.v)
        self.assertEqual('w', self.dadoRabbet.matrix.w)
        self.assertEqual('origin', self.dadoRabbet.matrix.origin)

    def testCanSetUUID(self):
        self.dadoRabbet.uuid = 'NEW_UUID'
        self.assertEqual('NEW_UUID', self.dadoRabbet.getUUID())

    def testCanSetUUIDInConstructor(self):
        dadoRabbet = DadoRabbet('NEW_UUID')
        self.assertEqual('NEW_UUID', dadoRabbet.getUUID())

    class TestMatrix:
        def __str__(self):
            return 'M'

    def testStringRepresentationIsAsExpected(self):
        self.dadoRabbet.matrix = self.TestMatrix()
        expected = 'Dado/Rabbet: LineMachining: Machining {NO_UUID}: matrix(M); width(0.00), length(0.00), startAddition(0.00), endAddition(0.00); depth(0.00)'
        self.assertEqual(expected, str(self.dadoRabbet))

    def testStringRepresentationWithUUIDIsAsExpected(self):
        self.dadoRabbet.matrix = self.TestMatrix()
        self.dadoRabbet.uuid = '{UUID}'
        expected = 'Dado/Rabbet: LineMachining: Machining {UUID}: matrix(M); width(0.00), length(0.00), startAddition(0.00), endAddition(0.00); depth(0.00)'
        self.assertEqual(expected, str(self.dadoRabbet))

    def testMove(self):
        self.dadoRabbet.matrix.origin = Vector(50.0, 60.0, 70.0)
        self.dadoRabbet.moveBy(Point(10.0, 20))
        self.assertEqual(Vector(60, 80, 70), self.dadoRabbet.getMatrix().origin)

    def testGetEndPoint(self):
        self.dadoRabbet.length = 100
        self.assertEqual(Vector(100, 0, 0), self.dadoRabbet.getEndPoint())

    def testDefaultDadoHasNoStartOrEndAddition(self):
        self.assertEqual(0.0, self.dadoRabbet.startAddition)
        self.assertEqual(0.0, self.dadoRabbet.endAddition)
