# -*- coding: utf-8 -*-

# Copyright 2011-2012, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.backend.workpiece.machining import Machining
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.point import Point


class Drilling(Machining):
    def __init__(self, uuid=''):
        super().__init__(uuid)
        self.diameter = 0
        self.depth = 0
        self.drillThrough = False

    def __str__(self):
        result = ' '.join((
            'Drilling:',
            super().__str__() + ';',
            'diameter({d:.{cp}f}),'.format(d=self.diameter, cp=2),
            'depth({d:.{cp}f})'.format(d=self.depth, cp=2),
        ))
        return result

    def getDiameter(self):
        return self.diameter

    def getDepth(self):
        return self.depth


class Tests(unittest.TestCase):
    def setUp(self):
        self.drilling = Drilling()

    def testDefaultDrillingHasZeroDiameter(self):
        self.assertEqual(0, self.drilling.getDiameter())

    def testDefaultDrillingHasZeroDepth(self):
        self.assertEqual(0, self.drilling.getDepth())

    def testDefaultDrillingHasDefaultMatrix(self):
        self.assertEqual(Matrix(), self.drilling.getMatrix())

    def testDefaultDrillingHasEmptyUUID(self):
        self.assertEqual('', self.drilling.uuid)

    def testCanSetDiameter(self):
        self.drilling.diameter = 5
        self.assertEqual(5, self.drilling.getDiameter())

    def testCanSetDepth(self):
        self.drilling.depth = 15
        self.assertEqual(15, self.drilling.getDepth())

    def testCanSetMatrix(self):
        matrix = Matrix('u', 'v', 'w', 'origin')
        self.drilling.matrix = matrix
        self.assertEqual('u', self.drilling.matrix.u)
        self.assertEqual('v', self.drilling.matrix.v)
        self.assertEqual('w', self.drilling.matrix.w)
        self.assertEqual('origin', self.drilling.matrix.origin)

    def testCanSetUUID(self):
        self.drilling.uuid = 'NEW_UUID'
        self.assertEqual('NEW_UUID', self.drilling.getUUID())

    def testCanSetUUIDInConstructor(self):
        drilling = Drilling('NEW_UUID')
        self.assertEqual('NEW_UUID', drilling.getUUID())

    class TestMatrix:
        def __str__(self):
            return 'M'

    def testStringRepresentationIsAsExpected(self):
        self.drilling.matrix = self.TestMatrix()
        expected = 'Drilling: Machining {NO_UUID}: matrix(M); diameter(0.00), depth(0.00)'
        self.assertEqual(expected, str(self.drilling))

    def testStringRepresentationWithUUIDIsAsExpected(self):
        self.drilling.matrix = self.TestMatrix()
        self.drilling.uuid = '{UUID}'
        expected = 'Drilling: Machining {UUID}: matrix(M); diameter(0.00), depth(0.00)'
        self.assertEqual(expected, str(self.drilling))

    def testMoveDrilling(self):
        from interiorcad.interface.geometry.vector import Vector
        self.drilling.matrix.origin = Vector(50.0, 60.0, 70.0)
        self.drilling.moveBy(Point(10.0, 20))
        self.assertEqual(Vector(60, 80, 70), self.drilling.getMatrix().origin)
