# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest

from interiorcad.backend.workpiece.machining import Machining
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector


class LineMachining(Machining):
    def __init__(self, uuid=''):
        super().__init__(uuid)
        self.width = 0
        self.length = 0
        self.startInset = 0
        self.endInset = 0
        self.startAddition = 0
        self.endAddition = 0

    def __str__(self):
        result = ' '.join((
            'LineMachining:',
            super().__str__() + ';',
            'width({:.{}f}),'.format(self.width, 2),
            'length({:.{}f}),'.format(self.length, 2),
            'startAddition({:.{}f}),'.format(self.startAddition, 2),
            'endAddition({:.{}f})'.format(self.endAddition, 2)
        ))
        return result

    def getWidth(self):
        return self.width

    def getLength(self):
        return self.length

    def getStartPoint(self):
        return self.matrix.origin

    def getEndPoint(self):
        return self.matrix.origin + self.matrix.u * self.length

    def getRealLength(self):
        return math.fsum((self.length, self.startAddition - self.startInset, self.endAddition - self.endInset))


class Tests(unittest.TestCase):
    def setUp(self):
        self.dadoRabbet = LineMachining()

    def testDefaultHasZeroWidth(self):
        self.assertEqual(0, self.dadoRabbet.getWidth())

    def testDefaultHasZeroLength(self):
        self.assertEqual(0, self.dadoRabbet.getLength())

    def testDefaultHasDefaultMatrix(self):
        self.assertEqual(Matrix(), self.dadoRabbet.getMatrix())

    def testDefaultHasEmptyUUID(self):
        self.assertEqual('', self.dadoRabbet.uuid)

    def testCanSetWidth(self):
        self.dadoRabbet.width = 15
        self.assertEqual(15, self.dadoRabbet.getWidth())

    def testCanSetLength(self):
        self.dadoRabbet.length = 5
        self.assertEqual(5, self.dadoRabbet.getLength())

    def testCanSetMatrix(self):
        matrix = Matrix('u', 'v', 'w', 'origin')
        self.dadoRabbet.matrix = matrix
        self.assertEqual('u', self.dadoRabbet.matrix.u)
        self.assertEqual('v', self.dadoRabbet.matrix.v)
        self.assertEqual('w', self.dadoRabbet.matrix.w)
        self.assertEqual('origin', self.dadoRabbet.matrix.origin)

    def testCanSetUUID(self):
        self.dadoRabbet.uuid = 'NEW_UUID'
        self.assertEqual('NEW_UUID', self.dadoRabbet.getUUID())

    def testCanSetUUIDInConstructor(self):
        dadoRabbet = LineMachining('NEW_UUID')
        self.assertEqual('NEW_UUID', dadoRabbet.getUUID())

    class TestMatrix:
        def __str__(self):
            return 'M'

    def testStringRepresentationIsAsExpected(self):
        self.dadoRabbet.matrix = self.TestMatrix()
        expected = 'LineMachining: Machining {NO_UUID}: matrix(M); width(0.00), length(0.00), startAddition(0.00), endAddition(0.00)'
        self.assertEqual(expected, str(self.dadoRabbet))

    def testStringRepresentationWithUUIDIsAsExpected(self):
        self.dadoRabbet.matrix = self.TestMatrix()
        self.dadoRabbet.uuid = '{UUID}'
        expected = 'LineMachining: Machining {UUID}: matrix(M); width(0.00), length(0.00), startAddition(0.00), endAddition(0.00)'
        self.assertEqual(expected, str(self.dadoRabbet))

    def testMove(self):
        self.dadoRabbet.matrix.origin = Vector(50.0, 60.0, 70.0)
        self.dadoRabbet.moveBy(Point(10.0, 20))
        self.assertEqual(Vector(60, 80, 70), self.dadoRabbet.getMatrix().origin)

    def testGetEndPoint(self):
        self.dadoRabbet.length = 100
        self.assertEqual(Vector(100, 0, 0), self.dadoRabbet.getEndPoint())

    def testDefaultDadoHasNoStartOrEndAddition(self):
        self.assertEqual(0.0, self.dadoRabbet.startAddition)
        self.assertEqual(0.0, self.dadoRabbet.endAddition)
