# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest

from interiorcad.backend.ncexport.linemachininganalyzer import LineMachiningAnalyzer
from interiorcad.backend.workpiece.dadorabbet import DadoRabbet
from interiorcad.backend.workpiece.shape import Shape
from interiorcad.interface.geometry import isNearlyEqual, EPSILON
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vertex import Vertex


class MachiningAnalyzer:
    def isFromTop(self, drilling):
        return drilling.matrix.w.z > 0 and not self.isHorizontal(drilling)

    def isFromBottom(self, drilling):
        return drilling.matrix.w.z < 0 and not self.isHorizontal(drilling)

    def isHorizontal(self, drilling):
        return isNearlyEqual(drilling.matrix.w.z, 0)

    def isFiveAxisFromTop(self, drilling):
        return self.isFromTop(drilling) and not self.isVerticalFromTop(drilling)

    def isFiveAxisFromBottom(self, drilling):
        return self.isFromBottom(drilling) and not self.isVerticalFromBottom(drilling)

    def isVertical(self, machining):
        return isNearlyEqual(math.fabs(machining.matrix.w.z), 1)

    def isVerticalFromTop(self, drilling):
        return isNearlyEqual(drilling.matrix.w.z, 1)

    def isVerticalFromBottom(self, drilling):
        return isNearlyEqual(drilling.matrix.w.z, -1)

    def isRabbet(self, dadorabbet, workpieceShape):
        return LineMachiningAnalyzer().isRabbet(dadorabbet, workpieceShape)

    def isDado(self, dadorabbet, workpieceShape):
        return LineMachiningAnalyzer().isDado(dadorabbet, workpieceShape)


class MachiningsOnWorkpieceGetter:
    def __init__(self, workpiece):
        self.workpiece = workpiece
        self.machiningAnalyzer = MachiningAnalyzer()

    def setMachiningAnalyzer(self, ma):
        self.machiningAnalyzer = ma

    def hasMachiningsFromTop(self):
        return any(self.getAllMachiningsFromTop())

    def hasMachiningsFromBottom(self):
        return any(self.getAllMachiningsFromBottom())

    def hasMachiningsHorizontal(self):
        return any(self.getAllHorizontalMachinings())

    def getAllMachinings(self):
        return self.workpiece.getAllMachinings()

    def getAllMachiningsFromBottom(self):
        return self.__getMachiningsByCondition(self.workpiece.getAllMachinings(), self.machiningAnalyzer.isFromBottom)

    def getAllMachiningsFromTop(self):
        return self.__getMachiningsByCondition(self.workpiece.getAllMachinings(), self.machiningAnalyzer.isFromTop)

    def getAllHorizontalMachinings(self):
        return self.__getMachiningsByCondition(self.workpiece.getAllMachinings(), self.machiningAnalyzer.isHorizontal)

    def __getMachiningsByCondition(self, machinings, condition):
        return [machining for machining in machinings if condition(machining)]

    def isRectangular(self):
        return self.workpiece.finishedSizeShape.polyline.isRectangular()


class TestMachiningCreator(unittest.TestCase):
    class TestVector:
        def __init__(self, x=0, y=0, z=0):
            self.x = x
            self.y = y
            self.z = z

    class TestMatrix:
        def __init__(self, wx=0, wy=0, wz=0):
            self.w = TestMachiningCreator.TestVector(wx, wy, wz)

    class TestMachining():
        def __init__(self, wx=0, wy=0, wz=0):
            self.matrix = TestMachiningCreator.TestMatrix(wx, wy, wz)
            self.width = 10

        def getMatrix(self):
            return self.matrix


class TestsDadoRabbetDetection(unittest.TestCase):
    def setUp(self):
        self.analyzer = MachiningAnalyzer()
        self.shape = Shape()
        poly = self.shape.polyline
        poly.addVertex(Vertex(Vector(0, 0)))
        poly.addVertex(Vertex(Vector(0, 50)))
        poly.addVertex(Vertex(Vector(50, 100)))
        poly.addVertex(Vertex(Vector(200, 100)))
        poly.addVertex(Vertex(Vector(200, 0)))
        poly.addVertex(Vertex(Vector(0, 0)))
        self.dadoRabbet = DadoRabbet()
        self.dadoRabbet.length = 1
        self.dadoRabbet.width = 10

    def testVerticalRabbetOnLeftWorkpieceSideShouldBeDetectedAsRabbet(self):
        self.dadoRabbet.matrix.u = Vector(0, 1, 0)
        self.dadoRabbet.matrix.origin = Vector(195, 0)
        self.assertTrue(self.analyzer.isRabbet(self.dadoRabbet, self.shape))
        self.assertFalse(self.analyzer.isDado(self.dadoRabbet, self.shape))

    def VectortestDadoShouldNotBeDetectedAsRabbet(self):
        self.dadoRabbet.matrix.origin = Vector(100, 0)
        self.assertFalse(self.analyzer.isRabbet(self.dadoRabbet, self.shape))
        self.assertTrue(self.analyzer.isDado(self.dadoRabbet, self.shape))

    def testHorizontalRabbetOnTopWorkpieceSideShouldBeDetectedAsRabbet(self):
        self.dadoRabbet.matrix.origin = Vector(0, 95)
        self.assertTrue(self.analyzer.isRabbet(self.dadoRabbet, self.shape))
        self.assertFalse(self.analyzer.isDado(self.dadoRabbet, self.shape))

    def testDiagonalRabbetOnLeftTopWorkpieceSideShouldBeDetectedAsRabbet(self):
        self.dadoRabbet.matrix.u = Vector(1, 1, 0)
        self.dadoRabbet.matrix.origin = Vector(0, 42.929)
        self.assertTrue(self.analyzer.isRabbet(self.dadoRabbet, self.shape))
        self.assertFalse(self.analyzer.isDado(self.dadoRabbet, self.shape))


class EpsilonTests(unittest.TestCase):
    def setUp(self):
        self.analyzer = MachiningAnalyzer()

    def testHoritontalNotFromTop(self):
        machining = TestMachiningCreator.TestMachining(wz=EPSILON / 2)
        self.assertFalse(self.analyzer.isFromTop(machining))
        self.assertFalse(self.analyzer.isFromBottom(machining))
        self.assertTrue(self.analyzer.isHorizontal(machining))

    def testHoritontalNotFromBottom(self):
        machining = TestMachiningCreator.TestMachining(wz=-EPSILON / 2)
        self.assertFalse(self.analyzer.isFromBottom(machining))
        self.assertFalse(self.analyzer.isFromTop(machining))
        self.assertTrue(self.analyzer.isHorizontal(machining))
