# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.backend.datatracker.datatrackeritemtypes import ITEMTYPE_DRILLING, ITEMTYPE_DADORABBET, ITEMTYPE_MITER, ITEMTYPE_CONTOUR, ITEMTYPE_CHAMFER, ITEMTYPE_CENTERPUNCH, ITEMTYPE_NCMACRO, ITEMTYPE_GRIDDRILLING, ITEMTYPE_NULL_MACHINING, ITEMTYPE_PROFILE
from interiorcad.backend.workpiece.centerpunch import CenterPunch
from interiorcad.backend.workpiece.chamfer import Chamfer
from interiorcad.backend.workpiece.complexgriddrilling import ComplexGridDrilling, GridPointsCoordinateSystem
from interiorcad.backend.workpiece.contour import Contour, MultiUsageRouting
from interiorcad.backend.workpiece.dadorabbet import DadoRabbet
from interiorcad.backend.workpiece.drilling import Drilling
from interiorcad.backend.workpiece.miter import Miter
from interiorcad.backend.workpiece.ncmacro import NCMacro
from interiorcad.backend.workpiece.nullmachining import NullMachining
from interiorcad.backend.workpiece.profile import Profile
from interiorcad.interface.geometry.point import Point


class MachiningCreator:
    def __init__(self, dataCollector):
        self.dataCollector = dataCollector

    def __createAllMachinings(self, typedRoutings):
        CREATORS = {
            ITEMTYPE_DRILLING: self.createDrilling,
            ITEMTYPE_DADORABBET: self.createDadoRabbet,
            ITEMTYPE_MITER: self.createMiter,
            ITEMTYPE_CONTOUR: self.createContour,
            ITEMTYPE_CHAMFER: self.createChamfer,
            ITEMTYPE_CENTERPUNCH: self.createCenterPunch,
            ITEMTYPE_NCMACRO: self.createNCMacro,
            ITEMTYPE_PROFILE: self.createProfile,
            ITEMTYPE_GRIDDRILLING: self.createGridDrilling,
            ITEMTYPE_NULL_MACHINING: self.createNullMachining,
        }

        for objectType, uuid in typedRoutings:
            if objectType in CREATORS:
                yield CREATORS[objectType].__call__(uuid)
            else:
                assert False, "Internal error: Unknown Routing Type"

    def createMachinings(self, customPartUUID):
        typedRoutings = self.dataCollector.getTypedRoutings(customPartUUID)
        return list(self.__createAllMachinings(typedRoutings))

    def createDrilling(self, uuid):
        drilling = Drilling(uuid)
        drilling.depth = self.dataCollector.getMachiningDepth(uuid)
        drilling.diameter = self.dataCollector.getDrillingDiameter(uuid)
        drilling.matrix = self.dataCollector.getMachiningMatrix(uuid)
        drilling.drillThrough = self.dataCollector.getDrillingThrough(uuid)
        return drilling

    def createGridDrilling(self, gridDrillingUUID):
        gridDrilling = ComplexGridDrilling(gridDrillingUUID)
        gridDrilling.depth = self.dataCollector.getMachiningDepth(gridDrillingUUID)
        gridDrilling.diameter = self.dataCollector.getDrillingDiameter(gridDrillingUUID)
        gridDrilling.matrix = self.dataCollector.getMachiningMatrix(gridDrillingUUID)

        rawGridPoints = self.dataCollector.getGridPoints(gridDrillingUUID)
        gridDrilling.gridPoints = [Point(rawGridPoint['x'], rawGridPoint['y']) for rawGridPoint in rawGridPoints]

        rawGridPointsCoordinateSystem = self.dataCollector.getGridPointsCoordinateSystem(gridDrillingUUID)
        if rawGridPointsCoordinateSystem:
            gridDrilling.gridPointsCoordinateSystem = GridPointsCoordinateSystem(
                xaxis=Point(rawGridPointsCoordinateSystem['xaxis']['x'], rawGridPointsCoordinateSystem['xaxis']['y']),
                yaxis=Point(rawGridPointsCoordinateSystem['yaxis']['x'], rawGridPointsCoordinateSystem['yaxis']['y']),
                origin=Point(rawGridPointsCoordinateSystem['origin']['x'], rawGridPointsCoordinateSystem['origin']['y']),
            )

        gridDrilling.drillThrough = self.dataCollector.getDrillingThrough(gridDrillingUUID)

        return gridDrilling

    def createCenterPunch(self, uuid):
        centerPunch = CenterPunch(uuid)
        centerPunch.diameter = self.dataCollector.getDrillingDiameter(uuid)
        centerPunch.matrix = self.dataCollector.getMachiningMatrix(uuid)
        return centerPunch

    def __setPropertiesForLineMachining(self, uuid, machining):
        machining.width = self.dataCollector.getMachiningWidth(uuid)
        machining.length = self.dataCollector.getLineMachiningLength(uuid)
        machining.matrix = self.dataCollector.getMachiningMatrix(uuid)
        machining.startAddition = self.dataCollector.getLineMachiningStartAddition(uuid)
        machining.endAddition = self.dataCollector.getLineMachiningEndAddition(uuid)

    def createDadoRabbet(self, uuid):
        machining = DadoRabbet(uuid)
        self.__setPropertiesForLineMachining(uuid, machining)
        machining.startInset = self.dataCollector.getDadoRabbetStartInset(uuid)
        machining.endInset = self.dataCollector.getDadoRabbetEndInset(uuid)
        machining.depth = self.dataCollector.getMachiningDepth(uuid)
        return machining

    def createMiter(self, uuid):
        miter = Miter(uuid)
        self.__setPropertiesForLineMachining(uuid, miter)
        miter.angle = self.dataCollector.getAngle(uuid)
        return miter

    def createChamfer(self, uuid):
        chamfer = Chamfer(uuid)
        self.__setPropertiesForLineMachining(uuid, chamfer)
        chamfer.angle = self.dataCollector.getAngle(uuid)
        return chamfer

    def createContour(self, uuid):
        contour = Contour(uuid)
        contour.width = self.dataCollector.getMachiningWidth(uuid)
        contour.depth = self.dataCollector.getMachiningDepth(uuid)
        contour.path = self.dataCollector.getContourPath(uuid)
        contour.matrix = self.dataCollector.getMachiningMatrix(uuid)
        contour.radiusCompensation = self.dataCollector.getRadiusCompensation(uuid)
        contour.approach = self.dataCollector.getApproach(uuid)
        contour.departure = self.dataCollector.getDeparture(uuid)
        contour.hasMachineOptions = self.dataCollector.hasContourMachineOptions(uuid)
        contour.selectedMachine = self.dataCollector.getContourSelectedMachine(uuid)
        contour.clearOut = self.dataCollector.getContourClearOut(uuid, 0)
        contour.cutThrough = self.dataCollector.getContourCutThrough(uuid, 0)
        self.__createMultiUsageContour(uuid, contour)

        return contour

    def __createMultiUsageContour(self, uuid, contour):
        count = self.dataCollector.getContourMultiUsageEntries(uuid)
        contour.multiUsageRoutings = [
            MultiUsageRouting(
                toolName=self.dataCollector.getContourRouter(uuid, index),  # Careful: it is really the tool name, not its ID
                depthPerStep=self.dataCollector.getContourDepthPerStep(uuid, index),
                distance=self.dataCollector.getContourDistance(uuid, index),
                feedRate=self.dataCollector.getContourFeedRate(uuid, index),
                depth=self.dataCollector.getContourDepth(uuid, index),
                clearOut=self.dataCollector.getContourClearOut(uuid, index),
                cutThrough=self.dataCollector.getContourCutThrough(uuid, index),
            )
            for index in range(count)
        ]

    def createProfile(self, uuid):
        profile = Profile(uuid)
        profile.path = self.dataCollector.getContourPath(uuid)
        profile.matrix = self.dataCollector.getMachiningMatrix(uuid)
        return profile

    def createNCMacro(self, uuid):
        macro = NCMacro(uuid)
        macro.matrix = self.dataCollector.getMachiningMatrix(uuid)
        macro.name = self.dataCollector.getMacroName(uuid)
        macro.deltax = self.dataCollector.getDeltaX(uuid)
        macro.deltay = self.dataCollector.getDeltaY(uuid)
        macro.depth = self.dataCollector.getMachiningDepth(uuid)
        macro.code = self.dataCollector.getNCMacroCode(uuid)
        macro.parameters = tuple(self.dataCollector.getNCMacroParametersAsStrings(uuid))
        return macro

    def createNullMachining(self, uuid):
        # We do not read the diameter and depth that are in the datatracker, as they are of very limited usefulness (both are 9.0)
        machining = NullMachining(uuid)
        machining.matrix = self.dataCollector.getMachiningMatrix(uuid)
        return machining
