# -*- coding: utf-8 -*-

# Copyright 2011-2014, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.point import Point
from interiorcad.backend.workpiece.linemachining import LineMachining
from interiorcad.interface.geometry.vector import Vector


class Miter(LineMachining):
    def __init__(self, uuid=''):
        super().__init__(uuid)
        self.angle = 0

    def __str__(self):
        result = ' '.join((
            'Miter:',
            super().__str__() + ';',
            'angle({:.{}f})'.format(self.angle, 2),
        ))
        return result

    def getAngle(self):
        return self.angle


class Tests(unittest.TestCase):
    def setUp(self):
        self.miter = Miter()

    def testDefaultHasZeroWidth(self):
        self.assertEqual(0, self.miter.getWidth())

    def testDefaultHasZeroLength(self):
        self.assertEqual(0, self.miter.getLength())

    def testDefaultHasDefaultMatrix(self):
        self.assertEqual(Matrix(), self.miter.getMatrix())

    def testDefaultHasEmptyUUID(self):
        self.assertEqual('', self.miter.uuid)

    def testCanSetWidth(self):
        self.miter.width = 15
        self.assertEqual(15, self.miter.getWidth())

    def testCanSetLength(self):
        self.miter.length = 5
        self.assertEqual(5, self.miter.getLength())

    def testCanSetAngle(self):
        self.miter.angle = 5
        self.assertEqual(5, self.miter.getAngle())

    def testCanSetMatrix(self):
        matrix = Matrix('u', 'v', 'w', 'origin')
        self.miter.matrix = matrix
        self.assertEqual('u', self.miter.matrix.u)
        self.assertEqual('v', self.miter.matrix.v)
        self.assertEqual('w', self.miter.matrix.w)
        self.assertEqual('origin', self.miter.matrix.origin)

    def testCanSetUUID(self):
        self.miter.uuid = 'NEW_UUID'
        self.assertEqual('NEW_UUID', self.miter.getUUID())

    def testCanSetUUIDInConstructor(self):
        dadoRabbet = Miter('NEW_UUID')
        self.assertEqual('NEW_UUID', dadoRabbet.getUUID())

    class TestMatrix:
        def __str__(self):
            return 'M'

    def testStringRepresentationIsAsExpected(self):
        self.miter.matrix = self.TestMatrix()
        expected = 'Miter: LineMachining: Machining {NO_UUID}: matrix(M); width(0.00), length(0.00), startAddition(0.00), endAddition(0.00); angle(0.00)'
        self.assertEqual(expected, str(self.miter))

    def testStringRepresentationWithUUIDIsAsExpected(self):
        self.miter.matrix = self.TestMatrix()
        self.miter.uuid = '{UUID}'
        expected = 'Miter: LineMachining: Machining {UUID}: matrix(M); width(0.00), length(0.00), startAddition(0.00), endAddition(0.00); angle(0.00)'
        self.assertEqual(expected, str(self.miter))

    def testMoveDrilling(self):
        self.miter.matrix.origin = Vector(50.0, 60.0, 70.0)
        self.miter.moveBy(Point(10.0, 20))
        self.assertEqual(Vector(60, 80, 70), self.miter.getMatrix().origin)

    def testGetEndPoint(self):
        self.miter.length = 100
        self.assertEqual(Vector(100, 0, 0), self.miter.getEndPoint())
