# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import unittest

from interiorcad.backend.workpiece import Workpiece
from interiorcad.backend.workpiece.testworkpieces import TestWorkpieces
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector


class WorkpieceMover:
    def __init__(self, workpiece):
        assert isinstance(workpiece, Workpiece), 'WorkpieceMover does only work with Workpiece (not with Surface any more)!'
        self.workpiece = workpiece

    def moveBottomLeftTo(self, pos):
        original = self.workpiece.bottomLeft()
        delta = self.getDelta(original, pos)
        self.moveBy(delta)

    def getDelta(self, original, new):
        return new - original

    def moveBy(self, delta):
        movableProperties = itertools.chain(
            self.workpiece.getAllBounds(),
            self.workpiece.getAllMachinings(),
            self.workpiece.getAllSizings(),
        )

        for movableProperty in movableProperties:
            movableProperty.moveBy(delta)


class WorkpieceMoverTest(unittest.TestCase):
    def setUp(self):
        self.wp = TestWorkpieces.createRectangularWithVerticalDrillings()
        self.mover = WorkpieceMover(self.wp)

    def testWorkpieceMoverShouldMoveFinishedSizePolylineToZero(self):
        self.mover.moveBottomLeftTo(Point(0, 0))
        polyline = self.wp.finishedSizeShape.polyline
        self.assertEqual(Point(0, 0), polyline.vertices[0].position)

    def testWorkpieceMoverShouldMoveSawnSizePolyline(self):
        self.mover.moveBottomLeftTo(Point(0, 0))
        polyline = self.wp.sawnSizeShape.polyline
        self.assertEqual(Point(2, 3), polyline.vertices[0].position)

    def testWorkpieceMoverMovesDrillingToZero(self):
        self.mover.moveBottomLeftTo(Point(0, 0))
        drilling = self.wp.drillings[0]
        self.assertEqual(Vector(0, 0, 20), drilling.matrix.origin)

    def testWorkpieceMoverMovesDrillingToTenTen(self):
        self.mover.moveBottomLeftTo(Point(10, 10))
        drilling = self.wp.drillings[0]
        self.assertEqual(Vector(10, 10, 20), drilling.matrix.origin)

    def testWorkpieceMoverMovesFinishedSizeBounds(self):
        self.mover.moveBottomLeftTo(Point(10, 20))
        self.assertEqual(10, self.wp.finishedSizeBounds.left)
        self.assertEqual(110, self.wp.finishedSizeBounds.right)
        self.assertEqual(20, self.wp.finishedSizeBounds.bottom)
        self.assertEqual(220, self.wp.finishedSizeBounds.top)

    def testWorkpieceMoverMovesSawnSizeBounds(self):
        self.mover.moveBottomLeftTo(Point(10, 20))
        self.assertEqual(12, self.wp.sawnSizeBounds.left)
        self.assertEqual(108, self.wp.sawnSizeBounds.right)
        self.assertEqual(23, self.wp.sawnSizeBounds.bottom)
        self.assertEqual(218, self.wp.sawnSizeBounds.top)
