# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.backend.workpiece import GRAIN_CROSSWISE
from interiorcad.backend.workpiece.machininganalyzer import MachiningsOnWorkpieceGetter
from interiorcad.backend.workpiece.shape import Shape
from interiorcad.backend.workpiece.testworkpieces import TestWorkpieces
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.geometry.size import Size
from interiorcad.interface.surface.machiningconverter import ForceClassMachiningConverter
from interiorcad.interface.surface.sizing import Sizing2D


class WorkpieceReader:
    def __init__(self, workpiece):
        self.workpiece = workpiece
        self.finishedSizeSizing = None
        self.sawnSizeSizing = None
        self.horizontalDrillings = []
        self.analyzer = MachiningsOnWorkpieceGetter(workpiece)

    def getFinishedSize(self):
        bounds = self.workpiece.finishedSizeBounds
        return Size(bounds.getLength(), bounds.getWidth())

    def getThickness(self):
        return self.workpiece.thickness

    def getName(self):
        return self.workpiece.name

    def getUUID(self):
        return self.workpiece.uuid

    def getCount(self):
        return self.workpiece.count

    def getGrain(self):
        return self.workpiece.grain

    def getFinishedSizedShape(self):
        return self.workpiece.finishedSizeShape

    def getAllMachiningsUnconverted(self):
        return self.analyzer.getAllMachinings()

    def getAllMachiningsFromTopUnconverted(self):
        return self.analyzer.getAllMachiningsFromTop()

    def getAllMachiningsFromBottomUnconverted(self):
        return self.analyzer.getAllMachiningsFromBottom()

    def getAllHorizontalMachiningsUnconverted(self):
        return self.analyzer.getAllHorizontalMachinings()

    # TODO InnoWeek 2021-12: that should be somewhere in Surface, not in Workpiece -- NG
    def __convertToSizing2D(self, shape):
        return ForceClassMachiningConverter(Sizing2D).convertMachining(shape)

    def getFinishedSizeShape(self):
        return self.workpiece.finishedSizeShape

    def getFinishedSizeSizing(self):
        return self.__convertToSizing2D(self.workpiece.finishedSizeShape)

    def getSawnSizeSizing(self):
        return self.__convertToSizing2D(self.workpiece.sawnSizeShape)

    def getJointingSizeSizing(self):
        return self.__convertToSizing2D(self.workpiece.jointingSizeShape)

    def getPreliminaryRectangularSizing(self, sizingAddition):
        bounds = self.getPreliminaryRectangularSizingBounds(sizingAddition)
        shape = Shape.createRectangularSizingFromBounds(bounds)
        return self.__convertToSizing2D(shape)

    def getPreliminaryRectangularSizingBounds(self, sizingAddition):
        bounds = copy.deepcopy(self.getSawnSizingBoundsFromWorkpiece())
        bounds.expand(sizingAddition / 2)
        return bounds

    def getPreliminaryJointingRectangularSizing(self, sizingAddition):
        shape = Shape.createRectangularSizingFromBounds(self.getPreliminaryJointingRectangularSizingBounds(sizingAddition))
        return self.__convertToSizing2D(shape)

    def getPreliminaryJointingRectangularSizingBounds(self, sizingAddition):
        bounds = copy.deepcopy(self.getJointingSizingBoundsFromWorkpiece())
        bounds.expand(sizingAddition / 2)
        return bounds

    def getFinishedSizingBoundsFromWorkpiece(self):
        return self.workpiece.finishedSizeBounds

    def getSawnSizingBoundsFromWorkpiece(self):
        return self.workpiece.sawnSizeBounds

    def getJointingSizingBoundsFromWorkpiece(self):
        return self.workpiece.jointingSizeBounds

    def __boundsToSize(self, bounds):
        return Size(bounds.getLength(), bounds.getWidth())

    def getJointingSizeFromWorkpiece(self):
        return self.__boundsToSize(self.workpiece.jointingSizeBounds)

    def getSawnSizeFromWorkpiece(self):
        return self.__boundsToSize(self.workpiece.sawnSizeBounds)

    def getFinishedSizeAddition(self):
        diffLeft = self.workpiece.sawnSizeBounds.left - self.workpiece.finishedSizeBounds.left
        diffRight = self.workpiece.finishedSizeBounds.right - self.workpiece.sawnSizeBounds.right
        diffTop = self.workpiece.finishedSizeBounds.top - self.workpiece.sawnSizeBounds.top
        diffBottom = self.workpiece.sawnSizeBounds.bottom - self.workpiece.finishedSizeBounds.bottom
        return Rectangle(left=diffLeft, bottom=diffBottom, right=diffRight, top=diffTop)

    def getJointingSizeAddition(self, jointingSize):
        diffLeft = jointingSize.left - self.workpiece.finishedSizeBounds.left
        diffRight = self.workpiece.finishedSizeBounds.right - jointingSize.right
        diffTop = self.workpiece.finishedSizeBounds.top - jointingSize.top
        diffBottom = jointingSize.bottom - self.workpiece.finishedSizeBounds.bottom
        return Rectangle(left=diffLeft, bottom=diffBottom, right=diffRight, top=diffTop)

    def isRectangular(self):
        return self.workpiece.isRectangular()

    def isPolygonal(self):
        return self.workpiece.isPolygonal()

    def isMirrored(self):
        return self.workpiece.isMirrored

    def getEdges(self):
        return self.workpiece.edges

    def isGrainCrosswise(self):
        return self.workpiece.grain == GRAIN_CROSSWISE


class WorkpieceReaderTest_RectWorkpieceWithoutDrillings(unittest.TestCase):
    def setUp(self):
        self.workpiece = TestWorkpieces.createRectangularWithoutDrillings()
        self.reader = WorkpieceReader(self.workpiece)

    def testGetFinishedSizeSizing(self):
        finishedSizing = self.reader.getFinishedSizeSizing()
        expectedSizing = Sizing2D(Polyline(CornerVertex(Point(0, 0)),
                                           CornerVertex(Point(0, 200)),
                                           CornerVertex(Point(100, 200)),
                                           CornerVertex(Point(100, 0)),
                                           CornerVertex(Point(0, 0))))
        self.assertEqual(expectedSizing, finishedSizing)

    def testGetSawnSizeSizing(self):
        finishedSizing = self.reader.getSawnSizeSizing()
        expectedSizing = Sizing2D(Polyline(CornerVertex(Point(2, 3)),
                                           CornerVertex(Point(2, 198)),
                                           CornerVertex(Point(98, 198)),
                                           CornerVertex(Point(98, 3)),
                                           CornerVertex(Point(2, 3))))
        self.assertEqual(expectedSizing, finishedSizing)

    def testGetSawnSizingBoundsFromWorkpiece(self):
        bounds = self.reader.getSawnSizingBoundsFromWorkpiece()
        self.assertEqual(bounds, Rectangle(2, 3, 98, 198))

    def testGetFinishedSizingBoundsFromWorkpiece(self):
        bounds = self.reader.getFinishedSizingBoundsFromWorkpiece()
        self.assertEqual(bounds, Rectangle(0, 0, 100, 200))

    def testGetName(self):
        self.assertEqual('TestWorkpiece', self.reader.getName())

    def testGetFinishedSizeAddition(self):
        self.assertEqual(Rectangle(2, 3, 2, 2), self.reader.getFinishedSizeAddition())

    def testWorkpieceIsRectangular(self):
        self.assertTrue(self.reader.isRectangular())

    def testWorkpieceIsPolygonal(self):
        self.assertTrue(self.reader.isPolygonal())

    def testWorkpieceIsNotMirrored(self):
        self.assertFalse(self.reader.isMirrored())
