# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.vertex import Vertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polylinesegment import PolylineSegment
from interiorcad.interface.geometry.polylinesegments import PolylineSegments
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.geometry.cornervertex import CornerVertex


class Shape:
    def __init__(self, polyline=None):
        self.polyline = polyline if polyline else Polyline()

    def __bool__(self):
        return bool(self.polyline)

    def isValid(self):
        return self.polyline.isValid()

    def moveBy(self, delta):
        self.polyline.moveBy(delta)

    def rawString(self):
        return 'Shape(polyline={})'

    def isRectangular(self):
        return self.polyline.isRectangular()

    def isPolygonal(self):
        return self.polyline.isPolygonal()

    def getSegments(self):
        return self.polyline.getSegments()

    def __eq__(self, other):
        return self.polyline == other.polyline

    def __str__(self):
        return self.rawString().format(str(self.polyline))

    def __repr__(self):
        return self.rawString().format(repr(self.polyline))

    @staticmethod
    def createRectangularSizingFromBounds(sizingBounds):
        return Shape(Polyline.createRectangularClockwiseFromBounds(sizingBounds))


class ShapeTest(unittest.TestCase):
    def setUp(self):
        self.shape = Shape()

    def testShapeCreation(self):
        self.assertIsInstance(self.shape, Shape)

    def testDefaultShapeHasNoVertices(self):
        self.assertEqual(0, self.shape.polyline.numVertices())

    def testDefaultShapeIsNotValid(self):
        self.assertFalse(self.shape.isValid())
        self.assertFalse(self.shape)

    def testShapeWithOneVertexIsValid(self):
        self.shape.polyline.addVertex('1')
        self.assertTrue(self.shape.isValid())
        self.assertTrue(self.shape)

    def testCanMoveEmptyShape(self):
        self.shape.moveBy(Point(10, 20.0))
        self.assertEqual(Shape(), self.shape)

    def testShapeWithRectangularPolylineIsRectangular(self):
        class MockRectangularPolyline:
            def isRectangular(self):
                return True

        self.shape.polyline = MockRectangularPolyline()
        self.assertTrue(self.shape.isRectangular())


class ShapeAdvancedTest(unittest.TestCase):
    def setUp(self):
        self.shape = Shape()
        self.shape.polyline.addVertex(Vertex(Point(0, 0)))
        self.shape.polyline.addVertex(Vertex(Point(0, 50)))
        self.shape.polyline.addVertex(Vertex(Point(100, 50)))
        self.shape.polyline.addVertex(Vertex(Point(100, 0)))

    def testMoveShapeRightAndTop(self):
        self.shape.moveBy(Point(10.0, 20.0))
        self.assertEqual(Point(10, 20), self.shape.polyline.vertices[0].position)
        self.assertEqual(Point(10, 70), self.shape.polyline.vertices[1].position)
        self.assertEqual(Point(110, 70), self.shape.polyline.vertices[2].position)
        self.assertEqual(Point(110, 20), self.shape.polyline.vertices[3].position)

    def testMoveShapeLeftAndDown(self):
        self.shape.moveBy(Point(-10, -20))
        self.assertEqual(Point(-10, -20), self.shape.polyline.vertices[0].position)
        self.assertEqual(Point(-10, 30), self.shape.polyline.vertices[1].position)
        self.assertEqual(Point(90, 30), self.shape.polyline.vertices[2].position)
        self.assertEqual(Point(90, -20), self.shape.polyline.vertices[3].position)

    def testGetSegmentsOfOpenShapeShouldReturnAllSegments(self):
        expected = PolylineSegments()
        expected.append(PolylineSegment(Vertex(Point(0, 0)), Vertex(Point(0, 50))))
        expected.append(PolylineSegment(Vertex(Point(0, 50)), Vertex(Point(100, 50))))
        expected.append(PolylineSegment(Vertex(Point(100, 50)), Vertex(Point(100, 0))))
        expected.append(PolylineSegment(Vertex(Point(100, 0)), Vertex(Point(0, 0))))
        self.assertEqual(expected, self.shape.getSegments())


class ShapeStringRepresentation(unittest.TestCase):
    def testStringRepresentation(self):
        f = Shape('poly')
        self.assertEqual('Shape(polyline=poly)', str(f))
        self.assertEqual("Shape(polyline='poly')", repr(f))


class CreationTests(unittest.TestCase):
    def testcreateRectangularSizingFromBoundsShouldBeCorrect(self):
        sizingBounds = Rectangle(11, 12, 91, 92)

        expectedRectangularSizingPolyline = Polyline(
            CornerVertex(Point(11, 12)),
            CornerVertex(Point(11, 92)),
            CornerVertex(Point(91, 92)),
            CornerVertex(Point(91, 12)),
        )

        rectangularShape = Shape.createRectangularSizingFromBounds(sizingBounds)
        self.assertEqual(expectedRectangularSizingPolyline, rectangularShape.polyline)
        self.assertIsInstance(rectangularShape, Shape)
