# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.backend.workpiececreationstrategy.constants import SizeMode
from interiorcad.backend.workpiececreationstrategy.size import DefaultSizeStrategy
from interiorcad.backend.workpiececreationstrategy.sizing import DefaultSizingStrategy


# Terminology:
# - size: the dimension of the workpiece
# - sizing the dimension/path of the contour used for sizing

class WorkpieceCreationStrategy:
    @staticmethod
    def createSizeStrategy(workpiece, configuration):
        return DefaultSizeStrategy(workpiece, configuration)

    @staticmethod
    def createSizingStrategy(workpiece, configuration):
        return DefaultSizingStrategy(workpiece, configuration)


class MainSurfaceSizingDeciderTests(unittest.TestCase):
    class MockConfiguration:
        jointing = False

    class MockWorkpiece:
        isRectangularBool = True

        def isRectangular(self):
            return self.isRectangularBool

    def setUp(self):
        self.workpiece = self.MockWorkpiece()
        self.configuration = self.MockConfiguration()

    def getMainSizeMode(self, needsOpposite=False):
        strategy = WorkpieceCreationStrategy.createSizeStrategy(self.workpiece, self.configuration)
        return strategy.getMainSizeMode(needsOpposite)

    def testRectangularAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testRectangularAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testRectangularInaccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, self.getMainSizeMode())

    def testRectangularInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testPolylineAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testPolylineAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, self.getMainSizeMode())

    def testPolylineInaccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, self.getMainSizeMode())

    def testPolylineInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())


class MainSurfaceNeedingOppositeSizingDeciderTests(unittest.TestCase):
    class MockConfiguration:
        jointing = False

    class MockWorkpiece:
        isRectangularBool = True

        def isRectangular(self):
            return self.isRectangularBool

    def setUp(self):
        self.workpiece = self.MockWorkpiece()
        self.configuration = self.MockConfiguration()

    def getMainSizeMode(self, needsOpposite=True):
        strategy = WorkpieceCreationStrategy.createSizeStrategy(self.workpiece, self.configuration)
        return strategy.getMainSizeMode(needsOpposite)

    def testRectangularAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testRectangularAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testRectangularInaccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, self.getMainSizeMode())

    def testRectangularInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testPolylineAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testPolylineAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.PRELIMINARY_RECTANGULAR_SIZE, self.getMainSizeMode())

    def testPolylineInaccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.PRELIMINARY_RECTANGULAR_SIZE, self.getMainSizeMode())

    def testPolylineInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())


class MainSurfaceNeedingOppositeJointingSizingDeciderTests(unittest.TestCase):
    class MockConfiguration:
        jointing = True

    class MockWorkpiece:
        isRectangularBool = True

        def isRectangular(self):
            return self.isRectangularBool

    def setUp(self):
        self.workpiece = self.MockWorkpiece()
        self.configuration = self.MockConfiguration()

    def getMainSizeMode(self, needsOpposite=True):
        strategy = WorkpieceCreationStrategy.createSizeStrategy(self.workpiece, self.configuration)
        return strategy.getMainSizeMode(needsOpposite)

    def testRectangularAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testRectangularAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testRectangularInaccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_JOINTING_AND_ADDITION, self.getMainSizeMode())

    def testRectangularInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testPolylineAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())

    def testPolylineAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_JOINTING_AND_ADDITION, self.getMainSizeMode())

    def testPolylineInaccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_JOINTING_AND_ADDITION, self.getMainSizeMode())

    def testPolylineInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getMainSizeMode())


class OppositeSurfaceSizingDeciderTests(unittest.TestCase):
    class MockConfiguration:
        jointing = False

    class MockWorkpiece:
        def isRectangular(self):
            return self.isRectangularBool

        def isPolygonal(self):
            return self.isRectangular()

    def setUp(self):
        self.workpiece = self.MockWorkpiece()
        self.configuration = self.MockConfiguration()
        self.needsMain = True

    def getOppositeSizeMode(self):
        strategy = WorkpieceCreationStrategy.createSizeStrategy(self.workpiece, self.configuration)
        return strategy.getOppositeSizeMode(self.needsMain)

    def testRectangularAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testRectangularAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testRectangularInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testPolylineAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testPolylineAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, self.getOppositeSizeMode())

    def testPolylineInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testRectangularInaccurateSawInlineSizing_WithoutMainSide(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, self.getOppositeSizeMode())

    def testRectangularInaccurateSawInlineSizing_WithMainSide(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = True
        self.assertEqual(SizeMode.SAWN_SIZE, self.getOppositeSizeMode())

    def testPolylineInaccurateSawInlineSizing_WithoutMainSide(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, self.getOppositeSizeMode())

    def testPolylineInaccurateSawInlineSizing_WithMainSide(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = True
        self.assertEqual(SizeMode.FINAL_SIZING_SIZE, self.getOppositeSizeMode())


class OppositeSurfaceJointingSizingDeciderTests(unittest.TestCase):
    class MockConfiguration:
        jointing = True

    class MockWorkpiece:
        def isRectangular(self):
            return self.isRectangularBool

        def isPolygonal(self):
            return self.isRectangular()

    def setUp(self):
        self.workpiece = self.MockWorkpiece()
        self.configuration = self.MockConfiguration()
        self.needsMain = True

    def getOppositeSizeMode(self):
        strategy = WorkpieceCreationStrategy.createSizeStrategy(self.workpiece, self.configuration)
        return strategy.getOppositeSizeMode(self.needsMain)

    def testRectangularAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testRectangularAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testRectangularInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testPolylineAccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testPolylineAccurateSawInlineSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = True
        self.configuration.separateSizing = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_JOINTING_AND_ADDITION, self.getOppositeSizeMode())

    def testPolylineInaccurateSawSeparateSizing(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = True
        self.assertEqual(SizeMode.FINISHED_SIZE, self.getOppositeSizeMode())

    def testRectangularInaccurateSawInlineSizing_WithoutMainSide(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_JOINTING_AND_ADDITION, self.getOppositeSizeMode())

    def testRectangularInaccurateSawInlineSizing_WithMainSide(self):
        self.workpiece.isRectangularBool = True
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = True
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_JOINTING, self.getOppositeSizeMode())

    def testPolylineInaccurateSawInlineSizing_WithoutMainSide(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = False
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_JOINTING_AND_ADDITION, self.getOppositeSizeMode())

    def testPolylineInaccurateSawInlineSizing_WithMainSide(self):
        self.workpiece.isRectangularBool = False
        self.configuration.accurateSaw = False
        self.configuration.separateSizing = False
        self.needsMain = True
        self.assertEqual(SizeMode.FINAL_SIZING_SIZE_WITH_JOINTING, self.getOppositeSizeMode())
