# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

import xml.etree.ElementTree as ET


FIRST_ITEM = 0
NOT_FOUND_INDEX = -1


class XMLParser:
    def parseFromFile(self, fullXMLFilepath):
        tree = ET.parse(fullXMLFilepath)
        self.root = tree.getroot()

    def parseFromString(self, xmlString):
        self.root = ET.fromstring(xmlString)

    def getSimpleValue(self, tagName):
        if self.__tagNameContainsAttribute(tagName):
            return self.__getValueForAttribute(tagName)

        elements = self.root.findall('./' + tagName)
        if elements:
            text = elements[FIRST_ITEM].text
            # We have found an element, so the string is empty, not None as if there would be not element.
            return text if text is not None else ''
        else:
            return None

    def __getValueForAttribute(self, tagNameWithAttribute):
        attributeName = self.__getAttributeNameFrom(tagNameWithAttribute)
        tagNameWithoutAttribute = self.__getTagNameWithoutAttribute(tagNameWithAttribute)

        if self.__tagNameWithAttributeFoundInXMLTree(tagNameWithoutAttribute, attributeName):
            return self.root.findall('./' + tagNameWithoutAttribute + '[@' + attributeName + ']')[FIRST_ITEM].attrib[attributeName]
        else:
            return None

    def __tagNameWithAttributeFoundInXMLTree(self, tagNameWithoutAttribute, attributeName):
        return len(self.root.findall('./' + tagNameWithoutAttribute + '[@' + attributeName + ']')) > 0

    def __getTagNameWithoutAttribute(self, tagName):
        indexOfAttribeStart = tagName.rfind('$')
        return tagName[:indexOfAttribeStart]

    def __getAttributeNameFrom(self, tagName):
        indexOfattribeStart = tagName.rfind('$')
        return tagName[indexOfattribeStart + 1:len(tagName)]

    def __tagNameContainsAttribute(self, tagName):
        return tagName.find('$') != NOT_FOUND_INDEX

    def getAllAttributesOfAllChildrenWithTag(self, xpathOfParent, tagName):
        elements = []
        if len(self.root.findall('./' + xpathOfParent)) > 0:
            for parentNode in self.root.findall('./' + xpathOfParent):
                for element in list(parentNode):
                    if element.tag == tagName:
                        elements.append(dict(element.items()))

        return tuple(elements)

    def getSimpleValueAsBool(self, xpath, default):
        value = self.getSimpleValue(xpath)
        if value is None:
            value = bool(default)
        elif value == "0":
            value = False
        elif value == "1":
            value = True
        else:
            value = bool(value)
        return value

    def getSimpleValueAsString(self, xpath, default):
        value = self.getSimpleValue(xpath)
        if value is None:
            value = str(default)
        return value

    def getSimpleValueAsInt(self, xpath, default):
        value = self.getSimpleValue(xpath)
        if value:
            return int(value)
        else:
            return int(default)

    def getSimpleValueAsFloat(self, xpath, default):
        value = self.getSimpleValue(xpath)
        if value:
            return float(value)
        else:
            return float(default)


class SimpleTests(unittest.TestCase):
    TEST_XML = r"""<?xml version="1.0" ?>
    <tests>
        <unit-tests>
            <boolean>
                <false>0</false>
                <true>1</true>
            </boolean>
            <string>
                <abcd>abcd</abcd>
                <empty-string/>
            </string>
            <int year="2012"/>
            <float pi="3.14"/>
        </unit-tests>
    </tests>"""

    def setUp(self):
        self.xmlParser = XMLParser()
        self.xmlParser.parseFromString(SimpleTests.TEST_XML)

    def testZeroShouldBeFalseAsBoolean(self):
        self.assertFalse(self.xmlParser.getSimpleValueAsBool("unit-tests/boolean/false", True))

    def testOneShouldBeTrueAsBoolean(self):
        self.assertTrue(self.xmlParser.getSimpleValueAsBool("unit-tests/boolean/true", False))

    def testADefaultOfZeroShouldBeFalseAsBoolean(self):
        self.assertFalse(self.xmlParser.getSimpleValueAsBool("unit-tests/entry-does-not-exist", False))

    def testADefaultOfOneShouldBeTrueAsBoolean(self):
        self.assertTrue(self.xmlParser.getSimpleValueAsBool("unit-tests/entry-does-not-exist", True))

    def testAStringShouldbeReturned(self):
        self.assertEqual("abcd", self.xmlParser.getSimpleValueAsString("unit-tests/string/abcd", None))

    def testADefaultStringShouldbeReturned(self):
        self.assertEqual("I am the default!", self.xmlParser.getSimpleValueAsString("unit-tests/entry-does-not-exist", "I am the default!"))

    def testAnIntShouldBeReturned(self):
        self.assertEqual(2012, self.xmlParser.getSimpleValueAsInt("unit-tests/int/$year", None))

    def testADefaultIntShouldBeReturned(self):
        self.assertEqual(2013, self.xmlParser.getSimpleValueAsInt("unit-tests/entry-does-not-exist", 2013))

    def testAFloatShouldBeReturned(self):
        self.assertEqual(3.14, self.xmlParser.getSimpleValueAsFloat("unit-tests/float/$pi", None))

    def testADefaultFloatShouldBeReturned(self):
        self.assertEqual(3.14, self.xmlParser.getSimpleValueAsFloat("unit-tests/entry-does-not-exist", 3.14))

    def testAnEmptyStringShouldBeReturnedAsSuch(self):
        self.assertEqual('', self.xmlParser.getSimpleValueAsString('unit-tests/string/empty-string', 'WE DO NOT WANT THE DEFAULT'))

    def testNotExistingStringShouldReturnTheDefault(self):
        self.assertEqual('Default', self.xmlParser.getSimpleValueAsString('unit-tests/string/not-Existing-string', 'Default'))

    def testAnEmptyStringAsIntShouldBeReturnedTheDefault(self):
        self.assertEqual(4321, self.xmlParser.getSimpleValueAsInt('unit-tests/string/empty-string', 4321))

    def testAnEmptyStringAsFloatShouldBeReturnedTheDefault(self):
        self.assertEqual(3.14, self.xmlParser.getSimpleValueAsFloat('unit-tests/string/empty-string', 3.14))

    def testEmptyString(self):
        self.assertEqual('', self.xmlParser.getSimpleValue('unit-tests/string/empty-string'))

    def testNotExistingString(self):
        self.assertEqual(None, self.xmlParser.getSimpleValue('unit-tests/string/not-Existing-string'))