# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.interface.geometry.edges import EdgeData


class ArdisExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'Ardis'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = ';'

        self.EXTENSION = self.getCuttinglistConfiguration().get('file-extension', 'csv')
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

        # grain 1=L or 2=W must be in the grain direction - 0=blank = no grain - 3=E = L or W but consistent within the job
        self.ARDIS_GRAIN_DEFINITION = self.getCuttinglistConfiguration().get('ardis-grain-definition', '0,1,2').split(',')

    def getDecimalSeparator(self, **keywords):
        return '.'

    def getHeaderLine(self, **keywords):
        return (
            'PartMat',
            'PartD',  # grain
            'PartL',
            'PartW',
            'PartQty',
            'PartRef',
            'PartDue',
            'PartRemark',
            'PartLMin',
            'PartWMin',
            'PartEdge1',
            'PartEdge2',
            'PartEdge3',
            'PartEdge4',
            'PartEdgeSeq',
            'PartExt01',
            'PartExt02',
            'PartExt03',
            'PartExt04',
            'PartExt05',
            'PartExt06',
            'PartExt07',
            'PartExt08',
            'PartExt09',
            'PartExt10',
            'PartExt11',
            'PartExt12',
            'PartExt13',
            'PartExt14',
            'PartExt15',
            'PartExt16',
            'PartExt17',
            'PartExt18',
            'PartExt19',
            'PartRemark2',
        )

    def getGrain(self, grain):
        # grain 1=L or 2=W must be in the grain direction - 0=blank = no grain - 3=E = L or W but consistent within the job
        if grain == self.GRAIN_NONE:  # Ardis 0
            return self.ARDIS_GRAIN_DEFINITION[0]
        elif grain == self.GRAIN_CROSSWISE:  # Ardis 2
            return self.ARDIS_GRAIN_DEFINITION[2]
        else:  # Ardis 1 / self.GRAIN_LENGTHWISE
            return self.ARDIS_GRAIN_DEFINITION[1]

    def getExportSpecificEdgeCode(self, **keywords):
        def edgeJointIsMiter():
            for edge in entry.edges.rectangular:  # irrelevant of lengthFollowsGrainDirection
                if edge.jointStart == EdgeData.JOINT_MITER or edge.jointEnd == EdgeData.JOINT_MITER:
                    return True

        entry = keywords['entry']
        edgecode = entry.edgesGrainDirection.edgecode if keywords['lengthFollowsGrainDirection'] else entry.edges.edgecode

        if not edgecode:
            return ''
        elif edgeJointIsMiter():
            return '0'  # 45° /
        elif edgecode.startswith('3') or edgecode.startswith('4'):
            return '1'  # Vertical first |
        else:
            return '2'  # Horizontal first -

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']
        workpiece = keywords['workpiece']

        unitname = self.asString(data=positionInfo['CabinetName'], **keywords)
        unitwidth = self.asCoordinate(key='unit_unitwidth', uuid=workpiece.parentUUID, **keywords)
        unitdepth = self.asCoordinate(key='unit_unitdepth', uuid=workpiece.parentUUID, **keywords)
        unitheight = self.asCoordinate(key='unit_unitheight', uuid=workpiece.parentUUID, **keywords)

        edgecode = entry.edges.edgecode  # lengthFollowsGrainDirection = False

        return (
            self.asString(data=entry.itemID, **keywords),
            self.getGrain(entry.grain),
            self.asCoordinate(data=entry.boardLength, **keywords),
            self.asCoordinate(data=entry.boardWidth, **keywords),
            self.asInt(data=entry.count, **keywords),
            self.asString(data=entry.name, **keywords),
            self.asString(data=projectInfo['DeliveryDate'], **keywords),
            self.asStringSingleLine(key='part_comment', **keywords),
            '',
            '',
            self.asString(data=entry.edges.front.supplierID, **keywords),
            self.asString(data=entry.edges.back.supplierID, **keywords),
            self.asString(data=entry.edges.left.supplierID, **keywords),
            self.asString(data=entry.edges.right.supplierID, **keywords),
            self.getExportSpecificEdgeCode(lengthFollowsGrainDirection=False, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),  # PartExt01
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),
            self.asString(data=projectInfo['ProjectID'], **keywords),
            self.asString(data=projectInfo['ProjectCommission'], **keywords),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),
            self.asString(data=positionInfo['PositionNumber'], **keywords),
            self.asString(data=positionInfo['PositionDescription'], **keywords),
            self.asCoordinate(data=entry.edges.front.thickness, **keywords),  # PartExt09
            self.asCoordinate(data=entry.edges.back.thickness, **keywords),
            self.asCoordinate(data=entry.edges.left.thickness, **keywords),
            self.asCoordinate(data=entry.edges.right.thickness, **keywords),

            self.asString(key='part_covering1', **keywords),  # PartExt13
            self.asString(key='part_covering2', **keywords),
            self.asString(key='part_laminate1', **keywords),
            self.asString(key='part_laminate2', **keywords),
            self.asString(key='part_laminate3', **keywords),
            '{unitname}'.format(unitname=unitname),  # PartExt18
            '{width} x {height} x {depth}'.format(width=unitwidth, height=unitheight, depth=unitdepth) if unitname else '',
            self.asString(data=edgecode, **keywords) if edgecode else '',
        )
