# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import datetime
import io
import itertools
import logging
import os
from collections import OrderedDict

import interiorcad.backend.clipboard
from interiorcad.backend.cuttinglistentry import WorkpieceEntry, FittingEntry, CoveringEntry, BoughtInEntry, PIOContainerEntry
from interiorcad.backend.datatracker import datatracker
from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.joint import JointingInfo
from interiorcad.backend.localization import getLocalizedString, CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP, CUTTINGLISTEXPORT_MESSAGES, CUTTINGLIST_RESOURCEGROUP, getConstructionGroupName, getConstructionGroupComponentName, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.backend.loggingofexcludedcgcs import LoggingOfExcludedConstructionGroupComponents
from interiorcad.backend.sorter.entrycriteriasorter import EntryCriteriaSorter
from interiorcad.backend.systemlocale import SystemLocale
from interiorcad.backend.workpiece import Workpiece
from interiorcad.backend.workpiece.creator import WorkpieceCreator
from interiorcad.interface.configuration import CuttingListGlobalExportConfiguration
from interiorcad.interface.cuttinglistconfiguration import CuttingListConfiguration
from interiorcad.interface.geometry import EPSILON_DIGITS
from interiorcad.interface.utilities.formatter import Formatter


class CuttinglistExportBase:
    OUTPUT_END_OF_LINE = '\r\n'
    OUTPUT_ENCODING = 'windows-1252'
    OUTPUT_ENCODING_ERROR_MODE = 'ignore'

    CARRIAGE_RETURN_REPLACEMENT = ' '

    ASCENDING = EntryCriteriaSorter.ASCENDING
    DESCENDING = EntryCriteriaSorter.DESCENDING

    GRAIN_LENGTHWISE, GRAIN_CROSSWISE, GRAIN_NONE = 0, 1, 2

    def __init__(self, **keywords):
        self.configuration = keywords.get('configuration', CuttingListGlobalExportConfiguration())
        self.cuttinglistConfiguration = keywords.get('cuttinglistConfiguration', CuttingListConfiguration())
        self.useCuttinglistConfiguration(**keywords)
        self.formatter = Formatter(
            coordinatePrecision=self.getCoordinatePrecision(),
            anglePrecision=self.getAnglePrecision(),
            decimalSeparator=self.getDecimalSeparator(**keywords),
            significantDecimalsOnly=self.shouldKeepSignificantDecimalsOnly()
        )
        logging.raiseExceptions = False
        if keywords.get('activateLogging', True) and not logging.getLogger().hasHandlers():
            self.logFileName = None
            self.loggingHandler = None
            self.initLogging(**keywords)

        self.DEFAULT_SORTING_CRITERIA = 'workpiecesfirst itemid length width grain description position cabinetname'
        self.POSITION_SORTING_CRITERIA = 'position cabinetname workpiecesfirst itemid length width grain description'
        self.CONFIGURED_SORTING_CRITERIA = self.cuttinglistConfiguration.get('sorting-criteria', self.getDefaultSortingCriteria(**keywords) if hasattr(self, 'getDefaultSortingCriteria') else self.DEFAULT_SORTING_CRITERIA)
        # getDefaultSortingCriteria() is only a compatibility function for old code and individual exportstarters (up to interiorcad 2018 F3); do NOT use for new code

        self.SORTING_CRITERIA_DEFINITIONS = OrderedDict((
            ('workpiecesfirst', {'entry': lambda entry: entry.__class__.__name__, 'type': 'string', 'order': self.DESCENDING}),  # to be sure that Custom Parts are in front of Coverings and Fittings
            ('itemid', {'entry': lambda entry: entry.itemID or '', 'type': 'string', 'order': self.ASCENDING}),
            ('length', {'entry': lambda entry: round(entry.boardLength or 0.0, EPSILON_DIGITS), 'type': 'float', 'order': self.DESCENDING}),
            ('width', {'entry': lambda entry: round(entry.boardWidth or 0.0, EPSILON_DIGITS), 'type': 'float', 'order': self.DESCENDING}),
            ('grain', {'entry': lambda entry: entry.grain or 0, 'type': 'int', 'order': self.ASCENDING}),
            ('description', {'entry': lambda entry: entry.description or '', 'type': 'string', 'order': self.ASCENDING}),
            ('position', {'entry': lambda entry: entry.positionInfo['PositionNumber'] or '0.0', 'type': 'string', 'order': self.ASCENDING}),
            ('cabinetname', {'entry': lambda entry: entry.positionInfo['CabinetName'] or '', 'type': 'string', 'order': self.ASCENDING}),
            ('name', {'entry': lambda entry: entry.name, 'type': 'string', 'order': self.ASCENDING}),
            ('count', {'entry': lambda entry: entry.count, 'type': 'int', 'order': self.DESCENDING}),
        ))

        self.CUMULATE_COVERINGS = None
        self.CUMULATE_FITTINGS = None
        self.CUMULATE_WORKPIECES = None

        self.uuidsAndCounts = None
        self.fittingsUUIDsAndCounts = None

    def useCuttinglistConfiguration(self, **_keywords):
        self.OUTPUT_FILE_NAME = self.getCuttinglistConfiguration().get('output-file-name', '{ExportFormat}')
        self.OUTPUT_FILE_NAME_POSTFIX_POSITIONS = self.getCuttinglistConfiguration().get('output-file-name-postfix-positions', '')
        self.OUTPUT_FILE_NAME_POSTFIX_TIMESTAMP = self.getCuttinglistConfiguration().get('output-file-name-postfix-timestamp', '')

    def getCuttinglistConfiguration(self, **_keywords):
        return self.cuttinglistConfiguration

    def initLogging(self, **keywords):
        logging.raiseExceptions = False
        logFileDirectory = ExportHelper.createExportPathWithoutWorkpiece(self.configuration)
        if os.path.basename(logFileDirectory) == '_':  # In case of empty {Position...}_{Position...} at the end of the path. Assumes Project Details before DocumentName, Position Details and CabinetName.
            logFileDirectory = os.path.dirname(logFileDirectory)
        ExportHelper.CreatePathIfNotExists(logFileDirectory)
        self.logFileName = os.path.join(logFileDirectory, self.configuration.logFileName)
        self.createLogFileArchive(**keywords)
        self.loggingHandler = logging.FileHandler(filename=self.logFileName, encoding=self.LOG_FILE_ENCODING, mode='a', delay=True)
        self.loggingHandler.setFormatter(logging.Formatter(fmt='%(levelname)s:%(message)s'))
        logging.root.addHandler(self.loggingHandler)
        logging.root.setLevel(logging.INFO)

    def removeLoggingHandle(self):
        if self.loggingHandler:
            self.loggingHandler.flush()
            logging.root.removeHandler(self.loggingHandler)
            self.loggingHandler.close()
            self.loggingHandler = None

    LOG_FILE_ENCODING = 'utf-8-sig'

    def createLogFileArchive(self, **keywords):
        # Create 'XG Export Log Archive.txt', move older messages and delete existing 'XG Export Log.txt'
        if os.path.isfile(self.logFileName):
            logFileNameArchive = os.path.splitext(self.logFileName)[0] + ' Archive' + os.path.splitext(self.logFileName)[1]
            logFileTime = datetime.datetime.fromtimestamp(os.path.getmtime(self.logFileName)).strftime('%Y-%m-%d %H:%M:%S')

            if not os.path.isfile(logFileNameArchive):
                with open(logFileNameArchive, mode='a', encoding=self.LOG_FILE_ENCODING) as f:
                    headline = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 55, 'This is the archive for all previous messages of XG Export Log.txt in chronological order.')
                    f.write(headline + '\n' + '\n')

            # We are using errors='surrogateescape' to avoid losing data if a user had written a comment in the log file with the wrong encoding
            with open(self.logFileName, mode='r', encoding=self.LOG_FILE_ENCODING, errors='surrogateescape') as f:
                logFileLines = f.readlines()

            with open(logFileNameArchive, mode='a', encoding=self.LOG_FILE_ENCODING, errors='surrogateescape') as f:
                f.write(logFileTime + '\n')
                f.writelines(logFileLines)
                f.write('\n')

            os.remove(self.logFileName)

    def getDecimalSeparator(self, **_keywords):
        return SystemLocale.getDecimalSeparator()

    def getFieldSeparator(self, **_keywords):
        return SystemLocale.getFieldSeparator()

    def openFile(self, exportPath):
        # We do not want that Python changes line-ends without our control.
        return open(exportPath, 'w', newline='', encoding=self.OUTPUT_ENCODING, errors=self.OUTPUT_ENCODING_ERROR_MODE)

    def writeLines(self, **keywords):
        if keywords['lines']:
            outputFile = keywords['outputFile']
            lines = self.OUTPUT_END_OF_LINE.join(keywords['lines'])
            outputFile.write(lines + self.OUTPUT_END_OF_LINE)

    # TODO IT-226 Move into ExportHelper?  --mjk
    def getProjectInfo(self, **keywords):
        return ExportHelper.createProjectInfoOfFirstSelectedCustomPart()

    def prepareExportPath(self, **keywords):
        exportPath = ExportHelper.createExportPathForFirstSelectedCustomPart(self.configuration)
        ExportHelper.CreatePathIfNotExists(exportPath)
        outputName = self.getOutputFileName(exportPath=exportPath, **keywords)
        return os.path.join(exportPath, outputName)

    def getPositionsOfAllCustomParts(self, **_keywords):
        positions = []
        for uuid, _count in ExportDataCollector.getAllCustomPartUUIDsAndCounts():
            positions.append(datatracker.getItemAsString(uuid, 0, 'jobd_positionnumber'))
        positions = sorted(list(set(positions)))
        positions = ','.join(str(x) for x in positions)
        return positions

    def getOutputFileNameVariables(self, **keywords):
        outputFileNameVariables = {'ExportFormat': self.EXPORT_FORMAT}
        outputFileNameVariables.update(ExportHelper.getProjectOutputFileNameVariables(self, **keywords))
        outputFileNameVariables.update({'PositionNumbers': self.getPositionsOfAllCustomParts(**keywords)})
        ExportHelper.updateDictionaryWithDateTime(outputFileNameVariables)
        return outputFileNameVariables

    def getOutputFileNameFromConfigurationForCuttinglistExport(self, **keywords):
        outputFileName = self.OUTPUT_FILE_NAME
        if self.OUTPUT_FILE_NAME_POSTFIX_POSITIONS:
            outputFileName += '_' + self.OUTPUT_FILE_NAME_POSTFIX_POSITIONS
        outputFileName = ExportHelper.createReducedOutputFileNameVariables(self, outputFileName=outputFileName, **keywords)

        if not outputFileName:  # Fallback if the users OUTPUT_FILE_NAME and OUTPUT_FILE_NAME_POSTFIX_POSITIONS is completely based on empty project details
            outputFileName = '{ExportFormat}'

        if self.OUTPUT_FILE_NAME_POSTFIX_TIMESTAMP:
            outputFileName += '_' + self.OUTPUT_FILE_NAME_POSTFIX_TIMESTAMP
        return outputFileName

    def getOutputFileName(self, **keywords):
        return self.getOutputFileNameFromConfigurationForCuttinglistExport(**keywords).format_map(self.getOutputFileNameVariables(**keywords)) + '.' + self.EXTENSION

    def getNCProgramDisplayNameFromPath(self, **keywords):
        displayNameDefinition = keywords['definition']
        path = keywords['path']

        if displayNameDefinition == 'pathonly':
            ncProgramDisplayName = os.path.dirname(path)
        else:
            if 'path' in displayNameDefinition:
                ncProgramDisplayName = path
            else:
                ncProgramDisplayName = os.path.basename(path)

            if 'extension' not in displayNameDefinition:
                ncProgramDisplayName = os.path.splitext(ncProgramDisplayName)[0]

        return ncProgramDisplayName

    def getNCProgramDisplayName(self, **keywords):
        ncProgramPath = keywords['data']

        if isinstance(ncProgramPath, list):
            ncProgramDisplayNameList = []

            for path in ncProgramPath:
                ncProgramDisplayName = self.getNCProgramDisplayNameFromPath(path=path, **keywords)

                if ncProgramDisplayName and ncProgramDisplayName not in ncProgramDisplayNameList:
                    ncProgramDisplayNameList.append(ncProgramDisplayName)

            return ','.join(ncProgramDisplayNameList)

        elif ncProgramPath:
            return self.getNCProgramDisplayNameFromPath(path=ncProgramPath, **keywords)

        else:
            return ''

    def doCuttinglistExport(self, **keywords):
        self.doFileExport(**keywords)

    def doExport(self, **keywords):
        raise NotImplementedError

    def doFileExport(self, **keywords):
        exportPath = self.prepareExportPath(**keywords)

        projectInfo = self.getProjectInfo()
        self.uuidsAndCounts = tuple(ExportDataCollector.getAllCustomPartUUIDsAndCounts())
        self.fittingsUUIDsAndCounts = tuple(ExportDataCollector.generateExportedFittingsUUIDsAndCounts())

        try:
            with self.openFile(exportPath) as outputFile:
                LoggingOfExcludedConstructionGroupComponents().createLoggingOfExcludedConstructionGroupComponents(excludedConstructionGroups=self.getExcludedConstructionGroups())
                self.doExport(outputFile=outputFile, projectInfo=projectInfo)
        except PermissionError as e:
            message = getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 3, "Cannot write {file}. Is the file opened by another program? Has the current user permissions to write the file into the folder?").format(file=exportPath)
            raise RuntimeError(message) from e

    def doClipboardExport(self, **keywords):
        projectInfo = self.getProjectInfo()
        self.uuidsAndCounts = tuple(ExportDataCollector.getAllCustomPartUUIDsAndCounts())
        self.fittingsUUIDsAndCounts = tuple(ExportDataCollector.generateExportedFittingsUUIDsAndCounts())

        with io.StringIO() as output:
            LoggingOfExcludedConstructionGroupComponents().createLoggingOfExcludedConstructionGroupComponents(excludedConstructionGroups=self.getExcludedConstructionGroups())
            self.doExport(outputFile=output, projectInfo=projectInfo)
            interiorcad.backend.clipboard.setText(output.getvalue())

    def getCoordinatePrecision(self, **keywords):
        return self.configuration.coordinatePrecision

    def getAnglePrecision(self, **keywords):
        return self.configuration.angularPrecision

    def shouldKeepSignificantDecimalsOnly(self, **_keywords):
        return self.configuration.significantDecimalsOnly

    def shouldCreateSeparateSizing(self):
        return self.configuration.separateSizing

    def shouldCreateSizingForRectangularPart(self):
        return not self.configuration.accurateSaw

    def shouldCreateSizing(self, **keywords):
        return self.shouldCreateSeparateSizing() or self.shouldCreateSizingForRectangularPart() or not keywords['workpiece'].isRectangular()

    def getJointingInfo(self):
        return JointingInfo.createFromConfiguration(self.configuration)

    def getsizingAddition(self):
        return self.configuration.sizingAddition

    def getAdditionalCoveringAddition(self):
        return self.configuration.additionalCoveringAddition

    def getExcludedConstructionGroups(self):  # Currently only construction group components are supported.
        return self.configuration.excludedConstructionGroups

    def sizingIsNcCompatible(self):
        return self.configuration.sizingText == 'nc-compatible'

    def getSizingText(self):
        return self.configuration.sizingText

    def getSetName(self):
        return self.configuration.setName

    def getDataAsString(self, **keywords):
        return ExportHelper.getDataAsString(**keywords)

    def getDataAsInt(self, **keywords):
        return ExportHelper.getDataAsInt(**keywords)

    def getDataAsFloat(self, **keywords):
        return ExportHelper.getDataAsFloat(**keywords)

    def formatAsString(self, **keywords):
        return self.formatter.formatAsString(data=keywords['data'])

    def formatAsStringSingleLine(self, **keywords):
        return self.CARRIAGE_RETURN_REPLACEMENT.join(self.formatter.formatAsString(data=keywords['data']).split())

    def formatAsInt(self, **keywords):
        return self.formatter.formatAsInt(data=keywords['data'])

    def formatAsCoordinate(self, **keywords):
        return self.formatter.formatAsCoordinate(data=keywords['data'])

    def formatAsSignedCoordinate(self, **keywords):
        return self.formatter.formatAsSignedCoordinate(data=keywords['data'])

    def formatAsAngle(self, **keywords):
        return self.formatter.formatAsAngle(data=keywords['data'])

    def formatAsArea(self, **keywords):
        return self.formatter.formatAsArea(data=keywords['data'])

    def formatAsVolume(self, **keywords):
        return self.formatter.formatAsVolume(data=keywords['data'])

    def asString(self, **keywords):
        keywords['data'] = self.getDataAsString(**keywords)
        return self.formatAsString(**keywords)

    def asStringSingleLine(self, **keywords):
        keywords['data'] = self.getDataAsString(**keywords)
        return self.formatAsStringSingleLine(**keywords)

    def asInt(self, **keywords):
        keywords['data'] = self.getDataAsInt(**keywords)
        return self.formatAsInt(**keywords)

    def asCoordinate(self, **keywords):
        keywords['data'] = self.getDataAsFloat(**keywords)
        return self.formatAsCoordinate(**keywords)

    def asAngle(self, **keywords):
        keywords['data'] = self.getDataAsFloat(**keywords)
        return self.formatAsAngle(**keywords)

    def asArea(self, **keywords):
        keywords['data'] = self.getDataAsFloat(**keywords)
        return self.formatAsArea(**keywords)

    def asVolume(self, **keywords):
        keywords['data'] = self.getDataAsFloat(**keywords)
        return self.formatAsVolume(**keywords)

    def openExportPath(self, **keywords):
        ExportHelper.openExportPath(self.configuration)
        self.openExportLogFile(**keywords)

    def openExportLogFile(self, **keywords):
        if hasattr(self, 'loggingHandler') and self.loggingHandler:
            self.removeLoggingHandle()
            if self.logFileName and os.path.exists(self.logFileName):
                # "Normalize" the path, as it contains forward slashes (/)
                interiorcad.backend.openservices.openFile(os.path.normpath(self.logFileName))

    def getLocalizedOutputFileName(self, index, default):
        return getLocalizedString(CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP, index, default)

    def getDocumentUUID(self, **_keywords):
        return ExportHelper.getDocumentUUID()

    # TODO IT-226 Move into ExportHelper?  --mjk
    def getPositionInfo(self, uuid):
        positionInfo = {}
        ExportHelper.updateDictionaryWithPartData(positionInfo, uuid)
        return positionInfo

    def getInfoForPositionNumber(self, **keywords):
        positionNumber = keywords['positionNumber']
        return ExportHelper.getPositionData(positionNumber)

    def isProjectIDValid(self, **keywords):
        # HACK: we do not want to consider the default project of interiorcad.sta as being valid (see also IC-6968)
        return keywords['ProjectID']

    def getConstructionGroupNameInclCoveringsAndFittings(self, **keywords):
        entry = keywords['entry']
        if entry.isCovering():
            return entry.description
        elif entry.isFitting():
            return self.getDataAsString(key='part_unitgroupname', **keywords)
        else:  # entry.isWorkpiece()
            return getConstructionGroupName(self.getDataAsInt(key='part_constructiongroupid', **keywords), self.getDataAsInt(key='part_constructiongroupcomponentid', **keywords))

    def getConstructionGroupComponentNameInclCoveringsAndFittings(self, **keywords):
        entry = keywords['entry']
        if entry.isCovering():
            return entry.description
        elif entry.isFitting():
            return self.getDataAsString(key='part_unitgroupname', **keywords)
        else:  # entry.isWorkpiece()
            return getConstructionGroupComponentName(self.getDataAsInt(key='part_constructiongroupcomponentid', **keywords))

    def hasCovering(self, **keywords):
        return any((
            self.getDataAsString(key='part_covering1', **keywords),
            self.getDataAsFloat(key='part_covering1thickness', **keywords),
            self.getDataAsString(key='part_covering2', **keywords),
            self.getDataAsFloat(key='part_covering2thickness', **keywords),
        ))

    def __generateWorkpieceEntry(self, uuidAndCount):
        uuid, count = uuidAndCount
        workpieceCreator = WorkpieceCreator(ExportDataCollector, self.getJointingInfo())
        workpiece = workpieceCreator.createNCWorkpiece(uuid)
        workpiece.count *= count
        entry = WorkpieceEntry(workpiece, self.configuration, (not hasattr(self, 'SCOPE_OF_EXPORT') or 'coverings' not in self.SCOPE_OF_EXPORT) and self.hasCovering(uuid=uuid))
        entry.positionInfo = self.getPositionInfo(uuid)
        return entry

    def generateUnsortedWorkpieceEntries(self, **_keywords):
        return map(self.__generateWorkpieceEntry, self.uuidsAndCounts)

    def generateUnsortedFittingEntries(self, **_keywords):
        def generateEntry(uuidAndCount):
            uuid, count = uuidAndCount
            fakeWorkpiece = Workpiece(uuid=uuid)
            fakeWorkpiece.customPartUUID = ExportDataCollector.getFittingParentUUID(uuid)
            fakeWorkpiece.cabinetUUID = ExportDataCollector.getUserPerceivedParentUUID(fakeWorkpiece.customPartUUID)
            # parentUUID is wrongly named, as it is the cabinet UUID, but we cannot rename it for compatibility reasons
            fakeWorkpiece.parentUUID = fakeWorkpiece.cabinetUUID
            entry = FittingEntry(uuid, fakeWorkpiece, count)
            entry.positionInfo = self.getPositionInfo(uuid=fakeWorkpiece.customPartUUID)
            return entry

        return map(generateEntry, self.fittingsUUIDsAndCounts)

    def generateUnsortedCoveringEntriesFromUUIDsAndCounts(self, **keywords):
        def generateEntry(uuid, keyCovering, keyCoveringThickness, indexTranslation, defaultTranslation, workpiece):
            covering = self.getDataAsString(uuid=uuid, key=keyCovering)
            coveringthickness = self.getDataAsFloat(uuid=uuid, key=keyCoveringThickness)
            if covering:
                coveringName = getLocalizedString(CUTTINGLIST_RESOURCEGROUP, indexTranslation, defaultTranslation)
                entry = CoveringEntry(workpiece, self.configuration, covering, coveringthickness)
                entry.positionInfo = self.getPositionInfo(uuid)
                entry.name = '{name} {covering}'.format(name=workpiece.name, covering=coveringName)
                entry.description = getLocalizedString(CUTTINGLIST_RESOURCEGROUP, 34, 'Covering')
                return entry

        def generateEntries(uuidAndCount):
            uuid, count = uuidAndCount
            workpieceCreator = WorkpieceCreator(ExportDataCollector, self.getJointingInfo())
            workpiece = workpieceCreator.createNCWorkpiece(uuid)
            workpiece.count *= count

            entry1 = generateEntry(uuid, 'part_covering1', 'part_covering1thickness', 25, 'Covering 1', workpiece)
            if entry1:
                yield entry1

            entry2 = generateEntry(uuid, 'part_covering2', 'part_covering2thickness', 26, 'Covering 2', workpiece)
            if entry2:
                yield entry2

        return itertools.chain.from_iterable(map(generateEntries, keywords['uuidsAndCounts']))

    def generateUnsortedCoveringEntries(self, **keywords):
        return self.generateUnsortedCoveringEntriesFromUUIDsAndCounts(uuidsAndCounts=self.uuidsAndCounts, **keywords)

    def generateUnsortedBoughtInEntries(self, **_keywords):
        def generateEntry(uuidAndCount):
            uuid, count = uuidAndCount
            fakeWorkpiece = Workpiece(uuid=uuid)
            fakeWorkpiece.customPartUUID = None
            fakeWorkpiece.cabinetUUID = ExportDataCollector.getUserPerceivedParentUUID(uuid)
            fakeWorkpiece.parentUUID = ExportDataCollector.getPIOContainerUUID(uuid)  # The corresponding drawer
            entry = BoughtInEntry(uuid, fakeWorkpiece, count)
            entry.positionInfo = self.getPositionInfo(uuid=fakeWorkpiece.cabinetUUID)
            # We must set 'CabinetName' explicitly, as getPositionInfo does not assume that the given UUID is one of a PIO container -- NG
            entry.positionInfo['CabinetName'] = ExportDataCollector.getUniqueName(uuid=fakeWorkpiece.cabinetUUID)
            return entry

        return map(generateEntry, ExportDataCollector.getAllBoughtInsUUIDsAndCounts())

    def generateUnsortedPIOContainersEntries(self, **_keywords):
        def generateEntry(uuidAndCount):
            uuid, count = uuidAndCount
            fakeWorkpiece = Workpiece(uuid=uuid)
            fakeWorkpiece.customPartUUID = None
            entry = PIOContainerEntry(uuid, fakeWorkpiece, count)
            entry.positionInfo = self.getPositionInfo(uuid=uuid)
            # We must set 'CabinetName' explicitly, as getPositionInfo does not assume that the given UUID is one of a PIO container -- NG
            entry.name = entry.positionInfo['CabinetName'] = ExportDataCollector.getUniqueName(uuid=uuid)
            return entry

        freeCustomParts, pioContainers, _childCustomParts = ExportDataCollector.getSplitCustomPartAndPIOContainerUUIDsAndCounts()
        return itertools.chain(
            map(generateEntry, pioContainers),
            map(self.__generateWorkpieceEntry, freeCustomParts) if 'standalone' in self.SCOPE_OF_EXPORT else {},
        )

    # Old name (with typo) up to interiorcad 2023 F4.1
    generateUnsortedPIOConatinersEntries = generateUnsortedPIOContainersEntries

    def generateUnsortedEntries(self, **keywords):
        return itertools.chain(
            self.generateUnsortedWorkpieceEntries(**keywords) if 'parts' in self.SCOPE_OF_EXPORT else {},
            self.generateUnsortedFittingEntries(**keywords) if 'fittings' in self.SCOPE_OF_EXPORT else {},
            self.generateUnsortedCoveringEntries(**keywords) if 'coverings' in self.SCOPE_OF_EXPORT else {},
            self.generateUnsortedBoughtInEntries(**keywords) if 'boughtins' in self.SCOPE_OF_EXPORT else {},
            self.generateUnsortedPIOContainersEntries(**keywords) if 'piocontainers' in self.SCOPE_OF_EXPORT else {},
        )

    # compatibility function for old code (up to interiorcad 2018 F3); do NOT use for new code
    def generateSortedWorkpieceAndFittingEntries(self, keywords):
        return self.generateSortedEntries(**keywords)

    def generateSortedEntries(self, **keywords):
        entries = self.generateUnsortedEntries(**keywords)
        entries = self.SortedEntriesByCriteria(entries=entries, **keywords)
        entries = self.getCumulatedEntries(entries=entries, **keywords)
        return entries

    def SortedEntriesByCriteria(self, **keywords):
        sorter = EntryCriteriaSorter(self.getSortingCriteria(**keywords))
        entries = list(keywords['entries'])
        sorter.sort(entries)
        return entries

    def getSortingCriteria(self, **keywords):
        return self.extractSortingCriteriaConfiguration(sortingCriteria=self.CONFIGURED_SORTING_CRITERIA, **keywords)

    def extractSortingCriteriaConfiguration(self, **keywords):
        sortingCriteria = []
        for criterium in keywords['sortingCriteria'].split():
            if criterium.lower() in self.SORTING_CRITERIA_DEFINITIONS:
                sortingCriteria.append(self.SORTING_CRITERIA_DEFINITIONS[criterium.lower()])
            else:
                message = getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 2, "Unknown sorting criterium: {criterium}! Accepted values: {criteria}")
                logging.warning(message.format(criterium=criterium, criteria=tuple(self.SORTING_CRITERIA_DEFINITIONS)))

        if self.CUMULATE_WORKPIECES and 'description' in self.CUMULATE_WORKPIECES and 'description' not in keywords['sortingCriteria']:
            sortingCriteria.append(self.SORTING_CRITERIA_DEFINITIONS['description'])

        return sortingCriteria

    def getCumulatedEntries(self, **keywords):
        if self.CUMULATE_WORKPIECES == 'materiallist-material-and-dimension':
            workpiecesAccumulationType = WorkpieceEntry.CUMULATE_SAME_MATERIAL_AND_DIMENSION
            workpieceMergeType = WorkpieceEntry.MERGE_FOR_MATERIAL_LIST
        elif self.CUMULATE_WORKPIECES == 'materiallist-material-and-dimension-and-description':
            workpiecesAccumulationType = WorkpieceEntry.CUMULATE_SAME_MATERIAL_AND_DIMENSION_AND_DESCRIPTION
            workpieceMergeType = WorkpieceEntry.MERGE_FOR_MATERIAL_LIST
        elif self.CUMULATE_WORKPIECES == 'partslist-material-and-dimension':
            workpiecesAccumulationType = WorkpieceEntry.CUMULATE_SAME_MATERIAL_AND_DIMENSION
            workpieceMergeType = WorkpieceEntry.MERGE_FOR_PARTS_LIST
        elif self.CUMULATE_WORKPIECES == 'partslist-material-and-dimension-and-description':
            workpiecesAccumulationType = WorkpieceEntry.CUMULATE_SAME_MATERIAL_AND_DIMENSION_AND_DESCRIPTION
            workpieceMergeType = WorkpieceEntry.MERGE_FOR_PARTS_LIST
        else:
            workpiecesAccumulationType = WorkpieceEntry.CUMULATE_NONE
            workpieceMergeType = None

        accumulationOptions = {
            'workpiecesAccumulationType': workpiecesAccumulationType,
            'workpiecesMergeType': workpieceMergeType,
            'shouldAccumulateFittings': self.CUMULATE_FITTINGS,
            'shouldAccumulateCoverings': self.CUMULATE_COVERINGS,
            'export': self,
        }

        pendingEntry = None

        for entry in keywords['entries']:
            if entry is not None and type(entry) == type(pendingEntry) and entry.canBeCumulated(pendingEntry, **accumulationOptions):
                self.mergeEntry(pendingEntry, entry, **accumulationOptions)
            else:
                if pendingEntry:
                    yield pendingEntry
                pendingEntry = entry

        if pendingEntry:
            yield pendingEntry

    def mergeEntry(self, pendingEntry, newEntry, **accumulationOptions):
        pendingEntry.count += newEntry.count
        pendingEntry.positionInfo = ExportHelper.mergePositionInfos(pendingEntry.positionInfo, newEntry.positionInfo, placeholder=getLocalizedString(CUTTINGLIST_RESOURCEGROUP, 31, 'n/a'))
        pendingEntry.name = ExportHelper.mergeSimpleStringsAsCommaSeparatedString(pendingEntry.name, newEntry.name)
        pendingEntry.description = ExportHelper.mergeSimpleStringsAsCommaSeparatedString(pendingEntry.description.strip(), newEntry.description.strip())

        self.mergeNCProgramPaths(pendingEntry, newEntry)
        pendingEntry.modifyAfterMerge(**accumulationOptions)

    def mergeNCProgramPaths(self, pendingEntry, newEntry):
        pendingEntry.ncProgramPathSizing = ExportHelper.mergeSimpleStringsAsList(pendingEntry.ncProgramPathSizing, newEntry.ncProgramPathSizing)
        pendingEntry.ncProgramPathMain = ExportHelper.mergeSimpleStringsAsList(pendingEntry.ncProgramPathMain, newEntry.ncProgramPathMain)
        pendingEntry.ncProgramPathOpposite = ExportHelper.mergeSimpleStringsAsList(pendingEntry.ncProgramPathOpposite, newEntry.ncProgramPathOpposite)
        pendingEntry.ncProgramNameSizing = ExportHelper.mergeSimpleStringsAsList(pendingEntry.ncProgramNameSizing, newEntry.ncProgramNameSizing)
        pendingEntry.ncProgramNameMain = ExportHelper.mergeSimpleStringsAsList(pendingEntry.ncProgramNameMain, newEntry.ncProgramNameMain)
        pendingEntry.ncProgramNameOpposite = ExportHelper.mergeSimpleStringsAsList(pendingEntry.ncProgramNameOpposite, newEntry.ncProgramNameOpposite)
