# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import csv

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.backend.filenamecleaner import FilenameCleaner


class BestOptFileExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'BestOpt'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = ' '
        # This exports takes care directly of quoting text fields. We must switch off the automatic quoting of the CSV export
        self.csvquoting = csv.QUOTE_NONE
        self.csvquotechar = None

        self.EXTENSION = 'PAR'
        self.BEST_OPT_ARTICLE = self.getCuttinglistConfiguration().get('best-opt-article', '{CabinetName}')
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')  # | partsandfittings | fittingsonly
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

    # modified CSVWriter, because csvdelimiter is used in strings, but should also be escaped!
    def createCSVWriter(self, **keywords):
        csvWriter = csv.writer(
            keywords['outputFile'],
            delimiter=self.csvdelimiter,
            lineterminator=self.OUTPUT_END_OF_LINE,
            quotechar=self.csvquotechar,
            quoting=self.csvquoting,
            dialect=self.csvdialect,
            escapechar=' '
        )
        return csvWriter

    def getDecimalSeparator(self, **keywords):
        return '.'

    # header
    def getHeaderLine(self, **keywords):
        projectInfo = self.getProjectInfo()
        if self.isProjectIDValid(**projectInfo):
            bestOptJobname = ' '.join((
                self.asString(data=projectInfo['ProjectID'], **keywords),
                self.asString(data=projectInfo['ProjectCommission'], **keywords),
                self.asString(data=projectInfo['ProjectConstruction'], **keywords)
            ))
        else:
            bestOptJobname = self.asString(data=projectInfo['DocumentName'], **keywords)

        return (
            '/N',
            bestOptJobname,
        )

    # elements
    def getBestOptArticleVariables(self, **keywords):
        projectInfo = self.getProjectInfo()
        positionInfo = keywords['positionInfo']
        return {
            'CabinetName': positionInfo['CabinetName'],
            'ProjectID': projectInfo['ProjectID'],
            'PositionNumber': positionInfo['PositionNumber'],
        }

    def deleteUnwantedUnderscores(self, element):
        return FilenameCleaner.getSimplifiedMultipleUnderscores(element).strip('_')

    def getBestOptArticle(self, **keywords):
        projectInfo = self.getProjectInfo()
        if not self.isProjectIDValid(**projectInfo) and (self.BEST_OPT_ARTICLE == '{PositionNumber}' or self.BEST_OPT_ARTICLE == '{ProjectID}_{PositionNumber}'):
            self.BEST_OPT_ARTICLE = '{CabinetName}'
        return self.deleteUnwantedUnderscores(self.BEST_OPT_ARTICLE.format_map(self.getBestOptArticleVariables(**keywords)))

    def quoteElement(self, element):
        quoteChar = '"'
        return quoteChar + element + quoteChar

    def getSimplifiedGrain(self, grain):
        # 0= part could be rotated, 1= part mustn't be rotated
        if grain == self.GRAIN_NONE:
            return 0
        else:  # self.GRAIN_LENGTHWISE, self.GRAIN_CROSSWISE:
            return 1

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']

        return (
            '/T',
            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardThickness, **keywords),
            self.asInt(data=entry.count, **keywords),
            self.getSimplifiedGrain(entry.grain),
            '0',
            '0',
            self.quoteElement(element=self.asString(data=entry.name, **keywords)),  # Teilebezeichnung
            self.quoteElement(element=self.asString(data=entry.itemID, **keywords)),  # Materialbezeichnung
            self.quoteElement(element=self.asString(data=self.getBestOptArticle(**keywords), **keywords)),  # ARTIKEL (Dieser Eintrag wird im Ausdruck als eindeutige Plattenkennung verwendet, max. 30 Zeichen)
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords)),  # Edge Left
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords)),  # Edge Right
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords)),  # Edge Front (Unten)
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords)),  # Edge Back (Oben)

            # Exact 10 additional fields (Zusatztext)
            self.quoteElement(element=self.asString(data=projectInfo['ProjectID'], **keywords)),
            self.quoteElement(element=self.asString(data=positionInfo['PositionNumber'], **keywords)),
            self.quoteElement(element=self.asString(data=positionInfo['PositionDescription'], **keywords)),
            self.quoteElement(element=self.asString(data=positionInfo['CabinetName'], **keywords)),
            self.quoteElement(element=self.asString(key='part_covering1', **keywords)),
            self.quoteElement(element=self.asString(key='part_covering2', **keywords)),
            self.quoteElement(element=self.asString(key='part_laminate1', **keywords)),
            self.quoteElement(element=self.asString(key='part_laminate2', **keywords)),
            self.quoteElement(element=self.asString(key='part_laminate3', **keywords)),
            self.quoteElement(element=self.asStringSingleLine(key='part_comment', **keywords)),

            # Frontinformations from interiorcad 2021 BF4.4 on (instead unused NC program names)
            '""',  # Frontinformation Left
            '""',  # Frontinformation Right
            '""',  # Frontinformation Front (Unten)
            '""',  # Frontinformation Back (Oben)
        )
