# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.interface.geometry.edges import EdgeData


class BormBusinessExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'BormBusiness'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = '\t'
        self.csvquotechar = '"'

        self.counter = 0

        self.EXTENSION = 'txt'
        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsandfittings')  # | partsonly | fittingsonly
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

        self.LENGTH_FOLLOWS_GRAIN_DIRECTION = self.getCuttinglistConfiguration().getAsBool('length-follows-grain-direction', True)

    def getDecimalSeparator(self, **keywords):
        return '.'

    def getHeaderLine(self, **keywords):
        pass

    # BormBusiness can handle dimensions as defined in interiorcad with separate parameter for grain crosswise and alternatively length following grain direction
    def getSimplifiedGrain(self, grain):  # For GrainDirection with switched length and width and rotated edges (length follows grain direction)
        if grain == self.GRAIN_LENGTHWISE or grain == self.GRAIN_CROSSWISE:
            return 0
        else:  # grain == self.GRAIN_NONE
            return -1

    def getGrain(self, grain):  # For Length/Width as defined in interiorcad, dimensions and edges stay the same.
        if grain == self.GRAIN_LENGTHWISE:
            return 0
        elif grain == self.GRAIN_CROSSWISE:  # Actually impossible, because using edgesGrainDirection.
            return 1
        else:  # grain == self.GRAIN_NONE
            return -1

    def entryDependingLengthFollowsGrainDirection(self, trueValue, falseValue):
        if self.LENGTH_FOLLOWS_GRAIN_DIRECTION:
            return trueValue
        else:
            return falseValue

    # Kantenstoßinformationen / Edge joint information
    def getJointAsNumberPerCorner(self, lhs, rhs, lengthwise):
        lhsJointEnd = lhs.jointEnd
        rhsJointStart = rhs.jointStart

        if lhsJointEnd == EdgeData.JOINT_NOEDGE and rhsJointStart == EdgeData.JOINT_NOEDGE:
            lhsSupplierID = lhs.supplierID
            rhsSupplierID = rhs.supplierID

            if not lhsSupplierID and not rhsSupplierID:
                return '00'
            elif lhsSupplierID and not rhsSupplierID:
                return '01'
            elif not lhsSupplierID and rhsSupplierID:
                return '04'
            else:
                assert False, "No joint but edges at both sides is not allowed"
        elif lhsJointEnd == EdgeData.JOINT_CONTINUOUS and rhsJointStart == EdgeData.JOINT_DISCONTINUOUS:
            return '02' if lengthwise else '03'
        elif lhsJointEnd == EdgeData.JOINT_DISCONTINUOUS and rhsJointStart == EdgeData.JOINT_CONTINUOUS:
            return '03' if lengthwise else '02'
        elif lhsJointEnd == rhsJointStart == EdgeData.JOINT_MITER:
            return '05'
        else:
            assert False, "Unsupported edge jointing case: " + str(lhs) + " " + str(rhs)

    def getExportSpecificEdgeCode(self, **keywords):
        entry = keywords['entry']
        edgesDependingLengthFollowsGrainDirection = entry.edgesGrainDirection if keywords['lengthFollowsGrainDirection'] else entry.edges

        return '.'.join((
            self.getJointAsNumberPerCorner(lhs=edgesDependingLengthFollowsGrainDirection.front, rhs=edgesDependingLengthFollowsGrainDirection.left, lengthwise=True),
            self.getJointAsNumberPerCorner(lhs=edgesDependingLengthFollowsGrainDirection.left, rhs=edgesDependingLengthFollowsGrainDirection.back, lengthwise=False),
            self.getJointAsNumberPerCorner(lhs=edgesDependingLengthFollowsGrainDirection.back, rhs=edgesDependingLengthFollowsGrainDirection.right, lengthwise=True),
            self.getJointAsNumberPerCorner(lhs=edgesDependingLengthFollowsGrainDirection.right, rhs=edgesDependingLengthFollowsGrainDirection.front, lengthwise=False),
        ))

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']
        workpiece = keywords['workpiece']

        self.counter += 1

        return (
            self.asString(data=projectInfo['DocumentName'], **keywords),
            self.asString(data=projectInfo['ProjectID'], **keywords),
            self.asString(data=projectInfo['ProjectCommission'], **keywords),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),
            self.asString(data=positionInfo['PositionDescription'], **keywords),

            self.asCoordinate(key='unit_unitwidth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelX
            self.asCoordinate(key='unit_unitdepth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelY
            self.asCoordinate(key='unit_unitheight', uuid=workpiece.parentUUID, **keywords),  # STIMoebelZ
            self.asString(data=positionInfo['PositionNumber'], **keywords),
            self.asString(data=entry.name, **keywords),  # PositionName / STK_BESTANDTEIL
            '',  # PositionAnzahl

            self.counter,
            self.asString(data=positionInfo['CabinetName'], **keywords),  # BaugruppenName / Name Objektgruppierung (STK_BESTANDTEIL)
            self.asString(data=entry.description, **keywords),  # Name
            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.boardLengthGrainDirection, entry.boardLength), **keywords),
            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.boardWidthGrainDirection, entry.boardWidth), **keywords),
            '',  # Flaeche

            self.asString(data=entry.itemID, **keywords),
            self.asCoordinate(data=entry.boardThickness, **keywords),
            self.asInt(data=entry.count, **keywords),

            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.front.supplierID, entry.edges.front.supplierID), **keywords),
            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.front.thickness, entry.edges.front.thickness), **keywords),
            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.back.supplierID, entry.edges.back.supplierID), **keywords),
            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.back.thickness, entry.edges.back.thickness), **keywords),
            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.left.supplierID, entry.edges.left.supplierID), **keywords),
            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.left.thickness, entry.edges.left.thickness), **keywords),
            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.right.supplierID, entry.edges.right.supplierID), **keywords),
            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.right.thickness, entry.edges.right.thickness), **keywords),

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),

            self.entryDependingLengthFollowsGrainDirection(self.getSimplifiedGrain(entry.grain), self.getGrain(entry.grain)),

            self.asString(key='part_laminate1', **keywords),
            self.asString(key='part_laminate2', **keywords),
            self.asString(key='part_laminate3', **keywords),
            '',  # Laminate 4
            '',  # Laminate 5
            '',  # Laminate 6

            self.asStringSingleLine(key='part_comment', **keywords),
            # End of Classic VectorWorksImportDefinition. The Classic export delivers a lot more values, but they are ignored.

            # The new Python based cuttinglist export adds new fields instead.
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),

            self.getExportSpecificEdgeCode(lengthFollowsGrainDirection=self.LENGTH_FOLLOWS_GRAIN_DIRECTION, **keywords),  # Kantenstoßinformationen / Edge joint information
        )
