# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.cuttinglistexport.schnittprofitexport import SchnittProfitExport


class CADmaticExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'CADmatic'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = self.getCuttinglistConfiguration().get('csv-field-delimiter', ';')

        self.EXTENSION = 'csv'
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')
        self.NC_PROGRAM_DISPLAY_PATH_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-path-definition', 'pathandnamewithextension')

    def getDecimalSeparator(self, **keywords):
        DECIMAL_SEPARATOR = self.getCuttinglistConfiguration().get('decimal-separator', 'system')
        if DECIMAL_SEPARATOR == '.':
            return '.'
        elif DECIMAL_SEPARATOR == ',':
            return ','
        else:  # SystemLocale.getDecimalSeparator()
            return super().getDecimalSeparator(**keywords)

    # header
    def getHeaderLine(self, **keywords):
        pass

    # elements
    def getSimplifiedGrain(self, grain):
        # 0= part could be rotated, 1= part mustn't be rotated
        if grain == self.GRAIN_NONE:
            return 0
        else:  # self.GRAIN_LENGTHWISE, self.GRAIN_CROSSWISE:
            return 1

    # Kantenstoßinformationen / Edge joint information is the same function as in SchnittProfitExport
    def getExportSpecificEdgeCode(self, **keywords):
        return SchnittProfitExport().getExportSpecificEdgeCode(**keywords)

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']
        workpiece = keywords['workpiece']

        return (
            # standard field, must be present in that order
            self.asString(data=entry.name, **keywords),
            self.asString(data=entry.itemID, **keywords),
            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asInt(data=entry.count, **keywords),
            '',
            self.asInt(data=self.getSimplifiedGrain(entry.grain), **keywords),
            self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),
            '0',
            self.asCoordinate(data=entry.finishedSizePartDimensionsGrainDirection.getLength(), **keywords),
            self.asCoordinate(data=entry.finishedSizePartDimensionsGrainDirection.getWidth(), **keywords),
            '',
            '',
            self.getExportSpecificEdgeCode(**keywords),  # Kantenstoßinformationen / Edge joint information
            '',

            # the following fields are optional (nearly the same as SchnittProfit)
            self.asString(data=projectInfo['ProjectID'], **keywords),
            self.asString(data=projectInfo['ProjectCommission'], **keywords),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),
            self.asString(data=projectInfo['DocumentName'], **keywords),

            self.asString(data=positionInfo['PositionNumber'], **keywords),
            self.asString(data=positionInfo['PositionDescription'], **keywords),

            self.asString(data=positionInfo['CabinetName'], **keywords),
            self.asCoordinate(key='unit_unitwidth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelX
            self.asCoordinate(key='unit_unitdepth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelY
            self.asCoordinate(key='unit_unitheight', uuid=workpiece.parentUUID, **keywords),  # STIMoebelZ

            self.asString(data=entry.description, **keywords),
            self.asStringSingleLine(key='part_comment', **keywords),

            self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),

            self.asString(key='part_laminate1', **keywords),
            self.asString(key='part_laminate2', **keywords),
            self.asString(key='part_laminate3', **keywords),

            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),
        )
