# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import os

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase


class CorporaExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'Corpora'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = '\t'
        self.counter = 0

        self.EXTENSION = 'txt'
        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsandfittings')  # | partsonly | fittingsonly
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

    def getHeaderLine(self, **_keywords):
        return None

    def createNCProgramBasePath(self, **keywords):
        entry = keywords['entry']
        path = entry.ncProgramPathMain
        if not path:
            path = entry.ncProgramPathOpposite
        if not path:
            path = entry.ncProgramPathSizing
        if isinstance(path, list):  # Cumulate parts is not allowed for CorporaExport.
            path = path[0]
        return os.path.dirname(path) if path else ''

    def getGrain(self, grain):
        # As interiorcad
        if grain == self.GRAIN_LENGTHWISE:
            return 0
        elif grain == self.GRAIN_CROSSWISE:
            return 1
        elif grain == self.GRAIN_NONE:
            return 2
        else:
            assert False, "Unknown grain"

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']
        workpiece = keywords['workpiece']
        self.counter += 1

        return (
            self.asString(data=projectInfo['DocumentName'], **keywords),

            self.asString(data=projectInfo['ProjectID'], **keywords),
            self.asString(data=projectInfo['ProjectCommission'], **keywords),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),

            self.asString(data=positionInfo['CabinetName'], **keywords),
            self.asCoordinate(key='unit_unitwidth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelX
            self.asCoordinate(key='unit_unitdepth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelY
            self.asCoordinate(key='unit_unitheight', uuid=workpiece.parentUUID, **keywords),  # STIMoebelZ

            self.asString(data=positionInfo['PositionNumber'], **keywords),
            self.asString(data=positionInfo['PositionDescription'], **keywords),

            self.asInt(data=self.counter, **keywords),  # STITeileNummer
            self.asString(data=entry.name, **keywords),  # STIBaugruppenName
            self.asString(data=entry.name, **keywords),  # STIName

            # Unlike Corpora.ef we do not swap length and width according to grain
            self.asCoordinate(data=entry.boardLength, **keywords),
            self.asCoordinate(data=entry.boardWidth, **keywords),
            '',  # STILaengeBreite
            self.asString(data=entry.itemID, **keywords),
            self.asCoordinate(data=entry.boardThickness, **keywords),
            self.asInt(data=entry.count, **keywords),

            self.asString(data=entry.edges.front.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.front.thickness, **keywords),
            self.asString(data=entry.edges.back.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.back.thickness, **keywords),
            self.asString(data=entry.edges.left.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.left.thickness, **keywords),
            self.asString(data=entry.edges.right.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.right.thickness, **keywords),

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),

            self.asInt(data=self.getGrain(entry.grain), **keywords),

            self.asString(key='part_laminate1', **keywords),
            self.asString(key='part_laminate2', **keywords),
            self.asString(key='part_laminate3', **keywords),
            '',  # STILaminate4
            '',  # STILaminate5
            '',  # STILaminate6

            self.asStringSingleLine(key='part_comment', **keywords),
            self.asString(data=entry.name, **keywords),  # STIBaugruppe
            '',  # NodeID
            '',  # NodeText
            '',  # NutLaenge

            # 42-74: up to 8 polygon points
            '',  # NumSegmentPts3
            '',  # PolyPktX1
            '',  # PolyPktY1
            '',  # PolySide1
            '',
            '',  # PolyPktX2
            '',  # PolyPktY2
            '',  # PolySide2
            '',
            '',  # PolyPktX3
            '',  # PolyPktY3
            '',  # PolySide3
            '',
            '',  # PolyPktX4
            '',  # PolyPktY4
            '',  # PolySide4
            '',
            '',  # PolyPktX5
            '',  # PolyPktY5
            '',  # PolySide5
            '',
            '',  # PolyPktX6
            '',  # PolyPktY6
            '',  # PolySide6
            '',
            '',  # PolyPktX7
            '',  # PolyPktY7
            '',  # PolySide7
            '',
            '',  # PolyPktX8
            '',  # PolyPktY8
            '',  # PolySide8
            '',

            # 75-82: chamfer lengths
            '',  # FaseBL
            '',  # FaseHL
            '',  # FaseBR
            '',  # FaseHR
            '',  # FaseBO
            '',  # FaseHO
            '',  # FaseBU
            '',  # FaseHU

            # 83-86: chamfer lengths on frame
            '',  # RahmenAussenHoehe
            '',  # RahmenAussenBreite
            '',  # RahmenInnenHoehe
            '',  # RahmenInnenBreite

            # 87: dado distance
            '',  # NutAbstand

            # 88-91: dado/rabbet lengths
            '',  # NutFalzL
            '',  # NutFalzR
            '',  # NutFalzO
            '',  # NutFalzU

            self.createNCProgramBasePath(**keywords),  # base path of nc-programs of nc-export
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),
        )
