# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections
import itertools
import math
import os

from interiorcad.backend.cmoduleadapters.contourpath import RADIUS_COMPENSATION_NONE, getOffsetContourPath
from interiorcad.backend.cmoduleadapters.materials.boards import getBoardType
from interiorcad.backend.datatracker.datatrackeritemtypes import ITEMTYPE_HINGE, ITEMTYPE_SLIDER, ITEMTYPE_UNDERFLOOR_SLIDER, ITEMTYPE_HANDLE, ITEMTYPE_LEG, ITEMTYPE_WOODEN_PIN, ITEMTYPE_SCREW, ITEMTYPE_SPECIAL_CONNECTOR, ITEMTYPE_CONNECTOR_HOUSING, ITEMTYPE_CONNECTOR_BOLT, ITEMTYPE_CABINET, ITEMTYPE_WOODENDRAWER, ITEMTYPE_SYSTEMDRAWER, ITEMTYPE_PLINTH3D, PIO_CONTAINER_TYPES
from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.datatracker.polygonconverter import PolygonConverter
from interiorcad.backend.localization import getLocalizedString, CUTTINGLIST_RESOURCEGROUP
from interiorcad.backend.ncexport.linemachininganalyzer import LineMachiningAnalyzer
from interiorcad.backend.workpiece import DadoRabbet, Drilling, CenterPunch, Miter, Chamfer, NCMacro, Contour
from interiorcad.backend.workpiece.complexgriddrilling import ComplexGridDrilling
from interiorcad.backend.workpiecefilters.emptyworkpiecefilter import EmptyWorkpieceFilter
from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.edges import EdgeData
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline

METER = 1000.0
SQUARE_METER = METER * METER
CUBIC_METER = SQUARE_METER * METER


class CorporaIntegrationExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'Corpora'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = '\t'
        self.counter = 0
        self.entries = ()

        self.EXTENSION = 'txt'
        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsandfittings')  # | partsonly | fittingsonly
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

        self.ADDITIONAL_TIME_INFORMATION = self.getCuttinglistConfiguration().getAsBool('additional-time-information', True)

        self.outputFittingUUIDs = None

    @property
    def columnCount(self):
        return 68

    def createNCProgramBasePath(self, **keywords):
        entry = keywords['entry']
        path = entry.ncProgramPathMain
        if not path:
            path = entry.ncProgramPathOpposite
        if not path:
            path = entry.ncProgramPathSizing
        if isinstance(path, list):  # Cumulate parts is not allowed for CorporaExport.
            path = path[0]
        return os.path.dirname(path) if path else ''

    def getGrain(self, grain):
        # As interiorcad
        if grain == self.GRAIN_LENGTHWISE:
            return 0
        elif grain == self.GRAIN_CROSSWISE:
            return 1
        elif grain == self.GRAIN_NONE:
            return 2
        else:
            assert False, "Unknown grain"

    def getJointAsNumber(self, jointId):
        # Code taken from CADmatic export, but with different return values
        if jointId == EdgeData.JOINT_MITER:
            return '0'
        elif jointId == EdgeData.JOINT_CONTINUOUS:
            return '2'
        elif jointId == EdgeData.JOINT_DISCONTINUOUS:
            return '1'
        elif not jointId or jointId == EdgeData.JOINT_NOEDGE:
            return '1'  # No value shown for no edge
        else:
            return '1'

    def getHeaderLine(self, **_keywords):
        title = [
            self.getLocalizedHeader(35, 'Document Name'),
            self.getLocalizedHeader(17, 'Project ID'),
            self.getLocalizedHeader(20, 'Cabinet Name'),
            self.getLocalizedHeader(114, "Cabinet width"),
            self.getLocalizedHeader(115, "Cabinet depth"),
            self.getLocalizedHeader(116, "Cabinet height"),
            self.getLocalizedHeader(18, 'Position Number'),
            self.getLocalizedHeader(19, 'Position Description'),
            self.getLocalizedHeader(117, 'Line (line number)'),
            self.getLocalizedHeader(7, 'Unique Name'),
            self.getLocalizedHeader(3, 'Length'),
            self.getLocalizedHeader(4, 'Width'),
            self.getLocalizedHeader(134, "Rectangular Area"),
            self.getLocalizedHeader(1, 'Supplier ID'),
            self.getLocalizedHeader(5, 'Thickness'),
            self.getLocalizedHeader(6, 'Count'),
            self.getLocalizedHeader(11, 'Edge ID Front 1'),
            self.getLocalizedHeader(21, 'Edge Thickness Front 1'),
            self.getLocalizedHeader(12, 'Edge ID Back 2'),
            self.getLocalizedHeader(22, 'Edge Thickness Back 2'),
            self.getLocalizedHeader(13, 'Edge ID Left 3'),
            self.getLocalizedHeader(23, 'Edge Thickness Left 3'),
            self.getLocalizedHeader(14, 'Edge ID Right 4'),
            self.getLocalizedHeader(24, 'Edge Thickness Right 4'),
            self.getLocalizedHeader(25, 'Covering 1'),
            self.getLocalizedHeader(26, 'Covering 2'),
            self.getLocalizedHeader(41, 'Grain'),
            self.getLocalizedHeader(27, 'Laminate 1'),
            self.getLocalizedHeader(28, 'Laminate 2'),
            self.getLocalizedHeader(29, 'Laminate 3'),
            self.getLocalizedHeader(30, 'Comment'),
            self.getLocalizedHeader(135, "Number of Polygon Points"),
            self.getLocalizedHeader(136, "1st polygon point X"),
            self.getLocalizedHeader(137, "1st polygon point Y"),
            self.getLocalizedHeader(138, "2nd polygon point X"),
            self.getLocalizedHeader(139, "2nd polygon point Y"),
            self.getLocalizedHeader(140, "3rd polygon point X"),
            self.getLocalizedHeader(141, "3rd polygon point Y"),
            self.getLocalizedHeader(142, "4th polygon point X"),
            self.getLocalizedHeader(143, "4th polygon point Y"),
            self.getLocalizedHeader(144, "5th polygon point X"),
            self.getLocalizedHeader(145, "5th polygon point Y"),
            self.getLocalizedHeader(146, "6th polygon point X"),
            self.getLocalizedHeader(147, "6th polygon point Y"),
            self.getLocalizedHeader(148, "7th polygon point X"),
            self.getLocalizedHeader(149, "7th polygon point Y"),
            self.getLocalizedHeader(150, "8th polygon point X"),
            self.getLocalizedHeader(151, "8th polygon point Y"),
            self.getLocalizedHeader(152, "Edge Jointing Front"),
            self.getLocalizedHeader(153, "Edge Jointing Back"),
            self.getLocalizedHeader(154, "Edge Jointing Left"),
            self.getLocalizedHeader(155, "Edge Jointing Right"),
            self.getLocalizedHeader(119, 'NC program path (base path)'),
            self.getLocalizedHeader(8, 'NC Program Name Sizing'),
            self.getLocalizedHeader(9, 'NC Program Name Main'),
            self.getLocalizedHeader(10, 'NC Program Name Opposite'),
            self.getLocalizedHeader(160, 'Length (workshop)'),
            self.getLocalizedHeader(161, 'Width (workshop)'),
            self.getLocalizedHeader(162, 'Has sizing'),
            self.getLocalizedHeader(163, 'Create separate sizing programs'),
            self.getLocalizedHeader(164, 'No sizing for rectangular parts'),
            self.getLocalizedHeader(165, 'Jointing during edge banding'),
            self.getLocalizedHeader(166, 'Jointing during edge banding (mininum)'),
            self.getLocalizedHeader(167, 'Jointing during edge banding (maximum)'),
            self.getLocalizedHeader(168, 'Cut size addition parts'),
            self.getLocalizedHeader(169, 'Additional covering addition'),
            self.getLocalizedHeader(170, 'Cut according to NC Output'),
            self.getLocalizedHeader(171, 'Cut (variable)'),
        ]

        assert len(title) == self.columnCount, 'Internal error: the number of column titles differ from the corresponding column number constant'
        return title

    def getElements(self, **keywords):
        def getShapeVertexPosition(index):
            if index < shapePolyline.numVertices():
                return shapePolyline.getVertex(index).position
            else:
                return Point(0.0, 0.0)

        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']
        workpiece = keywords['workpiece']
        self.counter += 1

        shapePolyline = workpiece.finishedSizeShape.polyline
        if entry.isWorkpiece() and workpiece.finishedSizeShape.isValid() and not workpiece.isRectangular() and workpiece.finishedSizeShape.polyline.numVertices() <= 8 and workpiece.isPolygonal():
            shapePolyline = workpiece.finishedSizeShape.polyline
        else:
            shapePolyline = Polyline()

        elements = [
            self.asString(data=projectInfo['DocumentName'], **keywords),

            self.asString(data=projectInfo['ProjectID'], **keywords),

            self.asString(key='unit_unitname', uuid=keywords['parentUUID'], **keywords),  # DO NOT use projectInfo['CabinetName']. as it can be out of order for fitings
            self.asCoordinate(key='unit_unitwidth', uuid=keywords['parentUUID'], **keywords),  # STIMoebelX
            self.asCoordinate(key='unit_unitdepth', uuid=keywords['parentUUID'], **keywords),  # STIMoebelY
            self.asCoordinate(key='unit_unitheight', uuid=keywords['parentUUID'], **keywords),  # STIMoebelZ

            self.asString(data=positionInfo['PositionNumber'], **keywords),
            self.asString(data=positionInfo['PositionDescription'], **keywords),

            self.asInt(data=self.counter, **keywords),  # STITeileNummer
            self.asString(data=entry.name, **keywords),  # STIName

            # Unlike Corpora.ef we do not swap length and width according to grain
            self.asCoordinate(data=entry.finishedSizePartDimensions.size.x, **keywords),
            self.asCoordinate(data=entry.finishedSizePartDimensions.size.y, **keywords),
            self.asCoordinate(data=entry.finishedSizePartDimensions.size.x * entry.finishedSizePartDimensions.size.y / SQUARE_METER, **keywords),  # STILaengeBreite
            self.asString(data=entry.itemID, **keywords),
            self.asCoordinate(data=entry.boardThickness, **keywords),
            self.asInt(data=entry.count, **keywords),

            self.asString(data=entry.edges.front.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.front.thickness, **keywords),
            self.asString(data=entry.edges.back.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.back.thickness, **keywords),
            self.asString(data=entry.edges.left.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.left.thickness, **keywords),
            self.asString(data=entry.edges.right.supplierID, **keywords),
            self.asCoordinate(data=entry.edges.right.thickness, **keywords),

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),

            self.asInt(data=self.getGrain(entry.grain), **keywords),

            self.asString(key='part_laminate1', **keywords),
            self.asString(key='part_laminate2', **keywords),
            self.asString(key='part_laminate3', **keywords),

            self.asStringSingleLine(key='part_comment', **keywords),

            self.asInt(data=shapePolyline.numVertices()),  # NumSegmentPts3
            self.asCoordinate(data=getShapeVertexPosition(0).x),  # PolyPktX1
            self.asCoordinate(data=getShapeVertexPosition(0).y),  # PolyPktY1
            self.asCoordinate(data=getShapeVertexPosition(1).x),  # PolyPktX2
            self.asCoordinate(data=getShapeVertexPosition(1).y),  # PolyPktY2
            self.asCoordinate(data=getShapeVertexPosition(2).x),  # PolyPktX3
            self.asCoordinate(data=getShapeVertexPosition(2).y),  # PolyPktY3
            self.asCoordinate(data=getShapeVertexPosition(3).x),  # PolyPktX4
            self.asCoordinate(data=getShapeVertexPosition(3).y),  # PolyPktY4
            self.asCoordinate(data=getShapeVertexPosition(4).x),  # PolyPktX5
            self.asCoordinate(data=getShapeVertexPosition(4).y),  # PolyPktY5
            self.asCoordinate(data=getShapeVertexPosition(5).x),  # PolyPktX6
            self.asCoordinate(data=getShapeVertexPosition(5).y),  # PolyPktY6
            self.asCoordinate(data=getShapeVertexPosition(6).x),  # PolyPktX7
            self.asCoordinate(data=getShapeVertexPosition(6).y),  # PolyPktY7
            self.asCoordinate(data=getShapeVertexPosition(7).x),  # PolyPktX8
            self.asCoordinate(data=getShapeVertexPosition(7).y),  # PolyPktY8

            self.getJointAsNumber(self.getDataAsString(data=entry.edges.front.jointStart, **keywords)),
            self.getJointAsNumber(self.getDataAsString(data=entry.edges.back.jointStart, **keywords)),
            self.getJointAsNumber(self.getDataAsString(data=entry.edges.left.jointStart, **keywords)),
            self.getJointAsNumber(self.getDataAsString(data=entry.edges.right.jointStart, **keywords)),

            self.createNCProgramBasePath(**keywords),  # base path of nc-programs of nc-export
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),
            self.asCoordinate(data=entry.boardLength, **keywords),
            self.asCoordinate(data=entry.boardWidth, **keywords),

            self.asInt(data=EmptyWorkpieceFilter.hasSizing(self, workpiece)),
            self.asInt(data=self.shouldCreateSeparateSizing()),
            self.asInt(data=self.shouldCreateSizingForRectangularPart()),
            self.asInt(data=self.getJointingInfo().jointing),
            self.asCoordinate(data=self.getJointingInfo().minJoint, **keywords),
            self.asCoordinate(data=self.getJointingInfo().maxJoint, **keywords),
            self.asCoordinate(data=self.getsizingAddition(), **keywords),
            self.asCoordinate(data=self.getAdditionalCoveringAddition(), **keywords),
            self.asInt(data=self.sizingIsNcCompatible()),
            self.asString(data=self.getSizingText(), **keywords),
        ]
        assert len(elements) == self.columnCount, 'Internal error: the number of columns differ from the corresponding column number constant'
        return elements

    def generateSortedEntries(self, **keywords):
        self.entries = tuple(super().generateSortedEntries(**keywords))
        return self.entries

    def generateFittingsUUIDS(self, **keywords):
        # TODO IT-303: we do not respect counts yet -- NG
        for customPartUUID in keywords['customPartUUIDs']:
            yield from ExportDataCollector.getFittingUUIDsFromCustomPart(customPartUUID)

    def countFilteredItems(self, filterLamdba, items):
        # We do not need to be more clever. According to https://stackoverflow.com/questions/3345785/getting-number-of-elements-in-an-iterator-in-python len(tuple( is the fastet anyway
        return len(tuple(filter(filterLamdba, items)))

    def getDrillingCount(self, **keywords):
        return self.countFilteredItems(lambda machining: isinstance(machining, Drilling) or isinstance(machining, CenterPunch), keywords['machinings'])

    def getGridDrillingCount(self, **keywords):
        # Code similar to interiorcad.backend.postprocessors.surfacepostprocessor.verticalgriddrillingsreplacer.VerticalGridDrillingsReplacer.processGridDrilling
        def processGridDrilling(machining):
            def key(point):
                return math.fsum((machining.gridPointsCoordinateSystem.xaxis.y * point.x, machining.gridPointsCoordinateSystem.yaxis.y * point.y))

            return len(tuple(itertools.groupby(sorted(machining.gridPoints, key=key), key=key)))

        complexGridDrillings = filter(lambda machining: isinstance(machining, ComplexGridDrilling), keywords['machinings'])
        return sum((processGridDrilling(machining) for machining in complexGridDrillings))

    def getLineMachiningAccumulatedLength(self, **keywords):
        # math.fsum is more precise than sum and gives 0.0 on an empty iterable :-)
        return math.fsum(self.getDataAsFloat(data=lineMachining.getRealLength()) for lineMachining in keywords['lineMachinings'])

    def __getSimpleFittingCount(self, fittingsUUIDs, itemType):
        return self.countFilteredItems(lambda uuid: ExportDataCollector.getPartType(uuid) == itemType, fittingsUUIDs)

    def __getMultipleFittingCount(self, fittingsUUIDs, itemTypes):
        return self.countFilteredItems(lambda uuid: ExportDataCollector.getPartType(uuid) in itemTypes, fittingsUUIDs)

    def getSliderCount(self, **keywords):
        return 2 * self.__getMultipleFittingCount(keywords['fittingsUUIDs'], (ITEMTYPE_SLIDER, ITEMTYPE_UNDERFLOOR_SLIDER))

    def getHingeCount(self, **keywords):
        return self.__getSimpleFittingCount(keywords['fittingsUUIDs'], ITEMTYPE_HINGE)

    def getHandleCount(self, **keywords):
        return self.__getSimpleFittingCount(keywords['fittingsUUIDs'], ITEMTYPE_HANDLE)

    def getLegCount(self, **keywords):
        return self.__getSimpleFittingCount(keywords['fittingsUUIDs'], ITEMTYPE_LEG)

    def getDowelCount(self, **keywords):
        return self.__getSimpleFittingCount(keywords['fittingsUUIDs'], ITEMTYPE_WOODEN_PIN)

    def getScrewCount(self, **keywords):
        return self.__getSimpleFittingCount(keywords['fittingsUUIDs'], ITEMTYPE_SCREW)

    def getSpecialConnectorCount(self, **keywords):
        # Even if in reality special connectors might consist of two items, we count them only as one.
        return self.__getSimpleFittingCount(keywords['fittingsUUIDs'], ITEMTYPE_SPECIAL_CONNECTOR)

    def getMultipleConnectorCount(self, **keywords):
        # Unlike for special connectors we do count bolt and connector separately
        return self.__getMultipleFittingCount(keywords['fittingsUUIDs'], (ITEMTYPE_CONNECTOR_HOUSING, ITEMTYPE_CONNECTOR_BOLT))

    def getMachinings(self, **keywords):
        for entry in keywords['currentEntries']:
            yield from entry.machinings

    def getDadoAndRabbets(self, **keywords):
        dados = []
        rabbets = []

        lineMachiningAnalyzer = LineMachiningAnalyzer()

        for entry in keywords['currentEntries']:
            dadoRabbets = filter(lambda machining: isinstance(machining, DadoRabbet), entry.machinings)
            for dadoRabbet in dadoRabbets:
                if lineMachiningAnalyzer.isDado(dadoRabbet, entry.finishedSizeShape):
                    dados.append(dadoRabbet)
                else:
                    rabbets.append(dadoRabbet)

        return dados, rabbets

    def getMiterLength(self, **keywords):
        return self.getLineMachiningAccumulatedLength(lineMachinings=filter(lambda machining: isinstance(machining, Miter), keywords['machinings']))

    def getChamferLength(self, **keywords):
        return self.getLineMachiningAccumulatedLength(lineMachinings=filter(lambda machining: isinstance(machining, Chamfer), keywords['machinings']))

    def getNCMacroCount(self, **keywords):
        return len(tuple(filter(lambda machining: isinstance(machining, NCMacro), keywords['machinings'])))

    def getEdgeLengths(self, **keywords):
        polygonalLengths = []
        polylineLengths = []

        for entry in keywords['currentEntries']:
            forcePolygonal = entry.isWorkpiece() and entry.isPolygonal()  # Older datatracker have no polygonal info and therefore assume false for isPolygonal
            for edge in entry.edges.polygonal:
                if edge.supplierID and not isNearlyEqual(edge.length, 0.0):
                    if forcePolygonal or edge.isPolygonal:
                        polygonalLengths.append(edge.length)
                    else:
                        polylineLengths.append(edge.length)

        return math.fsum(polygonalLengths), math.fsum(polylineLengths)

    def __isSolidMaterial(self, boardType):
        # See dev\SharedWithClassic\MaterialType.cpp. (But we do not support "glass")
        return boardType in ('solid', '3l', 'gw')

    def getSolidCurrentEntries(self, **keywords):
        return filter(lambda entry: self.__isSolidMaterial(getBoardType(entry.itemID)), keywords['currentEntries'])

    def getNonSolidCurrentEntries(self, **keywords):
        return itertools.filterfalse(lambda entry: self.__isSolidMaterial(getBoardType(entry.itemID)), keywords['currentEntries'])

    def __getBoardLength(self, filteredEntries):
        def longAndSmall(entry):
            yield entry.boardLength
            yield entry.boardLength
            yield entry.boardWidth
            yield entry.boardWidth

        return math.fsum(itertools.chain.from_iterable(longAndSmall(entry) for entry in filteredEntries))

    def getNonSolidBoardLength(self, **keywords):
        return self.__getBoardLength(self.getNonSolidCurrentEntries(**keywords))

    def getSolidBoardLength(self, **keywords):
        return self.__getBoardLength(self.getSolidCurrentEntries(**keywords))

    def getNonSolidBoardArea(self, **keywords):
        def area(entry):
            return entry.boardLength * entry.boardWidth

        return math.fsum(area(entry) for entry in self.getNonSolidCurrentEntries(**keywords))

    def getSolidBoardVolume(self, **keywords):
        def volume(entry):
            return entry.boardLength * entry.boardWidth * entry.boardThickness

        return math.fsum(volume(entry) for entry in self.getSolidCurrentEntries(**keywords))

    def getSolidCount(self, **keywords):
        return len(tuple(self.getSolidCurrentEntries(**keywords)))

    def getLaminateArea(self, **keywords):
        def generateLaminates(uuid):
            yield self.getDataAsString(key='part_laminate1', uuid=uuid, **keywords)
            yield self.getDataAsString(key='part_laminate2', uuid=uuid, **keywords)
            yield self.getDataAsString(key='part_laminate3', uuid=uuid, **keywords)

        def getLamitateArea(entry):
            if entry.isWorkpiece():
                bounds = entry.workpiece.finishedSizeBounds
                area = bounds.getLength() * bounds.getWidth()
                for laminate in generateLaminates(entry.uuid):
                    if laminate.strip():
                        yield area

        return 2.0 * math.fsum(itertools.chain.from_iterable(getLamitateArea(entry) for entry in keywords['currentEntries']))

    def getCoveringArea(self, **keywords):
        def getCoveringArea(entry):
            return entry.boardLength * entry.boardWidth

        return math.fsum(getCoveringArea(entry) for entry in keywords['currentCoveringEntries'])

    def getSizingPathLength(self, **keywords):
        def getSizingLength(entry):
            if entry.isWorkpiece() and entry.hasFinishedSizeShape() and EmptyWorkpieceFilter.hasSizing(self, entry.workpiece):
                return entry.finishedSizeShape.polyline.getLength()
            else:
                return 0.0

        return math.fsum(getSizingLength(entry) for entry in keywords['currentEntries'])

    def getContourPathLength(self, **keywords):
        def generageContourLengths(contour):
            # Some old MockDocument have not MultiUsageRoutings in the corresponding Datatracker file yet
            if not contour.hasMachineOptions:
                yield contour.path.getLength()
            else:
                for multiUsageRouting in contour.multiUsageRoutings:
                    if isNearlyEqual(multiUsageRouting.distance, 0.0) or contour.radiusCompensation == RADIUS_COMPENSATION_NONE:
                        yield contour.path.getLength()
                    else:
                        offsetPath = getOffsetContourPath(contour.path.vertices, contour.radiusCompensation, multiUsageRouting.distance)
                        newPath = PolygonConverter().convertPolyline(offsetPath)
                        yield newPath.getLength()

        return math.fsum(itertools.chain.from_iterable(generageContourLengths(contour) for contour in filter(lambda machining: isinstance(machining, Contour), keywords['machinings'])))

    TimeDataEntry = collections.namedtuple("TimeDataEntry", ("key", "data", "description"))

    def generateAdditionalTimeInformations(self, **keywords):
        parentUUID = keywords['parentUUID']
        customPartUUIDs = keywords['customPartUUIDs']

        currentEntries = tuple(filter(lambda entry: entry.isWorkpiece() and entry.uuid in customPartUUIDs, self.entries))
        machinings = tuple(self.getMachinings(currentEntries=currentEntries, **keywords))
        dados, rabbets = self.getDadoAndRabbets(currentEntries=currentEntries, **keywords)
        edgePolygonalLength, edgePolylineLength = self.getEdgeLengths(currentEntries=currentEntries, **keywords)

        currentCoveringEntries = self.generateUnsortedCoveringEntriesFromUUIDsAndCounts(**keywords)

        def makeTDE(key, data, description):
            assert key.startswith('$$dv'), "Internal error: the key should start with $$dv"
            return self.TimeDataEntry(key, data, description)

        # TODO IT-301/STORY IC-26641: I18N/L10N -- NG
        # Entries proposed by PinnCalc
        yield makeTDE('$$dv platte lfm gerade', self.asCoordinate(data=self.getNonSolidBoardLength(currentEntries=currentEntries, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für den Zuschnitt von Platten bei einem Kantenwinkel von 90°'))
        yield makeTDE('$$dv platte m2 eben', self.asArea(data=self.getNonSolidBoardArea(currentEntries=currentEntries, **keywords) / SQUARE_METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro m² bei einer ebenen Platte z.B. Spanplatte oder Tischlerplatte'))
        yield makeTDE('$$dv massiv lfm', self.asCoordinate(data=self.getSolidBoardLength(currentEntries=currentEntries, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm zum Bearbeiten von Massivholz'))
        # IC-27187: export cubic-meters as angle. Otherwise, it would rounded down to 0
        yield makeTDE('$$dv massiv m3', self.asVolume(data=self.getSolidBoardVolume(currentEntries=currentEntries, **keywords) / CUBIC_METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro m³ zum Bearbeiten von Massivholz'))
        yield makeTDE('$$dv massiv stk', self.asInt(data=self.getSolidCount(currentEntries=currentEntries, **keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück zum Bearbeiten von Massivholz'))
        yield makeTDE('$$dv belag1 m2', self.asArea(data=self.getCoveringArea(currentCoveringEntries=currentCoveringEntries, **keywords) / SQUARE_METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit für das Verarbeiten von Furnieren und Belägen pro m²'))
        yield makeTDE('$$dv oberf m2', self.asArea(data=self.getLaminateArea(currentEntries=currentEntries, **keywords) / SQUARE_METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit für das Lackieren pro m²'))
        yield makeTDE('$$dv kante1 lfm gebogen', self.asCoordinate(data=edgePolylineLength / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Verarbeiten von gebogenen Kanten'))
        yield makeTDE('$$dv kante1 lfm gerade', self.asCoordinate(data=edgePolygonalLength / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Verarbeiten von geraden Kanten'))
        yield makeTDE('$$dv sk1 stk', self.asInt(key='unit_drawercount', uuid=parentUUID, **keywords), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Fertigen eines Schubkastens'))
        yield makeTDE('$$dv auszug stk', self.asInt(data=self.getSliderCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren eines Auszuges'))
        yield makeTDE('$$dv band stk', self.asInt(data=self.getHingeCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren eines Bandes'))
        yield makeTDE('$$dv griff stk', self.asInt(data=self.getHandleCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren eines Griffs'))
        yield makeTDE('$$dv sockelfuss stk', self.asInt(data=self.getLegCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren eines Sockelfusses'))
        yield makeTDE('$$dv falz lfm gerade', self.asCoordinate(data=self.getLineMachiningAccumulatedLength(lineMachinings=rabbets, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Herstellen von geraden Fälzen'))
        yield makeTDE('$$dv nut lfm gerade', self.asCoordinate(data=self.getLineMachiningAccumulatedLength(lineMachinings=dados, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Herstellen von geraden Nuten'))
        yield makeTDE('$$dv bohrung stk', self.asInt(data=self.getDrillingCount(machinings=machinings, **keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit für eine einzelne Bohrung'))
        yield makeTDE('$$dv bohrunglr stk', self.asInt(data=self.getGridDrillingCount(machinings=machinings, **keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit für die Bohrung einer Lochreihe'))

        # Entries defined for interiorcad
        yield makeTDE('$$dv duebel stk', self.asInt(data=self.getDowelCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren eines Dübels'))
        yield makeTDE('$$dv fase lfm gerade', self.asCoordinate(data=self.getChamferLength(machinings=machinings, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Herstellen von geraden Fasen'))
        yield makeTDE('$$dv formatierung lfm', self.asCoordinate(data=self.getSizingPathLength(currentEntries=currentEntries, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Herstellen von gefrästen Formatierungen'))
        yield makeTDE('$$dv gehrung lfm gerade', self.asCoordinate(data=self.getMiterLength(machinings=machinings, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Herstellen von geraden Gehrungen'))
        yield makeTDE('$$dv konturzug lfm', self.asCoordinate(data=self.getContourPathLength(machinings=machinings, **keywords) / METER), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Herstellen von Konturzügen'))
        yield makeTDE('$$dv multiverbinder stk', self.asInt(data=self.getMultipleConnectorCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren eines Multiverbinders'))
        yield makeTDE('$$dv ncmakro stk', self.asInt(data=self.getNCMacroCount(machinings=machinings, **keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro lfm für das Herstellen von NC-Makros'))
        yield makeTDE('$$dv schraube stk', self.asInt(data=self.getScrewCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren einer Schraube'))
        yield makeTDE('$$dv systemverbinder stk', self.asInt(data=self.getSpecialConnectorCount(**keywords)), self.getLocalizedTimeDataDescription(None, 'interiorcad-Zeit pro Stück für das Montieren eines Systemverbinders'))

    def generateRowsWithTheRightNumberOfColumns(self, **keywords):
        projectInfo = keywords['projectInfo']
        parentName = keywords['parentName']
        parentUUID = keywords['parentUUID']
        positionInfo = self.getPositionInfo(uuid=parentUUID)
        unitWidth = self.asCoordinate(key='unit_unitwidth', uuid=parentUUID, **keywords)
        unitDepth = self.asCoordinate(key='unit_unitdepth', uuid=parentUUID, **keywords)
        unitHeight = self.asCoordinate(key='unit_unitheight', uuid=parentUUID, **keywords)

        for timeDataEntry in keywords['timeData']:
            self.counter += 1

            elements = (
                self.asString(data=projectInfo['DocumentName'], **keywords),
                self.asString(data=projectInfo['ProjectID'], **keywords),
                parentName,
                unitWidth,
                unitDepth,
                unitHeight,
                self.asString(data=positionInfo['PositionNumber'], **keywords),
                self.asString(data=positionInfo['PositionDescription'], **keywords),
                self.asInt(data=self.counter, **keywords),
                timeDataEntry.description,
                0,
                0,
                0,
                timeDataEntry.key,
                '',
                timeDataEntry.data,
            )

            assert len(elements) <= self.columnCount, "Internal error: too many columns"
            fullRow = elements + (('',) * (self.columnCount - len(elements)))
            assert len(fullRow) == self.columnCount, "Internal error: wrong number of columns"
            yield fullRow

    def writeAdditionalTimeInformationBatch(self, **keywords):
        parentName = ExportDataCollector.getUniqueName(keywords['parentUUID'])
        rows = self.generateRowsWithTheRightNumberOfColumns(parentName=parentName, timeData=self.generateAdditionalTimeInformations(**keywords), **keywords)
        self.writeRows(rows=rows, **keywords)

    def writeEntriesOfPIOContainer(self, **keywords):
        allExportedStuff = self.uuidsAndCounts
        allExportedFittings = self.fittingsUUIDsAndCounts

        self.uuidsAndCounts = keywords['uuidsAndCounts']

        fittingsUUIDs = set(self.generateFittingsUUIDS(**keywords))
        fittingsUUIDs -= self.outputFittingUUIDs
        self.fittingsUUIDsAndCounts = tuple((uuid, count) for uuid, count in allExportedFittings if uuid in fittingsUUIDs)

        super().writeEntriesAccordingScope(**keywords)
        if self.ADDITIONAL_TIME_INFORMATION:
            self.writeAdditionalTimeInformationBatch(fittingsUUIDs=fittingsUUIDs, **keywords)

        self.outputFittingUUIDs |= self.outputFittingUUIDs

        self.uuidsAndCounts = allExportedStuff
        self.fittingsUUIDsAndCounts = allExportedFittings

    def __filterDrawers(self, cabinetUUIDS, drawersAndCounts):
        for drawerUUID, drawerCount in drawersAndCounts:
            drawerParentUUID = ExportDataCollector.getUserParentPIOContainerUUID(drawerUUID)
            if drawerParentUUID not in cabinetUUIDS:
                yield drawerUUID, drawerCount

    def __sortPIOContainerByType(self, pioContainersAndCounts):
        def key(pioContainerAndCount):
            return ExportDataCollector.getPartType(pioContainerAndCount[0])

        groupedPioContainersAndCounts = dict((partType, tuple(containers)) for partType, containers in itertools.groupby(sorted(pioContainersAndCounts, key=key), key=key))

        cabinetUUIDs = ()
        if ITEMTYPE_CABINET in groupedPioContainersAndCounts:
            cabinets = groupedPioContainersAndCounts[ITEMTYPE_CABINET]
            cabinetUUIDs = tuple(entry[0] for entry in cabinets)
            yield from cabinets
        if ITEMTYPE_WOODENDRAWER in groupedPioContainersAndCounts:
            yield from self.__filterDrawers(cabinetUUIDs, groupedPioContainersAndCounts[ITEMTYPE_WOODENDRAWER])
        if ITEMTYPE_SYSTEMDRAWER in groupedPioContainersAndCounts:
            yield from self.__filterDrawers(cabinetUUIDs, groupedPioContainersAndCounts[ITEMTYPE_SYSTEMDRAWER])
        if ITEMTYPE_PLINTH3D in groupedPioContainersAndCounts:
            yield from groupedPioContainersAndCounts[ITEMTYPE_PLINTH3D]

        for partType, containers in groupedPioContainersAndCounts.items():
            if partType not in PIO_CONTAINER_TYPES:
                yield from containers

    def writeEntriesAccordingScope(self, **keywords):
        # We call super().writeEntriesAccordingScope only in one of the called functions

        self.outputFittingUUIDs = set()

        freeCustomParts, pioContainers, _childCustomParts = ExportDataCollector.getSplitCustomPartAndPIOContainerUUIDsAndCounts()

        groupedPioContainers = dict(self.__yieldElementsOutOfGroupBy(self.__getGroupsSortedByPositions(pioContainers)))
        groupedFreeCustomParts = dict(self.__yieldElementsOutOfGroupBy(self.__getGroupsSortedByPositions(freeCustomParts)))

        for position in sorted(set(groupedPioContainers.keys()) | set(groupedFreeCustomParts.keys())):
            self.writeEntryOfPositionGroup(groupedPioContainers.get(position, ()), groupedFreeCustomParts.get(position, ()), **keywords)

        self.outputFittingUUIDs = None

    def __yieldElementsOutOfGroupBy(self, groupedBy):
        for key, group in groupedBy:
            yield key, tuple(group)

    def __getGroupsSortedByPositions(self, uuids):
        def getPosition(uuidAndCount):
            uuid, _count = uuidAndCount
            return ExportDataCollector.getProjectPosition(uuid)

        sortedUUIDs = sorted(uuids, key=getPosition)
        return itertools.groupby(sortedUUIDs, key=getPosition)

    # TODO ISP-9: all parameters in **keywords ?
    def writeEntryOfPositionGroup(self, pioContainers, freeCustomParts, **keywords):
        for pioContainerUUID, pioContainerCount in self.__sortPIOContainerByType(pioContainers):
            customPartUUIDs = tuple(ExportDataCollector.getSubCustomPartsFromPIOContainer(pioContainerUUID))  # Be careful. the list might be empty!
            uuidsAndCounts = tuple((uuid, pioContainerCount) for uuid in customPartUUIDs)
            self.writeEntriesOfPIOContainer(
                parentUUID=pioContainerUUID,
                customPartUUIDs=customPartUUIDs,
                uuidsAndCounts=uuidsAndCounts,
                **keywords)

        for freeCustomPartUUID, freeCustomPartCount in freeCustomParts:
            self.writeEntriesOfPIOContainer(
                parentUUID=freeCustomPartUUID,
                customPartUUIDs=(freeCustomPartUUID,),
                uuidsAndCounts=((freeCustomPartUUID, freeCustomPartCount),),
                **keywords)

    def getLocalizedHeader(self, index, default):
        return getLocalizedString(CUTTINGLIST_RESOURCEGROUP, index, default)

    def getLocalizedTimeDataDescription(self, index, default):
        # TODO IT-301/STORY IC-26641: I18N/L10N -- NG
        return getLocalizedString(None, index, default)
