# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import csv
import sys

from interiorcad.backend.localization import getLocalizedString, CUTTINGLISTEXPORT_MESSAGES
from interiorcad.backend.sorter.datalistsorter import DataListSorter
from interiorcad.cuttinglistexport.base import CuttinglistExportBase


class CuttinglistExportCSVBase(CuttinglistExportBase):
    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdialect = csv.excel
        self.csvdelimiter = self.getFieldSeparator()
        self.csvquotechar = '"'
        self.csvquoting = csv.QUOTE_MINIMAL

    def doExport(self, **keywords):
        self.csvWriter = self.createCSVWriter(**keywords)
        self.writeListHeader(**keywords)
        if hasattr(self, 'SCOPE_OF_EXPORT'):
            self.writeEntriesAccordingScope(**keywords)
        else:
            self.writeListBody(**keywords)

    def createCSVWriter(self, **keywords):
        csvWriter = csv.writer(
            keywords['outputFile'],
            delimiter=self.csvdelimiter,
            lineterminator=self.OUTPUT_END_OF_LINE,
            quotechar=self.csvquotechar,
            quoting=self.csvquoting,
            dialect=self.csvdialect
        )
        return csvWriter

    def writeRows(self, **keywords):
        if keywords['rows']:
            for row in keywords['rows']:
                if row is not None:
                    rowAsTuple = tuple(row)
                    # HACK: We need the row again as a generator, to know how much was processed by csv.writerow ("row" itself cannot be used directly as it would be partially emptied by csv.writerow)
                    regeneratedRow = (element for element in rowAsTuple)
                    try:
                        self.csvWriter.writerow(regeneratedRow)
                    except csv.Error as e:
                        localizedMessage = getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 1, 'The following row contains illegal csv-characters or control characters. Possible conflict characters are {csvdelimiter} and {csvquotechar}. Search and replace them!').format(csvdelimiter=self.csvdelimiter, csvquotechar=self.csvquotechar)
                        leftOver = tuple(regeneratedRow)
                        message = '\n'.join((
                            '',
                            'ERROR: csv.Error: \'' + str(e) + '\'',
                            localizedMessage,
                            ("Field: " + rowAsTuple[-len(leftOver)-1]) if len(leftOver) < len(rowAsTuple) else "No field processed",
                            "Row: " + str(rowAsTuple),
                            getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 7, "The error may enventually be corrected by changing 'CSV quote character for strings' to another option than 'None'") if self.csvquotechar == None else '',
                            '',
                        ))
                        print(message, file=sys.stderr)
                        raise

    # compatibility function for old code; do NOT use for new code
    def writeEntries(self, **keywords):
        self.writeRows(rows=keywords['entry'], **keywords)

    def writeListHeader(self, **keywords):
        header = self.getHeaderLine(**keywords)
        rows = (header,) if header else None
        self.writeRows(rows=rows, **keywords)

    # compatibility function for old code; do NOT use for new code
    def writeListBody(self, **keywords):
        if 'rows' not in keywords:
            keywords['rows'] = self.generateWorkpieceEntries(**keywords)
        self.writeRows(**keywords)

    # compatibility function for old code; do NOT use for new code
    def writeListFittings(self, **keywords):
        if 'rows' not in keywords:
            keywords['rows'] = self.generateFittingEntries(**keywords)
        self.writeRows(**keywords)

    def writeEntriesAccordingScope(self, **keywords):
        keywords['entries'] = self.generateSortedEntries(**keywords)
        rows = list(filter(lambda entry: entry is not None, self.getUnsortedElements(**keywords)))
        self.writeRows(rows=rows, **keywords)

    # compatibility function for old code (up to interiorcad 2018 F3); do NOT use for new code (use writeEntriesAccordingScope instead)
    def writeWorkpiecesAndFittingsAccordingScope(self, **keywords):
        entries = self.generateUnsortedEntries(**keywords)
        rows = self.generateSortedElements(entries=entries, **keywords)
        self.writeRows(rows=rows, **keywords)

    def getUnsortedElements(self, **keywords):
        return (self.getElements(entry=entry, workpiece=entry.workpiece, positionInfo=entry.positionInfo, **keywords) for entry in keywords['entries'])

    # compatibility function for old code; do NOT use for new code
    def SortDataInLegacyWay(self, data):
        listSorter = DataListSorter(self.getSortingOrder())
        listSorter.sortList(data)

    # compatibility function for old code (up to interiorcad 2018 F3); do NOT use for new code
    def generateSortedElements(self, **keywords):
        if not hasattr(self, 'getSortingOrder'):
            keywords['entries'] = self.SortedEntriesByCriteria(**keywords)

        keywords['entries'] = self.getCumulatedEntries(**keywords)

        data = list(filter(lambda entry: entry is not None, self.getUnsortedElements(**keywords)))

        if hasattr(self, 'getSortingOrder'):
            self.SortDataInLegacyWay(data)

        return data

    def generateWorkpieceEntries(self, **keywords):
        entries = self.generateUnsortedWorkpieceEntries(**keywords)
        return self.generateSortedElements(entries=entries, **keywords)

    def generateFittingEntries(self, **keywords):
        entries = self.generateUnsortedFittingEntries(**keywords)
        return self.generateSortedElements(entries=entries, **keywords)
