# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import csv
import os.path
import string

from interiorcad.backend.cmoduleadapters.materials.boards import boards
from interiorcad.backend.cmoduleadapters.materials.coverings import coverings
from interiorcad.backend.cmoduleadapters.materials.edgebandings import edgebandings
from interiorcad.backend.cmoduleadapters.materials.finishings import finishings
from interiorcad.backend.cuttinglistentry import GRAIN_CROSSWISE
from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.dosfilenamecreator import DosFilenameCreator
from interiorcad.backend.filenameandpathfilter import FilenameAndPathFilter
from interiorcad.backend.filenamecleaner import FilenameCleaner
from interiorcad.backend.joint.calculator import JointCalculator
from interiorcad.backend.localization import getLocalizedString, CUTTINGLIST_RESOURCEGROUP
from interiorcad.backend.workpiece.reader import WorkpieceReader
from interiorcad.cuttinglistexport.ardisexport import ArdisExport
from interiorcad.cuttinglistexport.bormbusinessexport import BormBusinessExport
from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.cuttinglistexport.piosprofessionalexport import PiosProfessionalExport
from interiorcad.cuttinglistexport.schnittprofitexport import SchnittProfitExport
from interiorcad.cuttinglistexport.selcooptiplanningexport import SelcoOptiplanningExport
from interiorcad.interface.configuration import DATETIME_NOW
from interiorcad.interface.geometry.edges import EdgeData


class CustomExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'Custom'

    FIELD_TO_VALUE = {
        "project-id": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['ProjectID'], **keywords),
        "document-name": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['DocumentName'], **keywords),
        "project-commission": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['ProjectCommission'], **keywords),
        "project-construction": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['ProjectConstruction'], **keywords),
        "administrator": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['Administrator'], **keywords),
        "customer-number": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['CustomerNumber'], **keywords),
        "delivery-week": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['DeliveryWeek'], **keywords),
        "delivery-year": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['DeliveryYear'], **keywords),
        "delivery-date": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['DeliveryDate'], **keywords),
        "first-name": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['FirstName'], **keywords),
        "last-name": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['LastName'], **keywords),
        "street-name": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['StreetName'], **keywords),
        "postal-code": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['PostalCode'], **keywords),
        "city": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['City'], **keywords),
        "company": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['Company'], **keywords),
        "title": lambda entry, column, keywords, self: self.asString(data=keywords['projectInfo']['Title'], **keywords),

        "position-number": lambda entry, column, keywords, self: self.asString(data=keywords['positionInfo']['PositionNumber'], **keywords),
        "position-description": lambda entry, column, keywords, self: self.asString(data=keywords['positionInfo']['PositionDescription'], **keywords),
        "position-quantity": lambda entry, column, keywords, self: self.asString(data=keywords['positionInfo']['PositionQuantity'], **keywords),

        "cabinet-name": lambda entry, column, keywords, self: self.asString(data=keywords['positionInfo']['CabinetName'], **keywords),

        "item-id": lambda entry, column, keywords, self: self.asString(data=entry.itemID, **keywords),
        "board": lambda entry, column, keywords, self: self.asString(data=entry.itemID, **keywords),  # compatibility variable for getMaterialProperty(), because C++ returns hard coded value
        "description": lambda entry, column, keywords, self: self.asString(data=entry.description, **keywords),

        "length-workshop-strategy-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
        "width-workshop-strategy-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),

        "length-finished-size-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensionsGrainDirection.getLength(), **keywords),
        "width-finished-size-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensionsGrainDirection.getWidth(), **keywords),

        "length-finished-size-with-addition-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensionsWithAdditionGrainDirection.getLength(), **keywords),
        "width-finished-size-with-addition-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensionsWithAdditionGrainDirection.getWidth(), **keywords),

        "length-sawn-size-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensionsGrainDirection.getLength(), **keywords),
        "width-sawn-size-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensionsGrainDirection.getWidth(), **keywords),

        "length-sawn-size-with-addition-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensionsWithAdditionGrainDirection.getLength(), **keywords),
        "width-sawn-size-with-addition-grain-direction": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensionsWithAdditionGrainDirection.getWidth(), **keywords),

        "length-workshop-strategy": lambda entry, column, keywords, self: self.asCoordinate(data=entry.boardLength, **keywords),
        "width-workshop-strategy": lambda entry, column, keywords, self: self.asCoordinate(data=entry.boardWidth, **keywords),

        "length-finished-size": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensions.getLength(), **keywords),
        "width-finished-size": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensions.getWidth(), **keywords),

        "length-finished-size-with-addition": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensionsWithAddition.getLength(), **keywords),
        "width-finished-size-with-addition": lambda entry, column, keywords, self: self.asCoordinate(data=entry.finishedSizePartDimensionsWithAddition.getWidth(), **keywords),

        "length-sawn-size": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensions.getLength(), **keywords),
        "width-sawn-size": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensions.getWidth(), **keywords),

        "length-sawn-size-with-addition": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensionsWithAddition.getLength(), **keywords),
        "width-sawn-size-with-addition": lambda entry, column, keywords, self: self.asCoordinate(data=entry.sawnSizePartDimensionsWithAddition.getWidth(), **keywords),

        "joint-size-addition-front": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='front', **keywords), **keywords),
        "joint-size-addition-back": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='back', **keywords), **keywords),
        "joint-size-addition-left": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='left', **keywords), **keywords),
        "joint-size-addition-right": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='right', **keywords), **keywords),

        "joint-size-addition-grain-direction-front": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='front-grain-direction', **keywords), **keywords),
        "joint-size-addition-grain-direction-back": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='back-grain-direction', **keywords), **keywords),
        "joint-size-addition-grain-direction-left": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='left-grain-direction', **keywords), **keywords),
        "joint-size-addition-grain-direction-right": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointSizeAdditionPerEdge(edgeIdentifier='right-grain-direction', **keywords), **keywords),

        "thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.boardThickness, **keywords),
        "count": lambda entry, column, keywords, self: self.asInt(data=entry.count, **keywords),
        "name": lambda entry, column, keywords, self: self.asString(data=entry.name, **keywords),

        "edge-grain-direction-front-id": lambda entry, column, keywords, self: self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),
        "edge-grain-direction-front-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
        "edge-grain-direction-back-id": lambda entry, column, keywords, self: self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),
        "edge-grain-direction-back-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
        "edge-grain-direction-left-id": lambda entry, column, keywords, self: self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),
        "edge-grain-direction-left-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
        "edge-grain-direction-right-id": lambda entry, column, keywords, self: self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),
        "edge-grain-direction-right-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),

        "edge-part-dimensions-front-id": lambda entry, column, keywords, self: self.asString(data=entry.edges.front.supplierID, **keywords),
        "edge-part-dimensions-front-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edges.front.thickness, **keywords),
        "edge-part-dimensions-back-id": lambda entry, column, keywords, self: self.asString(data=entry.edges.back.supplierID, **keywords),
        "edge-part-dimensions-back-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edges.back.thickness, **keywords),
        "edge-part-dimensions-left-id": lambda entry, column, keywords, self: self.asString(data=entry.edges.left.supplierID, **keywords),
        "edge-part-dimensions-left-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edges.left.thickness, **keywords),
        "edge-part-dimensions-right-id": lambda entry, column, keywords, self: self.asString(data=entry.edges.right.supplierID, **keywords),
        "edge-part-dimensions-right-thickness": lambda entry, column, keywords, self: self.asCoordinate(data=entry.edges.right.thickness, **keywords),

        "edges-length": lambda entry, column, keywords, self: sum(entry.edges.getPolygonalExistingEdgeLengths()),
        "edges-count-polygonal": lambda entry, column, keywords, self: len(entry.edges.polygonal),
        "edges-count-rectangular": lambda entry, column, keywords, self: len(entry.edges.rectangular),

        "edge-ids-polygonal": lambda entry, column, keywords, self: self.asString(data=', '.join(entry.edges.getPolygonalExistingEdgeSupplierIDs()), **keywords),
        "edge-ids-polygonal-existing": lambda entry, column, keywords, self: self.asString(data=', '.join(entry.edges.getPolygonalExistingEdgeIDs()), **keywords),
        "edge-thicknesses-polygonal": lambda entry, column, keywords, self: ', '.join(self.asString(data=edgeThickness) for edgeThickness in entry.edges.getPolygonalExistingEdgeThicknesses()),
        "edge-lengths-polygonal": lambda entry, column, keywords, self: ', '.join(self.asString(data=edgeLength) for edgeLength in entry.edges.getPolygonalExistingEdgeLengths()),
        "edge-joints-polygonal": lambda entry, column, keywords, self: ', '.join('{start}-{end}'.format(start=self.getJointAsString(edgeJoints.start), end=self.getJointAsString(edgeJoints.end)) for edgeJoints in entry.edges.getPolygonalExistingEdgeJoints()),

        "edgecode-ardis-grain-direction": lambda entry, column, keywords, self: ArdisExport().getExportSpecificEdgeCode(lengthFollowsGrainDirection=True, **keywords),
        "edgecode-ardis-part-dimensions": lambda entry, column, keywords, self: ArdisExport().getExportSpecificEdgeCode(lengthFollowsGrainDirection=False, **keywords),
        "edgecode-borm-grain-direction": lambda entry, column, keywords, self: BormBusinessExport().getExportSpecificEdgeCode(lengthFollowsGrainDirection=True, **keywords),
        "edgecode-profacto-grain-direction": lambda entry, column, keywords, self: self.getEdgeCodeMappingWithParameter(column=column, edgecode=entry.edgesGrainDirection.edgecode, **keywords),
        "edgecode-profacto-part-dimensions": lambda entry, column, keywords, self: self.getEdgeCodeMappingWithParameter(column=column, edgecode=entry.edges.edgecode, **keywords),
        "edgecode-SchnittProfit-grain-direction": lambda entry, column, keywords, self: SchnittProfitExport().getExportSpecificEdgeCode(**keywords),
        "edgecode-SelcoOptiplanning-grain-direction": lambda entry, column, keywords, self: SelcoOptiplanningExport().getExportSpecificEdgeCode(edgecode=entry.edgesGrainDirection.edgecode, **keywords),

        "edgejoint-piosprofessional-grain-direction-front-left": lambda entry, column, keywords, self: PiosProfessionalExport().getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.front.jointEndTechnical, rhs=entry.edgesGrainDirection.left.jointStartTechnical),  # KanteAusbVL
        "edgejoint-piosprofessional-grain-direction-front-right": lambda entry, column, keywords, self: PiosProfessionalExport().getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.front.jointStartTechnical, rhs=entry.edgesGrainDirection.right.jointEndTechnical),  # KanteAusbVR
        "edgejoint-piosprofessional-grain-direction-back-right": lambda entry, column, keywords, self: PiosProfessionalExport().getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.back.jointEndTechnical, rhs=entry.edgesGrainDirection.right.jointStartTechnical),  # KanteAusbHR
        "edgejoint-piosprofessional-grain-direction-back-left": lambda entry, column, keywords, self: PiosProfessionalExport().getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.back.jointStartTechnical, rhs=entry.edgesGrainDirection.left.jointEndTechnical),  # KanteAusbHL

        "covering-1": lambda entry, column, keywords, self: self.asString(key='part_covering1', **keywords),
        "covering-1-thickness": lambda entry, column, keywords, self: self.asCoordinate(key='part_covering1thickness', **keywords),
        "covering-2": lambda entry, column, keywords, self: self.asString(key='part_covering2', **keywords),
        "covering-2-thickness": lambda entry, column, keywords, self: self.asCoordinate(key='part_covering2thickness', **keywords),

        "laminate-1": lambda entry, column, keywords, self: self.asString(key='part_laminate1', **keywords),
        "laminate-2": lambda entry, column, keywords, self: self.asString(key='part_laminate2', **keywords),
        "laminate-3": lambda entry, column, keywords, self: self.asString(key='part_laminate3', **keywords),

        "comment": lambda entry, column, keywords, self: self.asStringSingleLine(key='part_comment', **keywords),

        "grain-direction": lambda entry, column, keywords, self: self.asInt(data=entry.grain, **keywords),

        "nc-program-name-sizing": lambda entry, column, keywords, self: self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
        "nc-program-name-main": lambda entry, column, keywords, self: self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
        "nc-program-name-opposite": lambda entry, column, keywords, self: self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),

        "nc-program-path-sizing": lambda entry, column, keywords, self: self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
        "nc-program-path-main": lambda entry, column, keywords, self: self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathMain, **keywords),
        "nc-program-path-opposite": lambda entry, column, keywords, self: self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),

        "cabinet-width": lambda entry, column, keywords, self: self.asCoordinate(key='unit_unitwidth', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-depth": lambda entry, column, keywords, self: self.asCoordinate(key='unit_unitdepth', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-height": lambda entry, column, keywords, self: self.asCoordinate(key='unit_unitheight', uuid=keywords['workpiece'].parentUUID, **keywords),

        "cabinet-material-details": lambda entry, column, keywords, self: self.asString(key='unit_materialdetails', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-door-count": lambda entry, column, keywords, self: self.asString(key='unit_doorcount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-bifold-count": lambda entry, column, keywords, self: self.asString(key='unit_bifoldcount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-bifold-configuration": lambda entry, column, keywords, self: self.asString(key='unit_bifoldconf', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-door-flat-count": lambda entry, column, keywords, self: self.asString(key='unit_doorflatcount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-door-frame-count": lambda entry, column, keywords, self: self.asString(key='unit_doorframecount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-drawer-count": lambda entry, column, keywords, self: self.asString(key='unit_drawercount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-drawer-flat-count": lambda entry, column, keywords, self: self.asString(key='unit_drawerflatcount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-drawer-frame-count": lambda entry, column, keywords, self: self.asString(key='unit_drawerframecount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-case-mitre": lambda entry, column, keywords, self: self.asString(key='unit_casemitre', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-side-to-floor": lambda entry, column, keywords, self: self.asString(key='unit_sidestofloor', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-unit-id": lambda entry, column, keywords, self: self.asString(key='unit_unitid', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-flap-count": lambda entry, column, keywords, self: self.asString(key='unit_flapcount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "cabinet-flap-flat-count": lambda entry, column, keywords, self: self.asString(key='unit_flapflatcount', uuid=keywords['workpiece'].parentUUID, **keywords),
        "vw-document-path": lambda entry, column, keywords, self: self.asString(key='part_documentpath', uuid=self.getDocumentUUID(), **keywords),
        "vw-document-name": lambda entry, column, keywords, self: self.asString(key='part_documentname', uuid=self.getDocumentUUID(), **keywords),
        "construction-group-id-name": lambda entry, column, keywords, self: self.getDataAsInt(key='part_constructiongroupid', **keywords),
        "construction-group-component-id-name": lambda entry, column, keywords, self: self.getDataAsInt(key='part_constructiongroupcomponentid', **keywords),
        "empty-column": lambda entry, column, keywords, self: None,
        "cabinet-dimensions-string": lambda entry, column, keywords, self: self.getCabinetDimensionsString(**keywords),
        "sales-description": lambda entry, column, keywords, self: self.asString(key='part_description', **keywords),
        "sales-supplier": lambda entry, column, keywords, self: self.asString(key='part_supplier', **keywords),
        "sales-order-no": lambda entry, column, keywords, self: self.asString(key='part_ordernumber', **keywords),
        "sales-price": lambda entry, column, keywords, self: self.asCoordinate(key='sales_price', **keywords),

        "workshop-strategy-separate-sizing": lambda entry, column, keywords, self: self.shouldCreateSeparateSizing(),
        "workshop-strategy-sizing-rectangular": lambda entry, column, keywords, self: self.shouldCreateSizingForRectangularPart(),
        "workshop-strategy-jointing-edge-banding": lambda entry, column, keywords, self: self.getJointingInfo().jointing,
        "workshop-strategy-jointing-edge-banding-min": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointingInfo().minJoint, **keywords),
        "workshop-strategy-jointing-edge-banding-max": lambda entry, column, keywords, self: self.asCoordinate(data=self.getJointingInfo().maxJoint, **keywords),
        "workshop-strategy-cut-size-addition": lambda entry, column, keywords, self: self.asCoordinate(data=self.getsizingAddition(), **keywords),
        "workshop-strategy-covering-addition": lambda entry, column, keywords, self: self.asCoordinate(data=self.getAdditionalCoveringAddition(), **keywords),
        "workshop-strategy-construction-groups-enabled": lambda entry, column, keywords, self: self.asString(data=self.getIncludedConstructionGroupComponentsForExport(column=column), **keywords),
        "workshop-strategy-construction-groups-disabled": lambda entry, column, keywords, self: self.asString(data=self.getExcludedConstructionGroupComponentsForExport(column=column), **keywords),
        "workshop-strategy-set-name": lambda entry, column, keywords, self: self.getSetName(),
        "workshop-strategy-cut-size-configuration": lambda entry, column, keywords, self: self.getSizingText(),
        "workshop-strategy-sizing": lambda entry, column, keywords, self: self.shouldCreateSizing(**keywords),
        "has-machining": lambda entry, column, keywords, self: self.hasMachining(entry),

        # FUNCTIONS BELOW
        "map": lambda entry, column, keywords, self: self.mapData(column=column, **keywords),
        "constant-value": lambda entry, column, keywords, self: self.getConstantValue(column),
        "date-iso": lambda entry, column, keywords, self: self.getISODateAsString(),
        "cabinet-name-without-trailing-numbers": lambda entry, column, keywords, self: self.getCabinetNameWithoutTrailingNumbers(self.asString(key='unit_unitid', uuid=keywords['workpiece'].parentUUID, **keywords)),
        "entry-counter": lambda entry, column, keywords, self: self.getCounter(),
        "edges-count-front-back": lambda entry, column, keywords, self: self.countEdges(edges=(self.asString(data=entry.edges.front.supplierID, **keywords), self.asString(data=entry.edges.back.supplierID, **keywords)), **keywords),
        "edges-count-left-right": lambda entry, column, keywords, self: self.countEdges(edges=(self.asString(data=entry.edges.left.supplierID, **keywords), self.asString(data=entry.edges.right.supplierID, **keywords)), **keywords),
        "nc-program-base-path": lambda entry, column, keywords, self: self.createNCProgramBasePath(entry),
        "dos-file-name": lambda entry, column, keywords, self: self.asString(data=self.getDOSFileName(entry.name)),
        "nc-program-path-sizing-with-separator": lambda entry, column, keywords, self: self.getNCProgramDisplayNameWithSeparator(column=column, data=entry.ncProgramPathSizing, definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, **keywords),
        "nc-program-path-main-with-separator": lambda entry, column, keywords, self: self.getNCProgramDisplayNameWithSeparator(column=column, data=entry.ncProgramPathMain, definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, **keywords),
        "nc-program-path-opposite-with-separator": lambda entry, column, keywords, self: self.getNCProgramDisplayNameWithSeparator(column=column, data=entry.ncProgramPathOpposite, definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, **keywords),
        "bestopt-article": lambda entry, column, keywords, self: self.asString(data=self.getProjectInfoString(**keywords)),
        "bestopt-article-sanitized": lambda entry, column, keywords, self: self.asString(data=self.getProjectInfoStringSanitized(**keywords)),
        "name-shortened": lambda entry, column, keywords, self: self.asString(data=self.getFilteredPartName(entry.name)),
        "existing-nc-export-path-main": lambda entry, column, keywords, self: self.getExistingNCProgramNameForMainSide(**keywords),
        "existing-nc-export-path-opposite": lambda entry, column, keywords, self: self.getExistingNCProgramNameForOppositeSide(**keywords),
        "quote-function": lambda entry, column, keywords, self: self.quoteElement(column=column, **keywords),
        "quote-prefix-function": lambda entry, column, keywords, self: self.quoteElementPrefix(column=column, **keywords),
        "quote-suffix-function": lambda entry, column, keywords, self: self.quoteElementSuffix(column=column, **keywords),
        "ardis-grain-definition": lambda entry, column, keywords, self: self.getArdisGrain(**keywords),
        "function-concat": lambda entry, column, keywords, self: self.replacePlacerholders(column=column, **keywords),
        "material-properties": lambda entry, column, keywords, self: self.getMaterialProperty(column=column, **keywords),
    }

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.OUTPUT_END_OF_LINE = self.getConfigFor('end-of-line', 'CRLF', {'CR': '\r', 'LF': '\n'})
        self.CREATE_HEADER_LINE = self.getCuttinglistConfiguration().getAsBool('create-header', True)
        self.OUTPUT_ENCODING = self.getCuttinglistConfiguration().get('encoding', 'windows-1252')

        self.EXTENSION = self.getCuttinglistConfiguration().get('extension', '.csv')
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.OUTPUT_PART_NAME_FILTER = self.getCuttinglistConfiguration().get('output-part-name-filter', 'diacritics-replacements')

        self.csvquotechar = self.getConfigFor('string-delimiter', 'double-quotation-marks', {'double-quotation-marks': '\"', 'quotation-marks': '\''})
        if self.csvquotechar == 'none':
            self.csvquoting = csv.QUOTE_NONE
            self.csvquotechar = None

        self.decimalSeparator = self.getDecimalSeparator(**keywords)
        self.counter = 0
        self.EXPORT_TYPE = self.getCuttinglistConfiguration().get('export-location', 'file')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')
        self.NC_PROGRAM_DISPLAY_PATH_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-path-definition', 'pathandnamewithextension')
        self.CREATE_EMPTY_ROW_AFTER_HEADER = self.getCuttinglistConfiguration().getAsBool('empty-row-after-header', False)
        self.OPEN_EXPORT_PATH = self.getCuttinglistConfiguration().getAsBool('open-export-path', True)
        self.ARTICLE_DEFINITION = self.getCuttinglistConfiguration().get('article-variable', '{CabinetName}')
        self.ARDIS_GRAIN_DEFINITION = self.getCuttinglistConfiguration().get('ardis-grain-definition', '0,1,2').split(',')

    def doCuttinglistExport(self, **keywords):
        if self.EXPORT_TYPE == 'clipboard':
            self.doClipboardExport(**keywords)
        elif self.EXPORT_TYPE == "fileandclipboard":
            self.doClipboardExport(**keywords)
            super().doCuttinglistExport(**keywords)
            self.openExportPath()
        else:
            super().doCuttinglistExport(**keywords)
            self.openExportPath()

    def openExportPath(self, **keywords):
        if self.OPEN_EXPORT_PATH:
            super().openExportPath()

    def getConfigFor(self, configKey, defaultValue, mappingValues):
        configValue = self.getCuttinglistConfiguration().get(configKey, defaultValue)
        configValue = self.replaceValues(configValue, mappingValues)
        return configValue

    def replaceValues(self, value, mappingValues):
        newValue = value
        for key in mappingValues.keys():
            newValue = newValue.replace(key, mappingValues[key])
        return newValue

    def getDecimalSeparator(self, **keywords):
        return self.getConfigFor('decimal-separator', 'system', {'comma': ',', 'dot': '.', 'system': super(CustomExport, self).getDecimalSeparator(**keywords)})

    def getFieldSeparator(self, **keywords):
        return self.getConfigFor('field-separator', '\t', {'comma': ',', 'semicolon': ';', 'tab': '\t'})

    def getHeaderLine(self, **keywords):
        return () if not self.CREATE_HEADER_LINE else (
            entry.title for entry in self.getCuttinglistConfiguration().customColumns
        )

    def getElements(self, **keywords):
        return (
            self.getDataFrom(column, keywords) for column in self.getCuttinglistConfiguration().customColumns
        )

    def getDataFrom(self, column, keywords):
        entry = keywords['entry']
        data = column.data

        return self.FIELD_TO_VALUE[data](entry, column, keywords, self)

    def getISODateAsString(self):
        return DATETIME_NOW.strftime("%Y-%m-%d")

    def getCabinetNameWithoutTrailingNumbers(self, cabinetName):
        cabinetNameWithoutTrailingNumbers = cabinetName
        splitName = cabinetName.split('-')
        if len(splitName) > 1:
            if splitName[-1].isnumeric():
                cabinetNameWithoutTrailingNumbers = '-'.join(splitName[0:len(splitName) - 1])

        return cabinetNameWithoutTrailingNumbers

    def getCounter(self):
        self.counter += 1
        return self.counter

    def countEdges(self, **keywords):
        return sum(map(lambda edge: bool(edge), keywords['edges']))

    def getConstantValue(self, column):
        return self.getParameterValue(column, 'value', None)

    def quoteElement(self, **keywords):
        return self.getQuotedElementFrom((lambda element, quote: quote + element + quote), **keywords)

    def quoteElementPrefix(self, **keywords):
        return self.getQuotedElementFrom((lambda element, quote: quote + element), **keywords)

    def quoteElementSuffix(self, **keywords):
        return self.getQuotedElementFrom((lambda element, quote: element + quote), **keywords)

    def getQuotedElementFrom(self, concatFunc, **keywords):
        column = keywords['column']
        field = self.getParameterValue(column, 'field', None)
        quote = self.getParameterValue(column, 'value', None)

        element = str(self.FIELD_TO_VALUE[field](keywords['entry'], column, keywords, self)) if field is not None else None
        if element is not None and quote is not None:
            element = concatFunc(element, quote)

        return element

    def mapData(self, **keywords):
        column = keywords['column']
        data = column.data
        mapping = self.getMappingFrom(column)
        dataID = self.getMappingDataIDFrom(column)
        if dataID is not None:
            data = self.FIELD_TO_VALUE[dataID](keywords['entry'], column, keywords, self)

        mappedValue = mapping.get(str(data), data)

        return mappedValue

    def createNCProgramBasePath(self, entry):
        path = entry.ncProgramPathMain
        if not path:
            path = entry.ncProgramPathOpposite
        if not path:
            path = entry.ncProgramPathSizing
        if isinstance(path, list):
            path = path[0]
        return os.path.dirname(path) if path else ''

    def writeListHeader(self, **keywords):
        super().writeListHeader(**keywords)
        if self.CREATE_EMPTY_ROW_AFTER_HEADER:
            self.csvWriter.writerow('')

    def getJointSizeAdditionPerEdge(self, **keywords):
        if self.configuration.jointing:
            edgeIdentifier = keywords['edgeIdentifier']

            additionForEdges = WorkpieceReader(keywords['entry'].workpiece).getFinishedSizeAddition()
            jointSizeAddition = JointCalculator(self.configuration.minJoint, self.configuration.maxJoint).getJointAddition(additionForEdges)
            edgeFront = jointSizeAddition.bottom
            edgeBack = jointSizeAddition.top
            edgeLeft = jointSizeAddition.left
            edgeRight = jointSizeAddition.right

            if 'grain-direction' in edgeIdentifier and keywords['entry'].grain == GRAIN_CROSSWISE:
                if edgeIdentifier.startswith('front'):
                    return edgeRight
                elif edgeIdentifier.startswith('back'):
                    return edgeLeft
                elif edgeIdentifier.startswith('left'):
                    return edgeFront
                elif edgeIdentifier.startswith('right'):
                    return edgeBack

            else:
                if edgeIdentifier.startswith('front'):
                    return edgeFront
                elif edgeIdentifier.startswith('back'):
                    return edgeBack
                elif edgeIdentifier.startswith('left'):
                    return edgeLeft
                elif edgeIdentifier.startswith('right'):
                    return edgeRight

        else:
            return 0

    def getJointAsString(self, jointId):
        if jointId == EdgeData.JOINT_MITER:
            return self.getLocalizedValue(156, 'Miter')
        elif jointId == EdgeData.JOINT_CONTINUOUS:
            return self.getLocalizedValue(157, 'Continuous')
        elif jointId == EdgeData.JOINT_DISCONTINUOUS:
            return self.getLocalizedValue(158, 'Discontinuous')
        elif not jointId or jointId == EdgeData.JOINT_NOEDGE:
            return ''  # No value shown for no edge
        else:
            return self.getLocalizedValue(159, 'Unknown')

    def getEdgeCodeMappingWithParameter(self, **keywords):
        column = keywords['column']
        delimiter = self.getDelimiterFrom(column)
        mapping = self.getMappingFrom(column)

        return delimiter.join(element.translate(str.maketrans(mapping)) for element in keywords['edgecode'])

    def getDOSFileName(self, name):
        return DosFilenameCreator().createDosFilename(name)

    def getFilteredPartName(self, name):
        return FilenameAndPathFilter().filenameAndPathFilter(filterTerm=name, filterValue=self.OUTPUT_PART_NAME_FILTER, useIndividualFilters=True)

    def getNCProgramDisplayNameWithSeparator(self, **keywords):
        column = keywords['column']
        ncProgramPath = keywords['data']

        if len(column.parameters) >= 1 and ncProgramPath:
            separator = column.parameters[0].value
            dirName = os.path.dirname(ncProgramPath)
            baseName = os.path.basename(ncProgramPath)
            keywords['data'] = os.path.join(dirName, str(separator + baseName))  # New ncProgramPath

        return self.getNCProgramDisplayName(**keywords)

    def deleteUnwantedUnderscores(self, element):
        return FilenameCleaner.getSimplifiedMultipleUnderscores(element).strip('_')

    def getProjectInfosArray(self, **keywords):
        projectInfo = self.getProjectInfo()
        positionInfo = keywords['positionInfo']
        return {
            'CabinetName': positionInfo['CabinetName'],
            'ProjectID': projectInfo['ProjectID'],
            'PositionNumber': positionInfo['PositionNumber'],
        }

    def getProjectInfoStringSanitized(self, **keywords):
        return self.deleteUnwantedUnderscores(self.getProjectInfoString(**keywords))

    def getProjectInfoString(self, **keywords):
        projectInfo = self.getProjectInfo()
        if not projectInfo['ProjectID'] and (self.ARTICLE_DEFINITION == '{PositionNumber}' or self.ARTICLE_DEFINITION == '{ProjectID}_{PositionNumber}'):
            self.ARTICLE_DEFINITION = '{CabinetName}'
        return self.ARTICLE_DEFINITION.format_map(self.getProjectInfosArray(**keywords))

    def getExistingNCProgramNameForMainSide(self, **keywords):
        entry = keywords['entry']

        ncProgramNameMain = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords)
        ncProgramNameOpposite = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)

        if ncProgramNameMain:
            return ncProgramNameMain
        else:
            return ncProgramNameOpposite

    def getExistingNCProgramNameForOppositeSide(self, **keywords):
        entry = keywords['entry']

        ncProgramNameMain = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords)
        ncProgramNameOpposite = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)

        if ncProgramNameMain:
            return ncProgramNameOpposite
        else:
            return ''

    def getCabinetDimensionsString(self, **keywords):
        positionInfo = keywords['positionInfo']
        workpiece = keywords['workpiece']

        unitname = self.asString(data=positionInfo['CabinetName'], **keywords)
        unitwidth = self.asCoordinate(key='unit_unitwidth', uuid=workpiece.parentUUID, **keywords)
        unitdepth = self.asCoordinate(key='unit_unitdepth', uuid=workpiece.parentUUID, **keywords)
        unitheight = self.asCoordinate(key='unit_unitheight', uuid=workpiece.parentUUID, **keywords)
        return '{width} x {height} x {depth}'.format(width=unitwidth, height=unitheight, depth=unitdepth) if unitname else ''

    def getArdisGrain(self, **keywords):
        entry = keywords['entry']
        # grain 1=L or 2=W must be in the grain direction - 0=blank = no grain - 3=E = L or W but consistent within the job
        if entry.grain == self.GRAIN_NONE:  # Ardis 0
            return self.ARDIS_GRAIN_DEFINITION[0]
        elif entry.grain == self.GRAIN_CROSSWISE:  # Ardis 2
            return self.ARDIS_GRAIN_DEFINITION[2]
        else:  # Ardis 1 / self.GRAIN_LENGTHWISE
            return self.ARDIS_GRAIN_DEFINITION[1]

    def replacePlacerholders(self, **keywords):
        column = keywords['column']
        output = None
        if len(column.parameters) >= 1:
            stringWithPlacerholders = column.parameters[0].value
            placeholders = {}
            for field in string.Formatter().parse(stringWithPlacerholders):
                placeholder = self.getPlaceholdersFrom(field)
                placeholders[placeholder[1]] = tuple((placeholder[0], [placeholder[2]]))

            fieldValues = {key: self.FIELD_TO_VALUE[key](keywords['entry'], column, keywords, self) if key in self.FIELD_TO_VALUE else self.getMappedValueForPlaceholder(key, value[1])
                           for key, value in placeholders.items()}

            for fieldKey, fieldValue in fieldValues.items():
                if fieldValue == '' and 'I:' in placeholders[fieldKey][1][0]:
                    stringWithPlacerholders = stringWithPlacerholders.replace(placeholders[fieldKey][0] + '{' + placeholders[fieldKey][1][0] + '}', '', 1)

            stringWithPlacerholders = stringWithPlacerholders.replace('I:', '')
            output = stringWithPlacerholders.format_map(fieldValues)

        return output

    def getMappedValueForPlaceholder(self, key, value):
        if key is None:
            return None
        if 'I:' in value[0]:
            return ''
        else:
            return '{key} not found'.format(key=key)

    def getPlaceholdersFrom(self, field):
        stringBeforePlaceholder = field[0]
        placeholder1 = field[1]
        placeholder2 = field[2]

        placeholderWithoutModifier = placeholder1
        rawPlaceholder = placeholderWithoutModifier
        if placeholder2:
            placeholderWithoutModifier = placeholder2
            rawPlaceholder = str(placeholder1 + ':' + placeholder2)

        return tuple((stringBeforePlaceholder, placeholderWithoutModifier, rawPlaceholder))

    def getExcludedConstructionGroupComponentsForExport(self, **keywords):
        if self.getExcludedConstructionGroups():
            return self.filterConstructionGroupComponentIDsForExport(filterFunction=(lambda mapping, cg: mapping.get(str(cg), None) is not None), **keywords)

    def getIncludedConstructionGroupComponentsForExport(self, **keywords):
        if self.getExcludedConstructionGroups():
            return self.filterConstructionGroupComponentIDsForExport(filterFunction=(lambda mapping, cg: mapping.get(str(cg), None) is None), **keywords)

    def filterConstructionGroupComponentIDsForExport(self, **keywords):
        filterFunction = keywords['filterFunction']
        column = keywords['column']
        delimiter = self.getDelimiterFrom(column)

        cgcNameMappingFromColumn = self.getMappingFrom(column)  # Localized dictionary of all cgcs {id: 'name'}
        cgcNameMappingFromConfiguration = {entry.get('id'): entry.get('name') for entry in self.getExcludedConstructionGroups()}  # localized dictionary of excluded cgcid: cgcname

        filteredConstructionGroupComponents = set()
        for cgc in self.getConstructionGroupComponentsOfScope(**keywords):  # All cgcids of the included construction group components of the export
            if filterFunction(cgcNameMappingFromConfiguration, cgc):  # Depending on the filterFunction the included or excluded cgc names are added to the set of filtered construction group components
                filteredConstructionGroupComponents.add(cgcNameMappingFromColumn.get(str(cgc), cgcNameMappingFromConfiguration.get(cgc)))

        return delimiter.join(filteredConstructionGroupComponents)

    def getConstructionGroupComponentsOfScope(self, **keywords):
        constructionGroupComponents = {ExportDataCollector.getConstructionGroupComponentID(uuid) for uuid, _ in ExportDataCollector.getUnfiltertedAllCustomPartUUIDsAndCounts()}
        return constructionGroupComponents

    def hasMachining(self, entry):
        workpiece = entry.workpiece
        if workpiece.drillings:
            return True
        if workpiece.centerPunches:
            return True
        if workpiece.dadosrabbets:
            return True
        if workpiece.miters:
            return True
        if workpiece.contours:
            return True
        if workpiece.ncmacros:
            return True
        if workpiece.chamfers:
            return True
        return False

    def getMappingFrom(self, column):
        mapping = dict()
        for parameter in column.parameters:
            if parameter.key == 'mapping':
                mapping = parameter.mapping

        return mapping

    def getDelimiterFrom(self, column):
        delimiter = self.getParameterValue(column, 'delimiter', '')
        if delimiter == self.csvdelimiter:
            delimiter = '_'

        return delimiter

    def getMappingDataIDFrom(self, column):
        return self.getParameterValue(column, 'mapping', None)

    def getMaterialProperty(self, **keywords):
        column = keywords['column']
        materialKey = self.getParameterValue(column, 'material')
        materialPropertyKey = self.getParameterValue(column, 'property')
        materialID = self.FIELD_TO_VALUE[materialKey](keywords['entry'], column, keywords, self) if materialKey in self.FIELD_TO_VALUE else None
        data = None
        materialProperty = ''

        if materialID is not None:
            # materialKeys ('material') are returned hard coded from C++
            if 'board' in materialKey:
                data = boards.getBoardsData(materialID)
            if 'covering' in materialKey:
                data = coverings.getCoveringsData(materialID)
            if 'laminate' in materialKey:
                data = finishings.getFinishingsData(materialID)
            if 'edge' in materialKey:
                data = edgebandings.getEdgeBandingsData(materialID)

        if data is not None:
            if materialPropertyKey == 'description':
                materialProperty = self.asString(data=data.description, **keywords)
            if materialPropertyKey == 'supplier-id':
                materialProperty = self.asString(data=data.supplierid, **keywords)
            if materialPropertyKey == 'supplier':
                materialProperty = self.asString(data=data.supplier, **keywords)
            if materialPropertyKey == 'waste':
                materialProperty = self.asCoordinate(data=data.waste, **keywords)
            if materialPropertyKey == 'unit':
                materialProperty = self.asString(data=data.unit, **keywords)
            if materialPropertyKey == 'price':
                materialProperty = self.asCoordinate(data=data.price, **keywords)
            if materialPropertyKey == 'markup':
                materialProperty = self.asCoordinate(data=data.markup, **keywords)
            if materialPropertyKey == 'amount-per-unit':
                materialProperty = self.asCoordinate(data=data.amountperunit, **keywords)
            if materialPropertyKey == 'thickness':
                materialProperty = self.asCoordinate(data=data.thickness, **keywords)
            if materialPropertyKey == 'group':
                materialProperty = self.asString(data=data.group, **keywords)

        return materialProperty

    def getParameterValue(self, column, key, defaultValue=''):
        parameterValue = defaultValue
        for parameter in column.parameters:
            if parameter.key == key:
                parameterValue = parameter.value

        if parameterValue == self.csvdelimiter:
            parameterValue = ''
        if parameterValue == self.csvquotechar:
            parameterValue = ''
        if parameterValue == self.decimalSeparator:
            parameterValue = ''

        return parameterValue

    def getLocalizedValue(self, index, default):
        return getLocalizedString(CUTTINGLIST_RESOURCEGROUP, index, default)
