# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from openpyxl import Workbook

from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.localization import CUTTINGLISTEXPORT_MESSAGES
from interiorcad.cuttinglistexport.base import CuttinglistExportBase
from interiorcad.interface.configuration import DATETIME_NOW_UTC


class F4NestExport(CuttinglistExportBase):
    EXPORT_FORMAT = 'F4Nest'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = ';'

        self.EXTENSION = 'xlsx'
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')  # | partsandcoverings | coveringsonly
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        # We don't use self.CUMULATE_WORKPIECES, because all parts have to be one by one

    DATETIME_NOW_UTC_EXPORT = DATETIME_NOW_UTC  # Needed for patches for datetime consistent tests

    def doExport(self, **keywords):  # XLSX
        self.wb = Workbook()
        self.wb.properties.created = self.DATETIME_NOW_UTC_EXPORT
        self.wb.properties.modified = self.DATETIME_NOW_UTC_EXPORT

        self.ws = self.wb.active

        self.writeHeader(**keywords)
        self.writeBody(**keywords)

        self.wb.save(filename=keywords['outputFile'])

    def doFileExport(self, **keywords):
        exportPath = self.prepareExportPath(**keywords)

        projectInfo = self.getProjectInfo()
        self.uuidsAndCounts = tuple(ExportDataCollector.getAllCustomPartUUIDsAndCounts())

        try:
            self.doExport(outputFile=exportPath, projectInfo=projectInfo)
        except PermissionError as e:
            message = self.getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 3, "Cannot write {file}. Is the file opened by another program? Has the current user permissions to write the file into the folder?").format(file=exportPath)
            raise RuntimeError(message) from e

    # Header
    def writeHeader(self, **keywords):
        self.ws.append(self.getHeaderLine(**keywords))

    def getHeaderLine(self, **keywords):
        return (
            'Material',
            'ProgramName',
            'Mirror',
            'Quantity',
            'OptQuantity',
            'Angle',
            'Length',
            'Width',
            'LabelInfo1',
            'LabelInfo2',
            'LabelInfo3',
            'LabelInfo4',
            'LabelInfo5',
            'LabelInfo6',
            'LabelInfo7',
            'LabelInfo8',
            'LabelInfo9',
            'LabelInfo10',
            'LabelInfo11',
            'LabelInfo12',
            'LabelInfo13',
        )

    # Body
    def writeBody(self, **keywords):
        keywords['entries'] = self.generateSortedEntries(**keywords)
        rows = (self.getElements(entry=entry, workpiece=entry.workpiece, positionInfo=entry.positionInfo, **keywords) for entry in keywords['entries'])
        for row in rows:
            self.ws.append(row)

    # nearly identical with maestronestingexport, but reduced
    def getNCProgramNamesForMainOrOpposite(self, **keywords):
        entry = keywords['entry']
        side = keywords['side']
        path = keywords['path']

        ncProgramNameMain = entry.ncProgramNameMain
        ncProgramNameOpposite = entry.ncProgramNameOpposite

        ncProgramPathMain = entry.ncProgramPathMain
        ncProgramPathOpposite = entry.ncProgramPathOpposite

        if side == 'main' and ncProgramNameMain:
            return ncProgramNameMain if path == 'absolut' else ncProgramPathMain
        elif side == 'main':
            return ncProgramNameOpposite if path == 'absolut' else ncProgramPathOpposite
        elif side == 'opposite' and ncProgramNameMain:
            return ncProgramNameOpposite if path == 'absolut' else ncProgramPathOpposite
        else:
            return ''

    def getRotationAngle(self, **keywords):
        entry = keywords['entry']
        grain = entry.grain

        if grain == self.GRAIN_LENGTHWISE or grain == self.GRAIN_CROSSWISE:
            return 1  # Step 180°
        elif grain == self.GRAIN_NONE and entry.workpiece.isRectangular():
            return 2  # Step 90°
        elif grain == self.GRAIN_NONE and entry.workpiece.isPolygonal():
            return 3  # Free
        else:
            return 1

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']

        return (
            # Essential part information
            self.asString(data=entry.itemID, **keywords),
            self.getNCProgramNamesForMainOrOpposite(side='main', path='relative', **keywords),
            '0',  # Mirror not allowed
            self.asInt(data=entry.count, **keywords),
            '0',  # OptQuantity
            self.getRotationAngle(**keywords),
            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),

            # LabelInfos
            'positionx={posx}'.format(posx=self.asCoordinate(data=entry.boardLengthGrainDirection / 2, **keywords)),
            'positiony={posy}'.format(posy=self.asCoordinate(data=entry.boardWidthGrainDirection / 2, **keywords)),
            'edgeright={edge}'.format(edge=self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords)),
            'edgeleft={edge}'.format(edge=self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords)),
            'edgerear={edge}'.format(edge=self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords)),
            'edgefront={edge}'.format(edge=self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords)),
            'comment={}'.format(self.asStringSingleLine(key='part_comment', **keywords)),
            'project={}'.format(self.asString(data=projectInfo['ProjectID'], **keywords)),
            'customer={}'.format(self.asString(data=projectInfo['ProjectCommission'], **keywords)),
            'construction={}'.format(self.asString(data=projectInfo['ProjectConstruction'], **keywords)),
            'positionnumber={}'.format(self.asString(data=positionInfo['PositionNumber'], **keywords)),
            'positiondescription={}'.format(self.asString(data=positionInfo['PositionDescription'], **keywords)),
            'cabinetname={}'.format(self.asString(data=positionInfo['CabinetName'], **keywords)),
        )
