# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import csv
import logging

from interiorcad.backend.localization import getLocalizedString, CUTTINGLISTEXPORT_MESSAGES
from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase


class HHOSExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'HHOS'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvquoting = csv.QUOTE_NONE
        self.csvdelimiter = self.getCuttinglistConfiguration().get('csv-field-delimiter', ';')

        self.EXTENSION = 'csv'
        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')
        self.BAZ_PROGRAM_NAME = self.getCuttinglistConfiguration().get('baz-program-name', 'single')

        self.checkProjectID(**keywords)

    def checkProjectID(self, **keywords):
        projectInfo = self.getProjectInfo()
        if not projectInfo['ProjectID']:
            logging.warning(getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 5, "Mandatory Project ID not entered in Project Details, exporting Document Name as Project ID instead."))

    # Header
    def getHeaderLine(self, **keywords):
        return (
            'Auftrag',
            'Pos',  # Positionsnummer des Elements

            'Mat',
            'Teilbez',  # Bezeichnung des Elements
            'Stueck',

            'FLaenge',  # Länge des fertigen Elements (incl. Anleimer u. Umleimer)
            'FBreite',  # Breite des fertigen Elements (incl. Anleimer u. Umleimer)
            'RohLaenge',  # Länge des Elements nach dem Rohzuschnitt
            'RohBreite',  # Breite des Elements nach dem Rohzuschnitt
            'Drehbar',  # 0=no; 1=yes

            'KaVoBez',
            'KaVoTyp',  # A oder 0 = Anleimer; U oder 1 = Umleimer
            'KaVoRohdicke',
            # 'KaVoSaum',   # Besäumung des Elements an dieser Kante
            'KaHiBez',
            'KaHiTyp',  # A oder 0 = Anleimer; U oder 1 = Umleimer
            'KaHiRohdicke',
            # 'KaVoSaum',   # Besäumung des Elements an dieser Kante
            'KaLiBez',
            'KaLiTyp',  # A oder 0 = Anleimer; U oder 1 = Umleimer
            'KaLiRohdicke',
            # 'KaVoSaum',   # Besäumung des Elements an dieser Kante
            'KaReBez',
            'KaReTyp',  # A oder 0 = Anleimer; U oder 1 = Umleimer
            'KaReRohdicke',
            # 'KaVoSaum',   # Besäumung des Elements an dieser Kante

            'BelagI',
            'BelagA',

            'Info1',  # Zusatzinformationszeile 1-9
            'Info2',
            'Info3',
            'Info4',
            'Info5',
            'Info6',
            'Info7',
            'Info8',
            'Info9',

            'BAZPgm',  # Name des BAZ-Programms (255 Zeichen); Sizing, Main & Opposite gemergt

            'AEndDatum',  # Produktionsdatum Ende des Auftrags

            'AInfo1',  # Zusatzinformationszeile des Auftrags 1-6
            'AInfo2',
            'AInfo3',
            'AInfo4',
            'AInfo5',
            'AInfo6',

            'ElemIdent',  # Elementidentifikation (Elementnummer)
        )

    # Elements
    def getMandatoryProjectID(self, **keywords):
        projectInfo = keywords['projectInfo']
        projectID = projectInfo['ProjectID']

        if projectID:
            return projectID
        else:
            return projectInfo['DocumentName']

    def getSimplifiedGrain(self, grain):
        # If the part can be rotated
        if grain == self.GRAIN_NONE:
            return 1
        else:  # self.GRAIN_LENGTHWISE, self.GRAIN_CROSSWISE:
            return 0

    def getBAZProgramName(self, **keywords):
        entry = keywords['entry']

        if self.BAZ_PROGRAM_NAME == 'single':
            BAZProgramNameMain = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords)
            BAZProgramNameOpposite = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)

            if BAZProgramNameMain:
                return BAZProgramNameMain
            elif BAZProgramNameOpposite:
                return BAZProgramNameOpposite
            else:
                return ''

        else:  # 'merged'
            return self.mergeNCProgramnames(**keywords)  # NC-Program names merged into one field

    def mergeNCProgramnames(self, **keywords):
        entry = keywords['entry']
        reducedDisplayNameDefinition = self.NC_PROGRAM_DISPLAY_NAME_DEFINITION.replace('path', '')

        return ' / '.join(
            filter(lambda name: bool(name), (
                self.getNCProgramDisplayName(definition=reducedDisplayNameDefinition, data=entry.ncProgramPathSizing, **keywords),
                self.getNCProgramDisplayName(definition=reducedDisplayNameDefinition, data=entry.ncProgramPathMain, **keywords),
                self.getNCProgramDisplayName(definition=reducedDisplayNameDefinition, data=entry.ncProgramPathOpposite, **keywords))
            )
        )

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']

        return (
            self.asString(data=self.getMandatoryProjectID(**keywords), **keywords),
            self.asString(data=positionInfo['PositionNumber'], **keywords),

            self.asString(data=entry.itemID, **keywords),
            self.asString(data=entry.description, **keywords),  # Teilbez
            self.asInt(data=entry.count, **keywords),

            self.asCoordinate(data=entry.finishedSizePartDimensionsGrainDirection.getLength(), **keywords),
            self.asCoordinate(data=entry.finishedSizePartDimensionsGrainDirection.getWidth(), **keywords),
            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asInt(data=self.getSimplifiedGrain(entry.grain)),

            self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),
            'U',
            self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
            # joining during edge banding
            self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),
            'U',
            self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
            # joining during edge banding
            self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),
            'U',
            self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
            # joining during edge banding
            self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),
            'U',
            self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),
            # joining during edge banding

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),

            # Info1-9
            self.asStringSingleLine(key='part_comment', **keywords),
            self.asString(data=positionInfo['PositionDescription'], **keywords),
            self.asString(data=positionInfo['CabinetName'], **keywords),
            self.asString(key='part_laminate1', **keywords),
            self.asString(key='part_laminate2', **keywords),
            self.asString(key='part_laminate3', **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),

            self.getBAZProgramName(**keywords),  # BAZPgm

            self.asString(data=projectInfo['DeliveryDate'], **keywords),  # AEndDatum

            # AInfo1-6
            self.asString(data=projectInfo['DocumentName'], **keywords),
            self.asString(data=projectInfo['ProjectCommission'], **keywords),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),
            '',
            '',
            '',

            self.asString(data=entry.name, **keywords),  # ElemIdent
        )
