# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.cuttinglistexport.base import CuttinglistExportBase
from interiorcad.interface.geometry.edges import EdgeData


class KuhnleExport(CuttinglistExportBase):
    EXPORT_FORMAT = 'Kuhnle'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.counter = 0
        self.lastPositionNumber = 0

        self.EXTENSION = 'c2h'
        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA
        # Sorting Criteria 'cabinetname' is not part of getPositionLine and getElementLine.

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsandfittings')  # | partsonly | fittingsonly
        self.TEILETEXT = self.getCuttinglistConfiguration().get('teiletext', 'Name')  # alternativ: Description
        self.EXPORT_CABINET_NAME = self.getCuttinglistConfiguration().get('export-cabinet-name', 'none')
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')  # only minimal accumulation
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')
        self.NC_PROGRAM_NAME_OUTPUT = self.getCuttinglistConfiguration().get('nc-program-name-output', '')  # Default empty as NC Record, else Part Record

    def getDecimalSeparator(self, **keywords):
        return '.'

    def getHeaderLine(self, **keywords):
        return None

    # ElementLines
    def getElementPositionLine(self, **keywords):
        positionInfo = keywords['positionInfo']
        positionDescription = positionInfo['PositionDescription']
        if positionDescription == '':
            positionDescription = self.asString(data=positionInfo['CabinetName'], **keywords)

        return ';'.join((
            '"P"',
            '"{PosNr}"'.format(PosNr=self.asString(data=positionInfo['PositionNumber'], **keywords)),
            '"1"',
            '""',
            '""',
            '"1"',
            '"{PosText}"'.format(PosText=self.asString(data=positionDescription, **keywords)[:30]),  # Pos-Text
            '"{PosText}"'.format(PosText=self.asString(data=positionDescription, **keywords)),  # Kunden-Text
        ))

    def getPositionLine(self, **keywords):
        entry = keywords['entry']
        positionLine = None
        currentPositionNumber = entry.positionInfo['PositionNumber']
        if currentPositionNumber != self.lastPositionNumber:
            positionLine = self.getElementPositionLine(**keywords)
            self.lastPositionNumber = currentPositionNumber
        return positionLine

    def getElementPartCoverings(self, **keywords):
        return joinIfNotNone(' ',
            'v{}'.format(self.asString(key='part_covering1', **keywords)) if self.asString(key='part_covering1', **keywords) else None,
            'h{}'.format(self.asString(key='part_covering2', **keywords)) if self.asString(key='part_covering2', **keywords) else None,
            'b{}'.format(self.asString(key='part_laminate1', **keywords)) if self.asString(key='part_laminate1', **keywords) else None,
            'b{}'.format(self.asString(key='part_laminate2', **keywords)) if self.asString(key='part_laminate2', **keywords) else None,
            'b{}'.format(self.asString(key='part_laminate3', **keywords)) if self.asString(key='part_laminate3', **keywords) else None,
        )

    def getJointAsString(self, jointId):
        if jointId == EdgeData.JOINT_MITER:
            return 'G'
        elif jointId == EdgeData.JOINT_CONTINUOUS:
            return 'L'
        elif jointId == EdgeData.JOINT_DISCONTINUOUS:
            return 'K'

    def getExportSpecificEdgeCode(self, **keywords):
        edge = keywords['edge']
        supplierID = edge.supplierID

        if supplierID:
            return joinIfNotNone('',
                keywords['identifier'],
                self.getDataAsString(data=supplierID, **keywords),
                '/{thickness}'.format(thickness=self.asCoordinate(data=edge.thickness, **keywords)),
                ':U{joint}'.format(joint=self.getJointAsString(jointId=edge.jointStartTechnical)) if edge.jointStartTechnical else None,  # Anleimer/Umleimer + Kantenübergang am Ende der Kante entgegen dem Uhrzeigersinn.
            )

    def getElementPartEdges(self, **keywords):
        entry = keywords['entry']

        return joinIfNotNone(' ',
            # edges vorne/hinten/links/rechts
            self.getExportSpecificEdgeCode(identifier='l', edge=entry.edgesGrainDirection.front, **keywords),
            self.getExportSpecificEdgeCode(identifier='r', edge=entry.edgesGrainDirection.back, **keywords),
            self.getExportSpecificEdgeCode(identifier='o', edge=entry.edgesGrainDirection.left, **keywords),
            self.getExportSpecificEdgeCode(identifier='u', edge=entry.edgesGrainDirection.right, **keywords),
        )

    def getElementPartNCLines(self, **keywords):
        entry = keywords['entry']
        ncProgramFieldNames = self.NC_PROGRAM_NAME_OUTPUT.split(',')

        if any(name.strip() for name in ncProgramFieldNames):
            return ';'.join((
                '"{name}={value}"'.format(name=ncProgramFieldNames[0], value=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords)),
                '"{name}={value}"'.format(name=ncProgramFieldNames[1], value=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords)),
                '"{name}={value}"'.format(name=ncProgramFieldNames[2], value=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)),
            ))
        else:
            return None

    def getElementPartLine(self, **keywords):
        self.counter += 1
        entry = keywords['entry']
        positionInfo = keywords['positionInfo']

        teiletext = self.asString(data=entry.name, **keywords) if self.TEILETEXT == 'Name' else self.asString(data=entry.description, **keywords)
        werkstatttext = self.asStringSingleLine(key='part_comment', **keywords)
        cabinetName = self.asString(data=positionInfo['CabinetName'])

        if self.EXPORT_CABINET_NAME == 'Teiletext':
            teiletext = joinIfNotNone(', ',
                cabinetName if cabinetName else None,
                teiletext,
            )
        elif self.EXPORT_CABINET_NAME == 'Werkstatttext':
            if cabinetName or werkstatttext:
                werkstatttext = joinIfNotNone(', ',
                    cabinetName if cabinetName else None,
                    werkstatttext if werkstatttext else None,
                )
            else:
                werkstatttext = ''

        line = joinIfNotNone(';',
            '"T"',
            '"{}"'.format(self.asInt(data=self.counter, **keywords)),
            '"{}"'.format(self.asInt(data=entry.count, **keywords)),
            '"{}"'.format(self.asString(data=entry.description, **keywords)[:20]),  # Teiletyp
            '"{}"'.format(self.asString(data=entry.itemID, **keywords)),
            '"{}"'.format(self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords)),
            '"{}"'.format(self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords)),
            '"{}"'.format(self.asCoordinate(data=entry.boardThickness, **keywords)),
            '"{}"'.format(teiletext[:30]),  # Teiletext
            '"{}"'.format(self.getElementPartCoverings(**keywords) or ""),
            '"{}"'.format(self.getElementPartEdges(**keywords) or ""),
            '""',
            '"{}"'.format(werkstatttext),  # Werkstatttext
            self.getElementPartNCLines(**keywords) if self.NC_PROGRAM_NAME_OUTPUT else None,
        )
        return line

    def getElementNCLine(self, **keywords):
        line = ';'.join((
            '"N"',
            '"{ncProgram}"'.format(ncProgram=keywords['ncProgram']),
        ))
        return line if keywords['ncProgram'] != '' else None

    def getElementNCLineFromList(self, **keywords):
        ncPrograms = keywords['ncPrograms']

        if isinstance(ncPrograms, str):
            return self.getElementNCLine(ncProgram=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=ncPrograms, **keywords), **keywords)

        elif isinstance(ncPrograms, list) and any(ncPrograms):
            ncProgramDisplayNameList = []
            for element in ncPrograms:
                ncProgramDisplayNameList.append(self.getElementNCLine(ncProgram=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=element, **keywords), **keywords))
            return self.OUTPUT_END_OF_LINE.join(ncProgramDisplayNameList)

        else:
            return None

    def getElementNCLines(self, **keywords):
        entry = keywords['entry']

        return joinIfNotNone(self.OUTPUT_END_OF_LINE,
            self.getElementNCLineFromList(ncPrograms=entry.ncProgramPathSizing, **keywords),  # NC-DS
            self.getElementNCLineFromList(ncPrograms=entry.ncProgramPathMain, **keywords),  # NC-DS
            self.getElementNCLineFromList(ncPrograms=entry.ncProgramPathOpposite, **keywords),  # NC-DS
        )

    def getWorkpieceLines(self, **keywords):
        # Kuhnle Structure
        return joinIfNotNone(self.OUTPUT_END_OF_LINE,
            self.getPositionLine(**keywords),  # Positions-DS
            self.getElementPartLine(**keywords),  # Teile-DS
            self.getElementNCLines(**keywords) if not self.NC_PROGRAM_NAME_OUTPUT else None,  # NC-DS
        )

    def writeBody(self, **keywords):
        entries = self.generateSortedEntries(**keywords)

        lines = (self.getWorkpieceLines(entry=entry, workpiece=entry.workpiece, positionInfo=entry.positionInfo, **keywords) for entry in entries)
        self.writeLines(lines=lines, **keywords)

    # doExport
    def doExport(self, **keywords):
        # self.writeHeader(**keywords)
        self.writeBody(**keywords)
