# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import csv
import logging
from datetime import datetime

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.backend.localization import getLocalizedString, CUTTINGLISTEXPORT_MESSAGES


class MaestroOttimoCutExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'MaestroOttimoCut'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = ','
        self.csvquoting = csv.QUOTE_NONE
        self.csvquotechar = None

        self.EXTENSION = 'MLC'
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')  # | partsandcoverings | coveringssonly
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

        self.DELIVERY_DATE_FORMAT = self.getCuttinglistConfiguration().get('delivery-date-format', 'none')  # No default, user has to set up the format by himself.

        self.counter = 0

    def getDecimalSeparator(self, **keywords):
        return '.'

    # Dimensions must have 3 decimals
    def asPlusThreeCoordinate(self, **keywords):
        value = self.getDataAsFloat(**keywords)
        return '{0:.3f}'.format(value)

    def getFixedLengthAlphaNumeric(self, **keywords):
        value = (keywords['value'])[:keywords['length']]  # HACK: to avoid a C Assert in Python (Debug-Mode) (IC-11640)
        return '{value:<{length}}'.format(value=value, length=keywords['length'])[:keywords['length']]

    def removeCommasOfIDs(self, **keywords):
        id = keywords['id']
        if ',' in id:
            message = getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 4, "Illegal CSV character ',' removed from id '{id}'. Maintain your master data for full consistency!")
            logging.warning(message.format(id=id))
            return id.replace(',', '.')
        else:
            return id

    def quoteElement(self, element):
        quoteChar = '"'
        return quoteChar + element + quoteChar

    # doExport
    def doExport(self, **keywords):
        self.csvWriter = self.createCSVWriter(**keywords)

        self.writeFileHDR(**keywords),
        self.csvWriter.writerow('')  # empty row between HDR Body (Header & Elements)

        self.writeListHeader(**keywords)
        self.counter = 0
        self.writeEntriesAccordingScope(**keywords)

        self.csvWriter.writerow('')  # empty row between Body (Header & Elements) and OPT
        self.writeFileOPT(**keywords)

    # File Header
    def dateAsSerialNumber1900(self, dateAsString):
        dateAsObject = datetime(2000, 1, 1)

        if self.DELIVERY_DATE_FORMAT and self.DELIVERY_DATE_FORMAT != 'none':
            try:
                dateAsObject = datetime.strptime(dateAsString, self.DELIVERY_DATE_FORMAT)
            except ValueError:
                messageCantConvertString = "Can't convert String 'Delivery Date' to Datetime Object, because format doesn't fit the cuttinglist-configuration - using 2000-01-01 instead."
                if self.DELIVERY_DATE_FORMAT == '%d. %b. %Y':
                    try:
                        dateAsObject = datetime.strptime(dateAsString, '%d. %B %Y')  # Fallback for Mai/May, because it won't be shortened with '.'
                    except ValueError:
                        logging.warning(messageCantConvertString)
                else:
                    logging.warning(messageCantConvertString)
            except TypeError:
                # TODO IT-249 Alphanumeric formats %B/%B actually only work with english language, because we don't use locale! Using them return TypeError "attribute of type 'NoneType' is not callable".  --mjk
                logging.warning("Actually alphanumeric formats are only supported in english language! Using 2000-01-01 instead. Please use another format.".format(str(e)))
            except Exception as e:
                logging.warning("Error while converting String 'Delivery Date' to Datetime Object: {}. Using 2000-01-01 instead. Please try another format.".format(str(e)))
        else:
            logging.warning("Can't convert String 'Delivery Date' to Datetime Object, because 'Delivery Date Format' isn't set in the cuttinglist-configuration - using 2000-01-01 instead.")

        dateZero = datetime(1899, 12, 30)  # Note, not 31st Dec but 30th!
        delta = dateAsObject - dateZero

        return delta.days

    def getFileHDRLine(self, **keywords):
        projectInfo = keywords['projectInfo']

        deliveryDateAsString = self.asString(data=projectInfo['DeliveryDate'], **keywords)
        if deliveryDateAsString:
            deliveryDateAsSerialNumber = self.dateAsSerialNumber1900(dateAsString=deliveryDateAsString)
        else:
            deliveryDateAsSerialNumber = '36526'

        return (
            'HDR7',
            '"{filename}"'.format(filename=projectInfo['DocumentName'].replace(',', ''), ),
            '"{deliveryDate}"'.format(deliveryDate=deliveryDateAsSerialNumber),
            '"{deliveryDate}"'.format(deliveryDate=deliveryDateAsSerialNumber),
            '""',
            '""',
            '"  815"',
        )

    def writeFileHDR(self, **keywords):
        fileHeader = self.getFileHDRLine(**keywords)
        rows = (fileHeader,)
        self.writeRows(rows=rows, **keywords)

    # Header
    def getHeaderLine(self, **keywords):
        # basically the function writeEntriesAccordingScope() without writeRows().
        keywords['entries'] = self.generateSortedEntries(**keywords)
        rows = list(filter(lambda entry: entry is not None, self.getUnsortedElements(**keywords)))
        numberOfElements = len(rows)

        return (
            'SAG7',
            numberOfElements,
        )

    # Elements

    # Maestro Ottimo Cut uses rotated dimensions along grain direction
    # 0/1: part without grain / with grain
    def getSimplifiedGrain(self, grain):
        if grain == self.GRAIN_NONE:
            return 1
        else:  # self.GRAIN_LENGTHWISE, self.GRAIN_CROSSWISE:
            return 0

    def combineItemIDWithThickness(self, **keywords):
        entry = keywords['entry']

        return self.removeCommasOfIDs(id=entry.itemID, **keywords) + self.asString(data=entry.boardThickness, **keywords).rstrip('0').rstrip(self.getDecimalSeparator(**keywords))

    def getElements(self, **keywords):
        CODE_LENGTH = 20
        TEXT_LENGTH = 50

        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']

        self.counter += 1

        return (
            # Default fields
            self.asPlusThreeCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asPlusThreeCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asPlusThreeCoordinate(data=entry.boardThickness, **keywords),

            self.getFixedLengthAlphaNumeric(value=self.removeCommasOfIDs(id=entry.itemID, **keywords), length=CODE_LENGTH, **keywords),  # Color
            self.asInt(data=entry.count, **keywords),

            '0',  # Maxsurplus / Max.Überschuss / Überprod

            self.getSimplifiedGrain(entry.grain),
            self.getFixedLengthAlphaNumeric(value=entry.description, length=CODE_LENGTH, **keywords).replace(',', ';'),
            self.getFixedLengthAlphaNumeric(value=entry.name, length=TEXT_LENGTH, **keywords).replace(',', ';'),
            self.getFixedLengthAlphaNumeric(value=self.combineItemIDWithThickness(**keywords), length=TEXT_LENGTH, **keywords),

            '0',
            '0',
            self.asInt(data=self.counter, **keywords),
            '0',
            '0',
            '0',
            '0',
            '0',
            '0',
            '0',
            '0',
            '0.000000',
            '0',

            self.getFixedLengthAlphaNumeric(value=self.removeCommasOfIDs(id=entry.edgesGrainDirection.left.supplierID, **keywords), length=CODE_LENGTH, **keywords),
            self.getFixedLengthAlphaNumeric(value=self.removeCommasOfIDs(id=entry.edgesGrainDirection.front.supplierID, **keywords), length=CODE_LENGTH, **keywords),
            self.getFixedLengthAlphaNumeric(value=self.removeCommasOfIDs(id=entry.edgesGrainDirection.right.supplierID, **keywords), length=CODE_LENGTH, **keywords),
            self.getFixedLengthAlphaNumeric(value=self.removeCommasOfIDs(id=entry.edgesGrainDirection.back.supplierID, **keywords), length=CODE_LENGTH, **keywords),

            '0.000000',
            '0.000000',
            '0',
            '0',
            '0.000000',
            '0.000000',

            self.getFixedLengthAlphaNumeric(value=self.asStringSingleLine(key='part_comment', **keywords), length=TEXT_LENGTH, **keywords).replace(',', ';'),

            # Optional Fields. Will be imported as additional Info fields
            '18',  # Total number of used additional fields.

            self.quoteElement(element=self.asString(data=projectInfo['ProjectID'], **keywords)),
            self.quoteElement(element=self.asString(data=projectInfo['ProjectCommission'], **keywords)),
            self.quoteElement(element=self.asString(data=projectInfo['ProjectConstruction'], **keywords)),

            self.quoteElement(element=self.asString(data=positionInfo['PositionNumber'], **keywords)),
            self.quoteElement(element=self.asString(data=positionInfo['PositionDescription'], **keywords)),

            self.quoteElement(element=self.asString(data=positionInfo['CabinetName'], **keywords)),

            self.quoteElement(element=self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords)),
            self.quoteElement(element=self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords)),
            self.quoteElement(element=self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords)),
            self.quoteElement(element=self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords)),

            self.quoteElement(element=self.removeCommasOfIDs(id=self.asString(key='part_covering1', **keywords), **keywords)),
            self.quoteElement(element=self.removeCommasOfIDs(id=self.asString(key='part_covering2', **keywords), **keywords)),

            self.quoteElement(element=self.removeCommasOfIDs(id=self.asString(key='part_laminate1', **keywords), **keywords)),
            self.quoteElement(element=self.removeCommasOfIDs(id=self.asString(key='part_laminate2', **keywords), **keywords)),
            self.quoteElement(element=self.removeCommasOfIDs(id=self.asString(key='part_laminate3', **keywords), **keywords)),

            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords)),
            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords)),
            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)),
        )

    # File Footer
    def writeFileOPTLine(self, **keywords):
        return (
            'OPT7',
            '0',
        )

    def writeFileOPT(self, **keywords):
        fileFooter = self.writeFileOPTLine(**keywords)
        rows = (fileFooter,)
        self.writeRows(rows=rows, **keywords)
