# -*- coding: utf-8 -*-

# Copyright 2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase


class MagisOptiExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'MagisOpti'

    OUTPUT_ENCODING = 'utf-8-sig'

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.EXTENSION = 'csv'

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')  # | partsandcoverings | coveringsonly
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

        self.LENGTH_FOLLOWS_GRAIN_DIRECTION = self.getCuttinglistConfiguration().getAsBool('length-follows-grain-direction', True)
        self.MAGISOPTI_DESCRIPTION = self.getCuttinglistConfiguration().get('magisopti-description', 'name')  # Alternative: description

    def getDecimalSeparator(self, **keywords):
        return '.'

    # Header
    def getHeaderLine(self, **keywords):
        return None

    # Elements
    # Magis Opti can handle dimensions as defined in interiorcad with separate parameter for grain crosswise and alternatively length following grain direction
    def getSimplifiedGrain(self, grain):  # For GrainDirection with switched length and width and rotated edges (length follows grain direction)
        if grain == self.GRAIN_NONE:
            return 'None'
        else:  # self.GRAIN_LENGTHWISE (self.GRAIN_CROSSWISE actually impossible, because using LengthGrainDirection).
            return 'LeftRight'

    def getGrain(self, grain):  # For Length/Width as defined in interiorcad, dimensions and edges stay the same.
        if grain == self.GRAIN_LENGTHWISE:
            return 'LeftRight'
        elif grain == self.GRAIN_CROSSWISE:
            return 'UpDown'
        else:  # grain == self.GRAIN_NONE
            return 'None'

    def entryDependingLengthFollowsGrainDirection(self, trueValue, falseValue):
        if self.LENGTH_FOLLOWS_GRAIN_DIRECTION:
            return trueValue
        else:
            return falseValue

    def getMagisOptiName(self, **keywords):
        positionInfo = keywords['positionInfo']

        positionNumber = self.asString(data=positionInfo['PositionNumber'], **keywords)
        cabinetName = self.asString(data=positionInfo['CabinetName'], **keywords)

        return joinIfNotNone('_',
            positionNumber if positionNumber else None,
            cabinetName if cabinetName else None,
        )

    def getMagisOptiDescriptionEntry(self, **keywords):
        entry = keywords['entry']

        if 'description' in self.MAGISOPTI_DESCRIPTION:
            return entry.description
        else:  # name
            return entry.name

    def getMagisOptiDescription(self, **keywords):
        magisOptiDescription = self.asString(data=self.getMagisOptiDescriptionEntry(**keywords), **keywords)
        comment = self.asStringSingleLine(key='part_comment', **keywords)

        if comment and 'comment' in self.MAGISOPTI_DESCRIPTION:
            return magisOptiDescription + ' ({comment})'.format(comment=comment)
        else:
            return magisOptiDescription

    def getElements(self, **keywords):
        entry = keywords['entry']

        return (
            # Defaults
            self.asString(data=entry.itemID, **keywords),

            self.asInt(data=entry.count, **keywords),

            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.boardLengthGrainDirection, entry.boardLength), **keywords),
            self.asCoordinate(data=self.entryDependingLengthFollowsGrainDirection(entry.boardWidthGrainDirection, entry.boardWidth), **keywords),
            self.asCoordinate(data=entry.boardThickness, **keywords),

            self.entryDependingLengthFollowsGrainDirection(self.getSimplifiedGrain(entry.grain), self.getGrain(entry.grain)),

            self.getMagisOptiName(**keywords),
            self.getMagisOptiDescription(**keywords),

            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.left.supplierID, entry.edges.left.supplierID), **keywords),
            '0',
            self.asCoordinate(data=-self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.left.thickness, entry.edges.left.thickness), **keywords),

            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.back.supplierID, entry.edges.back.supplierID), **keywords),  # oben
            '0',
            self.asCoordinate(data=-self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.back.thickness, entry.edges.back.thickness), **keywords),

            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.right.supplierID, entry.edges.right.supplierID), **keywords),
            '0',
            self.asCoordinate(data=-self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.right.thickness, entry.edges.right.thickness), **keywords),

            self.asString(data=self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.front.supplierID, entry.edges.front.supplierID), **keywords),  # unten
            '0',
            self.asCoordinate(data=-self.entryDependingLengthFollowsGrainDirection(entry.edgesGrainDirection.front.thickness, entry.edges.front.thickness), **keywords),
        )
