# -*- coding: utf-8 -*-

# Copyright 2015-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import logging
import os.path

from interiorcad.backend.dosfilenamecreator import DosFilenameCreator
from interiorcad.backend.localization import getLocalizedString, CUTTINGLISTEXPORT_MESSAGES
from interiorcad.backend.workpiecefilters.emptyworkpiecefilter import EmptyWorkpieceFilter
from interiorcad.cuttinglistexport.base import CuttinglistExportBase


# Note: the comments "# EF:" are to document what the 3 OSD-EF-Exports did in comparison to this new export -- NG


class OSDExport(CuttinglistExportBase):
    EXPORT_FORMAT = 'OSD'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.counter = 0

        self.EXTENSION = 'stk'
        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA
        # Sorting Criteria 'cabinetname' is not part of getPositionDetailLine and getWorkpieceLines!

        cuttinglistConfiguration = self.getCuttinglistConfiguration()
        self.SCOPE_OF_EXPORT = cuttinglistConfiguration.get('scope-of-export', 'partsandfittings')  # | partsonly | fittingsonly
        self.CUMULATE_FITTINGS = cuttinglistConfiguration.getAsBool('cumulate-fittings', True)
        self.CUMULATE_WORKPIECES = cuttinglistConfiguration.get('cumulate-workpieces', 'none')  # only minimal acccumulation
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = cuttinglistConfiguration.get('nc-program-display-name-definition', 'pathandnamewithextension')

        # The Floats 5.3 and 9.1 refer to the program version. Additionally, based on version 9. 1, there are the function ranges S, XL and XXL, which interiorcad currently does not support.
        self.VERSION = cuttinglistConfiguration.get('version', 'XXL')  # '' | 5.3 | 9.1 | S | XL | XXL
        self.OSD_TEILEBEZ = cuttinglistConfiguration.get('osd-teilebez', 'name')  # alternativ: description
        self.OSD_BEARBEITUNGSHINWEIS = cuttinglistConfiguration.get('osd-bearbeitungshinweis', 'version-dependend-default')
        self.OSD_FREE_FIELD_FOR_CABINET_NAME = cuttinglistConfiguration.get('osd-free-field-for-cabinet-name', '')

    def isVersionOlderThanV9(self, **keywords):
        return self.VERSION == '' or self.VERSION == '5.3'

    VERSION_XXL = ('XXL', )
    VERSION_XL = (*VERSION_XXL, 'XL')
    VERSION_S = (*VERSION_XL, 'S')

    def selfVersionMeetsOSDRequiredVersion(self, requiredVersion):
        return self.VERSION in requiredVersion

    def asFivePlusOneCoordinate(self, **keywords):
        value = self.getDataAsFloat(**keywords)
        return '{:<5.1f}'.format(value)

    # doExport
    def doExport(self, **keywords):
        self.writeHeader(**keywords)
        self.writeBody(**keywords)

    # Header
    def writeHeader(self, **keywords):
        lines = itertools.chain(
            self.getVersionLines(**keywords),
            self.getProjectDetailLines(**keywords),
        )
        self.writeLines(lines=lines, **keywords)

    def getVersionLines(self, **keywords):
        # EF: Only in XG_NCEF_OSD.ef (5.3, needed for NC fields) and OSD-Klassik.ef (5.5) -- NG
        return (
            '//Version//',
            '#version',
            '{version}'.format(version=self.VERSION),
            '',
        )

    def getProjectDetailLinesV5(self, **keywords):
        projectInfo = keywords['projectInfo']

        return (
            '#{fieldname}'.format(fieldname='kunde'),
            self.asString(data=projectInfo['ProjectCommission'], **keywords),  # EF:STIAuftragKommission
            '#{fieldname}'.format(fieldname='projektbeschreibung1'),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),  # EF: STIAuftragBauvorhaben
        )

    def getProjectDetailLinesV9(self, **keywords):
        projectInfo = keywords['projectInfo']

        return list(filter(lambda line: line is not None, (
            '#{fieldname}'.format(fieldname='kommission') if self.selfVersionMeetsOSDRequiredVersion(requiredVersion=self.VERSION_XL) else None,
            self.asString(data=projectInfo['ProjectCommission'], **keywords) if self.selfVersionMeetsOSDRequiredVersion(requiredVersion=self.VERSION_XL) else None,  # EF:STIAuftragKommission
            '#{fieldname}'.format(fieldname='projektbeschreibung'),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),  # EF: STIAuftragBauvorhaben
        )))

    def getVersionDependingProjectDetailLines(self, **keywords):
        if self.isVersionOlderThanV9(**keywords):
            return self.getProjectDetailLinesV5(**keywords)
        else:
            return self.getProjectDetailLinesV9(**keywords)

    def getProjectDetailLines(self, **keywords):
        projectInfo = keywords['projectInfo']

        return itertools.chain(
            ('//Projektdaten//',
             '#projektnr',
             self.asString(data=projectInfo['ProjectID'], **keywords),  # EF: STIAuftragNummer
             ),
            self.getVersionDependingProjectDetailLines(**keywords),
            ('',),
        )

    # Body
    def writeBody(self, **keywords):
        def getPositionNumber(entry):
            return self.asString(data=entry.positionInfo['PositionNumber'], **keywords)

        entries = self.generateSortedEntries(**keywords)

        entriesByPosition = itertools.groupby(sorted(entries, key=getPositionNumber), key=getPositionNumber)
        for positionNumber, entries in entriesByPosition:
            self.writeLines(lines=self.getPositionDetailLines(positionNumber=positionNumber, **keywords), **keywords)
            for entry in entries:
                self.writeLines(lines=self.getWorkpieceLines(entry=entry, workpiece=entry.workpiece, **keywords), **keywords)
                self.writeLines(lines=self.getExportSpecificEdgeCode(entry=entry, workpiece=entry.workpiece, **keywords), **keywords) if self.selfVersionMeetsOSDRequiredVersion(requiredVersion=self.VERSION_XXL) else None
                self.writeLines(lines=self.getWorkpieceFreeFieldLines(entry=entry, **keywords), **keywords) if self.OSD_FREE_FIELD_FOR_CABINET_NAME else None
                self.writeLines(lines=self.getWorkpieceNCProgramLines(entry=entry, **keywords), **keywords)

    def getPositionDetailLines(self, **keywords):
        positionNumber = keywords['positionNumber']
        positionDescription, positionQuantity = self.getInfoForPositionNumber(positionNumber=positionNumber)
        return (
            '//Positionsdaten//',
            '#position',
            self.asString(data=positionNumber, **keywords),  # EF: STIPositionNummer
            '#{quantityName}'.format(quantityName='anzahl' if self.isVersionOlderThanV9() else 'menge'),
            self.asInt(data=float(positionQuantity), **keywords) if positionQuantity else '1',  # EF: Hard-coded to 1
            '#{positionDescriptionName}'.format(positionDescriptionName='positionsbeschreibung1' if self.isVersionOlderThanV9() else 'positionsbeschreibung'),
            self.asString(data=positionDescription, **keywords),  # EF: STIPositionName
            ''
            '',
        )

    def getOSDTeilebez(self, **keywords):
        entry = keywords['entry']

        if self.OSD_TEILEBEZ == 'description':
            return entry.description
        elif self.OSD_TEILEBEZ == 'shortname':
            return DosFilenameCreator().createDosFilename(entry.name)
        elif self.OSD_TEILEBEZ == 'shortdescription':
            return DosFilenameCreator().createDosFilename(entry.description)
        else:  # name
            return entry.name

    def getOSDBearbeitungshinweis(self, **keywords):  # EF: STIBemerkung | Bearbeitungshinweis
        cabinetName = self.asString(data=keywords['entry'].positionInfo['CabinetName'], **keywords)
        comment = self.asString(key='part_comment', **keywords)

        # bhinweis1: 20 characters in v5 as well as v9
        # bhinweis2: 20 characters in v5 and 2^15 characters in v9
        # bearbeitungshinweis: doesn't exist in v5(!) and 2^15 characters in v9
        if self.OSD_BEARBEITUNGSHINWEIS == 'comment-as-bhinweis1':
            return (
                '#bhinweis1',
                self.asString(key='part_comment', **keywords),
            )
        elif self.OSD_BEARBEITUNGSHINWEIS == 'cabinetname-as-bhinweis1-and-comment-as-bhinweis2':
            return (
                '#bhinweis1',
                cabinetName,
                '#bhinweis2',
                comment,
            )
        elif self.OSD_BEARBEITUNGSHINWEIS == 'comment-as-bearbeitungshinweis':
            return (
                '#bearbeitungshinweis',
                comment,
            )
        elif self.OSD_BEARBEITUNGSHINWEIS == 'cabinetname-and-comment-as-bearbeitungshinweis':
            return (
                '#bearbeitungshinweis',
                ' / '.join(
                    filter(lambda name: bool(name), (
                        cabinetName,
                        comment,
                    ))
                ),
            )
        else:  # 'version-dependend-default'
            return (
                '#{hinweis}'.format(hinweis='bhinweis1' if self.isVersionOlderThanV9(**keywords) else 'bearbeitungshinweis'),
                comment,
            )

    def getAdditionsForFinishedSizePartDimensionsWithAdditionGrainDirection(self, **keywords):
        entry = keywords['entry']
        workpiece = keywords['workpiece']

        # Actually version XXL is required. So far, only the version is supported, not the function range.
        if entry.isWorkpiece() and self.VERSION == 'XXL' and (EmptyWorkpieceFilter.hasSizing(self, workpiece) or self.hasCovering(**keywords)):
            return (
                '#rmz_hoehe',
                self.asFivePlusOneCoordinate(data=entry.finishedSizePartDimensionsWithAdditionGrainDirection.getLength() - entry.boardLengthGrainDirection, **keywords),
                '#rmz_breite',
                self.asFivePlusOneCoordinate(data=entry.finishedSizePartDimensionsWithAdditionGrainDirection.getWidth() - entry.boardWidthGrainDirection, **keywords),
                )
        else:
            return ''

    '''
    Definition OSD - interiorcad for "3-Maß-Übergabe" / "3-dimension-transfer"
    Bauteil/part
        Höhe/Height - length
        Breite/Width - width
        Tiefe/Depth - thickness
    Kanten/edges
        links/left - front 
        rechts/right - back
        oben/top - left
        unten/bottom - right
    Furnier/covering & Oberfläche/laminate
        1 vorne/front - 1
        2 hinten/back - 2
    '''

    def getWorkpieceLines(self, **keywords):
        self.counter += 1
        entry = keywords['entry']

        return itertools.chain(
            ('//Teiledaten//',
            '#lfdnr',
            self.asInt(data=self.counter, **keywords),  # EF: LaufendeNummer
            '#teilebez',
            self.asString(data=self.getOSDTeilebez(**keywords), **keywords),  # EF: STIName
            '#materialnr',
            self.asString(data=entry.itemID, **keywords),  # EF: STIArtikelNr
            '#stueckgesamt',
            self.asInt(data=entry.count, **keywords),  # EF: STIAnzahl
            '#hoehefertig',
            self.asFivePlusOneCoordinate(data=entry.boardLengthGrainDirection, **keywords),  # EF: STILaenge
            '#breitefertig',
            self.asFivePlusOneCoordinate(data=entry.boardWidthGrainDirection, **keywords),  # EF: STIBreite
            '#tiefefertig',
            self.asFivePlusOneCoordinate(data=entry.boardThickness, **keywords),),  # EF: STIHoehe
            self.getOSDBearbeitungshinweis(**keywords),  # EF: STIBemerkung
            self.getAdditionsForFinishedSizePartDimensionsWithAdditionGrainDirection(**keywords),
            # EF: Covering, Laminate and Edge are "optional" in OSD-Klassik.ef -- NG
            ('#tfurniermat1',
            self.asString(key='part_covering1', **keywords),  # EF: STIBelag1
            '#tfurniermat2',
            self.asString(key='part_covering2', **keywords),  # EF: STIBelag2
            '#tlackmat1',
            self.asString(key='part_laminate1', **keywords),  # EF: STIBelag1
            '#tlackmat2',
            self.asString(key='part_laminate2', **keywords),  # EF: STIBelag2
            '#tkantematlinks',
            self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),  # EF: STIKante1
            '#tkantematrechts',
            self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),  # EF: STIKante2
            '#tkantematoben',
            self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),  # EF: STIKante3
            '#tkantematunten',
            self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),  # EF: STIKante4
            '',),
        )

    # Export specific edge codes
    def getExportSpecificEdgeCodeCoveringCounter(self, **keywords):
        if self.hasCovering(**keywords):
            return (
                '//Kantenskizze-Furnier//',
                '#bearbeitungsreihenfolge',
                '1',  # 1 for coverings first, then edges; 2 for edges first, then coverings. Not yet implemented in interiorcad.
                '#furnierrichtung',
                '1',  # grain is always in length, because edgesGrainDirection is used.
                '',
            )
        else:
            return ''

    def getExportSpecificEdgeCodeCounter(self, **keywords):
        edge = keywords['edge']
        identifier = keywords['identifier']
        edgecode = keywords['entry'].edgesGrainDirection.edgecode

        if edge and identifier in edgecode:
            return (
                '//Kantenskizze-Kante{name}//'.format(name=keywords['name']),
                '#bearbeitungsreihenfolge',
                '{counter}'.format(counter=str(int(self.hasCovering(**keywords))+edgecode.find(identifier)+1)),
                '#gehrung_vorgaenger',
                '{miter:d}'.format(miter=(edge.jointEnd == 'Miter')),
                '#gehrung_nachfolger',
                '{miter:d}'.format(miter=(edge.jointStart == 'Miter')),
                '',
            )
        else:
            return ''

    def getExportSpecificEdgeCode(self, **keywords):
        entry = keywords['entry']

        return itertools.chain(
            self.getExportSpecificEdgeCodeCoveringCounter(**keywords),
            self.getExportSpecificEdgeCodeCounter(edge=entry.edgesGrainDirection.front, identifier='1', name='links', **keywords),
            self.getExportSpecificEdgeCodeCounter(edge=entry.edgesGrainDirection.back, identifier='2', name='rechts', **keywords),
            self.getExportSpecificEdgeCodeCounter(edge=entry.edgesGrainDirection.left, identifier='3', name='oben', **keywords),
            self.getExportSpecificEdgeCodeCounter(edge=entry.edgesGrainDirection.right, identifier='4', name='unten', **keywords),
        )

    # Free fields
    def getWorkpieceFreeFieldLines(self, **keywords):
        return (
            '//Freie-Felder//',
            '#{name}'.format(name=self.OSD_FREE_FIELD_FOR_CABINET_NAME),
            self.asString(data=keywords['entry'].positionInfo['CabinetName'], **keywords),
            '',
        )

    # NC program lines
    def getWorkpieceNCProgramLines(self, **keywords):
        if self.isVersionOlderThanV9(**keywords):
            return self.getWorkpieceNCProgramLinesV5(**keywords)
        else:
            return self.getWorkpieceNCProgramLinesV9(**keywords)

    def getWorkpieceNCProgramLinesV5(self, **keywords):
        # EF: NC programs were only in XG_NCEF_OSD.ef (in OSD-Klassic.ef there are exported in normal fields instead of the expected information -- NG
        entry = keywords['entry']

        return itertools.chain(
            ('//NC-Destination//',),
            # Separate sizing is not supported by OSD version 5.
            self.getNCExportLinesV5(osdPrefix='', path=entry.ncProgramPathMain, **keywords),
            self.getNCExportLinesV5(osdPrefix='ges_', path=entry.ncProgramPathOpposite, **keywords),
            (
                '#optimierung',
                self.asInt(data=1, **keywords),
                '',
            ),
        )

    def getNCExportLinesV5(self, **keywords):
        # OSD splits the path to the NC program file in a Windows specific way (especially the drive).
        # For a Mac, it means that the drive letter will be empty and path separated by / (instead of Windows \ )
        # Similar for Windows UNC paths, the drive will be the machine "\\example", not a letter "C:". (OSD ^might not like it either.)
        path = keywords['path']
        if isinstance(path, list):
            if len(path) > 1:
                message = getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 6, "Cumulated NC program paths are not allowed. They are reduced to the first entry. Workpiece: '{name}'.")
                logging.warning(message.format(name=keywords['entry'].workpiece.name))
            path = path[0]

        drive, purePath, filename, extension = self.splitPath(path=path)
        return (
            '#{prefix}laufwerk'.format(prefix=keywords['osdPrefix']),
            self.asString(data=drive),  # NC/EF: CNCFILEDRIVE or SECONDCNCFILEDRIVE
            '#{prefix}verzeichnis'.format(prefix=keywords['osdPrefix']),
            self.asString(data=purePath),  # NC/EF: CNCFILEPATH or SECONDCNCFILEPATH
            '#{prefix}ncprogramm'.format(prefix=keywords['osdPrefix']),
            self.asString(data=filename),  # NC/EF:CNCFILE or SECONDCNCFILE
            '#{prefix}ncprogrammext'.format(prefix=keywords['osdPrefix']),
            self.asString(data=extension),  # NC/EF: CNCPUREFILEEXTENSIONONLY or SECONDCNCPUREFILEEXTENSIONONLY
        )

    def splitPath(self, **keywords):
        path = keywords['path']
        filename, extension = os.path.splitext(os.path.basename(path))
        # drive can be empty or a UNC. Probably OSD might not like it -- NG
        drive, purePath = os.path.splitdrive(os.path.dirname(path))
        return drive, purePath, filename, extension[1:]

    def getWorkpieceNCProgramLinesV9(self, **keywords):
        entry = keywords['entry']

        return (
            '//NC-Destination//',
            '#ro_ncprogramm',
            self.asString(data=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords), **keywords),
            '#bo_ncprogramm',
            self.asString(data=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords), **keywords),
            '#bu_ncprogramm',
            self.asString(data=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)),
            '',
        )
