# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import logging

from interiorcad.backend.joint.calculator import JointCalculator
from interiorcad.backend.localization import getLocalizedString, CUTTINGLISTEXPORT_MESSAGES, CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP
from interiorcad.backend.workpiece.reader import WorkpieceReader
from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.interface.geometry.edges import EdgeData


class PiosProfessionalExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = getLocalizedString(CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP, 11, 'PIOS')

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = ';'
        self.counter = 0

        self.EXTENSION = 'csv'
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

        self.INDIVIDUAL_EXPORT_FORMAT = self.getCuttinglistConfiguration().get('individual-export-format', '')

        self.checkProjectID(**keywords)

    def getDecimalSeparator(self, **keywords):
        return '.'

    def getOutputFileNameVariables(self, **keywords):
        outputFileNameVariables = super().getOutputFileNameVariables(**keywords)
        if self.INDIVIDUAL_EXPORT_FORMAT:
            outputFileNameVariables['ExportFormat'] = self.INDIVIDUAL_EXPORT_FORMAT

        return outputFileNameVariables

    def checkProjectID(self, **_keywords):
        projectInfo = self.getProjectInfo()
        if not projectInfo['ProjectID']:
            logging.warning(getLocalizedString(CUTTINGLISTEXPORT_MESSAGES, 5, "Mandatory Project ID not entered in Project Details, exporting Document Name as Project ID instead."))

    # Header
    def getHeaderLine(self, **_keywords):
        return (
            'Aufkb',
            'Plakb',
            'Elnr',
            'Aufpos',
            'Stuck',
            'ZusLange',
            'ZusBreite',
            'Drehbar',
            'Teilbez',
            'Zusinfo',

            'TInfo1',
            'TInfo2',
            'TInfo3',
            'TInfo4',
            'TInfo5',
            'TInfo6',
            'TInfo7',
            'TInfo8',
            'TInfo9',

            'Belagikb',
            'Belagakb',

            'KantevKb',
            'KantehKb',
            'KantelKb',
            'KanterKb',
            'Kantevdicke',
            'Kantehdicke',
            'Kanteldicke',
            'Kanterdicke',
            'KantevlKb',
            'KantevrKb',
            'KantehlKb',
            'KantehrKb',
            'Kantevldicke',
            'Kantevrdicke',
            'Kantehldicke',
            'Kantehrdicke',
            'KanteAusbVL',
            'KanteAusbVR',
            'KanteAusbHR',
            'KanteAusbHL',

            'Gehrung1L',
            'Gehrung1Q',
            'Gehrung2L',
            'Gehrung2Q',
            'Gehrung3L',
            'Gehrung3Q',
            'Gehrung4L',
            'Gehrung4Q',
            'GWvorne',
            'GWhinten',
            'GWlinks',
            'GWrechts',
            'GWvoli',
            'GWhili',
            'GWhire',
            'GWvore',

            'KantevFuge',
            'KantehFuge',
            'KantelFuge',
            'KanterFuge',

            'Kantevsaum',
            'Kantehsaum',
            'Kantelsaum',
            'Kantersaum',
        )

    # Elements
    def getMandatoryProjectID(self, **keywords):
        projectInfo = keywords['projectInfo']
        projectID = projectInfo['ProjectID']

        if projectID:
            return projectID
        else:
            return projectInfo['DocumentName']

    def getEdgeJointingsBTLR(self, **keywords):
        if self.configuration.jointing:
            additionForEdges = WorkpieceReader(keywords['entry'].workpiece).getFinishedSizeAddition()
            jointSizeAddition = JointCalculator(self.configuration.minJoint, self.configuration.maxJoint).getJointAddition(additionForEdges)
            return jointSizeAddition.bottom, jointSizeAddition.top, jointSizeAddition.left, jointSizeAddition.right
        else:
            return 0, 0, 0, 0

    def getSimplifiedGrain(self, grain):
        # 0= part must not be rotated, 1= part can be rotated
        if grain == self.GRAIN_NONE:
            return 1
        else:  # self.GRAIN_LENGTHWISE, self.GRAIN_CROSSWISE:
            return 0

    def getJointAsNumberPerCorner(self, lhs, rhs):
        if lhs == EdgeData.JOINT_DISCONTINUOUS and rhs == EdgeData.JOINT_CONTINUOUS:
            return '1'
        elif lhs == EdgeData.JOINT_CONTINUOUS and rhs == EdgeData.JOINT_DISCONTINUOUS:
            return '2'
        elif lhs == EdgeData.JOINT_DISCONTINUOUS and rhs == EdgeData.JOINT_NOEDGE:
            return '3'
        elif lhs == EdgeData.JOINT_NOEDGE and rhs == EdgeData.JOINT_DISCONTINUOUS:
            return '4'
        elif lhs == rhs == EdgeData.JOINT_MITER:
            return '5'
        elif lhs == rhs == EdgeData.JOINT_NOEDGE:
            return '6'
        else:
            logging.info(f"Unsupported edge jointing case: {lhs} - {rhs}. Using 6 (no edge) instead.")
            return '6'

    def getElements(self, **keywords):
        entry = keywords['entry']
        positionInfo = keywords['positionInfo']
        projectInfo = keywords['projectInfo']

        self.counter += 1
        egdeBottomJointing, edgeTopJointing, edgeLeftJointing, edgeRightJointing = self.getEdgeJointingsBTLR(**keywords)

        return (
            self.asString(data=self.getMandatoryProjectID(**keywords), **keywords),
            self.asString(data=entry.itemID, **keywords),
            self.asInt(data=self.counter, **keywords),
            self.asString(data=positionInfo['PositionNumber'], **keywords),
            self.asInt(data=entry.count, **keywords),
            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asInt(data=self.getSimplifiedGrain(entry.grain)),  # 0 nicht drehbar, 1 drehbar
            self.asString(data=entry.name, **keywords),
            self.asStringSingleLine(key='part_comment', **keywords),

            # TInfo
            self.asString(data=positionInfo['PositionDescription'], **keywords),
            self.asString(data=entry.description),
            self.asString(data=positionInfo['CabinetName'], **keywords),
            self.asString(data=projectInfo['ProjectCommission'], **keywords),
            self.asString(data=projectInfo['ProjectConstruction'], **keywords),
            self.asString(data=projectInfo['Company'], **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),

            self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),
            '',
            '',
            '',
            '',
            '0',
            '0',
            '0',
            '0',
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.front.jointEndTechnical, rhs=entry.edgesGrainDirection.left.jointStartTechnical),  # KanteAusbVL
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.front.jointStartTechnical, rhs=entry.edgesGrainDirection.right.jointEndTechnical),  # KanteAusbVR
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.back.jointEndTechnical, rhs=entry.edgesGrainDirection.right.jointStartTechnical),  # KanteAusbHR
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.back.jointStartTechnical, rhs=entry.edgesGrainDirection.left.jointEndTechnical),  # KanteAusbHL

            '0',  # GehrungDXVL
            '0',  # GehrungDYVL
            '0',  # GehrungDXHL
            '0',  # GehrungDYHL
            '0',  # GehrungDXHR
            '0',  # GehrungDYHR
            '0',  # GehrungDXVR
            '0',  # GehrungDYVR
            '0',  # GehrungsWinkelV
            '0',  # GehrungsWinkelH
            '0',  # GehrungsWinkelL
            '0',  # GehrungsWinkelR
            '0',  # GehrungsWinkelVL
            '0',  # GehrungsWinkelHL
            '0',  # GehrungsWinkelHR
            '0',  # GehrungsWinkelVR

            self.formatAsCoordinate(data=egdeBottomJointing, **keywords),
            self.formatAsCoordinate(data=edgeTopJointing, **keywords),
            self.formatAsCoordinate(data=edgeLeftJointing, **keywords),
            self.formatAsCoordinate(data=edgeRightJointing, **keywords),

            self.formatAsCoordinate(data=self.configuration.sizingAddition) if self.shouldCreateSizing(**keywords) else '0',
            self.formatAsCoordinate(data=self.configuration.sizingAddition) if self.shouldCreateSizing(**keywords) else '0',
            self.formatAsCoordinate(data=self.configuration.sizingAddition) if self.shouldCreateSizing(**keywords) else '0',
            self.formatAsCoordinate(data=self.configuration.sizingAddition) if self.shouldCreateSizing(**keywords) else '0',
        )
