# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import logging

from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.cuttinglistexport.base import CuttinglistExportBase
from interiorcad.interface import configuration
from interiorcad.interface.geometry.edges import EdgeData


class SchellingHPOExport(CuttinglistExportBase):
    EXPORT_FORMAT = 'SchellingHPO'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.counter = 0
        self.createDates()

        self.EXTENSION = 'stk'
        self.CONFIGURED_SORTING_CRITERIA = self.cuttinglistConfiguration.get('sorting-criteria', self.getDefaultSortingCriteria(**keywords) if hasattr(self, 'getDefaultSortingCriteria') else self.POSITION_SORTING_CRITERIA)

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')  # | partsandfittings | fittingsonly
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

    def createDates(self):
        self.creationDate = configuration.DATETIME_NOW.date()
        # projectInfo = keywords['projectInfo']
        # deliveryDateRaw = projectInfo['DeliveryDate']
        self.deliveryDate = self.creationDate

    def doExport(self, **keywords):
        self.writeHeader(**keywords)
        self.writeBody(**keywords)
        self.writeFooter(**keywords)

    def getFixedLengthAlphaNumeric(self, **keywords):
        value = (keywords['value'])[:keywords['length']]  # HACK: to avoid a C Assert in Python (Debug-Mode) (IC-11640)
        return '{value:<{length}}'.format(value=value, length=keywords['length'])[:keywords['length']]

    def getFixedLengthInteger(self, **keywords):
        return '{value:0>{length}d}'.format(value=keywords['value'], length=keywords['length'])[:keywords['length']]

    def getFixedLengthCoordinate(self, **keywords):
        # Careful: Schelling HPO works with micrometers, not millimeters.
        return '{value:0>{length}.0f}'.format(value=keywords['value'] * 1000, length=keywords['length'])[:keywords['length']]

    # HeaderLines
    def writeHeader(self, **keywords):
        self.writeLines(lines=self.getHeaderLines(), **keywords)

    def getHeaderBaseLine(self, **keywords):
        TEXT_LENGTH = 20

        projectInfo = self.getProjectInfo()
        if self.isProjectIDValid(**projectInfo):
            cuttingListName = projectInfo['ProjectID']
            description = joinIfNotNone(' ',
                projectInfo['ProjectCommission'],
                projectInfo['ProjectConstruction']
            )
        else:
            cuttingListName = self.getOutputFileName()
            description = ''

        line = ' '.join((
            'L 01',
            self.getFixedLengthAlphaNumeric(value=cuttingListName, length=TEXT_LENGTH, **keywords),
            self.getFixedLengthAlphaNumeric(value=description, length=TEXT_LENGTH, **keywords),
            self.deliveryDate.isoformat(),
            self.creationDate.isoformat(),
            '0',
            '0',
            '0',
        ))

        assert len(line) == 74, 'L 01 line has not the fixed length of 74 characters'
        return line

    def getHeaderDatabaseLine(self, **keywords):
        line = ' '.join((
            'L 06',
            self.getFixedLengthAlphaNumeric(value='', length=20, **keywords),
            '1',
            '1',
            '0',
            '0',
            '0',
            '0',
            '0'
        ))
        assert len(line) == 39, 'L 06 line has not the fixed length of 39 characters'
        return line

    def getHeaderLines(self, **keywords):
        return (
            self.getHeaderBaseLine(**keywords),
            self.getHeaderDatabaseLine(**keywords),
        )

    # DatabaseLines
    def writeBody(self, **keywords):
        entries = self.generateSortedEntries(**keywords)
        lines = (self.getWorkpieceLines(entry=entry, workpiece=entry.workpiece, **keywords) for entry in entries)
        self.writeLines(lines=lines, **keywords)

    def getElementBaseLine(self, **keywords):
        self.counter += 1

        entry = keywords['entry']
        boardCount = self.getDataAsInt(data=entry.count, **keywords)
        itemID = self.asString(data=entry.itemID, **keywords)
        comment = self.asStringSingleLine(key='part_comment', **keywords)
        filename = self.getOutputFileName(**keywords)

        line = ' '.join((
            'F 01',
            self.getFixedLengthInteger(value=self.counter, length=4, **keywords),
            self.getFixedLengthInteger(value=0, length=4, **keywords),
            '0',
            self.getFixedLengthInteger(value=boardCount, length=6, **keywords),
            self.getFixedLengthInteger(value=0, length=6, **keywords),
            self.getFixedLengthAlphaNumeric(value=itemID, length=20, **keywords),
            self.getFixedLengthCoordinate(value=entry.boardThickness, length=8, **keywords),
            self.getFixedLengthCoordinate(value=entry.boardLengthGrainDirection, length=8, **keywords),
            self.getFixedLengthCoordinate(value=entry.boardWidthGrainDirection, length=8, **keywords),
            self.getFixedLengthCoordinate(value=entry.finishedSizePartDimensionsGrainDirection.getLength(), length=8, **keywords),
            self.getFixedLengthCoordinate(value=entry.finishedSizePartDimensionsGrainDirection.getWidth(), length=8, **keywords),
            self.getFixedLengthAlphaNumeric(value=entry.name, length=20, **keywords),
            '2',
            '9',
            '9',
            '0',
            '00',
            '000000',
            '000',
            '000000',
            '000',
            '1',
            '0',
            self.getFixedLengthAlphaNumeric(value=entry.positionInfo['PositionNumber'], length=20, **keywords),
            self.getFixedLengthAlphaNumeric(value='', length=20, **keywords),
            self.getFixedLengthAlphaNumeric(value=comment, length=40, **keywords),
            '1',
            '0',
            self.getFixedLengthAlphaNumeric(value='', length=20, **keywords),
            self.getFixedLengthAlphaNumeric(value=filename, length=20, **keywords),
            '000000001',
            self.deliveryDate.isoformat(),
            '0',
            '0',
            '000',
        ))
        assert len(line) == 312, 'F 01 line has not the fixed length of 312 characters'
        return line

    def getCoveringLine(self, **keywords):
        TEXT_LENGTH = 20

        line = ' '.join((
            'F {number:02d}'.format(number=keywords['number']),
            self.getFixedLengthAlphaNumeric(value=keywords['covering'], length=TEXT_LENGTH, **keywords),
            self.getFixedLengthCoordinate(value=keywords['thickness'], length=8, **keywords),
            self.getFixedLengthAlphaNumeric(value='', length=TEXT_LENGTH, **keywords),
            '00000000',
            '00000000',
            '0',
            '0',
        ))
        assert len(line) == 77, 'Covering Line F 07/08 has not the fixed length of 77 characters'
        return line

    def getCustomerTextLine(self, **keywords):
        line = ' '.join((
            'F {number:2d}'.format(number=keywords['number']),
            self.getFixedLengthAlphaNumeric(length=70, **keywords),
        ))
        assert len(line) == 75, 'Customer Line F 10-29 has not the fixed length of 75 characters'
        return line

    def getJointTypeAsNumberPerCorner(self, lhs, rhs):
        if lhs == EdgeData.JOINT_DISCONTINUOUS and rhs == EdgeData.JOINT_CONTINUOUS:
            return '1'
        elif lhs == EdgeData.JOINT_CONTINUOUS and rhs == EdgeData.JOINT_DISCONTINUOUS:
            return '2'
        elif lhs == rhs == EdgeData.JOINT_MITER:
            return '3'
        elif lhs == EdgeData.JOINT_NOEDGE and rhs == EdgeData.JOINT_DISCONTINUOUS:
            return '4'
        elif lhs == EdgeData.JOINT_DISCONTINUOUS and rhs == EdgeData.JOINT_NOEDGE:
            return '5'
        elif lhs == rhs == EdgeData.JOINT_NOEDGE:
            return '0'
        else:
            logging.info(f"Unsupported edge jointing case: {lhs} - {rhs}. Using 0 (without specification) instead.")
            return '0'

    def getJointTransitionAsNumberPerCorner(self):
        return '0'

    def getJointAsNumberPerCorner(self, lhs, rhs):
        return self.getJointTypeAsNumberPerCorner(lhs, rhs) + self.getJointTransitionAsNumberPerCorner()

    def getFormatPostprocessingLine(self, **keywords):
        TEXT_LENGTH = 20

        entry = keywords['entry']

        line = ' '.join((
            'F 80',
            '00000000',
            '00000000',
            '00000000',
            '00000000',
            '0000',
            # edges vorne/hinten/links/rechts
            self.getFixedLengthAlphaNumeric(value=entry.edgesGrainDirection.front.supplierID, length=TEXT_LENGTH, **keywords),
            self.getFixedLengthAlphaNumeric(value=entry.edgesGrainDirection.back.supplierID, length=TEXT_LENGTH, **keywords),
            self.getFixedLengthAlphaNumeric(value=entry.edgesGrainDirection.left.supplierID, length=TEXT_LENGTH, **keywords),
            self.getFixedLengthAlphaNumeric(value=entry.edgesGrainDirection.right.supplierID, length=TEXT_LENGTH, **keywords),
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.front.jointStartTechnical, rhs=entry.edgesGrainDirection.right.jointEndTechnical),  # EckLogoRV
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.front.jointEndTechnical, rhs=entry.edgesGrainDirection.left.jointStartTechnical),  # EckLogoLV
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.back.jointStartTechnical, rhs=entry.edgesGrainDirection.left.jointEndTechnical),  # EckLogoLH
            self.getJointAsNumberPerCorner(lhs=entry.edgesGrainDirection.back.jointEndTechnical, rhs=entry.edgesGrainDirection.right.jointStartTechnical),  # EckLogoRH
            '00000000',
            '00000000',
            self.getFixedLengthAlphaNumeric(value='', length=41, **keywords),
        ))
        assert len(line) == 201, 'F 80 line has not the fixed length of 201 characters'
        return line

    def getWorkpieceLines(self, **keywords):
        entry = keywords['entry']

        covering1 = self.asString(key='part_covering1', **keywords)
        covering2 = self.asString(key='part_covering2', **keywords)
        coveringthickness1 = self.getDataAsFloat(key='part_covering1thickness', **keywords)
        coveringthickness2 = self.getDataAsFloat(key='part_covering2thickness', **keywords)
        laminate1 = self.asString(key='part_laminate1', **keywords)
        laminate2 = self.asString(key='part_laminate2', **keywords)
        laminate3 = self.asString(key='part_laminate3', **keywords)

        description = self.asString(data=entry.description, **keywords)

        positionInfo = entry.positionInfo
        unitname = self.getDataAsString(data=positionInfo['CabinetName'], **keywords)

        return self.OUTPUT_END_OF_LINE.join((
            self.getElementBaseLine(**keywords),
            self.getCoveringLine(number=7, covering=covering1, thickness=coveringthickness1, **keywords),
            self.getCoveringLine(number=8, covering=covering2, thickness=coveringthickness2, **keywords),
            self.getCustomerTextLine(number=10, value=description, **keywords),
            self.getCustomerTextLine(number=11, value=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords), **keywords),
            self.getCustomerTextLine(number=12, value=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords), **keywords),
            self.getCustomerTextLine(number=13, value=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords), **keywords),
            self.getCustomerTextLine(number=14, value=covering1, **keywords),
            self.getCustomerTextLine(number=15, value=covering2, **keywords),
            self.getCustomerTextLine(number=16, value=laminate1, **keywords),
            self.getCustomerTextLine(number=17, value=laminate2, **keywords),
            self.getCustomerTextLine(number=18, value=laminate3, **keywords),
            self.getCustomerTextLine(number=19, value=unitname, **keywords),
            self.getCustomerTextLine(number=20, value=entry.positionInfo['PositionNumber'], **keywords),
            self.getCustomerTextLine(number=21, value=entry.positionInfo['PositionDescription'], **keywords),
            self.getFormatPostprocessingLine(**keywords),
        ))

    # Footer
    def writeFooter(self, **keywords):
        self.writeLines(lines=self.getFooter(**keywords), **keywords)

    def getFooter(self, **keywords):
        return 'L$',
