# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import os
import csv

from interiorcad.backend.cuttinglistentry import GRAIN_CROSSWISE, EDGE_CODE_CROSS_GRAIN_TRANSLATION
from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.dosfilenamecreator import DosFilenameCreator
from interiorcad.backend.localization import getLocalizedString, CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP
from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.interface.geometry.edges import EdgeData


class SchnittProfitExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = os.path.splitext(getLocalizedString(CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP, 4, "CutRite.pnx"))[0]

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = ';'  # standard of cut-rite is ',', but it's seems better using the standard csv-delimiter.
        # This exports takes care directly of quoting text fields. We must switch off the automatic quoting of the CSV export
        self.csvquoting = csv.QUOTE_NONE
        self.csvquotechar = None

        self.EXTENSION = 'pnx'
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')
        self.NC_PROGRAM_DISPLAY_PATH_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-path-definition', 'pathandnamewithextension')

        self.SHORT_PART_NAMES = self.getCuttinglistConfiguration().getAsBool('short-part-names', False)  # older versions only allow 12 characters as name
        self.EXPORT_COVERINGS_WITH_EDGES_OF_WORKPIECE = self.getCuttinglistConfiguration().getAsBool('export-coverings-with-edges-of-workpiece', False)

    def getDecimalSeparator(self, **keywords):
        return '.'

    # header
    def getHeaderLine(self, **keywords):
        return None

    # elements
    def quoteElement(self, element):
        quoteChar = '"'
        return quoteChar + element + quoteChar

    # SchnittProfit prefers rotated dimensions and edges over using cross grain (2)
    def getSimplifiedGrain(self, grain):
        # 0 = No grain/part can be rotated,
        # 1 = grain along the length of the board/part cannot be rotated
        # 2 = grain along the width of the board/part must be rotated (NOT USED!)
        # (Caution: 'Schnitt Profit.ef' defines it another way!)
        if grain == self.GRAIN_NONE:
            return 0
        else:  # self.GRAIN_LENGTHWISE, self.GRAIN_CROSSWISE:
            return 1

    # Kantenstoßinformationen / Edge joint information is the same function as in CADmaticExport
    def getJointAsNumber(self, jointId):
        if jointId == EdgeData.JOINT_MITER:
            return '2'
        elif jointId == EdgeData.JOINT_CONTINUOUS:
            return '1'
        elif jointId == EdgeData.JOINT_DISCONTINUOUS:
            return '0'
        elif not jointId or jointId == EdgeData.JOINT_NOEDGE:
            return '0'  # No value shown for no edge
        else:
            return '0'

    def getExportSpecificEdgeCode(self, **keywords):
        entry = keywords['entry']
        return ':'.join((
            ''.join((
                '0',
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.front.jointEnd, **keywords)),
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.front.jointStart, **keywords)),
            )) if self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords) else '',
            ''.join((
                '0',
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.back.jointEnd, **keywords)),
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.back.jointStart, **keywords)),
            )) if self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords) else '',
            ''.join((
                '0',
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.left.jointEnd, **keywords)),
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.left.jointStart, **keywords)),
            )) if self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords) else '',
            ''.join((
                '0',
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.right.jointEnd, **keywords)),
                self.getJointAsNumber(self.getDataAsString(data=entry.edgesGrainDirection.right.jointStart, **keywords)),
            )) if self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords) else '',
        ))

    # mainly duplicated from maestronestingexport & multicenterstudioexport, but reduced
    def getNCProgramNamesForMainOrOpposite(self, **keywords):
        entry = keywords['entry']
        side = keywords['side']

        ncProgramNameMain = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords)
        ncProgramNameOpposite = self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)

        if side == 'main' and ncProgramNameMain:
            return ncProgramNameMain
        elif side == 'main':
            return ncProgramNameOpposite
        elif side == 'opposite' and ncProgramNameMain:
            return ncProgramNameOpposite
        else:
            return ''

    def getEdgesOfWorkpieceForCovering(self, **keywords):
        entry = keywords['entry']

        if entry.isCovering() and self.EXPORT_COVERINGS_WITH_EDGES_OF_WORKPIECE:
            entry.edges = ExportDataCollector.getEdges(entry.workpieceUUID)
            entry.edges.ensureNotNoneRectangularEdges()

            entry.edgesGrainDirection = copy.deepcopy(entry.edges)

            if entry.grain == GRAIN_CROSSWISE:
                entry.edgesGrainDirection.polygonal = None  # TODO: correct polygonal edges information -- NG (IT-212)

                entry.edgesGrainDirection.front = entry.edges.right
                entry.edgesGrainDirection.back = entry.edges.left
                entry.edgesGrainDirection.left = entry.edges.front
                entry.edgesGrainDirection.right = entry.edges.back

                # Edges Front Back Left Right = 1234. After rotating 90° clockwise it must be Left Right Back Front = 3421
                entry.edgesGrainDirection.edgecode = entry.edgesGrainDirection.edgecode.translate(EDGE_CODE_CROSS_GRAIN_TRANSLATION)

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']
        workpiece = keywords['workpiece']

        self.getEdgesOfWorkpieceForCovering(**keywords)

        name = entry.name
        partName = DosFilenameCreator().createDosFilename(name) if self.SHORT_PART_NAMES else name

        return (
            # standard field, must be present in that order
            self.quoteElement(element=self.asString(data=partName, **keywords)),
            self.quoteElement(element=self.asString(data=entry.itemID, **keywords)),
            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asInt(data=entry.count, **keywords),
            '0',
            '0',
            self.asInt(data=self.getSimplifiedGrain(entry.grain), **keywords),
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords)),
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords)),
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords)),
            self.quoteElement(element=self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords)),

            # the following fields are optional
            self.quoteElement(element=self.asString(data=projectInfo['ProjectID'], **keywords)),
            self.quoteElement(element=self.asString(data=projectInfo['ProjectCommission'], **keywords)),
            self.quoteElement(element=self.asString(data=projectInfo['ProjectConstruction'], **keywords)),
            self.quoteElement(element=self.asString(data=projectInfo['DocumentName'], **keywords)),

            self.quoteElement(element=self.asString(data=positionInfo['PositionNumber'], **keywords)),
            self.quoteElement(element=self.asString(data=positionInfo['PositionDescription'], **keywords)),

            self.quoteElement(element=self.asString(data=positionInfo['CabinetName'], **keywords)),
            self.asCoordinate(key='unit_unitwidth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelX
            self.asCoordinate(key='unit_unitdepth', uuid=workpiece.parentUUID, **keywords),  # STIMoebelY
            self.asCoordinate(key='unit_unitheight', uuid=workpiece.parentUUID, **keywords),  # STIMoebelZ

            self.quoteElement(element=self.asString(data=entry.description, **keywords)),
            self.quoteElement(element=self.asStringSingleLine(key='part_comment', **keywords)),

            self.asCoordinate(data=entry.boardThickness, **keywords),

            self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),

            self.quoteElement(element=self.asString(data=entry.polygonalEdgesID, **keywords)),
            self.quoteElement(element=self.asString(data=entry.polygonalEdges, **keywords)),

            self.quoteElement(element=self.asString(key='part_covering1', **keywords)),
            self.quoteElement(element=self.asString(key='part_covering2', **keywords)),

            self.quoteElement(element=self.asString(key='part_laminate1', **keywords)),
            self.quoteElement(element=self.asString(key='part_laminate2', **keywords)),
            self.quoteElement(element=self.asString(key='part_laminate3', **keywords)),

            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords)),
            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords)),
            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)),
            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathSizing, **keywords)),
            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathMain, **keywords)),
            self.quoteElement(element=self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_PATH_DEFINITION, data=entry.ncProgramPathOpposite, **keywords)),

            self.getNCProgramNamesForMainOrOpposite(side='main', **keywords),
            self.getNCProgramNamesForMainOrOpposite(side='opposite', **keywords),

            self.getExportSpecificEdgeCode(**keywords),  # Kantenstoßinformationen / Edge joint information
        )
