# -*- coding: utf-8 -*-

# Copyright 2016-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools

from interiorcad.cuttinglistexport.base import CuttinglistExportBase


class SchreinersBueroExport(CuttinglistExportBase):
    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.FIELD_SEPARATOR = '\t'
        self.counter = 0

        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA
        # Sorting Criteria 'cabinetname' is not part of getPositionDetailLine and getWorkpieceLines!

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsandfittings')  # | partsonly | fittingsonly
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')  # only minimal acccumulation

        self.SCHREINERS_BUERO_USAGE = self.getCuttinglistConfiguration().get('schreiners-buero-usage', 'name')  # Verwendung

    def getDecimalSeparator(self, **keywords):
        return '.'

    def doExport(self, **keywords):
        self.writeHeader(**keywords)
        self.writeBody(**keywords)

    # header
    def writeHeader(self, **keywords):
        self.writeLines(lines=self.getHeaderVersionLine(**keywords), **keywords)

    def getHeaderVersionLine(self, **keywords):
        return (
            self.FIELD_SEPARATOR.join((
                'V',
                '1.0',
            )),
        )

    # body
    def writeBody(self, **keywords):
        def getPositionNumber(entry):
            return self.asString(data=entry.positionInfo['PositionNumber'], **keywords)

        entries = self.generateSortedEntries(**keywords)

        entriesByPosition = itertools.groupby(sorted(entries, key=getPositionNumber), key=getPositionNumber)
        for positionNumber, entriesRaw in entriesByPosition:
            entries = tuple(entriesRaw)
            self.writeLines(lines=self.getPositionDetailLines(parentUUID=entries[0].parentUUID, positionNumber=positionNumber, **keywords), **keywords)
            for entry in entries:
                self.writeLines(lines=self.getWorkpieceLines(entry=entry, workpiece=entry.workpiece, **keywords), **keywords)

    def getPositionDetailLines(self, **keywords):
        parentUUID = keywords['parentUUID']
        positionNumber = keywords['positionNumber']
        positionDescription, positionQuantity = self.getInfoForPositionNumber(positionNumber=positionNumber)
        return (
            self.FIELD_SEPARATOR.join((
                'P',
                self.asString(data=positionNumber, **keywords),  # EF: PositionNummer
                self.asInt(data=float(positionQuantity), **keywords) if positionQuantity else '1',  # EF: PositionAnzahl
                self.asString(data=positionDescription, **keywords),  # EF: PositionName
                self.asCoordinate(key='unit_unitwidth', uuid=parentUUID, **keywords),  # STIMoebelX
                self.asCoordinate(key='unit_unitheight', uuid=parentUUID, **keywords),  # STIMoebelY
                self.asCoordinate(key='unit_unitdepth', uuid=parentUUID, **keywords),  # STIMoebelZ
            )),
        )

    def getSchreinersBueroUsage(self, **keywords):  # Verwendung
        entry = keywords['entry']

        if self.SCHREINERS_BUERO_USAGE == 'description':
            return entry.description
        else:  # name
            return entry.name

    def getWorkpieceLines(self, **keywords):
        self.counter += 1
        entry = keywords['entry']

        return (
            self.FIELD_SEPARATOR.join((
                'T',
                self.asString(data=entry.positionInfo['PositionNumber'], **keywords),  # EF: PositionNummer
                self.asInt(data=self.counter, **keywords),  # EF: STITeileNummer
                self.asString(data=self.getConstructionGroupNameInclCoveringsAndFittings(**keywords)),  # EF: STIBaugruppenName
                self.asString(data=self.getSchreinersBueroUsage(**keywords), **keywords),  # EF: STIName -> Verwendung
                self.asString(data=entry.itemID, **keywords),  # EF: STIArtikelNr
                self.asInt(data=entry.count, **keywords),  # EF: STIAnzahl
                self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),  # EF: STILaenge
                self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),  # EF: STIBreite
                self.asCoordinate(data=entry.boardThickness, **keywords),  # EF: STIHoehe
                self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),
                self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
                self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),
                self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
                self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),
                self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
                self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),
                self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),
                self.asString(key='part_covering1', **keywords),  # EF: STIBelag1
                self.asString(key='part_covering2', **keywords),  # EF: STIBelag2
                self.asString(key='part_laminate1', **keywords),  # EF: STILaminate1
                self.asString(key='part_laminate2', **keywords),  # EF: STILaminate2
                self.asStringSingleLine(key='part_comment', **keywords),  # EF: STIBemerkung
            )),
        )
