# -*- coding: utf-8 -*-

# Copyright 2016-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools

from interiorcad.cuttinglistexport.base import CuttinglistExportBase


class SPlusExport(CuttinglistExportBase):
    EXPORT_FORMAT = 'S-Plus'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.FIELD_SEPARATOR = '\t'
        self.counter = 0

        self.EXTENSION = 'txt'
        self.CONFIGURED_SORTING_CRITERIA = self.POSITION_SORTING_CRITERIA

        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsandfittings')  # | partsonly | fittingsonly
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)

    def doExport(self, **keywords):
        self.writeHeader(**keywords)
        self.writeBody(**keywords)

    # header
    def writeHeader(self, **keywords):
        self.writeLines(lines=self.getHeaderProjectDetailsLine(**keywords), **keywords)

    def getHeaderProjectDetailsLine(self, **keywords):
        projectInfo = keywords['projectInfo']
        return (
            self.FIELD_SEPARATOR.join((
                'AUF',  # Prefix:STIAuftragNummer:"AUF\t"
                self.asString(data=projectInfo['ProjectID'], **keywords),  # STIAuftragNummer
                self.asString(data=projectInfo['Administrator'], **keywords),  # STIJobSachbearb
                self.asString(data=projectInfo['CustomerNumber'], **keywords),  # STIJobKundeNr
                self.asString(data=projectInfo['ProjectCommission'], **keywords),  # STIAuftragKomission
                self.asString(data=projectInfo['ProjectConstruction'], **keywords),  # STIAuftragBauvorhaben
                self.asString(data=projectInfo['DeliveryWeek'], **keywords),  # STIJobLieferKW
                self.asString(data=projectInfo['DeliveryYear'], **keywords),  # STIJobLieferJahr
                self.asString(data=projectInfo['DeliveryDate'], **keywords),  # STIJobFixtermin
                self.asString(data=projectInfo['ProjectID'], **keywords),  # STIJobFirma
                ' '.join((
                    self.asString(data=projectInfo['FirstName'], **keywords),  # STIJobVorname # Prefix:STIJobNachname:" "
                    self.asString(data=projectInfo['LastName'], **keywords),  # STIJobNachname
                )),
                self.asString(data=projectInfo['StreetName'], **keywords),  # STIJobStrasse
                'D',  # Prefix:STIJobPLZ:"D\t"
                self.asString(data=projectInfo['PostalCode'], **keywords),  # STIJobPLZ
                self.asString(data=projectInfo['City'], **keywords),  # STIJobOrt
            )),
        )

    # body
    def writeBody(self, **keywords):
        def getPositionNumber(entry):
            return self.asString(data=entry.positionInfo['PositionNumber'], **keywords)

        entries = self.generateSortedEntries(**keywords)

        entriesByPosition = itertools.groupby(sorted(entries, key=getPositionNumber), key=getPositionNumber)
        for positionNumber, entriesRaw in entriesByPosition:
            entries = tuple(entriesRaw)
            self.writeLines(lines=self.getPositionDetailLines(parentUUID=entries[0].parentUUID, positionNumber=positionNumber, **keywords), **keywords)
            for entry in entries:
                self.writeLines(lines=self.getWorkpieceLines(entry=entry, workpiece=entry.workpiece, **keywords), **keywords)

    def getPositionDetailLines(self, **keywords):
        parentUUID = keywords['parentUUID']
        positionNumber = keywords['positionNumber']
        positionDescription, positionQuantity = self.getInfoForPositionNumber(positionNumber=positionNumber)
        return (
            self.FIELD_SEPARATOR.join((
                'POS',  # Prefix:"POS\t"
                self.asString(data=positionNumber, **keywords),  # STIPositionNummer
                self.asInt(data=float(positionQuantity), **keywords) if positionQuantity else '1',  # STIPositionAnzahl
                'Stk',  # Postfix:STIPositionAnzahl:"Stk\t\t"
                '',
                self.asString(data=positionDescription, **keywords),  # STIPositionName
                self.asCoordinate(key='unit_unitwidth', uuid=parentUUID, **keywords),  # STIMoebelX
                self.asCoordinate(key='unit_unitdepth', uuid=parentUUID, **keywords),  # STIMoebelY
                self.asCoordinate(key='unit_unitheight', uuid=parentUUID, **keywords),  # STIMoebelZ
                '0',  # Postfix:STIMoebelZ:"0\t0\t0\t"
                '0',
                '0',
                self.asString(data=positionNumber, **keywords),  # STIPositionNummer
            )),
        )

    def getWorkpieceLines(self, **keywords):
        self.counter += 1
        entry = keywords['entry']

        return (
            self.FIELD_SEPARATOR.join((
                'STK',  # Prefix:"STK\t"
                self.asInt(data=self.counter, **keywords),  # STITeileNummer
                self.asInt(data=entry.count, **keywords),  # STIAnzahl
                '{unit}'.format(unit='m2' if not entry.isFitting() else 'Stk'),  # Postfix:STIAnzahl:"m2\t"
                self.asString(data=entry.itemID, **keywords),  # STIBestellOderArtikelNr
                self.asString(data=entry.name, **keywords),  # STIName
                self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),  # STILaenge
                self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),  # STIBreite
                self.asCoordinate(data=entry.boardThickness, **keywords),  # STIHoehe
                self.asString(key='part_covering1', **keywords),  # STIBelag1
                self.asString(key='part_covering2', **keywords),  # STIBelag2
                self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),  # STIKante1
                '0',  # Postfix:STIKante1:"0\t0\t"
                '0',
                self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),  # STIKantendicke1
                'N',  # Postfix:STIKantendicke1:"N\t"
                self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),  # STIKante2
                '0',  # Postfix:STIKante2:"0\t0\t"
                '0',
                self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),  # STIKantendicke2
                'N',  # Postfix:STIKantendicke2:"N\t"
                self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),  # STIKante3
                '0',  # Postfix:STIKante3:"0\t0\t"
                '0',
                self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),  # STIKantendicke3
                'N',  # Postfix:STIKantendicke3:"N\t"
                self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),  # STIKante4
                '0',  # Postfix:STIKante4:"0\t0\t"
                '0',
                self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),  # STIKantendicke4
                'N',  # Postfix:STIKantendicke4:"N\t"
                self.asString(key='part_laminate1', **keywords),  # STILaminate1
                'JJJJJJ',  # Postfix:STILaminate1:"JJJJJJ\t"
                self.asString(key='part_laminate2', **keywords),  # STILaminate2
                'JJJJJJ',  # Postfix:STILaminate2:"JJJJJJ\t"
                self.asString(key='part_laminate3', **keywords),  # STILaminate3
                'JJJJJJ',  # Postfix:STILaminate3:"JJJJJJ\t"
                self.asString(data=entry.name, **keywords),  # STIBaugruppe
                self.asString(data=entry.name, **keywords),  # STIBaugruppenName
                self.asStringSingleLine(key='part_comment', **keywords),  # EF: STIBemerkung
            )),
        )
