# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase
from interiorcad.backend.localization import getLocalizedString, CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP, CUTTINGLIST_RESOURCEGROUP


class SpreadsheetExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = getLocalizedString(CUTTINGLIST_OUTPUT_FILE_NAMES_RESOURCEGROUP, 10, 'Spreadsheet')

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.EXTENSION = 'csv'
        self.EXPORT_LOCATION = self.getCuttinglistConfiguration().get('export-location', 'file')  # | clipboard
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')  # | partsandfittings | fittingsonly
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_FITTINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-fittings', True)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')
        self.NC_PROGRAM_DISPLAY_NAME_DEFINITION = self.getCuttinglistConfiguration().get('nc-program-display-name-definition', 'namewithextension')

    def doCuttinglistExport(self, **keywords):
        if self.EXPORT_LOCATION == 'clipboard':
            self.doClipboardExport(**keywords)
        elif self.EXPORT_LOCATION == 'fileandclipboard':
            self.doFileExport(**keywords)
            self.doClipboardExport(**keywords)
        else:
            self.doFileExport(**keywords)

    def openExportPath(self, **keywords):
        if 'file' in self.EXPORT_LOCATION:
            super().openExportPath(**keywords)

    def getHeaderLine(self, **keywords):
        return (
            self.getLocalizedHeader(17, 'Project ID'),
            self.getLocalizedHeader(18, 'Position Number'),
            self.getLocalizedHeader(19, 'Position Description'),

            self.getLocalizedHeader(20, 'Cabinet Name'),

            self.getLocalizedHeader(1, 'Supplier ID'),
            self.getLocalizedHeader(2, 'Description'),

            self.getLocalizedHeader(3, 'Length'),
            self.getLocalizedHeader(4, 'Width'),
            self.getLocalizedHeader(5, 'Thickness'),

            self.getLocalizedHeader(6, 'Count'),
            self.getLocalizedHeader(7, 'Unique Name'),

            self.getLocalizedHeader(11, 'Edge ID Front 1'),
            self.getLocalizedHeader(21, 'Edge Thickness Front 1'),
            self.getLocalizedHeader(12, 'Edge ID Back 2'),
            self.getLocalizedHeader(22, 'Edge Thickness Back 2'),
            self.getLocalizedHeader(13, 'Edge ID Left 3'),
            self.getLocalizedHeader(23, 'Edge Thickness Left 3'),
            self.getLocalizedHeader(14, 'Edge ID Right 4'),
            self.getLocalizedHeader(24, 'Edge Thickness Right 4'),

            self.getLocalizedHeader(15, 'Polygonal Edges ID'),
            self.getLocalizedHeader(16, 'Polygonal Existing Edges'),

            self.getLocalizedHeader(25, 'Covering 1'),
            self.getLocalizedHeader(26, 'Covering 2'),

            self.getLocalizedHeader(27, 'Laminate 1'),
            self.getLocalizedHeader(28, 'Laminate 2'),
            self.getLocalizedHeader(29, 'Laminate 3'),

            self.getLocalizedHeader(30, 'Comment'),

            self.getLocalizedHeader(8, 'NC Program Name Sizing'),
            self.getLocalizedHeader(9, 'NC Program Name Main'),
            self.getLocalizedHeader(10, 'NC Program Name Opposite'),
        )

    def getElements(self, **keywords):
        entry = keywords['entry']
        projectInfo = keywords['projectInfo']
        positionInfo = keywords['positionInfo']

        return (
            self.asString(data=projectInfo['ProjectID'], **keywords),
            self.asString(data=positionInfo['PositionNumber'], **keywords),
            self.asString(data=positionInfo['PositionDescription'], **keywords),

            self.asString(data=positionInfo['CabinetName'], **keywords),

            self.asString(data=entry.itemID, **keywords),
            self.asString(data=entry.description, **keywords),

            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardThickness, **keywords),

            self.asInt(data=entry.count, **keywords),

            self.asString(data=entry.name, **keywords),

            self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
            self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
            self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
            self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),

            self.asString(data=entry.polygonalEdgesID, **keywords),
            self.asString(data=entry.polygonalEdges, **keywords),

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),

            self.asString(key='part_laminate1', **keywords),
            self.asString(key='part_laminate2', **keywords),
            self.asString(key='part_laminate3', **keywords),

            self.asStringSingleLine(key='part_comment', **keywords),

            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathSizing, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathMain, **keywords),
            self.getNCProgramDisplayName(definition=self.NC_PROGRAM_DISPLAY_NAME_DEFINITION, data=entry.ncProgramPathOpposite, **keywords),
        )

    def getLocalizedHeader(self, index, default):
        return getLocalizedString(CUTTINGLIST_RESOURCEGROUP, index, default)

    def doClipboardExport(self, **keywords):
        # Excel needs tabulators when pasting from clipboard (LibreOfficeCalc does not care, as it always shows the import dialog)
        self.csvdelimiter = '\t'
        super().doClipboardExport(**keywords)
