# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools

from interiorcad.cuttinglistexport.csvbase import CuttinglistExportCSVBase


class WebOptExport(CuttinglistExportCSVBase):
    EXPORT_FORMAT = 'WebOpt'

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.csvdelimiter = '\t'
        self.counter = 0

        self.EXTENSION = 'txt'
        self.EXPORT_BOARDS = self.getCuttinglistConfiguration().getAsBool('export-boards', True)
        self.SCOPE_OF_EXPORT = self.getCuttinglistConfiguration().get('scope-of-export', 'partsonly')
        self.CUMULATE_COVERINGS = self.getCuttinglistConfiguration().getAsBool('cumulate-coverings', False)
        self.CUMULATE_WORKPIECES = self.getCuttinglistConfiguration().get('cumulate-workpieces', 'none')

    # Project
    def joinProjectInfoWithCSVDelimiter(self, **keywords):
        projectInfo = self.getProjectInfo()
        return self.csvdelimiter.join((
            keywords['refText'],
            self.asString(data=projectInfo[keywords['refValue']], **keywords),
        ))

    def writeProjectInfo(self, **keywords):
        lines = ((
            'Projekt',
            self.joinProjectInfoWithCSVDelimiter(refText='ProjektID', refValue='ProjectID'),
            self.joinProjectInfoWithCSVDelimiter(refText='ProjektName', refValue='ProjectCommission'),
            self.joinProjectInfoWithCSVDelimiter(refText='Information', refValue='ProjectConstruction'),
            self.joinProjectInfoWithCSVDelimiter(refText='Beschreibung', refValue='DocumentName'),  # DeliveryDate could be useful either
        ))
        self.writeLines(lines=lines, **keywords)

    # Boards
    def writeBoardsHeader(self, **keywords):
        return (
            'Nr',
            'Anzahl',
            'XL',
            'YL',
            'ZL',
            'Material',
            'Information',
            'Beschreibung',
        )

    def getUniqueBoards(self, **keywords):
        entries = self.generateUnsortedEntries(**keywords)
        items = sorted(((entry.itemID, entry.boardThickness) for entry in entries), key=lambda entry: entry[0])
        groups = itertools.groupby(items, key=lambda entry: entry[0])
        boards = ((key, tuple(entries)[0][1]) for key, entries in groups)
        return sorted(boards, key=lambda board: board[0])

    def writeBoardsBody(self, **keywords):
        def getBoard(board):
            self.counter += 1
            return (
                self.counter,
                '0',  # 0 = unlimited
                '2800',  # length (not available in ic)
                '2070',  # width (not available in ic)
                self.asCoordinate(data=board[1], **keywords),  # thickness
                self.asString(data=board[0], **keywords),  # material
                '',  # complete name of material, optional
                '',  # no further comment
            )

        boards = self.getUniqueBoards(**keywords)
        rows = (getBoard(board) for board in boards)
        self.writeRows(rows=rows, **keywords)

    def writeBoards(self, **keywords):
        if self.EXPORT_BOARDS:
            self.writeLines(lines=('', 'Platten'), **keywords)
            self.writeBoardsHeader(**keywords)
            self.writeBoardsBody(**keywords)

    # Parts
    def getHeaderLine(self, **keywords):
        return (
            'Nr',
            'Anzahl',
            'XL',
            'YL',
            'ZL',
            'Material',
            'Information',
            'Beschreibung',
            'Drehbar',
            'Kante_V',
            'Kante_H',
            'Kante_L',
            'Kante_R',
            'Kante_V_Dicke',
            'Kante_H_Dicke',
            'Kante_L_Dicke',
            'Kante_R_Dicke',
            'Belag_I',
            'Belag_A',
        )

    def getSimplifiedGrain(self, grain):
        # N= part must not be rotated, Y= part can be rotated
        if grain == self.GRAIN_NONE:
            return 'Y'
        else:  # self.GRAIN_LENGTHWISE, self.GRAIN_CROSSWISE:
            return 'N'

    def getElements(self, **keywords):
        entry = keywords['entry']
        self.counter += 1

        positionInfo = keywords['positionInfo']
        unitUnitName = self.asString(data=positionInfo['CabinetName'], **keywords)
        partComment = self.asStringSingleLine(key='part_comment', **keywords)
        webOptDescription = ''
        if unitUnitName or partComment:
            if unitUnitName and partComment:
                webOptDescription = unitUnitName + ', ' + partComment
            elif unitUnitName:
                webOptDescription = unitUnitName
            elif partComment:
                webOptDescription = partComment

        return (
            self.counter,
            self.asInt(data=entry.count, **keywords),
            self.asCoordinate(data=entry.boardLengthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardWidthGrainDirection, **keywords),
            self.asCoordinate(data=entry.boardThickness, **keywords),
            self.asString(data=entry.itemID, **keywords),
            self.asString(data=entry.name, **keywords),
            webOptDescription,
            self.getSimplifiedGrain(entry.grain),

            self.asString(data=entry.edgesGrainDirection.front.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.back.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.left.supplierID, **keywords),
            self.asString(data=entry.edgesGrainDirection.right.supplierID, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.front.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.back.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.left.thickness, **keywords),
            self.asCoordinate(data=entry.edgesGrainDirection.right.thickness, **keywords),

            self.asString(key='part_covering1', **keywords),
            self.asString(key='part_covering2', **keywords),
        )

    def writeListHeader(self, **keywords):
        self.writeProjectInfo(**keywords)
        self.writeBoards(**keywords)
        self.writeLines(lines=('', 'Teile'), **keywords)
        super().writeListHeader(**keywords)
