# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import logging
import os.path
import unittest.mock
from datetime import datetime

import interiorcad.backend.exportsettingsfile
import interiorcad.interface.configurationkeys as ConfigurationKeys
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.backend.workpiececreationstrategy.constants import SizeMode


def createSizingModeFromText(sizingText):
    if sizingText == 'nc-compatible':
        sizing = SizeMode.NC_COMPATIBLE
    elif sizingText == 'sawn-size':
        sizing = SizeMode.SAWN_SIZE
    elif sizingText == 'sawn-size-with-addition':
        sizing = SizeMode.SAWN_SIZE_WITH_ADDITION
    elif sizingText == 'finished-size':
        sizing = SizeMode.FINISHED_SIZE
    elif sizingText == 'finished-size-with-addition':
        sizing = SizeMode.FINISHED_SIZE_WITH_ADDITION
    else:
        sizing = SizeMode.NC_COMPATIBLE
        logging.warning(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 1, "Cutting list sizing mode '{}' does not exist -- falling back to sizing mode 'nc-compatible'").format(sizingText))
    return sizing


DATETIME_NOW = datetime.now()
DATETIME_NOW_UTC = datetime.utcnow()


DEFAULT_CUSTOM_EXPORT_FOLDER_PATH = os.path.join('Export {ISODateTime}', '{DocumentName}')

# interiorcad 2020 F3: there used to be a class HardCodedConfiguration as base of all configurations. This class was removed and integrated only in NCGlobalExportConfiguration


class CommonConfiguration:
    def __init__(self):
        self.exportsettingsfile = interiorcad.backend.exportsettingsfile.ExportSettingsFile()

        self.jointing = self.exportsettingsfile.getSimpleValueAsBool(ConfigurationKeys.COMMON_JOINTING, False)
        self.minJoint = self.exportsettingsfile.getSimpleValueAsFloat(ConfigurationKeys.COMMON_JOINTING_MIN, 0.0)
        self.maxJoint = self.exportsettingsfile.getSimpleValueAsFloat(ConfigurationKeys.COMMON_JOINTING_MAX, 0.0)
        self.accurateSaw = self.exportsettingsfile.getSimpleValueAsBool(ConfigurationKeys.NC_OUTPUT_RECT_PART_SIZING, False)
        self.logFileName = self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.COMMON_LOG_FILENAME, "XG Export Log.txt")

        self.cuttingLists = [self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.CUTTING_OUTPUT_TARGET, "Spreadsheet")]
        self.sizingAddition = self.exportsettingsfile.getSimpleValueAsFloat(ConfigurationKeys.CUTTING_OUTPUT_CUT_SIZE_ADDITION, 0.0)
        self.separateSizing = self.exportsettingsfile.getSimpleValueAsBool(ConfigurationKeys.NC_OUTPUT_SEPARATE_SIZING, False)
        self.excludedConstructionGroups = self.exportsettingsfile.getAllAttributesOfAllChildrenWithTag(ConfigurationKeys.COMMON_CONSTRUCTION_GROUP, "excluded")
        self.setName = self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.COMMON_SET_NAME, None)

        self.basePathPrefix = '{TempDir}'
        self.setCustomizeablePath(DEFAULT_CUSTOM_EXPORT_FOLDER_PATH)

    def getPathForExport(self):
        return os.path.join(self.basePathPrefix, self.customizeablePath)

    def setBasePathPrefix(self, path):
        self.basePathPrefix = path

    def getBasePathPrefix(self):
        return self.basePathPrefix

    def getCustomizeablePath(self):
        return self.customizeablePath

    def setCustomizeablePath(self, path):
        self.customizeablePath = path

    # Old name of sizingAddition
    @property
    def cut_size_add(self):
        return self.sizingAddition


class NCGlobalExportConfiguration(CommonConfiguration):
    NULL, MIN_X, MAX_X, MIN_Y, MAX_Y = range(0, 5)

    def __init__(self):
        super().__init__()

        self.drillingStartPosition = [self.MAX_Y, self.MIN_X]
        self.minimumSegmentLength = self.exportsettingsfile.getSimpleValueAsFloat(ConfigurationKeys.NC_OUTPUT_MINIMUM_SEGMENT_LENGTH, 50.0)

        self.coordinatePrecision = self.exportsettingsfile.getSimpleValueAsInt(ConfigurationKeys.NC_OUTPUT_COORDINATE_PRECISION, 7)
        self.angularPrecision = self.exportsettingsfile.getSimpleValueAsInt(ConfigurationKeys.NC_OUTPUT_ANGULAR_PRECISION, 7)
        self.significantDecimalsOnly = self.exportsettingsfile.getSimpleValueAsBool(ConfigurationKeys.NC_OUTPUT_SIGNIFICANT_DECIMALS_ONLY, True)

        self.setBasePathPrefix(self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.NC_OUTPUT_EXPORT_FOLDER_PATH, "{DocumentDir}"))
        self.setCustomizeablePath(self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.NC_OUTPUT_CUSTOM_EXPORT_FOLDER_PATH, DEFAULT_CUSTOM_EXPORT_FOLDER_PATH))


class CuttingListGlobalExportConfiguration(CommonConfiguration):
    def __init__(self):
        super().__init__()

        self.sizingText = self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.CUTTING_OUTPUT_SIZING, 'nc-compatible')
        self.sizing = createSizingModeFromText(self.sizingText)

        self.additionalCoveringAddition = self.exportsettingsfile.getSimpleValueAsFloat(ConfigurationKeys.CUTTING_OUTPUT_ADDITIONAL_COVERING_ADDITION, 10.0)

        self.coordinatePrecision = self.exportsettingsfile.getSimpleValueAsInt(ConfigurationKeys.CUTTING_OUTPUT_COORDINATE_PRECISION, 1)
        self.angularPrecision = self.exportsettingsfile.getSimpleValueAsInt(ConfigurationKeys.CUTTING_OUTPUT_ANGULAR_PRECISION, 4)
        self.significantDecimalsOnly = self.exportsettingsfile.getSimpleValueAsBool(ConfigurationKeys.CUTTING_OUTPUT_SIGNIFICANT_DECIMALS_ONLY, True)

        self.setBasePathPrefix(self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.CUTTING_OUTPUT_EXPORT_FOLDER_PATH, "{DocumentDir}"))
        self.setCustomizeablePath(self.exportsettingsfile.getSimpleValueAsString(ConfigurationKeys.CUTTING_OUTPUT_CUSTOM_EXPORT_FOLDER_PATH, DEFAULT_CUSTOM_EXPORT_FOLDER_PATH))


class CommonConfigurationTests(unittest.TestCase):
    def setUp(self):
        self.config = CommonConfiguration()

    def testDefaultCommonConfigurationIsAsExpected(self):
        self.assertFalse(self.config.jointing)
        self.assertEqual(2.0, self.config.minJoint)
        self.assertEqual(5.0, self.config.maxJoint)
        self.assertFalse(self.config.accurateSaw)
        self.assertEqual('XG Export Log.txt', self.config.logFileName)
        self.assertEqual(['Spreadsheet (File)'], self.config.cuttingLists)
        self.assertEqual(10.0, self.config.sizingAddition)
        self.assertEqual(10.0, self.config.cut_size_add)

    def testDefaultOutputPathConfigurationIsAsExpected(self):
        self.assertEqual('{TempDir}', self.config.basePathPrefix)
        self.assertEqual(DEFAULT_CUSTOM_EXPORT_FOLDER_PATH, self.config.customizeablePath)
        self.assertEqual(os.path.join(self.config.basePathPrefix, self.config.customizeablePath), self.config.getPathForExport())


class NCConfigurationTests(unittest.TestCase):
    def setUp(self):
        self.config = NCGlobalExportConfiguration()

    def testFormerlyHardCodedConfiguration(self):
        # Test values that were formerly in the class HardCodedConfiguration
        self.assertEqual([4, 1], self.config.drillingStartPosition)
        self.assertEqual(51.0, self.config.minimumSegmentLength)

    def testCommonNCConfigurationShouldBeAsExpected(self):
        self.assertFalse(self.config.jointing)
        self.assertEqual(2.0, self.config.minJoint)
        self.assertEqual(5.0, self.config.maxJoint)
        self.assertFalse(self.config.accurateSaw)
        self.assertEqual('XG Export Log.txt', self.config.logFileName)
        self.assertEqual(['Spreadsheet (File)'], self.config.cuttingLists)
        self.assertEqual(10.0, self.config.sizingAddition)

    def testSpecificNCConfigurationShouldBeAsExpected(self):
        self.assertFalse(self.config.separateSizing)
        self.assertEqual(7, self.config.coordinatePrecision)
        self.assertEqual(7, self.config.angularPrecision)
        self.assertEqual('{DocumentDir}/ncexport', self.config.getBasePathPrefix())


class CuttingListConfigurationTests(unittest.TestCase):
    def setUp(self):
        self.config = CuttingListGlobalExportConfiguration()

    def testCommonCuttingListConfigurationIsAsExpected(self):
        self.assertFalse(self.config.jointing)
        self.assertEqual(2.0, self.config.minJoint)
        self.assertEqual(5.0, self.config.maxJoint)
        self.assertFalse(self.config.accurateSaw)
        self.assertEqual('XG Export Log.txt', self.config.logFileName)
        self.assertEqual(['Spreadsheet (File)'], self.config.cuttingLists)
        self.assertEqual(10.0, self.config.cut_size_add)
        self.assertEqual(10.0, self.config.sizingAddition)
        self.assertEqual(5.5, self.config.additionalCoveringAddition)

    def testSpecificCuttingListConfigurationIsAsExpected(self):
        self.assertEqual(SizeMode.NC_COMPATIBLE, self.config.sizing)
        self.assertEqual(1, self.config.coordinatePrecision)
        self.assertEqual(4, self.config.angularPrecision)
        self.assertEqual('{DocumentDir}/cuttinglistexport', self.config.getBasePathPrefix())


class CuttinglistSizingModeTest(unittest.TestCase):
    def testCreateFromText_NCCompatible(self):
        self.assertEqual(SizeMode.NC_COMPATIBLE, createSizingModeFromText('nc-compatible'))

    def testCreateFromText_SawnSize(self):
        self.assertEqual(SizeMode.SAWN_SIZE, createSizingModeFromText('sawn-size'))

    def testCreateFromText_SawnSizeWithAddition(self):
        self.assertEqual(SizeMode.SAWN_SIZE_WITH_ADDITION, createSizingModeFromText('sawn-size-with-addition'))

    def testCreateFromText_FinishedSize(self):
        self.assertEqual(SizeMode.FINISHED_SIZE, createSizingModeFromText('finished-size'))

    def testCreateFromText_FinishedSizeWithAddition(self):
        self.assertEqual(SizeMode.FINISHED_SIZE_WITH_ADDITION, createSizingModeFromText('finished-size-with-addition'))

    def testCreateFromText_UnknownTextShouldCreateNC_COMPATIBLE(self):
        with unittest.mock.patch('logging.warning') as _patcher:  # We do not want the logging warning in stderr
            self.assertEqual(SizeMode.NC_COMPATIBLE, createSizingModeFromText('this-option-is-not-known-by-parser'))

    def testCreateFromText_UnknownTextShouldLogWarning(self):
        with unittest.mock.patch('logging.warning') as patcher:
            createSizingModeFromText('this-option-is-not-known-by-parser')
            self.assertTrue(patcher.called)
