# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections

import interiorcad.backend.exportdatasafe
from interiorcad.interface.keyvalueconfiguration import KeyValueConfiguration
from interiorcad.backend.localization import getLocalizedString, CUTTINGLIST_RESOURCEGROUP

NC_CUTTINGLIST_CONFIGURATIONS = "cuttinglist-configuration"
NC_KEY_VALUE_PAIR = "keyvaluepair"
NC_CUSTOM_FIELD_CONFIGURATIONS = "custom-columns"
NC_COLUMN_NAME = "column"

MOCK_CUTTINGLIST_CONFIGURATION = """<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
    <interiorcad-export>
      <cuttinglist-configuration/>
    </interiorcad-export>"""


CustomColumn = collections.namedtuple('CustomColumn', 'data title type parameters')
CustomParameters = collections.namedtuple('CustomParameters', 'key value mapping')


class CuttingListConfiguration(KeyValueConfiguration):
    def getMockDefaultXMLCode(self):
        return MOCK_CUTTINGLIST_CONFIGURATION

    def readFullXMLFile(self):
        self.configurations = dict((item['key'], item['value']) for item in self.getAllAttributesOfAllChildrenWithTag(NC_CUTTINGLIST_CONFIGURATIONS, NC_KEY_VALUE_PAIR))
        self.customColumns = [
            CustomColumn(item['data'], self.getTitleFrom(item), item['type'], self.getParametersFrom(item['data'], item.get('id', '-1')))
            for item in self.getAllAttributesOfAllChildrenWithTag(NC_CUSTOM_FIELD_CONFIGURATIONS, NC_COLUMN_NAME)
        ]

    def getCuttingListConfigurationFileName(self):
        if interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS:
            return interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS[0][1]

    def getFileName(self):
        return self.getCuttingListConfigurationFileName()

    def getParametersFrom(self, attributeKey, fieldID):
        parameters = []
        if len(self.root.findall('./' + NC_CUSTOM_FIELD_CONFIGURATIONS)) > 0:
            for parentNode in self.root.findall('./' + NC_CUSTOM_FIELD_CONFIGURATIONS):
                for element in list(parentNode):
                    if 'data' in element.attrib:
                        hasValidID = element.attrib['id'] == fieldID if 'id' in element.attrib else True
                        if element.attrib['data'] == attributeKey and hasValidID:
                            for childElement in element.iter():
                                if childElement.tag == "parameter":
                                    if "key" in childElement.attrib:
                                        mapping = dict()

                                        if childElement.attrib["key"] == "mapping":
                                            for mapElement in childElement.iter():
                                                if mapElement.tag == "map":
                                                    for entry in mapElement.iter():
                                                        if 'key' in entry.attrib:
                                                            mapping[entry.attrib["key"]] = entry.attrib["value"]

                                        parameters.append(CustomParameters(childElement.attrib["key"], childElement.attrib["value"], mapping))

        return parameters

    def getTitleFrom(self, xmlAttribute):
        title = xmlAttribute['title']
        if 'resid' in xmlAttribute:
            title = getLocalizedString(CUTTINGLIST_RESOURCEGROUP, int(xmlAttribute['resid']), title)

        return title

