# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

"""Functions for testing clockwise-ness of polygons (from three points on)

See also:
    https://en.wikipedia.org/wiki/Curve_orientation
    http://stackoverflow.com/questions/1165647/how-to-determine-if-a-list-of-polygon-points-are-in-clockwise-order
"""

import math
import unittest

# The functions are low-level, as interiorcad.backend.cmoduleadapters.contourpath.getContourPathWithDistanceOldSchool is using them

def isClockwise(startPosition, middlePosition, endPosition):
    # http://stackoverflow.com/questions/1165647/how-to-determine-if-a-list-of-polygon-points-are-in-clockwise-order
    area = math.fsum((
        (middlePosition.x - startPosition.x) * (middlePosition.y + startPosition.y),
        (endPosition.x - middlePosition.x) * (middlePosition.y + endPosition.y),
        (startPosition.x - endPosition.x) * (startPosition.y + endPosition.y),
    ))
    return area > 0


def getIndexOfMostExtremePoint(positions):
    # https://en.wikipedia.org/wiki/Curve_orientation
    index = None

    for i in range(len(positions)):
        if index is None or positions[i].x < positions[index].x:
            index = i
        elif positions[i].x == positions[index].x and positions[i].y < positions[index].y:
            index = i

    return index


def arePositionsClockwise(positions):
    # https://en.wikipedia.org/wiki/Curve_orientation
    length = len(positions)
    if length > 2:
        index = getIndexOfMostExtremePoint(positions)
        assert index is not None, "Cannot find most extreme point!"

        previousIndex = (index - 1) % length
        if positions[previousIndex] == positions[index]:
            previousIndex = (previousIndex - 1) % length  # We assume that only two points can be the same

        nextIndex = (index + 1) % length
        if positions[nextIndex] == positions[index]:
            nextIndex = (nextIndex + 1) % length  # We assume that only two points can be the same

        return isClockwise(positions[previousIndex], positions[index], positions[nextIndex])


class ThreePointTests(unittest.TestCase):
    # The tests are taken from interiorcad.backend.ncexportmigrated.ncradiussegment.NCRadiusSegmentTests

    class MockPoint:
        def __init__(self, x, y):
            self.x = x
            self.y = y

    def testCWSegmentShouldBeClockwise(self):
        startPt = self.MockPoint(150.0, 100.0)
        middlePt = self.MockPoint(185.355339, 85.355339)
        endPt = self.MockPoint(200.0, 50.0)

        self.assertTrue(isClockwise(startPt, middlePt, endPt))

    def testCCWSegmentShouldNotBeClockwise(self):
        startPt = self.MockPoint(200.0, 50.0)
        middlePt = self.MockPoint(185.355339, 85.355339)
        endPt = self.MockPoint(150.0, 100.0)

        self.assertFalse(isClockwise(startPt, middlePt, endPt))

    def testBigCWSegmentShouldBeClockwise(self):
        startPt = self.MockPoint(150.0, 100.0)
        middlePt = self.MockPoint(216.513, 116.513)
        endPt = self.MockPoint(200.0, 50.0)

        self.assertTrue(isClockwise(startPt, middlePt, endPt))

    def testBigCCWSegmentShouldNotBeClockwise(self):
        startPt = self.MockPoint(200.0, 50.0)
        middlePt = self.MockPoint(216.513, 116.513)
        endPt = self.MockPoint(150.0, 100.0)

        self.assertFalse(isClockwise(startPt, middlePt, endPt))

    def testAlternateSegmentShouldBeClockwise(self):
        startPt = self.MockPoint(200.0, 50.0)
        middlePt = self.MockPoint(185.355339, 14.644)
        endPt = self.MockPoint(150.0, 0.0)

        self.assertTrue(isClockwise(startPt, middlePt, endPt))

    def testAlternateSegmentShouldNotBeClockwise(self):
        startPt = self.MockPoint(200.0, 50.0)
        middlePt = self.MockPoint(164.644, 35.355)
        endPt = self.MockPoint(150.0, 0.0)

        self.assertFalse(isClockwise(startPt, middlePt, endPt))

    def testAlternateBigSegmentShouldNotBeClockwise(self):
        startPt = self.MockPoint(200.0, 50.0)
        middlePt = self.MockPoint(133.487, 66.513)
        endPt = self.MockPoint(150.0, 0.0)

        self.assertFalse(isClockwise(startPt, middlePt, endPt))

    def testAlternateBigSegmentShouldBeClockwise(self):
        startPt = self.MockPoint(200.0, 66.513)
        middlePt = self.MockPoint(216.513, 0.0)
        endPt = self.MockPoint(150.0, 16.513)

        self.assertTrue(isClockwise(startPt, middlePt, endPt))


class PolygonTests(unittest.TestCase):
    # TODO IT-250 more tests, esp. with concave parts would be nice -- NG
    class MockPoint:
        def __init__(self, x, y):
            self.x = x
            self.y = y

        def __eq__(self, rhs):
            return self.x == rhs.x and self.y == rhs.y

    def testALineShouldBeNone(self):
        positions = [
            self.MockPoint(0, 0),
            self.MockPoint(0, 100),
        ]

        self.assertIsNone(arePositionsClockwise(positions))

    def testNonClosedRectangleCW(self):
        positions = [
            self.MockPoint(0, 0),
            self.MockPoint(0, 100),
            self.MockPoint(200, 100),
            self.MockPoint(200, 0),
        ]

        self.assertTrue(arePositionsClockwise(positions))

    def testClosedRectangleCW(self):
        positions = [
            self.MockPoint(0, 0),
            self.MockPoint(0, 100),
            self.MockPoint(200, 100),
            self.MockPoint(200, 0),
            self.MockPoint(0, 0),
        ]

        self.assertTrue(arePositionsClockwise(positions))

    def testNonClosedRectangleCCW(self):
        positions = [
            self.MockPoint(0, 0),
            self.MockPoint(200, 0),
            self.MockPoint(200, 100),
            self.MockPoint(0, 100),
        ]

        self.assertFalse(arePositionsClockwise(positions))

    def testClosedRectangleCCW(self):
        positions = [
            self.MockPoint(0, 0),
            self.MockPoint(200, 0),
            self.MockPoint(200, 100),
            self.MockPoint(0, 100),
            self.MockPoint(0, 0),
        ]

        self.assertFalse(arePositionsClockwise(positions))
