# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest
from collections import namedtuple

from interiorcad.interface.geometry import isNearlyEqual

EdgeJoints = namedtuple('EdgeJoints', 'start end')


class EdgeData:
    JOINT_MITER = "Miter"
    JOINT_CONTINUOUS = "Continuous"
    JOINT_DISCONTINUOUS = "Discontinuous"
    JOINT_NOEDGE = ""

    def __init__(self, supplierID='', thickness=0.0, length=0.0, jointStart=JOINT_NOEDGE, jointEnd=JOINT_NOEDGE, isPolygonal=False):
        self.supplierID = supplierID
        self.thickness = thickness
        self.length = length
        self.jointStart = jointStart
        self.jointEnd = jointEnd
        self.isPolygonal = isPolygonal  # Available in the datatracker from interiorcad 2022 F4 on

    def __repr__(self):
        return "EdgeData(supplierID='{supplierID}', thickness={thickness}, length={length}, jointStart={jointStart}, jointEnd={jointEnd})".format(
            supplierID=self.supplierID, thickness=self.thickness, length=self.length, jointStart=self.jointStart, jointEnd=self.jointEnd)

    def __eq__(self, rhs):
        return all((
            self.supplierID == rhs.supplierID,
            isNearlyEqual(self.thickness, rhs.thickness),
            isNearlyEqual(self.length, rhs.length),
            self.jointStart == rhs.jointStart,
            self.jointEnd == rhs.jointEnd,
            self.isPolygonal == rhs.isPolygonal
        ))

    @property
    def jointStartTechnical(self):
        return self.jointStart if self.jointStart != EdgeData.JOINT_NOEDGE or not self.supplierID else EdgeData.JOINT_DISCONTINUOUS

    @property
    def jointEndTechnical(self):
        return self.jointEnd if self.jointEnd != EdgeData.JOINT_NOEDGE or not self.supplierID else EdgeData.JOINT_DISCONTINUOUS


class Edges:
    def __init__(self, polygonal=None, rectangular=None, edgecode=''):
        self.polygonal = polygonal
        self.rectangular = rectangular
        self.edgecode = edgecode

    def __repr__(self):
        return "Edges(polygonal={polygonal}, rectangular={rectangular}, edgecode={edgecode})".format(polygonal=self.polygonal, rectangular=self.rectangular, edgecode=self.edgecode)

    def __eq__(self, rhs):
        return self.polygonal == rhs.polygonal and self.rectangular == rhs.rectangular

    def getPolygonalExistingEdgeIDs(self):
        return (str(index) for index, edge in enumerate(self.polygonal, start=1) if edge.supplierID)

    def getPolygonalExistingEdgeSupplierIDs(self):
        return (edge.supplierID for edge in self.polygonal if edge.supplierID)

    def getPolygonalExistingEdgeJoints(self):
        return (EdgeJoints(start=edge.jointStart, end=edge.jointEnd) for edge in self.polygonal if edge.supplierID)

    def getPolygonalExistingEdgeThicknesses(self):
        return (edge.thickness for edge in self.polygonal if edge.supplierID)

    def getPolygonalExistingEdgeLengths(self):
        return (edge.length for edge in self.polygonal if edge.supplierID)

    def getFirstValidPolygonalEdgeLegacy(self):
        polygonalEdge = EdgeData()
        for edge in self.polygonal:
            if edge.supplierID:
                polygonalEdge = edge
                break

        return polygonalEdge

    def getRectangularEdge(self, rectangularEdgeIndex):
        if rectangularEdgeIndex < 1 or rectangularEdgeIndex > 4:
            raise IndexError("rectangularEdgeIndex out of range: {}".format(rectangularEdgeIndex))

        valueByIndex = (None, lambda: self.front, lambda: self.back, lambda: self.left, lambda: self.right)
        return valueByIndex[rectangularEdgeIndex]()

    def __ensureNotNoneEdge(self, edge):
        if edge is not None:
            return edge
        else:
            return EdgeData()

    def ensureNotNoneRectangularEdges(self):
        self.rectangular = list(map(lambda edge: self.__ensureNotNoneEdge(edge), self.rectangular))

    @property
    def front(self):
        if self.rectangular:
            return self.rectangular[0]

    @front.setter
    def front(self, edge):
        if self.rectangular:
            self.rectangular[0] = edge

    @property
    def back(self):
        if self.rectangular:
            return self.rectangular[1]

    @back.setter
    def back(self, edge):
        if self.rectangular:
            self.rectangular[1] = edge

    @property
    def left(self):
        if self.rectangular:
            return self.rectangular[2]

    @left.setter
    def left(self, edge):
        if self.rectangular:
            self.rectangular[2] = edge

    @property
    def right(self):
        if self.rectangular:
            return self.rectangular[3]

    @right.setter
    def right(self, edge):
        if self.rectangular:
            self.rectangular[3] = edge


class EdgesCreationTest(unittest.TestCase):
    def setUp(self):
        self.edges = Edges()

    def testDefaultEdgesHaveNoRectangularAndPolygonalEdges(self):
        self.assertEqual(None, self.edges.polygonal)
        self.assertEqual(None, self.edges.rectangular)

    def testGetPolygonalExistingEdgeIDsShouldReturnEmptyStringWhenPolygonalIsEmpty(self):
        self.edges.polygonal = []
        self.assertEqual('', ', '.join(self.edges.getPolygonalExistingEdgeIDs()))

    def testGetPolygonalExistingEdgeIDsShouldReturnThreeEdges(self):
        self.edges.polygonal = [
            EdgeData('supplierID', 10, 500),
            EdgeData('supplierID', 10, 700),
            EdgeData('supplierID', 10, 200),
        ]
        self.assertEqual('1, 2, 3', ', '.join(self.edges.getPolygonalExistingEdgeIDs()))

    def testGetPolygonalExistingEdgeIDsShouldReturnFourEdges(self):
        self.edges.polygonal = [
            EdgeData('', 0, 0),
            EdgeData('supplierID', 10, 700),
            EdgeData('supplierID', 10, 200),
            EdgeData('supplierID', 10, 500),
            EdgeData('', 0, 0),
            EdgeData('supplierID', 10, 500),
        ]
        self.assertEqual('2, 3, 4, 6', ', '.join(self.edges.getPolygonalExistingEdgeIDs()))

    def testGetPolygonalSupplierIDShouldReturnEmptyStringWhenPolygonalIsEmpty(self):
        self.edges.polygonal = []
        self.assertEqual('', self.edges.getFirstValidPolygonalEdgeLegacy().supplierID)

    def testGetPolygonalSupplierIDShouldReturnCorrectValue(self):
        self.edges.polygonal = [
            EdgeData('', 0, 0),
            EdgeData('supplierID', 10, 700),
            EdgeData('supplierID', 10, 200),
            EdgeData('supplierID', 10, 500),
            EdgeData('', 0, 0),
            EdgeData('supplierID', 10, 500),
        ]
        self.assertEqual('supplierID', self.edges.getFirstValidPolygonalEdgeLegacy().supplierID)

    def testGetPolygonalExistingEdgeSupplierIDsShouldReturnFourEdges(self):
        self.edges.polygonal = [
            EdgeData('', 0, 0),
            EdgeData('supplierID2', 10, 700),
            EdgeData('supplierID3', 10, 200),
            EdgeData('supplierID4', 10, 500),
            EdgeData('', 0, 0),
            EdgeData('supplierID6', 10, 500),
        ]
        self.assertEqual('supplierID2, supplierID3, supplierID4, supplierID6', ', '.join(self.edges.getPolygonalExistingEdgeSupplierIDs()))


class EdgeJointTests(unittest.TestCase):
    def setUp(self):
        polygonalEdges = [
            EdgeData('', 0, 0),
            EdgeData('supplierID2', 10, 700, EdgeData.JOINT_NOEDGE, EdgeData.JOINT_MITER),
            EdgeData('supplierID3', 11, 200, EdgeData.JOINT_MITER, EdgeData.JOINT_CONTINUOUS),
            EdgeData('supplierID4', 12, 500, EdgeData.JOINT_DISCONTINUOUS, EdgeData.JOINT_NOEDGE),
            EdgeData('', 0, 0),
            EdgeData('supplierID6', 13, 500, EdgeData.JOINT_NOEDGE, EdgeData.JOINT_NOEDGE),
        ]
        self.edges = Edges(polygonal=polygonalEdges)

    def testGetPolygonalExistingEdgeJoints(self):
        expected = (
            EdgeJoints(start=EdgeData.JOINT_NOEDGE, end=EdgeData.JOINT_MITER),
            EdgeJoints(start=EdgeData.JOINT_MITER, end=EdgeData.JOINT_CONTINUOUS),
            EdgeJoints(start=EdgeData.JOINT_DISCONTINUOUS, end=EdgeData.JOINT_NOEDGE),
            EdgeJoints(start=EdgeData.JOINT_NOEDGE, end=EdgeData.JOINT_NOEDGE),
        )

        self.assertEqual(expected, tuple(self.edges.getPolygonalExistingEdgeJoints()))

    def testGetPolygonalExistingEdgeThicknesses(self):
        expected = (10, 11, 12, 13)

        self.assertEqual(expected, tuple(self.edges.getPolygonalExistingEdgeThicknesses()))

    def testGetPolygonalExistingEdgeLengths(self):
        expected = (700, 200, 500, 500)

        self.assertEqual(expected, tuple(self.edges.getPolygonalExistingEdgeLengths()))


class RectangularPropertyTests(unittest.TestCase):
    def setUp(self):
        self.edges = Edges(
            polygonal=['poly0', 'poly1', 'poly2', 'poly3'],
            rectangular=['edge1', 'edge2', 'edge3', 'edge4'],
        )

    def testGetFront(self):
        self.assertEqual('edge1', self.edges.front)

    def testGetBack(self):
        self.assertEqual('edge2', self.edges.back)

    def testGetLeft(self):
        self.assertEqual('edge3', self.edges.left)

    def testGetRight(self):
        self.assertEqual('edge4', self.edges.right)

    def testSetFront(self):
        self.edges.front = 'changed'
        self.assertEqual(['changed', 'edge2', 'edge3', 'edge4'], self.edges.rectangular)

    def testSetBack(self):
        self.edges.back = 'changed'
        self.assertEqual(['edge1', 'changed', 'edge3', 'edge4'], self.edges.rectangular)

    def testSetLeft(self):
        self.edges.left = 'changed'
        self.assertEqual(['edge1', 'edge2', 'changed', 'edge4'], self.edges.rectangular)

    def testSetRight(self):
        self.edges.right = 'changed'
        self.assertEqual(['edge1', 'edge2', 'edge3', 'changed'], self.edges.rectangular)


class EnsureNotNoneTests(unittest.TestCase):
    def setUp(self):
        self.edges = Edges(rectangular=[None, None, None, None])

    def testEnsureNotNoneRectangularEdges(self):
        self.edges.ensureNotNoneRectangularEdges()
        self.assertEqual([EdgeData(), EdgeData(), EdgeData(), EdgeData()], self.edges.rectangular)


class EdgesEqualityTests(unittest.TestCase):
    def testTwoEmptyEdgesShouldBeEqual(self):
        lhs = Edges()
        rhs = Edges()

        self.assertTrue(lhs == rhs)
        self.assertEqual(lhs, rhs)

    def testTwoSameEdgesShouldBeEqual(self):
        lhs = Edges(
            polygonal=['poly0', 'poly1', 'poly2', 'poly3'],
            rectangular=['edge1', 'edge2', 'edge3', 'edge4'],
        )
        rhs = Edges(
            polygonal=['poly0', 'poly1', 'poly2', 'poly3'],
            rectangular=['edge1', 'edge2', 'edge3', 'edge4'],
        )

        self.assertTrue(lhs == rhs)
        self.assertEqual(lhs, rhs)

    def testTwoEdgesDiffereingInPolygonalShouldNotBeEqual(self):
        lhs = Edges(
            polygonal=['poly0', 'poly1', 'poly2', 'poly3'],
            rectangular=['edge1', 'edge2', 'edge3', 'edge4'],
        )
        rhs = Edges(
            polygonal=['polyOther', 'poly1', 'poly2', 'poly3'],
            rectangular=['edge1', 'edge2', 'edge3', 'edge4'],
        )

        self.assertFalse(lhs == rhs)
        self.assertNotEqual(lhs, rhs)

    def testTwoEdgesDiffereingInRectangularShouldNotBeEqual(self):
        lhs = Edges(
            polygonal=['poly0', 'poly1', 'poly2', 'poly3'],
            rectangular=['edge1', 'edge2', 'edge3', 'edge4'],
        )
        rhs = Edges(
            polygonal=['poly0', 'poly1', 'poly2', 'poly3'],
            rectangular=['edge1Other', 'edge2', 'edge3', 'edge4'],
        )

        self.assertFalse(lhs == rhs)
        self.assertNotEqual(lhs, rhs)


class TechnicalJointingTests(unittest.TestCase):
    def testEdgeWithDefinedJointsShouldHaveTheSameTechnicalJoints(self):
        edgeData = EdgeData('supplierID', 10.0, 700.0, EdgeData.JOINT_CONTINUOUS, EdgeData.JOINT_DISCONTINUOUS)
        self.assertEqual(EdgeData.JOINT_CONTINUOUS, edgeData.jointStartTechnical)
        self.assertEqual(EdgeData.JOINT_DISCONTINUOUS, edgeData.jointEndTechnical)

    def testEdgeWithoutJointAtStartShouldHaveTechnicalJoints(self):
        edgeData = EdgeData('supplierID', 10.0, 700.0, EdgeData.JOINT_NOEDGE, EdgeData.JOINT_CONTINUOUS)
        self.assertEqual(EdgeData.JOINT_DISCONTINUOUS, edgeData.jointStartTechnical)
        self.assertEqual(EdgeData.JOINT_CONTINUOUS, edgeData.jointEndTechnical)

    def testEdgeWithoutJointAtEndShouldHaveTechnicalJoints(self):
        edgeData = EdgeData('supplierID', 10.0, 700.0, EdgeData.JOINT_CONTINUOUS, EdgeData.JOINT_NOEDGE)
        self.assertEqual(EdgeData.JOINT_CONTINUOUS, edgeData.jointStartTechnical)
        self.assertEqual(EdgeData.JOINT_DISCONTINUOUS, edgeData.jointEndTechnical)

    def testEdgeWithoutJointsShouldHaveTechnicalJoints(self):
        edgeData = EdgeData('supplierID', 10.0, 700.0, EdgeData.JOINT_NOEDGE, EdgeData.JOINT_NOEDGE)
        self.assertEqual(EdgeData.JOINT_DISCONTINUOUS, edgeData.jointStartTechnical)
        self.assertEqual(EdgeData.JOINT_DISCONTINUOUS, edgeData.jointEndTechnical)

    def testEmptyEdgeShouldHaveNoTechnicalJointings(self):
        edgeData = EdgeData('', 0.0, 0.0, EdgeData.JOINT_NOEDGE, EdgeData.JOINT_NOEDGE)
        self.assertEqual(EdgeData.JOINT_NOEDGE, edgeData.jointStartTechnical)
        self.assertEqual(EdgeData.JOINT_NOEDGE,  edgeData.jointEndTechnical)
