# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest

from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.vector import Vector


class Matrix:
    def __init__(self, u=None, v=None, w=None, offset=None):
        self.u = u
        self.v = v
        self.w = w
        self.origin = offset

        if u is None:
            self.u = Vector(x=1.0)
        if v is None:
            self.v = Vector(y=1.0)
        if w is None:
            self.w = Vector(z=1.0)
        if offset is None:
            self.origin = Vector()

    def __eq__(self, other):
        return (
            self.u == other.u and
            self.v == other.v and
            self.w == other.w and
            self.origin == other.origin
        )

    def rawString(self):
        return 'Matrix(u={}, v={}, w={}, origin={})'

    def __str__(self):
        return self.rawString().format(str(self.u), str(self.v), str(self.w),
                                       str(self.origin))

    def __repr__(self):
        return self.rawString().format(repr(self.u), repr(self.v),
                                       repr(self.w), repr(self.origin))

    def isZVectorPositiveInZDirection(self):
        return self.w.z > 0.0

    def __sub__(self, other):
        u = self.u - other.u
        v = self.v - other.v
        w = self.w - other.w
        offset = self.origin - other.origin
        return Matrix(u, v, w, offset)

    def __abs__(self):
        return Matrix(abs(self.u), abs(self.v), abs(self.w), abs(self.origin))

    def __round__(self, places):
        return Matrix(round(self.u, places), round(self.v, places),
                      round(self.w, places), round(self.origin, places))

    def __imul__(self, rhs):
        matrix = self * rhs
        self.u = matrix.u
        self.v = matrix.v
        self.w = matrix.w
        self.origin = matrix.origin

    # IT-228 (HORIZONTAL) UnitTests and ezdxf UnitTests -- NG
    def __mul__(self, rhs):
        matrix = Matrix()

        matrix.u.x = math.fsum((self.u.x * rhs.u.x, self.u.y * rhs.v.x, self.u.z * rhs.w.x))
        matrix.u.y = math.fsum((self.u.x * rhs.u.y, self.u.y * rhs.v.y, self.u.z * rhs.w.y))
        matrix.u.z = math.fsum((self.u.x * rhs.u.z, self.u.y * rhs.v.z, self.u.z * rhs.w.z))

        matrix.v.x = math.fsum((self.v.x * rhs.u.x, self.v.y * rhs.v.x, self.v.z * rhs.w.x))
        matrix.v.y = math.fsum((self.v.x * rhs.u.y, self.v.y * rhs.v.y, self.v.z * rhs.w.y))
        matrix.v.z = math.fsum((self.v.x * rhs.u.z, self.v.y * rhs.v.z, self.v.z * rhs.w.z))

        matrix.w.x = math.fsum((self.w.x * rhs.u.x, self.w.y * rhs.v.x, self.w.z * rhs.w.x))
        matrix.w.y = math.fsum((self.w.x * rhs.u.y, self.w.y * rhs.v.y, self.w.z * rhs.w.y))
        matrix.w.z = math.fsum((self.w.x * rhs.u.z, self.w.y * rhs.v.z, self.w.z * rhs.w.z))

        matrix.origin.x = math.fsum((self.origin.x * rhs.u.x, self.origin.y * rhs.v.x, self.origin.z * rhs.w.x, rhs.origin.x))
        matrix.origin.y = math.fsum((self.origin.x * rhs.u.y, self.origin.y * rhs.v.y, self.origin.z * rhs.w.y, rhs.origin.y))
        matrix.origin.z = math.fsum((self.origin.x * rhs.u.z, self.origin.y * rhs.v.z, self.origin.z * rhs.w.z, rhs.origin.z))

        return matrix

    # A matrix is flipped if the determinante is -1. Since we have a 3x3 matrix we can use the triple product(Spatprodukt) to determine whether the matrix is flipped.
    # See http://de.wikipedia.org/wiki/Spiegelungsmatrix for further information.
    def isFlipped(self):
        return (self.u.cross(self.v)).dot(self.w) < 0

    def getInverted(self):
        # https://fr.wikipedia.org/wiki/Matrice_inversible#Inversion_des_matrices_3%C3%973
        # (See also https://en.wikipedia.org/wiki/Invertible_matrix#Inversion_of_3_%C3%97_3_matrices )
        a = self.u.x
        b = self.v.x
        c = self.w.x
        d = self.u.y
        e = self.v.y
        f = self.w.y
        g = self.u.z
        h = self.v.z
        i = self.w.z

        determinant = math.fsum((a*e*i, b*f*g, c*d*h, -c*e*g, - f*h*a, - i*b*d))
        assert not isNearlyEqual(determinant, 0.0), "Determinant should not be nearly zero!"

        matrix = Matrix()

        matrix.u.x = math.fsum((e*i, -f*h)) / determinant
        matrix.v.x = math.fsum((c*h, -b*i)) / determinant
        matrix.w.x = math.fsum((b*f, -c*e)) / determinant
        matrix.u.y = math.fsum((f*g, -d*i)) / determinant
        matrix.v.y = math.fsum((a*i, -c*g)) / determinant
        matrix.w.y = math.fsum((c*d, -a*f)) / determinant
        matrix.u.z = math.fsum((d*h, -e*g)) / determinant
        matrix.v.z = math.fsum((b*g, -a*h)) / determinant
        matrix.w.z = math.fsum((a*e, -b*d)) / determinant

        # We handle the offset the same way VW does

        matrix.origin.x = - math.fsum((matrix.u.x * self.origin.x, matrix.v.x * self.origin.y, matrix.w.x * self.origin.z))
        matrix.origin.y = - math.fsum((matrix.u.y * self.origin.x, matrix.v.y * self.origin.y, matrix.w.y * self.origin.z))
        matrix.origin.z = - math.fsum((matrix.u.z * self.origin.x, matrix.v.z * self.origin.y, matrix.w.z * self.origin.z))

        return matrix

    def rotateZPlus90(self):
        ux = - self.u.y
        vx = - self.v.y
        wx = - self.w.y

        uy = self.u.x
        vy = self.v.x
        wy = self.w.x

        self.u.x, self.u.y = ux, uy
        self.v.x, self.v.y = vx, vy
        self.w.x, self.w.y = wx, wy

        self.origin.x, self.origin.y = -self.origin.y, self.origin.x

    def rotateZMinus90(self):
        ux = self.u.y
        vx = self.v.y
        wx = self.w.y

        uy = - self.u.x
        vy = - self.v.x
        wy = - self.w.x

        self.u.x, self.u.y = ux, uy
        self.v.x, self.v.y = vx, vy
        self.w.x, self.w.y = wx, wy

        self.origin.x, self.origin.y = self.origin.y, -self.origin.x

    def rotateZ180(self):
        self.u.x = - self.u.x
        self.v.x = - self.v.x
        self.w.x = - self.w.x

        self.u.y = - self.u.y
        self.v.y = - self.v.y
        self.w.y = - self.w.y

        self.origin.x, self.origin.y = - self.origin.x, -self.origin.y

    def transform3D(self, positionX, positionY, positionZ):
        return (
            math.fsum((self.u.x * positionX, self.v.x * positionY, self.w.x * positionZ, self.origin.x)),
            math.fsum((self.u.y * positionX, self.v.y * positionY, self.w.y * positionZ, self.origin.y)),
            math.fsum((self.u.z * positionX, self.v.z * positionY, self.w.z * positionZ, self.origin.z)),
        )

    def isIdentity(self):
        return (
            self.u == Vector(x=1.0) and
            self.v == Vector(y=1.0) and
            self.w == Vector(z=1.0) and
            self.origin == Vector()
        )

    def getTranslatedBefore(self, x, y, z):
        newMatrix = Matrix()
        newMatrix.origin = Vector(x, y, z)
        newMatrix = newMatrix * self
        return newMatrix


class MatrixCreationTest(unittest.TestCase):
    def setUp(self):
        self.m = Matrix()

    def testDefaultMatrixIsIdentityMatrix(self):
        self.assertEqual(Vector(1, 0, 0), self.m.u)
        self.assertEqual(Vector(0, 1, 0), self.m.v)
        self.assertEqual(Vector(0, 0, 1), self.m.w)
        self.assertEqual(Vector(0, 0, 0), self.m.origin)

    def testSetU(self):
        self.m.u = 'u'
        self.assertEqual('u', self.m.u)
        self.assertEqual(Vector(0, 1, 0), self.m.v)
        self.assertEqual(Vector(0, 0, 1), self.m.w)
        self.assertEqual(Vector(0, 0, 0), self.m.origin)

    def testSetV(self):
        self.m.v = 'v'
        self.assertEqual(Vector(1, 0, 0), self.m.u)
        self.assertEqual('v', self.m.v)
        self.assertEqual(Vector(0, 0, 1), self.m.w)
        self.assertEqual(Vector(0, 0, 0), self.m.origin)

    def testSetW(self):
        self.m.w = 'w'
        self.assertEqual(Vector(1, 0, 0), self.m.u)
        self.assertEqual(Vector(0, 1, 0), self.m.v)
        self.assertEqual('w', self.m.w)
        self.assertEqual(Vector(0, 0, 0), self.m.origin)

    def testSetOffset(self):
        self.m.origin = 'origin'
        self.assertEqual(Vector(1, 0, 0), self.m.u)
        self.assertEqual(Vector(0, 1, 0), self.m.v)
        self.assertEqual(Vector(0, 0, 1), self.m.w)
        self.assertEqual('origin', self.m.origin)

    def testSetAllVectorsInConstructor(self):
        m = Matrix('u', 'v', 'w', 'origin')
        self.assertEqual('u', m.u)
        self.assertEqual('v', m.v)
        self.assertEqual('w', m.w)
        self.assertEqual('origin', m.origin)

    def testStringRepresentation(self):
        m = Matrix('u', 'v', 'w', 'origin')
        self.assertEqual('Matrix(u=u, v=v, w=w, origin=origin)', str(m))
        self.assertEqual("Matrix(u='u', v='v', w='w', origin='origin')", repr(m))


class MatrixComparisonTest(unittest.TestCase):
    def testEmptyMatricesAreEqual(self):
        self.assertEqual(Matrix(), Matrix())

    def testNonEmptyMatricesAreEqual(self):
        m1 = Matrix(Vector(x=1), Vector(y=1), Vector(z=1), Vector(10, 20, 30))
        m2 = Matrix(Vector(x=1), Vector(y=1), Vector(z=1), Vector(10, 20, 30))
        self.assertEqual(m1, m2)

    def testSameMatrixIsEqualWithItself(self):
        m1 = Matrix()
        self.assertEqual(m1, m1)


class MatrixMathTest(unittest.TestCase):
    def setUp(self):
        self.m1 = Matrix(1, 2, 3, 4)
        self.m2 = Matrix(5, 5, 5, 5)
        self.mExpected = Matrix(4, 3, 2, 1)

    def testSubtractMatrices(self):
        self.assertEqual(self.mExpected, self.m2 - self.m1)

    def testAbsolutePositiveMatrix(self):
        self.assertEqual(self.m1, abs(self.m1))

    def testAbsoluteNegativeMatrix(self):
        m = Matrix(-1, -2, -3, -4)
        self.assertEqual(self.m1, abs(m))
        self.assertNotEqual(m, abs(m))

    def testRoundMatrix(self):
        m = Matrix(1.001, 2.002, 3.003, 4.004)
        expected = Matrix(1, 2, 3, 4)
        self.assertEqual(expected, round(m, 1))


# The following test classes should remain synchron to the tests in ezdxtmatrix44
class IsFlippedTests(unittest.TestCase):
    def testIsFlippedDiagonalCaseOne(self):
        matrix = Matrix(
            u=Vector(0.681508, 0.0, -0.731811),
            v=Vector(0.731811, 0.0, 0.681508),
            w=Vector(0.0, 1.0, 0.0)
        )

        self.assertTrue(matrix.isFlipped())

    def testIsNotFlippedCaseTwo(self):
        matrix = Matrix(
            u=Vector(0.731811, 0.0, 0.681508),
            v=Vector(0.681508, 0.0, -0.731811),
            w=Vector(0.0, 1.0, 0.0)
        )

        self.assertFalse(matrix.isFlipped())

    def testIsFlippedXAxisCaseThree(self):
        matrix = Matrix(
            u=Vector(1.0, 0.0, 0.0),
            v=Vector(0.0, 0.0, 1.0),
            w=Vector(0.0, 1.0, 0.0)
        )

        self.assertTrue(matrix.isFlipped())

    def testIsNotFlippedCaseFour(self):
        matrix = Matrix(
            u=Vector(-1.0, 0.0, 0.0),
            v=Vector(0.0, 0.0, 1.0),
            w=Vector(0.0, 1.0, 0.0)
        )

        self.assertFalse(matrix.isFlipped())

    def testIsNotFlippedCaseFive(self):
        matrix = Matrix(
            u=Vector(1.0, 0.0, 0.0),
            v=Vector(0.0, 0.0, -1.0),
            w=Vector(0.0, 1.0, 0.0)
        )

        self.assertFalse(matrix.isFlipped())

    def testIsFlippedYAxisCaseSix(self):
        matrix = Matrix(
            u=Vector(-1.0, 0.0, 0.0),
            v=Vector(0.0, 0.0, -1.0),
            w=Vector(0.0, 1.0, 0.0)
        )

        self.assertTrue(matrix.isFlipped())

    def testIsNotFlippedCaseSeven(self):
        matrix = Matrix(
            u=Vector(1.0, 0.0, 0.0),
            v=Vector(0.0, 0.0, 1.0),
            w=Vector(0.0, -1.0, 0.0)
        )

        self.assertFalse(matrix.isFlipped())

    def testIsFlippedZAxesCaseEight(self):
        matrix = Matrix(
            u=Vector(1.0, 0.0, 0.0),
            v=Vector(0.0, 0.0, -1.0),
            w=Vector(0.0, -1.0, 0.0)
        )

        self.assertTrue(matrix.isFlipped())


# The Tests should remain synchron not only to the tests in ezdxtmatrix44, but also to the C++ Tests of XGTransformMatrix using the fixture PythonMatrixInvertTestFixture
class InversionTests(unittest.TestCase):
    def testIdentity(self):
        matrix = Matrix()

        invertedMatrix = matrix.getInverted()

        expectedMatrix = Matrix()

        self.assertEqual(expectedMatrix, invertedMatrix)

    def testIdentityWithOffset(self):
        matrix = Matrix(offset=Vector(11, 22, 33))

        invertedMatrix = matrix.getInverted()

        expectedMatrix = Matrix(offset=Vector(-11, -22, -33))

        self.assertEqual(expectedMatrix, invertedMatrix)

    def testXYRotation45(self):
        matrix = Matrix()
        matrix.u.x = math.cos(math.pi / 4)
        matrix.u.y = math.sin(math.pi / 4)
        matrix.v.x = - math.sin(math.pi / 4)
        matrix.v.y = math.cos(math.pi / 4)

        invertedMatrix = matrix.getInverted()

        expectedMatrix = Matrix()
        expectedMatrix.u.x = math.cos(- math.pi / 4)
        expectedMatrix.u.y = math.sin(- math.pi / 4)
        expectedMatrix.v.x = - math.sin(- math.pi / 4)
        expectedMatrix.v.y = math.cos(- math.pi / 4)

        self.assertEqual(expectedMatrix, invertedMatrix)


class RotateTests(unittest.TestCase):
    def testIdentityRotatedPlus90(self):
        matrix = Matrix()
        matrix.rotateZPlus90()

        expectedMatrix = Matrix(Vector(0, 1, 0), Vector(-1, 0, 0), Vector(0, 0, 1))

        self.assertEqual(expectedMatrix, matrix)

    def testIdentityRotatedMinus90(self):
        matrix = Matrix()
        matrix.rotateZMinus90()

        expectedMatrix = Matrix(Vector(0, -1, 0), Vector(1, 0, 0), Vector(0, 0, 1))

        self.assertEqual(expectedMatrix, matrix)

    def testIdentityRotated180(self):
        matrix = Matrix()
        matrix.rotateZ180()

        expectedMatrix = Matrix(Vector(-1, 0, 0), Vector(0, -1, 0), Vector(0, 0, 1))

        self.assertEqual(expectedMatrix, matrix)


class transform3D(unittest.TestCase):
    def testExchangeXZWithoutOffset(self):
        matrix = Matrix(
            u=Vector(0, 0, 1),
            v=Vector(0, 1, 0),
            w=Vector(1, 0, 0)
        )

        self.assertEqual((33, 22, 11), matrix.transform3D(11, 22, 33))

    def testExchangeXZWithOffset(self):
        matrix = Matrix(
            u=Vector(0, 0, 1),
            v=Vector(0, 1, 0),
            w=Vector(1, 0, 0),
            offset=Vector(1000, 2000, 3000)
        )

        self.assertEqual((1033, 2022, 3011), matrix.transform3D(11, 22, 33))

    def testExchangeXYWithoutOffset(self):
        matrix = Matrix(
            u=Vector(0, 1, 0),
            v=Vector(1, 0, 0),
            w=Vector(0, 0, 1)
        )

        self.assertEqual((22, 11, 33), matrix.transform3D(11, 22, 33))

    def testExchangeXYWithOffset(self):
        matrix = Matrix(
            u=Vector(0, 1, 0),
            v=Vector(1, 0, 0),
            w=Vector(0, 0, 1),
            offset=Vector(1000, 2000, 3000)
        )

        self.assertEqual((1022, 2011, 3033), matrix.transform3D(11, 22, 33))


class IdentityTests(unittest.TestCase):
    def testDefaultMatrixIsIdentity(self):
        matrix = Matrix()
        self.assertTrue(matrix.isIdentity())

    def testDefaultWithAnOffsetMatrixIsNotIdentity(self):
        matrix = Matrix(offset=Vector(x=1, y=2, z=3))
        self.assertFalse(matrix.isIdentity())

    def testMatrixUIsNotIdentity(self):
        matrix = Matrix(u=Vector(x=1, y=2, z=3))
        self.assertFalse(matrix.isIdentity())

    def testMatrixVIsNotIdentity(self):
        matrix = Matrix(v=Vector(x=1, y=2, z=3))
        self.assertFalse(matrix.isIdentity())

    def testMatrixWIsNotIdentity(self):
        matrix = Matrix(w=Vector(x=1, y=2, z=3))
        self.assertFalse(matrix.isIdentity())
