# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.surface.acangles import calculateAAngleFromWVector, calculateCAngleToXYPlaneFromWVector


class Plane:
    # origin and normalVector were used up to interiorcad 2019 F3 (and in unit-tests)
    def __init__(self, origin=None, normalVector=None, *, sideID=None, predefined=False, matrix=None, wopMatrix=None):
        self.sideID = sideID
        self.predefined = predefined

        if matrix:
            self.matrix = matrix
        elif origin is None and normalVector is None:
            self.matrix = None
        else:
            u = normalVector.cross(Vector(0, 0, 1))
            v = u.cross(normalVector)

            self.matrix = Matrix(u, v, -normalVector, origin)

        if wopMatrix:
            self.wopMatrix = wopMatrix
        else:
            self.wopMatrix = self.matrix

    # New in interiorcad 2019 F4
    def isCompatiblePlane(self, otherPlane):
        return all((
                self.isPointOnPlane(otherPlane.matrix.origin),
                self.matrix.w == otherPlane.matrix.w,
        ))

    # Old name of function (up to interiorcad 2019 F3). Kept for source compatibility reasons. Use isCompatiblePlane instead
    # In interiorcad 2019 F3, this function checked also the complete matrix. The code is now reverted to the original design of upto interiorcad 2019 F2
    def isSamePlane(self, otherPlane):
        return self.isCompatiblePlane(otherPlane)

    def isPointOnPlane(self, point):
        # Point-Normal-Form of a plane - http://en.wikipedia.org/wiki/Plane_(geometry)
        # IC-24241: use normalized for (point - self.origin) to avoid enormous loss of precision if the resulting vector would be very small.
        # (Using normalized forces to check explicitly point == self.origin )
        return point == self.origin or isNearlyEqual(self.normalVector.dot((point - self.origin).normalized()), 0)

    def __str__(self):
        return "Plane: matrix=" + str(self.matrix)

    @property
    def normalVector(self):
        return -self.matrix.w

    @property
    def origin(self):
        return self.matrix.origin

    @staticmethod
    def createFromMachining(machining):
        return Plane.createFromMatrix(Plane.createPlaneMatrixFromMachining(machining))

    @staticmethod
    def createPlaneMatrixFromMachining(machining):
        normalVector = - machining.matrix.w

        matrix = Matrix()
        matrix.u = normalVector.cross(Vector(0, 0, 1))
        matrix.v = matrix.u.cross(normalVector)
        matrix.w = -copy.deepcopy(normalVector)
        matrix.origin = copy.deepcopy(machining.matrix.origin)

        return matrix

    @staticmethod
    def createFromMatrix(matrix):
        return Plane(matrix=matrix)

    @property
    def aAngle(self):
        return calculateAAngleFromWVector(wvector=-self.normalVector)

    @property
    def cAngleToXYPlane(self):
        return calculateCAngleToXYPlaneFromWVector(wvector=-self.normalVector)

    def isWopMatrixRightHanded(self):
        return not self.wopMatrix.isFlipped()


class IsSamePlaneTests(unittest.TestCase):
    def testDifferentOriginAndNormalVectorShouldNotBeSamePlane(self):
        plane1 = Plane(Vector(1, 2, 3), Vector(1, 0, 0))
        plane2 = Plane(Vector(0, 0, 0), Vector(0, 1, 0))
        self.assertFalse(plane1.isCompatiblePlane(plane2))

    def testSameOriginAndNormalVectorShouldBeSamePlane(self):
        plane1 = Plane(Vector(0, 0, 0), Vector(1, 0, 0))
        self.assertTrue(plane1.isCompatiblePlane(plane1))

    def testPlanesWithSameNormalVectorButDifferentOriginShouldBeOnSamePlane(self):
        plane1 = Plane(Vector(1, 2, 3), Vector(0, 0, 1))
        plane2 = Plane(Vector(0, 0, 3), Vector(0, 0, 1))
        self.assertTrue(plane1.isCompatiblePlane(plane2))

    def testParallelPlanesShouldNotBeOnSamePlane(self):
        plane1 = Plane(Vector(1, 2, 3), Vector(0, 0, 1))
        plane2 = Plane(Vector(1, 2, 5), Vector(0, 0, 1))
        self.assertFalse(plane1.isCompatiblePlane(plane2))


class IsPointOnPlaneTests(unittest.TestCase):
    def testPointIsOnPlane(self):
        plane = Plane(Vector(0, 0, 0), Vector(0, 0, 1))
        point = Vector(1, 2, 0)
        self.assertTrue(plane.isPointOnPlane(point))

    def testPointIsNotOnPlane(self):
        plane = Plane(Vector(0, 0, 0), Vector(0, 0, 1))
        point = Vector(1, 2, 3)
        self.assertFalse(plane.isPointOnPlane(point))


class AngleTests(unittest.TestCase):
    def testVerticalCAngle(self):
        plane = Plane(Vector(0, 0, 0), Vector(0, 0, 1))
        self.assertEqual(0.0, plane.cAngleToXYPlane)


class WopMatrixRightHanded(unittest.TestCase):
    def setUp(self):
        self.plane = Plane(wopMatrix=Matrix())

    def testIdentityMatrixShouldBeRightHanded(self):
        self.assertTrue(self.plane.isWopMatrixRightHanded())

    def testMatrixShouldNotBeRightHanded(self):
        self.plane.wopMatrix.w = Vector(0, 0, -1)
        self.assertFalse(self.plane.isWopMatrixRightHanded())
