# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest

from interiorcad.interface.geometry import isNearlyEqual


class Point:
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def __repr__(self):
        return str(self)

    def magnitude(self):
        return math.hypot(self.x, self.y)

    def distanceTo(self, other):
        return math.hypot(other.x - self.x, other.y - self.y)

    def mirrorOnZeroXAxis(self):
        self.y = - self.y

    def mirrorOnZeroYAxis(self):
        self.x = - self.x

    def rotateLeftAtZero(self):
        self.x, self.y = -self.y, self.x

    def rotateRightAtZero(self):
        self.x, self.y = self.y, -self.x

    def rotate180AtZero(self):
        self.x, self.y = -self.x, -self.y

    def normalize(self):
        point = self / self.magnitude()
        self.x, self.y = point.x, point.y

    def normalized(self):
        point = self / self.magnitude()
        return Point(point.x, point.y)

    def perp(self):
        return Point(self.y, -self.x)

    def dot(self, other):
        return self.x * other.x + self.y * other.y

    def __eq__(self, other):
        return isNearlyEqual(self.x, other.x) and isNearlyEqual(self.y, other.y)

    def __iadd__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __str__(self):
        return ''.join(('Point(', str(self.x), ', ', str(self.y), ')'))

    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)

    def __truediv__(self, scalar):
        if scalar != 0:
            return Point(self.x / scalar, self.y / scalar)

    def __mul__(self, rhs):
        if isinstance(rhs, Point):
            return Point(self.x * rhs.x, self.y * rhs.y)
        else:
            return Point(self.x * rhs, self.y * rhs)

    def __neg__(self):
        return Point(- self.x, - self.y)

    def multiplyByMatrix(self, matrix):
        x = math.fsum((matrix.u.x * self.x, matrix.v.x * self.y))
        y = math.fsum((matrix.u.y * self.x, matrix.v.y * self.y))
        self.x, self.y = x, y

    def asTuple(self):
        return self.x, self.y


class PointTest(unittest.TestCase):
    def testDefaultPoint2DIsZeroZero(self):
        self.assertEqual(0, Point().x)
        self.assertEqual(0, Point().y)

    def testCanSetXYInConstructor(self):
        self.assertEqual(5, Point(x=5).x)
        self.assertEqual(15, Point(y=15).y)
        self.assertEqual(25, Point(x=25, y=10).x)
        self.assertEqual(10, Point(25, 10).y)

    def testMagnitudeOfPointIsCorrect(self):
        self.assertEqual(2, Point(2, 0).magnitude())
        self.assertEqual(2, Point(0, 2).magnitude())
        self.assertEqual(math.sqrt(2), Point(1, 1).magnitude())

    def testDistanceOfTwoPointsIsCorrect(self):
        self.assertEqual(2, Point(0, 0).distanceTo(Point(0, 2)))
        self.assertEqual(2, Point(0, 0).distanceTo(Point(2, 0)))
        self.assertEqual(2, Point(2, 0).distanceTo(Point(0, 0)))
        self.assertEqual(2, Point(0, 2).distanceTo(Point(0, 0)))

    def testStringRepresentationOfPointIsAsExpected(self):
        self.assertEqual('Point(5, 10)', str(Point(5, 10)))

    def testPointsAreEqual(self):
        self.assertEqual(Point(5, 10), Point(5, 10))
        self.assertEqual(Point(5, 10), Point(y=10, x=5))
        self.assertEqual(Point(y=100.12), Point(0, 100.12))

    def testAddEqualPoint(self):
        p = Point(100, 200)
        p += Point(10, 20)
        self.assertEqual(Point(110, 220), p)

    def testCanAddPoints(self):
        point1 = Point(100, 200)
        point2 = Point(10, 20)
        self.assertEqual(Point(110, 220), point1 + point2)

    def testMirrorOnXAxis(self):
        p = Point(100, 200)
        p.mirrorOnZeroXAxis()
        self.assertEqual(Point(100, -200), p)

    def testRotateLeft(self):
        point = Point(100, 200)
        point.rotateLeftAtZero()
        self.assertEqual(Point(-200, 100), point)

    def testRotateRight(self):
        point = Point(100, 200)
        point.rotateRightAtZero()
        self.assertEqual(Point(200, -100), point)

    def testMirrorOnYAxis(self):
        p = Point(100, 200)
        p.mirrorOnZeroYAxis()
        self.assertEqual(Point(-100, 200), p)

    def testCanSubtractTwoPoints(self):
        point1 = Point(200, 100)
        point2 = Point(300, 200)
        self.assertEqual(Point(100, 100), point2 - point1)

    def testCanSubtractEmptyPoints(self):
        point1 = Point()
        point2 = Point()
        self.assertEqual(Point(), point2 - point1)

    def testCanDividePointWithScalar(self):
        point = Point(100, 100)
        self.assertEqual(Point(50, 50), point / 2)

    def testDividePointWithZeroScalar(self):
        point = Point(100, 100)
        self.assertIsNone(point / 0)

    def testMultiplyPointWithScalar(self):
        point = Point(100, 100)
        self.assertEqual(Point(200, 200), point * 2)

    def testMultiplyPointWithZeroScalar(self):
        point = Point(100, 100)
        self.assertEqual(Point(0, 0), point * 0)

    def testMultiplyPointWithPoint(self):
        point = Point(100, 100)
        point2 = Point(2, 3)
        self.assertEqual(Point(200, 300), point * point2)

    def testNormalizePointOnY(self):
        point = Point(0, 10)
        point.normalize()
        self.assertEqual(Point(0, 1), point)

    def testNormalizePointOnX(self):
        point = Point(10, 0)
        point.normalize()
        self.assertEqual(Point(1, 0), point)

    def testNormalizePoint(self):
        point = Point(10, 10)
        point.normalize()
        self.assertEqual(Point(math.sqrt(2) / 2, math.sqrt(2) / 2), point)

    def testNormalizedPointOnY(self):
        point = Point(0, 10)
        self.assertEqual(Point(0, 1), point.normalized())

    def testNormalizedPointOnX(self):
        point = Point(10, 0)
        self.assertEqual(Point(1, 0), point.normalized())

    def testNormalizedPoint(self):
        point = Point(10, 10)
        self.assertEqual(Point(math.sqrt(2) / 2, math.sqrt(2) / 2), point.normalized())

    def testPerpPoint(self):
        point = Point(1, 0)
        perpPt = point.perp()
        self.assertEqual(perpPt, Point(0, -1))

    def testDotOfParallelPoints(self):
        point1 = Point(1, 0)
        point2 = Point(1, 0)
        self.assertEqual(1, point1.dot(point2))

    def testDotOfParallelAndOppositePoints(self):
        point1 = Point(1, 0)
        point2 = Point(-1, 0)
        self.assertEqual(-1, point1.dot(point2))

    def testDotOfPerpendicularPoints(self):
        point1 = Point(1, 0)
        point2 = Point(0, 1)
        self.assertEqual(0, point1.dot(point2))

    def testNeg(self):
        point = Point(1, 2)
        self.assertEqual(Point(-1, -2), - point)
        self.assertEqual(Point(1, 2) * -1, - point)
