# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import copy
import math
import operator
import unittest.mock

from interiorcad.interface.geometry import EPSILON, isNearlyEqual, isNearlyGreater
from interiorcad.interface.geometry.center3points import getCenterOfThreePointsOnArc
from interiorcad.interface.geometry.clockwise import arePositionsClockwise
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polylinesegment import PolylineSegment
from interiorcad.interface.geometry.polylinesegments import PolylineSegments
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.geometry.vectormath import VectorMath


class Polyline:
    def __init__(self, *vertices):
        self.vertices = []
        for v in vertices:
            if isinstance(v, (list, tuple)):
                self.vertices.extend(v)
            else:
                self.vertices.append(v)

    @classmethod
    def createFromIterable(cls, iterable):
        return cls(*iterable)

    def __bool__(self):
        return self.isValid()

    def isValid(self):
        return self.numVertices() > 0

    def isClosed(self):
        return self.isValid() and (self.areStartAndEndPositionsMatching() or self.vertices[-1].visible)

    def numVertices(self):
        return len(self.vertices)

    def addVertex(self, vertex):
        self.vertices.append(vertex)

    def addVertices(self, vertices):
        self.vertices.extend(vertices)

    def moveBy(self, delta):
        for v in self.vertices:
            v.moveBy(delta)

    def mirrorOnZeroXAxis(self):
        for vertex in self.vertices:
            vertex.mirrorOnZeroXAxis()

    def mirrorOnZeroYAxis(self):
        for vertex in self.vertices:
            vertex.mirrorOnZeroYAxis()

    def rotateLeftAtZero(self):
        for vertex in self.vertices:
            vertex.rotateLeftAtZero()

    def rotateRightAtZero(self):
        for vertex in self.vertices:
            vertex.rotateRightAtZero()

    def rotate180AtZero(self):
        for vertex in self.vertices:
            vertex.rotate180AtZero()

    def getVertexIndicesNextTo(self, point):
        helpList = []
        for i, v in enumerate(self.vertices):
            distance = point.distanceTo(v.position)
            helpList.append((distance, i))
        helpList.sort(key=operator.itemgetter(0))
        return [element[1] for element in helpList]

    def getVertexIndexNextTo(self, point):
        if self.vertices:
            return self.getVertexIndicesNextTo(point)[0]

    def isPolygonal(self):
        return all((vertex.isCornerVertex() for vertex in self.vertices))

    def isRectangular(self):
        numVertices = len(self.vertices)
        isRectangular = numVertices == 4 or numVertices == 5

        if isRectangular:
            for index, vertex in enumerate(self.vertices):
                segment = self.getSegmentAt(index)
                isRectangular = isRectangular and vertex.isCornerVertex()
                if len(segment) > 1:
                    isRectangular = isRectangular and (isNearlyEqual(segment[0].position.x, segment[1].position.x) or isNearlyEqual(segment[0].position.y, segment[1].position.y))

        return isRectangular

    def getSegmentAt(self, index):
        if index == len(self.vertices) - 1:
            return [self.vertices[index], self.vertices[0]]
        else:
            return self.vertices[index:index + 2]

    def setStartingVertexTo(self, index):
        class VertexRearranger:
            def __init__(self, vertices):
                self.vertices = vertices
                self.firstAndLastIsTheSame = False
                self.openUpVertices()

            def getRearrangedVertices(self, startIndex):
                rearranged = self.vertices[startIndex:] + self.vertices[:startIndex]
                if self.firstAndLastIsTheSame:
                    rearranged.append(copy.deepcopy(rearranged[0]))
                return rearranged

            def openUpVertices(self):
                if len(self.vertices) > 1:
                    self.firstAndLastIsTheSame = self.vertices[0] == self.vertices[-1]
                    if self.firstAndLastIsTheSame:
                        del self.vertices[-1]

        if index is not None:
            rearranger = VertexRearranger(self.vertices[:])
            self.vertices = rearranger.getRearrangedVertices(index)

    def changeDirection(self):
        if len(self.vertices) > 1:
            self.vertices.reverse()
            if self.vertices[-1] != self.vertices[0]:
                self.vertices = self.vertices[-1:] + self.vertices[0:-1]

    def changeDirectionOfOpenPolygon(self):
        numVertices = len(self.vertices)
        if numVertices > 1:
            oldVisibilities = tuple((vertex.visible for vertex in self.vertices))
            for i in range(0, numVertices):
                self.vertices[(i + 1) % numVertices].visible = oldVisibilities[i]

            self.vertices.reverse()

    def changeDirectionWithVisibility(self):
        self.changeDirectionOfOpenPolygon()
        if len(self.vertices) > 1 and all(map(lambda vertex: vertex.visible, self.vertices)):
            self.vertices = self.vertices[-1:] + self.vertices[:-1]

    def insertVertexAfter(self, index, vertex):
        self.vertices.insert(index + 1, vertex)

    def getPolylineSegments(self):
        # This function DOES NOT close the polyline (see getSegments if you need a closed polyline)
        segments = PolylineSegments()
        for index in range(len(self.vertices)):
            segment = self.vertices[index:index + 2]
            if len(segment) > 1:
                segments.append(PolylineSegment(segment[0], segment[1]))
        return segments

    def getVertex(self, index):
        return self.vertices[index]

    def __eq__(self, other):
        return self.vertices == other.vertices

    def __str__(self):
        verticesText = ', '.join((
            "v{}={}".format(i, str(v)) for i, v in enumerate(self.vertices, 1)
        ))
        return 'Polyline({})'.format(verticesText)

    def __repr__(self):
        verticesText = ', '.join(("{}".format(repr(v)) for v in self.vertices))
        return 'Polyline({})'.format(verticesText)

    def __copy__(self):
        return Polyline(self.vertices[:])

    # See the source code of the Python module "copy" for the needed signature. (It is not well documented.)
    def __deepcopy__(self, memo):
        return Polyline([copy.deepcopy(vertex, memo) for vertex in self.vertices[:]])

    def polylineStartsWithCornerVertex(self):
        return self.vertices[0].isCornerVertex()

    # Compatibility function for old code; use preferably PolygonConverter.convertPolyline
    @staticmethod
    def createFromCornerVertexPositionTuples(positionTuples):
        return Polyline(tuple(CornerVertex(Point(position[0], position[1])) for position in positionTuples))

    def getSegments(self):
        # This function DOES close the polyline (use getPolylineSegments directly if you need the polyline as it is)
        segments = self.getPolylineSegments()
        if self.isValid():
            firstVertex = self.vertices[0]
            lastVertex = self.vertices[-1]
            if firstVertex != lastVertex:
                segments.append(PolylineSegment(lastVertex, firstVertex))
        return segments

    def areStartAndEndPositionsMatching(self):
        return self.isValid() and (self.vertices[0].position == self.vertices[-1].position)

    def isClockwise(self):
        return arePositionsClockwise([vertex.position for vertex in self.vertices])

    @classmethod
    def createRectangularClockwiseFromBounds(cls, bounds):
        left = bounds.bottomLeft().x
        right = bounds.topRight().x
        top = bounds.topRight().y
        bottom = bounds.bottomLeft().y

        return cls(
            CornerVertex(Point(left, bottom)),
            CornerVertex(Point(left, top)),
            CornerVertex(Point(right, top)),
            CornerVertex(Point(right, bottom)),
        )

    @classmethod
    def createRectangularCounterClockwiseFromBounds(cls, bounds):
        left = bounds.bottomLeft().x
        right = bounds.topRight().x
        top = bounds.topRight().y
        bottom = bounds.bottomLeft().y

        return cls(
            CornerVertex(Point(left, bottom)),
            CornerVertex(Point(right, bottom)),
            CornerVertex(Point(right, top)),
            CornerVertex(Point(left, top)),
        )

    def forceVisibility(self):
        for vertex in self.vertices:
            vertex.visible = True

    @staticmethod
    def calculateRadiusFrom3Points(point1, point2, point3):
        """http://de.wikipedia.org/wiki/Apollonisches_Problem#Drei_Punkte"""
        # Similar to C++ function ArcRadiusByThreePoints in dev\XG Custom Part\Source\CustomPartPolyline.cpp

        aMagnitude = (point3 - point1).magnitude()
        bNormalized = (point1 - point2).normalized()
        cNormalized = (point3 - point2).normalized()

        dotProduct = bNormalized.dot(cNormalized)
        alpha = math.acos(dotProduct)
        return aMagnitude / (2 * math.sin(alpha))

    # Similar code to interiorcad.backend.ncexportmigrated.ncradiussegment.NCRadiusSegment.isArcGreater180
    def __isArcGreater180(self, startPoint, radiusPoint, endPoint):
        firstVector = startPoint - radiusPoint
        secondVector = endPoint - radiusPoint
        return VectorMath.angleBetween(secondVector, firstVector) < 90.0

    def __generateLengthOfEachSegment(self):
        vertices = self.vertices
        i = 0
        while i < len(vertices):
            assert vertices[i].isCornerVertex() or vertices[i].isRadiusVertex(), "This over-simplistic code supports only corner and radius vertices!"
            j = (i + 1) % len(vertices)
            if vertices[i].visible:
                if vertices[i].isCornerVertex() and vertices[j].isCornerVertex():
                    yield (vertices[j].position - vertices[i].position).magnitude()
                    i += 1

                elif vertices[i].isCornerVertex() and vertices[j].isRadiusVertex():
                    k = (j + 1) % len(vertices)
                    if vertices[k].isCornerVertex():
                        radius = Polyline.calculateRadiusFrom3Points(vertices[i].position, vertices[j].position, vertices[k].position)
                        assert isNearlyGreater(radius, 0.0), "The radius should not be zero or negative"
                        center = getCenterOfThreePointsOnArc(vertices[i].position, vertices[j].position, vertices[k].position, radius)
                        startVector = vertices[i].position - center
                        startAngle = math.atan2(startVector.y, startVector.x)
                        endVector = vertices[k].position - center
                        endAngle = math.atan2(endVector.y, endVector.x)
                        isArcGreaterPi = self.__isArcGreater180(vertices[i].position, vertices[j].position, vertices[k].position)
                        deltaAngle = (2 * math.pi - (endAngle - startAngle)) if isArcGreaterPi else (endAngle - startAngle)
                        yield math.fabs(radius * deltaAngle)
                        i += 2
                    else:
                        assert False, "This over-simplistic code does not support the situation where the radius vertex is not followed by a corner vertex"

                else:
                    assert False, "This over-simplistic code does not support this situation"
            else:
                i += 1

    def getLength(self):
        return math.fsum(self.__generateLengthOfEachSegment())


class PolylineCreationTest(unittest.TestCase):
    def testCanCreatePolyline(self):
        poly = Polyline()
        self.assertFalse(poly)
        self.assertIsInstance(poly, Polyline)

    def testCanAddVertexInConstructor(self):
        poly = Polyline('v1')
        self.assertEqual(1, poly.numVertices())

    def testCanAddListOfVerticesInConstructor(self):
        poly = Polyline(['v1', 'v2'])
        self.assertEqual(2, poly.numVertices())

    def testCanAddMultipleVerticesInConstructor(self):
        poly = Polyline('v1', 'v2')
        self.assertEqual(2, poly.numVertices())

    def testCreateFromCornerVertexPositionTuples(self):
        positions = ((0, 0), (200, 0), (200, 100), (0, 100))
        polyline = Polyline.createFromCornerVertexPositionTuples(positions)

        expectedPoly = Polyline((
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(200, 0)),
            CornerVertex(Point(200, 100)),
            CornerVertex(Point(0, 100)),
        ))
        self.assertEqual(expectedPoly, polyline)

    def testIsClosedOnAPathReturningToStartShouldBeTrueEvenIfVisibilityIsFalse(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
            CornerVertex(Point(100, 100), visible=False),
        )
        self.assertTrue(polyline.areStartAndEndPositionsMatching())

    def testCopy(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
            CornerVertex(Point(100, 100), visible=False),
        )
        copiedPolyline = copy.copy(polyline)
        self.assertEqual(polyline, copiedPolyline)
        self.assertNotEqual(id(polyline), id(copiedPolyline))

    def testDeepCopy(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
            CornerVertex(Point(100, 100), visible=False),
        )
        copiedPolyline = copy.deepcopy(polyline)
        self.assertEqual(polyline, copiedPolyline)
        self.assertNotEqual(id(polyline), id(copiedPolyline))

    def testCreateRectangularClockwiseFromBoundsShouldBeCorrect(self):
        bounds = Rectangle(left=11, bottom=12, right=91, top=92)

        expectedRectangularSizingPolyline = Polyline(
            CornerVertex(Point(11, 12)),
            CornerVertex(Point(11, 92)),
            CornerVertex(Point(91, 92)),
            CornerVertex(Point(91, 12)),
        )

        polyline = Polyline.createRectangularClockwiseFromBounds(bounds)
        self.assertEqual(expectedRectangularSizingPolyline, polyline)
        self.assertIsInstance(polyline, Polyline)

    def testCreateRectangularCounterClockwiseFromBoundsShouldBeCorrect(self):
        bounds = Rectangle(left=11, bottom=12, right=91, top=92)

        expectedRectangularSizingPolyline = Polyline(
            CornerVertex(Point(11, 12)),
            CornerVertex(Point(91, 12)),
            CornerVertex(Point(91, 92)),
            CornerVertex(Point(11, 92)),
        )

        polyline = Polyline.createRectangularCounterClockwiseFromBounds(bounds)
        self.assertEqual(expectedRectangularSizingPolyline, polyline)
        self.assertIsInstance(polyline, Polyline)


class PolylineTest(unittest.TestCase):
    def setUp(self):
        self.poly = Polyline()

    def testAnEmptyPolylineIsInvalidAndFalse(self):
        self.assertFalse(self.poly)
        self.assertFalse(self.poly.isValid())

    def testANonEmptyPolylineIsvalidAndTrue(self):
        self.poly = Polyline(
            (CornerVertex(Point(100, 200))),
            (CornerVertex(Point(300, 400))),
        )
        self.assertTrue(self.poly)
        self.assertTrue(self.poly.isValid())

    def testDefaultPolylineHasZeroVertices(self):
        self.assertEqual(0, self.poly.numVertices())

    def testAddVertex(self):
        self.poly.addVertex('vertex')
        self.assertEqual(1, self.poly.numVertices())

    def testAddMultipleVertices(self):
        self.poly.addVertices(['vertex1', 'vertex2'])
        self.assertEqual(2, self.poly.numVertices())

    def testMoveEmptyPolygonShouldRemainEmpty(self):
        self.poly.moveBy(Point(10, 20))
        self.assertFalse(self.poly)
        self.assertEqual([], self.poly.vertices)

    def testMovePolygonWithOneVertex(self):
        self.poly.addVertex(CornerVertex(Point(100, 200)))
        self.poly.moveBy(Point(10, 20))
        self.assertEqual(Point(110, 220), self.poly.vertices[0].position)

    def testMovePolygonWithTwoVertices(self):
        self.poly.addVertex(CornerVertex(Point(100, 200)))
        self.poly.addVertex(CornerVertex(Point(300, 400)))
        self.poly.moveBy(Point(10, 20))
        self.assertEqual(Point(110, 220), self.poly.vertices[0].position)
        self.assertEqual(Point(310, 420), self.poly.vertices[1].position)

    def testMirrorOnZeroXAxisEmptyPolygonShouldRemainEmpty(self):
        self.poly.mirrorOnZeroXAxis()
        self.assertFalse(self.poly)
        self.assertEqual([], self.poly.vertices)

    def testMirrorOnZeroYAxisEmptyPolygonShouldRemainEmpty(self):
        self.poly.mirrorOnZeroYAxis()
        self.assertFalse(self.poly)
        self.assertEqual([], self.poly.vertices)

    def testMirrorOnZeroXAxisPolygonWithOneVertex(self):
        self.poly.addVertex(CornerVertex(Point(100, 200)))
        self.poly.mirrorOnZeroXAxis()
        self.assertEqual(Point(100, -200), self.poly.vertices[0].position)

    def testMirrorOnZeroYAxisPolygonWithOneVertex(self):
        self.poly.addVertex(CornerVertex(Point(100, 200)))
        self.poly.mirrorOnZeroYAxis()
        self.assertEqual(Point(-100, 200), self.poly.vertices[0].position)

    def testEmptyPolylineIsPolygonal(self):
        self.assertTrue(self.poly.isPolygonal())

    def testPolylineWithOneVertexIsPolygonal(self):
        self.poly.addVertex(CornerVertex())
        self.assertTrue(self.poly.isPolygonal())

    def testPolylineWithOneRadiusVertexIsNotPolygonal(self):
        self.poly.addVertex(RadiusVertex())
        self.assertFalse(self.poly.isPolygonal())

    def testPolylineWithVertexAndRadiusVertexIsNotPolygonal(self):
        self.poly.addVertex(CornerVertex())
        self.poly.addVertex(RadiusVertex())
        self.assertFalse(self.poly.isPolygonal())

    def testOpenPolylineRectangular(self):
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.assertTrue(self.poly.isRectangular())
        self.assertFalse(self.poly.areStartAndEndPositionsMatching())

    def testClosedPolylineRectangular(self):
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.assertTrue(self.poly.isRectangular())
        self.assertTrue(self.poly.areStartAndEndPositionsMatching())

    def testEmptyPolylineIsNotRectangular(self):
        self.assertFalse(self.poly.isRectangular())

    def testEmptyPolylineIsNotClosed(self):
        self.assertFalse(self.poly.areStartAndEndPositionsMatching())

    def testPolylineWithLastEdgeNotStraightIsNotRectangular(self):
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.poly.addVertex(CornerVertex(Point(100, 100)))
        self.poly.addVertex(CornerVertex(Point(50, 100)))
        self.assertFalse(self.poly.isRectangular())
        self.assertFalse(self.poly.areStartAndEndPositionsMatching())

    def testOpenPolylineRectangularWithMinimalDifference(self):
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 100 + EPSILON / 2)))
        self.poly.addVertex(CornerVertex(Point(100 + EPSILON / 2, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.assertTrue(self.poly.isRectangular())
        self.assertFalse(self.poly.areStartAndEndPositionsMatching())

    def testClosedPolylineRectangularWithMinimalDifference(self):
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 100 + EPSILON / 2)))
        self.poly.addVertex(CornerVertex(Point(100 + EPSILON / 2, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.assertTrue(self.poly.isRectangular())
        self.assertTrue(self.poly.areStartAndEndPositionsMatching())

    def testGetSegmentShouldWork(self):
        p0 = CornerVertex(Point(0, 0))
        p1 = CornerVertex(Point(100, 0))
        p2 = CornerVertex(Point(100, 100))
        p3 = CornerVertex(Point(50, 100))
        self.poly.addVertex(p0)
        self.poly.addVertex(p1)
        self.poly.addVertex(p2)
        self.poly.addVertex(p3)
        self.assertListEqual([p0, p1], self.poly.getSegmentAt(0))
        self.assertListEqual([p1, p2], self.poly.getSegmentAt(1))
        self.assertListEqual([p2, p3], self.poly.getSegmentAt(2))
        self.assertListEqual([p3, p0], self.poly.getSegmentAt(3))

    def testInsertPointAfter(self):
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.poly.addVertex(CornerVertex(Point(100, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 0)))

        expectedPoly = Polyline((
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(50, 50)),
            CornerVertex(Point(100, 100)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(0, 0)),
        ))
        vertex = CornerVertex(Point(50, 50))
        self.poly.insertVertexAfter(0, vertex)
        self.assertEqual(expectedPoly, self.poly)

    def testGetPolylineSegments(self):
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        self.poly.addVertex(CornerVertex(Point(100, 100)))
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.poly.addVertex(CornerVertex(Point(0, 0)))
        polySegmentList = [
            PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(100, 100))),
            PolylineSegment(CornerVertex(Point(100, 100)), CornerVertex(Point(100, 0))),
            PolylineSegment(CornerVertex(Point(100, 0)), CornerVertex(Point(0, 0)))
        ]
        outputSegmentList = self.poly.getPolylineSegments()
        self.assertListEqual(polySegmentList, outputSegmentList)


class PolylineBoolOperatorTest(unittest.TestCase):
    def setUp(self):
        self.poly = Polyline()
        self.otherPoly = Polyline()

    def testTwoEmptyPolylinesAreEqual(self):
        self.assertEqual(Polyline(), self.poly)

    def testTwoPolylinesWithSameVertexAreEqual(self):
        self.poly.addVertex('v1')
        self.otherPoly.addVertex('v1')
        self.assertEqual(self.otherPoly, self.poly)

    def testTwoPolylinesWithDifferentVertexAreNotEqual(self):
        self.poly.addVertex('v1')
        self.otherPoly.addVertex('v2')
        self.assertNotEqual(self.otherPoly, self.poly)


class PolylineStringRepresentationTest(unittest.TestCase):
    def testEmptyPolyline(self):
        self.poly = Polyline()
        self.assertEqual('Polyline()', str(self.poly))
        self.assertEqual('Polyline()', repr(self.poly))

    def testPolylineWithOneVertex(self):
        self.poly = Polyline()
        self.poly.addVertex('vertex')
        self.assertEqual('Polyline(v1=vertex)', str(self.poly))
        self.assertEqual("Polyline('vertex')", repr(self.poly))

    def testPolylineWithTwoVertices(self):
        self.poly = Polyline()
        self.poly.addVertices(['vertex1', 'vertex2'])
        self.assertEqual('Polyline(v1=vertex1, v2=vertex2)', str(self.poly))
        self.assertEqual("Polyline('vertex1', 'vertex2')", repr(self.poly))

    def testPolylineWithThreeVertices(self):
        self.poly = Polyline()
        self.poly.addVertices(['v1', 'v2', 'v3'])
        self.assertEqual('Polyline(v1=v1, v2=v2, v3=v3)', str(self.poly))
        self.assertEqual("Polyline('v1', 'v2', 'v3')", repr(self.poly))


class PolylineSetStartingVertexTest(unittest.TestCase):
    def setUp(self):
        self.poly = Polyline()
        self.poly.addVertex('0')
        self.poly.addVertex('1')

    def testTwoVerticesSetIndexToZero(self):
        self.poly.setStartingVertexTo(0)
        expected = ['0', '1']
        self.assertListEqual(expected, self.poly.vertices)

    def testTwoVerticesSetIndexToOne(self):
        self.poly.setStartingVertexTo(1)
        expected = ['1', '0']
        self.assertListEqual(expected, self.poly.vertices)

    def testThreeVerticesSetIndexToOne(self):
        self.poly.addVertex('2')
        self.poly.setStartingVertexTo(1)
        expected = ['1', '2', '0']
        self.assertListEqual(expected, self.poly.vertices)

    def testThreeVerticesClosedSetIndexToOne(self):
        self.poly.addVertex('0')
        self.poly.setStartingVertexTo(1)
        expected = ['1', '0', '1']
        self.assertListEqual(expected, self.poly.vertices)

    def testThreeVerticesClosedSetIndexToTwo(self):
        self.poly.addVertex('0')
        self.poly.setStartingVertexTo(2)
        expected = ['0', '1', '0']
        self.assertListEqual(expected, self.poly.vertices)

    def testFourVerticesOpenSetIndexToTwo(self):
        self.poly.addVertex('2')
        self.poly.addVertex('3')
        self.poly.setStartingVertexTo(2)
        expected = ['2', '3', '0', '1']
        self.assertListEqual(expected, self.poly.vertices)


class PolylineGetVertexIndexNextToTest(unittest.TestCase):
    def setUp(self):
        self.poly = Polyline()
        self.poly.addVertex(CornerVertex(Point(0.0, 0.0)))

    def testOneVertexMatchingExactly(self):
        self.assertEqual(0, self.poly.getVertexIndexNextTo(Point(0.0, 0.0)))

    def testOneVertexNotMatching(self):
        self.assertEqual(0, self.poly.getVertexIndexNextTo(Point(10.0, 0.0)))

    def testTwoVerticesMatchingExactlyY(self):
        self.poly.addVertex(CornerVertex(Point(0, 50)))
        self.assertEqual(0, self.poly.getVertexIndexNextTo(Point(0, 0)))

    def testTwoVerticesMatchingExactlyX(self):
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.assertEqual(1, self.poly.getVertexIndexNextTo(Point(100, 0)))

    def testTwoVerticesNotMatchingX(self):
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.assertEqual(0, self.poly.getVertexIndexNextTo(Point(10, 0)))

    def testTwoVerticesNotMatchingY(self):
        self.poly.addVertex(CornerVertex(Point(0, 50)))
        self.assertEqual(1, self.poly.getVertexIndexNextTo(Point(0, 40)))

    def testGetVertexIndicesNextTo(self):
        self.poly.addVertex(CornerVertex(Point(100, 0)))
        self.poly.addVertex(CornerVertex(Point(20, 0)))
        self.poly.addVertex(CornerVertex(Point(40, 0)))
        indices = self.poly.getVertexIndicesNextTo(Point(40, 0))
        expected = [3, 2, 0, 1]
        self.assertListEqual(expected, indices)


class PolylineChangeDirectionTest(unittest.TestCase):
    def setUp(self):
        self.poly = Polyline()
        self.poly.addVertex('1')

    def testOneVertex(self):
        expected = ['1']
        self.poly.changeDirection()
        self.assertListEqual(expected, self.poly.vertices)

    def testTwoVertices(self):
        self.poly.addVertex('2')
        expected = ['1', '2']
        self.poly.changeDirection()
        self.assertListEqual(expected, self.poly.vertices)

    def testThreeVertices(self):
        self.poly.addVertex('2')
        self.poly.addVertex('3')
        expected = ['1', '3', '2']
        self.poly.changeDirection()
        self.assertListEqual(expected, self.poly.vertices)

    def testFourVerticesClosed(self):
        self.poly.addVertex('2')
        self.poly.addVertex('3')
        self.poly.addVertex('1')
        expected = ['1', '3', '2', '1']
        self.poly.changeDirection()
        self.assertListEqual(expected, self.poly.vertices)

    def testStartPointIsCornerVertex(self):
        polylinePathStartsWithCorner = Polyline()
        polylinePathStartsWithCorner.addVertex(CornerVertex(0, 0))
        polylinePathStartsWithCorner.addVertex(CornerVertex(100, 0))
        polylinePathStartsWithCorner.addVertex(RadiusVertex(200, 10))
        polylinePathStartsWithCorner.addVertex(CornerVertex(100, 100))

        self.assertTrue(polylinePathStartsWithCorner.polylineStartsWithCornerVertex())

    def testStartPointIsNotCornerVertex(self):
        polylinePathStartsWithRadius = Polyline()
        polylinePathStartsWithRadius.addVertex(RadiusVertex(200, 10))
        polylinePathStartsWithRadius.addVertex(CornerVertex(100, 100))
        polylinePathStartsWithRadius.addVertex(CornerVertex(100, 0))
        polylinePathStartsWithRadius.addVertex(CornerVertex(0, 0))

        self.assertFalse(polylinePathStartsWithRadius.polylineStartsWithCornerVertex())


class GetSegmentsTest(unittest.TestCase):
    def setUp(self):
        self.polyline = Polyline()
        self.polyline.addVertex(CornerVertex(Point(0, 0)))
        self.polyline.addVertex(CornerVertex(Point(0, 50)))
        self.polyline.addVertex(CornerVertex(Point(100, 50)))
        self.polyline.addVertex(CornerVertex(Point(100, 0)))

    @staticmethod
    def readVertex(vertex):
        x = vertex[0]
        y = vertex[1]
        return x, y

    def testMoveShapeRightAndTop(self):
        self.polyline.moveBy(Point(10.0, 20.0))
        self.assertEqual(Point(10, 20), self.polyline.vertices[0].position)
        self.assertEqual(Point(10, 70), self.polyline.vertices[1].position)
        self.assertEqual(Point(110, 70), self.polyline.vertices[2].position)
        self.assertEqual(Point(110, 20), self.polyline.vertices[3].position)

    def testMoveShapeLeftAndDown(self):
        self.polyline.moveBy(Point(-10, -20))
        self.assertEqual(Point(-10, -20), self.polyline.vertices[0].position)
        self.assertEqual(Point(-10, 30), self.polyline.vertices[1].position)
        self.assertEqual(Point(90, 30), self.polyline.vertices[2].position)
        self.assertEqual(Point(90, -20), self.polyline.vertices[3].position)

    def testGetSegmentsOfOpenShapeShouldReturnAllSegments(self):
        expected = PolylineSegments()
        expected.append(PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 50))))
        expected.append(PolylineSegment(CornerVertex(Point(0, 50)), CornerVertex(Point(100, 50))))
        expected.append(PolylineSegment(CornerVertex(Point(100, 50)), CornerVertex(Point(100, 0))))
        expected.append(PolylineSegment(CornerVertex(Point(100, 0)), CornerVertex(Point(0, 0))))
        self.assertEqual(expected, self.polyline.getSegments())


class ChangeDirectionWithVisibilityTests(unittest.TestCase):
    def testOpenRectangularPolygon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(0, 100), visible=False),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 100), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(0, 0), visible=False),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testClosedRectangularPolygon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(0, 100), visible=True),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(0, 100), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testClosedRectangularPolygonStartAtMiddleOfSegment(self):  # Path of document of IC-15295
        polyline = Polyline((
            CornerVertex(position=Point(237, 134), visible=True),
            CornerVertex(position=Point(237, 268), visible=True),
            CornerVertex(position=Point(0, 268), visible=True),
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(0, 237), visible=True),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(237, 134), visible=True),
            CornerVertex(position=Point(0, 237), visible=True),
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(0, 268), visible=True),
            CornerVertex(position=Point(237, 268), visible=True),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testInvisibleFirstVertexRectangularPolygon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=False),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(0, 100), visible=True),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 100), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(200, 0), visible=False),
            CornerVertex(position=Point(0, 0), visible=True),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testInvisibleSecondVertexRectangularPolygon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(200, 100), visible=False),
            CornerVertex(position=Point(0, 100), visible=True),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 100), visible=False),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(0, 0), visible=True),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testOpenPentagon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(250, 50), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(0, 100), visible=False),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 100), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(250, 50), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(0, 0), visible=False),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testClosedPentagon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(250, 50), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(0, 100), visible=True),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(0, 100), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(250, 50), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testInvisibleSecondVertexPentagon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(250, 50), visible=False),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(0, 100), visible=True),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 100), visible=True),
            CornerVertex(position=Point(200, 100), visible=False),
            CornerVertex(position=Point(250, 50), visible=True),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(0, 0), visible=True),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)

    def testInvisibleFirstVertexPentagon(self):
        polyline = Polyline((
            CornerVertex(position=Point(0, 0), visible=False),
            CornerVertex(position=Point(200, 0), visible=True),
            CornerVertex(position=Point(250, 50), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(0, 100), visible=True),
        ))

        expectedPoly = Polyline((
            CornerVertex(position=Point(0, 100), visible=True),
            CornerVertex(position=Point(200, 100), visible=True),
            CornerVertex(position=Point(250, 50), visible=True),
            CornerVertex(position=Point(200, 0), visible=False),
            CornerVertex(position=Point(0, 0), visible=True),
        ))

        polyline.changeDirectionWithVisibility()

        self.assertEqual(expectedPoly, polyline)


class ClockwiseTests(unittest.TestCase):
    # TODO IT-250 more tests, esp. with concave parts would be nice -- NG
    def testRectangleCWShouldBeClockwise(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
        )
        self.assertTrue(polyline.isClockwise())

    def testRectangleCCWShouldNotBeClockwise(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(100, 0)),

        )
        self.assertFalse(polyline.isClockwise())


class IsClosedTests(unittest.TestCase):
    # The four tests are the same as the corresponding tests of the same names of interiorcad.interface.surface.contour2d.Contour2DTests
    def testIsClosedOnAnOpenPathShouldBeFalse(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=False),
        )
        self.assertFalse(polyline.isClosed())

    def testIsClosedOnANonePathShouldBeFalse(self):
        polyline = Polyline()
        self.assertFalse(polyline.isClosed())

    def testIsClosedOnAClosedPathShouldBeTrue(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
        )
        self.assertTrue(polyline.isClosed())

    def testIsClosedOnAPathReturningToStartShouldBeTrue(self):
        polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
            CornerVertex(Point(100, 100), visible=False),
        )
        self.assertTrue(polyline.isClosed())


class ForceVisibilityTests(unittest.TestCase):
    def setUp(self):
        self.polyline = Polyline()

    def testNoVertex(self):
        self.polyline.forceVisibility()
        self.assertEqual(Polyline(), self.polyline)

    def testOneVisibleVertex(self):
        self.polyline.vertices = [
            CornerVertex(Point(1, 11), visible=True)
        ]

        self.polyline.forceVisibility()

        expectedPolyline = Polyline(
            CornerVertex(Point(1, 11), visible=True)
        )

        self.assertEqual(expectedPolyline, self.polyline)

    def testOneInvisibleVertex(self):
        self.polyline.vertices = [
            CornerVertex(Point(1, 11), visible=False)
        ]

        self.polyline.forceVisibility()

        expectedPolyline = Polyline(
            CornerVertex(Point(1, 11), visible=True)
        )

        self.assertEqual(expectedPolyline, self.polyline)

    def testOpenPolygonWith3Vertices(self):
        self.polyline.vertices = [
            CornerVertex(Point(1, 11), visible=True),
            CornerVertex(Point(2, 22), visible=True),
            CornerVertex(Point(3, 33), visible=False),
        ]

        self.polyline.forceVisibility()

        expectedPolyline = Polyline(
            CornerVertex(Point(1, 11), visible=True),
            CornerVertex(Point(2, 22), visible=True),
            CornerVertex(Point(3, 33), visible=True),
        )

        self.assertEqual(expectedPolyline, self.polyline)


class CalculateRadiusFrom3PointsTests(unittest.TestCase):
    # Same test as CustomPartPolylineArcRadiusFunctionTest in dev\XG Custom Part\Source\CustomPartPolyline.cpp
    def testCalculateRadiusFrom3Points(self):
        p1 = Point(0, 0)
        p2 = Point(30, 30)
        p3 = Point(100, 0)

        radius = Polyline.calculateRadiusFrom3Points(p1, p2, p3)

        self.assertAlmostEqual(53.851648, radius, places=3)


class LengthTests(unittest.TestCase):
    def setUp(self):
        self.polyline = Polyline()

    def testEmptyPolylineShouldHaveALengthoFZero(self):
        self.assertEqual(0.0, self.polyline.getLength())

    def testOpenPolygonInXWith2Vertices(self):
        self.polyline.vertices = [
            CornerVertex(Point(11, 0), visible=True),
            CornerVertex(Point(33, 0), visible=False),
        ]

        self.assertEqual(22.0, self.polyline.getLength())

    def testOpenPolygonInYWith2Vertices(self):
        self.polyline.vertices = [
            CornerVertex(Point(0, 11), visible=True),
            CornerVertex(Point(0, 33), visible=False),
        ]

        self.assertEqual(22.0, self.polyline.getLength())

    def testTheLengthOfAHalfCircleShouldBeTheHalfCircumference(self):
        RADIUS = 33

        self.polyline.vertices = [
            CornerVertex(Point(-RADIUS, 0), visible=True),
            RadiusVertex(Point(0, RADIUS), radius=RADIUS, visible=True),
            CornerVertex(Point(RADIUS, 0), visible=False),
        ]

        halfCircumference = math.pi * RADIUS

        self.assertEqual(halfCircumference, self.polyline.getLength())

    def testTheLengthOfANegativeHalfCircleShouldBeTheHalfCircumference(self):
        RADIUS = 33

        self.polyline.vertices = [
            CornerVertex(Point(RADIUS, 0), visible=True),
            RadiusVertex(Point(0, -RADIUS), radius=RADIUS, visible=True),
            CornerVertex(Point(-RADIUS, 0), visible=False),
        ]

        halfCircumference = math.pi * RADIUS

        self.assertEqual(halfCircumference, self.polyline.getLength())

    def testTheLengthOfACircleShouldBeTheCircumference(self):
        RADIUS = 33

        self.polyline.vertices = [
            CornerVertex(Point(-RADIUS, 0), visible=True),
            RadiusVertex(Point(0, RADIUS), radius=RADIUS, visible=True),
            CornerVertex(Point(RADIUS, 0), visible=True),
            RadiusVertex(Point(0, -RADIUS), radius=RADIUS, visible=True),
        ]

        circumference = 2 * math.pi * RADIUS

        self.assertEqual(circumference, self.polyline.getLength())

    def testTheLengthOfAQuarterCircleShouldBeTheQuarterCircumference(self):
        # Short Arc
        RADIUS = 33
        FACTOR = math.sqrt(2) / 2

        self.polyline.vertices = [
            CornerVertex(Point(RADIUS, 0), visible=True),
            RadiusVertex(Point(RADIUS * FACTOR, RADIUS * FACTOR), radius=RADIUS, visible=True),
            CornerVertex(Point(0, RADIUS), visible=False),
        ]

        quarterCircumference = math.pi / 2.0 * RADIUS

        self.assertEqual(quarterCircumference, self.polyline.getLength())

    def testTheLengthOfAThreeQuarterCircleShouldBeTheThreeQuarterCircumference(self):
        # Long Arc
        RADIUS = 33
        FACTOR = math.sqrt(2) / 2

        self.polyline.vertices = [
            CornerVertex(Point(RADIUS, 0), visible=True),
            RadiusVertex(Point(-RADIUS * FACTOR, -RADIUS * FACTOR), radius=RADIUS, visible=True),
            CornerVertex(Point(0, RADIUS), visible=False),
        ]

        threeQuarterCircumference = 3.0 * math.pi / 2.0 * RADIUS

        self.assertEqual(threeQuarterCircumference, self.polyline.getLength())
