# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest

from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.rectangle import Rectangle


class PolylineSegment:
    def __init__(self, startVertex, endVertex):
        self.start = startVertex
        self.end = endVertex

    def calculateMidPoint(self):
        return CornerVertex((self.start.position + self.end.position) / 2)

    def isSegmentOnBounds(self, bounds):
        return (bounds.isPointOnRectangle(self.start.position) and
                bounds.isPointOnRectangle(self.end.position) and
                bounds.isPointOnRectangle(self.calculateMidPoint().position))

    def distance(self, point):
        x1 = self.start.position.x
        x2 = self.end.position.x
        y1 = self.start.position.y
        y2 = self.end.position.y

        b = math.hypot(x2 - x1, y2 - y1)
        if b < 0.000001:
            return 0

        # http://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
        a = (point.x - x1) * (y2 - y1) - (point.y - y1) * (x2 - x1)
        dist = math.fabs(a) / b
        return dist

    def hasOverlappingPoints(self):
        return self.start.position == self.end.position

    def __eq__(self, other):
        return other is not None and self.start == other.start and self.end == other.end

    def __str__(self):
        return 'PolylineSegment({}, {})'.format(self.start, self.end,)

    def __repr__(self):
        return str(self)

    def invert(self):
        self.start, self.end = self.end, self.start

    def mirrorOnXAxis(self, width):
        self.start.position.y = width - self.start.position.y
        self.end.position.y = width - self.end.position.y

    def mirrorOnYAxis(self, length):
        self.start.position.x = length - self.start.position.x
        self.end.position.x = length - self.end.position.x

    @staticmethod
    def createLineSegmentFromTwoTuples(start, end):
        return PolylineSegment(CornerVertex(Point(start[0], start[1])), CornerVertex(Point(end[0], end[1])))

    def asTuple(self):
        return self.start, self.end


class PolylineSegmentTest(unittest.TestCase):
    def testCalculateMidPoint(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(100, 100)))
        self.assertEqual(CornerVertex(Point(50.0, 50.0)), segment.calculateMidPoint())

    def testIsSegmentOnBounds(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100)))
        bounds = Rectangle(0, 0, 200, 100)
        self.assertTrue(segment.isSegmentOnBounds(bounds))

    def testIsSegmentNotOnBounds(self):
        segment = PolylineSegment(CornerVertex(Point(100, 100)), CornerVertex(Point(100, 0)))
        bounds = Rectangle(0, 0, 200, 100)
        self.assertFalse(segment.isSegmentOnBounds(bounds))

    def testDistancePointToLine(self):
        segment = PolylineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(100, 10)))
        point = Point(0, 0)
        self.assertEqual(10, segment.distance(point))

    def testDistancePointToLine2(self):
        segment = PolylineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(100, 10)))
        point = Point(-10, 0)
        self.assertEqual(10, segment.distance(point))

    def testDistancePointToLine3(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(70.711, 70.711)))
        point = Point(7.071, -7.071)
        self.assertAlmostEqual(10.0, segment.distance(point), places=3)

    def testDistancePointToLine4(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(70.711, 70.711)))
        point = Point(1.927, -12.215)
        self.assertAlmostEqual(10.0, segment.distance(point), places=3)

    def testNotOverlappingPoints(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(70.711, 70.711)))
        self.assertFalse(segment.hasOverlappingPoints())

    def testOverlappingPoints(self):
        segment = PolylineSegment(CornerVertex(Point(10.0, 20.0)), CornerVertex(Point(10.0, 20.0)))
        self.assertTrue(segment.hasOverlappingPoints())

    def testMirrorOnYAxisShouldNotChangeASegmentAlongTheYAxisOfMirroring(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100)))
        expectedSegment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100)))
        segment.mirrorOnYAxis(0)
        self.assertEqual(expectedSegment, segment)

    def testMirrorOnYAxisOnASegmentAlongTheYAxis(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100)))
        expectedSegment = PolylineSegment(CornerVertex(Point(100, 0)), CornerVertex(Point(100, 100)))
        segment.mirrorOnYAxis(100)
        self.assertEqual(expectedSegment, segment)

    def testMirrorOnYAxisOnASegmentAlongTheXAxis(self):
        segment = PolylineSegment(CornerVertex(Point(0, 10)), CornerVertex(Point(100, 10)))
        expectedSegment = PolylineSegment(CornerVertex(Point(100, 10)), CornerVertex(Point(0, 10)))
        segment.mirrorOnYAxis(100)
        self.assertEqual(expectedSegment, segment)

    def testMirrorOnXAxisShouldNotChangeASegmentAlongTheXAxisOfMirroring(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(100, 0)))
        expectedSegment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(100, 0)))
        segment.mirrorOnXAxis(0)
        self.assertEqual(expectedSegment, segment)

    def testMirrorOnXAxisOnASegmentAlongTheXAxis(self):
        segment = PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(100, 0)))
        expectedSegment = PolylineSegment(CornerVertex(Point(0, 100)), CornerVertex(Point(100, 100)))
        segment.mirrorOnXAxis(100)
        self.assertEqual(expectedSegment, segment)

    def testMirrorOnXAxisOnASegmentAlongTheYAxis(self):
        segment = PolylineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, 100)))
        expectedSegment = PolylineSegment(CornerVertex(Point(10, 100)), CornerVertex(Point(10, 0)))
        segment.mirrorOnXAxis(100)
        self.assertEqual(expectedSegment, segment)

    def testEqualToNoneShouldReturnFalse(self):
        segment = PolylineSegment(CornerVertex(Point(10, 0)), CornerVertex(Point(10, 100)))
        NOTHING = None  # avoid PyCharm's static analyzer complains about comparing with None
        self.assertFalse(segment == NOTHING)
        self.assertTrue(segment != NOTHING)
        self.assertNotEqual(segment, NOTHING)
        # noinspection PyPep8
        self.assertFalse(None == segment)  # This tests None.__eq__, not PolylineSegment.__eq__


class CreateLineSegmentFromTwoTuplesTests(unittest.TestCase):
    def setUp(self):
        self.polyline = PolylineSegment.createLineSegmentFromTwoTuples((1, 2), (3, 4))

    def testStartAndEndPointsShouldBeComparableToEachAPoint(self):
        self.assertTrue(Point(1, 2), self.polyline.start)
        self.assertTrue(Point(3, 4), self.polyline.end)

    def testMirrorOnXAxisShouldWorkWithoutRaisingAnyException(self):
        LENGTH = 5
        self.polyline.mirrorOnXAxis(LENGTH)
        self.assertTrue(Point(1, 5 - 2), self.polyline.start)
        self.assertTrue(Point(3, 5 - 4), self.polyline.end)

    def testMirrorOnYAxisShouldWorkWithoutRaisingAnyException(self):
        LENGTH = 5
        self.polyline.mirrorOnYAxis(LENGTH)
        self.assertTrue(Point(5 - 1, 2), self.polyline.start)
        self.assertTrue(Point(5 - 3, 4), self.polyline.end)
