# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry import isNearlyGreaterOrEqual, isNearlyLesserOrEqual, isNearlyGreater
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polylinesegment import PolylineSegment
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.rectangle import Rectangle


class PolylineSegments(list):
    def addSegments(self, segments):
        self._addObjectOrListOfObjects(segments)

    def _addObjectOrListOfObjects(self, objectOrList):
        if isinstance(objectOrList, (tuple, list)):
            self.extend(objectOrList)
        else:
            self.append(objectOrList)

    def filterSegmentsOnBounds(self, bounds):
        segmentsOnBounds = PolylineSegments()
        for segment in self:
            if segment.isSegmentOnBounds(bounds):
                segmentsOnBounds.append(segment)
        return segmentsOnBounds

    def filterSegmentWithHighestYValue(self):
        foundSegment = None
        for segment in self:
            if foundSegment is None:
                foundSegment = self[0]
            if isNearlyGreaterOrEqual(segment.start.position.y, foundSegment.start.position.y) and isNearlyGreaterOrEqual(segment.end.position.y, foundSegment.end.position.y):
                foundSegment = segment
        return foundSegment

    def filterSegmentWithLowestYValue(self):
        foundSegment = None
        for segment in self:
            if foundSegment is None:
                foundSegment = self[0]
            if isNearlyLesserOrEqual(segment.start.position.y, foundSegment.start.position.y) and isNearlyLesserOrEqual(segment.end.position.y, foundSegment.end.position.y):
                foundSegment = segment
        return foundSegment

    def filterSegmentsWithCornerVertices(self):
        segments = PolylineSegments()
        for segment in self:
            if segment.start.isCornerVertex() and segment.end.isCornerVertex():
                segments.append(segment)
        return segments

    def filterSegmentsGreaterThan(self, minLength):
        segments = PolylineSegments()
        for segment in self:
            vector = segment.end.position - segment.start.position
            if isNearlyGreater(vector.magnitude(), minLength):
                segments.append(segment)
        return segments


class PolylineSegmentsTest(unittest.TestCase):
    def setUp(self):
        self.segments = PolylineSegments()

    def testGetSegmentWithHighestYValue(self):
        expectedSegment = PolylineSegment(
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(100, 100))
        )
        self.segments.addSegments(expectedSegment)
        self.assertEqual(expectedSegment, self.segments.filterSegmentWithHighestYValue())

    def testGetSegmentWithLowestYValue(self):
        expectedSegment = PolylineSegment(
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(100, 0))
        )
        self.segments.addSegments(expectedSegment)
        self.assertEqual(expectedSegment, self.segments.filterSegmentWithHighestYValue())

    def testGetSegmentsOnBounds(self):
        self.segments.addSegments([
            PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100))),
            PolylineSegment(CornerVertex(Point(0, 100)), CornerVertex(Point(100, 100))),
            PolylineSegment(CornerVertex(Point(100, 100)), CornerVertex(Point(100, 0))),
            PolylineSegment(CornerVertex(Point(100, 0)), CornerVertex(Point(0, 0)))
        ])

        expectedSegments = ([
            PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100))),
            PolylineSegment(CornerVertex(Point(0, 100)), CornerVertex(Point(100, 100))),
            PolylineSegment(CornerVertex(Point(100, 0)), CornerVertex(Point(0, 0))),
        ])

        self.assertEqual(expectedSegments,
               self.segments.filterSegmentsOnBounds(Rectangle(0, 0, 200, 100)))

    def testFilterSegmentsWithCornerVertices(self):
        self.segments.addSegments(PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100))))
        self.segments.addSegments(PolylineSegment(CornerVertex(Point(0, 100)), RadiusVertex(Point(100, 100))))
        expectedSegments = PolylineSegments()
        expectedSegments.addSegments(PolylineSegment(CornerVertex(Point(0, 0)), CornerVertex(Point(0, 100))))
        self.assertEqual(expectedSegments, self.segments.filterSegmentsWithCornerVertices())
