# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest

from interiorcad.interface.geometry import EPSILON, isNearlyEqual, isNearlyLesser
from interiorcad.interface.geometry.point import Point


class Rectangle:
    def __init__(self, left=0, bottom=0, right=0, top=0):
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    @classmethod
    def createFromValues(cls, left, bottom, right, top):
        return cls(left, bottom, right, top)

    def __eq__(self, other):
        return all((
            self.left == other.left,
            self.top == other.top,
            self.right == other.right,
            self.bottom == other.bottom
        ))

    def isEmpty(self):
        return self.right - self.left < EPSILON and self.top - self.bottom < EPSILON

    def isDegenerated(self):
        return self.right - self.left < EPSILON or self.top - self.bottom < EPSILON

    def isPointOnRectangle(self, point):
        return any((
            isNearlyEqual(point.x, self.left),
            isNearlyEqual(point.x, self.right),
            isNearlyEqual(point.y, self.bottom),
            isNearlyEqual(point.y, self.top),
        ))

    def isPointStrictlyInside(self, point):
        return isNearlyLesser(self.left, point.x) and isNearlyLesser(point.x, self.right) and isNearlyLesser(self.bottom, point.y) and isNearlyLesser(point.y, self.top)

    def moveBy(self, delta):
        self.right += delta.x
        self.left += delta.x
        self.top += delta.y
        self.bottom += delta.y

    def mirrorOnZeroXAxis(self):
        self.top, self.bottom = -self.bottom, -self.top

    def mirrorOnZeroYAxis(self):
        self.left, self.right = -self.right, -self.left

    def mirrorOnZeroXAxisNoNegation(self):
        self.top, self.bottom = self.bottom, self.top

    def mirrorOnZeroYAxisNoNegation(self):
        self.left, self.right = self.right, self.left

    def rotateLeftAtZero(self):
        self.left, self.right, self.bottom, self.top = self.top, self.bottom, -self.left, -self.right

    def rotateRightAtZero(self):
        self.left, self.right, self.bottom, self.top = -self.bottom, -self.top, self.right, self.left

    def rotate180AtZero(self):
        self.left, self.right, self.bottom, self.top = -self.right, -self.left, -self.top, -self.bottom

    def rotateLeftAtZeroNoNegation(self):
        self.left, self.right, self.bottom, self.top = self.top, self.bottom, self.left, self.right

    def rotateRightAtZeroNoNegation(self):
        self.left, self.right, self.bottom, self.top = self.bottom, self.top, self.right, self.left

    def rotate180AtZeroNoNegation(self):
        self.left, self.right, self.bottom, self.top = self.right, self.left, self.top,  self.bottom

    def topRight(self):
        return Point(self.right, self.top)

    def bottomLeft(self):
        return Point(self.left, self.bottom)

    def getLength(self):
        return self.right - self.left

    def getWidth(self):
        return self.top - self.bottom

    def __rawString(self):
        return 'Rectangle(left={}, bottom={}, right={}, top={})'

    def __str__(self):
        return self.__rawString().format(str(self.left), str(self.bottom), str(self.right), str(self.top))

    def __repr__(self):
        return self.__rawString().format(repr(self.left), repr(self.bottom), repr(self.right), repr(self.top))

    def expand(self, addition):
        self.left -= addition
        self.right += addition
        self.top += addition
        self.bottom -= addition

    # Old name
    @property
    def edge1(self):
        return self.left

    # The old name did not exist. This is just for consistency
    @property
    def edge3(self):
        return self.top

    # Old name
    @property
    def edge2(self):
        return self.right

    # Old name
    @property
    def edge4(self):
        return self.bottom


class RectangleCreationTest(unittest.TestCase):
    def setUp(self):
        self.r = Rectangle()

    def testDefaultRectangleHasAllZeros(self):
        self.assertEqual(0, self.r.top)
        self.assertEqual(0, self.r.left)
        self.assertEqual(0, self.r.right)
        self.assertEqual(0, self.r.bottom)

    def testCanSetLeft(self):
        self.r.left = 1
        self.assertEqual(0, self.r.top)
        self.assertEqual(1, self.r.left)
        self.assertEqual(0, self.r.right)
        self.assertEqual(0, self.r.bottom)

    def testCanSetTop(self):
        self.r.top = 1
        self.assertEqual(1, self.r.top)
        self.assertEqual(0, self.r.left)
        self.assertEqual(0, self.r.right)
        self.assertEqual(0, self.r.bottom)

    def testCanSetRight(self):
        self.r.right = 1
        self.assertEqual(0, self.r.top)
        self.assertEqual(0, self.r.left)
        self.assertEqual(1, self.r.right)
        self.assertEqual(0, self.r.bottom)

    def testCanSetBottom(self):
        self.r.bottom = 1
        self.assertEqual(0, self.r.top)
        self.assertEqual(0, self.r.left)
        self.assertEqual(0, self.r.right)
        self.assertEqual(1, self.r.bottom)

    def testCanUseConstructorForInitialization(self):
        r = Rectangle(0, 10, 20, 30)
        self.assertEqual(0, r.left)
        self.assertEqual(10, r.bottom)
        self.assertEqual(30, r.top)
        self.assertEqual(20, r.right)

    def testLengthAndWidth(self):
        r = Rectangle(0, 11, 20, 33)
        self.assertEqual(20, r.getLength())
        self.assertEqual(22, r.getWidth())

    def testCreateFromValues(self):
        r = Rectangle.createFromValues(1, 2, 3, 4)
        self.assertEqual(1, r.left)
        self.assertEqual(2, r.bottom)
        self.assertEqual(3, r.right)
        self.assertEqual(4, r.top)


class RectangleTest(unittest.TestCase):
    def setUp(self):
        self.r = Rectangle()

    def testCanCompareRectangles(self):
        self.assertEqual(self.r, Rectangle())
        self.assertFalse(self.r == Rectangle(10))

    def testDefaultRectangleIsEmpty(self):
        self.assertTrue(self.r.isEmpty())

    def testRectangleIsNotEmpty(self):
        self.r.top = 100
        self.r.right = 100
        self.assertFalse(self.r.isEmpty())

    def testMoveByDelta(self):
        self.r.top = 50
        self.r.right = 20
        self.r.moveBy(Point(10, 50.0))
        expected = Rectangle(left=10, top=100, right=30, bottom=50)
        self.assertEqual(expected, self.r)

    def testRectanglesAreEqual(self):
        self.assertEqual(Rectangle(), self.r)

    def testIsPointOnRectangle(self):
        self.r.top = 200
        self.r.right = 100
        pointToCheck1 = Point(0, 0)
        pointToCheck2 = Point(50, 0)
        pointToCheck3 = Point(70, 200)
        pointToCheck4 = Point(100, 50)
        pointToCheck5 = Point(50, 50)
        self.assertTrue(self.r.isPointOnRectangle(pointToCheck1))
        self.assertTrue(self.r.isPointOnRectangle(pointToCheck2))
        self.assertTrue(self.r.isPointOnRectangle(pointToCheck3))
        self.assertTrue(self.r.isPointOnRectangle(pointToCheck4))
        self.assertFalse(self.r.isPointOnRectangle(pointToCheck5))

    def testRectangleHasStringRepresentation(self):
        self.r.left = '1'
        self.r.right = '2'
        self.r.top = '3'
        self.r.bottom = '4'
        expectedStr = 'Rectangle(left=1, bottom=4, right=2, top=3)'
        expectedRepr = "Rectangle(left='1', bottom='4', right='2', top='3')"
        self.assertEqual(expectedStr, str(self.r))
        self.assertEqual(expectedRepr, repr(self.r))

    def testTopRight(self):
        self.r.top = 1
        self.r.right = 2
        self.assertEqual(Point(2, 1), self.r.topRight())

    def testBottomLeft(self):
        self.r.bottom = 3
        self.r.left = 4
        self.assertEqual(Point(4, 3), self.r.bottomLeft())

    def testMirrorOnZeroXAxis(self):
        r = Rectangle(1, 3, 5, 7)
        r.mirrorOnZeroXAxis()
        expectedBounds = Rectangle(1, -7, 5, -3)
        self.assertEqual(expectedBounds, r)

    def testMirrorOnZeroYAxis(self):
        r = Rectangle(1, 3, 5, 7)
        r.mirrorOnZeroYAxis()
        expectedBounds = Rectangle(-5, 3, -1, 7)
        self.assertEqual(expectedBounds, r)

    def testMirrorOnZeroXAxisNoNegation(self):
        r = Rectangle(1, 3, 5, 7)
        r.mirrorOnZeroXAxisNoNegation()
        expectedBounds = Rectangle(1, 7, 5, 3)
        self.assertEqual(expectedBounds, r)

    def testMirrorOnZeroYAxisNoNegation(self):
        r = Rectangle(1, 3, 5, 7)
        r.mirrorOnZeroYAxisNoNegation()
        expectedBounds = Rectangle(5, 3, 1, 7)
        self.assertEqual(expectedBounds, r)

    def testRotateLeftAtZero(self):
        r = Rectangle(1, 3, 5, 7)
        r.rotateLeftAtZero()
        expectedBounds = Rectangle(7, -1, 3, -5)
        self.assertEqual(expectedBounds, r)

    def testRotateRightAtZero(self):
        r = Rectangle(1, 3, 5, 7)
        r.rotateRightAtZero()
        expectedBounds = Rectangle(-3, 5, -7, 1)
        self.assertEqual(expectedBounds, r)

    def testRotateLeftAtZeroNoNegation(self):
        r = Rectangle(1, 3, 5, 7)
        r.rotateLeftAtZeroNoNegation()
        expectedBounds = Rectangle(7, 1, 3, 5)
        self.assertEqual(expectedBounds, r)

    def testRotateRightAtZeroNoNegation(self):
        r = Rectangle(1, 3, 5, 7)
        r.rotateRightAtZeroNoNegation()
        expectedBounds = Rectangle(3, 5, 7, 1)
        self.assertEqual(expectedBounds, r)
        
    def testExpand(self):
        self.r.expand(100)

        expectedBounds = Rectangle(-100, -100, 100, 100)
        self.assertEqual(expectedBounds, self.r)

    def testDefaultRectangleIsDegenerated(self):
        self.assertTrue(self.r.isDegenerated())

    def testLineAlongXIsDegenerated(self):
        self.r.right = 100
        self.assertTrue(self.r.isDegenerated())

    def testLineAlongYIsNotValid(self):
        self.r.top = 100
        self.assertTrue(self.r.isDegenerated())

    def testRightSmallerThanLeftIsDegenrated(self):
        self.r.top = 100
        self.r.right = -100
        self.assertTrue(self.r.isDegenerated())

    def testTopSmallerThanBottomIsDegenerated(self):
        self.r.top = -100
        self.r.right = 100
        self.assertTrue(self.r.isDegenerated())

    def testRectangleIsNotDegenerated(self):
        self.r.top = 100
        self.r.right = 100
        self.assertFalse(self.r.isDegenerated())


class IsPointStrictlyInsideTest(unittest.TestCase):
    def setUp(self):
        self.rectangle = Rectangle(1, 2, 3, 5)

    def testPointOutsideRectangle(self):
        self.assertFalse(self.rectangle.isPointStrictlyInside(Point(0, 0)))
        self.assertFalse(self.rectangle.isPointStrictlyInside(Point(10, 20)))

    def testPointInsideRectangle(self):
        self.assertTrue(self.rectangle.isPointStrictlyInside(Point(2, 4)))

    def testPointOnRectangle(self):
        self.assertFalse(self.rectangle.isPointStrictlyInside(Point(1, 2)))
        self.assertFalse(self.rectangle.isPointStrictlyInside(Point(3, 5)))
        self.assertFalse(self.rectangle.isPointStrictlyInside(Point(2, 2)))
        self.assertFalse(self.rectangle.isPointStrictlyInside(Point(2, 5)))


class IsPointStrictlyInsideFloatTest(unittest.TestCase):
    def setUp(self):
        self.rectangle = Rectangle(0, 0, 1194.5000000000007, 305.3)

    def testPointOutsideRectangle(self):
        self.assertFalse(self.rectangle.isPointStrictlyInside(Point(1194.5000000000005, 165.29999999999927)))
