# -*- coding: utf-8 -*-
# Copyright 2015-2016, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.point import Point


class SimpleLineSegment:
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def getStartPoint(self):
        return self.start

    def getEndPoint(self):
        return self.end

    def getMiddlePoint(self):
        return (self.end - self.start) / 2 + self.start

    def getDirectionVector(self):
        vector = (self.end - self.start)
        vector.normalize()
        return vector

    def getPerpendicularVector(self):
        return self.getDirectionVector().perp()

    def getStartAndEndPoints(self):
        return self.start, self.end

    def moveByVector(self, perpendicularVector):
        self.start += perpendicularVector
        self.end += perpendicularVector

    def moveLeft(self, offset):
        perpendicularVector = - self.getPerpendicularVector() * offset
        self.moveByVector(perpendicularVector)

    def moveRight(self, offset):
        perpendicularVector = self.getPerpendicularVector() * offset
        self.moveByVector(perpendicularVector)

    def increaseAtStartAndEnd(self, offset):
        self.increaseAtStart(offset)
        self.increaseAtEnd(offset)

    def increaseAtStart(self, offset):
        self.start -= self.getDirectionVector() * offset

    def increaseAtEnd(self, offset):
        self.end += self.getDirectionVector() * offset

    @property
    def length(self):
        vector = (self.end - self.start)
        return vector.magnitude()

    def __eq__(self, rhs):
        return self.start == rhs.start and self.end == rhs.end

    def __repr__(self):
        return 'SimpleLineSegment(Point({startx}, {starty}), Point({endx}, {endy}))'.format(startx=self.start.x, starty=self.start.y, endx=self.end.x, endy=self.end.y)


class LineOffsetTests(unittest.TestCase):
    def setUp(self):
        self.lineFromLeftToRight = SimpleLineSegment(start=Point(0, 0), end=Point(100, 0))
        self.lineFromRightToLeft = SimpleLineSegment(start=Point(100, 0), end=Point(0, 0))
        self.lineFromBottomtoTop = SimpleLineSegment(start=Point(0, 0), end=Point(0, 100))
        self.lineFromTopToBottom = SimpleLineSegment(start=Point(0, 100), end=Point(0, 0))

    def testRightOffsetPointsOfLineFromLeftToRight(self):
        rightLine = self.lineFromLeftToRight
        rightLine.moveRight(offset=10)
        expectedRightLine = (Point(0, -10), Point(100, -10))
        self.assertEqual(expectedRightLine, rightLine.getStartAndEndPoints())

    def testLeftOffsetPointsOfLineFromLeftToRight(self):
        leftLine = self.lineFromLeftToRight
        leftLine.moveLeft(offset=10)
        expectedLeftLine = (Point(0, 10), Point(100, 10))
        self.assertEqual(expectedLeftLine, leftLine.getStartAndEndPoints())

    def testRightOffsetPointsOfLineFromRightToLeft(self):
        rightLine = self.lineFromRightToLeft
        rightLine.moveRight(offset=10)
        expectedRightLine = (Point(100, 10), Point(0, 10))
        self.assertEqual(expectedRightLine, rightLine.getStartAndEndPoints())

    def testLeftOffsetPointsOfLineFromRightToLeft(self):
        leftLine = self.lineFromRightToLeft
        leftLine.moveLeft(offset=10)
        expectedLeftLine = (Point(100, -10), Point(0, -10))
        self.assertEqual(expectedLeftLine, leftLine.getStartAndEndPoints())

    def testRightOffsetPointsOfLineFromBottomtoTop(self):
        rightLine = self.lineFromBottomtoTop
        rightLine.moveRight(offset=10)
        expectedRightLine = (Point(10, 0), Point(10, 100))
        self.assertEqual(expectedRightLine, rightLine.getStartAndEndPoints())

    def testLeftOffsetPointsOfLineFromBottomtoTop(self):
        leftLine = self.lineFromBottomtoTop
        leftLine.moveLeft(offset=10)
        expectedLeftLine = (Point(-10, 0), Point(-10, 100))
        self.assertEqual(expectedLeftLine, leftLine.getStartAndEndPoints())

    def testRightOffsetPointsOfLineFromTopToBottom(self):
        rightLine = self.lineFromTopToBottom
        rightLine.moveRight(offset=10)
        expectedLeftLine = (Point(-10, 100), Point(-10, 0))
        self.assertEqual(expectedLeftLine, rightLine.getStartAndEndPoints())

    def testLeftOffsetPointsOfLineFromTopToBottom(self):
        leftLine = self.lineFromTopToBottom
        leftLine.moveLeft(offset=10)
        expectedLeftLine = (Point(10, 100), Point(10, 0))
        self.assertEqual(expectedLeftLine, leftLine.getStartAndEndPoints())

    def testGetStartAndEndPointsOfIncreasedSegment(self):
        line = self.lineFromLeftToRight
        line.increaseAtStartAndEnd(offset=10)
        expectedLine = (Point(-10, 0), Point(110, 0))
        self.assertEqual(expectedLine, line.getStartAndEndPoints())

    def testGetStartPointsOfIncreasedSegment(self):
        line = self.lineFromLeftToRight
        line.increaseAtStart(offset=10)
        expectedLine = (Point(-10, 0), Point(100, 0))
        self.assertEqual(expectedLine, line.getStartAndEndPoints())

    def testGetEndPointsOfIncreasedSegment(self):
        line = self.lineFromLeftToRight
        line.increaseAtEnd(offset=10)
        expectedLine = (Point(0, 0), Point(110, 0))
        self.assertEqual(expectedLine, line.getStartAndEndPoints())

    def testGetMiddlePoint(self):
        self.assertEqual(Point(50, 0), self.lineFromLeftToRight.getMiddlePoint())

    def testLength(self):
        self.assertEqual(100, self.lineFromLeftToRight.length)
        self.assertEqual(100, self.lineFromRightToLeft.length)
        self.assertEqual(100, self.lineFromBottomtoTop.length)
        self.assertEqual(100, self.lineFromTopToBottom.length)

    def testEq(self):
        self.assertTrue(SimpleLineSegment(start=Point(0, 0), end=Point(100, 0)) == SimpleLineSegment(start=Point(0, 0), end=Point(100, 0)))
        self.assertEqual(SimpleLineSegment(start=Point(0, 0), end=Point(100, 0)), SimpleLineSegment(start=Point(0, 0), end=Point(100, 0)))
        self.assertFalse(SimpleLineSegment(start=Point(0, 0), end=Point(100, 0)) == SimpleLineSegment(start=Point(1, 0), end=Point(100, 0)))
        self.assertFalse(SimpleLineSegment(start=Point(0, 0), end=Point(100, 0)) == SimpleLineSegment(start=Point(0, 0), end=Point(100, 2)))
        self.assertFalse(SimpleLineSegment(start=Point(0, 0), end=Point(100, 0)) == SimpleLineSegment(start=Point(1, 0), end=Point(100, 2)))
