# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


from interiorcad.interface.geometry import isNearlyEqual
import math
import unittest


class Vector:
    def __init__(self, x=0, y=0, z=0):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)

    def dot(self, other):
        return math.fsum((self.x * other.x, self.y * other.y, self.z * other.z))

    def cross(self, other):
        return Vector(
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
            self.x * other.y - self.y * other.x
        )

    def length(self):
        return math.sqrt(self.dot(self))

    def magnitude(self):
        return self.length()

    def __eq__(self, other):
        return isNearlyEqual(self.x, other.x) and isNearlyEqual(self.y, other.y) and isNearlyEqual(self.z, other.z)

    def __str__(self):
        return 'Vector({}, {}, {})'.format(self.x, self.y, self.z)

    def __repr__(self):
        return str(self)

    def __sub__(self, other):
        return Vector(self.x - other.x, self.y - other.y, self.z - other.z)

    def __abs__(self):
        return Vector(abs(self.x), abs(self.y), abs(self.z))

    def __round__(self, places):
        p = places
        return Vector(round(self.x, p), round(self.y, p), round(self.z, p))

    def __mul__(self, rhs):
        return Vector(self.x * rhs, self.y * rhs, self.z * rhs)

    def __truediv__(self, rhs):
        return Vector(self.x / rhs, self.y / rhs, self.z / rhs)

    def __add__(self, rhs):
        return Vector(self.x + rhs.x, self.y + rhs.y, self.z + rhs.z)

    def __neg__(self):
        return Vector(- self.x, - self.y, - self.z)

    def normalized(self):
        length = self.length()
        return self / length

    def asTuple(self):
        return self.x, self.y, self.z


class TestVectorCreation(unittest.TestCase):
    def setUp(self):
        self.v = Vector()

    def testDefaultVectorShouldHaveZeroXYZ(self):
        self.assertEqual(0, self.v.x)
        self.assertEqual(0, self.v.y)
        self.assertEqual(0, self.v.z)

    def testCanSetX(self):
        self.v.x = 1
        self.assertEqual(1, self.v.x)
        self.assertEqual(0, self.v.y)
        self.assertEqual(0, self.v.z)

    def testCanSetY(self):
        self.v.y = 1
        self.assertEqual(0, self.v.x)
        self.assertEqual(1, self.v.y)
        self.assertEqual(0, self.v.z)

    def testCanSetZ(self):
        self.v.z = 1
        self.assertEqual(0, self.v.x)
        self.assertEqual(0, self.v.y)
        self.assertEqual(1, self.v.z)

    def testCanCreateVectorWithXYZNonZero(self):
        v = Vector(1, 2, 3)
        self.assertEqual(1, v.x)
        self.assertEqual(2, v.y)
        self.assertEqual(3, v.z)


class TestVectorDotProduct(unittest.TestCase):
    def testDotProductWithResultZero(self):
        v1 = Vector()
        v2 = Vector(1, 1, 1)
        self.assertEqual(0, v1.dot(v1))
        self.assertEqual(0, v1.dot(v2))
        self.assertEqual(0, v2.dot(v1))

    def testDotProductWithResultOne(self):
        v1 = Vector(1, 0, 0)
        v2 = Vector(0, 1, 0)
        v3 = Vector(0, 0, 1)
        self.assertEqual(1, v1.dot(v1))
        self.assertEqual(1, v2.dot(v2))
        self.assertEqual(1, v3.dot(v3))

    def testDotProductWithResultZero2(self):
        v1 = Vector(1, 0, 0)
        v2 = Vector(0, 1, 0)
        v3 = Vector(0, 0, 1)
        self.assertEqual(0, v1.dot(v2))
        self.assertEqual(0, v1.dot(v3))
        self.assertEqual(0, v2.dot(v1))
        self.assertEqual(0, v2.dot(v3))
        self.assertEqual(0, v3.dot(v1))
        self.assertEqual(0, v3.dot(v2))


class TestVectorLength(unittest.TestCase):
    def testVectorsWithLengthZero(self):
        self.assertEqual(0, Vector().length())

    def testVectorsWithLengthOne(self):
        self.assertEqual(1, Vector(1, 0, 0).length())
        self.assertEqual(1, Vector(0, 1, 0).length())
        self.assertEqual(1, Vector(0, 0, 1).length())

    def testVectorsWithLengthThree(self):
        self.assertEqual(3, Vector(3, 0, 0).length())
        self.assertEqual(3, Vector(0, 3, 0).length())
        self.assertEqual(3, Vector(0, 0, 3).length())

    def testVectorsWithLengthSquarerootOfTwo(self):
        sqrt2 = math.sqrt(2)
        self.assertEqual(sqrt2, Vector(1, 1, 0).length())
        self.assertEqual(sqrt2, Vector(1, 0, 1).length())
        self.assertEqual(sqrt2, Vector(0, 1, 1).length())

    def testVectorsWithLengthSquarerootOfThree(self):
        self.assertEqual(math.sqrt(3), Vector(1, 1, 1).length())

    def testVectorsWithOtherLengths(self):
        self.assertAlmostEqual(3.7416574, Vector(2, 3, 1).length())
        self.assertAlmostEqual(8.7965902, Vector(3.6, 6.9, 4.1).length())


class TestVectorMath(unittest.TestCase):
    def setUp(self):
        self.v1 = Vector(1, 2, 3)
        self.v2 = Vector(4, 4, 4)
        self.vExpected = Vector(3, 2, 1)

    def testSubtractVectors(self):
        self.assertEqual(self.vExpected, self.v2 - self.v1)

    def testSubtractAndAssignVectors(self):
        self.v2 -= self.v1
        self.assertEqual(self.vExpected, self.v2)

    def testAbsolutePositiveVector(self):
        self.assertEqual(self.v1, abs(self.v1))

    def testAbsoluteNegativeVector(self):
        v = Vector(-1, -2, -3)
        self.assertEqual(self.v1, abs(v))
        self.assertNotEqual(v, abs(v))

    def testRoundVector(self):
        v = Vector(1.01, 2.02, 3.03)
        expected = Vector(1, 2, 3)
        self.assertEqual(expected, round(v, 1))

    def testMul(self):
        self.assertEqual(Vector(3, 6, 9), self.v1 * 3)

    def testAdd(self):
        self.assertEqual(Vector(4, 5, 6), self.v1 + Vector(3, 3, 3))

    def testNeg(self):
        vector = Vector(1, 2, 4)
        self.assertEqual(Vector(-1, -2, -4), - vector)
        self.assertEqual(vector * -1, - vector)

    def testDiv(self):
        vector = Vector(2, 4, 6)
        vectorDiv2 = vector / 2
        self.assertEqual(Vector(1, 2, 3), vectorDiv2)

    def testANormalizedvectorShouldHaveALengthOf1(self):
        vector = Vector(1, 1, 1)
        self.assertEqual(1, vector.normalized().length())


class TestVectorStrRepresentation(unittest.TestCase):
    def testStrIsCorrect(self):
        v = Vector(0, 0, 0)
        self.assertEqual('Vector(0.0, 0.0, 0.0)', str(v))


class TestEqualnessOfVectors(unittest.TestCase):
    def testSameVectorsAreEqual(self):
        v = Vector(0, 0, 0)
        self.assertEqual(v, v)

    def testEqualVectorsAreEqual(self):
        v1 = Vector(1, 2, 3)
        v2 = Vector(1, 2, 3)
        self.assertEqual(v1, v2)

    def testNearlyEqualVectorsAreEqual(self):
        v1 = Vector(1.0000001, 2.0000001, 3.0000001)
        v2 = Vector(0.9999999, 1.9999999, 2.9999999)
        v3 = Vector(1, 2, 3)
        self.assertEqual(v1, v2)
        self.assertEqual(v1, v3)
        self.assertEqual(v2, v3)
