# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import math
import unittest

from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.vector import Vector


class VectorMath:
    @staticmethod
    def angleBetween(v1, v2):
        divisor = v1.magnitude() * v2.magnitude()
        degrees = 90.0
        if divisor > 0.0:
            dotProduct = v1.dot(v2)
            if isNearlyEqual(dotProduct, divisor):
                degrees = 0.0
            elif isNearlyEqual(-dotProduct, divisor):
                degrees = 180.0
            else:
                normalizedDotProduct = dotProduct / divisor
                degrees = math.degrees(math.acos(normalizedDotProduct))

        return degrees

    @staticmethod
    def distanceBetween(v1, v2):
        x = v1.x - v2.x
        y = v1.y - v2.y
        z = v1.z - v2.z
        return Vector(x, y, z).length()

    @staticmethod
    def getDegreeAngleOfProjectedVectorInXYPlane(v1):
        angleRad = math.atan2(v1.y, v1.x)
        return math.degrees(angleRad)


class getAngleInDegreesTest(unittest.TestCase):
    def testAngleOfXOnly_ShouldBeZero(self):
        v = Vector(1, 0, 0)
        self.assertAlmostEqual(0, VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(v), places=7)

    def testAngleOfYOnly_ShouldBe90(self):
        v = Vector(0, 1, 0)
        self.assertAlmostEqual(90, VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(v), places=7)

    def testAngleOfMinusXOnly_ShouldBe180(self):
        v = Vector(-1, 0, 0)
        self.assertAlmostEqual(180, VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(v), places=7)

    def testAngleOfMinusYOnly_ShouldBeMinus90(self):
        v = Vector(0, -1, 0)
        self.assertAlmostEqual(-90, VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(v), places=7)

    def testAngleOfMinusXNegativeY_ShouldBeMinus180(self):
        v = Vector(-1, -0.1e-9, 0)
        self.assertAlmostEqual(-180, VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(v), places=7)


class DistanceBetweenTest(unittest.TestCase):
    vm = VectorMath

    def setUp(self):
        self.v1 = Vector(10.0, 10.0, 0.0)
        self.v2 = Vector(10.0, 10.0, 0.0)

    def testDistanceXOnly(self):
        self.v1.x = 50.0
        self.assertAlmostEqual(40.0, self.vm.distanceBetween(self.v1, self.v2))

    def testDistanceYOnly(self):
        self.v2.y = 50.0
        self.assertAlmostEqual(40.0, self.vm.distanceBetween(self.v1, self.v2))

    def testDistanceZOnly(self):
        self.v2.z = 50.0
        self.assertAlmostEqual(50.0, self.vm.distanceBetween(self.v1, self.v2))


class AngleBetweenTest(unittest.TestCase):
    X_POS = Vector(1.0, 0.0, 0.0)
    X_NEG = Vector(-1.0, 0.0, 0.0)
    Y_POS = Vector(0.0, 1.0, 0.0)
    Y_NEG = Vector(0.0, -1.0, 0.0)
    Z_POS = Vector(0.0, 0.0, 1.0)
    Z_NEG = Vector(0.0, 0.0, -1.0)
    vm = VectorMath

    def testOrthogonalXY(self):
        self.assertEqual(90, self.vm.angleBetween(self.X_POS, self.Y_POS))
        self.assertEqual(90, self.vm.angleBetween(self.X_POS, self.Y_NEG))
        self.assertEqual(90, self.vm.angleBetween(self.Y_POS, self.X_NEG))
        self.assertEqual(90, self.vm.angleBetween(self.Y_NEG, self.X_NEG))

    def testOrthogonalXZ(self):
        self.assertEqual(90, self.vm.angleBetween(self.X_POS, self.Z_POS))
        self.assertEqual(90, self.vm.angleBetween(self.X_POS, self.Z_NEG))
        self.assertEqual(90, self.vm.angleBetween(self.Z_POS, self.X_NEG))
        self.assertEqual(90, self.vm.angleBetween(self.Z_NEG, self.X_NEG))

    def testOrthogonalYZ(self):
        self.assertEqual(90, self.vm.angleBetween(self.Y_POS, self.Z_POS))
        self.assertEqual(90, self.vm.angleBetween(self.Y_POS, self.Z_NEG))
        self.assertEqual(90, self.vm.angleBetween(self.Z_POS, self.Y_NEG))
        self.assertEqual(90, self.vm.angleBetween(self.Z_NEG, self.Y_NEG))

    def testParallelSameVectors(self):
        self.assertEqual(0, self.vm.angleBetween(self.X_POS, self.X_POS))
        self.assertEqual(0, self.vm.angleBetween(self.Y_POS, self.Y_POS))
        self.assertEqual(0, self.vm.angleBetween(self.Z_POS, self.Z_POS))

    def testParallelOppositeVectors(self):
        self.assertEqual(180, self.vm.angleBetween(self.X_POS, self.X_NEG))
        self.assertEqual(180, self.vm.angleBetween(self.X_NEG, self.X_POS))
        self.assertEqual(180, self.vm.angleBetween(self.Y_POS, self.Y_NEG))
        self.assertEqual(180, self.vm.angleBetween(self.Y_NEG, self.Y_POS))
        self.assertEqual(180, self.vm.angleBetween(self.Z_POS, self.Z_NEG))
        self.assertEqual(180, self.vm.angleBetween(self.Z_NEG, self.Z_POS))

    def testGetAngleBetweenTwo3DVectors(self):
        v1 = 0.17364817766704105
        v2 = 0.98480775301218859
        self.assertAlmostEqual(100.0, self.vm.angleBetween(self.X_POS, Vector(-v1, -v2)))
        self.assertAlmostEqual(100.0, self.vm.angleBetween(self.X_POS, Vector(-v1, v2)))
        self.assertAlmostEqual(80.0, self.vm.angleBetween(self.X_POS, Vector(v1, -v2)))
        self.assertAlmostEqual(80.0, self.vm.angleBetween(self.X_POS, Vector(v1, v2)))
        self.assertAlmostEqual(170.0, self.vm.angleBetween(self.X_POS, Vector(-v2, -v1)))
        self.assertAlmostEqual(170.0, self.vm.angleBetween(self.X_POS, Vector(-v2, v1)))
        self.assertAlmostEqual(10.0, self.vm.angleBetween(self.X_POS, Vector(v2, -v1)))
        self.assertAlmostEqual(10.0, self.vm.angleBetween(self.X_POS, Vector(v2, v1)))
