# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest

from interiorcad.interface.geometry.point import Point

CORNER_VERTEX = 0
RADIUS_VERTEX = 4


class Vertex:
    CORNER_VERTEX = 0
    RADIUS_VERTEX = 4

    def __init__(self, position=None, visible=True):
        self.position = Point()
        self.visible = visible
        if position is not None:
            self.position = position

    def moveBy(self, delta):
        self.position += delta

    def mirrorOnZeroXAxis(self):
        self.position.mirrorOnZeroXAxis()

    def mirrorOnZeroYAxis(self):
        self.position.mirrorOnZeroYAxis()

    def rotateLeftAtZero(self):
        self.position.rotateLeftAtZero()

    def rotateRightAtZero(self):
        self.position.rotateRightAtZero()

    def rotate180AtZero(self):
        self.position.rotate180AtZero()

    def __eq__(self, other):
        return isinstance(other, self.__class__) and (self.position == other.position and
            self.visible == other.visible)

    def __str__(self):
        return ('Vertex(position={}, visible={})'
                .format(str(self.position), str(self.visible)))

    def __repr__(self):
        return self.__str__()

    def asTuple(self):
        return self.position.x, self.position.y, self.type, self.radius, self.visible


class VertexTest(unittest.TestCase):
    def setUp(self):
        self.v = Vertex()

    def testDefaultVertexHasPositionZeroZero(self):
        self.assertEqual(Point(0, 0), self.v.position)

    def testDefaultVertexIsVisible(self):
        self.assertTrue(self.v.visible)

    def testMoveVertex(self):
        self.v.position = Point(100, 200)
        self.v.moveBy(Point(10, 20))
        self.assertEqual(Point(110, 220), self.v.position)

    def testMirrorOnXAxis(self):
        self.v.position = Point(100, 200)
        self.v.mirrorOnZeroXAxis()
        self.assertEqual(Point(100, -200), self.v.position)

    def testMirrorOnYAxis(self):
        self.v.position = Point(100, 200)
        self.v.mirrorOnZeroYAxis()
        self.assertEqual(Point(-100, 200), self.v.position)


class VertexCreationTest(unittest.TestCase):
    def testCanCreateVertex(self):
        v = Vertex()
        self.assertTrue(v)

    def testCanSetPositionInConstructor(self):
        v = Vertex('position')
        self.assertEqual('position', v.position)

    def testCanSetVisibilityInConstructor(self):
        v = Vertex(visible=False)
        self.assertFalse(v.visible)


class VertexBoolOperatorTest(unittest.TestCase):
    def setUp(self):
        self.v = Vertex('1')

    def testSameVertexAreEqual(self):
        self.assertEqual(self.v, self.v)

    def testDefaultVerticesAreEqual(self):
        self.assertEqual(Vertex(), Vertex())

    def testSameVerticesAreEqual(self):
        self.assertEqual(Vertex('1'), self.v)

    def testDifferentVerticesAreNotEqual(self):
        self.assertNotEqual(Vertex('2'), self.v)

    def testVerticesWithDifferentVisibilityAreNotEqual(self):
        self.v.visible = False
        self.assertNotEqual(Vertex('1'), self.v)


class VertexStringRepresentationTest(unittest.TestCase):
    def testEasyVertex(self):
        v = Vertex(position='pos', visible='vis')
        self.assertEqual('Vertex(position=pos, visible=vis)', str(v))
        self.assertEqual("Vertex(position=pos, visible=vis)", repr(v))
