# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import logging
import os.path

import interiorcad.backend.paths
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.backend.xmlparser import XMLParser


# TODO IT-248 / STORY IC-8196: the messages are still those of MachineConfiguration, including index number
class KeyValueConfiguration(XMLParser):
    def __init__(self, xmlCode=None, path=None):
        self.configurations = None
        super().__init__()
        if xmlCode:
            self.parseFromString(xmlCode)
        elif path:
            self.parseFromFile(path)
        else:
            machineConfigurationFile = self.getFileName()
            if machineConfigurationFile:
                self.__parseFromFileIfPossible(machineConfigurationFile)
            else:
                self.parseFromString(self.getMockDefaultXMLCode())

        self.readFullXMLFile()

    @classmethod
    def createFromRepositoryFile(cls, relativePath):
        """Create the Machine Configuration from a file in the git repository

        Use only for tests, not for production code!"""

        return cls(path=os.path.join(interiorcad.backend.paths.pythonRoot, relativePath))

    def __parseFromFileIfPossible(self, machineConfigurationFile):
        if machineConfigurationFile and os.path.isfile(machineConfigurationFile):
            self.parseFromFile(machineConfigurationFile)
        elif machineConfigurationFile:
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 21, "Could not find machine configuration XML file which is supposed to be located at: '{scriptPath}'.")
            logging.warning(message.format(scriptPath=machineConfigurationFile))

    def getFileName(self):
        return None

    def parseMockDefault(self):
        raise NotImplemented

    def readFullXMLFile(self):
        raise NotImplemented

    def getMockDefaultXMLCode(self):
        raise NotImplemented

    def getKeyValues(self):
        return self.configurations

    def get(self, key, default=None):
        if key in self.configurations:
            return self.configurations[key]
        elif default is not None:
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 30, '''Key '{key}' not defined in MachineConfiguration. Taking default '{default}' instead.''')
            logging.debug(message.format(key=key, default=default))
            return default
        else:
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 31, '''Key '{key}' not defined in MachineConfiguration. The key has no default value.''')
            raise Exception(message.format(key=key))

    def getIfNotEmpty(self, key, default=None):
        value = None
        if key in self.configurations:
            value = self.configurations[key] or default

        if value:
            return value
        elif default is not None:
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 30, '''Key '{key}' not defined in MachineConfiguration. Taking default '{default}' instead.''')
            logging.info(message.format(key=key, default=default))
            return default
        else:
            message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 31, '''Key '{key}' not defined in MachineConfiguration. The key has no default value.''')
            raise Exception(message.format(key=key))

    def getAsFloat(self, key, default=None):
        return float(self.get(key, default))

    def getAsInt(self, key, default=None):
        return int(self.get(key, default))

    def getAsBool(self, key, default=None):
        # Careful: we must first convert to int(), only then to bool(). The reason: '0' is True as the string is not empty.
        return bool(self.getAsInt(key, default))
