# -*- coding: utf-8 -*-

# Copyright 2016-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import logging
import unittest

from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.machineconfiguration.toolaxis import ToolAxis


class Tool:
    DRILL, SAW, ROUTER = range(3)
    VERTICAL, HORIZONTAL, FIVE_AXIS = range(3)
    DRILLING, DRILLING_THROUGH, CENTER_PUNCH, GRID_DRILLING, CIRCULAR_POCKET, DADO, RABBET, MITER, CONTOUR, SIZING = range(10)

    def __init__(self, name):
        self.name = name
        self.type = Tool.DRILL
        self.face = Tool.VERTICAL
        self.properties = {}
        self.routings = []
        self.axis = ToolAxis()
        self.exportMachinings = True  # Backward compatibility (new in interiorcad 2019 F3)

    def doesSupportRouting(self, routing):
        return routing in self.routings

    def getToolID(self):
        if 'tool-id' in self.properties:
            return self.properties['tool-id']

    def isToolIDValidIntDirect(self):
        toolID = self.getToolID()
        # valid ToolID as Int must be an Int and more than 1 digit
        try:
            int(toolID)
            return len(toolID) > 1
        except ValueError:
            return False

    def getPropertyAsBool(self, key, default=False):
        try:
            return bool(int(self.properties.get(key, default)))
        except ValueError:
            return default

    def getPropertyAsFloat(self, key, default=-0.0):
        try:
            return float(self.properties.get(key, default))
        except ValueError:
            return default

    def getPropertyAsInt(self, key, default=0):
        try:
            return int(self.properties.get(key, default))
        except ValueError:
            return default

    def getPropertyAsString(self, key, default=''):
        try:
            return self.properties.get(key, default)
        except ValueError:
            return default

    def getDrillingDiameter(self):
        return self.getPropertyAsFloat('drilling-diameter')

    def getRouterDiameter(self):
        routerDiameter = self.getPropertyAsFloat('router-diameter')
        if isNearlyEqual(routerDiameter, 0):
            logging.info("Router diameter should not be 0! This could calculate a wrong tool path and damage the part, router or machine.")

        return routerDiameter

    def getSawCutWidth(self):
        return self.getPropertyAsFloat('saw-cut-width')

    def getMaxSawDadoWidth(self):
        return self.getPropertyAsFloat('saw-max-dado-width')

    def getInstalledDiameter(self):
        installedDiameter = None
        if 'drilling-installed-diameter' in self.properties:
            installedDiameter = self.getPropertyAsFloat('drilling-installed-diameter')

        if not installedDiameter:
            installedDiameter = self.getDrillingDiameter()

        return installedDiameter

    def isAllowedForDefaultCenterPunch(self):
        return self.getPropertyAsBool('drilling-default-center-punch')

    def getSortableUsableLength(self):
        usableLength = self.getPropertyAsFloat('router-usable-length')
        if isNearlyEqual(usableLength, 0.0):
            usableLength = float('+inf')
        return usableLength

    def isRouterCentered(self):
        return self.getPropertyAsBool('router-measured-centered')

    def getRouterMaximumDepthPerStep(self):
        depthPerStep = self.getPropertyAsFloat('router-maximum-depth-per-step')
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = float('+inf')
        return depthPerStep

    def getRouterRotationDirection(self):
        return self.getPropertyAsString('router-rotation-direction')


class InitTests(unittest.TestCase):
    def setUp(self):
        self.tool = Tool(name='TestTool')

    def testExportMachiningShouldBeTrueByDefault(self):
        self.assertTrue(self.tool.exportMachinings)


class PropertyGetterTests(unittest.TestCase):
    def setUp(self):
        self.tool = Tool(name='TestTool')

    def testGetPropertyAsFloatShouldWorkCorrectlyForFloats(self):
        self.tool.properties['test'] = '-3.14'
        self.assertEqual(-3.14, self.tool.getPropertyAsFloat('test'))

    def testGetPropertyAsFloatShouldReturnDefaultForUnknownKey(self):
        self.assertEqual(3.14, self.tool.getPropertyAsFloat('unknown key', 3.14))

    def testGetPropertyAsFloatShouldReturnDefaultForValueNotConvertibleToFloat(self):
        self.tool.properties['test'] = 'I am not a number'
        self.assertEqual(3.14, self.tool.getPropertyAsFloat('test', 3.14))

    def testGetPropertyAsBoolShouldRteurnTrueForOne(self):
        self.tool.properties['test'] = '1'
        self.assertTrue(self.tool.getPropertyAsBool('test'))

    def testGetPropertyAsBoolShouldReturnFalseForZero(self):
        self.tool.properties['test'] = '0'
        self.assertFalse(self.tool.getPropertyAsBool('test'))

    def testGetPropertyAsBoolShouldReturnTrueDefaultForUnknownKey(self):
        self.assertTrue(self.tool.getPropertyAsBool('unknown key', True))

    def testGetPropertyAsBoolShouldReturnFalseDefaultForUnknownKey(self):
        self.assertFalse(self.tool.getPropertyAsBool('unknown key', False))

    def testGetPropertyAsBoolShouldReturnTrueDefaultForValueNotConvertibleToBool(self):
        self.tool.properties['test'] = 'I am not a bool'
        self.assertTrue(self.tool.getPropertyAsBool('test', True))

    def testGetPropertyAsBoolShouldReturnFalseDefaultForValueNotConvertibleToBool(self):
        self.tool.properties['test'] = 'I am not a bool'
        self.assertFalse(self.tool.getPropertyAsBool('test', False))


class InstalledDiameterTests(unittest.TestCase):
    def setUp(self):
        self.tool = Tool(name='TestTool')
        self.tool.properties['drilling-diameter'] = '8'

    def testValidInstalledDiameter(self):
        self.tool.properties['drilling-installed-diameter'] = '8.1'

        self.assertEqual(8.1, self.tool.getInstalledDiameter())

    def testNoInstalledDiameterShouldReturnNormalDiameter(self):
        self.assertEqual(8, self.tool.getInstalledDiameter())

    def testAnIntegerZeroInstalledDiameterShouldReturnNormalDiameter(self):
        self.tool.properties['drilling-installed-diameter'] = '0'

        self.assertEqual(8, self.tool.getInstalledDiameter())

    def testAnFloatZeroInstalledDiameterShouldReturnNormalDiameter(self):
        self.tool.properties['drilling-installed-diameter'] = '0.0'

        self.assertEqual(8, self.tool.getInstalledDiameter())

    def testAnEmptyInstalledDiameterShouldReturnNormalDiameter(self):
        self.tool.properties['drilling-installed-diameter'] = ''

        self.assertEqual(8, self.tool.getInstalledDiameter())
