# -*- coding: utf-8 -*-

# Copyright 2016-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest.mock

from interiorcad.interface.geometry import isNearlyEqual, isNearlyLesserOrEqual, isNearlyGreaterOrEqual, isNearlyGreater, EPSILON
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.machineconfiguration.toolaxis import ToolAxis


class ToolDatabase:
    def __init__(self):
        self.tools = ()

    def __len__(self):
        return len(self.tools)

    def __bool__(self):
        return bool(self.tools)

    def getToolForRouting(self, routing):
        for tool in self.tools:
            if tool.doesSupportRouting(routing):
                return tool

    def getToolForRoutings(self, routing):
        return filter(lambda tool: tool.doesSupportRouting(routing), self.tools)

    def getToolsOfType(self, toolType):
        return filter(lambda tool: tool.type == toolType, self.tools)

    def __getTypedToolsForRouting(self, routing, toolType):
        return filter(lambda tool: tool.type == toolType and tool.doesSupportRouting(routing), self.tools)

    # TODO IT-251: there is a lot of code in this module with "chosenTool". Can we refactor without making reading the code more complicated -- NG
    def __getDadoSawWithCondition(self, condition, dadoWidth):
        chosenSawCutWidth = float('-inf')  # Start very low, as the sawCutWidth can be 0 ("don't care"), especially on old installations
        chosenTool = None

        for tool in filter(condition, self.__getTypedToolsForRouting(Tool.DADO, Tool.SAW)):
            maxDadoWidth = tool.getMaxSawDadoWidth()
            sawCutWidth = tool.getSawCutWidth()
            if isNearlyLesserOrEqual(sawCutWidth, dadoWidth) and (isNearlyEqual(maxDadoWidth, 0) or isNearlyLesserOrEqual(dadoWidth, maxDadoWidth)):
                if isNearlyGreater(sawCutWidth, chosenSawCutWidth):
                    chosenTool = tool
                    chosenSawCutWidth = sawCutWidth

        return chosenTool

    def __getDadoSawWithConditionAndFallback(self, condition, dadoWidth):
        tool = self.__getDadoSawWithCondition(condition, dadoWidth)
        if not tool:
            tool = self.getDadoSawWithCRotation(dadoWidth)
        return tool

    def __getCRotatedDadoSawOnFace(self, face, dadoWidth):
        return self.__getDadoSawWithCondition(lambda tool: tool.axis.cRotating and tool.face == face, dadoWidth)

    def getDadoSawAlongX(self, dadoWidth):
        return self.__getDadoSawWithConditionAndFallback(lambda tool: tool.axis.isXAllowed() and not tool.axis.cRotating, dadoWidth)

    def getDadoSawAlongY(self, dadoWidth):
        return self.__getDadoSawWithConditionAndFallback(lambda tool: tool.axis.isYAllowed() and not tool.axis.cRotating, dadoWidth)

    def __getWithPrioritizedFace(self, faces, getter):
        tool = None

        for face in faces:
            tool = getter(face)
            if tool:
                break

        return tool

    def getDadoSawWithCRotation(self, dadoWidth):
        return self.__getWithPrioritizedFace(faces=(Tool.VERTICAL, Tool.FIVE_AXIS), getter=lambda face: self.__getCRotatedDadoSawOnFace(face, dadoWidth))

    def getHorizontalDadoSawWithCRotation(self, dadoWidth):
        return self.__getWithPrioritizedFace(faces=(Tool.HORIZONTAL, Tool.FIVE_AXIS), getter=lambda face: self.__getCRotatedDadoSawOnFace(face, dadoWidth))

    # TODO IT-265: can this be replaced by __getBestToolForMilling? -- NG
    def __getBestRouterForDadoForFace(self, face, dadoWidth, depth):
        chosenDiameter = 0
        chosenTool = None

        for tool in self.__getTypedToolsForRouting(Tool.DADO, Tool.ROUTER):
            if tool.face == face:
                routerDiameter = tool.getPropertyAsFloat('router-diameter')
                usableLength = tool.getSortableUsableLength()
                if isNearlyGreater(routerDiameter, chosenDiameter) and isNearlyGreaterOrEqual(dadoWidth, routerDiameter) and isNearlyLesserOrEqual(depth, usableLength):
                    chosenTool = tool
                    chosenDiameter = routerDiameter

        return chosenTool

    def getBestRouterForDado(self, dadoWidth, depth):
        return self.__getWithPrioritizedFace(faces=(Tool.VERTICAL, Tool.FIVE_AXIS), getter=lambda face: self.__getBestRouterForDadoForFace(face=face, dadoWidth=dadoWidth, depth=depth))

    def getBestRouterForHorizontalDado(self, dadoWidth, depth):
        return self.__getWithPrioritizedFace(faces=(Tool.HORIZONTAL, Tool.FIVE_AXIS), getter=lambda face: self.__getBestRouterForDadoForFace(face=face, dadoWidth=dadoWidth, depth=depth))

    def __getBestToolForMilling(self, face, diameter, depth, routing, getter):
        def possibleRouters():
            for tool in self.getToolForRoutings(routing):
                if tool.face == face:
                    minimumDiameter = getter(tool)
                    usableLength = tool.getSortableUsableLength()
                    if diameter >= minimumDiameter and depth <= usableLength:
                        yield tool

        routers = tuple(possibleRouters())
        if routers:
            return max(routers, key=lambda tool: (getter(tool), -tool.getSortableUsableLength()))

    def getBestToolForCircularPocket(self, diameter, depth):
        return self.__getWithPrioritizedFace(faces=(Tool.VERTICAL, Tool.FIVE_AXIS), getter=lambda face: self.__getBestToolForMilling(face=face, diameter=diameter, depth=depth, routing=Tool.CIRCULAR_POCKET, getter=lambda tool: tool.getPropertyAsFloat('router-minimum-pocket-diameter')))

    def getBestToolForHorizontalCircularPocket(self, diameter, depth):
        return self.__getWithPrioritizedFace(faces=(Tool.HORIZONTAL, Tool.FIVE_AXIS), getter=lambda face: self.__getBestToolForMilling(face=face, diameter=diameter, depth=depth, routing=Tool.CIRCULAR_POCKET, getter=lambda tool: tool.getPropertyAsFloat('router-minimum-pocket-diameter')))

    def getBestToolForContour(self, width, depth):
        return self.__getWithPrioritizedFace(faces=(Tool.VERTICAL, Tool.FIVE_AXIS), getter=lambda face: self.__getBestToolForMilling(face=face, diameter=width, depth=depth, routing=Tool.CONTOUR, getter=lambda tool: tool.getRouterDiameter()))

    def getBestToolForHorizontalContour(self, width, depth):
        return self.__getWithPrioritizedFace(faces=(Tool.HORIZONTAL, Tool.FIVE_AXIS), getter=lambda face: self.__getBestToolForMilling(face=face, diameter=width, depth=depth, routing=Tool.CONTOUR, getter=lambda tool: tool.getRouterDiameter()))

    def __getToolByName(self, face, name, routing):
        for tool in self.getToolForRoutings(routing):
            if tool.face == face and tool.name == name:
                return tool

        return None

    def getToolByNameForVerticalContour(self, toolName):
        return self.__getWithPrioritizedFace(faces=(Tool.VERTICAL, Tool.FIVE_AXIS), getter=lambda face: self.__getToolByName(face, toolName, Tool.CONTOUR))

    def getToolByNameForHorizontalContour(self, toolName):
        return self.__getWithPrioritizedFace(faces=(Tool.HORIZONTAL, Tool.FIVE_AXIS), getter=lambda face: self.__getToolByName(face, toolName, Tool.CONTOUR))

    def getBestToolForRabbet(self, width, depth):
        chosenDiameter = float('+inf')
        chosenTool = None

        rabbetTools = tuple(self.getToolForRoutings(Tool.RABBET))
        for tool in rabbetTools:
            toolDiameter = tool.getPropertyAsFloat('router-diameter')
            if width < toolDiameter < chosenDiameter and isNearlyLesserOrEqual(depth, tool.getSortableUsableLength()):
                chosenTool = tool
                chosenDiameter = toolDiameter

        # FIXME IT-273: that is buggy in relation of usable length, see for what we did for sizings IC-23741 and IC-22425 -- NG
        if not chosenTool and rabbetTools:
            chosenTool = max(rabbetTools, key=lambda tool: tool.getPropertyAsFloat('router-diameter'))

        return chosenTool

    # millThroughDepth new in interiorcad 2020 F8
    # jointingRouter new in interiorcad 2020 F9.2 resp. 2021 F2.3
    def getBestToolForSizing(self, thickness, millThroughDepth=None, jointingRouter=False):
        chosenUsableLength = float('+inf')
        chosenTool = None

        allSizingTools = self.__getTypedToolsForRouting(Tool.SIZING, Tool.ROUTER)
        sizingTools = filter(lambda tool: tool.getPropertyAsBool('router-for-jointing') == jointingRouter, allSizingTools)

        for tool in sizingTools:
            referenceThickness = thickness if millThroughDepth is None or tool.isRouterCentered() else millThroughDepth
            usableLength = tool.getSortableUsableLength()

            if (referenceThickness <= usableLength) and (usableLength < chosenUsableLength or chosenTool is None):
                chosenTool = tool
                chosenUsableLength = usableLength

        return chosenTool

    def getFallbackToolForSizing(self):
        sizingTools = tuple(self.__getTypedToolsForRouting(Tool.SIZING, Tool.ROUTER))
        if sizingTools:
            return max(sizingTools, key=lambda tool: tool.getSortableUsableLength())

    def __getRoutingsWithConditionAndDiameter(self, routing, condition, diameter):
        for tool in self.getToolForRoutings(routing):
            if condition(tool) and isNearlyEqual(tool.getDrillingDiameter(), diameter):
                return tool

    def getGridDrillingAlongX(self, diameter):
        return self.__getRoutingsWithConditionAndDiameter(Tool.GRID_DRILLING, lambda tool: tool.axis.isXAllowed(), diameter)

    def getGridDrillingAlongY(self, diameter):
        return self.__getRoutingsWithConditionAndDiameter(Tool.GRID_DRILLING, lambda tool: tool.axis.isYAllowed(), diameter)

    def getGridDrillingAlongCRotated(self, diameter):
        return self.__getRoutingsWithConditionAndDiameter(Tool.GRID_DRILLING, lambda tool: tool.axis.cRotating, diameter)

    def getSupportedGridDrillingDiametersAlongX(self):
        for tool in self.getToolForRoutings(Tool.GRID_DRILLING):
            if tool.axis.isXAllowed():
                yield tool.getDrillingDiameter()

    def getSupportedGridDrillingDiametersAlongY(self):
        for tool in self.getToolForRoutings(Tool.GRID_DRILLING):
            if tool.axis.isYAllowed():
                yield tool.getDrillingDiameter()

    def getSupportedGridDrillingDiametersAlongCRotated(self):
        for tool in self.getToolForRoutings(Tool.GRID_DRILLING):
            if tool.axis.cRotating:
                yield tool.getDrillingDiameter()

    def __getDrillForDrillingOnFace(self, face, diameter):
        return self.__getRoutingsWithConditionAndDiameter(Tool.DRILLING, lambda tool: tool.face == face, diameter)

    def __getDrillForCenterPunchOnFace(self, face, diameter):
        tool = self.__getRoutingsWithConditionAndDiameter(Tool.CENTER_PUNCH, lambda tool: tool.face == face, diameter)
        if not tool:
            tool = self.__getDefaultDrillForCenterPunchOnFace(face)
        return tool

    def __getDefaultDrillForCenterPunchOnFace(self, face):
        for tool in self.getToolForRoutings(Tool.CENTER_PUNCH):
            if tool.face == face and tool.isAllowedForDefaultCenterPunch():
                return tool

    def __getHorizontalCenterPunchDrillWithConditionAndFallback(self, diameter, condition):
        tool = self.__getHorizontalDrillWithConditionForRouting(diameter, condition, Tool.CENTER_PUNCH)
        if not tool:
            tool = self.__getHorizontalDrillWithConditionForRouting(diameter=diameter, condition=lambda tool: tool.axis.cRotating, routing=Tool.CENTER_PUNCH)
        if not tool:
            tool = self.__getHorizontalDefaultCenterPunchDrillWithCondition(Tool.HORIZONTAL, condition)
        if not tool:
            tool = self.__getHorizontalDefaultCenterPunchDrillWithCondition(Tool.HORIZONTAL, condition=lambda tool: tool.axis.cRotating)
        if not tool:
            tool = self.getFiveAxisCenterPunchDrill(diameter)
        return tool

    def getVerticalDrillThrough(self, diameter):
        tool = self.__getRoutingsWithConditionAndDiameter(Tool.DRILLING_THROUGH, lambda tool: tool.face == Tool.VERTICAL, diameter)
        if not tool:
            tool = self.__getRoutingsWithConditionAndDiameter(Tool.DRILLING_THROUGH, lambda tool: tool.face == Tool.FIVE_AXIS, diameter)
        return tool

    def getVerticalDrill(self, diameter):
        tool = self.__getDrillForDrillingOnFace(Tool.VERTICAL, diameter)
        if not tool:
            tool = self.getFiveAxisDrill(diameter)
        return tool

    def __getHorizontalDrillWithConditionForRouting(self, diameter, condition, routing):
        for tool in self.getToolForRoutings(routing):
            if tool.face == Tool.HORIZONTAL and condition(tool) and isNearlyEqual(tool.getDrillingDiameter(), diameter):
                return tool

    def __getHorizontalDrillWithConditionAndFallback(self, diameter, condition):
        tool = self.__getHorizontalDrillWithConditionForRouting(diameter, condition, Tool.DRILLING)
        if not tool:
            tool = self.getHorizontalDrillC(diameter)
        return tool

    def getHorizontalDrillC(self, diameter):
        tool = self.__getHorizontalDrillWithConditionForRouting(diameter=diameter, condition=lambda tool: tool.axis.cRotating, routing=Tool.DRILL)
        if not tool:
            tool = self.getFiveAxisDrill(diameter)
        return tool

    def getHorizontalDrillXMinus(self, diameter):
        return self.__getHorizontalDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isXMinusAllowed() and not tool.axis.cRotating)

    def getHorizontalDrillYMinus(self, diameter):
        return self.__getHorizontalDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isYMinusAllowed() and not tool.axis.cRotating)

    def getHorizontalDrillXPlus(self, diameter):
        return self.__getHorizontalDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isXPlusAllowed() and not tool.axis.cRotating)

    def getHorizontalDrillYPlus(self, diameter):
        return self.__getHorizontalDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isYPlusAllowed() and not tool.axis.cRotating)

    def getFiveAxisDrill(self, diameter):
        return self.__getDrillForDrillingOnFace(Tool.FIVE_AXIS, diameter)

    def getVerticalCenterPunchDrill(self, diameter):
        tool = self.__getDrillForCenterPunchOnFace(Tool.VERTICAL, diameter)
        if not tool:
            tool = self.getFiveAxisCenterPunchDrill(diameter)
        return tool

    def __getHorizontalDefaultCenterPunchDrillWithCondition(self, diameter, condition):
        for tool in self.getToolForRoutings(Tool.CENTER_PUNCH):
            if tool.face == Tool.HORIZONTAL and tool.isAllowedForDefaultCenterPunch() and condition(tool):
                return tool

    def getHorizontalCenterPunchDrillC(self, diameter):
        tool = self.__getHorizontalDrillWithConditionForRouting(diameter=diameter, condition=lambda tool: tool.axis.cRotating, routing=Tool.CENTER_PUNCH)
        if not tool:
            tool = self.__getHorizontalDefaultCenterPunchDrillWithCondition(Tool.HORIZONTAL, condition=lambda tool: tool.axis.cRotating)
        if not tool:
            tool = self.getFiveAxisCenterPunchDrill(diameter)
        return tool

    def getHorizontalCenterPunchDrillXMinus(self, diameter):
        return self.__getHorizontalCenterPunchDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isXMinusAllowed() and not tool.axis.cRotating)

    def getHorizontalCenterPunchDrillYMinus(self, diameter):
        return self.__getHorizontalCenterPunchDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isYMinusAllowed() and not tool.axis.cRotating)

    def getHorizontalCenterPunchDrillXPlus(self, diameter):
        return self.__getHorizontalCenterPunchDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isXPlusAllowed() and not tool.axis.cRotating)

    def getHorizontalCenterPunchDrillYPlus(self, diameter):
        return self.__getHorizontalCenterPunchDrillWithConditionAndFallback(diameter=diameter, condition=lambda tool: tool.axis.isYPlusAllowed() and not tool.axis.cRotating)

    def getFiveAxisCenterPunchDrill(self, diameter):
        return self.__getDrillForCenterPunchOnFace(Tool.FIVE_AXIS, diameter)


class PocketDiameterToolsTests(unittest.TestCase):
    DEPTH = 8

    def createMockTool(self):
        tool = unittest.mock.Mock()
        tool.mockProperties = {}
        tool.doesSupportRouting.return_value = True
        tool.getPropertyAsFloat.side_effect = lambda key: tool.mockProperties[key]
        tool.getSortableUsableLength.return_value = float('+inf')  # As it is an old test, it does not test the usable length.
        tool.face = Tool.VERTICAL
        return tool

    def setUp(self):
        self.tool1 = self.createMockTool()
        self.tool1.mockProperties['router-minimum-pocket-diameter'] = 10

        self.tool2 = self.createMockTool()
        self.tool2.mockProperties['router-minimum-pocket-diameter'] = 20

        self.tool3 = self.createMockTool()
        self.tool3.mockProperties['router-minimum-pocket-diameter'] = 30

        self.tool4 = self.createMockTool()
        self.tool4.mockProperties['router-minimum-pocket-diameter'] = 40

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3, self.tool4)

    def testTooSmall(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=6, depth=self.DEPTH)
        self.assertIs(tool, None)

    def testSmallPocket(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=12, depth=self.DEPTH)
        self.assertIs(tool, self.tool1)

    def testMiddlePocket(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=22, depth=self.DEPTH)
        self.assertIs(tool, self.tool2)

    def testMiddlePocketWithEqualDiameter(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=20, depth=self.DEPTH)
        self.assertIs(tool, self.tool2)

    def testLargePocket(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=32, depth=self.DEPTH)
        self.assertIs(tool, self.tool3)

    def testHugePocket(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=42, depth=self.DEPTH)
        self.assertIs(tool, self.tool4)


class PocketDepthToolsTests(unittest.TestCase):
    DIAMETER = 12

    def createMockTool(self):
        tool = unittest.mock.Mock()
        tool.mockProperties = {'router-minimum-pocket-diameter': 10}
        tool.doesSupportRouting.return_value = True
        tool.getPropertyAsFloat.side_effect = lambda key: tool.mockProperties[key]
        tool.face = Tool.VERTICAL
        return tool

    def setUp(self):
        self.tool1 = self.createMockTool()
        self.tool1.getSortableUsableLength.return_value = 10

        self.tool2 = self.createMockTool()
        self.tool2.getSortableUsableLength.return_value = 20

        self.tool3 = self.createMockTool()
        self.tool3.getSortableUsableLength.return_value = 30

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3)

    def testSmallDepth(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=self.DIAMETER, depth=6)
        self.assertIs(tool, self.tool1)

    def testMiddleDepth(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=self.DIAMETER, depth=12)
        self.assertIs(tool, self.tool2)

    def testLargeDepth(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=self.DIAMETER, depth=22)
        self.assertIs(tool, self.tool3)

    def testTooDeep(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=self.DIAMETER, depth=32)
        self.assertIs(tool, None)


class PocketMixedDiameterAndDepthToolsTests(unittest.TestCase):
    def createMockTool(self):
        tool = unittest.mock.Mock()
        tool.mockProperties = {}
        tool.doesSupportRouting.return_value = True
        tool.getPropertyAsFloat.side_effect = lambda key: tool.mockProperties[key]
        tool.face = Tool.VERTICAL
        return tool

    def setUp(self):
        self.tool1 = self.createMockTool()
        self.tool1.getSortableUsableLength.return_value = 10
        self.tool1.mockProperties['router-minimum-pocket-diameter'] = 10

        self.tool2 = self.createMockTool()
        self.tool2.getSortableUsableLength.return_value = 20
        self.tool2.mockProperties['router-minimum-pocket-diameter'] = 30

        self.tool3 = self.createMockTool()
        self.tool3.getSortableUsableLength.return_value = 40
        self.tool3.mockProperties['router-minimum-pocket-diameter'] = 20

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3)

    def testSmallDepth(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=12, depth=6)
        self.assertIs(tool, self.tool1)

    def testMiddleDepthNoAvaibleTool(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=12, depth=12)
        self.assertIs(tool, None)

    def testLargeDepth(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=22, depth=22)
        self.assertIs(tool, self.tool3)

    def testTooDeepForLargestRouter(self):
        tool = self.toolDatabase.getBestToolForCircularPocket(diameter=32, depth=18)
        self.assertIs(tool, self.tool2)


class DadoRoutersTests(unittest.TestCase):
    DEPTH = 10

    ### IT-164: can the Mock be changed so that getPropertyAsFloat returned the mocked value only for 'router-minimum-pocket-diameter' -- NG
    def setUp(self):
        self.tool1 = unittest.mock.Mock()
        self.tool1.type = Tool.ROUTER
        self.tool1.face = Tool.VERTICAL
        self.tool1.doesSupportRouting.return_value = True
        self.tool1.getPropertyAsFloat.return_value = 10
        self.tool1.getSortableUsableLength.return_value = 50

        self.tool2 = unittest.mock.Mock()
        self.tool2.type = Tool.ROUTER
        self.tool2.face = Tool.VERTICAL
        self.tool2.doesSupportRouting.return_value = True
        self.tool2.getPropertyAsFloat.return_value = 20
        self.tool2.getSortableUsableLength.return_value = 50

        self.tool3 = unittest.mock.Mock()
        self.tool3.type = Tool.ROUTER
        self.tool3.face = Tool.VERTICAL
        self.tool3.doesSupportRouting.return_value = True
        self.tool3.getPropertyAsFloat.return_value = 30
        self.tool3.getSortableUsableLength.return_value = 50

        self.tool4 = unittest.mock.Mock()
        self.tool4.type = Tool.ROUTER
        self.tool4.face = Tool.VERTICAL
        self.tool4.doesSupportRouting.return_value = True
        self.tool4.getPropertyAsFloat.return_value = 40
        self.tool4.getSortableUsableLength.return_value = 50

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3, self.tool4)

    def testTooSmall(self):
        tool = self.toolDatabase.getBestRouterForDado(dadoWidth=6, depth=self.DEPTH)
        self.assertIs(tool, None)

    def testSmallDado(self):
        tool = self.toolDatabase.getBestRouterForDado(dadoWidth=12, depth=self.DEPTH)
        self.assertIs(tool, self.tool1)

    def testMiddleDado(self):
        tool = self.toolDatabase.getBestRouterForDado(dadoWidth=22, depth=self.DEPTH)
        self.assertIs(tool, self.tool2)

    def testLargeDado(self):
        tool = self.toolDatabase.getBestRouterForDado(dadoWidth=32, depth=self.DEPTH)
        self.assertIs(tool, self.tool3)

    def testHugeDado(self):
        tool = self.toolDatabase.getBestRouterForDado(dadoWidth=42, depth=self.DEPTH)
        self.assertIs(tool, self.tool4)

    def testExactDiameterShouldBeAllowed(self):
        tool = self.toolDatabase.getBestRouterForDado(dadoWidth=10, depth=self.DEPTH)
        self.assertIs(tool, self.tool1)

    def testNearlyExactDiameterShouldBeAllowed(self):
        tool = self.toolDatabase.getBestRouterForDado(dadoWidth=10 - EPSILON / 2, depth=self.DEPTH)
        self.assertIs(tool, self.tool1)


class RabbetToolsTests(unittest.TestCase):
    DEPTH = 10

    # TODO IT-163: can the Mock be changed so that getPropertyAsFloat returned the mocked value only for 'router-minimum-pocket-diameter' -- NG
    def setUp(self):
        self.tool1 = unittest.mock.Mock()
        self.tool1.doesSupportRouting.return_value = True
        self.tool1.getPropertyAsFloat.return_value = 16
        self.tool1.getSortableUsableLength.return_value = 100

        self.tool2 = unittest.mock.Mock()
        self.tool2.doesSupportRouting.return_value = True
        self.tool2.getPropertyAsFloat.return_value = 25
        self.tool2.getSortableUsableLength.return_value = 100

        self.tool3 = unittest.mock.Mock()
        self.tool3.doesSupportRouting.return_value = True
        self.tool3.getPropertyAsFloat.return_value = 50
        self.tool3.getSortableUsableLength.return_value = 100

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3)

    def testSmallRabbet(self):
        tool = self.toolDatabase.getBestToolForRabbet(width=10, depth=self.DEPTH)
        self.assertIs(tool, self.tool1)

    def testMiddleRabbet(self):
        tool = self.toolDatabase.getBestToolForRabbet(width=21, depth=self.DEPTH)
        self.assertIs(tool, self.tool2)

    def testBigRabbet(self):
        tool = self.toolDatabase.getBestToolForRabbet(width=38, depth=self.DEPTH)
        self.assertIs(tool, self.tool3)

    def testRabbetWiderThanLargestTool(self):
        tool = self.toolDatabase.getBestToolForRabbet(width=60, depth=self.DEPTH)
        self.assertIs(tool, self.tool3)

    def testNoToolsAtAll(self):
        self.toolDatabase = ToolDatabase()
        tool = self.toolDatabase.getBestToolForRabbet(width=10, depth=self.DEPTH)
        self.assertIsNone(tool)

    def testRabbetWiderThanTwoLargestIdenticalTools(self):   # IC-9348
        tool3bis = copy.copy(self.tool3)
        self.toolDatabase.tools += tool3bis,
        tool = self.toolDatabase.getBestToolForRabbet(width=60, depth=self.DEPTH)
        self.assertTrue(tool is self.tool3 or tool is tool3bis)


class SizingToolsTests(unittest.TestCase):
    def setUp(self):
        self.tool1 = self.__createSizingRouter(name='Tool 1', usableLength=16)
        self.tool2 = self.__createSizingRouter(name='Tool 2', usableLength=25)
        self.tool3 = self.__createSizingRouter(name='Tool 3', usableLength=50)

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3)

    def __createSizingRouter(self, name, usableLength):
        tool = Tool(name)
        tool.face = Tool.VERTICAL
        tool.routings = [Tool.SIZING]
        tool.properties = {
            'router-for-jointing': False,
            'router-measured-centered': False,
            'router-usable-length': usableLength,
        }
        tool.type = Tool.ROUTER
        return tool

    def testThinWorkpiece(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=10)
        self.assertIs(tool, self.tool1)

    def testStandardWorkpiece(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=20)
        self.assertIs(tool, self.tool2)

    def testThickWorkpiece(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=38)
        self.assertIs(tool, self.tool3)

    def testWorkpieceThickerThanToolWithLargestUsableLength(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=60)
        self.assertIsNone(tool)

    def testNoToolsAtAll(self):
        self.toolDatabase = ToolDatabase()
        tool = self.toolDatabase.getBestToolForSizing(thickness=10)
        self.assertIsNone(tool)

    def testAdditionIsNeededForNonCentered(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=16, millThroughDepth=18)
        self.assertIs(tool, self.tool2)

    def testAdditionIsNotNeededForCentered(self):
        self.tool1.properties['router-measured-centered'] = True
        tool = self.toolDatabase.getBestToolForSizing(thickness=16, millThroughDepth=18)
        self.assertIs(tool, self.tool1)

    def testARouterForJointingShouldNotBeChosenIfItIsNotWanted(self):
        self.tool2.properties['router-for-jointing'] = True
        tool = self.toolDatabase.getBestToolForSizing(thickness=16, millThroughDepth=18)
        self.assertIsNot(tool, self.tool2)
        self.assertIs(tool, self.tool3)

    def testARouterForJointingShouldBeChosenIfItIsWanted(self):
        self.tool2.properties['router-for-jointing'] = True
        tool = self.toolDatabase.getBestToolForSizing(thickness=16, millThroughDepth=18, jointingRouter=True)
        self.assertIs(tool, self.tool2)


class SizingZeroUsableLengthToolsTests(unittest.TestCase):
    def setUp(self):
        self.tool1 = self.__createSizingRouter(name='Tool 1', usableLength=16)
        self.tool2 = self.__createSizingRouter(name='Tool 2', usableLength=25)
        self.tool3 = self.__createSizingRouter(name='Tool 3', usableLength=0)

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3)

    def __createSizingRouter(self, name, usableLength):
        tool = Tool(name)
        tool.face = Tool.VERTICAL
        tool.routings = [Tool.SIZING]
        tool.properties = {
            'router-for-jointing': False,
            'router-measured-centered': False,
            'router-usable-length': usableLength,
        }
        tool.type = Tool.ROUTER
        return tool

    def testThinWorkpiece(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=10)
        self.assertIs(tool, self.tool1)

    def testStandardWorkpiece(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=20)
        self.assertIs(tool, self.tool2)

    def testThickWorkpiece(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=38)
        self.assertIs(tool, self.tool3)

    def testWorkpieceThickerThanToolWithLargestUsableLength(self):
        tool = self.toolDatabase.getBestToolForSizing(thickness=60)
        self.assertIs(tool, self.tool3)

    def testNoToolsAtAll(self):
        self.toolDatabase = ToolDatabase()
        tool = self.toolDatabase.getBestToolForSizing(thickness=10)
        self.assertIsNone(tool)


class ToolGetterTests(unittest.TestCase):
    DIAMETER = 5

    def __createSimpleDrill(self, name, face, routing):
        tool = Tool(name)
        tool.face = face
        tool.routings = [routing]
        tool.properties['drilling-diameter'] = self.DIAMETER
        tool.type = Tool.DRILL
        return tool

    def createSimpleHorizontalDrill(self, name, axis, routing):
        tool = self.__createSimpleDrill(name, Tool.HORIZONTAL, routing)
        tool.axis = axis
        return tool

    def createSimpleVerticalDrill(self, name, routing):
        return self.__createSimpleDrill(name, Tool.VERTICAL, routing)

    def createSimpleFiveAxisDrill(self, name, routing):
        return self.__createSimpleDrill(name, Tool.FIVE_AXIS, routing)

    def createSimpleSaw(self, name, axis, face=Tool.VERTICAL):
        tool = Tool(name)
        tool.face = face
        tool.routings = [Tool.DADO]
        tool.axis = axis
        tool.type = Tool.SAW
        tool.properties = {
            'saw-cut-width': 4,
            'saw-max-dado-width': 25,
        }
        return tool

    def createSimpleToolAxis(self, x, y, cRotating):
        toolAxis = ToolAxis()
        toolAxis.x = x
        toolAxis.y = y
        toolAxis.cRotating = cRotating
        return toolAxis

    def createSimpleRouter(self, name, face=Tool.VERTICAL):
        tool = Tool(name)
        tool.face = face
        tool.routings = [Tool.DADO]
        tool.type = Tool.ROUTER
        return tool

    def asDefaultCenterPunchTool(self, tool):
        tool.properties.update({'drilling-default-center-punch': 1})
        return tool


class VerticalDrillTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.DRILLING),
            self.createSimpleVerticalDrill('Vertical', Tool.DRILLING),
        )

    def testAVerticalToolShouldBeFound(self):
        tool = self.toolDatabase.getVerticalDrill(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.VERTICAL, tool.face)


class VerticalDrillFiveAxisFallbackTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.DRILLING),
        )

    def testAVerticalToolShouldBeFound(self):
        tool = self.toolDatabase.getVerticalDrill(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)


class VerticalDrillThroughTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.DRILLING_THROUGH),
            self.createSimpleVerticalDrill('Wrong Vertical Tool', Tool.DRILLING),
            self.createSimpleVerticalDrill('Vertical', Tool.DRILLING_THROUGH),
        )

    def testAVerticalDrillThroughToolShouldBeFound(self):
        tool = self.toolDatabase.getVerticalDrillThrough(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.VERTICAL, tool.face)
        self.assertIn(Tool.DRILLING_THROUGH, tool.routings)


class VerticalDrillThroughFiveAxisFallbackTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Wrong Five Axis Tool', Tool.DRILLING),
            self.createSimpleFiveAxisDrill('Five Axis', Tool.DRILLING_THROUGH),
        )

    def testAVerticalDrillThroughToolShouldBeFound(self):
        tool = self.toolDatabase.getVerticalDrillThrough(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)
        self.assertIn(Tool.DRILLING_THROUGH, tool.routings)


class HorizontalDrillTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleHorizontalDrill('Default Axis', ToolAxis(), Tool.DRILLING),
            self.createSimpleHorizontalDrill('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.DRILLING),
            self.createSimpleHorizontalDrill('X Minus', self.createSimpleToolAxis(x=ToolAxis.MINUS, y=ToolAxis.NONE, cRotating=False), Tool.DRILLING),
            self.createSimpleHorizontalDrill('X Plus', self.createSimpleToolAxis(x=ToolAxis.PLUS, y=ToolAxis.NONE, cRotating=False), Tool.DRILLING),
            self.createSimpleHorizontalDrill('Y Minus', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.MINUS, cRotating=False), Tool.DRILLING),
            self.createSimpleHorizontalDrill('Y Plus', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.PLUS, cRotating=False), Tool.DRILLING),
        )

    def testAnCHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalDrillC(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.cRotating)

    def testAnXMinusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalDrillXMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.MINUS, tool.axis.x)
        self.assertFalse(tool.axis.cRotating)

    def testAnYMinusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalDrillYMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.MINUS, tool.axis.y)
        self.assertFalse(tool.axis.cRotating)

    def testAnXPlusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalDrillXPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.PLUS, tool.axis.x)
        self.assertFalse(tool.axis.cRotating)

    def testAnYPlusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalDrillYPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.PLUS, tool.axis.y)
        self.assertFalse(tool.axis.cRotating)


class HorizontalDrillCRotatedFallbackTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.DRILLING),
            self.createSimpleHorizontalDrill('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.DRILLING),
        )

    def testAnXMinusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillXMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnXPlusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillXPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYMinusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillYMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYPlusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillYPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnCHorizontalToolShouldNotFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillC(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.HORIZONTAL, tool.face)


class HorizontalDrillFiveAxisFallbackTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.DRILLING),
        )

    def testAnXMinusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillXMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnXPlusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillXPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnYMinusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillYMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnYPlusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillYPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnCHorizontalToolShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalDrillC(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)


class DrillingsShouldBeNearlyEqualDrillingDiameterTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleVerticalDrill('Vertical', Tool.DRILLING),
            self.createSimpleHorizontalDrill('X Plus', self.createSimpleToolAxis(x=ToolAxis.PLUS, y=ToolAxis.NONE, cRotating=False), Tool.DRILLING),
            self.createSimpleHorizontalDrill('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.DRILLING),
            self.createSimpleFiveAxisDrill('Five Axis', Tool.DRILLING),
        )

    def testVerticalDrillingIsNEarlyEqualDrillingDiameter(self):
        tool = self.toolDatabase.getVerticalDrill(5.000000000000189)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.VERTICAL, tool.face)

    def testHorizontalXPlusDrillingIsNEarlyEqualDrillingDiameter(self):
        tool = self.toolDatabase.getHorizontalDrillXPlus(5.000000000000189)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.PLUS, tool.axis.x)
        self.assertFalse(tool.axis.cRotating)

    def testHorizontalCRotatedDrillingIsNEarlyEqualDrillingDiameter(self):
        tool = self.toolDatabase.getHorizontalDrillC(5.000000000000189)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.cRotating)

    def testFiveAxisDrillingIsNEarlyEqualDrillingDiameter(self):
        tool = self.toolDatabase.getFiveAxisDrill(5.000000000000189)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)


class DadoAxisBoundSawTests(ToolGetterTests):
    DADO_WIDTH = 20

    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleSaw('Default Axis', ToolAxis()),
            self.createSimpleSaw('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True)),
            self.createSimpleSaw('X', self.createSimpleToolAxis(x=ToolAxis.MINUS, y=ToolAxis.NONE, cRotating=False)),
            self.createSimpleSaw('Y', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.MINUS, cRotating=False)),
        )

    def testXShouldNotFindTheCRotatingSaw(self):
        tool = self.toolDatabase.getDadoSawAlongX(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertFalse(tool.axis.cRotating)

    def testYShouldNotFindTheCRotatingSaw(self):
        tool = self.toolDatabase.getDadoSawAlongY(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertFalse(tool.axis.cRotating)


class DadoAxisBoundSawTestsWithExactWidth(ToolGetterTests):
    DADO_WIDTH = 4

    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleSaw('Default Axis', ToolAxis()),
            self.createSimpleSaw('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True)),
            self.createSimpleSaw('X', self.createSimpleToolAxis(x=ToolAxis.MINUS, y=ToolAxis.NONE, cRotating=False)),
            self.createSimpleSaw('Y', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.MINUS, cRotating=False)),
        )

    def testXShouldFindTheXAxisSaw(self):
        tool = self.toolDatabase.getDadoSawAlongX(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertFalse(tool.axis.cRotating)

    def testYShouldFindTheYAxisSaw(self):
        tool = self.toolDatabase.getDadoSawAlongY(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertFalse(tool.axis.cRotating)

    def testCShouldFindTheCRotatingSaw(self):
        tool = self.toolDatabase.getDadoSawWithCRotation(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.cRotating)


class DadoVerticalSawFallbackTests(ToolGetterTests):
    DADO_WIDTH = 20

    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleSaw('Five Axis', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.FIVE_AXIS),
            self.createSimpleSaw('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.VERTICAL),
        )

    def testXShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getDadoSawAlongX(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.VERTICAL, tool.face)

    def testYShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getDadoSawAlongY(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.VERTICAL, tool.face)


class DadoFiveAxisSawFallbackTests(ToolGetterTests):
    DADO_WIDTH = 20

    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleSaw('Five Axis', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.FIVE_AXIS),
        )

    def testXShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getDadoSawAlongX(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)

    def testYShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getDadoSawAlongY(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)


class DadoSawWidthTests(ToolGetterTests):
    DADO_WIDTH = 20

    def setUp(self):
        self.toolDatabase = ToolDatabase()

        tool1 = self.createSimpleSaw('2mm', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.VERTICAL)
        tool1.properties['saw-cut-width'] = 2

        tool2 = self.createSimpleSaw('4mm', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.VERTICAL)
        tool2.properties['saw-cut-width'] = 4

        self.toolDatabase.tools = (
            tool1,
            tool2,
        )

    def testTheWiderSawShouldBeUsed(self):
        tool = self.toolDatabase.getDadoSawWithCRotation(self.DADO_WIDTH)
        self.assertIsNotNone(tool)
        self.assertEqual("4mm", tool.name)

    def testNearlyTheSawWidthCutShouldBeRecognized(self):
        tool = self.toolDatabase.getDadoSawWithCRotation(2.0 - EPSILON / 2)
        self.assertIsNotNone(tool)
        self.assertEqual("2mm", tool.name)
        self.assertEqual(2.0, tool.getSawCutWidth())


class VerticalCenterPunchDrillTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.CENTER_PUNCH),
            self.createSimpleVerticalDrill('Vertical', Tool.CENTER_PUNCH),
        )

    def testAVerticalToolShouldBeFound(self):
        tool = self.toolDatabase.getVerticalCenterPunchDrill(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.VERTICAL, tool.face)


class VerticalCenterPunchDrillFiveAxisFallbackTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.CENTER_PUNCH),
        )

    def testAVerticalToolShouldBeFound(self):
        tool = self.toolDatabase.getVerticalCenterPunchDrill(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)


class HorizontalCenterPunchDrillTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleHorizontalDrill('Default Axis', ToolAxis(), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('X Minus', self.createSimpleToolAxis(x=ToolAxis.MINUS, y=ToolAxis.NONE, cRotating=False), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('X Plus', self.createSimpleToolAxis(x=ToolAxis.PLUS, y=ToolAxis.NONE, cRotating=False), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('Y Minus', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.MINUS, cRotating=False), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('Y Plus', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.PLUS, cRotating=False), Tool.CENTER_PUNCH),
        )

    def testAnCHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillC(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.cRotating)

    def testAnXMinusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.MINUS, tool.axis.x)
        self.assertFalse(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYMinusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.MINUS, tool.axis.y)
        self.assertFalse(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnXPlusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.PLUS, tool.axis.x)
        self.assertFalse(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYPlusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.PLUS, tool.axis.y)
        self.assertFalse(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)


class HorizontalCenterPunchDrillCRotatingFallbackTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.CENTER_PUNCH),
        )

    def testAnXMinusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnXPlusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYMinusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYPlusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnCHorizontalToolShouldNotFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillC(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.HORIZONTAL, tool.face)
        self.assertTrue(tool.axis.cRotating)


class HorizontalCenterPunchDrillFiveAxisFallbackTests(ToolGetterTests):
    def setUp(self):
        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleFiveAxisDrill('Five Axis', Tool.CENTER_PUNCH),
        )

    def testAnXMinusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnXPlusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnYMinusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYMinus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnYPlusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYPlus(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnCHorizontalToolShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillC(self.DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)


class HorizontalCenterPunchDrillFirstDefaultTests(ToolGetterTests):
    SEARCHED_DIAMETER = 8

    def setUp(self):
        assert self.SEARCHED_DIAMETER != self.DIAMETER, "Searched diameter cannot be the same as the tool diameter!"

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.createSimpleHorizontalDrill('Default Axis', ToolAxis(), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('X Minus', self.createSimpleToolAxis(x=ToolAxis.MINUS, y=ToolAxis.NONE, cRotating=False), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('X Plus', self.createSimpleToolAxis(x=ToolAxis.PLUS, y=ToolAxis.NONE, cRotating=False), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('Y Minus', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.MINUS, cRotating=False), Tool.CENTER_PUNCH),
            self.createSimpleHorizontalDrill('Y Plus', self.createSimpleToolAxis(x=ToolAxis.NONE, y=ToolAxis.PLUS, cRotating=False), Tool.CENTER_PUNCH),
        )
        for tool in self.toolDatabase.tools:
            tool.properties['drilling-default-center-punch'] = "1"

    def testAnCHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillC(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.cRotating)

    def testAnXMinusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXMinus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.MINUS, tool.axis.x)
        self.assertFalse(tool.axis.cRotating)

    def testAnYMinusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYMinus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.MINUS, tool.axis.y)
        self.assertFalse(tool.axis.cRotating)

    def testAnXPlusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXPlus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.PLUS, tool.axis.x)
        self.assertFalse(tool.axis.cRotating)

    def testAnYPlusHorizontalToolShouldBeFound(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYPlus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(ToolAxis.PLUS, tool.axis.y)
        self.assertFalse(tool.axis.cRotating)


class HorizontalCenterPunchDrillLastHorizontalDefaultTests(ToolGetterTests):
    SEARCHED_DIAMETER = 8

    def setUp(self):
        assert self.SEARCHED_DIAMETER != self.DIAMETER, "Searched diameter cannot be the same as the tool diameter!"

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.asDefaultCenterPunchTool(self.createSimpleFiveAxisDrill('Five Axis', Tool.CENTER_PUNCH)),
            self.asDefaultCenterPunchTool(self.createSimpleHorizontalDrill('C Rotating', self.createSimpleToolAxis(x=ToolAxis.BOTH, y=ToolAxis.BOTH, cRotating=True), Tool.CENTER_PUNCH)),
        )

    def testAnXMinusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXMinus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnXPlusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXPlus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isXAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYMinusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYMinus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnYPlusHorizontalDrillingShouldFallbackToCRotatedIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYPlus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.isYAllowed())
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)

    def testAnCHorizontalToolShouldNotFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillC(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertTrue(tool.axis.cRotating)
        self.assertEqual(Tool.HORIZONTAL, tool.face)


class HorizontalCenterPunchDrillLastFiveAxisDefaultTests(ToolGetterTests):
    SEARCHED_DIAMETER = 8

    def setUp(self):
        assert self.SEARCHED_DIAMETER != self.DIAMETER, "Searched diameter cannot be the same as the tool diameter!"

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (
            self.asDefaultCenterPunchTool(self.createSimpleFiveAxisDrill('Five Axis', Tool.CENTER_PUNCH)),
        )

    def testAnXMinusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXMinus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnXPlusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillXPlus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnYMinusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYMinus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnYPlusHorizontalDrillingShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillYPlus(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)

    def testAnCHorizontalToolShouldFallbackToFiveAxisIfNoOtherToolIsAvailable(self):
        tool = self.toolDatabase.getHorizontalCenterPunchDrillC(self.SEARCHED_DIAMETER)
        self.assertIsNotNone(tool)
        self.assertEqual(Tool.FIVE_AXIS, tool.face)


class ContourDiameterToolsTests(unittest.TestCase):
    DEPTH = 8

    def createMockTool(self):
        tool = unittest.mock.Mock()
        tool.doesSupportRouting.return_value = True
        tool.getSortableUsableLength.return_value = float('+inf')  # Do not test the usable length.
        tool.face = Tool.VERTICAL
        return tool

    def setUp(self):
        self.tool1 = self.createMockTool()
        self.tool1.getRouterDiameter.return_value = 10

        self.tool2 = self.createMockTool()
        self.tool2.getRouterDiameter.return_value = 20

        self.tool3 = self.createMockTool()
        self.tool3.getRouterDiameter.return_value = 30

        self.tool4 = self.createMockTool()
        self.tool4.getRouterDiameter.return_value = 40

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3, self.tool4)

    def testTooSmall(self):
        tool = self.toolDatabase.getBestToolForContour(width=6, depth=self.DEPTH)
        self.assertIs(tool, None)

    def testSmallContour(self):
        tool = self.toolDatabase.getBestToolForContour(width=12, depth=self.DEPTH)
        self.assertIs(tool, self.tool1)

    def testMiddleContour(self):
        tool = self.toolDatabase.getBestToolForContour(width=22, depth=self.DEPTH)
        self.assertIs(tool, self.tool2)

    def testMiddleContourWithEqualDiameter(self):
        tool = self.toolDatabase.getBestToolForContour(width=20, depth=self.DEPTH)
        self.assertIs(tool, self.tool2)

    def testLargeContour(self):
        tool = self.toolDatabase.getBestToolForContour(width=32, depth=self.DEPTH)
        self.assertIs(tool, self.tool3)

    def testHugeContour(self):
        tool = self.toolDatabase.getBestToolForContour(width=42, depth=self.DEPTH)
        self.assertIs(tool, self.tool4)


class ContourDepthToolsTests(unittest.TestCase):
    WIDTH = 12

    def createMockTool(self):
        tool = unittest.mock.Mock()
        tool.doesSupportRouting.return_value = True
        tool.getRouterDiameter.return_value = 10
        tool.face = Tool.VERTICAL
        return tool

    def setUp(self):
        self.tool1 = self.createMockTool()
        self.tool1.getSortableUsableLength.return_value = 10

        self.tool2 = self.createMockTool()
        self.tool2.getSortableUsableLength.return_value = 20

        self.tool3 = self.createMockTool()
        self.tool3.getSortableUsableLength.return_value = 30

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3)

    def testSmallDepth(self):
        tool = self.toolDatabase.getBestToolForContour(width=self.WIDTH, depth=6)
        self.assertIs(tool, self.tool1)

    def testMiddleDepth(self):
        tool = self.toolDatabase.getBestToolForContour(width=self.WIDTH, depth=12)
        self.assertIs(tool, self.tool2)

    def testLargeDepth(self):
        tool = self.toolDatabase.getBestToolForContour(width=self.WIDTH, depth=22)
        self.assertIs(tool, self.tool3)

    def testTooDeep(self):
        tool = self.toolDatabase.getBestToolForContour(width=self.WIDTH, depth=32)
        self.assertIs(tool, None)


class ContourMixedDiameterAndDepthToolsTests(unittest.TestCase):
    def createMockTool(self):
        tool = unittest.mock.Mock()
        tool.doesSupportRouting.return_value = True
        tool.face = Tool.VERTICAL
        return tool

    def setUp(self):
        self.tool1 = self.createMockTool()
        self.tool1.getSortableUsableLength.return_value = 10
        self.tool1.getRouterDiameter.return_value = 10

        self.tool2 = self.createMockTool()
        self.tool2.getSortableUsableLength.return_value = 20
        self.tool2.getRouterDiameter.return_value = 30

        self.tool3 = self.createMockTool()
        self.tool3.getSortableUsableLength.return_value = 40
        self.tool3.getRouterDiameter.return_value = 20

        self.toolDatabase = ToolDatabase()
        self.toolDatabase.tools = (self.tool1, self.tool2, self.tool3)

    def testSmallDepth(self):
        tool = self.toolDatabase.getBestToolForContour(width=12, depth=6)
        self.assertIs(tool, self.tool1)

    def testMiddleDepthNoAvaibleTool(self):
        tool = self.toolDatabase.getBestToolForContour(width=12, depth=12)
        self.assertIs(tool, None)

    def testLargeDepth(self):
        tool = self.toolDatabase.getBestToolForContour(width=22, depth=22)
        self.assertIs(tool, self.tool3)

    def testTooDeepForLargestRouter(self):
        tool = self.toolDatabase.getBestToolForContour(width=32, depth=18)
        self.assertIs(tool, self.tool2)
