# -*- coding: utf-8 -*-

# Copyright 2016-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from xml.etree import ElementTree
import unittest

from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.machineconfiguration.tooldatabase import ToolDatabase
from interiorcad.interface.machineconfiguration.toolaxis import ToolAxis


class ToolDatabaseReader:
    def __init__(self, toolDatabase):
        self.toolDatabase = toolDatabase

    def process(self, parentElement):
        toolElements = parentElement.findall('tool')
        self.toolDatabase.tools = tuple(self.readTools(toolElements))

    def readTools(self, toolElements):
        for toolElement in toolElements:
            if self.__StringToBool(toolElement.get('active', False)):
                yield self.readTool(toolElement)

    def readTool(self, toolElement):
        tool = Tool(name=toolElement.get('name'))
        tool.exportMachinings = self.__StringToBool(toolElement.get('export-machinings', True))  # Backward compatibility (new in interiorcad 2019 F3)
        self.readToolType(tool, toolElement)
        self.readAxis(tool, toolElement)
        self.readFace(tool, toolElement)
        self.readRoutings(tool, toolElement)
        self.readProperties(tool, toolElement)
        return tool

    def __StringToBool(self, value):
        return bool(int(value))

    def __getAttributeOfChild(self, toolElement, name):
        element = toolElement.find(name)
        if element is not None:
            return element.get(name, '').lower()

    def readToolType(self, tool, toolElement):
        attribute = self.__getAttributeOfChild(toolElement, 'type')
        if attribute == 'drill':
            tool.type = Tool.DRILL
        elif attribute == 'saw':
            tool.type = Tool.SAW
        elif attribute == 'router':
            tool.type = Tool.ROUTER
        else:
            tool.type = Tool.DRILL

    def readAxis(self, tool, toolElement):
        axisElement = toolElement.find('axis')
        if axisElement is not None:
            tool.axis.x = self.__AxisStringToInteger(axisElement.get('x', '').lower())
            tool.axis.y = self.__AxisStringToInteger(axisElement.get('y', '').lower())
            tool.axis.cRotating = self.__StringToBool(axisElement.get('rotating'))

    def __AxisStringToInteger(self, value):
        if value == 'none':
            return ToolAxis.NONE
        elif value == 'plus':
            return ToolAxis.PLUS
        elif value == 'minus':
            return ToolAxis.MINUS
        elif value == 'both':
            return ToolAxis.BOTH
        else:
            return ToolAxis.NONE

    def readRoutings(self, tool, toolElement):
        routingElement = toolElement.find('routing')
        if routingElement is not None:
            for key, value in routingElement.items():
                if not self.__StringToBool(value):
                    pass
                elif key == "drilling":
                    tool.routings.append(Tool.DRILLING)
                elif key == "drilling-through":
                    tool.routings.append(Tool.DRILLING_THROUGH)
                elif key == "center-punch":
                    tool.routings.append(Tool.CENTER_PUNCH)
                elif key == "grid-drilling":
                    tool.routings.append(Tool.GRID_DRILLING)
                elif key == "circular-pocket":
                    tool.routings.append(Tool.CIRCULAR_POCKET)
                elif key == "dado":
                    tool.routings.append(Tool.DADO)
                elif key == "rabbet":
                    tool.routings.append(Tool.RABBET)
                elif key == "miter":
                    tool.routings.append(Tool.MITER)
                elif key == "contour":
                    tool.routings.append(Tool.CONTOUR)
                elif key == "sizing":
                    tool.routings.append(Tool.SIZING)

    def readFace(self, tool, toolElement):
        attribute = self.__getAttributeOfChild(toolElement, 'face')
        if attribute == 'vertical':
            tool.face = Tool.VERTICAL
        elif attribute == 'horizontal':
            tool.face = Tool.HORIZONTAL
        elif attribute == '5axis':
            tool.face = Tool.FIVE_AXIS
        else:
            tool.face = Tool.VERTICAL

    def readProperties(self, tool, toolElement):
        propertyElements = toolElement.findall('properties/property')
        tool.properties = dict((element.get('key', 'UNNAMED'), element.get('value')) for element in propertyElements)


class ReadTests(unittest.TestCase):
    def setUp(self):
        XML_TOOLS = """<tool-database>
            <tool name="drill-vertical" active="1">
                <type type="Drill" />
                <face face="Vertical" />
            </tool>
            <tool name="drill-horizontal" active="1">
                <type type="Drill" />
                <face face="Horizontal" />
            </tool>
            <tool name="drill-five-axis" active="1">
                <type type="Drill" />
                <face face="5Axis" />
            </tool>
            <tool name="saw" active="1">
                <type type="Saw" />
            </tool>
            <tool name="router" active="1">
                <type type="Router" />
            </tool>
            <tool name="empty-tool" active="1" />
            <tool name="not activated" />
            <tool name="deactivated" active="0" />
            <some-noise/>
        </tool-database>"""

        self.root = ElementTree.fromstring(XML_TOOLS)
        self.reader = ToolDatabaseReader(ToolDatabase())
        self.reader.process(self.root)

    def testToolsAvailable(self):
        self.assertEqual(6, len(self.reader.toolDatabase))

    def testToolsAreNamed(self):
        names = tuple(tool.name for tool in self.reader.toolDatabase.tools)
        self.assertEqual(("drill-vertical", "drill-horizontal", "drill-five-axis", "saw", "router", "empty-tool"), names)

    def testToolsAreTyped(self):
        names = tuple(tool.type for tool in self.reader.toolDatabase.tools)
        self.assertEqual((Tool.DRILL, Tool.DRILL, Tool.DRILL, Tool.SAW, Tool.ROUTER, Tool.DRILL), names)

    def testToolsAreFaced(self):
        names = tuple(tool.face for tool in self.reader.toolDatabase.tools)
        self.assertEqual((Tool.VERTICAL, Tool.HORIZONTAL, Tool.FIVE_AXIS, Tool.VERTICAL, Tool.VERTICAL, Tool.VERTICAL), names)


class ReadExportMachiningsTests(unittest.TestCase):
    def setUp(self):
        XML_TOOLS = """<tool-database>
            <tool name="export-machinings" active="1" export-machinings="1" />
            <tool name="implicit-export-machinings" active="1" />
            <tool name="do-not-export-machinings" active="1" export-machinings="0" />
            <some-noise/>
        </tool-database>"""

        self.root = ElementTree.fromstring(XML_TOOLS)
        self.reader = ToolDatabaseReader(ToolDatabase())
        self.reader.process(self.root)

    def testToolsAvailable(self):
        self.assertEqual(3, len(self.reader.toolDatabase))

    def testToolsAreNamed(self):
        names = tuple(tool.name for tool in self.reader.toolDatabase.tools)
        self.assertEqual(("export-machinings", "implicit-export-machinings", "do-not-export-machinings",), names)

    def testToolsHaveExportMachinings(self):
        names = tuple(tool.exportMachinings for tool in self.reader.toolDatabase.tools)
        self.assertEqual((True, True, False), names)


class PropertiesReadTests(unittest.TestCase):
    def setUp(self):
        XML_TOOLS = """<tool-database>
            <tool name="test" active="1">
                <properties>
                     <property key="a" value="A" />
                     <property key="b" value="B" />
                     <property key="c" value="C" />
                     <property key="d" value="D" />
                </properties>
            </tool>
        </tool-database>"""

        self.root = ElementTree.fromstring(XML_TOOLS)
        self.reader = ToolDatabaseReader(ToolDatabase())
        self.reader.process(self.root)

    def testToolsHaveProperties(self):
        tool = self.reader.toolDatabase.tools[0]

        expectedProperties = {'a': 'A', 'b': 'B', 'c': 'C', 'd': 'D'}
        self.assertEqual(expectedProperties, tool.properties)


class AllRoutingsReadTests(unittest.TestCase):
    def setUp(self):
        XML_TOOLS = """<tool-database>
            <tool name="test" active="1">
                 <routing
                  center-punch="1"
                  circular-pocket="1"
                  contour="1"
                  dado="1"
                  drilling="1"
                  drilling-through="1"
                  grid-drilling="1"
                  miter="1"
                  rabbet="1"
                  sizing="1"
                  />
            </tool>
        </tool-database>"""

        self.root = ElementTree.fromstring(XML_TOOLS)
        self.reader = ToolDatabaseReader(ToolDatabase())
        self.reader.process(self.root)

    def testToolsHaveRoutings(self):
        tool = self.reader.toolDatabase.tools[0]
        tool.routings.sort()

        expectedRoutings = [Tool.DRILLING, Tool.DRILLING_THROUGH, Tool.CENTER_PUNCH, Tool.GRID_DRILLING, Tool.CIRCULAR_POCKET, Tool.DADO, Tool.RABBET, Tool.MITER, Tool.CONTOUR, Tool.SIZING]
        self.assertEqual(expectedRoutings, tool.routings)


class NoRoutingsReadTests(unittest.TestCase):
    def setUp(self):
        XML_TOOLS = """<tool-database>
            <tool name="test" active="1">
                 <routing
                  center-punch="0"
                  circular-pocket="0"
                  contour="0"
                  dado="0"
                  drilling="0"
                  drilling-through="0"
                  grid-drilling="0"
                  miter="0"
                  rabbet="0"
                  sizing="0"
                  />
            </tool>
        </tool-database>"""

        self.root = ElementTree.fromstring(XML_TOOLS)
        self.reader = ToolDatabaseReader(ToolDatabase())
        self.reader.process(self.root)

    def testToolsHaveRoutings(self):
        tool = self.reader.toolDatabase.tools[0]
        self.assertEqual([], tool.routings)


class AxisReadTests(unittest.TestCase):
    def setUp(self):
        XML_TOOLS = """<tool-database>
            <tool name="test" active="1">
                <axis rotating="1" x="Both" y="Both"/>
            </tool>
        </tool-database>"""

        self.root = ElementTree.fromstring(XML_TOOLS)
        self.reader = ToolDatabaseReader(ToolDatabase())
        self.reader.process(self.root)

    def testToolsDefineAxis(self):
        tool = self.reader.toolDatabase.tools[0]

        expectedAxis = ToolAxis()
        expectedAxis.x = ToolAxis.BOTH
        expectedAxis.y = ToolAxis.BOTH
        expectedAxis.cRotating = True
        self.assertEqual(expectedAxis, tool.axis)
