# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest

from interiorcad.interface.machiningsexporter.machiningsneedingwastetobeclearedexporter import MachiningsNeedingWasteToBeClearedExporter
from interiorcad.interface.surface.machiningdirection import MachiningDirection


class HorizontalMachiningsExporter(MachiningsNeedingWasteToBeClearedExporter):
    def createStop(self, **keywords):
        self.export.generateNCStopBeforeNextHorizontalDrilling = False
        return super().createStop(**keywords)

    def createMachinings(self, **keywords):
        return self.createHorizontalMachinings(**keywords)

    def createHorizontalMachinings(self, **keywords):
        surface = keywords['surface']
        return ''.join((self.createHorizontalMachining(machining=machining, drilling=machining, boardThickness=surface.thickness, **keywords) for machining in self.machinings))

    def getHorizontalDrillingZValue(self, **keywords):
        return self.export.getHorizontalDrillingZValue(**keywords)

    def createHorizontalMachining(self, **keywords):
        drilling = keywords['drilling']
        zPosition = self.getHorizontalDrillingZValue(zPosition=drilling.zPosition, **keywords)
        return self.machiningCreator.createRotated(zPosition=zPosition, **keywords)

    def getAllMachiningPositionsThatCouldBeInsideBounds(self):
        return (machining.position for machining in self.machinings)

    def areConditionsRelevantForAStop(self, **keywords):
        return self.export.generateNCStopBeforeNextHorizontalDrilling and super().areConditionsRelevantForAStop(**keywords)


class HorizontalMachiningsExporter_SeparateCallsAndZValueConvertersForMachiningsOnBounds(HorizontalMachiningsExporter):
    def createHorizontalMachinings(self, **keywords):
        surface = keywords['surface']
        bounds = surface.getFinishedSizeBounds()
        return super().createHorizontalMachinings(bounds=bounds, **keywords)

    def getLeftAndRightSidesHorizontalDrillingZValue(self, **keywords):
        return self.export.getLeftAndRightSidesHorizontalDrillingZValue(**keywords)

    def getFrontAndBackSidesHorizontalDrillingZValue(self, **keywords):
        return self.export.getFrontAndBackSidesHorizontalDrillingZValue(**keywords)

    def createHorizontalMachining(self, **keywords):
        drilling = keywords['drilling']
        direction = drilling.direction
        if keywords['bounds'].isPointStrictlyInside(drilling.position):
            zPosition = self.getHorizontalDrillingZValue(zPosition=drilling.zPosition, **keywords)
            return self.machiningCreator.createRotated(zPosition=zPosition, **keywords)
        elif MachiningDirection().isFromLeft(direction, self.export.SURFACE_ORIGIN):
            zPosition = self.getLeftAndRightSidesHorizontalDrillingZValue(zPosition=drilling.zPosition, **keywords)
            return self.machiningCreator.createFromLeft(zPosition=zPosition, **keywords)
        elif MachiningDirection().isFromRight(direction, self.export.SURFACE_ORIGIN):
            zPosition = self.getLeftAndRightSidesHorizontalDrillingZValue(zPosition=drilling.zPosition, **keywords)
            return self.machiningCreator.createFromRight(zPosition=zPosition, **keywords)
        elif MachiningDirection().isFromFront(direction, self.export.SURFACE_ORIGIN):
            zPosition = self.getFrontAndBackSidesHorizontalDrillingZValue(zPosition=drilling.zPosition, **keywords)
            return self.machiningCreator.createFromFront(zPosition=zPosition, **keywords)
        elif MachiningDirection().isFromBack(direction, self.export.SURFACE_ORIGIN):
            zPosition = self.getFrontAndBackSidesHorizontalDrillingZValue(zPosition=drilling.zPosition, **keywords)
            return self.machiningCreator.createFromBack(zPosition=zPosition, **keywords)
        else:
            zPosition = self.getHorizontalDrillingZValue(zPosition=drilling.zPosition, **keywords)
            return self.machiningCreator.createRotated(zPosition=zPosition, **keywords)


class HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds(HorizontalMachiningsExporter_SeparateCallsAndZValueConvertersForMachiningsOnBounds):
    def getLeftAndRightSidesHorizontalDrillingZValue(self, **keywords):
        return self.getHorizontalDrillingZValue(**keywords)

    def getFrontAndBackSidesHorizontalDrillingZValue(self, **keywords):
        return self.getHorizontalDrillingZValue(**keywords)


class HorizontalMachiningsExporterTests(unittest.TestCase):
    class MockDrilling:
        def __init__(self):
            self.position = 'position'

    def setUp(self):
        self.exporter = HorizontalMachiningsExporter(None)

    def testGetAllMachiningPositionsWithoutDrillingsReturnsEmptyList(self):
        self.assertEqual((), tuple(self.exporter.getAllMachiningPositionsThatCouldBeInsideBounds()))

    def testGetAllMachiningPositionsWithOneDrillingReturnsPosition(self):
        self.exporter.machinings.append(self.MockDrilling())
        self.assertEqual(('position',), tuple(self.exporter.getAllMachiningPositionsThatCouldBeInsideBounds()))
