# -*- coding: utf-8 -*-

# Copyright 2011-2014, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vertex import Vertex
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.surface import Surface
from interiorcad.interface.surface.sizing import Sizing2D


class MachiningsNeedingManualInterventionExporter:
    def __init__(self):
        self.stopAlreadyCreated = []
        self.toolIDGetter = None
        self.machiningCreator = None
        self.machinings = []

    def createStopIfNeeded(self, **keywords):
        if self.needsStop(**keywords):
            return self.createStop(**keywords)
        else:
            return ''

    def create(self, **keywords):
        return self.createStopIfNeeded(**keywords) + self.createMachinings(**keywords)

    def needsStop(self, **keywords):
        if self.isStopAlreadyCreated(**keywords):
            return False
        else:
            needStop = self.areConditionsRelevantForAStop(**keywords)
            if needStop:
                self.setStopAlreadyCreated(**keywords)
            return needStop

    def isStopAlreadyCreated(self, **keywords):
        surface = keywords['surface']
        return surface.uuid in self.stopAlreadyCreated

    def setStopAlreadyCreated(self, **keywords):
        surface = keywords['surface']
        self.stopAlreadyCreated.append(surface.uuid)


class Tests(unittest.TestCase):
    MOCK_STOP = 'STOP'

    class MockMachiningsExporter(MachiningsNeedingManualInterventionExporter):
        def areConditionsRelevantForAStop(self, **keywords):
            return self.myAreConditionsRelevantForAStop

        def createMachinings(self, dummySurface):
            return ''

        def createStop(self, **keywords):
            return Tests.MOCK_STOP

    def setUp(self):
        polyline = Polyline(
            Vertex(Point(0.0, 0.0)),
            Vertex(Point(0.0, 100.0)),
            Vertex(Point(100.0, 100.0)),
            Vertex(Point(100.0, 0.0))
        )
        self.surface = Surface()
        self.surface.sizing = Sizing2D(polyline)
        self.surface.sawnSize = Point(5, 5)
        self.exporter = self.MockMachiningsExporter()

    def testSurfaceWithoutRelevantPositionShouldNotNeedAStop(self):
        self.exporter.myAreConditionsRelevantForAStop = False
        self.assertFalse(self.exporter.needsStop(surface=self.surface))

    def testSurfaceWithRelevantPositionShouldNeedAStop(self):
        self.exporter.myAreConditionsRelevantForAStop = True
        self.assertTrue(self.exporter.needsStop(surface=self.surface))

    def testSurfaceWithoutRelevantPositionShouldNotStop(self):
        self.exporter.myAreConditionsRelevantForAStop = False
        self.assertEqual('', self.exporter.createStopIfNeeded(surface=self.surface))

    def testSurfaceWithRelevantPositionShouldStop(self):
        self.exporter.myAreConditionsRelevantForAStop = True
        self.assertEqual(Tests.MOCK_STOP, self.exporter.createStopIfNeeded(surface=self.surface))

    def testSurfaceWithRelevantPositionShouldRegisterTheSurfaceUUID(self):
        self.exporter.myAreConditionsRelevantForAStop = True
        self.exporter.createStopIfNeeded(surface=self.surface)
        self.assertIn(self.surface.uuid, self.exporter.stopAlreadyCreated)
