# -*- coding: utf-8 -*-

# Copyright 2011-2014, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vertex import Vertex
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.surface import Surface
from interiorcad.interface.surface.sizing import Sizing2D

from interiorcad.interface.machiningsexporter.machiningsneedingmanualinterventionexporter import MachiningsNeedingManualInterventionExporter


class MachiningsNeedingWasteToBeClearedExporter(MachiningsNeedingManualInterventionExporter):
    def __init__(self, export):
        super().__init__()
        self.export = export

    def createStop(self, **keywords):
        return self.export.stopForRemovingWaste(**keywords)

    def __isAnyMachiningPositionInsideBounds(self, surface, machiningPositions):
        bounds = surface.getFinishedSizeBounds()
        return any(bounds.isPointStrictlyInside(position) for position in machiningPositions)

    def areConditionsRelevantForAStop(self, **keywords):
        surface = keywords['surface']
        machiningPositions = self.getAllMachiningPositionsThatCouldBeInsideBounds()
        return all((surface.hasSizing(), self.__isAnyMachiningPositionInsideBounds(surface, machiningPositions)))


class Tests(unittest.TestCase):
    class MockMachiningsExporter(MachiningsNeedingWasteToBeClearedExporter):
        def __init__(self):
            super().__init__(None)
            self.positions = []

        def addPosition(self, position):
            self.positions.append(position)

        def getAllMachiningPositionsThatCouldBeInsideBounds(self):
            return self.positions

    def setUp(self):
        polyline = Polyline(Vertex(Point(0.0, 0.0)),
                            Vertex(Point(0.0, 100.0)),
                            Vertex(Point(100.0, 100.0)),
                            Vertex(Point(100.0, 0.0)))
        self.surface = Surface()
        self.surface.sizing = Sizing2D(polyline)
        self.surface.sawnSize = Point(5, 5)
        self.exporter = self.MockMachiningsExporter()

    def testSurfaceWithoutHorizontalDrillingsDoesNotNeedStop(self):
        self.assertFalse(self.exporter.needsStop(surface=self.surface))

    def testSurfaceWithHoriztontalDrillingAtBottomLeftDoesNotNeedStop(self):
        self.exporter.addPosition(Point(0, 0))
        self.assertFalse(self.exporter.needsStop(surface=self.surface))

    def testSurfaceWithHorizontalDrillingInsideBoundsNeedsStop(self):
        self.exporter.addPosition(Point(1, 1))
        self.assertTrue(self.exporter.needsStop(surface=self.surface))

    def testSurfaceWithHorizontalDrillingsAtTopRightDoesNotNeedStop(self):
        self.exporter.addPosition(Point(5, 5))
        self.assertFalse(self.exporter.needsStop(surface=self.surface))

    def testSurfaceWithHorizontalDrillingsInsideAndOutsideNeedsStop(self):
        self.exporter.addPosition(Point(5, 5))
        self.exporter.addPosition(Point(2, 2))
        self.assertTrue(self.exporter.needsStop(surface=self.surface))

    def testSurfaceWithHorizontalDrillingOutsideOfBoundsDoesNotNeedStop(self):
        self.exporter.addPosition(Point(-1, -1))
        self.assertFalse(self.exporter.needsStop(surface=self.surface))

    def testExporterShouldNeedStopOnlyOnce(self):
        self.exporter.addPosition(Point(1, 1))
        self.assertTrue(self.exporter.needsStop(surface=self.surface))
        self.assertFalse(self.exporter.needsStop(surface=self.surface))
