# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import interiorcad.backend.exportdatasafe
from interiorcad.backend.datatracker.exportdatacollector import ExportDataCollector
from interiorcad.backend.joint import JointingInfo
from interiorcad.backend.localization import getLocalizedString, LICENSE_ERRORS_RESOURCEGROUP
from interiorcad.backend.loggingofexcludedcgcs import LoggingOfExcludedConstructionGroupComponents
from interiorcad.backend.surfacecreator import SurfaceCreator, SurfaceCreatorOptions
from interiorcad.backend.workpiece.creator import WorkpieceCreator
from interiorcad.backend.workpiececreationstrategy.creator import WorkpieceCreationStrategy
from interiorcad.interface.ncexport.programnames import ProgramNamesHandler


class BaseExporter:
    def __init__(self, **keywords):
        self.postProcessor = keywords['postProcessor']
        self.workpiecePostProcessor = keywords['workpiecePostProcessor']
        self.ncexport = keywords['ncexport']
        self.workpieceFilters = keywords['workpieceFilters']
        self.jointingInfo = JointingInfo.createFromConfiguration(self.getConfiguration())
        self.surfaceOrigin = keywords['surfaceOrigin']
        self.firstSide = None

    def export(self, **_keywords):
        if self.ncexport.LICENSE_INFO and self.ncexport.LICENSE_INFO.twowop:
            creator = WorkpieceCreator(ExportDataCollector, self.jointingInfo)
            allCustomParts = ExportDataCollector.getAllCustomPartUUIDsAndCounts()
            self.createAndExportWorkpieces(creator, allCustomParts)
        else:
            raise RuntimeError(getLocalizedString(LICENSE_ERRORS_RESOURCEGROUP, 1, "License Error: 2WOP not activated!"))

    def getConfiguration(self):
        return self.ncexport.configuration

    def createAndExportWorkpieces(self, creator, uuidsAndCounts):
        workpieces = creator.createCountedWorkpieces(uuidsAndCounts)
        self.exportWorkpieces(workpieces)

    def getExcludedConstructionGroups(self):  # Currently only construction group components are supported.
        return self.ncexport.configuration.excludedConstructionGroups

    def exportWorkpieces(self, workpieces):
        LoggingOfExcludedConstructionGroupComponents().createLoggingOfExcludedConstructionGroupComponents(excludedConstructionGroups=self.getExcludedConstructionGroups())
        filteredWorkpieces = self.workpieceFilters.filterWorkpieces(workpieces)
        for workpiece in filteredWorkpieces:
            self.exportWorkpiece(workpiece)

    def exportWorkpiece(self, workpiece):
        self.workpiecePostProcessor.process(workpiece)
        programNamesHandler = ProgramNamesHandler(workpiece.uuid)
        surfaceCreator = self.getSurfaceCreator(workpiece)

        self.exportWorkpieceSurfaces(workpiece, surfaceCreator, programNamesHandler)

        if hasattr(self.ncexport, 'EXPORT_FORMAT'):
            programNamesHandler.setNCExportName(self.ncexport.EXPORT_FORMAT)
        else:
            programNamesHandler.setNCExportName(self.ncexport.__class__.__name__)

        programNamesHandler.write()

    def getSurfaceCreator(self, workpiece):
        sizeStrategy = WorkpieceCreationStrategy.createSizeStrategy(workpiece, self.getConfiguration())
        sizingStrategy = WorkpieceCreationStrategy.createSizingStrategy(workpiece, self.getConfiguration())
        options = SurfaceCreatorOptions(self.surfaceOrigin, self.ncexport.GRAIN_CONSIDERATION_MODE, self.ncexport.ALWAYS_CREATE_MAIN_SIDE, self.ncexport.ALLOW_HORIZONTAL_MACHININGS_ON_MAIN_SIDE)
        creator = SurfaceCreator(workpiece, self.getConfiguration(), sizeStrategy, sizingStrategy, options)
        creator.setPostProcessor(self.postProcessor)
        return creator
