# -*- coding: utf-8 -*-

# Copyright 2011-2016, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest.mock
import logging

from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.interface.machineconfiguration.tool import Tool


class DadoCreatorBase:
    def __init__(self, ncExportBase, workpieceName):
        self.configuration = ncExportBase.configuration
        self.ncExportBase = ncExportBase
        self.workpieceName = workpieceName

    def createDado(self, **keywords):
        dado = keywords['dado']

        if dado.tool and dado.tool.type == Tool.ROUTER:
            depth = self.ncExportBase.getDadoRabbetZValue(depth=dado.getDepth(), **keywords)
            return self.createVerticalDadoWithRouter(depth=depth, **keywords)
        elif dado.tool and dado.tool.type == Tool.SAW:
            depth = self.ncExportBase.getDadoZValue(depth=dado.getDepth(), **keywords)
            return self.createVerticalDadoWithSaw(depth=depth, **keywords)
        else:
            ### IT-164: das sollte vielleicht besser ein Filter sein... -- NG
            self.warnThatToolIsNotAvailable(dado)
            return ''

    def getMachineConfiguration(self):
        return self.ncExportBase.getMachineConfiguration()

    def createVerticalDadoWithRouter(self, **keywords):
        return self.ncExportBase.createVerticalDadoWithRouter(**keywords)

    def logWarning(self, dado, message):
        message = message.format(length=dado.getLength(), width=dado.getWidth(), depth=dado.getDepth(), workpiece=self.workpieceName)
        logging.warning(message)


class XAxisDadoCreator(DadoCreatorBase):
    def createVerticalDadoWithSaw(self, **keywords):
        return self.ncExportBase.createVerticalDadoWithSawAlongXAxis(**keywords)

    def warnThatToolIsNotAvailable(self, dado):
        self.logWarning(dado, getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 16, "No tool available for vertical dado along X axis. (Dado length: {length}, width: {width}, depth: {depth}, Workpiece: {workpiece})"))


class YAxisDadoCreator(DadoCreatorBase):
    def createVerticalDadoWithSaw(self, **keywords):
        return self.ncExportBase.createVerticalDadoWithSawAlongYAxis(**keywords)

    def warnThatToolIsNotAvailable(self, dado):
        self.logWarning(dado, getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 17, "No tool available for vertical dado along Y axis. (Dado length: {length}, width: {width}, depth: {depth}, Workpiece: {workpiece})"))


class CRoratedDadoCreator(DadoCreatorBase):
    def createVerticalDadoWithSaw(self, **keywords):
        return self.ncExportBase.createVerticalDadoWithSawCRotated(**keywords)

    def warnThatToolIsNotAvailable(self, dado):
        self.logWarning(dado, getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 15, "No tool available for vertical C rotated dado. (Dado length: {length}, width: {width}, depth: {depth}, Workpiece: {workpiece})"))


class ToolDatabaseTests(unittest.TestCase):
    class TestDecider(DadoCreatorBase):
        def __init__(self):
            self.ncExportBase = unittest.mock.Mock()
            self.warnThatSawIsNotAvailableWasCalled = False
            self.createVerticalDadoWithSawAlongAxisWasCalled = False
            self.createVerticalDadoWithRouterWasCalled = False

        def createVerticalDadoWithSaw(self, **keywords):
            self.createVerticalDadoWithSawAlongAxisWasCalled = True

        def createVerticalDadoWithRouter(self, **keywords):
            self.createVerticalDadoWithRouterWasCalled = True

        def warnThatToolIsNotAvailable(self, dummyDado):
            self.warnThatSawIsNotAvailableWasCalled = True

    def setUp(self):
        self.saw = Tool('Saw')
        self.saw.type = Tool.SAW
        self.saw.properties['saw-cut-width'] = 3

        self.router = Tool('Router')
        self.router.type = Tool.ROUTER

        self.decider = ToolDatabaseTests.TestDecider()

        self.dado = unittest.mock.Mock()
        self.dado.width = 10
        self.dado.tool = self.saw

    def testIfNoSawIsAvailableTheFunctionCreateVerticalDadoWithSawAlongAxisShouldCallTheCorrespondingWarningFunction(self):
        self.dado.tool = None
        result = self.decider.createDado(dado=self.dado)
        self.assertTrue(self.decider.warnThatSawIsNotAvailableWasCalled)
        self.assertEqual('', result)

    def testDadoMadeWithASawShouldCallCreateVerticalDadoWithSawAlongAxisWasCalled(self):
        self.decider.sawWidth = self.dado.width
        self.decider.createDado(dado=self.dado)
        self.assertTrue(self.decider.createVerticalDadoWithSawAlongAxisWasCalled)
        self.assertFalse(self.decider.createVerticalDadoWithRouterWasCalled)

    def testDadoMadeWithASawShouldCallCreateVerticalDadoWithRouterAlongAxisWasCalled(self):
        self.dado.tool = self.router
        self.decider.createDado(dado=self.dado)
        self.assertFalse(self.decider.createVerticalDadoWithSawAlongAxisWasCalled)
        self.assertTrue(self.decider.createVerticalDadoWithRouterWasCalled)
