# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import logging
import math
import unittest

from interiorcad.interface.geometry import isNearlyEqual


class MultiSawDado:
    def __init__(self, ncExport):
        self.ncExport = ncExport

    def segmentMoveRight(self, segment, offset):
        if self.ncExport.isStandardOrigin():
            segment.moveRight(offset=offset)
        else:
            segment.moveLeft(offset=offset)

    def segmentMoveLeft(self, segment, offset):
        if self.ncExport.isStandardOrigin():
            segment.moveLeft(offset=offset)
        else:
            segment.moveRight(offset=offset)

    def getEquallyDistributedClearOutSawPositions(self, **keywords):
        start = keywords['sawWidth'] / 2
        value = keywords['gapToClearOutWidth'] / keywords['insideClearOutSawCountRoundedUp']
        for i in range(int(keywords['insideClearOutSawCountRoundedUp'])):
            yield start + (i + 0.5) * value

    def getCenteredClearOutSawPosition(self, **keywords):
        yield keywords['dadoWidth'] / 2

    def getInsideClearOutSawPositions(self, **keywords):
        sawWidth = keywords['sawWidth']
        overlapWidth = keywords['overlapFactor'] * sawWidth
        gapToClearOutWidth = keywords['dadoWidth'] - sawWidth
        insideClearOutSawCountRoundedUp = math.ceil(gapToClearOutWidth / (sawWidth - overlapWidth))

        if insideClearOutSawCountRoundedUp == 2:
            yield from self.getCenteredClearOutSawPosition(**keywords)
        elif insideClearOutSawCountRoundedUp > 2:
            yield from self.getEquallyDistributedClearOutSawPositions(gapToClearOutWidth=gapToClearOutWidth, insideClearOutSawCountRoundedUp=insideClearOutSawCountRoundedUp, **keywords)

    def __createMultipleVerticalMiddleSawsGenerator(self, **keywords):
        createFunction = keywords['createFunction']
        sawWidth = keywords['dadoSawWidth']
        width = keywords['dado'].getWidth()

        baseSegmentForInsideClearOutSawPositions = copy.copy(keywords['unmodifiedLineSegment'])
        self.segmentMoveLeft(segment=baseSegmentForInsideClearOutSawPositions, offset=width / 2)

        for delta in self.getInsideClearOutSawPositions(sawWidth=sawWidth, overlapFactor=self.ncExport.DADO_SAW_OVERLAP_FACTOR, dadoWidth=width):
            segment = copy.copy(baseSegmentForInsideClearOutSawPositions)
            self.segmentMoveRight(segment=segment, offset=delta)
            yield createFunction(segment=segment, radiusCompensation=self.ncExport.RADIUS_COMPENSATION_NONE, **keywords)

    def __createMultipleVerticalSawGenerator(self, **keywords):
        width = keywords['dado'].getWidth()
        unmodifiedLineSegment = self.ncExport.getDadoLineSegmentWithSecurityDistance(**keywords)
        segment = copy.copy(unmodifiedLineSegment)
        self.segmentMoveLeft(segment=segment, offset=width / 2)

        createFunction = keywords['createFunction']
        yield createFunction(segment=segment, radiusCompensation=self.ncExport.RADIUS_COMPENSATION_RIGHT, **keywords)

        yield from self.__createMultipleVerticalMiddleSawsGenerator(unmodifiedLineSegment=unmodifiedLineSegment, **keywords)

        segment = copy.copy(unmodifiedLineSegment)
        self.segmentMoveRight(segment=segment, offset=width / 2)
        yield createFunction(segment=segment, radiusCompensation=self.ncExport.RADIUS_COMPENSATION_LEFT, **keywords)

    def __createMultipleVerticalCenteredSawGenerator(self, **keywords):
        width = keywords['dado'].getWidth()
        sawWidth = keywords['dadoSawWidth']
        unmodifiedLineSegment = self.ncExport.getDadoLineSegmentWithSecurityDistance(**keywords)

        segment = copy.copy(unmodifiedLineSegment)
        self.segmentMoveLeft(segment=segment, offset=(width - sawWidth) / 2)

        createFunction = keywords['createFunction']
        yield createFunction(segment=segment, radiusCompensation=self.ncExport.RADIUS_COMPENSATION_NONE, **keywords)

        yield from self.__createMultipleVerticalMiddleSawsGenerator(unmodifiedLineSegment=unmodifiedLineSegment, **keywords)

        # IC-18589: create the right segment only if it is different than the left one (a saw has no radius compensation, unlike a router)
        if not isNearlyEqual(width, sawWidth):
            segment = copy.copy(unmodifiedLineSegment)
            self.segmentMoveRight(segment=segment, offset=(width - sawWidth) / 2)
            yield createFunction(segment=segment, radiusCompensation=self.ncExport.RADIUS_COMPENSATION_NONE, **keywords)

    def createMultipleVerticalSawCuts(self, **keywords):
        dadoSawWidth = keywords['dado'].tool.getSawCutWidth()

        if not isNearlyEqual(dadoSawWidth, 0):
            return '\n'.join(self.__createMultipleVerticalSawGenerator(dadoSawWidth=dadoSawWidth, **keywords))
        else:
            warning = self.ncExport.getLocalizedMessage(50, "Suppressing dado, as width of saw is not defined! Workpiece: {workpiece}")
            logging.warning(warning.format(workpiece=keywords['surface'].workpieceName))

    def createMultipleVerticalCenteredSawCuts(self, **keywords):
        dadoSawWidth = keywords['dado'].tool.getSawCutWidth()

        if not isNearlyEqual(dadoSawWidth, 0):
            return '\n'.join(self.__createMultipleVerticalCenteredSawGenerator(dadoSawWidth=dadoSawWidth, **keywords))
        else:
            warning = self.ncExport.getLocalizedMessage(50, "Suppressing dado, as width of saw is not defined! Workpiece: {workpiece}")
            logging.warning(warning.format(workpiece=keywords['surface'].workpieceName))

    def createSingleVerticalSawCut(self, **keywords):
        dado = keywords['dado']
        segment = self.ncExport.getDadoLineSegmentWithSecurityDistance(**keywords)

        return keywords['createFunction'](width=dado.getWidth(), segment=segment, radiusCompensation=self.ncExport.RADIUS_COMPENSATION_NONE, **keywords)


class SawTests(unittest.TestCase):
    def setUp(self):
        self.multiSawDado = MultiSawDado(self)

    def getInsideClearOutSawPositions(self, **keywords):
        return tuple(self.multiSawDado.getInsideClearOutSawPositions(**keywords))

    def testNoInsideClearOutSawNecessary(self):
        positions = self.getInsideClearOutSawPositions(dadoWidth=7, sawWidth=4, overlapFactor=0.15)
        self.assertEqual((), positions)

    def testOneInsideClearOutSawNecessary(self):
        positions = self.getInsideClearOutSawPositions(dadoWidth=9, sawWidth=4, overlapFactor=0.15)
        self.assertEqual((4.5,), positions)

    def testOddNumberOfInsideClearOutSawNecessary(self):
        positions = self.getInsideClearOutSawPositions(dadoWidth=12, sawWidth=4, overlapFactor=0.15)
        self.assertEqual((3.333333333333333, 6.0, 8.666666666666666), positions)

    def testEvenNumberOfInsideClearOutSawNecessary(self):
        positions = self.getInsideClearOutSawPositions(dadoWidth=16, sawWidth=4, overlapFactor=0.15)
        self.assertEqual((3.5, 6.5, 9.5, 12.5), positions)
