# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import uuid

from interiorcad.interface.geometry.edges import Edges
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.geometry.size import Size
from interiorcad.interface.surface.centerpunch2D import CenterPunch2D, VerticalCenterPunch, HorizontalCenterPunch, FiveAxisCenterPunch
from interiorcad.interface.surface.contour2d import Contour2D, VerticalContour2D
from interiorcad.interface.surface.dado2d import Dado2D, VerticalDado2D
from interiorcad.interface.surface.drilling2d import Drilling2D, VerticalDrilling, HorizontalDrilling, FiveAxisDrilling
from interiorcad.interface.surface.griddrilling2D import GridDrilling2D
from interiorcad.interface.surface.miter2d import Miter2D
from interiorcad.interface.surface.ncmacro2d import NCMacro2D
from interiorcad.interface.surface.pocket2d import Pocket2D
from interiorcad.interface.surface.rabbet2d import Rabbet2D
from interiorcad.interface.surface.sizing import Sizing2D

SURFACE_SIDE_TYPE_UNKNOWN, SURFACE_SIDE_TYPE_SIZING, SURFACE_SIDE_TYPE_MAIN, SURFACE_SIDE_TYPE_OPPOSITE, SURFACE_SIDE_TYPE_BOTH_SIDED = range(5)
GRAIN_LENGTHWISE, GRAIN_CROSSWISE, GRAIN_NONE = 0, 1, 2


class Surface:
    def __init__(self):
        self.uuid = uuid.uuid4()
        self.sawnSize = Size(0, 0)
        self.thickness = 0
        self.machinings = []
        self.sizing = Sizing2D(None)
        self.sizingBounds = Rectangle()
        self.finishedSizeAddition = Rectangle()
        self.sizingAddition = 0
        self.finishedSizeSizing = None
        self.finalFinishedSizeShape = None
        self.workpieceName = ''
        self.workpieceUUID = None
        self.edges = Edges()
        self.count = 1
        self.sideType = SURFACE_SIDE_TYPE_UNKNOWN
        self.isFirstSide = None

    def __getClassMachinings(self, wantedMachiningClasses):
        return list(filter(lambda machining: isinstance(machining, wantedMachiningClasses), self.machinings))

    def __setClassMachinings(self, wantedMachiningClasses, machinings):
        nonClassDrillings = itertools.filterfalse(lambda machining: isinstance(machining, wantedMachiningClasses), self.machinings)
        self.machinings = list(itertools.chain(machinings, nonClassDrillings))

    @property
    def verticalDrillingsAndCenterPunches(self):
        return self.__getClassMachinings((VerticalDrilling, VerticalCenterPunch))

    @property
    def horizontalDrillingsAndCenterPunches(self):
        return self.__getClassMachinings((HorizontalDrilling, HorizontalCenterPunch))

    @property
    def fiveAxisDrillingsAndCenterPunches(self):
        return self.__getClassMachinings((FiveAxisDrilling, FiveAxisCenterPunch))

    @property
    def verticalGridDrillings(self):
        return self.__getClassMachinings((GridDrilling2D,))

    def getVerticalDrillings(self):
        return self.__getClassMachinings((VerticalDrilling,))

    def getVerticalCenterPunches(self):
        return self.__getClassMachinings((VerticalCenterPunch,))

    def getHorizontalDrillings(self):
        return self.__getClassMachinings((HorizontalDrilling,))

    def getHorizontalCenterPunches(self):
        return self.__getClassMachinings((HorizontalCenterPunch,))

    def getFiveAxisDrillings(self):
        return self.__getClassMachinings((FiveAxisDrilling,))

    def getFiveAxisCenterPunches(self):
        return self.__getClassMachinings((FiveAxisCenterPunch,))

    def setPockets(self, pockets):
        self.pockets = pockets

    def addVerticalDrillings(self, drillings):
        self.machinings.extend(drillings)

    def setVerticalDrillings(self, drillings):
        self.__setClassMachinings((VerticalDrilling,), drillings)

    def setVerticalCenterPunches(self, centerPunches):
        self.__setClassMachinings((VerticalCenterPunch,), centerPunches)

    def addVerticalCenterPunches(self, centerPunches):
        self.machinings.extend(centerPunches)

    def setVerticalDrillingsAndCenterPunches(self, drillings):
        self.__setClassMachinings((VerticalDrilling, VerticalCenterPunch), drillings)

    def setVerticalGridDrillings(self, drillings):
        self.__setClassMachinings((GridDrilling2D,), drillings)

    def setHorizontalDrillings(self, drillings):
        self.__setClassMachinings((HorizontalDrilling,), drillings)

    def setHorizontalCenterPunches(self, centerPunches):
        self.__setClassMachinings((HorizontalCenterPunch,), centerPunches)

    def addHorizontalDrillings(self, drillings):
        self.machinings.extend(drillings)

    def setHorizontalDrillingsAndCenterPunches(self, drillings):
        self.__setClassMachinings((HorizontalDrilling, HorizontalCenterPunch), drillings)

    def setFiveAxisDrillings(self, drillings):
        self.__setClassMachinings((FiveAxisDrilling,), drillings)

    def setFiveAxisCenterPunches(self, centerPunches):
        self.__setClassMachinings((FiveAxisCenterPunch,), centerPunches)

    def addFiveAxisDrillings(self, drillings):
        self.machinings.extend(drillings)

    def setFiveAxisDrillingsAndCenterPunches(self, drillings):
        self.__setClassMachinings((FiveAxisDrilling, FiveAxisCenterPunch), drillings)

    @property
    def dadosVertical(self):
        return self.__getClassMachinings((VerticalDado2D,))

    @dadosVertical.setter
    def dadosVertical(self, machinings):
        self.__setClassMachinings((VerticalDado2D,), machinings)

    def getAllDados(self):
        return self.__getClassMachinings((Dado2D,))

    @property
    def rabbetsVertical(self):
        return self.__getClassMachinings((Rabbet2D,))

    @rabbetsVertical.setter
    def rabbetsVertical(self, machinings):
        self.__setClassMachinings((Rabbet2D,), machinings)

    def hasRabbet(self):
        return bool(self.rabbetsVertical)

    @property
    def mitersVertical(self):
        return self.__getClassMachinings((Miter2D,))

    @mitersVertical.setter
    def mitersVertical(self, machinings):
        self.__setClassMachinings((Miter2D,), machinings)

    @property
    def contoursVertical(self):
        return self.__getClassMachinings((VerticalContour2D,))

    @contoursVertical.setter
    def contoursVertical(self, machinings):
        self.__setClassMachinings((VerticalContour2D,), machinings)

    def getAllContours(self):
        return self.__getClassMachinings((Contour2D,))

    def setAllContours(self, machinings):
        self.__setClassMachinings((Contour2D,), machinings)

    @property
    def pockets(self):
        return self.__getClassMachinings((Pocket2D,))

    @pockets.setter
    def pockets(self, machinings):
        self.__setClassMachinings((Pocket2D,), machinings)

    def addPockets(self, machinings):
        self.machinings.extend(machinings)

    @property
    def ncmacros(self):
        return self.__getClassMachinings((NCMacro2D,))

    @ncmacros.setter
    def ncmacros(self, machinings):
        self.__setClassMachinings((NCMacro2D,), machinings)

    def hasSizing(self):
        return self.sizing

    # Compatibility variable for old code (until interiorcad 2018 F7. Back then it was never mirrored); do NOT use for new code! Use directly finalFinishedSizeShape
    @property
    def workpieceFinishedSizeShape(self):
        return self.finalFinishedSizeShape

    def bottomLeft(self):
        return Point(0, 0)

    def topRight(self):
        return Point(self.getFinishedSizeX(), self.getFinishedSizeY())

    def topLeft(self):
        return Point(0, self.getFinishedSizeY())

    def mirrorOnXAxis(self):
        for machining in self.getAllMachiningsButSizing():
            machining.mirrorOnXAxis(self)

        self.mirrorSizingOnXAxis()
        self.mirrorSizingBoundsOnXAxis()
        self.mirrorFinishedSizeAdditionOnXAxis()

        self.mirrorEdgesOnXAxis()

    def mirrorOnYAxis(self):
        for machining in self.getAllMachiningsButSizing():
            machining.mirrorOnYAxis(self)

        self.mirrorSizingOnYAxis()
        self.mirrorSizingBoundsOnYAxis()
        self.mirrorFinishedSizeAdditionOnYAxis()

        self.mirrorEdgesOnYAxis()

    def mirrorOnXYPlane(self):
        for machining in self.getAllMachiningsButSizing():
            machining.mirrorOnXYPlane(self)

    def rotateLeft(self):
        for machining in self.getAllMachiningsButSizing():
            machining.rotateLeft(self)

        self.rotateSizingLeft()
        self.rotateSizingBoundsLeft()

        self.rotateFinishedSizeAdditionLeft()
        self.rotateSawnSizeLeft()
        self.rotateGrain()

        self.rotateEdgesLeft()

    def rotateRight(self):
        for machining in self.getAllMachiningsButSizing():
            machining.rotateRight(self)

        self.rotateSizingRight()
        self.rotateSizingBoundsRight()

        self.rotateFinishedSizeAdditionRight()
        self.rotateSawnSizeRight()
        self.rotateGrain()

        self.rotateEdgesRight()

    def rotate180(self):
        for machining in self.getAllMachiningsButSizing():
            machining.rotate180(self)

        self.rotateSizing180()
        self.rotateSizingBounds180()
        self.rotateFinishedSizeAddition180()

        self.rotateSawnSize180()

        self.rotateEdges180()

    def mirrorSizingOnXAxis(self):
        pointToMove = Point(0, self.getFinishedSize().y)
        approachStart = self.bottomLeft()
        self.sizing.mirrorOnXAxis(pointToMove, approachStart)
        self.finishedSizeSizing.mirrorOnXAxis(pointToMove, approachStart)
        self.finalFinishedSizeShape.mirrorOnXAxis(pointToMove, approachStart)

    def mirrorSizingOnYAxis(self):
        pointToMove = Point(self.getFinishedSize().x, 0)
        approachStart = self.bottomLeft()
        self.sizing.mirrorOnYAxis(pointToMove, approachStart)
        self.finishedSizeSizing.mirrorOnYAxis(pointToMove, approachStart)
        self.finalFinishedSizeShape.mirrorOnYAxis(pointToMove, approachStart)

    def rotateSawnSizeLeft(self):
        self.sawnSize.x, self.sawnSize.y = self.sawnSize.y, self.sawnSize.x

    def rotateSawnSizeRight(self):
        self.sawnSize.x, self.sawnSize.y = self.sawnSize.y, self.sawnSize.x

    def rotateSawnSize180(self):
        pass

    def rotateSizingLeft(self):
        pointToMove = Point(self.getFinishedSizeY(), 0)
        approachStart = self.bottomLeft()
        self.sizing.rotateLeft(pointToMove, approachStart)
        self.finishedSizeSizing.rotateLeft(pointToMove, approachStart)
        self.finalFinishedSizeShape.rotateLeft(pointToMove, approachStart)

    def rotateSizingRight(self):
        pointToMove = Point(0, self.getFinishedSizeX())
        approachStart = self.bottomLeft()
        self.sizing.rotateRight(pointToMove, approachStart)
        self.finishedSizeSizing.rotateRight(pointToMove, approachStart)
        self.finalFinishedSizeShape.rotateRight(pointToMove, approachStart)

    def rotateSizing180(self):
        pointToMove = self.getFinishedSize()
        approachStart = self.bottomLeft()
        self.sizing.rotate180(pointToMove, approachStart)
        self.finishedSizeSizing.rotate180(pointToMove, approachStart)
        self.finalFinishedSizeShape.rotate180(pointToMove, approachStart)

    def mirrorSizingBoundsOnXAxis(self):
        self.sizingBounds.mirrorOnZeroXAxis()
        self.sizingBounds.moveBy(Point(0, self.getFinishedSize().y))

    def mirrorSizingBoundsOnYAxis(self):
        self.sizingBounds.mirrorOnZeroYAxis()
        self.sizingBounds.moveBy(Point(self.getFinishedSize().x, 0))

    def rotateSizingBoundsLeft(self):
        self.sizingBounds.rotateLeftAtZero()
        self.sizingBounds.moveBy(Point(self.getFinishedSizeY(), 0))

    def rotateSizingBoundsRight(self):
        self.sizingBounds.rotateRightAtZero()
        self.sizingBounds.moveBy(Point(0, self.getFinishedSizeX()))

    def rotateSizingBounds180(self):
        self.sizingBounds.rotate180AtZero()
        self.sizingBounds.moveBy(self.getFinishedSize())

    def mirrorFinishedSizeAdditionOnXAxis(self):
        self.finishedSizeAddition.mirrorOnZeroXAxisNoNegation()

    def mirrorFinishedSizeAdditionOnYAxis(self):
        self.finishedSizeAddition.mirrorOnZeroYAxisNoNegation()

    def rotateFinishedSizeAdditionLeft(self):
        self.finishedSizeAddition.rotateLeftAtZeroNoNegation()

    def rotateFinishedSizeAdditionRight(self):
        self.finishedSizeAddition.rotateRightAtZeroNoNegation()

    def rotateFinishedSizeAddition180(self):
        self.finishedSizeAddition.rotate180AtZeroNoNegation()

    def rotateGrain(self):
        if self.grain == GRAIN_LENGTHWISE:
            self.grain = GRAIN_CROSSWISE
        elif self.grain == GRAIN_CROSSWISE:
            self.grain = GRAIN_LENGTHWISE
        # GRAIN_NONE always remain the same

    def mirrorEdgesOnXAxis(self):
        polygonalEdges = self.edges.polygonal
        if polygonalEdges and len(polygonalEdges) == 4:
            polygonalEdges[2], polygonalEdges[0] = polygonalEdges[0], polygonalEdges[2]

        # Be careful that the indices in the polygonal case are different from the rectangular case
        rectangularEdges = self.edges.rectangular
        if rectangularEdges:
            rectangularEdges[1], rectangularEdges[0] = rectangularEdges[0], rectangularEdges[1]

    def mirrorEdgesOnYAxis(self):
        polygonalEdges = self.edges.polygonal
        if polygonalEdges and len(polygonalEdges) == 4:
            polygonalEdges[1], polygonalEdges[3] = polygonalEdges[3], polygonalEdges[1]

        # Be careful that the indices in the polygonal case are different from the rectangular case
        rectangularEdges = self.edges.rectangular
        if rectangularEdges:
            rectangularEdges[2], rectangularEdges[3] = rectangularEdges[3], rectangularEdges[2]

    def rotateEdgesLeft(self):
        polygonalEdges = self.edges.polygonal
        if polygonalEdges and len(polygonalEdges) == 4:
            polygonalEdges[0], polygonalEdges[1], polygonalEdges[2], polygonalEdges[3] = polygonalEdges[3], polygonalEdges[0], polygonalEdges[1], polygonalEdges[2]

        # Be careful that the indices in the polygonal case are different from the rectangular case
        rectangularEdges = self.edges.rectangular
        if rectangularEdges:
            self.edges.left, self.edges.right, self.edges.front, self.edges.back = self.edges.back, self.edges.front, self.edges.left, self.edges.right

    def rotateEdgesRight(self):
        polygonalEdges = self.edges.polygonal
        if polygonalEdges and len(polygonalEdges) == 4:
            polygonalEdges[0], polygonalEdges[1], polygonalEdges[2], polygonalEdges[3] = polygonalEdges[1], polygonalEdges[2], polygonalEdges[3], polygonalEdges[0]

        # Be careful that the indices in the polygonal case are different from the rectangular case
        rectangularEdges = self.edges.rectangular
        if rectangularEdges:
            self.edges.left, self.edges.right, self.edges.front, self.edges.back = self.edges.front, self.edges.back, self.edges.right, self.edges.left

    def rotateEdges180(self):
        polygonalEdges = self.edges.polygonal
        if polygonalEdges and len(polygonalEdges) == 4:
            polygonalEdges[0], polygonalEdges[1], polygonalEdges[2], polygonalEdges[3] = polygonalEdges[1], polygonalEdges[0], polygonalEdges[3], polygonalEdges[2]

        # Be careful that the indices in the polygonal case are different from the rectangular case
        rectangularEdges = self.edges.rectangular
        if rectangularEdges:
            self.edges.left, self.edges.right, self.edges.front, self.edges.back = self.edges.right, self.edges.left, self.edges.back, self.edges.front

    def getAllBounds(self):
        return list()

    def getAllMachiningsButSizing(self):
        return self.machinings

    def getAllMachinings(self):
        return itertools.chain(
            self.getAllMachiningsButSizing(),
            (self.sizing,),
        )

    def setSizing(self, newSizing, newSizingBounds):
        self.sizing = newSizing
        self.sizingBounds = newSizingBounds
        if self.sizing is None:
            self.sizing = Sizing2D(None)
        if self.sizingBounds is None:
            self.sizingBounds = Rectangle()

    def getSawnSizeRectangularArea(self):
        return self.sawnSize.x * self.sawnSize.y

    def getFinishedSizeBounds(self):
        return Rectangle(0, 0, self.getFinishedSize().x, self.getFinishedSize().y)

    def getFinishedSizeX(self):
        return self.sawnSize.x + self.finishedSizeAddition.left + self.finishedSizeAddition.right

    def getFinishedSizeY(self):
        return self.sawnSize.y + self.finishedSizeAddition.top + self.finishedSizeAddition.bottom

    def getAdditionX(self):
        return self.sizingAddition - self.finishedSizeAddition.left

    def getAdditionY(self):
        return self.sizingAddition - self.finishedSizeAddition.bottom

    def getAdditionXMirrored(self):
        return self.sizingAddition - self.finishedSizeAddition.right

    def getAdditionYMirrored(self):
        return self.sizingAddition - self.finishedSizeAddition.top

    def getFinishedSize(self):
        return Size(self.getFinishedSizeX(), self.getFinishedSizeY())

    def getEdgesAdditionX(self):
        return 2 * self.sizingAddition - self.finishedSizeAddition.left - self.finishedSizeAddition.right

    def getEdgesAdditionY(self):
        return 2 * self.sizingAddition - self.finishedSizeAddition.top - self.finishedSizeAddition.bottom

    def getFinishedSizeBoundingPoints(self):
        return (
            (0, 0),
            (self.getFinishedSizeX(), 0),
            (self.getFinishedSizeX(), self.getFinishedSizeY()),
            (0, self.getFinishedSizeY()),
        )

    def changeSizingDirection(self):
        if self.hasSizing():
            self.sizing.changeDirection()

    def resetSizing(self):
        self.sizing = Sizing2D(None)
        self.sizingBounds = Rectangle()
