# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.vector import Vector


def normalizeAngle(angle):
    angle %= 360.0  # Guaranteed to be positive, unlike math.fmod

    if isNearlyEqual(0.0, angle) or isNearlyEqual(360.0, angle):
        angle = 0.0

    return angle


def calculateCAngleToXYPlaneFromWVector(wvector):
    # CAREFUL: this calculation does not work if the W-Vector is (0, 0, 1) as the projection on X/Y is (0,0) and the resulting angle is undefined.
    cAngle = VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(-wvector)
    return normalizeAngle(cAngle)


# Old name of function (up to interiorcad 2019 F3). Kept for source compatibility reasons.
def calculateCAngleFromWVector(wvector):
    return calculateCAngleToXYPlaneFromWVector(wvector)


def calculateCAngleInPlaneFromMatrix(matrix):
    cangle = VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(matrix.u)
    if matrix.isFlipped():
        cangle += 180
    return normalizeAngle(cangle)


# Old name of function (up to interiorcad 2019 F3). Kept for source compatibility reasons.
def calculateCAngleFromMatrix(matrix):
    return calculateCAngleInPlaneFromMatrix(matrix)


def calculateAAngleFromWVector(wvector):
    aAngle = VectorMath.angleBetween(wvector, Vector(0, 0, 1))
    return normalizeAngle(aAngle)


class FormerDrillingAngleTests(unittest.TestCase):
    def calculateAnglesAC(self, wvector):
        return calculateCAngleFromWVector(wvector=wvector), calculateAAngleFromWVector(wvector=wvector)

    def testHorizontalDrillingFromRight(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(1, 0, 0))

        self.assertEqual(180, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromRight(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.866025, -0.5, 0))

        self.assertAlmostEqual(150, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testHorizontalDrillingFromLeft(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(-1, 0, 0))

        self.assertEqual(0, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromLeft(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(-0.866025, -0.5, 0))

        self.assertAlmostEqual(30, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testHorizontalDrillingFromBack(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0, 1, 0))

        self.assertEqual(270.0, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromBack(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.5, 0.866025, 0))

        self.assertAlmostEqual(240, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testHorizontalDrillingFromFront(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0, -1, 0))

        self.assertEqual(90.0, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromFront(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.5, -0.866025, 0))

        self.assertAlmostEqual(120, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testVerticalDrillingFromTop(self):
        _cAngle, aAngle = self.calculateAnglesAC(Vector(0, 0, 1))

        self.assertEqual(0.0, aAngle)

    def test5AxisFromTopLeft(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(-0.5, 0, 0.8660254038))

        self.assertEqual(0, cAngle)
        self.assertAlmostEqual(30.0, aAngle, places=8)

    def test5Axis(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.29619813272602386, -0.81379768134937380, 0.50000000000000011))

        self.assertAlmostEqual(110, cAngle, places=8)
        self.assertAlmostEqual(60, aAngle, places=8)


class NormalizeAngleTests(unittest.TestCase):
    def testNormalValue(self):
        self.assertEqual(12.34, normalizeAngle(12.34))

    def testZero(self):
        self.assertEqual(0.0, normalizeAngle(0.0))

    def test360ShouldNormalizeToZero(self):
        self.assertEqual(0.0, normalizeAngle(360.0))

    def testTooBigShouldNormalize(self):
        # Using %, math.fmod or substracting manually 360 brings precision errors.
        self.assertAlmostEqual(12.34, normalizeAngle(360.0 + 12.34))

    def testNegativeShouldNormailzeToPositiveValue(self):
        self.assertEqual(360-12.34, normalizeAngle(-12.34))
