# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest.mock

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.machineconfiguration import machineconfigurationkeys
from interiorcad.interface.surface.drillmachining2d import DrillMachining2D


class CenterPunch2D(DrillMachining2D):
    def __init__(self, position=None, direction=None, diameter=0, zPosition=0, installedDiameter=0):
        super().__init__(position, direction, diameter, zPosition, installedDiameter)

    def __str__(self):
        return ' '.join((
            'CenterPunch 2D:',
            super().__str__() + ";",
        ))

    def setConfiguratedDepth(self, machineConfiguration):
        self.depth = self.actualDepth = machineConfiguration.getAsFloat(machineconfigurationkeys.DEFAULT_CENTER_PUNCH_DEPTH_KEY, 0.5)


class VerticalCenterPunch(CenterPunch2D):
    def assignTool(self, _surface, machineConfiguration):
        self.setConfiguratedDepth(machineConfiguration)
        self.setTool(machineConfiguration.toolDatabase.getVerticalCenterPunchDrill(self.diameter))


class HorizontalCenterPunch(CenterPunch2D):
    def assignTool(self, _surface, machineConfiguration):
        self.setConfiguratedDepth(machineConfiguration)
        if self.isXMinus():
            return self.setTool(machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillXMinus(self.diameter))
        elif self.isYMinus():
            return self.setTool(machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillYMinus(self.diameter))
        elif self.isXPlus():
            return self.setTool(machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillXPlus(self.diameter))
        elif self.isYPlus():
            return self.setTool(machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillYPlus(self.diameter))
        else:
            return self.setTool(machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillC(self.diameter))


class FiveAxisCenterPunch(CenterPunch2D):
    def assignTool(self, _surface, machineConfiguration):
        self.setConfiguratedDepth(machineConfiguration)
        self.setTool(machineConfiguration.toolDatabase.getFiveAxisCenterPunchDrill(self.diameter))


class CenterPunch2DTest(unittest.TestCase):
    def setUp(self):
        self.centerPunch = CenterPunch2D()

    def testInitSupportsInstalledDiameter(self):  # IC-14362
        self.centerPunch = CenterPunch2D(installedDiameter=8.5)
        self.assertEqual(8.5, self.centerPunch.installedDiameter)


class HorizontalCenterPunchToolAssigningTest(unittest.TestCase):
    def setUp(self):
        self.drilling = HorizontalCenterPunch(Point(0, 0), Vector(0.7, 0.7, 0))
        self.machineConfiguration = unittest.mock.Mock()
        self.machineConfiguration.toolDatabase = unittest.mock.Mock()
        self.surface = unittest.mock.Mock()

    def testFreeHorizontalDrillingShouldCallGetHorizontalDrill(self):
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillC.called)

    def testHorizontalDrillingXMinusShouldCallGetHorizontalDrillXMinus(self):
        self.drilling.matrix.w = Vector(1, 0, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillXMinus.called)

    def testHorizontalDrillingYMinusShouldCallGetHorizontalDrillYMinus(self):
        self.drilling.matrix.w = Vector(0, 1, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillYMinus.called)

    def testHorizontalDrillingXPlusShouldCallGetHorizontalDrillXPlus(self):
        self.drilling.matrix.w = Vector(-1, 0, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillXPlus.called)

    def testHorizontalDrillingYPlusShouldCallGetHorizontalDrillYPlus(self):
        self.drilling.matrix.w = Vector(0, -1, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalCenterPunchDrillYPlus.called)
