# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.surface.exportmethodnames import ExportKeys
from interiorcad.interface.surface.linemachining2d import LineMachining2D


class Chamfer2D(LineMachining2D):
    def __init__(self, position=None, width=0, length=0, directionVector=Point(1, 0), angle=0):
        position = position or Point(0, 0)
        super().__init__(position, width, length, directionVector)
        self.angle = angle
        self.segment = None

    def readFrom(self, machining3D):
        super().readFrom(machining3D)
        self.angle = machining3D.getAngle()

    def assignTool(self, surface, machineConfiguration):
        self.actualDepth = copy.copy(self.depth)  # TODO IT-265/InnoWeek 2021-12: what do we want for saw actualDepth -- NG
        self.tool = None   # TODO InnoWeek 2021-12: we have no tools available for chamfers -- NG
        self.assignToolIDFromTool()

    def getExportKey(self):
        return ExportKeys.CHAMFER

    def __str__(self):
        result = ' '.join((
            'Chamfer 2D:',
            super().__str__() + ";",
            'angle({d:.{cp}f})'.format(d=self.angle, cp=2),
        ))
        return result

    def __repr__(self):
        return self.__str__()

    def getAngle(self):
        return self.angle

    def __eq__(self, rhs):
        return super().__eq__(rhs) and self.angle == rhs.angle


class Tests(unittest.TestCase):
    def setUp(self):
        self.chamfer = Chamfer2D(Point(0, 0))

    def testDefaultDrillingHasZeroWidth(self):
        self.assertEqual(0, self.chamfer.getWidth())

    def testDefaultDrillingHasZeroLength(self):
        self.assertEqual(0, self.chamfer.getLength())

    def testCanSetWidth(self):
        self.chamfer.width = 15
        self.assertEqual(15, self.chamfer.getWidth())

    def testCanSetLength(self):
        self.chamfer.length = 5
        self.assertEqual(5, self.chamfer.getLength())

    def testCompleteInit(self):
        self.chamfer = Chamfer2D(Point(1, 2), 3, 5, Point(0, 20))
        self.assertEqual(Point(1, 2), self.chamfer.position)
        self.assertEqual(3.0, self.chamfer.width)
        self.assertEqual(5.0, self.chamfer.length)
        self.assertEqual(Point(0, 1), self.chamfer.directionVector)
        self.assertEqual(0.0, self.chamfer.angle)

    def testStartPointShouldBePosition(self):
        self.assertEqual(self.chamfer.position, self.chamfer.getStartPoint())

    def testEndPointShouldBeAtTheLengthOnTheXAxis(self):
        self.assertEqual(Point(self.chamfer.position.x + self.chamfer.length, self.chamfer.position.y), self.chamfer.getEndPoint())

    def testEndPointShouldBeAtTheLengthOnTheYAxis(self):
        self.assertEqual(Point(self.chamfer.position.x, self.chamfer.position.y + self.chamfer.length), self.chamfer.getEndPoint())

    def testEquality(self):
        self.assertTrue(Chamfer2D(Point(0, 0)) == self.chamfer)
        self.assertEqual(Chamfer2D(Point(0, 0)), self.chamfer)
        self.assertFalse(Chamfer2D(Point(0, 0)) != self.chamfer)

    def testUnequalityOfPosition(self):
        self.assertFalse(Chamfer2D(Point(1, 2)) == self.chamfer)
        self.assertTrue(Chamfer2D(Point(1, 2)) != self.chamfer)

    def testUnequalityOfWidth(self):
        self.assertFalse(Chamfer2D(Point(0, 0), width=20) == self.chamfer)
        self.assertTrue(Chamfer2D(Point(0, 0), width=20) != self.chamfer)

    def testUnequalityOfLength(self):
        self.assertFalse(Chamfer2D(Point(0, 0), length=20) == self.chamfer)
        self.assertTrue(Chamfer2D(Point(0, 0), length=20) != self.chamfer)

    def testUnequalityOfDirectionVector(self):
        self.assertFalse(Chamfer2D(Point(0, 0), directionVector=Point(0, -1)) == self.chamfer)
        self.assertTrue(Chamfer2D(Point(0, 0), directionVector=Point(0, -1)) != self.chamfer)

    def testInvertOfADadoAlongXAxis(self):
        chamfer = Chamfer2D(Point(0, 0), directionVector=Point(1, 0), length=10)
        chamfer.invert()
        self.assertEqual(Chamfer2D(Point(10, 0), directionVector=Point(-1, 0), length=10), chamfer)

    def testInvertOfADadoAlongYAxis(self):
        chamfer = Chamfer2D(Point(0, 0), directionVector=Point(0, 1), length=10)
        chamfer.invert()
        self.assertEqual(Chamfer2D(Point(0, 10), directionVector=Point(0, -1), length=10), chamfer)

    def testEqShouldRejectUnequalAngles(self):
        chamfer1 = Chamfer2D(angle=21)
        chamfer2 = Chamfer2D(angle=42)
        self.assertFalse(chamfer1 == chamfer2)
