# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import operator
import unittest

from interiorcad.backend.cmoduleadapters.contourpath import getOffsetContourPath
from interiorcad.backend.datatracker.polygonconverter import PolygonConverter
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.surface.actualdepths import getVerticalMillingActualDepth
from interiorcad.interface.surface.contourmachining2d import ContourMachining2D
from interiorcad.interface.surface.exportmethodnames import ExportKeys


class Contour2D(ContourMachining2D):
    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    def __init__(self, path=None, width=0, depth=0, radiusCompensation=0):
        super().__init__(path)
        self.depth = depth
        self.width = width
        self.radiusCompensation = radiusCompensation
        self.approach = ''
        self.departure = ''
        self.hasMachineOptions = False
        self.selectedMachine = ''
        self.clearOut = False
        self.multiUsageRoutings = None

    def readFrom(self, contour3D):
        super().readFrom(contour3D)

        self.width = contour3D.width
        self.radiusCompensation = contour3D.radiusCompensation
        self.approach = contour3D.approach
        self.departure = contour3D.departure
        self.hasMachineOptions = contour3D.hasMachineOptions
        self.selectedMachine = contour3D.selectedMachine
        self.clearOut = contour3D.clearOut
        self.multiUsageRoutings = copy.deepcopy(contour3D.multiUsageRoutings)

    def getExportKey(self):
        return ExportKeys.CONTOUR

    # TODO CONTOUR2.0 IT-233: Multi-Contour -- NG
    def __str__(self):
        result = ' '.join((
            'Contour 2D:',
            super().__str__(),
            'width({width:.{cp}f}),'.format(width=self.width, cp=2),
            'radiusCompensation({rc:d}),'.format(rc=self.radiusCompensation),
            'approach({})'.format(self.approach),
            'departure({})'.format(self.departure),
        ))
        return result

    def __repr__(self):
        return self.__str__()

    # TODO CONTOUR2.0 IT-233: Multi-Contour -- NG
    def __eq__(self, rhs):
        return all((
            super().__eq__(rhs),
            self.width == rhs.width,
            self.radiusCompensation == rhs.radiusCompensation,
            self.approach == rhs.approach,
            self.departure == rhs.departure,
            self.depth == rhs.depth  # TODO IT-265 why can super().__eq__ not replace this without breaking tests? -- NG
        ))

    @staticmethod
    def getRadiusCompensation(radiusCompensation, none, left, right):
        if radiusCompensation == Contour2D.RADIUS_COMPENSATION_NONE:
            return none
        elif radiusCompensation == Contour2D.RADIUS_COMPENSATION_LEFT:
            return left
        elif radiusCompensation == Contour2D.RADIUS_COMPENSATION_RIGHT:
            return right

    @staticmethod
    def getInvertedRadiusCompensation(radiusCompensation):
        if radiusCompensation == Contour2D.RADIUS_COMPENSATION_LEFT:
            return Contour2D.RADIUS_COMPENSATION_RIGHT
        elif radiusCompensation == Contour2D.RADIUS_COMPENSATION_RIGHT:
            return Contour2D.RADIUS_COMPENSATION_LEFT
        else:
            return radiusCompensation

    def areApproachDepartureValid(self):
        return self.radiusCompensation != self.RADIUS_COMPENSATION_NONE or 'arc' not in self.approach or 'arc' not in self.departure

    def forceApproachDepartureNotToBeAnArc(self):
        self.approach = self.approach.replace('arc', 'linear')
        self.departure = self.departure.replace('arc', 'linear')

    def hasValidMachineOptions(self):
        return self.hasMachineOptions and bool(self.multiUsageRoutings)

    def assignTool(self, surface, machineConfiguration):
        if self.hasValidMachineOptions():
            # The tool in pass 0 is the main tool, especially for sorting
            self.assignToolFromName(machineConfiguration=machineConfiguration, toolName=self.multiUsageRoutings[0].toolName)
            self.setActualDepth(surface, machineConfiguration, self.depth)
        else:
            self.assignToolAutomatically(surface, machineConfiguration)
            self.assignToolIDFromTool()

    def getToolFromName(self, machineConfiguration, toolName):
        # We must check if the tool works for vertical or horizontal, as the PIO does not care and does not filter the tools -- NG
        if self.isVertical():
            return machineConfiguration.toolDatabase.getToolByNameForVerticalContour(toolName)
        else:
            return machineConfiguration.toolDatabase.getToolByNameForHorizontalContour(toolName)

    def getFirstInvalidToolName(self, machineConfiguration):
        for routing in self.multiUsageRoutings:
            tool = self.getToolFromName(machineConfiguration, routing.toolName)
            if tool is None:
                return routing.toolName

    def assignToolFromName(self, machineConfiguration, toolName):
        self.tool = self.getToolFromName(machineConfiguration, toolName)
        self.assignToolIDFromTool()

    def assignToolAutomatically(self, _surface, _machineConfiguration):
        self.actualDepth = None
        self.tool = None

    def setActualDepth(self, _surface, _machineConfiguration, _depth):
        raise NotImplementedError

    def createCurrentMultiUsageRouting(self, passIndex, machineConfiguration):
        # This function does NOT modify the path, as it is meant for WOPs using the original path and the distance value
        contourPass = copy.deepcopy(self)
        contourPass.originalContour = self

        # Leave only the relevant pass data, not the others -- NG
        contourPass.currentMultiUsageRouting = contourPass.multiUsageRoutings[passIndex]
        del contourPass.multiUsageRoutings

        contourPass.assignToolFromName(machineConfiguration=machineConfiguration, toolName=contourPass.currentMultiUsageRouting.toolName)
        contourPass.width = contourPass.tool.getRouterDiameter() if contourPass.tool else None
        contourPass.depth = contourPass.currentMultiUsageRouting.depth
        contourPass.clearOut = contourPass.currentMultiUsageRouting.clearOut

        return contourPass

    def createCurrentMultiUsageRoutingMovingPath(self, passIndex, machineConfiguration, extraMirroring=False, distanceAdapter=None):
        # This function DOES modify the path, as it is meant for stupid exports that cannot just use the original path and the distance value
        matrix = self.matrix

        if self.radiusCompensation != self.RADIUS_COMPENSATION_NONE:
            isRight = operator.xor(operator.xor(self.radiusCompensation == self.RADIUS_COMPENSATION_RIGHT, matrix.isFlipped()), extraMirroring)
            adaptedRadiusCompensation = self.RADIUS_COMPENSATION_RIGHT if isRight else self.RADIUS_COMPENSATION_LEFT
        else:
            adaptedRadiusCompensation = self.radiusCompensation

        contourPass = self.createCurrentMultiUsageRouting(passIndex, machineConfiguration)
        distanceDelta = distanceAdapter(contour=contourPass, tool=contourPass.tool, radiusCompensation=adaptedRadiusCompensation) if distanceAdapter else 0.0
        offsetPath = getOffsetContourPath(
            vertices=contourPass.originalContour.pathInMatrixPlane.vertices[:],
            radiusCompensation=adaptedRadiusCompensation,
            distance=contourPass.currentMultiUsageRouting.distance + distanceDelta
        )
        contourPass.setPathInMatrixPlane(matrix, list(PolygonConverter().convertVertices(offsetPath)))

        return contourPass

    def createCurrentRoutingMovingPath(self, extraMirroring=False, distanceAdapter=None):
        # This function DOES modify the path, as it is meant for stupid exports that cannot just use the original path and the distance value
        matrix = self.matrix

        if self.radiusCompensation != self.RADIUS_COMPENSATION_NONE:
            isRight = operator.xor(operator.xor(self.radiusCompensation == self.RADIUS_COMPENSATION_RIGHT, matrix.isFlipped()), extraMirroring)
            adaptedRadiusCompensation = self.RADIUS_COMPENSATION_RIGHT if isRight else self.RADIUS_COMPENSATION_LEFT
        else:
            adaptedRadiusCompensation = self.radiusCompensation

        contourPass = copy.deepcopy(self)
        contourPass.originalContour = self
        distanceDelta = distanceAdapter(contour=self, tool=contourPass.tool, radiusCompensation=adaptedRadiusCompensation) if distanceAdapter else -0.0
        offsetPath = getOffsetContourPath(
            vertices=contourPass.originalContour.pathInMatrixPlane.vertices[:],
            radiusCompensation=adaptedRadiusCompensation,
            distance=distanceDelta
        )
        contourPass.setPathInMatrixPlane(matrix, list(PolygonConverter().convertVertices(offsetPath)))

        return contourPass

    @property
    def currentFeedRate(self):
        if hasattr(self, 'currentMultiUsageRouting'):
            return self.currentMultiUsageRouting.feedRate
        else:
            return 0.0

    @property
    def currentDepthPerStep(self):
        if hasattr(self, 'currentMultiUsageRouting'):
            return self.currentMultiUsageRouting.depthPerStep
        else:
            return 0.0


class VerticalContour2D(Contour2D):
    def assignToolAutomatically(self, surface, machineConfiguration):
        self.setActualDepth(surface, machineConfiguration, self.depth)
        self.tool = machineConfiguration.toolDatabase.getBestToolForContour(self.width, self.actualDepth)

    def setActualDepth(self, surface, machineConfiguration, depth):
        self.actualDepth = getVerticalMillingActualDepth(machineConfiguration, surface.thickness, depth)


class HorizontalContour2D(Contour2D):
    def assignToolAutomatically(self, surface, machineConfiguration):
        self.setActualDepth(surface, machineConfiguration, self.depth)
        self.tool = machineConfiguration.toolDatabase.getBestToolForHorizontalContour(self.width, self.actualDepth)

    def setActualDepth(self, _surface, _machineConfiguration, depth):
        self.actualDepth = copy.copy(depth)


class Contour2DTests(unittest.TestCase):
    def setUp(self):
        self.polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
        )
        self.contour = Contour2D(self.polyline, 10.0, 5.0, Contour2D.RADIUS_COMPENSATION_NONE)

    def testGetRadiusCompensationAsStringNone(self):
        self.assertEqual("OK", Contour2D.getRadiusCompensation(Contour2D.RADIUS_COMPENSATION_NONE, "OK", "BAD", "ALSO BAD"))

    def testGetRadiusCompensationAsStringLeft(self):
        self.assertEqual("OK", Contour2D.getRadiusCompensation(Contour2D.RADIUS_COMPENSATION_LEFT, "BAD", "OK", "ALSO BAD"))

    def testGetRadiusCompensationAsStringRight(self):
        self.assertEqual("OK", Contour2D.getRadiusCompensation(Contour2D.RADIUS_COMPENSATION_RIGHT, "BAD", "ALSO BAD", "OK"))

    def testAContourShouldBeEqualToItself(self):
        self.assertEqual(self.contour, self.contour)

    def testAContourShouldNotBeEqualToAnotherWithADifferentPath(self):
        self.assertNotEqual(self.contour, Contour2D(Polyline(), 10.0, 5.0, Contour2D.RADIUS_COMPENSATION_NONE))

    def testAContourShouldNotBeEqualToAnotherWithADifferentWidth(self):
        self.assertNotEqual(self.contour, Contour2D(self.polyline, 20.0, 5.0, Contour2D.RADIUS_COMPENSATION_NONE))

    def testAContourShouldNotBeEqualToAnotherWithADifferentDepth(self):
        self.assertNotEqual(self.contour, Contour2D(self.polyline, 10.0, 8.0, Contour2D.RADIUS_COMPENSATION_NONE))

    def testAContourShouldNotBeEqualToAnotherWithADifferentRadiusCompensation(self):
        self.assertNotEqual(self.contour, Contour2D(self.polyline, 10.0, 5.0, Contour2D.RADIUS_COMPENSATION_RIGHT))

    def testIsClosedOnAnOpenPathShouldBeFalse(self):
        self.contour.path = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=False),
        )
        self.assertFalse(self.contour.isClosed())

    def testIsClosedOnANonePathShouldBeFalse(self):
        self.contour.path = None
        self.assertFalse(self.contour.isClosed())

    def testIsClosedOnAClosedPathShouldBeTrue(self):
        self.assertTrue(self.contour.isClosed())

    def testIsClosedOnAPathReturningToStartShouldBeTrue(self):
        self.contour.path = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
            CornerVertex(Point(100, 100), visible=False),
        )
        self.assertTrue(self.contour.isClosed())

    def testGetInvertedRadiusCompensationNone(self):
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_NONE, Contour2D.getInvertedRadiusCompensation(Contour2D.RADIUS_COMPENSATION_NONE))

    def testGetInvertedRadiusCompensationLeft(self):
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_RIGHT, Contour2D.getInvertedRadiusCompensation(Contour2D.RADIUS_COMPENSATION_LEFT))

    def testGetInvertedRadiusCompensationRight(self):
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_LEFT, Contour2D.getInvertedRadiusCompensation(Contour2D.RADIUS_COMPENSATION_RIGHT))

    def testIsValid(self):
        self.assertTrue(self.contour.isValid())

    def testNoPathShouldNotBeIsValid(self):
        self.contour.path = None
        self.assertFalse(self.contour.isValid())

    def testEmptyPathShouldNotBeIsValid(self):
        self.contour.path = Polyline()
        self.assertFalse(self.contour.isValid())
