# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.surface.machining2d import Machining2D


class ContourMachining2D(Machining2D):
    def __init__(self, path=None):
        super().__init__()
        self.path = path

    def readFrom(self, contour3D):
        super().readFrom(contour3D)

        self.setPathInMatrixPlane(contour3D.matrix, contour3D.path.vertices[:])

    def setPathInMatrixPlane(self, matrix, vertices):
        self.pathInMatrixPlane = Polyline(vertices)
        self.path = Polyline([self.__convertVertex(vertex, matrix, self.position) for vertex in vertices])  # Legacy, only valid for XY-plane

    def __convertVertex(self, vertex, matrix, offset):
        convertedVertex = copy.deepcopy(vertex)
        convertedVertex.position.multiplyByMatrix(matrix)
        convertedVertex.moveBy(offset)
        return convertedVertex

    def transformedPolyline(self, matrix, offset):
        return Polyline([self.__convertVertex(vertex, matrix, offset) for vertex in self.pathInMatrixPlane.vertices[:]])

    def isValid(self):
        return self.path and self.path.isValid()

    def isClosed(self):
        return self.path and self.path.isClosed()

    def __str__(self):
        result = ' '.join((
            'ContourMAchining 2D:',
            'path({}),'.format(self.path),
         ))
        return result

    def __repr__(self):
        return self.__str__()

    def __eq__(self, rhs):
        return super().__eq__(rhs) and self.path == rhs.path

    def mirrorOnXAxis(self, surface):
        super().mirrorOnXAxis(surface)
        self.setPathInMatrixPlane(self.matrix, self.pathInMatrixPlane.vertices[:])

    def mirrorOnYAxis(self, surface):
        super().mirrorOnYAxis(surface)
        self.setPathInMatrixPlane(self.matrix, self.pathInMatrixPlane.vertices[:])

    def mirrorOnXYPlane(self, surface):
        super().mirrorOnXYPlane(surface)
        self.setPathInMatrixPlane(self.matrix, self.pathInMatrixPlane.vertices[:])

    def rotateLeft(self, surface):
        super().rotateLeft(surface)
        self.setPathInMatrixPlane(self.matrix, self.pathInMatrixPlane.vertices[:])

    def rotateRight(self, surface):
        super().rotateRight(surface)
        self.setPathInMatrixPlane(self.matrix, self.pathInMatrixPlane.vertices[:])

    def rotate180(self, surface):
        super().rotate180(surface)
        self.setPathInMatrixPlane(self.matrix, self.pathInMatrixPlane.vertices[:])


class ContourMachining2DTests(unittest.TestCase):
    def setUp(self):
        self.polyline = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
        )
        self.contourmachining = ContourMachining2D(self.polyline)

    def testAContourShouldBeEqualToItself(self):
        self.assertEqual(self.contourmachining, self.contourmachining)

    def testAContourShouldNotBeEqualToAnotherWithADifferentPath(self):
        self.assertNotEqual(self.contourmachining, ContourMachining2D(Polyline()))

    def testIsClosedOnAnOpenPathShouldBeFalse(self):
        self.contourmachining.path = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=False),
        )
        self.assertFalse(self.contourmachining.isClosed())

    def testIsClosedOnANonePathShouldBeFalse(self):
        self.contourmachining.path = None
        self.assertFalse(self.contourmachining.isClosed())

    def testIsClosedOnAClosedPathShouldBeTrue(self):
        self.assertTrue(self.contourmachining.isClosed())

    def testIsClosedOnAPathReturningToStartShouldBeTrue(self):
        self.contourmachining.path = Polyline(
            CornerVertex(Point(100, 100), visible=True),
            CornerVertex(Point(100, 0), visible=True),
            CornerVertex(Point(0, 0), visible=True),
            CornerVertex(Point(100, 100), visible=False),
        )
        self.assertTrue(self.contourmachining.isClosed())

    def testIsValid(self):
        self.assertTrue(self.contourmachining.isValid())

    def testNoPathShouldNotBeIsValid(self):
        self.contourmachining.path = None
        self.assertFalse(self.contourmachining.isValid())

    def testEmptyPathShouldNotBeIsValid(self):
        self.contourmachining.path = Polyline()
        self.assertFalse(self.contourmachining.isValid())
