# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest.mock

from interiorcad.backend.localization import LICENSE_ERRORS_RESOURCEGROUP, getLocalizedString
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.surface.actualdepths import getVerticalMillingActualDepth
from interiorcad.interface.surface.dadorabbet2d import DadoRabbet2D
from interiorcad.interface.surface.exportmethodnames import ExportKeys


class Dado2D(DadoRabbet2D):
    def __init__(self, position=Point(0, 0), width=0, depth=0, length=0, directionVector=Point(1, 0)):
        super().__init__(position, width, depth, length, directionVector)

    def getExportKey(self):
        return ExportKeys.DADO

    def __str__(self):
        return "Dado2D: " + super().__str__()

    def assignRouterToolOnly(self, surface, machineConfiguration):
        pass


class VerticalDado2D(Dado2D):
    def __getSaw(self, machineConfiguration):
        if self.isAlongXAxis():
            return machineConfiguration.toolDatabase.getDadoSawAlongX(self.width)
        elif self.isAlongYAxis():
            return machineConfiguration.toolDatabase.getDadoSawAlongY(self.width)
        else:
            return machineConfiguration.toolDatabase.getDadoSawWithCRotation(self.width)

    def assignTool(self, surface, machineConfiguration):
        sawTool = self.__getSaw(machineConfiguration)
        routerActualDepth, routerTool = self.__getRouterToolData(surface, machineConfiguration)

        # TODO IT-265: what do we want for saw actualDepth, if the dado is sawn-through? -- NG
        if self.isInset() and machineConfiguration.getAsBool('prefer-routing-for-inset-dados', False):
            self.tool = routerTool if routerTool else sawTool
            self.actualDepth = routerActualDepth if routerTool else copy.copy(self.depth)
        else:
            self.tool = sawTool if sawTool else routerTool
            self.actualDepth = copy.copy(self.depth)if sawTool else routerActualDepth

        self.assignToolIDFromTool()

    def assignRouterToolOnly(self, surface, machineConfiguration):
        routerActualDepth, routerTool = self.__getRouterToolData(surface, machineConfiguration)

        self.tool = routerTool
        self.actualDepth = routerActualDepth if routerTool else copy.copy(self.depth)

        self.assignToolIDFromTool()

    def __getRouterToolData(self, surface, machineConfiguration):
        routerActualDepth = getVerticalMillingActualDepth(machineConfiguration, surface.thickness, self.depth)
        routerTool = machineConfiguration.toolDatabase.getBestRouterForDado(self.width, routerActualDepth)
        return routerActualDepth, routerTool

    def getStarterError(self, workpieceName):
        if self.tool and self.tool.type == Tool.ROUTER:
            message = getLocalizedString(LICENSE_ERRORS_RESOURCEGROUP, 6, "License Error: routed Vertical Dado not activated. (Workpiece: {workpiece}, length: {length}, width: {width}, depth: {depth})")
            return message.format(workpiece=workpieceName, length=self.length, width=self.width, depth=self.depth)


class HorizontalDado2D(Dado2D):
    def assignTool(self, _surface, machineConfiguration):
        sawTool = machineConfiguration.toolDatabase.getHorizontalDadoSawWithCRotation(self.width)
        routerTool = self.__getRouterTool(machineConfiguration)

        if self.isInset() and machineConfiguration.getAsBool('prefer-routing-for-inset-dados', False):
            self.tool = routerTool if routerTool else sawTool
        else:
            self.tool = sawTool if sawTool else routerTool

        self.tool = routerTool  # FIXME IT-222 (HORIZONTAL): force using the router for now, as the NC exports cannot saw a horizontal dado (except woodWOP, but badly too) -- NG
        self.actualDepth = copy.copy(self.depth)

        self.assignToolIDFromTool()

    def assignRouterToolOnly(self, surface, machineConfiguration):
        self.tool = self.__getRouterTool(machineConfiguration)
        self.actualDepth = copy.copy(self.depth)

        self.assignToolIDFromTool()

    def __getRouterTool(self, machineConfiguration):
        routerTool = machineConfiguration.toolDatabase.getBestRouterForHorizontalDado(self.width, self.depth)
        return routerTool

    def getStarterError(self, workpieceName):
        message = getLocalizedString(LICENSE_ERRORS_RESOURCEGROUP, 5, "License Error: Horizontal Dado not activated. (Workpiece: {workpiece}, length: {length}, width: {width}, depth: {depth})")
        return message.format(workpiece=workpieceName, length=self.length, width=self.width, depth=self.depth)


class Tests(unittest.TestCase):
    def setUp(self):
        self.dado = Dado2D(Point(0, 0))

    def testEqualityOfDados(self):
        self.assertTrue(Dado2D() == Dado2D())
        self.assertEqual(Dado2D(), Dado2D())


class VerticalToolAssigningTest(unittest.TestCase):
    def setUp(self):
        self.dado = VerticalDado2D(Point(0, 0), directionVector=Point(0.7, 0.7))
        self.machineConfiguration = unittest.mock.Mock()
        self.machineConfiguration.toolDatabase = unittest.mock.Mock()
        self.machineConfiguration.toolDatabase.getDadoSawAlongX = unittest.mock.Mock()
        self.machineConfiguration.toolDatabase.getDadoSawAlongY = unittest.mock.Mock()
        self.machineConfiguration.toolDatabase.getDadoSawWithCRotation = unittest.mock.Mock()
        self.machineConfiguration.toolDatabase.getBestRouterForDado = unittest.mock.Mock()
        self.surface = unittest.mock.Mock()
        self.surface.thickness = 19

    def testDadoXShouldCallGetDadoSawAlongX(self):
        self.dado.matrix.u.x = 1
        self.dado.matrix.u.y = 0

        self.dado.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getDadoSawAlongX.called)
        self.assertFalse(self.machineConfiguration.toolDatabase.getDadoSawAlongY.called)
        self.assertFalse(self.machineConfiguration.toolDatabase.getDadoSawWithCRotation.called)
        self.assertEqual(self.machineConfiguration.toolDatabase.getDadoSawAlongX(), self.dado.tool)

    def testDadoYShouldCallGetDadoSawAlongY(self):
        self.dado.matrix.u.x = 0
        self.dado.matrix.u.y = 1

        self.dado.assignTool(self.surface, self.machineConfiguration)
        self.assertFalse(self.machineConfiguration.toolDatabase.getDadoSawAlongX.called)
        self.assertTrue(self.machineConfiguration.toolDatabase.getDadoSawAlongY.called)
        self.assertFalse(self.machineConfiguration.toolDatabase.getDadoSawWithCRotation.called)
        self.assertEqual(self.machineConfiguration.toolDatabase.getDadoSawAlongY(), self.dado.tool)

    def testFreeDadoShouldCallOnlyGetDadoSawWithCRotation(self):
        self.dado.matrix.u.x = 0.5
        self.dado.matrix.u.y = 0.5

        self.dado.assignTool(self.surface, self.machineConfiguration)
        self.assertFalse(self.machineConfiguration.toolDatabase.getDadoSawAlongX.called)
        self.assertFalse(self.machineConfiguration.toolDatabase.getDadoSawAlongY.called)
        self.assertTrue(self.machineConfiguration.toolDatabase.getDadoSawWithCRotation.called)
        self.assertEqual(self.machineConfiguration.toolDatabase.getDadoSawWithCRotation(), self.dado.tool)

    def testDadoXWithoutSawShouldCallGetDadoRouter(self):
        self.dado.matrix.u.x = 1
        self.dado.matrix.u.y = 0

        self.machineConfiguration.toolDatabase.getDadoSawAlongX.return_value = None
        self.dado.assignTool(self.surface, self.machineConfiguration)
        self.assertEqual(self.machineConfiguration.toolDatabase.getBestRouterForDado(), self.dado.tool)

    def testDadoYWithoutSawShouldCallGetDadoRouter(self):
        self.dado.matrix.u.x = 0
        self.dado.matrix.u.y = 1

        self.machineConfiguration.toolDatabase.getDadoSawAlongY.return_value = None
        self.dado.assignTool(self.surface, self.machineConfiguration)
        self.assertEqual(self.machineConfiguration.toolDatabase.getBestRouterForDado(), self.dado.tool)
