# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.surface.linemachining2d import LineMachining2D


class DadoRabbet2D(LineMachining2D):
    def __init__(self, position=Point(0, 0), width=0, depth=0, length=0, directionVector=Point(1, 0)):
        super().__init__(position, width, length, directionVector, depth=depth)

    def __str__(self):
        result = ' '.join((
            'DadoRabbet2D:',
            super().__str__() + ";",
        ))
        return result

    def __repr__(self):
        return self.__str__()

    def getDepth(self):
        return self.depth

    def mirrorOnXAxis(self, surface):
        super().mirrorOnXAxis(surface)
        if hasattr(self, "segment") and self.segment:
            self.segment.mirrorOnXAxis(surface.getFinishedSize().y)
            self.segment.invert()

    def mirrorOnYAxis(self, surface):
        super().mirrorOnYAxis(surface)
        if hasattr(self, "segment") and self.segment:
            self.segment.mirrorOnYAxis(surface.getFinishedSize().x)
            self.segment.invert()


class Tests(unittest.TestCase):
    def setUp(self):
        self.dadoRabbet = DadoRabbet2D(Point(0, 0))

    def testDefaultDrillingHasZeroWidth(self):
        self.assertEqual(0, self.dadoRabbet.getWidth())

    def testDefaultDrillingHasZeroDepth(self):
        self.assertEqual(0, self.dadoRabbet.getDepth())

    def testDefaultDrillingHasZeroLength(self):
        self.assertEqual(0, self.dadoRabbet.getLength())

    def testCanSetWidth(self):
        self.dadoRabbet.width = 15
        self.assertEqual(15, self.dadoRabbet.getWidth())

    def testCanSetDepth(self):
        self.dadoRabbet.depth = 15
        self.assertEqual(15, self.dadoRabbet.getDepth())

    def testCanSetLength(self):
        self.dadoRabbet.length = 5
        self.assertEqual(5, self.dadoRabbet.getLength())

    def testCompleteInit(self):
        self.dadoRabbet = DadoRabbet2D(Point(1, 2), 3, 4, 5, Point(0, 20))
        self.assertEqual(Point(1, 2), self.dadoRabbet.position)
        self.assertEqual(3.0, self.dadoRabbet.width)
        self.assertEqual(4.0, self.dadoRabbet.depth)
        self.assertEqual(5.0, self.dadoRabbet.length)
        self.assertEqual(Point(0, 1), self.dadoRabbet.directionVector)

    def testStartPointShouldBePosition(self):
        self.assertEqual(self.dadoRabbet.position, self.dadoRabbet.getStartPoint())

    def testEndPointShouldBeAtTheLengthOnTheXAxis(self):
        self.assertEqual(Point(self.dadoRabbet.position.x + self.dadoRabbet.length, self.dadoRabbet.position.y), self.dadoRabbet.getEndPoint())

    def testEndPointShouldBeAtTheLengthOnTheYAxis(self):
        self.assertEqual(Point(self.dadoRabbet.position.x, self.dadoRabbet.position.y + self.dadoRabbet.length), self.dadoRabbet.getEndPoint())

    def testIsOnXAxis(self):
        self.dadoRabbet.matrix.u.x = 1
        self.dadoRabbet.matrix.u.y = 0

        self.assertTrue(self.dadoRabbet.isAlongXAxis())
        self.assertFalse(self.dadoRabbet.isAlongYAxis())

    def testIsOnYAxis(self):
        self.dadoRabbet.matrix.u.x = 0
        self.dadoRabbet.matrix.u.y = 1

        self.assertFalse(self.dadoRabbet.isAlongXAxis())
        self.assertTrue(self.dadoRabbet.isAlongYAxis())

    def testEquality(self):
        self.assertTrue(DadoRabbet2D(Point(0, 0)) == self.dadoRabbet)
        self.assertEqual(DadoRabbet2D(Point(0, 0)), self.dadoRabbet)
        self.assertFalse(DadoRabbet2D(Point(0, 0)) != self.dadoRabbet)

    def testUnequalityOfPosition(self):
        self.assertFalse(DadoRabbet2D(Point(1, 2)) == self.dadoRabbet)
        self.assertTrue(DadoRabbet2D(Point(1, 2)) != self.dadoRabbet)

    def testUnequalityOfWidth(self):
        self.assertFalse(DadoRabbet2D(Point(0, 0), width=20) == self.dadoRabbet)
        self.assertTrue(DadoRabbet2D(Point(0, 0), width=20) != self.dadoRabbet)

    def testUnequalityOfLength(self):
        self.assertFalse(DadoRabbet2D(Point(0, 0), length=20) == self.dadoRabbet)
        self.assertTrue(DadoRabbet2D(Point(0, 0), length=20) != self.dadoRabbet)

    def testUnequalityOfDepth(self):
        self.assertFalse(DadoRabbet2D(Point(0, 0), depth=20) == self.dadoRabbet)
        self.assertTrue(DadoRabbet2D(Point(0, 0), depth=20) != self.dadoRabbet)

    def testUnequalityOfDirectionVector(self):
        self.assertFalse(DadoRabbet2D(Point(0, 0), directionVector=Point(0, -1)) == self.dadoRabbet)
        self.assertTrue(DadoRabbet2D(Point(0, 0), directionVector=Point(0, -1)) != self.dadoRabbet)

    def testInvertOfADadoAlongXAxis(self):
        dadoRabbet = DadoRabbet2D(Point(0, 0), directionVector=Point(1, 0), length=10)
        dadoRabbet.invert()
        self.assertEqual(DadoRabbet2D(Point(10, 0), directionVector=Point(-1, 0), length=10), dadoRabbet)

    def testInvertOfADadoAlongYAxis(self):
        dadoRabbet = DadoRabbet2D(Point(0, 0), directionVector=Point(0, 1), length=10)
        dadoRabbet.invert()
        self.assertEqual(DadoRabbet2D(Point(0, 10), directionVector=Point(0, -1), length=10), dadoRabbet)

    def testDefaultShouldNotBeInset(self):
        self.assertFalse(self.dadoRabbet.isInset())
        self.assertFalse(self.dadoRabbet.isInsetBoth())
        self.assertFalse(self.dadoRabbet.isInsetStart())
        self.assertFalse(self.dadoRabbet.isInsetEnd())

    def testIsInsetBoth(self):
        dadoRabbet = DadoRabbet2D(position=Point(0, 0), )
        dadoRabbet.startInset = 10
        dadoRabbet.endInset = 20
        self.assertTrue(dadoRabbet.isInset())
        self.assertTrue(dadoRabbet.isInsetBoth())
        self.assertTrue(dadoRabbet.isInsetStart())
        self.assertTrue(dadoRabbet.isInsetEnd())

    def testIsInsetAtStart(self):
        dadoRabbet = DadoRabbet2D(position=Point(0, 0), )
        dadoRabbet.startInset = 10
        self.assertTrue(dadoRabbet.isInset())
        self.assertFalse(dadoRabbet.isInsetBoth())
        self.assertTrue(dadoRabbet.isInsetStart())
        self.assertFalse(dadoRabbet.isInsetEnd())

    def testIsInsetAtEnd(self):
        dadoRabbet = DadoRabbet2D(position=Point(0, 0), )
        dadoRabbet.endInset = 20
        self.assertTrue(dadoRabbet.isInset())
        self.assertFalse(dadoRabbet.isInsetBoth())
        self.assertFalse(dadoRabbet.isInsetStart())
        self.assertTrue(dadoRabbet.isInsetEnd())
