# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest.mock

from interiorcad.backend.localization import LICENSE_ERRORS_RESOURCEGROUP, getLocalizedString
from interiorcad.interface.geometry import isNearlyGreaterOrEqual
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.surface.actualdepths import getVerticalDrillingActualDepth
from interiorcad.interface.surface.drillmachining2d import DrillMachining2D


class Drilling2D(DrillMachining2D):
    @classmethod
    def createFromTop(cls, position=None, diameter=0, depth=0, zPosition=0, installedDiameter=0):
        if position is None:
            position = Point(0, 0)
        direction = Vector(0, 0, 1)

        return cls(position, direction, diameter, depth, zPosition, installedDiameter)

    def __init__(self, position=None, direction=None, diameter=0, depth=0, zPosition=0, installedDiameter=0):
        super().__init__(position, direction, diameter, zPosition, installedDiameter, depth=depth)
        self.drillThrough = False

    def readFrom(self, machining3D):
        super().readFrom(machining3D)
        self.drillThrough = machining3D.drillThrough

    def __str__(self):
        result = ' '.join((
            'Drilling 2D:',
            super().__str__() + ";",
        ))
        return result

    def __repr__(self):
        return self.__str__()


class VerticalDrilling(Drilling2D):
    def assignTool(self, surface, machineConfiguration):
        boardThickness = surface.thickness
        if isNearlyGreaterOrEqual(self.depth, boardThickness):
            self.actualDepth = getVerticalDrillingActualDepth(machineConfiguration, boardThickness, self.depth)
            tool = machineConfiguration.toolDatabase.getVerticalDrillThrough(self.diameter)
        else:
            self.actualDepth = copy.copy(self.depth)
            tool = machineConfiguration.toolDatabase.getVerticalDrill(self.diameter)
        self.setTool(tool)


class HorizontalDrilling(Drilling2D):
    def assignTool(self, _surface, machineConfiguration):
        self.actualDepth = copy.copy(self.depth)

        if self.isXMinus():
            self.setTool(machineConfiguration.toolDatabase.getHorizontalDrillXMinus(self.diameter))
        elif self.isYMinus():
            self.setTool(machineConfiguration.toolDatabase.getHorizontalDrillYMinus(self.diameter))
        elif self.isXPlus():
            self.setTool(machineConfiguration.toolDatabase.getHorizontalDrillXPlus(self.diameter))
        elif self.isYPlus():
            self.setTool(machineConfiguration.toolDatabase.getHorizontalDrillYPlus(self.diameter))
        else:
            self.setTool(machineConfiguration.toolDatabase.getHorizontalDrillC(self.diameter))


class FiveAxisDrilling(Drilling2D):
    def assignTool(self, _surface, machineConfiguration):
        self.actualDepth = copy.copy(self.depth)
        self.setTool(machineConfiguration.toolDatabase.getFiveAxisDrill(self.diameter))

    def getStarterError(self, workpieceName):
        message = getLocalizedString(LICENSE_ERRORS_RESOURCEGROUP, 4, "License Error: five-axis Drilling not activated. (Workpiece: {workpiece}, diameter: {diameter})")
        return message.format(workpiece=workpieceName, diameter=self.diameter)


class Drilling2DTest(unittest.TestCase):
    def setUp(self):
        self.d = Drilling2D(Point(0, 0), Vector(0, 0, 1))

    def testDefaultDrillingHasPositionZeroZero(self):
        self.assertEqual(Point(0, 0), self.d.position)

    def testDefaultDrillingHasVectorFacingDown(self):
        self.assertEqual(Vector(0, 0, 1), self.d.direction)

    def testDefaultDrillingHasZeroDiameter(self):
        self.assertEqual(0, self.d.diameter)

    def testCanSetDiameter(self):
        self.d.diameter = 'new diameter'
        self.assertEqual('new diameter', self.d.diameter)

    def testDefaultDrillingHasZeroDepth(self):
        self.assertEqual(0, self.d.depth)

    def testCanSetDepthInConstructor(self):
        d = Drilling2D.createFromTop(depth='new depth')
        self.assertEqual('new depth', d.depth)

    def testDefaultDrillingHasZeroZPosition(self):
        self.assertEqual(0, self.d.zPosition)


class DrillingAngleTests(unittest.TestCase):
    def calculateAnglesAC(self, direction):
        drilling = Drilling2D(direction=direction)
        return drilling.cAngle, drilling.aAngle

    def testHorizontalDrillingFromRight(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(1, 0, 0))

        self.assertEqual(180, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromRight(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.866025, -0.5, 0))

        self.assertAlmostEqual(150, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testHorizontalDrillingFromLeft(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(-1, 0, 0))

        self.assertEqual(0, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromLeft(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(-0.866025, -0.5, 0))

        self.assertAlmostEqual(30, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testHorizontalDrillingFromBack(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0, 1, 0))

        self.assertEqual(270.0, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromBack(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.5, 0.866025, 0))

        self.assertAlmostEqual(240, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testHorizontalDrillingFromFront(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0, -1, 0))

        self.assertEqual(90.0, cAngle)
        self.assertEqual(90.0, aAngle)

    def testRotatedHorizontalDrillingFromFront(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.5, -0.866025, 0))

        self.assertAlmostEqual(120, cAngle, places=4)
        self.assertEqual(90.0, aAngle)

    def testVerticalDrillingFromTop(self):
        _cAngle, aAngle = self.calculateAnglesAC(Vector(0, 0, 1))

        self.assertEqual(0.0, aAngle)

    def test5AxisFromTopLeft(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(-0.5, 0, 0.8660254038))

        self.assertEqual(0, cAngle)
        self.assertAlmostEqual(30.0, aAngle, places=8)

    def test5Axis(self):
        cAngle, aAngle = self.calculateAnglesAC(Vector(0.29619813272602386, -0.81379768134937380, 0.50000000000000011))

        self.assertAlmostEqual(110, cAngle, places=8)
        self.assertAlmostEqual(60, aAngle, places=8)


class HorizontalDrillingToolAssigningTest(unittest.TestCase):
    def setUp(self):
        self.drilling = HorizontalDrilling(Point(0, 0), Vector(0.7, 0.7, 0))
        self.machineConfiguration = unittest.mock.Mock()
        self.machineConfiguration.toolDatabase = unittest.mock.Mock()
        self.surface = unittest.mock.Mock()

    def testFreeHorizontalDrillingShouldCallGetHorizontalDrill(self):
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalDrillC.called)

    def testHorizontalDrillingXMinusShouldCallGetHorizontalDrillXMinus(self):
        self.drilling.matrix.w = Vector(1, 0, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalDrillXMinus.called)

    def testHorizontalDrillingYMinusShouldCallGetHorizontalDrillYMinus(self):
        self.drilling.matrix.w = Vector(0, 1, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalDrillYMinus.called)

    def testHorizontalDrillingXPlusShouldCallGetHorizontalDrillXPlus(self):
        self.drilling.matrix.w = Vector(-1, 0, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalDrillXPlus.called)

    def testHorizontalDrillingYPlusShouldCallGetHorizontalDrillYPlus(self):
        self.drilling.matrix.w = Vector(0, -1, 0)
        self.drilling.assignTool(self.surface, self.machineConfiguration)
        self.assertTrue(self.machineConfiguration.toolDatabase.getHorizontalDrillYPlus.called)
