# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import itertools
import logging
import math
import unittest
from collections import OrderedDict

from interiorcad.backend.localization import getLocalizedString, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.interface.configuration import NCGlobalExportConfiguration
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector


class DrillingReader:
    @staticmethod
    def getDirectionValue(drilling):
        if drilling.direction == Vector(0, 0, 1):
            value = -2
        elif drilling.direction == Vector(0, 0, -1):
            value = -1
        else:
            value = math.atan2(-drilling.direction.y, drilling.direction.x)
            value = math.fmod(value + math.pi * 3, math.pi * 2)

        return value

    @staticmethod
    def getDiameter(drilling):
        return drilling.diameter

    def setSortCode(self, startCode):
        self.startCode = startCode

    def getPositionCodeValue(self, drilling, code):
        value = 0
        if code == NCGlobalExportConfiguration.MIN_X:
            value += drilling.position.x
        elif code == NCGlobalExportConfiguration.MAX_X:
            value -= drilling.position.x
        elif code == NCGlobalExportConfiguration.MIN_Y:
            value += drilling.position.y
        elif code == NCGlobalExportConfiguration.MAX_Y:
            value -= drilling.position.y
        elif code != NCGlobalExportConfiguration.NULL:
            assert False, "Invalid Drilling Position Code"
        return value

    def getPositionCoded(self, drilling):
        value = 0
        if len(self.startCode) > 0:
            value += self.getPositionCodeValue(drilling, self.startCode[0]) * 100000

        if len(self.startCode) > 1:
            value += self.getPositionCodeValue(drilling, self.startCode[1])

        return value


class DrillingSorter:
    @staticmethod
    def getSortingByDrillingDirection(x, y):
        sortDirection = [NCGlobalExportConfiguration.MIN_X]
        if math.fabs(x) < math.fabs(y):
            if y < 0:
                sortDirection = [NCGlobalExportConfiguration.MAX_X]
        else:
            if x > 0:
                sortDirection = [NCGlobalExportConfiguration.MAX_Y]
            else:
                sortDirection = [NCGlobalExportConfiguration.MIN_Y]

        return sortDirection

    @staticmethod
    def calculateStartPosition(drillings, startPositionCode):
        startPosition = Point(0, 0)
        if len(drillings) > 0 and startPositionCode is not None and len(startPositionCode) > 0:
            drillingreader = DrillingReader()

            if drillingreader.getDirectionValue(drillings[0]) < 0:
                drillingreader.setSortCode(startPositionCode)
            else:
                sortDirection = DrillingSorter.getSortingByDrillingDirection(drillings[0].direction.x, drillings[0].direction.y)
                drillingreader.setSortCode(sortDirection)

            startDrillings = sorted(drillings, key=drillingreader.getPositionCoded)
            startPosition = startDrillings[0].position

        return startPosition

    @staticmethod
    def sortDistance(drillings, startPositionCode):
        sorter = SorterByPosition()
        sorter.setObjects(drillings)
        sorter.initialPosition = DrillingSorter.calculateStartPosition(drillings, startPositionCode)
        sorter.process()
        return sorter.getObjects()

    @staticmethod
    def __sortByOrderedCriteriaHelper(drillings, startPositionCode, keys):
        if keys:
            key = keys[0]

            if key == 'distance':
                sortedDrillings = DrillingSorter.sortDistance(tuple(drillings), startPositionCode)
                # We cannot group the drillings sorted by distance correctly, so each drilling is its own group
                groups = ((drilling,) for drilling in sortedDrillings)
            else:
                sortedDrillings = sorted(drillings, key=key)
                groups = (group for _key, group in itertools.groupby(sortedDrillings, key=key))

            drillings = itertools.chain.from_iterable((DrillingSorter.__sortByOrderedCriteriaHelper(group, startPositionCode, keys[1:]) for group in groups))

        return drillings

    @staticmethod
    def getSortingCriteriaKeys(drillingOrderings):
        SORTING_CRITERIA = OrderedDict((
            ('direction', DrillingReader.getDirectionValue),
            ('toolid', lambda drilling: drilling.toolID or " "),
            ('diameter', DrillingReader.getDiameter),
            ('distance', 'distance'),
        ))
        # 'Direction ToolID Diameter Distance'   # old sorting criteria

        keys = []
        for ordering in drillingOrderings:
            if ordering.lower() in SORTING_CRITERIA:
                keys.append(SORTING_CRITERIA[ordering.lower()])
            else:
                message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 42, "Unknown drilling sorting ordering: {ordering}! Accepted values: {orderings}")
                logging.warning(message.format(ordering=ordering, orderings=tuple(SORTING_CRITERIA)))
        return keys

    @staticmethod
    def sortByOrderingCriteria(drillings, startPositionCode, drillingOrderings):
        keys = DrillingSorter.getSortingCriteriaKeys(drillingOrderings)
        return list(DrillingSorter.__sortByOrderedCriteriaHelper(drillings, startPositionCode, keys))


class TestDrilling:
    def __init__(self, direction=None, diameter=0, x=0, y=0, toolID="1"):
        self.direction = direction or Vector(0, 0, 1)
        self.diameter = diameter
        self.position = Point(x, y)
        self.toolID = toolID

    def getDiameter(self):
        return self.diameter

    def getPosition(self):
        return self.position

    def __eq__(self, other):
        return self.direction == other.direction and self.diameter == other.diameter and self.position == other.position and self.toolID == other.toolID

    def __repr__(self):
        return 'TestDrilling: Pos {position}, Dir {direction}, Diameter {diameter}, ToolID {toolID}'.format(position=self.position, direction=self.direction, diameter=self.diameter, toolID=self.toolID)

    @staticmethod
    def createTestDrillings():
        return [
            TestDrilling(Vector(0, 0, 1), 5, 10, 10),
            TestDrilling(Vector(0, 0, 1), 5, 100, 10),
            TestDrilling(Vector(0, 0, 1), 5, 20, 70),
            TestDrilling(Vector(0, 0, 1), 5, 200, 30),
            TestDrilling(Vector(0, 0, 1), 5, 50, 70),
            TestDrilling(Vector(0, -1, 0), 8, 20, 10),
            TestDrilling(Vector(0, -1, 0), 8, 200, 10),
            TestDrilling(Vector(0, -1, 0), 8, 200, 35),
            TestDrilling(Vector(0, 0, 1), 2, 10, 12),
            TestDrilling(Vector(0, 0, 1), 3, 13, 10),
            TestDrilling(Vector(0, 0, 1), 3, 15, 50),
            TestDrilling(Vector(0, 0, 1), 3, 15, 500),
            TestDrilling(Vector(0, 0, 1), 3, 5, 50),
            TestDrilling(Vector(0, 0, 1), 8, 18, 10),
            TestDrilling(Vector(0, 0, 1), 8, 100, 15),
            TestDrilling(Vector(math.sqrt(2) / 2, math.sqrt(2) / 2, 0), 5, 200, 40),
            TestDrilling(Vector(-1, 0, 0), 5, 200, 40),
            TestDrilling(Vector(0, 1, 0), 5, 200, 40),
            TestDrilling(Vector(0, -1, 0), 5, 20, 5),
            TestDrilling(Vector(0, -1, 0), 5, 200, 15),
            TestDrilling(Vector(1, 0, 0), 5, 200, 40)
        ]


class SortByToolIDAndDistance(unittest.TestCase):
    def setUp(self):
        self.drilling1 = TestDrilling(diameter=5, toolID="DefaultLongLongTextForAToolID_9")
        self.drilling2 = TestDrilling(diameter=5, toolID="DefaultLongLongTextForAToolID_1")
        self.drillings = [self.drilling1, self.drilling2]

    def sortByToolIDDiameterAndDistance(self, drillings):
        drillingOrderings = ('ToolID', 'Diameter', 'Distance')
        return list(DrillingSorter.sortByOrderingCriteria(drillings, '', drillingOrderings))

    def testSortByToolID(self):
        sortedDrillings = self.sortByToolIDDiameterAndDistance(self.drillings)
        self.assertListEqual([self.drilling2, self.drilling1], sortedDrillings)

    def testSortByToolID2(self):
        self.drilling1.toolID = "VerticalCenterPunchToolID_10"
        self.drilling2.toolID = "DefaultVerticalCenterPunchToolID"
        sortedDrillings = self.sortByToolIDDiameterAndDistance(self.drillings)
        self.assertListEqual([self.drilling2, self.drilling1], sortedDrillings)


class TestSorting(unittest.TestCase):
    def testFormerSortByFacingIndex(self):
        drillings = TestDrilling.createTestDrillings()
        self.assertEqual(len(drillings), 21)

        drillingOrderings = ('Direction', )
        sortedDrillings = list(DrillingSorter.sortByOrderingCriteria(drillings, startPositionCode='', drillingOrderings=drillingOrderings))

        self.assertEqual(len(sortedDrillings), 21)
        for point in sortedDrillings[:12]:
            self.assertEqual(point.direction, Vector(0, 0, 1))
        self.assertEqual(sortedDrillings[12].direction, Vector(-1, 0, 0))
        self.assertEqual(sortedDrillings[13].direction, Vector(0, 1, 0))
        self.assertEqual(sortedDrillings[14].direction, Vector(math.sqrt(2) / 2, math.sqrt(2) / 2, 0))
        self.assertEqual(sortedDrillings[15].direction, Vector(1, 0, 0))
        for point in sortedDrillings[16:21]:
            self.assertEqual(point.direction, Vector(0, -1, 0))

    def testFormerSortByDiameter(self):
        drillings = TestDrilling.createTestDrillings()

        drillingOrderings = ('Diameter', )
        sortedDrillings = list(DrillingSorter.sortByOrderingCriteria(drillings, startPositionCode='', drillingOrderings=drillingOrderings))

        self.assertEqual(len(sortedDrillings), 21)
        self.assertEqual(sortedDrillings[0].getDiameter(), 2)
        for point in sortedDrillings[1:5]:
            self.assertEqual(point.getDiameter(), 3)
        for point in sortedDrillings[5:16]:
            self.assertEqual(point.getDiameter(), 5)
        for point in sortedDrillings[16:]:
            self.assertEqual(point.getDiameter(), 8)

    def testSortByDistance(self):
        drillings = TestDrilling.createTestDrillings()

        DrillingSorter.startPositionNext = Point(0, 0)
        sortedDrillings = DrillingSorter.sortDistance(drillings, [NCGlobalExportConfiguration.NULL])
        self.assertEqual(len(sortedDrillings), 21)

        self.assertEqual(sortedDrillings[0].position, Point(10, 10))
        self.assertEqual(sortedDrillings[1].position, Point(10, 12))
        self.assertEqual(sortedDrillings[2].position, Point(13, 10))
        self.assertEqual(sortedDrillings[3].position, Point(18, 10))
        self.assertEqual(sortedDrillings[4].position, Point(20, 10))
        self.assertEqual(sortedDrillings[5].position, Point(20, 5))
        self.assertEqual(sortedDrillings[6].position, Point(15, 50))
        self.assertEqual(sortedDrillings[7].position, Point(5, 50))
        self.assertEqual(sortedDrillings[8].position, Point(20, 70))
        self.assertEqual(sortedDrillings[9].position, Point(50, 70))
        self.assertEqual(sortedDrillings[10].position, Point(100, 15))
        self.assertEqual(sortedDrillings[11].position, Point(100, 10))
        self.assertEqual(sortedDrillings[12].position, Point(200, 10))
        self.assertEqual(sortedDrillings[13].position, Point(200, 15))
        self.assertEqual(sortedDrillings[14].position, Point(200, 30))
        self.assertEqual(sortedDrillings[15].position, Point(200, 35))
        self.assertEqual(sortedDrillings[16].position, Point(200, 40))
        self.assertEqual(sortedDrillings[17].position, Point(200, 40))
        self.assertEqual(sortedDrillings[18].position, Point(200, 40))
        self.assertEqual(sortedDrillings[19].position, Point(200, 40))
        self.assertEqual(sortedDrillings[20].position, Point(15, 500))

    def testSortSurfaceDirectionDiameterDistance(self):
        drillings = TestDrilling.createTestDrillings()

        drillingOrderings = ('Direction', 'Diameter', 'Distance')
        startPositionCode = [NCGlobalExportConfiguration.MAX_Y, NCGlobalExportConfiguration.MIN_X]
        sortedDrillings = list(DrillingSorter.sortByOrderingCriteria(drillings, startPositionCode=startPositionCode, drillingOrderings=drillingOrderings))

        self.assertEqual(len(sortedDrillings), 21)

        self.assertEqual(sortedDrillings[0], TestDrilling(Vector(0, 0, 1), 2, 10, 12))

        self.assertEqual(sortedDrillings[1], TestDrilling(Vector(0, 0, 1), 3, 15, 500))
        self.assertEqual(sortedDrillings[2], TestDrilling(Vector(0, 0, 1), 3, 15, 50))
        self.assertEqual(sortedDrillings[3], TestDrilling(Vector(0, 0, 1), 3, 5, 50))
        self.assertEqual(sortedDrillings[4], TestDrilling(Vector(0, 0, 1), 3, 13, 10))

        self.assertEqual(sortedDrillings[5], TestDrilling(Vector(0, 0, 1), 5, 20, 70))
        self.assertEqual(sortedDrillings[6], TestDrilling(Vector(0, 0, 1), 5, 50, 70))
        self.assertEqual(sortedDrillings[7], TestDrilling(Vector(0, 0, 1), 5, 10, 10))
        self.assertEqual(sortedDrillings[8], TestDrilling(Vector(0, 0, 1), 5, 100, 10))
        self.assertEqual(sortedDrillings[9], TestDrilling(Vector(0, 0, 1), 5, 200, 30))

        self.assertEqual(sortedDrillings[10], TestDrilling(Vector(0, 0, 1), 8, 100, 15))
        self.assertEqual(sortedDrillings[11], TestDrilling(Vector(0, 0, 1), 8, 18, 10))

        self.assertEqual(sortedDrillings[12], TestDrilling(Vector(-1, 0, 0), 5, 200, 40))
        self.assertEqual(sortedDrillings[13], TestDrilling(Vector(0, 1, 0), 5, 200, 40))
        self.assertEqual(sortedDrillings[14], TestDrilling(Vector(math.sqrt(2) / 2, math.sqrt(2) / 2, 0), 5, 200, 40))
        self.assertEqual(sortedDrillings[15], TestDrilling(Vector(1, 0, 0), 5, 200, 40))

        self.assertEqual(sortedDrillings[16], TestDrilling(Vector(0, -1, 0), 5, 200, 15))
        self.assertEqual(sortedDrillings[17], TestDrilling(Vector(0, -1, 0), 5, 20, 5))

        self.assertEqual(sortedDrillings[18], TestDrilling(Vector(0, -1, 0), 8, 200, 10))
        self.assertEqual(sortedDrillings[19], TestDrilling(Vector(0, -1, 0), 8, 200, 35))
        self.assertEqual(sortedDrillings[20], TestDrilling(Vector(0, -1, 0), 8, 20, 10))

    def testNoSort(self):
        drillings = TestDrilling.createTestDrillings()

        drillingOrderings = ()
        startPositionCode = ''
        sortedDrillings = list(DrillingSorter.sortByOrderingCriteria(drillings, startPositionCode=startPositionCode, drillingOrderings=drillingOrderings))
        self.assertEqual(TestDrilling.createTestDrillings(), sortedDrillings)


# ## From sortbyposition.py
# I merged the sorting package together into one module
# because sorting went out of control - we should consider rewriting the package.
class ObjectWithPosition:
    def __init__(self, position):
        self.position = position

    def __lt__(self, other):
        return self.position < other.position

    def __str__(self):
        return ''.join(('ObjectWithPosition: ', str(self.position)))


class SorterByPosition:
    class SortByDistanceHelper:
        def __init__(self, initialObject):
            self.initialObject = initialObject

        def distanceTo(self, objectWithPosition):
            return self.initialObject.position.distanceTo(objectWithPosition.position)

    def __init__(self, initialPosition=Point(0, 0)):
        self.objects = []
        self.initialPosition = initialPosition

    def setObjects(self, objects):
        self.objects = objects

    def addObject(self, objectWithPosition):
        self.objects.append(objectWithPosition)

    def process(self):
        self.sortByLastObjectsPosition(self.initialPosition)

    def sortByLastObjectsPosition(self, initialPosition):
        sortedObjects = [ObjectWithPosition(initialPosition)]
        sortedObjects = self.getSortedByLastObjectsPosition(sortedObjects)
        sortedObjects.pop(0)
        self.objects = sortedObjects

    def getSortedByLastObjectsPosition(self, sortable):
        sortedObjects = sortable
        remainingObjects = self.objects
        while len(remainingObjects) > 0:
            lastIndex = len(sortedObjects) - 1
            remainingObjects = self.sortByDistance(remainingObjects, sortedObjects[lastIndex], reverse=True)
            sortedObjects.append(remainingObjects.pop())
        return sortedObjects

    def sortByDistance(self, sortable, objectWithPosition, reverse=False):
        sortHelper = self.SortByDistanceHelper(objectWithPosition)
        return sorted(sortable, key=sortHelper.distanceTo, reverse=reverse)

    def getObjects(self):
        return self.objects


class SortByDistance_WithFewObjects(unittest.TestCase):
    def setUp(self):
        self.sorter = SorterByPosition()
        self.object1 = ObjectWithPosition(Point(0, 0))
        self.object2 = ObjectWithPosition(Point(1, 0))

    def testSortNoObject(self):
        self.sorter.process()
        self.assertEqual([], self.sorter.getObjects())

    def testSortOneObject(self):
        self.sorter.addObject(self.object1)
        self.sorter.process()
        self.assertEqual([self.object1], self.sorter.getObjects())

    def testSortTwoObjects_InOrder(self):
        self.sorter.setObjects(self.twoObjectsSorted())
        self.sorter.process()
        self.assertEqual(self.twoObjectsSorted(), self.sorter.getObjects())

    def testSortTwoObjects_NotInOrder(self):
        self.sorter.setObjects(self.twoObjectsUnsorted())
        self.sorter.process()
        self.assertEqual(self.twoObjectsSorted(), self.sorter.getObjects())

    def testSortTwoObjects_WithDifferentStartPosition(self):
        self.sorter = SorterByPosition(Point(1, 1))
        self.sorter.setObjects(self.twoObjectsSorted())
        self.sorter.process()
        self.assertEqual([self.object2, self.object1], self.sorter.getObjects())

    def twoObjectsSorted(self):
        return [self.object1, self.object2]

    def twoObjectsUnsorted(self):
        return [self.object2, self.object1]


class SortByDistance_WithSomeObjects(unittest.TestCase):
    def setUp(self):
        self.sorter = SorterByPosition()
        self.object1 = ObjectWithPosition(Point(10, 19))
        self.object2 = ObjectWithPosition(Point(12, 20))
        self.object3 = ObjectWithPosition(Point(15, 18))
        self.object4 = ObjectWithPosition(Point(19, 16))
        self.object5 = ObjectWithPosition(Point(25, 17))

    def testSortFiveObjects_InOrder(self):
        self.sorter.setObjects(self.fiveObjectsSorted())
        self.sorter.initialPosition = Point(0, 0)
        self.sorter.process()
        self.assertEqual(self.fiveObjectsSorted(), self.sorter.getObjects())

    def testSortFiveObjects_NotInOrder(self):
        objectsNotInOrder = [self.object3, self.object2, self.object5, self.object1, self.object4]
        self.sorter.setObjects(objectsNotInOrder)
        self.sorter.initialPosition = Point(0, 0)
        self.sorter.process()
        self.assertEqual(self.fiveObjectsSorted(), self.sorter.getObjects())

    def fiveObjectsSorted(self):
        return [self.object1, self.object2, self.object3, self.object4, self.object5]
