# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest.mock

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.surface.machining2d import Machining2D
from interiorcad.interface.surface.acangles import calculateCAngleFromWVector
from interiorcad.interface.surface.exportmethodnames import ExportKeys
from interiorcad.interface.geometry import isNearlyEqual, EPSILON


class DrillMachining2D(Machining2D):
    def __init__(self, position=None, direction=None, diameter=0, zPosition=0, installedDiameter=0, *, depth=0):
        super().__init__(position, depth=depth)
        if direction:
            self.matrix.w = direction
        self.diameter = diameter
        self.matrix.origin.z = zPosition
        self.installedDiameter = installedDiameter or self.diameter

    def readFrom(self, machining3D):
        super().readFrom(machining3D)
        self.diameter = machining3D.getDiameter()

    def getExportKey(self):
        return ExportKeys.DRILLING

    def __str__(self):
        PRECISION = 2
        return ' '.join((
            'DrillMachining 2D:',
            super().__str__() + ";",
            'direction({}),'.format(self.direction),
            'diameter({d:.{cp}f}),'.format(d=self.diameter, cp=PRECISION),
            'installedDiameter({d:.{cp}f}),'.format(d=self.installedDiameter, cp=PRECISION),
            'zPosition({z:.{cp}f})'.format(z=self.zPosition, cp=PRECISION),
        ))

    def setTool(self, tool):
        if tool:
            self.tool = tool
            self.toolID = tool.getToolID()
            self.installedDiameter = tool.getInstalledDiameter()
        else:
            self.tool = None
            self.toolID = None
            self.installedDiameter = self.diameter

    def isXMinus(self):
        return isNearlyEqual(self.direction.dot(Vector(1, 0, 0)), 1)

    def isYMinus(self):
        return isNearlyEqual(self.direction.dot(Vector(0, 1, 0)), 1)

    def isXPlus(self):
        return isNearlyEqual(self.direction.dot(Vector(-1, 0, 0)), 1)

    def isYPlus(self):
        return isNearlyEqual(self.direction.dot(Vector(0, -1, 0)), 1)

    @property
    def cAngle(self):
        ### IT-144: it is bad that we have only the W vector of the matrix (mislabeled "direction". More correct would be to use the U vector of the matrix. -- NG
        return calculateCAngleFromWVector(wvector=self.direction)

    @property
    def direction(self):
        return self.matrix.w

    @classmethod
    def createFromDrilling(cls, drilling):
        newDrilling = cls(
            position=drilling.position,
            direction=drilling.direction,
            diameter=drilling.diameter,
            zPosition=drilling.zPosition,
            installedDiameter=drilling.installedDiameter
        )
        if hasattr(drilling, 'tool'):
            newDrilling.setTool(drilling.tool)
        if hasattr(drilling, 'depth'):
            newDrilling.depth = drilling.depth
        return newDrilling


class DrillMachining2DTest(unittest.TestCase):
    def setUp(self):
        self.d = DrillMachining2D(Point(0, 0), Vector(0, 0, 1))

    def testDefaultHasPositionZeroZero(self):
        self.assertEqual(Point(0, 0), self.d.position)

    def testDefaultHasVectorFacingDown(self):
        self.assertEqual(Vector(0, 0, 1), self.d.direction)

    def testDefaultHasZeroDiameter(self):
        self.assertEqual(0, self.d.diameter)

    def testCanSetDiameterInConstructor(self):
        d = DrillMachining2D(None, None, diameter='new diameter')
        self.assertEqual('new diameter', d.diameter)

    def testCanSetDiameter(self):
        self.d.diameter = 'new diameter'
        self.assertEqual('new diameter', self.d.diameter)

    def testCanSetZPositionInConstructor(self):
        d = DrillMachining2D(None, None, zPosition='new z position')
        self.assertEqual('new z position', d.zPosition)

    def testDefaultinstalledDiameterIsDiameter(self):
        d = DrillMachining2D(None, None, 5)
        self.assertEqual(5, d.installedDiameter)

    def testinstalledDiameterIsDiameterIfValueIsZero(self):
        d = DrillMachining2D(None, None, 5, installedDiameter=0)
        self.assertEqual(5, d.installedDiameter)

    def testinstalledDiameterIsValueIfValueIsGreaterThanZero(self):
        d = DrillMachining2D(None, None, 5, installedDiameter=7.1)
        self.assertEqual(7.1, d.installedDiameter)


class DrillMachining2DReadFromTest(unittest.TestCase):
    class TestPoint:
        def __init__(self):
            self.x = 'test_point_x'
            self.y = 'test_point_y'

    class TestVector:
        def __init__(self):
            self.x = 'testvector_x'
            self.y = 'testvector_y'
            self.z = 1

    class TestOffsetVector:
        def __init__(self):
            self.x = 'offset_x'
            self.y = 'offset_y'
            self.z = 'offset_z'

    class TestCenterPunch3D:
        def __init__(self):
            self.matrix = unittest.mock.Mock()
            self.matrix.origin = DrillMachining2DReadFromTest.TestOffsetVector()
            self.diameter = 'diameter'
            self.depth = 'depth'

        def getMatrix(self):
            return self.matrix

        def getDiameter(self):
            return self.diameter

        def getDepth(self):
            return self.depth

    def setUp(self):
        self.machining2D = DrillMachining2D()
        self.machining3D = self.TestCenterPunch3D()

    def testShouldReadAllParametersCorrectly(self):
        self.machining2D.readFrom(self.machining3D)
        self.assertEqual('offset_x', self.machining2D.position.x)
        self.assertEqual('offset_y', self.machining2D.position.y)
        self.assertEqual('diameter', self.machining2D.diameter)
        self.assertEqual('offset_z', self.machining2D.zPosition)


class IsNearlyEqualTests(unittest.TestCase):
    def testIsXMinus(self):
        self.drilling = DrillMachining2D(direction=Vector(1, 0, 0))
        self.assertTrue(self.drilling.isXMinus())
        self.assertFalse(self.drilling.isXPlus())
        self.assertFalse(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isYPlus())

    def testIsXMinusEpsilon(self):
        self.drilling = DrillMachining2D(direction=Vector(1 - EPSILON / 2, 0, 0))
        self.assertTrue(self.drilling.isXMinus())
        self.assertFalse(self.drilling.isXPlus())
        self.assertFalse(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isYPlus())

    def testIsXPlus(self):
        self.drilling = DrillMachining2D(direction=Vector(-1, 0, 0))
        self.assertTrue(self.drilling.isXPlus())
        self.assertFalse(self.drilling.isXMinus())
        self.assertFalse(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isYPlus())

    def testIsXPlusEpsilon(self):
        self.drilling = DrillMachining2D(direction=Vector(-1 + EPSILON / 2, 0, 0))
        self.assertTrue(self.drilling.isXPlus())
        self.assertFalse(self.drilling.isXMinus())
        self.assertFalse(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isYPlus())

    def testIsYMinus(self):
        self.drilling = DrillMachining2D(direction=Vector(0, 1, 0))
        self.assertTrue(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isYPlus())
        self.assertFalse(self.drilling.isXMinus())

    def testIsYMinusEpsilon(self):
        self.drilling = DrillMachining2D(direction=Vector(0, 1 - EPSILON / 2, 0))
        self.assertTrue(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isYPlus())
        self.assertFalse(self.drilling.isXMinus())
        self.assertFalse(self.drilling.isXPlus())

    def testIsYPlus(self):
        self.drilling = DrillMachining2D(direction=Vector(0, -1, 0))
        self.assertTrue(self.drilling.isYPlus())
        self.assertFalse(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isXMinus())
        self.assertFalse(self.drilling.isXPlus())

    def testIsYPlusEpsilon(self):
        self.drilling = DrillMachining2D(direction=Vector(0, -1 + EPSILON / 2, 0))
        self.assertTrue(self.drilling.isYPlus())
        self.assertFalse(self.drilling.isYMinus())
        self.assertFalse(self.drilling.isXMinus())
        self.assertFalse(self.drilling.isXPlus())
