# -*- coding: utf-8 -*-

# Copyright 2015-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import itertools
import unittest

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.surface.drilling2d import Drilling2D
from interiorcad.interface.surface.exportmethodnames import ExportKeys
from interiorcad.interface.surface.machining2d import Machining2D


class GridDrilling2D(Machining2D):
    def __init__(self, position, drillingDirection, gridNormalizedDirection, gridDistance, countDrillings, diameter, depth):
        super().__init__(position=position, depth=depth)
        self.drillingDirection = drillingDirection
        self.gridNormalizedDirection = gridNormalizedDirection
        self.gridDistance = gridDistance
        self.countDrillings = countDrillings
        self.diameter = diameter

    # TODO: grid drillings have no UUID (esp. a stable one for tests) (IT-195)
    @staticmethod
    def createFromSingleDrilling(drilling, gridNormalizedDirection, gridDistance, countDrillings):
        gridDrilling = GridDrilling2D(
            position=drilling.position,
            drillingDirection=drilling.direction,
            gridNormalizedDirection=gridNormalizedDirection,
            gridDistance=gridDistance,
            countDrillings=countDrillings,
            diameter=drilling.diameter,
            depth=drilling.depth
        )
        gridDrilling.matrix = drilling.matrix
        return gridDrilling

    def getDrillingPositions(self):
        return itertools.accumulate(itertools.chain((self.position,), itertools.repeat(self.gridNormalizedDirection * self.gridDistance, self.countDrillings - 1)))

    def getGridLength(self):
        return self.gridDistance * (self.countDrillings - 1)

    def getDeltaVector(self):
        return self.gridNormalizedDirection * self.gridDistance

    def __createDrilling(self, position):
        drilling = Drilling2D.createFromTop(position=position, diameter=self.diameter, depth=self.depth)

        if hasattr(self, 'tool'):
            drilling.setTool(self.tool)

        return drilling

    def getStartDrilling(self):
        return self.__createDrilling(position=self.position)

    def getEndDrilling(self):
        position = self.position
        position += (self.gridNormalizedDirection * self.getGridLength())
        return self.__createDrilling(position=position)

    def getExportKey(self):
        return ExportKeys.GRIDDRILLING

    def __eq__(self, rhs):
        return all((
            super().__eq__(rhs),
            self.position == rhs.position,
            self.drillingDirection == rhs.drillingDirection,
            self.gridNormalizedDirection == rhs.gridNormalizedDirection,
            self.gridDistance == rhs.gridDistance,
            self.countDrillings == rhs.countDrillings,
            self.diameter == rhs.diameter,
        ))

    def __repr__(self):
        PRECISION = 2
        result = ' '.join((
            'GridDrilling2D:',
            'position={},'.format(self.position),
            'drillingDirection={},'.format(self.drillingDirection),
            'gridNormalizedDirection={},'.format(self.gridNormalizedDirection),
            'gridDistance={},'.format(self.gridDistance),
            'countDrillings={},'.format(self.countDrillings),
            'diameter={d:.{cp}f},'.format(d=self.diameter, cp=PRECISION),
        ))
        return result

    def assignTool(self, _surface, machineConfiguration):
        self.actualDepth = self.depth

        if self.isAlongXAxis():
            self.tool = machineConfiguration.toolDatabase.getGridDrillingAlongX(self.diameter)
        elif self.isAlongYAxis():
            self.tool = machineConfiguration.toolDatabase.getGridDrillingAlongY(self.diameter)
        else:
            self.tool = machineConfiguration.toolDatabase.getGridDrillingAlongCRotated(self.diameter)

    def isAlongXAxis(self):
        return self.gridNormalizedDirection == Vector(1.0, 0.0, 0.0) or self.gridNormalizedDirection == Vector(-1.0, 0.0, 0.0)

    def isAlongYAxis(self):
        return self.gridNormalizedDirection == Vector(0.0, 1.0, 0.0) or self.gridNormalizedDirection == Vector(0.0, -1.0, 0.0)

    def getXYAngle(self):
        return VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(self.gridNormalizedDirection)

    def getInstalledDiameter(self):
        if hasattr(self, 'tool'):
            return self.tool.getInstalledDiameter()

        return -0.0

    def getToolID(self):
        if hasattr(self, 'tool'):
            return self.tool.getToolID()

        return ''

    def mirrorOnXAxis(self, surface):
        if self.isAlongYAxis():
            self.position = self.getEndDrilling().position
        else:
            assert self.isAlongXAxis(), "Grid drilling not along the X or Y axis are not implemented yet!"

        super().mirrorOnXAxis(surface)

    def mirrorOnYAxis(self, surface):
        if self.isAlongXAxis():
            self.position = self.getEndDrilling().position
        else:
            assert self.isAlongYAxis(), "Grid drilling not along the X or Y axis are not implemented yet!"

        super().mirrorOnYAxis(surface)

    def rotate180(self, surface):
        self.position = self.getEndDrilling().position
        super().rotate180(surface)


class VerticalTests(unittest.TestCase):
    def testOneDrillingShouldBeAtStartPosition(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=1, diameter=5, depth=8)
        self.assertEqual((Point(0, 0),), tuple(gridDrilling.getDrillingPositions()))
        self.assertEqual(Point(0, 0), gridDrilling.getStartDrilling().position)
        self.assertEqual(0, gridDrilling.getGridLength())

    def testTwoDrillingsInXGridDrilling(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=2, diameter=5, depth=8)
        self.assertEqual((Point(0, 0), Point(32, 0)), tuple(gridDrilling.getDrillingPositions()))
        self.assertEqual(Point(0, 0), gridDrilling.getStartDrilling().position)
        self.assertEqual(32, gridDrilling.getGridLength())

    def testTwoDrillingsInYGridDrilling(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=2, diameter=5, depth=8)
        self.assertEqual((Point(0, 0), Point(0, 32)), tuple(gridDrilling.getDrillingPositions()))
        self.assertEqual(Point(0, 0), gridDrilling.getStartDrilling().position)
        self.assertEqual(32, gridDrilling.getGridLength())

    def testFiveDrillingsInXGridDrilling(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=5, diameter=5, depth=8)
        self.assertEqual((Point(0, 0), Point(32, 0), Point(64, 0), Point(96, 0), Point(128, 0)), tuple(gridDrilling.getDrillingPositions()))
        self.assertEqual(Point(0, 0), gridDrilling.getStartDrilling().position)
        self.assertEqual(128, gridDrilling.getGridLength())

    def testFiveDrillingsInYGridDrilling(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=5, diameter=5, depth=8)
        self.assertEqual((Point(0, 0), Point(0, 32), Point(0, 64), Point(0, 96), Point(0, 128)), tuple(gridDrilling.getDrillingPositions()))
        self.assertEqual(Point(0, 0), gridDrilling.getStartDrilling().position)
        self.assertEqual(128, gridDrilling.getGridLength())


class MirrorTests(unittest.TestCase):
    class MockSurface:
        def getFinishedSizeX(self):
            return 600

        def getFinishedSizeY(self):
            return 400

    def testFiveDrillingsInXGridDrillingMirroredOnXAxis(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=5, diameter=5, depth=8)
        expectedGridDrilling = GridDrilling2D(position=Point(0, 400), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=5, diameter=5, depth=8)

        gridDrilling.mirrorOnXAxis(self.MockSurface())
        self.assertEqual(expectedGridDrilling, gridDrilling)

    def testFiveDrillingsInYGridDrillingMirroredOnXAxis(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=5, diameter=5, depth=8)
        expectedGridDrilling = GridDrilling2D(position=Point(0, 400 - 4 * 32), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=5, diameter=5, depth=8)

        gridDrilling.mirrorOnXAxis(self.MockSurface())
        self.assertEqual(expectedGridDrilling, gridDrilling)

    def testFiveDrillingsInXGridDrillingMirroredOnYAxis(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=5, diameter=5, depth=8)
        expectedGridDrilling = GridDrilling2D(position=Point(600 - 4 * 32, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(1.0, 0.0, 0.0), gridDistance=32, countDrillings=5, diameter=5, depth=8)

        gridDrilling.mirrorOnYAxis(self.MockSurface())
        self.assertEqual(expectedGridDrilling, gridDrilling)

    def testFiveDrillingsInYGridDrillingMirroredOnYAxis(self):
        gridDrilling = GridDrilling2D(position=Point(0, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=5, diameter=5, depth=8)
        expectedGridDrilling = GridDrilling2D(position=Point(600, 0), drillingDirection=Vector(0, 0, 1), gridNormalizedDirection=Vector(0, 1), gridDistance=32, countDrillings=5, diameter=5, depth=8)

        gridDrilling.mirrorOnYAxis(self.MockSurface())
        self.assertEqual(expectedGridDrilling, gridDrilling)
