# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.backend.ncexport.linemachininganalyzer import LineMachiningAnalyzer
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.simplelinesegment import SimpleLineSegment
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.surface.machining2d import Machining2D


class LineMachining2D(Machining2D):
    def __init__(self, position=None, width=0, length=0, directionVector=Point(1, 0), *, depth=0):
        super().__init__(position, depth=depth),
        self.width = width
        self.length = length
        self.startInset = 0
        self.endInset = 0
        if directionVector:
            directionVector.normalize()
            self.matrix.u.x = directionVector.x
            self.matrix.u.y = directionVector.y

    def readFrom(self, machining3D):
        super().readFrom(machining3D)
        self.width = machining3D.getWidth()
        self.length = machining3D.getRealLength()
        self.position -= self.directionVector * (machining3D.startAddition - machining3D.startInset)
        self.startInset = machining3D.startInset
        self.endInset = machining3D.endInset

    def __str__(self):
        result = ' '.join((
            'LineMachining2D:',
            super().__str__() + ";",
            'width({d:.{cp}f}),'.format(d=self.width, cp=2),
            'length({d:.{cp}f}),'.format(d=self.length, cp=2),
            'v({x:.{cp}f} : {y:.{cp}f})'.format(x=self.directionVector.x, y=self.directionVector.y, cp=2),
        ))
        return result

    def __repr__(self):
        return self.__str__()

    def getWidth(self):
        return self.width

    def getLength(self):
        return self.length

    def getDeltaVector(self):
        return self.directionVector * self.length

    # Deprecated, please use directionVector
    def getDirectionVector(self):
        return Point(self.matrix.u.x, self.matrix.u.y)

    def getStartPoint(self):
        return self.position

    def getEndPoint(self):
        return self.position + self.getDeltaVector()

    def isAlongXAxis(self):
        return self.directionVector == Point(1, 0) or self.directionVector == Point(-1, 0)

    def isAlongYAxis(self):
        return self.directionVector == Point(0, 1) or self.directionVector == Point(0, -1)

    def __eq__(self, rhs):
        return super().__eq__(rhs) and self.width == rhs.width and self.length == rhs.length and self.directionVector == rhs.directionVector  and self.startInset == rhs.startInset and self.endInset == rhs.endInset

    def invert(self):
        self.position = self.getEndPoint()

        self.matrix.u.x *= -1
        self.matrix.u.y *= -1
        self.matrix.v.x *= -1
        self.matrix.v.y *= -1

        self.startInset, self.endInset = self.endInset, self.startInset

    def getSegmentOfMachining(self, workpieceShape, expectedDistanceFromLineToSegment):
        segments = LineMachiningAnalyzer().generateFoundSegmentsAtDistance(self, workpieceShape, expectedDistanceFromLineToSegment)
        segment = next(iter(segments), None)
        if segment:
            return copy.deepcopy(segment)

    def getStartAndEndPoints(self):
        start = copy.copy(self.getStartPoint())
        end = copy.copy(self.getEndPoint())
        return start, end

    def getLineSegment(self):
        return SimpleLineSegment(self.getStartPoint(), self.getEndPoint())

    def getLineSegmentWithSecurityDistance(self, securityDistance):
        # TODO IT-164/IT-255: ist das nicht teilweise duplizierten Code mit interiorcad.backend.ncexportmigrated.dadotoncsegments.__getRightAndLeftLinesWithSecurityDistance -- NG

        segment = self.getLineSegment()
        if not self.isInsetStart():
            segment.increaseAtStart(offset=securityDistance)
        if not self.isInsetEnd():
            segment.increaseAtEnd(offset=securityDistance)
        return segment

    def __getStartAndEndPointsModifieddByVector(self, perpendicularVector):
        start = copy.copy(self.getStartPoint()) + perpendicularVector
        end = copy.copy(self.getEndPoint()) + perpendicularVector

        return start, end

    def getStartAndEndPointsModifiedLeftHandByDistance(self, **keywords):
        directionVector = self.directionVector
        perpendicularVector = - directionVector.perp() * keywords['correctionDistance']

        return self.__getStartAndEndPointsModifieddByVector(perpendicularVector)

    def getStartAndEndPointsModifiedRightHandByDistance(self, **keywords):
        directionVector = self.directionVector
        perpendicularVector = directionVector.perp() * keywords['correctionDistance']

        return self.__getStartAndEndPointsModifieddByVector(perpendicularVector)

    def getXYAngle(self):
        return VectorMath.getDegreeAngleOfProjectedVectorInXYPlane(self.directionVector)

    def moveRight(self, distance):
        rabbetLineSegment = self.getLineSegment()
        rabbetLineSegment.moveRight(distance)
        self.position = rabbetLineSegment.start  # We do not care about the end, as the direction vector and the length have not changed

    def moveLeft(self, distance):
        rabbetLineSegment = self.getLineSegment()
        rabbetLineSegment.moveLeft(distance)
        self.position = rabbetLineSegment.start  # We do not care about the end, as the direction vector and the length have not changed

    @property
    def directionVector(self):
        return Point(self.matrix.u.x, self.matrix.u.y)

    def isInsetStart(self):
        return not isNearlyEqual(self.startInset, 0)

    def isInsetEnd(self):
        return not isNearlyEqual(self.endInset, 0)

    def isInset(self):
        return self.isInsetStart() or self.isInsetEnd()

    def isInsetBoth(self):
        return self.isInsetStart() and self.isInsetEnd()


class XYAngleTests(unittest.TestCase):
    def setUp(self):
        self.lineFromLeftToRight = LineMachining2D(position=Point(0, 0), length=100, directionVector=Point(1, 0))
        self.lineFromRightToLeft = LineMachining2D(position=Point(100, 0), length=100, directionVector=Point(-1, 0))
        self.lineFromBottomtoTop = LineMachining2D(position=Point(0, 0), length=100, directionVector=Point(0, 1))
        self.lineFromTopToBottom = LineMachining2D(position=Point(0, 100), length=100, directionVector=Point(0, -1))

    def testXYOfLineFromLeftToRight(self):
        self.assertEqual(0, self.lineFromLeftToRight.getXYAngle())

    def testXYAngleOfLineFromRightToLeft(self):
        self.assertEqual(180, self.lineFromRightToLeft.getXYAngle())

    def testXYOfLineFromBottomtoTop(self):
        self.assertEqual(90, self.lineFromBottomtoTop.getXYAngle())

    def testXYAngleOfLineFromTopToBottom(self):
        self.assertEqual(-90, self.lineFromTopToBottom.getXYAngle())


class Tests(unittest.TestCase):
    def setUp(self):
        self.machining = LineMachining2D(Point(0, 0))

    def testDefaultDrillingHasZeroWidth(self):
        self.assertEqual(0, self.machining.getWidth())

    def testDefaultDrillingHasZeroLength(self):
        self.assertEqual(0, self.machining.getLength())

    def testCanSetWidth(self):
        self.machining.width = 15
        self.assertEqual(15, self.machining.getWidth())

    def testCanSetLength(self):
        self.machining.length = 5
        self.assertEqual(5, self.machining.getLength())

    def testCompleteInit(self):
        self.machining = LineMachining2D(Point(1, 2), 3, 5, Point(0, 20))
        self.assertEqual(Point(1, 2), self.machining.position)
        self.assertEqual(3.0, self.machining.width)
        self.assertEqual(5.0, self.machining.length)
        self.assertEqual(Point(0, 1), self.machining.directionVector)

    def testStartPointShouldBePosition(self):
        self.assertEqual(self.machining.position, self.machining.getStartPoint())

    def testEndPointShouldBeAtTheLengthOnTheXAxis(self):
        self.assertEqual(Point(self.machining.position.x + self.machining.length, self.machining.position.y), self.machining.getEndPoint())

    def testEndPointShouldBeAtTheLengthOnTheYAxis(self):
        self.assertEqual(Point(self.machining.position.x, self.machining.position.y + self.machining.length), self.machining.getEndPoint())

    def testIsOnXAxis(self):
        self.machining.matrix.u.x = 1
        self.machining.matrix.u.y = 0

        self.assertTrue(self.machining.isAlongXAxis())
        self.assertFalse(self.machining.isAlongYAxis())

    def testIsOnYAxis(self):
        self.machining.matrix.u.x = 0
        self.machining.matrix.u.y = 1

        self.assertFalse(self.machining.isAlongXAxis())
        self.assertTrue(self.machining.isAlongYAxis())

    def testEquality(self):
        self.assertTrue(LineMachining2D(Point(0, 0)) == self.machining)
        self.assertEqual(LineMachining2D(Point(0, 0)), self.machining)
        self.assertFalse(LineMachining2D(Point(0, 0)) != self.machining)

    def testUnequalityOfPosition(self):
        self.assertFalse(LineMachining2D(Point(1, 2)) == self.machining)
        self.assertTrue(LineMachining2D(Point(1, 2)) != self.machining)

    def testUnequalityOfWidth(self):
        self.assertFalse(LineMachining2D(Point(0, 0), width=20) == self.machining)
        self.assertTrue(LineMachining2D(Point(0, 0), width=20) != self.machining)

    def testUnequalityOfLength(self):
        self.assertFalse(LineMachining2D(Point(0, 0), length=20) == self.machining)
        self.assertTrue(LineMachining2D(Point(0, 0), length=20) != self.machining)

    def testUnequalityOfDirectionVector(self):
        self.assertFalse(LineMachining2D(Point(0, 0), directionVector=Point(0, -1)) == self.machining)
        self.assertTrue(LineMachining2D(Point(0, 0), directionVector=Point(0, -1)) != self.machining)

    def testInvertOfADadoAlongXAxis(self):
        dadoRabbet = LineMachining2D(Point(0, 0), directionVector=Point(1, 0), length=10)
        dadoRabbet.invert()
        self.assertEqual(LineMachining2D(Point(10, 0), directionVector=Point(-1, 0), length=10), dadoRabbet)

    def testInvertOfADadoAlongYAxis(self):
        dadoRabbet = LineMachining2D(Point(0, 0), directionVector=Point(0, 1), length=10)
        dadoRabbet.invert()
        self.assertEqual(LineMachining2D(Point(0, 10), directionVector=Point(0, -1), length=10), dadoRabbet)

    def testMoveLeftX(self):
        dadoRabbet = LineMachining2D(Point(0, 0), directionVector=Point(1, 0), length=10)
        dadoRabbet.moveLeft(20)
        self.assertEqual(LineMachining2D(Point(0, 20), directionVector=Point(1, 0), length=10), dadoRabbet)

    def testMoveLeftY(self):
        dadoRabbet = LineMachining2D(Point(0, 0), directionVector=Point(0, 1), length=10)
        dadoRabbet.moveLeft(20)
        self.assertEqual(LineMachining2D(Point(-20, 0), directionVector=Point(0, 1), length=10), dadoRabbet)

    def testMoveRightX(self):
        dadoRabbet = LineMachining2D(Point(0, 0), directionVector=Point(1, 0), length=10)
        dadoRabbet.moveRight(20)
        self.assertEqual(LineMachining2D(Point(0, -20), directionVector=Point(1, 0), length=10), dadoRabbet)

    def testMoveRightY(self):
        dadoRabbet = LineMachining2D(Point(0, 0), directionVector=Point(0, 1), length=10)
        dadoRabbet.moveRight(20)
        self.assertEqual(LineMachining2D(Point(20, 0), directionVector=Point(0, 1), length=10), dadoRabbet)

    def testGetStartAndEndPointsModifiedLeftHandByDistance(self):
        self.machining.length = 100
        start, end = self.machining.getStartAndEndPointsModifiedLeftHandByDistance(correctionDistance=5)
        self.assertEqual(Point(0, 5), start)
        self.assertEqual(Point(100, 5), end)

    def testGetStartAndEndPointsModifiedRightHandByDistance(self):
        self.machining.length = 100
        start, end = self.machining.getStartAndEndPointsModifiedRightHandByDistance(correctionDistance=5)
        self.assertEqual(Point(0, -5), start)
        self.assertEqual(Point(100, -5), end)


class InsetTests(unittest.TestCase):
    def testInvertOfAlongXAxis(self):
        machining = LineMachining2D(Point(0, 0), directionVector=Point(1, 0), length=10)
        machining.invert()
        self.assertEqual(LineMachining2D(Point(10, 0), directionVector=Point(-1, 0), length=10), machining)

    def testInvertOfAlongYAxis(self):
        dadoRabbet = LineMachining2D(Point(0, 0), directionVector=Point(0, 1), length=10)
        dadoRabbet.invert()
        self.assertEqual(LineMachining2D(Point(0, 10), directionVector=Point(0, -1), length=10), dadoRabbet)

    def testDefaultShouldNotBeInset(self):
        machining = LineMachining2D()
        self.assertFalse(machining.isInset())
        self.assertFalse(machining.isInsetBoth())
        self.assertFalse(machining.isInsetStart())
        self.assertFalse(machining.isInsetEnd())

    def testIsInsetBoth(self):
        machining = LineMachining2D(position=Point(0, 0), )
        machining.startInset = 10
        machining.endInset = 20
        self.assertTrue(machining.isInset())
        self.assertTrue(machining.isInsetBoth())
        self.assertTrue(machining.isInsetStart())
        self.assertTrue(machining.isInsetEnd())

    def testIsInsetAtStart(self):
        machining = LineMachining2D(position=Point(0, 0), )
        machining.startInset = 10
        self.assertTrue(machining.isInset())
        self.assertFalse(machining.isInsetBoth())
        self.assertTrue(machining.isInsetStart())
        self.assertFalse(machining.isInsetEnd())

    def testIsInsetAtEnd(self):
        machining = LineMachining2D(position=Point(0, 0), )
        machining.endInset = 20
        self.assertTrue(machining.isInset())
        self.assertFalse(machining.isInsetBoth())
        self.assertFalse(machining.isInsetStart())
        self.assertTrue(machining.isInsetEnd())
