# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.surface.acangles import calculateAAngleFromWVector, calculateCAngleToXYPlaneFromWVector


class Machining2D:
    def __init__(self, position=None, *, depth=None):
        self.matrix = Matrix()
        self.position = position if position else Point(0, 0)
        self.toolID = None
        self.uuid = None
        self.depth = depth
        self.actualDepth = None

    def moveBy(self, delta):
        self.position += delta

    def readFrom(self, machining3D):
        self.matrix = copy.copy(machining3D.getMatrix())
        if hasattr(machining3D, 'uuid'):
            self.uuid = machining3D.uuid
        if hasattr(machining3D, 'depth'):
            self.depth = machining3D.depth

    def __getPosition3DOnPlaneAsTuple(self, plane):
        position3D = self.matrix.origin
        assert plane.isPointOnPlane(position3D), "Trying to get machining on a plane that does not belong to the machining"

        vector = position3D - plane.wopMatrix.origin

        if plane.predefined:
            return vector.dot(plane.wopMatrix.u), vector.dot(plane.wopMatrix.v), vector.dot(plane.wopMatrix.w)
        else:
            planeNormalVector = - plane.wopMatrix.w
            uVector = planeNormalVector.cross(Vector(0, 0, 1)).normalized()
            vVector = uVector.cross(planeNormalVector).normalized()
            return vector.dot(uVector), vector.dot(vVector), vector.dot(-planeNormalVector)

    def getPositionOnPlane(self, plane):
        x, y, _z = self.__getPosition3DOnPlaneAsTuple(plane)
        return Point(x, y)

    def getPosition3DOnPlane(self, plane):
        return Vector(*self.__getPosition3DOnPlaneAsTuple(plane))

    def __str__(self):
        result = ' '.join((
            'Machining 2D:',
            'position({})'.format(self.position),
            'depth({})'.format(self.depth),
        ))
        return result

    def __eq__(self, rhs):
        return self.position == rhs.position and self.depth == rhs.depth

    def assignToolIDFromTool(self):
        self.toolID = self.tool.getToolID() if self.tool else None

    def mirrorOnXAxis(self, surface):
        self.matrix.origin.y = surface.getFinishedSizeY() - self.matrix.origin.y

        self.matrix.u.y *= -1
        self.matrix.v.y *= -1
        self.matrix.w.y *= -1

    def mirrorOnYAxis(self, surface):
        self.matrix.origin.x = surface.getFinishedSizeX() - self.matrix.origin.x

        self.matrix.u.x *= -1
        self.matrix.v.x *= -1
        self.matrix.w.x *= -1

    def mirrorOnXYPlane(self, surface):
        self.matrix.origin.z = surface.thickness - self.matrix.origin.z

        self.matrix.u.z *= -1
        self.matrix.v.z *= -1
        self.matrix.w.z *= -1

    def rotateLeft(self, surface):
        self.matrix.rotateZPlus90()
        self.matrix.origin.x += surface.getFinishedSizeY()  # getFinishedSizeY as the finished size is not rotated yet

    def rotateRight(self, surface):
        self.matrix.rotateZMinus90()
        self.matrix.origin.y += surface.getFinishedSizeX()  # getFinishedSizeX as the finished size is not rotated yet

    def rotate180(self, surface):
        self.matrix.rotateZ180()
        self.matrix.origin.x += surface.getFinishedSizeX()
        self.matrix.origin.y += surface.getFinishedSizeY()

    @property
    def zPosition(self):
        return self.matrix.origin.z

    @property
    def position(self):
        return Point(self.matrix.origin.x, self.matrix.origin.y)

    @position.setter
    def position(self, point):
        self.matrix.origin.x = point.x
        self.matrix.origin.y = point.y

    def isVertical(self):
        return isNearlyEqual(self.matrix.w.z, 1.0)

    def isHorizontal(self):
        return isNearlyEqual(self.matrix.w.z, 0)

    def convertPositionIn3D(self, position, zPosition):
        return self.matrix.transform3D(position.x, position.y, zPosition)

    @property
    def aAngle(self):
        return calculateAAngleFromWVector(wvector=self.matrix.w)

    @property
    def cAngleToXYPlane(self):
        return calculateCAngleToXYPlaneFromWVector(wvector=self.matrix.w)

    def getStarterError(self, _workpieceName):
        return None


class Tests(unittest.TestCase):
    def setUp(self):
        self.machining = Machining2D(Point(0, 0))

    def testDefaultDrillingHasPositionZeroZero(self):
        self.assertEqual(Point(0, 0), self.machining.position)

    def testCanSetPosition(self):
        self.machining.position = Point(1, 2)
        self.assertEqual(self.machining.matrix.origin.x, 1)
        self.assertEqual(self.machining.matrix.origin.y, 2)

    def testEquality(self):
        self.assertTrue(Machining2D(Point(0, 0)) == self.machining)
        self.assertEqual(Machining2D(Point(0, 0)), self.machining)
        self.assertFalse(Machining2D(Point(0, 0)) != self.machining)

    def testUnequality(self):
        self.assertFalse(Machining2D(Point(1, 2)) == self.machining)
        self.assertTrue(Machining2D(Point(1, 2)) != self.machining)

    def testMoveBy(self):
        self.machining.moveBy(Point(1, 2))
        self.assertEqual(Machining2D(Point(1, 2)), self.machining)

    def testIsVerticalFromTop(self):
        self.machining.matrix = Matrix()
        self.assertTrue(self.machining.isVertical())

    def testIsVerticalFromBottom(self):
        self.machining.matrix = Matrix(w=Vector(0, 0, -1))
        self.assertFalse(self.machining.isVertical())

    def testIsVerticalFromHorizontal(self):
        self.machining.matrix = Matrix(w=Vector(1, 0, 0))
        self.assertFalse(self.machining.isVertical())
