# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections
import copy

from interiorcad.backend.workpiece import Drilling, DadoRabbet, Miter, Contour, CenterPunch, NCMacro, Chamfer
from interiorcad.backend.workpiece.complexgriddrilling import ComplexGridDrilling
from interiorcad.backend.workpiece.machininganalyzer import MachiningAnalyzer
from interiorcad.backend.workpiece.nullmachining import NullMachining
from interiorcad.backend.workpiece.profile import Profile
from interiorcad.interface.surface import Rabbet2D, VerticalDado2D, Miter2D, VerticalContour2D, VerticalCenterPunch, HorizontalCenterPunch, FiveAxisCenterPunch, NCMacro2D
from interiorcad.interface.surface.chamfer2d import Chamfer2D
from interiorcad.interface.surface.complexgriddrilling2D import ComplexGridDrilling2D
from interiorcad.interface.surface.contour2d import HorizontalContour2D
from interiorcad.interface.surface.dado2d import HorizontalDado2D
from interiorcad.interface.surface.drilling2d import VerticalDrilling, HorizontalDrilling, FiveAxisDrilling
from interiorcad.interface.surface.nullmachining2d import NullMachining2D
from interiorcad.interface.surface.profile2d import Profile2D

Replacement2D = collections.namedtuple("Replacement2D", "vertical horizontal fiveaxis")


CLASSES_3D_2D = (
    # DadoRabbet is a special case, see MachiningConverter.createDadoRabbetClass2D
    (Drilling, Replacement2D(VerticalDrilling, HorizontalDrilling, FiveAxisDrilling)),
    (Miter, Replacement2D(Miter2D, None, None)),
    (Contour, Replacement2D(VerticalContour2D, HorizontalContour2D, None)),
    (CenterPunch, Replacement2D(VerticalCenterPunch, HorizontalCenterPunch, FiveAxisCenterPunch)),
    (NCMacro, Replacement2D(NCMacro2D, NCMacro2D, NCMacro2D)),
    (Chamfer, Replacement2D(Chamfer2D, None, None)),
    (Profile, Replacement2D(Profile2D, None, None)),
    (ComplexGridDrilling, Replacement2D(ComplexGridDrilling2D, None, None)),
    (NullMachining, Replacement2D(NullMachining2D, NullMachining2D, NullMachining2D))  # Keep entry last
)


class MachiningConverter:
    def __init__(self, workpieceShape):
        self.machiningAnalyzer = MachiningAnalyzer()
        self.workpieceShape = workpieceShape

    def convertMachinings(self, machinings3D):
        return [self.convertMachining(machining3D) for machining3D in machinings3D]

    def convertMachining(self, machining3D):
        machining2D = self.createMachiningClass2D(machining3D)()
        machining2D.readFrom(copy.deepcopy(machining3D))
        return machining2D

    def createMachiningClass2D(self, machining3D):
        if isinstance(machining3D, DadoRabbet):
            return self.createDadoRabbetClass2D(machining3D)
        else:
            for class3D, replacement2D in CLASSES_3D_2D:
                if isinstance(machining3D, class3D):
                    return self.createClass2D(machining3D, replacement2D)

            # TODO InnoWeek 2021-12: assert? I18N/L10N? -- NG
            assert False, "Internal error. Machining Class unknown: [class3D}]".format(class3D=machining3D.__class__.__name__)

    def createDadoRabbetClass2D(self, machining3D):
        if self.machiningAnalyzer.isVertical(machining3D):
            if self.machiningAnalyzer.isRabbet(machining3D, self.workpieceShape):
                return Rabbet2D
            else:
                return VerticalDado2D
        elif self.machiningAnalyzer.isHorizontal(machining3D):
            return HorizontalDado2D

        # TODO InnoWeek 2021-12: assert? I18N/L10N? -- NG
        assert False, "Internal error. Situation for Dado/Rabbet unsupported"

    def createClass2D(self, machining3D, replacement2D):
        if self.machiningAnalyzer.isVertical(machining3D):
            class2D = replacement2D.vertical
        elif self.machiningAnalyzer.isHorizontal(machining3D):
            class2D = replacement2D.horizontal
        else:
            class2D = replacement2D.fiveaxis

        if class2D is not None:
            return class2D
        else:
            # TODO InnoWeek 2021-12: assert? I18N/L10N? -- NG
            assert False, "Internal error. Situation for Machining Class unsupported: {class3D}".format(class3D=machining3D.__class__.__name__)


class ForceClassMachiningConverter(MachiningConverter):
    def __init__(self, cls):
        super().__init__(None)
        self.cls = cls

    def createMachiningClass2D(self, _machining3D):
        return self.cls
