# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import operator
import unittest

from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector


def simpleSideIDGenerator(lastUsedSideId):
    sideIDIndex = lastUsedSideId
    while True:
        sideIDIndex += 1
        yield sideIDIndex


class MachiningPlanes:
    def __init__(self, lastUsedSideId=None, sideIDGenerator=None):
        assert operator.xor(lastUsedSideId is not None, sideIDGenerator is not None), 'Not both parameters can be None!'
        self.planes = []
        if sideIDGenerator is None:
            self.sideIDGenerator = simpleSideIDGenerator(lastUsedSideId)
        else:
            self.sideIDGenerator = sideIDGenerator

    def createPlanes(self, machinings):
        for machining in machinings:
            self.createPlane(machining)

    def createPlane(self, machining):
        plane = Plane.createFromMachining(machining)
        if not self.exists(plane):
            self.registerPlane(plane)

    def registerPlane(self, plane):
        plane.sideID = next(self.sideIDGenerator)
        self.planes.append(plane)

    def exists(self, otherPlane):
        return any(plane.isCompatiblePlane(otherPlane) for plane in self.planes)

    def findPlane(self, searchedPlane):
        foundPlane = None
        for plane in self.planes:
            if searchedPlane.isCompatiblePlane(plane):
                foundPlane = plane
                break

        return foundPlane


class Tests(unittest.TestCase):
    def testRegisterAndFind(self):
        plane1 = Plane(Vector(1, 2, 3), Vector(1, 0, 0))
        plane2 = Plane(Vector(0, 0, 0), Vector(0, 1, 0))

        creator = MachiningPlanes(lastUsedSideId=0)
        creator.registerPlane(plane1)
        creator.registerPlane(plane2)

        searchedPlane = Plane(Vector(0, 0, 0), Vector(0, 1, 0))

        foundPlane = creator.findPlane(searchedPlane)
        self.assertEqual(Vector(0, 0, 0), foundPlane.origin)
        self.assertEqual(Vector(0, 1, 0), foundPlane.normalVector)
        self.assertEqual(2, foundPlane.sideID)

    def testSideID(self):
        plane1 = Plane(Vector(1, 2, 3), Vector(1, 0, 0))
        plane2 = Plane(Vector(0, 0, 0), Vector(0, 1, 0))

        creator = MachiningPlanes(lastUsedSideId=0)
        creator.registerPlane(plane1)
        creator.registerPlane(plane2)

        self.assertEqual(1, plane1.sideID)
        self.assertEqual(2, plane2.sideID)

    def testSpecificSideIDGenerator(self):
        def mockSideIDGenerator():
            sideIDIndex = 64
            while True:
                sideIDIndex += 1
                yield chr(sideIDIndex)

        plane1 = Plane(Vector(1, 2, 3), Vector(1, 0, 0))
        plane2 = Plane(Vector(0, 0, 0), Vector(0, 1, 0))

        creator = MachiningPlanes(sideIDGenerator=mockSideIDGenerator())
        creator.registerPlane(plane1)
        creator.registerPlane(plane2)

        self.assertEqual('A', plane1.sideID)
        self.assertEqual('B', plane2.sideID)
