# -*- coding: utf-8 -*-

# Copyright 2017, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import operator

from interiorcad.interface.machineconfiguration.tool import Tool


class MachiningSorter:
    def __init__(self, export):
        self.export = export

    def sortMachiningsBySortingName(self, machinings):
        machinings = tuple(machinings)
        for machiningType in self.export.MACHININGS_CONFIGURATION:
            machiningClasses = tuple(creator.machiningClass for creator in self.export.SINGLE_MACHINING_CREATORS if creator.sortingName == machiningType)

            yield from filter(lambda machining: isinstance(machining, machiningClasses), machinings)
        # We assume that there is no machining that was not filtered.

    def sortMachiningsBySortingFace(self, machinings):
        machinings = tuple(machinings)
        for face in self.export.SORTING_FACE_VERTICAL, self.export.SORTING_FACE_HORIZONTAL, self.export.SORTING_FACE_FIVE_AXIS:
            machiningClasses = tuple(creator.machiningClass for creator in self.export.SINGLE_MACHINING_CREATORS if creator.sortingFace == face)

            yield from filter(lambda machining: isinstance(machining, machiningClasses), machinings)
        # We assume that there is no machining that was not filtered.

    def sortMachiningsByToolType(self, machinings):
        TOOL_TYPES = (Tool.ROUTER, Tool.SAW, Tool.DRILL, None)  # We want drillings that are exported without an explicit tool to be last.

        machinings = tuple(machinings)
        for toolType in TOOL_TYPES:
            for machining in machinings:
                if hasattr(machining, 'tool') and machining.tool and machining.tool.type == toolType:
                    yield machining
                elif toolType is None and (not hasattr(machining, 'tool') or not machining.tool):
                    yield machining

    def sortMachiningByToolUsage(self, machinings):
        remainMachinings = list(self.sortMachiningsByToolType(machinings))

        while remainMachinings:
            referenceMachining = remainMachinings[0]
            yield referenceMachining
            indicesToDelete = [0]
            for i in range(1, len(remainMachinings)):
                # FIXME: can the if-cascade be simplified? -- NG (InnoWeek 2017-12)
                if operator.xor(hasattr(referenceMachining, 'tool'), hasattr(remainMachinings[i], 'tool')):
                    pass
                elif not hasattr(referenceMachining, 'tool') and not hasattr(remainMachinings[i], 'tool'):
                    yield remainMachinings[i]
                    indicesToDelete.append(i)
                elif referenceMachining.tool == remainMachinings[i].tool:
                    yield remainMachinings[i]
                    indicesToDelete.append(i)

            for i in reversed(indicesToDelete):
                del remainMachinings[i]

    def sortMachiningsByToolDimension(self, machinings):
        ROUTER_SORT, SAW_SORT, DRILLING_SORT, UNKNOWN_SORT = range(4)

        sortingEntries = []
        for machining in tuple(machinings):
            if hasattr(machining, 'tool') and machining.tool:
                if machining.tool.type == Tool.ROUTER:
                    sortingEntries.append(((ROUTER_SORT, -machining.tool.getRouterDiameter()), machining))
                elif machining.tool.type == Tool.SAW:
                    sortingEntries.append(((SAW_SORT, machining.tool.getSawCutWidth()), machining))
                elif machining.tool.type == Tool.DRILL:
                    sortingEntries.append(((DRILLING_SORT, machining.tool.getDrillingDiameter()), machining))
                else:
                    assert False, 'Tool has no supported type!'
            else:
                if hasattr(machining, 'diameter'):
                    sortingEntries.append(((UNKNOWN_SORT, machining.diameter), machining))
                else:
                    sortingEntries.append(((UNKNOWN_SORT, float('+inf')), machining))

        return [entry[1] for entry in sorted(sortingEntries, key=lambda entry: entry[0])]

    def sortMachiningsBasically(self, machinings):
        machinings = self.sortMachiningsBySortingName(machinings)
        return self.sortMachiningsBySortingFace(machinings)
