# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.backend.localization import getLocalizedString, LICENSE_ERRORS_RESOURCEGROUP
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.surface.exportmethodnames import ExportKeys
from interiorcad.interface.surface.linemachining2d import LineMachining2D


class Miter2D(LineMachining2D):
    def __init__(self, position=None, width=0, length=0, directionVector=Point(1, 0), angle=0):
        position = position or Point(0, 0)
        super().__init__(position, width, length, directionVector)
        self.angle = angle
        self.segment = None
        self.surface = None  # DO NOT USE! Only meant for supporting workpieceShape in old code.

    def readFrom(self, machining3D):
        super().readFrom(machining3D)
        self.angle = machining3D.getAngle()

    def assignTool(self, surface, machineConfiguration):
        self.actualDepth = copy.copy(self.depth)  # TODO IT-265: what do we want for saw actualDepth -- NG
        self.tool = machineConfiguration.toolDatabase.getToolForRouting(Tool.MITER)
        self.assignToolIDFromTool()
        self.surface = surface  # DO NOT USE! Only meant for supporting workpieceShape in old code.

    def getExportKey(self):
        return ExportKeys.MITER

    def __str__(self):
        result = ' '.join((
            'Miter 2D:',
            super().__str__() + ";",
            'angle({d:.{cp}f})'.format(d=self.angle, cp=2),
        ))
        return result

    def __repr__(self):
        return self.__str__()

    def getAngle(self):
        return self.angle

    def __eq__(self, rhs):
        return super().__eq__(rhs) and self.angle == rhs.angle

    # Compatibility variable for old code (until 2018 F7); do NOT use for new code! Use directly finalFinishedSizeShape from Surface (but not through self.surface)
    @property
    def workpieceShape(self):
        if self.surface:
            return self.surface.finalFinishedSizeShape

    def getStarterError(self, workpieceName):
        message = getLocalizedString(LICENSE_ERRORS_RESOURCEGROUP, 2, "License Error: Miter not activated. (Workpiece: {workpiece}, length: {length}, width: {width})")
        return message.format(workpiece=workpieceName, length=self.length, width=self.width)


class Tests(unittest.TestCase):
    def setUp(self):
        self.miter = Miter2D(Point(0, 0))

    def testDefaultDrillingHasZeroWidth(self):
        self.assertEqual(0, self.miter.getWidth())

    def testDefaultDrillingHasZeroLength(self):
        self.assertEqual(0, self.miter.getLength())

    def testCanSetWidth(self):
        self.miter.width = 15
        self.assertEqual(15, self.miter.getWidth())

    def testCanSetLength(self):
        self.miter.length = 5
        self.assertEqual(5, self.miter.getLength())

    def testCompleteInit(self):
        self.miter = Miter2D(Point(1, 2), 3, 5, Point(0, 20))
        self.assertEqual(Point(1, 2), self.miter.position)
        self.assertEqual(3.0, self.miter.width)
        self.assertEqual(5.0, self.miter.length)
        self.assertEqual(Point(0, 1), self.miter.directionVector)
        self.assertEqual(0.0, self.miter.angle)

    def testStartPointShouldBePosition(self):
        self.assertEqual(self.miter.position, self.miter.getStartPoint())

    def testEndPointShouldBeAtTheLengthOnTheXAxis(self):
        self.assertEqual(Point(self.miter.position.x + self.miter.length, self.miter.position.y), self.miter.getEndPoint())

    def testEndPointShouldBeAtTheLengthOnTheYAxis(self):
        self.assertEqual(Point(self.miter.position.x, self.miter.position.y + self.miter.length), self.miter.getEndPoint())

    def testIsOnXAxis(self):
        self.miter.matrix.u.x = 1
        self.miter.matrix.u.y = 0

        self.assertTrue(self.miter.isAlongXAxis())
        self.assertFalse(self.miter.isAlongYAxis())

    def testIsOnYAxis(self):
        self.miter.matrix.u.x = 0
        self.miter.matrix.u.y = 1

        self.assertFalse(self.miter.isAlongXAxis())
        self.assertTrue(self.miter.isAlongYAxis())

    def testEquality(self):
        self.assertTrue(Miter2D(Point(0, 0)) == self.miter)
        self.assertEqual(Miter2D(Point(0, 0)), self.miter)
        self.assertFalse(Miter2D(Point(0, 0)) != self.miter)

    def testUnequalityOfPosition(self):
        self.assertFalse(Miter2D(Point(1, 2)) == self.miter)
        self.assertTrue(Miter2D(Point(1, 2)) != self.miter)

    def testUnequalityOfWidth(self):
        self.assertFalse(Miter2D(Point(0, 0), width=20) == self.miter)
        self.assertTrue(Miter2D(Point(0, 0), width=20) != self.miter)

    def testUnequalityOfLength(self):
        self.assertFalse(Miter2D(Point(0, 0), length=20) == self.miter)
        self.assertTrue(Miter2D(Point(0, 0), length=20) != self.miter)

    def testUnequalityOfDirectionVector(self):
        self.assertFalse(Miter2D(Point(0, 0), directionVector=Point(0, -1)) == self.miter)
        self.assertTrue(Miter2D(Point(0, 0), directionVector=Point(0, -1)) != self.miter)

    def testInvertOfADadoAlongXAxis(self):
        dadoRabbet = Miter2D(Point(0, 0), directionVector=Point(1, 0), length=10)
        dadoRabbet.invert()
        self.assertEqual(Miter2D(Point(10, 0), directionVector=Point(-1, 0), length=10), dadoRabbet)

    def testInvertOfADadoAlongYAxis(self):
        dadoRabbet = Miter2D(Point(0, 0), directionVector=Point(0, 1), length=10)
        dadoRabbet.invert()
        self.assertEqual(Miter2D(Point(0, 10), directionVector=Point(0, -1), length=10), dadoRabbet)

    def testEqShouldRejectUnequalAngles(self):
        miter1 = Miter2D(angle=21)
        miter2 = Miter2D(angle=42)
        self.assertFalse(miter1 == miter2)
