# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections
import copy
import unittest

from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.surface.acangles import calculateCAngleFromWVector, calculateCAngleInPlaneFromMatrix, normalizeAngle
from interiorcad.interface.surface.exportmethodnames import ExportKeys
from interiorcad.interface.surface.machining2d import Machining2D

Parameter = collections.namedtuple('Parameter', 'name value')


class NCMacro2D(Machining2D):
    def __init__(self, position=None, name=None, deltax=0, deltay=0, depth=0):
        super().__init__(position, depth=depth)
        self.name = name
        self.deltax = deltax
        self.deltay = deltay
        self.code = None
        self.clamexcangle = None  # Not a property, as its calculation might trigger logging warnings too.
        self.clamexcangle180 = None  # Variant, where the +180° is added for non-vertical cases. Not a property, as its calculation might trigger logging warnings too.

    def readFrom(self, machining3D):
        super().readFrom(machining3D)
        self.name = machining3D.name
        self.deltax = machining3D.deltax
        self.deltay = machining3D.deltay
        self.code = machining3D.code
        self.parameters = tuple(Parameter(*parameter) for parameter in machining3D.parameters)

    def getVariables(self, formatter):
        variables = dict(vars(self))
        del variables['parameters']  # We do not want to have any recursive problem.
        variables['x'] = formatter.formatAsCoordinate(data=self.position.x)
        variables['y'] = formatter.formatAsCoordinate(data=self.position.y)
        variables['z'] = formatter.formatAsCoordinate(data=self.zPosition)
        variables['deltax'] = formatter.formatAsCoordinate(data=self.deltax)
        variables['deltay'] = formatter.formatAsCoordinate(data=self.deltay)
        variables['depth'] = formatter.formatAsCoordinate(data=self.depth)
        variables['cangle'] = formatter.formatAsAngle(data=self.cAngle)
        variables['aangle'] = formatter.formatAsAngle(data=self.aAngle)
        return variables

    def getExportKey(self):
        return ExportKeys.NCMACRO

    def __str__(self):
        PRECISION = 2
        result = ' '.join((
            'NC Macro 2D:',
            super().__str__() + ";",
            'name={}'.format(self.name),
            'deltax={d:.{cp}f}'.format(d=self.deltax, cp=PRECISION),
            'deltay={d:.{cp}f}'.format(d=self.deltay, cp=PRECISION),
            'depth={d:.{cp}f}'.format(d=self.depth, cp=PRECISION),
        ))
        return result

    def __repr__(self):
        return self.__str__()

    @property
    def cAngle(self):
        if self.matrix.w != Vector(0, 0, 1):
            return normalizeAngle(calculateCAngleFromWVector(self.matrix.w) - 90.0)
        else:
            return calculateCAngleInPlaneFromMatrix(self.matrix)

    def assignTool(self, _surface, _machineConfiguration):
        # The tool cannot come from the tool database but must be part of the parameters of the NC macro
        self.actualDepth = copy.copy(self.depth)

    def hasCode(self):
        return bool(self.code)

    def isCodeContaining(self, text: str) -> bool:
        return self.code and text in self.code


class Tests(unittest.TestCase):
    def setUp(self):
        self.ncmacro = NCMacro2D()

    def testDefaultHasNoCode(self):
        self.assertFalse(self.ncmacro.hasCode())

    def testEmptyCodeIsNoCode(self):
        self.ncmacro.code = ''
        self.assertFalse(self.ncmacro.hasCode())

    def testSomeCodeIsCode(self):
        self.ncmacro.code = 'The quick bron fox jumps over the lazy dog!'
        self.assertTrue(self.ncmacro.hasCode())
