# -*- coding: utf-8 -*-

# Copyright 2011-2019, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH


import unittest.mock

from interiorcad.interface.surface.linemachiningorienter import LineMachiningOrienter
from interiorcad.interface.surface.contour2d import Contour2D
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.surface.rabbet2d import Rabbet2D


class RabbetAdjuster(LineMachiningOrienter):
    def __init__(self, finalFinishedSizeShape, oppositeDirection=False):
        super().__init__(finalFinishedSizeShape)
        self.oppositeDirection = oppositeDirection

    def __getOffsetVector(self, rabbet2D):
        offsetVector = rabbet2D.directionVector.perp()
        offsetVector *= rabbet2D.width / 2

        if self.oppositeDirection:
            offsetVector = - offsetVector

        return offsetVector

    def __getAdjustedRabbetAtEdge(self, rabbet2D, orienter, radiusCompensation, multiplicator):
        rabbet2D.segment = rabbet2D.getSegmentOfMachining(self.finalFinishedSizeShape, rabbet2D.width / 2)
        if self.hasSegment(rabbet2D):
            orienter(rabbet2D)
            rabbet2D.radiusCompensation = radiusCompensation
        rabbet2D.moveBy(self.__getOffsetVector(rabbet2D) * multiplicator)
        return rabbet2D

    def getAdjustedRabbetAtEdgeCW(self, rabbet2D):
        return self.__getAdjustedRabbetAtEdge(rabbet2D, self.orientToCW, Contour2D.RADIUS_COMPENSATION_LEFT, 1)

    def getAdjustedRabbetAtEdgeCCW(self, rabbet2D):
        return self.__getAdjustedRabbetAtEdge(rabbet2D, self.orientToCCW, Contour2D.RADIUS_COMPENSATION_RIGHT, -1)


class Tests(unittest.TestCase):
    def setUp(self):
        self.rabbetPlusX = self.createMockRabbet(directionVector=Point(1, 0))
        self.rabbetMinusX = self.createMockRabbet(directionVector=Point(-1, 0))
        self.rabbetPlusY = self.createMockRabbet(directionVector=Point(0, 1))
        self.rabbetMinusY = self.createMockRabbet(directionVector=Point(0, -1))
        self.adjuster = RabbetAdjuster(unittest.mock.Mock())
        self.adjuster.orientToCW = unittest.mock.Mock()
        self.adjuster.orientToCCW = unittest.mock.Mock()

    def createMockRabbet(self, directionVector):
        rabbet = Rabbet2D(width=8, directionVector=directionVector)
        rabbet.getSegmentOfMachining = unittest.mock.Mock(return_value=unittest.mock.Mock())
        return rabbet

    def testPlusXCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCW(self.rabbetPlusX)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_LEFT, result.radiusCompensation)
        self.assertEqual(Point(0, -4), result.position)
        self.assertTrue(self.adjuster.orientToCW.called)
        self.assertFalse(self.adjuster.orientToCCW.called)

    def testPlusXCCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCCW(self.rabbetPlusX)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_RIGHT, result.radiusCompensation)
        self.assertEqual(Point(0, 4), result.position)
        self.assertFalse(self.adjuster.orientToCW.called)
        self.assertTrue(self.adjuster.orientToCCW.called)

    def testMinusXCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCW(self.rabbetMinusX)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_LEFT, result.radiusCompensation)
        self.assertEqual(Point(0, 4), result.position)
        self.assertTrue(self.adjuster.orientToCW.called)
        self.assertFalse(self.adjuster.orientToCCW.called)

    def testMinusXCCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCCW(self.rabbetMinusX)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_RIGHT, result.radiusCompensation)
        self.assertEqual(Point(0, -4), result.position)
        self.assertFalse(self.adjuster.orientToCW.called)
        self.assertTrue(self.adjuster.orientToCCW.called)

    def testPlusYCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCW(self.rabbetPlusY)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_LEFT, result.radiusCompensation)
        self.assertEqual(Point(4, 0), result.position)
        self.assertTrue(self.adjuster.orientToCW.called)
        self.assertFalse(self.adjuster.orientToCCW.called)

    def testPlusYCCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCCW(self.rabbetPlusY)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_RIGHT, result.radiusCompensation)
        self.assertEqual(Point(-4, 0), result.position)
        self.assertFalse(self.adjuster.orientToCW.called)
        self.assertTrue(self.adjuster.orientToCCW.called)

    def testMinusYCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCW(self.rabbetMinusY)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_LEFT, result.radiusCompensation)
        self.assertEqual(Point(-4, 0), result.position)
        self.assertTrue(self.adjuster.orientToCW.called)
        self.assertFalse(self.adjuster.orientToCCW.called)

    def testMinusYCCW(self):
        result = self.adjuster.getAdjustedRabbetAtEdgeCCW(self.rabbetMinusY)
        self.assertEqual(Contour2D.RADIUS_COMPENSATION_RIGHT, result.radiusCompensation)
        self.assertEqual(Point(4, 0), result.position)
        self.assertFalse(self.adjuster.orientToCW.called)
        self.assertTrue(self.adjuster.orientToCCW.called)
