# -*- coding: utf-8 -*-

# Copyright 2011-2021, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.polylinesegment import PolylineSegment
from interiorcad.interface.geometry.polylinesegments import PolylineSegments
from interiorcad.interface.geometry.vertex import Vertex


class Shape2D:
    def __init__(self, polyline=None):
        self.polyline = polyline if polyline else Polyline()

    def __bool__(self):
        return bool(self.polyline)

    def isValid(self):
        return self.polyline.isValid()

    def isClosed(self):
        return self.polyline and self.polyline.isClosed()

    def moveBy(self, delta):
        self.polyline.moveBy(delta)

    def rawString(self):
        return 'Shape2D(polyline={})'

    def isRectangular(self):
        return self.polyline.isRectangular()

    def isPolygonal(self):
        return self.polyline.isPolygonal()

    def getSegments(self):
        return self.polyline.getSegments()

    def __eq__(self, other):
        return self.polyline == other.polyline

    def __str__(self):
        return self.rawString().format(str(self.polyline))

    def __repr__(self):
        return self.rawString().format(repr(self.polyline))

    def mirrorOnZeroXAxis(self):
        return self.polyline.mirrorOnZeroXAxis()

    def rotateLeftAtZero(self):
        return self.polyline.rotateLeftAtZero()

    def rotateRightAtZero(self):
        return self.polyline.rotateRightAtZero()

    def rotate180AtZero(self):
        return self.polyline.rotate180AtZero()

    def mirrorOnZeroYAxis(self):
        return self.polyline.mirrorOnZeroYAxis()

    def getVertexIndicesNextTo(self, point):
        return self.polyline.getVertexIndicesNextTo(point)

    def setApproachTo(self, point):
        vertexIndex = self.polyline.getVertexIndexNextTo(point)
        self.polyline.setStartingVertexTo(vertexIndex)

    def changeDirection(self):
        if self.isValid():
            if self.isClosed():
                self.polyline.changeDirection()
            else:
                self.polyline.changeDirectionOfOpenPolygon()

    def __transform(self, transformer, moveByDelta, approachStartPoint):
        if self.isValid():
            transformer()
            self.moveBy(moveByDelta)
            self.setApproachTo(approachStartPoint)

    def mirrorOnXAxis(self, moveByDelta, approachStartPoint):
        self.__transform(self.mirrorOnZeroXAxis, moveByDelta, approachStartPoint)
        self.changeDirection()

    def mirrorOnYAxis(self, moveByDelta, approachStartPoint):
        self.__transform(self.mirrorOnZeroYAxis, moveByDelta, approachStartPoint)
        self.changeDirection()

    def rotateLeft(self, moveByDelta, approachStartPoint):
        self.__transform(self.rotateLeftAtZero, moveByDelta, approachStartPoint)

    def rotateRight(self, moveByDelta, approachStartPoint):
        self.__transform(self.rotateRightAtZero, moveByDelta, approachStartPoint)

    def rotate180(self, moveByDelta, approachStartPoint):
        self.__transform(self.rotate180AtZero, moveByDelta, approachStartPoint)

    def forceVisibility(self):
        if self.polyline:
            self.polyline.forceVisibility()


class Shape2DTest(unittest.TestCase):
    def setUp(self):
        self.shape = Shape2D()

    def testShape2DCreation(self):
        self.assertIsInstance(self.shape, Shape2D)

    def testDefaultShape2DHasNoVertices(self):
        self.assertEqual(0, self.shape.polyline.numVertices())

    def testDefaultShape2DIsNotValid(self):
        self.assertFalse(self.shape.isValid())
        self.assertFalse(self.shape)

    def testShape2DWithOneVertexIsValid(self):
        self.shape.polyline.addVertex('1')
        self.assertTrue(self.shape.isValid())
        self.assertTrue(self.shape)

    def testCanMoveEmptyShape2D(self):
        self.shape.moveBy(Point(10, 20.0))
        self.assertEqual(Shape2D(), self.shape)

    def testShape2DWithRectangularPolylineIsRectangular(self):
        class MockRectangularPolyline:
            def isRectangular(self):
                return True

        self.shape.polyline = MockRectangularPolyline()
        self.assertTrue(self.shape.isRectangular())


class Shape2DAdvancedTest(unittest.TestCase):
    def setUp(self):
        self.shape = Shape2D()
        self.shape.polyline.addVertex(Vertex(Point(0, 0)))
        self.shape.polyline.addVertex(Vertex(Point(0, 50)))
        self.shape.polyline.addVertex(Vertex(Point(100, 50)))
        self.shape.polyline.addVertex(Vertex(Point(100, 0)))

    def testMoveShape2DRightAndTop(self):
        self.shape.moveBy(Point(10.0, 20.0))
        self.assertEqual(Point(10, 20), self.shape.polyline.vertices[0].position)
        self.assertEqual(Point(10, 70), self.shape.polyline.vertices[1].position)
        self.assertEqual(Point(110, 70), self.shape.polyline.vertices[2].position)
        self.assertEqual(Point(110, 20), self.shape.polyline.vertices[3].position)

    def testMoveShape2DLeftAndDown(self):
        self.shape.moveBy(Point(-10, -20))
        self.assertEqual(Point(-10, -20), self.shape.polyline.vertices[0].position)
        self.assertEqual(Point(-10, 30), self.shape.polyline.vertices[1].position)
        self.assertEqual(Point(90, 30), self.shape.polyline.vertices[2].position)
        self.assertEqual(Point(90, -20), self.shape.polyline.vertices[3].position)

    def testGetSegmentsOfOpenShape2DShouldReturnAllSegments(self):
        expected = PolylineSegments()
        expected.append(PolylineSegment(Vertex(Point(0, 0)), Vertex(Point(0, 50))))
        expected.append(PolylineSegment(Vertex(Point(0, 50)), Vertex(Point(100, 50))))
        expected.append(PolylineSegment(Vertex(Point(100, 50)), Vertex(Point(100, 0))))
        expected.append(PolylineSegment(Vertex(Point(100, 0)), Vertex(Point(0, 0))))
        self.assertEqual(expected, self.shape.getSegments())


class Shape2DStringRepresentation(unittest.TestCase):
    def testStringRepresentation(self):
        f = Shape2D('poly')
        self.assertEqual('Shape2D(polyline=poly)', str(f))
        self.assertEqual("Shape2D(polyline='poly')", repr(f))


class IsClosedTests(unittest.TestCase):
    def setUp(self):
        self.shape = Shape2D()

    # The four tests are the same as the corresponding tests of the same names of interiorcad.interface.surface.contour2d.Contour2DTests
    def testIsClosedOnAnOpenPathShouldBeFalse(self):
        self.shape.polyline = Polyline(
            Vertex(Point(100, 100), visible=True),
            Vertex(Point(100, 0), visible=True),
            Vertex(Point(0, 0), visible=False),
        )
        self.assertFalse(self.shape.isClosed())

    def testIsClosedOnANonePathShouldBeFalse(self):
        self.assertFalse(self.shape.isClosed())

    def testIsClosedOnAClosedPathShouldBeTrue(self):
        self.shape.polyline = Polyline(
            Vertex(Point(100, 100), visible=True),
            Vertex(Point(100, 0), visible=True),
            Vertex(Point(0, 0), visible=True),
        )
        self.assertTrue(self.shape.isClosed())

    def testIsClosedOnAPathReturningToStartShouldBeTrue(self):
        self.shape.polyline = Polyline(
            Vertex(Point(100, 100), visible=True),
            Vertex(Point(100, 0), visible=True),
            Vertex(Point(0, 0), visible=True),
            Vertex(Point(100, 100), visible=False),
        )
        self.assertTrue(self.shape.isClosed())


class ClosedPolylineChangeDirectionTests(unittest.TestCase):
    # The tests are ported from interiorcad.interface.geometry.polyline.PolylineChangeDirectionTest
    def setUp(self):
        self.vertex1 = Vertex(Point(1, 11))
        self.vertex2 = Vertex(Point(2, 12))
        self.vertex3 = Vertex(Point(3, 13))

        self.poly = Polyline()
        self.poly.addVertex(self.vertex1)
        self.shape = Shape2D(polyline=self.poly)

    def testOneVertex(self):
        expected = [self.vertex1]
        self.shape.changeDirection()
        self.assertListEqual(expected, self.shape.polyline.vertices)

    def testTwoVertices(self):
        self.poly.addVertex(self.vertex2)
        expected = [self.vertex1, self.vertex2]
        self.shape.changeDirection()
        self.assertListEqual(expected, self.shape.polyline.vertices)

    def testThreeVertices(self):
        self.poly.addVertex(self.vertex2)
        self.poly.addVertex(self.vertex3)
        expected = [self.vertex1, self.vertex3, self.vertex2]
        self.shape.changeDirection()
        self.assertListEqual(expected, self.shape.polyline.vertices)

    def testFourVerticesClosed(self):
        self.poly.addVertex(self.vertex2)
        self.poly.addVertex(self.vertex3)
        self.poly.addVertex(self.vertex1)
        expected = [self.vertex1, self.vertex3, self.vertex2, self.vertex1]
        self.shape.changeDirection()
        self.assertListEqual(expected, self.shape.polyline.vertices)


class OpenPolylineChangeDirectionTests(unittest.TestCase):
    def setUp(self):
        self.shape = Shape2D()

    def test3PointsOpenPolygon(self):
        self.shape.polyline = Polyline(
            Vertex(Point(11, 22)),
            Vertex(Point(111, 222)),
            Vertex(Point(1111, 2222), visible=False),
        )

        self.shape.changeDirection()

        expected = [
            Vertex(Point(1111, 2222)),
            Vertex(Point(111, 222)),
            Vertex(Point(11, 22), visible=False),
        ]

        self.assertEqual(expected, self.shape.polyline.vertices)


class ForceVisibilityTests(unittest.TestCase):
    def setUp(self):
        self.shape = Shape2D()

    def testNoVertex(self):
        self.shape.forceVisibility()
        self.assertEqual(Polyline(), self.shape.polyline)

    def testOneVisibleVertex(self):
        self.shape.polyline = Polyline(
            Vertex(Point(1, 11), visible=True)
        )

        self.shape.forceVisibility()

        expectedPolyline = Polyline(
            Vertex(Point(1, 11), visible=True)
        )

        self.assertEqual(expectedPolyline, self.shape.polyline)

    def testOneInvisibleVertex(self):
        self.shape.polyline = Polyline(
            Vertex(Point(1, 11), visible=False)
        )

        self.shape.forceVisibility()

        expectedPolyline = Polyline(
            Vertex(Point(1, 11), visible=True)
        )

        self.assertEqual(expectedPolyline, self.shape.polyline)

    def testOpenPolygonWith3Vertices(self):
        self.shape.polyline = Polyline(
            Vertex(Point(1, 11), visible=True),
            Vertex(Point(2, 22), visible=True),
            Vertex(Point(3, 33), visible=False),
        )

        self.shape.forceVisibility()

        expectedPolyline = Polyline(
            Vertex(Point(1, 11), visible=True),
            Vertex(Point(2, 22), visible=True),
            Vertex(Point(3, 33), visible=True),
        )

        self.assertEqual(expectedPolyline, self.shape.polyline)
