# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import unittest

from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.vertex import Vertex
from interiorcad.interface.surface.actualdepths import getVerticalMillingActualDepth
from interiorcad.interface.surface.shape2d import Shape2D


class Sizing2D(Shape2D):
    def __init__(self, polyline=None):
        super().__init__(polyline)
        self.needsFlyingApproach = False
        self.depth = None
        self.actualDepth = None
        self.matrix = Matrix()

    def readFrom(self, sizing3D):
        self.polyline = copy.copy(sizing3D.polyline)

    def assignTool(self, surface, machineConfiguration):
        self.depth = surface.thickness
        millThroughDepth = getVerticalMillingActualDepth(machineConfiguration, surface.thickness, self.depth)

        # Careful: the selection of the sizing router is independent of its 'router-rotation-direction' property: right-handed or left-handed. (ID-684)
        # It is also independent of the machine-configuration entry 'sizing-contour-tool-is-right-hand' (Instead NCExportbase.isSizingMillingDirectionLeft will take care of it)
        self.tool = None
        if self.isRectangular():
            self.tool = machineConfiguration.toolDatabase.getBestToolForSizing(self.depth, millThroughDepth, jointingRouter=True)
        if not self.tool:
            self.tool = machineConfiguration.toolDatabase.getBestToolForSizing(self.depth, millThroughDepth, jointingRouter=False)

        if self.tool and self.tool.isRouterCentered():
            self.actualDepth = surface.thickness / 2
        else:
            # TODO IT-265: support NCExportBase.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS -- NG
            self.actualDepth = millThroughDepth

        self.toolID = self.tool.getToolID() if self.tool else None

    def assignFallbackTool(self, _surface, machineConfiguration):
        self.tool = machineConfiguration.toolDatabase.getFallbackToolForSizing()
        self.toolID = self.tool.getToolID() if self.tool else None

    def insertVertexAfter(self, index, vertex):
        return self.polyline.insertVertexAfter(index, vertex)

    def getVertex(self, index):
        return self.polyline.getVertex(index)

    def __eq__(self, other):
        return self.polyline == other.polyline and self.depth == self.depth

    def rawString(self):
        return 'Sizing2D(polyline={})'

    def __str__(self):
        return self.rawString().format(str(self.polyline))

    def __repr__(self):
        return self.rawString().format(repr(self.polyline))

    def isVertical(self):
        return True  # A sizing is always vertical


class Sizing2DTests(unittest.TestCase):
    def setUp(self):
        self.sizing = Sizing2D()

    def testAnEmptySizingIsInvalidAndFalse(self):
        self.assertFalse(self.sizing.isValid())
        self.assertFalse(self.sizing)

    def aNonEmptySizingIsValidAndTrue(self):
        polyline = Polyline(
            Vertex(Point(0.0, 0.0)),
            Vertex(Point(100.0, 100.0)),
            Vertex(Point(100.0, 0.0))
        )
        self.sizing = Sizing2D(polyline)
        self.assertTrue(self.sizing.isValid())
        self.assertTrue(self.sizing)

    def testCanSetPolyline(self):
        self.sizing.polyline = 'new polyline'
        self.assertEqual('new polyline', self.sizing.polyline)

    def testNoPolylineMeansInvalidSizing(self):
        self.assertFalse(self.sizing.isValid())
        self.sizing.polyline.addVertex('vertex')
        self.assertTrue(self.sizing.isValid())

    def testSetApproachToValidPointOnSizing(self):
        polyline = Polyline(Vertex(Point(0.0, 0.0)),
                            Vertex(Point(100.0, 100.0)),
                            Vertex(Point(100.0, 0.0)))

        expectedPolyline = Polyline(Vertex(Point(100.0, 100.0)),
                                    Vertex(Point(100.0, 0.0)),
                                    Vertex(Point(0.0, 0.0)))

        sizing = Sizing2D(polyline)
        sizing.setApproachTo(Point(105, 105))
        self.assertEqual(expectedPolyline, sizing.polyline)

    def testSizingsAreEqual(self):
        self.assertEqual(self.sizing, Sizing2D())

    def testChangeDirection(self):
        self.sizing.polyline.addVertex(Vertex(Point(0, 0)))
        self.sizing.polyline.addVertex(Vertex(Point(50, 0)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 0)))
        self.sizing.changeDirection()
        self.assertEqual(Point(0, 0), self.sizing.polyline.vertices[0].position)
        self.assertEqual(Point(100, 0), self.sizing.polyline.vertices[1].position)
        self.assertEqual(Point(50, 0), self.sizing.polyline.vertices[2].position)


class XAXisMirrorTests(unittest.TestCase):
    def setUp(self):
        self.sizing = Sizing2D()

    def testMirrorOnZeroXAxisEmptySizing(self):
        self.sizing.mirrorOnZeroXAxis()
        self.assertEqual(Sizing2D(), self.sizing)

    def testMirrorOnZeroXAxisSizing(self):
        self.sizing.polyline.addVertex(Vertex(Point(0, 0)))
        self.sizing.polyline.addVertex(Vertex(Point(0, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 0)))
        self.sizing.mirrorOnZeroXAxis()
        self.assertEqual(Point(0, 0), self.sizing.polyline.vertices[0].position)
        self.assertEqual(Point(0, -50), self.sizing.polyline.vertices[1].position)
        self.assertEqual(Point(100, -50), self.sizing.polyline.vertices[2].position)
        self.assertEqual(Point(100, 0), self.sizing.polyline.vertices[3].position)

    def testMirrorOnXAxisEmptySizing(self):
        self.sizing.mirrorOnXAxis(Point(0, -100), None)
        self.assertEqual(Sizing2D(), self.sizing)

    def testMirrorOnXAxisSizing(self):
        self.sizing.polyline.addVertex(Vertex(Point(0, 0)))
        self.sizing.polyline.addVertex(Vertex(Point(0, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 0)))
        self.sizing.mirrorOnXAxis(Point(0, 50), Point(0, 0))
        self.assertEqual(Point(0, 0), self.sizing.polyline.vertices[0].position)
        self.assertEqual(Point(0, 50), self.sizing.polyline.vertices[1].position)
        self.assertEqual(Point(100, 50), self.sizing.polyline.vertices[2].position)
        self.assertEqual(Point(100, 0), self.sizing.polyline.vertices[3].position)


class YAXisMirrorTests(unittest.TestCase):
    def setUp(self):
        self.sizing = Sizing2D()

    def testMirrorOnZeroYAxisEmptySizing(self):
        self.sizing.mirrorOnZeroYAxis()
        self.assertEqual(Sizing2D(), self.sizing)

    def testMirrorOnZeroYAxisSizing(self):
        self.sizing.polyline.addVertex(Vertex(Point(0, 0)))
        self.sizing.polyline.addVertex(Vertex(Point(0, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 0)))
        self.sizing.mirrorOnZeroYAxis()
        self.assertEqual(Point(0, 0), self.sizing.polyline.vertices[0].position)
        self.assertEqual(Point(0, 50), self.sizing.polyline.vertices[1].position)
        self.assertEqual(Point(-100, 50), self.sizing.polyline.vertices[2].position)
        self.assertEqual(Point(-100, 0), self.sizing.polyline.vertices[3].position)

    def testMirrorOnYAxisEmptySizing(self):
        self.sizing.mirrorOnYAxis(Point(-100, 0), None)
        self.assertEqual(Sizing2D(), self.sizing)

    def testMirrorOnYAxisSizing(self):
        self.sizing.polyline.addVertex(Vertex(Point(0, 0)))
        self.sizing.polyline.addVertex(Vertex(Point(0, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 50)))
        self.sizing.polyline.addVertex(Vertex(Point(100, 0)))
        self.sizing.mirrorOnYAxis(Point(100, 0), Point(0, 0))
        self.assertEqual(Point(0, 0), self.sizing.polyline.vertices[0].position)
        self.assertEqual(Point(0, 50), self.sizing.polyline.vertices[1].position)
        self.assertEqual(Point(100, 50), self.sizing.polyline.vertices[2].position)
        self.assertEqual(Point(100, 0), self.sizing.polyline.vertices[3].position)
