# -*- coding: utf-8 -*-

# Copyright 2011-2020, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import unittest

from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.polyline import Polyline
from interiorcad.interface.geometry.radiusvertex import RadiusVertex
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.surface import Surface
from interiorcad.interface.surface.sizing import Sizing2D


class SurfaceApproach:
    def __init__(self, surface, searchMaxY=True):
        self.surface = surface
        self.discardedSizings = []
        self.searchMaxY = searchMaxY

    def setApproach(self, minimumSegmentLength):
        self.setApproachAtVertex(self.__getApproachVertex(minimumSegmentLength))

    def __getApproachVertex(self, minimumSegmentLength):
        return self.__getApproachForCommonShapes(minimumSegmentLength)

    def setApproachAtVertex(self, approachVertex):
        if approachVertex is None:
            self.discardedSizings.append(self.surface.sizing)
            self.surface.sizing = Sizing2D(None)
        elif self.surface.hasSizing():
            self.surface.sizing.setApproachTo(approachVertex.position)
            self.surface.sizing.forceVisibility()

    def getSegmentOfExtremeYValue(self, approach):
        allSegments = approach.surface.sizing.getSegments()
        reducedSegments = allSegments.filterSegmentsWithCornerVertices()
        boundSegments = reducedSegments.filterSegmentsOnBounds(approach.surface.sizingBounds)
        return approach.__filterSegmentByExtremeYValue(boundSegments)

    def __getApproachForCommonShapes(self, minimumSegmentLength):
        approachVertex = self.__approachOnStraightSegmentOnBounds()
        if approachVertex is None:
            approachVertex = self.__approachOnPointOnBounds()
        if approachVertex is None:
            approachVertex = self.__approachOnStraightSegment(minimumSegmentLength)
        return approachVertex

    def __approachOnStraightSegmentOnBounds(self):
        segment = self.__filterSegmentOnBoundsWithExtremeYValue()
        approachVertex = self.__setApproachToMidOfSegment(segment)
        return approachVertex

    def __approachOnStraightSegment(self, minimumSegmentLength):
        allSegments = self.surface.sizing.getSegments()
        reducedSegments = allSegments.filterSegmentsWithCornerVertices()
        reducedSegments = reducedSegments.filterSegmentsGreaterThan(minimumSegmentLength)
        segment = self.__filterSegmentByExtremeYValue(reducedSegments)
        approachVertex = self.__setApproachToMidOfSegment(segment)
        if approachVertex is not None:
            self.surface.sizing.needsFlyingApproach = True
        return approachVertex

    def __filterSegmentOnBoundsWithExtremeYValue(self):
        allSegments = self.surface.sizing.getSegments()
        segmentsOnBounds = allSegments.filterSegmentsOnBounds(self.surface.sizingBounds)
        segment = self.__filterSegmentByExtremeYValue(segmentsOnBounds)
        return segment

    def __filterSegmentByExtremeYValue(self, reducedSegments):
        if self.searchMaxY:
            return reducedSegments.filterSegmentWithHighestYValue()
        else:
            return reducedSegments.filterSegmentWithLowestYValue()

    def __setApproachToMidOfSegment(self, segment):
        approachVertex = None
        if segment is not None:
            approachVertex = segment.calculateMidPoint()
            if approachVertex:
                topLeftVertexIndex = self.surface.sizing.polyline.getVertexIndexNextTo(segment.start.position)
                if topLeftVertexIndex is not None:
                    self.surface.sizing.insertVertexAfter(topLeftVertexIndex, approachVertex)
        return approachVertex

    def __approachOnPointOnBounds(self):
        approachVertex = None
        vertexIndices = self.surface.sizing.getVertexIndicesNextTo(self.surface.sizingBounds.topRight())
        for vertexIndex in vertexIndices:
            vertex = self.surface.sizing.getVertex(vertexIndex)
            if self.__isValidApproachVertex(vertex):
                approachVertex = vertex
                break
        return approachVertex

    def __isValidApproachVertex(self, vertex):
        bounds = self.surface.sizingBounds
        return bounds.isPointOnRectangle(vertex.position)


class SurfaceApproachTest(unittest.TestCase):
    def setUp(self):
        self.surface = Surface()

    def rectangularSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(100, 100)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(0, 0)),
        )))

    def expectedRectangularSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(50, 100)),
            CornerVertex(Point(100, 100)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(50, 100)),
        )))

    def polygonalSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(100, 100)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(50, 50)),
            CornerVertex(Point(0, 0)),
        )))

    def expectedPolygonalSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(50, 100)),
            CornerVertex(Point(100, 100)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(50, 50)),
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(50, 100)),
        )))

    def edgedPolygonalSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(50, 100)),
            CornerVertex(Point(70, 70)),
            CornerVertex(Point(100, 60)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(0, 0)),
        )))

    def expectededgedPolygonalSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(25, 100)),
            CornerVertex(Point(50, 100)),
            CornerVertex(Point(70, 70)),
            CornerVertex(Point(100, 60)),
            CornerVertex(Point(100, 0)),
            CornerVertex(Point(0, 0)),
            CornerVertex(Point(0, 100)),
            CornerVertex(Point(25, 100)),
        )))

    def polylineRoundedOnEachEdgeSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(200.0, 100.0)),
            RadiusVertex(Point(29.3, 29.3), radius=100.0),
            CornerVertex(Point(100.0, 200.0)),
            CornerVertex(Point(100.0, 500.0)),
            RadiusVertex(Point(29.3, 670.7), radius=100.0),
            CornerVertex(Point(200.0, 600.0)),
            CornerVertex(Point(800.0, 600.0)),
            RadiusVertex(Point(970.7, 670.7), radius=100.0),
            CornerVertex(Point(900.0, 500.0)),
            CornerVertex(Point(900.0, 200.0)),
            RadiusVertex(Point(970.7, 29.3), radius=100.0),
            CornerVertex(Point(200.0, 100.0)),
        )))

    def expectedPolylineRoundedOnEachEdgeSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(500.0, 600.0)),
            CornerVertex(Point(800.0, 600.0)),
            RadiusVertex(Point(970.7, 670.7), radius=100.0),
            CornerVertex(Point(900.0, 500.0)),
            CornerVertex(Point(900.0, 200.0)),
            RadiusVertex(Point(970.7, 29.3), radius=100.0),
            CornerVertex(Point(200.0, 100.0)),
            RadiusVertex(Point(29.3, 29.3), radius=100.0),
            CornerVertex(Point(100.0, 200.0)),
            CornerVertex(Point(100.0, 500.0)),
            RadiusVertex(Point(29.3, 670.7), radius=100.0),
            CornerVertex(Point(200.0, 600.0)),
            CornerVertex(Point(500.0, 600.0)),

        )))

    def circularSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(100.0, 0.0)),
            RadiusVertex(Point(29.3, 29.3), radius=100.0),
            CornerVertex(Point(0.0, 100.0)),
            RadiusVertex(Point(29.3, 170.7), radius=100.0),
            CornerVertex(Point(100.0, 200.0)),
            RadiusVertex(Point(170.7, 170.7), radius=100.0),
            CornerVertex(Point(200.0, 100.0)),
            RadiusVertex(Point(170.7, 29.3), radius=100.0),
            CornerVertex(Point(100.0, 0.0)),
        )))

    def circularSizingNotOnBounds(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(120.0, 20.0)),
            RadiusVertex(Point(49.3, 49.3), radius=100.0),
            CornerVertex(Point(20.0, 120.0)),
            RadiusVertex(Point(49.3, 190.7), radius=100.0),
            CornerVertex(Point(120.0, 220.0)),
            RadiusVertex(Point(190.7, 190.7), radius=100.0),
            CornerVertex(Point(220.0, 120.0)),
            RadiusVertex(Point(190.7, 49.3), radius=100.0),
            CornerVertex(Point(120.0, 20.0)),
        )))

    def expectedCircularSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(100.0, 200.0)),
            RadiusVertex(Point(170.7, 170.7), radius=100.0),
            CornerVertex(Point(200.0, 100.0)),
            RadiusVertex(Point(170.7, 29.3), radius=100.0),
            CornerVertex(Point(100.0, 0.0)),
            RadiusVertex(Point(29.3, 29.3), radius=100.0),
            CornerVertex(Point(0.0, 100.0)),
            RadiusVertex(Point(29.3, 170.7), radius=100.0),
            CornerVertex(Point(100.0, 200.0)),
        )))

    def polylineRoundedSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(0.0, 0.0)),
            CornerVertex(Point(0.0, 100.0)),
            CornerVertex(Point(100.0, 100.0)),
            RadiusVertex(Point(156.0, 87.0), radius=120.0),
            CornerVertex(Point(200.0, 50.0)),
            CornerVertex(Point(200.0, 0.0)),
            CornerVertex(Point(0.0, 0.0)),
        )))

    def expectedPolylineRoundedSizing(self):
        return Sizing2D(Polyline((
            CornerVertex(Point(50.0, 100.0)),
            CornerVertex(Point(100.0, 100.0)),
            RadiusVertex(Point(156.0, 87.0), radius=120.0),
            CornerVertex(Point(200.0, 50.0)),
            CornerVertex(Point(200.0, 0.0)),
            CornerVertex(Point(0.0, 0.0)),
            CornerVertex(Point(0.0, 100.0)),
            CornerVertex(Point(50.0, 100.0)),
        )))

    def testRectangularSurfaceSetsApproachToMidOfTopSegment(self):
        self.surface.sizingBounds = Rectangle(0, 0, 100, 100)
        self.surface.sizing = self.rectangularSizing()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(self.expectedRectangularSizing(), self.surface.sizing)

    def testPolygonalSurfaceSetsApproachToMidOfTopSegment(self):
        self.surface.sizingBounds = Rectangle(0, 0, 100.0, 100.0)
        self.surface.sizing = self.polygonalSizing()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(self.expectedPolygonalSizing(), self.surface.sizing)

    def testPolygonalSurfaceSetsApproachToRightOnBounds(self):
        self.surface.sizingBounds = Rectangle(0, 0, 100.0, 100.0)
        self.surface.sizing = self.edgedPolygonalSizing()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(self.expectededgedPolygonalSizing(), self.surface.sizing)

    def testPolylineWithoutValidApproachIsDiscarded(self):
        self.surface.sizing = self.circularSizingNotOnBounds()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(Sizing2D(), self.surface.sizing)

    def testPolylineWithoutValidApproachShouldBeLogged(self):
        self.surface.sizing = self.circularSizingNotOnBounds()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(1, len(surfaceApproach.discardedSizings))

    def testPolylineWithoutStraightSegmentsIsApproachedFlying(self):
        self.surface.sizing = self.polylineRoundedOnEachEdgeSizing()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        expectedSizing = self.expectedPolylineRoundedOnEachEdgeSizing()
        self.assertEqual(expectedSizing, self.surface.sizing)
        self.assertTrue(self.surface.sizing.needsFlyingApproach)

    def testCircularPolylineShouldBeApproachedCorrectly(self):
        self.surface.sizingBounds = Rectangle(0, 0, 200.0, 200.0)
        self.surface.sizing = self.circularSizing()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(self.expectedCircularSizing(), self.surface.sizing)

    def testPolylineWithRoundedEdgeShouldHaveApproachInMidOfTopMostSegment(self):
        self.surface.sizingBounds = Rectangle(0, 0, 200.0, 100.0)
        self.surface.sizing = self.polylineRoundedSizing()
        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(self.expectedPolylineRoundedSizing(), self.surface.sizing)

    def testComplexRealLifeSizing(self):  # IC-15550: Bauteil-4449 (the Vertices are from the path in the OIP, not from the Datatracker)
        self.surface.sizingBounds = Rectangle(0, -403.7, 2536, 578.9)
        self.surface.sizing = Sizing2D(Polyline((
            CornerVertex(Point(0.0, 0.0)),
            CornerVertex(Point(371.8, 578.9)),
            RadiusVertex(Point(1257.6, 318.1), radius=1638),
            CornerVertex(Point(2143.7, 577.8)),
            CornerVertex(Point(2536, -33.1)),
            RadiusVertex(Point(1991.4, -291.5), radius=2364),
            CornerVertex(Point(1399.1, -403.7)),
            CornerVertex(Point(1396.8, 365.8)),
            RadiusVertex(Point(669.3, -294), radius=2326),
            CornerVertex(Point(0.0, 0.0)),
        )))

        expectedSizing = Sizing2D(Polyline((
            CornerVertex(Point(2536, -33.1)),
            RadiusVertex(Point(1991.4, -291.5), radius=2364.00),
            CornerVertex(Point(1399.1, -403.7)),
            CornerVertex(Point(1396.8, 365.8)),
            RadiusVertex(Point(669.3, -294), radius=2326.00),
            CornerVertex(Point(0.0, 0.0)),
            CornerVertex(Point(371.8, 578.9)),
            RadiusVertex(Point(1257.6, 318.1), radius=1638.00),
            CornerVertex(Point(2143.7, 577.8)),
            CornerVertex(Point(2536, -33.1)),
        )))

        surfaceApproach = SurfaceApproach(self.surface)
        surfaceApproach.setApproach(50.0)
        self.assertEqual(expectedSizing, self.surface.sizing)
