# -*- coding: utf-8 -*-

# Copyright 2011-2022, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import math
import unittest


class Formatter:
    def __init__(self, coordinatePrecision=1, anglePrecision=4, decimalSeparator='.', significantDecimalsOnly=True):
        self.coordinatePrecision = coordinatePrecision
        self.anglePrecision = anglePrecision
        self.decimalSeparator = decimalSeparator
        self.significantDecimalsOnly = significantDecimalsOnly
        self.areaPrecision = coordinatePrecision
        self.volumePrecision = anglePrecision

    def __keepSignificantDecimalsOnly(self, data):
        if self.significantDecimalsOnly and self.decimalSeparator in data:
            data = data.rstrip('0')
            if data.endswith(self.decimalSeparator):
                data = data[0:-len(self.decimalSeparator)]
        return data

    def __formatAsFloat(self, data, precision):
        result = '{data:.{precision}f}'.format(data=data, precision=precision)
        if self.decimalSeparator != '.':
            result = result.replace('.', self.decimalSeparator)
        return self.__keepSignificantDecimalsOnly(result)

    def __formatAsSignedFloat(self, data, precision):
        result = '{data:+.{precision}f}'.format(data=data, precision=precision)
        if self.decimalSeparator != '.':
            result = result.replace('.', self.decimalSeparator)
        return self.__keepSignificantDecimalsOnly(result)

    def formatAsString(self, data):
        return str(data or "")

    def formatAsInt(self, data):
        return str(int(data or 0))

    def formatAsCoordinate(self, data):
        return self.__formatAsFloat(data=data or 0.0, precision=self.coordinatePrecision)

    def formatAsSignedCoordinate(self, data):
        return self.__formatAsSignedFloat(data=data or 0.0, precision=self.coordinatePrecision)

    def formatAsAngle(self, data):
        return self.__formatAsFloat(data=data or 0.0, precision=self.anglePrecision)

    def formatAsArea(self, data):
        return self.__formatAsFloat(data=data or 0.0, precision=self.areaPrecision)

    def formatAsVolume(self, data):
        return self.__formatAsFloat(data=data or 0.0, precision=self.volumePrecision)


class CuttinglistExportBaseTests(unittest.TestCase):
    def setUp(self):
        self.formatter = Formatter()

    def testFormatAsString(self):
        self.assertEqual('abcd', self.formatter.formatAsString(data='abcd'))

    def testFormatAsInt(self):
        self.assertEqual('3', self.formatter.formatAsInt(data=3))

    def testFormatAsCoordinateWithDefaultPrecision(self):
        self.assertEqual('3.1', self.formatter.formatAsCoordinate(data=3.111))

    def testNegativeFormatAsCoordinateWithDefaultPrecision(self):
        self.assertEqual('-3.1', self.formatter.formatAsCoordinate(data=-3.111))

    def testFormatAsSignedCoordinateWithDefaultPrecision(self):
        self.assertEqual('+3.1', self.formatter.formatAsSignedCoordinate(data=3.111))

    def testNegativeFormatAsSignedCoordinateWithDefaultPrecision(self):
        self.assertEqual('-3.1', self.formatter.formatAsSignedCoordinate(data=-3.111))

    def testFormatAsAngleWithDefaultPrecision(self):
        self.assertEqual('3.111', self.formatter.formatAsAngle(data=3.111))

    def testFormatAsAreaWithDefaultPrecision(self):
        self.assertEqual('3.1', self.formatter.formatAsArea(data=3.111))

    def testFormatAsVolumeWithDefaultPrecision(self):
        self.assertEqual('3.1416', self.formatter.formatAsVolume(data=math.pi))

    def testFormatAsCoordinateWithCommaAsDecimalCharacter(self):
        self.formatter.decimalSeparator = ','
        self.assertEqual('3,1', self.formatter.formatAsCoordinate(data=3.1))

    def testFormatAsAngleWithDifferentDecimalCharacter(self):
        self.formatter.decimalSeparator = 'A'
        self.assertEqual('3A1', self.formatter.formatAsAngle(data=3.1))


class CuttinglistExportBaseWithCustomPrecisions(unittest.TestCase):
    def setUp(self):
        self.formatter = Formatter(coordinatePrecision=5, anglePrecision=3)

    def testAsCoordinateWithDefaultPrecision(self):
        self.assertEqual('3.11111', self.formatter.formatAsCoordinate(data=3.111111))

    def testAsAngleWithDefaultPrecision(self):
        self.assertEqual('3.111', self.formatter.formatAsAngle(data=3.111111))

    def testRemoveTrailingDecimalZeroByDefaultForCoordiantes(self):
        self.assertEqual('3.11', self.formatter.formatAsCoordinate(data=3.11))

    def testRemoveTrailingDecimalZeroByDefaultForAngles(self):
        self.assertEqual('3.11', self.formatter.formatAsAngle(data=3.11))

    def testRelevantZerosShouldNotBeRemovedForCoordinates(self):
        self.assertEqual('30', self.formatter.formatAsCoordinate(data=30))

    def testRelevantZerosShouldNotBeRemovedForAngles(self):
        self.assertEqual('30', self.formatter.formatAsAngle(data=30))


class TestsWithTrailingDecimalZeros(unittest.TestCase):
    def setUp(self):
        self.formatter = Formatter(coordinatePrecision=5, anglePrecision=3, significantDecimalsOnly=False)

    def testAsCoordinateWithDefaultPrecision(self):
        self.assertEqual('3.11111', self.formatter.formatAsCoordinate(data=3.111111))

    def testAsAngleWithDefaultPrecision(self):
        self.assertEqual('3.111', self.formatter.formatAsAngle(data=3.111111))

    def testRemoveTrailingDecimalZeroByDefaultForCoordiantes(self):
        self.assertEqual('3.11000', self.formatter.formatAsCoordinate(data=3.11))

    def testRemoveTrailingDecimalZeroByDefaultForAngles(self):
        self.assertEqual('3.110', self.formatter.formatAsAngle(data=3.11))

    def testRelevantZerosShouldNotBeRemovedForCoordinates(self):
        self.assertEqual('30.00000', self.formatter.formatAsCoordinate(data=30))

    def testRelevantZerosShouldNotBeRemovedForAngles(self):
        self.assertEqual('30.000', self.formatter.formatAsAngle(data=30))


class TestsWithoutDecimals(unittest.TestCase):
    def setUp(self):
        self.formatter = Formatter(coordinatePrecision=0, anglePrecision=0, significantDecimalsOnly=True)

    def testAsCoordinateShouldNotRemoveRelevantTrailingZeros(self):  # IC-22965
        self.assertEqual('30', self.formatter.formatAsCoordinate(data=30.0))

    def testAsAngleShouldNotRemoveRelevantTrailingZeros(self):
        self.assertEqual('30', self.formatter.formatAsAngle(data=30.0))
