# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import datetime
import itertools
import logging
import math
import operator
import os
import runpy
from collections import OrderedDict, namedtuple

import interiorcad.backend.exportdatasafe
import interiorcad.backend.openservices
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.cmoduleadapters.packagemanager import getLicenseInfo
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.dosfilenamecreator import DosFilenameCreator
from interiorcad.backend.exportmainpath import getNCExportInAuxiliaryFilesPath
from interiorcad.backend.filenameandpathfilter import FilenameAndPathFilter
from interiorcad.backend.localization import getLocalizedString, NCEXPORT_NCPROGRAMS_RESOURCEGROUP, NCEXPORT_MESSAGES_RESOURCEGROUP, FILE_NAME_POSTFIXES_RESOURCEGROUP, CUTTINGLIST_RESOURCEGROUP, SIDE_NAMES_RESOURCEGROUP, NCEXPORTBASE_MESSAGES, COMMON_MESSAGES_RESOURCEGROUP
from interiorcad.backend.ncexportmigrated.dadotoncsegments import getDadoContourCW, getDadoContourCCW
from interiorcad.backend.ncexportmigrated.nclineandradiussegment import NCLineAndRadiusSegment
from interiorcad.backend.ncexportmigrated.rabbettoncsegments import RabbetToNCSegments
from interiorcad.backend.postprocessors import PostProcessors
from interiorcad.backend.postprocessors.piercingverticalmachiningssidechanger import PiercingVerticalMachiningsSideChanger
from interiorcad.backend.postprocessors.surfacepostprocessor.chamferwarner import ChamferWarner
from interiorcad.backend.postprocessors.surfacepostprocessor.clamexcangleprocessor import CLAMEX_C_ANGLE, ClamexCAngleProcessor, CLAMEX_C_ANGLE_180
from interiorcad.backend.postprocessors.surfacepostprocessor.contourapproachfilter import ContourApproachFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.contourexportfilter import ContourExportFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.contourtoolfilter import ContourToolFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.dadossorter import DadosSorter
from interiorcad.backend.postprocessors.surfacepostprocessor.drillingssorter import DrillingsSorter
from interiorcad.backend.postprocessors.surfacepostprocessor.licensefilter import LicenseFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.nonverticaldrillingsfilter import NonVerticalDrillingsFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.pocketfilter import PocketFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.profilewarner import ProfileWarner
from interiorcad.backend.postprocessors.surfacepostprocessor.rabbetssorter import RabbetsSorter
from interiorcad.backend.postprocessors.surfacepostprocessor.rabbettoolfilter import RabbetToolFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.sizingapproachprocessor import SizingApproachProcessor
from interiorcad.backend.postprocessors.surfacepostprocessor.sizingcontourprocessor import SizingContourProcessor
from interiorcad.backend.postprocessors.surfacepostprocessor.sizingthreesidecontourprocessor import SizingThreeSideContourProcessor
from interiorcad.backend.postprocessors.surfacepostprocessor.sizingtoolfilter import SizingToolFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.toolassigner import ToolAssigner
from interiorcad.backend.postprocessors.surfacepostprocessor.toolexportmachiningsfilter import ToolExportMachiningsFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.verticalcenterpunchfilter import VerticalCenterPunchFilter, HorizontalCenterPunchFilter, FiveAxisCenterPunchFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.verticaldrillingsfilter import VerticalDrillingsFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.verticalgriddrillingsfilter import VerticalGridDrillingsFilter
from interiorcad.backend.postprocessors.surfacepostprocessor.verticalgriddrillingsreplacer import VerticalGridDrillingsReplacer
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.backend.workpiecefilters import WorkpieceFilters
from interiorcad.backend.workpiecefilters.emptyworkpiecefilter import EmptyWorkpieceFilter
from interiorcad.backend.workpiecefilters.invalidpathworkpiecefilter import InvalidPathWorkpieceFilter
from interiorcad.backend.workpiecefilters.nullmachiningfilter import NullMachiningFilter
from interiorcad.backend.workpiecefilters.smallworkpiecefilter import SmallWorkpieceFilter
from interiorcad.backend.workpiecefilters.workpiecemachiningsfilter import WorkpieceMachiningsFilter
from interiorcad.backend.zvalue.configuration import ZValueConverterConfiguration
from interiorcad.backend.zvalue.converter import ZValueConverter
from interiorcad.interface.configuration import NCGlobalExportConfiguration
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.machineconfiguration import MachineConfiguration
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.machiningsexporter.contoursexporter import ContoursExporter
from interiorcad.interface.machiningsexporter.horizontaldrillingsexporter import HorizontalMachiningsExporter
from interiorcad.interface.machiningsexporter.horizontalmachiningcreator import HorizontalDrillingsAndCenterPunchesCreator
from interiorcad.interface.ncexport.dadocreator import XAxisDadoCreator, YAxisDadoCreator, CRoratedDadoCreator
from interiorcad.interface.ncexport.generalexporter import GeneralExporter
from interiorcad.interface.surface import Sizing2D, Miter2D, Rabbet2D, GridDrilling2D, SURFACE_SIDE_TYPE_MAIN, SURFACE_SIDE_TYPE_OPPOSITE, SURFACE_SIDE_TYPE_SIZING, SURFACE_SIDE_TYPE_BOTH_SIDED
from interiorcad.interface.surface.centerpunch2D import CenterPunch2D, VerticalCenterPunch, HorizontalCenterPunch, FiveAxisCenterPunch
from interiorcad.interface.surface.contour2d import Contour2D, VerticalContour2D, HorizontalContour2D
from interiorcad.interface.surface.dado2d import VerticalDado2D, HorizontalDado2D
from interiorcad.interface.surface.drilling2d import Drilling2D, VerticalDrilling, HorizontalDrilling, FiveAxisDrilling
from interiorcad.interface.surface.machiningsorter import MachiningSorter
from interiorcad.interface.surface.ncmacro2d import Parameter, NCMacro2D
from interiorcad.interface.surface.origins import SURFACE_ORIGIN_FRONT_LEFT, isStandardOrigin, isOriginFront
from interiorcad.interface.surface.pocket2d import VerticalPocket2D, HorizontalPocket2D
from interiorcad.interface.surface.rabbetadjuster import RabbetAdjuster
from interiorcad.interface.utilities.formatter import Formatter


# noinspection PyUnusedLocal
class NCExportBase:
    OUTPUT_END_OF_LINE = '\r\n'
    OUTPUT_ENCODING = 'windows-1252'
    OUTPUT_ENCODING_ERROR_MODE = 'ignore'

    CARRIAGE_RETURN_REPLACEMENT = ' '

    def __init__(self, **keywords):
        self.configuration = keywords.get('configuration', NCGlobalExportConfiguration())
        self.machineconfiguration = keywords.get('machineConfiguration', MachineConfiguration())
        self.logFileName = None
        self.loggingHandler = None
        self.generateNCStopBeforeNextContour = None
        self.generateNCStopBeforeNextHorizontalDrilling = None

        if keywords.get('activateLogging', True):
            self.initLogging(**keywords)

        self.initializeMachineCreators(**keywords)
        self.useMachineConfiguration(**keywords)
        self.formatter = Formatter(
            coordinatePrecision=self.getCoordinatePrecision(),
            anglePrecision=self.getAnglePrecision(),
            decimalSeparator=self.getDecimalSeparator(**keywords),
            significantDecimalsOnly=self.shouldKeepSignificantDecimalsOnly()
        )

    def initializeMachineCreators(self, **keywords):
        # This list does NOT contain any of the new horizontal machinings, like contours, dados and pockets
        self.LEGACY_MACHINING_CREATORS = OrderedDict((
            ('sizing', (self.createSizing, None, None)),
            ('contours', (self.createVerticalContours, None, None)),
            ('miters', (self.createVerticalMiters, None, None)),
            ('dados', (self.createVerticalDados, None, None)),
            ('rabbets', (self.createVerticalRabbets, None, None)),
            ('pockets', (self.createVerticalPockets, None, None)),
            ('griddrillings', (self.createVerticalGridDrillings, None, None)),
            ('drillingsandcenterpunches', (self.createVerticalDrillingsAndCenterPunches, self.createHorizontalDrillingsAndCenterPunches, self.createFiveAxisDrillingsAndCenterPunches)),
            ('ncmacros', (self.createNCMacros, None, None)),  # NC Macro is considered as vertical only, when using this structure-- NG
        ))
        # 'Sizing Contours Miters Dados Rabbets Pockets GridDrillings DrillingsAndCenterPunches NCMacros'   # new default
        # 'GridDrillings DrillingsAndCenterPunches Pockets Sizing Rabbets Dados Miters Contours NCMacros'   # old order

        CreatorClass = namedtuple("CreatorClass", "machiningClass sortingName sortingFace creator")
        self.SORTING_FACE_VERTICAL, self.SORTING_FACE_HORIZONTAL, self.SORTING_FACE_FIVE_AXIS = range(3)

        self.SINGLE_MACHINING_CREATORS = (
            CreatorClass(Sizing2D, 'sizing', self.SORTING_FACE_VERTICAL, self.__createSizingHack),
            CreatorClass(VerticalContour2D, 'contours', self.SORTING_FACE_VERTICAL, self.__createVerticalContourHack),
            CreatorClass(Miter2D, 'miters', self.SORTING_FACE_VERTICAL, self.createVerticalMiterBase),
            CreatorClass(VerticalDado2D, 'dados', self.SORTING_FACE_VERTICAL, self.createVerticalDadoBase),
            CreatorClass(Rabbet2D, 'rabbets', self.SORTING_FACE_VERTICAL, self.createVerticalRabbetBase),
            CreatorClass(VerticalPocket2D, 'pockets', self.SORTING_FACE_VERTICAL, self.createVerticalPocketBase),
            CreatorClass(GridDrilling2D, 'griddrillings', self.SORTING_FACE_VERTICAL, self.createVerticalGridDrillingBase),
            CreatorClass(VerticalDrilling, 'drillingsandcenterpunches', self.SORTING_FACE_VERTICAL, self.createVerticalDrillingAndCenterPunchBase),
            CreatorClass(VerticalCenterPunch, 'drillingsandcenterpunches', self.SORTING_FACE_VERTICAL, self.createVerticalDrillingAndCenterPunchBase),

            CreatorClass(HorizontalDrilling, 'drillingsandcenterpunches', self.SORTING_FACE_HORIZONTAL, self.__createHorizontalDrillingAndCenterPunchBaseHack),
            CreatorClass(HorizontalCenterPunch, 'drillingsandcenterpunches', self.SORTING_FACE_HORIZONTAL, self.__createHorizontalDrillingAndCenterPunchBaseHack),
            CreatorClass(HorizontalDado2D, 'dados', self.SORTING_FACE_HORIZONTAL, self.createHorizontalDadoBase),
            CreatorClass(HorizontalContour2D, 'contours', self.SORTING_FACE_HORIZONTAL, self.createHorizontalContourBase),
            CreatorClass(HorizontalPocket2D, 'pockets', self.SORTING_FACE_HORIZONTAL, self.createHorizontalPocketBase),

            CreatorClass(FiveAxisDrilling, 'drillingsandcenterpunches', self.SORTING_FACE_FIVE_AXIS, self.createFiveAxisDrillingAndCenterPunchBase),
            CreatorClass(FiveAxisCenterPunch, 'drillingsandcenterpunches', self.SORTING_FACE_FIVE_AXIS, self.createFiveAxisDrillingAndCenterPunchBase),

            CreatorClass(NCMacro2D, 'ncmacros', self.SORTING_FACE_VERTICAL, self.createNCMacroBase),
        )

        self.SURFACE_ORIGIN = SURFACE_ORIGIN_FRONT_LEFT

    def getDecimalSeparator(self, **keywords):
        return "."

    def useMachineConfiguration(self, **keywords):
        machineConfiguration = self.getMachineConfiguration()
        self.INDIVIDUAL_EXPORT_FORMAT = machineConfiguration.get('individual-export-format', '')

        self.PIERCING_DRILLING_ADDITION = machineConfiguration.getAsFloat(MachineConfigurationKeys.PIERCING_DRILLING_ADDITION_KEY, 2.0)
        self.PIERCING_POCKET_ADDITION = machineConfiguration.getAsFloat(MachineConfigurationKeys.PIERCING_POCKET_ADDITION_KEY, 3.0)
        self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER = machineConfiguration.getAsFloat(MachineConfigurationKeys.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER_KEY, 0)
        self.CREATE_NC_STOP_FOR_CONTOURS = machineConfiguration.getAsBool(MachineConfigurationKeys.CREATE_NC_STOP_FOR_CONTOURS_KEY, True)
        self.ALLOW_SMALL_VERTICAL_DRILLINGS_WITHOUT_TOOL_ID = machineConfiguration.getAsBool(MachineConfigurationKeys.ALLOW_SMALL_VERTICAL_DRILLINGS_WITHOUT_TOOL_ID_KEY, False)

        # We still use the property 'sizing-contour-tool-is-right-hand' for compatibility reasons (changed in Epic ID-681)
        self.SIZING_MILLING_DIRECTION = machineConfiguration.get(MachineConfigurationKeys.SIZING_CONTOUR_TOOL_IS_RIGHT_HAND_KEY, 'opposite')

        self.SIZING_DEPTH_ADJUSTMENT = machineConfiguration.getAsFloat(MachineConfigurationKeys.SIZING_DEPTH_ADJUSTMENT_KEY, 3.0)
        self.SIZING_ENTRY_OF_PROFILE = machineConfiguration.get(MachineConfigurationKeys.SIZING_ENTRY_OF_PROFILE_KEY, 'arc piercing')
        self.CREATE_SIZING_WITH_MACRO = machineConfiguration.getAsBool(MachineConfigurationKeys.CREATE_SIZING_WITH_MACRO_KEY, False)  # TODO IT-252: would it not better if each export would handle it? (also as documentation) -- NG
        self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR = False
        self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS = True

        self.SECURITY_DISTANCE_FOR_DADO_SAW = machineConfiguration.getAsFloat(MachineConfigurationKeys.SECURITY_DISTANCE_FOR_DADO_SAW_KEY, 100)

        # interiorcad 2020 F8: not used anymore, see interiorcad.interface.surface.centerpunch2D.CenterPunch2D.setConfiguratedDepth
        self.DEFAULT_CENTER_PUNCH_DEPTH = machineConfiguration.getAsFloat(MachineConfigurationKeys.DEFAULT_CENTER_PUNCH_DEPTH_KEY, 0.5)

        self.EXPORT_VERTICAL_DRILLINGS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_VERTICAL_DRILLINGS_KEY, True)
        self.EXPORT_HORIZONTAL_DRILLINGS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_HORIZONTAL_DRILLINGS_KEY, True)
        self.EXPORT_FIVE_AXIS_DRILLINGS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_FIVE_AXIS_DRILLINGS_KEY, True)
        self.EXPORT_VERTICAL_RABBETS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_VERTICAL_RABBETS_KEY, True)
        self.EXPORT_VERTICAL_DADOS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_VERTICAL_DADOS_KEY, True)
        self.EXPORT_HORIZONTAL_DADOS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_HORIZONTAL_DADOS_KEY, True)
        self.EXPORT_VERTICAL_CONTOURS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_VERTICAL_CONTOURS_KEY, True)
        self.EXPORT_HORIZONTAL_CONTOURS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_HORIZONTAL_CONTOURS_KEY, True)
        self.EXPORT_MITERS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_MITERS_KEY, True)
        self.EXPORT_SIZINGS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_SIZINGS_KEY, True)
        self.EXPORT_VERTICAL_NC_MACROS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_VERTICAL_NC_MACROS_KEY, True)
        self.EXPORT_HORIZONAL_NC_MACROS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_HORIZONTAL_NC_MACROS_KEY, self.EXPORT_VERTICAL_NC_MACROS)
        self.EXPORT_FIVE_AXIS_NC_MACROS = machineConfiguration.getAsBool(MachineConfigurationKeys.EXPORT_FIVE_AXIS_NC_MACROS_KEY, self.EXPORT_VERTICAL_NC_MACROS)

        self.FILE_NAME_POSTFIX_SIZING = machineConfiguration.getIfNotEmpty(MachineConfigurationKeys.FILE_NAME_POSTFIX_SIZING_KEY,
            getLocalizedString(FILE_NAME_POSTFIXES_RESOURCEGROUP, 1, 'Sizing'))
        self.FILE_NAME_POSTFIX_MAIN = machineConfiguration.getIfNotEmpty(MachineConfigurationKeys.FILE_NAME_POSTFIX_MAIN_KEY,
            getLocalizedString(FILE_NAME_POSTFIXES_RESOURCEGROUP, 2, 'Main'))
        self.FILE_NAME_POSTFIX_OPPOSITE = machineConfiguration.getIfNotEmpty(MachineConfigurationKeys.FILE_NAME_POSTFIX_OPPOSITE_KEY,
            getLocalizedString(FILE_NAME_POSTFIXES_RESOURCEGROUP, 3, 'Opposite'))

        self.checkThatFilePostFixesAreDifferents()
        self.OUTPUT_FILE_NAME = machineConfiguration.get(MachineConfigurationKeys.OUTPUT_FILE_NAME_KEY, '{DosFileNameDependingCabinetName}_{WorkpieceName}_{filePostfix}')
        self.USE_DOS_FILENAMES = machineConfiguration.getAsBool(MachineConfigurationKeys.USE_DOS_FILENAMES_KEY, False)

        self.USE_INDIVIDUAL_OUTPUT_NAME_FILTERS = machineConfiguration.getAsBool('use-individual-output-name-filters', True)
        self.OUTPUT_FILE_NAME_FILTER = machineConfiguration.get('output-file-name-filter', 'diacritics-replacements')

        self.ALLOW_VERTICAL_PIERCING_MACHININGS_TO_CHANGE_SIDES = machineConfiguration.getAsBool(MachineConfigurationKeys.ALLOW_VERTICAL_PIERCING_DRILLINGS_TO_CHANGE_SIDES_KEY, False)

        self.VERTICAL_GRID_DRILLING_CONFIGURATION = {
            'gridDrillingXDistance': machineConfiguration.getAsFloat(MachineConfigurationKeys.GRID_DRILLING_X_DISTANCE_KEY, 32.0),
            'gridDrillingXMinimumNumberOfDrillings': machineConfiguration.getAsInt(MachineConfigurationKeys.GRID_DRILLING_X_MINIMUM_NUMBER_OF_DRILLINGS_KEY, 2),
            'gridDrillingYDistance': machineConfiguration.getAsFloat(MachineConfigurationKeys.GRID_DRILLING_Y_DISTANCE_KEY, 32.0),
            'gridDrillingYMinimumNumberOfDrillings': machineConfiguration.getAsInt(MachineConfigurationKeys.GRID_DRILLING_Y_MINIMUM_NUMBER_OF_DRILLINGS_KEY, 2),
        }

        MACHINING_CONFIGURATION = machineConfiguration.get(MachineConfigurationKeys.MACHINING_CONFIGURATION_KEY, '')
        self.MACHININGS_CONFIGURATION = self.extractMachiningOrderingConfiguration(machiningOrdering=MACHINING_CONFIGURATION)

        # 'Direction Diameter Distance'   # good readable, optimization in wop-software mandatory
        # 'Direction Distance'   # optimized for short distances
        # 'Direction ToolID Diameter Distance'   # drill on spindle
        DRILLING_ORDERINGS_RAW = machineConfiguration.get(MachineConfigurationKeys.DRILLING_ORDERINGS_KEY, 'Direction Diameter Distance')

        self.DRILLING_ORDERINGS = DRILLING_ORDERINGS_RAW.split()

        self.DADO_REPLACEMENT = machineConfiguration.get('dado-replacement', 'contour-CW')  # contour-CW & contour-CCW
        self.RABBET_REPLACEMENT = machineConfiguration.get('rabbet-replacement', 'contour-CW')  # contour-CW & contour-CCW
        self.POCKET_DIRECTION = machineConfiguration.get('pocket-direction', 'pocket-same')

        self.SECURITY_DISTANCE_ADDITION_FOR_DADO_RABBET = machineConfiguration.getAsFloat('security-distance-addition-for-rabbet', 10)

        self.EXPORT_EMPTY_WORKPIECE = machineConfiguration.getAsBool('export-empty-workpiece', False)
        self.MIN_LENGTH_OF_WORKPIECE = machineConfiguration.getAsFloat('min-length-of-workpiece', 150)
        self.MIN_WIDTH_OF_WORKPIECE = machineConfiguration.getAsFloat('min-width-of-workpiece', 70)

        self.MACHINING_SORTING_STRATEGY = machineConfiguration.get('machining-sorting-strategy', 'tool-usage').lower()  # tool-type, tool-dimension, basic
        self.FORCE_SIZING_PLACEMENT = machineConfiguration.get('force-sizing-placement', 'no')  # first, last, no = machining-configuration

        self.GRAIN_CONSIDERATION_MODE = machineConfiguration.get('grain-consideration-mode', 'ignore')  # rotate-left | rotate-right

        self.ALWAYS_CREATE_MAIN_SIDE = machineConfiguration.getAsBool('always-create-main-side', False)
        self.ALLOW_HORIZONTAL_MACHININGS_ON_MAIN_SIDE = machineConfiguration.getAsBool('allow-horizontal-machinings-on-main-side', True)

        self.ALLOW_CONTOUR_ON_OTHER_THAN_SELECTED_MACHINE = machineConfiguration.getAsBool('allow-contour-on-other-than-selected-machine', False)

        self.LICENSE_INFO = getLicenseInfo()

        self.VERTICAL_CLAMEX_MAX_DISTANCE_FROM_EDGE = machineConfiguration.getAsFloat('vertical-clamex-max-distance-from-edge', 38.0)  # We choose 38.0 as dafault, as it is a distance most tools can handle without crashing into the workpiece
        self.VERTICAL_CLAMEX_INNER_PLACEMENT_EXPORT = machineConfiguration.get('vertical-clamex-inner-placement-export', 'none')

    def extractMachiningOrderingConfiguration(self, **keywords):
        machiningOrderings = []
        for ordering in keywords['machiningOrdering'].split():
            if ordering.lower() in self.LEGACY_MACHINING_CREATORS:
                machiningOrderings.append(ordering.lower())
            else:
                message = getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 41, "Unknown machining ordering: {ordering}! Accepted values: {orderings}")
                logging.warning(message.format(ordering=ordering, orderings=tuple(self.LEGACY_MACHINING_CREATORS)))

        machiningOrderings.extend(filter(lambda key: key not in machiningOrderings, self.LEGACY_MACHINING_CREATORS))

        return machiningOrderings

    def initLogging(self, **keywords):
        logging.raiseExceptions = False
        logFileDirectory = ExportHelper.createExportPathWithoutWorkpiece(self.configuration)
        if os.path.basename(logFileDirectory) == '_':  # In case of empty {Position...}_{Position...} at the end of the path. Assumes Project Details before DocumentName, Position Details and CabinetName.
            logFileDirectory = os.path.dirname(logFileDirectory)
        ExportHelper.CreatePathIfNotExists(logFileDirectory)
        self.logFileName = os.path.join(logFileDirectory, self.configuration.logFileName)
        self.createLogFileArchive(**keywords)
        self.loggingHandler = logging.FileHandler(filename=self.logFileName, encoding=self.LOG_FILE_ENCODING, mode='a', delay=True)
        self.loggingHandler.setFormatter(logging.Formatter(fmt='%(levelname)s:%(message)s'))
        logging.root.addHandler(self.loggingHandler)
        logging.root.setLevel(logging.INFO)

    def removeLoggingHandle(self):
        if self.loggingHandler:
            self.loggingHandler.flush()
            logging.root.removeHandler(self.loggingHandler)
            self.loggingHandler.close()
            self.loggingHandler = None

    LOG_FILE_ENCODING = 'utf-8-sig'

    def createLogFileArchive(self, **keywords):
        # Create 'XG Export Log Archive.txt', move older messages and delete existing 'XG Export Log.txt'
        if os.path.isfile(self.logFileName):
            logFileNameArchive = os.path.splitext(self.logFileName)[0] + ' Archive' + os.path.splitext(self.logFileName)[1]
            logFileTime = datetime.datetime.fromtimestamp(os.path.getmtime(self.logFileName)).strftime('%Y-%m-%d %H:%M:%S')

            if not os.path.isfile(logFileNameArchive):
                with open(logFileNameArchive, mode='a', encoding=self.LOG_FILE_ENCODING) as f:
                    headline = getLocalizedString(COMMON_MESSAGES_RESOURCEGROUP, 1, 'This is the archive for all previous messages of XG Export Log.txt in chronological order.')
                    f.write(headline + '\n' + '\n')

            # We are using errors='surrogateescape' to avoid losing data if a user had written a comment in the log file with the wrong encoding
            with open(self.logFileName, mode='r', encoding=self.LOG_FILE_ENCODING, errors='surrogateescape') as f:
                logFileLines = f.readlines()

            with open(logFileNameArchive, mode='a', encoding=self.LOG_FILE_ENCODING, errors='surrogateescape') as f:
                f.write(logFileTime + '\n')
                f.writelines(logFileLines)
                f.write('\n')

            os.remove(self.logFileName)

    def checkThatFilePostFixesAreDifferents(self):
        if self.FILE_NAME_POSTFIX_SIZING == self.FILE_NAME_POSTFIX_MAIN or self.FILE_NAME_POSTFIX_MAIN == self.FILE_NAME_POSTFIX_OPPOSITE or self.FILE_NAME_POSTFIX_OPPOSITE == self.FILE_NAME_POSTFIX_SIZING:
            raise Exception(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 2, "At least two File Name Postfixes are equal!"))

    def isFunctionSupportedByWop(self, **keywords):
        return hasattr(self, keywords['function'])

    def getCoordinatePrecision(self, **keywords):
        return self.configuration.coordinatePrecision

    def getAnglePrecision(self, **keywords):
        return self.configuration.angularPrecision

    def shouldKeepSignificantDecimalsOnly(self, **keywords):
        return self.configuration.significantDecimalsOnly

    def getMachineConfiguration(self, **keywords):
        return self.machineconfiguration

    def getDataAsString(self, **keywords):
        return ExportHelper.getDataAsString(**keywords)

    def getDataAsInt(self, **keywords):
        return ExportHelper.getDataAsInt(**keywords)

    def getDataAsFloat(self, **keywords):
        return ExportHelper.getDataAsFloat(**keywords)

    def formatAsString(self, **keywords):
        return self.formatter.formatAsString(data=keywords['data'])

    def formatAsStringSingleLine(self, **keywords):
        return self.CARRIAGE_RETURN_REPLACEMENT.join(self.formatter.formatAsString(data=keywords['data']).split())

    def formatAsInt(self, **keywords):
        return self.formatter.formatAsInt(data=keywords['data'])

    def formatAsCoordinate(self, **keywords):
        return self.formatter.formatAsCoordinate(data=keywords['data'])

    def formatAsSignedCoordinate(self, **keywords):
        return self.formatter.formatAsSignedCoordinate(data=keywords['data'])

    def formatAsAngle(self, **keywords):
        return self.formatter.formatAsAngle(data=keywords['data'])

    def formatAsCoordinateAddition(self, **keywords):
        lhs = keywords['lhs']
        rhs = keywords['rhs']
        if isNearlyEqual(rhs, 0) or rhs is None:
            return self.formatAsCoordinate(data=lhs)
        elif rhs > 0:
            return self.formatAsCoordinate(data=lhs) + '+' + self.formatAsCoordinate(data=rhs)
        else:
            return self.formatAsCoordinate(data=lhs) + '-' + self.formatAsCoordinate(data=-rhs)

    def asString(self, **keywords):
        keywords['data'] = self.getDataAsString(**keywords)
        return self.formatAsString(**keywords)

    def asStringSingleLine(self, **keywords):
        keywords['data'] = self.getDataAsString(**keywords)
        return self.formatAsStringSingleLine(**keywords)

    def asInt(self, **keywords):
        keywords['data'] = self.getDataAsInt(**keywords)
        return self.formatAsInt(**keywords)

    def asCoordinate(self, **keywords):
        keywords['data'] = self.getDataAsFloat(**keywords)
        return self.formatAsCoordinate(**keywords)

    def asAngle(self, **keywords):
        keywords['data'] = self.getDataAsFloat(**keywords)
        return self.formatAsAngle(**keywords)

    # TODO IT-226 Move into ExportHelper?  --mjk
    def getProjectInfo(self, **keywords):
        return ExportHelper.createProjectInfoOfFirstSelectedCustomPart()

    # TODO IT-226 Move into ExportHelper?  --mjk
    def getPositionInfo(self, uuid):
        positionInfo = {}
        ExportHelper.updateDictionaryWithPartData(positionInfo, uuid)
        return positionInfo

    GRAIN_LENGTHWISE, GRAIN_CROSSWISE, GRAIN_NONE = 0, 1, 2

    def getSimplifiedGrain(self, grain):
        if grain == self.GRAIN_NONE:
            return 0
        else:  # self.GRAIN_LENGTHWISE / self.GRAIN_CROSSWISE
            return 1

    def getWorkpieceCommentLabels(self, **_keywords):
        def getLocalizedCommentLabel(index, default):
            return getLocalizedString(CUTTINGLIST_RESOURCEGROUP, index, default)

        return {
            'WorkpieceLabel': getLocalizedCommentLabel(7, 'Unique Name'),
            'CommentLabel': getLocalizedCommentLabel(30, 'Comment'),
            'ItemIDLabel': getLocalizedCommentLabel(1, 'Board'),
            'CountLabel': getLocalizedCommentLabel(6, 'Count'),

            'ProjectIDLabel': getLocalizedCommentLabel(17, 'Project ID'),
            'PositionNumberLabel': getLocalizedCommentLabel(18, 'Position Number'),
            'CabinetNameLabel': getLocalizedCommentLabel(20, 'Cabinet Name'),

            'Edge1Label': getLocalizedCommentLabel(11, 'Edge ID Front 1'),
            'Edge2Label': getLocalizedCommentLabel(12, 'Edge ID Back 2'),
            'Edge3Label': getLocalizedCommentLabel(13, 'Edge ID Left 3'),
            'Edge4Label': getLocalizedCommentLabel(14, 'Edge ID Right 4'),
            'EdgeFrontLabel': getLocalizedCommentLabel(11, 'Edge ID Front 1'),
            'EdgeBackLabel': getLocalizedCommentLabel(12, 'Edge ID Back 2'),
            'EdgeLeftLabel': getLocalizedCommentLabel(13, 'Edge ID Left 3'),
            'EdgeRightLabel': getLocalizedCommentLabel(14, 'Edge ID Right 4'),
        }

    def getWorkpieceCommentProperties(self, **keywords):
        surface = keywords['surface']
        edges = surface.edges
        uuid = surface.workpieceUUID
        keywords['uuid'] = uuid
        projectInfo = self.getProjectInfo()
        positionInfo = self.getPositionInfo(uuid)

        # Same rows/functions as in prepareFilePath()
        filePathWithoutFilename = self.createExportFilePath(workpieceUUID=surface.workpieceUUID, **keywords)
        fileName = self.createFileName(name=surface.workpieceName, extension=self.getNCFileExtension(**keywords), **keywords)
        filepath = os.path.join(filePathWithoutFilename, fileName)

        return {
            'WorkpieceName': surface.workpieceName,
            'Comment': self.asStringSingleLine(key='part_comment', **keywords),
            'ItemID': self.asString(key='part_itemid', **keywords),
            'Count': self.asInt(data=surface.count),

            'ProjectID': projectInfo['ProjectID'],
            'ProjectCommission': projectInfo['ProjectCommission'],
            'ProjectConstruction': projectInfo['ProjectConstruction'],
            'PositionNumber': positionInfo['PositionNumber'],
            'PositionDescription': positionInfo['PositionDescription'],
            'CabinetName': positionInfo['CabinetName'],

            'Edge1': edges.front.supplierID,
            'Edge2': edges.back.supplierID,
            'Edge3': edges.left.supplierID,
            'Edge4': edges.right.supplierID,
            'EdgeFront': edges.front.supplierID,
            'EdgeBack': edges.back.supplierID,
            'EdgeLeft': edges.left.supplierID,
            'EdgeRight': edges.right.supplierID,
            'EdgeFrontThickness': self.asCoordinate(data=edges.front.thickness, **keywords),
            'EdgeBackThickness': self.asCoordinate(data=edges.back.thickness, **keywords),
            'EdgeLeftThickness': self.asCoordinate(data=edges.left.thickness, **keywords),
            'EdgeRightThickness': self.asCoordinate(data=edges.right.thickness, **keywords),

            'Covering1': self.asString(key='part_covering1', **keywords),
            'Covering2': self.asString(key='part_covering2', **keywords),

            'Laminate1': self.asString(key='part_laminate1', **keywords),
            'Laminate2': self.asString(key='part_laminate2', **keywords),
            'Laminate3': self.asString(key='part_laminate3', **keywords),

            'NCProgramName': os.path.splitext(fileName)[0],
            'NCProgramFilename': fileName,
            'NCProgramFolder': filePathWithoutFilename,
            'NCProgramPath': filepath,

            'HasGrain': self.asInt(data=self.getSimplifiedGrain(grain=surface.grain)),
            'IsFirstSide': self.asInt(data=surface.isFirstSide),
            'HasMainSide': self.asInt(data=surface.hasMainSide),
            'HasOppositeSide': self.asInt(data=surface.hasOppositeSide),
        }

    def getWorkpieceComments(self, **keywords):
        comments = self.getWorkpieceCommentLabels(**keywords)
        comments.update(self.getWorkpieceCommentProperties(**keywords))

        ExportHelper.updateDictionaryWithPartData(comments, keywords['surface'].workpieceUUID)
        comments.pop('SideName', None)  # not used as comment
        comments.pop('SideNameNesting', None)  # not used as comment
        comments.pop('ExportFormat', None)  # not used as comment

        ExportHelper.updateDictionaryWithDateTime(comments)

        return comments

    def getHorizontalMachiningsExporter(self, **keywords):
        return HorizontalMachiningsExporter(self)

    def createZValueConverter(self, machineConfigurationKey, defaultPosition, defaultDirection):
        value = self.getMachineConfiguration().get(machineConfigurationKey, False)
        return ZValueConverter(ZValueConverterConfiguration.getOriginFrom(value, defaultPosition), ZValueConverterConfiguration.getDirectionFrom(value, defaultDirection))

    def createMachinings(self, **keywords):
        # We can only use the NewSchool sorting if no user customization was done on createVerticalMachinings, createHorizontalMachinings and createFiveAxisMachinings
        # (We must check that the codes (__code__) are equal, as the function are automatically different if there are not the same class.)
        if all((
            self.createVerticalMachinings.__code__ is NCExportBase.createVerticalMachinings.__code__,
            self.createHorizontalMachinings.__code__ is NCExportBase.createHorizontalMachinings.__code__,
            self.createFiveAxisMachinings.__code__ is NCExportBase.createFiveAxisMachinings.__code__
        )):
            return self.createMachiningsOneByOne(**keywords)
        else:
            return self.createMachiningsLegacyWay(**keywords)

    def guarantSizingPlacement(self, **keywords):
        def helper():
            sizings = filter(lambda machining: isinstance(machining, Sizing2D), machinings)
            otherMachinings = itertools.filterfalse(lambda machining: isinstance(machining, Sizing2D), machinings)
            return sizings, otherMachinings

        machinings = list(keywords['machinings'])
        if self.FORCE_SIZING_PLACEMENT == 'first':
            sizings, otherMachinings = helper()
            return list(itertools.chain(sizings, otherMachinings))
        elif self.FORCE_SIZING_PLACEMENT == 'last':
            sizings, otherMachinings = helper()
            return list(itertools.chain(otherMachinings, sizings))
        else:
            return machinings

    # TODO: it would be much better to make a Surface-Post-Processor out of the function, but currently Surface.machinings does not include sizing -- NG (InnoWeek 2017-12)
    def sortMachinings(self, **keywords):
        sorter = MachiningSorter(self)
        machinings = sorter.sortMachiningsBasically(keywords['machinings'])
        if self.MACHINING_SORTING_STRATEGY == 'tool-usage':
            keywords['machinings'] = sorter.sortMachiningByToolUsage(machinings)
        elif self.MACHINING_SORTING_STRATEGY == 'tool-type':
            keywords['machinings'] = sorter.sortMachiningsByToolType(machinings)
        elif self.MACHINING_SORTING_STRATEGY == 'tool-dimension':
            keywords['machinings'] = sorter.sortMachiningsByToolDimension(machinings)
        else:  # basic
            keywords['machinings'] = machinings
        return self.guarantSizingPlacement(**keywords)

    def createMachiningsOneByOne(self, **keywords):
        def generator():
            for machining in self.sortMachinings(machinings=keywords['surface'].getAllMachinings(), **keywords):
                for creator in self.SINGLE_MACHINING_CREATORS:
                    if isinstance(machining, creator.machiningClass):
                        if creator.creator:
                            yield self.__createMachining(machining=machining, exportMethod=creator.creator, **keywords)
                        break

        return ''.join(filter(lambda line: line is not None, generator()))

    # Compatibility function for old code using createMachinings; do NOT use for new code!
    def createMachiningsLegacyWay(self, **keywords):
        return joinIfNotNone('',
            self.createVerticalMachinings(**keywords),
            self.createHorizontalMachinings(**keywords),
            self.createFiveAxisMachinings(**keywords),
        )

    # Compatibility function for old code; do NOT use for new code!
    def __createOrderedMachinings(self, **keywords):
        def creator(creators, configurationIndex):
            return creators[configurationIndex](**keywords) if creators[configurationIndex] else None

        configurationIndex = keywords['configurationIndex']
        lines = (creator(self.LEGACY_MACHINING_CREATORS[machining], configurationIndex) for machining in self.MACHININGS_CONFIGURATION)
        return ''.join(filter(lambda line: line is not None, lines))

    # Compatibility function for old code; do NOT use for new code!
    def createVerticalMachinings(self, **keywords):
        return self.__createOrderedMachinings(configurationIndex=0, **keywords)

    # Compatibility function for old code; do NOT use for new code!
    # This code does NOT call the new horizontal machinings, like contours, dados and pockets
    def createHorizontalMachinings(self, **keywords):
        return self.__createOrderedMachinings(configurationIndex=1, **keywords)

    # Compatibility function for old code; do NOT use for new code!
    def createFiveAxisMachinings(self, **keywords):
        return self.__createOrderedMachinings(configurationIndex=2, **keywords)

    def getDrillType(self, **keywords):
        defaultDrillType = str(keywords['defaultDrillType'])
        drilling = keywords['drilling']
        if drilling.tool:
            return drilling.tool.getPropertyAsString('drilling-drill-type', defaultDrillType)
        else:
            return defaultDrillType

    def createVerticalDrillingsAndCenterPunches(self, **keywords):
        return self.__createMachinings(keywords['surface'].verticalDrillingsAndCenterPunches, exportMethod=self.createVerticalDrillingAndCenterPunchBase, **keywords)

    def createOneDrillingOrCenterPunch(self, drillingFunction, centerPunchFunction, **keywords):
        drilling = keywords['drilling']
        if isinstance(drilling, Drilling2D):
            return drillingFunction(**keywords)
        elif isinstance(drilling, CenterPunch2D):
            return centerPunchFunction(**keywords)
        else:
            raise TypeError('Machining is supposed to be drilled, but is neither a Drilling nor a Center Punch')

    def createVerticalDrillingAndCenterPunchBase(self, **keywords):
        return self.createOneDrillingOrCenterPunch(self.createVerticalDrillingBase, self.createVerticalCenterPunchBase, **keywords)

    def __createHorizontalDrillingAndCenterPunchBaseHack(self, **keywords):
        # TODO: can it be implemented in a less hacky way -- NG (InnoWeek 2017-12)
        exporter = self.getHorizontalMachiningsExporter(**keywords)
        exporter.machiningCreator = HorizontalDrillingsAndCenterPunchesCreator(self)
        exporter.machinings = (keywords['machining'],)
        del keywords['drilling']
        del keywords['machining']
        return exporter.create(**keywords)

    def createHorizontalDrillingsAndCenterPunches(self, **keywords):
        exporter = self.getHorizontalMachiningsExporter(**keywords)
        exporter.machiningCreator = HorizontalDrillingsAndCenterPunchesCreator(self)
        exporter.machinings = keywords['surface'].horizontalDrillingsAndCenterPunches
        return exporter.create(**keywords)

    def createFiveAxisDrillingsAndCenterPunches(self, **keywords):
        return self.__createMachinings(keywords['surface'].fiveAxisDrillingsAndCenterPunches, exportMethod=self.createFiveAxisDrillingAndCenterPunchBase, **keywords)

    def createFiveAxisDrillingAndCenterPunchBase(self, **keywords):
        return self.createOneDrillingOrCenterPunch(self.createFiveAxisDrillingBase, self.createFiveAxisCenterPunchBase, **keywords)

    def createVerticalPockets(self, **keywords):
        return self.__createMachinings(keywords['surface'].pockets, exportMethod=self.createVerticalPocketBase, **keywords)

    def createVerticalGridDrillings(self, **keywords):
        return self.__createMachinings(keywords['surface'].verticalGridDrillings, exportMethod=self.createVerticalGridDrillingBase, **keywords)

    def createVerticalRabbets(self, **keywords):
        return self.__createMachinings(keywords['surface'].rabbetsVertical, exportMethod=self.createVerticalRabbetBase, **keywords)

    # TODO ISP-20: better name, as this function is used for circular pockets too --NG
    def isDadoRabbetReplacementCW(self, machining, machiningReplacement, **keywords):
        if machiningReplacement.endswith('-opposite'):
            return (machining.tool.getRouterRotationDirection() == 'right'
                or not bool(machining.tool.getRouterRotationDirection()))  # Support older (test) tools that do not have a Router Rotation Direction
        elif machiningReplacement.endswith('-same'):
            return (machining.tool.getRouterRotationDirection() == 'left'
                and bool(machining.tool.getRouterRotationDirection()))  # Support older (test) tools that do not have a Router Rotation Direction
        elif machiningReplacement.endswith('-CW'):
            return True
        elif machiningReplacement.endswith('-CCW'):
            return False
        else:
            assert False, "Unknown suffix for determining CW or CCW!"

    def orientRabbets(self, **keywords):
        oppositeDirection = not self.isStandardOrigin(**keywords)
        rabbetAdjuster = RabbetAdjuster(keywords['surface'].finalFinishedSizeShape, oppositeDirection)

        for rabbet in keywords['surface'].rabbetsVertical:
            # The direction of the contour is opposite of the direction of the workpiece path.
            if self.isDadoRabbetReplacementCW(rabbet, self.RABBET_REPLACEMENT):
                rabbetAdjuster.getAdjustedRabbetAtEdgeCCW(rabbet)
            else:  # contour-ccw
                rabbetAdjuster.getAdjustedRabbetAtEdgeCW(rabbet)

        # Do not reassign to keywords['surface'].rabbetsVertical as it would change the order of machinings

    def createVerticalDados(self, **keywords):
        return self.__createMachinings(keywords['surface'].dadosVertical, exportMethod=self.createVerticalDadoBase, **keywords)

    def createVerticalMiters(self, **keywords):
        return self.__createMachinings(keywords['surface'].mitersVertical, exportMethod=self.createVerticalMiterBase, **keywords)

    def createFiveAxisDrillings(self, **keywords):
        return self.__createMachinings(keywords['surface'].getFiveAxisDrillings(), exportMethod=self.createFiveAxisDrillingBase, **keywords)

    def createFiveAxisCenterPunches(self, **keywords):
        return self.__createMachinings(keywords['surface'].getFiveAxisCenterPunches(), exportMethod=self.createFiveAxisCenterPunchBase, **keywords)

    def getSizingZValue(self, **keywords):
        # TODO: Wie soll man's testen? --ng IT-202
        sizing = keywords['surface'].sizing
        if sizing and sizing.tool and sizing.tool.isRouterCentered():
            keywords['depth'] = keywords['surface'].thickness / 2
            return self.getMillingZValue(**keywords)
        else:
            return self.getMillThroughZValue(**keywords)

    def getFinishedSizeShape(self, **keywords):
        return keywords['surface'].finalFinishedSizeShape

    def getMillThroughZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        if self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS:
            keywords['depth'] = boardThickness + self.SIZING_DEPTH_ADJUSTMENT
        else:
            keywords['depth'] = boardThickness - self.SIZING_DEPTH_ADJUSTMENT
        return self.getMillingZValue(**keywords)

    def getMillingZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        depth = keywords['depth']
        return self.zValueMillingConverter.convertWithThickness(boardThickness, depth)

    def getDrillingZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        depth = keywords['depth']
        return self.zValueVerticalDrillingConverter.convertWithThickness(boardThickness, depth)

    def getHorizontalDrillingZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        zPosition = keywords['zPosition']
        return self.zValueHorizontalDrillingConverter.convertWithThickness(boardThickness, zPosition)

    def getFiveAxisDrillingZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        zPosition = keywords['zPosition']
        return self.zValueFiveAxisDrillingConverter.convertWithThickness(boardThickness, zPosition)

    def getPocketZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        depth = keywords['depth']
        return self.zValueVerticalPocketConverter.convertWithThickness(boardThickness, depth)

    def getPocketPiercingZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        return self.getPocketZValue(depth=boardThickness + self.PIERCING_POCKET_ADDITION, **keywords)

    def getDadoZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        depth = keywords['depth']
        return self.zValueDadoSawConverter.convertWithThickness(boardThickness, depth)

    def stopForRemovingWaste(self, **keywords):
        return self.createNCStop(comment=getLocalizedString(NCEXPORT_NCPROGRAMS_RESOURCEGROUP, 1, 'Please remove waste'), **keywords)

    def stopBeforeSizing(self, **keywords):
        return self.createNCStop(comment=getLocalizedString(NCEXPORT_NCPROGRAMS_RESOURCEGROUP, 2, 'Please check approach'), **keywords)

    def stopForInspectingContours(self, **keywords):
        if self.CREATE_NC_STOP_FOR_CONTOURS:
            return self.createNCStop(comment=getLocalizedString(NCEXPORT_NCPROGRAMS_RESOURCEGROUP, 3, 'Please inspect contours'), **keywords)
        else:
            return ''

    def getApproachMode(self, **keywords):
        approachDeparture = keywords['approachDeparture']
        if approachDeparture == 'linear piercing':
            return self.APPROACH_MODE_LINEAR_PIERCING
        elif approachDeparture == 'linear flying':
            return self.APPROACH_MODE_LINEAR_FLYING
        elif approachDeparture == 'arc piercing':
            return self.APPROACH_MODE_ARC_PIERCING
        elif approachDeparture == 'arc flying':
            return self.APPROACH_MODE_ARC_FLYING
        elif approachDeparture == 'without':
            return self.APPROACH_MODE_WITHOUT
        else:
            return keywords['defaultApproachMode']

    def getSizingApproachMode(self, **keywords):
        return self.getApproachMode(approachDeparture=self.SIZING_ENTRY_OF_PROFILE, defaultApproachMode=self.APPROACH_MODE_SIZING_DEFAULT, **keywords)

    def createNCPolylineSizingContourFromSegmentList(self, **keywords):
        sizing = keywords['surface'].sizing
        return self.createContourFromSegmentList(
            createStartBlock=self.createSizingStartBlock,
            createEndBlock=self.createSizingEndBlock,
            machining=sizing,
            toolID=sizing.toolID,
            **keywords)

    def createRabbetContourFromSegmentList(self, **keywords):
        return self.createContourFromSegmentList(
            createStartBlock=self.createRabbetContourStartBlock,
            createEndBlock=self.createRabbetContourEndBlock,
            toolID=keywords['machining'].toolID,
            **keywords)

    def createDadoContourFromSegmentList(self, **keywords):
        return self.createContourFromSegmentList(
            createStartBlock=self.createDadoContourStartBlock,
            createEndBlock=self.createDadoContourEndBlock,
            toolID=keywords['machining'].toolID,
            **keywords)

    def createNCPolylineContourFromSegmentList(self, **keywords):
        return self.createContourFromSegmentList(
            createStartBlock=self.createContourStartBlock,
            createEndBlock=self.createContourEndBlock,
            toolID=keywords['machining'].toolID,
            **keywords)

    def generateContourFromSegmentList(self, **keywords):
        segmentList = keywords['segmentList']
        if segmentList:
            yield keywords['createStartBlock'](startVertex=segmentList[0].getStartVertex(), **keywords)
            for segmentID, segment in enumerate(segmentList, start=1):
                yield self.createContourSegment(segmentID=segmentID, segment=segment, **keywords)
            yield keywords['createEndBlock'](**keywords)

    def createContourFromSegmentList(self, **keywords):
        return '\n'.join(self.generateContourFromSegmentList(**keywords))

    def createContourSegment(self, **keywords):
        if keywords['segment'].isLineSegment():
            return self.createLineSegment(**keywords)
        else:
            return self.createRadiusSegment(**keywords)

    def createSizingContourDefinition(self, **keywords):
        if self.EXPORT_SIZINGS:
            zValue = self.getSizingZValue(**keywords)
            return self.createDefinitionForOneSizingContour(zValue=zValue, **keywords)

    def createDefinitionForOneSizingContour(self, **keywords):
        surface = keywords['surface']
        if surface.hasSizing():
            return self.createSizingDefinitionForOnePolyline(polyline=surface.sizing.polyline, **keywords)
        else:
            return ''

    def createSizingDefinitionForOnePolyline(self, **keywords):
        if not self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR:  # TODO IT-251: remove this if in interiorcad 2020 F4 -- NG
            segmentList = NCLineAndRadiusSegment.generateLineAndRadiusSegments(keywords['polyline'])
        else:
            segmentList = NCLineAndRadiusSegment.generateVisibleLineAndRadiusSegments(keywords['polyline'])
        keywords['surface'].sizing.segmentCount = len(segmentList)
        return self.createNCPolylineSizingContourFromSegmentList(segmentList=segmentList, **keywords)

    def createSizingContourExecution(self, **keywords):
        surface = keywords['surface']
        if surface.hasSizing():
            zValue = self.getSizingZValue(**keywords)
            return self.createStopBeforeSizing(**keywords) + self.createSizingContourBlock(sizing=surface.sizing, zValue=zValue, **keywords)
        return ''

    def createSizingContourDefinitionAsExecution(self, **keywords):
        zValue = self.getSizingZValue(**keywords)
        ### IT-115: warum wird nicht ein ähnlicher Mechanismus genutzt wie alle andere Bearbeitungen, die ein Stop-Befehl triggern? -- NG
        return self.createStopBeforeSizing(**keywords) + self.createDefinitionForOneSizingContour(zValue=zValue, **keywords)

    def createRabbetContourDefinitions(self, **keywords):
        surface = keywords['surface']
        rabbets = keywords['rabbets'] if 'rabbets' in keywords else surface.rabbetsVertical
        lines = ''.join((self.createDefinitionForOneRabbet(machining=rabbet, rabbet=rabbet, **keywords) for rabbet in rabbets))
        if lines:
            return lines

    def createDadoContourDefinitions(self, **keywords):
        def isDadoRouted(dado):
            return hasattr(dado, 'tool') and dado.tool and dado.tool.type == Tool.ROUTER

        surface = keywords['surface']
        dados = keywords['dados'] if 'dados' in keywords else surface.dadosVertical
        lines = ''.join((self.createDefinitionForOneDado(machining=dado, dado=dado, **keywords) for dado in dados if isDadoRouted(dado)))
        if lines:
            return lines

    def generateRabbetSegments(self, **keywords):
        return tuple(RabbetToNCSegments.generateRabbetSegmentsWithInsets(rabbet=keywords['rabbet']))

    def createDefinitionForOneRabbet(self, **keywords):
        segmentList = self.generateRabbetSegments(**keywords)
        keywords['rabbet'].segmentCount = len(segmentList)
        return self.createRabbetContourFromSegmentList(segmentList=segmentList, **keywords)

    def createDefinitionForOneDado(self, **keywords):
        segmentList = self.generateDadoSegments(**keywords)
        keywords['dado'].segmentCount = len(segmentList)
        return self.createDadoContourFromSegmentList(segmentList=segmentList, **keywords)

    def generateDadoSegments(self, **keywords):
        dado = keywords['dado']

        assert self.DADO_REPLACEMENT.startswith('contour'), 'This function only supports "contour-..."'

        routerRadius = dado.tool.getRouterDiameter() / 2
        xoredCondition = keywords['xoredCondition'] if 'xoredCondition' in keywords else False

        # No security distance is used here, as the WOP will set it later
        if operator.xor(self.isDadoRabbetReplacementCW(dado, self.DADO_REPLACEMENT), xoredCondition):
            return getDadoContourCW(dado, securityDistance=0, routerRadius=routerRadius)
        else:
            return getDadoContourCCW(dado, securityDistance=0, routerRadius=routerRadius)

    def getDadoRadiusCompensation(self, **keywords):
        dado = keywords['dado']

        if self.isDadoRabbetReplacementCW(dado, self.DADO_REPLACEMENT):
            return self.RADIUS_COMPENSATION_RIGHT
        else:  # contour-CCW
            return self.RADIUS_COMPENSATION_LEFT

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']
        segmentList = NCLineAndRadiusSegment.generateVisibleLineAndRadiusSegments(contour.path)
        contour.segmentCount = len(segmentList)
        return self.createNCPolylineContourFromSegmentList(segmentList=segmentList, **keywords)

    def createContourDefinitions(self, **keywords):
        surface = keywords['surface']
        contours = keywords['contours'] if 'contours' in keywords else surface.contoursVertical

        lines = ''.join((self.createDefinitionForOneContour(contour=contour, machining=contour, **keywords) for contour in contours))
        if lines:
            return lines

    def createStopBeforeSizing(self, **keywords):
        surface = keywords['surface']
        stopBeforeSizing = ''
        if surface.hasSizing() and surface.sizing.needsFlyingApproach:
            stopBeforeSizing = self.stopBeforeSizing(**keywords)
        return stopBeforeSizing

    def createVerticalDrillingBase(self, **keywords):
        drilling = keywords['drilling']
        surfaceThickness = keywords['surface'].thickness
        drillsThrough = drilling.depth >= surfaceThickness
        if drillsThrough:
            return self.createVerticalDrillingPiercing(depth=self.getDrillingPiercingZValue(**keywords), **keywords)
        else:
            return self.createVerticalDrilling(depth=self.getDrillingZValue(depth=drilling.depth, **keywords), **keywords)

    def createVerticalGridDrillingBase(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        surfaceThickness = keywords['surface'].thickness
        assert gridDrilling.depth < surfaceThickness, "Drill-through currently not supported for grid drillings"
        depth = self.getDrillingZValue(depth=gridDrilling.depth, **keywords)

        if gridDrilling.isAlongXAxis() and self.areVerticalGridDrillingsAlongXAxisSupported(**keywords):
            return self.createVerticalGridDrillingAlongXAxis(depth=depth, **keywords)
        elif gridDrilling.isAlongYAxis() and self.areVerticalGridDrillingsAlongYAxisSupported(**keywords):
            return self.createVerticalGridDrillingAlongYAxis(depth=depth, **keywords)
        elif self.areVerticalGridDrillingsAlongCRotatedAxisSupported(**keywords):
            return self.createVerticalGridDrillingAlongCRotatedAxis(depth=depth, **keywords)
        else:
            raise Exception(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 29, 'Internal Error: Grid Drilling was found but it is not supported by export!'))

    def areVerticalGridDrillingsAlongXAxisSupported(self, **keywords):
        return self.isFunctionSupportedByWop(function='createVerticalGridDrillingAlongXAxis')

    def areVerticalGridDrillingsAlongYAxisSupported(self, **keywords):
        return self.isFunctionSupportedByWop(function='createVerticalGridDrillingAlongYAxis')

    def areVerticalGridDrillingsAlongCRotatedAxisSupported(self, **keywords):
        return self.isFunctionSupportedByWop(function='createVerticalGridDrillingAlongCRotatedAxis')

    def createVerticalCenterPunchBase(self, **keywords):
        drilling = keywords['drilling']
        return self.createVerticalCenterPunch(depth=self.getDrillingZValue(depth=drilling.depth, **keywords), **keywords)

    def createVerticalCenterPunch(self, **keywords):
        return self.createVerticalDrilling(**keywords)

    def createFiveAxisDrillingBase(self, **keywords):
        drilling = keywords['drilling']
        zPosition = self.getFiveAxisDrillingZValue(zPosition=drilling.zPosition, **keywords)
        return self.createFiveAxisDrilling(zPosition=zPosition, **keywords)

    def createFiveAxisCenterPunchBase(self, **keywords):
        drilling = keywords['drilling']
        zPosition = self.getFiveAxisDrillingZValue(zPosition=drilling.zPosition, **keywords)
        return self.createFiveAxisCenterPunch(zPosition=zPosition, **keywords)

    def createFiveAxisCenterPunch(self, **keywords):
        return self.createFiveAxisDrilling(**keywords)

    def getRouterDepthPerStep(self, **keywords):
        tool = keywords['machining'].tool
        if tool:
            depthPerStep = tool.getRouterMaximumDepthPerStep()
            if math.isfinite(depthPerStep):
                return depthPerStep

        return 0.0

    def getCountOfStepsForDepth(self, **keywords):
        depthPerStep = keywords['depthPerStep']
        machiningDepth = keywords['machiningDepth']

        useSteps = machiningDepth > depthPerStep > 0.0 and not isNearlyEqual(depthPerStep, 0.0) and not isNearlyEqual(machiningDepth, depthPerStep)
        if useSteps:
            numberOfStepsDivider = math.ceil(machiningDepth / depthPerStep)
            numberOfSteps = max(numberOfStepsDivider, 1)
        else:
            numberOfSteps = 1

        return numberOfSteps

    def getPocketEntranceRadius(self, **keywords):  # Could be used by createPocket() as well as createVerticalContour() -> createFreeFormPocket()
        if 'pocket' in keywords:
            pocket = keywords['pocket']
            pocketDiameter = pocket.diameter
            routerDiameter = pocket.tool.getRouterDiameter()
            securityAddition = routerDiameter / 2
            if pocketDiameter / 2 > routerDiameter + securityAddition:
                return routerDiameter
            else:
                return (routerDiameter + securityAddition * ((pocketDiameter - routerDiameter) / routerDiameter)) / 2
        else:
            return keywords['machining'].tool.getRouterDiameter()

    def getPocketRadiusCompensation(self, **keywords):
        pocket = keywords['pocket']

        if self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION):
            return self.RADIUS_COMPENSATION_RIGHT
        else:  # contour-CCW
            return self.RADIUS_COMPENSATION_LEFT

    def createVerticalPocketBase(self, **keywords):
        pocketDepth = keywords['pocket'].depth
        surfaceThickness = keywords['surface'].thickness
        if pocketDepth < surfaceThickness:
            depth = self.getPocketZValue(depth=pocketDepth, **keywords)
            return self.createVerticalPocket(depth=depth, **keywords)
        else:
            depth = self.getPocketPiercingZValue(**keywords)
            return self.createVerticalPocketPiercing(depth=depth, **keywords)

    def createHorizontalPocketBase(self, **keywords):
        if not self.isFunctionSupportedByWop(function='createHorizontalPocket'):
            surface = keywords['surface']
            pocket = keywords['pocket']
            message = self.getLocalizedMessage(61, "The current export does not support routed horizontal pockets. (Diameter: {diameter}, depth: {depth}, Workpiece: {workpiece})")
            logging.warning(message.format(diameter=pocket.diameter, depth=pocket.depth, workpiece=surface.workpieceName))
        else:
            return self.createHorizontalPocket(**keywords)

    def createVerticalDadoBase(self, **keywords):
        surface = keywords['surface']
        dado = keywords['dado']

        if dado.isAlongXAxis():
            dadoCreator = XAxisDadoCreator(self, surface.workpieceName)
        elif dado.isAlongYAxis():
            dadoCreator = YAxisDadoCreator(self, surface.workpieceName)
        else:
            dadoCreator = CRoratedDadoCreator(self, surface.workpieceName)

        return dadoCreator.createDado(**keywords)

    def createHorizontalDadoBase(self, **keywords):
        surface = keywords['surface']
        dado = keywords['dado']
        message = None

        if dado.tool and dado.tool.type == Tool.ROUTER:
            if not self.isFunctionSupportedByWop(function='createHorizontalDadoWithRouter'):
                message = self.getLocalizedMessage(62, "The current export does not support routed horizontal dados. (Length: {length}, width: {width}, depth: {depth}, Workpiece: {workpiece})")
            else:
                return self.createHorizontalDadoWithRouter(**keywords)
        elif dado.tool and dado.tool.type == Tool.SAW:
            if not self.isFunctionSupportedByWop(function='createHorizontalDadoWithSaw'):
                message = self.getLocalizedMessage(63, "The current export does not support sawn horizontal dados. (Length: {length}, width: {width}, depth: {depth}, Workpiece: {workpiece})")
            else:
                return self.createHorizontalDadoWithSaw(**keywords)
        else:
            message = self.getLocalizedMessage(65, "Suppressing horizontal dado, as no tool was found for it. (Length: {length}, width: {width}, depth: {depth}, Workpiece: {workpiece})")  # TODO IT-227  all dados should be filtered (also vertical) -- NG

        if message:
            logging.warning(message.format(length=dado.length, width=dado.width, depth=dado.depth, workpiece=surface.workpieceName))

    def createVerticalRabbetBase(self, **keywords):
        zValue = self.getDadoRabbetZValue(depth=keywords['rabbet'].depth, **keywords)
        contourDistances = self.getRabbetContourDistances(**keywords)
        return ''.join(self.createVerticalRabbet(contourDistance=contourDistance, zValue=zValue, **keywords) for contourDistance in contourDistances)

    def createVerticalRabbetOffsetByContourDistance(self, **keywords):
        rabbet = keywords['rabbet']
        temporaryRabbet = copy.deepcopy(rabbet)

        if operator.xor(self.isDadoRabbetReplacementCW(rabbet, self.RABBET_REPLACEMENT), not self.isStandardOrigin(**keywords)):
            temporaryRabbet.moveRight(keywords['contourDistance'])
        else:
            temporaryRabbet.moveLeft(keywords['contourDistance'])

        keywords['rabbet'] = temporaryRabbet
        return self.createDefinitionForOneRabbet(**keywords)

    def getRabbetContourDistances(self, **keywords):
        OVERLAP = 0.15
        rabbet = keywords['rabbet']
        toolWidth = rabbet.tool.getRouterDiameter()
        if not isNearlyEqual(toolWidth, 0.0):
            passes = math.floor(rabbet.width / (toolWidth * (1 - OVERLAP))) + 1
            for i in range(passes):
                yield rabbet.width - rabbet.width / passes * (i + 1)
        else:
            workpieceName = keywords['surface'].workpieceName
            rawMessage = self.getLocalizedMessage(48, "Rabbet router diameter is 0. (Workpiece: {workpiece}, Rabbet length: {length}, width: {width}, depth: {depth})")
            message = rawMessage.format(length=rabbet.getLength(), width=rabbet.getWidth(), depth=rabbet.getDepth(), workpiece=workpieceName)
            logging.warning(message)
            yield 0.0

    def getDadoRabbetSecurityDistance(self, **keywords):
        radius = keywords['machining'].tool.getPropertyAsFloat('router-diameter') / 2
        return radius + self.SECURITY_DISTANCE_ADDITION_FOR_DADO_RABBET

    def getDadoLineSegmentWithSecurityDistance(self, **keywords):
        return keywords['dado'].getLineSegmentWithSecurityDistance(securityDistance=self.SECURITY_DISTANCE_FOR_DADO_SAW)

    def __getWarningMessageForMiterIfWrong(self, **keywords):
        warning = None

        miter = keywords['miter']

        if not self.isFunctionSupportedByWop(function='createVerticalMiter'):
            warning = self.getLocalizedMessage(22, 'The current export does not support miters. (Workpiece: {workpiece})')
        elif not hasattr(miter, 'segment') or not miter.segment:
            message = self.getLocalizedMessage(59, "Suppressing miter, as the corresponding edge segement cannot be found! Please check the VW document! Workpiece: {workpiece}")
        elif not hasattr(miter, 'toolID') or not miter.toolID:
            warning = self.getLocalizedMessage(24, 'A miter would have been exported, but no tool ID has been assigned to the miter saw. (Workpiece: {workpiece})')

        if warning:
            miter = keywords['miter']
            surface = keywords['surface']
            return warning.format(length=miter.getLength(), width=miter.getWidth(), angle=miter.getAngle(), workpiece=surface.workpieceName)

    def createVerticalMiterBase(self, **keywords):
        warning = self.__getWarningMessageForMiterIfWrong(**keywords)

        if not warning:
            return self.createVerticalMiter(**keywords)
        else:
            logging.warning(warning)
            return ''

    def isValidContour(self, **keywords):
        def __getCurrentExportScriptName():
            currentScriptPath = interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS[0][0]
            filename, extension = os.path.splitext(os.path.basename(currentScriptPath))
            return filename

        contour = keywords['contour']
        return (
            not contour.hasMachineOptions
            or not contour.selectedMachine
            or contour.selectedMachine == __getCurrentExportScriptName()
            or self.ALLOW_CONTOUR_ON_OTHER_THAN_SELECTED_MACHINE
        )

    def generateCommonMultiUsageContour(self, **keywords):
        contour = keywords['contour']
        for i, data in enumerate(contour.multiUsageRoutings):
            contourPass = keywords['contourPassCreator'](passIndex=i)
            if contourPass.isValid():
                if contourPass.toolID:
                    contourPass.setActualDepth(keywords['surface'], self.getMachineConfiguration(**keywords), contourPass.depth)
                    keywords['machining'] = keywords['contour'] = contourPass
                    yield keywords['creator'](contourDistance=contourPass.currentMultiUsageRouting.distance, **keywords)
                else:
                    if contour.isVertical():
                        message = self.getLocalizedMessage(69, "The current export does not know the tool name '{toolName}' for vertical contours: (Depth: {depth}, Workpiece: {workpiece})")
                    else:
                        message = self.getLocalizedMessage(70, "The current export does not know the tool name '{toolName}' for horizontal contours: (Depth: {depth}, Workpiece: {workpiece})")
                    logging.warning(message.format(toolName=contour.multiUsageRoutings[i].toolName, depth=contourPass.depth, workpiece=keywords['surface'].workpieceName))
            else:
                # TODO CONTOUR2.0 IT-238: better message? -- NG
                message = self.getLocalizedMessage(72, "The pass {i} of the contour does not define a valid contour: (Depth: {depth}, Tool name: {toolName}, Workpiece: {workpiece})")
                logging.warning(message.format(i=i, toolName=contour.multiUsageRoutings[i].toolName, depth=contourPass.depth, workpiece=keywords['surface'].workpieceName))

    def generateMultiUsageContour(self, **keywords):
        return self.generateCommonMultiUsageContour(contourPassCreator=lambda passIndex: keywords['contour'].createCurrentMultiUsageRouting(passIndex=passIndex, machineConfiguration=self.getMachineConfiguration(**keywords)), **keywords)

    def generateMultiUsageContourNoBuiltInDistanceSupport(self, **keywords):
        def contourPassCreator(passIndex):
            return keywords['contour'].createCurrentMultiUsageRoutingMovingPath(passIndex=passIndex, machineConfiguration=self.getMachineConfiguration(**keywords), extraMirroring=not self.isStandardOrigin(**keywords))

        return self.generateCommonMultiUsageContour(contourPassCreator=contourPassCreator, **keywords)

    def generateContourNoBuiltInDepthSteps(self, **keywords):
        contour = keywords['contour']

        machiningDepth = contour.actualDepth
        depthPerStep = contour.currentDepthPerStep
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        yield from self.generateCommonNoBuiltInDepthSteps(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords)

    def generateCommonDepthStepsForNoBuiltInDepthSteps(self, **keywords):
        depthPerStep = keywords['depthPerStep']
        machiningDepth = keywords['machiningDepth']

        if 'sizing' in keywords and not self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS:
            machiningDepth = keywords['surface'].thickness - self.SIZING_DEPTH_ADJUSTMENT

        useSteps = machiningDepth > depthPerStep > 0.0 and not isNearlyEqual(depthPerStep, 0.0) and not isNearlyEqual(machiningDepth, depthPerStep)
        if useSteps:
            numberOfSteps = math.ceil(machiningDepth / depthPerStep)
            rawDepthStep = machiningDepth / numberOfSteps
            assert 0 < rawDepthStep <= depthPerStep or isNearlyEqual(rawDepthStep, depthPerStep), "Calculated step should not be 0 and not be bigger than maximum step!"
            startDepth = rawDepthStep

            for i in range(0, numberOfSteps):
                yield startDepth + i * rawDepthStep
        else:
            yield machiningDepth

    def generateCommonNoBuiltInDepthSteps(self, **keywords):
        boardThickness = keywords['surface'].thickness
        for machiningDepth in self.generateCommonDepthStepsForNoBuiltInDepthSteps(**keywords):
            keywords['machiningDepth'] = machiningDepth
            keywords['zValue'] = keywords['zValueConverter'](boardThickness, machiningDepth)
            yield keywords['creator'](**keywords)

    def createVerticalContourBase(self, **keywords):
        contour = keywords['contour']
        if contour.hasValidMachineOptions():
            if not self.isFunctionSupportedByWop(function='createVerticalMultiUsageContour'):
                surface = keywords['surface']
                message = self.getLocalizedMessage(67, "The current export does not support vertical contours with machine options. (Width: {width}, depth: {depth}, Workpiece: {workpiece})")
                logging.warning(message.format(width=contour.width, depth=contour.depth, workpiece=surface.workpieceName))
            else:
                # TODO CONTOUR2.0 IT-233: should we support contours with just *one* pass by using createVerticalContour directly -- NG
                return self.createVerticalMultiUsageContour(machining=contour, **keywords)
        else:
            return self.createVerticalContour(machining=contour, **keywords)

    def createHorizontalContourBase(self, **keywords):
        contour = keywords['contour']
        if contour.hasValidMachineOptions():
            if contour.hasValidMachineOptions():
                if not self.isFunctionSupportedByWop(function='createHorizontalMultiUsageContour'):
                    surface = keywords['surface']
                    message = self.getLocalizedMessage(68, "The current export does not support horizontal contours with machine options. (Width: {width}, depth: {depth}, Workpiece: {workpiece})")
                    logging.warning(message.format(width=contour.width, depth=contour.depth, workpiece=surface.workpieceName))
                else:
                    # TODO CONTOUR2.0 IT-233: should we support contours with just *one* pass by using createHorizontalContour directly -- NG
                    return self.createHorizontalMultiUsageContour(**keywords)
        else:
            if not self.isFunctionSupportedByWop(function='createHorizontalContour'):
                surface = keywords['surface']
                message = self.getLocalizedMessage(64, "The current export does not support routed horizontal contours. (Width: {width}, depth: {depth}, Workpiece: {workpiece})")
                logging.warning(message.format(width=contour.width, depth=contour.depth, workpiece=surface.workpieceName))
            else:
                return self.createHorizontalContour(**keywords)

    def getDrillingPiercingZValue(self, **keywords):
        return self.getDrillingZValue(depth=keywords['surface'].thickness + self.PIERCING_DRILLING_ADDITION, **keywords)

    def getContourZValue(self, **keywords):
        if keywords['depth'] >= keywords['surface'].thickness:
            return self.getMillThroughZValue(**keywords)
        else:
            return self.getMillingZValue(**keywords)

    def getDadoRabbetZValue(self, **keywords):
        return self.getMillingZValue(**keywords)

    def createHorizontalCenterPunchFromRight(self, **keywords):
        return self.createHorizontalDrillingFromRight(**keywords)

    def createHorizontalCenterPunchFromLeft(self, **keywords):
        return self.createHorizontalDrillingFromLeft(**keywords)

    def createHorizontalCenterPunchFromFront(self, **keywords):
        return self.createHorizontalDrillingFromFront(**keywords)

    def createHorizontalCenterPunchFromBack(self, **keywords):
        return self.createHorizontalDrillingFromBack(**keywords)

    def createRotatedHorizontalCenterPunch(self, **keywords):
        return self.createRotatedHorizontalDrilling(**keywords)

    def export(self, **keywords):
        generalExporter = self.getSurfacesExporter(
            ncexport=self,
            postProcessor=self.createPostProcessor(**keywords),
            workpiecePostProcessor=self.createWorkpiecePostProcessor(**keywords),
            workpieceFilters=self.createWorkpieceFilters(**keywords),
            surfaceOrigin=self.SURFACE_ORIGIN,
            **keywords)
        generalExporter.export(**keywords)

    def getSurfacesExporter(self, **keywords):
        return GeneralExporter(**keywords)

    def isCabinetNameInExportPath(self, **keywords):
        return '{CabinetName}' in self.configuration.getPathForExport()

    def getPIOContainerFileNamePrefix(self, **keywords):  # Legacy -> DosFileNameDependingCabinetName
        if not self.isCabinetNameInExportPath(**keywords):
            return keywords['CabinetName'] + '_'
        else:
            return ''

    def prepareFilePath(self, **keywords):
        surface = keywords['surface']
        filePathWithoutFilename = self.createExportFilePath(workpieceUUID=surface.workpieceUUID, **keywords)
        fileName = self.createFileName(name=surface.workpieceName, extension=self.getNCFileExtension(**keywords), **keywords)
        filepath = os.path.join(filePathWithoutFilename, fileName)
        ExportHelper.CreatePathIfNotExists(filePathWithoutFilename)
        return filepath

    def reduceFileNameByCustomizeablePath(self, **keywords):
        outputFileName = keywords['outputFileName']
        outputFileNameKeys = outputFileName.split('_')
        customizeablePath = self.configuration.customizeablePath

        for key in outputFileNameKeys:
            if key in customizeablePath:
                outputFileName = outputFileName.replace(key, '')

        return outputFileName

    # Default before interiorcad 2019 F3
    # Changed in interiorcad 2023 F5 to need CabinetName as parameter, has we do not have a file(name)Prefix anymore.
    def createDosDependingFileName(self, **keywords):
        if not self.USE_DOS_FILENAMES:
            return self.getPIOContainerFileNamePrefix(**keywords) if keywords['CabinetName'] else ''

    def createPositionNumberForOutputFileNameVariables(self, **keywords):
        positionInfo = self.getPositionInfo(uuid=keywords['surface'].workpieceUUID)
        positionNumber = positionInfo['PositionNumber']
        if self.shouldKeepSignificantDecimalsOnly() and positionNumber.endswith('.0'):
            return positionNumber.rstrip('0').rstrip('.')
        else:
            return positionNumber

    def getPositionOutputFileNameVariables(self, **keywords):
        surface = keywords['surface']
        positionInfo = self.getPositionInfo(uuid=surface.workpieceUUID)
        return {
            'WorkpieceName': surface.workpieceName,
            'CabinetName': positionInfo['CabinetName'],
            'PositionNumber': self.createPositionNumberForOutputFileNameVariables(**keywords),
            'Count': self.asInt(data=surface.count),
        }

    def getOutputFileNameVariables(self, **keywords):
        outputFileNameVariables = dict()
        outputFileNameVariables.update(ExportHelper.getProjectOutputFileNameVariables(self, **keywords))
        outputFileNameVariables.update(self.getPositionOutputFileNameVariables(**keywords))
        outputFileNameVariables['DosFileNameDependingCabinetName'] = self.createDosDependingFileName(CabinetName=outputFileNameVariables['CabinetName'], **keywords)
        outputFileNameVariables.update({'filePostfix': keywords['filePostfix']})
        return outputFileNameVariables

    def getOutputFileNameFromConfiguration(self, **keywords):
        outputFileName = self.OUTPUT_FILE_NAME
        if '{WorkpieceName}' not in outputFileName:  # Fallback if the users OUTPUT_FILE_NAME is customized and returns an empty or not unique name
            outputFileName += '_{WorkpieceName}'
        if '{filePostfix}' not in outputFileName:
            outputFileName += '_{filePostfix}'
        outputFileName = ExportHelper.createReducedOutputFileNameVariables(self, outputFileName=outputFileName, **keywords)
        outputFileName = self.reduceFileNameByCustomizeablePath(outputFileName=outputFileName, **keywords)
        return outputFileName

    def createFileName(self, **keywords):
        fileName = self.getOutputFileNameFromConfiguration(**keywords).format_map(self.getOutputFileNameVariables(**keywords))

        if self.USE_DOS_FILENAMES:
            fileName = DosFilenameCreator().createDosFilename(fileName)  # FilenameCleaner() is called inside DosFilenameCreator()
        else:
            fileName = FilenameAndPathFilter().filenameAndPathFilter(filterTerm=fileName, filterValue=self.OUTPUT_FILE_NAME_FILTER, useIndividualFilters=self.USE_INDIVIDUAL_OUTPUT_NAME_FILTERS)

        fileName += keywords['extension']
        return fileName

    def createExportFilePath(self, **keywords):
        exportPathDict = ExportHelper.createDictionaryForExportPath()
        ExportHelper.updateDictionaryWithPartData(exportPathDict, keywords['workpieceUUID'])
        exportPathDict['SideName'] = self.getSideName(**keywords)
        exportPathDict['SideNameNesting'] = self.getSideNameNesting(**keywords)
        exportPathDict['ExportFormat'] = self.EXPORT_FORMAT if not self.INDIVIDUAL_EXPORT_FORMAT else self.INDIVIDUAL_EXPORT_FORMAT
        exportPath = ExportHelper.createExportPathFromConfigurationWithDictionary(self.configuration, exportPathDict)

        if self.USE_DOS_FILENAMES and '{DosFileNameDependingCabinetName}' in self.OUTPUT_FILE_NAME:
            CabinetName = exportPathDict['CabinetName']
            CabinetName = DosFilenameCreator().filenameCleanerInclDeleteDosExcludes(filename=CabinetName)
            if not self.isCabinetNameInExportPath(**keywords):
                exportPath = os.path.join(exportPath, CabinetName)

        return exportPath

    def getCustomUserVariablesFilePath(self, **keywords):
        return os.path.join(getNCExportInAuxiliaryFilesPath(), keywords['customUserVariableFileName'])

    def getCustomUserVariablesFile(self, **keywords):
        filePath = self.getCustomUserVariablesFilePath(**keywords)

        try:
            with open(filePath, 'r', encoding='utf-8-sig') as f:
                return ''.join(f.readlines())
        except FileNotFoundError:
            with open(filePath, 'w', encoding='utf-8-sig') as f:
                f.write(keywords['customUserVariableFileDefaultText'])

        return ''

    def getMappedCustomUserVariablesFromFile(self, **keywords):
        workpieceComments = self.getWorkpieceComments(**keywords)
        customUserVariables = self.getCustomUserVariablesFile(**keywords)

        try:
            customUserVariables = customUserVariables.format_map(workpieceComments)
            return customUserVariables
        except KeyError as e:
            path = self.getCustomUserVariablesFilePath(**keywords)
            message = self.getLocalizedMessage(77, "Key {key} of custom file {path} is wrong! Accepted keys: {keys}.")
            raise KeyError(message.format(key=str(e), keys=tuple(workpieceComments.keys()), path=path)) from e

    def getMappedCustomUserVariablesFromStringWithKeys(self, **keywords):
        workpieceComments = self.getWorkpieceComments(**keywords)
        customUserVariables = keywords['stringWithKeys']

        try:
            customUserVariables = customUserVariables.format_map(workpieceComments)
            return customUserVariables
        except KeyError as e:
            message = self.getLocalizedMessage(80, "Key {key} of custom field '{field}' is wrong! Accepted keys: {keys}.")
            raise KeyError(message.format(key=str(e), keys=tuple(workpieceComments.keys()), field=customUserVariables)) from e

    def getSideName(self, **keywords):
        sideType = keywords['surface'].sideType
        if sideType == SURFACE_SIDE_TYPE_MAIN:
            return getLocalizedString(SIDE_NAMES_RESOURCEGROUP, 2, 'Main')
        elif sideType == SURFACE_SIDE_TYPE_OPPOSITE:
            return getLocalizedString(SIDE_NAMES_RESOURCEGROUP, 3, 'Opposite')
        elif sideType == SURFACE_SIDE_TYPE_SIZING:
            return getLocalizedString(SIDE_NAMES_RESOURCEGROUP, 1, 'Sizing')
        elif sideType == SURFACE_SIDE_TYPE_BOTH_SIDED:
            return getLocalizedString(SIDE_NAMES_RESOURCEGROUP, 4, 'BothSided')
        else:
            assert False, 'Unknown surface side type:' + str(sideType)

    def getSideNameNesting(self, **keywords):
        if keywords['surface'].isFirstSide:
            return getLocalizedString(SIDE_NAMES_RESOURCEGROUP, 5, 'Nesting')
        else:
            return self.getSideName(**keywords)

    def createSide(self, **keywords):
        self.adaptFinalFinishedSizeShape(**keywords)
        self.orientMachinings(**keywords)
        fileContent = self.createExportFileContentFor(**keywords)
        self.writeFileContent(fileContent=fileContent, **keywords)

    # fileContent should not be part of **keywords to avoid recursion problems
    def writeFileContent(self, fileContent, **keywords):
        filePath = self.prepareFilePath(**keywords)
        self.writeToFile(fileContent, filePath)

        realFilePath = self.postProcessFile(filePath=filePath, **keywords)

        keywords['ncProgramNameWriteBackFunc'](realFilePath)

    # New in Epic ID-681 (replacing isSizingToolRightHand)
    def isSizingMillingDirectionLeft(self, **keywords):
        if not keywords['sizing']:
            return True
        elif self.SIZING_MILLING_DIRECTION == 'opposite':
            return (keywords['sizing'].tool.getRouterRotationDirection() == 'right'
                or not bool(keywords['sizing'].tool.getRouterRotationDirection()))  # Support older (test) tools that do not have a Router Rotation Direction
        elif self.SIZING_MILLING_DIRECTION == 'same':
            return (keywords['sizing'].tool.getRouterRotationDirection() == 'left'
                and bool(keywords['sizing'].tool.getRouterRotationDirection()))  # Support older (test) tools that do not have a Router Rotation Direction
        elif self.SIZING_MILLING_DIRECTION == '1':
            return True
        elif self.SIZING_MILLING_DIRECTION == '0':
            return False
        else:
            assert False, "Unknown SIZING_MILLING_DIRECTION"

    def orientMachinings(self, **keywords):
        self.orientSizing(**keywords)
        self.orientMiters(**keywords)
        self.orientRabbets(**keywords)

    def isStandardOrigin(self, **_keywords):
        return isStandardOrigin(self.SURFACE_ORIGIN)

    def orientSizing(self, **keywords):
        surface = keywords['surface']
        if surface.hasSizing():
            if operator.xor(self.isSizingMillingDirectionLeft(sizing=surface.sizing, **keywords), not self.isStandardOrigin(**keywords)):
                surface.changeSizingDirection()

    def adaptFinalFinishedSizeShape(self, **keywords):
        if not self.isStandardOrigin(**keywords):
            keywords['surface'].finalFinishedSizeShape.changeDirection()

    def createExportFileContentFor(self, **keywords):
        self.generateNCStopBeforeNextContour = True
        self.generateNCStopBeforeNextHorizontalDrilling = True
        return ''.join((
            self.createHead(**keywords),
            self.createMachinings(**keywords),
            self.createFoot(**keywords),
        ))

    def postProcessFile(self, **keywords):
        return keywords['filePath']

    def createWorkpiecePostProcessor(self, **keywords):
        postProcessor = PostProcessors()
        if self.ALLOW_VERTICAL_PIERCING_MACHININGS_TO_CHANGE_SIDES:
            postProcessor.append(PiercingVerticalMachiningsSideChanger())
        return postProcessor

    def createWorkpieceFilters(self, **keywords):
        workpieceFilters = WorkpieceFilters(
            InvalidPathWorkpieceFilter(),
            NullMachiningFilter(),
        )

        if not isNearlyEqual(self.MIN_LENGTH_OF_WORKPIECE, 0) or not isNearlyEqual(self.MIN_WIDTH_OF_WORKPIECE, 0):
            workpieceFilters.append(SmallWorkpieceFilter(self.configuration, self.MIN_LENGTH_OF_WORKPIECE, self.MIN_WIDTH_OF_WORKPIECE))

        exportMachiningsInfo = {
            'verticalDrillings': self.EXPORT_VERTICAL_DRILLINGS,
            'horizontalDrillings': self.EXPORT_HORIZONTAL_DRILLINGS,
            'fiveAxisDrillings': self.EXPORT_FIVE_AXIS_DRILLINGS,
            'miters': self.EXPORT_MITERS,
            'verticalContours': self.EXPORT_VERTICAL_CONTOURS,
            'horizontalContours': self.EXPORT_HORIZONTAL_CONTOURS,
            'verticalDados': self.EXPORT_VERTICAL_DADOS,
            'horizontalDados': self.EXPORT_HORIZONTAL_DADOS,
            'verticalRabbets': self.EXPORT_VERTICAL_RABBETS,
            'verticalNCMacros': self.EXPORT_VERTICAL_NC_MACROS,
            'horizontalNCMacros': self.EXPORT_HORIZONAL_NC_MACROS,
            'fiveAxisNCMacros': self.EXPORT_FIVE_AXIS_NC_MACROS,
        }
        workpieceFilters.append(WorkpieceMachiningsFilter(exportMachiningsInfo))

        if not self.EXPORT_EMPTY_WORKPIECE:
            workpieceFilters.append(EmptyWorkpieceFilter(self.configuration))

        return workpieceFilters

    def createToolDatabasePostProcessors(self, **keywords):
        machineConfiguration = self.getMachineConfiguration()

        yield ChamferWarner()
        yield ProfileWarner()

        xAxisGridDrillDiameters = tuple(machineConfiguration.toolDatabase.getSupportedGridDrillingDiametersAlongX())
        yAxisGridDrillDiameters = tuple(machineConfiguration.toolDatabase.getSupportedGridDrillingDiametersAlongY())
        cRotatedAxisDrillDiameters = tuple(machineConfiguration.toolDatabase.getSupportedGridDrillingDiametersAlongCRotated())
        yield VerticalGridDrillingsReplacer(
            xaxis=self.areVerticalGridDrillingsAlongXAxisSupported(), xAxisGridDrillDiameters=xAxisGridDrillDiameters,
            yaxis=self.areVerticalGridDrillingsAlongYAxisSupported(), yAxisGridDrillDiameters=yAxisGridDrillDiameters,
            crotatedaxis=self.areVerticalGridDrillingsAlongCRotatedAxisSupported(), crotatedAxisGridDrillingDiameters=cRotatedAxisDrillDiameters
        )
        yield VerticalGridDrillingsFilter(xaxis=self.areVerticalGridDrillingsAlongXAxisSupported(), yaxis=self.areVerticalGridDrillingsAlongYAxisSupported(), xAxisGridDrillDiameters=xAxisGridDrillDiameters, yAxisGridDrillDiameters=yAxisGridDrillDiameters, **self.VERTICAL_GRID_DRILLING_CONFIGURATION)

        yield ToolAssigner(machineConfiguration)
        yield from self.getPostAssignToolProcessors(**keywords)

        yield ClamexCAngleProcessor(self.VERTICAL_CLAMEX_MAX_DISTANCE_FROM_EDGE, self.VERTICAL_CLAMEX_INNER_PLACEMENT_EXPORT)

        yield PocketFilter(machineConfiguration)
        yield VerticalDrillingsFilter(allowSmallDrillingsWithoutToolID=self.ALLOW_SMALL_VERTICAL_DRILLINGS_WITHOUT_TOOL_ID)
        yield NonVerticalDrillingsFilter()

        yield VerticalCenterPunchFilter()
        yield HorizontalCenterPunchFilter()
        yield FiveAxisCenterPunchFilter()

        yield ToolExportMachiningsFilter()

        yield DrillingsSorter(self.configuration.drillingStartPosition, self.DRILLING_ORDERINGS)
        yield DadosSorter()

        yield RabbetToolFilter()
        yield RabbetsSorter()

        yield ContourExportFilter(self)
        yield ContourToolFilter(machineConfiguration)
        yield ContourApproachFilter()

        yield LicenseFilter(self.LICENSE_INFO)

        yield SizingToolFilter(self.getMachineConfiguration(**keywords))

        if not self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR:
            yield SizingApproachProcessor(self.configuration.minimumSegmentLength, isOriginFront(self.SURFACE_ORIGIN))
        else:
            yield SizingThreeSideContourProcessor(isOriginFront(self.SURFACE_ORIGIN))

        yield SizingContourProcessor()

    def createPostProcessor(self, **keywords):
        postProcessor = PostProcessors()
        postProcessor.extend(self.getPrePostProcessors(**keywords))
        postProcessor.extend(self.createToolDatabasePostProcessors(**keywords))
        postProcessor.extend(self.getPostPostProcessors(**keywords))
        return postProcessor

    def getPrePostProcessors(self, **_keywords):
        return ()

    def getPostAssignToolProcessors(self, **_keywords):
        return ()

    def getPostPostProcessors(self, **_keywords):
        return ()

    def openFileForWritingText(self, filepath):
        return open(filepath, mode='wt', newline=self.OUTPUT_END_OF_LINE, encoding=self.OUTPUT_ENCODING, errors=self.OUTPUT_ENCODING_ERROR_MODE)

    def writeToFile(self, content, filepath):
        with self.openFileForWritingText(filepath) as exportFile:
            exportFile.write(content)

    # The function does more than its (old) name, but we cannot change the name, as the function is called directly by every NC export starter.
    def openOutputPath(self, **keywords):
        ExportHelper.openExportPath(self.configuration)
        self.openExportLogFile(**keywords)

    def openExportLogFile(self, **keywords):
        if self.loggingHandler:
            self.removeLoggingHandle()
            if self.logFileName and os.path.exists(self.logFileName):
                # "Normalize" the path, as it contains forward slashes (/)
                interiorcad.backend.openservices.openFile(os.path.normpath(self.logFileName))

    def exportCuttingLists(self, **keywords):
        if interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS:
            interiorcad.backend.exportdatasafe.FIRST_EXPORT = False  # Explicitly set to False for cuttinglistexport inside ncexport.
            interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS = interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS[1:]
            runpy.run_path(interiorcad.backend.exportdatasafe.SCRIPTS_AND_CONFIGURATIONS[0][0])
        else:
            logging.warning(self.getLocalizedMessage(32, "Export of cutting list wanted, but no path given!"))

    def __createVerticalContourHack(self, **keywords):
        # TODO: can it be implemented in a less hacky way -- NG (InnoWeek 2017-12)
        class OneContourExporter(ContoursExporter):
            def getMachinings(self, **keywords):
                return machining,

            def areConditionsRelevantForAStop(self, **keywords):
                return ncstop

        machining = keywords['machining']
        ncstop = self.generateNCStopBeforeNextContour

        del keywords['contour']
        del keywords['machining']
        exporter = OneContourExporter(self)
        return exporter.create(**keywords)

    def createVerticalContours(self, **keywords):
        exporter = ContoursExporter(self)
        return exporter.create(**keywords)

    def createNCMacros(self, **keywords):
        return self.__createMachinings(keywords['surface'].ncmacros, exportMethod=self.createNCMacroBase, **keywords)

    def createNCMacroBase(self, **keywords):
        ncmacro = keywords['ncmacro']
        if not ncmacro.hasCode() and not ncmacro.name:
            warning = self.getLocalizedMessage(39, "Export of NC macro suppressed, as the NC macro has an empty name. Workpiece: {workpiece}")
            logging.warning(warning.format(workpiece=keywords['surface'].workpieceName, name=ncmacro.name))
            return ''
        elif ncmacro.hasCode():
            return self.createNCMacroCode(**keywords)
        else:
            if hasattr(self, 'createNCMacro'):  # up to interiorcad 2020 F1 - do NOT use for new code
                return self.createNCMacro(**keywords)
            elif hasattr(self, 'createNCMacroRecordFormat'):  # from interiorcad 2020 F2
                return self.createNCMacroRecordFormat(**keywords)
            else:
                self.warnNCMacroNotSupported(**keywords)
                return ''

    def warnNCMacroNotSupported(self, **keywords):
        warning = self.getLocalizedMessage(38, "NC macro found, but this NC export does not support NC macros. Workpiece: {workpiece}, NC Macro Name: {name}")
        logging.warning(warning.format(workpiece=keywords['surface'].workpieceName, name=keywords['ncmacro'].name))

    def getNCMacroCookedVariablesForCode(self, **keywords):
        """Cooked variables, only for code-based NCMacros"""

        variables = self.getNCMacroCookedVariables(**keywords)
        variables[CLAMEX_C_ANGLE] = keywords['ncmacro'].clamexcangle
        variables[CLAMEX_C_ANGLE_180] = keywords['ncmacro'].clamexcangle180

        return variables

    def getNCMacroCookedVariables(self, **keywords):
        """Cooked variables, only for both RecordFormat-based and code-based NCMacros"""

        ncmacro = keywords['ncmacro']
        surface = keywords['surface']

        variables = ncmacro.getVariables(formatter=self)
        variables['thickness'] = self.formatAsCoordinate(data=surface.thickness)
        variables['name'] = ncmacro.name if ncmacro.name else getLocalizedString(NCEXPORTBASE_MESSAGES, 1, 'NC Macro')

        return variables

    def getNCMacroCookedParametersWithSubstitutedValues(self, **keywords):
        for parameter in keywords['ncmacro'].parameters:
            if ':' in parameter.value:  # parameter values in the form of "key: value" will be reduced to "value" only.
                cookedValue = parameter.value[parameter.value.find(':') + 1:].lstrip()
            else:
                cookedValue = parameter.value
            yield Parameter(name=parameter.name, value=cookedValue)

    def createNCMacroCode(self, **keywords):
        ncmacro = keywords['ncmacro']
        variables = self.getNCMacroCookedVariablesForCode(**keywords)
        variables.update(self.getNCMacroCookedParametersWithSubstitutedValues(**keywords))
        code = ncmacro.code.format_map(variables)
        # TODO IT-247 to \n or not to \n ? -- NG
        return code + '\n'

    # Function for older createNCMacroRecordFormat(); it's recommended to use the newer and more flexible createNCMacroCode() instead.
    def getNCMacroCookedParameters(self, **keywords):
        variables = self.getNCMacroCookedVariables(**keywords)
        for parameter in keywords['ncmacro'].parameters:
            yield Parameter(name=parameter.name.format_map(variables), value=parameter.value.format_map(variables))

    def __createSizingHack(self, **keywords):
        del keywords['machining']  # TODO: unfortunately the code called by createSizing sets 'machining' by itself -- NG (InnoWeek 2017-12)
        return self.createSizing(**keywords)

    def createSizing(self, **keywords):
        if self.EXPORT_SIZINGS and keywords['surface'].hasSizing():
            return self.createSizingContour(**keywords)

    def __createMachinings(self, machinings, **keywords):
        return ''.join((self.__createMachining(machining=machining, **keywords) for machining in machinings))

    def __createMachining(self, **keywords):
        machining = keywords['machining']
        if hasattr(machining, 'getExportKey'):  # Sizing2D and its code works differently for defining 'sizing'
            keywords[machining.getExportKey()] = machining
        lines = keywords['exportMethod'](**keywords)
        self.generateNCStopBeforeNextContour = not isinstance(machining, Contour2D)  # FIXME: it seems curios to have a generic function that do something special for a certain class -- NG (InnoWeek 2017-12)
        return lines

    DIN_RADIUS_COMPENSATION_NONE = 0
    DIN_RADIUS_COMPENSATION_LEFT = 1
    DIN_RADIUS_COMPENSATION_RIGHT = 2

    def getDinRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=self.DIN_RADIUS_COMPENSATION_NONE, left=self.DIN_RADIUS_COMPENSATION_LEFT, right=self.DIN_RADIUS_COMPENSATION_RIGHT)

    def createFoot(self, **keywords):
        return self.createEndOfFile(**keywords)

    def getLocalizedMessage(self, index, defaultText):
        return getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, index, defaultText)
