# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import json
import logging
import math
import operator
import os.path
from collections import OrderedDict

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.exportmainpath import getNCExportInAuxiliaryFilesPath
from interiorcad.backend.jsonfilewriter import JSONFileWriter
from interiorcad.backend.localization import getLocalizedString, BIESSE_CIX_NAME_MESSAGES, BIESSE_CIX_MESSAGES, NCEXPORT_MESSAGES_RESOURCEGROUP
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual, isNearlyGreater, isNearlyGreaterOrEqual, isNearlyLesserOrEqual, isNearlyLesser
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.machiningsexporter.horizontaldrillingsexporter import HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds
from interiorcad.interface.surface.acangles import normalizeAngle
from interiorcad.interface.surface.contour2d import Contour2D
from interiorcad.interface.surface.machiningplanes import MachiningPlanes
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.interface.surface.origins import SURFACE_ORIGIN_BACK_LEFT, SURFACE_ORIGIN_BACK_RIGHT, SURFACE_ORIGIN_FRONT_LEFT, SURFACE_ORIGIN_FRONT_RIGHT
from interiorcad.ncexport.base import NCExportBase


class BiesseCIXExport(NCExportBase):
    EXPORT_FORMAT = 'BiesseCIX'

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.lastPocketIndex = None
        self.lastSizingIndex = None
        self.lastRabbetIndex = None
        self.lastContourIndex = None
        self.lastDadoIndex = None
        self.lastSequentialID = None
        self.geometryReferenceCounts = None
        self.machiningPlanes = None

        machineConfiguration = self.getMachineConfiguration()
        self.ZERO_OFFSET_Y_IS_SIZING_ADDITION = machineConfiguration.getAsBool('zero-offset-y-is-sizing-addition', True)
        self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR = machineConfiguration.getAsBool('create-sizing-with-three-side-contour', False)
        self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS = machineConfiguration.getAsBool('sizing-mill-through-for-non-rectangular-parts', True)
        self.SIZING_FIRST_PASS_REMAINING_THICKNESS = machineConfiguration.getAsFloat('sizing-first-pass-remaining-thickness', 2)
        self.SIZING_PASS_STRATEGY = machineConfiguration.get('sizing-pass-strategy', 'standard')
        self.SIZING_PASS_STRATEGY_AREA_THRESHOLD = machineConfiguration.getAsFloat('sizing-pass-strategy-area-threshold', 0.1)

        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 2)
        self.ADDITIONAL_DEPTH_FOR_MITER = machineConfiguration.getAsFloat(MachineConfigurationKeys.ADDITIONAL_DEPTH_FOR_MITER_KEY, 10)
        self.DADO_PRESAW_DEPTH = machineConfiguration.getAsFloat('dado-presaw-depth', 0)

        self.SIZING_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('sizing-approach-departure-distance', 30)
        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('contour-approach-departure-distance', 30)

        self.OPTIMISE_SUCTION_CUPS = machineConfiguration.getAsBool('optimise-suction-cups', True)  # 1 = YES = automatic placement

        # Hood positions: value from 1 to 6, 0 enables automatic positioning of the hoods
        self.HOOD_POSITION_DRILLING = machineConfiguration.getAsInt('hood-position-drilling', 0)
        self.HOOD_POSITION_MILLINGS = machineConfiguration.getAsInt('hood-position-millings', 0)
        self.HOOD_POSITION_SAW_CUTS = machineConfiguration.getAsInt('hood-position-saw-cuts', 0)

        idTypeDefinition = machineConfiguration.get('id-type-definition', 'geoname georefname simplesequence')  # "geoname", "geosequence", "georefname", "georefsequence", "sawsequence", "simplesequence",
        self.ID_TYPES = idTypeDefinition.split()

        self.WOP_WORKING_AREA_ORIGINS_LIST = machineConfiguration.get('wop-working-area-origins-list', '1')
        self.USE_INDIVIDUAL_WORK_TABLE_ORIGINS = machineConfiguration.getAsBool('use-individual-work-table-origins', False)

        try:
            self.WOP_WORKING_AREA_ORIGIN = self.WOP_WORKING_AREA_ORIGINS_LIST.split(',')[0]
        except ValueError:
            logging.warning(getLocalizedString(BIESSE_CIX_MESSAGES, 1, "Incorrect List of origins. Assuming origin 1"))
            self.WOP_WORKING_AREA_ORIGINS_LIST = '1'
            self.WOP_WORKING_AREA_ORIGIN = '1'

        self.SURFACE_ORIGIN_RAW = machineConfiguration.get('point-of-origin', 'back left')
        self.SURFACE_ORIGIN, self.WOP_ORIGIN = self.getSurfaceOrigin(surfaceOriginRaw=self.SURFACE_ORIGIN_RAW)

        self.RADIUS_SEGMENT_MACRO = machineConfiguration.get('radius-segment-macro', '')

        self.createZValueConverters(**keywords)

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)

    def getSurfaceOrigin(self, **keywords):
        surfaceOriginRaw = keywords['surfaceOriginRaw'].lower()
        if surfaceOriginRaw and 'back' in surfaceOriginRaw and 'left' in surfaceOriginRaw:
            return SURFACE_ORIGIN_BACK_LEFT, self.WOP_ORIGIN_BACK_LEFT
        elif surfaceOriginRaw and 'back' in surfaceOriginRaw and 'right' in surfaceOriginRaw:
            return SURFACE_ORIGIN_BACK_RIGHT, self.WOP_ORIGIN_BACK_RIGHT
        elif surfaceOriginRaw and 'front' in surfaceOriginRaw and 'left' in surfaceOriginRaw:
            return SURFACE_ORIGIN_FRONT_LEFT, self.WOP_ORIGIN_FRONT_LEFT
        elif surfaceOriginRaw and 'front' in surfaceOriginRaw and 'right' in surfaceOriginRaw:
            return SURFACE_ORIGIN_FRONT_RIGHT, self.WOP_ORIGIN_FRONT_RIGHT
        elif surfaceOriginRaw == 'working area origin':
            return self.getMathematicalOrigin(**keywords)
        else:
            logging.warning(getLocalizedString(BIESSE_CIX_MESSAGES, 2, "Unknown surface origin mode. Assuming back/left"))
            return SURFACE_ORIGIN_FRONT_LEFT, self.WOP_ORIGIN_FRONT_LEFT

    def defaultWorkTableOriginsDictionary(self):
        return json.loads(r"""{
            "_ReadMe": "Note the original JSON and predefined file syntax when making changes. Only the following abbreviations are allowed: BL = Back Left, FL = Front Left, FR = Front Right, BR = Back Right",
            "_Origin Number": "Mathematical Origin (corner as reference point), Origin of the work table",
            "1": "BL,BL",
            "2": "BR,BR",
            "3": "BL,BL",
            "4": "BR,BR",
            "5": "BL,BL",
            "6": "BR,BR",
            "7": "BL,BL",
            "8": "BR,BR",
            "9": "BL,BL",
            "10": "BR,BR",
            "11": "BL,BL",
            "12": "BR,BR",
            "13": "FL,FL",
            "14": "FR,FR",
            "15": "FL,FL",
            "16": "FR,FR"
        }""", object_pairs_hook=OrderedDict)

    def useIndividualWorkTableOrigins(self):
        return self.USE_INDIVIDUAL_WORK_TABLE_ORIGINS and self.WOP_WORKING_AREA_ORIGINS_LIST and self.SURFACE_ORIGIN_RAW == 'working area origin'

    def getWorkTableOriginsDictionary(self, **_keywords):
        filePath = os.path.join(getNCExportInAuxiliaryFilesPath(), 'BiesseCIX Work Table Origins Dictionary.json')

        WorkTableOriginsDict = self.defaultWorkTableOriginsDictionary()
        try:
            with open(filePath, 'r') as handle:
                WorkTableOriginsDict = json.load(handle, object_pairs_hook=OrderedDict)
        except FileNotFoundError:
            if self.useIndividualWorkTableOrigins():
                JSONFileWriter().writeFile(filePath, WorkTableOriginsDict)
        except Exception as e:
            logging.warning(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 81, "Error while reading the JSON file '{filepath}': {exception}").format(filepath=filePath, exception=str(e)))
            logging.info(getLocalizedString(NCEXPORT_MESSAGES_RESOURCEGROUP, 82, "Note the original JSON syntax when making changes. The JSON file will be re-created if you delete it."))

        return WorkTableOriginsDict

    def getWorkTableOrigins(self, **keywords):
        SURFACE_AND_WOP_ORIGIN_TRANSLATIONS = {
            'FL': (SURFACE_ORIGIN_FRONT_LEFT, self.WOP_ORIGIN_FRONT_LEFT),
            'FR': (SURFACE_ORIGIN_FRONT_RIGHT, self.WOP_ORIGIN_FRONT_RIGHT),
            'BL': (SURFACE_ORIGIN_BACK_LEFT, self.WOP_ORIGIN_BACK_LEFT),
            'BR': (SURFACE_ORIGIN_BACK_RIGHT, self.WOP_ORIGIN_BACK_RIGHT),
        }

        surfaceAndWopOrigin, workTableOrigin = self.getWorkTableOriginsDictionary()[self.WOP_WORKING_AREA_ORIGIN].split(',')
        if 'key' in keywords:
            key = keywords['key']
            if key == 'surface and wop origin':
                return SURFACE_AND_WOP_ORIGIN_TRANSLATIONS[surfaceAndWopOrigin]
            elif key == 'work table origin':
                return workTableOrigin
            else:
                assert False, 'Key must be either "surface and wop origin" or "work table origin"!'
        else:
            return SURFACE_AND_WOP_ORIGIN_TRANSLATIONS[surfaceAndWopOrigin], workTableOrigin

    def getMathematicalOrigin(self, **keywords):
        try:
            return self.getWorkTableOrigins(key='surface and wop origin', **keywords)
        except:
            logging.warning(getLocalizedString(BIESSE_CIX_MESSAGES, 3, "Unknown surface origin number. Assuming back/left"))
            return SURFACE_ORIGIN_BACK_LEFT, self.WOP_ORIGIN_BACK_LEFT

    def getNCFileExtension(self, **_keywords):
        return '.cix'

    SIDE_MAIN = 0
    SIDE_LEFT = 1
    SIDE_FRONT = 2
    SIDE_RIGHT = 3
    SIDE_BACK = 4
    SIDE_OPPOSITE = 5

    WOP_ORIGIN_BACK_LEFT = 1
    WOP_ORIGIN_FRONT_LEFT = 2
    WOP_ORIGIN_FRONT_RIGHT = 3
    WOP_ORIGIN_BACK_RIGHT = 4

    REPEAT_NO = 'rpNO'
    REPEAT_X = 'rpX'
    REPEAT_Y = 'rpY'
    REPEAT_XY = 'rpXY'
    REPEAT_CIRCUMFERENCE = 'rpCIR'
    REPEAT_ANGLED_STRAIGHT_LINE = 'rpAL'

    INCLINATION_NO = 'azrNO'
    INCLINATION_ABSOLUTE = 'azrABS'
    INCLINATION_INCREMEMTAL = 'azrINC'

    FLAT_DRILL_TYPE = 0
    PIERCING_DRILL_TYPE = 1

    TYPE_DRILLING = 0
    TYPE_ROUTING = 1
    TYPE_CUTTING = 2

    CUT_XY = 'cutXY'
    CUT_XA = 'cutXA'
    CUT_YA = 'cutYA'
    CUT_LA = 'cutLA'

    # TODO: why is left and right documented the other way round for CRC, but it seems to work that way nevertheless -- NG (IT-210)
    # Careful: in BiesseWorks the German translation of Left and right seem to be interchanged (IT-244) -- NG
    RADIUS_COMPENSATION_NONE = 0  # Central
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    CW = 'dirCW'
    CCW = 'dirCCW'

    APPROACH_NONE = 0
    APPROACH_CURVE = 1
    APPROACH_LINE = 2
    APPROACH_HELIX = 5
    APPROACH_CORRECTED_3DLINE = 7

    APPROACH_MODE_NONE = APPROACH_NONE
    APPROACH_MODE_LINEAR_PIERCING = APPROACH_LINE
    APPROACH_MODE_LINEAR_FLYING = APPROACH_CORRECTED_3DLINE
    APPROACH_MODE_ARC_PIERCING = APPROACH_CURVE
    APPROACH_MODE_ARC_FLYING = APPROACH_HELIX
    APPROACH_MODE_WITHOUT = APPROACH_NONE
    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_LINEAR_FLYING

    # CIX specific macros and Geometry
    def getCIXBoolean(self, **keywords):
        return 'YES' if keywords['boolean'] else 'NO'

    def getSequentialID(self, **_keywords):
        self.lastSequentialID += 1  # BiesseWorks and bSolid number automatically "P1000" as string
        return 'P{sequence}'.format(sequence=self.lastSequentialID)

    def generateSimpleIDTuple(self, **keywords):
        if 'simplesequence' in self.ID_TYPES:
            return 'ID', '"{id}"'.format(id=self.getSequentialID(**keywords))

    def getGeometryID(self, **keywords):
        if 'geosequence' in self.ID_TYPES:
            identifier = self.getSequentialID(**keywords)
        else:
            identifier = keywords['proposedNamedID']

        assert identifier, "ID cannot be empty for a GEO, as it cannot be referenced that way"
        return identifier

    def generateGeometryReferenceIDTuple(self, **keywords):
        identifier = None
        if 'georefname' in self.ID_TYPES:
            gid = keywords['gid']
            if gid in self.geometryReferenceCounts:
                self.geometryReferenceCounts[gid] += 1
                count = self.geometryReferenceCounts[gid]
            else:
                count = 1
                self.geometryReferenceCounts[gid] = count
            identifier = "{gid}_{count}".format(gid=keywords['gid'], count=count)
        elif 'georefsequence' in self.ID_TYPES:
            identifier = self.getSequentialID(**keywords)

        if identifier:
            return 'ID', '"{identifier}"'.format(identifier=identifier)

    def getSawnMachiningID(self, **keywords):
        # CUT_X, CUT_Y and CUT_G do support ID, neither in BiesseWorks (error message) nor in bSolid
        if 'sawsequence' in self.ID_TYPES:
            return self.getSequentialID(**keywords)

    def createCIXMacro(self, **keywords):
        def generate():
            yield 'BEGIN MACRO'
            yield '\tNAME={name}'.format(name=keywords['macroName'])
            for name, value in filter(lambda entry: entry is not None, keywords['namesAndValues']):
                yield '\tPARAM,NAME={name},VALUE={value}'.format(name=name, value=value)
            yield 'END MACRO'
            yield ''
            yield ''

        return '\n'.join((generate()))

    def createCIXGeoMacro(self, **keywords):
        macroName = 'GEO'
        namesAndValues = (
            ('ID', '"{id}"'.format(id=keywords['gid'])),
            ('SIDE', keywords['sideID'] if 'sideID' in keywords else self.SIDE_MAIN),
            ('CRN', '"{stop}"'.format(stop=keywords['wopOrigin'] if 'wopOrigin' in keywords else self.WOP_ORIGIN)),
            ('DP', self.formatAsCoordinate(data=keywords['depth'])),
            ('RTY', self.REPEAT_NO),
            ('XRC', 0),
            ('YRC', 0),
            ('DX', 0),
            ('DY', 0),
            ('R', 0),
            ('A', 0),
            ('DA', 0),
            ('RDL', self.getCIXBoolean(boolean=False)),
            ('NRP', 0),
            ('ARP', 0),
            ('LRP', 0),
            ('ER', self.getCIXBoolean(boolean=True)),
            ('RV', self.getCIXBoolean(boolean=False)),
            ('COW', self.getCIXBoolean(boolean=False)),  # For Skipper machine only
            ('LAY', '"{layer}"'.format(layer=macroName)),  # the convention is to use the macro name as layer name
            ('CRC', keywords['radiusCompensation']),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createCIXStartPointMacro(self, **keywords):
        macroName = 'START_POINT'
        namesAndValues = (
            self.generateSimpleIDTuple(**keywords),
            ('X', self.formatAsCoordinate(data=keywords['x'])),
            ('Y', self.formatAsCoordinate(data=keywords['y'])),
            ('Z', 0),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createCIXEndPathMacro(self, **keywords):
        macroName = 'ENDPATH'
        namesAndValues = (
            self.generateSimpleIDTuple(**keywords),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createCIXCircleCR(self, **keywords):
        center = keywords['center']

        macroName = 'CIRCLE_CR'
        namesAndValues = (
            self.generateSimpleIDTuple(**keywords),
            ('XC', self.formatAsCoordinate(data=center.x)),
            ('YC', self.formatAsCoordinate(data=center.y)),
            ('R', self.formatAsCoordinate(data=keywords['radius'])),
            ('AS', 0),
            ('DIR', self.CW),
            ('FD', 0),
            ('SP', 0),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createCIXStraightSide(self, **keywords):
        surface = keywords['surface']
        plane = keywords['plane']

        sideID = plane.sideID
        assert sideID <= 65535, "Too many virtual sides"

        # We cannot use zValueConverter, as zPosition is considered from bottom, but ZValueConverter expects a value from top
        planePosition = copy.deepcopy(plane.wopMatrix.origin)
        planePosition.y = surface.getFinishedSizeY() - plane.wopMatrix.origin.y  # relative to CRN = self.WOP_ORIGIN_BACK_LEFT
        planePosition.z = surface.thickness - plane.wopMatrix.origin.z

        macroName = 'WFL'
        namesAndValues = (
            ('ID', sideID),
            ('X', self.formatAsCoordinate(data=planePosition.x)),
            ('Y', self.formatAsCoordinate(data=planePosition.y)),
            ('Z', self.formatAsCoordinate(data=planePosition.z)),
            ('AZ', self.formatAsCoordinate(data=90 - plane.aAngle)),
            ('AR', self.formatAsCoordinate(data=normalizeAngle(360 - plane.cAngleToXYPlane + 90))),
            ('L', self.formatAsCoordinate(data=keywords['fakeVirtualFaceLength'])),
            ('H', self.formatAsCoordinate(data=keywords['fakeVirtualFaceHeight'])),
            ('VRT', self.getCIXBoolean(boolean=isNearlyEqual(plane.aAngle, 90.0))),
            ('VF', self.getCIXBoolean(boolean=True)),
            ('AFL', self.getCIXBoolean(boolean=False)),
            ('AFH', self.getCIXBoolean(boolean=False)),
            ('UCS', self.getCIXBoolean(boolean=True)),
            ('RV', self.getCIXBoolean(boolean=False)),
            ('FRC', self.WOP_ORIGIN_FRONT_LEFT),
            ('LAY', '"{layer}"'.format(layer=macroName)),  # the convention is to use the macro name as layer name
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createCIXWait(self, **keywords):
        macroName = 'WAIT'
        namesAndValues = (
            # According to documentation there is no ID
            ('TYP', 'stNT'),
            ('OG', 1),
            ('RT', '0'),
            ('MR', 'mrrNO'),
            ('UK', self.getCIXBoolean(boolean=False)),
            ('SWS', self.getCIXBoolean(boolean=False)),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createISOComment(self, **keywords):
        macroName = 'ISO'
        namesAndValues = (
            # Apparently ISO has no ID (It is not well documented)
            ('ISO', '"; {comment}"'.format(comment=keywords['comment'])),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    # Header
    def createHead(self, **keywords):
        surface = keywords['surface']

        self.lastPocketIndex = 0
        self.lastSizingIndex = 0
        self.lastRabbetIndex = 0
        self.lastContourIndex = 0
        self.lastDadoIndex = 0
        self.lastNCMacroIndex = 0
        self.lastSequentialID = 1000
        self.geometryReferenceCounts = {}
        self.machiningPlanes = MachiningPlanes(lastUsedSideId=self.SIDE_OPPOSITE)
        self.machiningPlanes.planes.extend(self.__createPredefinedPlanes(**keywords))

        return joinIfNotNone('\n',
            self.createVersionHeader(**keywords),
            self.createNCWorkpiece(**keywords),
            self.createOffset(**keywords) if self.hasOffset(**keywords) else None,
            self.createSizingContourDefinition(**keywords),
            self.createRabbetContourDefinitions(**keywords),
            self.createDadoContourDefinitions(dados=surface.getAllDados(), **keywords),
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords),
        )

    def createVersionHeader(self, **_keywords):
        return '\n'.join((
            'BEGIN ID CID3',
            '\tREL= 5.0',
            'END ID',
            '',
        ))

    def createNCWorkpiece(self, **keywords):
        surface = keywords['surface']
        workpieceComments = self.getWorkpieceComments(**keywords)

        return '\n'.join((
            'BEGIN MAINDATA',
            '\tLPX = {x}'.format(x=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            '\tLPY = {y}'.format(y=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            '\tLPZ = {z}'.format(z=self.formatAsCoordinate(data=surface.thickness)),
            '\tORLST = "{origin}"'.format(origin=self.WOP_WORKING_AREA_ORIGINS_LIST),
            '\tSIMMETRY = 0',
            '\tTLCHK = 0',
            '\tTOOLING = ""',
            '\tCUSTSTR =$B$KBsExportToNcRoverNET.XncExtraPanelData$V""',
            '\tFCN = 1.000000',  # Millimeter
            '\tXCUT = 0',
            '\tYCUT = 0',
            '\tJIGTH = 0',
            '\tCKOP = 0',
            '\tUNIQUE = 0',
            '\tMATERIAL = "{material}"'.format(material=workpieceComments['ItemID']),
            '\tPUTLST = ""',
            '\tOPPWKRS = 0',
            '\tUNICLAMP = 0',
            '\tCHKCOLL = 0',
            '\tWTPIANI = 0',
            '\tCOLLTOOL = 0',
            '\tCALCEDTH = 0',
            '\tENABLELABEL = 0',
            '\tLOCKWASTE = 0',
            '\tLOADEDGEOPT = 0',
            '\tITLTYPE = 0',
            '\tRUNPAV = {cups:d}'.format(cups=self.OPTIMISE_SUCTION_CUPS),
            '\tFLIPEND = 0',
            'END MAINDATA',
            '',
        ))

    def hasOffset(self, **keywords):
        surface = keywords['surface']
        return not isNearlyEqual(surface.getAdditionX(), 0) or (not isNearlyEqual(surface.getAdditionY(), 0) and self.ZERO_OFFSET_Y_IS_SIZING_ADDITION)

    def getOffsetX(self, **keywords):
        surface = keywords['surface']
        surfaceAdditionX = surface.getAdditionX()

        def defaultSurfaceDependentOffsetX():
            if self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT:
                return surfaceAdditionX
            else:
                return -surfaceAdditionX

        if self.useIndividualWorkTableOrigins():
            workTableOrigins = self.getWorkTableOrigins(key='work table origin', **keywords)
            if 'L' in workTableOrigins:  # Left
                return surfaceAdditionX
            elif 'R' in workTableOrigins:  # Right
                return -surfaceAdditionX
            else:
                logging.warning(getLocalizedString(BIESSE_CIX_MESSAGES, 4, "Can't find work table origin in file 'BiesseCIX Work Table Origins Dictionary.json'. Using default surface-dependent offset."))
                return defaultSurfaceDependentOffsetX()
        else:
            return defaultSurfaceDependentOffsetX()

    def getOffsetY(self, **keywords):
        surface = keywords['surface']
        surfaceAdditionY = surface.getAdditionY()

        def defaultSurfaceDependentOffsetY():
            if self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT:  # Back
                return surfaceAdditionY
            else:  # Front
                return -surfaceAdditionY

        if self.useIndividualWorkTableOrigins():
            workTableOrigins = self.getWorkTableOrigins(key='work table origin', **keywords)
            if 'B' in workTableOrigins:  # Back
                return surfaceAdditionY
            elif 'F' in workTableOrigins:  # Front
                return -surfaceAdditionY
            else:
                logging.warning(getLocalizedString(BIESSE_CIX_MESSAGES, 4, "Can't find work table origin in file 'BiesseCIX Work Table Origins Dictionary.json'. Using default surface-dependent offset."))
                return defaultSurfaceDependentOffsetY()
        else:
            return defaultSurfaceDependentOffsetY()

    def createOffset(self, **keywords):
        macroName = 'OFFSET'
        namesAndValues = (
            # According to documentation, there is no ID
            ('X', self.getOffsetX(**keywords)),
            ('Y', self.getOffsetY(**keywords) if self.ZERO_OFFSET_Y_IS_SIZING_ADDITION else 0.0),
            ('Z', 0),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    # Machinings
    # Drillings
    def createCommonDrilling(self, **keywords):
        drilling = keywords['drilling']

        macroName = keywords['macroName']
        namesAndValues = (
            ('SIDE', keywords['side']),
            ('CRN', '"{stop}"'.format(stop=keywords['wopOrigin'] if 'wopOrigin' in keywords else self.WOP_ORIGIN)),
            ('X', self.formatAsCoordinate(data=keywords['x'])),
            ('Y', self.formatAsCoordinate(data=keywords['y'])),
            ('Z', 0),
            ('DP', self.formatAsCoordinate(data=keywords['depth'] if 'depth' in keywords else drilling.depth)),
            ('DIA', self.formatAsCoordinate(data=drilling.installedDiameter)),
            ('THR', self.getCIXBoolean(boolean=keywords['drillThrough'])),
            ('RTY', keywords['repetitionType'] if 'repetitionType' in keywords else self.REPEAT_NO),
            ('DX', keywords['dx'] if 'dx' in keywords else 32),
            ('DY', keywords['dy'] if 'dy' in keywords else 32),
            ('R', 32),
            ('A', 0),
            ('DA', 45),
            ('NRP', keywords['countDrillings'] if 'countDrillings' in keywords else 0),
            ('ISO', '""'),
            ('OPT', self.getCIXBoolean(boolean=True)),
            ('AZ', 0),
            ('AR', 0),
            ('AP', self.getCIXBoolean(boolean=False)),
            ('CKA', self.INCLINATION_NO),
            ('XRC', 0),
            ('YRC', 0),
            ('ARP', self.formatAsAngle(data=keywords['angle'] if 'angle' in keywords else 0.0)),
            ('LRP', 0),
            ('ER', self.getCIXBoolean(boolean=True)),
            ('MD', self.getCIXBoolean(boolean=False)),
            ('COW', self.getCIXBoolean(boolean=False)),  # For Skipper machine only
            ('A21', 0),
            ('TOS', self.getCIXBoolean(boolean=False)),
            ('VTR', 0),
            ('S21', -1),
            self.generateSimpleIDTuple(**keywords),
            ('AZS', 0),
            ('MAC', '"{machine}"'.format(machine='')),
            ('TNM', '"{tool}"'.format(tool=drilling.toolID if drilling.toolID else '')),
            ('TTP', self.getDrillType(**keywords)),
            ('TCL', self.TYPE_DRILLING),
            ('RSP', 0),  # automatic speed
            ('IOS', 0),  # automatic speed
            ('WSP', 0),  # automatic speed
            ('SPI', '"{spindle}"'.format(spindle='')),
            ('DDS', 3 if keywords['drillThrough'] else 0),
            ('DSP', 0),  # automatic speed
            ('BFC', self.getCIXBoolean(boolean=False)),
            ('SHP', self.HOOD_POSITION_DRILLING),
            ('EA21', self.getCIXBoolean(boolean=False)),
            ('CEN', '"{center}"'.format(center='')),  # center in sense of machine, not in sense of middle
            ('AGG', '"{aggregate}"'.format(aggregate='')),
            ('LAY', '"{layer}"'.format(layer=macroName)),  # the convention is to use the macro name as layer name
            ('PRS', self.getCIXBoolean(boolean=False)),
        )

        return self.createCIXMacro(namesAndValues=namesAndValues, **keywords)

    def createVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']
        return self.createCommonDrilling(macroName='BV', side=self.SIDE_MAIN, x=drilling.position.x, y=drilling.position.y, defaultDrillType=self.FLAT_DRILL_TYPE, drillThrough=False, **keywords)

    def getDrillingPiercingZValue(self, **keywords):
        return self.PIERCING_DRILLING_ADDITION

    def createVerticalDrillingPiercing(self, **keywords):
        drilling = keywords['drilling']
        return self.createCommonDrilling(macroName='BV', side=self.SIDE_MAIN, x=drilling.position.x, y=drilling.position.y, defaultDrillType=self.PIERCING_DRILL_TYPE, drillThrough=True, **keywords)

    def createVerticalGridDrilling(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        drilling = gridDrilling.getStartDrilling()
        drilling.uuid = None
        dx = gridDrilling.gridNormalizedDirection.x * gridDrilling.gridDistance
        dy = gridDrilling.gridNormalizedDirection.y * gridDrilling.gridDistance
        countDrillings = gridDrilling.countDrillings

        return self.createCommonDrilling(
            macroName='BV',
            side=self.SIDE_MAIN,
            drilling=drilling,
            x=drilling.position.x,
            y=drilling.position.y,
            defaultDrillType=self.FLAT_DRILL_TYPE,
            drillThrough=False,
            dx=dx,
            dy=dy,
            countDrillings=countDrillings,
            **keywords
        )

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        return self.createVerticalGridDrilling(repetitionType=self.REPEAT_X, **keywords)

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        return self.createVerticalGridDrilling(repetitionType=self.REPEAT_Y, **keywords)

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        return self.createVerticalGridDrilling(repetitionType=self.REPEAT_XY, angle=normalizeAngle(keywords['gridDrilling'].getXYAngle()), **keywords)

    def getOverlapDistance(self, **keywords):
        machining = keywords['machining']
        routerRadius = machining.tool.getRouterDiameter() / 2

        if hasattr(machining, 'diameter'):  # pocket
            machiningDiameter = machining.diameter
            return machiningDiameter / math.ceil(machiningDiameter / routerRadius)
        else:  # contour
            return routerRadius  # Normally depends on the variable 'maximum horizontal step' of the tool, but is actually limited to its radius.

    def createCIXPockMacro(self, **keywords):  # Used by createPocketCommon() as well as createVerticalContour() -> createFreeFormPocket()
        machining = keywords['machining']

        if 'contour' in keywords and hasattr(machining, 'path'):
            reverseMachining = not machining.path.isClockwise()  # contour.path must be independent of POCKET_DIRECTION and exported as defined.
        else:
            reverseMachining = not self.isDadoRabbetReplacementCW(machining, self.POCKET_DIRECTION)

        macroName = 'POCK'
        namesAndValues = (
            ('GID', '"{gid}"'.format(gid=machining.gid)),
            ('ISL', '""'),
            ('DIA', self.formatAsCoordinate(data=machining.diameter)) if hasattr(machining, 'diameter') else None,  # pocket|contour
            ('DP', self.formatAsCoordinate(data=keywords['depth'])),
            ('ISO', '""'),
            ('OPT', self.getCIXBoolean(boolean=True)),
            ('TYP', 'ptOUT'),
            ('DLT', self.formatAsCoordinate(data=self.getOverlapDistance(**keywords))),
            ('N', -1),  # Not documented for POCK
            ('A', 0),
            ('TC', self.getCIXBoolean(boolean=True)),  # tool compensation
            ('CKI', self.getCIXBoolean(boolean=False)),
            ('ZST', self.formatAsCoordinate(data=keywords['depthPerStep'])),
            ('RV', self.getCIXBoolean(boolean=reverseMachining)),  # While the geometry remains CW, the machining only needs to be reversed if the direction is CCW.
            ('RRV', self.getCIXBoolean(boolean=False)),
            ('COW', self.getCIXBoolean(boolean=False)),  # For Skipper machine only
            ('OVM', self.formatAsCoordinate(data=keywords['contourDistance'] if 'contourDistance' in keywords else 0)),  # pocket|contour
            ('A21', 0),
            ('Z', 0),
            ('TOS', self.getCIXBoolean(boolean=False)),
            ('S21', -1),
            self.generateGeometryReferenceIDTuple(**keywords),
            ('DSP', 100),  # lead out speed
            ('RSP', 0),  # automatic rotation speed
            ('IOS', 100),  # lead in speed
            ('WSP', 0),  # automatic working speed
            ('TNM', '"{tool}"'.format(tool=machining.toolID if machining.toolID else '')),
            ('TTP', 100),  # CANDELA
            ('TCL', self.TYPE_ROUTING),
            ('TIN', keywords['approach']),
            ('AIN', 45),
            ('CIN', self.getCIXBoolean(boolean=False)),
            ('TOU', keywords['approach']),
            ('AOU', 45),
            ('COU', self.getCIXBoolean(boolean=False)),
            ('DIN', self.formatAsCoordinate(data=self.getPocketEntranceRadius(**keywords))),
            ('DOU', self.formatAsCoordinate(data=self.getPocketEntranceRadius(**keywords))),
            ('SDS', 0),  # automatic speed
            ('PRP', 0),
            ('SPI', '"{spindle}"'.format(spindle='')),
            ('SHP', self.HOOD_POSITION_MILLINGS),
            ('EA21', self.getCIXBoolean(boolean=False)),
            ('CEN', '"{center}"'.format(center='')),  # center in sense of machine, not in sense of middle
            ('AGG', '"{aggregate}"'.format(aggregate='')),
            ('LAY', '"{layer}"'.format(layer=macroName)),  # the convention is to use the macro name as layer name
            ('AZ', 0),  # Not documented for POCK
            ('AR', 0),  # Not documented for POCK
            ('CKA', self.INCLINATION_NO),  # Not documented for POCK
            ('BFC', self.getCIXBoolean(boolean=False)),  # Not documented for POCK
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createPocketCommon(self, **keywords):
        pocket = keywords['pocket']
        pocket.gid = keywords['gid']

        depthPerStep = self.getRouterDepthPerStep(machiningDepth=pocket.actualDepth, **keywords)

        return ''.join((
            self.createCIXGeoMacro(radiusCompensation=self.RADIUS_COMPENSATION_RIGHT, **keywords),
            self.createCIXCircleCR(center=keywords['position'], radius=pocket.diameter / 2, **keywords),
            self.createCIXEndPathMacro(**keywords),
            self.createCIXPockMacro(depthPerStep=depthPerStep, approach=self.APPROACH_CURVE, **keywords),
        ))

    def createVerticalPocket(self, **keywords):
        self.lastPocketIndex += 1
        proposedNamedID = self.getLocalizedMachiningName(5, "Pocket {index}").format(index=self.lastPocketIndex)
        gid = self.getGeometryID(proposedNamedID=proposedNamedID, **keywords)

        return self.createPocketCommon(clearanceType='ptOUT', gid=gid, position=keywords['pocket'].position, **keywords)  # ic 2021 BF4.4: clearanceType not used anymore. Compatibility variable for old code; do NOT use for new code!

    def createVerticalPocketPiercing(self, **keywords):
        self.lastPocketIndex += 1
        proposedNamedID = self.getLocalizedMachiningName(6, "Piercing Pocket {index}").format(index=self.lastPocketIndex)
        gid = self.getGeometryID(proposedNamedID=proposedNamedID, **keywords)

        # clearanceType = 'ptFSH' if keywords['pocket'].diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER else 'ptOUT'
        # TODO IT-293: Doesn't work, because POCK is always cleared out. Not clearing out must be executed as contour. ptFSH clears out from the inside to the outside, ptOUT clears out from the outside to the inside.  --mjk

        return self.createPocketCommon(clearanceType='ptOUT', gid=gid, position=keywords['pocket'].position, **keywords)  # ic 2021 BF4.4: clearanceType not used anymore. Compatibility variable for old code; do NOT use for new code!

    def createHorizontalPocket(self, **keywords):
        self.lastPocketIndex += 1
        proposedNamedID = self.getLocalizedMachiningName(7, "Horizontal Pocket {index}").format(index=self.lastPocketIndex)
        gid = self.getGeometryID(proposedNamedID=proposedNamedID, **keywords)

        plane, planeIsNew, positionOnPlane = self.__calculatePlane(**keywords)
        fakeDimension = keywords['pocket'].diameter

        return ''.join((
            self.createCIXStraightSide(plane=plane, fakeVirtualFaceLength=fakeDimension, fakeVirtualFaceHeight=fakeDimension, **keywords) if planeIsNew else '',
            self.createPocketCommon(wopOrigin=self.WOP_ORIGIN_FRONT_LEFT, clearanceType='ptOUT', gid=gid, sideID=plane.sideID, position=positionOnPlane, depth=keywords['pocket'].depth, **keywords),  # ic 2021 BF4.4: clearanceType not used anymore. Compatibility variable for old code; do NOT use for new code!
        ))

    def getHorizontalMachiningsExporter(self, **keywords):
        return HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds(self)

    def createHorizontalDrillingOnBounds(self, **keywords):
        return self.createCommonDrilling(macroName='BH', drillThrough=False, depth=keywords['drilling'].depth, defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)

    def createHorizontalDrillingFromFront(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        wopOrigin = self.WOP_ORIGIN_FRONT_LEFT if self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT else self.WOP_ORIGIN_FRONT_RIGHT
        return self.createHorizontalDrillingOnBounds(side=self.SIDE_FRONT, x=drilling.position.x, y=zPosition, wopOrigin=wopOrigin, **keywords)

    def createHorizontalDrillingFromRight(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        wopOrigin = self.WOP_ORIGIN_FRONT_LEFT if self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_RIGHT else self.WOP_ORIGIN_FRONT_RIGHT
        return self.createHorizontalDrillingOnBounds(side=self.SIDE_RIGHT, x=drilling.position.y, y=zPosition, wopOrigin=wopOrigin, **keywords)

    def createHorizontalDrillingFromBack(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        wopOrigin = self.WOP_ORIGIN_FRONT_LEFT if self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_RIGHT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT else self.WOP_ORIGIN_FRONT_RIGHT
        return self.createHorizontalDrillingOnBounds(side=self.SIDE_BACK, x=drilling.position.x, y=zPosition, wopOrigin=wopOrigin, **keywords)

    def createHorizontalDrillingFromLeft(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        wopOrigin = self.WOP_ORIGIN_FRONT_LEFT if self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT else self.WOP_ORIGIN_FRONT_RIGHT
        return self.createHorizontalDrillingOnBounds(side=self.SIDE_LEFT, x=drilling.position.y, y=zPosition, wopOrigin=wopOrigin, **keywords)

    def createRotatedHorizontalDrilling(self, **keywords):
        return self.createFiveAxisDrilling(**keywords)

    def __createPredefinedPlanes(self, **keywords):
        size = keywords['surface'].getFinishedSize()

        yield Plane(
            predefined=True, sideID=self.SIDE_FRONT,
            matrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_RIGHT,
            matrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_BACK,
            matrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(size.x, size.y, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_LEFT,
            matrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
        )

    def __calculatePlane(self, **keywords):
        # Biesse CIX expects that the planes are created with origin at front left. Therefore, we must put the drilling back in the front left coordinates
        machiningOriginFrontLeft = self.getMachiningOriginFrontLeft(**keywords)

        machiningPlane = Plane.createFromMachining(machiningOriginFrontLeft)
        plane = self.machiningPlanes.findPlane(machiningPlane)

        planeIsNew = plane is None
        if planeIsNew:
            self.machiningPlanes.registerPlane(machiningPlane)
            plane = machiningPlane

        assert plane is not None, "No plane found for machining"

        positionOnPlane = machiningOriginFrontLeft.getPositionOnPlane(plane)

        return plane, planeIsNew, positionOnPlane

    def getMachiningOriginFrontLeft(self, **keywords):
        surface = keywords['surface']
        machining = keywords['machining']

        machiningOriginFrontLeft = copy.deepcopy(machining)
        if self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT:
            machiningOriginFrontLeft.mirrorOnXAxis(surface)
            machiningOriginFrontLeft.mirrorOnYAxis(surface)
        elif self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT:
            machiningOriginFrontLeft.mirrorOnXAxis(surface)
        elif self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_RIGHT:
            machiningOriginFrontLeft.mirrorOnYAxis(surface)

        return machiningOriginFrontLeft

    def createFiveAxisDrilling(self, **keywords):
        drilling = keywords['drilling']
        depth = drilling.depth

        plane, planeIsNew, positionOnPlane = self.__calculatePlane(**keywords)

        fakeDimension = keywords['drilling'].diameter

        return ''.join((
            self.createCIXStraightSide(plane=plane, fakeVirtualFaceLength=fakeDimension, fakeVirtualFaceHeight=fakeDimension, **keywords) if planeIsNew else '',
            self.createCommonDrilling(
                macroName='BG',
                side=plane.sideID,
                x=positionOnPlane.x,
                y=positionOnPlane.y,
                drillThrough=False,
                depth=depth,
                defaultDrillType=self.FLAT_DRILL_TYPE,
                wopOrigin=self.WOP_ORIGIN_FRONT_LEFT,
                **keywords)
        ))

    # Dados
    def getStartAndEndDistanceSecurityForDadoAlongAxis(self, **keywords):
        dado = keywords['dado']
        startInset, endInset = 0, 0
        if not dado.isInsetStart():
            startInset = self.SECURITY_DISTANCE_FOR_DADO_SAW
        if not dado.isInsetEnd():
            endInset = self.SECURITY_DISTANCE_FOR_DADO_SAW
        return startInset, endInset

    def getPreSawDepth(self, **keywords):
        preSawDepth = keywords['preSawDepth']
        if preSawDepth:
            verticalRuns = 2
            finalRun = keywords['sawDepth'] - preSawDepth + keywords['additionalDepth']
            return verticalRuns, finalRun
        else:
            return 0, 0

    def createVerticalDadoWithSawAlongAxis(self, **keywords):
        dado = keywords['dado']
        start, end = dado.getStartAndEndPoints()

        startInset, endInset = self.getStartAndEndDistanceSecurityForDadoAlongAxis(**keywords)
        if operator.xor(operator.xor(isNearlyLesser(end.x, start.x), isNearlyLesser(end.y, start.y)), keywords['reverseCuttingDirection']):
            startInset, endInset = endInset, startInset

        verticalRuns, finalRun = self.getPreSawDepth(sawDepth=keywords['depth'], preSawDepth=self.DADO_PRESAW_DEPTH, additionalDepth=0, **keywords)

        macroName = keywords['macroName']
        namesAndValues = (
            ('ID', '"{id}"'.format(id=keywords['identifier'])) if 'identifier' in keywords and keywords['identifier'] else None,  # CUT_X, CUT_Y and CUT_G do support ID, neither in BiesseWorks (error message) nor in bSolid
            ('SIDE', self.SIDE_MAIN),
            ('CRN', '"{stop}"'.format(stop=self.WOP_ORIGIN)),
            ('X', self.formatAsCoordinate(data=min(start.x, end.x))),
            ('Y', self.formatAsCoordinate(data=min(start.y, end.y))),
            ('Z', 0),
            ('DP', self.formatAsCoordinate(data=keywords['depth'])),
            ('L', self.formatAsCoordinate(data=dado.length)),
            ('NRP', 0),
            ('D', 0),
            ('ISO', '""'),
            ('OPT', self.getCIXBoolean(boolean=True)),
            ('TH', 0),  # Do not select saw tool by saw blade thickness
            ('THR', self.getCIXBoolean(boolean=False)),
            ('RV', self.getCIXBoolean(boolean=keywords['reverseCuttingDirection'])),
            ('COW', self.getCIXBoolean(boolean=False)),  # For Skipper machine only
            ('TTK', self.formatAsCoordinate(data=dado.width)),
            ('OVM', 0),
            ('TOS', self.getCIXBoolean(boolean=False)),
            ('VTR', verticalRuns),
            ('GIP', self.getCIXBoolean(boolean=False)),
            ('TNM', '"{tool}"'.format(tool=dado.toolID)),
            ('TTP', 200),  # CUTT0
            ('TCL', self.TYPE_CUTTING),
            ('RSP', 0),  # automatic speed
            ('IOS', 0),  # automatic speed
            ('WSP', 0),  # automatic speed
            ('SPI', '"{spindle}"'.format(spindle='')),
            ('BFC', self.getCIXBoolean(boolean=False)),
            ('SHP', self.HOOD_POSITION_SAW_CUTS),
            ('BRC', self.getCIXBoolean(boolean=keywords['bladeRadiusCorrection'] if 'bladeRadiusCorrection' in keywords else False)),
            ('BDR', self.getCIXBoolean(boolean=True)),
            ('PRV', self.getCIXBoolean(boolean=True)),
            ('NRV', self.getCIXBoolean(boolean=False)),
            ('DIN', self.formatAsCoordinate(data=startInset)),
            ('DOU', self.formatAsCoordinate(data=endInset)),
            ('CRC', 0),  # Central
            ('DSP', 0),  # automatic speed
            ('CEN', '"{center}"'.format(center='')),  # center in sense of machine, not in sense of middle
            ('AGG', '"{aggregate}"'.format(aggregate='')),
            ('LAY', '"{layer}"'.format(layer=macroName)),  # the convention is to use the macro name as layer name
            ('DVR', self.formatAsCoordinate(data=finalRun)),
        )

        return self.createCIXMacro(namesAndValues=namesAndValues, **keywords)

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        self.lastDadoIndex += 1
        identifier = self.getSawnMachiningID(**keywords)

        reverseCuttingDirection = self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_RIGHT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT
        return self.createVerticalDadoWithSawAlongAxis(macroName='CUT_X', identifier=identifier, reverseCuttingDirection=reverseCuttingDirection, bladeRadiusCorrection=keywords['dado'].isInset(), **keywords)

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        identifier = self.getSawnMachiningID(**keywords)

        reverseCuttingDirection = self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT
        return self.createVerticalDadoWithSawAlongAxis(macroName='CUT_Y', identifier=identifier, reverseCuttingDirection=reverseCuttingDirection, bladeRadiusCorrection=keywords['dado'].isInset(), **keywords)

    def createCIXCutG(self, **keywords):
        start = keywords['start']
        end = keywords['end']

        macroName = 'CUT_G'
        namesAndValues = (
            ('SIDE', self.SIDE_MAIN),
            ('CRN', '"{stop}"'.format(stop=self.WOP_ORIGIN)),
            ('X', self.formatAsCoordinate(data=start.x)),
            ('Y', self.formatAsCoordinate(data=start.y)),
            ('Z', 0),
            ('DP', self.formatAsCoordinate(data=keywords['depth'])),
            ('TYP', self.CUT_XY),
            ('ANG', self.formatAsAngle(data=keywords['machining'].getXYAngle())),
            ('XE', self.formatAsCoordinate(data=end.x)),
            ('YE', self.formatAsCoordinate(data=end.y)),
            ('ZE', 0),
            ('AZ', self.formatAsCoordinate(data=keywords['angle'])),  # Angle of the saw. Not documented for cuttings (CUT_G)
            ('ISO', '""'),
            ('OPT', self.getCIXBoolean(boolean=True)),
            ('TH', 0),  # Do not select saw tool by saw blade thickness
            ('RTY', self.REPEAT_NO),
            ('DX', 0),
            ('DY', 0),
            ('R', 0),
            ('A', 0),
            ('DA', 0),
            ('RDL', self.getCIXBoolean(boolean=False)),
            ('NRP', 0),
            ('CKA', keywords['inclination']),
            ('THR', self.getCIXBoolean(boolean=keywords['sawThrough'])),
            ('RV', self.getCIXBoolean(boolean=False)),
            ('XRC', 0),
            ('YRC', 0),
            ('ARP', 0),
            ('LRP', 0),
            ('ER', self.getCIXBoolean(boolean=True)),
            ('COW', self.getCIXBoolean(boolean=False)),  # For Skipper machine only
            ('TTK', self.formatAsCoordinate(data=keywords['width'])),
            ('OVM', 0),
            ('TOS', self.getCIXBoolean(boolean=False)),
            ('VTR', keywords['verticalRuns']),
            ('GIP', self.getCIXBoolean(boolean=False)),
            ('ID', '"{id}"'.format(id=keywords['identifier'])) if 'identifier' in keywords and keywords['identifier'] else None,  # CUT_X, CUT_Y and CUT_G do support ID, neither in BiesseWorks (error message) nor in bSolid
            ('TNM', '"{tool}"'.format(tool=keywords['machining'].toolID)),
            ('TTP', 200),  # CUTT0
            ('TCL', self.TYPE_CUTTING),
            ('RSP', 0),  # automatic speed
            ('IOS', 0),  # automatic speed
            ('WSP', 0),  # automatic speed
            ('SPI', '"{spindle}"'.format(spindle='')),
            ('BFC', self.getCIXBoolean(boolean=False)),
            ('SHP', self.HOOD_POSITION_SAW_CUTS),
            ('BRC', self.getCIXBoolean(boolean=keywords['bladeRadiusCorrection'] if 'bladeRadiusCorrection' in keywords else False)),
            ('BDR', self.getCIXBoolean(boolean=True)),
            ('PRV', self.getCIXBoolean(boolean=True)),
            ('NRV', self.getCIXBoolean(boolean=keywords['reverseCuttingDirection'])),
            ('DIN', self.formatAsCoordinate(data=keywords['startInset'])),
            ('DOU', self.formatAsCoordinate(data=keywords['endInset'])),
            ('CRC', keywords['radiusCompensation']),
            ('DSP', 0),  # automatic speed
            ('CEN', '"{center}"'.format(center='')),  # center in sense of machine, not in sense of middle
            ('AGG', '"{aggregate}"'.format(aggregate='')),
            ('LAY', '"{layer}"'.format(layer=macroName)),  # the convention is to use the macro name as layer name
            ('DVR', self.formatAsCoordinate(data=keywords['finalRun'])),
            ('ETB', self.getCIXBoolean(boolean=False)),  # Not documented for cuttings
            ('KDT', self.getCIXBoolean(boolean=False)),  # Not documented for cuttings
            ('IMS', 0),  # Not documented for cuttings
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createVerticalDadoWithSawCRotated(self, **keywords):
        dado = keywords['dado']
        start, end = dado.getStartAndEndPoints()
        startInset, endInset = self.getStartAndEndDistanceSecurityForDadoAlongAxis(**keywords)
        verticalRuns, finalRun = self.getPreSawDepth(sawDepth=keywords['depth'], preSawDepth=self.DADO_PRESAW_DEPTH, additionalDepth=0, **keywords)

        return self.createCIXCutG(
            identifier=self.getSawnMachiningID(**keywords),
            start=start,
            end=end,
            startInset=startInset,
            endInset=endInset,
            angle=0,
            inclination=self.INCLINATION_NO,
            sawThrough=False,
            radiusCompensation=self.RADIUS_COMPENSATION_NONE,
            width=dado.width,
            reverseCuttingDirection=False,
            verticalRuns=verticalRuns,
            finalRun=finalRun,
            bladeRadiusCorrection=keywords['dado'].isInset(),
            **keywords
        )

    # Miters
    def createVerticalMiter(self, **keywords):
        miter = keywords['miter']
        start, end = miter.getStartAndEndPoints()
        verticalRuns, finalRun = self.getPreSawDepth(sawDepth=keywords['surface'].thickness, preSawDepth=self.MITER_PRE_SAW_DEPTH_FROM_TOP, additionalDepth=self.ADDITIONAL_DEPTH_FOR_MITER, **keywords)

        angle = 90 - miter.getAngle()
        if not self.isStandardOrigin(**keywords):
            angle = - angle

        return self.createCIXCutG(
            start=start,
            end=end,
            startInset=self.SECURITY_DISTANCE_FOR_DADO_SAW,
            endInset=self.SECURITY_DISTANCE_FOR_DADO_SAW,
            angle=angle,
            inclination=self.INCLINATION_ABSOLUTE,
            depth=self.ADDITIONAL_DEPTH_FOR_MITER,
            sawThrough=True,
            radiusCompensation=self.RADIUS_COMPENSATION_LEFT,
            width=0,  # No width, as we do not want a slanted dado with a defined width
            reverseCuttingDirection=True,
            verticalRuns=verticalRuns,
            finalRun=finalRun,
            **keywords
        )

    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCW)

    def generateDadoSegments(self, **keywords):
        xoredCondition = not self.isStandardOrigin(**keywords) and keywords['sideID'] == self.SIDE_MAIN
        return super().generateDadoSegments(xoredCondition=xoredCondition, **keywords)

    # Contour Definitions
    def getSizingRadiusCompensation(self, **keywords):
        return self.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else self.RADIUS_COMPENSATION_LEFT

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=self.RADIUS_COMPENSATION_NONE, left=self.RADIUS_COMPENSATION_LEFT, right=self.RADIUS_COMPENSATION_RIGHT)

    def createCommonContourStartBlock(self, **keywords):
        machining = keywords['machining']
        machining.gid = keywords['gid']

        vertex = keywords['startVertex']

        return ''.join((
            self.createCIXGeoMacro(**keywords),
            self.createCIXStartPointMacro(x=vertex.position.x, y=vertex.position.y, **keywords),
        ))

    def createCommonContourEndBlock(self, **keywords):
        return self.createCIXEndPathMacro(**keywords)

    def createSizingStartBlock(self, **keywords):
        self.lastSizingIndex += 1
        proposedNamedID = self.getLocalizedMachiningName(2, "Sizing {index}").format(index=self.lastSizingIndex)
        gid = self.getGeometryID(proposedNamedID=proposedNamedID, **keywords)

        depth = self.getSizingZValue(**keywords)
        radiusCompensation = self.getSizingRadiusCompensation(sizing=keywords['surface'].sizing, **keywords)

        return self.createCommonContourStartBlock(gid=gid, depth=depth, radiusCompensation=radiusCompensation, **keywords)

    def createSizingEndBlock(self, **keywords):
        return self.createCommonContourEndBlock(**keywords)

    def createContourStartBlock(self, **keywords):
        self.lastContourIndex += 1
        proposedNamedID = self.getLocalizedMachiningName(4, "Contour {index}").format(index=self.lastContourIndex)
        gid = self.getGeometryID(proposedNamedID=proposedNamedID, **keywords)

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        return self.createCommonContourStartBlock(gid=gid, radiusCompensation=radiusCompensation, **keywords)

    def createContourEndBlock(self, **keywords):
        return self.createCommonContourEndBlock(**keywords)

    def createDadoContourStartBlock(self, **keywords):
        self.lastDadoIndex += 1
        proposedNamedID = self.getLocalizedMachiningName(1, "Dado {index}").format(index=self.lastDadoIndex)
        gid = self.getGeometryID(proposedNamedID=proposedNamedID, **keywords)

        radiusCompensation = self.getDadoRadiusCompensation(**keywords)

        return self.createCommonContourStartBlock(gid=gid, depth=keywords['dado'].depth, radiusCompensation=radiusCompensation, **keywords)

    def createDadoContourEndBlock(self, **keywords):
        return self.createCommonContourEndBlock(**keywords)

    def createRabbetContourStartBlock(self, **keywords):
        self.lastRabbetIndex += 1
        proposedNamedID = self.getLocalizedMachiningName(3, "Rabbet {index}").format(index=self.lastRabbetIndex)
        gid = self.getGeometryID(proposedNamedID=proposedNamedID, **keywords)

        depth = keywords['rabbet'].depth
        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        return self.createCommonContourStartBlock(gid=gid, depth=depth, radiusCompensation=radiusCompensation, **keywords)

    def createRabbetContourEndBlock(self, **keywords):
        return self.createCommonContourEndBlock(**keywords)

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        position = segment.getEndVertex().position

        macroName = 'LINE_EP'
        namesAndValues = (
            self.generateSimpleIDTuple(**keywords),
            ('XE', self.formatAsCoordinate(data=position.x)),
            ('YE', self.formatAsCoordinate(data=position.y)),
            ('ZS', 0),
            ('ZE', 0),
            ('SC', 'scOFF'),
            ('FD', 0),  # automatic speed
            ('SP', 0),  # automatic speed
            ('SOL', 0),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createRadiusSegmentAsARCEPCE(self, **keywords):
        segment = keywords['segment']

        if keywords['machining'].isVertical() and not self.isStandardOrigin(**keywords):
            direction = segment.getDirection(self.CCW, self.CW, self.CCW, self.CW)
        else:
            direction = segment.getDirection(self.CW, self.CCW, self.CW, self.CCW)

        position = segment.getEndVertex().position
        center = segment.getCenter()

        macroName = 'ARC_EPCE'
        namesAndValues = (
            self.generateSimpleIDTuple(**keywords),
            ('XE', self.formatAsCoordinate(data=position.x)),
            ('YE', self.formatAsCoordinate(data=position.y)),
            ('ZE', 0),
            ('XC', self.formatAsCoordinate(data=center.x)),
            ('YC', self.formatAsCoordinate(data=center.y)),
            ('DIR', direction),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createRadiusSegmentAsARCIPEP(self, **keywords):
        segment = keywords['segment']

        secondPosition = segment.radiusVertex.position
        endPosition = segment.getEndVertex().position

        macroName = 'ARC_IPEP'
        namesAndValues = (
            self.generateSimpleIDTuple(**keywords),
            ('XE', self.formatAsCoordinate(data=endPosition.x)),
            ('YE', self.formatAsCoordinate(data=endPosition.y)),
            ('ZE', 0),
            ('X2', self.formatAsCoordinate(data=secondPosition.x)),
            ('Y2', self.formatAsCoordinate(data=secondPosition.y)),
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=namesAndValues, **keywords)

    def createRadiusSegment(self, **keywords):
        if self.RADIUS_SEGMENT_MACRO == 'ARC_IPEP':
            return self.createRadiusSegmentAsARCIPEP(**keywords)
        else:
            return self.createRadiusSegmentAsARCEPCE(**keywords)

    # Contour Executions
    def createCIXRoutGMacro(self, **keywords):
        machining = keywords['machining']
        approachDepartureDistance = keywords['approachDepartureDistance']

        macroName = 'ROUTG'
        namesAndValues = (
            self.generateGeometryReferenceIDTuple(gid=machining.gid, **keywords),
            ('GID', '"{gid}"'.format(gid=machining.gid)),
            ('WSP', self.formatAsCoordinate(data=keywords['feedRate'] if 'feedRate' in keywords else 0)),  # 0 is automatic speed
            ('TNM', '"{tool}"'.format(tool=machining.toolID)),
            ('CRC', keywords['radiusCompensation']),
            ('TIN', keywords['approach']),
            ('AIN', 45),
            ('DIN', self.formatAsCoordinate(data=approachDepartureDistance)),
            ('CIN', self.getCIXBoolean(boolean=False)),
            ('GIN', 0),
            ('TBI', self.getCIXBoolean(boolean=False)),
            ('TLI', 0),
            ('TQI', 0),
            ('TOU', keywords['departure']),
            ('AOU', 45),
            ('DOU', self.formatAsCoordinate(data=approachDepartureDistance)),
            ('PRP', 0),
            ('COU', self.getCIXBoolean(boolean=False)),
            ('GOU', 0),
            ('TBO', self.getCIXBoolean(boolean=False)),
            ('TLO', 0),
            ('TQO', 0),
            ('DP', self.formatAsCoordinate(data=keywords['depth'])) if 'depth' in keywords else None,
            ('SHP', self.HOOD_POSITION_MILLINGS),
            ('UDT', self.getCIXBoolean(boolean=False)),
            ('OVM', self.formatAsCoordinate(data=keywords['contourDistance'] if 'contourDistance' in keywords else 0)),
            ('VTR', self.formatAsCoordinate(data=keywords['numberOfSteps'])),
            ('DVR', self.formatAsCoordinate(data=keywords['finalRun'])) if 'finalRun' in keywords and not isNearlyEqual(keywords['finalRun'], 0.0) else None,
            ('THR', self.getCIXBoolean(boolean=keywords['throughRouting'])) if 'throughRouting' in keywords and keywords['throughRouting'] else None,
            ('ZS', self.formatAsCoordinate(data=keywords['startDepth'])) if 'startDepth' in keywords else None,
            ('ZE', self.formatAsCoordinate(data=keywords['endDepth'])) if 'endDepth' in keywords else None,
        )

        return self.createCIXMacro(macroName=macroName, namesAndValues=filter(lambda entry: entry is not None, namesAndValues), **keywords)

    def getRouterDepthPerStep(self, **keywords):
        depthPerStep = super().getRouterDepthPerStep(**keywords)

        # BiesseCIX does not allow the depth-per-step to be deeper than the depth of the machining
        if isNearlyGreater(depthPerStep, keywords['machiningDepth']):
            return keywords['machiningDepth']

        return depthPerStep

    def createFreeFormPocket(self, **keywords):
        return ''.join((
            self.createCIXPockMacro(**keywords),
        ))

    def createCommonContour(self, **keywords):
        return ''.join((
            self.createCIXRoutGMacro(**keywords),
            self.createCIXEndPathMacro(**keywords),  # ROUTG too needs an ENDPATH macro, not only the contour definition
        ))

    def isSizingStrategyUsingTwoPasses(self, **keywords):
        area = keywords['surface'].getSawnSizeRectangularArea()

        return (
            self.SIZING_PASS_STRATEGY == 'force-two-passes'
            or (self.SIZING_PASS_STRATEGY == 'two-passes-for-small-workpieces' and area < 1000000 * self.SIZING_PASS_STRATEGY_AREA_THRESHOLD)
        )

    def createSizingContour(self, **keywords):
        return self.createSizingContourExecution(**keywords)

    def createSizingContourBlock(self, **keywords):
        machining = sizing = keywords['sizing']
        surface = keywords['surface']

        radiusCompensation = self.getSizingRadiusCompensation(**keywords)
        sizingApproachMode = self.getSizingApproachMode(**keywords)

        depthPerStep = self.getRouterDepthPerStep(machining=sizing, machiningDepth=sizing.actualDepth, **keywords)
        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=sizing.actualDepth, depthPerStep=depthPerStep, **keywords)
        finalRun = 0  # Automatic

        if self.isSizingStrategyUsingTwoPasses(**keywords) and surface.isFirstSide:
            finalRun = self.SIZING_FIRST_PASS_REMAINING_THICKNESS + self.SIZING_DEPTH_ADJUSTMENT
            if isNearlyLesserOrEqual(finalRun, depthPerStep) and not isNearlyEqual(depthPerStep, 0.0):
                message = self.getLocalizedMessage(76, "Router with insufficient depth per step to be able to make nesting remaining-thickness sizing. Workpiece: {workpiece}, Tool name: '{toolName}', Tool ID: {toolID}")
                logging.error(message.format(
                    workpiece=surface.workpieceName,
                    toolName=machining.tool.name if hasattr(machining.tool, 'name') else None,
                    toolID=machining.toolID
                ))

            numberOfSteps += 1

        if self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS and isNearlyGreaterOrEqual(self.SIZING_DEPTH_ADJUSTMENT, 0.0) and not machining.tool.isRouterCentered():
            depth = self.SIZING_DEPTH_ADJUSTMENT
            throughRouting = True
        else:
            depth = keywords['zValue']
            throughRouting = False

        return self.createCommonContour(
            machining=keywords['sizing'],
            radiusCompensation=radiusCompensation,
            approach=sizingApproachMode,
            departure=sizingApproachMode,
            approachDepartureDistance=self.SIZING_APPROACH_DEPARTURE_DISTANCE,
            depth=depth,
            numberOfSteps=numberOfSteps,
            finalRun=finalRun,
            throughRouting=throughRouting,
            **keywords
        )

    def createVerticalContour(self, **keywords):
        contour = keywords['contour']

        approachMode = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departureMode = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        depthPerStep = contour.currentDepthPerStep
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=contour.actualDepth, **keywords)

        keywords['depth'] = contour.actualDepth
        keywords['numberOfSteps'] = self.getCountOfStepsForDepth(machiningDepth=contour.actualDepth, depthPerStep=depthPerStep, **keywords)

        # TODO CONTOUR2.0 IT-238 Currently feedRate and depthPerStep are only implemented directly within createVerticalContour, not other contour using machinings -- NG
        if not isNearlyEqual(contour.currentFeedRate, 0.0):
            # TODO CONTOUR2.0 IT-238 feedRate is only 6 or 8 in the AcceptanceTests but WSP (of ROUTG) is documented as "mm/min" -- NG
            keywords['feedRate'] = contour.currentFeedRate

        if contour.clearOut:
            return self.createFreeFormPocket(
                depthPerStep=depthPerStep,
                approach=approachMode,
                clearanceType='ptOUT',
                gid=contour.gid,
                **keywords
            )
        else:
            return self.createCommonContour(
                radiusCompensation=self.getWopRadiusCompensation(**keywords),
                approach=approachMode,
                departure=departureMode,
                approachDepartureDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
                **keywords
            )

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createVerticalContour, **keywords))

    def createHorizontalContour(self, **keywords):
        return self.createVerticalContour(**keywords)

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createHorizontalContour, **keywords))

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']

        # TODO (HORIZONTAL) IT-222/IT-230: unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if contour.isVertical():
            sideID = self.SIDE_MAIN
            plane = None  # TODO (HORIZONTAL) IT-228/IT-230: can we be more precise? -- NG
            planeIsNew = False
            depth = contour.actualDepth
            zValue = self.getContourZValue(depth=keywords['contour'].depth, **keywords)
            wopOrigin = self.WOP_ORIGIN
        else:
            plane, planeIsNew, positionOnPlane = self.__calculatePlane(**keywords)
            sideID = plane.sideID

            contourOriginFrontLeft = self.getMachiningOriginFrontLeft(**keywords)
            matrix = plane.wopMatrix * contourOriginFrontLeft.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"

            contour.path = contourOriginFrontLeft.transformedPolyline(matrix.getInverted(), positionOnPlane)

            zValue = depth = - contour.depth
            wopOrigin = self.WOP_ORIGIN_FRONT_LEFT

        return joinIfNotNone('',
            self.createCIXStraightSide(plane=plane, fakeVirtualFaceLength=1, fakeVirtualFaceHeight=1, **keywords) if planeIsNew else '',
            super().createDefinitionForOneContour(sideID=sideID, depth=depth, zValue=zValue, wopOrigin=wopOrigin, **keywords),
        )

    def createDefinitionForOneDado(self, **keywords):
        dado = keywords['dado']
        if dado.isVertical():
            plane = None  # TODO (HORIZONTAL) IT-230: can we be more precise? -- NG
            planeIsNew = False
            sideID = self.SIDE_MAIN
            keywords['zValue'] = self.getDadoRabbetZValue(depth=keywords['machining'].depth, **keywords)
            wopOrigin = self.WOP_ORIGIN
        else:
            plane, planeIsNew, positionOnPlane = self.__calculatePlane(**keywords)

            dadoOriginFrontLeft = self.getMachiningOriginFrontLeft(**keywords)
            matrix = plane.wopMatrix * dadoOriginFrontLeft.matrix.getInverted()

            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            dado.matrix = matrix  # TODO (HORIZONTAL) IT-230: unschönes Design -- NG
            dado.matrix.origin = positionOnPlane

            sideID = plane.sideID
            keywords['zValue'] = - dado.depth
            wopOrigin = self.WOP_ORIGIN_FRONT_LEFT

        return joinIfNotNone('',
            self.createCIXStraightSide(plane=plane, fakeVirtualFaceLength=dado.getLength(), fakeVirtualFaceHeight=dado.getWidth(), **keywords) if planeIsNew else '',
            super().createDefinitionForOneDado(sideID=sideID, wopOrigin=wopOrigin, **keywords),
        )

    def createCommonDadoWithRouter(self, **keywords):
        dado = keywords['dado']
        keywords['depth'] = dado.depth

        radiusCompensation = self.getDadoRadiusCompensation(**keywords)

        approachMode = self.APPROACH_MODE_LINEAR_FLYING if dado.isInsetBoth() else self.APPROACH_MODE_LINEAR_PIERCING

        depthPerStep = self.getRouterDepthPerStep(machiningDepth=dado.actualDepth, **keywords)
        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=dado.actualDepth, depthPerStep=depthPerStep, **keywords)

        return self.createCommonContour(
            radiusCompensation=radiusCompensation,
            approach=approachMode,
            departure=approachMode,
            approachDepartureDistance=self.getDadoRabbetSecurityDistance(**keywords),
            numberOfSteps=numberOfSteps,
            **keywords
        )

    def createVerticalDadoWithRouter(self, **keywords):
        return self.createCommonDadoWithRouter(**keywords)

    def createHorizontalDadoWithRouter(self, **keywords):
        return self.createCommonDadoWithRouter(depth=keywords['dado'].depth, **keywords)

    def createVerticalRabbet(self, **keywords):
        rabbet = keywords['rabbet']
        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        depthPerStep = self.getRouterDepthPerStep(machiningDepth=rabbet.actualDepth, **keywords)
        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=rabbet.actualDepth, depthPerStep=depthPerStep, **keywords)

        return self.createCommonContour(
            radiusCompensation=radiusCompensation,
            approach=self.APPROACH_MODE_LINEAR_PIERCING,
            departure=self.APPROACH_MODE_LINEAR_PIERCING,
            approachDepartureDistance=self.getDadoRabbetSecurityDistance(**keywords),
            depth=keywords['rabbet'].depth,
            numberOfSteps=numberOfSteps,
            **keywords
        )

    def createNCStop(self, **keywords):
        return ''.join((
            self.createISOComment(**keywords),
            self.createCIXWait(**keywords)
        ))

    # NC Macro
    def getNCMacroCookedVariables(self, **keywords):
        variables = super().getNCMacroCookedVariables(**keywords)

        ncmacro = keywords['ncmacro']
        if ncmacro.isVertical():
            sideid = self.SIDE_MAIN
            sideCAngle = variables['cangle']
        else:
            plane = keywords['plane']
            sideid = plane.sideID

            machiningOriginFrontLeft = self.getMachiningOriginFrontLeft(**keywords)

            angle = VectorMath.angleBetween(machiningOriginFrontLeft.matrix.u, plane.wopMatrix.u)

            if not isNearlyEqual(angle, 0.0):
                crossProduct = machiningOriginFrontLeft.matrix.u.cross(plane.wopMatrix.u)
                if not isNearlyEqual(crossProduct.length(), 0.0) and crossProduct.normalized() == plane.wopMatrix.w:
                    angle = - angle

            sideCAngle = self.formatter.formatAsAngle(normalizeAngle(angle))

        self.lastNCMacroIndex += 1
        variables['index'] = self.lastNCMacroIndex

        variables['sideid'] = sideid
        variables['sidex'] = self.formatter.formatAsCoordinate(keywords['positionOnPlane'].x)
        variables['sidey'] = self.formatter.formatAsCoordinate(keywords['positionOnPlane'].y)
        variables['sidecangle'] = sideCAngle

        return variables

    def createNCMacroCode(self, **keywords):
        ncmacro = keywords['ncmacro']

        if ncmacro.isVertical():
            plane = None
            planeIsNew = False
            machiningOriginFrontLeft = self.getMachiningOriginFrontLeft(**keywords)
            positionOnPlane = machiningOriginFrontLeft.position
        else:
            plane, planeIsNew, positionOnPlane = self.__calculatePlane(**keywords)

        return joinIfNotNone('',
            self.createCIXStraightSide(plane=plane, fakeVirtualFaceLength=ncmacro.deltax, fakeVirtualFaceHeight=ncmacro.deltay, **keywords) if planeIsNew else '',
            super().createNCMacroCode(plane=plane, positionOnPlane=positionOnPlane, **keywords),
        )

    # Footer
    def createEndOfFile(self, **_keywords):
        return ''

    def getLocalizedMachiningName(self, index, default):
        return getLocalizedString(BIESSE_CIX_NAME_MESSAGES, index, default)
