# -*- coding: utf-8 -*-

# Copyright 2015-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections
import copy
import math
import re

import ezdxf

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.ncexportmigrated.rabbettoncsegments import RabbetToNCSegments
from interiorcad.interface.geometry import isNearlyGreater
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.simplelinesegment import SimpleLineSegment
from interiorcad.backend.ncexportmigrated.dadotoncsegments import getMiddleLine, getPerimeterCW, getPerimeterCCW, getDadoPerimeterCCW, getDadoPerimeterCW, getStartEndLine, getDadoContourCW, getDadoContourCCW, getDadoRoundedContourCW, getDadoRoundedContourCCW
from interiorcad.backend.ncexportmigrated.pockettoncsegments import generateTwoSegmentsCCW, generateTwoSegmentsCW, generateTwoSegmentsCCWWithApproach, generateTwoSegmentsCWWithApproach
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.ncexport.approachcreator import createArcApproach, createArcDeparture, createLinearApproach, createLinearDeparture
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.ncexport.base import NCExportBase


SimpleContourPoint = collections.namedtuple('SimpleContourPoint', 'x y radius')


### IT-145: is there a good idea to avoid  the """return ''""" without having to change 90% of our NCExportBase code? -- NG
class DXFExport(NCExportBase):
    EXPORT_FORMAT = 'DXF'

    RADIUS_COMPENSATION_NONE = NCExportBase.DIN_RADIUS_COMPENSATION_NONE
    RADIUS_COMPENSATION_LEFT = NCExportBase.DIN_RADIUS_COMPENSATION_LEFT
    RADIUS_COMPENSATION_RIGHT = NCExportBase.DIN_RADIUS_COMPENSATION_RIGHT

    def __init__(self, **keywords):
        super().__init__(**keywords)
        machineConfiguration = self.getMachineConfiguration()
        self.DXF_VERSION = machineConfiguration.get('dxf-version', 'AC1024')
        self.WORKPIECE_LAYER_NAME = machineConfiguration.get('workpiece-layer-name', 'Workpiece')
        self.EXPORT_CIRCLE_AS_CONTOUR_START_POINT = machineConfiguration.getAsBool('export-circle-as-contour-start-point', False)
        self.EXPORT_DRILLINGS_AS_CIRCLES = machineConfiguration.getAsBool('export-drillings-as-circles', True)
        self.EXPORT_HORIZONTAL_DRILLINGS_AS_LINES = machineConfiguration.getAsBool('export-horizontal-drillings-as-lines', True)
        self.EXPORT_POCKETS_AS_CONTOURS = machineConfiguration.getAsBool('export-pockets-as-contours', True)
        self.POCKET_REPLACEMENT = machineConfiguration.get('pocket-replacement', 'circle-CW')
        self.POCKET_PIERCING_REPLACEMENT = machineConfiguration.get('pocket-piercing-replacement', 'circle-CW')
        self.MITER_CW = machineConfiguration.getAsBool('miter-clock-wise', False)
        self.USE_ATTRIBUTE_THICKNESS = machineConfiguration.getAsBool('use-attribute-thickness', True)  # TODO: do we really want it switched on by default? (Does it work correctly in EnRoute) (IT-199)
        self.SIZING_ENTRY_OF_PROFILE = machineConfiguration.get(MachineConfigurationKeys.SIZING_ENTRY_OF_PROFILE_KEY, 'straightaway')
        self.SIZING_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('sizing-approach-departure-distance', 30)
        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('contour-approach-departure-distance', 30)
        self.RADIUS_COMPENSATION_NAMES = machineConfiguration.get('radius-compensation-names', '')

        self.createZValueConverters(**keywords)
        self.createDXFColorNumbersForMachinings(**keywords)

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)  # actually not integrated in machine-configuration
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueWorkpieceConverter = self.createZValueConverter('workpiece-z-reference', ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)

    def createDXFColorNumbersForMachinings(self, **_keywords):
        machineConfiguration = self.getMachineConfiguration()
        self.CONTOUR_DXF_COLOR_NUMBER = machineConfiguration.get('contour-dxf-color-number', '')
        self.CONTOUR_START_POINT_DXF_COLOR_NUMBER = machineConfiguration.get('contour-start-point-dxf-color-number', '')
        self.DADO_DXF_COLOR_NUMBER = machineConfiguration.get('dado-dxf-color-number', '')
        self.HORIZONTAL_DRILLINGS_DXF_COLOR_NUMBER = machineConfiguration.get('horizontal-drillings-dxf-color-number', '')
        self.MITER_DXF_COLOR_NUMBER = machineConfiguration.get('miter-dxf-color-number', '')
        self.POCKETS_DXF_COLOR_NUMBER = machineConfiguration.get('pockets-dxf-color-number', '')
        self.VERTICAL_DRILLINGS_DXF_COLOR_NUMBER = machineConfiguration.get('vertical-drillings-dxf-color-number', '')
        self.WORKPIECE_DXF_COLOR_NUMBER = machineConfiguration.get('workpiece-dxf-color-number', '')

    # helper
    def getDecimalSeparator(self, **keywords):
        return self.getMachineConfiguration().get('decimal-separator', '.')

    def writeToFile(self, content, filepath):
        # drawing.saveas do not set 'newline', making the AcceptanceTests unstable between Mac and Windows. So we have to open the file on our own.
        with self.openFileForWritingText(filepath) as exportFile:
            self.drawing.write(exportFile)

    def getNCFileExtension(self, **_keywords):
        return '.dxf'

    def createLayer(self, **keywords):
        layerName = keywords['layerName']
        assert layerName, 'A layer name cannot be empty!'
        if layerName not in self.drawing.layers:
            self.drawing.layers.new(name=layerName)

    def getCookedLayerName(self, **keywords):
        return keywords['template'].format_map(keywords['variables'])

    def createDXFAttributeThickness(self, **keywords):
        if self.USE_ATTRIBUTE_THICKNESS:
            keywords['dxfattribs'].update({'thickness': keywords['dxfThickness']})

    def createDXFAttributeColor(self, **keywords):
        dxfColor = keywords['dxfColor']
        if dxfColor != '':
            keywords['dxfattribs'].update({'color': dxfColor})

    def getRadiusCompensationString(self, **keywords):
        radiusCompensation = keywords['radiusCompensation']

        if self.RADIUS_COMPENSATION_NAMES:
            return self.RADIUS_COMPENSATION_NAMES.split(',')[radiusCompensation]
        else:
            return str(radiusCompensation)

    # header
    def getWorkpieceZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        return self.zValueWorkpieceConverter.convertWithThickness(boardThickness, boardThickness)

    def createWorkpieceLayer(self, **keywords):
        thickness = self.getWorkpieceZValue(**keywords)
        layerName = self.getCookedLayerName(template=self.WORKPIECE_LAYER_NAME, variables={'thickness': thickness}, **keywords)
        if layerName != '':
            self.createLayer(layerName=layerName)

            dxfattribs = {'layer': layerName}
            self.createDXFAttributeThickness(dxfattribs=dxfattribs, dxfThickness=thickness, **keywords)
            self.createDXFAttributeColor(dxfattribs=dxfattribs, dxfColor=self.WORKPIECE_DXF_COLOR_NUMBER, **keywords)

            surface = keywords['surface']
            modelspace = self.drawing.modelspace()
            polyline = modelspace.add_lwpolyline(surface.getFinishedSizeBoundingPoints(), dxfattribs=dxfattribs)
            polyline.closed = True

    def createHead(self, **keywords):
        ### IT-145: ezdxf.new is slow, as it reads a file. (Perhaps only relevant for AcceptanceTests) -- NG
        self.drawing = ezdxf.new(dxfversion=self.DXF_VERSION)
        self.createWorkpieceLayer(**keywords)
        return ''

    # machinings
    DXF_DUMMY_WIDTH = 0

    # drillings
    def createVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']
        depth = self.formatAsCoordinate(data=keywords['depth'])
        layerName = self.getCookedLayerName(template=drilling.toolID, variables={'depth': depth}, **keywords)
        self.createLayer(layerName=layerName)

        dxfattribs = {'layer': layerName}
        self.createDXFAttributeThickness(dxfattribs=dxfattribs, dxfThickness=depth, **keywords)
        self.createDXFAttributeColor(dxfattribs=dxfattribs, dxfColor=self.VERTICAL_DRILLINGS_DXF_COLOR_NUMBER, **keywords)

        modelspace = self.drawing.modelspace()
        if self.EXPORT_DRILLINGS_AS_CIRCLES:
            modelspace.add_circle((drilling.position.x, drilling.position.y), radius=drilling.diameter / 2, dxfattribs=dxfattribs)
        else:
            modelspace.add_point((drilling.position.x, drilling.position.y), dxfattribs=dxfattribs)

        return ''

    def createVerticalDrillingPiercing(self, **keywords):
        return self.createVerticalDrilling(**keywords)

    def createRotatedHorizontalDrilling(self, **keywords):
        drilling = keywords['drilling']

        startPosition = drilling.position
        endPosition = copy.copy(startPosition)
        endPosition.x += drilling.depth * math.cos(math.radians(drilling.cAngle))
        endPosition.y += drilling.depth * math.sin(math.radians(drilling.cAngle))
        zPosition = self.formatAsCoordinate(data=keywords['zPosition'])

        variables = {
            'depth': self.formatAsCoordinate(data=drilling.depth),
            'z': zPosition,
            'aangle': self.formatAsAngle(data=drilling.aAngle),
            'cangle': self.formatAsAngle(data=drilling.cAngle),
        }
        layerName = self.getCookedLayerName(template=drilling.toolID, variables=variables, **keywords)
        self.createLayer(layerName=layerName)

        dxfattribs = {'layer': layerName}
        self.createDXFAttributeThickness(dxfattribs=dxfattribs, dxfThickness=zPosition, **keywords)
        self.createDXFAttributeColor(dxfattribs=dxfattribs, dxfColor=self.HORIZONTAL_DRILLINGS_DXF_COLOR_NUMBER, **keywords)

        modelspace = self.drawing.modelspace()
        if self.EXPORT_HORIZONTAL_DRILLINGS_AS_LINES:
            modelspace.add_line(
                start=self.roundPoint(point=(startPosition.x, startPosition.y)),
                end=self.roundPoint(point=(endPosition.x, endPosition.y)),
                dxfattribs=dxfattribs
            )
        else:
            drillingSegment = SimpleLineSegment(start=Point(startPosition.x, startPosition.y), end=Point(endPosition.x, endPosition.y))
            rightSegment = copy.copy(drillingSegment)
            rightSegment.moveRight(drilling.diameter / 2)
            leftSegment = copy.copy(drillingSegment)
            leftSegment.moveLeft(drilling.diameter / 2)
            points = (
                rightSegment.getStartPoint().asTuple(),
                rightSegment.getEndPoint().asTuple(),
                leftSegment.getEndPoint().asTuple(),
                leftSegment.getStartPoint().asTuple(),
            )
            polyline = modelspace.add_lwpolyline(self.roundPoints(points=points), dxfattribs=dxfattribs)
            polyline.closed = True

        return ''

    # roundPoints help to get stable tests by rounding point positions. But interiorcad users should get the whole precision, that is why this function does nothing.
    def roundPoints(self, **keywords):
        return keywords['points']

    # roundPoint help to get stable tests by rounding point positions. But interiorcad users should get the whole precision, that is why this function does nothing.
    def roundPoint(self, **keywords):
        return keywords['point']

    # workings with contours
    def createCircleForContourStartPoint(self, **keywords):
        if self.EXPORT_CIRCLE_AS_CONTOUR_START_POINT:
            if 'approach' in keywords and 'flying' in keywords['approach']:
                if self.zValueMillingConverter.originReferenceSystem == ZValueConstants.ORIGIN_ON_TOP:
                    thickness = 0
                else:
                    thickness = self.getWorkpieceZValue(**keywords)
                self.createDXFAttributeThickness(dxfThickness=thickness, **keywords)

            keywords['dxfColor'] = self.CONTOUR_START_POINT_DXF_COLOR_NUMBER
            self.createDXFAttributeColor(**keywords)

            modelspace = self.drawing.modelspace()
            modelspace.add_circle(keywords['point'], radius=keywords['machining'].tool.getRouterDiameter() / 2, dxfattribs=keywords['dxfattribs'])

        return ''

    def createDXFLWPolyline(self, **keywords):
        zValue = self.formatAsCoordinate(data=keywords['zValue'])

        radiusCompensationString = self.getRadiusCompensationString(**keywords)
        variables = {
            'depth': zValue,
            'radiuscompensation': radiusCompensationString,
            'rc': radiusCompensationString,  # Layer names are long enough; we do not want that users have to always type the long {radiuscompensation}
        }
        if 'layerNameVariables' in keywords:
            variables.update(keywords['layerNameVariables'])
        layerName = self.getCookedLayerName(template=keywords['machining'].toolID, variables=variables, **keywords)
        self.createLayer(layerName=layerName)

        dxfattribs = {'layer': layerName}
        self.createDXFAttributeThickness(dxfattribs=dxfattribs, dxfThickness=zValue, **keywords)
        self.createDXFAttributeColor(dxfattribs=dxfattribs, **keywords)

        modelspace = self.drawing.modelspace()
        self.dxfPolyline = modelspace.add_lwpolyline(keywords['points'], dxfattribs=dxfattribs)

        with self.dxfPolyline.points() as points:
            self.createCircleForContourStartPoint(point=points[0][:2], dxfattribs=dxfattribs, **keywords)

        return ''

    def createCommonStartBlock(self, **keywords):
        position = keywords['startVertex'].position
        return self.createDXFLWPolyline(points=((position.x, position.y),), **keywords)

    def createCommonStartBlockWithArcApproach(self, **keywords):
        isClockwise, _approachDelta, approachPoint, _centerPoint, _tangentialAngle = createArcApproach(left=keywords['radiusCompensation'] == self.RADIUS_COMPENSATION_LEFT, **keywords)

        startVertex = keywords['startVertex']

        rawPoints = (
            SimpleContourPoint(approachPoint.x, approachPoint.y, None),
            SimpleContourPoint(startVertex.position.x, startVertex.position.y, keywords['securityDistance']),
        )

        bulgeFactor = -1 if isClockwise else 1
        return self.createDXFLWPolyline(points=self.generateDXFPointsFromSimpleContourPoints(points=rawPoints, bulgeFactor=bulgeFactor), **keywords)

    def createCommonEndBlockWithArcDeparture(self, **keywords):
        isClockwise, _4departureDelta, departurePoint, _centerPoint = createArcDeparture(left=keywords['radiusCompensation'] == self.RADIUS_COMPENSATION_LEFT, **keywords)

        endVertex = keywords['segmentList'][-1].getEndVertex()
        securityDistance = keywords['securityDistance']

        rawPoints = (
            SimpleContourPoint(endVertex.position.x, endVertex.position.y, None),
            SimpleContourPoint(departurePoint.x, departurePoint.y, securityDistance),
        )

        bulgeFactor = -1 if isClockwise else 1
        with self.dxfPolyline.points() as points:
            points[-1:] = list(self.generateDXFPointsFromSimpleContourPoints(points=rawPoints, bulgeFactor=bulgeFactor))

        return ''

    def createCommonStartBlockWithLinearApproach(self, **keywords):
        _approachDelta, approachPoint = createLinearApproach(ncexport=self, **keywords)

        startVertex = keywords['startVertex']

        rawPoints = (
            SimpleContourPoint(approachPoint.x, approachPoint.y, None),
            SimpleContourPoint(startVertex.position.x, startVertex.position.y, None),
        )

        return self.createDXFLWPolyline(points=self.generateDXFPointsFromSimpleContourPoints(points=rawPoints, bulgeFactor=None), **keywords)

    def createCommonEndBlockWithLinearDeparture(self, **keywords):
        _departureDelta, departurePoint = createLinearDeparture(ncexport=self, **keywords)

        endVertex = keywords['segmentList'][-1].getEndVertex()

        rawPoints = (
            SimpleContourPoint(endVertex.position.x, endVertex.position.y, None),
            SimpleContourPoint(departurePoint.x, departurePoint.y, None),
        )

        with self.dxfPolyline.points() as points:
            points[-1:] = list(self.generateDXFPointsFromSimpleContourPoints(points=rawPoints, bulgeFactor=None))

        return ''

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        position = segment.getEndVertex().position
        self.dxfPolyline.append_points((self.roundPoint(point=(position.x, position.y)),))
        return ''

    def getBulge(self, startPosition, endPosition, radius, isBigAngle, bulgeFactor):
        # For the math, see the formulas of the drawing under "Version 1" of http://lee-mac.com/bulgeconversion.html
        calcAngle = (endPosition - startPosition).magnitude() / radius / 2
        if calcAngle < -1:
            calcAngle = -1
        elif calcAngle > 1:
            calcAngle = 1
        angle = 2 * math.asin(calcAngle)
        if isBigAngle:
            angle = 2 * math.pi - angle
        bulge = math.tan(angle / 4)
        return bulge * bulgeFactor

    def getDxfPolylineData(self, **keywords):
        CW = -1, False
        CCW = 1, False
        CW180 = -1, True
        CCW180 = 1, True

        segment = keywords['segment']
        bulgeFactor, isBigAngle = segment.getDirection(CW, CCW, CW180, CCW180)
        startPosition = segment.getStartVertex().position
        endPosition = segment.getEndVertex().position
        radius = segment.getRadius()

        bulge = self.getBulge(startPosition, endPosition, radius, isBigAngle, bulgeFactor)

        return startPosition, endPosition, bulge

    def createRadiusSegment(self, **keywords):
        startPosition, endPosition, bulge = self.getDxfPolylineData(**keywords)

        with self.dxfPolyline.points() as points:
            points[-1:] = [
                (startPosition.x, startPosition.y, self.DXF_DUMMY_WIDTH, self.DXF_DUMMY_WIDTH, bulge),
                (endPosition.x, endPosition.y),
            ]
        return ''

    # sizing
    ### IT-145: why do we have similar code for radius compensation for all types (Sizing, Rabbet, Contour) for all WOPs? Would it be easier to tell centrally for a WOP how none, left, right is designed (especially as most of the time it is DIN 0/1/2)? -- NG.
    def createSizingContour(self, **keywords):
        radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(sizing=keywords['surface'].sizing, **keywords) else DXFExport.DIN_RADIUS_COMPENSATION_LEFT
        return self.createSizingContourDefinitionAsExecution(radiusCompensation=radiusCompensation, **keywords)

    def createCommonStartBlockWithPotentialApproach(self, **keywords):
        if 'arc' in keywords['approach']:
            return self.createCommonStartBlockWithArcApproach(**keywords)
        elif 'linear' in keywords['approach']:
            return self.createCommonStartBlockWithLinearApproach(**keywords)
        else:
            return self.createCommonStartBlock(**keywords)

    def createCommondEndBlockWithPotentialDeparture(self, **keywords):
        if 'arc' in keywords['departure']:
            return self.createCommonEndBlockWithArcDeparture(**keywords)
        elif 'linear' in keywords['departure']:
            return self.createCommonEndBlockWithLinearDeparture(**keywords)
        elif keywords['allowClosingPolygon']:
            self.dxfPolyline.closed = True

        return ''

    def createSizingStartBlock(self, **keywords):
        return self.createCommonStartBlockWithPotentialApproach(
            approach=self.SIZING_ENTRY_OF_PROFILE,
            securityDistance=self.SIZING_APPROACH_DEPARTURE_DISTANCE,
            dxfColor=self.CONTOUR_DXF_COLOR_NUMBER,
            **keywords)

    def createSizingEndBlock(self, **keywords):
        return self.createCommondEndBlockWithPotentialDeparture(departure=self.SIZING_ENTRY_OF_PROFILE, securityDistance=self.SIZING_APPROACH_DEPARTURE_DISTANCE, allowClosingPolygon=True, **keywords)

    # rabbets
    def createVerticalRabbet(self, **keywords):
        return self.createVerticalRabbetOffsetByContourDistance(**keywords)

    def generateRabbetSegments(self, **keywords):
        rabbet = keywords['rabbet']
        securityDistance = self.getDadoRabbetSecurityDistance(**keywords)
        if self.RABBET_REPLACEMENT.startswith('rounded-'):
            routerRadius = rabbet.tool.getRouterDiameter() / 2
            return tuple(RabbetToNCSegments.generateRabbetRoundedSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=securityDistance, straightSecurityDistance=securityDistance, routerRadius=routerRadius))
        elif self.RABBET_REPLACEMENT.startswith('contour-'):
            return tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=securityDistance, straightSecurityDistance=securityDistance))
        else:
            return tuple(RabbetToNCSegments.generateRabbetSegmentsWithInsets(rabbet=rabbet))

    def createRabbetContourStartBlock(self, **keywords):
        radiusCompensation = self.getDinRadiusCompensation(**keywords)
        return self.createCommonStartBlock(
            radiusCompensation=radiusCompensation,
            dxfColor=self.CONTOUR_DXF_COLOR_NUMBER,
            **keywords)

    def createRabbetContourEndBlock(self, **_keywords):
        return ''

    # contours
    def createVerticalContour(self, **keywords):
        zValue = self.getContourZValue(depth=keywords['contour'].depth, **keywords)
        return self.createDefinitionForOneContour(zValue=zValue, **keywords)

    def createContourStartBlock(self, **keywords):
        radiusCompensation = self.getDinRadiusCompensation(**keywords)
        return self.createCommonStartBlockWithPotentialApproach(
            radiusCompensation=radiusCompensation,
            approach=keywords['contour'].approach,
            securityDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
            dxfColor=self.CONTOUR_DXF_COLOR_NUMBER,
            **keywords)

    def createContourEndBlock(self, **keywords):
        radiusCompensation = self.getDinRadiusCompensation(**keywords)
        return self.createCommondEndBlockWithPotentialDeparture(
            radiusCompensation=radiusCompensation,
            departure=keywords['contour'].departure,
            securityDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
            allowClosingPolygon=False,
            **keywords)

    def __createDummyContourEndBlock(self, **_keywords):
        return ''

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createVerticalContour, **keywords))

    # pockets
    def generateClearOutData(self, **keywords):
        # Code ported from Masterwork Classic template (KREISTASCHE), radius refactored ("Introduce Variable") and rotationFactor added
        pocket = keywords['pocket']
        rotationFactor = keywords['rotationFactor']

        routerDiameter = pocket.tool.getRouterDiameter()
        assert isNearlyGreater(routerDiameter, 0), "The pocket tool '{toolName}' must have a diameter greater than 0.".format(toolName=pocket.tool.name)

        yield SimpleContourPoint(pocket.position.x, pocket.position.y, None)
        for i in range(1, int(pocket.diameter // routerDiameter) - 1):
            radius = i * routerDiameter / 2
            if radius < pocket.diameter:
                yield SimpleContourPoint(pocket.position.x - radius, pocket.position.y, None)
                yield SimpleContourPoint(pocket.position.x, pocket.position.y - radius * rotationFactor, radius)
                yield SimpleContourPoint(pocket.position.x + radius, pocket.position.y, radius)
                yield SimpleContourPoint(pocket.position.x, pocket.position.y + radius * rotationFactor, radius)
                yield SimpleContourPoint(pocket.position.x - radius, pocket.position.y, radius)

        radius = pocket.diameter / 2 - routerDiameter / 2
        if pocket.diameter > routerDiameter * (pocket.diameter // routerDiameter):
            yield SimpleContourPoint(pocket.position.x - radius, pocket.position.y, None)
            yield SimpleContourPoint(pocket.position.x, pocket.position.y - radius * rotationFactor, radius)
            yield SimpleContourPoint(pocket.position.x + radius, pocket.position.y, radius)
            yield SimpleContourPoint(pocket.position.x, pocket.position.y + radius * rotationFactor, radius)
            yield SimpleContourPoint(pocket.position.x - radius, pocket.position.y, radius)
        else:
            yield SimpleContourPoint(pocket.position.x - radius, pocket.position.y, radius)

    def generateDXFPointsFromSimpleContourPoints(self, **keywords):
        points = keywords['points']
        for i in range(len(points) - 1):
            if points[i + 1].radius:
                bulge = self.getBulge(
                    startPosition=Point(points[i].x, points[i].y),
                    endPosition=Point(points[i + 1].x, points[i + 1].y),
                    radius=points[i + 1].radius,
                    isBigAngle=False,
                    bulgeFactor=keywords['bulgeFactor']
                )
            else:
                bulge = 0
            yield points[i].x, points[i].y, self.DXF_DUMMY_WIDTH, self.DXF_DUMMY_WIDTH, bulge

        yield points[-1].x, points[-1].y

    def pocketGeometryShouldBeCreatedInCWDirection(self, **keywords):
        pocketReplacement = keywords['pocketReplacement']
        pocket = keywords['pocket']

        return pocketReplacement.endswith('-CW') or (pocketReplacement.endswith('-opposite') and not self.isDadoRabbetReplacementCW(pocket, self.POCKET_REPLACEMENT)) or (pocketReplacement.endswith('-same') and self.isDadoRabbetReplacementCW(pocket, self.POCKET_REPLACEMENT))

    def createVerticalPocketSpiral(self, **keywords):
        rotationFactor = -1 if self.pocketGeometryShouldBeCreatedInCWDirection(**keywords) else 1

        spiralData = tuple(self.generateClearOutData(rotationFactor=rotationFactor, **keywords))

        self.createDXFLWPolyline(
            points=self.generateDXFPointsFromSimpleContourPoints(points=spiralData, bulgeFactor=rotationFactor),
            radiusCompensation=self.DIN_RADIUS_COMPENSATION_NONE,
            zValue=keywords['depth'],
            dxfColor=self.CONTOUR_DXF_COLOR_NUMBER,
            **keywords)

    def createVerticalPocketCircle(self, **keywords):
        pocketReplacement = keywords['pocketReplacement']
        if self.pocketGeometryShouldBeCreatedInCWDirection(**keywords):
            if 'approach' in pocketReplacement:
                generator = generateTwoSegmentsCWWithApproach
            else:
                generator = generateTwoSegmentsCW
            radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_RIGHT
        else:
            if 'approach' in pocketReplacement:
                generator = generateTwoSegmentsCCWWithApproach
            else:
                generator = generateTwoSegmentsCCW
            radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_LEFT

        segmentList = tuple(generator(pocket=keywords['pocket']))
        self.createContourFromSegmentList(
            createStartBlock=self.createCommonStartBlock,
            createEndBlock=self.__createDummyContourEndBlock,
            radiusCompensation=radiusCompensation,
            segmentList=segmentList,
            zValue=keywords['depth'],
            dxfColor=self.CONTOUR_DXF_COLOR_NUMBER,
            **keywords)

    def createVerticalPocketCode(self, **keywords):
        pocketReplacement = keywords['pocketReplacement']

        if 'spiral' in pocketReplacement:
            self.createVerticalPocketSpiral(**keywords)

        if 'circle' in pocketReplacement:
            self.createVerticalPocketCircle(**keywords)

        return ''

    def createVerticalPocketAsCircleCode(self, **keywords):
        pocket = keywords['pocket']

        depth = self.formatAsCoordinate(data=keywords['depth'])
        radiusCompensationString = self.getRadiusCompensationString(radiusCompensation=self.RADIUS_COMPENSATION_NONE, **keywords)
        variables = {
            'depth': depth,
            'radiuscompensation': radiusCompensationString,
            'rc': radiusCompensationString,  # Layer names are long enough; we do not want that users have to always type the long {radiuscompensation}
        }
        layerName = self.getCookedLayerName(template=pocket.toolID, variables=variables, **keywords)
        self.createLayer(layerName=layerName)

        dxfattribs = {'layer': layerName}
        self.createDXFAttributeThickness(dxfattribs=dxfattribs, dxfThickness=depth, **keywords)
        self.createDXFAttributeColor(dxfattribs=dxfattribs, dxfColor=self.POCKETS_DXF_COLOR_NUMBER, **keywords)

        modelspace = self.drawing.modelspace()
        modelspace.add_circle((pocket.position.x, pocket.position.y), radius=pocket.diameter / 2, dxfattribs=dxfattribs)

        return ''

    def createVerticalPocket(self, **keywords):
        if self.EXPORT_POCKETS_AS_CONTOURS:
            return self.createVerticalPocketCode(pocketReplacement=self.POCKET_REPLACEMENT, **keywords)
        else:
            return self.createVerticalPocketAsCircleCode(**keywords)

    def createVerticalPocketPiercing(self, **keywords):
        if self.EXPORT_POCKETS_AS_CONTOURS:
            if keywords['pocket'].diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER:
                return self.createVerticalPocketCode(pocketReplacement=self.POCKET_PIERCING_REPLACEMENT, **keywords)
            else:
                return self.createVerticalPocketCode(pocketReplacement=self.POCKET_REPLACEMENT, **keywords)
        else:
            return self.createVerticalPocketAsCircleCode(**keywords)

    # dados
    REGEXP_PERIMETER_WITH_SECURITY_DISTANCE = re.compile(r'perimeter(-[A-Za-z]+?)-with-security-distance', re.A)

    def createVerticalDadoWithAnyTool(self, **keywords):
        dado = keywords['dado']
        securityDistance = self.getDadoRabbetSecurityDistance(**keywords)
        routerRadius = dado.tool.getRouterDiameter() / 2
        if self.DADO_REPLACEMENT == 'middle-line':
            segmentList = getMiddleLine(dado)
            radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_NONE
        elif self.DADO_REPLACEMENT.startswith('perimeter-') and self.DADO_REPLACEMENT.endswith('-with-security-distance'):
            match = self.REGEXP_PERIMETER_WITH_SECURITY_DISTANCE.match(self.DADO_REPLACEMENT)
            cw = self.isDadoRabbetReplacementCW(dado, match.group(1))

            if cw:
                segmentList = getDadoPerimeterCW(dado, securityDistance=securityDistance)
                radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_RIGHT
            else:
                segmentList = getDadoPerimeterCCW(dado, securityDistance=securityDistance)
                radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_LEFT
        else:
            cw = self.isDadoRabbetReplacementCW(dado, self.DADO_REPLACEMENT)

            if self.DADO_REPLACEMENT.startswith('perimeter-') and cw:
                segmentList = getPerimeterCW(dado)
                radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_RIGHT
            elif self.DADO_REPLACEMENT.startswith('perimeter-') and not cw:
                segmentList = getPerimeterCCW(dado)
                radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_LEFT
            elif self.DADO_REPLACEMENT.startswith('contour-') and cw:
                segmentList = getDadoContourCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
                radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_RIGHT
            elif self.DADO_REPLACEMENT.startswith('contour-') and not cw:
                segmentList = getDadoContourCCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
                radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_LEFT
            elif self.DADO_REPLACEMENT.startswith('rounded-contour-') and cw:
                segmentList = getDadoRoundedContourCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
                radiusCompensation = self.RADIUS_COMPENSATION_RIGHT
            elif self.DADO_REPLACEMENT.startswith('rounded-contour-') and not cw:
                segmentList = getDadoRoundedContourCCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
                radiusCompensation = self.RADIUS_COMPENSATION_LEFT
            else:
                assert False, "Unknown replacement"

        if 'perimeter' in self.DADO_REPLACEMENT and dado.tool.type == Tool.SAW:  # for saw only
            dxfColor = self.DADO_DXF_COLOR_NUMBER
        else:
            dxfColor = self.CONTOUR_DXF_COLOR_NUMBER

        return self.createContourFromSegmentList(
            createStartBlock=self.createCommonStartBlock,
            createEndBlock=self.__createDummyContourEndBlock,
            radiusCompensation=radiusCompensation,
            segmentList=segmentList,
            zValue=keywords['depth'],
            layerNameVariables={
                'width': self.formatAsCoordinate(data=dado.width),
                'aangle': self.formatAsAngle(data=90),  # Compatibility to miter, to be able to use the same saw
            },
            dxfColor=dxfColor,
            **keywords)

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        return self.createVerticalDadoWithAnyTool(**keywords)

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        return self.createVerticalDadoWithAnyTool(**keywords)

    def createVerticalDadoWithSawCRotated(self, **keywords):
        return self.createVerticalDadoWithAnyTool(**keywords)

    def createVerticalDadoWithRouter(self, **keywords):
        return self.createVerticalDadoWithAnyTool(**keywords)

    # miters
    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjusterFunction = adjuster.getMiterOrientedToCW if self.MITER_CW else adjuster.getMiterOrientedToCCW
        adjuster.orient(surface.mitersVertical, adjusterFunction)

    def createVerticalMiter(self, **keywords):
        surface = keywords['surface']
        miter = keywords['miter']
        segmentList = getStartEndLine(miter)
        radiusCompensation = DXFExport.DIN_RADIUS_COMPENSATION_LEFT if self.MITER_CW else DXFExport.DIN_RADIUS_COMPENSATION_RIGHT

        return self.createContourFromSegmentList(
            createStartBlock=self.createCommonStartBlock,
            createEndBlock=self.__createDummyContourEndBlock,
            radiusCompensation=radiusCompensation,
            segmentList=segmentList,
            zValue=self.zValueMiterSawConverter.convertWithThickness(surface.thickness, surface.thickness),
            layerNameVariables={'aangle': self.formatAsAngle(data=miter.getAngle())},
            dxfColor=self.MITER_DXF_COLOR_NUMBER,
            **keywords)

    def createNCMacroCode(self, **keywords):
        self.warnNCMacroNotSupported(**keywords)
        return ''

    #
    # Everything below is still from template.py
    #

    def createFiveAxisDrilling(self, **_keywords):
        return ''

    def createNCStop(self, **_keywords):
        return ''

    def createEndOfFile(self, **_keywords):
        return ''
