# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.localization import getLocalizedString, F4INTEGRATE_COMMENT_MESSAGES
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.backend.zvalue.converter import ZValueConverter
from interiorcad.interface import configuration
from interiorcad.interface.geometry import isNearlyEqual, isNearlyGreater
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.surface import Contour2D, HorizontalDrilling, HorizontalCenterPunch, FiveAxisDrilling, FiveAxisCenterPunch, NCMacro2D
from interiorcad.interface.surface.acangles import normalizeAngle
from interiorcad.interface.surface.contour2d import HorizontalContour2D
from interiorcad.interface.surface.dado2d import HorizontalDado2D
from interiorcad.interface.surface.machiningplanes import MachiningPlanes
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.interface.surface.pocket2d import HorizontalPocket2D
from interiorcad.ncexport.base import NCExportBase


class F4IntegrateExport(NCExportBase):
    EXPORT_FORMAT = 'F4Integrate'

    OUTPUT_ENCODING = 'utf-8-sig'

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.machiningPlanes = None
        self.currentPlane = None
        self.profileNameIndex = None

        machineConfiguration = self.getMachineConfiguration()
        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 2)
        self.ADDITIONAL_DEPTH_FOR_MITER = machineConfiguration.getAsFloat(MachineConfigurationKeys.ADDITIONAL_DEPTH_FOR_MITER_KEY, 5)
        self.DADO_PRESAW_DEPTH = machineConfiguration.getAsFloat('dado-presaw-depth', 0)
        self.MACHINING_AREA_DEFAULT = machineConfiguration.get('machining-area-default', 'AD1')
        self.SHOW_EDGES_MESSAGE_IN_F4OPERATE = machineConfiguration.getAsBool('show-edges-message-in-f4operate', True)
        self.USE_CUSTOM_USER_WORKPIECE_COMMENT = machineConfiguration.getAsBool('use-custom_user-workpiece-comment', False)
        self.CREATE_LABEL_CALL = machineConfiguration.getAsBool('create-label-call', False)
        self.CREATE_SIZING_AS_NESTING_GEOMETRY = machineConfiguration.getAsBool('create-sizing-as-nesting-geometry', False)

        self.createZValueConverters(**keywords)

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueHorizontalDrillingConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueVerticalPocketConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueFiveAxisDrillingConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMillingConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueDadoSawConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMiterPreSawConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMiterSawConverter = ZValueConverter(ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)

    def getNCFileExtension(self, **_keywords):
        return '.f4g'

    # Header
    def createHead(self, **keywords):
        self.currentPlane = None
        self.generatePlanes(**keywords)

        self.profileNameIndex = 0

        return joinIfNotNone('\n',
            self.createVersion(**keywords),
            self.createHeader(**keywords),
            self.createLabelCall(**keywords) if self.CREATE_LABEL_CALL else None,
            self.createEdgesMessage(**keywords) if self.SHOW_EDGES_MESSAGE_IN_F4OPERATE else None,
            '',
        )

    def createVersion(self, **keywords):
        # The variable must be kept and not to be renamed, to be found by tools/versionChanger
        majorVersion = 2024
        return '\n'.join((
            '[version]',
            "articleid='{id}'".format(id=self.getWorkpieceComments(**keywords)['WorkpieceName']),
            'revision={revision}'.format(revision=majorVersion),
            "date='{timestamp}'".format(timestamp=configuration.DATETIME_NOW.strftime('%d/%m/%Y %H:%M:%S')),
            '[/version]',
        ))

    def createCustomUserWorkpieceComment(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='F4Integrate Workpiece Comment.txt',
            customUserVariableFileDefaultText=self.createEdgesTextRaw(**keywords) + ' | {Comment}',
            **keywords)

    def createHeader(self, **keywords):
        surface = keywords['surface']

        customUserWorkpieceComment = self.createCustomUserWorkpieceComment(**keywords) if self.USE_CUSTOM_USER_WORKPIECE_COMMENT else self.getWorkpieceComments(**keywords)['Comment']

        return '\n'.join((
            '[header]',
            'l={length}'.format(length=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            'h={width}'.format(width=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            't={thick}'.format(thick=self.formatAsCoordinate(data=surface.thickness)),
            'isinch=0',
            'area={area}'.format(area=self.MACHINING_AREA_DEFAULT),
            'offsetx={x}'.format(x=self.formatAsCoordinate(data=surface.getAdditionX())),
            'offsety={y}'.format(y=self.formatAsCoordinate(data=surface.getAdditionY())),
            'offsetz=0',
            'rawxm=0',  # raw dimensions are added to offsets
            'rawym=0',
            'rawzm=0',
            'rawxp=0',
            'rawyp=0',
            'rawzp=0',
            "outfit='{outfit}'".format(outfit='Default'),  # the option is not documented # TODO IT-284 F4INTEGRATE: option should be machine configuration -- NG
            "comment='{comment}'".format(comment=customUserWorkpieceComment),
            '[/header]',
            '',
        ))

    def customUserLabelCallFileDefaultText(self, **_keywords):
        return '\n'.join((
            'G107P"F4_template"X(x/2)Y(y/2)Z0C0:',
            '{{material}}=("{ItemID}");',
            '{{name}}=("{NCProgramFilename}");',
            '{{edgeright}}={EdgeRightThickness};',
            '{{edgeleft}}={EdgeLeftThickness};',
            '{{edgerear}}={EdgeBackThickness};',
            '{{edgefront}}={EdgeFrontThickness};',
            '{{date}}=("{FreeDateTime:%d %m %Y %H:%M:%S}");',
            '{{path}}=("{NCProgramFolder}");',
            '{{comment}}=("{Comment}");',
            '{{project}}=("{ProjectID}");',
            '{{customer}}=("{ProjectCommission}");',
            '{{construction}}=("{ProjectConstruction}");',
            '{{positionnumber}}=("{PositionNumber}");',
            '{{positiondescription}}=("{PositionDescription}");',
            '{{cabinetname}}=("{CabinetName}")',
            '',
        ))

    def createLabelCall(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='F4Integrate Label call.txt',
            customUserVariableFileDefaultText=self.customUserLabelCallFileDefaultText(**keywords),
            **keywords)

    def createEdgesTextRaw(self, **_keywords):
        return ' | '.join((
            'KL = {EdgeLeftThickness}mm',
            'KV = {EdgeFrontThickness}mm',
            'KR = {EdgeRightThickness}mm',
            'KH = {EdgeBackThickness}mm',
        ))

    def createEdgesMessage(self, **keywords):
        workpieceComments = self.getWorkpieceComments(**keywords)

        return ''.join((
            'M50',
            "S('{textToShow}')".format(textToShow=self.createEdgesTextRaw(**keywords).format_map(workpieceComments)),
            'T0',  # OK button
            self.getMachiningComment(comment=self.getLocalizedMachiningComment(21, "Edges")),
        )) + '\n'

    SIDE_MAIN = 'TOP'
    SIDE_LEFT = 'LEFT'
    SIDE_RIGHT = 'RIGHT'
    SIDE_FRONT = 'FRONT'
    SIDE_BACK = 'BACK'
    SIDE_OPPOSITE = 'BASE'

    def __createPredefinedPlanes(self, **keywords):
        size = keywords['surface'].getFinishedSize()

        yield Plane(
            predefined=True, sideID=self.SIDE_FRONT,
            matrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_RIGHT,
            matrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_BACK,
            matrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_LEFT,
            matrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )

    def generatePlanes(self, **keywords):
        def sideIDGenerator():
            sideIDIndex = 0
            while True:
                sideIDIndex += 1
                yield "plane{index}".format(index=sideIDIndex)

        self.machiningPlanes = MachiningPlanes(sideIDGenerator=sideIDGenerator())
        self.machiningPlanes.planes.extend(self.__createPredefinedPlanes(**keywords))
        self.machiningPlanes.createPlanes(filter(lambda machining: machining.matrix.w != Vector(0, 0, 1),
            filter(lambda machining: isinstance(machining, (
                HorizontalDrilling, HorizontalCenterPunch, HorizontalDado2D, HorizontalContour2D, HorizontalPocket2D,
                FiveAxisDrilling, FiveAxisCenterPunch, NCMacro2D
            )), keywords['surface'].machinings),
        ))

    # Machinings
    ABSOLUTE_COORDINATES = 'G90'
    RELATIVE_COORDINATES = 'G91'

    RADIUS_COMPENSATION_NONE = 'G40'
    RADIUS_COMPENSATION_LEFT = 'G41'
    RADIUS_COMPENSATION_RIGHT = 'G42'

    def isValidToolId(self, **keywords):
        return 'tool' in keywords and keywords['tool'] and keywords['tool'].getToolID()

    def isValidNumericalToolId(self, **keywords):
        return 'tool' in keywords and keywords['tool'] and keywords['tool'].isToolIDValidIntDirect()

    def getToolSelection(self, **keywords):
        return "T" + self.getWOPToolID(**keywords)

    def getWOPToolID(self, **keywords):
        if self.isValidNumericalToolId(**keywords):
            return '{toolID}'.format(toolID=keywords['tool'].getToolID())
        elif self.isValidToolId(**keywords):
            return '"{toolID}"'.format(toolID=keywords['tool'].getToolID())
        else:
            assert False, "Something is wrong with the Tool ID"

    def getMachiningComment(self, **keywords):
        return ' #{comment}'.format(comment=keywords['comment'])

    def changePlane(self, **keywords):
        newPlane = keywords['plane']
        if newPlane != self.currentPlane:
            self.currentPlane = newPlane
            if newPlane is None:
                return self.SIDE_MAIN
            elif newPlane.predefined:
                return newPlane.sideID
            else:
                return self.createFictiveFaceDefinition(**keywords)

    def createFictiveFaceDefinition(self, **keywords):
        plane = keywords['plane']
        return ''.join((
            self.ABSOLUTE_COORDINATES,
            'G500',
            'T"{sideID}"'.format(sideID=plane.sideID),
            'X{x}'.format(x=self.formatAsAngle(data=plane.wopMatrix.origin.x)),
            'Y{y}'.format(y=self.formatAsAngle(data=plane.wopMatrix.origin.y)),
            'Z{z}'.format(z=self.formatAsAngle(data=plane.wopMatrix.origin.z)),
            'U{u}'.format(u=self.formatAsAngle(data=plane.wopMatrix.u.x)),
            'V{v}'.format(v=self.formatAsAngle(data=plane.wopMatrix.u.y)),
            'W{w}'.format(w=self.formatAsAngle(data=plane.wopMatrix.u.z)),
            'I{i}'.format(i=self.formatAsAngle(data=plane.wopMatrix.v.x)),
            'J{j}'.format(j=self.formatAsAngle(data=plane.wopMatrix.v.y)),
            'K{k}'.format(k=self.formatAsAngle(data=plane.wopMatrix.v.z)),
        ))

    def createNCStop(self, **keywords):
        return ''.join((
            'M50',
            "S('{comment}')".format(comment=keywords['comment']),
            'T0',  # OK button
            self.getMachiningComment(comment=self.getLocalizedMachiningComment(16, "NC-Stop")),
        )) + '\n'

    # Drillings
    FLAT_DRILL_TYPE = 0
    PIERCING_DRILL_TYPE = 1

    def __createCommonDrillingCode(self, **keywords):
        drilling = keywords['drilling']
        isToolValid = self.isValidToolId(tool=drilling.tool, **keywords)
        return joinIfNotNone('',
            self.ABSOLUTE_COORDINATES,
            'G81',
            'D{diameter};{drillType}'.format(diameter=self.formatAsCoordinate(data=drilling.installedDiameter), drillType=keywords['drillType']) if not isToolValid else None,
            'X{x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            'Y{y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            'Z{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            self.getToolSelection(tool=drilling.tool, **keywords) if isToolValid else None,
            self.getMachiningComment(**keywords),
        )

    def __createCommonDrilling(self, **keywords):
        return joinIfNotNone('\n',
            self.changePlane(**keywords),
            self.__createCommonDrillingCode(**keywords),
            '',
        )

    def createVerticalDrilling(self, **keywords):
        position = keywords['drilling'].position
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)
        return self.__createCommonDrilling(
            x=position.x,
            y=position.y,
            drillType=drillType,
            comment=self.getLocalizedMachiningComment(1, "Drilling"),
            plane=None,
            **keywords)

    def createVerticalDrillingPiercing(self, **keywords):
        position = keywords['drilling'].position
        drillType = self.getDrillType(defaultDrillType=self.PIERCING_DRILL_TYPE, **keywords)
        return self.__createCommonDrilling(
            x=position.x,
            y=position.y,
            drillType=drillType,
            comment=self.getLocalizedMachiningComment(3, "Drilling through"),
            plane=None,
            **keywords)

    def createVerticalCenterPunch(self, **keywords):
        position = keywords['drilling'].position
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)
        return self.__createCommonDrilling(
            x=position.x,
            y=position.y,
            drillType=drillType,
            comment=self.getLocalizedMachiningComment(2, "Center Punch"),
            plane=None,
            **keywords)

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        isToolValid = self.isValidToolId(tool=gridDrilling.tool, **keywords)
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)
        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            joinIfNotNone('',
                'G120',
                'D{diameter};{drillType}'.format(diameter=self.formatAsCoordinate(data=startDrilling.installedDiameter), drillType=drillType) if not isToolValid else None,
                'X{xs};{xe}'.format(xs=self.formatAsCoordinate(data=startDrilling.position.x), xe=self.formatAsCoordinate(data=endDrilling.position.x)),
                'Y{y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
                'Z{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
                'P{step}'.format(step=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                self.getToolSelection(tool=gridDrilling.tool, **keywords) if isToolValid else None,
                self.getMachiningComment(comment=self.getLocalizedMachiningComment(4, "Grid Drilling X"), **keywords),
            ),
            '',
        )

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        isToolValid = self.isValidToolId(tool=gridDrilling.tool, **keywords)
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)
        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            joinIfNotNone('',
                'G121',
                'D{diameter};{drillType}'.format(diameter=self.formatAsCoordinate(data=startDrilling.installedDiameter), drillType=drillType) if not isToolValid else None,
                'X{x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
                'Y{ys};{ye}'.format(ys=self.formatAsCoordinate(data=startDrilling.position.y), ye=self.formatAsCoordinate(data=endDrilling.position.y)),
                'Z{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
                'P{step}'.format(step=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                self.getToolSelection(tool=gridDrilling.tool, **keywords) if isToolValid else None,
                self.getMachiningComment(comment=self.getLocalizedMachiningComment(5, "Grid Drilling Y"), **keywords),
            ),
            '',
        )

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        isToolValid = self.isValidToolId(tool=gridDrilling.tool, **keywords)
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)
        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            joinIfNotNone('',
                'G124',
                'D{diameter};{drillType}'.format(diameter=self.formatAsCoordinate(data=startDrilling.installedDiameter), drillType=drillType) if not isToolValid else None,
                'X{xs}'.format(xs=self.formatAsCoordinate(data=startDrilling.position.x)),
                'Y{ys}'.format(ys=self.formatAsCoordinate(data=startDrilling.position.y)),
                'Z{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
                'J{xe}'.format(xe=self.formatAsCoordinate(data=endDrilling.position.x)),
                'K{ye}'.format(ye=self.formatAsCoordinate(data=endDrilling.position.y)),
                'P{step}'.format(step=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                self.getToolSelection(tool=gridDrilling.tool, **keywords) if isToolValid else None,
                self.getMachiningComment(comment=self.getLocalizedMachiningComment(25, "Grid Drilling X/Y"), **keywords),
            ),
            '',
        )

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']

        machiningPlane = Plane.createFromMachining(machining)

        plane = self.machiningPlanes.findPlane(machiningPlane)
        assert plane is not None, "No plane found for machining"

        return plane, machining.getPositionOnPlane(plane)

    def __createCommonFiveAxisDrilling(self, **keywords):
        drilling = keywords['drilling']
        if drilling.direction == Vector(0, 0, 1):
            x = drilling.position.x
            y = drilling.position.y
            plane = None
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            x = positionOnPlane.x
            y = positionOnPlane.y

        depth = - keywords['drilling'].depth
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)
        return self.__createCommonDrilling(plane=plane, x=x, y=y, depth=depth, drillType=drillType, **keywords)

    def createRotatedHorizontalDrilling(self, **keywords):
        return self.__createCommonFiveAxisDrilling(comment=self.getLocalizedMachiningComment(17, "Horizontal Drilling"), **keywords)

    def createFiveAxisDrilling(self, **keywords):
        return self.__createCommonFiveAxisDrilling(comment=self.getLocalizedMachiningComment(18, "5 Axis Drilling"), **keywords)

    # Pockets
    def createDepthPerStepForPockets(self, **keywords):
        pocket = keywords['pocket']

        actualDepth = pocket.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=actualDepth, **keywords)
        if isNearlyEqual(depthPerStep, 0.0) or isNearlyGreater(depthPerStep, actualDepth):  # needed for approach with helix
            zComplement = ';' + self.formatAsCoordinate(data=actualDepth)
        else:
            zComplement = ';' + self.formatAsCoordinate(data=depthPerStep)

        return zComplement

    def __createCommonPocket(self, **keywords):
        pocket = keywords['pocket']
        toolID = self.getWOPToolID(tool=pocket.tool, **keywords)

        if (pocket.installedDiameter - pocket.tool.getRouterDiameter()) / 10 < 0.75:
            sideDistanceFactor = '((' + self.formatAsCoordinate(data=pocket.installedDiameter) + '-TDIA({toolID}))/10)'.format(toolID=toolID)
        else:
            sideDistanceFactor = 0.75

        approachCode = self.getCommonContourApproachDepartureCode(
            approachDepartureParameter='I',
            approachDepartureMode=self.APPROACH_MODE_ARC_FLYING_WITHOUT_AUTOMATIC,  # Temporary workaround for pockets as long as automatic damages the pocket inside F4Integrate
            radiusCompensation=self.getPocketRadiusCompensation(**keywords),
            machiningDepth=pocket.depth,
            sideDistance='(TDIA({toolID})/2*{factor})'.format(toolID=toolID, factor=sideDistanceFactor),  # Using the formula as Felder suggests is more reproducible than the value only
            **keywords)

        departureCode = self.getCommonContourApproachDepartureCode(
            approachDepartureParameter='Q',
            approachDepartureMode=self.APPROACH_MODE_WITHOUT,
            **keywords)

        return ''.join((
            self.ABSOLUTE_COORDINATES,
            'G232',
            'X{x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            'Y{y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            'Z{depth}{complement}'.format(depth=self.formatAsCoordinate(data=keywords['depth']), complement=(self.createDepthPerStepForPockets(**keywords))),
            'D{diameter}'.format(diameter=self.formatAsCoordinate(data=pocket.installedDiameter)),
            'W{direction:d}'.format(direction=self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
            self.getToolSelection(tool=pocket.tool, **keywords),
            approachCode,
            departureCode,
            'J1',
            self.getMachiningComment(**keywords),
        ))

    def createVerticalPocket(self, **keywords):
        pocket = keywords['pocket']

        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            self.__createCommonPocket(
                x=pocket.position.x,
                y=pocket.position.y,
                comment=self.getLocalizedMachiningComment(6, "Circular Pocket"),
                **keywords),
            '',
        )

    def createVerticalPocketPiercing(self, **keywords):
        pocket = keywords['pocket']

        return ''.join((
            'G125',  # The macro G125 will only work correctly in F4Integrate 1.3, esp. due to automatic approach and departure
            self.getPocketRadiusCompensation(**keywords).replace('G', 'L'),  # L42 is not documented
            'X{x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'Y{y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'Z{depth}{complement}'.format(depth=self.formatAsCoordinate(data=keywords['depth']), complement=(self.createDepthPerStepForPockets(**keywords))),
            'R{diameter}'.format(diameter=self.formatAsCoordinate(data=pocket.installedDiameter)),  # Documented as radius, but F4Integrate expects diameter
            self.getToolSelection(tool=pocket.tool, **keywords),
            'D{clockwise:d}'.format(clockwise=self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
            'I0',
            'Q0',
            self.getMachiningComment(comment=self.getLocalizedMachiningComment(7, "Circular Pocket Piercing"), **keywords),
        )) + '\n'

    def createHorizontalPocket(self, **keywords):
        plane, positionOnPlane = self.__calculatePlane(**keywords)

        return joinIfNotNone('\n',
            self.changePlane(plane=plane, **keywords),
            self.__createCommonPocket(
                x=positionOnPlane.x,
                y=positionOnPlane.y,
                depth=-keywords['pocket'].depth,
                comment=self.getLocalizedMachiningComment(19, "Horizontal Circular Pocket"),
                **keywords),
            '',
        )

    # Contours
    APPROACH_DEPARTURE_NONE = 0
    APPROACH_DEPARTURE_EXTEND = 1
    APPROACH_DEPARTURE_ARC = 2
    APPROACH_DEPARTURE_LINEAR = 4

    ApproachMode = collections.namedtuple('ApproachMode', ('type', 'automatic', 'flying'))

    APPROACH_MODE_LINEAR_PIERCING = ApproachMode(APPROACH_DEPARTURE_LINEAR, True, False)
    APPROACH_MODE_LINEAR_FLYING = ApproachMode(APPROACH_DEPARTURE_LINEAR, False, True)  # We do not want automatic, due to side distance
    APPROACH_MODE_EXTEND = ApproachMode(APPROACH_DEPARTURE_EXTEND, True, False)
    APPROACH_MODE_ARC_PIERCING = ApproachMode(APPROACH_DEPARTURE_ARC, True, False)
    APPROACH_MODE_ARC_FLYING = ApproachMode(APPROACH_DEPARTURE_ARC, True, True)
    APPROACH_MODE_ARC_FLYING_WITHOUT_AUTOMATIC = ApproachMode(APPROACH_DEPARTURE_ARC, False, True)  # Temporary workaround for pockets as long as automatic damages the pocket inside F4Integrate
    APPROACH_MODE_WITHOUT = ApproachMode(APPROACH_DEPARTURE_NONE, None, False)

    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_WITHOUT

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        position = segment.getEndVertex().position
        return ''.join((
            ' ' * 2,
            self.ABSOLUTE_COORDINATES,
            'G1',
            'X{x}'.format(x=self.formatAsCoordinate(data=position.x)),
            'Y{y}'.format(y=self.formatAsCoordinate(data=position.y)),
        ))

    def createRadiusSegment(self, **keywords):
        segment = keywords['segment']
        CW = 'G2'
        CCW = 'G3'
        SMALL_ARC_FACTOR = 1
        LARGE_ARC_FACTOR = -1
        multiplicationFactor, operationCode = segment.getDirection((SMALL_ARC_FACTOR, CW), (SMALL_ARC_FACTOR, CCW), (LARGE_ARC_FACTOR, CW), (LARGE_ARC_FACTOR, CCW))
        position = segment.getEndVertex().position
        radius = segment.getRadius() * multiplicationFactor

        return joinIfNotNone('',
            ' ' * 2,
            self.ABSOLUTE_COORDINATES,
            operationCode,
            'X{x}'.format(x=self.formatAsCoordinate(data=position.x)),
            'Y{y}'.format(y=self.formatAsCoordinate(data=position.y)),
            'R{r}'.format(r=self.formatAsAngle(data=radius)),
        )

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']

        if machining.isVertical():
            left = self.RADIUS_COMPENSATION_LEFT
            right = self.RADIUS_COMPENSATION_RIGHT
        else:
            plane, _positionOnPlane = self.__calculatePlane(**keywords)
            rightHanded = plane.isWopMatrixRightHanded()
            left = self.RADIUS_COMPENSATION_LEFT if rightHanded else self.RADIUS_COMPENSATION_RIGHT
            right = self.RADIUS_COMPENSATION_RIGHT if rightHanded else self.RADIUS_COMPENSATION_LEFT

        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=self.RADIUS_COMPENSATION_NONE, left=left, right=right)

    def getSideDistanceAsStringForContourApproachDeparture(self, **keywords):
        if self.sizingIsContourCreator(**keywords) and keywords['approachDepartureMode'].type == self.APPROACH_DEPARTURE_LINEAR and self.CREATE_SIZING_AS_NESTING_GEOMETRY:  # APPROACH_MODE_LINEAR_PIERCING is already filtered.
            return self.formatAsCoordinate(data=1.0)
        elif not keywords['approachDepartureAutomatic'] and 'sideDistance' in keywords:
            if isinstance(keywords['sideDistance'], str):
                return keywords['sideDistance']
            else:
                return self.formatAsCoordinate(data=keywords['sideDistance'])
        else:
            return self.formatAsCoordinate(data=0.0)

    def getCommonContourApproachDepartureCode(self, **keywords):
        approachDepartureType, approachDepartureAutomatic, approachDepartureFlying = keywords['approachDepartureMode']
        distance = keywords['distanceFactor'] * keywords['machining'].tool.getRouterDiameter() / 2 if 'distanceFactor' in keywords else 0

        if approachDepartureType == self.APPROACH_DEPARTURE_NONE:
            return keywords['approachDepartureParameter'] + self.formatAsInt(data=approachDepartureType)
        elif approachDepartureType == self.APPROACH_DEPARTURE_LINEAR:
            return keywords['approachDepartureParameter'] + joinIfNotNone(';',
                self.formatAsInt(data=approachDepartureType),
                '{automatic:d}'.format(automatic=approachDepartureAutomatic),
                '{distance}'.format(distance=self.formatAsCoordinate(data=distance if not approachDepartureAutomatic else 0)),
                self.getSideDistanceAsStringForContourApproachDeparture(approachDepartureAutomatic=approachDepartureAutomatic, **keywords),
                '{z}'.format(z='({depth}+1)'.format(depth=self.formatAsCoordinate(data=keywords['machiningDepth'])) if approachDepartureFlying else self.formatAsCoordinate(data=0)),
                keywords['radiusCompensation'][2],
                '{endInCenter}'.format(endInCenter=0),
            )
        elif approachDepartureType == self.APPROACH_DEPARTURE_ARC:
            return keywords['approachDepartureParameter'] + joinIfNotNone(';',
                self.formatAsInt(data=approachDepartureType),
                '{automatic:d}'.format(automatic=approachDepartureAutomatic),
                '{distance}'.format(distance=self.formatAsCoordinate(data=distance if not approachDepartureAutomatic else 0)),
                self.getSideDistanceAsStringForContourApproachDeparture(approachDepartureAutomatic=approachDepartureAutomatic, **keywords),
                keywords['radiusCompensation'][2],
                '{z}'.format(z='({depth}+1)'.format(depth=self.formatAsCoordinate(data=keywords['machiningDepth'])) if approachDepartureFlying else self.formatAsCoordinate(data=0)),
            )
        elif approachDepartureType == self.APPROACH_DEPARTURE_EXTEND:
            return keywords['approachDepartureParameter'] + joinIfNotNone(';',
                self.formatAsInt(data=approachDepartureType),
                '{automatic:d}'.format(automatic=approachDepartureAutomatic),
                '{distance}'.format(distance=self.formatAsCoordinate(data=distance if not approachDepartureAutomatic else 0)),
            )
        else:
            assert False, "Not supported approach/departure type"

    def getCommonContourApproachCode(self, **keywords):
        return self.getCommonContourApproachDepartureCode(approachDepartureMode=keywords['approachMode'], distanceFactor=2, **keywords)

    def getCommonContourDepartureCode(self, **keywords):
        return self.getCommonContourApproachDepartureCode(approachDepartureMode=keywords['departureMode'], distanceFactor=2, **keywords)

    def sizingIsContourCreator(self, **keywords):
        return 'contourCreator' in keywords and keywords['contourCreator'] == 'Sizing'

    def createCommonContourStartBlock(self, **keywords):
        position = keywords['startVertex'].position

        return joinIfNotNone('',
            'M21' if self.CREATE_SIZING_AS_NESTING_GEOMETRY and self.sizingIsContourCreator(**keywords) else None,
            self.ABSOLUTE_COORDINATES,
            keywords['radiusCompensation'],
            'G100',
            'X{x}'.format(x=self.formatAsCoordinate(data=position.x)),
            'Y{y}'.format(y=self.formatAsCoordinate(data=position.y)),
            'Z{z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            self.getToolSelection(tool=keywords['machining'].tool, **keywords),
            'J{deactivated:d}'.format(deactivated=keywords['deactivated'] if 'deactivated' in keywords else False),
            'L"{profileName}"'.format(profileName=keywords['profileName']) if 'profileName' in keywords and keywords["profileName"] else None,
            'D{distance}'.format(distance=self.formatAsCoordinate(data=keywords['contourDistance'])) if 'contourDistance' in keywords and not isNearlyEqual(keywords['contourDistance'], 0.0) else None,
            'F{feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            self.getCommonContourApproachCode(approachDepartureParameter='Q', **keywords),
            self.getCommonContourDepartureCode(approachDepartureParameter='W', **keywords),
            self.getMachiningComment(**keywords),
        )

    def createRabbetContourStartBlock(self, **keywords):
        return self.createCommonContourStartBlock(
            radiusCompensation=self.getWopRadiusCompensation(**keywords),
            comment=self.getLocalizedMachiningComment(12, "Rabbet"),
            approachMode=self.APPROACH_MODE_EXTEND,
            departureMode=self.APPROACH_MODE_EXTEND,
            **keywords)

    def createRabbetContourEndBlock(self, **_keywords):
        return '\n'

    def createVerticalRabbet(self, **keywords):
        return self.createDefinitionForOneRabbet(**keywords)

    def createContourStartBlock(self, **keywords):
        contour = keywords['contour']

        if contour.clearOut:
            self.profileNameIndex += 1
            contour.profileName = self.getLocalizedMachiningComment(23, "Profile {index}").format(index=self.profileNameIndex)
        else:
            contour.profileName = None

        return self.createCommonContourStartBlock(
            radiusCompensation=self.getWopRadiusCompensation(**keywords),
            comment=self.getLocalizedMachiningComment(14, "Contour"),
            approachMode=self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords),
            departureMode=self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords),
            # TODO CONTOUR2.0 IT-243/IT-286 Currently feedRate is only implemented directly within createContourStartBlock, not other contour using machinings --mjk/NG
            feedRate=contour.currentFeedRate,
            profileName=contour.profileName,
            deactivated=contour.clearOut,
            **keywords)

    def createFreeFormPocket(self, **keywords):
        return joinIfNotNone('',
            self.ABSOLUTE_COORDINATES,
            'G210',
            'P"{profileName}"'.format(profileName=keywords['profileName']),
            # 'X1',  # "start from internal", but in reality it is CW/CCW that is changed, leading to errors at exit.
            'Z{z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            self.getToolSelection(tool=keywords['machining'].tool, **keywords),
            # Careful: L defines a distance, but it does not mean that not everything will be removed. The difference is just how.
            'L{distance}'.format(distance=self.formatAsCoordinate(data=keywords['contourDistance'])) if 'contourDistance' in keywords and not isNearlyEqual(keywords['contourDistance'], 0.0) else None,
            'F{feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            self.getCommonContourApproachCode(approachDepartureParameter='I', **keywords),
            self.getCommonContourDepartureCode(approachDepartureParameter='Q', **keywords),  # Yes, Q is exit in G210, even if it is entry in G100
            self.getMachiningComment(**keywords),
        )

    def createContourEndBlock(self, **keywords):
        contour = keywords['contour']

        return joinIfNotNone('\n',
            self.createNCStop(
                comment=self.getLocalizedMachiningComment(24, "Emptying a contour with a distance is currently not supported by Emptying of F4Integrate"),
                **keywords) if contour.clearOut and not isNearlyEqual(keywords['contourDistance'], 0.0) else None,
            self.createFreeFormPocket(
                radiusCompensation=self.getWopRadiusCompensation(**keywords),
                comment=self.getLocalizedMachiningComment(22, "Emptying Contour"),
                approachMode=self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords),
                departureMode=self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords),
                profileName=contour.profileName,
                **keywords) if contour.clearOut else None,
            '',
            '',
        )

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']

        # TODO (HORIZONTAL) IT-222/IT-230/IT-285: unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if not contour.isVertical():
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.wopMatrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1) or matrix.w == Vector(0, 0, -1), "Matrix must be 2D!"
            contour.path = contour.transformedPolyline(matrix.getInverted(), positionOnPlane)
            keywords['plane'] = plane

        return super().createDefinitionForOneContour(**keywords)

    def createVerticalContour(self, **keywords):
        keywords['creator'] = self.createDefinitionForOneContour
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            *self.generateContourNoBuiltInDepthSteps(**keywords)
        )

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createVerticalContour, **keywords))

    def createHorizontalContour(self, **keywords):
        keywords['creator'] = self.createDefinitionForOneContour
        keywords['zValueConverter'] = lambda _boardThickness, currentDepth: - currentDepth

        plane, _positionOnPlane = self.__calculatePlane(**keywords)

        return joinIfNotNone('\n',
            self.changePlane(plane=plane, **keywords),
            *self.generateContourNoBuiltInDepthSteps(**keywords)
        )

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createHorizontalContour, **keywords))

    def createCommonDadoWithRouter(self, **keywords):
        return self.createContourFromSegmentList(
            createStartBlock=self.createDadoContourStartBlock,
            createEndBlock=self.createDadoContourEndBlock,
            radiusCompensation=self.getDadoRadiusCompensation(**keywords),
            segmentList=self.generateDadoSegments(**keywords),
            zValue=keywords['depth'],
            **keywords)

    def createSizingDefinitionForOnePolyline(self, **keywords):
        sizing = keywords['surface'].sizing

        keywords['creator'] = super().createSizingDefinitionForOnePolyline
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
        keywords['sizing'] = sizing

        machiningDepth = sizing.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machining=sizing, machiningDepth=machiningDepth, **keywords)

        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            *self.generateCommonNoBuiltInDepthSteps(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords),
        )

    def createSizingStartBlock(self, **keywords):
        approachMode = self.getSizingApproachMode(**keywords)
        return self.createCommonContourStartBlock(
            radiusCompensation=self.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else self.RADIUS_COMPENSATION_LEFT,
            comment=self.getLocalizedMachiningComment(13, "Sizing"),
            approachMode=approachMode,
            departureMode=approachMode,
            contourCreator='Sizing',
            **keywords)

    def createSizingEndBlock(self, **_keywords):
        return '\n'

    def createSizingContour(self, **keywords):
        return self.createSizingContourDefinitionAsExecution(**keywords)

    def createDadoContourStartBlock(self, **keywords):
        dado = keywords['dado']
        if dado.isInsetBoth():
            approachDepatureMode = self.APPROACH_MODE_LINEAR_FLYING
            sideDistance = (dado.width - dado.tool.getRouterDiameter()) / 2.0
        else:
            approachDepatureMode = self.APPROACH_MODE_EXTEND
            sideDistance = 0

        return self.createCommonContourStartBlock(
            approachMode=approachDepatureMode,
            departureMode=approachDepatureMode,
            sideDistance=sideDistance,
            **keywords)

    def createDadoContourEndBlock(self, **_keywords):
        return '\n'

    def createVerticalDadoWithRouter(self, **keywords):
        def creatorHelper(**kwargs):
            kwargs['depth'] = kwargs['zValue']
            del kwargs['zValue']  # TODO IT-265/IT-285: can we do it better? -- NG
            return self.createCommonDadoWithRouter(**kwargs)

        keywords['creator'] = creatorHelper
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
        machiningDepth = keywords['dado'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            *self.generateCommonNoBuiltInDepthSteps(machiningDepth=machiningDepth, depthPerStep=depthPerStep, comment=self.getLocalizedMachiningComment(11, "Routed Dado"), **keywords)
        )

    def createHorizontalDadoWithRouter(self, **keywords):
        def creatorHelper(**kwargs):
            kwargs['depth'] = kwargs['zValue']
            del kwargs['zValue']  # TODO IT-265/IT-285: can we do it better? -- NG
            return self.createCommonDadoWithRouter(**kwargs)

        dado = keywords['dado']

        plane, positionOnPlane = self.__calculatePlane(**keywords)

        matrix = plane.wopMatrix * dado.matrix.getInverted()
        assert matrix.w == Vector(0, 0, 1) or matrix.w == Vector(0, 0, -1), "Matrix must be 2D!"
        dado.matrix = matrix  # TODO (HORIZONTAL) IT-229/IT-285: unschönes Design -- NG
        dado.matrix.origin = positionOnPlane

        keywords['creator'] = creatorHelper
        keywords['zValueConverter'] = lambda _boardThickness, currentDepth: -currentDepth
        machiningDepth = keywords['dado'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return joinIfNotNone('\n',
            self.changePlane(plane=plane, **keywords),
            *self.generateCommonNoBuiltInDepthSteps(sideID=plane.sideID, machiningDepth=machiningDepth, depthPerStep=depthPerStep, comment=self.getLocalizedMachiningComment(20, "Horizontal Routed Dado"), **keywords)
        )

    def getStartAndEndDistanceSecurityForDado(self, **keywords):
        dado = keywords['dado']
        startInset, endInset = 0, 0
        if not dado.isInsetStart():
            startInset = self.SECURITY_DISTANCE_FOR_DADO_SAW
        if not dado.isInsetEnd():
            endInset = self.SECURITY_DISTANCE_FOR_DADO_SAW
        return startInset, endInset

    def createCommonVerticalDadoWithSaw(self, **keywords):
        dado = keywords['dado']

        surface = keywords['surface']
        preSawDepth = self.zValueDadoSawConverter.convertWithThickness(surface.thickness, self.DADO_PRESAW_DEPTH)
        isPreSaw = not isNearlyEqual(self.DADO_PRESAW_DEPTH, 0.0)

        startInset, endInset = self.getStartAndEndDistanceSecurityForDado(**keywords)

        return joinIfNotNone('',
            'G111',
            'X' + keywords['xValues'],
            'Y' + keywords['yValues'],
            'Z{z}'.format(z=self.formatAsCoordinate(data=keywords['depth'])),
            self.getToolSelection(tool=dado.tool, **keywords),
            'V{width}'.format(width=self.formatAsCoordinate(data=dado.width)),  # V is documented for G110, but not for G111
            'W{preSawDepth}'.format(preSawDepth=self.formatAsCoordinate(data=preSawDepth)) if isPreSaw else None,
            'A1',
            'I{startInset}'.format(startInset=self.formatAsCoordinate(data=startInset)) if startInset and not isNearlyEqual(startInset, 0.0) else None,
            'Q{endInset}'.format(endInset=self.formatAsCoordinate(data=endInset)) if endInset and not isNearlyEqual(endInset, 0.0) else None,
            self.RADIUS_COMPENSATION_NONE.replace('G', 'D'),
            self.getMachiningComment(**keywords),
        )

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        dado = keywords['dado']
        start = dado.getStartPoint()
        end = dado.getEndPoint()

        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            self.createCommonVerticalDadoWithSaw(
                xValues='{xs};{xe}'.format(xs=self.formatAsCoordinate(data=start.x), xe=self.formatAsCoordinate(data=end.x)),
                yValues='{y}'.format(y=self.formatAsCoordinate(data=start.y)),
                comment=self.getLocalizedMachiningComment(8, "Sawn Dado X"),
                **keywords),
            ''
        )

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        dado = keywords['dado']
        start = dado.getStartPoint()
        end = dado.getEndPoint()

        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            self.createCommonVerticalDadoWithSaw(
                xValues='{x}'.format(x=self.formatAsCoordinate(data=start.x)),
                yValues='{ys};{ye}'.format(ys=self.formatAsCoordinate(data=start.y), ye=self.formatAsCoordinate(data=end.y)),
                comment=self.getLocalizedMachiningComment(9, "Sawn Dado Y"),
                **keywords),
            ''
        )

    def createVerticalDadoWithSawCRotated(self, **keywords):
        dado = keywords['dado']
        start = dado.getStartPoint()
        end = dado.getEndPoint()

        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            self.createCommonVerticalDadoWithSaw(
                xValues='{xs};{xe}'.format(xs=self.formatAsCoordinate(data=start.x), xe=self.formatAsCoordinate(data=end.x)),
                yValues='{ys};{ye}'.format(ys=self.formatAsCoordinate(data=start.y), ye=self.formatAsCoordinate(data=end.y)),
                comment=self.getLocalizedMachiningComment(10, "Sawn Dado"),
                **keywords),
            ''
        )

    # Miter
    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCCW)

    def createVerticalMiter(self, **keywords):
        miter = keywords['miter']
        start = miter.getStartPoint()
        end = miter.getEndPoint()

        surface = keywords['surface']
        boardThickness = surface.thickness

        preSawDepth = self.zValueMiterPreSawConverter.convertWithThickness(boardThickness, self.MITER_PRE_SAW_DEPTH_FROM_TOP)
        depth = self.zValueMiterSawConverter.convertWithThickness(boardThickness, boardThickness + self.ADDITIONAL_DEPTH_FOR_MITER)
        isPreSaw = not isNearlyEqual(self.MITER_PRE_SAW_DEPTH_FROM_TOP, 0.0)

        return joinIfNotNone('\n',
            self.changePlane(plane=None, **keywords),
            joinIfNotNone('',
                'G111',
                self.getToolSelection(tool=miter.tool, **keywords),
                'X{xs};{xe}'.format(xs=self.formatAsCoordinate(data=start.x), xe=self.formatAsCoordinate(data=end.x)),
                'Y{ys};{ye}'.format(ys=self.formatAsCoordinate(data=start.y), ye=self.formatAsCoordinate(data=end.y)),
                'Z{z};1'.format(z=self.formatAsCoordinate(data=depth)),  # ;1 is for "move starting point"
                'K(-T)',
                'B{miterAngle}'.format(miterAngle=self.formatAsAngle(data=normalizeAngle(90.0 - miter.getAngle()))),
                'J0',  # Depth along Z axis, not along cutting
                'W(T{preSawDepth})'.format(preSawDepth=self.formatAsSignedCoordinate(data=preSawDepth)) if isPreSaw else None,
                'A0',
                self.RADIUS_COMPENSATION_RIGHT.replace('G', 'D'),
                self.getMachiningComment(comment=self.getLocalizedMachiningComment(15, "Miter"), **keywords),
            ),
            '',
        )

    # NC Macro
    def getNCMacroCookedVariables(self, **keywords):
        variables = super().getNCMacroCookedVariables(**keywords)

        ncmacro = keywords['ncmacro']
        if ncmacro.isVertical():
            sideID = self.SIDE_MAIN
            sideX = ncmacro.position.x
            sideY = ncmacro.position.y
            sideCAngle = variables['cangle']
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            sideID = plane.sideID
            sideX = positionOnPlane.x
            sideY = positionOnPlane.y

            angle = VectorMath.angleBetween(ncmacro.matrix.u, plane.wopMatrix.u)

            if not plane.isWopMatrixRightHanded():
                angle = 360.0 - angle

            if not isNearlyEqual(angle, 0.0):
                crossProduct = ncmacro.matrix.u.cross(plane.wopMatrix.u)
                if not isNearlyEqual(crossProduct.length(), 0.0) and crossProduct.normalized() == plane.wopMatrix.w:
                    angle = - angle

            sideCAngle = self.formatter.formatAsAngle(normalizeAngle(angle))

        variables['sideid'] = sideID
        variables['sidex'] = self.formatter.formatAsCoordinate(sideX)
        variables['sidey'] = self.formatter.formatAsCoordinate(sideY)
        variables['sidecangle'] = sideCAngle

        return variables

    def getPlaneForNCMacro(self, **keywords):
        ncmacro = keywords['ncmacro']

        if ncmacro.isVertical() or ncmacro.code.partition('\n')[0] == self.SIDE_MAIN:
            return None
        else:
            plane, _positionOnPlane = self.__calculatePlane(**keywords)
            return plane

    def createNCMacroCode(self, **keywords):
        plane = self.getPlaneForNCMacro(**keywords)

        return joinIfNotNone('\n',
            self.changePlane(plane=plane, **keywords),
            super().createNCMacroCode(**keywords) if not keywords['ncmacro'].code.startswith(self.SIDE_MAIN) else super().createNCMacroCode(**keywords).partition('\n')[2]
        )

    # Footer
    def createEndOfFile(self, **_keywords):
        return ''

    def getLocalizedMachiningComment(self, index, default):
        return getLocalizedString(F4INTEGRATE_COMMENT_MESSAGES, index, default)
