# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections
import logging

import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
import interiorcad.backend.zvalue.constants as ZValueConstants
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual, isNearlyGreater
from interiorcad.interface.geometry.simplelinesegment import SimpleLineSegment
from interiorcad.interface.ncexport.approachcreator import createArcApproach, createArcDeparture, createLinearApproach, createLinearDeparture
from interiorcad.ncexport.base import NCExportBase
from interiorcad.interface.ncexport.multisawdado import MultiSawDado
from interiorcad.backend.ncexportmigrated.dadotoncsegments import getDadoContourCW, getDadoContourCCW, getDadoRoundedContourCW, getDadoRoundedContourCCW
from interiorcad.backend.ncexportmigrated.rabbettoncsegments import RabbetToNCSegments
from interiorcad.interface.machiningsexporter.horizontaldrillingsexporter import HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.backend.localization import getLocalizedString, GANNOMAT_MESSAGES_RESOURCEGROUP
from interiorcad.interface.ncexport.gannomatexporter import GannoMATExporter
from interiorcad.interface.surface.machiningdirection import MachiningDirection
from interiorcad.interface.surface.contour2d import Contour2D


# GannoMAT considers x as width and y as length, unlike interiorcad (X as length, Y as width).
# Therefore, the "layers" (left, front, rear, right) are different from expected by interiorcad's point of view


class GannoMATExport(NCExportBase):
    EXPORT_FORMAT = 'GannoMAT'

    OUTPUT_ENCODING = 'utf-8'

    def __init__(self, **keywords):
        super().__init__(**keywords)

        machineConfiguration = self.getMachineConfiguration()
        self.FILE_NAME_POSTFIX_BOTH_SIDED = ''
        self.VERSION = machineConfiguration.getAsFloat('gannomat-ascii-version', 1.3)

        self.DADO_PRESAW_DEPTH = machineConfiguration.getAsFloat('dado-presaw-depth', 0)
        self.DADO_SAW_OVERLAP_FACTOR = self.getDadoSawOverlapFactor(**keywords)
        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('contour-approach-departure-distance', 30)

        self.ALLOW_SIZING_WITH_CONTOUR = False  # Sizing with contour is functionally not possible at the moment, because the workpiece does not have any offsets. This key could be deleted if sizing with contour is possible.  --mjk
        self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR = True  # If sizing with contour is possible, this would be the only option for interiorcad, because the workpiece can not be rotated while running the nc program.
        self.CREATE_SIZING_WITH_MACRO_FUNCTION = machineConfiguration.get('create-sizing-with-macro-function', '3')  # 2 = two sided, 3 = three sided, 4 = four sided

        self.CLEAR_OUT_POCKET_OVERLAP_FACTOR = machineConfiguration.getAsInt('clear-out-pocket-overlap-factor', 70)  # "70% overlapping is recommended."
        self.USE_CUSTOM_USER_WORKPIECE_COMMENT = machineConfiguration.getAsBool('use-custom-user-workpiece-comment', False)

        self.createZValueConverters(**keywords)
        self.checkOnlyOneRouter(**keywords)

        self.lastRouterToolID = None

    def getDadoSawOverlapFactor(self, **_keywords):
        factor = self.getMachineConfiguration().getAsFloat('dado-saw-overlap-factor', 0.30)  # MultiSawDado needs the mathematical factor between 0 and 1 for version <= 1.2.
        if self.VERSION <= 1.2:
            return factor
        elif factor <= 1:  # assuming mathematical factor of version <= 1.2 with value smaller than 1, but needs prozentual value between 1 and 100 for version 1.3.
            return factor * 100
        else:  # assuming actual version 1.3 with prozentual value between 1 and 100.
            return factor

    # actually ZValueConverters are not integrated in machine-configuration
    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)

    def getNCFileExtension(self, **_keywords):
        return '.ascii'

    def checkThatFilePostFixesAreDifferents(self):
        pass  # irrelevant

    def checkOnlyOneRouter(self, **_keywords):
        if self.VERSION < 1.2 and len(tuple(self.getMachineConfiguration().toolDatabase.getToolsOfType(Tool.ROUTER))) > 1:
            logging.warning(getLocalizedString(GANNOMAT_MESSAGES_RESOURCEGROUP, 1, "GannoMAT ASCII Postprocessor 1.1 supports only one router!"))

    def getSurfacesExporter(self, **keywords):
        # The concept of GannoMAT does not work well with the default GeneralExporter, so we do not allow to use it, not even optionally
        return GannoMATExporter(**keywords)

    def getWorkpieceComments(self, **keywords):
        workpieceComments = {}
        for key, value in super().getWorkpieceComments(**keywords).items():
            if isinstance(value, str) and ';' in value:
                workpieceComments[key] = value.replace(';', '_')
            else:
                workpieceComments[key] = value

        return workpieceComments

    # Header
    def createHead(self, **keywords):
        self.lastRouterToolID = None
        return joinIfNotNone('\n',
            self.createVersionDefinition(**keywords),
            self.createWorkpieceDefinition(**keywords),
            self.createWorkpieceName(**keywords),
            self.createWorkpieceComment(**keywords),
            '',
        )

    def createVersionDefinition(self, **_keywords):
        return 'V; {version:05.2f}'.format(version=self.VERSION)

    def createWorkpieceDefinition(self, **keywords):
        surface = keywords['surface']
        return '; '.join((
            'P',
            '{x}'.format(x=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            '{y}'.format(y=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            '{z}'.format(z=self.formatAsCoordinate(data=surface.thickness)),
            'MM'
        ))

    def createWorkpieceName(self, **keywords):
        surface = keywords['surface']
        extendedWorkpieceName = self.createFileName(name=surface.workpieceName, extension='', **keywords)

        return '; '.join((
            'N',
            '{workpieceName}'.format(workpieceName=extendedWorkpieceName),
        ))

    def customUserWorkpieceCommentFileDefaultText(self, **_keywords):
        return '\n'.join((
            '; The complete comment "C" must be in a single row. In the GannoMAT Editor, multiple lines of text can be specified by separating by "|".',
            '; Lines starting with ";" are ignored by the GannoMAT ASCII Postprocessor. This note could be deleted after changing the file.',
            '; '.join((
                'C',
                '|'.join((
                    '{WorkpieceLabel}: {WorkpieceName}',
                    '{CommentLabel}: {Comment}',
                    '{ItemIDLabel}: {ItemID}',
                    '{Edge1Label}: {Edge1}',
                    '{Edge2Label}: {Edge2}',
                    '{Edge3Label}: {Edge3}',
                    '{Edge4Label}: {Edge4}',
                    '{CabinetNameLabel}: {CabinetName}',
                    '{ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}',
                    '{PositionNumberLabel}: {PositionNumber} {PositionDescription}',
                    '{CountLabel}: {Count}'
                ))
            ))
        ))

    def createCustomUserWorkpieceComment(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='GannoMAT Workpiece Description.txt',
            customUserVariableFileDefaultText=self.customUserWorkpieceCommentFileDefaultText(**keywords),
            **keywords)

    def createWorkpieceComment(self, **keywords):
        if self.VERSION >= 1.2:
            customUserWorkpieceComment = self.createCustomUserWorkpieceComment(**keywords) if self.USE_CUSTOM_USER_WORKPIECE_COMMENT else None

            if customUserWorkpieceComment:
                return customUserWorkpieceComment

            else:
                projectInfo = self.getProjectInfo()
                workpieceComments = self.getWorkpieceComments(**keywords)
                isProjectIDValid = ExportHelper.isProjectIDValid(self, **projectInfo)

                return '; '.join((
                    'C',
                    joinIfNotNone('|',
                        '{WorkpieceLabel}: {WorkpieceName}'.format_map(workpieceComments),
                        '{CommentLabel}: {Comment}'.format_map(workpieceComments) if workpieceComments['Comment'] else None,
                        '{ItemIDLabel}: {ItemID}'.format_map(workpieceComments),
                        '{Edge1Label}: {Edge1}'.format_map(workpieceComments) if workpieceComments['Edge1'] else None,  # front
                        '{Edge2Label}: {Edge2}'.format_map(workpieceComments) if workpieceComments['Edge2'] else None,  # back
                        '{Edge3Label}: {Edge3}'.format_map(workpieceComments) if workpieceComments['Edge3'] else None,  # left
                        '{Edge4Label}: {Edge4}'.format_map(workpieceComments) if workpieceComments['Edge4'] else None,  # right
                        '{CabinetNameLabel}: {CabinetName}'.format_map(workpieceComments) if workpieceComments['CabinetName'] else None,
                        '{ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}'.format_map(workpieceComments) if isProjectIDValid else None,
                        '{PositionNumberLabel}: {PositionNumber} {PositionDescription}'.format_map(workpieceComments) if isProjectIDValid else None,
                        '{CountLabel}: {Count}'.format_map(workpieceComments),
                    ),
                ))

    # Machinings
    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    REFERENCE_TOP_LEFT = 1
    REFERENCE_BOTTOM_LEFT = 4

    # In GannoMat the main side is bottom, not top
    LAYER_TOP = 'TOP'
    LAYER_BOTTOM = 'BOT'
    LAYER_LEFT = 'LEF'
    LAYER_RIGHT = 'RIG'
    LAYER_FRONT = 'FRO'
    LAYER_REAR = 'REA'

    def getLayer(self, **keywords):
        direction = keywords['machining'].matrix.w
        machineDirection = MachiningDirection()
        if machineDirection.isVerticalFromTop(direction):
            return self.LAYER_BOTTOM
        elif machineDirection.isVerticalFromBottom(direction):
            return self.LAYER_TOP
        elif machineDirection.isFromLeft(direction):
            return self.LAYER_FRONT
        elif machineDirection.isFromRight(direction):
            return self.LAYER_REAR
        elif machineDirection.isFromFront(direction):
            return self.LAYER_LEFT
        elif machineDirection.isFromBack(direction):
            return self.LAYER_RIGHT
        else:
            return ''

    def createRouterToolChange(self, **keywords):
        # Only supported from Gannomat ASCII Import version 1.2 on
        if self.VERSION >= 1.2:
            toolID = keywords['toolID']
            if toolID != self.lastRouterToolID:
                self.lastRouterToolID = toolID
                return '; '.join((
                    'T',
                    '{id}'.format(id=toolID),
                ))

    # Drillings
    def createCommonDrilling(self, **keywords):
        drilling = keywords['drilling']
        return '; '.join((
            'H',
            '{layer}'.format(layer=keywords['layer']),
            '{reference}'.format(reference=self.REFERENCE_BOTTOM_LEFT),
            '{x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            '{y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            '{z}'.format(z=self.formatAsCoordinate(data=keywords['z'])),
            '{diameter}'.format(diameter=self.formatAsCoordinate(data=drilling.installedDiameter)),
            '0',  # "entspänen" from Version 1.2 on
            '0',
        )) + '\n'

    def createVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']
        return self.createCommonDrilling(
            layer=self.getLayer(**keywords),
            x=drilling.position.y,
            y=drilling.position.x,
            z=keywords['depth'],
            **keywords
        )

    def createVerticalDrillingPiercing(self, **keywords):
        return self.createVerticalDrilling(**keywords)

    def createVerticalGridDrillingAlongAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        deltaVector = gridDrilling.getDeltaVector()

        startDrilling = keywords['gridDrilling'].getStartDrilling()

        numberOfRows = gridDrilling.countDrillings if gridDrilling.isAlongYAxis() else 1
        numberOfColumns = gridDrilling.countDrillings if gridDrilling.isAlongXAxis() else 1
        rowsDistance = deltaVector.y
        columnDistance = deltaVector.x

        return '; '.join((
            'R',
            '{layer}'.format(layer=self.getLayer(**keywords)),
            '{reference}'.format(reference=self.REFERENCE_BOTTOM_LEFT),
            '1',
            '{x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.y)),
            '{y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.x)),
            '{z}'.format(z=self.formatAsCoordinate(data=keywords['depth'])),
            '{diameter}'.format(diameter=self.formatAsCoordinate(data=startDrilling.installedDiameter)),
            '{a1}'.format(a1=self.formatAsCoordinate(data=rowsDistance)),
            '{c}'.format(c=self.formatAsInt(data=numberOfRows)),
            '{a2}'.format(a2=self.formatAsCoordinate(data=columnDistance)),
            '{r}'.format(r=self.formatAsInt(data=numberOfColumns)),
            # "entspänen" from Version 1.2 on
        )) + '\n'

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        return self.createVerticalGridDrillingAlongAxis(**keywords)

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        return self.createVerticalGridDrillingAlongAxis(**keywords)

    def areVerticalGridDrillingsAlongXAxisSupported(self, **keywords):
        return self.VERSION >= 1.2

    def areVerticalGridDrillingsAlongYAxisSupported(self, **keywords):
        return self.VERSION >= 1.2

    def getHorizontalMachiningsExporter(self, **keywords):
        return HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds(self)

    def createHorizontalDrillingFromFront(self, **keywords):
        drilling = keywords['drilling']
        return self.createCommonDrilling(
            layer=self.getLayer(**keywords),
            x=drilling.depth,
            y=drilling.position.x,
            z=keywords['zPosition'],
            **keywords
        )

    def createHorizontalDrillingFromRight(self, **keywords):
        drilling = keywords['drilling']
        return self.createCommonDrilling(
            layer=self.getLayer(**keywords),
            x=drilling.position.y,
            y=drilling.depth,
            z=keywords['zPosition'],
            **keywords
        )

    def createHorizontalDrillingFromBack(self, **keywords):
        drilling = keywords['drilling']
        return self.createCommonDrilling(
            layer=self.getLayer(**keywords),
            x=drilling.depth,
            y=drilling.position.x,
            z=keywords['zPosition'],
            **keywords
        )

    def createHorizontalDrillingFromLeft(self, **keywords):
        drilling = keywords['drilling']
        return self.createCommonDrilling(
            layer=self.getLayer(**keywords),
            x=drilling.position.y,
            y=drilling.depth,
            z=keywords['zPosition'],
            **keywords
        )

    def createRotatedHorizontalDrilling(self, **_keywords):
        return ''

    def createFiveAxisDrilling(self, **_keywords):
        return ''

    # Pockets
    MILLING_CIRCULAR_POCKET = 2
    MILLING_LINE = 3
    MILLING_ARC_SEGMENT_P2P = 6

    def createVerticalPocket(self, **keywords):
        pocket = keywords['pocket']

        machiningDepth = pocket.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)
        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords)

        return joinIfNotNone('\n',
            self.createRouterToolChange(toolID=pocket.toolID, **keywords),
            '; '.join((
                'M',
                self.getLayer(**keywords),
                str(self.REFERENCE_BOTTOM_LEFT * 100 + self.MILLING_CIRCULAR_POCKET),
                '{x}'.format(x=self.formatAsCoordinate(data=pocket.position.y)),
                '{y}'.format(y=self.formatAsCoordinate(data=pocket.position.x)),
                '{z}'.format(z=self.formatAsCoordinate(data=keywords['depth'])),
                '{diameter}'.format(diameter=self.formatAsCoordinate(data=pocket.diameter)),
                '{count}'.format(count=self.formatAsInt(data=numberOfSteps)),
                '0',  # All piercing pockets have to be chipped completely
                '{overlap}'.format(overlap=self.CLEAR_OUT_POCKET_OVERLAP_FACTOR),
                '{direction}'.format(direction=self.MILLING_DIRECTION_SAME if self.POCKET_DIRECTION.endswith('-same') else self.MILLING_DIRECTION_OPPOSITE),
            )),
            '',
        )

    def createVerticalPocketPiercing(self, **keywords):
        return self.createVerticalPocket(**keywords)

    # Contours
    APPROACH_DEPARTURE_NONE = ''
    APPROACH_DEPARTURE_LINEAR = 'linear'
    APPROACH_DEPARTURE_ARC = 'arc'

    ApproachMode = collections.namedtuple('ApproachMode', ('movement', 'flying'))

    APPROACH_MODE_LINEAR_PIERCING = ApproachMode(APPROACH_DEPARTURE_LINEAR, False)
    APPROACH_MODE_LINEAR_FLYING = ApproachMode(APPROACH_DEPARTURE_LINEAR, True)
    APPROACH_MODE_ARC_PIERCING = ApproachMode(APPROACH_DEPARTURE_ARC, False)
    APPROACH_MODE_ARC_FLYING = ApproachMode(APPROACH_DEPARTURE_ARC, True)
    APPROACH_MODE_WITHOUT = ApproachMode(APPROACH_DEPARTURE_NONE, False)

    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_LINEAR_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_WITHOUT

    APPROACH_TYPE = 'approach'
    DEPARTURE_TYPE = 'departure'

    APPROACH_DEPARTURE_FLYING_Z_VALUE = 0  # Starting with negative values 'above the part' doesn't work with ASCII Postprocessor 1.4 actually.

    MILLING_DIRECTION_SAME = 'GL'
    MILLING_DIRECTION_OPPOSITE = 'GG'

    def getWopRadiusCompensation(self, **keywords):
        return keywords['machining'].radiusCompensation

    def getSizingMillingDirectionForMacro(self, **keywords):
        if self.SIZING_MILLING_DIRECTION == 'opposite':
            return self.MILLING_DIRECTION_OPPOSITE
        elif self.SIZING_MILLING_DIRECTION == 'same':
            return self.MILLING_DIRECTION_SAME
        elif self.SIZING_MILLING_DIRECTION == '1':
            right = not self.isSizingMillingDirectionLeft(**keywords)
            return self.MILLING_DIRECTION_OPPOSITE if right else self.MILLING_DIRECTION_SAME
        elif self.SIZING_MILLING_DIRECTION == '0':
            left = self.isSizingMillingDirectionLeft(**keywords)
            return self.MILLING_DIRECTION_OPPOSITE if left else self.MILLING_DIRECTION_SAME
        else:
            assert False, "Unknown SIZING_MILLING_DIRECTION"

    # Macro (F) is supported from on ASCII-Post version 1.2.
    def createSizingWithMacro(self, **keywords):
        surface = keywords['surface']
        sizing = surface.sizing

        machiningDepth = sizing.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machining=sizing, machiningDepth=machiningDepth, **keywords)
        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords)

        return joinIfNotNone('\n',
            self.createRouterToolChange(toolID=sizing.toolID, **keywords),
            '; '.join((
                'F',
                '{x}'.format(x=self.formatAsCoordinate(data=surface.getFinishedSizeY() + (2 * surface.getAdditionY()))),
                '{y}'.format(y=self.formatAsCoordinate(data=surface.getFinishedSizeX() + (2 * surface.getAdditionX()))),
                self.CREATE_SIZING_WITH_MACRO_FUNCTION,  # 2 = two sided, 3 = three sided, 4 = four sided
                self.getSizingMillingDirectionForMacro(**keywords),
                '95',
                '{numberOfSteps}'.format(numberOfSteps=numberOfSteps),
            )),
            '',
        )

    def createSizingContour(self, **keywords):
        if keywords['surface'].hasSizing():
            if self.VERSION < 1.2:
                logging.warning(getLocalizedString(GANNOMAT_MESSAGES_RESOURCEGROUP, 2, "GannoMAT ASCII Postprocessor 1.1 supports no sizing!"))
            else:
                finishedSizeShape = self.getFinishedSizeShape(**keywords)
                if finishedSizeShape.isRectangular():
                    return self.createSizingWithMacro(**keywords)
                else:
                    if self.ALLOW_SIZING_WITH_CONTOUR:
                        warning = getLocalizedString(GANNOMAT_MESSAGES_RESOURCEGROUP, 4, "The sizing of the polygonal part {workpiece} with contour has no offsets in GannoMAT Editor. You must change the nc-program manually!")
                        logging.warning(warning.format(workpiece=keywords['surface'].workpieceName))

                        radiusCompensation = self.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(sizing=keywords['surface'].sizing, **keywords) else self.RADIUS_COMPENSATION_LEFT

                        # As GannoMAT has the main side on BOT (bottom), we need to adapt direction and eventually the radius compensation
                        radiusCompensation = radiusCompensation if MachiningDirection().isVerticalFromBottom(keywords['surface'].sizing.matrix.w) else Contour2D.getInvertedRadiusCompensation(radiusCompensation)
                        return self.createSizingContourDefinitionAsExecution(radiusCompensation=radiusCompensation, **keywords)
                    else:
                        warning = getLocalizedString(GANNOMAT_MESSAGES_RESOURCEGROUP, 3, "GannoMAT Editor does not support sizing of polygonal parts with contour. Workpiece {workpiece} is exported without sizing!")
                        logging.warning(warning.format(workpiece=keywords['surface'].workpieceName))

    def createVerticalContour(self, **keywords):
        contour = keywords['contour']

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        # As GannoMAT has the main side on BOT (bottom), we need to adapt direction and eventually the radius compensation
        radiusCompensation = radiusCompensation if MachiningDirection().isVerticalFromBottom(contour.matrix.w) else Contour2D.getInvertedRadiusCompensation(radiusCompensation)

        depthPerStep = contour.currentDepthPerStep
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=contour.actualDepth, **keywords)

        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=contour.actualDepth, depthPerStep=depthPerStep, **keywords)

        return self.createDefinitionForOneContour(
            radiusCompensation=radiusCompensation,
            numberOfSteps=numberOfSteps,
            zValue=self.getContourZValue(depth=keywords['contour'].depth, **keywords),
            **keywords)

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createVerticalContour, **keywords))

    def createNCStop(self, **_keywords):
        return ''

    SAW_DIRECTION_GANNOMAT_X = 1
    SAW_DIRECTION_GANNOMAT_Y = 2

    def __createVerticalDadoWithSawAlongAxisVersionDependingAdditionalParameters(self, **keywords):
        dado = keywords['dado']

        return '; '.join((
            '{width}'.format(width=self.formatAsCoordinate(data=dado.getWidth())),
            '0',
            '{presawdepth}'.format(presawdepth=self.formatAsCoordinate(data=self.DADO_PRESAW_DEPTH)),
            '1',
            '{overlap}'.format(overlap=self.formatAsCoordinate(data=self.DADO_SAW_OVERLAP_FACTOR)),
            '0',
            '0',
            '{tool}'.format(tool=dado.toolID),
        ))

    def __createVerticalDadoWithSawAlongAxis(self, **keywords):
        line = keywords['segment']
        realStart = min(line.start, line.end, key=keywords['key'])

        return joinIfNotNone('; ',
            'S',
            '{layer}'.format(layer=keywords['layer']),
            str(self.REFERENCE_BOTTOM_LEFT),
            '{x}'.format(x=self.formatAsCoordinate(data=realStart.y)),
            '{y}'.format(y=self.formatAsCoordinate(data=realStart.x)),
            '{z}'.format(z=self.formatAsCoordinate(data=keywords['depth'])),
            '{length}'.format(length=self.formatAsCoordinate(data=line.length)),
            '{direction}'.format(direction=keywords['direction']),
            '1',
            self.__createVerticalDadoWithSawAlongAxisVersionDependingAdditionalParameters(**keywords) if self.VERSION >= 1.3 else None,
        )

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        if self.VERSION >= 1.3:
            return self.__createVerticalDadoWithSawAlongAxis(
                segment=self.getDadoLineSegmentWithSecurityDistance(**keywords),
                key=lambda point: point.x,
                layer=self.getLayer(**keywords),
                direction=self.SAW_DIRECTION_GANNOMAT_Y,
                **keywords
            ) + '\n'

        else:
            multiSawDado = MultiSawDado(self)
            return multiSawDado.createMultipleVerticalCenteredSawCuts(
                createFunction=self.__createVerticalDadoWithSawAlongAxis,
                key=lambda point: point.x,
                layer=self.getLayer(**keywords),
                direction=self.SAW_DIRECTION_GANNOMAT_Y,
                **keywords
            ) + '\n'

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        if self.VERSION >= 1.3:
            return self.__createVerticalDadoWithSawAlongAxis(
                segment=self.getDadoLineSegmentWithSecurityDistance(**keywords),
                key=lambda point: point.y,
                layer=self.getLayer(**keywords),
                direction=self.SAW_DIRECTION_GANNOMAT_X,
                **keywords
            ) + '\n'

        else:
            multiSawDado = MultiSawDado(self)
            return multiSawDado.createMultipleVerticalCenteredSawCuts(
                createFunction=self.__createVerticalDadoWithSawAlongAxis,
                key=lambda point: point.y,
                layer=self.getLayer(**keywords),
                direction=self.SAW_DIRECTION_GANNOMAT_X,
                **keywords
            ) + '\n'

    def createVerticalDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        routerRadius = dado.tool.getRouterDiameter() / 2

        cw = self.isDadoRabbetReplacementCW(dado, self.DADO_REPLACEMENT)

        # No security distance is used here, as GannoMAT will set it later
        if self.DADO_REPLACEMENT.startswith('contour-') and cw:
            segmentList = getDadoContourCW(dado, securityDistance=0, routerRadius=routerRadius)
            radiusCompensation = self.RADIUS_COMPENSATION_RIGHT
        elif self.DADO_REPLACEMENT.startswith('contour-') and not cw:
            segmentList = getDadoContourCCW(dado, securityDistance=0, routerRadius=routerRadius)
            radiusCompensation = self.RADIUS_COMPENSATION_LEFT
        elif self.DADO_REPLACEMENT.startswith('rounded-contour-') and cw:
            segmentList = getDadoRoundedContourCW(dado, securityDistance=0, routerRadius=routerRadius)
            radiusCompensation = self.RADIUS_COMPENSATION_RIGHT
        else:  # self.DADO_REPLACEMENT.startswith('rounded-contour-') and not cw
            segmentList = getDadoRoundedContourCCW(dado, securityDistance=0, routerRadius=routerRadius)
            radiusCompensation = self.RADIUS_COMPENSATION_LEFT

        machiningDepth = dado.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)
        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords)

        # FIXME: missing  if MachiningDirection().isVerticalFromBottom(keywords['machining'].matrix.w)  --  NG (IT-210)
        radiusCompensation = Contour2D.getInvertedRadiusCompensation(radiusCompensation)

        return self.createContourFromSegmentList(
            createStartBlock=self.createDadoContourStartBlock,
            createEndBlock=self.createDadoContourEndBlock,
            radiusCompensation=radiusCompensation,
            segmentList=segmentList,
            zValue=keywords['depth'],
            toolID=dado.toolID,
            numberOfSteps=numberOfSteps,
            **keywords)

    def createVerticalRabbet(self, **keywords):
        machiningDepth = keywords['rabbet'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)
        numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords)

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        # FIXME: missing  if MachiningDirection().isVerticalFromBottom(keywords['machining'].matrix.w)  --  NG (IT-210)
        # As GannoMAT has the main side on BOT (bottom), we need to adapt direction and eventually the radius compensation
        radiusCompensation = Contour2D.getInvertedRadiusCompensation(radiusCompensation)

        return self.createVerticalRabbetOffsetByContourDistance(numberOfSteps=numberOfSteps, radiusCompensation=radiusCompensation, **keywords)

    def generateRabbetSegments(self, **keywords):
        rabbet = keywords['rabbet']
        securityDistance = self.getDadoRabbetSecurityDistance(**keywords)
        if self.RABBET_REPLACEMENT.startswith('rounded-'):
            routerRadius = rabbet.tool.getRouterDiameter() / 2
            return tuple(RabbetToNCSegments.generateRabbetRoundedSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=securityDistance, straightSecurityDistance=securityDistance, routerRadius=routerRadius))
        else:
            return tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=securityDistance, straightSecurityDistance=securityDistance))

    def __createCommonContourStartBlock(self, **keywords):
        return joinIfNotNone('\n',
            self.createRouterToolChange(**keywords),
            '; '.join((
                '{',
                self.getLayer(**keywords),
                '0',
                'auto',
                'c',  # only used from version 1.2 on
            )),
        )

    def __createCommonContourEndBlock(self, **_keywords):
        return '}\n'

    def createCommonStartBlockWithArcApproach(self, **keywords):
        isLeft = keywords['radiusCompensation'] == (self.RADIUS_COMPENSATION_LEFT if MachiningDirection().isVerticalFromBottom(keywords['machining'].matrix.w) else self.RADIUS_COMPENSATION_RIGHT)

        isClockwise, _approachDelta, approachPoint, centerPoint, _tangentialAngle = createArcApproach(left=isLeft, **keywords)

        return '\n'.join((
            self.__createCommonContourStartBlock(**keywords),
            self.createSimpleRadiusSegment(
                startPosition=approachPoint,
                endPosition=keywords['startVertex'].position,
                centerPosition=centerPoint,
                rotationArcDirection='CCW' if isClockwise else 'CW',  # As GannoMAT has the main side on BOT (bottom), we need to adapt direction and eventually the radius compensation
                **keywords),
        ))

    def createCommonEndBlockWithArcDeparture(self, **keywords):
        isLeft = keywords['radiusCompensation'] == (self.RADIUS_COMPENSATION_LEFT if MachiningDirection().isVerticalFromBottom(keywords['machining'].matrix.w) else self.RADIUS_COMPENSATION_RIGHT)

        isClockwise, _departureDelta, departurePoint, centerPoint = createArcDeparture(left=isLeft, **keywords)

        return '\n'.join((
            self.createSimpleRadiusSegment(
                startPosition=keywords['segmentList'][-1].getEndVertex().position,
                endPosition=departurePoint,
                centerPosition=centerPoint,
                rotationArcDirection='CCW' if isClockwise else 'CW',  # As GannoMAT has the main side on BOT (bottom), we need to adapt direction and eventually the radius compensation
                **keywords),
            self.__createCommonContourEndBlock(**keywords),
        ))

    def createCommonStartBlockWithLinearApproach(self, **keywords):
        _approachDelta, approachPoint = createLinearApproach(ncexport=self, **keywords)

        return '\n'.join((
            self.__createCommonContourStartBlock(**keywords),
            self.createSimpleLineSegment(startPosition=approachPoint, endPosition=keywords['startVertex'].position, **keywords),
        ))

    def createCommonEndBlockWithLinearDeparture(self, **keywords):
        _departureDelta, departurePoint = createLinearDeparture(ncexport=self, **keywords)

        return '\n'.join((
            self.createSimpleLineSegment(startPosition=keywords['segmentList'][-1].getEndVertex().position, endPosition=departurePoint, **keywords),
            self.__createCommonContourEndBlock(**keywords),
        ))

    def createCommonStartBlockWithPotentialApproach(self, **keywords):
        approachDepartureMovement, _approachDepartureFlying = keywords['approachDepartureMode']

        if 'arc' in approachDepartureMovement:
            return self.createCommonStartBlockWithArcApproach(**keywords)
        elif 'linear' in approachDepartureMovement:
            return self.createCommonStartBlockWithLinearApproach(**keywords)
        else:
            return self.__createCommonContourStartBlock(**keywords)

    def createCommondEndBlockWithPotentialDeparture(self, **keywords):
        approachDepartureMovement, _approachDepartureFlying = keywords['approachDepartureMode']

        if 'arc' in approachDepartureMovement:
            return self.createCommonEndBlockWithArcDeparture(**keywords)
        elif 'linear' in approachDepartureMovement:
            return self.createCommonEndBlockWithLinearDeparture(**keywords)
        else:
            return self.__createCommonContourEndBlock(**keywords)

    def createSizingStartBlock(self, **keywords):
        return self.__createCommonContourStartBlock(**keywords)

    def createSizingEndBlock(self, **_keywords):
        return self.__createCommonContourEndBlock()

    def getApproachDepartureModeAndSideDistanceForDado(self, **keywords):
        dado = keywords['dado']

        if dado.isInsetBoth():
            approachDepartureMode = self.APPROACH_MODE_LINEAR_FLYING
            sideDistance = (dado.width - dado.tool.getRouterDiameter()) / 2.0
        else:
            approachDepartureMode = self.APPROACH_MODE_LINEAR_PIERCING
            sideDistance = 0

        return approachDepartureMode, sideDistance

    def createDadoContourStartBlock(self, **keywords):
        approachDepartureMode, sideDistance = self.getApproachDepartureModeAndSideDistanceForDado(**keywords)

        return self.createCommonStartBlockWithPotentialApproach(
            approachDepartureType=self.APPROACH_TYPE,
            approachDepartureMode=approachDepartureMode,
            securityDistance=self.getDadoRabbetSecurityDistance(**keywords),
            sideDistance=sideDistance,
            **keywords
        )

    def createDadoContourEndBlock(self, **keywords):
        approachDepartureMode, sideDistance = self.getApproachDepartureModeAndSideDistanceForDado(**keywords)

        return self.createCommondEndBlockWithPotentialDeparture(
            approachDepartureType=self.DEPARTURE_TYPE,
            approachDepartureMode=approachDepartureMode,
            securityDistance=self.getDadoRabbetSecurityDistance(**keywords),
            sideDistance=sideDistance,
            **keywords
        )

    def createContourStartBlock(self, **keywords):
        contour = keywords['contour']

        return self.createCommonStartBlockWithPotentialApproach(
            approachDepartureType=self.APPROACH_TYPE,
            approachDepartureMode=self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords),
            securityDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
            **keywords
        )

    def createContourEndBlock(self, **keywords):
        contour = keywords['contour']

        return self.createCommondEndBlockWithPotentialDeparture(
            approachDepartureType=self.DEPARTURE_TYPE,
            approachDepartureMode=self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords),
            securityDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
            **keywords
        )

    def createRabbetContourStartBlock(self, **keywords):
        return self.__createCommonContourStartBlock(**keywords)

    def createRabbetContourEndBlock(self, **_keywords):
        return self.__createCommonContourEndBlock()

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        return self.createSimpleLineSegment(
            startPosition=segment.getStartVertex().position,
            endPosition=segment.getEndVertex().position,
            **keywords
        )

    def createSimpleLineSegment(self, **keywords):
        startPoint = keywords['startPosition']
        endPoint = keywords['endPosition']

        if 'approachDepartureMode' in keywords:
            _approachDepartureMovement, approachDepartureFlying = keywords['approachDepartureMode']
            approachDepartureType = keywords['approachDepartureType']

            if 'sideDistance' in keywords and approachDepartureFlying and approachDepartureType:
                sideDistance = keywords['sideDistance']
                if isNearlyGreater(sideDistance, 0):
                    line = SimpleLineSegment(startPoint, endPoint)
                    if keywords['radiusCompensation'] == self.RADIUS_COMPENSATION_LEFT:  # already on BOT (bottom)
                        line.moveRight(sideDistance)
                    else:
                        line.moveLeft(sideDistance)

                    if approachDepartureType == self.APPROACH_TYPE:
                        startPoint = SimpleLineSegment.getStartPoint(line)
                    else:  # self.DEPARTURE_TYPE
                        endPoint = SimpleLineSegment.getEndPoint(line)

        else:
            approachDepartureFlying = False
            approachDepartureType = ''

        return '; '.join((
            'M',
            self.getLayer(**keywords),
            str(self.REFERENCE_BOTTOM_LEFT * 100 + self.MILLING_LINE),
            '{x}'.format(x=self.formatAsCoordinate(data=startPoint.y)),
            '{y}'.format(y=self.formatAsCoordinate(data=startPoint.x)),
            '{z}'.format(z=self.formatAsCoordinate(data=self.APPROACH_DEPARTURE_FLYING_Z_VALUE if approachDepartureType == self.APPROACH_TYPE and approachDepartureFlying else keywords['zValue'])),  # depth at the starting point
            '{x}'.format(x=self.formatAsCoordinate(data=endPoint.y)),
            '{y}'.format(y=self.formatAsCoordinate(data=endPoint.x)),
            '{rc}'.format(rc=keywords['radiusCompensation']),
            '{count}'.format(count=self.formatAsInt(data=keywords['numberOfSteps'] if 'numberOfSteps' in keywords else 1)),
            '{z}'.format(z=self.formatAsCoordinate(data=self.APPROACH_DEPARTURE_FLYING_Z_VALUE if approachDepartureType == self.DEPARTURE_TYPE and approachDepartureFlying else keywords['zValue'])),  # optional depth at the end point
        ))

    def createRadiusSegment(self, **keywords):
        segment = keywords['segment']

        # As GannoMAT has the main side on BOT (bottom), we need to adapt direction and eventually the radius compensation
        rotationArcDirection = segment.getDirection('CCW', 'CW', 'CCW', 'CW')

        return self.createSimpleRadiusSegment(
            startPosition=segment.getStartVertex().position,
            endPosition=segment.getEndVertex().position,
            centerPosition=segment.getCenter(),
            rotationArcDirection=rotationArcDirection,
            **keywords
        )

    def createSimpleRadiusSegment(self, **keywords):
        # GannoMAT needs rather precise coordinate values to do its math, so we output the coordinates with angle precision.
        if 'approachDepartureMode' in keywords:
            _approachDepartureMovement, approachDepartureFlying = keywords['approachDepartureMode']
            approachDepartureType = keywords['approachDepartureType']
        else:
            approachDepartureFlying = False
            approachDepartureType = ''

        return '; '.join((
            'M',
            self.getLayer(**keywords),
            str(self.REFERENCE_BOTTOM_LEFT * 100 + self.MILLING_ARC_SEGMENT_P2P),
            '{x}'.format(x=self.formatAsAngle(data=keywords['centerPosition'].y)),
            '{y}'.format(y=self.formatAsAngle(data=keywords['centerPosition'].x)),
            '{z}'.format(z=self.formatAsCoordinate(data=self.APPROACH_DEPARTURE_FLYING_Z_VALUE if approachDepartureType == 'approach' and approachDepartureFlying else keywords['zValue'])),  # depth at the starting point
            '{x}'.format(x=self.formatAsAngle(data=keywords['startPosition'].y)),
            '{y}'.format(y=self.formatAsAngle(data=keywords['startPosition'].x)),
            '{x}'.format(x=self.formatAsAngle(data=keywords['endPosition'].y)),
            '{y}'.format(y=self.formatAsAngle(data=keywords['endPosition'].x)),
            keywords['rotationArcDirection'],
            '{rc}'.format(rc=keywords['radiusCompensation']),
            '{count}'.format(count=self.formatAsInt(data=keywords['numberOfSteps'] if 'numberOfSteps' in keywords else 1)),
            '{z}'.format(z=self.formatAsCoordinate(data=self.APPROACH_DEPARTURE_FLYING_Z_VALUE if approachDepartureType == 'departure' and approachDepartureFlying else keywords['zValue'])),  # optional depth at the end point
        ))

    # Miters
    def orientMiters(self, **keywords):
        pass

    # NC Macros
    def createNCMacroRecordFormat(self, **keywords):
        keywords['ncmacro'].face = self.getLayer(**keywords)
        return '; '.join(parameter.value for parameter in self.getNCMacroCookedParameters(**keywords)) + '\n'

    # Footer
    def createEndOfFile(self, **_keywords):
        return ''
