# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import itertools
import math

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.localization import getLocalizedString, IMAWOP_DESCRIPTIONS_RESOURCEGROUP, FILE_NAME_POSTFIXES_RESOURCEGROUP, IMAWOP_NAME_MESSAGES
from interiorcad.backend.postprocessors.surfacepostprocessor.drillingswithtoolidtofiveaxisdrillingsconverter import DrillingsWithToolIDToFiveAxisDrillingsConverter
from interiorcad.backend.utilities.iterationtools import grouper
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.machiningsexporter.horizontaldrillingsexporter import HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds
from interiorcad.interface.ncexport.oneprogramexporter import OneProgramExporter
from interiorcad.interface.surface.acangles import calculateCAngleFromWVector, normalizeAngle
from interiorcad.interface.surface.contour2d import Contour2D
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.ncexport.base import NCExportBase


class IMAWOPExport(NCExportBase):
    EXPORT_FORMAT = 'IMAWOP'

    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    APPROACH_TYPE_RAPID = -1
    APPROACH_TYPE_NONE = 0
    APPROACH_TYPE_FLYING = 1

    APPROACH_TANGENTIAL = 0
    APPROACH_ARC = 1

    APPROACH_MODE_LINEAR_PIERCING = (APPROACH_TANGENTIAL, APPROACH_TYPE_NONE, True)
    APPROACH_MODE_LINEAR_FLYING = (APPROACH_TANGENTIAL, APPROACH_TYPE_FLYING, True)
    APPROACH_MODE_ARC_PIERCING = (APPROACH_ARC, APPROACH_TYPE_NONE, True)
    APPROACH_MODE_ARC_FLYING = (APPROACH_ARC, APPROACH_TYPE_FLYING, True)
    APPROACH_MODE_WITHOUT = (APPROACH_TANGENTIAL, APPROACH_TYPE_NONE, False)
    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_LINEAR_FLYING

    def __init__(self, **keywords):
        super().__init__(**keywords)
        self.imawopVersion = self.getIMAWOPVersion(**keywords)
        machineConfiguration = self.getMachineConfiguration()
        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 3)
        self.MITER_SAW_DEPTH_FROM_BOTTOM = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_SAW_DEPTH_FROM_BOTTOM_KEY, 10)
        self.SHOW_PROGRAMINFO = machineConfiguration.getAsBool('show-programinfo', False)
        self.DIAGNOSE_ACTIVATED = machineConfiguration.getAsBool('diagnose-activated', True)
        self.CALL_PRODUCTION = machineConfiguration.getAsBool('call-production', True)
        self.PRODUCTION_QUANTITY_INFINITE = machineConfiguration.getAsBool('production-quantity-infinite', False)
        self.PREDEFINE_LAYER = machineConfiguration.getAsBool('predefine-layer', False)
        self.SELECT_MIRRORING = machineConfiguration.getAsBool('select-mirroring', False)
        self.WORKTABLE_CONFIGURATION_TYPE = machineConfiguration.get('worktable-configuration-type', '')
        self.WORKPIECE_AUTO_DETACH = machineConfiguration.get('workpiece-auto-detach', 'never')  # options 'never', 'without sizing' and 'always'
        self.WORKPIECE_AUTO_STOPS = machineConfiguration.getAsBool('workpiece-auto-stops', True)
        self.SAFETY_CLEARANCE = machineConfiguration.getAsInt('safety-clearance', '20')
        self.CREATE_DRILLING_TECHNOLOGY_DATA = machineConfiguration.getAsBool('create-drilling-technology-data', True)

        self.USE_APPROACH_TYPE_RAPID_INSTEAD_NONE = machineConfiguration.getAsBool('use-approach-type-rapid-instead-none', False)  # special approach type for faster movements
        self.useApproachTypeRapidInsteadNone(**keywords)

        self.ONE_PROGRAM = machineConfiguration.getAsBool('one-program', False)
        self.FILE_NAME_POSTFIX_BOTH_SIDED = machineConfiguration.get('file-name-postfix-both-sided', getLocalizedString(FILE_NAME_POSTFIXES_RESOURCEGROUP, 4, 'BothSided'))
        self.USE_CUSTOM_USER_VARIABLES = machineConfiguration.getAsBool('use-custom-user-variables', False)
        self.USE_CUSTOM_USER_WORKPIECE_COMMENT = machineConfiguration.getAsBool('use-custom-user-workpiece-comment', False)

        self.createZValueConverters(**keywords)

        self.lastContourIndex = None
        self.lastSizingIndex = None
        self.lastRabbetIndex = None
        self.lastDadoIndex = None

    def useApproachTypeRapidInsteadNone(self, **_keywords):
        if self.USE_APPROACH_TYPE_RAPID_INSTEAD_NONE:
            self.APPROACH_MODE_LINEAR_PIERCING = (self.APPROACH_TANGENTIAL, self.APPROACH_TYPE_RAPID, True)
            self.APPROACH_MODE_ARC_PIERCING = (self.APPROACH_ARC, self.APPROACH_TYPE_RAPID, True)
            self.APPROACH_MODE_WITHOUT = (self.APPROACH_TANGENTIAL, self.APPROACH_TYPE_RAPID, False)

    def getIMAWOPVersion(self, **_keywords):
        machineConfigurationIMAWOPVersion = self.getMachineConfiguration().get('imawop-version')
        try:
            int(machineConfigurationIMAWOPVersion)
            if len(machineConfigurationIMAWOPVersion) == 1 and int(machineConfigurationIMAWOPVersion) >= 9:
                machineConfigurationIMAWOPVersion = '01'
            elif len(machineConfigurationIMAWOPVersion) > 1 and int(machineConfigurationIMAWOPVersion) >= 90:
                machineConfigurationIMAWOPVersion = '01'
            return machineConfigurationIMAWOPVersion + '0' if len(machineConfigurationIMAWOPVersion) == 1 else machineConfigurationIMAWOPVersion
        except (TypeError, ValueError):
            return machineConfigurationIMAWOPVersion

    def useContourDefinition(self, **_keywords):
        machineConfigurationIMAWOPVersion = self.getMachineConfiguration().get('imawop-version')
        if machineConfigurationIMAWOPVersion == '01':
            return True

        try:
            int(machineConfigurationIMAWOPVersion)
            if len(machineConfigurationIMAWOPVersion) == 1 and int(machineConfigurationIMAWOPVersion) >= 6:
                return True
            elif len(machineConfigurationIMAWOPVersion) > 1 and int(machineConfigurationIMAWOPVersion) >= 60:
                return True
            return False
        except (TypeError, ValueError):
            return False

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMiterPreSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_PRE_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)

    def getNCFileExtension(self, **_keywords):
        return '.fmc'

    def getSurfacesExporter(self, **keywords):
        if self.ONE_PROGRAM:
            return OneProgramExporter(**keywords)
        else:
            return super().getSurfacesExporter(**keywords)

    def createMainAndOppositeSidesTogether(self, mainSurface, oppositeSurface, **keywords):
        self.orientMachinings(surface=mainSurface, **keywords)
        yield self.createHead(surface=mainSurface, **keywords)
        yield self.createMachinings(surface=mainSurface, **keywords)

        self.orientMachinings(surface=oppositeSurface, **keywords)
        yield self.createTurnOver(surface=oppositeSurface, **keywords)
        yield self.createMachinings(surface=oppositeSurface, **keywords)
        yield self.createFoot(surface=oppositeSurface, **keywords)

    def createTurnOver(self, **keywords):
        surface = keywords['surface']
        self.drillingGroupID = 2

        insetRabbets = tuple(filter(lambda rabbet: rabbet.isInset(), surface.rabbetsVertical))

        return joinIfNotNone('\n',
            self.createSeparatorComment(),
            self.createNCStop(comment=self.getLocalizedDescription(26, 'Please turn over part!'), **keywords),
            self.createZeroOffset(additionX=surface.getAdditionXMirrored(), additionY=surface.getAdditionY(), **keywords),

            self.createSizingContourDefinition(**keywords),
            self.createRabbetContourDefinitions(rabbets=insetRabbets, **keywords),
            self.createDadoContourDefinitions(dados=surface.getAllDados(), **keywords),
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords),
        )

    def createZeroOffset(self, **keywords):
        return '\n'.join((
            '[SWNPV_{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={description}'.format(description=self.getLocalizedDescription(27, 'Zero offset')),  # Nullpunktverschiebung
            'KETTE=0',
            'NPVX={npvx}'.format(npvx=self.formatAsCoordinate(data=keywords['additionX'])),
            'NPVY={npvy}'.format(npvy=self.formatAsCoordinate(data=keywords['additionY'])),
            'NPVZ=0',
            '',
            '',
        ))

    def getWorkpieceCommentProperties(self, **keywords):
        dictionary = super().getWorkpieceCommentProperties(**keywords)
        dictionary['VersionIMAWOP'] = self.imawopVersion

        return dictionary

    # Header
    def createLayer(self, **keywords):
        surface = keywords['surface']
        name = surface.workpieceName
        if self.PREDEFINE_LAYER:
            return '\n'.join((
                '[LAYER]',
                'AN=1',
                'FCOL=0',
                'IAN=0',
                'IAUS=0',
                'LS=-1',
                'LT=0',
                'NAME={}'.format(name),
                '',
                '',
            ))

    def createProduz(self, **keywords):
        if self.CALL_PRODUCTION:
            return '\n'.join((
                '[PRODUZ{}]'.format(self.imawopVersion),
                'AFB=1',
                'BEZB={produce}'.format(produce=self.getLocalizedDescription(6, 'Produce')),
                'BTG=SRC',
                'JOK=2',
                'JON=1',
                'KAF=',
                'KOM=(NULL)',
                'STZ={count}'.format(count=keywords['surface'].count if not self.PRODUCTION_QUANTITY_INFINITE else 0),
                'WDA=1',
                '',
                '',
            ))

    def createWOPVariable(self, **keywords):
        for index, variable in enumerate(keywords['variables'], start=1):
            if variable:
                yield 'VAR{index}={name}'.format(index=index, name=variable[0])
                yield 'WERT{index}={value}'.format(index=index, value=variable[1])

    def createOneVariableDefinition(self, **keywords):
        return '\n'.join(
            itertools.chain(
                (
                    '[SWVARD{}]'.format(self.imawopVersion),
                    'AFB=1',
                    'BEZB={description}'.format(description=self.getLocalizedDescription(8, 'Define variables')),
                ),
                self.createWOPVariable(**keywords),
                (
                    '',
                    '',
                )
            )
        )

    def createVariableDefinition(self, **keywords):
        groupedVariables = grouper(keywords['variables'], 10)
        return ''.join((self.createOneVariableDefinition(variables=variables) for variables in groupedVariables))

    def shouldAutoDetachWorkpiece(self, **keywords):
        # options 'never', 'always' and 'without sizing'
        # Auto LOESEN
        #     return True
        #         always
        #         without sizing & workpiece has NO sizing
        #     return False
        #         never
        #         without sizing & workpiece has sizing
        return self.WORKPIECE_AUTO_DETACH == 'always' or (self.WORKPIECE_AUTO_DETACH == 'without sizing' and not keywords['surface'].hasSizing())

    def getWorktableConfiguration(self, **_keywords):
        return self.WORKTABLE_CONFIGURATION_TYPE if self.WORKTABLE_CONFIGURATION_TYPE else 0

    def formatGetEdgesAdditions(self, **keywords):
        addition = keywords['addition']
        return '+' + self.formatAsCoordinate(data=addition) if not isNearlyEqual(addition, 0) else ''

    def createWorkpiece(self, **keywords):
        surface = keywords['surface']
        return '\n'.join((
            '[PGKOPF{}]'.format(self.imawopVersion),
            'AAE={stops:d}'.format(stops=self.WORKPIECE_AUTO_STOPS),
            'ABSTS={distance}'.format(distance=self.SAFETY_CLEARANCE),
            'AFB=1',
            'AXV={addX}'.format(addX=self.formatAsCoordinate(data=surface.getAdditionX())),
            'AYV={addY}'.format(addY=self.formatAsCoordinate(data=surface.getAdditionY())),
            'DIAGNOSE_EIN={diagnose:d}'.format(diagnose=self.DIAGNOSE_ACTIVATED),
            'FAGRP= ',
            'FFarbe=8',
            'FTB={width}'.format(width=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            'FTD={thick}'.format(thick=self.formatAsCoordinate(data=surface.thickness)),
            'FTL={length}'.format(length=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            'HOLKA=-1',
            'HAUB={hood:d}'.format(hood=True),  # Parameter needed for older ImaWOP 3 (not in GUI of ImaWOP 8)
            'KOM1={workpiece} {name}'.format(workpiece=self.getLocalizedDescription(4, 'Workpiece'), name=surface.workpieceName),
            'KOM2={side} {postfix}'.format(side=self.getLocalizedDescription(5, 'Side'), postfix=keywords['filePostfix']),
            'LFarbe=0',
            'LOESEN={detach:d}'.format(detach=self.shouldAutoDetachWorkpiece(**keywords)),
            'MPB={type}'.format(type=self.getWorktableConfiguration(**keywords)),
            'NLB=0',
            'PROGRAMMINFO_ZEIGEN={programinfo:d}'.format(programinfo=self.SHOW_PROGRAMINFO),
            'RTB=B{addY}'.format(addY=self.formatGetEdgesAdditions(addition=surface.getEdgesAdditionY())),
            'RTL=L{addX}'.format(addX=self.formatGetEdgesAdditions(addition=surface.getEdgesAdditionX())),
            'SPGLWKS={mirror:d}'.format(mirror=self.SELECT_MIRRORING),
            'VERSION=Version number',  # ImaWOP 7.0 writes new programs that way: 'Version number'
            'WERKZEUGINFO_ZEIGEN=0',
            'WKDS=0',
            'ZSCHABLO=0',
            '',
            '',
        ))

    def getVacuumSuctionCupY(self, **keywords):
        machineConfiguration = self.getMachineConfiguration()
        VACUUM_SUCTION_CUP_STANDARD = machineConfiguration.get('vacuum-suction-cup-standard', None)
        VACUUM_SUCTION_CUP_Y_HALF = machineConfiguration.get('vacuum-suction-cup-y-half', None)
        VACUUM_SUCTION_CUP_Y_HALF_SWITCH = machineConfiguration.getAsFloat('vacuum-suction-cup-y-half-switch', 0)
        VACUUM_SUCTION_CUP_Y_QUARTER = machineConfiguration.get('vacuum-suction-cup-y-quarter', None)
        VACUUM_SUCTION_CUP_Y_QUARTER_SWITCH = machineConfiguration.getAsFloat('vacuum-suction-cup-y-quarter-switch', 0)

        surface = keywords['surface']
        finishedSizeY = surface.getFinishedSizeY()
        if finishedSizeY < VACUUM_SUCTION_CUP_Y_HALF_SWITCH:
            if finishedSizeY < VACUUM_SUCTION_CUP_Y_QUARTER_SWITCH:
                vacuum_suction_cup_y = VACUUM_SUCTION_CUP_Y_QUARTER
            else:
                vacuum_suction_cup_y = VACUUM_SUCTION_CUP_Y_HALF
        else:
            vacuum_suction_cup_y = VACUUM_SUCTION_CUP_STANDARD

        return vacuum_suction_cup_y

    def createVacuumSuctionCup(self, **keywords):
        vacuum_suction_cup_y = self.getVacuumSuctionCupY(**keywords)
        if vacuum_suction_cup_y:
            return '\n'.join((
                '[SAUGAUFR]',
                'ANSG=1',
                'MAXBK=0',
                'OPTIMIERUNGSART=AUTO',
                'SBD={cup}'.format(cup=vacuum_suction_cup_y),
                'SGCNR=1',
                'WAN=1',
                '',
                '',
            ))

    def createSingleWorkpieceComment(self, comment):
        return '\n'.join((
            '[SWKOMT{}]'.format(self.imawopVersion),
            'BEZB={comment}'.format(comment=comment),
            '',
        ))

    def createSeparatorComment(self):
        return self.createSingleWorkpieceComment(comment='--------------------------------------')

    def customUserWorkpieceCommentFileDefaultText(self, **_keywords):
        return '\n'.join((
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={WorkpieceLabel}: {WorkpieceName}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={CommentLabel}: {Comment}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={ItemIDLabel}: {ItemID}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={EdgeFrontLabel}: {EdgeFront}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={EdgeBackLabel}: {EdgeBack}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={EdgeLeftLabel}: {EdgeLeft}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={EdgeRightLabel}: {EdgeRight}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={CabinetNameLabel}: {CabinetName}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={PositionNumberLabel}: {PositionNumber} {PositionDescription}',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={CountLabel}: {Count}',
            '',
            '',
        ))

    def createCustomUserWorkpieceComment(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='IMAWOP Workpiece Comments.txt',
            customUserVariableFileDefaultText=self.customUserWorkpieceCommentFileDefaultText(**keywords),
            **keywords)

    def createWorkpieceComment(self, **keywords):
        customUserWorkpieceComment = self.createCustomUserWorkpieceComment(**keywords) if self.USE_CUSTOM_USER_WORKPIECE_COMMENT else None

        if customUserWorkpieceComment:
            return customUserWorkpieceComment + ('\n\n' if not customUserWorkpieceComment.endswith('\n') else '')

        else:
            projectInfo = self.getProjectInfo()
            workpieceComments = self.getWorkpieceComments(**keywords)
            isProjectIDValid = ExportHelper.isProjectIDValid(self, **projectInfo)

            return joinIfNotNone('\n',
                self.createSingleWorkpieceComment(comment='{WorkpieceLabel}: {WorkpieceName}'.format_map(workpieceComments)),
                self.createSingleWorkpieceComment(comment='{CommentLabel}: {Comment}'.format_map(workpieceComments)) if workpieceComments['Comment'] else None,
                self.createSingleWorkpieceComment(comment='{ItemIDLabel}: {ItemID}'.format_map(workpieceComments)),
                self.createSingleWorkpieceComment(comment='{Edge1Label}: {Edge1}'.format_map(workpieceComments)) if workpieceComments['Edge1'] else None,  # front
                self.createSingleWorkpieceComment(comment='{Edge2Label}: {Edge2}'.format_map(workpieceComments)) if workpieceComments['Edge2'] else None,  # back
                self.createSingleWorkpieceComment(comment='{Edge3Label}: {Edge3}'.format_map(workpieceComments)) if workpieceComments['Edge3'] else None,  # left
                self.createSingleWorkpieceComment(comment='{Edge4Label}: {Edge4}'.format_map(workpieceComments)) if workpieceComments['Edge4'] else None,  # right
                self.createSingleWorkpieceComment(comment='{CabinetNameLabel}: {CabinetName}'.format_map(workpieceComments)) if workpieceComments['CabinetName'] else None,
                self.createSingleWorkpieceComment(comment='{ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}'.format_map(workpieceComments)) if isProjectIDValid else None,
                self.createSingleWorkpieceComment(comment='{PositionNumberLabel}: {PositionNumber} {PositionDescription}'.format_map(workpieceComments)) if isProjectIDValid else None,
                self.createSingleWorkpieceComment(comment='{CountLabel}: {Count}'.format_map(workpieceComments)) if not self.CALL_PRODUCTION else None,
                '',
            )

    def customUserVariableFileDefaultText(self, **_keywords):
        return '\n'.join((
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB={}'.format(self.getLocalizedDescription(31, 'Examples for <Define variable> and <Define variables (10)>')),
            '',
            '[SWVD1_{VersionIMAWOP}]',
            'AFB=1',
            'BEZB={}'.format(self.getLocalizedDescription(32, 'Define variable')),
            'VAR=SingleExampleVariable',
            'WERT=12.3',
            '',
            '[SWVARD{VersionIMAWOP}]',
            'AFB=1',
            'BEZB={}'.format(self.getLocalizedDescription(33, 'Define variables')),
            'VAR1=ExampleVariable1',
            'VAR2=(NULL)',
            'WERT1=1.2',
            'WERT2=(NULL)',
            '',
            '[SWKOMT{VersionIMAWOP}]',
            'BEZB=--------------------------------------',
            '',
            '',
        ))

    def createCustomUserVariables(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='IMAWOP Variables.txt',
            customUserVariableFileDefaultText=self.customUserVariableFileDefaultText(**keywords),
            **keywords)

    def createHead(self, **keywords):
        self.drillingGroupID = 1

        useContourDefinition = self.useContourDefinition(**keywords)
        surface = keywords['surface']

        self.lastContourIndex = 0
        self.lastSizingIndex = 0
        self.lastRabbetIndex = 0
        self.lastDadoIndex = 0

        insetRabbets = tuple(filter(lambda rabbet: rabbet.isInset(), surface.rabbetsVertical))

        return joinIfNotNone('',
            self.createLayer(**keywords),
            self.createProduz(**keywords),
            self.createWorkpiece(**keywords),
            self.createVacuumSuctionCup(**keywords),
            self.createSeparatorComment() + '\n',
            self.createWorkpieceComment(**keywords),
            self.createSeparatorComment() + '\n',
            self.createCustomUserVariables(**keywords) if self.USE_CUSTOM_USER_VARIABLES else None,
            self.createSizingContourDefinition(**keywords) if useContourDefinition else None,
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords) if useContourDefinition else None,
            self.createRabbetContourDefinitions(rabbets=insetRabbets, **keywords) if useContourDefinition else None,
            self.createDadoContourDefinitions(dados=surface.getAllDados(), **keywords) if useContourDefinition else None,
            )

    # Machinings
    def getPostAssignToolProcessors(self, **_keywords):
        return DrillingsWithToolIDToFiveAxisDrillingsConverter(toolValidator=self.isValidToolId),

    def createSizingContour(self, **keywords):
        sizing = keywords['surface'].sizing

        approachType, approachOnset, useApproachDistance = self.getSizingApproachMode(**keywords)
        if approachOnset == sizing.needsFlyingApproach and self.APPROACH_TYPE_NONE:
            approachOnset = self.APPROACH_TYPE_FLYING

        keywords['approach'] = approachType
        keywords['departure'] = approachType
        keywords['onset'] = approachOnset
        keywords['approachDistance'] = '2*WKZR' if useApproachDistance else 0
        keywords['departureDistance'] = '2*WKZR' if useApproachDistance else 0

        if self.useContourDefinition(**keywords):
            if sizing:
                depthPerStep = self.getRouterDepthPerStep(machining=sizing, machiningDepth=sizing.actualDepth, **keywords)

                return self.__createVerticalCommonContourExecute(
                    radiusCompensation=self.__getSizingRadiusCompensation(**keywords),
                    machining=sizing,
                    zValue=self.getSizingZValue(**keywords),
                    depthPerStep=depthPerStep,
                    **keywords)
        else:
            return self.createSizingContourDefinitionAsExecution(**keywords)

    def createSizingDefinitionForOnePolyline(self, **keywords):
        if self.useContourDefinition(**keywords):
            return super().createSizingDefinitionForOnePolyline(**keywords)
        else:
            keywords['sizing'] = sizing = keywords['surface'].sizing

            creator = super().createSizingDefinitionForOnePolyline
            keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)

            machiningDepth = sizing.actualDepth
            depthPerStep = self.getRouterDepthPerStep(machining=sizing, machiningDepth=machiningDepth, **keywords)

            return '\n'.join(self.generateCommonNoBuiltInDepthSteps(creator=creator, machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    def getHorizontalDrillingsExporter(self, **_keywords):
        return HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds(self)

    def getHorizontalMachiningsExporter(self, **keywords):
        return HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds(self)

    def isValidToolId(self, **keywords):
        return 'tool' in keywords and keywords['tool'] and len(keywords['tool'].getToolID()) > 2

    # Drillings
    def createDrillingTechnologyData(self, **keywords):
        if self.CREATE_DRILLING_TECHNOLOGY_DATA:
            return joinIfNotNone('\n',
                'LGEAN={lgean}'.format(lgean=keywords['lgean']),
                'FAN={fan}'.format(fan=keywords['fan']),
                'LGEAB={lgeab}'.format(lgeab=keywords['lgeab']) if 'lgeab' in keywords else None,
                'F={f}'.format(f=keywords['f']),
                'ZSM={zsm}'.format(zsm=keywords['zsm']) if 'zsm' in keywords else None,
                'N={n}'.format(n=keywords['n']) if 'n' in keywords else None,
                'S={s}'.format(s=keywords['s']) if 's' in keywords else None,
            )

    def createVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']
        x = drilling.position.x
        y = drilling.position.y
        if 'description' not in keywords:
            keywords['description'] = self.getLocalizedDescription(9, 'Drilling')
        installedDiameter = self.formatAsCoordinate(data=drilling.installedDiameter)
        depth = self.formatAsCoordinate(data=keywords['depth'])
        return joinIfNotNone('\n',
            '[VBDMES{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={descrDrilling} Ø{dm} {descrDepth}{depth}'.format(descrDrilling=keywords['description'], dm=installedDiameter, descrDepth=self.getLocalizedDescription(12, 'D'), depth=depth),
            'DM={d}'.format(d=installedDiameter),
            self.createDrillingTechnologyData(lgean=4, fan=2000, lgeab=4, f=8000, zsm=0, n=9000),
            'GRP={group}'.format(group=self.drillingGroupID),
            'MRICHT=0',
            'PKTBEZUG_XY=0',
            'TASTEIN=-1',
            'TI={depth}'.format(depth=depth),
            'X={x}'.format(x=self.formatAsCoordinate(data=x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=y)),
            '',
            '',
        )

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        return joinIfNotNone('\n',
            '[VBLAEN{}]'.format(self.imawopVersion),
            'SPX={x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
            'EPX={x}'.format(x=self.formatAsCoordinate(data=endDrilling.position.x)),
            'Y1={y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
            'RASTER={distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'D={diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
            'MITTE=0',
            'EPY={y}'.format(y=self.formatAsCoordinate(data=endDrilling.position.y)),
            self.createDrillingTechnologyData(lgean=1, fan=1000, lgeab=4, f=3000, s=9000),
            'TASTEIN=-1',
            'MRICHT=0',
            'AFB=1',
            'GRP={group}'.format(group=self.drillingGroupID),
            '',
            '',
        )

    def getDrillingPiercingZValue(self, **keywords):
        return -self.PIERCING_DRILLING_ADDITION if self.PIERCING_DRILLING_ADDITION > 0 else self.getDrillingZValue(depth=keywords['surface'].thickness + self.PIERCING_DRILLING_ADDITION, **keywords)

    def createVerticalDrillingPiercing(self, **keywords):
        return self.createVerticalDrilling(description=self.getLocalizedDescription(11, 'Drilling through'), **keywords)

    def createVerticalCenterPunch(self, **keywords):
        return super().createVerticalCenterPunch(description=self.getLocalizedDescription(10, 'Center Punch'), **keywords)

    CLEAR_POCKET = 1
    DONT_CLEAR_POCKET = 0

    def createVerticalPocketCode(self, **keywords):
        pocket = keywords['pocket']
        toolID = pocket.toolID
        diameter = self.formatAsCoordinate(data=pocket.diameter)
        return '\n'.join((
            '[ZYKTFR{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={description} Ø{diameter}'.format(description=keywords['description'], diameter=diameter),
            'DM={diameter}'.format(diameter=diameter),
            'F=AUTO',
            'FAN=AUTO',
            'GEGENL={direction:d}'.format(direction=self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
            'LGEZU={depthPerStep}'.format(depthPerStep=self.formatAsCoordinate(data=keywords['depthPerStep'])),
            'MPX={x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'MPY={y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'N=AUTO',
            'PKTBEZUG_XY=0',
            'RAEUMEN={}'.format(keywords['clear']),
            'SWKZID={}'.format(toolID),
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'TOMBA=1',
            'WERKZEUGNAME={}'.format(toolID),
            'WKZAKTUELL=1',
            'WKZID={}'.format(toolID),
            '',
            '',
        ))

    def createVerticalPocket(self, **keywords):
        pocket = keywords['pocket']
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=pocket.actualDepth, **keywords)

        return self.createVerticalPocketCode(
            clear=self.CLEAR_POCKET,
            description=self.getLocalizedDescription(13, 'Circular Pocket'),
            depthPerStep=depthPerStep,
            **keywords
        )

    def createVerticalPocketPiercing(self, **keywords):
        pocket = keywords['pocket']
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=pocket.actualDepth, **keywords)

        clear = self.DONT_CLEAR_POCKET if pocket.diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER else self.CLEAR_POCKET
        return self.createVerticalPocketCode(
            clear=clear,
            description=self.getLocalizedDescription(14, 'Circular Pocket through'),
            depthPerStep=depthPerStep,
            **keywords
        )

    def createCuff(self, **keywords):
        pocket = keywords['pocket']
        toolID = pocket.toolID
        diameterAsString = self.formatAsCoordinate(data=pocket.diameter)
        return '\n'.join((
            '[ZYSTUL{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={description} Ø{diameter}'.format(description=keywords['description'], diameter=diameterAsString),
            'BRICHT=1',  # from front
            'F=AUTO',
            'FAN=AUTO',
            'GEGENL=1',
            'LGET1={length}'.format(length=diameterAsString),
            'LGET2={width}'.format(width=diameterAsString),
            'LGEZU={depthPerStep}'.format(depthPerStep=self.formatAsCoordinate(data=keywords['depthPerStep'])),
            'MPX={x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'MPY={y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'MPZ={y}'.format(y=self.formatAsCoordinate(data=pocket.zPosition)),
            'N=AUTO',
            'PKTBEZUG_XY=0',
            'R={radius}'.format(radius=pocket.diameter / 2),
            'SWKZID={}'.format(toolID),
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=pocket.depth)),
            'TOMBA=1',
            'WERKZEUGNAME={}'.format(toolID),
            'WKZAKTUELL=1',
            'WKZID={}'.format(toolID),
            '',
            '',
        ))

    def createChangePosition(self, **keywords):
        pocket = keywords['pocket']
        description = self.getLocalizedDescription(28, "Change Position (Angle {cAngle})").format(cAngle=self.formatAsAngle(data=keywords['cAngle']))
        return '\n'.join((
            '[SWLAGE{}]'.format(self.imawopVersion),
            'AFB=1',
            'APX={x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'APY={y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'BEZB={description}'.format(description=description),
            'DPX={x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'DPY={y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'DWI={cAngle}'.format(cAngle=self.formatAsAngle(data=keywords['cAngle'])),
            'GDW=',
            'KM=0',
            '',
            '',
        ))

    def createHorizontalPocket(self, **keywords):
        pocket = keywords['pocket']
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=pocket.actualDepth, **keywords)
        cAngle = normalizeAngle(calculateCAngleFromWVector(pocket.matrix.w) - 90)
        return joinIfNotNone('\n',
            self.createChangePosition(cAngle=cAngle, **keywords) if not isNearlyEqual(cAngle, 0) else None,
            self.createCuff(description=self.getLocalizedDescription(29, "Horizontal Circular Pocket"), depthPerStep=depthPerStep, **keywords),
            self.createChangePosition(cAngle=0, **keywords) if not isNearlyEqual(cAngle, 0) else None,
        )

    FROM_LEFT, FROM_RIGHT, FROM_FRONT, FROM_BACK = range(4)

    def createHorizontalDrillingCode_X(self, **keywords):
        # TODO IT-278 Move up macroName & xPos´, because they might be referenced before assignment  --mjk
        if keywords['kind'] == self.FROM_LEFT:
            macroName = '[HBXPLU{}]'.format(self.imawopVersion)
            xPos = '0'
        elif keywords['kind'] == self.FROM_RIGHT:
            macroName = '[HBXMIN{}]'.format(self.imawopVersion)
            xPos = 'L'

        return joinIfNotNone('\n',
            '{}'.format(macroName),
            'AFB=1',
            'DM={d}'.format(d=self.formatAsCoordinate(data=keywords['installedDiameter'])),
            self.createDrillingTechnologyData(lgean=4, fan=2000, f=4000, n=8000),
            'GRP=1',
            'KETTE=1',
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'X1={x}'.format(x=xPos),
            'X2=-1',
            'Y1={y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            'Y2=(NULL)',
            'Y3=(NULL)',
            'Y4=(NULL)',
            'Y5=(NULL)',
            'Y6=(NULL)',
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['z'])),
            '',
            '',
        )

    def createHorizontalDrillingFromLeft(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        return self.createHorizontalDrillingCode_X(kind=self.FROM_LEFT, y=drilling.position.y, z=zPosition, depth=drilling.depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter, **keywords)

    def createHorizontalDrillingFromRight(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        return self.createHorizontalDrillingCode_X(kind=self.FROM_RIGHT, y=drilling.position.y, z=zPosition, depth=drilling.depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter, **keywords)

    def createHorizontalDrillingCode_Y(self, **keywords):
        # TODO IT-278 Move up macroName & yPos´, because they might be referenced before assignment  --mjk
        if keywords['kind'] == self.FROM_FRONT:
            macroName = '[HBYPLU{}]'.format(self.imawopVersion)
            yPos = '0'
        elif keywords['kind'] == self.FROM_BACK:
            macroName = '[HBYMIN{}]'.format(self.imawopVersion)
            yPos = 'B'

        return joinIfNotNone('\n',
            '{}'.format(macroName),
            'AFB=1',
            'DM={d}'.format(d=self.formatAsCoordinate(data=keywords['installedDiameter'])),
            self.createDrillingTechnologyData(lgean=4, fan=2000, f=4000, n=8000),
            'GRP=1',
            'KETTE=1',
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'X1={x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            'X2=(NULL)',
            'X3=(NULL)',
            'X4=(NULL)',
            'X5=(NULL)',
            'X6=(NULL)',
            'Y1={y}'.format(y=yPos),
            'Y2=-1',
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['z'])),
            '',
            '',
        )

    def createHorizontalDrillingFromFront(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        return self.createHorizontalDrillingCode_Y(kind=self.FROM_FRONT, x=drilling.position.x, z=zPosition, depth=drilling.depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter)

    def createHorizontalDrillingFromBack(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        return self.createHorizontalDrillingCode_Y(kind=self.FROM_BACK, x=drilling.position.x, z=zPosition, depth=drilling.depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter)

    def createRotatedHorizontalDrilling(self, **keywords):
        # This function is only called when the horizontal driller is allowed to rotate (otherwise it is handled by the filter DrillingsWithToolIDToFiveAxisDrillingsConverter)
        drilling = keywords['drilling']
        description = self.getLocalizedDescription(24, 'Horizontal Drilling')
        installedDiameter = self.formatAsCoordinate(data=drilling.installedDiameter)
        depth = self.formatAsCoordinate(data=drilling.depth)
        return joinIfNotNone('\n',
            '[HBEINW{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={descrDrilling} Ø{dm} {descrDepth}{depth}'.format(descrDrilling=description, dm=installedDiameter, descrDepth=self.getLocalizedDescription(12, 'D'), depth=depth),
            'DM={d}'.format(d=installedDiameter),
            self.createDrillingTechnologyData(lgean=4, fan=2000, f=4000, n=8000, zsm=0),
            'GRP=1',
            'TI={depth}'.format(depth=depth),
            'TWKL=0',
            'WKL={c}'.format(c=self.formatAsAngle(data=drilling.cAngle % 360)),
            'X={x}'.format(x=self.formatAsCoordinate(data=drilling.position.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=drilling.position.y)),
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['zPosition'])),
            '',
            '',
        )

    def createFiveAxisDrilling(self, **keywords):
        drilling = keywords['drilling']
        description = self.getLocalizedDescription(3, '5 axis drilling')
        installedDiameter = self.formatAsCoordinate(data=drilling.installedDiameter)
        depth = self.formatAsCoordinate(data=drilling.depth)
        return '\n'.join((
            '[ROTNRT{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={descrDrilling} Ø{dm} {descrDepth}{depth}'.format(descrDrilling=description, dm=installedDiameter, descrDepth=self.getLocalizedDescription(12, 'D'), depth=depth),
            'BT={depth}'.format(depth=depth),
            'BWKL=90-{a}'.format(a=self.formatAsAngle(data=90 - drilling.aAngle)),
            'CWKL={c}-90'.format(c=self.formatAsAngle(data=drilling.cAngle % 360)),
            'WKZID={toolID}'.format(toolID=drilling.toolID),
            'X={x}'.format(x=self.formatAsCoordinate(data=drilling.position.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=drilling.position.y)),
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['zPosition'])),
            '',
            '',
        ))

    # Contours
    def __createVerticalCommonContourStartBlockDefinition(self, **keywords):
        return joinIfNotNone('\n',
            '[KDEFANFC]',
            # HP_RADIUS=0
            # JUS = 0
            'BEZB={description}'.format(description=keywords['description']),
            'KNAME={name}'.format(name=keywords['machining'].name),
            # PKTBEZUG_XY = 0
            # SICHTBAR = 1
            # TH = -10
            # TNAM=Courier New
        )

    def createVerticalCommonContourExecutionInASingleStep(self, **keywords):
        toolID = keywords['machining'].toolID

        return joinIfNotNone('\n',
            '[KTFRAE{}]'.format(self.imawopVersion),
            'AFB=1',
            #### 'BEZB={description}'.format(description=keywords['description']),
            'F={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            'KNAME={name}'.format(name=keywords['machining'].name),
            'LGEAB={}'.format(keywords['departureDistance']) if 'departureDistance' in keywords else None,
            'LGEAN={}'.format(keywords['approachDistance']) if 'approachDistance' in keywords else None,
            'SWKZID={}'.format(toolID),
            'TRKOR={}'.format(keywords['radiusCompensation']),
            'TYPAB={}'.format(keywords['departure']),
            'TYPAN={}'.format(keywords['approach']),
            'TYPEIN={}'.format(keywords['onset']),
            'WERKZEUGNAME={}'.format(toolID),
            'WKZID={}'.format(toolID),
            'WKZAKTUELL=1',
            'KD={kd}'.format(kd=self.formatAsCoordinate(data=-keywords['contourDistance'])) if 'contourDistance' in keywords else None,
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            '',
            '',
        )

    def createVerticalCommonContourExecutionWithDepthPerStep(self, **keywords):
        toolID = keywords['machining'].toolID

        return joinIfNotNone('\n',
            '[KTFRZU{}]'.format(self.imawopVersion),
            'AFB=1',
            #### 'BEZB={description}'.format(description=keywords['description']),
            'F={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            'KNAME={name}'.format(name=keywords['machining'].name),
            'LGEAB={}'.format(keywords['departureDistance']) if 'departureDistance' in keywords else None,
            'LGEAN={}'.format(keywords['approachDistance']) if 'approachDistance' in keywords else None,
            'SWKZID={}'.format(toolID),
            'TRKOR={}'.format(keywords['radiusCompensation']),
            'TYPAB={}'.format(keywords['departure']),
            'TYPAN={}'.format(keywords['approach']),
            'TYPEIN={}'.format(keywords['onset']),
            'WERKZEUGNAME={}'.format(toolID),
            'WKZID={}'.format(toolID),
            'WKZAKTUELL=1',
            'KD={kd}'.format(kd=self.formatAsCoordinate(data=-keywords['contourDistance'])) if 'contourDistance' in keywords else None,
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            'ZUSTANZ={count}'.format(count=keywords['numberOfSteps']),
            'ZZUST1={startDepth}'.format(startDepth=self.formatAsCoordinate(data=keywords['startDepth'])),
            '',
            '',
        )

    def createHorizontalCommonContourExecution(self, **keywords):
        toolID = keywords['machining'].toolID

        return joinIfNotNone('\n',
            '[KTHFRA{}]'.format(self.imawopVersion),
            'AFB=1',
            #### 'BEZB={description}'.format(description=keywords['description']),
            'F={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            'KNAME={name}'.format(name=keywords['machining'].name),
            'LGEAB={}'.format(keywords['departureDistance']) if 'departureDistance' in keywords else None,
            'LGEAN={}'.format(keywords['approachDistance']) if 'approachDistance' in keywords else None,
            'SWKZID={}'.format(toolID),
            'TRKOR={}'.format(keywords['radiusCompensation']),
            'TYPAB={}'.format(keywords['departure']),
            'TYPAN={}'.format(keywords['approach']),
            'TYPEIN={}'.format(keywords['onset']),
            'WERKZEUGNAME={}'.format(toolID),
            'WKZID={}'.format(toolID),
            'WKZAKTUELL=1',
            'KD={kd}'.format(kd=self.formatAsCoordinate(data=-keywords['contourDistance'])) if 'contourDistance' in keywords else None,
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            '',
            '',
        )

    def __createVerticalCommonContourStartBlockExecution(self, **keywords):
        toolID = keywords['machining'].toolID

        return joinIfNotNone('\n',
            '[KAFRAE{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={description}'.format(description=keywords['description']),
            'F={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            'LGEAB={}'.format(keywords['departureDistance']) if 'departureDistance' in keywords else None,
            'LGEAN={}'.format(keywords['approachDistance']) if 'approachDistance' in keywords else None,
            'SWKZID={}'.format(toolID),
            'TRKOR={}'.format(keywords['radiusCompensation']),
            'TYPAB={}'.format(keywords['departure']),
            'TYPAN={}'.format(keywords['approach']),
            'TYPEIN={}'.format(keywords['onset']),
            'WERKZEUGNAME={}'.format(toolID),
            'WKZID={}'.format(toolID),
            'WKZAKTUELL=1',
            'KD={kd}'.format(kd=self.formatAsCoordinate(data=-keywords['contourDistance'])) if 'contourDistance' in keywords else None,
       )

    def __createFiveAxisCommonContourStartBlockExecution(self, **keywords):
        machining = keywords['machining']
        toolID = machining.toolID

        return joinIfNotNone('\n',
            '[KAHFRA{}]'.format(self.imawopVersion),
            'AFB=1',
            'AGGD=0',  # The angle should not follow the contour
            'AGGFWKL={cangle}'.format(cangle=self.formatAsAngle(data=(calculateCAngleFromWVector(wvector=machining.matrix.w) - 90) % 360)),
            'AGGO=0',
            'BEZB={description}'.format(description=keywords['description']),
            'F={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            'LGEAB={}'.format(keywords['departureDistance']) if 'departureDistance' in keywords else None,
            'LGEAN={}'.format(keywords['approachDistance']) if 'approachDistance' in keywords else None,
            'NGW={aangle}'.format(aangle=self.formatAsAngle(data=machining.aAngle)),
            'SWKZID={}'.format(toolID),
            'TRKOR={}'.format(keywords['radiusCompensation']),
            'TYPAB={}'.format(keywords['departure']),
            'TYPAN={}'.format(keywords['approach']),
            'TYPEIN={}'.format(keywords['onset']),
            'WERKZEUGNAME={}'.format(toolID),
            'WKZID={}'.format(toolID),
            'WKZAKTUELL=1',
       )

    def __createCommonContourStartBlock(self, **keywords):
        if self.useContourDefinition(**keywords):
            return self.__createVerticalCommonContourStartBlockDefinition(**keywords)
        elif keywords['machining'].isVertical():
            return self.__createVerticalCommonContourStartBlockExecution(**keywords)
        else:
            return self.__createFiveAxisCommonContourStartBlockExecution(**keywords)

    def createCommonRouterDefinitionDescription(self, description):
        return ' '.join((
            description,
            self.getLocalizedDescription(15, 'Router definition'),
        ))

    def createSizingStartBlock(self, **keywords):
        sizing = keywords['surface'].sizing

        self.lastSizingIndex += 1
        sizing.name = self.getLocalizedMachiningName(2, "Sizing {index}").format(index=self.lastSizingIndex)

        return '\n'.join((
            self.__createCommonContourStartBlock(
                description=self.createCommonRouterDefinitionDescription(description=self.getLocalizedDescription(16, 'Sizing')),
                radiusCompensation=self.__getSizingRadiusCompensation(**keywords),
                **keywords
            ),
            self.createContourFirstCornerPolylineVertex(**keywords),
        ))

    def createContourStartBlock(self, **keywords):
        contour = keywords['contour']

        radiusCompensation = self.getWopRadiusCompensation(**keywords)
        approachType, approachOnset, useApproachDistance = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departureType, _departureOnset, useDepartureDistance = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        if not contour.isVertical() or (0 < keywords['zValue'] < keywords['surface'].thickness):
            description = self.getLocalizedDescription(17, 'Contour')
        else:
            description = self.getLocalizedDescription(18, 'Contour through')

        self.lastContourIndex += 1
        contour.name = self.getLocalizedMachiningName(4, "Contour {index}").format(index=self.lastContourIndex)

        return '\n'.join((
            self.__createCommonContourStartBlock(
                description=self.createCommonRouterDefinitionDescription(description=description),
                radiusCompensation=radiusCompensation,
                approach=approachType,
                departure=departureType,
                onset=approachOnset,
                approachDistance='2*WKZR' if useApproachDistance else 0,
                departureDistance='2*WKZR' if useDepartureDistance else 0,
                # TODO CONTOUR2.0 IT-243 Currently feedRate and depthPerStep is only implemented directly within createContourStartBlock, not other contour using machinings --mjk
                feedRate=contour.currentFeedRate,
                **keywords
            ),
            self.createContourFirstCornerPolylineVertex(**keywords),
        ))

    def createRabbetContourStartBlock(self, **keywords):
        rabbet = keywords['rabbet']

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        self.lastRabbetIndex += 1
        rabbet.name = self.getLocalizedMachiningName(3, "Rabbet {index}").format(index=self.lastRabbetIndex)

        if 'zValue' not in keywords:
            keywords['zValue'] = self.zValueMillingConverter.convertWithThickness(keywords['surface'].thickness, rabbet.actualDepth)

        return '\n'.join((
            self.__createCommonContourStartBlock(
                description=self.createCommonRouterDefinitionDescription(description=self.getLocalizedDescription(20, 'Rabbet')),
                radiusCompensation=radiusCompensation,
                **keywords
            ),
            self.createContourFirstCornerPolylineVertex(**keywords),
        ))

    def createDadoContourStartBlock(self, **keywords):
        dado = keywords['dado']

        self.lastDadoIndex += 1
        dado.name = self.getLocalizedMachiningName(1, "Dado {index}").format(index=self.lastDadoIndex)

        if 'zValue' not in keywords:
            keywords['zValue'] = self.zValueMillingConverter.convertWithThickness(keywords['surface'].thickness, dado.actualDepth)

        return '\n'.join((
            self.__createCommonContourStartBlock(
                description=self.createCommonRouterDefinitionDescription(description=self.getLocalizedDescription(19, 'Dado')),
                **keywords
            ),
            self.createContourFirstCornerPolylineVertex(**keywords),
        ))

    def __getSizingRadiusCompensation(self, **keywords):
        keywords['sizing'] = keywords['surface'].sizing
        return self.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else self.RADIUS_COMPENSATION_LEFT

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=IMAWOPExport.RADIUS_COMPENSATION_NONE, left=IMAWOPExport.RADIUS_COMPENSATION_LEFT, right=IMAWOPExport.RADIUS_COMPENSATION_RIGHT)

    def createSizingEndBlock(self, **keywords):
        if self.useContourDefinition(**keywords):
            return '\n'.join((
                '[KDEFENDE]',
                '',
                '',
            ))

        else:
            return '\n'.join((
                '[KOABFA{}]'.format(self.imawopVersion),
                'AFB=1',
                'AGGD=AGGDREHBAR',
                'AGGFWKL=AGGFWKL',
                'AGGO=AGGOFFSET',
                '',
                '',
            ))

    def createSimpleCommonContourEndBlock(self, **keywords):
        if self.useContourDefinition(**keywords):
            macroName = 'KDEFENDE'
        else:
            macroName = 'KOABFA{version}'.format(version=self.imawopVersion)

        return '\n'.join((
            '[{name}]'.format(name=macroName),
            '',
            '',
        ))

    def createContourEndBlock(self, **keywords):
        return self.createSimpleCommonContourEndBlock(**keywords)

    def createRabbetContourEndBlock(self, **keywords):
        return self.createSimpleCommonContourEndBlock(**keywords)

    def createDadoContourEndBlock(self, **keywords):
        return self.createSimpleCommonContourEndBlock(**keywords)

    def createContourFirstCornerPolylineVertex(self, **keywords):
        vertex = keywords['startVertex']
        x, y = vertex.position.asTuple()
        z = keywords['zValue']

        return joinIfNotNone('\n',
            'X={x}'.format(x=self.formatAsCoordinate(data=x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=y)),
            'Z={z}'.format(z=self.formatAsCoordinate(data=z)) if not self.useContourDefinition(**keywords) else None,
            '',
        )

    def createLineSegment(self, **keywords):
        endPosition = keywords['segment'].getEndVertex().position

        return '\n'.join((
            '[G1]',
            'AFB=1',
            'AGGDREHBAR=AGGDREHBAR',
            'AGGFWKL=AGGWKL',
            'AGGGEDREHT=0',
            'AGGOFFSET=AGGOFFSET',
            'EB=EBG',
            'F=V',
            'PKTBEZUG_XY=0',
            'X={x}'.format(x=self.formatAsCoordinate(data=endPosition.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=endPosition.y)),
            'Z=Z',
            '',
        ))

    def createRadiusSegment(self, **keywords):
        IMAWOP_CW = (1, '2')
        IMAWOP_CCW = (1, '3')
        IMAWOP_CW180 = (-1, '2')
        IMAWOP_CCW180 = (-1, '3')

        segment = keywords['segment']

        multiplicationFactor, directionCode = segment.getDirection(IMAWOP_CW, IMAWOP_CCW, IMAWOP_CW180, IMAWOP_CCW180)
        radius = segment.getRadius() * multiplicationFactor

        endPosition = segment.getEndVertex().position

        return '\n'.join((
            '[G{}]'.format(directionCode),
            'AFB=1',
            'AGGDREHBAR=AGGDREHBAR',
            'AGGFWKL=AGGWKL',
            'AGGGEDREHT=0',
            'AGGOFFSET=AGGOFFSET',
            'EB=EBG',
            'F=V',
            'MPX=X',
            'MPY=Y',
            'PKTBEZUG_XY=0',
            'RAD={radius}'.format(radius=self.formatAsCoordinate(data=radius)),
            'X={x}'.format(x=self.formatAsCoordinate(data=endPosition.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=endPosition.y)),
            'Z=Z',
            ''
        ))

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']

        machiningPlane = Plane.createFromMachining(machining)

        assert machiningPlane is not None, "No plane found for machining"

        return machiningPlane, machining.getPositionOnPlane(machiningPlane)

    def getIMAWOPVersionForPlanes(self):
        if self.imawopVersion == '70':
            return '71'
        elif self.imawopVersion == '60':
            return '61'
        else:
            return self.imawopVersion

    def createPlane(self, **keywords):
        plane = keywords['plane']

        return '\n'.join((
            '[SEBENE{}]'.format(self.getIMAWOPVersionForPlanes()),
            'AFB=1',
            'BEZB={description}'.format(description=self.getLocalizedDescription(30, "Inclined plane definition")),
            'EA={angle}'.format(angle=self.formatAsCoordinate(data=plane.aAngle)),
            'EC={angle}'.format(angle=self.formatAsCoordinate(data=normalizeAngle(plane.cAngleToXYPlane - 90))),
            'EU={x}'.format(x=self.formatAsCoordinate(data=plane.origin.x)),
            'EV={y}'.format(y=self.formatAsCoordinate(data=plane.origin.y)),
            'EW={z}'.format(z=self.formatAsCoordinate(data=plane.origin.z)),
            'FREB_EIN=1',
            'ProjMode=2004',
            'SAVEHIDDEN=1',
            '',
            '',
        ))

    def createDefaultPlane(self, **_keywords):
        return '\n'.join((
            '[SEBENE{}]'.format(self.getIMAWOPVersionForPlanes()),
            'AFB=1',
            'BEZB={description}'.format(description=self.getLocalizedDescription(30, "Inclined plane definition")),
            'EA=0',
            'EC=0',
            'EU=0',
            'EV=0',
            'EW=0',
            'FREB_EIN=0',
            'ProjMode=2004',
            'SAVEHIDDEN=1',
            '',
            '',
        ))

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']
        # TODO (HORIZONTAL) IT-222/IT-230: unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if contour.isVertical():
            plane = None

            if 'zValue' not in keywords:
                keywords['zValue'] = self.getContourZValue(depth=keywords['contour'].depth, **keywords)

        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            contour.path = contour.transformedPolyline(matrix.getInverted(), positionOnPlane)

            if 'zValue' not in keywords:
                keywords['zValue'] = - keywords['contour'].depth

        return joinIfNotNone('\n',
            self.createPlane(plane=plane, **keywords) if plane and not self.useContourDefinition(**keywords) else None,
            super().createDefinitionForOneContour(**keywords),
            self.createDefaultPlane(**keywords) if plane and not self.useContourDefinition(**keywords) else None,
        )

    def createFreeFormPocket(self, **keywords):
        toolID = keywords['machining'].toolID
        machiningDepth = keywords['contour'].actualDepth
        depthPerStep = keywords['depthPerStep']

        # TODO CONTOUR2.0 Innoweek 2021/1: should this be a generalised function? Also used in __createVerticalCommonContourExecute() -- NG/MJK
        useSteps = (machiningDepth > depthPerStep > 0.0) and not isNearlyEqual(depthPerStep, 0.0) and not isNearlyEqual(machiningDepth, depthPerStep)
        if useSteps:
            numberOfStepsDivider = math.ceil(machiningDepth / depthPerStep)
            rawDepthStep = machiningDepth / numberOfStepsDivider
            assert 0 < rawDepthStep <= depthPerStep or isNearlyEqual(rawDepthStep, depthPerStep), "Calculated step should not be 0 and not be bigger than maximum step!"
        else:
            rawDepthStep = None

        return joinIfNotNone('\n',
            '[ZYKPTA{}]'.format(self.imawopVersion),
            'F={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            'LGEAN={}'.format(keywords['approachDistance']) if 'approachDistance' in keywords else None,
            'NAME={name}'.format(name=keywords['machining'].name),
            'OffSXY={}'.format(self.formatAsCoordinate(data=keywords['contourDistance'])) if 'contourDistance' in keywords else None,
            'SWKZID={}'.format(toolID),
            'WERKZEUGNAME={}'.format(toolID),
            'TI={}'.format(self.formatAsCoordinate(data=machiningDepth)),
            'TYPEIN={}'.format(keywords['onset']),
            'WKZID={}'.format(toolID),
            'WKZAKTUELL=1',
            'ZUSTZ={}'.format(self.formatAsCoordinate(data=rawDepthStep)) if useSteps else None,
            '',
            '',
        )

    def __createVerticalContourExecute(self, **keywords):
        contour = keywords['contour']
        machiningDepth = contour.actualDepth

        depthPerStep = contour.currentDepthPerStep
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        approachType, approachOnset, useApproachDistance = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departureType, _departureOnset, useDepartureDistance = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        if 'zValue' not in keywords:
            keywords['zValue'] = self.zValueMillingConverter.convertWithThickness(keywords['surface'].thickness, contour.actualDepth)

        if contour.clearOut:
            return self.createFreeFormPocket(
                depthPerStep=depthPerStep,
                onset=approachOnset,
                approachDistance='2*WKZR' if useApproachDistance else 0,
                **keywords)
        else:
            return self.__createVerticalCommonContourExecute(
                depthPerStep=depthPerStep,
                radiusCompensation=self.getWopRadiusCompensation(**keywords),
                approach=approachType,
                departure=departureType,
                onset=approachOnset,
                approachDistance='2*WKZR' if useApproachDistance else 0,
                departureDistance='2*WKZR' if useDepartureDistance else 0,
                **keywords)

    def __createVerticalCommonContourExecute(self, **keywords):
        machining = keywords['machining']
        machiningDepth = machining.actualDepth
        depthPerStep = keywords['depthPerStep']

        # TODO CONTOUR2.0 IT-237/IT-272 : should this be a generalised function? Also used in createFreeFormPocket() -- NG
        useSteps = (machiningDepth > depthPerStep > 0.0) and not isNearlyEqual(depthPerStep, 0.0) and not isNearlyEqual(machiningDepth, depthPerStep)
        if useSteps:
            numberOfStepsDivider = math.ceil(machiningDepth / depthPerStep)
            rawDepthStep = machiningDepth / numberOfStepsDivider
            numberOfSteps = numberOfStepsDivider
            assert 0 < rawDepthStep <= depthPerStep or isNearlyEqual(rawDepthStep, depthPerStep), "Calculated step should not be 0 and not be bigger than maximum step!"
            startDepth = self.zValueMillingConverter.convertWithThickness(keywords['surface'].thickness, rawDepthStep)
            creator = self.createVerticalCommonContourExecutionWithDepthPerStep
        else:
            numberOfSteps = 0
            startDepth = 0
            creator = self.createVerticalCommonContourExecutionInASingleStep

        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)

        return creator(
            numberOfSteps=numberOfSteps,
            startDepth=startDepth,
            **keywords)

    def createVerticalContour(self, **keywords):
        if self.useContourDefinition(**keywords):
            return self.__createVerticalContourExecute(**keywords)
        else:
            keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
            keywords['creator'] = self.createDefinitionForOneContour
            return '\n'.join(self.generateContourNoBuiltInDepthSteps(**keywords))

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createVerticalContour, **keywords))

    def createHorizontalContour(self, **keywords):
        contour = keywords['contour']

        keywords['zValueConverter'] = lambda _boardThickness, currentDepth: - currentDepth

        if self.useContourDefinition(**keywords):
            radiusCompensation = self.getWopRadiusCompensation(**keywords)
            approachType, approachOnset, useApproachDistance = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
            departureType, _departureOnset, useDepartureDistance = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

            plane, _positionOnPlane = self.__calculatePlane(**keywords)

            keywords['creator'] = self.createHorizontalCommonContourExecution

            return joinIfNotNone('\n',
                self.createPlane(plane=plane, **keywords) if plane and self.useContourDefinition(**keywords) else None,
                '\n'.join(self.generateContourNoBuiltInDepthSteps(
                    radiusCompensation=radiusCompensation,
                    approach=approachType,
                    departure=departureType,
                    onset=approachOnset,
                    approachDistance='2*WKZR' if useApproachDistance else 0,
                    departureDistance='2*WKZR' if useDepartureDistance else 0,
                    **keywords)),
                self.createDefaultPlane(**keywords) if plane and self.useContourDefinition(**keywords) else None,
            )

        else:
            keywords['creator'] = self.createDefinitionForOneContour
            return '\n'.join(self.generateContourNoBuiltInDepthSteps(**keywords))

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createHorizontalContour, **keywords))

    def createNCStop(self, **keywords):
        surface = keywords['surface']
        finishedSizeX = self.formatAsCoordinate(data=surface.getFinishedSizeX())
        stopDistance = self.getMachineConfiguration().getAsFloat('stop-distance', 1000)
        stopDistanceX = self.formatAsCoordinate(data=stopDistance)
        return '\n'.join((
            self.createSingleWorkpieceComment(comment=keywords['comment']),
            '[SNHALT{}]'.format(self.imawopVersion),
            'AFB=1',
            'APX={finishedSizeX}+{stopDistanceX}'.format(finishedSizeX=finishedSizeX, stopDistanceX=stopDistanceX),
            'APY=0',
            '',
            '',
        ))

    # Dados
    def createVerticalDadoWithSaw(self, **keywords):
        dado = keywords['dado']
        start, end = self.getDadoLineSegmentWithSecurityDistance(**keywords).getStartAndEndPoints()
        toolID = dado.toolID
        return '\n'.join((
            '[ZYSNUT{}]'.format(self.imawopVersion),
            'ABSTN=10',
            'AFB=1',
            'BEZB={description}'.format(description=keywords['description']),
            'EPX={x}'.format(x=self.formatAsCoordinate(data=end.x)),
            'EPY={y}'.format(y=self.formatAsCoordinate(data=end.y)),
            'F=AUTO',
            'FAN=AUTO',
            'GEGENL=0',
            'N=AUTO',
            'NB={width}'.format(width=self.formatAsCoordinate(data=dado.getWidth())),
            'PKTBEZUG_XY=0',
            'SPX={x}'.format(x=self.formatAsCoordinate(data=start.x)),
            'SPY={y}'.format(y=self.formatAsCoordinate(data=start.y)),
            'SWKZID={}'.format(toolID),
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'TOMBA=3',
            'TRKOR=0',
            'TWKL=0',
            'TYPA=1',
            'TYPN=1',
            'WERKZEUGNAME={}'.format(toolID),
            'WKZID={}'.format(toolID),
            'Z=D/2',
            '',
            '',
        ))

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        return self.createVerticalDadoWithSaw(description=self.getLocalizedDescription(21, 'Dado X'), **keywords)

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        return self.createVerticalDadoWithSaw(description=self.getLocalizedDescription(22, 'Dado Y'), **keywords)

    def createVerticalDadoWithSawCRotated(self, **keywords):
        return self.createVerticalDadoWithSaw(description=self.getLocalizedDescription(25, 'Dado C Rotated'), **keywords)

    def generateDadoSegments(self, **keywords):
        if not keywords['dado'].isVertical():
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            dado = copy.deepcopy(keywords['dado'])
            matrix = plane.matrix * dado.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            dado.matrix = matrix  # TODO (HORIZONTAL) IT-230/IT-272: unschönes Design -- NG
            dado.matrix.origin = positionOnPlane
            keywords['dado'] = dado

        return super().generateDadoSegments(**keywords)

    def createCommonDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        radiusCompensation = self.getDadoRadiusCompensation(**keywords)

        machiningDepth = dado.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return '\n'.join(
            self.generateCommonNoBuiltInDepthSteps(
                machiningDepth=machiningDepth,
                depthPerStep=depthPerStep,
                radiusCompensation=radiusCompensation,
                segmentList=self.generateDadoSegments(**keywords),
                **keywords))

    def createVerticalDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        approachType, approachOnset, _useApproachDistance = self.APPROACH_MODE_LINEAR_PIERCING
        keywords['approach'] = approachType
        keywords['departure'] = approachType
        keywords['onset'] = self.APPROACH_TYPE_FLYING if dado.isInsetBoth() else approachOnset

        if self.useContourDefinition(**keywords):
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=dado.actualDepth, **keywords)
            return self.__createVerticalCommonContourExecute(
                depthPerStep=depthPerStep,
                radiusCompensation = self.getDadoRadiusCompensation(**keywords),
                zValue=self.zValueMillingConverter.convertWithThickness(keywords['surface'].thickness, dado.actualDepth),
                **keywords)

        else:
            def creator(**kwargs):
                zValue = kwargs['zValue']
                del kwargs['zValue']  # TODO IT-272: can we do it better? -- NG
                return self.createContourFromSegmentList(
                    createStartBlock=self.createDadoContourStartBlock,
                    createEndBlock=self.createDadoContourEndBlock,
                    zValue=zValue,
                    **kwargs)

            return self.createCommonDadoWithRouter(
                creator=creator,
                zValueConverter=lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth),
                **keywords)

    def createHorizontalDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        approachType, approachOnset, _useApproachDistance = self.APPROACH_MODE_LINEAR_PIERCING
        keywords['approach'] = approachType
        keywords['departure'] = approachType
        keywords['onset'] = self.APPROACH_TYPE_FLYING if dado.isInsetBoth() else approachOnset

        if self.useContourDefinition(**keywords):
            def creator(**kwargs):
                return self.createHorizontalCommonContourExecution(
                    **kwargs)

        else:
            def creator(**kwargs):
                zValue = kwargs['zValue']
                del kwargs['zValue']  # TODO IT-272: can we do it better? -- NG
                # ImaWOP needs a createPlane before each contour to have the contour on the right plane
                return joinIfNotNone('\n',
                    self.createPlane(plane=plane, **keywords),
                    self.createContourFromSegmentList(
                        createStartBlock=self.createDadoContourStartBlock,
                        createEndBlock=self.createDadoContourEndBlock,
                        zValue=zValue,
                        **kwargs),
                    self.createDefaultPlane(**keywords),
                )

        plane, _positionOnPlane = self.__calculatePlane(**keywords)

        matrix = plane.matrix * dado.matrix.getInverted()
        assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
        dado.matrix = matrix  # TODO (HORIZONTAL) IT-230: unschönes Design -- NG

        return joinIfNotNone('\n',
            self.createPlane(plane=plane, **keywords) if plane and self.useContourDefinition(**keywords) else None,
            self.createCommonDadoWithRouter(
                creator=creator,
                zValueConverter=lambda _boardThickness, currentDepth: -currentDepth,
                **keywords),
            self.createDefaultPlane(**keywords) if plane and self.useContourDefinition(**keywords) else None,
        )

    # Rabbets
    def createNonInsetVerticalRabbet(self, **keywords):
        rabbet = keywords['rabbet']
        start = rabbet.getStartPoint()
        end = rabbet.getEndPoint()
        toolID = rabbet.toolID

        return '\n'.join((
            '[ZYFALZ{}]'.format(self.imawopVersion),
            'AFB=1',
            'EPX={x}'.format(x=self.formatAsCoordinate(data=end.x)),
            'EPY={y}'.format(y=self.formatAsCoordinate(data=end.y)),
            'F=AUTO',
            'FAN=AUTO',
            'KD={kd}'.format(kd=self.formatAsCoordinate(data=-keywords['contourDistance'])),
            'LGEA=2*WKZR',
            'SPX={x}'.format(x=self.formatAsCoordinate(data=start.x)),
            'SPY={y}'.format(y=self.formatAsCoordinate(data=start.y)),
            'SWKZID={}'.format(toolID),
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=keywords['zValue'])),
            'TRKOR={rc}'.format(rc=self.getWopRadiusCompensation(**keywords)),
            'TYPEIN={onset}'.format(onset=keywords['onset']),
            'WERKZEUGNAME={}'.format(toolID),
            'WKZAKTUELL=1',
            'WKZID={}'.format(toolID),
            '',
            '',
        ))

    def createInsetVerticalRabbet(self, **keywords):
        return self.createDefinitionForOneRabbet(**keywords)

    def createVerticalRabbet(self, **keywords):
        rabbet = keywords['rabbet']

        approachType, approachOnset, _useApproachDistance = self.APPROACH_MODE_LINEAR_PIERCING
        keywords['approach'] = approachType
        keywords['departure'] = approachType
        keywords['onset'] = approachOnset

        machiningDepth = rabbet.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        if rabbet.isInset() and self.useContourDefinition(**keywords):
            return self.__createVerticalCommonContourExecute(
                radiusCompensation=(self.getWopRadiusCompensation(**keywords)),
                depthPerStep=depthPerStep,
                **keywords)

        if rabbet.isInset():
            creator = self.createInsetVerticalRabbet
        else:
            creator = self.createNonInsetVerticalRabbet

        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)

        return '\n'.join(self.generateCommonNoBuiltInDepthSteps(creator=creator, machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    # Miters
    def createVerticalMiter(self, **keywords):
        surface = keywords['surface']
        miter = keywords['miter']
        toolID = miter.toolID

        start = miter.getStartPoint()
        end = miter.getEndPoint()

        preSawDepth = self.zValueMiterPreSawConverter.convertWithThickness(surface.thickness, self.MITER_PRE_SAW_DEPTH_FROM_TOP)
        sawDepth = self.zValueMiterSawConverter.convertWithThickness(surface.thickness, surface.thickness + self.MITER_SAW_DEPTH_FROM_BOTTOM)

        return '\n'.join((
            '[ZYSCHN{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={description}'.format(description=self.getLocalizedDescription(23, 'Miter')),
            'EPX={x}'.format(x=self.formatAsCoordinate(data=end.x)),
            'EPY={y}'.format(y=self.formatAsCoordinate(data=end.y)),
            'F=AUTO',
            'FAN=AUTO',
            'GEGENL=0',
            'KD=0',
            'LGEA=WKZR',
            'LGEAN={depth}'.format(depth=self.formatAsCoordinate(data=preSawDepth)),
            'N=AUTO',
            'PKTBEZUG_XY=0',
            'RF=AUTO',
            'SPX={x}'.format(x=self.formatAsCoordinate(data=start.x)),
            'SPY={y}'.format(y=self.formatAsCoordinate(data=start.y)),
            'SWKZID={}'.format(toolID),
            'TI={depth}'.format(depth=self.formatAsCoordinate(data=sawDepth)),
            'TOMBA=3',
            'TRKOR={}'.format(self.RADIUS_COMPENSATION_RIGHT),
            'TWKL={angle}'.format(angle=self.formatAsAngle(data=90 - miter.getAngle())),
            'TYPN=1',
            'WERKZEUGNAME={}'.format(toolID),
            'WKZAKTUELL=1',
            'WKZID={}'.format(toolID),
            '',
            '',
        ))

    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCCW)

    # NC Macros
    def getNCMacroCookedVariables(self, **keywords):
        variables = super().getNCMacroCookedVariables(**keywords)

        variables['version'] = self.imawopVersion

        return variables

    def createNCMacroInclude(self, **keywords):
        ncmacro = keywords['ncmacro']
        variables = tuple(self.getNCMacroCookedParameters(**keywords))

        return self.createVariableDefinition(variables=variables, **keywords) + '\n'.join((
            '[SWINCL{}]'.format(self.imawopVersion),
            'AFB=1',
            'BEZB={description}'.format(description=self.getLocalizedDescription(7, 'Include')),
            'DWKL={angle}'.format(angle=self.formatAsAngle(data=ncmacro.cAngle)),
            'KMI=0',
            'NAME={name}'.format(name=ncmacro.name),
            'PKTBEZUG_XY=0',
            'POSX={x}'.format(x=self.formatAsCoordinate(data=ncmacro.position.x)),
            'POSY={y}'.format(y=self.formatAsCoordinate(data=ncmacro.position.y)),
            '',
            '',
        ))

    def createNCMacroInternal(self, **keywords):
        parameters = keywords['parameters']
        wopMacro = parameters[0]
        parameters = parameters[1:]
        return '\n'.join(itertools.chain(
            (
                '{macro}'.format(macro=wopMacro.name),
            ),
            ('{recordName}={recordValue}'.format(recordName=parameter.name, recordValue=parameter.value) for parameter in parameters),
            (
                '',
                '',
            )
        ))

    def createNCMacroRecordFormat(self, **keywords):
        parameters = tuple(self.getNCMacroCookedParameters(**keywords))
        if parameters and parameters[0].name and parameters[0].name[0] == '[' and parameters[0].name[-1] == ']':
            return self.createNCMacroInternal(parameters=parameters, **keywords)
        else:
            return self.createNCMacroInclude(parameters=parameters, **keywords)

    # Footer
    def createEndOfFile(self, **keywords):
        machineConfiguration = self.getMachineConfiguration()
        parking_position_short = 5
        parking_position_long = 2
        parking_position_switch = machineConfiguration.getAsFloat('parking-position-switch', 1750)

        surface = keywords['surface']
        # if surface.getFinishedSizeX() < parking_position_switch:
        if surface.sawnSize.x < parking_position_switch:
            parking_position = parking_position_short
        else:
            parking_position = parking_position_long

        return '\n'.join((
            self.createSeparatorComment(),
            '[PGENDE{}]'.format(self.imawopVersion),
            'AFB=1',
            'PP={}'.format(parking_position),
            'BEZB=Programm-Ende',
            '',
        ))

    def getLocalizedDescription(self, index, default):
        return getLocalizedString(IMAWOP_DESCRIPTIONS_RESOURCEGROUP, index, default)

    def getLocalizedMachiningName(self, index, default):
        return getLocalizedString(IMAWOP_NAME_MESSAGES, index, default)
