# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import logging
import math
import os
import sys

import interiorcad.backend.openservices
import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.localization import getLocalizedString, MAESTRO_NAME_MESSAGES, XCONVERTER_MESSAGES_RESOURCEGROUP, MAESTRO_PROGRAM_COMMENTS_RESOURCEGROUP, MAESTRO_MULTI_CENTER_STUDIO_PROGRAM_COMMENTS_RESOURCEGROUP, FILE_NAME_POSTFIXES_RESOURCEGROUP
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual, isNearlyLesserOrEqual, isNearlyGreater
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.ncexport.multisawdado import MultiSawDado
from interiorcad.interface.ncexport.nestingexporter import NestingExport
from interiorcad.interface.ncexport.oneprogramexporter import OneProgramExporter
from interiorcad.interface.surface import SURFACE_SIDE_TYPE_OPPOSITE
from interiorcad.interface.surface.acangles import normalizeAngle
from interiorcad.interface.surface.contour2d import Contour2D
from interiorcad.interface.surface.machiningplanes import MachiningPlanes
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.ncexport.base import NCExportBase

# subprocess does not work on Mac under VWi
if sys.platform == "win32":
    import subprocess


XCS_EXTENSION = '.xcs'
PGMX_EXTENSION = '.pgmx'


class MaestroPostProcessing:
    def __init__(self, filePath, machineconfiguration, toolConfigurationFile):
        assert sys.platform == "win32", "XConverter works only under Windows!"
        self.originalFilePath = os.path.abspath(filePath)
        self.xconverterPath = self.__getXConverterPath(self, machineconfiguration)
        self.toolFilePath = self.__getTLGXFilePath(self, toolConfigurationFile)

    @staticmethod
    def __getTLGXFilePath(self, toolConfigurationFile):
        if not toolConfigurationFile:
            logging.info(self.getLocalizedXConverterMessage(9, "The path to the Tool configuration file (*.tlgx) in the configuration is empty, only XCS files are created."))
            return ''

        if not toolConfigurationFile.endswith('.tlgx'):
            toolConfigurationFile += '.tlgx'

        tlgxPath = os.path.join(os.path.dirname(self.xconverterPath), 'Tlgx', toolConfigurationFile)  # If a component is an absolute path, all previous components are thrown away and joining continues
        if not os.path.exists(tlgxPath):
            logging.error(self.getLocalizedXConverterMessage(10, "Could not find Tool configuration file (*.tlgx) according to configuration. Adjust the wrong path '{path}'.").format(path=tlgxPath))

        return tlgxPath

    @staticmethod
    def __getXConverterPath(self, machineconfiguration):
        xconverter = machineconfiguration.get('xconverter')
        if not xconverter:
            logging.info(self.getLocalizedXConverterMessage(7, "The path to XConverter in the configuration is empty, only XCS files are created."))
        elif not os.path.exists(xconverter):
            logging.error(self.getLocalizedXConverterMessage(4, "Could not find XConverter according to configuration. Adjust the wrong path '{path}'.").format(path=xconverter))
            xconverter = ''
        return xconverter

    @staticmethod
    def getPGMXFileNameOutOfXCSFileName(fileName, newSuffix):
        result = fileName
        if fileName.endswith(XCS_EXTENSION):
            result = fileName[:-4] + newSuffix
        return result

    def __callMaestroPostProcessor(self):
        xconverterDirectoryPath = os.path.dirname(self.xconverterPath)
        outputPath = self.getPGMXFileNameOutOfXCSFileName(self.originalFilePath, PGMX_EXTENSION)

        try:
            subprocess.check_call((
                self.xconverterPath,
                '-s',
                '-ow',
                '-m', '0',
                '-i', '{inputPath}'.format(inputPath=self.originalFilePath),
                '-t', self.toolFilePath,
                '-o', '{outputPath}'.format(outputPath=outputPath),
                '-report',
            ))
        except subprocess.CalledProcessError:
            message = self.getLocalizedXConverterMessage(5, "XConverter failed while processing file '{file}'.")
            logging.error(message.format(path=self.xconverterPath, file=self.originalFilePath))
        except OSError:
            message = self.getLocalizedXConverterMessage(6, "Cannot start XConverter at path '{path}'.")
            logging.error(message.format(path=self.xconverterPath))

    @staticmethod
    def __openFolderContaining(filePath):
        interiorcad.backend.openservices.openPath(os.path.dirname(filePath))

    def __checkIfPGMFileExistsOrRaiseException(self):
        pgmFilePath = self.getPGMXFileNameOutOfXCSFileName(self.originalFilePath, PGMX_EXTENSION)
        if not os.path.exists(pgmFilePath):
            self.__openFolderContaining(pgmFilePath)
            logging.error(self.getLocalizedXConverterMessage(1, "XConverter did not generate file '{}'").format(pgmFilePath))

    def __deleteNotNeededFiles(self):
        pgmFilePath = self.getPGMXFileNameOutOfXCSFileName(self.originalFilePath, PGMX_EXTENSION)
        if os.path.exists(pgmFilePath):
            try:
                os.remove(self.originalFilePath)
            except OSError:
                logging.info(self.getLocalizedXConverterMessage(2, "The file '{}' could not be deleted.").format(self.originalFilePath))

    def __CheckLogFile(self):
        if os.path.exists(self.toolFilePath):
            logFileDirectory = os.path.split(os.path.dirname(self.toolFilePath))[0]
        else:
            logFileDirectory = os.path.dirname(self.xconverterPath)

        # TODO: logFileDirectory only exists for version <= 17. version >17 uses C:\Users\Public\Documents\SCM Group\Maestro\Log\…logx.
        logFilePath = os.path.join(
            os.path.dirname(logFileDirectory),
            'Temp',
            'LastImportErrors.log'
        )

        if os.path.exists(logFilePath):
            with open(logFilePath, mode='r', encoding='utf-8-sig', errors='surrogateescape') as logFile:
                logFileLines = logFile.readlines()
                basicMessage = self.getLocalizedXConverterMessage(8, "Feedback of log file from XConverter:")
                for line in logFileLines:
                    logging.warning(basicMessage + ' ' + line.rstrip())

    def process(self):
        if self.xconverterPath:
            self.__callMaestroPostProcessor()
            self.__checkIfPGMFileExistsOrRaiseException()
            self.__deleteNotNeededFiles()
            self.__CheckLogFile()

    def getNewFilePath(self):
        if self.xconverterPath:
            return self.getPGMXFileNameOutOfXCSFileName(self.originalFilePath, PGMX_EXTENSION)
        else:
            return self.originalFilePath

    def getLocalizedXConverterMessage(self, index, default):
        return getLocalizedString(XCONVERTER_MESSAGES_RESOURCEGROUP, index, default)


class OneProgramBottomDadoCensor:
    def __init__(self, oneProgram, machineConfiguration):
        self.oneProgram = oneProgram
        self.machineConfiguration = machineConfiguration

    def process(self, surface):
        if self.oneProgram and surface.sideType == SURFACE_SIDE_TYPE_OPPOSITE:
            for dado in surface.dadosVertical:
                if dado.tool and dado.tool.type == Tool.SAW:
                    if not dado.isAlongXAxis() and not dado.isAlongYAxis():
                        dado.assignRouterToolOnly(surface, self.machineConfiguration)


class MaestroExport(NCExportBase):
    EXPORT_FORMAT = 'MaestroCNC'

    OUTPUT_ENCODING = 'utf-8-sig'

    RADIUS_COMPENSATION_AUTOMATIC = -1
    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    WORKPLANE_TOP = 'Top'
    WORKPLANE_BOTTOM = 'Bottom'
    WORKPLANE_LEFT = 'Left'
    WORKPLANE_RIGHT = 'Right'
    WORKPLANE_FRONT = 'Front'
    WORKPLANE_BACK = 'Back'

    STROKE_STRAIGHT_LINE = 0
    STROKE_LIFT_SHIFT_PLUNGE = 1

    def __init__(self, **keywords):
        super().__init__(**keywords)

        machineConfiguration = self.getMachineConfiguration()
        # INFO: SCM 2023-03-16: "Der einzige wirkliche Schnitt ist im Moment der Sprung über die Version 17, da diese die letzte Version war, die noch nicht .net 4.7.2 vorausgesetzt hat welches sich auf XP nicht mehr installieren lässt."
        self.VERSION = machineConfiguration.getAsInt('maestro-cnc-version', 18)  # 17 is the old version, from 18 newer version (ISP-7: 29 is the newest version from on 2022-09-28)
        self.TOOL_CONFIGURATION_FILE = machineConfiguration.get('tool-configuration-file', 'def')
        self.DADO_SAW_OVERLAP_FACTOR = machineConfiguration.getAsFloat('dado-saw-overlap-factor', 0.15)
        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 5)
        self.ADDITIONAL_DEPTH_FOR_MITER = machineConfiguration.getAsFloat(MachineConfigurationKeys.ADDITIONAL_DEPTH_FOR_MITER_KEY, 15)

        self.MACHINE_PARAMETERS_BLOCKING = machineConfiguration.getAsInt('machine-parameters-blocking', 10)
        self.MACHINE_PARAMETERS_MECHANICAL_OTIONS = machineConfiguration.get('machine-parameters-mechanical-otions', '196608')
        self.MACHINE_PARAMETERS_CONTINUOUS_CYCLE = machineConfiguration.getAsBool('machine-parameters-continuous-cycle', False)
        self.MACHINE_PARAMETERS_TECHNOLOGICAL_MIRROR = machineConfiguration.getAsBool('machine-parameters-technological-mirror', False)
        self.MACHINE_PARAMETERS_AUTO_SETUP = machineConfiguration.getAsBool('machine-parameters-auto-setup', False)
        self.MACHINE_PARAMETERS_PRESERVE_WORKING_SEQUENCE = machineConfiguration.getAsBool('machine-parameters-preserve-working-sequence', False)
        self.MACHINE_PARAMETERS_MULTIPLE_DRILLINGS_CAN_CHANGE_WORKING_SEQUENCE = machineConfiguration.getAsBool('machine-parameters-multiple-drillings-can-change-working-sequence', False)
        self.MACHINE_PARAMETERS_ENABLE_OPTIMIZATION = machineConfiguration.getAsBool('machine-parameters-enable-optimization', False)
        self.MACHINE_PARAMETERS_ENABLE_DYNAMIC_AREAS = machineConfiguration.getAsBool('machine-parameters-enable-dynamic-areas', False)
        self.MACHINE_PARAMETERS_COMMENT = machineConfiguration.get('machine-parameters-comment', '')
        self.MACHINE_PARAMETERS_DESCRIPTION = machineConfiguration.get('machine-parameters-description', '')
        self.CUSTOM_USER_WORKPIECE_NAME = machineConfiguration.get('custom-user-workpiece-name', '{WorkpieceName}')

        self.MACHINING_AREA_X = machineConfiguration.getAsFloat('machining-area-x', 0)
        self.MACHINING_AREA_DEFAULT = machineConfiguration.get('machining-area-default', 'AB')

        self.INDIVIDUAL_HOOD_CONTROL_FUNCTIONS = machineConfiguration.getAsBool('individual-hood-control-functions', False)
        self.SUCTION_CUP_CODE_DEFAULT = machineConfiguration.get('suction-cup-code-default', '')
        self.PARKING_POSITION_DISTANCE = machineConfiguration.getAsInt('parking-position-distance', 1000)
        self.PARKING_POSITION_DISTANCE_FOR_TURN_OVER = machineConfiguration.getAsInt('parking-position-distance-for-turn-over', 1000)
        self.PARKING_POSITION_RELATIVE = machineConfiguration.getAsBool('parking-position-relative', False)
        self.PARKING_POSITION_Y = machineConfiguration.get('parking-position-y', 'finished-size-y')
        self.SECURITY_DISTANCE_FACTOR_FOR_SIZING = machineConfiguration.getAsFloat('security-distance-factor-for-sizing', 1.5)
        self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET = machineConfiguration.getAsFloat('security-distance-factor-for-dado-rabbet', 1.5)
        self.SECURITY_DISTANCE_FACTOR_FOR_CONTOUR = machineConfiguration.getAsFloat('security-distance-factor-for-contour', 1.5)
        self.USE_DYNAMIC_RADIUS_MULTIPLICATOR_FOR_CIRCULAR_POCKET = machineConfiguration.getAsBool('use-dynamic-radius-multiplicator-for-circular-pocket', True)

        self.ONE_PROGRAM = machineConfiguration.getAsBool('one-program', False)
        self.FILE_NAME_POSTFIX_BOTH_SIDED = machineConfiguration.get('file-name-postfix-both-sided', getLocalizedString(FILE_NAME_POSTFIXES_RESOURCEGROUP, 4, 'BothSided'))
        self.CREATE_NC_STOP_FOR_TURN_OVER = machineConfiguration.getAsBool('create-nc-stop-for-turn-over', True)

        self.SIZING_FIRST_PASS_REMAINING_THICKNESS = machineConfiguration.getAsFloat('sizing-first-pass-remaining-thickness', 2)
        self.SIZING_PASS_STRATEGY = machineConfiguration.get('sizing-pass-strategy', 'standard')
        self.SIZING_PASS_STRATEGY_AREA_THRESHOLD = machineConfiguration.getAsFloat('sizing-pass-strategy-area-threshold', 0.1)

        # MultiCenter-Studio (MCS)
        self.NON_RECTANGULAR_WORKPIECE_AS_EXTRUDE = machineConfiguration.getAsBool('non-rectangular-workpiece-as-extrude', False)
        self.MCS_DONT_FORMAT_DURING_NESTING = machineConfiguration.getAsBool('mcs-dont-format-during-nesting', False)

        self.ROTATE_WORKPIECE_180 = machineConfiguration.getAsBool('rotate-workpiece-180', False)

        self.createZValueConverters(**keywords)

        self.checkConfiguration(**keywords)

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMiterPreSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_PRE_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)

    def checkConfiguration(self, **_keywords):
        if self.MCS_DONT_FORMAT_DURING_NESTING:
            if self.configuration.separateSizing:
                logging.error(self.getLocalizedMultiCenterStudioMessage(2, "Separate sizing is not supported in MultiCenter-Studio"))
            if self.ONE_PROGRAM:
                logging.error(self.getLocalizedMultiCenterStudioMessage(3, "'Export all machinings in a single NC program...' is not supported in MultiCenter-Studio"))
            if not self.configuration.accurateSaw:
                logging.error(self.getLocalizedMultiCenterStudioMessage(4, "'No Sizing for Rectangular Parts' should be set for MultiCenter-Studio"))

    def getNCFileExtension(self, **_keywords):
        return XCS_EXTENSION

    def getCSharpBoolean(self, **keywords):
        if keywords['data']:
            return 'true'
        else:
            return 'false'

    def postProcessFile(self, **keywords):
        filePath = keywords['filePath']
        newFilePath = filePath
        toolConfigurationFile = self.TOOL_CONFIGURATION_FILE
        if sys.platform == "win32":
            postProcessor = MaestroPostProcessing(filePath, self.getMachineConfiguration(), toolConfigurationFile)
            postProcessor.process()
            newFilePath = postProcessor.getNewFilePath()
        return newFilePath

    def getSurfacesExporter(self, **keywords):
        if self.MCS_DONT_FORMAT_DURING_NESTING:  # does not support two sides in one program
            return NestingExport(**keywords)
        elif self.ONE_PROGRAM:
            return OneProgramExporter(**keywords)
        else:
            return super().getSurfacesExporter(**keywords)

    def createSide(self, **keywords):
        if self.ROTATE_WORKPIECE_180:
            surface = keywords['surface']
            surface.rotate180()

        return super().createSide(**keywords)

    def createMainAndOppositeSidesTogether(self, mainSurface, oppositeSurface, **keywords):
        if self.ROTATE_WORKPIECE_180:
            mainSurface.rotate180()

        self.orientMachinings(surface=mainSurface, **keywords)
        yield self.createHead(surface=mainSurface, **keywords)
        yield self.createMachinings(surface=mainSurface, **keywords)

        oppositeSurface.mirrorOnYAxis()  # We must mirror back (again)
        oppositeSurface.mirrorOnXAxis()

        if self.ROTATE_WORKPIECE_180:
            oppositeSurface.rotate180()

        self.orientMachinings(surface=oppositeSurface, **keywords)
        yield self.createTurnOver(surface=oppositeSurface, **keywords)
        yield self.createMachinings(surface=oppositeSurface, **keywords)
        yield self.createFoot(surface=oppositeSurface, **keywords)

    def createParkingPosition(self, **keywords):  # This function calls the Null operation (XN) in Maestro, not the Machine parking P
        surface = keywords['surface']
        parkingPositionDistance = keywords['parkingPositionDistance']

        if parkingPositionDistance > 0:
            if self.MACHINING_AREA_X and self.MACHINING_AREA_X < keywords['surface'].getFinishedSizeX() + parkingPositionDistance:
                distanceX = self.MACHINING_AREA_X
            else:
                distanceX = surface.getFinishedSizeX() + parkingPositionDistance
            distanceX = self.formatAsCoordinate(data=distanceX)
        else:  # < 0
            distanceX = parkingPositionDistance

        if self.PARKING_POSITION_Y == 'zero':
            distanceY = 0
        else:  # 'finished-size-y'
            distanceY = surface.getFinishedSizeY()
        # Could be extended for working-area-y if requested.
        distanceY = self.formatAsCoordinate(data=distanceY)

        return 'CreateNullOperation(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name'] + ' X=' + str(distanceX) + ' Y=' + str(distanceY)),
            '{x}'.format(x=distanceX),
            '{y}'.format(y=distanceY),
            '{q}'.format(q=int(self.PARKING_POSITION_RELATIVE)),
        ) + ');'

    def createTurnOver(self, **keywords):
        surface = keywords['surface']
        comment = self.getLocalizedComment(1, "Please turn over part!")
        name = self.getLocalizedMachiningName(18, "Null operation Turnover")
        self.mainWorkplane = self.WORKPLANE_BOTTOM

        return joinIfNotNone('\n',
            self.createParkingPosition(name=name, parkingPositionDistance=self.PARKING_POSITION_DISTANCE_FOR_TURN_OVER, **keywords) if self.PARKING_POSITION_DISTANCE_FOR_TURN_OVER else None,
            self.createNCStop(comment=comment, releaseWorkpiece=True) if self.CREATE_NC_STOP_FOR_TURN_OVER else None,

            'CreateWorkplan("{name}");'.format(name=self.getLocalizedMachiningName(17, "Opposite Side")),
            self.setWorkpieceSetupPosition(additionX=surface.getAdditionX(), additionY=surface.getAdditionYMirrored(), **keywords),

            self.selectWorkplane(workplane=self.WORKPLANE_BOTTOM, **keywords),

            self.createSizingContourDefinition(**keywords),
            self.createRabbetContourDefinitions(**keywords),
            self.createDadoContourDefinitions(dados=surface.getAllDados(), **keywords),
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords),
        )

    # Header
    def createHead(self, **keywords):
        surface = keywords['surface']

        self.lastDrillingIndex = 0
        self.lastPocketIndex = 0
        self.lastSizingIndex = 0
        self.lastRabbetIndex = 0
        self.lastContourIndex = 0
        self.lastDadoIndex = 0
        self.lastMiterIndex = 0
        self.lastGeometryIndex = 0
        self.lastNCStopIndex = 0
        self.lastNCMacroIndex = 0
        self.lastWorkplaneIndex = 0
        self.mainWorkplane = self.WORKPLANE_TOP

        # TODO IT-261: we should also collect every create planes, not to have to create a plane per machining
        self.predefinedMachiningPlanes = MachiningPlanes(sideIDGenerator=lambda: None)
        self.predefinedMachiningPlanes.planes.extend(self.__createPredefinedPlanes(**keywords))

        return joinIfNotNone('\n',
            self.setMachineParameters(**keywords),
            self.createFinishedAndRawWorkpiece(**keywords),
            self.setWorkpieceSetupPosition(additionX=surface.getAdditionX(), additionY=surface.getAdditionY(), **keywords),
            self.setHoodControlFunctions(**keywords),
            self.setSuctionCups(**keywords),
            self.createSizingContourDefinition(**keywords),
            self.createRabbetContourDefinitions(**keywords),
            self.createDadoContourDefinitions(dados=surface.getAllDados(), **keywords),
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords),
            '',
        )

    def getMachiningArea(self, **_keywords):
        return self.MACHINING_AREA_DEFAULT  # ### MAESTRO: needs to be an option handled similar to Xilog -- NG

    def setMachiningParameters(self, **keywords):
        return 'SetMachiningParameters(' + joinIfNotNone(', ',
            '"{executionFields}"'.format(executionFields=self.getMachiningArea(**keywords)),
            '{repetitions}'.format(repetitions=keywords['surface'].count),
            '{tableOptions}'.format(tableOptions=self.MACHINE_PARAMETERS_BLOCKING),
            '{mechanicalOptions}'.format(mechanicalOptions=self.MACHINE_PARAMETERS_MECHANICAL_OTIONS),
            '{continuousCycle}'.format(continuousCycle=self.getCSharpBoolean(data=self.MACHINE_PARAMETERS_CONTINUOUS_CYCLE)),
        ) + ');'

    def setMachineParameterTechnologicalMirror(self, **_keywords):
        return 'SetTechnologicalMirror(true);'

    def setMachineParameterAutoSetup(self, **_keywords):
        return 'SetAutoSetup(true);'

    def setMachineParameterPreserveWorkingSequence(self, **_keywords):
        return 'SetPreserveWorkingSequence(true);'

    def setMachineParameterMultipleDrillingsCanChangeWorkingSequence(self, **_keywords):
        return 'SetMultipleDrillingsCanChangeWorkingSequence(true);'

    def setMachineParameterEnableOptimization(self, **_keywords):
        return 'SetOptimization(true);'

    def setMachineParameterEnableDynamicAreas(self, **_keywords):
        return 'SetEnableDynamicAreas(true);'

    def setMachineParameterComment(self, **keywords):
        return 'SetComment("{comment}");'.format(comment=self.getMappedCustomUserVariablesFromStringWithKeys(stringWithKeys=self.MACHINE_PARAMETERS_COMMENT, **keywords))

    def setMachineParameterDescription(self, **keywords):
        return 'SetDescription("{description}");'.format(description=self.getMappedCustomUserVariablesFromStringWithKeys(stringWithKeys=self.MACHINE_PARAMETERS_DESCRIPTION, **keywords))

    def setMachineParameters(self, **keywords):
        return joinIfNotNone('\n',
            self.setMachiningParameters(**keywords),
            self.setMachineParameterTechnologicalMirror(**keywords) if self.MACHINE_PARAMETERS_TECHNOLOGICAL_MIRROR else None,
            self.setMachineParameterAutoSetup(**keywords) if self.MACHINE_PARAMETERS_AUTO_SETUP else None,
            self.setMachineParameterPreserveWorkingSequence(**keywords) if self.MACHINE_PARAMETERS_PRESERVE_WORKING_SEQUENCE else None,
            self.setMachineParameterMultipleDrillingsCanChangeWorkingSequence(**keywords) if self.MACHINE_PARAMETERS_MULTIPLE_DRILLINGS_CAN_CHANGE_WORKING_SEQUENCE else None,
            self.setMachineParameterEnableOptimization(**keywords) if self.MACHINE_PARAMETERS_ENABLE_OPTIMIZATION else None,
            self.setMachineParameterEnableDynamicAreas(**keywords) if self.MACHINE_PARAMETERS_ENABLE_DYNAMIC_AREAS else None,
            self.setMachineParameterComment(**keywords) if self.MACHINE_PARAMETERS_COMMENT else None,
            self.setMachineParameterDescription(**keywords) if self.MACHINE_PARAMETERS_DESCRIPTION else None,
            '',
        )

    def createFinishedAndRawWorkpiece(self, **keywords):
        surface = keywords['surface']
        if self.MCS_DONT_FORMAT_DURING_NESTING and not self.NON_RECTANGULAR_WORKPIECE_AS_EXTRUDE and surface.sizing and not surface.sizing.isRectangular():
            message = self.getLocalizedMultiCenterStudioMessage(1, "Sizing of non-rectangular parts in MultiCenter-Studio must use 'Export non-rectangular parts as piece with extrusion' of the machine-configuration. Workpiece: {workpiece}").format(workpiece=surface.workpieceName)
            raise RuntimeError(message)
        elif self.NON_RECTANGULAR_WORKPIECE_AS_EXTRUDE and surface.sizing and not surface.sizing.isRectangular():
            return joinIfNotNone('\n',
                self.createSizingDefinitionForOnePolyline(polyline=surface.finishedSizeSizing.polyline, **keywords),
                self.createFinishedWorkpieceFromExtrusion(**keywords),
                '',
                self.createSizingDefinitionForOnePolyline(polyline=surface.finishedSizeSizing.polyline, **keywords),
                self.createRawWorkpieceFromExtrusion(**keywords),
            )
        else:
            return joinIfNotNone('\n',
                self.createFinishedWorkpieceBox(**keywords),
                self.createRawWorkpiece(**keywords),
            )

    def getCustomUserWorkpieceName(self, **keywords):
        return self.getMappedCustomUserVariablesFromStringWithKeys(stringWithKeys=self.CUSTOM_USER_WORKPIECE_NAME, **keywords)

    def createFinishedWorkpieceBox(self, **keywords):
        surface = keywords['surface']
        return 'CreateFinishedWorkpieceBox(' + joinIfNotNone(', ',
            '"{name}"'.format(name=self.getCustomUserWorkpieceName(**keywords)),
            '{dx}'.format(dx=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            '{dy}'.format(dy=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            '{dz}'.format(dz=self.formatAsCoordinate(data=surface.thickness)),
        ) + ');'

    def createFinishedWorkpieceFromExtrusion(self, **keywords):
        surface = keywords['surface']
        return 'CreateFinishedWorkpieceFromExtrusion(' + joinIfNotNone(', ',
            '"{name}"'.format(name=self.getCustomUserWorkpieceName(**keywords)),
            '{dz}'.format(dz=self.formatAsCoordinate(data=surface.thickness)),
        ) + ');'

    def createRawWorkpiece(self, **keywords):
        surface = keywords['surface']
        return 'CreateRawWorkpiece(' + joinIfNotNone(', ',
            '"{name}"'.format(name=self.getCustomUserWorkpieceName(**keywords)),
            '{bx1}'.format(bx1=self.formatAsCoordinate(data=surface.getAdditionX())),
            '{bx2}'.format(bx2=self.formatAsCoordinate(data=surface.getAdditionXMirrored())),
            '{by1}'.format(by1=self.formatAsCoordinate(data=surface.getAdditionY())),
            '{by2}'.format(by2=self.formatAsCoordinate(data=surface.getAdditionYMirrored())),
            '{bz1}'.format(bz1=self.formatAsCoordinate(data=0)),
            '{bz2}'.format(bz2=self.formatAsCoordinate(data=0)),
        ) + ');'

    def createRawWorkpieceFromExtrusion(self, **keywords):
        surface = keywords['surface']
        return 'CreateRawWorkpieceFromExtrusion(' + joinIfNotNone(', ',
            '"{name}"'.format(name=self.getCustomUserWorkpieceName(**keywords)),
            '{dz}'.format(dz=self.formatAsCoordinate(data=surface.thickness)),
        ) + ');'

    def setWorkpieceSetupPosition(self, **keywords):
        return 'SetWorkpieceSetupPosition(' + joinIfNotNone(', ',
            '{x}'.format(x=self.formatAsCoordinate(data=keywords['additionX'])),
            '{y}'.format(y=self.formatAsCoordinate(data=keywords['additionY'])),
            '{z}'.format(z=self.formatAsCoordinate(data=0)),
            '{zRot}'.format(zRot=self.formatAsAngle(data=0)),
        ) + ');'

    def setHoodControlFunctions(self, **keywords):
        if self.INDIVIDUAL_HOOD_CONTROL_FUNCTIONS:
            comment = '// Use the individual hood position methods with care. The methods are used only once after the program header without any further influence on other functions.'
            return self.getMappedCustomUserVariablesFromFile(
                customUserVariableFileName='Maestro Hood Position Methods.txt',
                customUserVariableFileDefaultText=comment + '\n' + self.defaultHoodControlFunctions(**keywords),
                **keywords)
        else:
            return self.defaultHoodControlFunctions(**keywords)

    def defaultHoodControlFunctions(self, **_keywords):
        # Both set to automatic for the whole NC program.
        return '\n'.join((
            'ResetPneumaticHood();',
            'ResetAuxiliaryHood();',
        ))

    def setSuctionCups(self, **_keywords):
        # Currently positions only one bar with a suction cup manually without any collision control.
        if self.SUCTION_CUP_CODE_DEFAULT:
            return '\n'.join((
                'SetBarPosition(1,150);',
                'SetSuctionCupPosition(1,150,0, "{suctionCupCode}" );'.format(suctionCupCode=self.SUCTION_CUP_CODE_DEFAULT),
            ))

    # Machinings
    AUTOMATIC = -1

    FLAT_DRILL_TYPE = 'P'
    PIERCING_DRILL_TYPE = 'L'

    DRILLING_PROCESS_TYPE = 'TypeOfProcess.Drilling'
    GENERAL_ROUTING_PROCESS_TYPE = 'TypeOfProcess.GeneralRouting'

    def getPostAssignToolProcessors(self, **keywords):
        return OneProgramBottomDadoCensor(oneProgram=self.ONE_PROGRAM, machineConfiguration=self.getMachineConfiguration(**keywords)),

    def selectWorkplane(self, **keywords):
        return 'SelectWorkplane(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['workplane']),
        ) + ');'

    def getNewWorkplaneName(self, **_keywords):
        self.lastWorkplaneIndex += 1
        return self.getLocalizedMachiningName(20, "Workplane {index}").format(index=self.lastWorkplaneIndex)

    def getNewGeometryName(self, **_keywords):
        self.lastGeometryIndex += 1
        return self.getLocalizedMachiningName(1, "Geometry {index}").format(index=self.lastGeometryIndex)

    def getToolHead(self, **keywords):
        machining = keywords['machining']
        if hasattr(machining, 'tool') and machining.tool:
            return machining.tool.getPropertyAsString('work-head-number', self.AUTOMATIC)
        else:
            return self.AUTOMATIC

    def createDrill(self, **keywords):
        drilling = keywords['drilling']
        toolID = drilling.toolID
        return 'CreateDrill(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
            '{x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            '{y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            '{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'] if 'depth' in keywords else drilling.depth)),  # parameter 'passing' isn't documented, therefore we use keywords['depth'] instead of drilling.depth
            '{diameter}'.format(diameter=self.formatAsCoordinate(data=drilling.installedDiameter)),
            '"{description}"'.format(description=''),
            '{typeOfProcess}'.format(typeOfProcess=self.DRILLING_PROCESS_TYPE),
            '"{tool}"'.format(tool=self.AUTOMATIC if not toolID else toolID),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            '{dischargeSteps}'.format(dischargeSteps=0),
            '{rotSpeed}'.format(rotSpeed=self.AUTOMATIC),
            '{boringSpeed}'.format(boringSpeed=self.AUTOMATIC),
            '"{kindOfHole}"'.format(kindOfHole=self.getDrillType(defaultDrillType=self.AUTOMATIC, **keywords)),  # the parameter is documented as "double" but the demos uses string
            '{taperHeight}'.format(taperHeight=self.formatAsCoordinate(data=0)),
            'null',  # null to use Maestro CNC -> Options -> Parameters -> Machinings -> "Safety distance from machining plane"
        ) + ');'

    def createSlantedDrill(self, **keywords):
        drilling = keywords['drilling']
        return 'CreateSlantedDrill(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
            '{x}'.format(x=self.formatAsCoordinate(data=drilling.position.x)),
            '{y}'.format(y=self.formatAsCoordinate(data=drilling.position.y)),
            '{z}'.format(z=self.formatAsCoordinate(data=drilling.zPosition)),
            '{angleA}'.format(angleA=self.formatAsAngle(data=normalizeAngle(drilling.cAngle - 90))),
            '{angleB}'.format(angleB=self.formatAsAngle(data=drilling.aAngle)),
            '{depth}'.format(depth=self.formatAsCoordinate(data=drilling.depth)),
            '{diameter}'.format(diameter=self.formatAsCoordinate(data=drilling.installedDiameter)),
            '"{description}"'.format(description=''),
            '{typeOfProcess}'.format(typeOfProcess=self.DRILLING_PROCESS_TYPE),
            '"{tool}"'.format(tool=drilling.toolID),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            '{dischargeSteps}'.format(dischargeSteps=0),
            '{rotSpeed}'.format(rotSpeed=self.AUTOMATIC),
            '{boringSpeed}'.format(boringSpeed=self.AUTOMATIC),
            '"{kindOfHole}"'.format(kindOfHole=self.getDrillType(defaultDrillType=self.AUTOMATIC, **keywords)),  # the parameter is documented as "double" but the demos uses string
            '{taperHeight}'.format(taperHeight=self.formatAsCoordinate(data=0)),
            'null',  # null to use Maestro CNC -> Options -> Parameters -> Machinings -> "Safety distance from machining plane"
        ) + ');'

    def createCircleCenterRadius(self, **keywords):
        pocket = keywords['pocket']
        return 'CreateCircleCenterRadius(' + joinIfNotNone(', ',
            '"{name}"'.format(name=self.getNewGeometryName(**keywords)),
            '{centerX}'.format(centerX=self.formatAsCoordinate(data=pocket.position.x)),
            '{centerY}'.format(centerY=self.formatAsCoordinate(data=pocket.position.y)),
            '{radius}'.format(radius=self.formatAsCoordinate(data=pocket.diameter / 2)),
            '{isClockwise}'.format(isClockwise=self.getCSharpBoolean(data=keywords['cw'])),
        ) + ');'

    def createContourParallelStrategy(self, **keywords):
        return 'CreateContourParallelStrategy(' + joinIfNotNone(', ',
            '{insideToOutSide}'.format(insideToOutSide=self.getCSharpBoolean(data=True)),
            '{rotationDirection:d}'.format(rotationDirection=not keywords['directionCW']),
            '{allowMultiplePasses}'.format(allowMultiplePasses=self.getCSharpBoolean(data=True)),
            '{cuttingDepth}'.format(cuttingDepth=self.formatAsCoordinate(data=keywords['depthPerStep'])),
            '{finishCuttingDepth}'.format(finishCuttingDepth=self.formatAsCoordinate(data=keywords['endDepth'])),
            '{strokeConnectionType}'.format(strokeConnectionType=self.formatAsInt(data=self.STROKE_STRAIGHT_LINE)),
            '{isHelicStrategy}'.format(isHelicStrategy=self.getCSharpBoolean(data=True)),
            '{isApproachRetractPassStrategy}'.format(isApproachRetractPassStrategy=self.getCSharpBoolean(data=False)) if self.VERSION > 17 else None,
            '{allowanceSide}'.format(allowanceSide=self.formatAsCoordinate(data=keywords['contourDistance']) if 'contourDistance' in keywords else 0) if self.VERSION > 17 else None,
        ) + ');'

    def createContourPocket(self, **keywords):
        pocket = keywords['pocket']
        return 'CreateContourPocket(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
            '{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '"{description}"'.format(description=''),
            '{typeOfProcess}'.format(typeOfProcess=self.GENERAL_ROUTING_PROCESS_TYPE),  # ### MAESTRO: or TypeOfProcess.ConcentricalPocket like documented? -- NG
            '"{tool}"'.format(tool=pocket.toolID),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            '{inputSpeed}'.format(inputSpeed=-1),
            '{rotSpeed}'.format(rotSpeed=-1),
            '{speed}'.format(speed=-1),
            '{overlap}'.format(overlap=50),
            '{finalPass}'.format(finalPass=self.getCSharpBoolean(data=True)),
        ) + ');'

    def __createVerticalDrillingCommon(self, **keywords):
        drilling = keywords['drilling']
        return self.createDrill(x=drilling.position.x, y=drilling.position.y, **keywords) + '\n'

    def createVerticalDrilling(self, **keywords):
        self.lastDrillingIndex += 1
        name = self.getLocalizedMachiningName(10, "Vertical Drilling {index}").format(index=self.lastDrillingIndex)
        return self.__createVerticalDrillingCommon(name=name, **keywords)

    def createVerticalDrillingPiercing(self, **keywords):
        self.lastDrillingIndex += 1
        name = self.getLocalizedMachiningName(11, "Vertical Piercing Drilling {index}").format(index=self.lastDrillingIndex)
        return self.__createVerticalDrillingCommon(name=name, **keywords)

    def __getDepthPerStepData(self, **keywords):
        if 'depth' in keywords:  # TODO IT-271 HACK: can we remove this hack -- NG
            del keywords['depth']

        machiningDepth = keywords['machining'].actualDepth

        depthPerStep = keywords['depthPerStep']
        useSteps = (machiningDepth > depthPerStep > 0.0) and not isNearlyEqual(depthPerStep, 0.0) and not isNearlyEqual(machiningDepth, depthPerStep)
        if useSteps:
            numberOfSteps = math.ceil(machiningDepth / depthPerStep)
            rawDepthStep = machiningDepth / numberOfSteps
            assert 0 < rawDepthStep <= depthPerStep or isNearlyEqual(rawDepthStep, depthPerStep), "Calculated step should not be 0 and not be bigger than maximum step!"
            return self.getContourZValue(depth=rawDepthStep, **keywords)

    def __createPocketCommon(self, **keywords):
        pocket = keywords['pocket']

        if self.USE_DYNAMIC_RADIUS_MULTIPLICATOR_FOR_CIRCULAR_POCKET:
            pocketDiameter = pocket.diameter
            routerDiameter = pocket.tool.getRouterDiameter()
            assert isNearlyGreater(routerDiameter, 0), "The pocket tool '{toolName}' must have a diameter greater than 0.".format(toolName=pocket.tool.name)

            approachDistance = 1 + ((pocketDiameter - routerDiameter) / 2) / routerDiameter
            if approachDistance > 2:
                approachDistance = 2
        else:
            approachDistance = 0

        depthPerStep = self.__getDepthPerStepData(depthPerStep=self.getRouterDepthPerStep(**keywords), **keywords)
        if not depthPerStep:
            depthPerStep = pocket.actualDepth

        endDepth = 0  # We must set endDepth to 0 or the helix motion will be done directly to the end depth instead of being progressively. (Not documented that way)

        cw = self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)

        return joinIfNotNone('\n',
            self.resetApproachStrategyApproach(**keywords),
            self.resetRetractStrategyDeparture(**keywords),
            self.setApproachStrategyApproach(approach=False, onset=False, distance=approachDistance, **keywords),
            self.createCircleCenterRadius(cw=cw, **keywords),
            # self.setRetractStrategyDeparture(approach=False, onset=False, distance=approachDistance, **keywords),  # Actually a technical problem in Maestro CNC --mjk
            keywords['pocketStrategy'](depthPerStep=depthPerStep, endDepth=endDepth, **keywords),
            keywords['pocketMiller'](**keywords),
            '',
            '',
        )

    def __createCommonClearOutPocket(self, **keywords):
        return self.__createPocketCommon(
            pocketMiller=self.createContourPocket,
            pocketStrategy=self.createContourParallelStrategy,
            directionCW=self.isDadoRabbetReplacementCW(keywords['pocket'], self.POCKET_DIRECTION),
            **keywords)

    def __createVerticalCircularPocket(self, **keywords):
        cw = self.isDadoRabbetReplacementCW(keywords['pocket'], self.POCKET_DIRECTION)
        return self.__createPocketCommon(
            pocketMiller=self.createRoughFinish,
            pocketStrategy=self.createHelicMillingStrategy,
            radiusCompensation=self.RADIUS_COMPENSATION_RIGHT if cw else self.RADIUS_COMPENSATION_LEFT,
            **keywords)

    def createVerticalPocket(self, **keywords):
        self.lastPocketIndex += 1
        name = self.getLocalizedMachiningName(8, "Pocket {index}").format(index=self.lastPocketIndex)
        return self.__createCommonClearOutPocket(name=name, **keywords)

    def createVerticalPocketPiercing(self, **keywords):
        self.lastPocketIndex += 1
        name = self.getLocalizedMachiningName(9, "Piercing Pocket {index}").format(index=self.lastPocketIndex)
        if keywords['pocket'].diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER:
            return self.__createVerticalCircularPocket(name=name, **keywords)
        else:
            return self.__createCommonClearOutPocket(name=name, **keywords)

    def createWorkplane(self, **keywords):
        plane = keywords['plane']
        assert not plane.predefined, "A predefined plane should not be created explicitely!"

        point1 = plane.origin
        point2 = plane.origin + plane.matrix.u
        point3 = plane.origin + plane.matrix.v

        name = plane.sideID
        assert name, "The plane name may not be empty."

        return (
            'CreateWorkplane(' +
            ','.join((
                '"{name}"'.format(name=name),
                self.formatAsCoordinate(data=point1.x),
                self.formatAsCoordinate(data=point1.y),
                self.formatAsCoordinate(data=point1.z),
                self.formatAsCoordinate(data=point2.x),
                self.formatAsCoordinate(data=point2.y),
                self.formatAsCoordinate(data=point2.z),
                self.formatAsCoordinate(data=point3.x),
                self.formatAsCoordinate(data=point3.y),
                self.formatAsCoordinate(data=point3.z),
            )) + ');'
        )

    def __createPredefinedPlanes(self, **keywords):
        size = keywords['surface'].getFinishedSize()

        yield Plane(
            predefined=True, sideID=self.WORKPLANE_FRONT,
            matrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.WORKPLANE_RIGHT,
            matrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.WORKPLANE_BACK,
            matrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(size.x, size.y, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.WORKPLANE_LEFT,
            matrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
        )

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']

        machiningPlane = Plane.createFromMachining(machining)

        plane = self.predefinedMachiningPlanes.findPlane(machiningPlane)
        if plane:
            machiningPlane = plane

        return machiningPlane, machining.getPositionOnPlane(machiningPlane)

    def createHorizontalPocket(self, **keywords):
        pocket = keywords['pocket']

        self.lastPocketIndex += 1
        name = self.getLocalizedMachiningName(21, "Horizontal Pocket {index}").format(index=self.lastPocketIndex)

        plane, positionOnPlane = self.__calculatePlane(**keywords)
        pocket.matrix.origin = Vector(positionOnPlane.x, positionOnPlane.y, 0.0)  # TODO IT-222 (HORIZONTAL): unschönes Design -- NG

        if plane and not plane.predefined:
            plane.sideID = self.getNewWorkplaneName(**keywords)

        return joinIfNotNone('\n',
            self.createWorkplane(plane=plane, **keywords) if plane and not plane.predefined else None,
            self.selectWorkplane(workplane=plane.sideID, **keywords) if plane and plane.predefined else None,
            self.__createCommonClearOutPocket(name=name, depth=pocket.depth, **keywords),
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            ''
            '',
        )

    def createHorizontalDrillingOnBounds(self, **keywords):
        self.lastDrillingIndex += 1
        name = self.getLocalizedMachiningName(12, "Horizontal Drilling {index}").format(index=self.lastDrillingIndex)
        return '\n'.join((
            self.selectWorkplane(workplane=keywords['horizontalPlane'], **keywords),
            self.createDrill(name=name, **keywords),
            self.selectWorkplane(workplane=self.WORKPLANE_TOP, **keywords),
            '',
        ))

    def createPattern(self, **keywords):
        return 'CreatePattern(' + joinIfNotNone(', ',
            '{numberOfRows}'.format(numberOfRows=self.formatAsInt(data=keywords['numberOfRows'])),
            '{numberOfColumns}'.format(numberOfColumns=self.formatAsInt(data=keywords['numberOfColumns'])),
            '{rowsDistance}'.format(rowsDistance=self.formatAsCoordinate(data=keywords['rowsDistance'])),
            '{columnDistance}'.format(columnDistance=self.formatAsCoordinate(data=keywords['columnDistance'])),
            '{rotationAngle}'.format(rotationAngle=self.formatAsAngle(data=keywords['rotationAngle'])),
            '{rowLayoutAngle}'.format(rowLayoutAngle=self.formatAsAngle(data=90)),
        ) + ');'

    def resetPattern(self):
        return 'ResetPattern();'

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']

        self.lastDrillingIndex += 1
        name = self.getLocalizedMachiningName(13, "Vertical Grid Drilling along X Axis {index}").format(index=self.lastDrillingIndex)
        rotationAngle = normalizeAngle(gridDrilling.getXYAngle())

        return '\n'.join((
            self.createPattern(numberOfRows=1, numberOfColumns=gridDrilling.countDrillings, rowsDistance=0, columnDistance=gridDrilling.gridDistance, rotationAngle=rotationAngle, **keywords),
            self.__createVerticalDrillingCommon(drilling=keywords['gridDrilling'].getStartDrilling(), name=name, **keywords),
            self.resetPattern(),
            '',
        ))

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']

        self.lastDrillingIndex += 1
        name = self.getLocalizedMachiningName(14, "Vertical Grid Drilling along Y Axis {index}").format(index=self.lastDrillingIndex)
        rotationAngle = normalizeAngle(gridDrilling.getXYAngle() - 90)  # different from GridDrillingAlongXAxis, because the native GridDrillingY in Maestro CNC is already rotated 90°.

        return '\n'.join((
            self.createPattern(numberOfRows=gridDrilling.countDrillings, numberOfColumns=1, rowsDistance=gridDrilling.gridDistance, columnDistance=0, rotationAngle=rotationAngle, **keywords),
            self.__createVerticalDrillingCommon(drilling=keywords['gridDrilling'].getStartDrilling(), name=name, **keywords),
            self.resetPattern(),
            '',
        ))

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']

        self.lastDrillingIndex += 1
        name = self.getLocalizedMachiningName(23, "Vertical Grid Drilling in X/Y plane {index}").format(index=self.lastDrillingIndex)
        rotationAngle = normalizeAngle(gridDrilling.getXYAngle())

        return '\n'.join((
            self.createPattern(numberOfRows=1, numberOfColumns=gridDrilling.countDrillings, rowsDistance=1, columnDistance=gridDrilling.gridDistance, rotationAngle=rotationAngle, **keywords),
            self.__createVerticalDrillingCommon(drilling=keywords['gridDrilling'].getStartDrilling(), name=name, **keywords),
            self.resetPattern(),
            '',
        ))

    def createRotatedHorizontalDrilling(self, **keywords):
        plane, positionOnPlane = self.__calculatePlane(**keywords)
        if plane.predefined:
            return self.createHorizontalDrillingOnBounds(horizontalPlane=plane.sideID, x=positionOnPlane.x, y=positionOnPlane.y, **keywords)
        else:
            self.lastDrillingIndex += 1
            name = self.getLocalizedMachiningName(15, "C rotated Horizontal Drilling {index}").format(index=self.lastDrillingIndex)

            return self.createSlantedDrill(name=name, **keywords) + '\n'

    def createFiveAxisDrilling(self, **keywords):
        self.lastDrillingIndex += 1
        name = self.getLocalizedMachiningName(16, "Five Axis Drilling {index}").format(index=self.lastDrillingIndex)

        return self.createSlantedDrill(name=name, **keywords) + '\n'

    # contours
    APPROACH_DEPARTURE_LINEAR = True
    APPROACH_DEPARTURE_ARC = False

    ONSET_OUTSET_PIERCING = True
    ONSET_OUTSET_FLYING = False

    APPROACH_MODE_LINEAR_PIERCING = (APPROACH_DEPARTURE_LINEAR, ONSET_OUTSET_PIERCING)
    APPROACH_MODE_LINEAR_FLYING = (APPROACH_DEPARTURE_LINEAR, ONSET_OUTSET_FLYING)
    APPROACH_MODE_ARC_PIERCING = (APPROACH_DEPARTURE_ARC, ONSET_OUTSET_PIERCING)
    APPROACH_MODE_ARC_FLYING = (APPROACH_DEPARTURE_ARC, ONSET_OUTSET_FLYING)
    APPROACH_MODE_WITHOUT = (None, None)
    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_LINEAR_FLYING

    def resetApproachStrategyApproach(self, **_keywords):
        return 'ResetApproachStrategy();'

    def resetRetractStrategyDeparture(self, **_keywords):
        return 'ResetRetractStrategy();'

    def setApproachStrategyApproach(self, **keywords):
        return 'SetApproachStrategy(' + joinIfNotNone(', ',
            '{isLinear}'.format(isLinear=self.getCSharpBoolean(data=keywords['approach'])),
            '{isQuote}'.format(isQuote=self.getCSharpBoolean(data=keywords['onset'])),
            '{distance}'.format(distance=self.formatAsCoordinate(data=keywords['distance'])),
        ) + ');'

    def setRetractStrategyDeparture(self, **keywords):
        return 'SetRetractStrategy(' + joinIfNotNone(', ',
            '{isLinear}'.format(isLinear=self.getCSharpBoolean(data=keywords['approach'])),
            '{isQuote}'.format(isQuote=self.getCSharpBoolean(data=keywords['onset'])),
            '{distance}'.format(distance=self.formatAsCoordinate(data=keywords['distance'])),
            '{overlapLength}'.format(overlapLength=self.formatAsCoordinate(data=0)),
        ) + ');'

    def setCompensationMode(self, **keywords):  # Necessary for rabbets with offsets and MultiUsageContours, but not yet documented.  --mjk
        return 'SetCompensationMode({mode});'.format(mode=self.getCSharpBoolean(data=False)) if 'contourDistance' in keywords and keywords['contourDistance'] else None

    def createHelicMillingStrategy(self, **keywords):
        return 'CreateHelicMillingStrategy(' + joinIfNotNone(', ',
            '{cuttingDepth}'.format(cuttingDepth=self.formatAsCoordinate(data=keywords['depthPerStep'])),
            '{allowsFinishCutting}'.format(allowsFinishCutting=self.getCSharpBoolean(data=True)),
            '{finishCuttingDepth}'.format(finishCuttingDepth=self.formatAsCoordinate(data=keywords['endDepth'])),
        ) + ');'

    def createUnidirectionalMillingStrategy(self, **keywords):
        return 'CreateUnidirectionalMillingStrategy(' + joinIfNotNone(', ',
            '{allowMultiplePasses}'.format(allowMultiplePasses=self.getCSharpBoolean(data=True)),
            '{cuttingDepth}'.format(cuttingDepth=self.formatAsCoordinate(data=keywords['depthPerStep'])),
            '{finishCuttingDepth}'.format(finishCuttingDepth=self.formatAsCoordinate(data=keywords['endDepth'])),
            '{strokeConnectionType}'.format(strokeConnectionType=self.formatAsInt(data=self.STROKE_LIFT_SHIFT_PLUNGE)),
            '{isApproachRetractPassStrategy}'.format(isApproachRetractPassStrategy=self.getCSharpBoolean(data=True)) if self.VERSION > 17 else None,
        ) + ');'

    def createPolyline(self, **keywords):
        vertex = keywords['startVertex']
        return 'CreatePolyline(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['machining'].contourName),
            '{startX}'.format(startX=self.formatAsCoordinate(data=vertex.position.x)),
            '{startY}'.format(startY=self.formatAsCoordinate(data=vertex.position.y)),
        ) + ');'

    def __createCommonContourStartBlock(self, **keywords):
        keywords['machining'].contourName = self.getNewGeometryName(**keywords)
        return self.createPolyline(**keywords)

    def __createCommonContourEndBlock(self, **keywords):
        if keywords['isClosed']:
            return 'ClosePolyline(' + joinIfNotNone(', ',
                '"{nameElement}"'.format(nameElement=keywords['machining'].contourName),
            ) + ');\n'
        else:
            return ''

    def createSizingStartBlock(self, **keywords):
        return self.__createCommonContourStartBlock(**keywords)

    def createSizingEndBlock(self, **keywords):
        return self.__createCommonContourEndBlock(isClosed=True, **keywords)

    def createDadoContourStartBlock(self, **keywords):
        dado = keywords['dado']

        if dado.isVertical():
            plane = None
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * dado.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            dado.matrix = matrix  # TODO (HORIZONTAL) IT-230: unschönes Design -- NG
            dado.matrix.origin = positionOnPlane

            if not plane.predefined:
                plane.sideID = self.getNewWorkplaneName(**keywords)

        return joinIfNotNone('\n',
            self.createWorkplane(plane=plane, **keywords) if plane and not plane.predefined else None,
            self.selectWorkplane(workplane=plane.sideID, **keywords) if plane and plane.predefined else None,
            self.__createCommonContourStartBlock(**keywords),
        )

    def createDadoContourEndBlock(self, **keywords):
        return joinIfNotNone('\n',
            self.__createCommonContourEndBlock(isClosed=False, **keywords),
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            '',
        )

    def createContourStartBlock(self, **keywords):
        contour = keywords['contour']

        if contour.isVertical():
            plane = None
        else:
            plane, _positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            contour.matrix = matrix  # TODO (HORIZONTAL) IT-230: unschönes Design -- NG

            if not plane.predefined:
                plane.sideID = self.getNewWorkplaneName(**keywords)

        return joinIfNotNone('\n',
            self.createWorkplane(plane=plane, **keywords) if plane and not plane.predefined else None,
            self.selectWorkplane(workplane=plane.sideID, **keywords) if plane and plane.predefined else None,
            self.__createCommonContourStartBlock(**keywords),
        )

    def createContourEndBlock(self, **keywords):
        return joinIfNotNone('\n',
            self.__createCommonContourEndBlock(isClosed=keywords['contour'].isClosed(), **keywords),
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            '',
        )

    def createRabbetContourStartBlock(self, **keywords):
        return self.__createCommonContourStartBlock(**keywords)

    def createRabbetContourEndBlock(self, **keywords):
        return self.__createCommonContourEndBlock(isClosed=False, **keywords)

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        position = segment.getEndVertex().position
        return 'AddSegmentToPolyline(' + joinIfNotNone(', ',
            '{endX}'.format(endX=self.formatAsCoordinate(data=position.x)),
            '{endY}'.format(endY=self.formatAsCoordinate(data=position.y)),
            # Parameter nameElement is optional
        ) + ');'

    def createRadiusSegment(self, **keywords):
        CW = (True, False)
        CCW = (False, False)
        CW_GT_180 = (True, True)
        CCW_GT_180 = (False, True)

        segment = keywords['segment']
        isClockwise, isOver180 = segment.getDirection(CW, CCW, CW_GT_180, CCW_GT_180)
        position = segment.getEndVertex().position
        radius = segment.getRadius()

        return 'AddArc2PointRadiusToPolyline(' + joinIfNotNone(', ',
            '{endX}'.format(endX=self.formatAsCoordinate(data=position.x)),
            '{endY}'.format(endY=self.formatAsCoordinate(data=position.y)),
            '{radius}'.format(radius=self.formatAsAngle(data=radius)),
            '{isClockwise}'.format(isClockwise=self.getCSharpBoolean(data=isClockwise)),
            '{isOver180}'.format(isOver180=self.getCSharpBoolean(data=isOver180)),
            # Parameter nameElement is optional
        ) + ');'

    def generateDadoSegments(self, **keywords):
        if not keywords['dado'].isVertical():
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            dado = copy.deepcopy(keywords['dado'])
            matrix = plane.matrix * dado.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            dado.matrix = matrix  # TODO (HORIZONTAL) IT-230: unschönes Design -- NG
            dado.matrix.origin = positionOnPlane
            keywords['dado'] = dado

        return super().generateDadoSegments(**keywords)

    def selectGeometry(self, **keywords):
        return 'SelectGeometry(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
        ) + ');'

    def createRoughFinish(self, **keywords):
        overMaterial = keywords['contourDistance'] if 'contourDistance' in keywords else 0

        return 'CreateRoughFinish(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
            '{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '"{description}"'.format(description=''),
            '{typeOfProcess}'.format(typeOfProcess=self.GENERAL_ROUTING_PROCESS_TYPE),
            '"{tool}"'.format(tool=keywords['machining'].toolID),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            '{correction}'.format(correction=keywords['radiusCompensation']),
            '{inputSpeed}'.format(inputSpeed=self.AUTOMATIC),
            '{rotSpeed}'.format(rotSpeed=self.AUTOMATIC),
            # TODO CONTOUR2.0 IT-238 feedRate only as "speed" or should it be also "inputSpeed"? -- NG
            '{speed}'.format(speed=self.formatAsCoordinate(data=keywords['feedRate']) if 'feedRate' in keywords and keywords['feedRate'] else self.AUTOMATIC),
            '{overMaterial}'.format(overMaterial=self.formatAsCoordinate(data=overMaterial)),
        ) + ');'

    def __getSizingRadiusCompensation(self, **keywords):
        return self.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else self.RADIUS_COMPENSATION_LEFT

    def createSizingContour(self, **keywords):
        if not self.MCS_DONT_FORMAT_DURING_NESTING or not keywords['nestable']:
            return self.createSizingContourExecution(**keywords)

    def isSizingStrategyUsingTwoPasses(self, **keywords):
        area = keywords['surface'].getSawnSizeRectangularArea()

        return (
            self.SIZING_PASS_STRATEGY == 'force-two-passes'
            or (self.SIZING_PASS_STRATEGY == 'two-passes-for-small-workpieces' and area < 1000000 * self.SIZING_PASS_STRATEGY_AREA_THRESHOLD)
        )

    def createSizingContourBlock(self, **keywords):
        machining = sizing = keywords['sizing']
        surface = keywords['surface']

        self.lastSizingIndex += 1
        name = self.getLocalizedMachiningName(5, "Sizing {index}").format(index=self.lastSizingIndex)

        approach, onset = self.getSizingApproachMode(**keywords)

        routerDepthPerStep = self.getRouterDepthPerStep(machining=sizing, **keywords)

        depthPerStep = self.__getDepthPerStepData(machining=sizing, depthPerStep=routerDepthPerStep, **keywords)

        if self.isSizingStrategyUsingTwoPasses(**keywords) and surface.isFirstSide:
            finalDepth = self.SIZING_FIRST_PASS_REMAINING_THICKNESS + self.SIZING_DEPTH_ADJUSTMENT
            if isNearlyLesserOrEqual(finalDepth, routerDepthPerStep):
                message = self.getLocalizedMessage(76, "Router with insufficient depth per step to be able to make nesting remaining-thickness sizing. Workpiece: {workpiece}, Tool name: '{toolName}', Tool ID: {toolID}")
                logging.error(message.format(
                    workpiece=surface.workpieceName,
                    toolName=machining.tool.name if hasattr(machining.tool, 'name') else None,
                    toolID=machining.toolID
                ))

            if depthPerStep is None:
                depthPerStep = surface.thickness - self.SIZING_FIRST_PASS_REMAINING_THICKNESS

            endDepth = finalDepth
            unidirectionalMillingStrategyNeeded = True
        elif depthPerStep is not None:
            endDepth = 0  # We must set endDepth to 0 or the motion will be done directly to the end depth instead of being progressively. (Not documented that way)
            unidirectionalMillingStrategyNeeded = bool(depthPerStep)
        else:
            endDepth = sizing.actualDepth
            unidirectionalMillingStrategyNeeded = False

        return joinIfNotNone('\n',
            self.resetApproachStrategyApproach(**keywords),
            self.resetRetractStrategyDeparture(**keywords),
            self.setApproachStrategyApproach(approach=approach, onset=onset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_SIZING, **keywords),
            self.selectGeometry(name=sizing.contourName, **keywords),
            self.setRetractStrategyDeparture(approach=approach, onset=onset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_SIZING, **keywords),
            self.createUnidirectionalMillingStrategy(depthPerStep=depthPerStep, endDepth=endDepth, **keywords) if unidirectionalMillingStrategyNeeded else None,
            self.createRoughFinish(name=name, machining=sizing, depth=keywords['zValue'], radiusCompensation=self.__getSizingRadiusCompensation(**keywords), **keywords),
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            '',
            '',
        )

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=self.RADIUS_COMPENSATION_NONE, left=self.RADIUS_COMPENSATION_LEFT, right=self.RADIUS_COMPENSATION_RIGHT)

    def createVerticalRabbet(self, **keywords):
        self.lastRabbetIndex += 1
        name = self.getLocalizedMachiningName(6, "Rabbet {index}").format(index=self.lastRabbetIndex)
        approach, onset = self.APPROACH_MODE_LINEAR_PIERCING

        depthPerStep = self.getRouterDepthPerStep(**keywords)

        depthPerStep = self.__getDepthPerStepData(depthPerStep=depthPerStep, **keywords)
        if depthPerStep is not None:
            endDepth = 0  # We must set endDepth to 0 or the  motion will be done directly to the end depth instead of being progressively. (Not documented that way)
        else:
            endDepth = keywords['rabbet'].actualDepth

        return joinIfNotNone('\n',
            self.resetApproachStrategyApproach(**keywords),
            self.resetRetractStrategyDeparture(**keywords),
            self.setApproachStrategyApproach(approach=approach, onset=onset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET, **keywords),
            self.selectGeometry(name=keywords['rabbet'].contourName, **keywords),
            self.setRetractStrategyDeparture(approach=approach, onset=onset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET, **keywords),
            self.setCompensationMode(**keywords),
            self.createUnidirectionalMillingStrategy(depthPerStep=depthPerStep, endDepth=endDepth, **keywords) if depthPerStep else None,
            self.createRoughFinish(name=name, depth=keywords['zValue'], radiusCompensation=self.getWopRadiusCompensation(**keywords), **keywords),
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            '',
            '',
        )

    def createFreeFormPocket(self, **keywords):
        return joinIfNotNone('\n',
            self.createContourParallelStrategy(**keywords),
            self.createContourPocket(pocket=keywords['contour'], **keywords),
        )

    def createVerticalContour(self, **keywords):
        contour = keywords['contour']

        approachType, approachOnset = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departureType, departureOnset = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        depthPerStep = contour.currentDepthPerStep
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = self.getRouterDepthPerStep(**keywords)

        depthPerStep = self.__getDepthPerStepData(depthPerStep=depthPerStep, **keywords)
        if depthPerStep is not None:
            endDepth = 0  # We must set endDepth to 0 or the  motion will be done directly to the end depth instead of being progressively. (Not documented that way)
        else:
            endDepth = contour.actualDepth

        directionCW = contour.path.isClockwise()

        self.lastContourIndex += 1

        if contour.clearOut:
            name = self.getLocalizedMachiningName(22, "Free Form Pocket {index}").format(index=self.lastContourIndex)
            contourCode = self.createFreeFormPocket(name=name, depth=contour.actualDepth, depthPerStep=depthPerStep, endDepth=endDepth, directionCW=directionCW, **keywords)
        else:
            name = self.getLocalizedMachiningName(7, "Contour {index}").format(index=self.lastContourIndex)
            contourCode = self.createRoughFinish(name=name, depth=contour.actualDepth, radiusCompensation=self.getWopRadiusCompensation(**keywords), feedRate=contour.currentFeedRate, **keywords)

        # TODO CONTOUR2.0 IT-238/IT-271 Currently feedRate are only implemented directly within createVerticalContour, not other contour using machinings -- NG
        return joinIfNotNone('\n',
            self.resetApproachStrategyApproach(**keywords),
            self.resetRetractStrategyDeparture(**keywords),
            self.setApproachStrategyApproach(approach=approachType, onset=approachOnset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_CONTOUR, **keywords) if approachType is not None else None,
            self.selectGeometry(name=keywords['contour'].contourName, **keywords),
            self.setRetractStrategyDeparture(approach=departureType, onset=departureOnset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_CONTOUR, **keywords) if departureType is not None else None,
            self.setCompensationMode(**keywords),
            self.createUnidirectionalMillingStrategy(depthPerStep=depthPerStep, endDepth=endDepth, **keywords) if depthPerStep else None,
            contourCode,
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            '',
            '',
        )

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createVerticalContour, **keywords))

    def createHorizontalContour(self, **keywords):
        return self.createVerticalContour(**keywords)

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createHorizontalContour, **keywords))

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']
        # TODO (HORIZONTAL): IT-222/IT-230 unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if not contour.isVertical():
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            contour.path = contour.transformedPolyline(matrix.getInverted(), positionOnPlane)

        return super().createDefinitionForOneContour(**keywords)

    # miter
    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCW)

    def createSegment(self, **keywords):
        start = keywords['segment'].getStartPoint()
        end = keywords['segment'].getEndPoint()
        return 'CreateSegment(' + joinIfNotNone(', ',
            '"{name}"'.format(name=self.getNewGeometryName(**keywords)),
            '{startX}'.format(startX=self.formatAsCoordinate(data=start.x)),
            '{startY}'.format(startY=self.formatAsCoordinate(data=start.y)),
            '{endX}'.format(endX=self.formatAsCoordinate(data=end.x)),
            '{endY}'.format(endY=self.formatAsCoordinate(data=end.y)),
        ) + ');'

    def createChamfer(self, **keywords):  # with cutter
        return 'CreateChamfer(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
            '{chamferWidth}'.format(chamferWidth=keywords['chamferWidth']),
            '{chamferHeight}'.format(chamferHeight=keywords['chamferHeight']),
            '{overcutLength}'.format(overcutLength=self.ADDITIONAL_DEPTH_FOR_MITER),
            '{toolPosition}'.format(toolPosition=0),  # left at top
            '"{description}"'.format(description=''),
            '{typeOfProcess}'.format(typeOfProcess=self.GENERAL_ROUTING_PROCESS_TYPE),
            '"{tool}"'.format(tool=keywords['toolID']),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            # The rest of the parameters are optional
        ) + ');'

    def createSawCut(self, **keywords):  # with saw incl. pre saw
        start = keywords['start']
        end = keywords['end']

        return 'CreateMacro(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
            '"{macroName}"'.format(macroName='SawCut'),
            '{startX}'.format(startX=self.formatAsCoordinate(data=start.x)),
            '{startY}'.format(startY=self.formatAsCoordinate(data=start.y)),
            '{endX}'.format(endX=self.formatAsCoordinate(data=end.x)),
            '{endY}'.format(endY=self.formatAsCoordinate(data=end.y)),
            '{angle}'.format(angle=self.formatAsAngle(data=keywords['angle'])),
            '1',  # IType = Incisitiontype (pre saw type). 1 = Incisition on Top with the Blade
            '{depth}'.format(depth=keywords['surface'].thickness),
            '{preSawDepth}'.format(preSawDepth=keywords['preSawDepth']),
            '{overDepth}'.format(overDepth=keywords['overDepth']),
            '3',  # IFDepth
            '0.05',  # IDist
            '{approach}'.format(approach='null'),  # OLengthIn (empty=auto)
            '{departure}'.format(departure='null'),  # OLengthOut (empty=auto)
            'null',  # FeedIn
            'null',  # FeedCut
            '2',  # FeedDia
            '"{head}"'.format(head=self.getToolHead(**keywords)),  # HeadS: Head of saw
            '"{tool}"'.format(tool=keywords['toolID']),  # ToolS: ToolID of saw
            'null',  # RotSS
            '"{head}"'.format(head=self.getToolHead(**keywords)),  # HeadD: Head of cutter
            '"{tool}"'.format(tool=''),  # ToolD: ToolID of cutter. Actually we forward nothing, because we don't support the use of a cutter.
            'null',  # RotSD
            '{zReference}'.format(zReference='0'),  # Z_REFERENCE RefPos Auto 0 / Top 1 / Bottom 2
            '0',  # IDDist
            '{info}'.format(info=self.getCSharpBoolean(data=False)),  # Info
            '{radiusCompensation}'.format(radiusCompensation=self.RADIUS_COMPENSATION_AUTOMATIC),
            '{hood}'.format(hood='null'),
            '{sawOn}'.format(sawOn=self.getCSharpBoolean(data=True)),  # SawOn
            '0',  # DepthCor
            '{natEnd}'.format(natEnd=self.getCSharpBoolean(data=False)),  # NatEnd
            'null',  # CAngle
        ) + ');'

    def createVerticalMiter(self, **keywords):
        self.lastMiterIndex += 1
        miterName = self.getLocalizedMachiningName(3, "Miter {index}").format(index=self.lastMiterIndex)

        surface = keywords['surface']
        miter = keywords['miter']

        start = miter.getEndPoint()  # start/end reversed because of the use in CreateSawCut
        end = miter.getStartPoint()
        angle = miter.getAngle()
        preSawDepth = self.zValueMiterPreSawConverter.convertWithThickness(surface.thickness, self.MITER_PRE_SAW_DEPTH_FROM_TOP)
        # sawDepth = self.zValueMiterSawConverter.convertWithThickness(surface.thickness, surface.thickness + self.MITER_SAW_DEPTH_FROM_BOTTOM)

        return '\n'.join((
            # The Maestro working CreateChamfer uses the cutter instead of the saw!
            # self.createSegment(segment=miter.getLineSegment(), **keywords),  # not needed for createSawCut
            # self.createChamfer(name=miterName, chamferWidth=miter.width, chamferHeight=surface.thickness, toolID=miter.toolID, **keywords),
            # The Maestro macro SawCut uses the saw incl. pre saw
            self.createSawCut(name=miterName, start=start, end=end, angle=angle, preSawDepth=preSawDepth, toolID=miter.toolID, overDepth=self.ADDITIONAL_DEPTH_FOR_MITER, **keywords),
            '',
        ))

    def createNCStop(self, **keywords):
        self.lastNCStopIndex += 1
        name = self.getLocalizedMachiningName(2, "NC Stop {index}").format(index=self.lastNCStopIndex)

        return 'CreateMessage(' + joinIfNotNone(', ',
            '"{name}"'.format(name=name),
            '"{text}"'.format(text=keywords['comment']),
            '{waitForUserInput}'.format(waitForUserInput=self.getCSharpBoolean(data=True)),
            '{releaseWorkpiece}'.format(releaseWorkpiece=self.getCSharpBoolean(data=keywords['releaseWorkpiece']) if 'releaseWorkpiece' in keywords else self.getCSharpBoolean(data=False)),
        ) + ');\n'

    def createTechnologyDataForDadoWithSawAlongAxis(self, **_keywords):
        return joinIfNotNone(', ',
            '{hood}'.format(hood='null'),  # Technology data 'Haube'
            '{feed}'.format(feed='10'),  # Technology data 'Vorschub'
            '{speed}'.format(speed='null'),  # Technology data 'Drehzahl'
        )

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        def __getYforRNT(y, width):
            return y - width / 2

        self.lastDadoIndex += 1
        dadoName = self.getLocalizedMachiningName(4, "Dado {index}").format(index=self.lastDadoIndex)

        start, end = self.getDadoLineSegmentWithSecurityDistance(**keywords).getStartAndEndPoints()
        dado = keywords['dado']
        width = dado.getWidth()
        return 'CreateMacro(' + joinIfNotNone(', ',
            '"{name}"'.format(name=dadoName),
            '"{macroName}"'.format(macroName='RNT'),
            '{startX}'.format(startX=self.formatAsCoordinate(data=min(start.x, end.x))),
            '{startY}'.format(startY=self.formatAsCoordinate(data=__getYforRNT(start.y, width))),
            '{startZ}'.format(startZ=self.formatAsCoordinate(data=keywords['depth'])),
            self.createTechnologyDataForDadoWithSawAlongAxis(**keywords),
            '{endX}'.format(endX=self.formatAsCoordinate(data=max(start.x, end.x))),
            '{a}'.format(a=self.formatAsCoordinate(data=width)),
            '{boolean}'.format(boolean=self.getCSharpBoolean(data=True)),  # Laengenkorrektur
            '"{tool}"'.format(tool=dado.toolID),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            'false, true, false, null, null, null, null, null, false',
            '"{face}"'.format(face=self.mainWorkplane),
        ) + ');\n'

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        if self.ONE_PROGRAM and self.mainWorkplane == self.WORKPLANE_BOTTOM:
            return self.createVerticalDadoWithSawCRotated(**keywords)

        def __getXforRNTY(x, width):
            return x - width / 2

        self.lastDadoIndex += 1
        dadoName = self.getLocalizedMachiningName(4, "Dado {index}").format(index=self.lastDadoIndex)

        start, end = self.getDadoLineSegmentWithSecurityDistance(**keywords).getStartAndEndPoints()
        dado = keywords['dado']
        width = dado.getWidth()
        return 'CreateMacro(' + joinIfNotNone(', ',
            '"{name}"'.format(name=dadoName),
            '"{macroName}"'.format(macroName='RNTY'),
            '{startX}'.format(startX=self.formatAsCoordinate(data=__getXforRNTY(start.x, width))),
            '{startY}'.format(startY=self.formatAsCoordinate(data=min(start.y, end.y))),
            '{startZ}'.format(startZ=self.formatAsCoordinate(data=keywords['depth'])),
            self.createTechnologyDataForDadoWithSawAlongAxis(**keywords),
            '{endY}'.format(endY=self.formatAsCoordinate(data=max(start.y, end.y))),
            '{a}'.format(a=self.formatAsCoordinate(data=width)),
            '{boolean}'.format(boolean=self.getCSharpBoolean(data=True)),  # Laengenkorrektur
            '"{tool}"'.format(tool=dado.toolID),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            # The rest of the parameters are optional
        ) + ');\n'

    def createVerticalDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        self.lastDadoIndex += 1
        name = self.getLocalizedMachiningName(4, "Dado {index}").format(index=self.lastDadoIndex)

        radiusCompensation = self.getDadoRadiusCompensation(**keywords)
        approach, onset = self.APPROACH_MODE_LINEAR_FLYING if dado.isInsetBoth() else self.APPROACH_MODE_LINEAR_PIERCING

        depthPerStep = self.getRouterDepthPerStep(**keywords)

        depthPerStep = self.__getDepthPerStepData(depthPerStep=depthPerStep, **keywords)
        if depthPerStep is not None:
            endDepth = 0  # We must set endDepth to 0 or the  motion will be done directly to the end depth instead of being progressively. (Not documented that way)
        else:
            endDepth = dado.actualDepth

        return joinIfNotNone('\n',
            self.resetApproachStrategyApproach(**keywords),
            self.resetRetractStrategyDeparture(**keywords),
            self.setApproachStrategyApproach(approach=approach, onset=onset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET, **keywords),
            self.selectGeometry(name=dado.contourName, **keywords),
            self.setRetractStrategyDeparture(approach=approach, onset=onset, distance=self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET, **keywords),
            self.createUnidirectionalMillingStrategy(depthPerStep=depthPerStep, endDepth=endDepth, **keywords) if depthPerStep else None,
            self.createRoughFinish(name=name, radiusCompensation=radiusCompensation, **keywords),
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            '',
            '',
        )

    def createHorizontalDadoWithRouter(self, **keywords):
        return self.createVerticalDadoWithRouter(depth=keywords['dado'].depth, **keywords)

    def createSlot(self, **keywords):
        return 'CreateSlot(' + joinIfNotNone(', ',
            '"{name}"'.format(name=keywords['name']),
            '{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '"{description}"'.format(description=''),
            '{typeOfProcess}'.format(typeOfProcess=self.GENERAL_ROUTING_PROCESS_TYPE),
            '"{tool}"'.format(tool=keywords['machining'].toolID),
            '"{head}"'.format(head=self.getToolHead(**keywords)),
            '{correction}'.format(correction=str(keywords['radiusCompensation']) + '3'),  # depth / correction in length
            # The further parameters are optional
        ) + ');\n'

    def __createSingleVerticalSawCutCRotated(self, **keywords):
        self.lastDadoIndex += 1
        dadoName = self.getLocalizedMachiningName(4, "Dado {index}").format(index=self.lastDadoIndex)

        return '\n'.join((
            self.createSegment(**keywords),
            self.createSlot(name=dadoName, **keywords),
            '',
        ))

    def createVerticalDadoWithSawCRotated(self, **keywords):
        multiSawDado = MultiSawDado(self)
        return multiSawDado.createMultipleVerticalSawCuts(createFunction=self.__createSingleVerticalSawCutCRotated, **keywords) + '\n'

    # NC Macro
    def getNCMacroCookedVariables(self, **keywords):
        variables = super().getNCMacroCookedVariables(**keywords)

        plane = keywords['plane']
        ncmacro = keywords['ncmacro']
        if ncmacro.isVertical():
            sideCAngle = variables['cangle']
        else:
            angle = VectorMath.angleBetween(ncmacro.matrix.u, plane.wopMatrix.u)

            if not isNearlyEqual(angle, 0.0):
                crossProduct = ncmacro.matrix.u.cross(plane.wopMatrix.u)
                if not isNearlyEqual(crossProduct.length(), 0.0) and crossProduct.normalized() == plane.wopMatrix.w:
                    angle = - angle

            sideCAngle = self.formatter.formatAsAngle(normalizeAngle(angle))

        self.lastNCMacroIndex += 1
        variables['index'] = self.lastNCMacroIndex

        variables['sideid'] = plane.sideID
        variables['sidex'] = self.formatter.formatAsCoordinate(keywords['positionOnPlane'].x)
        variables['sidey'] = self.formatter.formatAsCoordinate(keywords['positionOnPlane'].y)
        variables['sidecangle'] = sideCAngle

        return variables

    def createNCMacroCode(self, **keywords):
        ncmacro = keywords['ncmacro']
        if ncmacro.isVertical():
            plane = Plane(sideID=self.mainWorkplane, predefined=True)
            positionOnPlane = ncmacro.position
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            if plane and not plane.predefined:
                plane.sideID = self.getNewWorkplaneName(**keywords)

        return joinIfNotNone('\n',
            self.createWorkplane(plane=plane, **keywords) if plane and not plane.predefined else None,
            self.selectWorkplane(workplane=plane.sideID, **keywords) if plane and plane.predefined else None,
            super().createNCMacroCode(plane=plane, positionOnPlane=positionOnPlane, **keywords),
            self.selectWorkplane(workplane=self.mainWorkplane, **keywords),
            ''
            '',
        )

    def createRecordFormatParameters(self, **keywords):
        return ', '.join((
            '{recordValue}'.format(recordValue=parameter.value) for parameter in keywords['parameters']
        ))

    def createNCMacroRecordFormat(self, **keywords):
        ncmacro = keywords['ncmacro']

        if ncmacro.isVertical():
            plane = Plane(sideID=self.mainWorkplane, predefined=True)
            positionOnPlane = ncmacro.position
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)

        parameters = tuple(self.getNCMacroCookedParameters(plane=plane, positionOnPlane=positionOnPlane, **keywords))

        macroName = ncmacro.name
        wopParameters = self.createRecordFormatParameters(parameters=parameters, **keywords)

        # self.lastNCMacroIndex is incremented by getNCMacroCookedVariables
        name = "{macroName} {index}".format(macroName=macroName, index=self.lastNCMacroIndex)

        return 'CreateMacro(' + joinIfNotNone(', ',
            '"{name}"'.format(name=name),
            '"{macroName}"'.format(macroName=macroName),
            wopParameters,
        ) + ');\n'

    # Footer
    def createEndOfFile(self, **keywords):
        name = self.getLocalizedMachiningName(19, "Null operation End")
        return self.createParkingPosition(name=name, parkingPositionDistance=self.PARKING_POSITION_DISTANCE, **keywords) if self.PARKING_POSITION_DISTANCE else ''

    # Localization
    def getLocalizedMachiningName(self, index, default):
        return getLocalizedString(MAESTRO_NAME_MESSAGES, index, default)

    def getLocalizedComment(self, index, default):
        return getLocalizedString(MAESTRO_PROGRAM_COMMENTS_RESOURCEGROUP, index, default)

    def getLocalizedMultiCenterStudioMessage(self, index, default):
        return getLocalizedString(MAESTRO_MULTI_CENTER_STUDIO_PROGRAM_COMMENTS_RESOURCEGROUP, index, default)
