# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import getpass
import itertools
import logging
import math

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.localization import getLocalizedString, MASTERWORK_DESCRIPTIONS_RESOURCEGROUP
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface import configuration
from interiorcad.interface.geometry import isNearlyEqual
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.machiningsexporter.horizontaldrillingsexporter import HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds
from interiorcad.interface.ncexport.approachcreator import createArcApproach
from interiorcad.interface.surface import Contour2D, Sizing2D, Drilling2D, Pocket2D, GridDrilling2D, Rabbet2D, Dado2D, CenterPunch2D
from interiorcad.interface.surface.machiningplanes import MachiningPlanes
from interiorcad.ncexport.base import NCExportBase


class MasterworkExport(NCExportBase):
    EXPORT_FORMAT = 'Masterwork'

    APPROACH_MODE_LINEAR_PIERCING = 1
    APPROACH_MODE_LINEAR_FLYING = 2
    APPROACH_MODE_ARC_PIERCING = 3
    APPROACH_MODE_ARC_FLYING = APPROACH_MODE_ARC_PIERCING  # We cannot make a difference in Masterwork
    APPROACH_MODE_WITHOUT = 0
    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_LINEAR_FLYING

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.CURRENT_USER = getpass.getuser()

        # FIXME IT-259: the keys must be defined in the XML file -- NG
        self.SIZING_APPROACH_DEPARTURE_DISTANCE = self.getMachineConfiguration().getAsFloat('sizing-approach-departure-distance', 30)
        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = self.getMachineConfiguration().getAsFloat('contour-approach-departure-distance', 30)

        self.TARGET_MACHINE = self.getMachineConfiguration().get('target-machine', '6 PROJECT MW310 WINDOWS')

        self.createZValueConverters(**keywords)

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)

    def getNCFileExtension(self, **_keywords):
        return '.tlf'

    # Geometrical Entities
    def createGeometricalEntityCIR(self, **keywords):
        return '\n'.join((
            'CIR',
            'INDEX',
            '{index}'.format(index=keywords['index']),
            'LINKS',
            '1',
            'L',
            '2',
            'DIAM_UT',
            '0',
            'X',
            '{x} {xRef}'.format(x=self.formatAsCoordinate(data=keywords['x']), xRef=keywords['xRef']),
            'Y',
            '{y} {yRef}'.format(y=self.formatAsCoordinate(data=keywords['y']), yRef=keywords['yRef']),
            'XEXPR',
            '',
            'YEXPR',
            '',
            'R',
            '{radius}'.format(radius=self.formatAsCoordinate(data=keywords['diameter'] / 2)),
            '\\CIR',
            '',
        ))

    def createGeometricalEntityLIN(self, **keywords):
        return '\n'.join((
            'LIN',
            'INDEX',
            '{index}'.format(index=keywords['index']),
            'LINKS',
            '1',
            'L',
            '2',
            'DIAM_UT',
            '0',
            'X1',
            '{x} I'.format(x=self.formatAsCoordinate(data=keywords['machiningStartPosition'].x)),
            'Y1',
            '{y} I'.format(y=self.formatAsCoordinate(data=keywords['machiningStartPosition'].y)),
            'X1EXPR',
            '',
            'Y1EXPR',
            '',
            'X2',
            '{x} I'.format(x=self.formatAsCoordinate(data=keywords['machiningEndPosition'].x)),
            'Y2',
            '{y} I'.format(y=self.formatAsCoordinate(data=keywords['machiningEndPosition'].y)),
            'X2EXPR',
            '',
            'Y2EXPR',
            '',
            'LUNG',
            '0',
            'ANG',
            '0',
            '\\LIN',
            '',
        ))

    # Header
    def createHead(self, **keywords):
        self.lastSideMainIndex = -1

        self.predefinedMachiningPlanes = MachiningPlanes(sideIDGenerator=lambda: None)
        self.predefinedMachiningPlanes.planes.extend(self.createPredefinedPlanes(**keywords))

        return '\n'.join((
            self.setTargetMachine(**keywords),
            self.createGeneralParameters(**keywords),
            self.createEntities(**keywords),
        ))

    def setTargetMachine(self, **_keywords):
        return '\n'.join((
            'TARGET',
            self.TARGET_MACHINE,
        ))

    # General parameters
    def createGeneralParameters(self, **keywords):
        return '\n'.join((
            'PARGEN',
            self.createGeneralParametersCommon(**keywords),
            '\\PARGEN',
        ))

    def createGeneralParametersCommon(self, **keywords):
        surface = keywords['surface']

        creationDate = configuration.DATETIME_NOW.strftime("%d.%m.%Y")
        additionX = surface.getAdditionX()
        additionY = surface.getAdditionY()

        return '\n'.join((
            'LIMMINX',
            '0',
            'LIMMINY',
            '0',
            'LIMMAXX',
            '{length}'.format(length=self.formatAsCoordinate(data=surface.getFinishedSizeX() + surface.getEdgesAdditionX())),
            'LIMMAXY',
            '{width}'.format(width=self.formatAsCoordinate(data=surface.getFinishedSizeY() + surface.getEdgesAdditionY())),
            'RAGGIORACCORDO',
            '10',
            'CURLAYER',
            '1',
            'LAYER',  # Layers are not handled; this part is present only for compatibility toward an older TLF format.
            '1',
            'CAD',
            '1',
            '\\LAYER',
            'LAYER',
            '2',
            'CAM',
            '1',
            '\\LAYER',
            'NUMPROG',
            '1',
            'DESCR',
            surface.workpieceName,
            'NUMTABUTE',
            '1',
            'NUMTABVENT',
            '1',
            'LUNG',
            '{length}'.format(length=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            'LARG',
            '{width}'.format(width=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            'ALT',
            '{thickness}'.format(thickness=self.formatAsCoordinate(data=surface.thickness)),
            'AUTORE',
            '{currentUser} @ Vectorworks interiorcad'.format(currentUser=self.CURRENT_USER),
            'DATACREAZ',
            '{date}'.format(date=creationDate),
            'DATAMOD',
            '{date}'.format(date=creationDate),
            'ORDINECOMP',
            '0|1|2|3|4|5|6|',
            'ORIGLUNG',
            '-1',
            'ORIGLARG',
            '-1',
            'ORIGALT',
            '-1',
            'FLAGOTTFORI',
            '1',
            'FLAGOTTLAV',
            '1',
            'DEFAULT_ALT_CUFFIA',
            '0',
            'OFFSET_DIMA_X',
            '{additionX}'.format(additionX=self.formatAsCoordinate(data=additionX)),
            'OFFSET_DIMA_XEXPR',
            '{additionX}'.format(additionX=self.formatAsCoordinate(data=additionX)),
            'OFFSET_DIMA_Y',
            '{additionY}'.format(additionY=self.formatAsCoordinate(data=additionY)),
            'OFFSET_DIMA_YEXPR',
            '{additionY}'.format(additionY=self.formatAsCoordinate(data=additionY)),
            'OFFSET_DIMA_Z',
            '0',
            'OFFSET_DIMA_ZEXPR',
            '0',
            'QUOTA_X_BREAK',
            '0',
            'TOLL_BREAK',
            '0',
            'QPARK_U1',
            '',
            'QPARK_U2',
            '',
            'QCARIC_U1',
            '',
            'QCARIC_U2',
            '',
            'QCARIC_V',
            '',
            'BATT_AUX',
            '1',
            'MOT_OTTIM',
            '',
            'VEL_ROT',
            '0',
            'PRG_BLOCCO',
            '',
            'PRG_SBLOCCO',
            '',
            'PLATEAU',
            '',
            'SCARAUT_UTEFOR',
        ))

    def createDefinitionForOneMachining(self, **keywords):
        def isDadoRouted(dado):
            return hasattr(dado, 'tool') and dado.tool and dado.tool.type == Tool.ROUTER

        surface = keywords['surface']
        self.currentPlane = None

        # Similar to surface.getAllMachinings, but here the sizing is the first.
        allMachinings = itertools.chain(
            (surface.sizing,),
            surface.getAllMachiningsButSizing(),
        )

        for machining in allMachinings:
            if isinstance(machining, Sizing2D):
                if surface.hasSizing():
                    yield self.createSizingContourDefinition(**keywords)
            elif isinstance(machining, Pocket2D):  # Careful a Pocket2D ist a Drilling2D
                yield self.createDefinitionForOneDrilling(machining=machining, drilling=machining, pocket=machining, **keywords)
            elif isinstance(machining, (Drilling2D, CenterPunch2D)):
                yield self.createDefinitionForOneDrilling(machining=machining, drilling=machining, **keywords)
            elif isinstance(machining, GridDrilling2D):
                yield self.createDefinitionForOneGridDrilling(machining=machining, drilling=machining, **keywords)
            elif isinstance(machining, Rabbet2D):
                yield self.createDefinitionForOneRabbet(machining=machining, rabbet=machining, **keywords)
            elif isinstance(machining, Dado2D):
                if machining.isVertical() and isDadoRouted(machining):  # TODO IT-260: implement horizontal dado definitions
                    yield self.createDefinitionForOneDado(machining=machining, dado=machining, **keywords)
            elif isinstance(machining, Contour2D):
                if machining.isVertical():  # TODO IT-260: implement horizontal contour definitions
                    yield self.createDefinitionForOneContour(contour=machining, machining=machining, **keywords)

    # Section containing geometrical entities description
    def createEntities(self, **keywords):
        self.currentPlane = None

        # Generate all the lines explicitly first, so that self.currentPlane is correct for self.selectWorkplane
        lines = ''.join(line for line in self.createDefinitionForOneMachining(**keywords))

        return joinIfNotNone('\n',
            'ENTITIES',
            lines,
            self.selectWorkplane(plane=self.SIDE_INCLINED_FACE_FRONT, **keywords),
            '\\ENTITIES',
            '',
        )

    # Machinings
    SIDE_MAIN, SIDE_BACK, SIDE_FRONT, SIDE_LEFT, SIDE_RIGHT, SIDE_INCLINED_FACE_BACK, SIDE_INCLINED_FACE_FRONT = range(0, 7)

    # only used for default and legacy drills, replaced with property 'drilling-drill-type'
    FLAT_DRILL_TYPE = 0  # Standard
    PIERCING_DRILL_TYPE = 2  # V-Point
    WIDE_DRILL_TYPE = 3  # Flat

    DRILLING_SPEED_FAST = 0
    DRILLING_SPEED_SLOW_FAST = 1
    DRILLING_SPEED_FAST_SLOW = 2
    DRILLING_SPEED_SLOW_FAST_SLOW = 3

    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    DRILLING_FEED_RATE_IN_DEPTH = 1.5
    ROUTER_FEED_RATE = 8

    def getPostPostProcessors(self, **keywords):
        # TODO IT-260 avoid duplicate Code with sortMachinings -- NG
        this = self

        def getMachiningSide(machining):
            if machining.isVertical():
                return self.SIDE_MAIN
            else:
                plane, _positionOnPlane = self.__calculatePlane(machining=machining, **keywords)
                return plane.sideID if plane.sideID else 9999  # TODO IT-260: support machinings on non-predefined faces (a.k.a. five-axis machinings)

        class MachiningPlaneSorter:
            def __init__(self, ncexport):
                self.ncexport = ncexport

            def process(self, surface):
                self.ncexport.predefinedMachiningPlanes = MachiningPlanes(sideIDGenerator=lambda: None)
                self.ncexport.predefinedMachiningPlanes.planes.extend(self.ncexport.createPredefinedPlanes(surface=surface, **keywords))

                machinings = surface.machinings
                sortedDrillings = sorted(machinings, key=getMachiningSide)
                groups = (group for _key, group in itertools.groupby(sortedDrillings, key=getMachiningSide))
                surface.machinings = list(itertools.chain.from_iterable(groups))

                del self.ncexport.predefinedMachiningPlanes

        return (
            MachiningPlaneSorter(this),
        )

    def selectWorkplane(self, **keywords):
        plane = keywords['plane']
        oldPlane = self.currentPlane if self.currentPlane is not None else -1
        self.currentPlane = plane

        if self.currentPlane is not None:
            return '\n'.join((
                'PLANE {plane}'.format(plane=sideID) for sideID in range(oldPlane + 1, self.currentPlane + 1)
            ))
        else:
            return 'PLANE NONE'  # IT-260: this should not happen, when five-axis are correctly supported -- NG

    def changeWorkplane(self, **keywords):
        if keywords['plane'] != self.currentPlane:
            return joinIfNotNone('\n',
                self.selectWorkplane(**keywords),
            )

    def createMachiningsOneByOne(self, **keywords):
        self.currentPlane = None

        # Generate all the lines explicitly first, so that self.currentPlane is correct for self.selectWorkplane
        lines = super().createMachiningsOneByOne(**keywords)

        return joinIfNotNone('\n',
            'WORKS',
            lines,
            self.selectWorkplane(plane=self.SIDE_INCLINED_FACE_FRONT, **keywords),
            '\\WORKS'
            '',
            '',
        )

    def sortMachinings(self, **keywords):
        # TODO IT-260 avoid duplicate Code with getPostPostProcessor -- NG
        def getMachiningSide(machining):
            if machining.isVertical():
                return self.SIDE_MAIN
            else:
                plane, _positionOnPlane = self.__calculatePlane(machining=machining, **keywords)
                return plane.sideID if plane.sideID else 9999  # TODO IT-260: support machinings on non-predefined faces (a.k.a. five-axis machinings)

        machinings = super().sortMachinings(**keywords)

        sortedDrillings = sorted(machinings, key=getMachiningSide)
        groups = (group for _key, group in itertools.groupby(sortedDrillings, key=getMachiningSide))
        return list(itertools.chain.from_iterable(groups))

    # Drillings
    def createDrillingDefinitions(self, **keywords):
        raise NotImplementedError  # See createDefinitionForOneMachining

    def createHorizontalDrillingDefinitions(self, **keywords):
        raise NotImplementedError  # See createDefinitionForOneMachining

    def createGridDrillingDefinitions(self, **keywords):
        raise NotImplementedError  # See createDefinitionForOneMachining

    def createPocketDefinitions(self, **keywords):
        raise NotImplementedError  # See createDefinitionForOneMachining

    def createDefinitionForOneDrilling(self, **keywords):
        surface = keywords['surface']
        drilling = keywords['drilling']
        machining = keywords['machining']

        diameter = drilling.installedDiameter
        xRef = 'I'
        yRef = 'I'
        zPosition = drilling.zPosition

        if drilling.isVertical():
            self.lastSideMainIndex += 1
            machining.index = self.lastSideMainIndex

            x = drilling.position.x
            y = drilling.position.y

            sideID = self.SIDE_MAIN

        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            if hasattr(plane, 'lastSideIndex'):
                plane.lastSideIndex += 1
            else:
                plane.lastSideIndex = 0

            drilling.index = plane.lastSideIndex
            sideID = plane.sideID

            if plane.sideID == self.SIDE_RIGHT:  # isXMinus
                x = surface.getFinishedSizeY() - drilling.position.y
                xRef = 'F'
                y = surface.thickness - zPosition
                yRef = 'F'
            elif plane.sideID == self.SIDE_BACK:  # isYMinus
                x = drilling.position.x
                xRef = 'I'
                y = surface.thickness - zPosition
                yRef = 'F'
            elif plane.sideID == self.SIDE_LEFT:  # isXPlus
                x = drilling.position.y
                xRef = 'I'
                y = surface.thickness - zPosition
                yRef = 'F'
            elif plane.sideID == self.SIDE_FRONT:  # isYPlus
                x = drilling.position.x
                xRef = 'I'
                y = zPosition
                yRef = 'I'
            else:
                # TODO IT-260: support rotated horizontal and other five-axis drillings
                logging.error("Rotated horizontal drillings aren't supported yet!")
                x, y = positionOnPlane.x, positionOnPlane.y

        drilling.newOrigin = Vector(x, y, 0)  # TODO IT-260: unschönes Design -- NG

        return joinIfNotNone('\n',
            self.changeWorkplane(plane=sideID),
            self.createGeometricalEntityCIR(index=machining.index, x=x, xRef=xRef, y=y, yRef=yRef, diameter=diameter, **keywords),
        )

    def createDefinitionForOneGridDrilling(self, **keywords):
        machining = keywords['machining']

        self.lastSideMainIndex += 1
        machining.index = self.lastSideMainIndex

        machiningStartPosition = machining.getStartDrilling().position
        machiningEndPosition = machining.getEndDrilling().position

        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createGeometricalEntityLIN(index=machining.index, machiningStartPosition=machiningStartPosition, machiningEndPosition=machiningEndPosition, **keywords),
        )

    def createCommonDrilling(self, **keywords):
        surface = keywords['surface']
        drilling = keywords['drilling']

        executionIndex = keywords['drilling'].index

        installedDiameter = self.formatAsCoordinate(data=drilling.installedDiameter)
        # depth = self.formatAsCoordinate(data=drilling.depth)
        depth = self.formatAsCoordinate(data=keywords['depth'] if 'depth' in keywords else drilling.depth)

        if 'description' not in keywords:
            keywords['description'] = self.getLocalizedDescription(1, 'Drilling')

        return '\n'.join((
            'FOR_SING',
            'ENT',
            '{index}'.format(index=executionIndex),
            'ENTORIG',
            '{index}'.format(index=executionIndex),
            'DESCRIZIONE',
            '{description} Ø{diameter} x {depth}'.format(description=keywords['description'], diameter=installedDiameter, depth=depth),
            'NUMUTENSILE',
            '0',
            'VELROTAZIONE',
            '0',
            'VELENTRATA',
            '{feedRateInDepth}'.format(feedRateInDepth=self.formatAsCoordinate(data=self.DRILLING_FEED_RATE_IN_DEPTH)),
            'PROFLAVORAZIONE',
            '{depth}'.format(depth=depth),
            'PROFLAVORAZIONEEXPR',
            '',
            'ANGTILTING',
            '90',
            'ANGINDEX',
            '0',
            'ORDER',
            '0',
            'SEQUENZA',
            '0',
            'SPESS',
            '{thickness}'.format(thickness=self.formatAsCoordinate(data=surface.thickness)),
            'ALT_CUFFIA',
            '0',
            'IDPEZZONESTING',
            '0',
            'ENABLEDEXPR',
            '',
            'VELLAVORAZIONE',
            '0',
            'DIAMETRO',
            '{diameter}'.format(diameter=installedDiameter),
            'OFFSET',
            '0',
            'OFFSETEXPR',
            '0',
            'RALLFORO',
            '{drillingSpeed}'.format(drillingSpeed=keywords['drillingSpeed']),
            'VELRALL',
            '1',
            'TOPDIST',
            '3',
            'BOTTOMDIST',
            '3',
            'TIPOPUNTA',
            '{drilltype}'.format(drilltype=self.getDrillType(**keywords)),
            '\\FOR_SING',
            '',
        ))

    def createVerticalDrilling(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=self.DRILLING_SPEED_FAST, **keywords),
        )

    def createVerticalCenterPunch(self, **keywords):
        return super().createVerticalCenterPunch(description=self.getLocalizedDescription(3, 'Center Punch'), **keywords)

    def createVerticalDrillingPiercing(self, **keywords):
        drillingSpeed = self.DRILLING_SPEED_SLOW_FAST_SLOW if self.PIERCING_DRILLING_ADDITION > 0 else self.DRILLING_SPEED_FAST
        description = self.getLocalizedDescription(2, 'Drilling through')

        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=drillingSpeed, description=description, **keywords),
        )

    def createVerticalGridDrillingCommon(self, **keywords):
        surface = keywords['surface']
        gridDrilling = keywords['gridDrilling']

        executionIndex = gridDrilling.index

        return '\n'.join((
            'FOR_LINEA',
            'ENT',
            '{index}'.format(index=executionIndex),
            'ENTORIG',
            '{index}'.format(index=executionIndex),
            'DESCRIZIONE',
            '{description}'.format(description=self.getLocalizedDescription(4, 'Grid Drilling')),
            'NUMUTENSILE',
            '0',
            'VELROTAZIONE',
            '0',
            'VELENTRATA',
            '{feedRateInDepth}'.format(feedRateInDepth=self.formatAsCoordinate(data=self.DRILLING_FEED_RATE_IN_DEPTH)),
            'PROFLAVORAZIONE',
            '{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'] if 'depth' in keywords else gridDrilling.depth)),
            'PROFLAVORAZIONEEXPR',
            '',
            'ANGTILTING',
            '90',
            'ANGINDEX',
            '0',
            'ORDER',
            '0',
            'SEQUENZA',
            '0',
            'SPESS',
            '{thickness}'.format(thickness=self.formatAsCoordinate(data=surface.thickness)),
            'ALT_CUFFIA',
            '0',
            'IDPEZZONESTING',
            '0',
            'ENABLEDEXPR',
            '',
            'VELLAVORAZIONE',
            '0',
            'DIAMETRO',
            '{diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
            'PASSO',
            '{distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
            'PASSOEXPR',
            '',
            'MODO',
            '1',
            'CENTRATO',
            '0',
            'OFFSET',
            '0',
            'RALLFORO',
            '{drillingSpeed}'.format(drillingSpeed=keywords['drillingSpeed']),
            'VELRALL',
            '1',
            'TOPDIST',
            '3',
            'BOTTOMDIST',
            '3',
            'TIPOPUNTA',
            '{drilltype}'.format(drilltype=self.getDrillType(drilling=keywords['gridDrilling'], **keywords)),
            '\\FOR_LINEA',
        ))

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createVerticalGridDrillingCommon(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=self.DRILLING_SPEED_FAST, **keywords),
            ''
        )

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createVerticalGridDrillingCommon(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=self.DRILLING_SPEED_FAST, **keywords),
            ''
        )

    def getHorizontalDrillingsExporter(self, **_keywords):
        return HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds(self)

    def getHorizontalMachiningsExporter(self, **_keywords):
        return HorizontalMachiningsExporter_SeparateCallForMachiningsOnBounds(self)

    def createHorizontalDrillingFromFront(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_FRONT),
            self.createCommonDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=self.DRILLING_SPEED_FAST, **keywords),
        )

    def createHorizontalDrillingFromRight(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_RIGHT),
            self.createCommonDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=self.DRILLING_SPEED_FAST, **keywords),
        )

    def createHorizontalDrillingFromBack(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_BACK),
            self.createCommonDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=self.DRILLING_SPEED_FAST, **keywords),
        )

    def createHorizontalDrillingFromLeft(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_LEFT),
            self.createCommonDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, drillingSpeed=self.DRILLING_SPEED_FAST, **keywords),
        )

    def createRotatedHorizontalDrilling(self, **_keywords):
        return ''

    def createFiveAxisDrilling(self, **_keywords):
        return ''

    def createPredefinedPlanes(self, **keywords):
        size = keywords['surface'].getFinishedSize()

        yield Plane(
            predefined=True, sideID=self.SIDE_FRONT,
            matrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_RIGHT,
            matrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, size.y, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_BACK,
            matrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_LEFT,
            matrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
        )

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']

        machiningPlane = Plane.createFromMachining(machining)

        plane = self.predefinedMachiningPlanes.findPlane(machiningPlane)
        if plane:
            machiningPlane = plane

        return machiningPlane, machining.getPositionOnPlane(machiningPlane)

    # Pockets
    def createCommonPocket(self, **keywords):
        surface = keywords['surface']
        pocket = keywords['pocket']

        executionIndex = pocket.index

        # depth = self.formatAsCoordinate(data=drilling.depth)
        depth = self.formatAsCoordinate(data=keywords['depth'] if 'depth' in keywords else pocket.depth)

        return '\n'.join((
            'SVUOTAMENTO',
            'ENT',
            '{index}'.format(index=executionIndex),
            'ENTORIG',
            '{index}'.format(index=executionIndex),
            'DESCRIZIONE',
            '{description} Ø{diameter} x {depth}'.format(description=keywords['description'], diameter=pocket.diameter, depth=depth),
            'NUMUTENSILE',
            '{toolID}'.format(toolID=pocket.toolID),
            'VELROTAZIONE',
            '22',
            'VELENTRATA',
            '5',
            'PROFLAVORAZIONE',
            '{depth}'.format(depth=depth),
            'PROFLAVORAZIONEEXPR',
            '',
            'ANGTILTING',
            '90',
            'ANGINDEX',
            '0',
            'ORDER',
            '0',
            'SEQUENZA',
            '0',
            'SPESS',
            '{thickness}'.format(thickness=self.formatAsCoordinate(data=surface.thickness)),
            'ALT_CUFFIA',
            '0',
            'IDPEZZONESTING',
            '0',
            'ENABLEDEXPR',
            '',
            'VELLAVORAZIONE',
            '{feedRate}'.format(feedRate=self.formatAsCoordinate(data=self.ROUTER_FEED_RATE)),
            'PROFPASSATA',
            '0',
            'PROF_PASS_OTT',
            'TIPOSVUOT',
            '1',
            'V1_X',
            '{x} I'.format(x=self.formatAsCoordinate(data=pocket.newOrigin.x)),
            'V1_Y',
            '{y} I'.format(y=self.formatAsCoordinate(data=pocket.newOrigin.y)),
            'R',
            '{radius}'.format(radius=self.formatAsCoordinate(data=pocket.diameter / 2.0)),
            'ANG',
            '0',
            'ANG_DEFL',
            '0',
            'VERSSVUOT',
            '0',
            '\\SVUOTAMENTO',
            '',
        ))

    def createVerticalPocket(self, **keywords):
        description = self.getLocalizedDescription(5, 'Pocket')
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonPocket(description=description, **keywords)
        )

    def createVerticalPocketPiercing(self, **keywords):
        description = self.getLocalizedDescription(6, 'Piercing Pocket')
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonPocket(description=description, **keywords)
        )

    def createHorizontalPocket(self, **keywords):
        plane, _positionOnPlane = self.__calculatePlane(**keywords)

        description = self.getLocalizedDescription(5, 'Pocket')
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=plane.sideID),
            self.createCommonPocket(description=description, **keywords)
        )

    # Contours
    def createContourFirstCornerPolylineVertex(self, **keywords):
        machining = keywords['machining']
        vertex = keywords['startVertex']

        executionIndex = machining.index

        x, y = vertex.position.asTuple()
        # z = keywords['zValue']

        return '\n'.join((
            'POL',
            'INDEX',
            '{index}'.format(index=executionIndex),
            'LINKS',
            '1',
            'L',
            '2',
            'DIAM_UT',
            '0',
            'X',
            '{x} I'.format(x=self.formatAsCoordinate(data=x)),
            'Y',
            '{y} I'.format(y=self.formatAsCoordinate(data=y)),
            'Z',
            '0 0',  # Starting point Z + 0 or 1. 0=z not used by milling, 1=z used by milling.
            'XEXPR',
            '',
            'YEXPR',
            '',
            'ZEXPR',
            '',
        ))

    def createCommonSegment(self, **keywords):
        position = keywords['position']

        return '\n'.join((
            'V',
            'X',
            '{x} I'.format(x=self.formatAsCoordinate(data=position.x)),
            'Y',
            '{y} I'.format(y=self.formatAsCoordinate(data=position.y)),
            'Z',
            '0 0',
            'XEXPR',
            '',
            'YEXPR',
            '',
            'ZEXPR',
            '',
            'B',
            '{bulge}'.format(bulge=keywords['bulge']),  # Bulge = tan(ang/4) for arcs, 0 for segments
            'PARTDEF',
            '0',
            'S',
            '0',
            '\\V',
        ))

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        position = segment.getEndVertex().position

        return self.createCommonSegment(position=position, bulge=0, **keywords)

    # Copy of interiorcad.ncexport.dxf.DXFExport
    def getBulge(self, startPosition, endPosition, radius, isBigAngle, bulgeFactor):
        # For the math, see the formulas of the drawing under "Version 1" of http://lee-mac.com/bulgeconversion.html
        calcAngle = (endPosition - startPosition).magnitude() / radius / 2
        if calcAngle < -1:
            calcAngle = -1
        elif calcAngle > 1:
            calcAngle = 1
        angle = 2 * math.asin(calcAngle)
        if isBigAngle:
            angle = 2 * math.pi - angle
        bulge = math.tan(angle / 4)
        return bulge * bulgeFactor

    def createRadiusSegment(self, **keywords):
        CW = -1, False
        CCW = 1, False
        CW180 = -1, True
        CCW180 = 1, True

        segment = keywords['segment']
        bulgeFactor, isBigAngle = segment.getDirection(CW, CCW, CW180, CCW180)
        startPosition = segment.getStartVertex().position
        endPosition = segment.getEndVertex().position
        radius = segment.getRadius()

        bulge = self.getBulge(startPosition, endPosition, radius, isBigAngle, bulgeFactor)

        return self.createCommonSegment(position=endPosition, bulge=bulge, **keywords)

    def __createCommonStartBlock(self, **keywords):
        machining = keywords['machining']

        self.lastSideMainIndex += 1
        machining.index = self.lastSideMainIndex

        machining.originalStartPoint = keywords['segmentList'][0].getStartVertex().position

        approachMode = keywords['approachMode']
        if approachMode == self.APPROACH_MODE_ARC_PIERCING:
            _isClockwise, _approachDelta, approachPoint, _centerPoint, _tangentialAngle = createArcApproach(left=keywords['radiusCompensation'] == self.RADIUS_COMPENSATION_LEFT, **keywords)
            machining.approachPoint = approachPoint
        elif approachMode == self.APPROACH_MODE_LINEAR_FLYING and keywords['isClosed']:
            machining.slopeLength = keywords['securityDistance']

        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createContourFirstCornerPolylineVertex(**keywords),
        )

    def __createCommonEndBlock(self, **keywords):
        return '\n'.join((
            'C',
            '{isClosed}'.format(isClosed=int(keywords['isClosed'])),
            'P0DEF',
            '0',
            '\\POL',
            '',
        ))

    def getSizingRadiusCompensation(self, **keywords):
        return self.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else self.RADIUS_COMPENSATION_LEFT

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=self.RADIUS_COMPENSATION_NONE, left=self.RADIUS_COMPENSATION_LEFT, right=self.RADIUS_COMPENSATION_RIGHT)

    def createSizingStartBlock(self, **keywords):
        return self.__createCommonStartBlock(
            approachMode=self.getSizingApproachMode(**keywords),  # TODO IT-259: do we need a special case for flying approach like in interiorcad.ncexport.woodwop.WoodWOPExport.createSizingContourBlock  -- NG
            radiusCompensation=self.getSizingRadiusCompensation(sizing=keywords['surface'].sizing, **keywords),
            securityDistance=self.SIZING_APPROACH_DEPARTURE_DISTANCE,
            isClosed=True,
            **keywords)

    def createSizingEndBlock(self, **keywords):
        return self.__createCommonEndBlock(isClosed=True, **keywords)

    def createContourStartBlock(self, **keywords):
        contour = keywords['contour']
        return self.__createCommonStartBlock(
            approachMode=self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords),
            radiusCompensation=self.getWopRadiusCompensation(**keywords),
            securityDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
            isClosed=contour.isClosed(),
            **keywords)

    def createContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(isClosed=keywords['contour'].isClosed(), **keywords)

    def createRabbetContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(
            approachMode=self.APPROACH_MODE_CONTOUR_DEFAULT,  # FIXME IT-259: which to use?
            radiusCompensation=self.getWopRadiusCompensation(**keywords),
            securityDistance=self.getDadoRabbetSecurityDistance(**keywords),
            isClosed=False,
            **keywords)

    def createRabbetContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(isClosed=False, **keywords)

    def createDadoContourStartBlock(self, **keywords):
        dado = keywords['dado']

        radiusCompensation = self.getDadoRadiusCompensation(**keywords)
        approachMode = self.APPROACH_MODE_LINEAR_FLYING if dado.isInsetBoth() else self.APPROACH_MODE_WITHOUT  # FIXME IT-259: right?

        return self.__createCommonStartBlock(
            approachMode=approachMode,
            radiusCompensation=radiusCompensation,
            securityDistance=self.getDadoRabbetSecurityDistance(**keywords),
            isClosed=dado.isInsetBoth(),
            **keywords)

    def createDadoContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(isClosed=keywords['dado'].isInsetBoth(), **keywords)

    def createCommonContour(self, **keywords):
        machining = keywords['machining']

        executionIndex = machining.index

        startPoint = machining.originalStartPoint

        hasApproachPoint = hasattr(machining, 'approachPoint')
        hasSlopeLength = hasattr(machining, 'slopeLength')

        depthPerStep = keywords['depthPerStep'] if 'depthPerStep' in keywords else 0.0
        feedRate = keywords['feedRate'] if 'feedRate' in keywords and keywords['feedRate'] else self.ROUTER_FEED_RATE

        return joinIfNotNone('\n',
            'FRES',
            'ENT',
            '{index}'.format(index=executionIndex),
            'ENTORIG',
            '{index}'.format(index=executionIndex),
            'DESCRIZIONE',
            '{description}'.format(description=keywords['description']),
            'NUMUTENSILE',
            '{toolID}'.format(toolID=machining.toolID),
            'VELROTAZIONE',
            '22',
            'VELENTRATA',
            '5',
            'PROFLAVORAZIONE',
            '{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'PROFLAVORAZIONEEXPR',
            '',
            'ANGTILTING',
            '90',
            'ANGINDEX',
            '0',
            'ORDER',
            '0',
            'SEQUENZA',
            '0',
            'SPESS',
            '{thickness}'.format(thickness=self.formatAsCoordinate(data=keywords['surface'].thickness)),
            'ALT_CUFFIA',
            '0',
            'IDPEZZONESTING',
            '0',
            'ENABLEDEXPR',
            '',
            'VELFRESATURA',
            '{feedRate}'.format(feedRate=self.formatAsCoordinate(data=feedRate)),
            'PUNTOPARTENZAX',
            '{x} I'.format(x=self.formatAsCoordinate(data=startPoint.x)),
            'PUNTOPARTENZAY',
            '{y} I'.format(y=self.formatAsCoordinate(data=startPoint.y)),
            'INIZRACX' if hasApproachPoint else None,
            '{x}'.format(x=self.formatAsCoordinate(data=machining.approachPoint.x)) if hasApproachPoint else None,
            'INIZRACY' if hasApproachPoint else None,
            '{y}'.format(y=self.formatAsCoordinate(data=machining.approachPoint.y)) if hasApproachPoint else None,
            'LUNGPENDENZA' if hasSlopeLength else None,
            '{distance}'.format(distance=self.formatAsCoordinate(data=machining.slopeLength)) if hasSlopeLength else None,
            'NONGENERAQ' if 'contourDistance' not in keywords or isNearlyEqual(keywords['contourDistance'], 0.0) else None,
            'VERSOPERC',
            '{clock:d}'.format(clock=keywords['isClockwise']),
            'RACCORDO',
            '0',
            'COMPENSAZ',
            '{rc}'.format(rc=keywords['radiusCompensation']),
            'RAGGIOCOMP',
            '{distance}'.format(distance=self.formatAsCoordinate(data=keywords['contourDistance'] if 'contourDistance' in keywords else 0.0)),
            'PROFPASSATA',
            '{depthPerStep}'.format(depthPerStep=self.formatAsCoordinate(data=depthPerStep)),
            'SORMONTO',
            '0',
            'PROFINIZ',
            '0',
            'OUTPUT_COMPIL',
            '0',
            'ANG_DEFL',
            '0',
            'ISO',
            '',
            'PROFILO_UTE',
            'STANDARD',
            'RIFDELPROF',
            '0',
            'QMAGGIORATAZ',
            '0',
            'QAVVICINXY',
            '0',
            'TCHIUPRESS',
            '0',
            'PROFILATURA',
            '0',
            '\\FRES',
            ''
        )

    def createSizingContour(self, **keywords):
        return self.createSizingContourExecution(**keywords)

    def createSizingContourBlock(self, **keywords):
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonContour(
                description=self.getLocalizedDescription(7, 'Sizing'),
                depth=self.getSizingZValue(**keywords),
                radiusCompensation=self.getSizingRadiusCompensation(**keywords),
                isClockwise=not self.isSizingMillingDirectionLeft(**keywords),
                machining=keywords['surface'].sizing,
                **keywords),
        )

    def createVerticalContour(self, **keywords):
        contour = keywords['contour']

        depthPerStep = contour.currentDepthPerStep

        # TODO CONTOUR2.0 IT-238/IT-260 : should this be a generalised function? -- NG
        useSteps = contour.depth > depthPerStep > 0.0 and not isNearlyEqual(depthPerStep, 0.0) and not isNearlyEqual(contour.depth, depthPerStep)
        if useSteps:
            numberOfSteps = math.ceil(contour.depth / depthPerStep)
            rawDepthStep = contour.depth / numberOfSteps
            assert 0 < rawDepthStep <= depthPerStep or isNearlyEqual(rawDepthStep, depthPerStep), "Calculated step should not be 0 and not be bigger than maximum step!"
            depthPerStep = self.getContourZValue(depth=rawDepthStep, **keywords)
        else:
            depthPerStep = None

        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonContour(
                description=self.getLocalizedDescription(8, 'Contour'),
                depth=self.getContourZValue(depth=contour.depth, **keywords),
                radiusCompensation=self.getWopRadiusCompensation(**keywords),
                isClockwise=bool(contour.path.isClockwise()),  # isClockwise returns None for just a line (VWI-28896, IC-22050). Masterworks seems to prefer 0
                depthPerStep=depthPerStep,
                feedRate=contour.currentFeedRate,
                **keywords),
        )

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createVerticalContour, **keywords))

    def createVerticalRabbet(self, **keywords):
        rabbet = keywords['rabbet']
        radiusCompensation = self.getWopRadiusCompensation(**keywords)
        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonContour(
                description=self.getLocalizedDescription(9, 'Rabbet'),
                depth=self.getContourZValue(depth=rabbet.depth, **keywords),
                radiusCompensation=radiusCompensation,
                isClockwise=radiusCompensation == self.RADIUS_COMPENSATION_RIGHT,
                **keywords),
        )

    def createNCStop(self, **_keywords):
        return ''

    def createVerticalDadoWithRouter(self, **keywords):
        radiusCompensation = self.getDadoRadiusCompensation(**keywords)

        return joinIfNotNone('\n',
            self.changeWorkplane(plane=self.SIDE_MAIN),
            self.createCommonContour(
                description=self.getLocalizedDescription(10, 'Dado'),
                radiusCompensation=radiusCompensation,
                isClockwise=radiusCompensation == self.RADIUS_COMPENSATION_RIGHT,
                **keywords),
        )

    def createVerticalDadoWithSawAlongXAxis(self, **_keywords):
        return ''

    def createVerticalDadoWithSawAlongYAxis(self, **_keywords):
        return ''

    def createVerticalDadoWithSawCRotated(self, **_keywords):
        return ''

    # Miter
    def orientMiters(self, **keywords):
        pass

    # NC Macro
    def createNCMacro(self, **keywords):
        pass

    # Footer
    def createEndOfFile(self, **_keywords):
        return 'EOF' + '\n'

    def getLocalizedDescription(self, index, default):
        return getLocalizedString(MASTERWORK_DESCRIPTIONS_RESOURCEGROUP, index, default)
