# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import logging
import math

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.localization import getLocalizedString, NCHOPSPROGRAMVARIABLES_RESOURCEGROUP
from interiorcad.backend.postprocessors.surfacepostprocessor.addsurfacedeltatodados import AddSurfaceDeltaToDados
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual, isNearlyGreaterOrEqual
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.surface.acangles import normalizeAngle
from interiorcad.interface.surface.contour2d import Contour2D
from interiorcad.interface.surface.machiningplanes import MachiningPlanes
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.ncexport.base import NCExportBase


class NCHopsExport(NCExportBase):
    EXPORT_FORMAT = 'NCHops'

    LAST_Z_POSITION = '_Z'

    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    APPROACH_NONE = 0
    APPROACH_LINEAR = 1
    APPROACH_RADIAL = 2
    APPROACH_SIDE = 3
    APPROACH_LINEAR_FLYING = -1
    APPROACH_RADIAL_FLYING = -2
    APPROACH_SIDE_FLYING = -3

    APPROACH_MODE_NONE = APPROACH_NONE
    APPROACH_MODE_LINEAR_PIERCING = APPROACH_LINEAR
    APPROACH_MODE_LINEAR_FLYING = APPROACH_LINEAR_FLYING
    APPROACH_MODE_ARC_PIERCING = APPROACH_RADIAL
    APPROACH_MODE_ARC_FLYING = APPROACH_RADIAL_FLYING
    APPROACH_MODE_WITHOUT = APPROACH_NONE
    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_LINEAR_PIERCING

    def __init__(self, **keywords):
        super().__init__(**keywords)

        # NC-Hops has a convention for length, width, thickness (the German ones are the same as the English one), but they could also be individual
        machineConfiguration = self.getMachineConfiguration()
        self.INDIVIDUAL_PART_DIMENSION_VARIABLES = machineConfiguration.get('individual-part-dimension-variables', '')
        self.prepareLocalizationOfPartDimensionVariables(**keywords)
        self.VAR_LENGTH = self.localizedPartDimensionVariableNames['LENGTH']
        self.VAR_WIDTH = self.localizedPartDimensionVariableNames['WIDTH']
        self.VAR_THICKNESS = self.localizedPartDimensionVariableNames['THICKNESS']
        # Careful: NPVX, NPVY are special variables in NCHOPS that define the offset
        self.VAR_ORIGINX = self.getLocalizedVariable(4, 'NPVX')
        self.VAR_ORIGINY = self.getLocalizedVariable(5, 'NPVY')

        self.ZERO_OFFSET_Y_IS_SIZING_ADDITION = machineConfiguration.getAsBool('zero-offset-y-is-sizing-addition', True)
        self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR = machineConfiguration.getAsBool('create-sizing-with-three-side-contour', False)
        self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS = machineConfiguration.getAsBool('sizing-mill-through-for-non-rectangular-parts', True)
        self.SIZING_FIRST_PASS_REMAINING_THICKNESS = machineConfiguration.getAsFloat('sizing-first-pass-remaining-thickness', 2)
        self.SIZING_PASS_STRATEGY = machineConfiguration.get('sizing-pass-strategy', 'standard')
        self.SIZING_PASS_STRATEGY_AREA_THRESHOLD = machineConfiguration.getAsFloat('sizing-pass-strategy-area-threshold', 0.1)

        self.VERSION = machineConfiguration.getAsInt('nc-hops-version', 7)
        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 1)
        self.MITER_SAW_DEPTH_FROM_BOTTOM = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_SAW_DEPTH_FROM_BOTTOM_KEY, 3)
        self.MITER_PRE_SAW_OFFSET = machineConfiguration.getAsFloat('miter-pre-saw-offset', 0.05)
        self.HOOD_POSITIONING_MACRO = machineConfiguration.get('hood', None)
        self.HOOD_POSITION_MAXIMUM = machineConfiguration.getAsInt('hood-position-maximum', 9)
        self.USE_FINISHED_PART = machineConfiguration.getAsBool('use-finished-part', True)
        self.DADO_MODE = machineConfiguration.getAsInt('dado-mode', 0)  # Options: 0 = Synchronous rotating / 1 = Contra rotating
        self.USE_RADIAL_POCKET_DEPARTURE = machineConfiguration.get('use-radial-pocket-departure', 'always')  # never | large pockets | always
        self.CONTOUR_CORNER_RADIUS = machineConfiguration.get('contour-corner-radius', '0')  # I's normally real, but it also could be used with a string variable

        self.CLEAR_OUT_POCKET_OVERLAP_FACTOR = machineConfiguration.getAsInt('clear-out-pocket-overlap-factor', 35)  # "A minimum of 35% overlapping is recommended."

        self.PARKING_POSITION_SWITCH = machineConfiguration.getAsFloat('parking-position-switch', 1750)
        self.PARKING_POSITION_INTEGRATION = machineConfiguration.getAsBool('parking-position-integration', False)

        self.BETTER_NEST_OPTIONS = machineConfiguration.getAsBool('better-nest-options', False)
        self.USE_CUSTOM_USER_VARIABLES = machineConfiguration.getAsBool('use-custom-user-variables', False)
        self.CREATE_LABEL_CALL = machineConfiguration.get('create-label-call', '')
        self.CUSTOM_USER_SIZING_MACRO = machineConfiguration.get('custom-user-sizing-macro', '')

        self.createZValueConverters(**keywords)

        self.checkConfiguration(**keywords)

    def checkConfiguration(self, **_keywords):
        if self.CREATE_SIZING_WITH_MACRO and self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR:
            logging.warning(self.getLocalizedMessage(73, "Simultaneous sizing with three-sided contour and with macro is not possible."))
            self.CREATE_SIZING_WITH_MACRO = ''

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMiterPreSawConverterUpToVersion4 = self.createZValueConverter('miter-pre-saw-up-to-version-4-z-reference', ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMiterPreSawConverterFromVersion5 = self.createZValueConverter('miter-pre-saw-from-version-5-z-reference', ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)

    def getIndividualPartDimensionVariables(self):
        individualPartDimensionVariables = self.INDIVIDUAL_PART_DIMENSION_VARIABLES.split()
        self.localizedPartDimensionVariableNames = {
            'LENGTH': individualPartDimensionVariables[0],
            'WIDTH': individualPartDimensionVariables[1],
            'THICKNESS': individualPartDimensionVariables[2],
        }

    def getLocalizedPartDimensionVariables(self):
        LOCALIZED_PART_DIMENSION_VARIABLES = (
            ('LENGTH', 1, 'LPX'),
            ('WIDTH', 2, 'LPY'),
            ('THICKNESS', 3, 'LPZ'),
        )

        self.localizedPartDimensionVariableNames = {}
        for name, index, default in LOCALIZED_PART_DIMENSION_VARIABLES:
            self.localizedPartDimensionVariableNames[name] = self.getLocalizedVariable(index, default)

    def prepareLocalizationOfPartDimensionVariables(self, **_keywords):
        if self.INDIVIDUAL_PART_DIMENSION_VARIABLES:
            self.getIndividualPartDimensionVariables()
        else:
            self.getLocalizedPartDimensionVariables()

    def createMacroNameVersionAdditionBlank(self):
        if self.VERSION >= 5:
            return '_V5'
        else:
            return ''

    def createMacroNameVersionAdditionV4(self):
        if self.VERSION >= 5:
            return '_V5'
        else:
            return '_V4'

    def getSimplifiedGrainDirection(self, **keywords):
        surface = keywords['surface']
        grain = surface.grain

        if grain == self.GRAIN_NONE:
            return 0
        elif grain == self.GRAIN_LENGTHWISE:
            return 1
        else:  # self.GRAIN_CROSSWISE
            return 2

    def getRotationAngle(self, **keywords):
        surface = keywords['surface']
        grain = surface.grain

        if grain == self.GRAIN_LENGTHWISE or grain == self.GRAIN_CROSSWISE:
            return 180
        elif grain == self.GRAIN_NONE and surface.workpieceFinishedSizeShape.isRectangular():
            return 90
        elif grain == self.GRAIN_NONE and not surface.workpieceFinishedSizeShape.isRectangular():  # isPolygonal() as well as not CornerVertex()
            return 5
        else:
            return 0

    def getWorkpieceCommentProperties(self, **keywords):
        surface = keywords['surface']

        dictionary = super().getWorkpieceCommentProperties(**keywords)
        dictionary['IsRectangular'] = self.asInt(data=surface.workpieceFinishedSizeShape.isRectangular())
        dictionary['GrainDirectionNCHops'] = self.asInt(data=self.getSimplifiedGrainDirection(**keywords))
        dictionary['StepAngleNCHops'] = self.asInt(data=self.getRotationAngle(**keywords))

        return dictionary

    # header
    def createHead(self, **keywords):
        self.currentPlane = None
        self.lastContourDefinitionIndex = 0
        self.previousToolID = 'firstToolIDCall'
        self.previousFeedRate = None
        self.previousToolIDForHood = 'firstHoodCall'
        surface = keywords['surface']

        self.predefinedMachiningPlanes = MachiningPlanes(sideIDGenerator=lambda: None)
        self.predefinedMachiningPlanes.planes.extend(self.__createPredefinedPlanes(**keywords))

        return joinIfNotNone('',
            self.createHeader(**keywords),
            self.createNCWorkpiece(**keywords),
            self.createWorkpieceDefinition(**keywords),
            self.createSizingContourDefinition(**keywords) if self.VERSION >= 5 else None,
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords) if self.VERSION >= 5 else None,
            self.createRabbetContourDefinitions(**keywords) if self.VERSION >= 5 else None,
            self.createDadoContourDefinitions(dados=surface.getAllDados(), **keywords) if self.VERSION >= 5 else None,
        )

    def createSizingContour(self, **keywords):
        radiusCompensation = self.__getSizingRadiusCompensation(**keywords)
        if self.VERSION >= 5:
            return self.createSizingContourExecution(radiusCompensation=radiusCompensation, **keywords)
        else:
            approach = self.getSizingApproachMode(**keywords)
            return self.createSizingContourDefinitionAsExecution(approach=approach, departure=approach, radiusCompensation=radiusCompensation, **keywords)

    def getNCFileExtension(self, **_keywords):
        return '.hop'

    def getPostPostProcessors(self, **keywords):
        return AddSurfaceDeltaToDados(),

    def createHeader(self, **keywords):
        surface = keywords['surface']
        fileName = self.createFileName(name=surface.workpieceName, extension='', **keywords)
        return '\n'.join((
            ';WZGV={}'.format(self.getMachineConfiguration().get('wzgv')),
            ';MASCHINE={}'.format(self.getMachineConfiguration().get('machine')),
            ';NCNAME={}'.format(fileName),
            ';DX=0',
            ';DY=0',
            ';DZ=0',
            '',
            '',
        ))

    def customUserVariableFileDefaultText(self, **_keywords):
        return '\n'.join((
            '\tWKZ_B := 53300;Bohrer für senkrecht (Fläche)',
            '\tWKZ_B_H := 53300;Bohrer für waagrecht (Seite)',
            '\tWKZ_B_S := 53300;Bohrer für Schräge',
            '\tWKZ_F := 43300;Nutrad für senkrecht (Fläche)',
            '\tWKZ_F_H := 43100;Nutrad für waagrecht (Seiten)',
            '\tWKZ_F_S := 43100;Nutrad für Schräge',
            '\tWKZ_FINGER := 139;Fingerfräser exklusiv für senkrecht (Fläche)',
        ))
        # Actually default tools for NC-Macro 3D of new Clamex_Scheibe from direkt cnc

    def createCustomUserVariables(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='NC-Hops Variables.txt',
            customUserVariableFileDefaultText=self.customUserVariableFileDefaultText(**keywords),
            **keywords)

    def createNCWorkpiece(self, **keywords):
        surface = keywords['surface']

        return joinIfNotNone('\n',
            'VARS',
            '\t{} := {};*VAR*'.format(self.VAR_LENGTH, self.formatAsCoordinate(data=surface.getFinishedSizeX())),  # if self.USE_FINISHED_PART else surface.sawnSize.x  --mjk
            '\t{} := {};*VAR*'.format(self.VAR_WIDTH, self.formatAsCoordinate(data=surface.getFinishedSizeY())),  # if self.USE_FINISHED_PART else surface.sawnSize.y  --mjk
            '\t{} := {};*VAR*'.format(self.VAR_THICKNESS, self.formatAsCoordinate(data=surface.thickness)),
            '\t{} := {};*VAR*'.format(self.VAR_ORIGINX, self.formatAsCoordinate(data=surface.getAdditionX())),
            '\t{} := {};*VAR*'.format(self.VAR_ORIGINY, self.formatAsCoordinate(data=surface.getAdditionY()) if self.ZERO_OFFSET_Y_IS_SIZING_ADDITION else 0),
            self.createCustomUserVariables(**keywords) if self.USE_CUSTOM_USER_VARIABLES else None,
            '',
            '',
        )

    def __createFinishedPartOptionalParamaters(self, **_keywords):
        if self.VERSION >= 5:
            return ','.join((
                '0',
                '{ox}'.format(ox=self.VAR_ORIGINX),
                '{oy}'.format(oy=self.VAR_ORIGINY),
                '0',
                "''",
                '0',
                '0',
                '0',
            ))
        else:
            return '3'

    def createFinishedPart(self, **keywords):
        return joinIfNotNone(',',
            'Fertigteil ({l}'.format(l=self.VAR_LENGTH),
            '{w}'.format(w=self.VAR_WIDTH),
            '{t}'.format(t=self.VAR_THICKNESS),
            '0',
            self.__createFinishedPartOptionalParamaters(**keywords) + ')',
        )

    def createRawPart(self, **_keywords):
        return '\n'.join((
            joinIfNotNone(',',
                'Rohteil ({l}+2*{ox}'.format(l=self.VAR_LENGTH, ox=self.VAR_ORIGINX),
                '{w}+2*{oy}'.format(w=self.VAR_WIDTH, oy=self.VAR_ORIGINY),
                '{t}'.format(t=self.VAR_THICKNESS),
                '0',
                '3' + ')',
                ),
            joinIfNotNone(',',
                'SetUrsprung ({ox}'.format(ox=self.VAR_ORIGINX),
                '{oy}'.format(oy=self.VAR_ORIGINY) + ')',
                ),
        ))

    def createZTextOptionalParameter(self):
        if self.VERSION >= 7:
            return ',0,1'
        elif self.VERSION >= 5:
            return ',0'
        else:
            return ''

    def createLabelCall(self, **keywords):
        if self.CREATE_LABEL_CALL:
            return '\n' + '{call}'.format(call=self.getMappedCustomUserVariablesFromStringWithKeys(stringWithKeys=self.CREATE_LABEL_CALL, **keywords))

    def createZRechteckOptionalParameter(self):
        if self.VERSION >= 7:
            return ',0,0,0,1'
        elif self.VERSION >= 5:
            return ',0'
        else:
            return ''

    def createWorkpieceComment(self, **keywords):
        projectInfo = self.getProjectInfo()
        workpieceComments = self.getWorkpieceComments(**keywords)
        workpieceComments['ZTextOptParameter'] = self.createZTextOptionalParameter()
        workpieceComments['ZRechteckOptParameter'] = self.createZRechteckOptionalParameter()

        return joinIfNotNone('\n',
            "ZText ('{WorkpieceName}',_rx/2,_ry/2,50,0,9,0,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments),
            "ZText ('{CommentLabel}: {Comment}',_rx,_ry+120-30-10,20,0,5,255,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if workpieceComments['Comment'] else None,
            "ZText ('{ItemIDLabel}: {ItemID}',0,_ry+50,30,0,1,255,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments),
            "ZText ('{Edge1}',_rx/2,-25,20,0,9,0,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if workpieceComments['Edge1'] else None,  # front
            "ZText ('{Edge2}',_rx/2,_ry+20,20,0,9,0,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if workpieceComments['Edge2'] else None,  # back
            "ZText ('{Edge3}',-30,_ry/2,20,90,9,0,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if workpieceComments['Edge3'] else None,  # left
            "ZText ('{Edge4}',_rx+30,_ry/2,20,90,9,0,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if workpieceComments['Edge4'] else None,  # right
            "ZText ('{CabinetNameLabel}: {CabinetName}',10,-60,20,0,1,255,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if workpieceComments['CabinetName'] else None,
            "ZText ('{ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}',10,-60-30,20,0,1,255,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if ExportHelper.isProjectIDValid(self, **projectInfo) else None,
            "ZText ('{PositionNumberLabel}: {PositionNumber} {PositionDescription}',10,-60-60,20,0,1,255,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments) if ExportHelper.isProjectIDValid(self, **projectInfo) else None,
            "ZText ('{CountLabel}: {Count}', 10, -60-90,20,0,1,255,'',0,0,0,0{ZTextOptParameter})".format_map(workpieceComments),
            "ZRechteck (-120,-140,_rx+(2*120),_ry+(2*120),0,0{ZRechteckOptParameter})".format_map(workpieceComments),
        )

    def createLoadingDeviceIfAvailable(self, **_keywords):
        loadingdevice = self.getMachineConfiguration().get('loading-device', None)
        return 'CALL {} (VAL LINIE:=2 )'.format(loadingdevice) if loadingdevice else None

    def createLaserMeasuringIfActivated(self, **_keywords):
        laserMeasuring = self.getMachineConfiguration().getAsBool('laser-measuring', False)
        return 'CALL HH_MeaQuote ( VAL X:=25)' if laserMeasuring else None

    def createBetterNestTrennerInnenAussen(self, **_keywords):
        return 'CALL BN_TrennerInnenAussen ()' if self.BETTER_NEST_OPTIONS else None

    def createWorkpieceDefinition(self, **keywords):
        return joinIfNotNone('\n',
            'START',
            self.createFinishedPart(**keywords) if self.USE_FINISHED_PART else self.createRawPart(**keywords),
            self.createWorkpieceComment(**keywords),
            self.createLabelCall(**keywords),
            self.createLoadingDeviceIfAvailable(**keywords),
            self.createLaserMeasuringIfActivated(**keywords),
            self.createBetterNestTrennerInnenAussen(**keywords),
            '',
            '',
        )

    # Machinings

    SIDE_MAIN, SIDE_FRONT, SIDE_LEFT, SIDE_BACK, SIDE_RIGHT = range(0, 5)
    SIDE_FREE = 'F'

    def createToolChange(self, **keywords):
        toolID = keywords['toolID']
        toolType = keywords['toolType']
        currentFeedRate = '_V' if 'feedRate' not in keywords or isNearlyEqual(keywords['feedRate'], 0.0) else self.formatAsCoordinate(data=keywords['feedRate'])
        if toolID != self.previousToolID or self.previousFeedRate != currentFeedRate:
            self.previousToolID = toolID
            self.previousFeedRate = currentFeedRate
            return '{toolType} ({toolID},_VE,{feedRate},_VA,_SD,_ANF{optParameter})'.format(
                toolType=toolType,
                toolID=toolID,
                feedRate=currentFeedRate,
                optParameter=self.toolChangeOptionalParameter()
            )

    def toolChangeOptionalParameter(self):
        if self.VERSION >= 7:
            return ",'1'"
        elif self.VERSION >= 5:
            return ",''"
        else:  # VERSION <= 4
            return ""

    def createHoodPositioning(self, **keywords):
        if self.HOOD_POSITIONING_MACRO:
            toolID = keywords['toolID']
            if toolID != self.previousToolIDForHood:
                self.previousToolIDForHood = toolID
                return 'CALL {macro} ( VAL '.format(macro=self.HOOD_POSITIONING_MACRO) + joinIfNotNone(',',
                    'NEXTTC:=1' if self.VERSION > 6 else None,
                    'ALLSTEPS:=1' if self.VERSION > 6 else None,
                    'POS:={position}'.format(position=keywords['hoodPosition']),
                ) + ')'

    def createHoodPositioningWithToolChange(self, **keywords):
        return joinIfNotNone('\n',
            self.createHoodPositioning(**keywords) if self.VERSION <= 6 else None,
            self.createToolChange(**keywords),
            self.createHoodPositioning(**keywords) if self.VERSION > 6 else None,
        )

    def __getDepthOfVerticalDrilling(self, **keywords):
        ### IT-131: a correct ZValueConverter should be used here -- NG
        return -keywords['depth']

    def __createCommonVerticalDrillingOptionalParameter(self):
        if self.VERSION >= 5:
            return '0,0,0,0,0'

    def __createCommonVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']

        return joinIfNotNone('\n',
            self.createToolChange(toolType='WZB', toolID=drilling.toolID),
            self.changeView(plane=None, **keywords),
            'Bohrung (' + joinIfNotNone(',',
                '{x}'.format(x=self.formatAsCoordinate(data=drilling.position.x)),
                '{y}'.format(y=self.formatAsCoordinate(data=drilling.position.y)),
                '{d}'.format(d=self.formatAsCoordinate(data=drilling.installedDiameter)),
                '{z}'.format(z=self.formatAsCoordinate(data=self.__getDepthOfVerticalDrilling(**keywords))),
                '{type}'.format(type=self.getDrillType(**keywords)),
                '0',
                '0',
                self.__createCommonVerticalDrillingOptionalParameter(),
            ) + ')',
            '',
            )

    # only used for legacy drills, replaced with property 'drilling-drill-type'
    FLAT_DRILL_TYPE = 10
    PIERCING_DRILL_TYPE = 20
    WIDE_DRILL_TYPE = 30

    WIDE_DRILLING_DIAMETER_THRESHOLD = 14  # Typischerweise: ab 15 mm Forstner-Bohrer, bis 14 mm Spiral-Bohrer

    def createVerticalDrilling(self, **keywords):
        return self.__createCommonVerticalDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)

    def createVerticalDrillingPiercing(self, **keywords):
        return self.__createCommonVerticalDrilling(defaultDrillType=self.PIERCING_DRILL_TYPE, **keywords)

    def __createVerticalGridDrillingAlongXAxisOptionalParameters(self):
        if self.VERSION >= 5:
            return ','.join((
                'ESXY:=1',
                'ESX:=1',
                'ESY:=5',
                'ESD:=0',
                'USE2:=0',
            ))

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)
        return joinIfNotNone('\n',
            self.createToolChange(toolType='WZB', toolID=gridDrilling.getToolID()),
            self.changeView(plane=None, **keywords),
            'CALL _Rbohx_einpass{version} ( VAL '.format(version=self.createMacroNameVersionAdditionBlank()) + joinIfNotNone(',',
                'EINPASSEN:=0',
                'SPX:={x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
                'SPY:={y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
                'SPY2:={y}'.format(y=self.formatAsCoordinate(data=0)),
                'RASTER:={distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                'EPX:={x}'.format(x=self.formatAsCoordinate(data=endDrilling.position.x)),
                'T:={depth}'.format(depth=self.__getDepthOfVerticalDrilling(**keywords)),
                'D:={diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
                'TLF:={type}'.format(type=drillType),
                self.__createVerticalGridDrillingAlongXAxisOptionalParameters(),
            ) + ')',
            '',
        )

    def __createVerticalGridDrillingAlongYAxisOptionalParameters(self):
        if self.VERSION >= 5:
            return ','.join((
                'ESXY:=1',
                'ESX:=3',
                'ESY:=3',
                'ESD:=0',
                'USE2:=0',
            ))

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)
        return joinIfNotNone('\n',
            self.createToolChange(toolType='WZB', toolID=gridDrilling.getToolID()),
            self.changeView(plane=None, **keywords),
            'CALL _Rbohy_einpass{version} ( VAL '.format(version=self.createMacroNameVersionAdditionBlank()) + joinIfNotNone(',',
                'EINPASSEN:=0',
                'SPX:={x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
                'SPX2:={x}'.format(x=self.formatAsCoordinate(data=0)),
                'SPY:={y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
                'RASTER:={distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                'EPY:={y}'.format(y=self.formatAsCoordinate(data=endDrilling.position.y)),
                'T:={depth}'.format(depth=self.__getDepthOfVerticalDrilling(**keywords)),
                'D:={diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
                'TLF:={type}'.format(type=drillType),
                self.__createVerticalGridDrillingAlongYAxisOptionalParameters(),
            ) + ')',
            '',
        )

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        startDrilling = gridDrilling.getStartDrilling()
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)
        return joinIfNotNone('\n',
            self.createToolChange(toolType='WZB', toolID=gridDrilling.getToolID()),
            self.changeView(plane=None, **keywords),
            'CALL _Rbohr_winkel{version} ( VAL '.format(version=self.createMacroNameVersionAdditionBlank()) + joinIfNotNone(',',
                'SPX:={x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
                'SPY:={y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
                'WINKEL:={angle}'.format(angle=self.formatAsAngle(data=normalizeAngle(gridDrilling.getXYAngle()))),
                'RASTER:={distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                'ANZAHL:={count}'.format(count=self.formatAsInt(data=gridDrilling.countDrillings)),
                'T:={depth}'.format(depth=self.__getDepthOfVerticalDrilling(**keywords)),
                'D:={diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
                'TLF:={type}'.format(type=drillType),
                self.__createVerticalGridDrillingAlongCRotatedAxisOptionalParameters(),
            ) + ')',
            '',
            )

    def __createVerticalGridDrillingAlongCRotatedAxisOptionalParameters(self):
        if self.VERSION >= 5:
            return ','.join((
                'ABSTAND:=0',
                'ESXY:=1',
                'ESD:=0',
            ))

    # pockets
    POCKET_NON_PIERCING, POCKET_PIERCING = range(2)
    POCKET_NON_PIERCING_DEPART_WITHOUT, POCKET_NON_PIERCING_DEPART_LINEAR, POCKET_NON_PIERCING_DEPART_RADIAL = range(3)
    POCKET_PIERCING_DEPART_WITHOUT, POCKET_PIERCING_DEPART_RADIAL = range(2)

    def createVerticalPocket(self, **keywords):
        return self.createVerticalPocketCode(kind=self.POCKET_NON_PIERCING, **keywords)

    def createVerticalPocketPiercing(self, **keywords):
        return self.createVerticalPocketCode(kind=self.POCKET_PIERCING, **keywords)

    def createVerticalPocketCode(self, **keywords):
        pocket = keywords['pocket']
        toolID = pocket.toolID

        if keywords['kind'] == self.POCKET_PIERCING and pocket.diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER:
            pocketCallCode = self.createVerticalPocketPiercingCode(**keywords)
        else:
            pocketCallCode = self.createVerticalPocketNonPiercingCode(**keywords)

        return joinIfNotNone('\n',
            self.createHoodPositioningWithToolChange(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolType='WZF', toolID=toolID, **keywords),
            self.changeView(plane=None, **keywords),
            pocketCallCode,
            '',
        )

    def useRadialPocketDeparture(self, **keywords):
        pocket = keywords['pocket']
        return self.USE_RADIAL_POCKET_DEPARTURE == 'always' or (self.USE_RADIAL_POCKET_DEPARTURE == 'big pockets' and pocket.tool.getRouterDiameter() * 2 < pocket.diameter)

    def __createVerticalPocketNonPiercingCodeOptionalParameters(self, **keywords):
        if self.VERSION >= 5:
            return ','.join((
                'AB:={ab}'.format(ab=self.POCKET_NON_PIERCING_DEPART_RADIAL if self.useRadialPocketDeparture(**keywords) else self.POCKET_NON_PIERCING_DEPART_WITHOUT),
                'ESXY:={esxy}'.format(esxy=keywords['esxy'] if 'esxy' in keywords else 1),
                'ESMD:=0',
                'LASER:=0',
                'INTERPOL:=1',
            ))
        else:  # VERSION <= 4
            assert 'esxy' not in keywords or keywords['esxy'] == 1, "NC-Hops 4 does not support an explicit ESXY"
            return ','.join((
                'AB:=0',
                'INTERPOL:=0',
            ))

    def getRouterDepthPerStep(self, **keywords):
        depthPerStep = super().getRouterDepthPerStep(**keywords)

        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = keywords['machiningDepth']

        return depthPerStep

    def createVerticalPocketNonPiercingCode(self, **keywords):
        pocket = keywords['pocket']
        depth = keywords['depth']
        return ''.join((
            'CALL _Kreistasche{version} ( VAL '.format(version=self.createMacroNameVersionAdditionV4()) + joinIfNotNone(',',
                'X_MITTE:={x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
                'Y_MITTE:={y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
                'TIEFE:={depth}'.format(depth=self.formatAsCoordinate(data=-depth)),
                'ZUSTELLUNG:={dps}'.format(dps=self.formatAsCoordinate(data=-self.getRouterDepthPerStep(machiningDepth=depth, **keywords))),
                'RADIUS:={r}'.format(r=self.formatAsCoordinate(data=pocket.diameter / 2)),
                'ABF:=_ANF',
                'UMKEHREN:={direction:d}'.format(direction=not self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
                self.__createVerticalPocketNonPiercingCodeOptionalParameters(**keywords),
            ) + ')',
            '',
        ))

    def createVerticalPocketPiercingCodeMacroNameVersionAddition(self):
        if self.VERSION >= 5:
            return '_V5'
        elif self.VERSION == 4:
            return '_V4'
        else:
            return '_z'

    def __createVerticalPocketPiercingCodeOptionalParameters(self, **keywords):
        pocket = keywords['pocket']

        if self.VERSION >= 5:
            return ','.join((
                'AB:={ab}'.format(ab=self.POCKET_PIERCING_DEPART_RADIAL if self.useRadialPocketDeparture(**keywords) else self.POCKET_PIERCING_DEPART_WITHOUT),
                'AUFMASS:=0',
                'BEARB_UMKEHREN:={direction:d}'.format(direction=not self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
                'WINKEL:=180',
                'ANF:=_ANF',
                'ABF:=_ANF',
                'RAMPE:=1',
                'INTERPOL:=1',
                'ESXY:=1',
                'ESMD:=0',
                'LASER:=0',
            ))
        elif self.VERSION == 4:
            return ','.join((
                'AB:=2',
                'AUFMASS:=0',
                'BEARB_UMKEHREN:={direction:d}'.format(direction=not self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
                'WINKEL:=180',
                'ANF:=_ANF',
                'ABF:=_ANF',
                'RAMPE:=1',
                'INTERPOL:=0',
            ))
        else:
            return 'ANF_BEW:=2'

    def createVerticalPocketPiercingCode(self, **keywords):
        pocket = keywords['pocket']
        depth = keywords['depth']
        return ''.join((
            'CALL _Kreisbahn{version} ( VAL '.format(version=self.createVerticalPocketPiercingCodeMacroNameVersionAddition()) + joinIfNotNone(',',
                'X_MITTE:={x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
                'Y_MITTE:={y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
                'TIEFE:={depth}'.format(depth=self.formatAsCoordinate(data=-depth)),
                'ZUTIEFE:={dps}'.format(dps=self.formatAsCoordinate(data=-self.getRouterDepthPerStep(machiningDepth=depth, **keywords))),
                'RADIUS:={r}'.format(r=self.formatAsAngle(data=pocket.diameter / 2)),
                'RADIUSKORREKTUR:={}'.format(self.RADIUS_COMPENSATION_LEFT),
                self.__createVerticalPocketPiercingCodeOptionalParameters(**keywords),
            ) + ')',
            '',
        ))

    def createFreeView(self, **keywords):  # NC-Hops call the German "Ebene" "View" in English
        plane = keywords['plane']
        return (
            'EbeneF (' +
            ','.join((
                self.formatAsCoordinate(data=plane.origin.x),
                self.formatAsCoordinate(data=plane.origin.y),
                self.formatAsCoordinate(data=plane.origin.z),
                self.formatAsAngle(data=plane.aAngle),
                self.formatAsAngle(data=normalizeAngle(plane.cAngleToXYPlane-90)),
                '0',
                '0',
            )) + ')'
        )

    def createMainView(self, **_keywords):  # NC-Hops call the German "Ebene" "View" in English
        return 'Ebene0()'

    def selectPredefinedView(self, **keywords):  # NC-Hops call the German "Ebene" "View" in English
        return 'Ebene{sideID}()'.format(sideID=keywords['plane'].sideID)

    def changeView(self, **keywords):
        if 'plane' in keywords:
            newPlane = keywords['plane']
        else:
            newPlane = None

        if newPlane != self.currentPlane:
            self.currentPlane = newPlane
            if newPlane is None:
                return self.createMainView(**keywords)
            elif newPlane.predefined:
                return self.selectPredefinedView(**keywords)
            else:  # not newPlane.predefined
                # TODO keywords['plane'].sideID = self.SIDE_FREE here or in calculatePlane?
                return self.createFreeView(**keywords)

    def __createPredefinedPlanes(self, **keywords):
        size = keywords['surface'].getFinishedSize()

        yield Plane(  # Front
            predefined=True, sideID=self.SIDE_FRONT,
            matrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )
        yield Plane(  # Right
            predefined=True, sideID=self.SIDE_RIGHT,
            matrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
        )
        yield Plane(  # Back
            predefined=True, sideID=self.SIDE_BACK,
            matrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)  # TODO: ist x really 0.0 and not size.x
            ),
        )
        yield Plane(  # Left
            predefined=True, sideID=self.SIDE_LEFT,
            matrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']

        machiningPlane = Plane.createFromMachining(machining)

        plane = self.predefinedMachiningPlanes.findPlane(machiningPlane)
        if plane:
            machiningPlane = plane

        return machiningPlane, machining.getPositionOnPlane(machiningPlane)

    def createHorizontalPocket(self, **keywords):
        pocket = keywords['pocket']

        plane, positionOnPlane = self.__calculatePlane(**keywords)
        pocket.matrix.origin = Vector(positionOnPlane.x, positionOnPlane.y, 0.0)  # TODO IT-220 (HORIZONTAL): unschönes Design -- NG

        return joinIfNotNone('\n',
            self.createHoodPositioningWithToolChange(hoodPosition=9, toolType='WZF', toolID=pocket.toolID, **keywords),
            self.changeView(plane=plane, **keywords),
            self.createVerticalPocketNonPiercingCode(depth=pocket.depth, esxy=0, **keywords),
            ''
            '',
        )

    # rotated and five axis drillings
    def createFiveAxisDrilling(self, **keywords):
        drilling = keywords['drilling']
        cAngle = drilling.cAngle % 360  # NCHops works well with negative angles but its dialog of HorzB recognizes 270 but not -90 as being from front
        aAngle = -(90 - drilling.aAngle)
        return self.createNCMacroHorizontalDrilling(positionX=drilling.position.x, positionY=drilling.position.y, diameter=drilling.diameter, depth=-drilling.depth, aAngle=aAngle, cAngle=cAngle, **keywords)

    def createRotatedHorizontalDrilling(self, **keywords):
        drilling = keywords['drilling']
        aAngle = 0.0
        cAngle = (drilling.cAngle % 360)
        return self.createNCMacroHorizontalDrilling(depth=-drilling.depth, aAngle=aAngle, cAngle=cAngle, **keywords)

    def createNCMacroHorizontalDrillingOptionalParameter(self):
        if self.VERSION >= 5:
            return '0,0,0,0'

    def createNCMacroHorizontalDrilling(self, **keywords):
        drilling = keywords['drilling']

        return joinIfNotNone('\n',
            self.createToolChange(toolType='WZB', toolID=drilling.toolID),
            self.changeView(plane=None, **keywords),
            'HorzB (' + joinIfNotNone(',',
                '{x}'.format(x=self.formatAsCoordinate(data=drilling.position.x)),
                '{y}'.format(y=self.formatAsCoordinate(data=drilling.position.y)),
                '{z}'.format(z=self.formatAsCoordinate(data=keywords['zPosition'])),
                '{d}'.format(d=self.formatAsCoordinate(data=drilling.installedDiameter)),
                '{depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
                '0',
                '{a}'.format(a=self.formatAsAngle(data=keywords['aAngle'])),
                '{c}'.format(c=self.formatAsAngle(data=keywords['cAngle'])),
                self.createNCMacroHorizontalDrillingOptionalParameter(),
                ) + ')',
            '',
        )

    # contours
    def __createCommonStartBlock(self, **keywords):
        return self.createContourFirstCornerPolygonPoint(**keywords)

    def createSizingStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def createRabbetContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def createDadoContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def createContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def __createCommonEndBlock(self, **keywords):
        if self.VERSION >= 5:
            return ''
        else:
            return 'EP ({departure},_ANF,0)'.format(departure=keywords['departure']) + '\n'

    def createSizingEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    def createRabbetContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    def createDadoContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    def createContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    def createContourDefinitionOptionalParameter(self, **keywords):
        if self.VERSION >= 7:
            return '0,0'
        else:
            return None

    def createContourFirstCornerPolygonPointV5(self, **keywords):
        vertex = keywords['startVertex']

        self.lastContourDefinitionIndex += 1
        keywords['machining'].wopContourDefinitionIndex = self.lastContourDefinitionIndex

        return 'KB (' + joinIfNotNone(',',
                    "'{index}'".format(index=self.lastContourDefinitionIndex),
                    "''",
                    '{x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
                    '{y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
                    '{z}'.format(z=self.formatAsCoordinate(data=-keywords['zValue'])),
                    "''",
                    self.createContourDefinitionOptionalParameter(**keywords),
                ) + ')'

    def createContourFirstCornerPolygonPointV4(self, **keywords):
        vertex = keywords['startVertex']

        if 'toolID' not in keywords:
            keywords['toolID'] = keywords['machining'].toolID

        return joinIfNotNone('\n',
            self.createHoodPositioningWithToolChange(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolType='WZF', **keywords),
            self.changeView(**keywords),  # plane is already in keywords
            'SP (' + ','.join((
                '{x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
                '{y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
                '{z}'.format(z=self.formatAsCoordinate(data=-keywords['zValue'])),
                '{rc}'.format(rc=keywords['radiusCompensation']),
                '{approach}'.format(approach=keywords['approach']),
                '_ANF',
                '{distance}'.format(distance=0 if 'contourDistance' not in keywords else self.formatAsCoordinate(data=keywords['contourDistance'])),
                '0',
                '0',
            )) + ')'
            )

    def createContourFirstCornerPolygonPoint(self, **keywords):
        return self.createContourFirstCornerPolygonPointV5(**keywords) if self.VERSION >= 5 else self.createContourFirstCornerPolygonPointV4(**keywords)

    def createLineSegmentV5(self, **keywords):
        vertex = keywords['segment'].getEndVertex()

        return 'KG01 (' + joinIfNotNone(',',
                    "''",
                    '{x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
                    '{y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
                    '{z}'.format(z=NCHopsExport.LAST_Z_POSITION),
                    "''",
                    self.createContourDefinitionOptionalParameter(**keywords),
                ) + ')'

    def createLineSegmentV4(self, **keywords):
        vertex = keywords['segment'].getEndVertex()

        return 'G01 (' + ','.join((
            '{x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
            '{y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
            '{z}'.format(z=NCHopsExport.LAST_Z_POSITION),
            '0',
        )) + ')'

    def createLineSegment(self, **keywords):
        return self.createLineSegmentV5(**keywords) if self.VERSION >= 5 else self.createLineSegmentV4(**keywords)

    def createRadiusSegmentV5(self, **keywords):
        NCHOPS_CW = (1, "_KG02R_V5")
        NCHOPS_CCW = (1, "_KG03R_V5")
        NCHOPS_CW180 = (-1, "_KG02R_V5")
        NCHOPS_CCW180 = (-1, "_KG03R_V5")

        segment = keywords['segment']
        vertex = segment.getEndVertex()
        multiplicationFactor, directionCode = segment.getDirection(NCHOPS_CW, NCHOPS_CCW, NCHOPS_CW180, NCHOPS_CCW180)

        return joinIfNotNone(',',
            "CALL {direction} ( VAL EN:=''".format(direction=directionCode),
            'EX:={x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
            'EY:={y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
            'EZ:={z}'.format(z=NCHopsExport.LAST_Z_POSITION),
            'R:={radius}'.format(radius=self.formatAsCoordinate(data=segment.getRadius() * multiplicationFactor)),
            "EI:=''",
            'ESXY:=0',
            'ESD:=0',
        ) + ')'

    def createRadiusSegmentV4(self, **keywords):
        NCHOPS_CW = (1, "G02R")
        NCHOPS_CCW = (1, "G03R")
        NCHOPS_CW180 = (-1, "G02R")
        NCHOPS_CCW180 = (-1, "G03R")

        segment = keywords['segment']
        vertex = segment.getEndVertex()

        multiplicationFactor, directionCode = segment.getDirection(NCHOPS_CW, NCHOPS_CCW, NCHOPS_CW180, NCHOPS_CCW180)
        radius = segment.getRadius() * multiplicationFactor

        return '{direction} ('.format(direction=directionCode) + ','.join((
            '{x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
            '{y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
            '{z}'.format(z=NCHopsExport.LAST_Z_POSITION),
            '{radius}'.format(radius=self.formatAsCoordinate(data=radius)),
            '0',
        )) + ')'

    def createRadiusSegment(self, **keywords):
        return self.createRadiusSegmentV5(**keywords) if self.VERSION >= 5 else self.createRadiusSegmentV4(**keywords)

    def createFreeFormPocket(self, **keywords):
        plunge = 1 if keywords['approach'] in [self.APPROACH_LINEAR_FLYING, self.APPROACH_RADIAL_FLYING, self.APPROACH_SIDE_FLYING] else 0

        return joinIfNotNone(', ',
            "CALL _ExecutePocket_V5 ( VAL NAMEN:='{id}'".format(id=keywords['machining'].wopContourDefinitionIndex),
            'AA := {distance}'.format(distance=0 if 'contourDistance' not in keywords else self.formatAsCoordinate(data=keywords['contourDistance'])),
            'UEBERLAPPUNG := {overlap}'.format(overlap=self.CLEAR_OUT_POCKET_OVERLAP_FACTOR),
            'MODE := 0',
            'ANGLE := 0',
            'TIEFE := {depth}'.format(depth=keywords['depthAsString']),
            'ESMD := 0',
            'ANZAHL := {numberOfSteps}'.format(numberOfSteps=keywords['numberOfSteps']),
            'MAXZ := {depthPerStep}'.format(depthPerStep=keywords['depthPerStep']),
            'RD := 0',
            'FLIEGENDEINTAUCHEN := {plunge}'.format(plunge=plunge),
            'MAXEINTAUCHLAENGE := _ANF' + ')'
            )

    def createContourStartPointV5(self, **keywords):
        return joinIfNotNone(',',
            "KSP ('???'",
            "'???'",
            '{depth}'.format(depth=keywords['depthAsString']),
            '{rc}'.format(rc=keywords['radiusCompensation']),
            '{approach}'.format(approach=keywords['approach']),
            '_ANF',
            '{distance}'.format(distance=0 if 'contourDistance' not in keywords else self.formatAsCoordinate(data=keywords['contourDistance'])),
            '0',
            '0',
            '0',
            # Start optional parameters
            '0',
            '0',
            '{numberOfSteps}'.format(numberOfSteps=keywords['numberOfSteps'] if 'numberOfSteps' in keywords else 1),
            '{depthPerStep}'.format(depthPerStep=keywords['forceDepthPerStep'] if 'forceDepthPerStep' in keywords else 0),  # Depth per step
            '0',
            '0',
            '0',
            '0',
            '0',
            '0',
            '0',
            '0',
            '0',
        ) + ')'

    def createContourStartPointV4(self, **keywords):
        vertex = keywords['startVertex']
        return joinIfNotNone(',',
            'SP ({x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
            '{y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
            '{depth}'.format(depth=keywords['depthAsString']),  # {z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])), ?
            '{rc}'.format(rc=keywords['radiusCompensation']),
            '{approach}'.format(approach=keywords['approach']),
            '_ANF',
            '{distance}'.format(distance=0 if 'contourDistance' not in keywords else self.formatAsCoordinate(data=keywords['contourDistance'])),
            '0',
            '0',
        ) + ')'

    def createCommonContourExecution(self, **keywords):
        contourExecutionID = keywords['index']
        return joinIfNotNone('\n',
            self.createContourStartPointV5(**keywords) if self.VERSION >= 5 else self.createContourStartPointV4(**keywords),
            "KonturFraesen ('{index}',0,{count},0,{cornerradius},0,{useZ:d},0,0,0,0)".format(index=contourExecutionID, count=keywords['segmentCount'], cornerradius=self.CONTOUR_CORNER_RADIUS, useZ='forceDepth' not in keywords),
            "EP ({departure},_ANF,0)".format(departure=keywords['departure']),
            )

    # Currently not used, because the call of __CreateCommonContourBlock requires V5!
    def createSizingWithMacroV4(self, **keywords):
        if keywords['sizing'].tool.isRouterCentered():
            depth = '/2'  # Better readable than zValue, which represents the calculated thickness / 2
        else:
            depth = -self.SIZING_DEPTH_ADJUSTMENT

        return joinIfNotNone(',',
            'CALL _Format_V4 ( VAL ECKENVERRUNDUNG:={cornerradius}'.format(cornerradius=self.CONTOUR_CORNER_RADIUS),
            'RICHTUNG:={direction}'.format(direction='1' if self.isSizingMillingDirectionLeft(**keywords) else '0'),
            'TIEFE:=-_RZ{depth}'.format(depth=depth),
            'AUFMASS:=0',
            'ANF:=_ANF' + ')'
        )

    def createSizingWithMacroV5(self, **keywords):
        if keywords['sizing'].tool.isRouterCentered():
            depth = keywords['zValue']  # Calculated thickness / 2
        else:
            depth = -self.SIZING_DEPTH_ADJUSTMENT

        return joinIfNotNone(',',
            'CALL _Format_V5 ( VAL ECKENVERRUNDUNG:={cornerradius}'.format(cornerradius=self.CONTOUR_CORNER_RADIUS),
            'TIEFE:={depth}'.format(depth=depth),  # Could also be written as '_LPZ/2', '+/-' depending on the z-reference.
            'AUFMASS:=0',
            'ANF:=_ANF',
            'ESMD:=1',
            'GGL:={direction}'.format(direction='2' if self.isSizingMillingDirectionLeft(**keywords) else '1'),
            'LASER:=0',
            'DEPTH:=0' + ')'
        )

    def createSizingWithMacro(self, **keywords):
        if self.CUSTOM_USER_SIZING_MACRO:
            return self.getMappedCustomUserVariablesFromStringWithKeys(stringWithKeys=self.CUSTOM_USER_SIZING_MACRO, **keywords)
        else:
            return self.createSizingWithMacroV5(**keywords) if self.VERSION >= 5 else self.createSizingWithMacroV4(**keywords)  # Actually returns every time V5, because the call of __CreateCommonContourBlock requires V5

    def createBetterNestKontur(self, **keywords):
        return 'CALL BN_NestKontur ()' if 'sizing' in keywords and self.BETTER_NEST_OPTIONS else None

    def __CreateCommonContourBlock(self, **keywords):
        finishedSizeShape = self.getFinishedSizeShape(**keywords)
        contourExecutionID = keywords['machining'].wopContourDefinitionIndex

        if self.CREATE_SIZING_WITH_MACRO and 'sizing' in keywords and finishedSizeShape.isRectangular():  # HOLZHER Evolution Format-Macro
            contourCode = self.createSizingWithMacro(**keywords)
        else:
            if 'clearOut' in keywords and keywords['clearOut']:
                contourCode = self.createFreeFormPocket(**keywords)
            else:
                contourCode = self.createCommonContourExecution(index=contourExecutionID, **keywords)

        return joinIfNotNone('\n',
            self.createBetterNestKontur(**keywords),
            self.createHoodPositioningWithToolChange(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolType='WZF', **keywords),
            self.changeView(**keywords),  # plane should already be in keywords
            contourCode,
            '',
            '')

    def isSizingStrategyUsingTwoPasses(self, **keywords):
        area = keywords['surface'].getSawnSizeRectangularArea()

        return (
            self.SIZING_PASS_STRATEGY == 'force-two-passes'
            or (self.SIZING_PASS_STRATEGY == 'two-passes-for-small-workpieces' and area < 1000000 * self.SIZING_PASS_STRATEGY_AREA_THRESHOLD)
        )

    def createSizingContourBlock(self, **keywords):
        machining = sizing = keywords['sizing']
        surface = keywords['surface']
        toolID = sizing.toolID
        approach = self.getSizingApproachMode(**keywords) if not sizing.needsFlyingApproach else -abs(self.getSizingApproachMode(**keywords))

        zValue = keywords['zValue']
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=zValue, machining=sizing, **keywords)

        if self.VERSION >= 5 and self.isSizingStrategyUsingTwoPasses(**keywords) and surface.isFirstSide:
            if isNearlyGreaterOrEqual(depthPerStep, self.SIZING_FIRST_PASS_REMAINING_THICKNESS + self.SIZING_DEPTH_ADJUSTMENT):
                machiningDepth = surface.thickness - self.SIZING_FIRST_PASS_REMAINING_THICKNESS
                numberOfStepsDivider = math.ceil(machiningDepth / depthPerStep)
                newDepthPerStep = machiningDepth / numberOfStepsDivider

                return self.__CreateCommonContourBlock(
                    approach=approach,
                    departure=approach,
                    depthAsString=self.formatAsCoordinate(data=-zValue),
                    forceDepth=-zValue,
                    forceDepthPerStep=newDepthPerStep,
                    segmentCount=sizing.segmentCount,
                    toolID=toolID,
                    machining=sizing,
                    plane=None,
                    **keywords)
            else:
                message = self.getLocalizedMessage(
                    76, "Router with insufficient depth per step to be able to make nesting remaining-thickness sizing. Workpiece: {workpiece}, Tool name: '{toolName}', Tool ID: {toolID}"
                ).format(
                    workpiece=surface.workpieceName,
                    toolName=machining.tool.name if hasattr(machining.tool, 'name') else None,
                    toolID=machining.toolID
                )
                logging.error(message)
                return self.createNCError(comment=message)
        else:
            machiningDepth = zValue
            numberOfSteps = self.getCountOfStepsForDepth(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords)

            return self.__CreateCommonContourBlock(
                approach=approach,
                departure=approach,
                depthAsString=self.formatAsCoordinate(data=-zValue),
                forceDepth=-zValue,
                numberOfSteps=numberOfSteps,
                segmentCount=sizing.segmentCount,
                toolID=toolID,
                machining=sizing,
                plane=None,
                **keywords)

    def __getSizingRadiusCompensation(self, **keywords):
        sizing = keywords['surface'].sizing
        return self.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(sizing=sizing, **keywords) else self.RADIUS_COMPENSATION_LEFT

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=self.RADIUS_COMPENSATION_NONE, left=self.RADIUS_COMPENSATION_LEFT, right=self.RADIUS_COMPENSATION_RIGHT)

    def createVerticalContour(self, **keywords):
        contour = keywords['contour']

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        approach = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departure = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        if self.VERSION >= 5:
            depth = -contour.actualDepth
            # TODO CONTOUR2.0 IT-237: is forceDepth really needed? Or is there a better way to do it? -- NG
            keywords['forceDepth'] = depth
            depthPerStep = contour.currentDepthPerStep
            if isNearlyEqual(depthPerStep, 0.0):
                depthPerStep = self.getRouterDepthPerStep(machiningDepth=contour.actualDepth, **keywords)

            return self.__CreateCommonContourBlock(
                approach=approach,
                departure=departure,
                depthAsString=self.formatAsCoordinate(data=depth),
                radiusCompensation=radiusCompensation,
                segmentCount=contour.segmentCount,
                toolID=contour.toolID,
                feedRate=contour.currentFeedRate,
                numberOfSteps=self.getCountOfStepsForDepth(machiningDepth=contour.actualDepth, depthPerStep=depthPerStep, **keywords),
                depthPerStep=depthPerStep,
                clearOut=contour.clearOut,
                plane=None,
                **keywords)
        else:
            return self.createDefinitionForOneContour(
                approach=approach,
                departure=departure,
                radiusCompensation=radiusCompensation,
                feedRate=contour.currentFeedRate,
                plane=None,
                **keywords)

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createVerticalContour, **keywords))

    def createHorizontalContour(self, **keywords):
        contour = keywords['contour']

        plane, _positionOnPlane = self.__calculatePlane(**keywords)

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        approach = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departure = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        if self.VERSION >= 5:
            depth = -contour.depth
            depthAsString = self.formatAsCoordinate(data=depth)
            # TODO CONTOUR2.0 IT-237: is forceDepth really needed? Or is there a better way to do it? -- NG
            keywords['forceDepth'] = depth

            depthPerStep = contour.currentDepthPerStep
            if isNearlyEqual(depthPerStep, 0.0):
                depthPerStep = self.getRouterDepthPerStep(machiningDepth=contour.actualDepth, **keywords)

            contourCode = self.__CreateCommonContourBlock(
                approach=approach,
                departure=departure,
                depthAsString=depthAsString,
                radiusCompensation=radiusCompensation,
                segmentCount=contour.segmentCount,
                toolID=contour.toolID,
                feedRate=contour.currentFeedRate,
                numberOfSteps=self.getCountOfStepsForDepth(machiningDepth=contour.actualDepth, depthPerStep=depthPerStep, **keywords),
                plane=plane,
                **keywords)
        else:
            contourCode = self.createDefinitionForOneContour(
                approach=approach,
                departure=departure,
                radiusCompensation=radiusCompensation,
                feedRate=contour.currentFeedRate,
                plane=plane,
                **keywords)

        return joinIfNotNone('\n',
            contourCode,
            '',
        )

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createHorizontalContour, **keywords))

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']
        # TODO IT-220/IT-230 (HORIZONTAL): unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if contour.isVertical():
            zValue = self.getContourZValue(depth=keywords['contour'].depth, **keywords)
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            if plane:
                keywords['plane'] = plane

            matrix = plane.matrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            contour.path = contour.transformedPolyline(matrix.getInverted(), positionOnPlane)
            zValue = keywords['contour'].depth

        return super().createDefinitionForOneContour(zValue=zValue, **keywords)

    def createNCStop(self, **keywords):
        return 'Hinweismeldung ("{comment}")\n'.format(comment=keywords['comment'])

    def createNCError(self, **keywords):
        return 'Fehlermeldung ("{comment}")\n'.format(comment=keywords['comment'])

    # dado
    def __DadoMacroOptionalESMDParameters(self, **_keywords):
        if self.VERSION >= 5:
            return 'ESMD:=0'

    def __DadoMacroOptionalCRotatedParameters(self, **_keywords):
        if self.VERSION >= 5:
            return 'SPEGA:=0,EPEGA:=0,ESMD:=0,ESXY1:=1,ESXY2:=1'

    def getStartAndEndDistanceForDadoAlongAxis(self, **keywords):
        dado = keywords['dado']
        startInset, endInset = dado.startInset, dado.endInset
        if not dado.isInsetStart():
            startInset = -self.SECURITY_DISTANCE_FOR_DADO_SAW
        if not dado.isInsetEnd():
            endInset = -self.SECURITY_DISTANCE_FOR_DADO_SAW
        return startInset, endInset

    def getLeftAndRightDistanceForDadoAlongXAxis(self, **keywords):
        dado = keywords['dado']
        startInset, endInset = self.getStartAndEndDistanceForDadoAlongAxis(**keywords)
        start, end = dado.getStartAndEndPoints()
        return (startInset, endInset) if end.x >= start.x else (endInset, startInset)

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        dado = keywords['dado']
        toolID = dado.toolID
        start = dado.getStartPoint()
        left, right = self.getLeftAndRightDistanceForDadoAlongXAxis(**keywords)
        return joinIfNotNone('\n',
            self.createHoodPositioningWithToolChange(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolType='WZS', toolID=toolID, **keywords),
            self.changeView(plane=None, **keywords),
            ' '.join((
                'CALL _Nuten_X{version}'.format(version=self.createMacroNameVersionAdditionBlank()),
                '( VAL ' + joinIfNotNone(",",
                    'NB:={}'.format(self.formatAsCoordinate(data=dado.getWidth())),
                    ### IT-131: warum nutzen wir hier explizit -(minus) und nicht nur den Wert, der durch ZValueConverter berechnet wurde -- NG
                    'NT:={}'.format(self.formatAsCoordinate(data=-keywords['depth'])),
                    'EBENE:=1',
                    'ARAND:={}'.format(self.formatAsCoordinate(data=start.y - dado.getWidth() / 2)),
                    'ALINKS:={}'.format(self.formatAsCoordinate(data=left)),
                    'ARECHTS:={}'.format(self.formatAsCoordinate(data=right)),
                    'RK:={mode}'.format(mode=self.DADO_MODE),
                    self.__DadoMacroOptionalESMDParameters(**keywords),
                ) + ')'
            )),
            '',
        )

    def getBottomAndTopDistanceForDadoAlongYAxis(self, **keywords):
        dado = keywords['dado']
        startInset, endInset = self.getStartAndEndDistanceForDadoAlongAxis(**keywords)
        start, end = dado.getStartAndEndPoints()
        return (startInset, endInset) if end.y >= start.y else (endInset, startInset)

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        dado = keywords['dado']
        toolID = dado.toolID
        start = dado.getStartPoint()
        bottom, top = self.getBottomAndTopDistanceForDadoAlongYAxis(**keywords)
        return joinIfNotNone('\n',
            self.createHoodPositioningWithToolChange(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolType='WZS', toolID=toolID, **keywords),
            self.changeView(plane=None, **keywords),
            ' '.join((
                'CALL _Nuten_Y{version}'.format(version=self.createMacroNameVersionAdditionBlank()),
                '( VAL ' + joinIfNotNone(",",
                    'NB:={}'.format(self.formatAsCoordinate(data=dado.getWidth())),
                    ### IT-131: warum nutzen wir explizit - und nicht nur ein ZValueConverter -- NG
                    'NT:={}'.format(self.formatAsCoordinate(data=-keywords['depth'])),
                    'EBENE:=0',
                    'ARAND:={}'.format(self.formatAsCoordinate(data=start.x - dado.getWidth() / 2)),
                    'AUNTEN:={}'.format(self.formatAsCoordinate(data=bottom)),
                    'AOBEN:={}'.format(self.formatAsCoordinate(data=top)),
                    'RK:={mode}'.format(mode=self.DADO_MODE),
                    self.__DadoMacroOptionalESMDParameters(**keywords),
                ) + ')'
            )),
            '',
        )

    def createVerticalDadoWithSawCRotated(self, **keywords):
        dado = keywords['dado']
        start, end = self.getDadoLineSegmentWithSecurityDistance(**keywords).getStartAndEndPoints()
        toolID = dado.toolID
        return joinIfNotNone('\n',
            self.createHoodPositioningWithToolChange(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolType='WZS', toolID=toolID, **keywords),
            self.changeView(plane=None, **keywords),
            ' '.join((
                'CALL _nuten_frei{version}'.format(version=self.createMacroNameVersionAdditionBlank()),
                '( VAL ' + joinIfNotNone(",",
                    'X1:={x}'.format(x=self.formatAsCoordinate(data=start.x)),
                    'Y1:={y}'.format(y=self.formatAsCoordinate(data=start.y)),
                    'X2:={x}'.format(x=self.formatAsCoordinate(data=end.x)),
                    'Y2:={y}'.format(y=self.formatAsCoordinate(data=end.y)),
                    'NB:={}'.format(self.formatAsCoordinate(data=dado.getWidth())),
                    ### IT-131: warum nutzen wir explizit - und nicht nur ein ZValueConverter -- NG
                    'TIEFE:={}'.format(self.formatAsCoordinate(data=-keywords['depth'])),
                    'LAGE:=0',
                    'RK:={mode}'.format(mode=self.DADO_MODE),
                    self.__DadoMacroOptionalCRotatedParameters(**keywords),
                ) + ')'
            )),
            '',
        )

    def createVerticalDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        radiusCompensation = self.getDadoRadiusCompensation(**keywords)

        approach = self.APPROACH_MODE_LINEAR_FLYING if dado.isInsetBoth() or self.BETTER_NEST_OPTIONS else self.APPROACH_MODE_LINEAR_PIERCING

        if self.VERSION >= 5:
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=dado.actualDepth, **keywords)

            return self.__CreateCommonContourBlock(
                approach=approach,
                departure=approach,
                depthAsString='{depth}'.format(depth=self.formatAsCoordinate(data=dado.depth)),
                numberOfSteps=self.getCountOfStepsForDepth(machiningDepth=dado.actualDepth, depthPerStep=depthPerStep, **keywords),
                radiusCompensation=radiusCompensation,
                segmentCount=dado.segmentCount,
                toolID=dado.toolID,
                plane=None,
                **keywords)
        else:
            return self.createContourFromSegmentList(
                createStartBlock=self.createDadoContourStartBlock,
                createEndBlock=self.createDadoContourEndBlock,
                radiusCompensation=radiusCompensation,
                segmentList=self.generateDadoSegments(**keywords),
                zValue=dado.depth,
                approach=approach,
                departure=approach,
                plane=None,
                **keywords)

    def createDefinitionForOneDado(self, **keywords):
        return super().createDefinitionForOneDado(zValue=keywords['dado'].depth, **keywords)

    def generateDadoSegments(self, **keywords):
        if not keywords['dado'].isVertical():
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            dado = copy.deepcopy(keywords['dado'])
            matrix = plane.matrix * dado.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            dado.matrix = matrix  # TODO (HORIZONTAL) IT-230: unschönes Design -- NG
            dado.matrix.origin = positionOnPlane
            keywords['dado'] = dado

        return super().generateDadoSegments(**keywords)

    def createHorizontalDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        plane, _positionOnPlane = self.__calculatePlane(**keywords)

        radiusCompensation = self.getDadoRadiusCompensation(**keywords)

        approach = self.APPROACH_MODE_LINEAR_FLYING if dado.isInsetBoth() else self.APPROACH_MODE_LINEAR_PIERCING

        if self.VERSION >= 5:
            contourCode = self.__CreateCommonContourBlock(
                approach=approach,
                departure=approach,
                depthAsString='{depth}'.format(depth=self.formatAsCoordinate(data=dado.depth)),
                radiusCompensation=radiusCompensation,
                segmentCount=dado.segmentCount,
                toolID=dado.toolID,
                plane=plane,
                **keywords)
        else:
            contourCode = self.createContourFromSegmentList(
                createStartBlock=self.createDadoContourStartBlock,
                createEndBlock=self.createDadoContourEndBlock,
                radiusCompensation=radiusCompensation,
                segmentList=self.generateDadoSegments(**keywords),
                zValue=dado.depth,
                approach=approach,
                departure=approach,
                plane=plane,
                **keywords)

        return joinIfNotNone('\n',
            self.changeView(plane=plane, **keywords),
            contourCode,
            '',
        )

    # rabbet
    def generateRabbetContourPoints(self, **keywords):
        segments = self.generateRabbetSegments(**keywords)
        iterator = iter(segments)

        segment = next(iterator)
        yield segment.startVertex.position
        yield segment.endVertex.position

        for segment in iterator:
            yield segment.endVertex.position

    def __createVerticalRabbetLinesOptionalParameters(self):
        return '0,0,0,0,0,0,0,0,0,0,0,0,0,0' if self.VERSION >= 5 else None

    def createOneVerticalRabbetLines(self, **keywords):
        rabbet = keywords['rabbet']
        toolID = rabbet.toolID

        ### IT-134: warum wird depth hier berechnet, anstatt ein korrekten ZValueConverter zu nutzen? (siehe auch andere IT-131 Kommentar in dieser Datei) -- NG
        depth = - rabbet.depth

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        points = self.generateRabbetContourPoints(**keywords)
        iterator = iter(points)

        yield self.createHoodPositioning(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolID=toolID, **keywords) if self.VERSION <= 6 else None  # TODO IT-219 Refactoring?  --mjk
        yield self.createToolChange(toolType='WZF', toolID=toolID)
        yield self.createHoodPositioning(hoodPosition=self.HOOD_POSITION_MAXIMUM, toolID=toolID, **keywords) if self.VERSION > 6 else None
        yield self.changeView(plane=None, **keywords)

        point = next(iterator)
        yield joinIfNotNone(',',
            'SP({x}'.format(x=self.formatAsCoordinate(data=point.x)),  # works well for NC-Hops 4
            '{y}'.format(y=self.formatAsCoordinate(data=point.y)),
            '{depth}'.format(depth=self.formatAsCoordinate(data=depth)),
            '{}'.format(radiusCompensation),
            '1',
            '_ANF',
            '{distance}'.format(distance=self.formatAsCoordinate(data=keywords['contourDistance'])),
            '0',
            self.__createVerticalRabbetLinesOptionalParameters(),
            '0)'
        )

        for point in iterator:
            yield ','.join((
                'G01({x}'.format(x=self.formatAsCoordinate(data=point.x)),
                '{y}'.format(y=self.formatAsCoordinate(data=point.y)),
                '{depth}'.format(depth=self.formatAsCoordinate(data=depth)),
                '0)'
            ))

        yield 'EP(1,_ANF,0)'  # works well for NC-Hops 4
        yield ''

    def createDefinitionForOneRabbet(self, **keywords):
        return super().createDefinitionForOneRabbet(zValue=self.getDadoRabbetZValue(depth=keywords['machining'].depth, **keywords), **keywords)

    def createVerticalRabbet(self, **keywords):
        rabbet = keywords['rabbet']

        approach = self.APPROACH_MODE_LINEAR_FLYING if self.BETTER_NEST_OPTIONS else self.APPROACH_MODE_LINEAR_PIERCING

        if self.VERSION >= 5:
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=rabbet.actualDepth, **keywords)
            return self.__CreateCommonContourBlock(
                approach=approach,
                departure=approach,
                depthAsString='{depth}'.format(depth=self.formatAsCoordinate(data=-rabbet.depth)),
                numberOfSteps=self.getCountOfStepsForDepth(machiningDepth=rabbet.actualDepth, depthPerStep=depthPerStep, **keywords),
                radiusCompensation=self.getWopRadiusCompensation(**keywords),
                segmentCount=rabbet.segmentCount,
                toolID=rabbet.toolID,
                plane=None,
                **keywords)
        else:
            lines = self.createOneVerticalRabbetLines(**keywords)
            return joinIfNotNone('\n', *lines)

    def getHoodPosition(self, **keywords):
        correctedAngle = keywords['correctedAngle']

        assert correctedAngle >= 0, "Angle should not be negative!"
        assert isinstance(correctedAngle, float), "Angle should be a number!"

        if correctedAngle <= 22.5:
            hoodPosition = round(8 / 9 * self.HOOD_POSITION_MAXIMUM)
        elif correctedAngle <= 45.0:
            hoodPosition = round(7 / 9 * self.HOOD_POSITION_MAXIMUM)
        elif correctedAngle <= 67.5:
            hoodPosition = round(6 / 9 * self.HOOD_POSITION_MAXIMUM)
        elif correctedAngle > 67.5:
            hoodPosition = round(3 / 9 * self.HOOD_POSITION_MAXIMUM)
        else:
            hoodPosition = 1

        return hoodPosition

    # miter
    def __createHoodPositioningForVerticalMiter(self, **keywords):
        hoodPosition = self.getHoodPosition(**keywords)
        return self.createHoodPositioning(hoodPosition=hoodPosition, **keywords)

    def getSawMacroAndPreSawDepth(self, **keywords):
        boardThickness = keywords['boardThickness']
        if self.VERSION >= 5:
            return self.zValueMiterPreSawConverterFromVersion5.convertWithThickness(boardThickness, self.MITER_PRE_SAW_DEPTH_FROM_TOP)
        else:
            return self.zValueMiterPreSawConverterUpToVersion4.convertWithThickness(boardThickness, self.MITER_PRE_SAW_DEPTH_FROM_TOP)

    def __createSawCallForMiter(self, **keywords):
        boardThickness = keywords['boardThickness']
        miter = keywords['miter']

        start = miter.getStartPoint()
        end = miter.getEndPoint()

        preSawDepth = self.getSawMacroAndPreSawDepth(**keywords)
        sawDepth = self.zValueMiterSawConverter.convertWithThickness(boardThickness, boardThickness + self.MITER_SAW_DEPTH_FROM_BOTTOM)

        return ' '.join((
            'CALL _saege_frei{version}'.format(version=self.createMacroNameVersionAdditionBlank()),
            '( VAL',
            ','.join((
                'SX:={x}'.format(x=self.formatAsCoordinate(data=start.x)),
                'SY:={y}'.format(y=self.formatAsCoordinate(data=start.y)),
                'SZ:={z}'.format(z=self.formatAsCoordinate(data=preSawDepth)),
                'EX:={x}'.format(x=self.formatAsCoordinate(data=end.x)),
                'EY:={y}'.format(y=self.formatAsCoordinate(data=end.y)),
                'EZ:={z}'.format(z=self.formatAsCoordinate(data=sawDepth)),
                'BL:={rc}'.format(rc=self.RADIUS_COMPENSATION_RIGHT),
                'EINPASSEN:=0',
                'EL:={inset}'.format(inset=self.SECURITY_DISTANCE_FOR_DADO_SAW),
                'AL:={inset}'.format(inset=self.SECURITY_DISTANCE_FOR_DADO_SAW),
                'PARALLEL:=0',
                'K:={mode}'.format(mode=2 if preSawDepth else 1),  # 2 with preSaw, 1 without preSaw when depth = 0
                'KW:={aangle}'.format(aangle=self.formatAsAngle(data=-keywords['correctedAngle'])),
                'BH:=0',
                'RITZVERSATZ:={rv}'.format(rv=self.MITER_PRE_SAW_OFFSET),
                'ESZ:=0',
                'ESXY1:=1',
                'ESXY2:=1'
            )),
            ')',
        ))

    def createVerticalMiter(self, **keywords):
        surface = keywords['surface']
        miter = keywords['miter']
        toolID = miter.toolID

        correctedAngle = 90 - abs(miter.getAngle())

        return joinIfNotNone('\n',
            self.__createHoodPositioningForVerticalMiter(correctedAngle=correctedAngle, toolID=toolID, **keywords) if self.VERSION <= 6 else None,  # TODO IT-219 Refactoring?  --mjk
            self.createToolChange(toolType='WZS', toolID=toolID),
            self.__createHoodPositioningForVerticalMiter(correctedAngle=correctedAngle, toolID=toolID, **keywords) if self.VERSION > 6 else None,
            self.changeView(plane=None, **keywords),
            self.__createSawCallForMiter(boardThickness=surface.thickness, correctedAngle=correctedAngle, **keywords),
            '',
        )

    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCCW)

    # NC Macro
    def getNCMacroCookedVariables(self, **keywords):
        variables = super().getNCMacroCookedVariables(**keywords)

        ncmacro = keywords['ncmacro']
        if ncmacro.code.startswith('Ebene0()'):
            sideID = self.SIDE_MAIN
            sideX = ncmacro.position.x
            sideY = ncmacro.position.y
            sideCAngle = variables['cangle']
            ncmacro.code = ''.join(keywords['ncmacro'].code.splitlines(keepends=True)[1:])  # Remove first line of ncmacro.code to use changeView() correctly.
        elif ncmacro.isVertical():
            sideID = self.SIDE_MAIN
            sideX = ncmacro.position.x
            sideY = ncmacro.position.y
            sideCAngle = variables['cangle']
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            sideID = plane.sideID if plane.sideID else self.SIDE_FREE  # It's only the sideID, coordinates for positioning EbeneF(...) are not calculated.
            sideX = positionOnPlane.x
            sideY = positionOnPlane.y

            angle = VectorMath.angleBetween(ncmacro.matrix.u, plane.wopMatrix.u)

            if not isNearlyEqual(angle, 0.0):
                crossProduct = ncmacro.matrix.u.cross(plane.wopMatrix.u)
                if not isNearlyEqual(crossProduct.length(), 0.0) and crossProduct.normalized() == plane.wopMatrix.w:
                    angle = - angle

            sideCAngle = self.formatter.formatAsAngle(normalizeAngle(angle))

        variables['sideid'] = sideID
        variables['sidex'] = self.formatter.formatAsCoordinate(sideX)
        variables['sidey'] = self.formatter.formatAsCoordinate(sideY)
        variables['sidecangle'] = sideCAngle

        return variables

    def createNCMacroCode(self, **keywords):
        ncmacro = keywords['ncmacro']
        ncmacroCode = ncmacro.code

        if ncmacro.isVertical() or ncmacroCode.startswith('Ebene0()'):
            plane = None
        else:
            plane, _positionOnPlane = self.__calculatePlane(**keywords)

        return joinIfNotNone('\n',
            self.changeView(plane=plane, **keywords),
            super().createNCMacroCode(**keywords),
            '',
        )

    def createRecordFormatParameters(self, **keywords):
        return ','.join((
            '{recordName}:={recordValue}'.format(recordName=parameter.name, recordValue=parameter.value) for parameter in keywords['parameters']
        ))

    def createNCMacroRecordFormat(self, **keywords):
        ncmacro = keywords['ncmacro']
        toolCall = None
        parameters = tuple(self.getNCMacroCookedParameters(**keywords))
        if parameters and parameters[0].name.startswith('WZ'):
            toolCall = self.createToolChange(toolType=parameters[0].name, toolID=parameters[0].value)
            parameters = parameters[1:]
        wopParameters = self.createRecordFormatParameters(parameters=parameters, **keywords)

        return joinIfNotNone('\n',
            toolCall,
            joinIfNotNone(' ',
                'CALL',
                ncmacro.name,
                '(',
                'VAL' if wopParameters else None,
                wopParameters,
                ')',
            ),
            ''
        )

    # Footer
    PARKING_POSITION_RIGHT_REAR = 2
    PARKING_POSITION_MIDDLE_REAR = 3

    def createEndOfFile(self, **keywords):
        surface = keywords['surface']
        # if surface.getFinishedSizeX() < self.PARKING_POSITION_SWITCH:
        if surface.sawnSize.x < self.PARKING_POSITION_SWITCH:
            parkingPosition = self.PARKING_POSITION_MIDDLE_REAR
        else:
            parkingPosition = self.PARKING_POSITION_RIGHT_REAR

        return joinIfNotNone('\n',
            self.changeView(plane=None, **keywords),
            'CALL HH_Park ( VAL PARK:={position},X:=0,Y:=0)'.format(position=parkingPosition) if self.PARKING_POSITION_INTEGRATION else '',
        )

    def getLocalizedVariable(self, index, default):
        return getLocalizedString(NCHOPSPROGRAMVARIABLES_RESOURCEGROUP, index, default)
