# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import operator

from interiorcad.backend.ncexportmigrated.rabbettoncsegments import RabbetToNCSegments
from interiorcad.backend.zvalue.constants import ORIGIN_ON_TOP, Z_DIRECTION_DOWN
from interiorcad.backend.zvalue.converter import ZValueConverter
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.surface import Contour2D
from interiorcad.ncexport.svgcommon import SVGCommonExport


class ShaperOriginExport(SVGCommonExport):
    AUXILLARY_LINE_SVG_STROKE_COLOR = '#0068FF'
    AUXILLARY_LINE_SVG_FILL_COLOR = 'none'

    POCKET_SVG_STROKE_COLOR = None
    POCKET_SVG_FILL_COLOR = '#7F7F7F'

    EXTERIOR_CUT_STROKE_COLOR = '#000000'
    EXTERIOR_CUT_FILL_COLOR = '#000000'

    INTERIOR_CUT_STROKE_COLOR = '#000000'
    INTERIOR_CUT_FILL_COLOR = '#FFFFFF'

    RADIUS_COMPENSATION_NONE_STROKE_COLOR = '#7F7F7F'
    RADIUS_COMPENSATION_NONE_FILL_COLOR = 'none'

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.GZIPPED_SVG_FILES = False
        self.SURFACE_ORIGIN = self.getSurfaceOrigin(surfaceOriginRaw='top left')
        self.SVG_VERSION = '1.1'

        self.SIZING_ENTRY_OF_PROFILE = 'without'
        self.SIZING_APPROACH_DEPARTURE_DISTANCE = 0
        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = 0

        self.DEFAULT_SVG_FILL_COLOR = None
        self.DEFAULT_SVG_STROKE_COLOR = None

        self.ensureRoutersHavingRotationDirectionRight()

        self.createZValueConverters(**keywords)

    def ensureRoutersHavingRotationDirectionRight(self):
        machineConfiguration = self.getMachineConfiguration()
        for tool in machineConfiguration.toolDatabase.tools:
            if tool.type == Tool.ROUTER:
                tool.properties['router-rotation-direction'] = 'right'

    def createZValueConverters(self, **_keywords):
        topDown = ZValueConverter(ORIGIN_ON_TOP, Z_DIRECTION_DOWN)
        self.zValueVerticalDrillingConverter = topDown
        self.zValueHorizontalDrillingConverter = topDown
        self.zValueVerticalPocketConverter = topDown
        self.zValueFiveAxisDrillingConverter = topDown
        self.zValueMillingConverter = topDown
        self.zValueDadoSawConverter = topDown
        self.zValueMiterSawConverter = topDown

    def getDadoRabbetSecurityDistance(self, **keywords):
        # We want closed contours for dados (and in future also for rabbets), so there cannot be an approach or departure
        return 0.0

    def getPresentationAttributesForCommonPockets(self, **keywords):
        return self.getPresentationAttributes(stroke=self.POCKET_SVG_STROKE_COLOR, fill=self.POCKET_SVG_FILL_COLOR, **keywords)

    def getPresentationAttributesForVerticalDrillings(self, **keywords):
        return self.getPresentationAttributesForCommonPockets(**keywords)

    def getPresentationAttributesForHorizontalDrillings(self, **keywords):
        return self.getPresentationAttributes(stroke=self.AUXILLARY_LINE_SVG_STROKE_COLOR, fill=self.AUXILLARY_LINE_SVG_FILL_COLOR, **keywords)

    def getPresentationAttributesForPockets(self, **keywords):
        return self.getPresentationAttributesForCommonPockets(**keywords)

    def getPresentationAttributesForRabbets(self, **keywords):
        return self.getPresentationAttributesForCommonPockets(**keywords)

    def getPresentationAttributesForContour(self, **keywords):
        contour = keywords['contour']
        if contour.isClosed() and contour.clearOut:
            return self.getPresentationAttributesForCommonPockets(**keywords)
        elif contour.isClosed() and (contour.radiusCompensation != Contour2D.RADIUS_COMPENSATION_NONE):
            if operator.xor(self.isStandardOrigin(**keywords), operator.xor(contour.path.isClockwise(), contour.radiusCompensation == Contour2D.RADIUS_COMPENSATION_RIGHT)):
                return self.getPresentationAttributes(stroke=self.INTERIOR_CUT_STROKE_COLOR, fill=self.INTERIOR_CUT_FILL_COLOR, **keywords)
            else:
                return self.getPresentationAttributes(stroke=self.EXTERIOR_CUT_STROKE_COLOR, fill=self.EXTERIOR_CUT_FILL_COLOR, **keywords)
        else:
            return self.getPresentationAttributes(stroke=self.RADIUS_COMPENSATION_NONE_STROKE_COLOR, fill=self.RADIUS_COMPENSATION_NONE_FILL_COLOR, **keywords)

    def getPresentationAttributesForDados(self, **keywords):
        return self.getPresentationAttributesForCommonPockets(**keywords)

    def getPresentationAttributesForMiters(self, **keywords):
        return self.getPresentationAttributes(stroke=self.AUXILLARY_LINE_SVG_STROKE_COLOR, fill=self.AUXILLARY_LINE_SVG_FILL_COLOR, **keywords)

    def getPresentationAttributesForSizing(self, **keywords):
        return self.getPresentationAttributes(stroke=self.EXTERIOR_CUT_STROKE_COLOR, fill=self.EXTERIOR_CUT_FILL_COLOR, **keywords)

    def isDadoClosed(self, **_keywords):
        # We need dados to be defined as closed in ShaperOrigin to be able to declare them as pockets
        return True

    def isRabbetClosed(self, **_keywords):
        # We need dados to be defined as closed in ShaperOrigin to be able to declare them as pockets
        return True

    def generateRabbetSegments(self, **keywords):
        segments = tuple(RabbetToNCSegments.generateRabbetSegmentsClosedPerimeter(rabbet=keywords['rabbet']))

        self.updateViewBoxBySegments(segments=segments)

        return segments

    def createVerticalRabbet(self, **keywords):
        return self.createDefinitionForOneRabbet(
            presenterAttributeGenerator=self.getPresentationAttributesForRabbets,
            **keywords)

    def createNCStop(self, **keywords):
        return ''  # We do not want any XML comment. With a shaper origin, the user always checks what he is doing anyway.

    @staticmethod
    def __distanceAdapter(contour, tool, radiusCompensation):
        if not contour.isClosed() and tool and radiusCompensation != Contour2D.RADIUS_COMPENSATION_NONE:
            return tool.getRouterDiameter() / 2
        else:
            return 0.0

    def createVerticalContour(self, **keywords):
        keywords['machining'] = keywords['contour'] = keywords['contour'].createCurrentRoutingMovingPath(extraMirroring=not self.isStandardOrigin(**keywords),
            distanceAdapter=self.__distanceAdapter
        )
        return self.createCommonVerticalContour(**keywords)

    def createVerticalMultiUsageContour(self, **keywords):
        def contourPassCreator(passIndex):
            return keywords['contour'].createCurrentMultiUsageRoutingMovingPath(
                passIndex=passIndex,
                machineConfiguration=self.getMachineConfiguration(**keywords),
                extraMirroring=not self.isStandardOrigin(**keywords),
                distanceAdapter=self.__distanceAdapter
            )

        return '\n'.join(self.generateCommonMultiUsageContour(contourPassCreator=contourPassCreator, creator=self.createCommonVerticalContour, **keywords))
