# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

from interiorcad.backend.ncexportmigrated.rabbettoncsegments import RabbetToNCSegments
from interiorcad.ncexport.svgcommon import SVGCommonExport


class SVGExport(SVGCommonExport):
    def __init__(self, **keywords):
        super().__init__(**keywords)

        machineConfiguration = self.getMachineConfiguration()
        self.GZIPPED_SVG_FILES = machineConfiguration.getAsBool('gzip-compressed-files', False)
        self.SURFACE_ORIGIN = self.getSurfaceOrigin(surfaceOriginRaw=machineConfiguration.get('point-of-origin', 'top left'))
        self.SVG_VERSION = machineConfiguration.get('svg-version', '1.1')

        self.SIZING_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('sizing-approach-departure-distance', 30)
        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('contour-approach-departure-distance', 30)

        self.DEFAULT_SVG_FILL_COLOR = 'none'
        self.DEFAULT_SVG_STROKE_COLOR = 'black'
        self.readColorSettings(**keywords)

        self.createZValueConverters(**keywords)

    def readColorSettings(self, **_keywords):
        machineConfiguration = self.getMachineConfiguration()

        self.VERTICAL_DRILLINGS_SVG_STROKE_COLOR = machineConfiguration.get('vertical-drillings-svg-stroke-color', '')
        self.VERTICAL_DRILLINGS_SVG_FILL_COLOR = machineConfiguration.get('vertical-drillings-svg-fill-color', '')

        self.HORIZONTAL_DRILLINGS_SVG_STROKE_COLOR = machineConfiguration.get('horizontal-drillings-svg-stroke-color', '')
        self.HORIZONTAL_DRILLINGS_SVG_FILL_COLOR = machineConfiguration.get('horizontal-drillings-svg-fill-color', '')

        self.POCKETS_SVG_STROKE_COLOR = machineConfiguration.get('pockets-svg-stroke-color', '')
        self.POCKETS_SVG_FILL_COLOR = machineConfiguration.get('pockets-svg-fill-color', '')

        self.CONTOUR_SVG_STROKE_COLOR = machineConfiguration.get('contour-svg-stroke-color', '')
        self.CONTOUR_SVG_FILL_COLOR = machineConfiguration.get('contour-svg-fill-color', '')

        self.DADO_SVG_STROKE_COLOR = machineConfiguration.get('dado-svg-stroke-color', '')
        self.DADO_SVG_FILL_COLOR = machineConfiguration.get('dado-svg-fill-color', '')

        self.MITER_SVG_STROKE_COLOR = machineConfiguration.get('miter-svg-stroke-color', '')
        self.MITER_SVG_FILL_COLOR = machineConfiguration.get('miter-svg-fill-color', '')

        self.SIZING_SVG_STROKE_COLOR = machineConfiguration.get('sizing-svg-stroke-color', '')
        self.SIZING_SVG_FILL_COLOR = machineConfiguration.get('sizing-svg-fill-color', '')

        self.RABBET_SVG_STROKE_COLOR = machineConfiguration.get('rabbet-svg-stroke-color', '')
        self.RABBET_SVG_FILL_COLOR = machineConfiguration.get('rabbet-svg-fill-color', '')

    def getPresentationAttributesForVerticalDrillings(self, **keywords):
        return self.getPresentationAttributes(stroke=self.VERTICAL_DRILLINGS_SVG_STROKE_COLOR, fill=self.VERTICAL_DRILLINGS_SVG_FILL_COLOR, **keywords)

    def getPresentationAttributesForHorizontalDrillings(self, **keywords):
        return self.getPresentationAttributes(stroke=self.HORIZONTAL_DRILLINGS_SVG_STROKE_COLOR, fill=self.HORIZONTAL_DRILLINGS_SVG_FILL_COLOR, **keywords)

    def getPresentationAttributesForPockets(self, **keywords):
        return self.getPresentationAttributes(stroke=self.POCKETS_SVG_STROKE_COLOR, fill=self.POCKETS_SVG_FILL_COLOR, **keywords)

    def getPresentationAttributesForRabbets(self, **keywords):
        return self.getPresentationAttributes(fill=self.RABBET_SVG_FILL_COLOR, stroke=self.RABBET_SVG_STROKE_COLOR, **keywords)

    def getPresentationAttributesForContour(self, **keywords):
        return self.getPresentationAttributes(fill=self.CONTOUR_SVG_FILL_COLOR, stroke=self.CONTOUR_SVG_STROKE_COLOR, **keywords)

    def getPresentationAttributesForDados(self, **keywords):
        return self.getPresentationAttributes(fill=self.DADO_SVG_FILL_COLOR, stroke=self.DADO_SVG_STROKE_COLOR, **keywords)

    def getPresentationAttributesForMiters(self, **keywords):
        return self.getPresentationAttributes(fill=self.MITER_SVG_FILL_COLOR, stroke=self.MITER_SVG_STROKE_COLOR, **keywords)

    def getPresentationAttributesForSizing(self, **keywords):
        return self.getPresentationAttributes(fill=self.SIZING_SVG_FILL_COLOR, stroke=self.SIZING_SVG_STROKE_COLOR, **keywords)

    def isDadoClosed(self, **_keywords):
        return False

    def isRabbetClosed(self, **_keywords):
        return False

    def generateRabbetSegments(self, **keywords):
        rabbet = keywords['rabbet']
        securityDistance = self.getDadoRabbetSecurityDistance(**keywords)
        if self.RABBET_REPLACEMENT.startswith('rounded-'):
            routerRadius = rabbet.tool.getRouterDiameter() / 2
            segments = tuple(RabbetToNCSegments.generateRabbetRoundedSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=securityDistance, straightSecurityDistance=securityDistance, routerRadius=routerRadius))
        elif self.RABBET_REPLACEMENT.startswith('contour-'):
            segments = tuple(RabbetToNCSegments.generateRabbetSegmentsWithSecurityDistance(rabbet=rabbet, insetSecurityDistance=securityDistance, straightSecurityDistance=securityDistance))
        else:
            segments = tuple(RabbetToNCSegments.generateRabbetSegmentsWithInsets(rabbet=rabbet))

        self.updateViewBoxBySegments(segments=segments)

        return segments

    def createVerticalRabbet(self, **keywords):
        return self.createVerticalRabbetOffsetByContourDistance(
            presenterAttributeGenerator=self.getPresentationAttributesForRabbets,
            **keywords)

    def createNCStop(self, **keywords):
        return 2 * self.INDENT + '<!-- {comment} -->\n'.format(comment=self.getXMLEscaped(data=keywords['comment']))

    def createVerticalContour(self, **keywords):
        return self.createCommonVerticalContour(**keywords)

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createCommonVerticalContour, **keywords))
