# -*- coding: utf-8 -*-

# Copyright 2011-2023, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import datetime
import gzip
import html
import math
import operator
import textwrap

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.ncexportmigrated.dadotoncsegments import getMiddleLine, getPerimeterCW, getPerimeterCCW, getStartEndLine, getDadoContourCW, getDadoContourCCW, getDadoRoundedContourCW, getDadoRoundedContourCCW
from interiorcad.backend.ncexportmigrated.pockettoncsegments import generateTwoSegmentsCCW, generateTwoSegmentsCW
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyLesser, isNearlyGreater
from interiorcad.interface.geometry.cornervertex import CornerVertex
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.rectangle import Rectangle
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.ncexport.approachcreator import createArcDeparture, createLinearApproach, createLinearDeparture, createArcApproach
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.interface.surface.origins import SURFACE_ORIGIN_BACK_LEFT, SURFACE_ORIGIN_BACK_RIGHT, SURFACE_ORIGIN_FRONT_LEFT, SURFACE_ORIGIN_FRONT_RIGHT
from interiorcad.ncexport.base import NCExportBase


class SVGViewBox(Rectangle):
    """View Box for SVG

    Be careful that a view box in SVG has always its origin at top left"""

    # We derive a new class, as Rectangle is already a sort of hybrid class not respecting Clean Code principles

    def updateByPoint(self, point):
        if isNearlyLesser(point.x, self.left):
            self.left = point.x

        if isNearlyGreater(point.x, self.right):
            self.right = point.x

        if isNearlyLesser(point.y, self.top):
            self.top = point.y

        if isNearlyGreater(point.y, self.bottom):
            self.bottom = point.y

    def updateByStartEndPoints(self, start, end):
        self.updateByPoint(start)
        self.updateByPoint(end)

    def updateByLineSegment(self, lineSegment):
        start, end = lineSegment.getStartAndEndPoints()
        self.updateByStartEndPoints(start, end)

    @property
    def width(self):
        return self.right - self.left

    @property
    def height(self):
        return self.bottom - self.top

    def transformed(self, matrix):
        topRight = self.topRight()
        bottomLeft = self.bottomLeft()

        topRight.multiplyByMatrix(matrix)
        topRight += matrix.origin

        bottomLeft.multiplyByMatrix(matrix)
        bottomLeft += matrix.origin

        left = min(topRight.x, bottomLeft.x)
        right = max(topRight.x, bottomLeft.x)
        top = min(topRight.y, bottomLeft.y)
        bottom = max(topRight.y, bottomLeft.y)

        return SVGViewBox.createFromValues(left, bottom, right, top)


class SVGCommonExport(NCExportBase):
    EXPORT_FORMAT = 'SVG'

    # TODO IT-249: we do not support something else than UTF-8, as we do no set the encoding in <?xml
    OUTPUT_ENCODING = 'UTF-8'

    INDENT = 4 * ' '

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.currentPath = None
        self.gridDrillingIndex = None
        self.viewBox = None

        self.DEFAULT_SVG_FILL_COLOR = None
        self.DEFAULT_SVG_STROKE_COLOR = None

        machineConfiguration = self.getMachineConfiguration()
        self.POCKET_DIRECTION = machineConfiguration.get('pocket-clock-wise', 'pocket-same')
        self.MITER_CW = machineConfiguration.getAsBool('miter-clock-wise', False)

    def getSurfaceOrigin(self, **keywords):
        surfaceOriginRaw = keywords['surfaceOriginRaw'].lower()
        if surfaceOriginRaw and 'top' in surfaceOriginRaw and 'left' in surfaceOriginRaw:
            return SURFACE_ORIGIN_BACK_LEFT
        elif surfaceOriginRaw and 'top' in surfaceOriginRaw and 'right' in surfaceOriginRaw:
            return SURFACE_ORIGIN_BACK_RIGHT
        elif surfaceOriginRaw and 'bottom' in surfaceOriginRaw and 'left' in surfaceOriginRaw:
            return SURFACE_ORIGIN_FRONT_LEFT
        elif surfaceOriginRaw and 'bottom' in surfaceOriginRaw and 'right' in surfaceOriginRaw:
            return SURFACE_ORIGIN_FRONT_RIGHT
        else:
            return SURFACE_ORIGIN_FRONT_LEFT

    # Everything in the following function is still from template.py
    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)

    def getNCFileExtension(self, **_keywords):
        return '.svgz' if self.GZIPPED_SVG_FILES else '.svg'

    def openFileForWritingText(self, filepath):
        if self.GZIPPED_SVG_FILES:
            return gzip.open(filepath, mode='wt', compresslevel=9, newline=self.OUTPUT_END_OF_LINE, encoding=self.OUTPUT_ENCODING, errors=self.OUTPUT_ENCODING_ERROR_MODE)
        else:
            return super().openFileForWritingText(filepath)

    def getXMLEscaped(self, **keywords):
        if isinstance(keywords['data'], datetime.datetime):
            # Ignore explicitly datetime objects, otherwise they couldn't be formatted afterward anymore.
            return keywords['data']
        else:
            return html.escape(keywords['data'])

    def createExportFileContentFor(self, **keywords):
        lines = super().createExportFileContentFor(**keywords)

        # All the machinings must have been exported before the view box can be written out
        return joinIfNotNone('\n',
            '<?xml version="1.0" standalone="no"?>',
            '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">' if 'doctype' in self.SVG_VERSION else None,
            self.createRealHead(**keywords),
            lines,
            '</svg>',
            '',
        )

    # Header
    def createHead(self, **keywords):
        surface = keywords['surface']

        self.gridDrillingIndex = 0
        self.viewBox = SVGViewBox(
            left=0,
            bottom=surface.getFinishedSizeY(),
            right=surface.getFinishedSizeX(),
            top=0
        )

        return ''  # Defer until the very end where createRealHead is called

    def createRealHead(self, **keywords):
        transformationMatrix = self.getMainTransformationMatrix(**keywords)

        return joinIfNotNone('\n',
            self.createWorkpieceDefinition(transformationMatrix=transformationMatrix, **keywords),
            self.createWorkpieceComments(**keywords),
            self.createMainGroupStartTag(transformationMatrix=transformationMatrix, **keywords),
        )

    def createWorkpieceDefinition(self, **keywords):
        viewBox = self.viewBox.transformed(keywords['transformationMatrix'])

        return joinIfNotNone(' ',
            '<svg',
            'width="{width}mm"',
            'height="{height}mm"',
            'viewBox="{left} {top} {width} {height}"',
            'version="1.1"' if self.SVG_VERSION.startswith('1.1') else None,
            'xmlns="http://www.w3.org/2000/svg"',
            'xmlns:xlink="http://www.w3.org/1999/xlink"' if self.SVG_VERSION.startswith('1.1') else None,
            '>'
        ).format(
            left=math.ceil(viewBox.left),
            top=math.ceil(viewBox.top),
            width=math.ceil(viewBox.width),
            height=math.ceil(viewBox.height),
        )

    def createWorkpieceComments(self, **keywords):
        projectInfo = self.getProjectInfo()
        isProjectIDValid = ExportHelper.isProjectIDValid(self, **projectInfo)
        workpieceComments = dict(
            (self.getXMLEscaped(data=key), self.getXMLEscaped(data=value))
            for key, value in self.getWorkpieceComments(**keywords).items()
        )

        return joinIfNotNone('\n',
            self.INDENT + '<title>{WorkpieceName}</title>'.format_map(workpieceComments),
            self.INDENT + '<desc>',
            2 * self.INDENT + '{WorkpieceLabel}: {WorkpieceName}'.format_map(workpieceComments),
            2 * self.INDENT + '{CommentLabel}: {Comment}'.format_map(workpieceComments) if workpieceComments['Comment'] else None,
            2 * self.INDENT + '{ItemIDLabel}: {ItemID}'.format_map(workpieceComments),
            2 * self.INDENT + '{Edge1Label}: {Edge1}'.format_map(workpieceComments) if workpieceComments['Edge1'] else None,  # front
            2 * self.INDENT + '{Edge2Label}: {Edge2}'.format_map(workpieceComments) if workpieceComments['Edge2'] else None,  # back
            2 * self.INDENT + '{Edge3Label}: {Edge3}'.format_map(workpieceComments) if workpieceComments['Edge3'] else None,  # left
            2 * self.INDENT + '{Edge4Label}: {Edge4}'.format_map(workpieceComments) if workpieceComments['Edge4'] else None,  # right
            2 * self.INDENT + '{CabinetNameLabel}: {CabinetName}'.format_map(workpieceComments) if workpieceComments['CabinetName'] else None,
            2 * self.INDENT + '{ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}'.format_map(workpieceComments) if isProjectIDValid else None,
            2 * self.INDENT + '{PositionNumberLabel}: {PositionNumber} {PositionDescription}'.format_map(workpieceComments) if isProjectIDValid else None,
            2 * self.INDENT + '{CountLabel}: {Count}'.format_map(workpieceComments),
            self.INDENT + '</desc>',
        )

    def createMainGroupStartTag(self, **keywords):
        return joinIfNotNone(' ',
            self.INDENT + '<g',
            self.createMainTransformationMatrix(**keywords),
            'fill="{fill}"'.format(fill=self.DEFAULT_SVG_FILL_COLOR) if self.DEFAULT_SVG_FILL_COLOR else None,
            'stroke="{stroke}"'.format(stroke=self.DEFAULT_SVG_STROKE_COLOR) if self.DEFAULT_SVG_STROKE_COLOR else None,
        ) + '>'

    def getMainTransformationMatrix(self, **keywords):
        surface = keywords['surface']

        offsetX = surface.getFinishedSizeX()
        offsetY = surface.getFinishedSizeY()

        if self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT:
            return Matrix()  # the native origin of SVG is by default at top left of the SVG document
        elif self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT:
            return Matrix(u=Vector(1, 0, 0), v=Vector(0, -1, 0), offset=Vector(0, offsetY, 0))  # interiorcad default. We need an origin at bottom left of the SVG document
        elif self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT:
            return Matrix(u=Vector(-1, 0, 0), v=Vector(0, 1, 0), offset=Vector(offsetX, 0, 0))  # We need an origin at bottom left of the SVG document
        elif self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_RIGHT:
            return Matrix(u=Vector(-1, 0, 0), v=Vector(0, -1, 0), offset=Vector(offsetX, offsetY, 0))  # We need an origin at bottom right of the SVG document
        else:
            assert False, 'Unknown surface origin: {}'.format(self.SURFACE_ORIGIN)

    def createMainTransformationMatrix(self, **keywords):
        transformationMatrix = keywords['transformationMatrix']

        if not transformationMatrix.isIdentity():
            return 'transform="matrix({ux} {uy} {vx} {vy} {width} {height})"'.format(
                ux=self.formatAsCoordinate(data=transformationMatrix.u.x),
                uy=self.formatAsCoordinate(data=transformationMatrix.u.y),
                vx=self.formatAsCoordinate(data=transformationMatrix.v.x),
                vy=self.formatAsCoordinate(data=transformationMatrix.v.y),
                width=self.formatAsCoordinate(data=transformationMatrix.origin.x),
                height=self.formatAsCoordinate(data=transformationMatrix.origin.y),
            )

    # Machinings
    def getCookedName(self, **keywords):
        return self.getXMLEscaped(data=keywords['template'].format_map(keywords['variables']))

    def createSizingContour(self, **keywords):
        radiusCompensation = self.DIN_RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(sizing=keywords['surface'].sizing, **keywords) else self.DIN_RADIUS_COMPENSATION_LEFT
        return self.createSizingContourDefinitionAsExecution(radiusCompensation=radiusCompensation, **keywords)

    def getPresentationAttributes(self, **keywords):
        attributes = []

        fill = keywords['fill']
        if fill and fill != self.DEFAULT_SVG_FILL_COLOR:
            attributes.append('fill="{fill}"'.format(fill=fill))

        stroke = keywords['stroke']
        if stroke and stroke != self.DEFAULT_SVG_STROKE_COLOR:
            attributes.append('stroke="{stroke}"'.format(stroke=stroke))

        if attributes:
            return ' ' + ' '.join(attributes)  # TODO IT-271: do we really want the prepended space here or should the calling code add it? -- NG
        else:
            return ''

    def getPresentationAttributesForVerticalDrillings(self, **keywords):
        raise NotImplementedError

    def getPresentationAttributesForHorizontalDrillings(self, **keywords):
        raise NotImplementedError

    def getPresentationAttributesForPockets(self, **keywords):
        raise NotImplementedError

    def getPresentationAttributesForRabbets(self, **keywords):
        raise NotImplementedError

    def getPresentationAttributesForContour(self, **keywords):
        raise NotImplementedError

    def getPresentationAttributesForDados(self, **keywords):
        raise NotImplementedError

    def getPresentationAttributesForMiters(self, **keywords):
        raise NotImplementedError

    def getPresentationAttributesForSizing(self, **keywords):
        raise NotImplementedError

    def isDadoClosed(self, **_keywords):
        raise NotImplementedError

    def isRabbetClosed(self, **_keywords):
        raise NotImplementedError

    def createVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']
        title = self.getCookedName(
            template=drilling.toolID,
            variables={'depth': self.formatAsCoordinate(data=drilling.depth)},
            **keywords
        )
        return '\n'.join((
            2 * self.INDENT + '<circle cx="{x}" cy="{y}" r="{radius}"{presentation}>'.format(
                x=self.formatAsCoordinate(data=drilling.position.x),
                y=self.formatAsCoordinate(data=drilling.position.y),
                radius=self.formatAsCoordinate(data=drilling.diameter / 2),
                presentation=self.getPresentationAttributesForVerticalDrillings(**keywords)
            ),
            3 * self.INDENT + '<title>{title}</title>'.format(title=title),
            2 * self.INDENT + '</circle>',
            '',
        ))

    def createVerticalDrillingPiercing(self, **keywords):
        return self.createVerticalDrilling(**keywords)

    def createVerticalGridDrilling(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        self.gridDrillingIndex += 1
        symbolName = "GridDrilling{index}".format(index=self.gridDrillingIndex)
        title = self.getCookedName(
            template=gridDrilling.tool.getToolID(),
            variables={'depth': self.formatAsCoordinate(data=gridDrilling.depth)},
            **keywords
        )
        return '\n'.join((
            2 * self.INDENT + '<g>',
            3 * self.INDENT + '<title>{title}</title>'.format(symbolName=symbolName, title=title),
            3 * self.INDENT + '<defs>',
            4 * self.INDENT + '<circle id="{id}" r="{radius}"{presentation} />'.format(
                id=symbolName,
                radius=self.formatAsCoordinate(data=gridDrilling.diameter / 2),
                presentation=self.getPresentationAttributesForVerticalDrillings(**keywords)
            ),
            3 * self.INDENT + '</defs>',
            '\n'.join((
                ' ' * 12 + '<use x="{x}" y="{y}" {xlinknamespace}href="#{id}" />'.format(
                    xlinknamespace='xlink:' if self.SVG_VERSION.startswith('1.1') else '',
                    x=self.formatAsCoordinate(data=position.x),
                    y=self.formatAsCoordinate(data=position.y),
                    id=symbolName,
                ) for position in gridDrilling.getDrillingPositions()
            )),
            2 * self.INDENT + '</g>',
            ''
        ))

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def isDadoRabbetReplacementCW(self, machining, machiningReplacement, **keywords):
        if machiningReplacement == "1":
            return True
        elif machiningReplacement == "0":
            return False
        else:
            return super().isDadoRabbetReplacementCW(machining, machiningReplacement, **keywords)

    def generatePocketSegments(self, **keywords):
        pocket = keywords['pocket']
        cw = self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)
        generator = generateTwoSegmentsCW if cw else generateTwoSegmentsCCW
        return tuple(generator(pocket=pocket))

    def createVerticalPocket(self, **keywords):
        segmentList = self.generatePocketSegments(**keywords)
        cw = self.isDadoRabbetReplacementCW(keywords['pocket'], self.POCKET_DIRECTION)
        radiusCompensation = self.DIN_RADIUS_COMPENSATION_RIGHT if cw else self.DIN_RADIUS_COMPENSATION_LEFT

        return self.createContourFromSegmentList(
            createStartBlock=self.__createCommonStartBlock,
            createEndBlock=self.__createCommonEndBlock,
            radiusCompensation=radiusCompensation,
            segmentList=segmentList,
            isClosed=True,
            zValue=keywords['depth'],
            presenterAttributeGenerator=self.getPresentationAttributesForPockets,
            **keywords)

    def createVerticalPocketPiercing(self, **keywords):
        return self.createVerticalPocket(**keywords)

    def createRotatedHorizontalDrilling(self, **keywords):
        drilling = keywords['drilling']

        startPosition = drilling.position
        endPosition = copy.copy(startPosition)
        endPosition.x += drilling.depth * math.cos(math.radians(drilling.cAngle))
        endPosition.y += drilling.depth * math.sin(math.radians(drilling.cAngle))

        variables = {
            'depth': self.formatAsCoordinate(data=drilling.depth),
            'z': self.formatAsCoordinate(data=keywords['zPosition']),
            'aangle': self.formatAsAngle(data=drilling.aAngle),
            'cangle': self.formatAsAngle(data=drilling.cAngle),
        }

        title = self.getCookedName(
            template=drilling.toolID,
            variables=variables,
            **keywords
        )
        return '\n'.join((
            2 * self.INDENT + '<line x1="{x1}" y1="{y1}" x2="{x2}" y2="{y2}"{presentation}>'.format(
                x1=self.formatAsCoordinate(data=startPosition.x),
                y1=self.formatAsCoordinate(data=startPosition.y),
                x2=self.formatAsCoordinate(data=endPosition.x),
                y2=self.formatAsCoordinate(data=endPosition.y),
                presentation=self.getPresentationAttributesForHorizontalDrillings(**keywords)
            ),
            3 * self.INDENT + '<title>{title}</title>'.format(title=title),
            2 * self.INDENT + '</line>',
            '',
        ))

    # Everything in the following function is still from template.py
    def createFiveAxisDrilling(self, **_keywords):
        return ''

    def createVerticalRabbet(self, **keywords):
        raise NotImplementedError

    def updateViewBoxBySegments(self, **keywords):
        #  TODO IT-270 is the code too simplistic, even for inset and outset? -- NG
        for segment in keywords['segments']:
            self.viewBox.updateByStartEndPoints(segment.getStartVertex().position, segment.getEndVertex().position)

    def generateRabbetSegments(self, **keywords):
        raise NotImplementedError

    def createCommonVerticalContour(self, **keywords):
        return self.createDefinitionForOneContour(
            zValue=self.getContourZValue(depth=keywords['contour'].depth, **keywords),
            presenterAttributeGenerator=self.getPresentationAttributesForContour,
            **keywords)

    def generateDadoSegments(self, **keywords):
        dado = keywords['dado']
        routerRadius = dado.tool.getRouterDiameter() / 2
        securityDistance = self.getDadoRabbetSecurityDistance(**keywords)

        if self.DADO_REPLACEMENT == 'middle-line':
            segments = getMiddleLine(dado)
        else:
            cw = operator.xor(self.isDadoRabbetReplacementCW(dado, self.DADO_REPLACEMENT), not self.isStandardOrigin(**keywords))

            if self.DADO_REPLACEMENT.startswith('perimeter-') and cw:
                segments = getPerimeterCW(dado)
            elif self.DADO_REPLACEMENT.startswith('perimeter-') and not cw:
                segments = getPerimeterCCW(dado)
            elif self.DADO_REPLACEMENT.startswith( 'contour-') and cw:
                segments = getDadoContourCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
            elif self.DADO_REPLACEMENT.startswith('contour-') and not cw:
                segments = getDadoContourCCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
            elif self.DADO_REPLACEMENT.startswith('rounded-contour-') and cw:
                segments = getDadoRoundedContourCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
            elif self.DADO_REPLACEMENT.startswith('rounded-contour-') and not cw:
                segments = getDadoRoundedContourCCW(dado, securityDistance=securityDistance, routerRadius=routerRadius)
            else:
                assert False, "Unknown dado replacement!"

        self.updateViewBoxBySegments(segments=segments, **keywords)

        return segments

    def createVerticalDadoWithAnyTool(self, **keywords):
        dado = keywords['dado']

        if self.DADO_REPLACEMENT == 'middle-line':
            radiusCompensation = self.DIN_RADIUS_COMPENSATION_NONE
        else:
            cw = self.isDadoRabbetReplacementCW(dado, self.DADO_REPLACEMENT)
            radiusCompensation = self.DIN_RADIUS_COMPENSATION_RIGHT if cw else self.DIN_RADIUS_COMPENSATION_LEFT

        segments = self.generateDadoSegments(**keywords)

        return self.createContourFromSegmentList(
            createStartBlock=self.__createCommonStartBlock,
            createEndBlock=self.__createCommonEndBlock,
            radiusCompensation=radiusCompensation,
            segmentList=segments,
            isClosed=self.isDadoClosed(**keywords),
            zValue=dado.depth,
            additionalVariables={
                'width': self.formatAsCoordinate(data=dado.width),
                'aangle': self.formatAsAngle(data=90),  # Compatibility to miter, to be able to use the same saw
            },
            presenterAttributeGenerator=self.getPresentationAttributesForDados,
            **keywords)

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        return self.createVerticalDadoWithAnyTool(**keywords)

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        return self.createVerticalDadoWithAnyTool(**keywords)

    def createVerticalDadoWithRouter(self, **keywords):
        return self.createVerticalDadoWithAnyTool(**keywords)

    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        if operator.xor(self.MITER_CW, not self.isStandardOrigin(**keywords)):
            adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCW)
        else:
            adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCCW)

    def createVerticalMiter(self, **keywords):
        miter = keywords['miter']
        segmentList = getStartEndLine(miter)
        radiusCompensation = self.DIN_RADIUS_COMPENSATION_LEFT if self.MITER_CW else self.DIN_RADIUS_COMPENSATION_RIGHT

        return self.createContourFromSegmentList(
            createStartBlock=self.__createCommonStartBlock,
            createEndBlock=self.__createCommonEndBlock,
            radiusCompensation=radiusCompensation,
            segmentList=segmentList,
            isClosed=False,
            zValue=keywords['surface'].thickness,
            additionalVariables={'aangle': self.formatAsAngle(data=miter.getAngle())},
            presenterAttributeGenerator=self.getPresentationAttributesForMiters,
            **keywords)

    def createNCStop(self, **keywords):
        raise NotImplementedError

    def __createCommonStartBlock(self, **keywords):
        vertex = keywords['startVertex']
        # d supports no unit, that is the reason we have to define a viewBox in <svg>
        self.currentPath = 'M {x} {y}'.format(x=self.formatAsCoordinate(data=vertex.position.x), y=self.formatAsCoordinate(data=vertex.position.y))
        return ''

    # Needed for tests, as we have no paths that are so long
    def getPathWrapWidth(self):
        # In section 8.3.1 "General information about path data" of the SVG 1.1 (Second Edition), it is told that the path should be wrapped in lines of max. 255 characters
        return 200

    def __createCommonEndBlock(self, **keywords):
        if keywords['isClosed']:
            self.currentPath += ' z'

        variables = {
            'depth': self.formatAsCoordinate(data=keywords['zValue']),
            'radiuscompensation': keywords['radiusCompensation'],
            'rc': keywords['radiusCompensation'],  # Layer names are long enough; we do not want that users have to always type the long {radiuscompensation}
        }
        if 'additionalVariables' in keywords:
            variables.update(keywords['additionalVariables'])
        title = self.getCookedName(
            template=keywords['machining'].toolID,
            variables=variables,
            **keywords
        )

        presentation = keywords['presenterAttributeGenerator'](**keywords)
        path = textwrap.fill(self.currentPath, width=self.getPathWrapWidth(), subsequent_indent=12 * ' ', break_on_hyphens=False)
        return '\n'.join((
            2 * self.INDENT + '<path d="{d}"{presentation}>'.format(d=path, presentation=presentation),
            3 * self.INDENT + '<title>{title}</title>'.format(title=title),
            2 * self.INDENT + '</path>',
            '',
        ))

    def createContourFromSegmentList(self, **keywords):
        return ''.join(self.generateContourFromSegmentList(**keywords))

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        endPosition = segment.getEndVertex().position
        return self.createSimpleLineSegment(endPosition=endPosition, **keywords)

    def createSimpleLineSegment(self, **keywords):
        endPosition = keywords['endPosition']
        self.currentPath += ' '.join((
            '',
            'L',
            '{x}'.format(x=self.formatAsCoordinate(data=endPosition.x)),
            '{y}'.format(y=self.formatAsCoordinate(data=endPosition.y)),
        ))
        return ''

    def createRadiusSegment(self, **keywords):
        CW = (False, False)
        CCW = (False, True)
        CW180 = (True, False)
        CCW180 = (True, True)

        segment = keywords['segment']
        isLargeArc, sweep = segment.getDirection(CW, CCW, CW180, CCW180)
        radius = segment.getRadius()
        endPosition = segment.getEndVertex().position

        return self.createSimpleRadiusSegment(isLargeArc=isLargeArc, sweep=sweep, radius=radius, endPosition=endPosition, **keywords)

    def createSimpleRadiusSegment(self, **keywords):
        endPosition = keywords['endPosition']
        radius = keywords['radius']

        self.currentPath += ' '.join((
            '',
            'A',
            '{rx}'.format(rx=self.formatAsAngle(data=radius)),
            '{ry}'.format(ry=self.formatAsAngle(data=radius)),
            '{rotation}'.format(rotation=0),
            '{largeArc:d}'.format(largeArc=keywords['isLargeArc']),
            '{sweep:d}'.format(sweep=keywords['sweep']),
            '{x}'.format(x=self.formatAsCoordinate(data=endPosition.x)),
            '{y}'.format(y=self.formatAsCoordinate(data=endPosition.y)),
        ))
        return ''

    def createCommonStartBlockWithArcApproach(self, **keywords):
        isLeft = operator.xor(keywords['radiusCompensation'] == self.DIN_RADIUS_COMPENSATION_LEFT, not self.isStandardOrigin(**keywords))

        isClockwise, _approachDelta, approachPoint, centerPoint, _tangentialAngle = createArcApproach(left=isLeft, **keywords)
        endPosition = keywords['startVertex'].position
        radius = (approachPoint - centerPoint).magnitude()

        keywords['startVertex'] = CornerVertex(position=approachPoint)
        self.viewBox.updateByStartEndPoints(approachPoint, endPosition)

        return ''.join((
            self.__createCommonStartBlock(**keywords),
            self.createSimpleRadiusSegment(
                endPosition=endPosition,
                isLargeArc=False,
                sweep=not isClockwise,
                radius=radius,
                **keywords),
        ))

    def createCommonEndBlockWithArcDeparture(self, **keywords):
        isLeft = operator.xor(keywords['radiusCompensation'] == self.DIN_RADIUS_COMPENSATION_LEFT, not self.isStandardOrigin(**keywords))

        isClockwise, _departureDelta, departurePoint, centerPoint = createArcDeparture(left=isLeft, **keywords)
        radius = (departurePoint - centerPoint).magnitude()

        self.viewBox.updateByPoint(departurePoint)

        keywords['isClosed'] = False

        return ''.join((
            self.createSimpleRadiusSegment(
                endPosition=departurePoint,
                sweep=not isClockwise,
                isLargeArc=False,
                radius=radius,
                **keywords),
            self.__createCommonEndBlock(**keywords),
        ))

    def createCommonStartBlockWithLinearApproach(self, **keywords):
        _approachDelta, approachPoint = createLinearApproach(ncexport=self, **keywords)
        endPosition = keywords['startVertex'].position

        keywords['startVertex'] = CornerVertex(position=approachPoint)
        self.viewBox.updateByStartEndPoints(approachPoint, endPosition)

        return ''.join((
            self.__createCommonStartBlock(**keywords),
            self.createSimpleLineSegment(endPosition=endPosition, **keywords),
        ))

    def createCommonEndBlockWithLinearDeparture(self, **keywords):
        _departureDelta, departurePoint = createLinearDeparture(ncexport=self, **keywords)
        self.viewBox.updateByPoint(departurePoint)

        keywords['isClosed'] = False

        return ''.join((
            self.createSimpleLineSegment(endPosition=departurePoint, **keywords),
            self.__createCommonEndBlock(**keywords),
        ))

    def createCommonStartBlockWithPotentialApproach(self, **keywords):
        if 'arc' in keywords['approach']:
            return self.createCommonStartBlockWithArcApproach(**keywords)
        elif 'linear' in keywords['approach']:
            return self.createCommonStartBlockWithLinearApproach(**keywords)
        else:
            return self.__createCommonStartBlock(**keywords)

    def createCommondEndBlockWithPotentialDeparture(self, **keywords):
        if 'arc' in keywords['departure']:
            return self.createCommonEndBlockWithArcDeparture(**keywords)
        elif 'linear' in keywords['departure']:
            return self.createCommonEndBlockWithLinearDeparture(**keywords)
        else:
            return self.__createCommonEndBlock(**keywords)

    def createSizingStartBlock(self, **keywords):
        return self.createCommonStartBlockWithPotentialApproach(
            approach=self.SIZING_ENTRY_OF_PROFILE,
            securityDistance=self.SIZING_APPROACH_DEPARTURE_DISTANCE,
            **keywords)

    def createSizingEndBlock(self, **keywords):
        return self.createCommondEndBlockWithPotentialDeparture(
            departure=self.SIZING_ENTRY_OF_PROFILE,
            isClosed=True,
            securityDistance=self.SIZING_APPROACH_DEPARTURE_DISTANCE,
            presenterAttributeGenerator=self.getPresentationAttributesForSizing,
            **keywords)

    def createContourStartBlock(self, **keywords):
        return self.createCommonStartBlockWithPotentialApproach(
            approach=keywords['contour'].approach,
            radiusCompensation=self.getDinRadiusCompensation(**keywords),
            securityDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
            **keywords)

    def createContourEndBlock(self, **keywords):
        return self.createCommondEndBlockWithPotentialDeparture(
            departure=keywords['contour'].departure,
            isClosed=keywords['contour'].isClosed(),
            radiusCompensation=self.getDinRadiusCompensation(**keywords),
            securityDistance=self.CONTOUR_APPROACH_DEPARTURE_DISTANCE,
            **keywords)

    def createRabbetContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def createRabbetContourEndBlock(self, **keywords):
        radiusCompensation = self.getDinRadiusCompensation(**keywords)
        return self.__createCommonEndBlock(
            isClosed=self.isRabbetClosed(**keywords),
            radiusCompensation=radiusCompensation,
            **keywords)

    def createNCMacroCode(self, **keywords):
        return '\n'.join((
            2 * self.INDENT + '<g>',
            super().createNCMacroCode(**keywords),
            2 * self.INDENT + '</g>',
            ''
        ))

    # Footer
    def createEndOfFile(self, **_keywords):
        return self.createMainGroupEndTag()

    def createMainGroupEndTag(self, **_keywords):
        return self.INDENT + '</g>'
