# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import copy
import itertools
import logging
import math

import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.localization import getLocalizedString, WOODFLASH_SIDE_DESCRIPTIONS_RESOURCEGROUP, WOODFLASH_COMMENT_MESSAGES
from interiorcad.backend.postprocessors.surfacepostprocessor.drillingswithtoolidtofiveaxisdrillingsconverter import DrillingsWithToolIDToFiveAxisDrillingsConverter
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual, isNearlyGreater
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.point import Point
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.ncexport.approachcreator import createArcApproach, createArcDeparture, createLinearApproach, createLinearDeparture
from interiorcad.interface.ncexport.multisawdado import MultiSawDado
from interiorcad.interface.surface import FiveAxisDrilling, FiveAxisCenterPunch, HorizontalDrilling, HorizontalCenterPunch, NCMacro2D
from interiorcad.interface.surface.acangles import normalizeAngle
from interiorcad.interface.surface.contour2d import Contour2D, HorizontalContour2D
from interiorcad.interface.surface.dado2d import HorizontalDado2D
from interiorcad.interface.surface.machiningplanes import MachiningPlanes
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.interface.surface.pocket2d import HorizontalPocket2D
from interiorcad.ncexport.base import NCExportBase


class WoodFlashExport(NCExportBase):
    EXPORT_FORMAT = 'WoodFlash'

    RADIUS_COMPENSATION_NONE = 0
    RADIUS_COMPENSATION_LEFT = 1
    RADIUS_COMPENSATION_RIGHT = 2

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.machiningPlanes = None
        self.profileNameIndex = None

        machineConfiguration = self.getMachineConfiguration()
        self.VERSION = machineConfiguration.getAsInt('woodflash-version', 3)
        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 2)
        self.ADDITIONAL_DEPTH_FOR_MITER = machineConfiguration.getAsFloat(MachineConfigurationKeys.ADDITIONAL_DEPTH_FOR_MITER_KEY, 5)
        self.SIZING_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('approach-departure-distance', 30)
        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('contour-approach-departure-distance', 30)
        self.DADO_PRESAW_DEPTH = machineConfiguration.getAsFloat('dado-presaw-depth', 0)
        self.DADO_SAW_OVERLAP_FACTOR = machineConfiguration.getAsFloat('dado-saw-overlap-factor', 0.15)
        self.DADO_WIDTH_SUPPORTED = machineConfiguration.getAsBool('dado-width-supported', True)

        self.CREATE_SIZING_WITH_MACRO_NAME = machineConfiguration.get(MachineConfigurationKeys.CREATE_SIZING_WITH_MACRO_KEY, '0')  # Instead boolean CREATE_SIZING_WITH_MACRO. Note the old values 0|1.
        self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR = machineConfiguration.getAsBool('create-sizing-with-three-side-contour', False)
        self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS = machineConfiguration.getAsBool('sizing-mill-through-for-non-rectangular-parts', True)
        self.CREATE_SIZING_AS_NESTING_GEOMETRY = machineConfiguration.getAsBool('create-sizing-as-nesting-geometry', False)

        self.USE_SECURITY_DISTANCE_FOR_CONTINUOUS_DADO_SAW = machineConfiguration.getAsBool('use-security-distance-for-continuous-dado-saw', False)
        self.USE_VERTICAL_POCKET_MACRO_IN_WDFLASH2 = machineConfiguration.getAsBool('use-vertical-pocket-macro-in-wdflash2', False)
        self.CONVERT_DRILLINGS_WITH_TOOLID_TO_FIVE_AXIS_DRILLINGS = machineConfiguration.getAsBool('convert-drillings-with-toolid-to-five-axis-drillings', False)
        self.MACHINE_TYPE = machineConfiguration.get('machine-type', 'standard')
        self.USE_CUSTOM_USER_DESCRIPTION = machineConfiguration.getAsBool('use-custom-user-description', False)
        self.CREATE_LABEL_CALL = machineConfiguration.getAsBool('create-label-call', False)

        self.ROTATE_WORKPIECE_180 = machineConfiguration.getAsBool('rotate-workpiece-180', False)

        self.createZValueConverters(**keywords)

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_UP)

    def getNCFileExtension(self, **_keywords):
        return '.tcn'

    def createCommentWOBlanksLowerCase(self, **keywords):
        return 'W$={comment}'.format(comment=keywords['comment'].replace(' ', '_').lower())

    def createNameWOBlanksLowerCase(self, **keywords):
        return keywords['name'].replace(' ', '_').lower()

    def getPostAssignToolProcessors(self, **_keywords):
        if self.CONVERT_DRILLINGS_WITH_TOOLID_TO_FIVE_AXIS_DRILLINGS:
            return DrillingsWithToolIDToFiveAxisDrillingsConverter(toolValidator=self.isValidToolId),
        else:
            return ()

    def createSide(self, **keywords):
        if self.ROTATE_WORKPIECE_180:
            surface = keywords['surface']
            surface.rotate180()

        return super().createSide(**keywords)

    # Header
    def createHeader(self, **_keywords):
        self.profileNameIndex = 0

        TPA_ALBATROS_EDICAD_INDIVIDUAL_LABEL = self.getMachineConfiguration().get('tpa-albatros-edicad-individual-label', '')
        if TPA_ALBATROS_EDICAD_INDIVIDUAL_LABEL:
            return r'TPA\ALBATROS\EDICAD\{}'.format(TPA_ALBATROS_EDICAD_INDIVIDUAL_LABEL)
        else:
            if self.VERSION >= 3:
                return r'TPA\ALBATROS\EDICAD\02.00:955:r0w0h0s1'
            else:
                return r'TPA\ALBATROS\EDICAD\01.00:20:r0w0'

    def customUserUserDescriptionFileDefaultText(self, **_keywords):
        return ' '.join((
            '$=[name={WorkpieceName}]',
            '[comment={Comment}]',
            '[material={ItemID}]',
            '[project={ProjectID}]',
            '[customer={ProjectCommission}]',
            '[construction={ProjectConstruction}]',
            '[positionnumber={PositionNumber}]',
            '[positiondescription={PositionDescription}]',
            '[cabinetname={CabinetName}]',
            '[edgefront={EdgeFront}]',
            '[edgeback={EdgeBack}]',
            '[edgeleft={EdgeLeft}]',
            '[edgeright={EdgeRight}]',
            '[ncprogramname={NCProgramName}]',
            '[ncprogramfilename={NCProgramFilename}]',
            '[ncprogramfolder={NCProgramFolder}]',
            '[ncprogrampath={NCProgramPath}]',
        ))

    def createCustomUserDescription(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='Wood Flash Description.txt',
            customUserVariableFileDefaultText=self.customUserUserDescriptionFileDefaultText(**keywords),
            **keywords)

    def createWorkpieceComment(self, **keywords):
        projectInfo = self.getProjectInfo()
        workpieceComments = self.getWorkpieceComments(**keywords)
        isProjectIDValid = ExportHelper.isProjectIDValid(self, **projectInfo)

        customUserDescription = self.createCustomUserDescription(**keywords) if self.USE_CUSTOM_USER_DESCRIPTION else None

        if customUserDescription:
            return customUserDescription

        else:
            return joinIfNotNone('\n',
                '$={WorkpieceLabel}: {WorkpieceName}'.format_map(workpieceComments),
                '$={CommentLabel}: {Comment}'.format_map(workpieceComments) if workpieceComments['Comment'] else None,
                '$={ItemIDLabel}: {ItemID}'.format_map(workpieceComments),
                '$={Edge1Label}: {Edge1}'.format_map(workpieceComments) if workpieceComments['Edge1'] else None,  # front
                '$={Edge2Label}: {Edge2}'.format_map(workpieceComments) if workpieceComments['Edge2'] else None,  # back
                '$={Edge3Label}: {Edge3}'.format_map(workpieceComments) if workpieceComments['Edge3'] else None,  # left
                '$={Edge4Label}: {Edge4}'.format_map(workpieceComments) if workpieceComments['Edge4'] else None,  # right
                '$={CabinetNameLabel}: {CabinetName}'.format_map(workpieceComments) if workpieceComments['CabinetName'] else None,
                '$={ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}'.format_map(workpieceComments) if isProjectIDValid else None,
                '$={PositionNumberLabel}: {PositionNumber} {PositionDescription}'.format_map(workpieceComments) if isProjectIDValid else None,
                '$={CountLabel}: {Count}'.format_map(workpieceComments),
                '$=Wood Flash {version}'.format(version=self.VERSION),
            )

    def createSideEnumeration(self, **_keywords):
        return '::SIDE='

    def createVirtualSides(self, **keywords):
        self.generatePlanes(**keywords)
        lines = '\n'.join((self.createVirtualSideDefinition(plane=plane, **keywords) for plane in self.machiningPlanes.planes if not plane.predefined))

        if lines:
            return '\n'.join((
                'GEO{ ::NF=1',
                lines,
                '}GEO',
            ))

    def __createPredefinedPlanes(self, **keywords):
        size = keywords['surface'].getFinishedSize()

        yield Plane(
            predefined=True, sideID=self.SIDE_FRONT,
            matrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_RIGHT,
            matrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_BACK,
            matrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
        )
        yield Plane(
            predefined=True, sideID=self.SIDE_LEFT,
            matrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )

    def generatePlanes(self, **keywords):
        surface = keywords['surface']

        self.machiningPlanes = MachiningPlanes(lastUsedSideId=self.SIDE_LEFT)
        self.machiningPlanes.planes.extend(self.__createPredefinedPlanes(**keywords))
        self.machiningPlanes.createPlanes(filter(lambda machining: machining.matrix.w != Vector(0, 0, 1),
            filter(lambda machining: isinstance(machining, (
                HorizontalDrilling, HorizontalCenterPunch, HorizontalDado2D, HorizontalContour2D, HorizontalPocket2D,
                FiveAxisDrilling, FiveAxisCenterPunch, NCMacro2D
            )), surface.machinings),
        ))

    def createVirtualSideVector(self, **keywords):
        point = keywords['point']
        return '|'.join((
            '{x}'.format(x=self.formatAsCoordinate(data=point.x)),
            '{y}'.format(y=self.formatAsCoordinate(data=point.y)),
            '{z}'.format(z=self.formatAsCoordinate(data=point.z)),
        ))

    def createVirtualSideDefinition(self, **keywords):
        plane = keywords['plane']

        wopMatrix = plane.wopMatrix
        origin = wopMatrix.origin
        point1 = origin
        point2 = origin + wopMatrix.u
        point3 = origin + wopMatrix.v

        return '\n'.join((
            'GSIDE#{number}{{'.format(number=plane.sideID),
            '#1=' + self.createVirtualSideVector(point=point1, **keywords),
            '#2=' + self.createVirtualSideVector(point=point2, **keywords),
            '#3=' + self.createVirtualSideVector(point=point3, **keywords),
            '#P1=0',
            '#P2=0',
            '#P3=0',
            '}GSIDE',
        ))

    def createWorkpieceDefinition(self, **keywords):
        surface = keywords['surface']
        return ' '.join((
            '::UNm',
            'DL={length}'.format(length=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            'DH={width}'.format(width=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            'DS={thick}'.format(thick=self.formatAsCoordinate(data=surface.thickness)),
        ))

    def createVariablesSection(self, **keywords):
        surface = keywords['surface']

        if self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR and self.CREATE_SIZING_WITH_MACRO_NAME == '0':  # Contour
            useOffsetY = False
        elif self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR and self.CREATE_SIZING_WITH_MACRO_NAME == '1523':  # Macro G1037 / #1523 explicitly for Creator 950
            useOffsetY = False
        else:
            useOffsetY = True

        return '\n'.join((
            'EXE{',
            '#0=0',
            '#1=0',
            '#2=0',
            '#3=0',
            '#4=0',
            '}EXE',
            'OFFS{',
            '#0={x}|{x}'.format(x=self.formatAsCoordinate(data=surface.getAdditionX())),
            '#1={y}|{y}'.format(y=self.formatAsCoordinate(data=surface.getAdditionY()) if useOffsetY else '0'),
            '#2=0|0',
            '}OFFS',
            'VARV{',
            '#0=1|1',
            '#1=2|2',
            '#2=3|3',
            '#3=4|4',
            '#4=0|0',
            '#5=0|0',
            '#6=0|0',
            '#7=0|0',
            '}VARV',
            'VAR{',
            '}VAR',
            'SPEC{',
            '}SPEC',
            'INFO{',
            '}INFO',
            'OPTI{',
            ':: OPTDEF=1 OPTIMIZE=%;0 OPTMIN=0 OPT3=0 OPT0=0 OPTTOOL=0 OPT2=0 OPTX=0 OPTY=0 OPTR=0 OPT4=0 OPT6=0 OPT7=0 LSTCOD=0%1%2%3 LTOOLFR=0 LTOOLPN=0 OPTF1=0 OOO=0.5',
            '}OPTI',
            'LINK{',
            '}LINK',
        ))

    def createLabelCall(self, **_keywords):
        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            ' '.join((
                'W#102{ ::WTg',
                'W$=~label',
                '#1=l/2',
                '#2=h/2',
                '#205=1',
                '#9505=100',
                '}W',
            )),
        )

    def createHead(self, **keywords):
        self.currentSide = None
        return joinIfNotNone('\n',
            self.createHeader(**keywords),
            self.createWorkpieceComment(**keywords),
            self.createSideEnumeration(**keywords),
            self.createWorkpieceDefinition(**keywords),
            self.createVariablesSection(**keywords),
            self.createVirtualSides(**keywords),
            self.createLabelCall(**keywords) if self.CREATE_LABEL_CALL else None,
            ''
        )

    # Machinings

    SIDE_ABSOLUTE = 0  # Face 0 / SIDE#0
    SIDE_MAIN, SIDE_OPPOSITE, SIDE_FRONT, SIDE_RIGHT, SIDE_BACK, SIDE_LEFT = range(1, 7)
    FIRST_VIRTUAL_SIDE_ID = 7

    def isValidToolId(self, **keywords):
        return 'tool' in keywords and keywords['tool'] and keywords['tool'].isToolIDValidIntDirect()

    def getDefaultSideComment(self, **keywords):
        # The file format use "side", but the manual call it "face".
        side = keywords['side']
        if side == self.SIDE_MAIN:
            translation = 'top face'
        elif side == self.SIDE_OPPOSITE:
            translation = 'bottom face'
        elif side == self.SIDE_FRONT:
            translation = 'front face'
        elif side == self.SIDE_RIGHT:
            translation = 'right face'
        elif side == self.SIDE_BACK:
            translation = 'back face'
        elif side == self.SIDE_LEFT:
            translation = 'left face'
        elif side == self.SIDE_ABSOLUTE:
            side = 8
            translation = 'overall face'
        else:
            side = 7
            translation = 'fictive face'

        return getLocalizedString(WOODFLASH_SIDE_DESCRIPTIONS_RESOURCEGROUP, side, translation)

    def openSide(self, **keywords):
        assert self.currentSide is None, "A side is currently still open"
        side = keywords['side']
        self.currentSide = side
        comment = keywords['comment'] if 'comment' in keywords else self.getDefaultSideComment(**keywords)
        return joinIfNotNone('\n',
            'SIDE#{side}{{'.format(side=side),
            '$={comment}'.format(comment=comment) if comment else None,
        )

    def closeSide(self, **_keywords):
        if self.currentSide is not None:
            self.currentSide = None
            return '}SIDE'

    def changeSide(self, **keywords):
        if keywords['side'] != self.currentSide:
            return joinIfNotNone('\n',
                self.closeSide(**keywords),
                self.openSide(**keywords),
            )

    def createMachinings(self, **keywords):
        return joinIfNotNone('\n',
            super().createMachinings(**keywords).rstrip(),  # ### IT-119. wie können wir uns den rstrip sparen? Alle meine Versuche, es zu tun, schlugen fehl. -- NG
            self.closeSide(**keywords),
        )

    APPROACH_DEPARTURE_NOT_MANAGED = 0
    APPROACH_DEPARTURE_MANAGED = 1

    APPROACH_DEPARTURE_LINEAR = 0
    APPROACH_DEPARTURE_LEFT_ARC = 1
    APPROACH_DEPARTURE_RIGHT_ARC = 2
    APPROACH_DEPARTURE_3D_ARC = 3

    APPROACH_MODE_LINEAR_PIERCING = (APPROACH_DEPARTURE_MANAGED, APPROACH_DEPARTURE_LINEAR, False)
    APPROACH_MODE_LINEAR_FLYING = (APPROACH_DEPARTURE_MANAGED, APPROACH_DEPARTURE_LINEAR, True)
    APPROACH_MODE_ARC_PIERCING = (APPROACH_DEPARTURE_MANAGED, APPROACH_DEPARTURE_RIGHT_ARC, False)
    APPROACH_MODE_ARC_FLYING = (APPROACH_DEPARTURE_MANAGED, APPROACH_DEPARTURE_3D_ARC, True)
    APPROACH_MODE_WITHOUT = (APPROACH_DEPARTURE_NOT_MANAGED, APPROACH_DEPARTURE_LINEAR, False)
    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_WITHOUT

    DEFAULT_INITIAL_Z = 1

    def __createCommonContourApproach(self, **keywords):
        if keywords['approachManaged']:
            return joinIfNotNone(' ',
                '#8180={approachManaged}'.format(approachManaged=keywords['approachManaged']),
                '#8181={approachType}'.format(approachType=keywords['approachType']),
                '#8182={approachDistance}'.format(approachDistance=self.formatAsCoordinate(data=keywords['securityDistance'])),
                '#8184={z}'.format(z=self.DEFAULT_INITIAL_Z) if 'useApproachInitialZ' in keywords and keywords['useApproachInitialZ'] else None,
            )

    def __createCommonContourDeparture(self, **keywords):
        if keywords['departureManaged']:
            return joinIfNotNone(' ',
                '#8185={departureManaged}'.format(departureManaged=keywords['departureManaged']),
                '#8186={departureType}'.format(departureType=keywords['departureType']),
                '#8187={departureDistance}'.format(departureDistance=self.formatAsCoordinate(data=keywords['securityDistance'])),
                '#8189={z}'.format(z=self.DEFAULT_INITIAL_Z) if 'useDepartureInitialZ' in keywords and keywords['useDepartureInitialZ'] else None,
            )

    def useSeparateContourApproach(self, **keywords):
        return 'useSeparateContourApproach' in keywords and keywords['useSeparateContourApproach']

    def __createCommonContourSetup(self, **keywords):
        def sizingIsContourCreator(**keywords):
            return 'contourCreator' in keywords and keywords['contourCreator'] == 'Sizing'

        machining = keywords['machining']
        toolID = machining.toolID
        vertex = keywords['startVertex']
        approachDelta = keywords['approachDelta']

        separateApproachWithInitialZ = self.useSeparateContourApproach(**keywords) and 'useApproachInitialZ' in keywords and keywords['useApproachInitialZ']

        profileName = None
        if hasattr(machining, 'profileName'):
            profileName = machining.profileName

        return joinIfNotNone(' ',
            'W#89{ ::WTs',
            'WN={name}'.format(name=profileName) if profileName else None,  # Actually only needed as profile name for free-form pockets from interiorcad 2021 BF5 on.
            '#1={x}'.format(x=self.formatAsCoordinateAddition(lhs=vertex.position.x, rhs=approachDelta.x)),
            '#2={y}'.format(y=self.formatAsCoordinateAddition(lhs=vertex.position.y, rhs=approachDelta.y)),
            '#3={z}'.format(z=self.DEFAULT_INITIAL_Z if separateApproachWithInitialZ else self.formatAsCoordinate(data=keywords['zValue'])),
            '#205={toolID}'.format(toolID=toolID),
            '#40={rc}'.format(rc=keywords['radiusCompensation']),
            '#201=1',
            '#203=1',
            '#1001={type}'.format(type=self.MILLING_TYPE),
            '#2005={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) else None,
            '#8015=0',
            '#8101=0',
            '#8135=0',
            '#8136=0',
            self.__createCommonContourApproach(**keywords) if not self.useSeparateContourApproach(**keywords) else None,
            self.__createCommonContourDeparture(**keywords) if not self.useSeparateContourApproach(**keywords) else None,
            '#9511={nestingGeometry:d}'.format(nestingGeometry=self.CREATE_SIZING_AS_NESTING_GEOMETRY and sizingIsContourCreator(**keywords)),
            '}W',
        )

    def createSeparateArcContour(self, **keywords):
        # normally used by Wood Flash 2 and older TPA
        position = keywords['position']
        tangentialAngle = keywords['tangentialAngle'] if 'tangentialAngle' in keywords else ''

        if 'useApproachInitialZ' in keywords and keywords['useApproachInitialZ']:
            destinationZ = keywords['zValue']
        elif 'useDepartureInitialZ' in keywords and keywords['useDepartureInitialZ']:
            destinationZ = self.DEFAULT_INITIAL_Z
        else:
            destinationZ = None

        return joinIfNotNone(' ',
            'W#{code}{{ ::WTa'.format(code=2117 if 'approachType' in keywords else 2115),
            '#8015={type:d}'.format(type=0),  # 0=absolute 1=relative
            '#1={x}'.format(x=self.formatAsCoordinate(data=position.x)),
            '#2={y}'.format(y=self.formatAsCoordinate(data=position.y)),
            '#3={z}'.format(z=self.formatAsCoordinate(data=destinationZ)) if destinationZ is not None else None,
            '#8046={angle}'.format(angle=self.formatAsAngle(data=tangentialAngle)) if tangentialAngle else None,
            '#8019={direction}'.format(direction=1 if keywords['clockwise'] else 2),  # 1=CW 2=CCW
            '}W',
        )

    def createSeparateLinearContour(self, **keywords):
        # normally used by Wood Flash 2 and older TPA
        position = keywords['position']

        if 'useApproachInitialZ' in keywords and keywords['useApproachInitialZ']:
            destinationZ = keywords['zValue']
        elif 'useDepartureInitialZ' in keywords and keywords['useDepartureInitialZ']:
            destinationZ = self.DEFAULT_INITIAL_Z
        else:
            destinationZ = None

        return joinIfNotNone(' ',
            'W#2201{ ::WTl',
            '#8015={type:d}'.format(type=0),  # 0=absolute 1=relative
            '#1={x}'.format(x=self.formatAsCoordinate(data=position.x)),
            '#2={y}'.format(y=self.formatAsCoordinate(data=position.y)),
            '#3={z}'.format(z=self.formatAsCoordinate(data=destinationZ)) if destinationZ is not None else None,
            '}W',
        )

    def __createCommonContourStartBlock(self, **keywords):
        return joinIfNotNone('\n',
            self.changeSide(side=keywords['sideID'] if 'sideID' in keywords else self.SIDE_MAIN),
            self.__createCommonContourSetup(**keywords),
        )

    def __createCommonContourEndBlock(self, **_keywords):
        return ''

    def createCommonStartBlockWithArcApproach(self, **keywords):
        isClockwise, approachDelta, _approachPoint, _centerPoint, tangentialAngle = createArcApproach(left=keywords['radiusCompensation'] == self.RADIUS_COMPENSATION_LEFT, **keywords)

        return '\n'.join((
            self.__createCommonContourStartBlock(approachDelta=approachDelta, **keywords),
            self.createSeparateArcContour(position=keywords['startVertex'].position, tangentialAngle=tangentialAngle, clockwise=isClockwise,
                **keywords),
        ))

    def createCommonEndBlockWithArcDeparture(self, **keywords):
        isClockwise, _departureDelta, departurePoint, _centerPoint = createArcDeparture(left=keywords['radiusCompensation'] == self.RADIUS_COMPENSATION_LEFT, **keywords)

        return '\n'.join((
            self.createSeparateArcContour(position=departurePoint, clockwise=isClockwise, **keywords),
            self.__createCommonContourEndBlock(**keywords),
        ))

    def createCommonStartBlockWithLinearApproach(self, **keywords):
        approachDelta, _approachPoint = createLinearApproach(ncexport=self, **keywords)

        return '\n'.join((
            self.__createCommonContourStartBlock(approachDelta=approachDelta, **keywords),
            self.createSeparateLinearContour(position=keywords['segmentList'][0].getStartVertex().position, **keywords)
        ))

    def createCommonEndBlockWithLinearDeparture(self, **keywords):
        _departureDelta, departurePoint = createLinearDeparture(ncexport=self, **keywords)

        return '\n'.join((
            self.createSeparateLinearContour(position=departurePoint, **keywords),
            self.__createCommonContourEndBlock(**keywords),
        ))

    def createCommonStartBlockWithPotentialApproach(self, **keywords):
        if keywords['useSeparateContourApproach']:  # Compatibility function for Wood Flash 2; do NOT use for new code.
            approachType = keywords['approachType']
            if approachType == self.APPROACH_DEPARTURE_RIGHT_ARC or approachType == self.APPROACH_DEPARTURE_LEFT_ARC or approachType == self.APPROACH_DEPARTURE_3D_ARC:
                return self.createCommonStartBlockWithArcApproach(**keywords)
            elif approachType == self.APPROACH_DEPARTURE_LINEAR:
                return self.createCommonStartBlockWithLinearApproach(**keywords)

        return self.__createCommonContourStartBlock(approachDelta=Point(0, 0), **keywords)

    def createCommondEndBlockWithPotentialDeparture(self, **keywords):
        if keywords['useSeparateContourDeparture']:  # Compatibility function for Wood Flash 2; do NOT use for new code.
            departureType = keywords['departureType']
            if departureType == self.APPROACH_DEPARTURE_RIGHT_ARC or departureType == self.APPROACH_DEPARTURE_LEFT_ARC or departureType == self.APPROACH_DEPARTURE_3D_ARC:
                return self.createCommonEndBlockWithArcDeparture(**keywords)
            elif departureType == self.APPROACH_DEPARTURE_LINEAR:
                return self.createCommonEndBlockWithLinearDeparture(**keywords)

        return self.__createCommonContourEndBlock(**keywords)

    def __getSizingRadiusCompensation(self, **keywords):
        return WoodFlashExport.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else WoodFlashExport.RADIUS_COMPENSATION_LEFT

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']

        if machining.isVertical():
            left = self.RADIUS_COMPENSATION_LEFT
            right = self.RADIUS_COMPENSATION_RIGHT
        else:
            plane, _positionOnPlane = self.__calculatePlane(**keywords)
            rightHanded = plane.isWopMatrixRightHanded()
            left = self.RADIUS_COMPENSATION_LEFT if rightHanded else self.RADIUS_COMPENSATION_RIGHT
            right = self.RADIUS_COMPENSATION_RIGHT if rightHanded else self.RADIUS_COMPENSATION_LEFT

        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=self.RADIUS_COMPENSATION_NONE, left=left, right=right)

    def useSeparateContourDeparture(self, **keywords):
        return 'useSeparateContourDeparture' in keywords and keywords['useSeparateContourDeparture']

    def createSizingDefinitionForOnePolyline(self, **keywords):
        sizing = keywords['surface'].sizing

        keywords['creator'] = super().createSizingDefinitionForOnePolyline
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
        keywords['sizing'] = sizing

        machiningDepth = sizing.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machining=sizing, machiningDepth=machiningDepth, **keywords)

        return '\n'.join(self.generateCommonNoBuiltInDepthSteps(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    def createSizingStartBlock(self, **keywords):
        managed, approach, initialZ = self.getSizingApproachMode(**keywords)

        contourInfo = copy.copy(keywords)
        contourInfo['radiusCompensation'] = self.__getSizingRadiusCompensation(**keywords)
        contourInfo['approachManaged'] = managed
        contourInfo['approachType'] = approach
        contourInfo['useApproachInitialZ'] = initialZ
        contourInfo['securityDistance'] = self.SIZING_APPROACH_DEPARTURE_DISTANCE + keywords['sizing'].tool.getRouterDiameter() / 2
        contourInfo['contourCreator'] = 'Sizing'
        contourInfo['useSeparateContourApproach'] = self.VERSION <= 2

        if not self.useSeparateContourApproach(**contourInfo):
            contourInfo['departureManaged'] = managed
            contourInfo['departureType'] = approach
            contourInfo['useDepartureInitialZ'] = initialZ

        return self.createCommonStartBlockWithPotentialApproach(**contourInfo)

    def createContourStartBlock(self, **keywords):
        contour = keywords['contour']

        if contour.clearOut:
            self.profileNameIndex += 1
            contour.profileName = self.createNameWOBlanksLowerCase(name=self.getLocalizedMachiningComment(23, "Profile{index}").format(index=self.profileNameIndex), **keywords)
        else:
            contour.profileName = None

        approachManaged, approachType, useApproachInitialZ = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        contourInfo = copy.copy(keywords)
        contourInfo['radiusCompensation'] = self.getWopRadiusCompensation(**keywords)
        contourInfo['approachManaged'] = approachManaged
        contourInfo['approachType'] = approachType
        contourInfo['useApproachInitialZ'] = useApproachInitialZ
        contourInfo['securityDistance'] = self.CONTOUR_APPROACH_DEPARTURE_DISTANCE
        contourInfo['useSeparateContourApproach'] = self.VERSION <= 2
        # TODO CONTOUR2.0 IT-243 Currently feedRate is only implemented directly within createContourStartBlock, not other contour using machinings --mjk
        contourInfo['feedRate'] = contour.currentFeedRate

        if not self.useSeparateContourApproach(**contourInfo):
            departureManaged, departureType, useDepartureInitialZ = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
            contourInfo['departureManaged'] = departureManaged
            contourInfo['departureType'] = departureType
            contourInfo['useDepartureInitialZ'] = useDepartureInitialZ

        return self.createCommonStartBlockWithPotentialApproach(**contourInfo)

    def createRabbetContourStartBlock(self, **keywords):
        contourInfo = copy.copy(keywords)
        contourInfo['radiusCompensation'] = self.getWopRadiusCompensation(**keywords)
        contourInfo['approachManaged'] = self.APPROACH_DEPARTURE_MANAGED
        contourInfo['departureManaged'] = self.APPROACH_DEPARTURE_MANAGED
        contourInfo['approachType'] = self.APPROACH_DEPARTURE_LINEAR
        contourInfo['departureType'] = self.APPROACH_DEPARTURE_LINEAR
        contourInfo['securityDistance'] = self.getDadoRabbetSecurityDistance(**keywords)
        contourInfo['useSeparateContourApproach'] = self.VERSION <= 2
        return self.createCommonStartBlockWithPotentialApproach(**contourInfo)

    def createDadoContourStartBlock(self, **keywords):
        contourInfo = copy.copy(keywords)
        contourInfo['approachManaged'] = self.APPROACH_DEPARTURE_MANAGED
        contourInfo['departureManaged'] = self.APPROACH_DEPARTURE_MANAGED
        contourInfo['approachType'] = self.APPROACH_DEPARTURE_LINEAR
        contourInfo['departureType'] = self.APPROACH_DEPARTURE_LINEAR
        contourInfo['securityDistance'] = self.getDadoRabbetSecurityDistance(**keywords)
        contourInfo['useSeparateContourApproach'] = self.VERSION <= 2
        return self.createCommonStartBlockWithPotentialApproach(**contourInfo)

    def createSizingEndBlock(self, **keywords):
        managed, departure, initialZ = self.getSizingApproachMode(**keywords)

        contourInfo = copy.copy(keywords)
        contourInfo['radiusCompensation'] = self.__getSizingRadiusCompensation(**keywords)
        contourInfo['departureManaged'] = managed
        contourInfo['departureType'] = departure
        contourInfo['securityDistance'] = self.SIZING_APPROACH_DEPARTURE_DISTANCE + keywords['sizing'].tool.getRouterDiameter() / 2
        contourInfo['contourCreator'] = 'Sizing'
        contourInfo['useSeparateContourDeparture'] = self.VERSION <= 2
        contourInfo['useDepartureInitialZ'] = initialZ

        return self.createCommondEndBlockWithPotentialDeparture(**contourInfo)

    def createFreeFormPocket(self, **keywords):
        contour = keywords['contour']
        toolID = keywords['machining'].toolID

        # InnoWeek 2021-12: Would work better with depthPerStep = self.getRouterDepthPerStep(machiningDepth=machining.actualDepth, **keywords), but not possible with NoBuiltInDepthSteps  --mjk
        useSteps = False  # Just in case if we would integrate an alternative to NoBuiltInDepthSteps

        return joinIfNotNone(' ',
            'W#1041{ ::WTs',
            self.createCommentWOBlanksLowerCase(comment=self.getLocalizedMachiningComment(22, "Emptying Contour"), **keywords),
            '#8117={name}'.format(name=contour.profileName),
            '#8022={z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            '#8142=1',  # Ausräumen nach Außen
            '#8144={startZ}'.format(startZ=0) if useSteps else None,  # Anfangsmaß Z
            '#8148={useSteps}'.format(useSteps=1 if useSteps else 0),  # Freigabe nächster Durchlauf
            '#8145=0' if useSteps else None,  # Endmaß Z
            '#8146={depthPerStep}'.format(depthPerStep=keywords['depthPerStep']) if useSteps else None,  # Schritt Z
            '#8153={toolID}'.format(toolID=toolID),  # Fräser Setup
            '#8159=1',  # Restfläche ausräumen
            # Linie im Eingang und/oder Ausgang?
            '#8163={toolID}'.format(toolID=toolID),
            '}W',
        )

    def createContourEndBlock(self, **keywords):
        contour = keywords['contour']
        departureManaged, departureType, useDepartureInitialZ = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        contourInfo = copy.copy(keywords)
        contourInfo['radiusCompensation'] = self.getWopRadiusCompensation(**keywords)
        contourInfo['departureManaged'] = departureManaged
        contourInfo['departureType'] = departureType
        contourInfo['securityDistance'] = self.CONTOUR_APPROACH_DEPARTURE_DISTANCE
        contourInfo['useSeparateContourDeparture'] = self.VERSION <= 2
        contourInfo['useDepartureInitialZ'] = useDepartureInitialZ

        return joinIfNotNone('\n',
            self.createCommondEndBlockWithPotentialDeparture(**contourInfo),
            self.createFreeFormPocket(**keywords) + '\n' if contour.clearOut else None,
        )

    def createRabbetContourEndBlock(self, **keywords):
        contourInfo = copy.copy(keywords)
        contourInfo['departureType'] = self.APPROACH_DEPARTURE_LINEAR
        contourInfo['securityDistance'] = self.getDadoRabbetSecurityDistance(**keywords)
        contourInfo['useSeparateContourDeparture'] = self.VERSION <= 2
        return self.createCommondEndBlockWithPotentialDeparture(**contourInfo)

    def createDadoContourEndBlock(self, **keywords):
        contourInfo = copy.copy(keywords)
        contourInfo['departureType'] = self.APPROACH_DEPARTURE_LINEAR
        contourInfo['securityDistance'] = self.getDadoRabbetSecurityDistance(**keywords)
        contourInfo['useSeparateContourDeparture'] = self.VERSION <= 2
        return self.createCommondEndBlockWithPotentialDeparture(**contourInfo)

    def createRadiusSegment(self, **keywords):
        segment = keywords['segment']
        CW = 0
        CCW = 1
        SMALL_ARC_TYPE = 0
        LARGE_ARC_TYPE = 1
        arcSizeType, rotationArcDirection = segment.getDirection((SMALL_ARC_TYPE, CW), (SMALL_ARC_TYPE, CCW), (LARGE_ARC_TYPE, CW), (LARGE_ARC_TYPE, CCW))
        position = segment.getEndVertex().position

        # WoodFlash 2 uses the operation code #2111 for small arcs and #2112 for large arcs.
        # WoodFlash 3 however always uses the operation code #2111 and instead uses parameter #8050 to distinct between small and large arcs.
        # WoodFlash 3 does not know the operation code #2112 anymore, so that unfortunately no single code can be shared between WoodFlash 2 and 3.
        operationCode = 2112 if self.VERSION <= 2 and arcSizeType == LARGE_ARC_TYPE else 2111
        return joinIfNotNone(' ',
            'W#{code}{{ ::WTa'.format(code=operationCode),
            '#1={x}'.format(x=self.formatAsCoordinate(data=position.x)),
            '#2={y}'.format(y=self.formatAsCoordinate(data=position.y)),
            '#34={direction}'.format(direction=rotationArcDirection),
            '#8017={r}'.format(r=self.formatAsAngle(data=segment.getRadius())),
            '#8050={arcsizetype}'.format(arcsizetype=arcSizeType) if self.VERSION > 2 else None,
            '}W',
        )

    def createLineSegment(self, **keywords):
        segment = keywords['segment']
        position = segment.getEndVertex().position
        return ' '.join((
            'W#2201{ ::WTl',
            '#8015=0',
            '#1={x}'.format(x=self.formatAsCoordinate(data=position.x)),
            '#2={y}'.format(y=self.formatAsCoordinate(data=position.y)),
            '}W',
        ))

    def createSizingMacro1037And1510SpecificBlock(self, **_keywords):
        return joinIfNotNone(' ',
            '#8505={nestingGeometry:d}'.format(nestingGeometry=self.CREATE_SIZING_AS_NESTING_GEOMETRY),
            '#8506=1',
            '#8507=0',  # execute 2 steps
            '#8508=0',
            '#8509=0.5',
        )

    def createSizingMacro1523SpecificBlock(self, **_keywords):
        return joinIfNotNone(' ',
            '#8512={excludeRearEdge:d}'.format(excludeRearEdge=self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR),  # exclude rear side for sizing with three sides
        )

    def createSizingMacroSpecificBlock(self, **keywords):
        if keywords['macro'] == '1523':
            return self.createSizingMacro1523SpecificBlock(**keywords)
        else:  # for macros 1037 and 1510
            return self.createSizingMacro1037And1510SpecificBlock(**keywords)

    def createCommonSizingWithMacro(self, **keywords):
        return joinIfNotNone(' ',
            'W#{macro}{{ ::WT2'.format(macro=keywords['macro']),
            '#8098=squad.tmcr',
            '#8500={approachDistance}'.format(approachDistance=self.formatAsCoordinate(data=keywords['securityDistance'])),
            '#8501={departureDistance}'.format(departureDistance=self.formatAsCoordinate(data=keywords['securityDistance'])),
            '#8502={toolID}'.format(toolID=keywords['toolID']),
            '#8503={z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            '#8504=0',
            self.createSizingMacroSpecificBlock(**keywords),
            '}W',
        )

    def createSizingWithMacro(self, **keywords):
        def getSizingWithMacroName():
            if self.CREATE_SIZING_WITH_MACRO_NAME == '1':  # version-depending
                return '1510' if self.VERSION >= 3 else '1037'
            else:
                return self.CREATE_SIZING_WITH_MACRO_NAME

        surface = keywords['surface']
        if surface.hasSizing():
            keywords['machining'] = surface.sizing
            keywords['macro'] = getSizingWithMacroName()
            keywords['securityDistance'] = self.SIZING_APPROACH_DEPARTURE_DISTANCE + surface.sizing.tool.getRouterDiameter() / 2
            keywords['toolID'] = surface.sizing.toolID
            keywords['zValue'] = self.getSizingZValue(**keywords)

            return joinIfNotNone('\n',
                self.changeSide(side=self.SIDE_MAIN),
                self.createCommonSizingWithMacro(**keywords),
                '',
            )

    def createSizingContour(self, **keywords):
        finishedSizeShape = self.getFinishedSizeShape(**keywords)
        if self.CREATE_SIZING_WITH_MACRO_NAME != '0' and finishedSizeShape.isRectangular():  # TODO Note the old values 0|1
            return self.createSizingWithMacro(**keywords)
        else:
            return self.createSizingContourDefinitionAsExecution(**keywords)

    def createVerticalContour(self, **keywords):
        keywords['creator'] = self.createDefinitionForOneContour
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)

        return '\n'.join(self.generateContourNoBuiltInDepthSteps(**keywords))

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createVerticalContour, **keywords))

    def createHorizontalContour(self, **keywords):
        keywords['creator'] = self.createDefinitionForOneContour
        keywords['zValueConverter'] = lambda _boardThickness, currentDepth: - currentDepth

        return '\n'.join(self.generateContourNoBuiltInDepthSteps(**keywords))

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createHorizontalContour, **keywords))

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']

        # TODO (HORIZONTAL) IT-222/IT-230: unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if not contour.isVertical():
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.wopMatrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1) or matrix.w == Vector(0, 0, -1), "Matrix must be 2D!"
            contour.path = contour.transformedPolyline(matrix.getInverted(), positionOnPlane)
            keywords['sideID'] = plane.sideID

        return super().createDefinitionForOneContour(**keywords)

    def createVerticalRabbet(self, **keywords):
        keywords['creator'] = self.createVerticalRabbetOffsetByContourDistance
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
        machiningDepth = keywords['rabbet'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return '\n'.join(self.generateCommonNoBuiltInDepthSteps(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    def createCommonDadoWithRouter(self, **keywords):
        return self.createContourFromSegmentList(
            createStartBlock=self.createDadoContourStartBlock,
            createEndBlock=self.createDadoContourEndBlock,
            radiusCompensation=self.getDadoRadiusCompensation(**keywords),
            segmentList=self.generateDadoSegments(**keywords),
            zValue=keywords['depth'],
            **keywords)

    def createVerticalDadoWithRouter(self, **keywords):
        def creatorHelper(**kwargs):
            kwargs['depth'] = kwargs['zValue']
            del kwargs['zValue']  # TODO IT-265: can we do it better? -- NG
            return self.createCommonDadoWithRouter(**kwargs)

        keywords['creator'] = creatorHelper
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
        machiningDepth = keywords['dado'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return '\n'.join(self.generateCommonNoBuiltInDepthSteps(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    def createHorizontalDadoWithRouter(self, **keywords):
        def creatorHelper(**kwargs):
            kwargs['depth'] = kwargs['zValue']
            del kwargs['zValue']  # TODO IT-265: can we do it better? -- NG
            return self.createCommonDadoWithRouter(**kwargs)

        dado = keywords['dado']

        plane, positionOnPlane = self.__calculatePlane(**keywords)

        matrix = plane.wopMatrix * dado.matrix.getInverted()
        assert matrix.w == Vector(0, 0, 1) or matrix.w == Vector(0, 0, -1), "Matrix must be 2D!"
        dado.matrix = matrix  # TODO (HORIZONTAL) IT-229: unschönes Design -- NG
        dado.matrix.origin = positionOnPlane

        keywords['creator'] = creatorHelper
        keywords['zValueConverter'] = lambda _boardThickness, currentDepth: -currentDepth
        machiningDepth = keywords['dado'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return '\n'.join(self.generateCommonNoBuiltInDepthSteps(sideID=plane.sideID, machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    def __createCommonDrillingDiameterOrToolID(self, **keywords):
        drilling = keywords['drilling']

        if self.isValidToolId(tool=drilling.tool, **keywords):
            return '#205={toolID}'.format(toolID=drilling.toolID)
        else:
            return '#1002={diameter}'.format(diameter=self.formatAsCoordinate(data=drilling.installedDiameter))

    def __createCommonDrillingCode(self, **keywords):
        # Aus WoodFlash Doku: "Programming workings in the piece is always relative to one face and referred to the three-dimensional Cartesian coordinates of the face."
        return ' '.join((
            'W#81{ ::WTp',
            self.__createCommonDrillingDiameterOrToolID(**keywords),
            '#1={x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            '#2={y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            '#3={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '#8015=0',
            '#1001={type}'.format(type=keywords['drillType']),
            '}W',
        ))

    def __createCommonDrilling(self, **keywords):
        return joinIfNotNone('\n',
            self.changeSide(side=keywords['sideID']),
            self.__createCommonDrillingCode(**keywords),
            '',
        )

    def __createCommonVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']
        return self.__createCommonDrilling(sideID=self.SIDE_MAIN, x=drilling.position.x, y=drilling.position.y, **keywords)

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            ' '.join((
                'W#1001{ ::WT2',
                '#8098=fittingx.tmcr',
                '#6=1',
                '#8508=0',
                '#8509=0',
                '#8510={x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
                '#8511={x}'.format(x=self.formatAsCoordinate(data=endDrilling.position.x)),
                '#8512={distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                '#8513={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
                '#8517=0',
                '#8518={y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
                '#8520=1',
                '#8521=1',
                '#8522={diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
                '#8525={type}'.format(type=self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)),
                '}W',
            )),
            '',
        )

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            ' '.join((
                'W#1003{ ::WT2',
                '#8098=fittingy.tmcr',
                '#6=1',
                '#8508=0',
                '#8509=0',
                '#8510={y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
                '#8511={y}'.format(y=self.formatAsCoordinate(data=endDrilling.position.y)),
                '#8512={distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                '#8513={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
                '#8517=0',
                '#8518={x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
                '#8520=1',
                '#8521=1',
                '#8522={diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
                '#8525={type}'.format(type=self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)),
                '}W',
            )),
            '',
        )

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        startDrilling = gridDrilling.getStartDrilling()
        endDrilling = gridDrilling.getEndDrilling()
        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            ' '.join((
                'W#1005{ ::WT2',
                '#8098=repeatxy.tmcr',
                '#6=1',
                '#8020={x}'.format(x=self.formatAsCoordinate(data=startDrilling.position.x)),
                '#8021={y}'.format(y=self.formatAsCoordinate(data=startDrilling.position.y)),
                '#8022={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
                '#8512={distance}'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
                '#8515={type}'.format(type=self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, drilling=gridDrilling, **keywords)),
                '#8516={diameter}'.format(diameter=self.formatAsCoordinate(data=gridDrilling.getInstalledDiameter())),
                '#8518={x}'.format(x=self.formatAsCoordinate(data=endDrilling.position.x)),
                '#8519={y}'.format(y=self.formatAsCoordinate(data=endDrilling.position.y)),
                '}W',
            )),
            '',
        )

    FLAT_DRILL_TYPE = 0
    PIERCING_DRILL_TYPE = 1
    MILLING_TYPE = 100

    def createVerticalDrilling(self, **keywords):
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)
        return self.__createCommonVerticalDrilling(drillType=drillType, **keywords)

    def createVerticalDrillingPiercing(self, **keywords):
        drillType = self.getDrillType(defaultDrillType=self.PIERCING_DRILL_TYPE, **keywords)
        return self.__createCommonVerticalDrilling(drillType=drillType, **keywords)

    def __createSimplifiedLegacyLineSegment(self, **keywords):
        return ' '.join((
            'W#2201{ ::WTl',
            '#1={x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            '#2={y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            '#3={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '#8015=0',
            '#9022=0',
            '#8054=0',
            '#8055=0',
            '#8056=0',
            '}W',
        ))

    def __createSimplifiedLegacyRadiusSegment(self, **keywords):
        CW = 0
        return ' '.join((
            'W#2111{ ::WTa',
            '#1={x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            '#2={y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            '#34={direction}'.format(direction=CW),
            '#8015=0',
            '}W',
        ))

    def getPocketMillingDiameter(self, **keywords):
        pocket = keywords['pocket']
        diameter = pocket.tool.getRouterDiameter()
        if not isNearlyEqual(diameter, 0):
            return diameter
        else:
            warning = self.getLocalizedMessage(51, "Suppressing pocket, as diameter of router is not defined! Workpiece: {workpiece}")
            logging.warning(warning.format(workpiece=keywords['surface'].workpieceName))

    def getRouterDepthPerStep(self, **keywords):
        depthPerStep = super().getRouterDepthPerStep(**keywords)

        # TODO IT-271 Do we want it in woodFlash? We used to have this code in NCExportBase for woodWOP, which does not allow the depth-per-step to be deeper than the depth of the machining
        if isNearlyGreater(depthPerStep, keywords['machiningDepth']):
            return 0.0

        return depthPerStep

    def __createSingleCommonPocket(self, **keywords):
        pocket = keywords['pocket']
        approachManaged, approachType, _dummy = self.APPROACH_MODE_ARC_FLYING
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=pocket.actualDepth, **keywords)

        return joinIfNotNone(' ',
            'W#1035{ ::WT2',
            '#8098=cerchio.tmcr',
            '#6=1',
            '#8020={x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            '#8021={y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            '#8140=1',
            '#8142=1',
            '#8149=1',
            '#8143=10',
            '#8148=1',
            '#8145={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '#8146={depthPerStep}'.format(depthPerStep=depthPerStep) if not isNearlyEqual(depthPerStep, 0.0) else None,
            '#8147=1',
            '#8131={setup}'.format(setup=keywords['toolSetupType']),
            '#8150=1',
            '#8151=1',
            '#8153={toolid}'.format(toolid=pocket.toolID),
            '#8154=100',
            '#8174={managed:d}'.format(managed=approachManaged),
            '#8175={approachType}'.format(approachType=approachType),
            '#8157=2',  # interpolation speed
            '#8158=2',  # feed speed above workpiece
            '#8509={r}'.format(r=self.formatAsCoordinate(data=pocket.diameter / 2)),
            '#8512={direction:d}'.format(direction=not self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
            '}W',
        )

    def __createSingleVerticalPocket(self, **keywords):
        pocket = keywords['pocket']
        return self.__createSingleCommonPocket(x=pocket.position.x, y=pocket.position.y, toolSetupType=91, **keywords)

    def __simulateSingleVerticalPocket(self, **keywords):
        # This is legacy code from the classical VectorWOP export (section KREISTASCHE), except z (#3) is not given for each and every macro
        pocket = keywords['pocket']
        pocketMillingDiameter = self.getPocketMillingDiameter(**keywords)
        if pocketMillingDiameter:
            lines = [self.__createCommonContourSetup(startVertex=pocket, zValue=keywords['depth'], radiusCompensation=WoodFlashExport.RADIUS_COMPENSATION_NONE, approachManaged=False, departureManaged=False, **keywords)]

            if pocketMillingDiameter * 2 > pocket.diameter:
                lines.append(self.__createSimplifiedLegacyLineSegment(x=pocket.position.x - pocket.diameter / 2 + pocketMillingDiameter / 1.9, y=pocket.position.y, **keywords))
            else:
                lines.append(self.__createSimplifiedLegacyLineSegment(x=pocket.position.x - pocketMillingDiameter / 2, y=pocket.position.y, **keywords))
                for i in range(1, int(pocket.diameter // pocketMillingDiameter)):
                    lines.append(self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x + i * pocketMillingDiameter / 2, y=pocket.position.y, **keywords))
                    lines.append(self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x - i * pocketMillingDiameter / 2, y=pocket.position.y, **keywords))
            if pocket.diameter / 2 > (pocket.diameter // pocketMillingDiameter) * pocketMillingDiameter / 2:
                lines.append(self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x + pocket.diameter / 2 - pocketMillingDiameter / 2, y=pocket.position.y, **keywords))
                lines.append(self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x - pocket.diameter / 2 + pocketMillingDiameter / 2, y=pocket.position.y, **keywords))
                lines.append(self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x + pocket.diameter / 2 - pocketMillingDiameter / 2, y=pocket.position.y, **keywords))
            else:
                lines.append(self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x + pocket.diameter / 2 - pocketMillingDiameter / 2, y=pocket.position.y, **keywords))

            return '\n'.join(lines)

    def createVerticalPocket(self, **keywords):
        createFunction = self.__createSingleVerticalPocket if self.VERSION >= 3 or self.USE_VERTICAL_POCKET_MACRO_IN_WDFLASH2 else self.__simulateSingleVerticalPocket
        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            createFunction(approachDelta=Point(0, 0), **keywords),
            '',
        )

    def __simulateSingleVerticalPocketPiercing(self, **keywords):
        # This is legacy code from the classical VectorWOP export (section KREISTASCHEDURCHGEHEND)
        pocket = keywords['pocket']
        pocketMillingDiameter = self.getPocketMillingDiameter(**keywords)
        if pocketMillingDiameter:
            return '\n'.join((
                self.__createCommonContourSetup(startVertex=pocket, zValue=1, radiusCompensation=WoodFlashExport.RADIUS_COMPENSATION_NONE, approachManaged=False, departureManaged=False, **keywords),
                self.__createSimplifiedLegacyLineSegment(x=pocket.position.x - pocket.diameter / 2 + pocketMillingDiameter / 2, y=pocket.position.y, **keywords),
                self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x + pocket.diameter / 2 - pocketMillingDiameter / 2, y=pocket.position.y, **keywords),
                self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x - pocket.diameter / 2 + pocketMillingDiameter / 2, y=pocket.position.y, **keywords),
                self.__createSimplifiedLegacyRadiusSegment(x=pocket.position.x + pocket.diameter / 2 - pocketMillingDiameter / 2, y=pocket.position.y, **keywords),
            ))

    def __createSingleVerticalPocketPiercing(self, **keywords):
        pocket = keywords['pocket']

        return ' '.join((
            'W#1028{ ::WT2',
            '#8098=circle1.tmcr',
            '#8020={x}'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            '#8021={y}'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            '#8505={rc}'.format(rc=self.RADIUS_COMPENSATION_RIGHT),
            '#8509={r}'.format(r=self.formatAsCoordinate(data=pocket.diameter / 2)),
            '#8512={direction:d}'.format(direction=not self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
            '#8514={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '#8522={toolid}'.format(toolid=pocket.toolID),
            '#8524=1',
            '#8525=1',
            '#8526={r}'.format(r=self.formatAsCoordinate(data=self.getPocketEntranceRadius(**keywords))),
            '}W',
        ))

    def createVerticalPocketPiercing(self, **keywords):
        def createFunction3(**kwargs):
            # TODO IT-265: can it be implemented less stupidly -- NG
            kwargs['depth'] = kwargs['zValue']
            return self.__createSingleVerticalPocketPiercing(**kwargs)

        if keywords['pocket'].diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER:
            if self.VERSION >= 3 or self.USE_VERTICAL_POCKET_MACRO_IN_WDFLASH2:
                machiningDepth = keywords['pocket'].actualDepth
                depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)
                return joinIfNotNone('\n',
                    self.changeSide(side=self.SIDE_MAIN),
                    '\n'.join(self.generateCommonNoBuiltInDepthSteps(
                        creator=createFunction3,
                        machiningDepth=machiningDepth,
                        depthPerStep=depthPerStep,
                        zValueConverter=lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth),
                        approachDelta=Point(0, 0),
                        **keywords)
                    ),
                    '',
                )
            else:
                return joinIfNotNone('\n',
                    self.changeSide(side=self.SIDE_MAIN),
                    self.__simulateSingleVerticalPocketPiercing(approachDelta=Point(0, 0), **keywords),
                    '',
                )
        else:
            return self.createVerticalPocket(**keywords)

    def createHorizontalPocket(self, **keywords):
        pocket = keywords['pocket']

        plane, positionOnPlane = self.__calculatePlane(**keywords)

        # TODO IT-222 (HORIZONTAL): WoodFlash2 -- NG

        return joinIfNotNone('\n',
            self.changeSide(side=plane.sideID),
            self.__createSingleCommonPocket(x=positionOnPlane.x, y=positionOnPlane.y, toolSetupType=89, depth=-pocket.depth, **keywords),
            '',
        )

    def createHorizontalDrillingOnBounds(self, **keywords):
        depth = - keywords['drilling'].depth
        drillType = self.getDrillType(defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)
        return self.__createCommonDrilling(depth=depth, drillType=drillType, **keywords)

    def createRotatedHorizontalDrilling(self, **keywords):
        drilling = keywords['drilling']
        if drilling.direction == Vector(0, 0, 1):
            sideID = self.SIDE_MAIN
            x = drilling.position.x
            y = drilling.position.y
            predefined = True
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            sideID = plane.sideID
            predefined = plane.predefined
            x = positionOnPlane.x
            y = positionOnPlane.y

        if predefined:
            return self.createHorizontalDrillingOnBounds(sideID=sideID, x=x, y=y, **keywords)
        else:
            return self.createFiveAxisDrilling(**keywords)

    def createFiveAxisDrillingCode(self, **keywords):
        drilling = keywords['drilling']
        diameter = self.formatAsCoordinate(data=drilling.diameter)
        comment = 'Ø{d} x {depth}'.format(d=diameter, depth=self.formatAsCoordinate(data=abs(keywords['depth'])))
        return ' '.join((
            'W#91{ ::WTp',
            self.createCommentWOBlanksLowerCase(comment=comment, **keywords),
            '#1002={d}'.format(d=diameter),
            '#1={x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
            '#2={y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
            '#3={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            '#4={c}'.format(c=self.formatAsAngle(data=drilling.cAngle)),
            '#5=-{a}'.format(a=self.formatAsAngle(data=drilling.aAngle)),
            '#8015=0',
            '#205={toolid}'.format(toolid=drilling.toolID),
            '}W',
        ))

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']

        machiningPlane = Plane.createFromMachining(machining)

        plane = self.machiningPlanes.findPlane(machiningPlane)
        assert plane is not None, "No plane found for machining"

        return plane, machining.getPositionOnPlane(plane)

    def createFiveAxisDrilling(self, **keywords):
        drilling = keywords['drilling']
        if drilling.direction == Vector(0, 0, 1):
            sideID = self.SIDE_MAIN
            positionOnPlane = drilling.position
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            sideID = plane.sideID

        return joinIfNotNone('\n',
            self.changeSide(side=sideID),
            self.createFiveAxisDrillingCode(x=positionOnPlane.x, y=positionOnPlane.y, depth=-drilling.depth, **keywords),
            '',
        )

    # dado with saw
    SAW_ONE_STEP, SAW_TWO_STEPS = 0, 1
    FIXED_SAW, TILTABLE_SAW = 1, 2

    def usePreSaw(self, depth):
        return not isNearlyEqual(depth, 0)

    def __createCommonVerticalSawCutProperties(self, **keywords):
        dado = keywords['dado']
        surface = keywords['surface']
        start = keywords['segment'].getStartAndEndPoints()[0]
        depth = keywords['depth']
        preSawDepth = self.zValueDadoSawConverter.convertWithThickness(surface.thickness, self.DADO_PRESAW_DEPTH)
        usePreSaw = self.usePreSaw(depth=self.DADO_PRESAW_DEPTH)

        return joinIfNotNone(' ',
            '#8098=lame.tmcr',
            '#8503={width}'.format(width=self.formatAsCoordinate(data=keywords['width'])) if 'width' in keywords else None,
            '#8509={sawType}'.format(sawType=keywords['sawType']),
            '#{paramX}={x}'.format(paramX='8020' if self.MACHINE_TYPE == 'cexpress' else '8510', x=self.formatAsCoordinate(data=start.x)),
            '#{paramY}={y}'.format(paramY='8021' if self.MACHINE_TYPE == 'cexpress' else '8511', y=self.formatAsCoordinate(data=start.y)),
            '#{paramZ}={z}'.format(paramZ='8022' if self.MACHINE_TYPE == 'cexpress' else '8512', z=self.formatAsCoordinate(data=preSawDepth if usePreSaw else depth)),
            '#8513={z}'.format(z=self.formatAsCoordinate(data=depth)) if usePreSaw else None,
            '#8514=1',
            '#8515=1',
            '#8516={toolID}'.format(toolID=dado.toolID),
        )

    def __createCommonVerticalSawCutPropertiesAdditional(self, **keywords):
        dado = keywords['dado']
        usePreSaw = self.usePreSaw(depth=self.DADO_PRESAW_DEPTH)

        return joinIfNotNone(' ',
            '#8525={radiusCompensation}'.format(radiusCompensation=keywords['radiusCompensation']),
            '#8526={steps}'.format(steps=self.SAW_TWO_STEPS if usePreSaw else self.SAW_ONE_STEP),
            '#8527={mode:d}'.format(mode=dado.isInset() or self.USE_SECURITY_DISTANCE_FOR_CONTINUOUS_DADO_SAW),  # must always be true for Nesting  --mjk
        )

    def __createSingleVerticalSawCutAlongXAxis(self, **keywords):
        end = keywords['segment'].getStartAndEndPoints()[1]
        return joinIfNotNone(' ',
            'W#1050{ ::WT2',
            self.__createCommonVerticalSawCutProperties(sawType=self.FIXED_SAW, **keywords),
            '#8517={xe}'.format(xe=self.formatAsCoordinate(data=end.x)),
            self.__createCommonVerticalSawCutPropertiesAdditional(**keywords),
            '}W',
        )

    def __createSingleVerticalSawCutAlongYAxis(self, **keywords):
        end = keywords['segment'].getStartAndEndPoints()[1]
        return joinIfNotNone(' ',
            'W#1051{ ::WT2',
            self.__createCommonVerticalSawCutProperties(sawType=self.FIXED_SAW, **keywords),
            '#8518={ye}'.format(ye=self.formatAsCoordinate(data=end.y)),
            self.__createCommonVerticalSawCutPropertiesAdditional(**keywords),
            '}W',
        )

    def __createXYEndForSingleVerticalSawCutCRotated(self, **keywords):
        end = keywords['segment'].getStartAndEndPoints()[1]
        return joinIfNotNone(' ',
            '#8517={xe}'.format(xe=self.formatAsCoordinate(data=end.x)),
            '#8518={ye}'.format(ye=self.formatAsCoordinate(data=end.y)),
        )

    def __createCAngleAndLengthForSingleVerticalSawCutCRotated(self, **keywords):
        dado = keywords['dado']
        segment = keywords['segment']
        return joinIfNotNone(' ',
            '#8519={cangle}'.format(cangle=self.formatAsAngle(data=dado.getXYAngle())),
            '#8520={length}'.format(length=self.formatAsCoordinate(data=segment.length)),
        )

    def __createSingleVerticalSawCutCRotated(self, **keywords):
        return joinIfNotNone(' ',
            'W#1052{ ::WT2',
            self.__createCommonVerticalSawCutProperties(sawType=self.TILTABLE_SAW, **keywords),
            self.__createXYEndForSingleVerticalSawCutCRotated(**keywords) if self.VERSION >= 3 else None,
            self.__createCAngleAndLengthForSingleVerticalSawCutCRotated(**keywords) if self.VERSION < 3 else None,
            '#8521={angle}'.format(angle=self.formatAsAngle(data=90)),
            self.__createCommonVerticalSawCutPropertiesAdditional(**keywords),
            '#8533=1',
            '}W',
        )

    def getDadoLineSegmentWithSecurityDistance(self, **keywords):
        dado = keywords['dado']
        segment = dado.getLineSegment()

        if not dado.isInsetStart() and dado.isInsetEnd():
            segment.increaseAtStart(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
        elif not dado.isInsetEnd() and dado.isInsetStart():
            segment.increaseAtEnd(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
        elif self.USE_SECURITY_DISTANCE_FOR_CONTINUOUS_DADO_SAW and not dado.isInsetStart() and not dado.isInsetEnd():
            segment.increaseAtStart(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
            segment.increaseAtEnd(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
        return segment

    def __createVerticalDadoWithSawAlongAxis(self, **keywords):
        multiSawDado = MultiSawDado(self)
        createFunction = multiSawDado.createSingleVerticalSawCut if self.DADO_WIDTH_SUPPORTED else multiSawDado.createMultipleVerticalSawCuts

        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            createFunction(**keywords),
            '',
        )

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        return self.__createVerticalDadoWithSawAlongAxis(createFunction=self.__createSingleVerticalSawCutAlongXAxis, **keywords)

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        return self.__createVerticalDadoWithSawAlongAxis(createFunction=self.__createSingleVerticalSawCutAlongYAxis, **keywords)

    def createVerticalDadoWithSawCRotated(self, **keywords):
        return self.__createVerticalDadoWithSawAlongAxis(createFunction=self.__createSingleVerticalSawCutCRotated, **keywords)

    # miter
    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCW)

    @staticmethod
    def getSawCutDepth(**keywords):
        return math.hypot(keywords['width'], keywords['boardThickness'])

    def __createMiterSawCut(self, **keywords):
        miter = keywords['miter']
        start = keywords['startPoint']
        depth = keywords['depth']
        preSawDepth = keywords['preSawDepth']
        usePreSaw = self.usePreSaw(depth=preSawDepth)

        return joinIfNotNone(' ',
            'W#1052{ ::WT2',
            '#8098=lame.tmcr',
            '#8509={sawType}'.format(sawType=self.TILTABLE_SAW),
            '#8510={x}'.format(x=self.formatAsCoordinate(data=start.x)),
            '#8511={y}'.format(y=self.formatAsCoordinate(data=start.y)),
            '#8512={z}'.format(z=self.formatAsCoordinate(data=preSawDepth if usePreSaw else depth)),
            '#8515=1',
            '#8516={toolID}'.format(toolID=miter.toolID),
            '#8519={xyAngle}'.format(xyAngle=self.formatAsAngle(data=miter.getXYAngle())),
            '#8520={length}'.format(length=self.formatAsCoordinate(data=miter.length)),
            '#8521={miterAngle}'.format(miterAngle=self.formatAsAngle(data=miter.getAngle())),
            '#8525={radiusCompensation}'.format(radiusCompensation=self.RADIUS_COMPENSATION_LEFT),
            '#8526={steps}'.format(steps=self.SAW_TWO_STEPS if usePreSaw else self.SAW_ONE_STEP),
            '#8513={z}'.format(z=self.formatAsCoordinate(data=depth)) if usePreSaw else None,
            '}W',
        )

    def createVerticalMiter(self, **keywords):
        miter = keywords['miter']
        surface = keywords['surface']
        boardThickness = surface.thickness
        segment = miter.getLineSegment()
        segment.moveRight(offset=miter.width)
        start = segment.getStartPoint()

        preSawDepth = self.zValueMiterSawConverter.convertWithThickness(boardThickness, self.MITER_PRE_SAW_DEPTH_FROM_TOP)
        rawSawCutDepth = self.getSawCutDepth(width=miter.width, boardThickness=boardThickness, **keywords) + self.ADDITIONAL_DEPTH_FOR_MITER
        depth = self.zValueMiterSawConverter.convertWithThickness(boardThickness, rawSawCutDepth)

        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            self.__createMiterSawCut(startPoint=start, depth=depth, preSawDepth=preSawDepth, **keywords),
            '',
        )

    def createNCStop(self, **keywords):
        return joinIfNotNone('\n',
            self.changeSide(side=self.SIDE_MAIN),
            joinIfNotNone(' ',
                'W#102{',
                '::WTg',
                'WS=2',
                self.createCommentWOBlanksLowerCase(**keywords),
                '#1=0',
                '#2=0',
                '#201=1',
                '#205=1',
                '#9505=99',
                '}W',
            ),
            '',
        )

    # NC Macro
    def getNCMacroCookedVariables(self, **keywords):
        variables = super().getNCMacroCookedVariables(**keywords)

        ncmacro = keywords['ncmacro']
        if ncmacro.code.startswith('SIDE#0'):
            sideID = self.SIDE_ABSOLUTE
            sideX = ncmacro.position.x
            sideY = ncmacro.position.y
            sideCAngle = variables['cangle']
            ncmacro.code = ''.join(keywords['ncmacro'].code.splitlines(keepends=True)[1:])  # Remove first line of ncmacro.code to use changeSide() correctly.
        elif ncmacro.isVertical():
            sideID = self.SIDE_MAIN
            sideX = ncmacro.position.x
            sideY = ncmacro.position.y
            sideCAngle = variables['cangle']
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            sideID = plane.sideID
            sideX = positionOnPlane.x
            sideY = positionOnPlane.y

            angle = VectorMath.angleBetween(ncmacro.matrix.u, plane.wopMatrix.u)

            if not plane.isWopMatrixRightHanded():
                angle = 360.0 - angle

            if not isNearlyEqual(angle, 0.0):
                crossProduct = ncmacro.matrix.u.cross(plane.wopMatrix.u)
                if not isNearlyEqual(crossProduct.length(), 0.0) and crossProduct.normalized() == plane.wopMatrix.w:
                    angle = - angle

            sideCAngle = self.formatter.formatAsAngle(normalizeAngle(angle))

        variables['sideid'] = sideID
        variables['sidex'] = self.formatter.formatAsCoordinate(sideX)
        variables['sidey'] = self.formatter.formatAsCoordinate(sideY)
        variables['sidecangle'] = sideCAngle

        return variables

    def createNCMacroCode(self, **keywords):
        sideID = self.getPredefinedSideNumberForNCMacro(**keywords)
        return joinIfNotNone('\n',
            self.changeSide(side=sideID),
            super().createNCMacroCode(**keywords)
        )

    def prefixWithHash(self, **keywords):
        data = keywords['data']
        if not data.startswith('#'):
            data = '#{data}'.format(data=data)
        return data

    def getPredefinedSideNumberForNCMacro(self, **keywords):
        ncmacro = keywords['ncmacro']

        if ncmacro.code.startswith('SIDE#0'):
            return self.SIDE_ABSOLUTE
        elif ncmacro.isVertical():
            return self.SIDE_MAIN
        else:
            plane, _positionOnPlane = self.__calculatePlane(**keywords)
            return plane.sideID

    def createNCMacroRecordFormat(self, **keywords):
        parameters = tuple(self.getNCMacroCookedParameters(**keywords))

        if parameters and parameters[0].name == "SIDE":
            side = int(parameters[0].value)
            parameters = parameters[1:]
        else:
            side = self.getPredefinedSideNumberForNCMacro(**keywords)

        if parameters:
            macroNumber = parameters[0].name
            parameterType = parameters[0].value

            parameters = parameters[1:]

            return joinIfNotNone('\n',
                self.changeSide(side=side),
                ' '.join(itertools.chain(
                    (
                        'W{macro}{{ ::{type}'.format(macro=self.prefixWithHash(data=macroNumber), type=parameterType),
                    ),
                    ('{name}={value}'.format(name=self.prefixWithHash(data=parameter.name), value=parameter.value) for parameter in parameters),
                    (
                        '}W',
                    ),
                )),
                '',
            )
        else:
            warning = self.getLocalizedMessage(40, "Export of NC macro suppressed, as the NC macro had no parameters. Workpiece: {workpiece}, macro name: {name}")
            logging.warning(warning.format(workpiece=keywords['surface'].workpieceName, name=keywords['ncmacro'].name))
            return ''

    # Footer
    def createEndOfFile(self, **_keywords):
        return '\n'

    # Localization
    def getLocalizedMachiningComment(self, index, default):
        return getLocalizedString(WOODFLASH_COMMENT_MESSAGES, index, default)
