# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import functools
import itertools
import logging
import math

import interiorcad.backend.localization
import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.localization import getLocalizedString, FILE_NAME_POSTFIXES_RESOURCEGROUP, WOODWOP_MESSAGES
from interiorcad.backend.ncexport.ncsegmentsplitter import splitSectionsInXDistance
from interiorcad.backend.umlautscleaner import UmlautsCleaner
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual, isNearlyGreater, isNearlyLesser
from interiorcad.interface.geometry.matrix import Matrix
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.geometry.vectormath import VectorMath
from interiorcad.interface.machineconfiguration.tool import Tool
from interiorcad.interface.ncexport.oneprogramexporter import OneProgramExporter
from interiorcad.interface.surface import NCMacro2D
from interiorcad.interface.surface.acangles import calculateCAngleFromWVector, normalizeAngle
from interiorcad.interface.surface.contour2d import Contour2D, HorizontalContour2D
from interiorcad.interface.surface.dado2d import HorizontalDado2D
from interiorcad.interface.surface.machiningplanes import MachiningPlanes
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.ncexport.base import NCExportBase


# noinspection PyUnusedLocal
class WoodWOPExport(NCExportBase):
    EXPORT_FORMAT = 'woodWOP'

    DEFAULT_COORDINATE_SYSTEM = '00'

    RADIUS_COMPENSATION_NONE = 'NOWRK'
    RADIUS_COMPENSATION_LEFT = 'WRKL'
    RADIUS_COMPENSATION_RIGHT = 'WRKR'

    APPROACH_TYPE_NONE = 0
    APPROACH_TYPE_FLYING = 1

    QUARTER_ARC_APPROACH = 'SEI'
    QUARTER_ARC_DEPARTURE = 'SEI_AB'

    TANGENTIAL_APPROACH = 'TAN'
    TANGENTIAL_DEPARTURE = 'TAN_AB'

    VERTICAL_APPROACH = 'SEN'
    VERTICAL_DEPARTURE = 'SEN_AB'

    APPROACH_MODE_LINEAR_PIERCING = (TANGENTIAL_APPROACH, TANGENTIAL_DEPARTURE, APPROACH_TYPE_NONE)
    APPROACH_MODE_LINEAR_FLYING = (TANGENTIAL_APPROACH, TANGENTIAL_DEPARTURE, APPROACH_TYPE_FLYING)
    APPROACH_MODE_ARC_PIERCING = (QUARTER_ARC_APPROACH, QUARTER_ARC_DEPARTURE, APPROACH_TYPE_NONE)
    APPROACH_MODE_ARC_FLYING = (QUARTER_ARC_APPROACH, QUARTER_ARC_DEPARTURE, APPROACH_TYPE_FLYING)
    APPROACH_MODE_WITHOUT = (VERTICAL_APPROACH, VERTICAL_DEPARTURE, APPROACH_TYPE_NONE)
    APPROACH_MODE_WITHOUT_FLYING = (VERTICAL_APPROACH, VERTICAL_DEPARTURE, APPROACH_TYPE_FLYING)  # Used by Nesting only for closed contours, e.g. sizing

    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_LINEAR_FLYING

    def __init__(self, **keywords):
        super().__init__(**keywords)

        self.machiningPlanes = None

        # woodWOP has a convention for length, width, thickness (the German ones are not the same as the English one), but they could also be individual
        machineConfiguration = self.getMachineConfiguration()
        self.INDIVIDUAL_PART_DIMENSION_VARIABLES = machineConfiguration.get('individual-woodwop-variables', '')
        self.prepareLocalizationOfPartDimensionVariables(**keywords)
        self.VAR_LENGTH = self.localizedPartDimensionVariableNames['LENGTH']
        self.VAR_WIDTH = self.localizedPartDimensionVariableNames['WIDTH']
        self.VAR_THICKNESS = self.localizedPartDimensionVariableNames['THICKNESS']

        self.ZERO_OFFSET_Y_IS_SIZING_ADDITION = machineConfiguration.getAsBool('zero-offset-y-is-sizing-addition', True)
        self.CREATE_SIZING_AS_NESTING_GEOMETRY = machineConfiguration.getAsBool('create-sizing-as-nesting-geometry', False)
        self.CREATE_SIZING_WITH_MACRO_NAME = machineConfiguration.get('create-sizing-with-macro-name', '')  # instead CREATE_SIZING_WITH_MACRO
        self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR = machineConfiguration.getAsBool('create-sizing-with-three-side-contour', False)
        self.SIZING_MILL_THROUGH_FOR_NON_RECTANGULAR_PARTS = machineConfiguration.getAsBool('sizing-mill-through-for-non-rectangular-parts', True)

        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 2)
        self.MITER_SAW_DEPTH_FROM_BOTTOM = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_SAW_DEPTH_FROM_BOTTOM_KEY, 10)

        self.OPTIMIZING_MODE = machineConfiguration.get('optimizing-mode', '1')
        self.AVOID_CROSS_DRIVE_HEIGHT_BY_HORIZONTAL_PROCESSING = machineConfiguration.getAsBool('avoid-cross-drive-height-by-horizontal-processing', False)
        self.BLOCKWISE_OPTIMIZING_OF_HORIZONTAL_DRILLINGS = machineConfiguration.getAsBool('blockwise-optimizing-of-horizontal-drillings', False)
        self.AVOID_CROSS_DRIVE_HEIGHT_BY_FINISH_PROCESSING = machineConfiguration.getAsBool('avoid-cross-drive-height-by-finish-processing', False)
        self.CLEARANCE_MODE = machineConfiguration.getAsInt('clearance-mode', '1')
        self.CLEARANCE_VALUE = machineConfiguration.getAsInt('clearance-value', '1000')
        self.DELETE_ELEMENTS_THAT_CANNOT_BE_PROCESSED = machineConfiguration.getAsBool('delete-elements-that-cannot-be-processed', True)
        self.GENERATED_AS_IN_X_MIRRORED_PROGRAM = machineConfiguration.getAsBool('generated-as-in-x-mirrored-program', False)
        self.GENERATED_AS_IN_Y_MIRRORED_PROGRAM = machineConfiguration.getAsBool('generated-as-in-y-mirrored-program', False)
        self.GENERATED_AS_IN_X_AND_Y_MIRRORED_PROGRAM = machineConfiguration.getAsBool('generated-as-in-x-and-y-mirrored-program', False)
        self.SUBSTITUTE_UMLAUTS = machineConfiguration.getAsBool('substitute-umlauts', False)
        self.VACUUM_SUCTION_CUP_PROPOSAL = machineConfiguration.getAsBool('vacuum-suction-cup-proposal', False)
        self.MACHINE_TYPE = machineConfiguration.get('machine-type', '')
        self.USE_CALCULATED_VALUES_OF_WORKPIECE_DESCRIPTION = machineConfiguration.getAsBool('use-calculated-values-of-workpiece-description', True)

        self.ONE_PROGRAM = machineConfiguration.getAsBool('one-program', False)
        self.FILE_NAME_POSTFIX_BOTH_SIDED = machineConfiguration.get('file-name-postfix-both-sided', getLocalizedString(FILE_NAME_POSTFIXES_RESOURCEGROUP, 4, 'BothSided'))

        self.CONTOUR_APPROACH_DEPARTURE_DISTANCE = machineConfiguration.getAsFloat('contour-approach-departure-distance', 30)
        self.CONTOUR_APPROACH_DEPARTURE_INTEGRATION = machineConfiguration.get('contour-approach-departure-integration', 'ever')  # ever | never | only open
        self.CONTOUR_NEEDS_ADDITIONAL_STEP = machineConfiguration.getAsBool('contour-needs-additional-step', True)
        self.MAXIMUM_SEGMENT_X_DISTANCE = machineConfiguration.getAsFloat('maximum-segment-x-distance', 0)

        self.SIZING_APPROACH_DISTANCE = machineConfiguration.getAsInt('sizing-approach-distance', 0)
        self.SIZING_PASS_STRATEGY = machineConfiguration.get('sizing-pass-strategy', 'standard')
        self.SIZING_PASS_STRATEGY_AREA_THRESHOLD = machineConfiguration.getAsFloat('sizing-pass-strategy-area-threshold', 0.1)
        self.SIZING_FIRST_PASS_DISTANCE = machineConfiguration.getAsFloat('sizing-first-pass-distance', 0.0)
        self.SIZING_FIRST_PASS_REMAINING_THICKNESS = machineConfiguration.getAsFloat('sizing-first-pass-remaining-thickness', 2)

        self.DRILL_MODE = machineConfiguration.get('drill-mode', 'slow')  # Options: slow & fast
        self.CREATE_GRID_DRILLING_WITH_ANGLE_INSTEAD_XY_DIRECTION = machineConfiguration.getAsBool('create-grid-drilling-with-angle-instead-xy-direction', False)
        self.CREATE_ROTATED_HORIZONTAL_DRILLING_AS_HORIZONTAL_DRILLING = machineConfiguration.getAsBool('create-rotated-horizontal-drilling-as-horizontal-drilling', True)

        self.createZValueConverters(**keywords)

        self.DADO_MODE = machineConfiguration.get('dado-mode', 'GL')  # Options: GL & GGL
        self.DADO_PRESAW_DEPTH = machineConfiguration.getAsFloat('dado-presaw-depth', 0)
        self.DADO_PERCENTS_OF_SAW_WIDTH = machineConfiguration.getAsInt('dado-infeed-factor', 85)
        ### IT-110 - IT-105: Die letzte Konstante soll eine Konstante von NCExportBase sein.   -- NG

        self.CLEAR_OUT_POCKET_OVERLAP_FACTOR = machineConfiguration.getAsInt('clear-out-pocket-overlap-factor', 80)
        self.USE_CUSTOM_USER_VARIABLES = machineConfiguration.getAsBool('use-custom-user-variables', False)
        self.USE_CUSTOM_USER_WORKPIECE_COMMENT = machineConfiguration.getAsBool('use-custom-user-workpiece-comment', False)

        self.INDIVIDUAL_MACHININGS_COORDINATE_SYSTEM = machineConfiguration.get('individual-machinings-coordinate-system', '')
        self.MACHININGS_COORDINATE_SYSTEM = self.getMachiningsCoordinateSystem(**keywords)

        self.checkConfiguration(**keywords)

    def getMachiningsCoordinateSystem(self, **keywords):
        if self.INDIVIDUAL_MACHININGS_COORDINATE_SYSTEM:
            return self.INDIVIDUAL_MACHININGS_COORDINATE_SYSTEM
        else:
            return self.DEFAULT_COORDINATE_SYSTEM

    def checkConfiguration(self, **keywords):
        if self.CREATE_SIZING_WITH_MACRO_NAME and self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR:
            logging.warning(self.getLocalizedMessage(73, "Simultaneous sizing with three-sided contour and with macro is not possible."))
            self.CREATE_SIZING_WITH_MACRO_NAME = ''

    def createZValueConverters(self, **keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_BOTTOM, ZValueConstants.Z_DIRECTION_UP)

    def getIndividualPartDimensionVariables(self):
        individualPartDimensionVariables = self.INDIVIDUAL_PART_DIMENSION_VARIABLES.split()
        self.localizedPartDimensionVariableNames = {
            'LENGTH': individualPartDimensionVariables[0],
            'WIDTH': individualPartDimensionVariables[1],
            'THICKNESS': individualPartDimensionVariables[2],
        }

    def getLocalizedPartDimensionVariables(self):
        LOCALIZED_PART_DIMENSION_VARIABLES = (
            ('LENGTH', 1, 'l'),
            ('WIDTH', 2, 'w'),
            ('THICKNESS', 3, 't'),
        )

        self.localizedPartDimensionVariableNames = {}
        for name, index, default in LOCALIZED_PART_DIMENSION_VARIABLES:
            self.localizedPartDimensionVariableNames[name] = self.getLocalizedVariable(index, default)

    def prepareLocalizationOfPartDimensionVariables(self, **keywords):
        if self.INDIVIDUAL_PART_DIMENSION_VARIABLES:
            self.getIndividualPartDimensionVariables()
        else:
            self.getLocalizedPartDimensionVariables()

    def substituteUmlautsIfNeeded(self, **keywords):
        return UmlautsCleaner().clean(keywords['data']) if self.SUBSTITUTE_UMLAUTS else keywords['data']

    def getSurfacesExporter(self, **keywords):
        if self.ONE_PROGRAM:
            return OneProgramExporter(**keywords)
        else:
            return super().getSurfacesExporter(**keywords)

    def createMainAndOppositeSidesTogether(self, mainSurface, oppositeSurface, **keywords):
        self.orientMachinings(surface=mainSurface, **keywords)
        yield self.createHead(surface=mainSurface, **keywords)
        yield self.createMachinings(surface=mainSurface, **keywords)

        self.orientMachinings(surface=oppositeSurface, **keywords)
        yield self.createTurnOver(surface=oppositeSurface, **keywords)
        yield self.createMachinings(surface=oppositeSurface, **keywords)
        yield self.createFoot(surface=oppositeSurface, **keywords)

    def createTurnOver(self, **keywords):
        surface = keywords['surface']
        parkPositionXString = '{length}+{clearance}'.format(length=self.VAR_LENGTH, clearance=self.CLEARANCE_VALUE)
        comment = self.getLocalizedKMComment(16, 'Please turn over part!')

        additionY = surface.getAdditionY() if self.ZERO_OFFSET_Y_IS_SIZING_ADDITION else 0.0

        return joinIfNotNone('\n',
            self.createNCStop(comment=comment, releaseVacuum=True, additionX=surface.getAdditionXMirrored(), additionY=additionY, parkPositionXString=parkPositionXString, **keywords),
            self.createSizingContourDefinition(**keywords),
            self.createRabbetContourDefinitions(**keywords),
            self.createDadoContourDefinitions(**keywords),
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords),
        )

    def createSizingContour(self, **keywords):
        return self.createSizingContourExecution(**keywords)

    def getNCFileExtension(self, **keywords):
        return '.mpr'

    def getWorkpieceComments(self, **keywords):
        workpieceComments = {}
        for key, value in super().getWorkpieceComments(**keywords).items():
            if isinstance(value, str) and self.SUBSTITUTE_UMLAUTS:
                workpieceComments[key] = self.substituteUmlautsIfNeeded(data=value)
            else:
                workpieceComments[key] = value

        return workpieceComments

    # Header
    def createHead(self, **keywords):
        surface = keywords['surface']
        self.lastContourDefinitionIndex = 0
        return joinIfNotNone('',
            self.createHeader(**keywords),
            self.createNCWorkpiece(**keywords),
            self.createCoordinateSystem(**keywords),
            self.createWorkpieceDefinition(**keywords),
            self.createWorkpieceComment(**keywords),
            self.createSizingContourDefinition(**keywords),
            self.createRabbetContourDefinitions(**keywords),
            self.createDadoContourDefinitions(dados=surface.getAllDados(), **keywords),
            self.createContourDefinitions(contours=surface.getAllContours(), **keywords),
        )

    def calculatedValuesOfWorkpieceDescription(self, **keywords):
        surface = keywords['surface']
        if self.USE_CALCULATED_VALUES_OF_WORKPIECE_DESCRIPTION:
            return joinIfNotNone('\n',
                '_BSX={}'.format(self.formatAsCoordinate(data=surface.getFinishedSizeX())),
                '_BSY={}'.format(self.formatAsCoordinate(data=surface.getFinishedSizeY())),
                '_BSZ={}'.format(self.formatAsCoordinate(data=surface.thickness)),
                '',
                '',
            )

    def createHeader(self, **keywords):
        count = keywords['surface'].count
        return joinIfNotNone('\n',
            '[H',
            'VERSION="4.0"',
            'OP="{}"'.format(self.OPTIMIZING_MODE),
            'WRK2="1"' if self.DELETE_ELEMENTS_THAT_CANNOT_BE_PROCESSED else None,
            'O2="1"' if self.AVOID_CROSS_DRIVE_HEIGHT_BY_HORIZONTAL_PROCESSING else None,
            'O4="1"' if self.BLOCKWISE_OPTIMIZING_OF_HORIZONTAL_DRILLINGS else None,
            'O5="1"' if self.AVOID_CROSS_DRIVE_HEIGHT_BY_FINISH_PROCESSING else None,
            'FM="{}"'.format(self.CLEARANCE_MODE),
            '',
            'GP="{:d}"'.format(self.GENERATED_AS_IN_X_MIRRORED_PROGRAM),
            'GY="{:d}"'.format(self.GENERATED_AS_IN_Y_MIRRORED_PROGRAM),
            'GXY="{:d}"'.format(self.GENERATED_AS_IN_X_AND_Y_MIRRORED_PROGRAM),
            'NP="1"',
            '',
            'FW="{}"'.format(self.CLEARANCE_VALUE),
            'INCH="0"',
            'HP="1"',
            'UP="0"',
            'DW="0"',
            'MAT="{mat}"'.format(mat=self.MACHINE_TYPE) if self.MACHINE_TYPE else None,
            'MI="0"',
            'ANZ="{count}"'.format(count=count) if count > 1 else None,
            '',
            self.calculatedValuesOfWorkpieceDescription(**keywords),
            '',
        )

    def customUserWorkpieceCommentFileDefaultText(self, **keywords):
        return '\n'.join((
            '<101 \\{}\\'.format(self.getLocalizedMacroComment(1, 'Kommentar')),
            'KM="{WorkpieceLabel}: {WorkpieceName}"',
            'KM="{CommentLabel}: {Comment}"',
            'KM="{ItemIDLabel}: {ItemID}"',
            'KM="{EdgeFrontLabel}: {EdgeFront}"',
            'KM="{EdgeBackLabel}: {EdgeBack}"',
            'KM="{EdgeLeftLabel}: {EdgeLeft}"',
            'KM="{EdgeRightLabel}: {EdgeRight}"',
            'KM="{CabinetNameLabel}: {CabinetName}"',
            'KM="{ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}"',
            'KM="{PositionNumberLabel}: {PositionNumber} {PositionDescription}"',
            '',
            '',
        ))

    def createCustomUserWorkpieceComment(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='woodWOP Workpiece Comment.txt',
            customUserVariableFileDefaultText=self.customUserWorkpieceCommentFileDefaultText(**keywords),
            **keywords)

    def createWorkpieceComment(self, **keywords):
        customUserWorkpieceComment = self.createCustomUserWorkpieceComment(**keywords) if self.USE_CUSTOM_USER_WORKPIECE_COMMENT else None

        if customUserWorkpieceComment:
            return customUserWorkpieceComment + ('\n' if not customUserWorkpieceComment.endswith('\n') else '')

        else:
            projectInfo = self.getProjectInfo()
            workpieceComments = self.getWorkpieceComments(**keywords)
            isProjectIDValid = ExportHelper.isProjectIDValid(self, **projectInfo)

            return joinIfNotNone('\n',
                '<101 \\{}\\'.format(self.getLocalizedMacroComment(1, 'Kommentar')),
                'KM="{WorkpieceLabel}: {WorkpieceName}"'.format_map(workpieceComments),
                'KM="{CommentLabel}: {Comment}"'.format_map(workpieceComments) if workpieceComments['Comment'] else None,
                'KM="{ItemIDLabel}: {ItemID}"'.format_map(workpieceComments),
                'KM="{Edge1Label}: {Edge1}"'.format_map(workpieceComments) if workpieceComments['Edge1'] else None,  # front
                'KM="{Edge2Label}: {Edge2}"'.format_map(workpieceComments) if workpieceComments['Edge2'] else None,  # back
                'KM="{Edge3Label}: {Edge3}"'.format_map(workpieceComments) if workpieceComments['Edge3'] else None,  # left
                'KM="{Edge4Label}: {Edge4}"'.format_map(workpieceComments) if workpieceComments['Edge4'] else None,  # right
                'KM="{CabinetNameLabel}: {CabinetName}"'.format_map(workpieceComments) if workpieceComments['CabinetName'] else None,
                'KM="{ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}"'.format_map(workpieceComments) if isProjectIDValid else None,
                'KM="{PositionNumberLabel}: {PositionNumber} {PositionDescription}"'.format_map(workpieceComments) if isProjectIDValid else None,
                '',
                '',
            )

    def customUserVariableFileDefaultText(self, **keywords):
        return '\n'.join((
            'example="12.3"',
            'KM="Example variable. See MPR format description -> Variable table for details."',
        ))

    def createCustomUserVariables(self, **keywords):
        return self.getMappedCustomUserVariablesFromFile(
            customUserVariableFileName='woodWOP Variables.txt',
            customUserVariableFileDefaultText=self.customUserVariableFileDefaultText(**keywords),
            **keywords)

    def createNCWorkpiece(self, **keywords):
        surface = keywords['surface']
        return joinIfNotNone('\n',
            '[001',
            '{}="{}"'.format(self.VAR_LENGTH, self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            'KM="{}"'.format(self.getLocalizedKMComment(2, 'Length')),
            '{}="{}"'.format(self.VAR_WIDTH, self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            'KM="{}"'.format(self.getLocalizedKMComment(3, 'Width')),
            '{}="{}"'.format(self.VAR_THICKNESS, self.formatAsCoordinate(data=surface.thickness)),
            'KM="{}"'.format(self.getLocalizedKMComment(4, 'Thickness')),
            self.createCustomUserVariables(**keywords) if self.USE_CUSTOM_USER_VARIABLES else None,
            '',
            '',
        )

    def __createPredefinedPlanes(self, **keywords):
        size = keywords['surface'].getFinishedSize()

        yield Plane(  # Front
            predefined=True, sideID='A00',
            matrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, -1.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
        )
        yield Plane(  # Right
            predefined=True, sideID='B00',
            matrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, 1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(1.0, 0.0, 0.0),
                offset=Vector(size.x, 0.0, 0.0)
            ),
        )
        yield Plane(  # Back
            predefined=True, sideID='C00',
            matrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(-1.0, 0.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(0.0, 1.0, 0.0),
                offset=Vector(size.x, size.y, 0.0)
            ),
        )
        yield Plane(  # Left
            predefined=True, sideID='D00',
            matrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, 0.0, 0.0)
            ),
            wopMatrix=Matrix(
                u=Vector(0.0, -1.0, 0.0),
                v=Vector(0.0, 0.0, 1.0),
                w=Vector(-1.0, 0.0, 0.0),
                offset=Vector(0.0, size.y, 0.0)
            ),
        )

    def generatePlanes(self, **keywords):
        def woodWOPSideIDGenerator():
            # According to MPR documentation 00 to 03 are reserved.
            # woodWOP uses a base-36 number
            for firstDigit in '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ':
                for secondDigit in '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ' if firstDigit != '0' else '456789ABCDEFGHIJKLMNOPQRSTUVWXYZ':
                    yield "F" + firstDigit + secondDigit

            assert False, 'Too many coordinate systems!'

        surface = keywords['surface']
        self.machiningPlanes = MachiningPlanes(sideIDGenerator=woodWOPSideIDGenerator())
        self.machiningPlanes.planes.extend(self.__createPredefinedPlanes(**keywords))

        self.machiningPlanes.createPlanes(filter(lambda machining: machining.matrix.w != Vector(0, 0, 1),
            filter(lambda machining: isinstance(machining, (HorizontalDado2D, HorizontalContour2D, NCMacro2D)), surface.machinings),
        ))

    def createCoordinateSystem(self, **keywords):
        self.generatePlanes(**keywords)

        return '\n'.join(
            itertools.chain(
                (self.createDefaultCoordinateSystem(**keywords),),  # Keep calling the old function in case somebody would have overwritten it.
                (self.createOneCoordinateSystem(plane=plane, **keywords) for plane in self.machiningPlanes.planes if not plane.predefined),
                ('',),
            ),
        )

    def __createCoordinatesMacro(self, **keywords):
        return '<00 \\{}\\'.format(self.getLocalizedMacroComment(2, 'Koordinatensystem'))

    def createDefaultCoordinateSystem(self, **keywords):
        return '\n'.join((
            '[K',
            self.__createCoordinatesMacro(**keywords),
            'NR="{sideID}"'.format(sideID=self.DEFAULT_COORDINATE_SYSTEM),
            'XP="0"',
            'XF="1"',
            'YP="0"',
            'YF="1"',
            'ZP="0"',
            'ZF="1"',
            'D1="0"',
            'KI="0"',
            'D2="0"',
            'MI="0"',
            '',
        ))

    def createOneCoordinateSystem(self, **keywords):
        plane = keywords['plane']
        return '\n'.join((
            '[K',
            self.__createCoordinatesMacro(**keywords),
            'NR="{sideID}"'.format(sideID=plane.sideID),
            'XP="{x}"'.format(x=self.formatAsCoordinate(data=plane.origin.x)),
            'XF="1"',
            'YP="{y}"'.format(y=self.formatAsCoordinate(data=plane.origin.y)),
            'YF="1"',
            'ZP="{z}"'.format(z=self.formatAsCoordinate(data=plane.origin.z)),
            'ZF="1"',
            'D1="{angle}"'.format(angle=self.formatAsCoordinate(data=normalizeAngle(plane.cAngleToXYPlane - 90))),
            'KI="{angle}"'.format(angle=self.formatAsCoordinate(data=plane.aAngle)),
            'D2="0"',
            'MI="0"',
            '',
        ))

    def createWorkpieceDefinition(self, **keywords):
        surface = keywords['surface']
        return '\n'.join((
            '<100 \\{}\\'.format(self.getLocalizedMacroComment(3, 'Werkstck')),  # No, it is not a typo; there is no u umlaut.
            'LA="{}"'.format(self.VAR_LENGTH),
            'BR="{}"'.format(self.VAR_WIDTH),
            'DI="{}"'.format(self.VAR_THICKNESS),
            'FNX="{}"'.format(self.formatAsCoordinate(data=surface.getAdditionX())),
            'FNY="{}"'.format(self.formatAsCoordinate(data=surface.getAdditionY() if self.ZERO_OFFSET_Y_IS_SIZING_ADDITION else 0.0)),
            'AX="{}"'.format(self.formatAsCoordinate(data=surface.getEdgesAdditionX())),
            'AY="{}"'.format(self.formatAsCoordinate(data=surface.getEdgesAdditionY())),
            '',
            '',
        ))

    @functools.lru_cache()
    def createKATAndMNMCategories(self, index, defaultText):
        return '\n'.join((
            'KAT="{}"'.format(self.getLocalizedKATCategories(index, defaultText)),
            'MNM="{}"'.format(self.getLocalizedMNMCategories(index, defaultText)),
        ))

    def __createKMCommentIfNeeded(self, **keywords):
        if 'comment' in keywords and keywords['comment']:
            return 'KM="{}"'.format(keywords['comment'])

    def createBlockMacroForVacuumSuctionCups(self, **keywords):
        return '\n'.join((
            '<121 \\{}\\'.format(self.getLocalizedMacroComment(14, 'Block')),
            'XP="0"',
            'YP="0"',
            'ZP="0"',
            'NM=""',
            'DP="{quantity}"'.format(quantity=keywords['quantityX']),
            '',
        ))

    def getDimensionsDependingOnQuantityForVacuumSuctionCups(self, **keywords):
        quantity = keywords['quantity']
        cupDistance = keywords['cupDistance']
        finishedSize = keywords['finishedSize']

        a2, a3 = 0, 0
        if quantity == 1:
            a1 = finishedSize / 2
        elif quantity == 2:
            a1 = cupDistance
            a2 = finishedSize - cupDistance
        else:
            a1 = cupDistance
            a2 = finishedSize / 2
            a3 = finishedSize - cupDistance
        return a1, a2, a3

    def createSingleVacuumCupParalellToYAxis(self, **keywords):
        ya1, ya2, ya3 = self.getDimensionsDependingOnQuantityForVacuumSuctionCups(quantity=keywords['quantityY'], cupDistance=90, finishedSize=keywords['surface'].getFinishedSizeY(), **keywords)

        return '\n'.join((
            '<115 \\{}\\'.format(self.getLocalizedMacroComment(13, 'SaugerK')),
            'XA="{x}"'.format(x=keywords['x']),  # X coordinate of console
            'AN="{quantity}"'.format(quantity=keywords['quantityY']),  # Number of cups in Y
            'YA1="{y}"'.format(y=ya1),  # Y coordinate of 1 cup
            'YA2="{y}"'.format(y=ya2),  # Y coordinate of 2 cup
            'YA3="{y}"'.format(y=ya3),  # Y coordinate of 3 cup
            '',
        ))

    def createMultipleVacuumCupsParalellToYAxis(self, **keywords):
        xa1, xa2, xa3 = self.getDimensionsDependingOnQuantityForVacuumSuctionCups(quantity=keywords['quantityX'], cupDistance=97.5, finishedSize=keywords['surface'].getFinishedSizeX(), **keywords)

        return joinIfNotNone('\n',
            self.createSingleVacuumCupParalellToYAxis(x=xa1, **keywords),
            self.createSingleVacuumCupParalellToYAxis(x=xa2, **keywords) if xa2 else '',
            self.createSingleVacuumCupParalellToYAxis(x=xa3, **keywords) if xa3 else '',
        )

    def getQuantityForVacuumSuctionCups(self, **keywords):
        finishedSize = keywords['finishedSize']
        dim2 = keywords['dim2']
        dim3 = keywords['dim3']

        if finishedSize < dim2:
            return 1
        elif finishedSize < dim3:
            return 2
        else:
            return 3

    def createVacuumSuctionCups(self, **keywords):
        # standard vacuum suction cup size = 115 x 140 mm
        # default dimensions of automatic vacuum cup positioning
        surface = keywords['surface']
        quantityX = self.getQuantityForVacuumSuctionCups(finishedSize=surface.getFinishedSizeX(), dim2=310, dim3=520, **keywords)
        quantityY = self.getQuantityForVacuumSuctionCups(finishedSize=surface.getFinishedSizeY(), dim2=400, dim3=560, **keywords)

        return '\n'.join((
            self.createBlockMacroForVacuumSuctionCups(quantityX=quantityX, **keywords),
            self.createMultipleVacuumCupsParalellToYAxis(quantityX=quantityX, quantityY=quantityY, **keywords),
        ))

    # Machinings
    # Drillings
    def __createCommonDrillingDiameterOrToolID(self, **keywords):
        drilling = keywords['drilling']
        toolID = drilling.toolID
        return (
            '{toolLabel}="{toolID}"'.format(toolLabel=keywords['toolLabel'], toolID=toolID)
            if toolID and (hasattr(drilling, 'tool') and drilling.tool.isToolIDValidIntDirect()) else
            'DU="{diameter}"'.format(diameter=self.formatAsCoordinate(data=drilling.installedDiameter))
        )

    def __createCommonDrillingProperties(self, **keywords):
        drilling = keywords['drilling']
        depth = keywords['depth'] if 'depth' in keywords else drilling.depth

        return joinIfNotNone('\n',
            'XA="{x}"'.format(x=self.formatAsCoordinate(data=drilling.position.x)),
            'YA="{y}"'.format(y=self.formatAsCoordinate(data=drilling.position.y)),
            'ZA="{z}"'.format(z=self.formatAsCoordinate(data=keywords['zPosition'])) if 'zPosition' in keywords else None,
            self.__createCommonDrillingDiameterOrToolID(**keywords),
            'TI="{depth}"'.format(depth=self.formatAsCoordinate(data=depth)),
        )

    def __createCommonVerticalDrilling(self, **keywords):
        return joinIfNotNone('\n',
            '<102 \\{}\\'.format(self.getLocalizedMacroComment(4, 'BohrVert')),
            self.__createCommonDrillingProperties(toolLabel='TNO', **keywords),
            'BM="{}"'.format(keywords['drillMode']),
            'AN="1"',
            'F_="STANDARD"',
            'S_="2"',
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            self.createKATAndMNMCategories(1, 'Bohren vertikal'),
            self.__createKMCommentIfNeeded(**keywords),
            '??="1"',
        )

    DRILL_MODE_SLOW_FAST = 'LS'
    DRILL_MODE_SLOW_FAST_SLOW = 'LSL'
    DRILL_MODE_FAST_FAST = 'SS'
    DRILL_MODE_FAST_FAST_FAST = 'SSS'

    def getDrillMode(self, **keywords):
        type = keywords['type']

        if self.DRILL_MODE == 'slow' and type == 'blind-hole':
            return self.DRILL_MODE_SLOW_FAST
        elif self.DRILL_MODE == 'slow' and type == 'through-hole':
            return self.DRILL_MODE_SLOW_FAST_SLOW
        elif self.DRILL_MODE == 'fast' and type == 'blind-hole':
            return self.DRILL_MODE_FAST_FAST
        elif self.DRILL_MODE == 'fast' and type == 'through-hole':
            return self.DRILL_MODE_FAST_FAST_FAST
        else:
            message = getLocalizedString(WOODWOP_MESSAGES, 1, "Drill mode not valid, using LS instead. Workpiece: {workpiece}")
            logging.info(message.format(workpiece=keywords['surface'].workpieceName))
            return self.DRILL_MODE_SLOW_FAST

    def createVerticalDrilling(self, **keywords):
        return '\n'.join((
            self.__createCommonVerticalDrilling(drillMode=self.getDrillMode(type='blind-hole', **keywords), **keywords),
            '',
            '',
        ))

    def createVerticalDrillingPiercing(self, **keywords):
        return '\n'.join((
            self.__createCommonVerticalDrilling(drillMode=self.getDrillMode(type='through-hole', **keywords) if self.PIERCING_DRILLING_ADDITION > 0 else self.getDrillMode(type='blind-hole', **keywords), **keywords),
            '',
            '',
        ))

    def createAngleParametersForVerticalGridDrilling(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        if self.CREATE_GRID_DRILLING_WITH_ANGLE_INSTEAD_XY_DIRECTION:
            return '\n'.join((
                'WI="{angle}"'.format(angle=normalizeAngle(gridDrilling.getXYAngle())),
            ))
        else:
            return '\n'.join((
                'XR="{x}"'.format(x=self.formatAsCoordinate(data=gridDrilling.gridNormalizedDirection.x)),
                'YR="{y}"'.format(y=self.formatAsCoordinate(data=gridDrilling.gridNormalizedDirection.y)),
            ))

    def createVerticalGridDrilling(self, **keywords):
        comment = self.getLocalizedKMComment(14, 'Vertical Grid Drilling')
        gridDrilling = keywords['gridDrilling']
        return '\n'.join((
            self.__createCommonVerticalDrilling(drilling=gridDrilling.getStartDrilling(), drillMode=self.getDrillMode(type='blind-hole', **keywords), comment=comment, **keywords),
            'LA="{length}"'.format(length=self.formatAsCoordinate(data=gridDrilling.getGridLength())),
            self.createAngleParametersForVerticalGridDrilling(**keywords),
            'AB="{distance}"'.format(distance=self.formatAsCoordinate(data=gridDrilling.gridDistance)),
            '',
            '',
        ))

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def createVerticalCenterPunch(self, **keywords):
        comment = self.getLocalizedKMComment(11, 'Vertical Center Punch')
        return super().createVerticalCenterPunch(comment=comment, **keywords)

    def createFiveAxisDrilling(self, **keywords):
        return self.__createCommonFiveAxisDrilling(**keywords)

    def createFiveAxisCenterPunch(self, **keywords):
        comment = self.getLocalizedKMComment(13, 'Five Axis Center Punch')
        return self.__createCommonFiveAxisDrilling(comment=comment, **keywords)

    def __createCommonFiveAxisDrilling(self, **keywords):
        drilling = keywords['drilling']

        return joinIfNotNone('\n',
            '<104 \\{}\\'.format(self.getLocalizedMacroComment(11, 'BohrUniv')),
            self.__createCommonDrillingProperties(toolLabel='TNO', **keywords),
            'AB="32"',
            'AN="1"',
            'CA="{c}"'.format(c=self.formatAsAngle(data=drilling.cAngle)),
            'WI="{a}"'.format(a=self.formatAsAngle(data=drilling.aAngle)),
            'BM="STD"',
            'F_="STANDARD"',
            'ZT="0"',
            'RM="0"',
            'VW="0"',
            'HP="0"',
            'SP="0"',
            'YVE="0"',
            'ASG="2"',
            self.createKATAndMNMCategories(8, 'Universalbohren'),
            self.__createKMCommentIfNeeded(**keywords),
            'ORI=""',
            'MX="0"',
            'MY="0"',
            'MZ="0"',
            'MXF="1"',
            'MYF="1"',
            'MZF="1"',
            'KO="00"',
            '',
            '',
        )

    def createNCStop(self, **keywords):
        return '\n'.join((
            '<117 \\{}\\'.format(self.getLocalizedMacroComment(5, 'NCStop')),
            'KM="{comment}"'.format(comment=self.substituteUmlautsIfNeeded(data=keywords['comment'])),
            'VL="{release:d}"'.format(release=keywords['releaseVacuum'] if 'releaseVacuum' in keywords else False),
            'XA="{parkx}"'.format(parkx=keywords['parkPositionXString'] if 'parkPositionXString' in keywords else '0'),
            'F_="30"',
            'YA="0"',
            '_Y="0"',
            '_X="{movex:d}"'.format(movex='parkPositionXString' in keywords),
            'XV="{xv}"'.format(xv=self.formatAsCoordinate(data=keywords['additionX'] if 'additionX' in keywords else 0.0)),
            'YV="{yv}"'.format(yv=self.formatAsCoordinate(data=keywords['additionY'] if 'additionY' in keywords else 0.0)),
            'WT="0.0"',
            'ZR="1"',
            self.createKATAndMNMCategories(2, 'NC-Stop'),
            '',
            '',
        ))

    def createHorizontalDrillingDirection(self, **keywords):
        drilling = keywords['drilling']
        angle = (drilling.cAngle % 360)
        if isNearlyEqual(angle, 0) or isNearlyEqual(angle, 360):
            direction = 'BM="XP"'
        elif isNearlyEqual(angle, 90):
            direction = 'BM="YP"'
        elif isNearlyEqual(angle, 180):
            direction = 'BM="XM"'
        elif isNearlyEqual(angle, 270):
            direction = 'BM="YM"'
        else:
            direction = '\n'.join((
                'BM="C"',
                'WI="{a}"'.format(a=self.formatAsAngle(data=angle)),
            ))

        return direction

    def createRotatedHorizontalDrilling(self, **keywords):
        drilling = keywords['drilling']
        if drilling.toolID and drilling.tool.face == Tool.FIVE_AXIS and not self.CREATE_ROTATED_HORIZONTAL_DRILLING_AS_HORIZONTAL_DRILLING:
            return self.__createCommonFiveAxisDrilling(**keywords)  # probably only used within woodWOP 5
        else:
            return self.__createCommonHorizontalDrilling(**keywords)  # default from woodWOP 6 on

    def createRotatedHorizontalCenterPunch(self, **keywords):
        comment = self.getLocalizedKMComment(12, 'Horizontal Center Punch')
        return super().createRotatedHorizontalCenterPunch(comment=comment, **keywords)

    def __createCommonHorizontalDrilling(self, **keywords):
        return joinIfNotNone('\n',
            '<103 \\{}\\'.format(self.getLocalizedMacroComment(6, 'BohrHoriz')),
            self.__createCommonDrillingProperties(toolLabel='T_', **keywords),
            self.createHorizontalDrillingDirection(**keywords),
            'AN="1"',
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            self.createKATAndMNMCategories(3, 'Horizontalbohren'),
            self.__createKMCommentIfNeeded(**keywords),
            '',
            '',
        )

    def getRouterDepthPerStep(self, **keywords):
        depthPerStep = super().getRouterDepthPerStep(**keywords)

        # woodWOP does not allow the depth-per-step to be deeper than the depth of the machining
        if isNearlyGreater(depthPerStep, keywords['machiningDepth']):
            return 0.0

        return depthPerStep

    # Pockets
    def createVerticalCommonPocket(self, **keywords):
        pocket = keywords['pocket']
        return '\n'.join((
            '<112 \\{}\\'.format(self.getLocalizedMacroComment(7, 'Tasche')),
            'XA="{x}"'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'YA="{y}"'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'LA="0"',
            'BR="0"',
            'RD="{diameter}/2"'.format(diameter=self.formatAsCoordinate(data=pocket.diameter)),
            'WI="0"',
            'TI="{depth}"'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'ZT="{stepDepth}"'.format(stepDepth=self.formatAsCoordinate(data=self.getRouterDepthPerStep(machiningDepth=keywords['depth'], **keywords))),
            'XY="{}"'.format(keywords['overlap']),
            'T_="{}"'.format(pocket.toolID),
            'F_="STANDARD"',
            'DS="{direction:d}"'.format(direction=not self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
            'OSZI="0"',
            'BL="0"',
            'OSZVS="0"',
            'HP="0"',
            'SP="0"',
            'YVE="0"',
            'ASG="2"',
            'KG="0"',
            'RP="STANDARD"',
            'MX="0"',
            'MY="0"',
            'MZ="0"',
            'MLM="0"',
            'MXR="STANDARD"',
            'MYR="STANDARD"',
            'MZR="STANDARD"',
            'MLR="STANDARD"',
            '_MXF="1"',
            'MXF="1"',
            '_MYF="1"',
            'MYF="1"',
            '_MZF="1"',
            'MZF="1"',
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            self.createKATAndMNMCategories(4, 'Tasche'),
            '',
            '',
        ))

    def createVerticalPocket(self, **keywords):
        return self.createVerticalCommonPocket(overlap=self.CLEAR_OUT_POCKET_OVERLAP_FACTOR, **keywords)

    def createVerticalPocketPiercing(self, **keywords):
        if keywords['pocket'].diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER:
            overlap = 0
        else:
            overlap = self.CLEAR_OUT_POCKET_OVERLAP_FACTOR
        return self.createVerticalCommonPocket(overlap=overlap, **keywords)

    def createHorizontalPocket(self, **keywords):
        pocket = keywords['pocket']
        depth = pocket.depth
        return '\n'.join((
            '<123 \\{}\\'.format(self.getLocalizedMacroComment(15, 'HTasc')),
            'XA="{x}"'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'YA="{y}"'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'ZA="{z}"'.format(z=self.formatAsCoordinate(data=pocket.zPosition)),
            'LA="0"',
            'BR="0"',
            'RD="{diameter}/2"'.format(diameter=self.formatAsCoordinate(data=pocket.diameter)),
            'WI="{a}"'.format(a=self.formatAsAngle(data=normalizeAngle(pocket.cAngle))),
            'TI="{depth}"'.format(depth=self.formatAsCoordinate(data=depth)),
            'ZT="{stepDepth}"'.format(stepDepth=self.formatAsCoordinate(data=self.getRouterDepthPerStep(machiningDepth=depth, **keywords))),
            'AB="{distance}"'.format(distance=self.formatAsCoordinate(data=self.getDadoRabbetSecurityDistance(**keywords))),
            'DS="{direction:d}"'.format(direction=not self.isDadoRabbetReplacementCW(pocket, self.POCKET_DIRECTION)),
            'T_="{toolID}"'.format(toolID=pocket.toolID),
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            self.createKATAndMNMCategories(11, 'Pocket Horizontal'),
            '',
            '',
        ))

    # This was the former code for horizontal pocket (until interiorcad 2020 F5.2)
    def createFixeAxisPocket(self, **keywords):
        pocket = keywords['pocket']
        depth = pocket.depth
        return '\n'.join((
            '<141 \\{}\\'.format(self.getLocalizedMacroComment(14, 'VTasche')),
            'XA="{x}"'.format(x=self.formatAsCoordinate(data=pocket.position.x)),
            'YA="{y}"'.format(y=self.formatAsCoordinate(data=pocket.position.y)),
            'Z_="{z}"'.format(z=self.formatAsCoordinate(data=pocket.zPosition)),
            'LA="0"',
            'BR="0"',
            'RD="{diameter}/2"'.format(diameter=self.formatAsCoordinate(data=pocket.diameter)),
            'WI="0"',
            'TI="{depth}"'.format(depth=self.formatAsCoordinate(data=depth)),
            'ZT="{stepDepth}"'.format(stepDepth=self.formatAsCoordinate(data=self.getRouterDepthPerStep(depth=depth, **keywords))),
            'T_="{toolID}"'.format(toolID=pocket.toolID),
            'A_="90"',
            'C_="{c}"'.format(c=self.formatAsAngle(data=calculateCAngleFromWVector(pocket.matrix.w) - 90.0)),
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            self.createKATAndMNMCategories(9, 'Vektor-Tasche'),
            '',
            '',
        ))

    # Contours
    def createSizingMacro3SFormatSpecificBlock(self, **keywords):
        return '\n'.join((
            'VA="Z_Mass {z}"'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            'VA="F 8"',
            'VA="Tool {toolID}"'.format(toolID=keywords['sizing'].toolID),
        ))

    def createSizingMacro3SBHXFormatSpecificBlock(self, **keywords):
        return '\n'.join((
            'VA="Z_Mass {z}"'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            'VA="Tool_1 {toolID}"'.format(toolID=keywords['sizing'].toolID),
            'VA="Tool_2 {toolID}"'.format(toolID=keywords['sizing'].toolID),
            'VA="F 8"',
        ))

    def createSizingMacro3S4SFormatBHX05XSpecificBlock(self, **keywords):
        return '\n'.join((
            'VA="Z_Mass {z}"'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),  # Z-Messure absolut
            'VA="Tool_1 {toolID}"'.format(toolID=keywords['sizing'].toolID),  # Toolnumber rightturned
            'VA="Tool_2 {toolID}"'.format(toolID=keywords['sizing'].toolID),  # Toolnumber leftturned
            'VA="F 8"',  # Feed in m/min
            'VA="Sides_3 1"',  # BOOL 0= 4 Sides | 1= 3 Sides
        ))

    def createSizingMacroFormatRSpecificBlock(self, **keywords):
        return '\n'.join((
            'VA="Abstand 0"',
            'VA="ZMass {z}"'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            'VA="Vorschub 10"',
            'VA="Werkzeug {toolID}"'.format(toolID=keywords['sizing'].toolID),
        ))

    def createSizingMacroSpecificBlock(self, **keywords):
        # instead CREATE_SIZING_WITH_MACRO
        if self.CREATE_SIZING_WITH_MACRO_NAME == '3s_format':
            return self.createSizingMacro3SFormatSpecificBlock(**keywords)
        elif self.CREATE_SIZING_WITH_MACRO_NAME == '3S_Format_BHX5X' or self.CREATE_SIZING_WITH_MACRO_NAME == '3S_Format_BHX200':
            return self.createSizingMacro3SBHXFormatSpecificBlock(**keywords)
        elif self.CREATE_SIZING_WITH_MACRO_NAME == '3S_4S_Format_BHX05X':
            return self.createSizingMacro3S4SFormatBHX05XSpecificBlock(**keywords)
        else:
            return self.createSizingMacroFormatRSpecificBlock(**keywords)

    def createSizingWithMacro(self, **keywords):
        return '\n'.join((
            '<139 \\{}\\'.format(self.getLocalizedMacroComment(12, 'Komponente')),
            'IN="{macro}.mpr"'.format(macro=self.CREATE_SIZING_WITH_MACRO_NAME),  # instead CREATE_SIZING_WITH_MACRO
            'XA="0.0"',
            'YA="0.0"',
            'ZA="0.0"',
            'EM="0"',
            self.createSizingMacroSpecificBlock(**keywords),
            'KAT="Komponentenmakro"',
            'MNM="{name}"'.format(name=self.CREATE_SIZING_WITH_MACRO_NAME),
            '',
            '',
        ))

    def getFlyingApproachFrom(self, **keywords):
        sizing = keywords['sizing']
        flyingApproach = '0'
        if sizing.needsFlyingApproach:
            flyingApproach = '1'
        return flyingApproach

    def __createCommonContourBlock(self, **keywords):
        contourExecutionID = keywords['machining'].wopContourDefinitionIndex

        # Feedrate F and Speed S are handled different in woodWOP. While Feedrate F should be "Standard", Speed S should be missing. Both will trigger their corresponding defaults in woodWOP.
        if 'contour' in keywords and keywords['contour'].currentFeedRate:
            contour = keywords['contour']
            feedRate = self.formatAsCoordinate(data=contour.currentFeedRate)
        else:
            feedRate = 'STANDARD'

        return '\n'.join((
            '<105 \\{}\\'.format(self.getLocalizedMacroComment(8, 'Konturfraesen')),
            'EA="{}:{}"'.format(contourExecutionID, keywords['startVertex']),
            'MDA="{:s}"'.format(keywords['approach']),
            'RK="{rk}"'.format(rk=keywords['radiusCompensation']),
            'EE="{}:{}"'.format(contourExecutionID, keywords['endVertex']),
            'MDE="{:s}"'.format(keywords['departure']),
            'EM="{}"'.format(keywords['flyingApproach']),
            'TNO="{}"'.format(keywords['machining'].toolID),
            'ZA="{z}"'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            'KM="{}"'.format(keywords['localizedComment']),
            self.createKATAndMNMCategories(5, 'Fr\u00e4sen'),
            'AB="{ab}"'.format(ab=self.formatAsCoordinate(data=keywords['contourDistance'] if 'contourDistance' in keywords else 0.0)),
            'F_="{feedRate}"'.format(feedRate=feedRate),
        ))

    def __createContourSizingSpecificBlock(self, **keywords):
        return '\n'.join((
            'RI="1"',
            'SM="0"',
            'AF="{distance}"'.format(distance=self.SIZING_APPROACH_DISTANCE),
            'VLS="0"',
            'VLE="0"',
            'HP="0"',
            'SP="0"',
            'YVE="0"',
            'ASG="2"',
            'RSEL="0"',
            'RWID="0"',
            'MX="0"',
            'MY="0"',
            'MZ="0"',
            'MXF="1"',
            'MYF="1"',
            'MZF="1"',
            '??="{execute:d}"'.format(execute=True),
        ))

    def __createContourSizingBlock(self, **keywords):
        machiningDepth = keywords['machining'].actualDepth
        return '\n'.join((
            self.__createCommonContourBlock(**keywords),
            self.__createCommonContourDepthPerStep(machiningDepth=machiningDepth, depthPerStep=self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords), **keywords),
            self.__createContourSizingSpecificBlock(**keywords),
            '',
            '',
        ))

    def __createContourSizingBlockInTwoPasses(self, **keywords):
        originalZValue = keywords['zValue']
        del keywords['zValue']

        machining = keywords['machining']
        surface = keywords['surface']

        machiningDepth = surface.thickness - self.SIZING_FIRST_PASS_REMAINING_THICKNESS
        firstZValue = self.getMillingZValue(depth=machiningDepth, **keywords)

        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        if isNearlyLesser(depthPerStep, machiningDepth) and not isNearlyEqual(depthPerStep, 0.0):
            message = self.getLocalizedMessage(76, "Router with insufficient depth per step to be able to make nesting remaining-thickness sizing. Workpiece: {workpiece}, Tool name: '{toolName}', Tool ID: {toolID}")
            logging.error(message.format(
                workpiece=surface.workpieceName,
                toolName=machining.tool.name if hasattr(machining.tool, 'name') else None,
                toolID=machining.toolID
            ))

        return '\n'.join((
            # First pass
            self.__createCommonContourBlock(zValue=firstZValue, contourDistance=self.SIZING_FIRST_PASS_DISTANCE, **keywords),
            self.__createCommonContourDepthPerStep(machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords),
            self.__createContourSizingSpecificBlock(**keywords),
            '',
            # Second pass
            self.__createCommonContourBlock(zValue=originalZValue, **keywords),
            self.__createContourSizingSpecificBlock(**keywords),
            '',
            '',
        ))

    def __createCommonContourSecurityDistances(self, contourSecurityDistanceInfo, **keywords):
        securityDistance = keywords['securityDistance']
        if self.CONTOUR_APPROACH_DEPARTURE_INTEGRATION == 'ever' or (self.CONTOUR_APPROACH_DEPARTURE_INTEGRATION == 'only open' and not contourSecurityDistanceInfo['isClosed']):
            return joinIfNotNone('\n',
                'VLS="{vls}"'.format(vls=self.formatAsCoordinate(data=securityDistance)) if contourSecurityDistanceInfo['useApproach'] else None,
                'VLE="{vle}"'.format(vle=self.formatAsCoordinate(data=securityDistance)) if contourSecurityDistanceInfo['useDeparture'] else None,
            )

    def __createDefaultSecurityDistanceInfo(self, **keywords):
        return {
            'isClosed': False,
            'useApproach': True,
            'useDeparture': True,
        }

    def __createCommonContourDepthPerStep(self, **keywords):
        depthPerStep = keywords['depthPerStep']
        boardThickness = keywords['surface'].thickness
        machiningDepth = keywords['machiningDepth']

        # TODO CONTOUR2.0 IT-237/IT-265 : should this be a generalised function? -- NG
        useSteps = (machiningDepth > depthPerStep > 0.0) and not isNearlyEqual(depthPerStep, 0.0) and not isNearlyEqual(machiningDepth, depthPerStep)
        if useSteps:
            numberOfStepsDivider = math.ceil(machiningDepth / depthPerStep)
            rawDepthStep = machiningDepth / numberOfStepsDivider

            if self.CONTOUR_NEEDS_ADDITIONAL_STEP:
                numberOfSteps = numberOfStepsDivider - 1
            else:
                numberOfSteps = numberOfStepsDivider

            assert 0 < rawDepthStep <= depthPerStep or isNearlyEqual(rawDepthStep, depthPerStep), "Calculated step should not be 0 and not be bigger than maximum step!"
            startDepth = self.zValueMillingConverter.convertWithThickness(boardThickness, rawDepthStep)
        else:
            numberOfSteps = 0
            startDepth = 0

        return joinIfNotNone('\n',
            'STUFEN="{useSteps:d}"'.format(useSteps=useSteps),
            'ZSTART="{z}"'.format(z=self.formatAsCoordinate(data=startDepth)),
            'ANZZST="{count}"'.format(count=self.formatAsInt(data=numberOfSteps)),
        )

    def __createContourBlock(self, **keywords):
        contour = keywords['contour']
        contourSecurityDistanceInfo = self.__createDefaultSecurityDistanceInfo(**keywords)
        contourSecurityDistanceInfo['isClosed'] = contour.isClosed()
        contourSecurityDistanceInfo['useApproach'] = contour.approach != 'without'
        contourSecurityDistanceInfo['useDeparture'] = contour.departure != 'without'

        boardThickness = keywords['surface'].thickness
        depthPerStep = contour.currentDepthPerStep
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=contour.actualDepth, **keywords)

        return joinIfNotNone('\n',
            self.__createCommonContourBlock(**keywords),
            self.__createCommonContourSecurityDistances(contourSecurityDistanceInfo, **keywords),
            self.__createCommonContourDepthPerStep(machiningDepth=contour.actualDepth, depthPerStep=depthPerStep, **keywords),
            '',
            '',
        )

    def __createRabbetContourBlock(self, **keywords):
        machiningDepth = keywords['rabbet'].actualDepth
        contourSecurityDistanceInfo = self.__createDefaultSecurityDistanceInfo(**keywords)
        return joinIfNotNone('\n',
            self.__createCommonContourBlock(**keywords),
            self.__createCommonContourSecurityDistances(contourSecurityDistanceInfo, **keywords),
            self.__createCommonContourDepthPerStep(machiningDepth=machiningDepth, depthPerStep=self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords), **keywords),
            '',
            '',
        )

    def __createDadoContourBlock(self, **keywords):
        machiningDepth = keywords['dado'].actualDepth
        contourSecurityDistanceInfo = self.__createDefaultSecurityDistanceInfo(**keywords)
        contourSecurityDistanceInfo['isClosed'] = keywords['dado'].isInsetBoth()
        return joinIfNotNone('\n',
            self.__createCommonContourBlock(**keywords),
            self.__createCommonContourSecurityDistances(contourSecurityDistanceInfo, **keywords),
            self.__createCommonContourDepthPerStep(machiningDepth=machiningDepth, depthPerStep=self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords), **keywords),
            '',
            '',
        )

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=WoodWOPExport.RADIUS_COMPENSATION_NONE, left=WoodWOPExport.RADIUS_COMPENSATION_LEFT, right=WoodWOPExport.RADIUS_COMPENSATION_RIGHT)

    def __getSizingRadiusCompensation(self, **keywords):
        return WoodWOPExport.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else WoodWOPExport.RADIUS_COMPENSATION_LEFT

    def getSizingApproachMode(self, **keywords):
        if self.SIZING_ENTRY_OF_PROFILE == 'vertical flying':
            return self.APPROACH_MODE_WITHOUT_FLYING
        else:
            return super().getSizingApproachMode(**keywords)

    def isSizingStrategyUsingTwoPasses(self, **keywords):
        area = keywords['surface'].getSawnSizeRectangularArea()

        return (
            self.SIZING_PASS_STRATEGY == 'force-two-passes'
            or (self.SIZING_PASS_STRATEGY == 'two-passes-for-small-workpieces' and area < 1000000 * self.SIZING_PASS_STRATEGY_AREA_THRESHOLD)
        )

    def createSizingContourBlock(self, **keywords):
        sizing = keywords['sizing']

        approachType, departureType, approachFlying = self.getSizingApproachMode(**keywords)
        if sizing.needsFlyingApproach and approachFlying == self.APPROACH_TYPE_NONE:
            approachFlying = self.APPROACH_TYPE_FLYING

        keywords['localizedComment'] = self.getLocalizedKMComment(6, 'Sizing')
        keywords['approach'] = approachType
        keywords['departure'] = departureType
        keywords['flyingApproach'] = approachFlying
        keywords['radiusCompensation'] = self.__getSizingRadiusCompensation(**keywords)
        keywords['startVertex'] = 0
        keywords['endVertex'] = sizing.segmentCount if self.CREATE_SIZING_WITH_THREE_SIDE_CONTOUR else 0  # endVertex=contour.segmentCount is NOT -1, as we must count the start point too.

        finishedSizeShape = self.getFinishedSizeShape(**keywords)

        if self.CREATE_SIZING_WITH_MACRO_NAME and finishedSizeShape.isRectangular():
            return self.createSizingWithMacro(**keywords)
        elif self.isSizingStrategyUsingTwoPasses(**keywords) and keywords['surface'].isFirstSide:
            return self.__createContourSizingBlockInTwoPasses(machining=sizing, **keywords)
        else:
            return self.__createContourSizingBlock(machining=sizing, **keywords)

    def createFreeFormPocket(self, **keywords):
        contour = keywords['contour']
        machining = keywords['machining']

        depthPerStep = contour.currentDepthPerStep
        if isNearlyEqual(depthPerStep, 0.0):
            depthPerStep = self.getRouterDepthPerStep(machiningDepth=contour.actualDepth, **keywords)

        return joinIfNotNone('\n',
            '<181 \\{}\\'.format(self.getLocalizedMacroComment(16, 'FreiFormTasche')),
            'EA="{id}:0"'.format(id=machining.wopContourDefinitionIndex),
            'AD="{ad}"'.format(ad=self.formatAsCoordinate(data=keywords['contourDistance'] if 'contourDistance' in keywords else 0.0)),
            'AZ="5"',
            'ZU="0.5"',
            'TI="{depth}"'.format(depth=self.formatAsCoordinate(data=contour.actualDepth)),
            'ZT="{depthPerStep}"'.format(depthPerStep=self.formatAsCoordinate(data=depthPerStep)),
            'XY="{overlap}"'.format(overlap=self.CLEAR_OUT_POCKET_OVERLAP_FACTOR),
            'DS="{direction:d}"'.format(direction=not contour.path.isClockwise()),  # "not" because interiorcad's True/False for CW/CCW are inverted against woodWOPs Clockwise = "0" & Counterclockwise = "1"
            'T_="{tool}"'.format(tool=machining.toolID),
            'F_="{feedRate}"'.format(feedRate='STANDARD' if isNearlyEqual(contour.currentFeedRate, 0.0) else self.formatAsCoordinate(data=contour.currentFeedRate)),
            self.createKATAndMNMCategories(12, 'Fr\u00e4sen Tasche Freiform'),
            '',
            '',
        )

    def createVerticalContour(self, **keywords):
        contour = keywords['contour']

        approachType, approachDepartureDummy, approachFlying = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departureApproachDummy, departureType, departureFlying = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        keywords['localizedComment'] = self.getLocalizedKMComment(10, 'Contour')
        keywords['approach'] = approachType
        keywords['departure'] = departureType
        keywords['flyingApproach'] = approachFlying
        keywords['radiusCompensation'] = self.getWopRadiusCompensation(**keywords)
        keywords['startVertex'] = 0
        keywords['endVertex'] = contour.segmentCount  # endVertex=contour.segmentCount is NOT -1, as we must count the start point too.
        keywords['securityDistance'] = self.CONTOUR_APPROACH_DEPARTURE_DISTANCE
        keywords['zValue'] = self.getMillingZValue(depth=contour.actualDepth, **keywords)

        if contour.clearOut:
            return self.createFreeFormPocket(**keywords)
        else:
            return self.__createContourBlock(**keywords)

    def createVerticalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createVerticalContour, **keywords))

    def createHorizontalCommonContourStartBlock(self, **keywords):
        machining = keywords['machining']
        contourExecutionID = machining.wopContourDefinitionIndex

        return joinIfNotNone('\n',
            '<133 \\{}\\'.format(self.getLocalizedMacroComment(8, 'Konturfraesen')),
            'EA="{}:{}"'.format(contourExecutionID, keywords['startVertex']),
            'MDA="{:s}"'.format(keywords['approach']),
            'RK="{rk}"'.format(rk=keywords['radiusCompensation']),
            'EE="{}:{}"'.format(contourExecutionID, keywords['endVertex']),
            'MDE="{:s}"'.format(keywords['departure']),
            'EM="{}"'.format(keywords['flyingApproach']),
            'TNO="{}"'.format(machining.toolID),
            'F_="{feedRate}"'.format(feedRate='STANDARD' if not hasattr(machining, 'currentFeedRate') or isNearlyEqual(machining.currentFeedRate, 0.0) else self.formatAsCoordinate(data=machining.currentFeedRate)),
            'AB="{ab}"'.format(ab=self.formatAsCoordinate(data=keywords['contourDistance'] if 'contourDistance' in keywords else 0.0)),
            'ZM="{z}"'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            self.__createCommonContourSecurityDistances(**keywords),
            'KM="{}"'.format(keywords['localizedComment']),
            self.createKATAndMNMCategories(10, 'Horizontal Fräsen'),
            '',
            '',
        )

    def createHorizontalContour(self, **keywords):
        contour = keywords['contour']

        approachType, approachDepartureDummy, approachFlying = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departureApproachDummy, departureType, departureFlying = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)

        keywords['localizedComment'] = self.getLocalizedKMComment(17, 'Horizontal Contour')
        keywords['approach'] = approachType
        keywords['departure'] = departureType
        keywords['flyingApproach'] = approachFlying
        keywords['radiusCompensation'] = self.getWopRadiusCompensation(**keywords)
        keywords['startVertex'] = 0
        keywords['endVertex'] = contour.segmentCount  # endVertex=contour.segmentCount is NOT -1, as we must count the start point too.
        keywords['securityDistance'] = self.CONTOUR_APPROACH_DEPARTURE_DISTANCE

        contourSecurityDistanceInfo = self.__createDefaultSecurityDistanceInfo(**keywords)
        contourSecurityDistanceInfo['isClosed'] = contour.isClosed()
        contourSecurityDistanceInfo['useApproach'] = contour.approach != 'without'
        contourSecurityDistanceInfo['useDeparture'] = contour.departure != 'without'

        keywords['contourSecurityDistanceInfo'] = contourSecurityDistanceInfo

        keywords['creator'] = self.createHorizontalCommonContourStartBlock
        keywords['zValueConverter'] = lambda _boardThickness, currentDepth: - currentDepth

        return '\n'.join(self.generateContourNoBuiltInDepthSteps(**keywords))

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContour(creator=self.createHorizontalContour, **keywords))

    def __getVertexIndexSequences(self, sections):  # FIXME IT-254: **keywords
        lastIndex = 0
        for section in sections:
            endIndex = lastIndex + len(section)
            yield lastIndex, endIndex
            lastIndex = endIndex

    def generateDadoSegments(self, **keywords):
        dado = keywords['dado']

        segments = super().generateDadoSegments(**keywords)

        sections = tuple(splitSectionsInXDistance(segments, maximumXDistance=self.MAXIMUM_SEGMENT_X_DISTANCE))

        keywords['dado'].vertexIndexSequences = tuple(self.__getVertexIndexSequences(sections))

        return tuple(itertools.chain.from_iterable(sections))

    def createDefinitionForOneDado(self, **keywords):
        dado = keywords['dado']
        if not dado.isVertical():
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * dado.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            dado.matrix = matrix  # TODO (HORIZONTAL) IT-229: unschönes Design -- NG
            dado.matrix.origin = positionOnPlane

            keywords['sideID'] = plane.sideID
            keywords['zValue'] = - dado.depth
        else:
            keywords['zValue'] = self.getDadoRabbetZValue(depth=keywords['machining'].depth, **keywords)

        return super().createDefinitionForOneDado(**keywords)

    def generateRabbetSegments(self, **keywords):
        sections = tuple(splitSectionsInXDistance(super().generateRabbetSegments(**keywords), maximumXDistance=self.MAXIMUM_SEGMENT_X_DISTANCE))

        keywords['rabbet'].vertexIndexSequences = tuple(self.__getVertexIndexSequences(sections))

        return tuple(itertools.chain.from_iterable(sections))

    def createDefinitionForOneRabbet(self, **keywords):
        return super().createDefinitionForOneRabbet(zValue=self.getDadoRabbetZValue(depth=keywords['machining'].depth, **keywords), **keywords)

    def createVerticalRabbet(self, **keywords):
        rabbet = keywords['rabbet']

        keywords['localizedComment'] = self.getLocalizedKMComment(7, 'Rabbet')
        if self.SIZING_ENTRY_OF_PROFILE == 'vertical flying':  # TODO IT-274 'vertical flying' as special approach for Nesting and override of dado() and rabbet(). Should it be separated?  --mjk
            keywords['flyingApproach'] = self.APPROACH_TYPE_FLYING
        else:
            keywords['flyingApproach'] = self.APPROACH_TYPE_NONE
        keywords['radiusCompensation'] = self.getWopRadiusCompensation(**keywords)
        keywords['securityDistance'] = self.getDadoRabbetSecurityDistance(**keywords)

        vertexIndexSequences = rabbet.vertexIndexSequences if hasattr(rabbet, 'vertexIndexSequences') else (0, rabbet.segmentCount)

        output = []

        for startVertex, endVertex in vertexIndexSequences:
            output += self.__createRabbetContourBlock(
                startVertex=startVertex,
                endVertex=endVertex,
                approach=self.TANGENTIAL_APPROACH if startVertex == 0 else self.QUARTER_ARC_APPROACH,  # QUARTER_ARC_APPROACH if MAXIMUM_SEGMENT_X_DISTANCE # 0.
                departure=self.TANGENTIAL_DEPARTURE if endVertex == rabbet.segmentCount else self.QUARTER_ARC_DEPARTURE,
                **keywords)

        return ''.join(output)

    def createVerticalDadoWithRouter(self, **keywords):
        dado = keywords['dado']

        keywords['localizedComment'] = self.getLocalizedKMComment(15, 'Dado')
        if dado.isInsetBoth() or self.SIZING_ENTRY_OF_PROFILE == 'vertical flying':  # TODO IT-274 'vertical flying' as special approach for Nesting and override of dado() and rabbet(). Should it be separated?  --mjk
            keywords['flyingApproach'] = self.APPROACH_TYPE_FLYING
        else:
            keywords['flyingApproach'] = self.APPROACH_TYPE_NONE
        keywords['radiusCompensation'] = self.getDadoRadiusCompensation(**keywords)
        keywords['securityDistance'] = self.getDadoRabbetSecurityDistance(**keywords)
        keywords['zValue'] = keywords['depth']

        vertexIndexSequences = dado.vertexIndexSequences if hasattr(dado, 'vertexIndexSequences') else (0, dado.segmentCount)

        output = []

        for startVertex, endVertex in vertexIndexSequences:
            output += self.__createDadoContourBlock(
                startVertex=startVertex,
                endVertex=endVertex,
                approach=self.TANGENTIAL_APPROACH,
                departure=self.TANGENTIAL_DEPARTURE,
                **keywords)

        return ''.join(output)

    def createHorizontalDadoWithRouter(self, **keywords):
        dado = keywords['dado']
        keywords['approach'] = self.TANGENTIAL_APPROACH
        keywords['departure'] = self.TANGENTIAL_DEPARTURE
        keywords['localizedComment'] = self.getLocalizedKMComment(15, 'Dado')
        keywords['flyingApproach'] = self.APPROACH_TYPE_FLYING if dado.isInsetBoth() else self.APPROACH_TYPE_NONE
        keywords['radiusCompensation'] = self.getDadoRadiusCompensation(**keywords)
        keywords['startVertex'] = 0
        keywords['endVertex'] = dado.segmentCount  # endVertex=contour.segmentCount is NOT -1, as we must count the start point too.
        keywords['securityDistance'] = self.getDadoRabbetSecurityDistance(**keywords)
        keywords['zValue'] = -dado.depth

        keywords['contourSecurityDistanceInfo'] = self.__createDefaultSecurityDistanceInfo(**keywords)

        return self.createHorizontalCommonContourStartBlock(**keywords)

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']
        machiningPlane = Plane.createFromMachining(machining)

        plane = self.machiningPlanes.findPlane(machiningPlane)
        assert plane is not None, "No plane found for machining"

        return plane, machining.getPosition3DOnPlane(plane)

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']
        # TODO IT-220 (HORIZONTAL): unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if contour.isVertical():
            zValue = self.getContourZValue(depth=keywords['contour'].depth, **keywords)
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            contour.path = contour.transformedPolyline(matrix.getInverted(), positionOnPlane)

            keywords['sideID'] = plane.sideID
            zValue = - contour.depth

        return super().createDefinitionForOneContour(zValue=zValue, **keywords)

    def createContourFirstCornerPolylineVertex(self, **keywords):
        def sizingIsContourCreator(**keywords):
            return 'contourCreator' in keywords and keywords['contourCreator'] == 'Sizing'

        machining = keywords['machining']
        vertex = keywords['startVertex']

        x, y = vertex.position.asTuple()
        z = keywords['zValue']

        return '\n'.join((
            '$E0',
            'KP {comment}{nest}'.format(comment=self.getLocalizedKMComment(8, 'Kontur'), nest=';NEST' if self.CREATE_SIZING_AS_NESTING_GEOMETRY and sizingIsContourCreator(**keywords) else ''),
            'X={xx}'.format(xx=self.formatAsCoordinate(data=x)),
            'Y={yy}'.format(yy=self.formatAsCoordinate(data=y)),
            'Z={zz}'.format(zz=self.formatAsCoordinate(data=z)),
            'KO={sideID}'.format(sideID=keywords['sideID'] if 'sideID' in keywords else self.MACHININGS_COORDINATE_SYSTEM),
        ))

    def createLineSegment(self, **keywords):
        machining = keywords['machining']
        segment = keywords['segment']

        x, y = segment.getEndVertex().position.asTuple()

        return joinIfNotNone('\n',
            '$E{}'.format(keywords['segmentID']),
            'KL',
            'X={xx}'.format(xx=self.formatAsCoordinate(data=x)),
            'Y={yy}'.format(yy=self.formatAsCoordinate(data=y)),
        )

    def createRadiusSegment(self, **keywords):
        WOODWOP_CW = 0
        WOODWOP_CCW = 1
        WOODWOP_CW_GT_180 = 2
        WOODWOP_CCW_GT_180 = 3

        machining = keywords['machining']
        segment = keywords['segment']

        directionCode = segment.getDirection(WOODWOP_CW, WOODWOP_CCW, WOODWOP_CW_GT_180, WOODWOP_CCW_GT_180)
        radius = segment.getRadius()

        x, y = segment.getEndVertex().position.asTuple()

        return joinIfNotNone('\n',
            '$E{}'.format(keywords['segmentID']),
            'KA',
            'X={xx}'.format(xx=self.formatAsCoordinate(data=x)),
            'Y={yy}'.format(yy=self.formatAsCoordinate(data=y)),
            'DS={}'.format(directionCode),
            'R={radius}'.format(radius=self.formatAsAngle(data=radius)),
        )

    def __createCommonStartBlock(self, **keywords):
        self.lastContourDefinitionIndex += 1
        keywords['machining'].wopContourDefinitionIndex = self.lastContourDefinitionIndex
        return '\n'.join((
            ']{index}'.format(index=self.lastContourDefinitionIndex),
            self.createContourFirstCornerPolylineVertex(**keywords)
        ))

    def createSizingStartBlock(self, **keywords):
        return self.__createCommonStartBlock(contourCreator='Sizing', **keywords)

    def createRabbetContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def createDadoContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def createContourStartBlock(self, **keywords):
        return self.__createCommonStartBlock(**keywords)

    def __createCommonEndBlock(self, **keywords):
        return '\n'

    def createSizingEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    def createRabbetContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    def createDadoContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    def createContourEndBlock(self, **keywords):
        return self.__createCommonEndBlock(**keywords)

    # Dado with saw
    def createVerticalDadoWithSaw(self, **keywords):
        dado = keywords['dado']
        segment = dado.getLineSegment()
        if not dado.isInsetStart() and dado.isInsetEnd():
            segment.increaseAtStart(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
        elif not dado.isInsetEnd() and dado.isInsetStart():
            segment.increaseAtEnd(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
        start, end = segment.getStartAndEndPoints()

        insetMode = 'MOD1' if dado.isInset() else 'MOD2'

        return joinIfNotNone('\n',
            '<109 \\{}\\'.format(self.getLocalizedMacroComment(9, 'Nuten')),
            'XA="{xa}"'.format(xa=self.formatAsCoordinate(data=start.x)),
            'YA="{ya}"'.format(ya=self.formatAsCoordinate(data=start.y)),
            'XE="{xe}"'.format(xe=self.formatAsCoordinate(data=end.x)),
            'YE="{ye}"'.format(ye=self.formatAsCoordinate(data=end.y)),
            'TI="{depth}"'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            'NB="{width}"'.format(width=self.formatAsCoordinate(data=dado.width)),
            'TV="{depth}"'.format(depth=self.DADO_PRESAW_DEPTH) if self.DADO_PRESAW_DEPTH != 0 else None,
            'XY="{}"'.format(self.DADO_PERCENTS_OF_SAW_WIDTH),
            'MN="{mode}"'.format(mode=self.DADO_MODE),
            'RK="{}"'.format(self.RADIUS_COMPENSATION_NONE),
            'EM="{mode}"'.format(mode=insetMode),
            'OP="1"',
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            'T_="{}"'.format(dado.toolID) if (dado.toolID and dado.toolID.lower() != 'auto') else None,
            self.createKATAndMNMCategories(6, 'Nuten'),
            '',
            '',
        )

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        return self.createVerticalDadoWithSaw(**keywords)

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        return self.createVerticalDadoWithSaw(**keywords)

    def createVerticalDadoWithSawCRotated(self, **keywords):
        return self.createVerticalDadoWithSaw(**keywords)

    # TODO IT-218 (HORIZONTAL): can this code be refactored to be common with createVerticalMiter -- NG
    def createHorizontalDadoWithSaw(self, **keywords):
        surface = keywords['surface']
        dado = keywords['dado']
        toolID = dado.toolID

        # TODO IT-218: duplicate code with createVerticalDadoWithSaw
        segment = dado.getLineSegment()
        if not dado.isInsetStart() and dado.isInsetEnd():
            segment.increaseAtStart(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
        elif not dado.isInsetEnd() and dado.isInsetStart():
            segment.increaseAtEnd(offset=self.SECURITY_DISTANCE_FOR_DADO_SAW)
        start, end = segment.getStartAndEndPoints()

        insetMode = 'MOD1' if dado.isInset() else 'MOD2'

        return '\n'.join((
            '<124 \\{}\\'.format(self.getLocalizedMacroComment(10, 'Nut_R')),
            'WI="{angle}"'.format(angle=self.formatAsAngle(data=0)),
            'WI2="{angle}"'.format(angle=self.formatAsAngle(data=0)),
            'XA="{xa}"'.format(xa=self.formatAsCoordinate(data=start.x)),
            'YA="{ya}"'.format(ya=self.formatAsCoordinate(data=start.y)),
            'XE="{xe}"'.format(xe=self.formatAsCoordinate(data=end.x)),
            'YE="{ye}"'.format(ye=self.formatAsCoordinate(data=end.y)),
            'ZA="{za}"'.format(za=self.formatAsCoordinate(data=dado.zPosition)),
            'RK="{}"'.format(self.RADIUS_COMPENSATION_LEFT),
            'EM="{mode}"'.format(mode=insetMode),
            'MV="GL"',
            'TI="{depth}"'.format(depth=self.formatAsCoordinate(data=dado.depth)),
            'TZ="{depth}"'.format(depth=self.formatAsCoordinate(data=0)),
            'XY="1"',
            'MN="GL"',
            'T_="{}"'.format(toolID),
            'HP="0"',
            'SP="0"',
            'YVE="0"',
            'ASG="2"',
            'MX="0"',
            'MY="0"',
            'MZ="0"',
            'MXF="1"',
            'MYF="1"',
            'MZF="1"',
            'XY2="{overlap}"'.format(overlap=self.formatAsCoordinate(data=85)),  # TODO IT-218: exists only from woodWOP 6 on -- NG
            'NB="{width}"'.format(width=self.formatAsCoordinate(data=dado.width)),  # TODO IT-218: exists only from woodWOP 6 on -- NG
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            'KM="{}"'.format(self.getLocalizedKMComment(18, 'Horizontal Dado')),
            self.createKATAndMNMCategories(7, 'Geschwenktes Nuten'),
            '',
            '',
        ))

    # Miter
    def createVerticalMiter(self, **keywords):
        surface = keywords['surface']
        miter = keywords['miter']
        toolID = miter.toolID

        start = miter.getStartPoint()
        end = miter.getEndPoint()

        preSawDepth = self.zValueMiterSawConverter.convertWithThickness(surface.thickness, self.MITER_PRE_SAW_DEPTH_FROM_TOP)
        sawDepth = self.zValueMiterSawConverter.convertWithThickness(surface.thickness, surface.thickness + self.MITER_SAW_DEPTH_FROM_BOTTOM)

        return '\n'.join((
            '<124 \\{}\\'.format(self.getLocalizedMacroComment(10, 'Nut_R')),
            'WI="{angle}"'.format(angle=self.formatAsAngle(data=miter.getAngle())),
            # TODO IT-218: define WI2 (angle in "newer" versions of woodWOP) -- NG
            'XA="{xa}"'.format(xa=self.formatAsCoordinate(data=start.x)),
            'YA="{ya}"'.format(ya=self.formatAsCoordinate(data=start.y)),
            'XE="{xe}"'.format(xe=self.formatAsCoordinate(data=end.x)),
            'YE="{ye}"'.format(ye=self.formatAsCoordinate(data=end.y)),
            'ZA="{}"'.format(self.VAR_THICKNESS),
            'RK="{}"'.format(self.RADIUS_COMPENSATION_LEFT),
            'EM="MOD2"',
            'MV="GL"',
            'Z_="{depth}"'.format(depth=self.formatAsCoordinate(data=sawDepth)),
            'VZ="{depth}"'.format(depth=self.formatAsCoordinate(data=preSawDepth)),
            'XY="1"',
            'MN="GGL"',
            'T_="{}"'.format(toolID),
            'HP="0"',
            'SP="0"',
            'YVE="0"',
            'ASG="2"',
            'MX="0"',
            'MY="0"',
            'MZ="0"',
            'MXF="1"',
            'MYF="1"',
            'MZF="1"',
            'KO="{}"'.format(self.MACHININGS_COORDINATE_SYSTEM),
            'KM="{}"'.format(self.getLocalizedKMComment(9, 'Miter')),
            self.createKATAndMNMCategories(7, 'Geschwenktes Nuten'),
            '',
            '',
        ))

    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCW)

    # NC Macro
    def getNCMacroCookedVariables(self, **keywords):
        variables = super().getNCMacroCookedVariables(**keywords)

        ncmacro = keywords['machining']
        if ncmacro.isVertical():
            sideID = self.DEFAULT_COORDINATE_SYSTEM
            sideX = ncmacro.position.x
            sideY = ncmacro.position.y
            sideZ = ncmacro.zPosition
            sideCAngle = variables['cangle']
            bAngle = 0
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)
            sideID = plane.sideID
            sideX = positionOnPlane.x
            sideY = positionOnPlane.y
            sideZ = positionOnPlane.z

            angle = VectorMath.angleBetween(ncmacro.matrix.u, plane.wopMatrix.u)

            if not isNearlyEqual(angle, 0.0):
                crossProduct = ncmacro.matrix.u.cross(plane.wopMatrix.u)
                if not isNearlyEqual(crossProduct.length(), 0.0) and crossProduct.normalized() == plane.wopMatrix.w:
                    angle = - angle

            sideCAngle = bAngle = self.formatter.formatAsAngle(normalizeAngle(angle))

        variables['sideid'] = sideID
        variables['sidex'] = self.formatter.formatAsCoordinate(sideX)
        variables['sidey'] = self.formatter.formatAsCoordinate(sideY)
        variables['sidez'] = self.formatter.formatAsCoordinate(sideZ)
        variables['sidecangle'] = sideCAngle
        variables['bangle'] = bAngle
        variables['coordinatesystem'] = self.MACHININGS_COORDINATE_SYSTEM

        return variables

    def createWoodWOPMacro(self, **keywords):
        parameters = keywords['parameters']
        wopMacro = parameters[0]
        parameters = parameters[1:]
        return '\n'.join(itertools.chain(
            (
                '<{id} \\{comment}\\'.format(id=wopMacro.name, comment=wopMacro.value),
            ),
            ('{recordName}="{recordValue}"'.format(recordName=parameter.name, recordValue=parameter.value) for parameter in parameters),
            (
                '',
                '',
            )
        ))

    def createWoodWOPComponent(self, **keywords):
        ncmacro = keywords['ncmacro']
        parameters = keywords['parameters']
        if parameters and 'XA' == parameters[0].name and 'YA' == parameters[1].name:
            xa = float(parameters[0].value)
            ya = float(parameters[1].value)
            parameters = parameters[2:]
        else:
            xa = ncmacro.position.x
            ya = ncmacro.position.y

        if parameters and 'ZA' == parameters[0].name:
            za = float(parameters[0].value)
            parameters = parameters[1:]
        else:
            za = ncmacro.zPosition

        return '\n'.join(itertools.chain(
            (
                '<139 \\Komponente\\',
                'IN="{name}"'.format(name=ncmacro.name),
                'XA="{x}"'.format(x=self.formatAsCoordinate(data=xa)),
                'YA="{y}"'.format(y=self.formatAsCoordinate(data=ya)),
                'ZA="{z}"'.format(z=self.formatAsCoordinate(data=za)),
            ),
            ('VA="{recordName} {recordValue}"'.format(recordName=parameter.name, recordValue=parameter.value) for parameter in parameters),
            (
                '',
                '',
            )
        ))

    def createNCMacroRecordFormat(self, **keywords):
        parameters = tuple(self.getNCMacroCookedParameters(**keywords))
        if parameters and all(map(str.isdigit, parameters[0].name)):
            return self.createWoodWOPMacro(parameters=parameters, **keywords)
        else:
            return self.createWoodWOPComponent(parameters=parameters, **keywords)

    # Footer
    def createEndOfFile(self, **keywords):
        return joinIfNotNone('\n',
            self.createVacuumSuctionCups(**keywords) if self.VACUUM_SUCTION_CUP_PROPOSAL else None,
            '!' + '\n'
        )

    # Helper
    def getLocalizedKMComment(self, index, default):
        return interiorcad.backend.localization.getLocalizedString(interiorcad.backend.localization.WOODWOP_KM_COMMENTS_RESOURCEGROUP, index, default)

    def getLocalizedVariable(self, index, defaultText):
        return interiorcad.backend.localization.getLocalizedString(interiorcad.backend.localization.WOODWOPPROGRAMVARIABLES_RESOURCEGROUP, index, defaultText)

    def getLocalizedMacroComment(self, index, defaultText):
        # WoodWOP's  macro comments are typically in German
        return interiorcad.backend.localization.getLocalizedString(interiorcad.backend.localization.WOODWOP_MACRO_COMMENTS_RESOURCEGROUP, index, defaultText)

    def getLocalizedKATCategories(self, index, defaultText):
        # WoodWOP's KAT categories are typically in German (The translated ones are in MNM)
        return interiorcad.backend.localization.getLocalizedString(interiorcad.backend.localization.WOODWOP_KAT_CATEGORIES_RESOURCEGROUP, index, defaultText)

    def getLocalizedMNMCategories(self, index, defaultText):
        # WoodWOP's MNM categories are typically translated (The untranslated ones are in KAT)
        return interiorcad.backend.localization.getLocalizedString(interiorcad.backend.localization.WOODWOP_MNM_CATEGORIES_RESOURCEGROUP, index, defaultText)
