# -*- coding: utf-8 -*-

# Copyright 2011-2024, extragroup GmbH, Pottkamp 19, 48149 Münster, Germany
# All rights reserved
#
# All information contained herein is, and remains the property of
# extragroup GmbH and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to extragroup GmbH
# and its suppliers and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is
# strictly forbidden unless prior written permission is obtained from
# extragroup GmbH

import collections
import itertools
import logging
import operator
import os
import sys

import interiorcad.backend.openservices
import interiorcad.backend.zvalue.constants as ZValueConstants
import interiorcad.interface.machineconfiguration.machineconfigurationkeys as MachineConfigurationKeys
from interiorcad.backend.datatracker.exporthelper import ExportHelper
from interiorcad.backend.localization import getLocalizedString, XILOGPROGRAMCOMMENTS_RESOURCEGROUP, WINXISO_MESSAGES_RESOURCEGROUP, XILOG_MESSAGES
from interiorcad.backend.postprocessors.surfacepostprocessor.drillingswithtoolidtofiveaxisdrillingsconverter import DrillingsWithToolIDToFiveAxisDrillingsConverter
from interiorcad.backend.utilities.stringconcatenation import joinIfNotNone
from interiorcad.interface.geometry import isNearlyEqual, isNearlyLesser
from interiorcad.interface.geometry.plane import Plane
from interiorcad.interface.geometry.vector import Vector
from interiorcad.interface.machiningsexporter.horizontaldrillingsexporter import HorizontalMachiningsExporter_SeparateCallsAndZValueConvertersForMachiningsOnBounds
from interiorcad.interface.ncexport.multisawdado import MultiSawDado
from interiorcad.interface.surface.acangles import normalizeAngle
from interiorcad.interface.surface.contour2d import Contour2D
from interiorcad.interface.surface.miterorienter import MiterOrienter
from interiorcad.interface.surface.origins import SURFACE_ORIGIN_BACK_LEFT, SURFACE_ORIGIN_FRONT_LEFT, SURFACE_ORIGIN_BACK_RIGHT, SURFACE_ORIGIN_FRONT_RIGHT
from interiorcad.ncexport.base import NCExportBase

# subprocess does not work on Mac under VWi
if sys.platform == "win32":
    import subprocess


XXL_EXTENSION = '.xxl'
PGM_EXTENSION = '.pgm'
INF_EXTENSION = '.inf'


class XilogPostProcessing:
    def __init__(self, filePath, machineconfiguration):
        assert sys.platform == "win32", "WinXISO works only under Windows!"
        self.originalFilePath = filePath
        self.winxisoPath = XilogPostProcessing.__getWinXIsoPath(machineconfiguration)

    @staticmethod
    def __getWinXIsoPath(machineconfiguration):
        winxiso = machineconfiguration.get('winxiso')
        if not winxiso:
            logging.info(getLocalizedString(WINXISO_MESSAGES_RESOURCEGROUP, 7, "The path to WinXISO in the configuration is empty, only XXL-files are created."))
        elif not os.path.exists(winxiso):
            logging.error(getLocalizedString(WINXISO_MESSAGES_RESOURCEGROUP, 4, "Could not find WinXISO according to configuration. Correct '{winxiso}'.").format(winxiso=winxiso))
            winxiso = ''
        return winxiso

    @staticmethod
    def getPGMFileNameOutOfXXLFileName(fileName, newSuffix):
        result = fileName
        if fileName.endswith(XXL_EXTENSION):
            result = fileName[:-4] + newSuffix
        return result

    def __callXilogPostProcessor(self):
        try:
            subprocess.check_call((
                self.winxisoPath,
                self.originalFilePath,
                '-i',
                '-s'
            ))
        except subprocess.CalledProcessError:
            message = getLocalizedString(WINXISO_MESSAGES_RESOURCEGROUP, 5, "WinXISO failed while processing file '{file}'.")
            logging.error(message.format(path=self.winxisoPath, file=self.originalFilePath))
        except OSError:
            message = getLocalizedString(WINXISO_MESSAGES_RESOURCEGROUP, 6, "Cannot start WinXISO at path '{path}'.")
            logging.error(message.format(path=self.winxisoPath))

    @staticmethod
    def __openFolderContaining(filePath):
        interiorcad.backend.openservices.openPath(os.path.dirname(filePath))

    def __checkIfPGMFileExistsOrRaiseException(self):
        pgmFilePath = XilogPostProcessing.getPGMFileNameOutOfXXLFileName(self.originalFilePath, PGM_EXTENSION)
        if not os.path.exists(pgmFilePath):
            XilogPostProcessing.__openFolderContaining(pgmFilePath)
            logging.error(getLocalizedString(WINXISO_MESSAGES_RESOURCEGROUP, 1, "Xilog WinXISO did not generate file '{}'").format(pgmFilePath))

    def __deleteNotNeededFiles(self):
        pgmFilePath = XilogPostProcessing.getPGMFileNameOutOfXXLFileName(self.originalFilePath, PGM_EXTENSION)
        if os.path.exists(pgmFilePath):
            try:
                os.remove(self.originalFilePath)
            except OSError:
                logging.info(getLocalizedString(WINXISO_MESSAGES_RESOURCEGROUP, 2, "The file '{}' could not be deleted.").format(self.originalFilePath))

        infFilePath = XilogPostProcessing.getPGMFileNameOutOfXXLFileName(self.originalFilePath, INF_EXTENSION)
        if os.path.exists(infFilePath):
            try:
                os.remove(infFilePath)
            except OSError:
                logging.info(getLocalizedString(WINXISO_MESSAGES_RESOURCEGROUP, 2, "The file '{}' could not be deleted.").format(infFilePath))

    def process(self):
        if self.winxisoPath:
            self.__callXilogPostProcessor()
            self.__checkIfPGMFileExistsOrRaiseException()
            self.__deleteNotNeededFiles()

    def getNewFilePath(self):
        if self.winxisoPath:
            return XilogPostProcessing.getPGMFileNameOutOfXXLFileName(self.originalFilePath, PGM_EXTENSION)
        else:
            return self.originalFilePath


class XilogExport(NCExportBase):
    EXPORT_FORMAT = 'Xilog'

    # Constants for entry side I (see Xilog documentation for XG0R)
    ENTRY_SIDE_PARALLEL_TO_X_AXIS_FIXED_ANGLE = 0
    ENTRY_SIDE_RIGHT_HAND_INTERPOLED_ANGLE = 1
    ENTRY_SIDE_LEFT_HAND_INTERPOLED_ANGLE = 2
    ENTRY_SIDE_PARALLEL_TO_PATH_FIXED_ANGLE = 3
    ENTRY_SIDE_PARALLEL_TO_PATH_INTERPOLED_ANGLE = 3

    MiterSawSpeeds = collections.namedtuple('MiterSawSpeeds', ('entry', 'presawing', 'sawing', 'exit'))
    MITER_SAW_SPEEDS = MiterSawSpeeds(entry=6, presawing=12, sawing=5, exit=10)
    SECURITY_DISTANCE_BEFORE_MITER = 100
    SECURITY_DISTANCE_AFTER_MITER = 125

    def __init__(self, **keywords):
        super().__init__(**keywords)

        machineConfiguration = self.getMachineConfiguration()
        self.CONFIGURATION_FILE = machineConfiguration.get('tool-configuration-file', 'def')

        # "wop-settings-..." are old deprecated names of the options
        self.MACHINE_OPTION_T = machineConfiguration.get('machine-options-t', machineConfiguration.get('wop-settings-t', 0))
        self.MACHINE_OPTION_V = machineConfiguration.get('machine-options-v', machineConfiguration.get('wop-settings-v', ''))

        self.MITER_PRE_SAW_DEPTH_FROM_TOP = machineConfiguration.getAsFloat(MachineConfigurationKeys.MITER_PRE_SAW_DEPTH_FROM_TOP_KEY, 2)
        self.ADDITIONAL_DEPTH_FOR_MITER = machineConfiguration.getAsFloat(MachineConfigurationKeys.ADDITIONAL_DEPTH_FOR_MITER_KEY, 5)
        self.MITER_SAW_REVERSE_ANGLE = machineConfiguration.get('miter-saw-reverse-angle', '')
        self.MITER_SAW_REVERSE_ENTRY_SIDE = machineConfiguration.getAsBool('miter-saw-reverse-entry-side', False)

        self.DADO_SAW_OVERLAP_FACTOR = machineConfiguration.getAsFloat('dado-saw-overlap-factor', 0.15)
        self.DADO_SAW_RADIUS_COMPENSATION_LEFT_RIGHT_FOR_RNT = machineConfiguration.getAsBool('dado-saw-radius-compensation-left-right-for-rnt', 0)
        self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET = machineConfiguration.getAsFloat('security-distance-factor-for-dado-rabbet', 2)
        self.SECURITY_DISTANCE_FACTOR_FOR_SIZING = machineConfiguration.getAsFloat('security-distance-factor-for-sizing', 2)
        self.SECURITY_DISTANCE_FACTOR_FOR_CONTOUR = machineConfiguration.getAsFloat('security-distance-factor-for-contour', 2)

        self.POCKET_MACRO = machineConfiguration.get('pocket-macro', 'KT')
        self.POCKET_SMALL_MACRO = machineConfiguration.get('pocket-small-macro', 'none')
        self.POCKET_PIERCING_MACRO = machineConfiguration.get('pocket-piercing-macro', 'KREIS')
        self.DADO_WITH_SAW_WIDTH_IDENTIFIER = machineConfiguration.get('dado-with-saw-width-identifier', 'L')

        self.MIRROR_IN_X = machineConfiguration.getAsBool('mirror-in-x', False)
        self.MIRROR_IN_Y = machineConfiguration.getAsBool('mirror-in-y', False)
        self.MIRROR_DIRECTION_OF_ROTATION = machineConfiguration.getAsBool('mirror-direction-of-rotation', False)
        self.MIRROR_RADIUS_CORRECTION = machineConfiguration.getAsBool('mirror-radius-correction', False)
        self.MIRROR_TOOL_CORRECTION = machineConfiguration.getAsBool('mirror-tool-correction', False)
        self.TOOL_RADIUS_COMPENSATION_SUPPORTS_DISTANCE_PARAMETER = machineConfiguration.getAsBool('tool-radius-compensation-supports-distance-parameter', False)

        self.createRadiusCompensation(**keywords)
        self.assignHorizontalFaces(**keywords)
        self.createZValueConverters(**keywords)

        self.MACHINING_AREA_FRONT_LEFT = machineConfiguration.get('machining-area-front-left', 'A')
        self.MACHINING_AREA_FRONT_RIGHT = machineConfiguration.get('machining-area-front-right', 'D')
        self.MACHINING_AREA_MINIMAL_COMBINED_LENGTH = machineConfiguration.getAsFloat('machining-area-minimal-combined-length', 1600)
        self.MACHINING_AREA_BACK_LEFT = machineConfiguration.get('machining-area-back-left', 'I')
        self.MACHINING_AREA_BACK_RIGHT = machineConfiguration.get('machining-area-back-right', 'L')
        self.MACHINING_AREA_MINIMAL_COMBINED_WIDTH = machineConfiguration.getAsFloat('machining-area-minimal-combined-width', 700)

        self.MACHINING_AREA_OFFSET_BY = machineConfiguration.getAsFloat('machining-area-offset-by', 0)
        self.MACHINING_AREA_OFFSET_BZ = machineConfiguration.getAsFloat('machining-area-offset-bz', 0)

        self.MACHINING_AREA_X = machineConfiguration.getAsFloat('machining-area-x', 0)
        self.WORKPIECE_COUNT_OVERWRITE = machineConfiguration.getAsBool('workpiece-count-overwrite', False)
        self.PARKING_POSITION_DISTANCE = machineConfiguration.get('parking-position-distance', '1000')

        self.ROTATION_ANGLE_CLOCKWISE = machineConfiguration.getAsBool('rotation-angle-clockwise', False)
        self.ROTATION_ANGLE_ADDITION = machineConfiguration.getAsFloat('rotation-angle-addition', 0)  # only relevant for machines with stops at the back and SX|SY # 0
        self.DRILLING_FEED_RATES = machineConfiguration.get('drilling-feed-rates', '')

        self.VACUUM_HOOD_POSITION = machineConfiguration.get('hood', '')

        self.SURFACE_ORIGIN = self.getSurfaceOrigin(surfaceOriginRaw=machineConfiguration.get('point-of-origin', 'front left'))

    def getSurfaceOrigin(self, **keywords):
        surfaceOriginRaw = keywords['surfaceOriginRaw'].lower()
        if surfaceOriginRaw and 'back' in surfaceOriginRaw and 'left' in surfaceOriginRaw:
            return SURFACE_ORIGIN_BACK_LEFT
        elif surfaceOriginRaw and 'front' in surfaceOriginRaw and 'left' in surfaceOriginRaw:
            return SURFACE_ORIGIN_FRONT_LEFT
        elif surfaceOriginRaw and 'back' in surfaceOriginRaw and 'right' in surfaceOriginRaw:
            return SURFACE_ORIGIN_BACK_RIGHT
        elif surfaceOriginRaw and 'front' in surfaceOriginRaw and 'right' in surfaceOriginRaw:
            return SURFACE_ORIGIN_FRONT_RIGHT
        else:
            return SURFACE_ORIGIN_FRONT_LEFT

    def createRadiusCompensation(self, **_keywords):
        # Constants for radius compensation C
        XilogExport.RADIUS_COMPENSATION_NONE = 0
        XilogExport.RADIUS_COMPENSATION_LENGTH = 3  # for inset dado with saw (RNT & RNTY)
        if not self.MIRROR_TOOL_CORRECTION:
            # Careful: the radius compensation 1 and 2 are opposite to DIN.
            XilogExport.RADIUS_COMPENSATION_LEFT = 2
            XilogExport.RADIUS_COMPENSATION_RIGHT = 1
        else:
            XilogExport.RADIUS_COMPENSATION_LEFT = 1
            XilogExport.RADIUS_COMPENSATION_RIGHT = 2

    def assignHorizontalFaces(self, **_keywords):
        self.SWITCH_HORIZONTAL_FACES = self.getMachineConfiguration().get('switch-horizontal-faces', 'switchNothing')
        if self.SWITCH_HORIZONTAL_FACES == 'switchAll' or self.SWITCH_HORIZONTAL_FACES == '1':
            self.SIDE_MAIN, self.SIDE_LEFT, self.SIDE_RIGHT, self.SIDE_BACK, self.SIDE_FRONT = range(1, 6)
        elif self.SWITCH_HORIZONTAL_FACES == 'switchLeftRight':
            self.SIDE_MAIN, self.SIDE_LEFT, self.SIDE_RIGHT, self.SIDE_FRONT, self.SIDE_BACK = range(1, 6)
        elif self.SWITCH_HORIZONTAL_FACES == 'switchFrontBack':
            self.SIDE_MAIN, self.SIDE_RIGHT, self.SIDE_LEFT, self.SIDE_BACK, self.SIDE_FRONT = range(1, 6)
        else:  # switchNothing or 0 (False)
            self.SIDE_MAIN, self.SIDE_RIGHT, self.SIDE_LEFT, self.SIDE_FRONT, self.SIDE_BACK = range(1, 6)

    def createZValueConverters(self, **_keywords):
        self.zValueVerticalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.createZValueHorizontalDrillingsConverters()
        self.zValueVerticalPocketConverter = self.createZValueConverter(MachineConfigurationKeys.VERTICAL_POCKET_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueFiveAxisDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.FIVE_AXIS_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMillingConverter = self.createZValueConverter(MachineConfigurationKeys.MILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueDadoSawConverter = self.createZValueConverter(MachineConfigurationKeys.DADO_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueMiterPreSawConverter = self.createZValueConverter(MachineConfigurationKeys.MITER_PRE_SAW_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)

    def createZValueHorizontalDrillingsConverters(self):
        zValueHorizontalDrillingConverter = self.createZValueConverter(MachineConfigurationKeys.HORIZONTAL_DRILLING_Z_REFERENCE_KEY, ZValueConstants.ORIGIN_ON_TOP, ZValueConstants.Z_DIRECTION_DOWN)
        self.zValueLeftAndRightSidesHorizontalDrillingConverter = zValueHorizontalDrillingConverter.createInverted() if self.MIRROR_IN_X else zValueHorizontalDrillingConverter
        self.zValueFrontAndBackSidesHorizontalDrillingConverter = zValueHorizontalDrillingConverter.createInverted() if self.MIRROR_IN_Y else zValueHorizontalDrillingConverter

    def createHead(self, **keywords):
        return joinIfNotNone('\n',
            self.createWorkpieceDefinition(**keywords),
            self.createWorkpieceComments(**keywords),
            self.createMirroringDefinitionForStartPointLeftBottom(enableMirroring=True, createAllParameters=True, **keywords),
            '',
            '',
        )

    def createMirroringDefinitionForStartPointLeftBottom(self, **keywords):
        enableMirroring = keywords['enableMirroring']  # True = enable | False = disable Mirroring
        createAllParameters = keywords['createAllParameters']  # True = create all Parameters | False = only used Parameters
        return joinIfNotNone('\n',
            '; {macro} {comment}'.format(macro=keywords['macro'], comment=self.getLocalizedComment(3, 'unmirror and remirror')) if ((self.MIRROR_IN_X or self.MIRROR_IN_Y) and not enableMirroring and 'macro' in keywords) else None,
            'SX={sx:d}'.format(sx=self.MIRROR_IN_X if enableMirroring else False) if (createAllParameters or self.MIRROR_IN_X) else None,
            'SY={sy:d}'.format(sy=self.MIRROR_IN_Y if enableMirroring else False) if (createAllParameters or self.MIRROR_IN_Y) else None,
        )

    def createSizingDefinitionForOnePolyline(self, **keywords):
        keywords['sizing'] = sizing = keywords['surface'].sizing

        creator = super().createSizingDefinitionForOnePolyline
        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)

        machiningDepth = sizing.actualDepth
        depthPerStep = self.getRouterDepthPerStep(machining=sizing, machiningDepth=machiningDepth, **keywords)

        return '\n'.join(self.generateCommonNoBuiltInDepthSteps(creator=creator, machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    def createSizingContour(self, **keywords):
        finishedSizeShape = self.getFinishedSizeShape(**keywords)
        if self.CREATE_SIZING_WITH_MACRO and finishedSizeShape.isRectangular():
            return self.createSizingWithMacroWithMirroringDefinitionNeutralization(**keywords)
        else:
            return self.createSizingContourDefinitionAsExecution(**keywords)

    def getPostAssignToolProcessors(self, **keywords):
        return DrillingsWithToolIDToFiveAxisDrillingsConverter(toolValidator=self.isValidToolId),

    def getHorizontalMachiningsExporter(self, **keywords):
        return HorizontalMachiningsExporter_SeparateCallsAndZValueConvertersForMachiningsOnBounds(self)

    def getNCFileExtension(self, **_keywords):
        return XXL_EXTENSION

    def getMachiningArea(self, **keywords):
        def getIndexedMachiningArea(area, index):
            return area if not len(area) > 1 else area[index]

        surface = keywords['surface']
        if surface.getFinishedSizeY() < self.MACHINING_AREA_MINIMAL_COMBINED_WIDTH or self.MACHINING_AREA_OFFSET_BY:
            if surface.getFinishedSizeX() < self.MACHINING_AREA_MINIMAL_COMBINED_LENGTH:
                machiningArea = self.MACHINING_AREA_FRONT_LEFT
            else:
                machiningArea = getIndexedMachiningArea(self.MACHINING_AREA_FRONT_LEFT, 0) + getIndexedMachiningArea(self.MACHINING_AREA_FRONT_RIGHT, -1)
        else:
            if surface.getFinishedSizeX() < self.MACHINING_AREA_MINIMAL_COMBINED_LENGTH:
                machiningArea = self.MACHINING_AREA_BACK_LEFT
            else:
                machiningArea = getIndexedMachiningArea(self.MACHINING_AREA_BACK_LEFT, 0) + getIndexedMachiningArea(self.MACHINING_AREA_BACK_RIGHT, -1)
        return machiningArea

    def createWorkpieceComments(self, **keywords):
        projectInfo = self.getProjectInfo()
        workpieceComments = self.getWorkpieceComments(**keywords)

        return joinIfNotNone('\n',
            '; {WorkpieceLabel}: {WorkpieceName}'.format_map(workpieceComments),
            '; {CommentLabel}: {Comment}'.format_map(workpieceComments) if workpieceComments['Comment'] else None,
            '; {ItemIDLabel}: {ItemID}'.format_map(workpieceComments),
            '; {Edge1Label}: {Edge1}'.format_map(workpieceComments) if workpieceComments['Edge1'] else None,  # front
            '; {Edge2Label}: {Edge2}'.format_map(workpieceComments) if workpieceComments['Edge2'] else None,  # back
            '; {Edge3Label}: {Edge3}'.format_map(workpieceComments) if workpieceComments['Edge3'] else None,  # left
            '; {Edge4Label}: {Edge4}'.format_map(workpieceComments) if workpieceComments['Edge4'] else None,  # right
            '; {CabinetNameLabel}: {CabinetName}'.format_map(workpieceComments) if workpieceComments['CabinetName'] else None,
            '; {ProjectIDLabel}: {ProjectID}, {ProjectCommission}, {ProjectConstruction}'.format_map(workpieceComments) if ExportHelper.isProjectIDValid(self, **projectInfo) else None,
            '; {PositionNumberLabel}: {PositionNumber} {PositionDescription}'.format_map(workpieceComments) if ExportHelper.isProjectIDValid(self, **projectInfo) else None,
        )

    def createWorkpieceDefinition(self, **keywords):
        surface = keywords['surface']
        sizingAdditionOffsetX = surface.getAdditionX() if not self.MIRROR_IN_X else surface.getAdditionXMirrored()
        sizingAdditionOffsetY = surface.getAdditionY() if not self.MIRROR_IN_Y else surface.getAdditionYMirrored()
        machiningAreaOffsetBY = self.MACHINING_AREA_OFFSET_BY if self.MACHINING_AREA_OFFSET_BY and surface.getFinishedSizeY() < self.MACHINING_AREA_MINIMAL_COMBINED_WIDTH else 0
        totalOffsetBY = sizingAdditionOffsetY + machiningAreaOffsetBY

        return joinIfNotNone(' ',
            'H',
            'DX={x}'.format(x=self.formatAsCoordinate(data=surface.getFinishedSizeX())),
            'DY={y}'.format(y=self.formatAsCoordinate(data=surface.getFinishedSizeY())),
            'DZ={z}'.format(z=self.formatAsCoordinate(data=surface.thickness)),
            '-{area}'.format(area=self.getMachiningArea(**keywords)),
            '/{configuration}'.format(configuration=self.CONFIGURATION_FILE),
            '*MM',
            'C={c}'.format(c=XilogExport.RADIUS_COMPENSATION_NONE),
            'T={t}'.format(t=self.MACHINE_OPTION_T),
            'R={r}'.format(r=surface.count if not self.WORKPIECE_COUNT_OVERWRITE else 9999),
            'BX={x}'.format(x=self.formatAsCoordinate(data=sizingAdditionOffsetX)),
            'BY={y}'.format(y=self.formatAsCoordinate(data=totalOffsetBY)),
            'BZ={z}'.format(z=self.formatAsCoordinate(data=self.MACHINING_AREA_OFFSET_BZ)),
            'V={v}'.format(v=self.MACHINE_OPTION_V) if self.MACHINE_OPTION_V else None,
            ';R original = {r}'.format(r=surface.count) if self.WORKPIECE_COUNT_OVERWRITE else None,
        )

    # machinings
    FLAT_DRILL_TYPE = 'P'
    PIERCING_DRILL_TYPE = 'L'

    def isValidToolId(self, **keywords):
        return 'tool' in keywords and keywords['tool'] and keywords['tool'].isToolIDValidIntDirect()

    def createHoodPositioning(self, **_keywords):
        if self.VACUUM_HOOD_POSITION:
            return 'E={e}'.format(e=self.VACUUM_HOOD_POSITION)

    def getDrillingFeedRates(self, **keywords):
        if not self.DRILLING_FEED_RATES:
            return None

        splitRates = self.DRILLING_FEED_RATES.split(',')
        drillingFeedRate = splitRates[0] if len(splitRates) == 1 else splitRates[0] if keywords['side'] == self.SIDE_MAIN else splitRates[1].lstrip()

        return f'V={drillingFeedRate}' if drillingFeedRate else None

    def createPerpendicularToSideDrillingCode(self, **keywords):
        # the meaning of x and y values depends on the side --- See documentation for "F" in Xilog help
        drillType = self.getDrillType(**keywords)
        repeat = keywords['repeat'] if 'repeat' in keywords else 1

        # tool on drillhead
        return ' '.join(
            itertools.chain(
                (
                    'XBO',
                    'X={x}'.format(x=self.formatAsCoordinate(data=keywords['x'])),
                    'Y={y}'.format(y=self.formatAsCoordinate(data=keywords['y'])),
                    'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['depth'])),
                    'D={diameter}'.format(diameter=self.formatAsCoordinate(data=keywords['installedDiameter'])),
                    'N={type}'.format(type=drillType),
                    'R={repeat}'.format(repeat=repeat),
                ),
                (self.getDrillingFeedRates(**keywords),) if self.getDrillingFeedRates(**keywords) else tuple(),
                keywords['extraParameters'] if 'extraParameters' in keywords else tuple(),
                (
                    'F={f}'.format(f=keywords['side']),
                ),
            )
        ) + '\n'

    def createCommonVerticalDrilling(self, **keywords):
        drilling = keywords['drilling']
        x = drilling.position.x
        y = drilling.position.y
        return self.createPerpendicularToSideDrillingCode(x=x, y=y, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter, side=self.SIDE_MAIN, **keywords)

    def createVerticalDrilling(self, **keywords):
        return self.createCommonVerticalDrilling(defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)

    def createVerticalDrillingPiercing(self, **keywords):
        return self.createCommonVerticalDrilling(defaultDrillType=self.PIERCING_DRILL_TYPE, **keywords)

    def createVerticalGridDrilling(self, **keywords):
        gridDrilling = keywords['gridDrilling']
        drilling = gridDrilling.getStartDrilling()
        deltaVector = gridDrilling.getDeltaVector()
        extraParameters = (
            'x={x}'.format(x=self.formatAsCoordinate(data=deltaVector.x)),
            'y={y}'.format(y=self.formatAsCoordinate(data=deltaVector.y)),
        )
        return self.createCommonVerticalDrilling(
            drilling=drilling,
            z=drilling.depth,
            repeat=gridDrilling.countDrillings,
            extraParameters=extraParameters,
            defaultDrillType=self.FLAT_DRILL_TYPE,
            **keywords
        )

    def createVerticalGridDrillingAlongXAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def createVerticalGridDrillingAlongYAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def createVerticalGridDrillingAlongCRotatedAxis(self, **keywords):
        return self.createVerticalGridDrilling(**keywords)

    def getDepthForDrilling(self, **keywords):
        # We use the ZValueConverter of vertical drillings, as it works the same way for each SCM and Morbidelli.
        return self.getDrillingZValue(depth=keywords['drilling'].depth, **keywords)

    def createFiveAxisDrilling(self, **keywords):
        depth = self.getDepthForDrilling(**keywords)
        return self.__createFiveAxisDrilling(depth=depth, **keywords)

    def getWopCAngle(self, **keywords):
        cAngle = keywords['cAngle']
        return normalizeAngle((- cAngle if self.ROTATION_ANGLE_CLOCKWISE else cAngle) + self.ROTATION_ANGLE_ADDITION)

    def __createFiveAxisDrilling(self, **keywords):
        drilling = keywords['drilling']

        depth = keywords['depth']
        return ' '.join((
            'XBR',
            'X={x}'.format(x=self.formatAsCoordinate(data=drilling.position.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=drilling.position.y)),
            'Z={depth}'.format(depth=self.formatAsCoordinate(data=depth)),
            'A={angle}#'.format(angle=self.formatAsAngle(data=self.getWopCAngle(cAngle=drilling.cAngle, **keywords))),
            'H={z}'.format(z=self.formatAsCoordinate(data=keywords['zPosition'])),
            'Q=0#',
            'B={angle}#'.format(angle=self.formatAsAngle(data=drilling.aAngle)),
            'T={toolID}'.format(toolID=drilling.toolID),
            'F={side}'.format(side=self.SIDE_MAIN),
            ';Ø{dm} x {depth}'.format(dm=self.formatAsCoordinate(data=drilling.installedDiameter), depth=self.formatAsCoordinate(data=abs(depth))),
        )) + '\n'

    def createNCStop(self, **keywords):
        return 'XMSG N="{comment}" Q=1\n'.format(comment=keywords['comment'])

    def createRotatedHorizontalCenterPunch(self, **keywords):
        # This function is only called when the horizontal driller is allowed to rotate (otherwise it is handled by the filter DrillingsWithToolIDToFiveAxisDrillingsConverter)
        return self.createFiveAxisCenterPunch(**keywords)

    def createRotatedHorizontalDrilling(self, **keywords):
        # This function is only called when the horizontal driller is allowed to rotate (otherwise it is handled by the filter DrillingsWithToolIDToFiveAxisDrillingsConverter)
        return self.createFiveAxisDrilling(**keywords)

    def getLeftAndRightSidesHorizontalDrillingZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        zPosition = keywords['zPosition']
        return self.zValueLeftAndRightSidesHorizontalDrillingConverter.convertWithThickness(boardThickness, zPosition)

    def getFrontAndBackSidesHorizontalDrillingZValue(self, **keywords):
        boardThickness = keywords['surface'].thickness
        zPosition = keywords['zPosition']
        return self.zValueFrontAndBackSidesHorizontalDrillingConverter.convertWithThickness(boardThickness, zPosition)

    def getHorizontalDrillingZValue(self, **keywords):
        return self.getFiveAxisDrillingZValue(**keywords)

    def createHorizontalDrillingFromRight(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        depth = self.getDepthForDrilling(**keywords)
        return self.createPerpendicularToSideDrillingCode(side=self.SIDE_RIGHT, x=zPosition, y=drilling.position.y, depth=depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter, defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)

    def createHorizontalDrillingFromLeft(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        depth = self.getDepthForDrilling(**keywords)
        return self.createPerpendicularToSideDrillingCode(side=self.SIDE_LEFT, x=zPosition, y=drilling.position.y, depth=depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter, defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)

    def createHorizontalDrillingFromBack(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        depth = self.getDepthForDrilling(**keywords)
        return self.createPerpendicularToSideDrillingCode(side=self.SIDE_BACK, x=drilling.position.x, y=zPosition, depth=depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter, defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)

    def createHorizontalDrillingFromFront(self, **keywords):
        zPosition = keywords['zPosition']
        drilling = keywords['drilling']
        depth = self.getDepthForDrilling(**keywords)
        return self.createPerpendicularToSideDrillingCode(side=self.SIDE_FRONT, x=drilling.position.x, y=zPosition, depth=depth, diameter=drilling.diameter, installedDiameter=drilling.installedDiameter, defaultDrillType=self.FLAT_DRILL_TYPE, **keywords)

    def __calculatePlane(self, **keywords):
        machining = keywords['machining']
        machiningPlane = Plane.createFromMachining(machining)

        if not self.isStandardOrigin(**keywords):
            machiningPlane.matrix.v = - machiningPlane.matrix.v

        return machiningPlane, machining.getPositionOnPlane(machiningPlane)

    def createPlane(self, **keywords):
        plane = keywords['plane']
        cAngle = self.getWopCAngle(cAngle=plane.cAngleToXYPlane - 90, **keywords)

        return joinIfNotNone(' ',
            'XPL',
            'X={x}'.format(x=self.formatAsCoordinate(data=plane.origin.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=plane.origin.y)),
            'Z={z}'.format(z=self.formatAsCoordinate(data=plane.origin.z)),
            'Q={cangle}'.format(cangle=self.formatAsCoordinate(data=cAngle)),
            'R={aangle}'.format(aangle=self.formatAsCoordinate(data=plane.aAngle)),
        )

    def createPlaneReset(self, **_keywords):
        return joinIfNotNone(' ',
            'XPL',
            'X={x}'.format(x=self.formatAsCoordinate(data=0)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=0)),
            'Z={z}'.format(z=self.formatAsCoordinate(data=0)),
            'Q={angle}'.format(angle=self.formatAsCoordinate(data=0)),
            'R={angle}'.format(angle=self.formatAsCoordinate(data=0)),
        )

    def createContourFirstCornerPolygonPoint(self, **keywords):
        machining = keywords['machining']
        toolID = machining.toolID
        vertex = keywords['startVertex']

        return joinIfNotNone(' ',
            'G0',
            'X={x}'.format(x=self.formatAsCoordinate(data=vertex.position.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=vertex.position.y)),
            'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['zValue'])),
            self.createHoodPositioning(**keywords),
            'T={toolID}'.format(toolID=toolID),
        )

    def createLineSegment(self, **keywords):
        segment = keywords['segment']

        position = segment.getEndVertex().position
        return joinIfNotNone(' ',
            'G1',
            'X={x}'.format(x=self.formatAsCoordinate(data=position.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=position.y)),
            'V={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) and keywords['segmentID'] == 1 else None,
        )

    def createRadiusSegment(self, **keywords):
        if operator.xor(self.MIRROR_DIRECTION_OF_ROTATION, not self.isStandardOrigin(**keywords)):
            XILOG_G_CW = 'G3'
            XILOG_G_CCW = 'G2'
        else:  # default
            XILOG_G_CW = 'G2'
            XILOG_G_CCW = 'G3'

        if self.MIRROR_RADIUS_CORRECTION:
            XILOG_CW = (1, XILOG_G_CW)
            XILOG_CCW = (-1, XILOG_G_CCW)
            XILOG_CW180 = (-1, XILOG_G_CW)
            XILOG_CCW180 = (1, XILOG_G_CCW)
        else:
            XILOG_CW = (-1, XILOG_G_CW)
            XILOG_CCW = (1, XILOG_G_CCW)
            XILOG_CW180 = (1, XILOG_G_CW)
            XILOG_CCW180 = (-1, XILOG_G_CCW)

        segment = keywords['segment']
        multiplicationFactor, directionCode = segment.getDirection(XILOG_CW, XILOG_CCW, XILOG_CW180, XILOG_CCW180)
        radius = segment.getRadius() * multiplicationFactor

        position = segment.getEndVertex().position
        return joinIfNotNone(' ',
            directionCode,
            'X={x}'.format(x=self.formatAsCoordinate(data=position.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=position.y)),
            'r={r}'.format(r=self.formatAsAngle(data=radius)),
            'V={feedRate}'.format(feedRate=self.formatAsCoordinate(data=keywords['feedRate'])) if 'feedRate' in keywords and not isNearlyEqual(keywords['feedRate'], 0.0) and keywords['segmentID'] == 1 else None,
        )

    def createToolRadiusCompensation(self, **keywords):
        if self.TOOL_RADIUS_COMPENSATION_SUPPORTS_DISTANCE_PARAMETER:
            distance = self.formatAsCoordinate(data=keywords['contourDistance'] if 'contourDistance' in keywords else 0.0)
        else:
            distance = '0.00'

        return 'C={c} S={distance}'.format(c=keywords['radiusCompensation'], distance=distance)

    def __createContourApproachOrDeparture(self, **keywords):
        return ' '.join((
            keywords['macro'],
            'G={approach:s}'.format(approach=keywords['approach']),
            'R={multiplier}'.format(multiplier=keywords['approachMultiplier']),
            'Q={onset}'.format(onset=keywords['onset']),
        ))

    def createContourApproach(self, **keywords):
        if keywords['machining'].isVertical():
            return self.__createContourApproachOrDeparture(macro='GIN', **keywords)

    def createContourDeparture(self, **keywords):
        if keywords['machining'].isVertical():
            return self.__createContourApproachOrDeparture(macro='GOUT', **keywords)

    def __getSizingRadiusCompensation(self, **keywords):
        return XilogExport.RADIUS_COMPENSATION_RIGHT if self.isSizingMillingDirectionLeft(**keywords) else XilogExport.RADIUS_COMPENSATION_LEFT

    def getWopRadiusCompensation(self, **keywords):
        machining = keywords['machining']
        return Contour2D.getRadiusCompensation(machining.radiusCompensation, none=XilogExport.RADIUS_COMPENSATION_NONE, left=XilogExport.RADIUS_COMPENSATION_LEFT, right=XilogExport.RADIUS_COMPENSATION_RIGHT)

    # Onset and Approaches, see Xilog documentation for GIN and GOUT
    ONSET_IN_QUOTA = '0#'
    ONSET_IN_DESCENT = '1#'

    LINEAR_APPROACH = '1#'
    QUARTER_ARC_APPROACH = '2#'
    LINEAR_FLYING_APPROACH = '3#'  # not documented in normal Xilog Plus documentation
    QUARTER_ARC_FLYING_APPROACH = '4#'  # not documented in normal Xilog Plus documentation

    APPROACH_MODE_LINEAR_PIERCING = (LINEAR_APPROACH, ONSET_IN_QUOTA, True)
    APPROACH_MODE_LINEAR_FLYING = (LINEAR_APPROACH, ONSET_IN_DESCENT, True)
    APPROACH_MODE_ARC_PIERCING = (QUARTER_ARC_APPROACH, ONSET_IN_QUOTA, True)
    APPROACH_MODE_ARC_FLYING = (QUARTER_ARC_APPROACH, ONSET_IN_DESCENT, True)
    APPROACH_MODE_WITHOUT = (LINEAR_APPROACH, ONSET_IN_DESCENT, False)
    APPROACH_MODE_SIZING_DEFAULT = APPROACH_MODE_ARC_PIERCING
    APPROACH_MODE_CONTOUR_DEFAULT = APPROACH_MODE_WITHOUT

    def __createCommonStartBlock(self, **keywords):
        return joinIfNotNone('\n',
            self.createToolRadiusCompensation(**keywords),
            self.createContourApproach(**keywords),
            self.createContourFirstCornerPolygonPoint(**keywords),
        )

    def createSizingWithMacroWithMirroringDefinitionNeutralization(self, **keywords):
        surface = keywords['surface']
        if surface.hasSizing():
            return joinIfNotNone('\n',
                self.createMirroringDefinitionForStartPointLeftBottom(enableMirroring=False, createAllParameters=False, macro='FORMAT', **keywords),
                self.createSizingWithMacro(**keywords),
                self.createMirroringDefinitionForStartPointLeftBottom(enableMirroring=True, createAllParameters=False, **keywords),
            ) + '\n'

    def createSizingWithMacro(self, **keywords):
        surface = keywords['surface']
        zValue = self.getSizingZValue(**keywords)
        return ' '.join((
            'FORMAT',
            'Z={z}'.format(z=self.formatAsCoordinate(data=zValue)),  # 'DZ+{z}' or '-DZ-{z}'.format(z=self.SIZING_DEPTH_ADJUSTMENT) would be better, but needs deeper implementation of zValueConverter
            'R=0.01',
            'L={direction:d}'.format(direction=not self.isSizingMillingDirectionLeft(sizing=surface.sizing, **keywords)),
            'T={toolID}'.format(toolID=surface.sizing.toolID),
        ))

    def formatWithHash(self, **keywords):
        return "{}#".format(self.formatAsCoordinate(**keywords))  # formatAsCoordinate uses data

    def createSizingStartBlock(self, **keywords):
        radiusCompensation = self.__getSizingRadiusCompensation(**keywords)
        approach, onset, _dummyUseApproachModifier = self.getSizingApproachMode(**keywords)
        return self.__createCommonStartBlock(radiusCompensation=radiusCompensation, approach=approach, onset=onset, approachMultiplier=self.formatWithHash(data=self.SECURITY_DISTANCE_FACTOR_FOR_SIZING), **keywords)

    def createDadoContourStartBlock(self, **keywords):
        approach, onset, _dummyUseApproachModifier = self.APPROACH_MODE_LINEAR_FLYING if keywords['dado'].isInsetBoth() else self.APPROACH_MODE_LINEAR_PIERCING
        return self.__createCommonStartBlock(approach=approach, onset=onset, approachMultiplier=self.formatWithHash(data=self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET), **keywords)

    def createRabbetContourStartBlock(self, **keywords):
        radiusCompensation = self.getWopRadiusCompensation(**keywords)
        approach, onset, _dummyUseApproachModifier = self.APPROACH_MODE_LINEAR_PIERCING
        return self.__createCommonStartBlock(radiusCompensation=radiusCompensation, approach=approach, onset=onset, approachMultiplier=self.formatWithHash(data=self.SECURITY_DISTANCE_FACTOR_FOR_DADO_RABBET), **keywords)

    def createContourStartBlock(self, **keywords):
        contour = keywords['contour']

        approachType, approachOnset, useApproachModifier = self.getApproachMode(approachDeparture=contour.approach, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        approachMultiplier = self.formatWithHash(data=self.SECURITY_DISTANCE_FACTOR_FOR_CONTOUR)

        if not useApproachModifier:
            # Compatibility to behaviour up to interiorcad 2018 F4
            approachMultiplier = self.formatWithHash(data=0)

        radiusCompensation = self.getWopRadiusCompensation(**keywords)

        return self.__createCommonStartBlock(radiusCompensation=radiusCompensation, approach=approachType, onset=approachOnset, approachMultiplier=approachMultiplier, **keywords)

    def __createCommonEndBlock(self, **keywords):
        return joinIfNotNone('\n',
            self.createContourDeparture(**keywords),
            ''
        )

    def createSizingEndBlock(self, **keywords):
        approach, onset, _dummyUseApproachModifier = self.getSizingApproachMode(**keywords)
        return self.__createCommonEndBlock(approach=approach, approachMultiplier="2#", onset=onset, **keywords)

    def createDadoContourEndBlock(self, **keywords):
        approach, onset, _dummyUseApproachModifier = self.APPROACH_MODE_LINEAR_FLYING if keywords['dado'].isInsetBoth() else self.APPROACH_MODE_LINEAR_PIERCING
        return self.__createCommonEndBlock(approach=approach, onset=onset, approachMultiplier="2#", **keywords)

    def createRabbetContourEndBlock(self, **keywords):
        approach, onset, _dummyUseApproachModifier = self.APPROACH_MODE_LINEAR_PIERCING
        return self.__createCommonEndBlock(approach=approach, onset=onset, approachMultiplier="2#", **keywords)

    def createContourEndBlock(self, **keywords):
        contour = keywords['contour']

        departureType, departureOnset, useApproachModifier = self.getApproachMode(approachDeparture=contour.departure, defaultApproachMode=self.APPROACH_MODE_CONTOUR_DEFAULT, **keywords)
        departureMultiplier = self.formatWithHash(data=self.SECURITY_DISTANCE_FACTOR_FOR_CONTOUR)

        if not useApproachModifier:
            # Similar to the behaviour of approach up to interiorcad 2018 F4
            departureMultiplier = self.formatWithHash(data=0)

        return self.__createCommonEndBlock(approach=departureType, onset=departureOnset, approachMultiplier=departureMultiplier, **keywords)

    def createDefinitionForOneContour(self, **keywords):
        contour = keywords['contour']
        # TODO (HORIZONTAL): IT-224/IT-231 unschön, dass man diese Aufrufe (nicht nur in der Funktion) haben muss und damit fast so wie 2 Konturzug-System hat: einmal vertikal und einmal horizontal -- NG
        if contour.isVertical():
            plane = None  # TODO (HORIZONTAL) IT-231: can we be more precise? -- NG
        else:
            plane, positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * contour.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            contour.path = contour.transformedPolyline(matrix.getInverted(), positionOnPlane)

        return joinIfNotNone('\n',
            self.createPlane(plane=plane, **keywords) if plane else None,
            super().createDefinitionForOneContour(feedRate=contour.currentFeedRate, **keywords),
            self.createPlaneReset(**keywords) if plane else None,
            '' if plane else None,
        )

    def createVerticalContour(self, **keywords):
        keywords['creator'] = self.createDefinitionForOneContour

        return '\n'.join(self.generateContourNoBuiltInDepthSteps(
            zValueConverter=lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth),
            **keywords))

    def createVerticalMultiUsageContour(self, **keywords):
        if self.TOOL_RADIUS_COMPENSATION_SUPPORTS_DISTANCE_PARAMETER:
            return '\n'.join(self.generateMultiUsageContour(creator=self.createVerticalContour, **keywords))
        else:
            return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createVerticalContour, **keywords))

    def createHorizontalContour(self, **keywords):
        keywords['creator'] = self.createDefinitionForOneContour

        return '\n'.join(self.generateContourNoBuiltInDepthSteps(
            zValueConverter=lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth),
            **keywords))

    def createHorizontalMultiUsageContour(self, **keywords):
        return '\n'.join(self.generateMultiUsageContourNoBuiltInDistanceSupport(creator=self.createHorizontalContour, **keywords))

    def getVerticalPocketMacroFromConfiguration(self, **keywords):
        pocket = keywords['pocket']
        if self.POCKET_SMALL_MACRO != 'none' and isNearlyLesser(pocket.diameter, 2 * pocket.tool.getRouterDiameter()):
            return self.POCKET_SMALL_MACRO
        else:
            return self.POCKET_MACRO

    def createVerticalPocket(self, **keywords):
        return self.createVerticalPocketCodeWithMirroringDefinitionNeutralization(macro=self.getVerticalPocketMacroFromConfiguration(**keywords), **keywords)

    def createVerticalPocketPiercing(self, **keywords):
        if keywords['pocket'].diameter >= self.VERTICAL_POCKET_PIERCING_CLEAR_UP_TO_DIAMETER:
            macro = self.POCKET_PIERCING_MACRO
        else:
            macro = self.getVerticalPocketMacroFromConfiguration(**keywords)
        return self.createVerticalPocketCodeWithMirroringDefinitionNeutralization(macro=macro, **keywords)

    def createVerticalPocketCodeWithMirroringDefinitionNeutralization(self, **keywords):
        return joinIfNotNone('\n',
            self.createMirroringDefinitionForStartPointLeftBottom(enableMirroring=False, createAllParameters=False, **keywords),
            self.createVerticalPocketCode(**keywords),
            self.createMirroringDefinitionForStartPointLeftBottom(enableMirroring=True, createAllParameters=False, **keywords),
        ) + '\n'

    def createVerticalPocketCode(self, **keywords):
        pocket = keywords['pocket']
        macro = keywords['macro']
        return joinIfNotNone(' ',
            '{macro}'.format(macro=macro),
            'X={dx}{x}'.format(dx='DX-' if self.MIRROR_IN_X else '', x=self.formatAsCoordinate(data=pocket.position.x)),
            'Y={dy}{y}'.format(dy='DY-' if self.MIRROR_IN_Y else '', y=self.formatAsCoordinate(data=pocket.position.y)),
            'Z={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            self.createHoodPositioning(**keywords),
            'D={diameter}'.format(diameter=self.formatAsCoordinate(data=pocket.diameter)),
            'G=0' if macro != 'KT' else None,
            'T={toolID}'.format(toolID=pocket.toolID),
        )

    def postProcessFile(self, **keywords):
        filePath = keywords['filePath']
        newFilePath = filePath
        if sys.platform == "win32":
            postProcessor = XilogPostProcessing(filePath, self.getMachineConfiguration())
            postProcessor.process()
            newFilePath = postProcessor.getNewFilePath()
        return newFilePath

    # dado
    def getRadiusCompensationOfVerticalDadoWithSaw(self, **keywords):
        return int(''.join((
            str(keywords['radiusCompensation'] if 'radiusCompensation' in keywords else ''),  # radiusCompensation only exists in  createSingleVerticalSawCutCRotated
            str(XilogExport.RADIUS_COMPENSATION_LENGTH if keywords['dado'].isInset() else XilogExport.RADIUS_COMPENSATION_NONE if 'radiusCompensation' not in keywords else ''),
        ))
        )

    def createVerticalDadoWithSawAlongXAxis(self, **keywords):
        def __getYforRNT(y, width):
            return y - width / 2

        if self.DADO_SAW_RADIUS_COMPENSATION_LEFT_RIGHT_FOR_RNT:
            if self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT:
                keywords['radiusCompensation'] = XilogExport.RADIUS_COMPENSATION_LEFT
            else:
                keywords['radiusCompensation'] = XilogExport.RADIUS_COMPENSATION_RIGHT

        dado = keywords['dado']
        start, end = self.getDadoLineSegmentWithSecurityDistance(**keywords).getStartAndEndPoints()
        width = dado.getWidth()

        if self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT:
            startX = min(start.x, end.x)
            endX = max(start.x, end.x)
        else:
            startX = max(start.x, end.x)
            endX = min(start.x, end.x)

        return joinIfNotNone(' ',
            'RNT',
            'X={x}'.format(x=self.formatAsCoordinate(data=startX)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=__getYforRNT(start.y, width))),
            'Z={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            self.createHoodPositioning(**keywords),
            'x={x}'.format(x=self.formatAsCoordinate(data=endX)),
            'C={c}'.format(c=self.getRadiusCompensationOfVerticalDadoWithSaw(**keywords)),
            '{identifier}={width}'.format(identifier=self.DADO_WITH_SAW_WIDTH_IDENTIFIER, width=self.formatAsCoordinate(data=width)),
            'T={toolID}'.format(toolID=dado.toolID),
        ) + '\n'

    def createVerticalDadoWithSawAlongYAxis(self, **keywords):
        def __getXforRNTY(x, width):
            return x - width / 2

        if self.DADO_SAW_RADIUS_COMPENSATION_LEFT_RIGHT_FOR_RNT:
            if self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_RIGHT:
                keywords['radiusCompensation'] = XilogExport.RADIUS_COMPENSATION_RIGHT
            else:
                keywords['radiusCompensation'] = XilogExport.RADIUS_COMPENSATION_LEFT

        dado = keywords['dado']
        start, end = self.getDadoLineSegmentWithSecurityDistance(**keywords).getStartAndEndPoints()
        width = dado.getWidth()

        if self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_FRONT_RIGHT:
            startY = min(start.y, end.y)
            endY = max(start.y, end.y)
        else:
            startY = max(start.y, end.y)
            endY = min(start.y, end.y)

        return joinIfNotNone(' ',
            'RNTY',
            'X={x}'.format(x=self.formatAsCoordinate(data=__getXforRNTY(start.x, width))),
            'Y={y}'.format(y=self.formatAsCoordinate(data=startY)),
            'Z={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            self.createHoodPositioning(**keywords),
            'y={y}'.format(y=self.formatAsCoordinate(data=endY)),
            'C={c}'.format(c=self.getRadiusCompensationOfVerticalDadoWithSaw(**keywords)),
            '{identifier}={width}'.format(identifier=self.DADO_WITH_SAW_WIDTH_IDENTIFIER, width=self.formatAsCoordinate(data=width)),
            'T={toolID}'.format(toolID=dado.toolID),
        ) + '\n'

    def __createSingleVerticalSawCutCRotated(self, **keywords):
        dado = keywords['dado']
        start, end = keywords['segment'].getStartAndEndPoints()
        return joinIfNotNone(' ',
            'NT',
            'X={x}'.format(x=self.formatAsCoordinate(data=start.x)),
            'Y={y}'.format(y=self.formatAsCoordinate(data=start.y)),
            'Z={depth}'.format(depth=self.formatAsCoordinate(data=keywords['depth'])),
            self.createHoodPositioning(**keywords),
            'x={x}'.format(x=self.formatAsCoordinate(data=end.x)),
            'y={y}'.format(y=self.formatAsCoordinate(data=end.y)),
            'C={c}'.format(c=self.getRadiusCompensationOfVerticalDadoWithSaw(**keywords)),
            'T={toolID}'.format(toolID=dado.toolID),
        )

    def createVerticalDadoWithSawCRotated(self, **keywords):
        multiSawDado = MultiSawDado(self)
        return multiSawDado.createMultipleVerticalSawCuts(createFunction=self.__createSingleVerticalSawCutCRotated, **keywords) + '\n'

    def generateDadoSegments(self, **keywords):
        xoredCondition = not self.isStandardOrigin(**keywords)
        return super().generateDadoSegments(xoredCondition=xoredCondition, **keywords)

    def createVerticalDadoWithRouter(self, **keywords):
        radiusCompensation = self.getDadoRadiusCompensation(**keywords)

        machiningDepth = keywords['dado'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return '\n'.join(
            self.generateCommonNoBuiltInDepthSteps(
                creator=self.createContourFromSegmentList,
                machiningDepth=machiningDepth,
                depthPerStep=depthPerStep,
                radiusCompensation=radiusCompensation,
                segmentList=self.generateDadoSegments(**keywords),
                createStartBlock=self.createDadoContourStartBlock,
                createEndBlock=self.createDadoContourEndBlock,
                zValueConverter=lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth),
                **keywords))

    def createHorizontalDadoWithRouter(self, **keywords):
        dado = keywords['dado']
        if dado.isVertical():
            plane = None  # TODO (HORIZONTAL) IT-231: can we be more precise? -- NG
        else:
            plane, _positionOnPlane = self.__calculatePlane(**keywords)

            matrix = plane.matrix * dado.matrix.getInverted()
            assert matrix.w == Vector(0, 0, 1), "Matrix must be 2D!"
            dado.matrix = matrix  # TODO (HORIZONTAL) IT-231: unschönes Design -- NG

        return joinIfNotNone('\n',
            self.createPlane(plane=plane, **keywords) if plane else None,
            self.createVerticalDadoWithRouter(**keywords),
            self.createPlaneReset(**keywords) if plane else None,
            '' if plane else None,
        )

    def createVerticalRabbet(self, **keywords):
        if self.TOOL_RADIUS_COMPENSATION_SUPPORTS_DISTANCE_PARAMETER:
            creator = self.createDefinitionForOneRabbet
        else:
            creator = self.createVerticalRabbetOffsetByContourDistance

        keywords['zValueConverter'] = lambda boardThickness, currentDepth: self.zValueMillingConverter.convertWithThickness(boardThickness, currentDepth)
        machiningDepth = keywords['rabbet'].actualDepth
        depthPerStep = self.getRouterDepthPerStep(machiningDepth=machiningDepth, **keywords)

        return '\n'.join(self.generateCommonNoBuiltInDepthSteps(creator=creator, machiningDepth=machiningDepth, depthPerStep=depthPerStep, **keywords))

    def getPreSawDepth(self, **keywords):
        return self.zValueMiterPreSawConverter.convertWithThickness(keywords['boardThickness'], self.MITER_PRE_SAW_DEPTH_FROM_TOP)

    def createCommentForMiter(self, **keywords):
        miter = keywords['miter']
        # Do use formatAsCoordinate, even if it is only a comment, otherwise Mac and Windows might get precision differences, which is bad for tests.
        return '; ' + self.getLocalizedComment(2, 'Miter {startx}, {starty} -> {endx}, {endy}').format(
            startx=self.formatAsCoordinate(data=miter.getStartPoint().x),
            starty=self.formatAsCoordinate(data=miter.getStartPoint().y),
            endx=self.formatAsCoordinate(data=miter.getEndPoint().x),
            endy=self.formatAsCoordinate(data=miter.getEndPoint().y),
        )

    def createPreSawForMiter(self, **keywords):
        miter = keywords['miter']
        angle = miter.getAngle()

        if self.MITER_SAW_REVERSE_ANGLE == 'mirrored':
            angle *= -1
        elif self.MITER_SAW_REVERSE_ANGLE == 'rotated':
            angle = 360 - angle

        leftHand = operator.xor(self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_LEFT or self.SURFACE_ORIGIN == SURFACE_ORIGIN_BACK_RIGHT, not self.MITER_SAW_REVERSE_ENTRY_SIDE)

        return '\n'.join((
            joinIfNotNone(' ',
                'XG0R',
                'X={x}'.format(x=keywords['startX']),
                'Y={y}'.format(y=keywords['startY']),
                'Z={z}'.format(z=self.formatAsCoordinate(data=keywords['preSawDepth'])),
                'H=DZ',
                self.createHoodPositioning(**keywords),
                'I={i}'.format(i=self.ENTRY_SIDE_LEFT_HAND_INTERPOLED_ANGLE if leftHand else self.ENTRY_SIDE_RIGHT_HAND_INTERPOLED_ANGLE),
                'V={v}'.format(v=self.formatAsCoordinate(data=XilogExport.MITER_SAW_SPEEDS.entry)),
                'B={angle}'.format(angle=self.formatAsAngle(data=angle)),
                'T={toolID}'.format(toolID=miter.toolID),
                'C={c}'.format(c=self.RADIUS_COMPENSATION_NONE),
            ),
            ' '.join((
                'XG1R',
                'X={x}'.format(x=keywords['endX']),
                'Y={y}'.format(y=keywords['endY']),
                'V={v}'.format(v=self.formatAsCoordinate(data=XilogExport.MITER_SAW_SPEEDS.presawing)),
            )),
        ))

    def createSawForMiter(self, **keywords):
        return '\n'.join((
            ' '.join((
                'XG1R',
                'Z={z}'.format(z='UEBERTIEFE'),
                'V={v}'.format(v=self.formatAsCoordinate(data=XilogExport.MITER_SAW_SPEEDS.sawing)),
            )),
            ' '.join((
                'XG1R',
                'X={x}'.format(x=keywords['startX']),
                'Y={y}'.format(y=keywords['startY']),
                'V={v}'.format(v=self.formatAsCoordinate(data=XilogExport.MITER_SAW_SPEEDS.exit)),
            )),
        ))

    def addCorrectionVariable(self, **keywords):
        coordinate = keywords['coordinate']
        vectorDimension = keywords['vectorDimension']
        variableName = keywords['variableName']

        if isNearlyEqual(coordinate, 0):
            if isNearlyEqual(vectorDimension, 1):
                result = variableName
            elif isNearlyEqual(vectorDimension, -1):
                result = '-' + variableName
            elif isNearlyEqual(vectorDimension, 0):
                result = '0'
            else:
                result = variableName + '*' + self.formatAsAngle(data=vectorDimension)
        else:
            result = self.formatAsCoordinate(data=coordinate)
            if isNearlyEqual(vectorDimension, 1):
                result += '+' + variableName
            elif isNearlyEqual(vectorDimension, -1):
                result += '-' + variableName
            elif not isNearlyEqual(vectorDimension, 0):
                result += '+' + variableName + '*' + self.formatAsAngle(data=vectorDimension)
        return result

    def getModifiedStartAndEndPointsForMiter(self, **keywords):
        miter = keywords['miter']

        directionVector = miter.directionVector
        start, end = miter.getStartAndEndPoints()
        start -= directionVector * self.SECURITY_DISTANCE_BEFORE_MITER
        end += directionVector * self.SECURITY_DISTANCE_AFTER_MITER

        perpendicularVector = -directionVector.perp()
        perpendicularVector.normalize()

        startX = self.addCorrectionVariable(coordinate=start.x, vectorDimension=perpendicularVector.x, variableName='KORR')
        startY = self.addCorrectionVariable(coordinate=start.y, vectorDimension=perpendicularVector.y, variableName='KORR')
        endX = self.addCorrectionVariable(coordinate=end.x, vectorDimension=perpendicularVector.x, variableName='KORR')
        endY = self.addCorrectionVariable(coordinate=end.y, vectorDimension=perpendicularVector.y, variableName='KORR')

        return startX, startY, endX, endY

    def createMiterVariables(self, **keywords):
        miter = keywords['miter']
        thickness = keywords['surface'].thickness
        overdepthFormula = self.zValueMiterSawConverter.convertVariableWithThickness(thickness, 'SAETIF+{addition}').format(addition=self.formatAsCoordinate(data=self.ADDITIONAL_DEPTH_FOR_MITER))

        angle = miter.getAngle()

        return '\n'.join((
            'L HBB =TOOL({toolID},9)'.format(toolID=miter.toolID),  # Half width of saw (see appendix B of Xilog Editor help (Xilog_Plus_Editor.chm))
            'L SBK =HBB/SIN({angle})'.format(angle=self.formatAsAngle(data=angle)),  # angle compensation of saw blade
            'L KORR ={width}-SBK'.format(width=self.formatAsCoordinate(data=miter.width)),  # compensation due slated sawing
            'L SAETIF =DZ/SIN({angle})'.format(angle=self.formatAsAngle(data=angle)),  # real saw depth (hypotenuse)
            'L UEBERTIEFE ={formula}'.format(formula=overdepthFormula),  # total depth for sawing through
        ))

    def createVerticalMiter(self, **keywords):
        startX, startY, endX, endY = self.getModifiedStartAndEndPointsForMiter(**keywords)
        preSawDepth = self.getPreSawDepth(boardThickness=keywords['surface'].thickness, **keywords)

        return '\n'.join((
            self.createMiterVariables(**keywords),
            self.createCommentForMiter(**keywords),
            self.createPreSawForMiter(startX=startX, startY=startY, endX=endX, endY=endY, preSawDepth=preSawDepth, **keywords),
            self.createSawForMiter(startX=startX, startY=startY, **keywords),
            '',
            '',
        ))

    def orientMiters(self, **keywords):
        surface = keywords['surface']
        adjuster = MiterOrienter(keywords['surface'].finalFinishedSizeShape)
        if self.isStandardOrigin(**keywords):
            adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCCW)
        else:
            adjuster.orient(surface.mitersVertical, adjuster.getMiterOrientedToCW)

    # NC Macro
    def createRecordFormatParameters(self, **keywords):
        return ' '.join((
            '{recordName}={recordValue}'.format(recordName=parameter.name, recordValue=parameter.value) for parameter in self.getNCMacroCookedParameters(**keywords)
        ))

    def createNCMacroRecordFormat(self, **keywords):
        ncmacro = keywords['ncmacro']
        return ' '.join((
            ncmacro.name,
            self.createRecordFormatParameters(**keywords)
        )) + '\n'

    # Footer
    def createParkingPosition(self, **keywords):
        if self.PARKING_POSITION_DISTANCE:
            try:  # != '' only for legacy machine-configurations till interiorcad 2020 F5
                parkingPositionDistance = int(self.PARKING_POSITION_DISTANCE)
            except ValueError:
                logging.info(getLocalizedString(XILOG_MESSAGES, 1, "Parking position distance must be a number for further calculations, assuming 0."))
                parkingPositionDistance = 0

            if self.MACHINING_AREA_X and self.MACHINING_AREA_X < keywords['surface'].getFinishedSizeX() + parkingPositionDistance:
                distance = self.MACHINING_AREA_X
            else:
                distance = 'DX+' + str(parkingPositionDistance)

            return 'N Q=1 X={distance}'.format(distance=distance)

        else:
            return ''

    def createEndOfFile(self, **keywords):
        return '\n'.join((
            '',
            'C=0',
            'SX=0',
            'SY=0 M=0',
            'O X=0 Y=0',
            self.createParkingPosition(**keywords),
        ))

    def getLocalizedComment(self, index, default):
        return getLocalizedString(XILOGPROGRAMCOMMENTS_RESOURCEGROUP, index, default)
